function mpc = case8387pegase
%CASE8387PEGASE 8387-bus PEGASE case, by Florin Capitanescu
%   Please read carefully until the end before using these data sets. These
%   data stem from the Pan European Grid Advanced Simulation and State
%   Estimation (PEGASE) project, funded by the European Union
%   (http://www.fp7-pegase.com/). These data are derived from the original
%   data [A] provided to Florin Capitanescu in 2013 under request by the
%   data owners in PEGASE project. The derivation consists in: (i) removing
%   data related to a transmission system operator, which were replaced by
%   equivalent loads at the end of the three interconnection lines with the
%   rest of the system, (ii) some asymmetrical PI branch models were
%   approximated as symmetrical, averaging parameters values, (iii)
%   modelling differences concerning the transformer transversal
%   susceptance, (iv) data anonymization, (v) data conversion from Eurostag
%   (.ech) fromat to my OPF data format and then to MATPOWER (.m) format,
%   and (vi) definition of various AC OPF problems (see [B] below) readily
%   to obtain in MATPOWER format.
%
%   This power system model contains 8,387 buses, 1,865 generators, 4,669
%   loads, 12,474 lines, 2,087 transformers (out of which 589 OLTC and 84
%   phases shifters) and 178 shunt capacitors/reactors. Also, 1,299 lines
%   and 529 transformers (out of which 17 phase shifters) have more than one
%   circuit in parallel. The system models 8 voltage levels: 750, 400, 380,
%   330, 220, 150, 120, and 110 kV. These data represent reasonably well the
%   size and complexity of the European transmission system. However, these
%   data are fictitious and include modifications made on purpose to real
%   world data. They can thus be used to test methods and algorithms, but
%   any claim should not be made regarding the operation or planning of the
%   European power system.
%
%   These specific data sets were publicly released for the main purpose of
%   enabling testing methods and algorithms on truly numerically challenging
%   and tough optimization problems (see details in [B], where two solvers
%   in MATPOWER version 4.1 failed to converge on two instances and my own
%   OPF failed to converge on another instance) and more comprehensive in
%   terms of grid models and versatile AC OPF formulations. For example, [B]
%   presented results for 16 various instances of AC OPF problems with
%   different objective functions, sets of decision variables (e.g. among:
%   generators active/reactive powers, ratio of LTC transformers, reactance
%   of shunt capacitors/reactors, and angle of phase shifter transformers)
%   and constraints. If you wish to include control variables like
%   transformer ratio, phase shifter angle and shunt reactance, please
%   contact me to share with you their limits and discrete steps. In my AC
%   OPF instances all thermal limits were expressed only in terms of
%   current.
%
%   When publishing results based on these data, please cite:
%
%   [A] F. Villella, S. Leclerc, I. Erlich, S. Rapoport, "PEGASE
%       pan-European test-beds for testing of algorithms on very large scale
%       power systems". IEEE PES Innovative Smart Grid Technologies Europe
%       (ISGT Europe), Berlin (Germany), 2012.
%       doi: 10.1109/ISGTEurope.2012.6465783
%
%   [B] F. Capitanescu, L. Wehenkel, "Experiments with the interior-point
%       method for solving large scale optimal power flow problems",
%       Electric Power Systems Research, vol. 95, pp. 276-283, 2013.
%       doi: 10.1016/j.epsr.2012.10.001
%
%   Besides this test case, all AC OPF instances presented in [B] can be
%   easily obtained using one or some combinations of following features:
%
%   1)  The angle of all 84 phase shifters is set to zero.
%   2)  Identical unity costs (c=1.0) are used for all generators to
%       minimize the sum of generators active powers (an unconventional
%       loss minimization via also active power redispatch) as OPF
%       objective function.
%   3)  The quadratic OPF objective function minimizes the sum of squares
%       of generator’ active power deviation with respect to the initial
%       value Pgi^0 (provided in the data set), i.e. (Pgi-Pgi^0)^2. Thus,
%       the coefficients of the fictitious cost function are in p.u.
%       (100MVA base): (0.01*Pgi^0)^2, -2*0.01*Pgi^0, and 0.0001.
%   4)  The thermal limit of line between nodes 746 and 5968 and can be
%       decreased on purpose from 2061.62 MVA to 1000 MVA to create congestion
%       (around 20% current overload) and a tough OPF problem, see [B].
%
%   If you face convergence issues with your own NLP solver, some
%   simplifications are suggested:
%    -  Some generators (around 40) with small active power range (<=20MW)
%       are not taken as control variable, their active power remain at the
%       initial value;
%    -  Some generators (around 40) with small reactive power range
%       (<=10MVAr) are not taken as control variable, their reactive power
%       remain at the initial value;
%    -  The reactance of 98 short lines could be raised to the minimum
%       threshold value of 0.0002 pu;
%    -  Increase the convergence tolerance of the NLP solver to 10^-5 on
%       primal and relative dual feasibility.
%
%   Contact: Florin Capitanescu fcapitanescu@yahoo.com
%
%   February 18th, 2021

%   MATPOWER
%   Copyright (c) 2021 by Florin Capitanescu
%   Licensed under the Creative Commons Attribution 4.0 International license,
%   https://creativecommons.org/licenses/by/4.0/

fixed = 0;  %% change to 1 to fix the generation of 615 smaller units that
            %% have no lower/upper bounds on active and reactive generation
            %% in the original data

%% MATPOWER Case Format : Version 2
mpc.version = '2';

%%-----  Power Flow Data  -----%%
%% system MVA base
mpc.baseMVA = 100;

%% bus data
%	bus_i	type	Pd	Qd	Gs	Bs	area	Vm	Va	baseKV	zone	Vmax	Vmin
mpc.bus = [
	1	1	378.4	98.5	0	0	1	1.038963	-34.89481	220	1	1.1	0.9;
	2	1	151	48.8	0	0	1	1.026375	-20.392332	220	1	1.1	0.9;
	3	1	171.4081	23.4	0	0	1	1.040071	-6.609053	220	1	1.127885	0.9;
	4	1	45.95095	14	0	0	1	1.01637	-46.441272	220	1	1.1	0.9;
	5	2	343.4	53.1	0	20	1	1.060105	-39.700511	380	1	1.126561	0.9;
	6	1	31.03441	10.3	0	0	1	0.901905	-32.067072	110	1	1.1	0.9;
	7	1	134	24.7	0	0	1	1.075492	-22.993847	220	1	1.16095	0.9;
	8	1	0	0	0	11.29798	1	1.030663	-24.28533	150	1	1.1	0.9;
	9	1	-6.10677	2	0	0	1	1.03402	-18.796018	150	1	1.105047	0.9;
	10	2	8.509435	5	0	0	1	1.045776	-45.266112	220	1	1.1	0.9;
	11	1	13.5	14.8	0	0	1	1.098764	-14.368752	380	1	1.166895	0.9;
	12	1	59.66615	26.7	0	0	1	1.064712	-44.0377	220	1	1.113369	0.9;
	13	1	67.67503	12.1	0	0	1	1.010989	7.736285	220	1	1.1	0.9;
	14	2	0	0	0	0	1	1.040474	18.686769	380	1	1.1	0.9;
	15	1	25.32808	7.5	0	0	1	0.993806	-34.020232	110	1	1.1	0.9;
	16	1	161.2	39.3	0	0	1	1.074018	-20.85146	220	1	1.153603	0.9;
	17	1	0	0	0	0	1	1.081379	-17.543279	220	1	1.146159	0.9;
	18	1	19.62176	7.1	0	0	1	1.017163	-21.39745	110	1	1.1	0.9;
	19	1	17.21909	5.7	0	0	1	0.927816	-27.872689	110	1	1.1	0.9;
	20	2	0	0	0	0	1	1.012955	14.035448	220	1	1.1	0.9;
	21	1	167.3	54.8	0	0	1	1.079094	-3.040599	380	1	1.140082	0.9;
	22	1	0	0	0	0	1	1.074452	14.420626	380	1	1.142106	0.9;
	23	1	0	0	0	0	1	0.96616	-25.110965	150	1	1.1	0.9;
	24	1	152.9	49.4	0	0	1	1.067288	-12.812268	380	1	1.121859	0.9;
	25	1	0	0	0	0	1	1.014514	-22.601714	150	1	1.1	0.9;
	26	1	69.67725	43.2	0	0	1	0.935328	-28.779899	150	1	1.1	0.879727;
	27	2	0	0	0	0	1	1.020864	-38.450258	220	1	1.1	0.9;
	28	2	0	0	0	0	1	1.025773	-34.375097	220	1	1.1	0.9;
	29	1	13.51498	4.9	0	0	1	1.044666	-20.014722	110	1	1.106205	0.9;
	30	1	10.51165	3.5	0	0	1	1.054655	-32.269646	110	1	1.123445	0.9;
	31	1	0	0	0.122998	0	1	1.06255	-16.574001	220	1	1.115122	0.9;
	32	1	9.210212	2.6	0	0	1	0.942396	-25.378952	110	1	1.1	0.9;
	33	1	42.74739	12.8	0	0	1	1.079439	20.319468	220	1	1.157714	0.9;
	34	2	0	0	0	0	1	1.023133	-34.021096	150	1	1.1	0.9;
	35	2	24.02664	7.8	0	0	1	0.952528	-26.572018	110	1	1.1	0.9;
	36	1	20.82309	7.9	0	0	1	1.087429	-44.289567	380	1	1.138606	0.9;
	37	1	0	0	0	0	1	1.077563	-21.020443	220	1	1.161391	0.9;
	38	1	10.11121	3.7	0	0	1	1.055964	-20.284863	110	1	1.113777	0.9;
	39	1	10.0111	2.7	0	0	1	0.939617	-24.979801	110	1	1.1	0.9;
	40	1	6.306993	0	0	0	1	0.96251	-23.689017	110	1	1.1	0.9;
	41	1	2.903219	0.9	0	0	1	0.989708	-20.928223	110	1	1.1	0.9;
	42	2	0	0	0	0	1	1.051136	-33.838556	220	1	1.113044	0.9;
	43	2	151.8	48.3	0	0	1	1.026	-42.739462	220	1	1.1	0.9;
	44	1	127.6	31.8	0	0	1	1.058132	0.522482	150	1	1.119723	0.9;
	45	1	48.65394	16.1	0	0	1	1.069015	-8.755465	220	1	1.141578	0.9;
	46	1	110.5	-26.8	0	0	1	1.063422	-33.39153	220	1	1.128492	0.9;
	47	1	0	0	0	0	1	0.993101	10.688212	220	1	1.1	0.9;
	48	1	140	79.7	0	0	1	1.023887	-15.442695	220	1	1.1	0.9;
	49	1	0	0	0	0	1	1.074444	-15.666053	380	1	1.135699	0.9;
	50	1	119.2	20.7	0	0	1	1.018017	-23.197334	220	1	1.125867	0.9;
	51	2	5.105661	1.3	0	0	1	1.030763	-16.065598	110	1	1.1	0.9;
	52	1	0	0	0	0	1	1.078773	46.429599	380	1	1.1	0.9;
	53	1	0	0	0	0	1	0.95334	-23.586564	110	1	1.1	0.9;
	54	1	187.9	110.2	0	20	1	1.059872	6.237875	380	1	1.113265	0.9;
	55	1	14.91654	5.5	0	0	1	1.00326	-17.973502	110	1	1.1	0.9;
	56	1	-18.5205	6.1	0	0	1	0.94589	-27.327121	150	1	1.1	0.9;
	57	1	4.605106	1.7	0	0	1	1.038632	-15.021604	110	1	1.13511	0.9;
	58	1	0	0	0.090701	0	1	1.079305	-18.429678	380	1	1.14929	0.9;
	59	1	10.81199	5.1	0	0	1	0.919535	-28.937152	110	1	1.1	0.9;
	60	1	62.86971	20.5	0	0	1	1.059071	-0.602036	150	1	1.121873	0.9;
	61	1	126.3	42.4	0	0	1	1.084891	2.141781	150	1	1.156303	0.9;
	62	1	6.907659	-4.6	0	0	1	1.02866	-36.521106	220	1	1.1	0.9;
	63	1	15.71743	5.2	0	0	1	0.933539	-34.456554	110	1	1.1	0.9;
	64	1	0	0	0	17.1282	1	1.024338	-23.11119	150	1	1.1	0.9;
	65	1	38.74295	6.1	0	0	1	1.054835	-48.187327	220	1	1.107549	0.9;
	66	1	22.12453	10.7	0	0	1	1.079567	20.328447	220	1	1.157811	0.9;
	67	1	0	0	0	0	1	1.048413	-45.550599	220	1	1.1	0.9;
	68	1	0	0	0	0	1	0.99755	-16.514886	110	1	1.1	0.9;
	69	1	138.8	136.7	0	0	1	1.014697	14.350572	380	1	1.1	0.9;
	70	2	137.3	66.4	0	0	1	1.022545	-21.297375	220	1	1.108244	0.9;
	71	2	67.77514	19.7	0	0	1	1.048726	61.695204	150	1	1.1	0.9;
	72	1	22.02442	5.1	0	0	1	1.016103	-29.430561	110	1	1.2	0.8;
	73	1	0	0	0	0	1	1.016746	14.422022	220	1	1.1	0.9;
	74	1	1.00111	0.3	0	0	1	1.037164	-22.840921	110	1	1.132836	0.9;
	75	1	0	0	0	0	1	1.077496	-17.641439	150	1	1.1	0.9;
	76	1	21.52386	5.1	0	0	1	1.035786	-10.805746	110	1	1.1	0.9;
	77	1	71.9798	28.8	0	0	1	1.052622	-36.234527	220	1	1.135	0.9;
	78	1	0	0	0.173359	0	1	1.027193	-5.300984	380	1	1.1	0.9;
	79	2	365.2	116.5	0	0	1	1.031659	-27.564002	220	1	1.110955	0.9;
	80	1	0	0	0	0	1	1.088263	1.178574	150	1	1.157145	0.9;
	81	1	192.9	73.3	0	0	1	1.060148	4.664268	380	1	1.126192	0.9;
	82	1	17.51942	6.3	0	0	1	1.055871	-16.94169	110	1	1.135196	0.9;
	83	1	33.13674	12	0	0	1	1.010453	-20.40704	110	1	1.1	0.9;
	84	1	40.44484	25	0	0	1	1.043878	-23.359092	150	1	1.1	0.9;
	85	2	212.9	61.7	0	0	1	1.068345	-14.0534	220	1	1.124136	0.9;
	86	1	0	0	0	0	1	1.07894	2.524891	380	1	1.126808	0.9;
	87	1	90.70056	25.5	0	0	1	1.038032	-1.648397	220	1	1.114867	0.9;
	88	1	176.3	-58.9	0	0	1	1.075423	-32.825831	380	1	1.151238	0.9;
	89	1	35.93985	22.3	0	0	1	1.010047	-23.837407	150	1	1.1	0.9;
	90	1	0	0	0	0	1	1.030319	-24.295172	150	1	1.1	0.9;
	91	1	0	0	0	0	1	1.01282	-47.531793	220	1	1.1	0.9;
	92	1	65.07214	40.3	0	0	1	0.998755	-26.993988	150	1	1.1	0.9;
	93	1	51.35694	-8.8	0	0	1	0.986974	-51.973397	220	1	1.1	0.9;
	94	1	0	0	0	0	1	1.068183	-32.344945	380	1	1.128981	0.9;
	95	1	0	0	0	0	1	1.101714	-15.920408	380	1	1.183108	0.9;
	96	2	0	0	0	0	1	1.103763	42.095641	380	1	1.119079	0.9;
	97	1	3.804218	1.3	0	0	1	1.022916	-14.269525	110	1	1.1	0.9;
	98	1	0	0	0	0	1	0.960682	-25.230096	150	1	1.1	0.9;
	99	1	61.06771	36	0	0	1	1.054988	-22.53124	220	1	1.1	0.9;
	100	1	0	0	0	0	1	1.066066	-8.275109	220	1	1.12853	0.9;
	101	1	0	0	0	0	1	1.075399	-13.687413	380	1	1.14595	0.9;
	102	1	0	0	0	0	1	1.081257	-19.059926	220	1	1.184353	0.9;
	103	1	1.801998	0.7	0	0	1	1.005392	-18.555746	110	1	1.1	0.9;
	104	1	40.0444	16	0	0	1	0.984345	-19.661056	110	1	1.1	0.9;
	105	1	0	0	0	0	1	1.071027	-9.623723	380	1	1.119605	0.9;
	106	1	0	0	0	0	1	1.070813	-14.444565	150	1	1.123004	0.9;
	107	1	45.15006	14.1	0	0	1	1.006189	-19.974072	110	1	1.1	0.9;
	108	1	0	0	0	0	1	1.004673	9.721829	220	1	1.1	0.9;
	109	2	0	0	0	0	1	1.066727	-18.880729	220	1	1.113308	0.9;
	110	2	0	0	0	0	1	1.066	8.596379	220	1	1.13202	0.9;
	111	2	8.209102	4.4	0	0	1	1.0287	-16.101178	110	1	1.1	0.9;
	112	1	0	0	0	0	1	1.060106	-26.65192	220	1	1.122189	0.9;
	113	1	23.72631	12.7	0	0	1	1.056511	-15.772611	220	1	1.129086	0.9;
	114	1	121.9	7.4	0	0	1	1.049178	-10.846798	220	1	1.11918	0.9;
	115	1	0	0	0.156015	0	1	1.098262	-42.669422	380	1	1.154381	0.9;
	116	1	47.3525	16.4	0	0	1	1.035033	-35.673812	110	1	1.1	0.9;
	117	1	0	0	0	0	1	1.09848	-39.509301	380	1	1.154681	0.9;
	118	1	38.14229	12.5	0	0	1	1.057885	2.038552	150	1	1.117601	0.9;
	119	1	130.6	84.9	0	0	1	1.05183	-20.341738	220	1	1.1	0.9;
	120	1	7.908769	2.6	0	0	1	1.027787	-31.846971	110	1	1.1	0.9;
	121	2	47.15228	78	0	0	1	1.029075	61.886161	150	1	1.1	0.9;
	122	1	-75.3836	7.8	0	0	1	1.095416	-37.547369	380	1	1.158685	0.9;
	123	1	0	0	0	0	1	1.037056	14.583899	380	1	1.1	0.9;
	124	1	0	0	0	0	1	1.025318	13.046206	220	1	1.1	0.9;
	125	2	100.9	49.3	0	0	1	1.002727	-0.386041	220	1	1.105252	0.9;
	126	2	0	0	0	0	1	1.080818	-28.948979	220	1	1.133107	0.9;
	127	1	0	0	0	0	1	1.067289	3.229482	150	1	1.13802	0.9;
	128	1	0	0	0	0	1	1.057126	19.944335	380	1	1.1	0.9;
	129	1	38.34251	5.6	0	0	1	1.079405	-8.263213	220	1	1.145959	0.9;
	130	1	269.6	65.7	0	0	1	1.096744	-22.741628	120	1	1.160482	0.9;
	131	2	160	3.3	0	0	1	1.095727	-11.840449	220	1	1.144905	0.9;
	132	1	32.33585	12	0	0	1	1.042972	-7.650175	220	1	1.132096	0.9;
	133	1	100.1	45.1	0	0	1	1.050753	-38.483146	380	1	1.115611	0.9;
	134	2	12.31365	3.9	0	0	1	1.0975	2.372802	220	1	1.16377	0.9;
	135	1	0	0	0	0	1	1.058294	-27.663837	150	1	1.129496	0.9;
	136	1	0	0	0	0	1	1.003067	-35.970977	110	1	1.1	0.9;
	137	1	0	0	0	0	1	0.980759	-25.149654	110	1	1.1	0.9;
	138	1	40.34473	25	0	0	1	1.014528	-24.673955	150	1	1.1	0.9;
	139	1	9.210212	3.3	0	0	1	1.034794	-8.963928	110	1	1.107242	0.9;
	140	1	166	60.3	0	0	1	1.016031	8.161181	220	1	1.1	0.9;
	141	1	0	0	0	0	1	0.992829	-23.492907	150	1	1.1	0.9;
	142	1	0	0	0	0	1	1.038358	-21.671873	150	1	1.11491	0.9;
	143	1	0	0	0	0	1	1.098909	-17.161843	380	1	1.167046	0.9;
	144	2	152.4	35.1	0	0	1	1.073994	-10.278393	220	1	1.135549	0.9;
	145	1	0	0	0	0	1	1.063202	-1.710233	150	1	1.139094	0.9;
	146	1	26.8	-3.7	0	-25.3011	1	1.070647	-16.970903	380	1	1.118094	0.9;
	147	2	0	0	0	0	1	1.030091	13.653817	220	1	1.1	0.9;
	148	1	257.5	131.1	0	0	1	1.035723	-4.054882	220	1	1.1	0.9;
	149	1	0	0	0	0	1	1.021161	-23.284148	150	1	1.1	0.9;
	150	1	0	0	0	0	1	0.937285	38.058108	220	1	1.1	0.9;
	151	1	100.8	13.2	0	0	1	1.095539	-24.360315	220	1	1.18781	0.9;
	152	2	113.9	11.9	0	0	1	1.0096	-42.882955	220	1	1.1	0.9;
	153	2	104.4	0.7	0	0	1	1.0196	15.723856	220	1	1.1	0.9;
	154	1	370.5	-17	0	0	1	1.046899	-27.358917	220	1	1.156542	0.9;
	155	1	9.910988	3.2	0	0	1	0.990146	-37.281745	110	1	1.1	0.9;
	156	2	0	0	0	0	1	1.008891	-22.243841	150	1	1.1	0.9;
	157	1	94.9	36.2	0	0	1	0.988844	-38.857087	220	1	1.1	0.9;
	158	1	81.0899	48	0	0	1	1.004099	-23.31839	150	1	1.1	0.9;
	159	2	210.4	159.2	0	0	1	1.089842	-45.007732	380	1	1.13289	0.9;
	160	1	0	0	0	0	1	0.991367	-23.761052	110	1	1.1	0.9;
	161	1	54.96094	16.9	0	0	1	1.03332	-35.055506	110	1	1.1	0.9;
	162	1	19.32142	26	0	0	1	1.049309	40.294048	220	1	1.1	0.9;
	163	1	0	0	0	0	1	1.021022	-22.119247	150	1	1.1	0.9;
	164	1	0	0	0	0	1	1.0523	-11.912149	220	1	1.124278	0.9;
	165	1	19.92209	6.6	0	0	1	0.958234	-25.691252	110	1	1.1	0.9;
	166	1	57.16338	2	0	0	1	0.992219	-33.930735	110	1	1.1	0.9;
	167	1	69.57714	16	0	0	1	1.07593	-25.498546	220	1	1.165972	0.9;
	168	1	36.84085	13.9	0	0	1	1.088144	-18.434766	220	1	1.161548	0.9;
	169	2	4.805328	1.3	0	0	1	0.998323	-23.519367	110	1	1.1	0.9;
	170	2	9.210212	6	0	0	1	0.993364	-16.827852	110	1	1.1	0.9;
	171	1	0	0	0	0	1	1.047656	-34.236842	220	1	1.110235	0.9;
	172	2	0	0	0	0	1	1.087684	66.734724	380	1	1.1	0.9;
	173	1	119.9	3.7	0	0	1	1.038146	25.439384	220	1	1.1	0.9;
	174	2	0	0	0	0	1	1.075867	11.667964	150	1	1.10778	0.9;
	175	2	75.78402	24.6	0	0	1	1.092354	-8.438261	220	1	1.153235	0.9;
	176	1	0	0	0	0	1	1.076831	-25.438083	220	1	1.166823	0.9;
	177	2	0	0	0.288019	0	1	1.105553	45.754712	380	1	1.119591	0.9;
	178	1	39.74406	11.8	0	0	1	1.054374	-15.988696	220	1	1.12666	0.9;
	179	1	86.79623	28.4	0	0	1	1.061803	0.70194	150	1	1.12403	0.9;
	180	1	22.42486	2.1	0	0	1	1.040788	61.564881	150	1	1.1	0.9;
	181	2	0	0	0	0	1	1.057864	-46.539464	220	1	1.107704	0.9;
	182	1	110.9	32.7	0	0	1	1.007606	-0.256966	220	1	1.1	0.9;
	183	1	0	0	0.191685	0	1	1.042712	-19.254167	220	1	1.11363	0.9;
	184	1	46.15117	18.2	0	0	1	1.063849	3.781789	150	1	1.134437	0.9;
	185	1	34.03774	12.3	0	0	1	1.062433	-16.925274	110	1	1.137647	0.9;
	186	2	322.4	84.4	0	0	1	1.03615	-13.393889	220	1	1.109357	0.9;
	187	1	33.43707	10	0	0	1	1.013491	-26.974477	110	1	1.1	0.9;
	188	2	0	0	0	0	1	0.972727	-31.899847	110	1	1.1	0.9;
	189	1	0	0	0	0	1	0.979203	-20.113238	110	1	1.1	0.9;
	190	1	75.28346	19.2	0	0	1	1.034151	9.011174	110	1	1.115164	0.9;
	191	2	219.6	77.5	0	0	1	1.040354	-4.000858	220	1	1.112689	0.9;
	192	1	37.3414	10.9	0	0	1	1.006456	-30.429716	110	1	1.2	0.8;
	193	2	340	93.8	0	0	1	1.056395	14.274392	380	1	1.126171	0.9;
	194	1	0	0	0	0	1	1.001811	-23.390712	110	1	1.1	0.9;
	195	1	155.6	54.7	0	0	1	1.071309	-23.375004	220	1	1.136527	0.9;
	196	1	105.2	0.1	0	0	1	1.081865	-3.59454	220	1	1.135314	0.9;
	197	2	80.9898	21	0	0	1	1.049023	-12.546491	220	1	1.106428	0.9;
	198	2	0	0	0	0	1	1.079341	-14.723245	380	1	1.12015	0.9;
	199	2	0	0	0	0	1	1.0985	-16.80129	380	1	1.164753	0.9;
	200	2	0	0	0	0	1	1.076684	-16.997007	380	1	1.138316	0.9;
	201	2	176.2	20.8	0	0	1	0.944667	65.220886	150	1	1.1	0.9;
	202	1	27.02997	8.8	0	0	1	1.022563	-16.762063	110	1	1.1	0.9;
	203	1	2.602886	0.5	0	0	1	1.005716	-15.671355	110	1	1.1	0.9;
	204	1	0	0	0	0	1	1.078195	-19.440094	220	1	1.181608	0.9;
	205	1	14.61621	4.8	0	0	1	1.003882	-29.788001	110	1	1.1	0.9;
	206	1	0	0	0	0	1	1.075366	-20.605091	220	1	1.122058	0.9;
	207	1	-23.426	57.4	0	0	1	1.086245	-9.25803	150	1	1.148859	0.9;
	208	1	426.6	93.7	0	0	1	1.086886	-25.343887	380	1	1.144748	0.9;
	209	2	154.2	23.3	0	0	1	1.011746	-42.601473	220	1	1.1	0.9;
	210	2	0	0	0	0	1	1.04248	17.34212	110	1	1.112015	0.9;
	211	1	0	0	0	0	1	1.031525	-17.322132	150	1	1.1	0.9;
	212	1	2.502775	3.8	0	0	1	0.947371	-27.095908	110	1	1.1	0.9;
	213	1	34.03774	4.4	0	0	1	1.001166	-37.186914	110	1	1.1	0.9;
	214	1	45.85083	7	0	0	1	1.0523	-49.135864	220	1	1.110936	0.9;
	215	1	0	0	0	0	1	1.063929	40.992984	380	1	1.124591	0.9;
	216	1	0	0	0	0	1	1.073552	-45.318439	220	1	1.130584	0.9;
	217	2	46.05106	16.9	0	0	1	1.036909	-22.83738	110	1	1.131861	0.9;
	218	1	0	0	0	12.42588	1	0.986553	-23.775778	150	1	1.1	0.9;
	219	1	0	0	0	0	1	1.007596	60.472961	220	1	1.1	0.9;
	220	1	167.9	18.2	0	0	1	1.073607	-1.57326	220	1	1.134657	0.9;
	221	1	0	0	0	0	1	1.048173	15.421454	150	1	1.1	0.9;
	222	1	0	0	0	0	1	1.056078	5.510922	150	1	1.109746	0.9;
	223	1	16.61843	5.5	0	0	1	0.975182	-26.101782	110	1	1.1	0.9;
	224	2	131.9	31.1	0	0	1	1.057909	-18.438186	220	1	1.2	0.8;
	225	1	0	0	0	0	1	0.990718	-19.403743	110	1	1.1	0.9;
	226	2	224.2	73.9	0	0	1	1.0115	12.538368	380	1	1.107643	0.9;
	227	1	0	0	0	0	1	1.020417	-38.452691	220	1	1.1	0.9;
	228	1	0	0	0	0	1	1.046646	10.107236	380	1	1.107593	0.9;
	229	1	0	0	0	0	1	1.058167	54.278935	220	1	1.106174	0.9;
	230	1	0	0	0	0	1	1.058609	-19.146934	220	1	1.111298	0.9;
	231	1	0	0	0.152776	0	1	1.064814	-24.013785	380	1	1.135156	0.9;
	232	2	0	0	0	0	1	1.102343	-13.747738	380	1	1.147516	0.9;
	233	1	0	0	0	0	1	1.078397	-35.325664	220	1	1.157332	0.9;
	234	1	113.9	24.5	0	0	1	1.017906	-41.855053	220	1	1.1	0.9;
	235	2	43.34806	26.9	0	0	1	1.033422	-25.170853	150	1	1.1	0.9;
	236	1	216.8	40.5	0	0	1	1.062213	-10.346926	220	1	1.121101	0.9;
	237	1	0	0	0	0	1	1.086366	-41.190921	380	1	1.146157	0.9;
	238	2	160.8	10.6	0	0	1	1.028136	10.353076	220	1	1.120336	0.9;
	239	1	38.74295	9	0	0	1	1.03375	-42.667185	220	1	1.1	0.9;
	240	1	350.4	29.3	0	0	1	1.087699	-26.505474	220	1	1.150201	0.9;
	241	1	0	0	0	-261.753	1	1.077426	-38.117944	380	1	1.155069	0.9;
	242	1	103.8	2.7	0	0	1	1.019651	-46.889667	220	1	1.1	0.9;
	243	2	0	0	0	0	1	1.079235	20.51598	220	1	1.16345	0.9;
	244	1	0	0	0	0	1	1.017773	-36.447882	110	1	1.1	0.9;
	245	1	0	0	0	0	1	1.08165	67.139659	380	1	1.1	0.9;
	246	1	96.7	26.8	0	0	1	1.079745	-9.793118	150	1	1.147271	0.9;
	247	1	0	0	0	0	1	1.013233	-22.996644	150	1	1.143953	0.9;
	248	1	0	0	0	0	1	1.012234	-22.01638	220	1	1.1	0.9;
	249	1	19.02109	6	0	0	1	0.979697	-23.10462	110	1	1.1	0.9;
	250	1	102.1	25.3	0	0	1	1.051763	1.603872	150	1	1.111364	0.9;
	251	1	0	0	0	0	1	1.026253	65.096126	220	1	1.1	0.9;
	252	1	0	0	0	0	1	0.999672	-21.935927	150	1	1.1	0.9;
	253	2	113.3	30.2	0	0	1	0.979909	-25.596375	110	1	1.1	0.9;
	254	1	8.309213	3	0	0	1	1.022513	-22.223609	110	1	1.112581	0.9;
	255	1	119.6	29.6	0	0	1	1.099521	-10.827073	220	1	1.156354	0.9;
	256	1	0	0	0	0	1	1.065974	11.928881	380	1	1.117088	0.9;
	257	2	49.9	144	0	0	1	1.104	-8.299411	380	1	1.150112	0.9;
	258	1	27.3303	11.9	0	0	1	1.043782	-12.042962	220	1	1.116684	0.9;
	259	1	191.2	-3.4	0	0	1	1.065007	-44.08172	220	1	1.113845	0.9;
	260	1	0	0	0.03458	0	1	1.000093	66.738694	380	1	1.1	0.9;
	261	1	177.1	66.9	0	0	1	1.099334	-14.390411	380	1	1.147872	0.9;
	262	1	0	0	0	0	1	1.070669	-14.380307	150	1	1.122123	0.9;
	263	2	0	0	0	0	1	1.090773	-29.430538	220	1	1.141373	0.9;
	264	1	53.45927	18.7	0	0	1	1.06625	-19.319712	220	1	1.1	0.9;
	265	1	0	0	0	0	1	1.01999	4.423568	220	1	1.1	0.9;
	266	1	13.81532	1.4	0	0	1	1.071346	-13.822146	220	1	1.13528	0.9;
	267	1	85.1	67	0	0	1	1.06169	-15.378716	150	1	1.120351	0.9;
	268	1	0	0	0	0	1	1.05581	-27.142614	220	1	1.117973	0.9;
	269	1	189.3	76.7	0	0	1	1.030364	28.27478	220	1	1.1	0.9;
	270	1	23.72631	9.4	0	0	1	1.074839	3.235018	150	1	1.144299	0.9;
	271	1	10.0111	-21	0	0	1	1.050548	-33.131623	220	1	1.134182	0.9;
	272	2	228.1	131.8	0	0	1	1.078658	-41.66976	380	1	1.120095	0.9;
	273	1	0	0	0	0	1	1.042556	-15.707649	150	1	1.1	0.9;
	274	2	0	0	0	0	1	0.97336	-19.305206	110	1	1.1	0.9;
	275	1	431.7	121.6	0	0	1	0.974351	-24.576868	220	1	1.1	0.9;
	276	1	2.202442	0.5	0	0	1	1.030897	-49.304332	150	1	1.1	0.9;
	277	1	103	12	0	0	1	1.057362	-31.516276	220	1	1.184252	0.9;
	278	1	0	0	0	0	1	1.085277	-17.81646	220	1	1.165895	0.9;
	279	1	0	0	0	0	1	1.091806	-20.656481	380	1	1.155557	0.9;
	280	1	0	0	0	40.46213	1	0.944457	-32.813558	150	1	1.1	0.831014;
	281	2	0	0	0	0	1	1.014045	13.924736	220	1	1.1	0.9;
	282	2	-463.61	253.81	0	0	1	1.092632	-13.322266	380	1	1.17531	0.9;
	283	1	42.04662	17.2	0	0	1	1.068698	-0.717571	150	1	1.13575	0.9;
	284	2	0	0	0	0	1	1.075726	8.619709	110	1	1.14437	0.9;
	285	1	163.1	46.3	0	0	1	1.054129	-15.632559	220	1	1.123665	0.9;
	286	1	0	0	0	0	1	1.046946	-32.258169	220	1	1.105409	0.9;
	287	1	0	0	0	0	1	1.075715	-3.694537	220	1	1.13559	0.9;
	288	1	0	0	0.041702	0	1	1.086367	1.518637	380	1	1.134899	0.9;
	289	1	0	0	0	0	1	1.031226	-24.032509	150	1	1.1	0.9;
	290	1	67.17448	-23.9	0	0	1	1.027328	64.861694	220	1	1.1	0.9;
	291	1	303.8	85.4	0	0	1	0.979764	-40.335625	220	1	1.1	0.9;
	292	2	0	0	0	0	1	1.087711	53.455386	380	1	1.114174	0.9;
	293	1	0	0	0	0	1	1.079163	20.557431	220	1	1.163549	0.9;
	294	1	0	0	0	0	1	1.056394	-11.319676	380	1	1.124142	0.9;
	295	2	291.5	51.3	0	0	1	0.948091	25.15516	220	1	1.1	0.9;
	296	1	24.22686	7.1	0	0	1	0.981243	-24.794944	110	1	1.1	0.9;
	297	1	108.6	2.6	0	0	1	1.046773	10.828043	380	1	1.108345	0.9;
	298	1	4.00444	2.4	0	0	1	0.991813	-25.115659	150	1	1.1	0.9;
	299	1	30.0333	10.9	0	0	1	1.023813	-21.256669	110	1	1.1	0.9;
	300	1	524.5	43	0	0	1	1.016255	-41.301074	220	1	1.125504	0.9;
	301	1	0	0	0	0	1	0.943888	-27.508957	110	1	1.1	0.9;
	302	1	4.404884	0	0	0	1	1.021895	-15.877938	110	1	1.1	0.9;
	303	2	7.408214	2.7	0	0	1	1.030147	-16.413037	110	1	1.132888	0.9;
	304	2	0	0	0	0	1	1.076818	-5.802663	220	1	1.136328	0.9;
	305	1	0	0	0	0	1	1.074351	-33.95989	220	1	1.154529	0.9;
	306	1	186.6	56.6	0	0	1	1.098289	-19.922406	380	1	1.151615	0.9;
	307	1	0	0	0	0	1	0.976967	-24.995005	110	1	1.1	0.9;
	308	2	48.2535	23.5	0	0	1	1.060331	-32.224317	220	1	1.135645	0.9;
	309	1	3.203552	1.2	0	0	1	1.038716	-18.255563	110	1	1.12359	0.9;
	310	1	372.8	147.5	0	0	1	1.064048	-2.428148	150	1	1.141954	0.9;
	311	1	0	0	0.087531	0	1	1.011521	-24.28785	220	1	1.125383	0.9;
	312	1	197.5	48.3	0	0	1	1.070747	-20.262446	220	1	1.135698	0.9;
	313	2	0	0	0	0	1	1.102211	-23.66592	380	1	1.15504	0.9;
	314	1	0	0	0	0	1	1.082248	7.031108	150	1	1.149448	0.9;
	315	1	110.4	26.2	0	0	1	1.076544	-23.052428	220	1	1.161458	0.9;
	316	2	0	0	0	0	1	1.029313	13.665017	220	1	1.106748	0.9;
	317	2	0	0	0	0	1	1.092132	-0.233458	380	1	1.153889	0.9;
	318	1	0	0	0	0	1	1.036616	-25.081018	220	1	1.1	0.9;
	319	1	0	0	0	0	1	1.070807	-15.141283	380	1	1.122154	0.9;
	320	2	0	0	0	0	1	0.975854	-20.255082	110	1	1.1	0.9;
	321	1	141.1	2.1	0	0	1	1.080003	-15.933285	220	1	1.123593	0.9;
	322	1	52.95872	32.8	0	0	1	1.032825	-25.473856	150	1	1.1	0.9;
	323	1	0	0	0	0	1	1.075258	-12.558053	220	1	1.130475	0.9;
	324	1	35.5394	11.5	0	0	1	1.068521	8.388578	110	1	1.138209	0.9;
	325	2	0	0	0	0	1	1.095395	17.445456	380	1	1.163437	0.9;
	326	2	48.95428	15.2	0	0	1	0.97953	61.92687	150	1	1.1	0.9;
	327	1	0	0	0	-16.4774	1	1.006927	-27.662808	150	1	1.194399	0.9;
	328	1	10.0111	4	0	0	1	1.055392	-35.677307	220	1	1.137158	0.9;
	329	1	0	0	0	0	1	1.035646	-15.919004	150	1	1.1	0.9;
	330	1	65.17226	-51.5	0	0	1	1.073567	-22.7269	220	1	1.160366	0.9;
	331	1	0	0	0	0	1	1.048695	-31.028999	220	1	1.109877	0.9;
	332	1	0	0	0	0	1	1.010246	9.382501	220	1	1.1	0.9;
	333	1	56.76294	23.2	0	0	1	1.066165	-0.886317	150	1	1.133361	0.9;
	334	2	373.7	125.6	0	0	1	1.092053	9.444908	380	1	1.126214	0.9;
	335	1	0	0	0	0	1	1.04291	-23.435765	150	1	1.1	0.9;
	336	1	0	0	0	0	1	0.955804	54.65206	220	1	1.1	0.9;
	337	1	0	0	0	0	1	1.057572	-30.223531	380	1	1.123319	0.9;
	338	2	0	0	0	0	1	1.064448	9.242033	150	1	1.1	0.9;
	339	2	328.3	96.3	0	0	1	0.999614	-6.673699	220	1	1.1	0.9;
	340	1	23.82642	7.9	0	0	1	0.965191	-27.482574	110	1	1.1	0.9;
	341	1	-95.7061	-2.8	0	0	1	1.01476	-29.052805	220	1	1.1	0.9;
	342	1	44.04884	0	0	0	1	1.028942	-30.323699	150	1	1.1	0.9;
	343	2	16.51831	3	0	0	1	1.090682	-21.668732	220	1	1.140834	0.9;
	344	1	19.02109	11.5	0	0	1	1.026573	-23.594182	150	1	1.1	0.9;
	345	1	0	0	0	0	1	1.004959	-22.886213	150	1	1.1	0.9;
	346	1	0	0	0	0	1	1.081256	-44.957472	380	1	1.137793	0.9;
	347	1	0	0	0	0	1	1.088634	21.61296	380	1	1.133022	0.9;
	348	2	0	0	0	0	1	1.086247	4.965551	150	1	1.152877	0.9;
	349	1	0	0	0	0	1	0.991411	8.387595	220	1	1.1	0.9;
	350	2	0	0	0	0	1	1.0236	-36.548388	150	1	1.1	0.9;
	351	1	27.93097	10.1	0	0	1	1.021416	-15.357982	110	1	1.122119	0.9;
	352	1	34.93874	21.6	0	0	1	1.050774	-20.66845	150	1	1.1	0.9;
	353	1	134.1	34.2	0	0	1	1.05597	-0.689213	150	1	1.118911	0.9;
	354	1	17.71965	5.9	0	0	1	0.996871	-21.657581	220	1	1.1	0.9;
	355	2	0	0	0	0	1	1.082	42.561895	380	1	1.1	0.9;
	356	1	0	0	0	0	1	1.065657	13.092137	380	1	1.115751	0.9;
	357	1	0	0	0	33.63425	1	0.961936	-29.660674	150	1	1.1	0.876388;
	358	1	77.78624	-40.8	0	0	1	1.049744	-12.44878	220	1	1.11459	0.9;
	359	1	11.91321	3.8	0	0	1	0.99307	-16.90292	110	1	1.1	0.9;
	360	1	5.00555	2	0	0	1	1.060595	-35.381005	220	1	1.141787	0.9;
	361	2	11.01221	1	0	0	1	0.997732	-25.836673	110	1	1.1	0.9;
	362	2	63.06993	40.6	0	0	1	1.021904	-13.644157	220	1	1.113518	0.9;
	363	1	0	0	0	0	1	1.040595	-33.842534	220	1	1.1	0.9;
	364	1	0	0	0.095843	0	1	1.058148	-24.550667	220	1	1.127763	0.9;
	365	2	13.6151	3.8	0	0	1	1.012912	-22.206507	110	1	1.1	0.9;
	366	1	254.7	79.4	0	0	1	1.056855	-44.244899	380	1	1.125112	0.9;
	367	2	-0.6	-0.5	0	0	1	0.949182	-31.519731	110	1	1.1	0.9;
	368	1	129.1	7.4	0	0	1	1.044489	-11.534227	220	1	1.116773	0.9;
	369	2	174.2	52	0	0	1	1.034669	-25.712266	220	1	1.132324	0.9;
	370	2	0	0	0	0	1	1.024636	-34.195277	110	1	1.1	0.9;
	371	2	0	0	0	0	1	1.014864	-6.165324	220	1	1.116168	0.9;
	372	1	20.52275	10.3	0	0	1	1.034874	-24.829211	150	1	1.1	0.9;
	373	1	4.705216	0.7	0	0	1	1.070185	-47.227929	220	1	1.126368	0.9;
	374	1	70.67836	13.5	0	0	1	1.069833	-6.520725	220	1	1.125092	0.9;
	375	1	187.1886	1.8	0	0	1	1.051469	-10.320102	220	1	1.121312	0.9;
	376	1	24.42708	8.7	0	0	1	0.974426	-21.82411	110	1	1.1	0.9;
	377	1	8.809768	3.2	0	0	1	0.970255	-18.625797	110	1	1.1	0.9;
	378	1	0	0	0	0	1	1.082652	16.184558	380	1	1.127847	0.9;
	379	1	33.73741	20.9	0	0	1	0.960682	-25.230096	150	1	1.1	0.9;
	380	1	35.83974	12.5	0	0	1	1.020192	-23.61097	110	1	1.118083	0.9;
	381	2	0	0	0.226002	0	1	1.070579	33.824652	380	1	1.1	0.9;
	382	1	0	0	0	0	1	1.040789	6.718977	380	1	1.1	0.9;
	383	1	123.1	6.5	0	0	1	1.082612	-17.660855	220	1	1.151019	0.9;
	384	1	0	0	0	0	1	1.093485	-31.813916	380	1	1.164568	0.9;
	385	1	0.100111	0	0	0	1	0.986531	-17.527618	110	1	1.1	0.9;
	386	1	26.92986	-22.4	0	0	1	1.047042	-44.072735	220	1	1.1	0.9;
	387	1	29.03219	9	0	0	1	1.013882	-21.68604	110	1	1.1	0.9;
	388	2	331.4	164.8	0	0	1	1.059877	7.883555	220	1	1.142203	0.9;
	389	1	0	0	0	0	1	1.065452	2.455898	380	1	1.144564	0.9;
	390	2	0	0	0	0	1	1.047842	20.80638	380	1	1.107802	0.9;
	391	1	33.13674	11.6	0	0	1	1.039563	-22.920885	110	1	1.143179	0.9;
	392	1	-264	63.3	0	0	1	1.093118	-14.837767	380	1	1.155904	0.9;
	393	1	69.07658	24	0	0	1	1.062436	-35.129169	220	1	1.143983	0.9;
	394	2	0	0	0	0	1	1.021182	61.01308	220	1	1.1	0.9;
	395	2	0	0	0	0	1	1.098682	62.268815	220	1	1.106208	0.9;
	396	2	0	0	0	0	1	1.085545	-32.943003	220	1	1.134776	0.9;
	397	1	0	0	0.209995	0	1	1.075203	-47.422751	380	1	1.132025	0.9;
	398	2	44.84972	7.1	0	0	1	1.0075	-42.326698	220	1	1.1	0.9;
	399	1	-12.614	4.1	0	0	1	1.015555	-23.274057	150	1	1.1	0.9;
	400	1	0	0	0	0	1	1.045135	-12.171396	220	1	1.117577	0.9;
	401	1	36.74074	10.3	0	0	1	1.058974	-40.304851	220	1	1.10776	0.9;
	402	2	105.9	45.7	0	0	1	1.045887	-23.426034	110	1	1.129667	0.9;
	403	2	2.202442	1.3	0	0	1	1.001545	-4.717098	110	1	1.1	0.9;
	404	1	0	0	0	0	1	0.983818	-25.515252	150	1	1.1	0.9;
	405	1	0	0	0	0	1	1.043068	-8.209284	380	1	1.129723	0.9;
	406	1	42.04662	43	0	0	1	1.035254	-20.111554	150	1	1.113029	0.9;
	407	1	0	0	0	0	1	1.062953	23.327858	380	1	1.1	0.9;
	408	2	104.4	27.6	0	0	1	1.012818	-36.91865	110	1	1.1	0.9;
	409	2	0	0	0	0	1	1.077842	-7.453377	380	1	1.13497	0.9;
	410	2	281.2	9.1	0	0	1	1.057826	-29.432357	220	1	1.156699	0.9;
	411	2	70.57825	28.9	0	0	1	0.996818	-23.598987	110	1	1.1	0.9;
	412	1	1.801998	0.7	0	0	1	0.994994	-13.474962	110	1	1.1	0.9;
	413	1	0	0	0.169733	0	1	1.057916	-21.135615	380	1	1.1	0.9;
	414	1	97	40.3	0	0	1	1.030089	-5.556067	220	1	1.109584	0.9;
	415	1	0	0	0	0	1	0.980117	-21.9188	110	1	1.1	0.9;
	416	1	0	0	0	0	1	1.017721	-26.785907	150	1	1.1	0.9;
	417	1	0	0	0	0	1	0.975061	-24.159228	150	1	1.1	0.9;
	418	1	24.32697	7.6	0	0	1	1.001541	-15.95114	110	1	1.1	0.9;
	419	1	44.74961	0.9	0	0	1	1.086771	-13.164256	220	1	1.132184	0.9;
	420	1	0	0	0.119	0	1	0.990433	-23.016834	220	1	1.1	0.9;
	421	1	0	0	0	0	1	0.999928	-23.123736	150	1	1.1	0.9;
	422	1	9.00999	6.8	0	0	1	0.948895	-24.265169	110	1	1.1	0.9;
	423	1	81.0899	48	0	0	1	1.005412	-23.167447	150	1	1.1	0.9;
	424	1	0	0	0	0	1	1.012315	-21.821988	150	1	1.1	0.9;
	425	1	0	0	0	0	1	1.060701	-20.695437	220	1	1.124154	0.9;
	426	1	336.2	131.6	0	0	1	1.081014	5.91166	150	1	1.150023	0.9;
	427	2	-6.60733	1.4	0	0	1	1.045996	-41.849136	220	1	1.110125	0.9;
	428	1	0	0	0	0	1	0.993561	-34.672866	110	1	1.1	0.9;
	429	1	0	0	0	0	1	1.088461	54.25258	380	1	1.119472	0.9;
	430	1	67.3	94.8	0	0	1	0.996111	7.446131	220	1	1.1	0.9;
	431	1	0	0	0	0	1	1.097108	-37.992251	380	1	1.158247	0.9;
	432	1	0	0	0	0	1	1.092705	45.840566	380	1	1.1	0.9;
	433	1	16.61843	3.1	0	0	1	1.024265	-14.345058	110	1	1.1	0.9;
	434	1	5.806438	1.1	0	0	1	1.063304	-5.672326	110	1	1.138715	0.9;
	435	1	0	0	0	0	1	1.064909	-43.610049	220	1	1.114499	0.9;
	436	1	0	0	0	0	1	0.968794	-21.067697	110	1	1.1	0.9;
	437	1	57.26349	15.6	0	0	1	1.046287	-45.347647	220	1	1.105343	0.9;
	438	2	0	0	0	0	1	0.989364	56.583338	220	1	1.1	0.9;
	439	1	0	0	0	0	1	1.077182	-30.675461	380	1	1.150514	0.9;
	440	1	7.00777	2	0	0	1	1.053109	3.801329	150	1	1.10919	0.9;
	441	2	10.41154	5.4	0	0	1	1.011891	3.678516	220	1	1.1	0.9;
	442	2	115.9	14.7	0	0	1	1.031	-43.432127	220	1	1.1	0.9;
	443	1	20.52275	4	0	0	1	1.060755	-21.985393	220	1	1.138282	0.9;
	444	2	0	0	0.265944	490.5867	1	1.099	35.801068	380	1	1.123069	0.9;
	445	1	68.17558	22	0	0	1	0.952381	-38.943375	110	1	1.1	0.9;
	446	1	23.22575	7.6	0	0	1	0.950541	-28.78371	110	1	1.1	0.9;
	447	1	0	0	0	0	1	1.029368	-23.366632	150	1	1.1	0.9;
	448	1	181.7	51	0	0	1	1.072133	47.570741	380	1	1.1	0.9;
	449	1	0	0	0	0	1	1.06281	9.349979	150	1	1.1	0.9;
	450	1	31.23463	11.3	0	0	1	1.031577	-23.207466	110	1	1.1	0.9;
	451	2	423.2	55.3	0	0	1	1.032798	-3.638804	380	1	1.117422	0.9;
	452	1	144.4	35.5	0	0	1	1.045347	-14.231817	220	1	1.1	0.9;
	453	1	0	0	0	0	1	1.012854	-27.407185	110	1	1.1	0.9;
	454	1	0	0	0	0	1	1.04021	-6.27181	220	1	1.129562	0.9;
	455	1	4.1	-607	0.248006	0	1	1.092736	23.67849	380	1	1.132148	0.9;
	456	1	18.62065	6.2	0	0	1	0.970219	-27.534951	110	1	1.1	0.9;
	457	1	0	0	0	0	1	1.100091	-8.858699	380	1	1.167824	0.9;
	458	1	23.42597	8.5	0	0	1	1.056915	-20.934087	110	1	1.114838	0.9;
	459	2	43.7485	13.7	0	0	1	0.960088	-20.167411	110	1	1.1	0.9;
	460	1	77.28568	7	0	0	1	1.05739	-54.985593	220	1	1.1	0.9;
	461	1	0	0	0	0	1	1.061946	-20.329748	220	1	1.125233	0.9;
	462	2	0	0	0.26688	0	1	1.078553	57.975376	380	1	1.1	0.9;
	463	1	10.11121	2.8	0	0	1	0.985078	-24.175429	110	1	1.1	0.9;
	464	1	57.46371	17.1	0	0	1	1.036071	-7.033356	220	1	1.125468	0.9;
	465	1	9.610656	3.1	0	0	1	0.950378	-34.336303	110	1	1.1	0.9;
	466	2	-195.216	-19	0	0	1	1.075364	-18.939255	220	1	1.126952	0.9;
	467	2	0	0	0	0	1	1.075818	-32.766717	220	1	1.177684	0.9;
	468	1	7.708547	16.7	0	0	1	0.986269	-27.842525	150	1	1.1	0.9;
	469	2	602.9	201.4	0	0	1	1.031395	-1.819847	380	1	1.1	0.9;
	470	1	207.2	9.4	0	0	1	1.048016	-46.06973	220	1	1.110116	0.9;
	471	1	118.5	20.4	0	0	1	1.045585	-26.971909	220	1	1.123231	0.9;
	472	1	14.71632	5.3	0	0	1	1.011979	-22.426698	110	1	1.110924	0.9;
	473	1	11.61288	7.2	0	0	1	1.033781	-25.505581	150	1	1.1	0.9;
	474	1	189.9	74.7	0	0	1	1.076315	5.25489	150	1	1.146115	0.9;
	475	1	36.54051	27.4	0	0	1	1.006839	14.481993	220	1	1.1	0.9;
	476	1	176.7	25.6	0	0	1	1.042456	-44.891763	220	1	1.1	0.9;
	477	1	277.7	45.4	0	0	1	1.033882	-42.508442	220	1	1.1	0.9;
	478	1	0	0	0	0	1	1.013266	-47.679933	220	1	1.1	0.9;
	479	1	0	0	0	0	1	1.058796	-15.004281	380	1	1.115606	0.9;
	480	1	0	0	0	0	1	1.07775	-15.093516	380	1	1.121097	0.9;
	481	1	0	0	0	0	1	1.019064	11.085555	220	1	1.1	0.9;
	482	1	16.51831	3.8	0	0	1	0.940317	-25.253513	110	1	1.1	0.9;
	483	1	11.51276	3.8	0	0	1	0.956985	-32.157302	110	1	1.1	0.9;
	484	1	87.79734	-1.4	0.085004	0	1	1.095133	-43.920539	380	1	1.152135	0.9;
	485	1	27.53052	17	0	0	1	1.000264	-22.280101	150	1	1.1	0.9;
	486	1	21.02331	13	0	0	1	0.968559	-29.923957	150	1	1.1	0.9;
	487	1	57.96427	18.9	0	0	1	1.05206	2.750607	150	1	1.109524	0.9;
	488	1	166.9	86.1	0	210.9002	1	1.005852	11.422371	220	1	1.1	0.9;
	489	1	2.00222	0.7	0	0	1	1.025532	-16.821682	110	1	1.127459	0.9;
	490	1	79.68835	39.6	0	0	1	1.077184	-47.177539	380	1	1.133741	0.9;
	491	1	0	0	0	0	1	1.058459	-17.763846	220	1	1.125595	0.9;
	492	1	0	0	0	-16.557	1	1.016452	-26.858608	150	1	1.183928	0.9;
	493	1	30.0333	14	0	0	1	1.059733	-35.465613	220	1	1.140988	0.9;
	494	1	30.9343	8.9	0	0	1	1.067795	11.013566	110	1	1.146091	0.9;
	495	1	0	0	0	0	1	1.070269	-10.116176	220	1	1.153027	0.9;
	496	1	0	0	0	0	1	1.058763	-41.33226	220	1	1.114597	0.9;
	497	1	17.71965	5.9	0	0	1	0.904142	-32.155554	110	1	1.1	0.9;
	498	1	180.8	90.8	0	0	1	0.991133	-39.099122	150	1	1.1	0.9;
	499	1	0	0	0	0	1	1.071157	-19.658311	220	1	1.136823	0.9;
	500	1	0	0	0	0	1	1.06005	-25.172393	220	1	1.124149	0.9;
	501	1	16.51831	9	0	0	1	1.061003	-21.745314	220	1	1.138404	0.9;
	502	1	0	0	0	0	1	1.030144	-9.265439	220	1	1.111886	0.9;
	503	2	0	0	0	0	1	1.04818	15.421511	150	1	1.1	0.9;
	504	1	42.24684	16.6	0	0	1	1.081374	3.930379	150	1	1.149446	0.9;
	505	1	3.804218	1.3	0	0	1	1.018785	-11.955025	110	1	1.1	0.9;
	506	1	0	0	0	0	1	1.052531	-10.711361	220	1	1.133335	0.9;
	507	1	0	0	0	0	1	1.015579	-3.731791	220	1	1.1	0.9;
	508	1	0	0	0	0	1	1.03851	-24.498833	150	1	1.1	0.9;
	509	2	58.56493	-0.1	0	0	1	1.040728	-46.512792	220	1	1.171698	0.9;
	510	1	198.4	32.7	0	0	1	1.018483	-22.201732	220	1	1.1	0.9;
	511	1	5.806438	3.8	0	0	1	0.951679	-27.11965	110	1	1.1	0.9;
	512	1	0	0	0	0	1	1.088721	-32.900237	380	1	1.137318	0.9;
	513	1	20.22242	7.3	0	0	1	1.036617	-22.89367	110	1	1.132621	0.9;
	514	1	0	0	0	0	1	1.092128	-20.100979	380	1	1.160976	0.9;
	515	1	47.95317	29.7	0	0	1	0.954046	-25.902941	150	1	1.1	0.9;
	516	1	120.6337	56.2	0	0	1	1.006567	12.647319	220	1	1.1	0.9;
	517	1	27.13008	8.9	0	0	1	1.051536	-11.788539	110	1	1.130226	0.9;
	518	1	10.71188	3.5	0	0	1	0.961477	-35.326019	110	1	1.1	0.9;
	519	1	82.89191	26.8	0	0	1	1.018149	-23.182896	110	1	1.1	0.9;
	520	1	0	0	0	0	1	1.046577	-11.266074	220	1	1.117796	0.9;
	521	1	108.6	38.7	0	0	1	1.014391	-14.608178	110	1	1.1	0.9;
	522	2	0	0	0	0	1	1.078858	20.442128	220	1	1.161277	0.9;
	523	2	166	-8.8	0	0	1	0.997091	20.068849	220	1	1.1	0.9;
	524	1	17.3192	6.6	0	0	1	1.077814	-47.108294	380	1	1.114056	0.9;
	525	1	322.7	11.9	0	0	1	1.043444	-38.335942	380	1	1.111349	0.9;
	526	1	51.15672	18.2	0	0	1	1.073726	-18.965303	220	1	1.113031	0.9;
	527	2	303.6	79.8	0	0	1	1.057636	-14.590902	220	1	1.111291	0.9;
	528	1	151.2	72.4	0	0	1	1.02113	5.007866	380	1	1.1	0.9;
	529	1	21.12342	8.3	0	0	1	1.075899	2.86124	150	1	1.145142	0.9;
	530	1	9.510545	3.2	0	0	1	1.017403	-22.354383	110	1	1.1	0.9;
	531	2	3.00333	0.2	0	0	1	1.0595	1.211549	220	1	1.126202	0.9;
	532	1	1.501665	0.8	0	0	1	0.984147	-25.3661	110	1	1.1	0.9;
	533	1	27.02997	9.8	0	0	1	1.04692	-21.588347	110	1	1.107946	0.9;
	534	2	300.1	126.3	0	0	1	1.094921	56.239847	380	1	1.114706	0.9;
	535	1	0	0	0	0	1	0.977568	-27.132905	110	1	1.1	0.9;
	536	1	0	0	0	0	1	1.03131	4.041293	220	1	1.1	0.9;
	537	1	0	0	0	0	1	1.072705	10.807187	380	1	1.123994	0.9;
	538	1	67.57492	17	0	0	1	1.013163	-8.636851	220	1	1.112696	0.9;
	539	1	47.55272	5.3	0	0	1	1.048075	-46.061821	220	1	1.111559	0.9;
	540	1	24.82753	7.6	0	0	1	0.936642	-24.930101	110	1	1.1	0.9;
	541	1	33.33696	15.7	0	0	1	1.090217	-43.441063	380	1	1.155752	0.9;
	542	1	0	0	0	0	1	0.991741	-20.692445	150	1	1.1	0.9;
	543	1	0	0	0	0	1	1.010913	-19.593834	220	1	1.1	0.9;
	544	1	0	0	0	0	1	1.040607	3.750387	380	1	1.110782	0.9;
	545	1	140.6	42.6	0	0	1	1.069518	-13.520167	220	1	1.132823	0.9;
	546	1	110.8	-8.8	0	0	1	1.070342	-42.699287	220	1	1.118873	0.9;
	547	1	48.9	28.6	0	0	1	1.077907	-15.344549	380	1	1.146577	0.9;
	548	1	33.23685	4.4	0	0	1	1.046028	-12.56487	380	1	1.112167	0.9;
	549	1	0	0	0	0	1	1.069062	31.546178	380	1	1.1	0.9;
	550	1	0	0	0	0	1	1.05654	-11.287673	330	1	1.148955	0.9;
	551	1	150.9	10.4	0	0	1	1.036054	-32.584651	220	1	1.131289	0.9;
	552	1	0	0	0	0	1	1.026875	-14.094244	220	1	1.1	0.9;
	553	1	0	0	0	0	1	1.02753	-50.882772	150	1	1.1	0.9;
	554	1	0	0	0.085901	0	1	1.075706	-10.638282	380	1	1.133284	0.9;
	555	1	0	0	0	0	1	1.016872	-25.693887	110	1	1.1	0.9;
	556	1	18.5	3.7	0	0	1	0.964544	-30.567292	150	1	1.1	0.9;
	557	2	0	0	0	0	1	1.096591	-10.595994	220	1	1.17165	0.9;
	558	2	0	0	0	0	1	1.079273	-12.189346	220	1	1.141595	0.9;
	559	2	8.809768	1.3	0	0	1	0.973935	-24.453497	110	1	1.1	0.9;
	560	2	0.900999	74.1	0	0	1	1.082367	-44.35254	380	1	1.118898	0.9;
	561	1	72.88081	14.4	0	0	1	1.057314	-24.09848	220	1	1.132006	0.9;
	562	1	15.41709	5.8	0	0	1	1.012667	-14.766434	110	1	1.1	0.9;
	563	1	21.42375	8.1	0	0	1	1.03714	-47.366002	220	1	1.1	0.9;
	564	1	0	0	0	0	1	1.099017	60.575265	380	1	1.116941	0.9;
	565	1	111.4	36.5	0	20	1	1.080951	-9.661245	380	1	1.145625	0.9;
	566	1	0	0	0	0	1	1.025239	-23.241188	220	1	1.1	0.9;
	567	1	0	0	0	0	1	1.002386	-26.000495	150	1	1.1	0.9;
	568	1	116.8	60.3	0	0	1	1.002737	-27.984949	150	1	1.1	0.9;
	569	2	0	0	0	0	1	0.992455	-30.836531	220	1	1.1	0.9;
	570	1	177.2	32.3	0	0	1	1.043347	-16.985065	220	1	1.1	0.9;
	571	1	0	0	0	0	1	1.063857	8.160507	380	1	1.125465	0.9;
	572	1	0	0	0	0	1	1.003202	-23.569718	150	1	1.1	0.9;
	573	1	0	0	0	0	1	1.035331	-21.602683	150	1	1.111941	0.9;
	574	1	2.502775	0	0	0	1	0.939513	-25.220281	110	1	1.1	0.9;
	575	1	0	0	0	0	1	1.009138	10.749861	220	1	1.1	0.9;
	576	2	7.908769	2.5	0	0	1	1.056636	-18.681614	110	1	1.118066	0.9;
	577	2	202	24.6	0	0	1	1.020676	15.654738	220	1	1.1	0.9;
	578	2	179.9	-18.1	0	0	1	1.050773	-29.649309	220	1	1.144685	0.9;
	579	1	102	26.6	0	0	1	1.023843	37.784045	220	1	1.1	0.9;
	580	2	46.55161	18.4	0	0	1	1.081123	-12.616963	220	1	1.135379	0.9;
	581	2	-24.6273	159.9	0	0	1	1.085368	-45.64343	380	1	1.12659	0.9;
	582	1	15.5172	6.3	0	0	1	1.072329	-0.479283	150	1	1.139169	0.9;
	583	1	26.32919	9.5	0	0	1	1.017177	-20.84962	110	1	1.109312	0.9;
	584	2	0	0	0	0	1	1.05456	-15.677227	150	1	1.106461	0.9;
	585	1	48.05328	15.8	0	0	1	1.06482	63.864513	150	1	1.1	0.9;
	586	2	0	0	0	0	1	1.054773	61.445595	220	1	1.108396	0.9;
	587	1	40.9454	19.2	0	0	1	1.064975	-43.910399	220	1	1.113835	0.9;
	588	1	17.21909	4.4	0	0	1	1.051264	7.851991	110	1	1.127189	0.9;
	589	2	79.98869	10.9	0	0	1	0.950047	64.765831	150	1	1.1	0.9;
	590	1	0	0	0	0	1	1.056145	-9.787783	220	1	1.12766	0.9;
	591	1	10.31143	2.3	0	0	1	0.962021	-13.410075	110	1	1.1	0.9;
	592	1	205.6	74	0	0	1	1.039073	-22.588184	220	1	1.1	0.9;
	593	2	0	0	0	0	1	1.029293	13.66667	220	1	1.106726	0.9;
	594	1	65.27237	30.9	0	0	1	1.010518	10.828212	220	1	1.1	0.9;
	595	1	258.4	66.7	0	0	1	1.060491	-16.808833	220	1	1.139974	0.9;
	596	1	0	0	0.174464	20	1	1.085071	1.446342	380	1	1.134312	0.9;
	597	1	584.3	166.3	0	0	1	1.082237	-40.354654	380	1	1.151867	0.9;
	598	1	0	0	0	0	1	1.081054	0.960033	380	1	1.132136	0.9;
	599	1	0	0	0	0	1	1.050487	9.632286	380	1	1.110115	0.9;
	600	1	0	0	0.119999	0	1	0.973603	-24.18199	220	1	1.1	0.9;
	601	1	38.04218	11.7	0	0	1	1.013055	-28.447765	110	1	1.1	0.9;
	602	1	0	0	0	0	1	1.05194	-17.59937	220	1	1.10781	0.9;
	603	2	0	0	0	0	1	1.07087	-5.228918	380	1	1.12423	0.9;
	604	1	0	0	0	0	1	1.052359	-27.841534	220	1	1.1	0.9;
	605	1	0	0	0	-15.6219	1	1.024517	-23.233757	150	1	1.1	0.9;
	606	1	5.506105	17.7	0	0	1	0.995949	-22.378624	150	1	1.1	0.9;
	607	1	0	0	0	10.91344	1	1.048697	-20.010721	150	1	1.1	0.9;
	608	2	0	-0.3	0	0	1	1.063973	-41.996651	380	1	1.129411	0.9;
	609	1	0	0	0	0	1	1.027265	12.552226	380	1	1.1	0.9;
	610	1	0	0	0	0	1	1.110436	-10.188562	380	1	1.170712	0.9;
	611	1	5.105661	-0.1	0	0	1	0.981868	-21.643463	110	1	1.1	0.9;
	612	1	109.3	35.4	0	0	1	1.069406	4.526731	150	1	1.139443	0.9;
	613	1	30.83419	10.8	0	0	1	0.996149	-25.902399	110	1	1.1	0.9;
	614	1	232.1	89.8	0	0	1	1.052765	-20.45451	220	1	1.112061	0.9;
	615	1	0	0	0	0	1	1.080299	-15.43131	220	1	1.137345	0.9;
	616	1	79.58824	50.5	0	0	1	1.043128	8.683305	380	1	1.1	0.9;
	617	1	0	0	0	0	1	1.088737	-32.903548	380	1	1.137371	0.9;
	618	2	0	0	0	0	1	1.084227	-9.495118	220	1	1.154876	0.9;
	619	1	169.8	55.5	0	0	1	1.000498	11.419469	220	1	1.1	0.9;
	620	1	60.0666	28	0	0	1	1.015238	-47.069623	220	1	1.1	0.9;
	621	1	0	0	0	0	1	1.100635	-16.433157	380	1	1.164547	0.9;
	622	2	42.34695	13.1	0	0	1	1.053044	-42.094203	220	1	1.108978	0.9;
	623	1	0	0	0	0	1	1.070885	-4.95474	220	1	1.136646	0.9;
	624	1	-3.40377	-1.3	0	0	1	1.048712	-22.206669	220	1	1.1	0.9;
	625	1	124.8	23.5	0	0	1	0.981202	-18.295145	220	1	1.1	0.9;
	626	1	22.42486	8.2	0	0	1	1.022882	-29.78219	110	1	1.1	0.9;
	627	2	0	0	0	0	1	1.09875	-36.043391	120	1	1.177368	0.9;
	628	1	0	0	0	0	1	1.094993	-36.282578	380	1	1.147488	0.9;
	629	1	0	0	0.124999	0	1	1.010466	-19.896256	220	1	1.1	0.9;
	630	1	96.40689	26.5	0	0	1	1.021471	-34.04119	220	1	1.1	0.9;
	631	2	13.01443	8.1	0	0	1	1.073545	-15.08163	150	1	1.12018	0.9;
	632	2	0	0	0	0	1	1.096316	-36.938434	380	1	1.161895	0.9;
	633	2	0	0	0	0	1	1.016727	-25.873368	220	1	1.1	0.9;
	634	1	36.54051	8.4	0	0	1	1.050616	-47.882481	220	1	1.1	0.9;
	635	1	12.7141	0.8	0	0	1	0.936202	-25.332066	110	1	1.1	0.9;
	636	1	20.32253	6.7	0	0	1	1.005987	-32.586588	110	1	1.1	0.9;
	637	1	0	0	0	0	1	1.007914	-31.413473	150	1	1.1	0.9;
	638	1	0	0	0	0	1	1.094656	-14.544391	380	1	1.152521	0.9;
	639	1	41.04551	14.3	0	0	1	1.05586	-27.16557	220	1	1.118401	0.9;
	640	1	114.8	22.8	0	0	1	1.056118	-39.867239	220	1	1.105695	0.9;
	641	1	83.99313	21.2	0	0	1	1.057489	-25.087192	220	1	1.14444	0.9;
	642	1	0	0	0	0	1	1.016196	-47.303864	220	1	1.1	0.9;
	643	2	0	0	0	-129.525	1	1.085395	-38.006581	380	1	1.164155	0.9;
	644	1	76.38469	5.1	0	0	1	1.05869	-41.127056	220	1	1.120693	0.9;
	645	2	423.4	215.2	0	0	1	1.055184	-0.174084	380	1	1.135238	0.9;
	646	1	110.4	41.9	0	0	1	1.049086	-0.355982	150	1	1.123976	0.9;
	647	2	0	0	0	0	1	0.991455	-4.755008	110	1	1.1	0.9;
	648	2	0	0	0	0	1	1.092237	-8.089901	380	1	1.128297	0.9;
	649	1	29.63285	10.8	0	0	1	1.036944	-19.884576	110	1	1.1	0.9;
	650	1	4.504995	-5.6	0	0	1	1.010496	-19.629822	110	1	1.1	0.9;
	651	2	152.3	46.2	0	0	1	1.039331	-24.598821	220	1	1.123711	0.9;
	652	2	0	0	0	0	1	1.043796	-18.556075	220	1	1.117708	0.9;
	653	1	0	0	0	0	1	0.984414	-26.529834	380	1	1.1	0.9;
	654	1	0	0	0	0	1	1.06525	-23.139584	220	1	1.122851	0.9;
	655	1	0	0	0	0	1	1.087705	1.081026	150	1	1.156613	0.9;
	656	1	30.83419	6.6	0	0	1	1.006528	-36.876182	110	1	1.1	0.9;
	657	1	0	0	0	-25.7162	1	1.103275	-16.201189	380	1	1.16219	0.9;
	658	1	400.3	189.9	0	0	1	1.061657	-5.79777	220	1	1.127936	0.9;
	659	2	34.93874	-32.3	0	0	1	1.097447	-7.91391	380	1	1.132198	0.9;
	660	1	0	0	0	0	1	1.040387	-20.642275	150	1	1.1	0.9;
	661	1	20.62287	6.5	0	0	1	0.998773	-37.10915	110	1	1.1	0.9;
	662	1	0	0	0	0	1	1.082696	-45.346098	380	1	1.141269	0.9;
	663	1	0	0	0	0	1	1.063546	19.952996	380	1	1.127169	0.9;
	664	1	58.36471	10.7	0	0	1	1.02225	2.801027	110	1	1.1	0.9;
	665	2	206	110.9	0	0	1	1.04125	-0.687841	220	1	1.111446	0.9;
	666	2	60.26682	3	0	0	1	1.033004	61.84178	150	1	1.1	0.9;
	667	1	8.108991	2	0	0	1	0.95955	-24.453893	110	1	1.1	0.9;
	668	2	11.21221	3.5	0	0	1	1.016545	-34.347516	110	1	1.1	0.9;
	669	2	1.301443	59.6	0	0	1	1.079789	-38.547778	380	1	1.141483	0.9;
	670	1	22.22464	13.7	0	0	1	1.010953	-22.101037	150	1	1.1	0.9;
	671	1	0	0	0	0	1	0.979566	-17.226484	110	1	1.1	0.9;
	672	1	0	0	0	0	1	1.071057	-4.956251	220	1	1.136827	0.9;
	673	1	5.00555	2	0	0	1	1.030998	-22.748654	220	1	1.110372	0.9;
	674	1	6.507215	4.1	0	0	1	1.053309	-15.190851	150	1	1.1	0.9;
	675	1	90.40023	15.2	0	0	1	1.014668	-20.259451	220	1	1.1	0.9;
	676	1	2.502775	1.3	0	0	1	0.941454	-27.824941	110	1	1.1	0.9;
	677	2	69.97758	21.3	0	0	1	0.989364	-24.886173	110	1	1.1	0.9;
	678	1	0	0	0	0	1	1.077267	-38.424914	220	1	1.137734	0.9;
	679	2	16.51831	7.8	0	0	1	0.995136	-36.974928	220	1	1.11826	0.9;
	680	1	0	0	0	0	1	1.073044	-23.237951	220	1	1.138188	0.9;
	681	2	142.6	50.6	0	0	1	1.023544	-39.765126	220	1	1.1	0.9;
	682	1	0	0	0	0	1	1.08267	-24.837295	220	1	1.122374	0.9;
	683	1	1.902109	0.6	0	0	1	0.995543	-22.609869	110	1	1.1	0.9;
	684	1	132.1	32.4	0	0	1	1.093563	-40.631783	380	1	1.134003	0.9;
	685	1	36.74074	8.9	0	0	1	1.010469	-12.668448	110	1	1.1	0.9;
	686	2	0	0	0	0	1	1.066816	-13.252332	380	1	1.132447	0.9;
	687	1	0	0	0	0	1	0.968559	-29.923957	150	1	1.1	0.9;
	688	1	0	0	0	0	1	1.082469	-24.635807	220	1	1.152099	0.9;
	689	1	0	0	0	0	1	1.054	-12.322977	110	1	1.116334	0.9;
	690	1	0	0	0	0	1	1.014676	-16.703663	110	1	1.112885	0.9;
	691	1	0	0	0	0	1	1.019974	-23.37169	150	1	1.1	0.9;
	692	2	0	0	0	0	1	1.081455	-28.862538	220	1	1.133577	0.9;
	693	2	223.4	107.8	0	0	1	1.000409	-1.210827	220	1	1.1	0.9;
	694	2	0	0	0	0	1	1.062591	-35.079996	220	1	1.144159	0.9;
	695	1	0	0	0	0	1	1.05028	-20.49844	220	1	1.13898	0.9;
	696	1	0	0	0	0	1	0.939995	-29.215258	150	1	1.1	0.891359;
	697	1	0	0	0	0	1	0.995714	-16.713646	150	1	1.1	0.9;
	698	2	153	40	0	0	1	1.065508	-10.9882	220	1	1.135997	0.9;
	699	1	0	0	0	0	1	1.085545	0.805565	220	1	1.184849	0.9;
	700	1	46.15117	20.9	0	0	1	1.03043	-35.060062	110	1	1.1	0.9;
	701	1	343.9	141.9	0	0	1	1.036593	-13.466759	220	1	1.1	0.9;
	702	2	0	0	0	0	1	0.988124	-16.638085	110	1	1.1	0.9;
	703	2	0	0	0	0	1	1.081084	5.919215	150	1	1.150496	0.9;
	704	2	142.2	29.6	0	0	1	1.063773	-12.379783	220	1	1.135275	0.9;
	705	2	0	0	0	0	1	1.078	-6.313339	220	1	1.136738	0.9;
	706	2	0	0	0	0	1	1.068842	18.5322	380	1	1.135758	0.9;
	707	2	0	0	0	0	1	1.024136	-40.423528	220	1	1.1	0.9;
	708	1	34.03774	11	0	0	1	0.998449	-40.215058	110	1	1.1	0.9;
	709	1	0	0	0	0	1	1.070257	-18.149929	220	1	1.1	0.9;
	710	1	0	0	0	0	1	1.032508	-22.742899	150	1	1.1	0.9;
	711	2	0	0	0	0	1	1.076152	3.335431	150	1	1.145865	0.9;
	712	1	0	0	0	0	1	1.057221	-29.951981	220	1	1.111664	0.9;
	713	1	0	0	0	0	1	1.041123	-18.611205	220	1	1.1	0.9;
	714	1	18.01998	6.4	0	0	1	0.935692	-25.509866	110	1	1.1	0.9;
	715	1	0	0	0	0	1	1.073369	-5.732901	220	1	1.127977	0.9;
	716	2	0	0	0	0	1	1.077179	9.472781	110	1	1.146118	0.9;
	717	2	52.05772	0	0	0	1	1.103342	-5.260136	380	1	1.148683	0.9;
	718	1	295.3	41.9	0	0	1	1.083378	-6.042288	150	1	1.138998	0.9;
	719	1	0	0	0	0	1	1.014657	-14.948117	220	1	1.1	0.9;
	720	2	0	0	0	0	1	1.073864	-13.952376	220	1	1.129758	0.9;
	721	2	0	0	0	0	1	1.014545	-39.036205	110	1	1.1	0.9;
	722	1	2	114.6	0	0	1	1.065212	-35.856476	220	1	1.125861	0.9;
	723	1	235.2	41.1	0	0	1	1.018784	14.701661	380	1	1.1	0.9;
	724	1	26.72964	10.8	0	0	1	0.990239	-26.326965	110	1	1.1	0.9;
	725	1	50.55605	18	0	0	1	1.04336	-12.076378	220	1	1.116287	0.9;
	726	1	9.710767	3	0	0	1	0.94674	-29.320349	110	1	1.1	0.9;
	727	1	0	0	0	0	1	1.082658	-40.003793	380	1	1.152365	0.9;
	728	2	200.4	60.5	0	0	1	1.038955	-24.396361	220	1	1.121942	0.9;
	729	1	0	0	0	0	1	1.096935	-7.047745	220	1	1.165831	0.9;
	730	1	124.5	29.6	0	0	1	1.05892	-10.601115	220	1	1.115942	0.9;
	731	1	0	0	0	0	1	1.031214	6.71888	380	1	1.1	0.9;
	732	2	222.5	262.8	0	0	1	1.085841	-34.273374	380	1	1.128044	0.9;
	733	1	24.52719	8.6	0	0	1	1.043736	-45.794562	220	1	1.1	0.9;
	734	1	0	0	0	0	1	1.076669	-0.969106	380	1	1.138592	0.9;
	735	1	0	0	0	10.61976	1	1.06309	-20.119799	150	1	1.1	0.9;
	736	2	0	0	0	0	1	1.059217	-20.297167	220	1	1.1	0.9;
	737	1	0	0	0	0	1	0.971733	-24.413773	150	1	1.1	0.9;
	738	1	287.7	16.3	0	0	1	1.072584	-45.683359	380	1	1.111267	0.9;
	739	2	22.12453	7.5	0	0	1	1.03953	-20.176584	110	1	1.146638	0.9;
	740	1	234.7	-12	0	0	1	1.092618	-30.870659	380	1	1.134719	0.9;
	741	1	333.1	17.5	0	0	1	1.03835	-42.812321	380	1	1.106323	0.9;
	742	1	0	0	0	0	1	1.039768	54.342291	220	1	1.1	0.9;
	743	1	0	0	0	0	1	1.090656	-0.669868	220	1	1.154776	0.9;
	744	1	0	0	0	-14.7192	1	1.016356	-26.875174	150	1	1.18734	0.9;
	745	1	121.9	-0.2	0	0	1	1.025139	-6.006674	220	1	1.1	0.9;
	746	2	0	0	0	0	1	1.084658	-2.629753	380	1	1.148477	0.9;
	747	1	0	0	0	0	1	1.012539	-31.933832	220	1	1.1	0.9;
	748	1	31.93541	7.7	0	0	1	1.053193	-28.380077	110	1	1.1	0.9;
	749	1	167.8	77	0	0	1	1.059471	-11.775101	380	1	1.128586	0.9;
	750	1	113.6	122.7	0	0	1	1.026817	16.096788	380	1	1.1	0.9;
	751	1	303.1	75.5	0	0	1	1.051214	-38.437215	380	1	1.116046	0.9;
	752	2	0	0	0	0	1	0.999909	-22.239846	110	1	1.1	0.9;
	753	2	0	0	0	0	1	1.013227	-47.46786	220	1	1.1	0.9;
	754	1	15.01665	10	0	0	1	1.028939	-20.791212	110	1	1.1	0.9;
	755	1	52.7585	17.2	0	0	1	1.054022	-36.296311	220	1	1.136256	0.9;
	756	1	0	0	0	0	1	1.077362	-18.302932	380	1	1.120425	0.9;
	757	2	0	0	0	0	1	1.066136	-38.885019	220	1	1.144107	0.9;
	758	1	615.8	286.7	0	0	1	1.068649	2.419407	150	1	1.140151	0.9;
	759	1	0	0	0	0	1	1.072841	-2.173491	380	1	1.135397	0.9;
	760	1	0	0	0	0	1	1.042197	-16.65121	150	1	1.1	0.9;
	761	1	0	0	0.148479	0	1	1.081358	-17.198493	380	1	1.151123	0.9;
	762	1	0	0	0	0	1	1.03341	-4.165138	220	1	1.1	0.9;
	763	1	0	0	0	0	1	1.005459	-35.742258	110	1	1.1	0.9;
	764	2	0	0	0	0	1	1.091289	-8.207248	380	1	1.128271	0.9;
	765	1	0	0	0	0	1	0.968077	-24.776211	150	1	1.1	0.9;
	766	1	159.2	59.9	0	0	1	1.095555	-13.408463	380	1	1.139749	0.9;
	767	2	62.7696	0.7	0	0	1	1.045867	61.77888	150	1	1.1	0.9;
	768	2	0	0	0	0	1	1.087553	-37.636356	380	1	1.128101	0.9;
	769	1	0	0	0	0	1	0.952473	-26.817877	110	1	1.1	0.9;
	770	2	34.43818	4.4	0	0	1	1.065727	-19.378858	220	1	1.12436	0.9;
	771	1	190	43.3	0	0	1	1.028902	-10.954463	220	1	1.1	0.9;
	772	1	49.95539	7.5	0	0	1	1.056166	-49.275668	220	1	1.113831	0.9;
	773	1	10.0111	3.3	0	0	1	1.003235	-29.621822	110	1	1.1	0.9;
	774	1	-28.9321	9.5	0	0	1	0.990393	-16.810805	150	1	1.1	0.9;
	775	1	21.72409	4.8	0	0	1	0.967182	-26.708669	110	1	1.1	0.9;
	776	2	0	0	0	0	1	1.081211	0.098725	380	1	1.14275	0.9;
	777	1	0	0	0	0	1	1.040712	-21.289009	150	1	1.117084	0.9;
	778	2	0	0	0	0	1	1.054546	-25.486784	110	1	1.116172	0.9;
	779	1	36.4404	12.5	0	0	1	0.962557	-25.771318	110	1	1.1	0.9;
	780	1	1.301443	0.6	0	0	1	0.942577	-27.496595	110	1	1.1	0.9;
	781	1	43.44817	26.9	0	0	1	1.030691	-21.834225	150	1	1.107752	0.9;
	782	1	0	0	0	0	1	0.978946	46.587878	220	1	1.1	0.9;
	783	1	25.12786	14	0	0	1	1.028193	-4.425097	220	1	1.1	0.9;
	784	1	13.81532	5	0	0	1	1.044959	-17.314953	110	1	1.128736	0.9;
	785	1	94.20444	5.5	0	0	1	1.07354	-42.317747	220	1	1.1221	0.9;
	786	1	110	3.4	0	0	1	1.0727	-23.386228	220	1	1.137907	0.9;
	787	2	120.7	-16.9	0	0	1	1.057058	-11.613944	220	1	1.136452	0.9;
	788	1	143.9	123.2	0	0	1	1.023641	5.267888	380	1	1.1	0.9;
	789	1	0	0	0	11.12461	1	1.038653	-24.43639	150	1	1.1	0.9;
	790	2	0	0	0	0	1	1.081263	-34.709792	380	1	1.159949	0.9;
	791	1	42.24684	13.9	0	0	1	0.984869	-22.171686	110	1	1.1	0.9;
	792	1	2.00222	0.6	0	0	1	0.979864	-25.175252	110	1	1.1	0.9;
	793	1	0	0	0	0	1	1.085208	8.56556	150	1	1.133641	0.9;
	794	1	7.708547	3.3	0	0	1	1.001774	-39.954836	110	1	1.1	0.9;
	795	1	0	0	0	0	1	1.091067	-15.940967	220	1	1.179934	0.9;
	796	1	0	0	0	0	1	1.039167	-15.877499	150	1	1.1	0.9;
	797	1	394.9	77	0	0	1	1.054342	-43.772308	380	1	1.122849	0.9;
	798	1	192.5	78.3	0	0	1	1.063145	-9.400205	220	1	1.130475	0.9;
	799	1	13.81532	5	0	0	1	1.037779	-22.836501	110	1	1.134963	0.9;
	800	1	23.02553	14.3	0	0	1	1.01978	-23.607485	150	1	1.1	0.9;
	801	2	0	0	0	0	1	1.032767	12.765979	220	1	1.1	0.9;
	802	1	0	0	0.34231	0	1	1.081228	-17.801039	380	1	1.143828	0.9;
	803	2	17.41931	11.9	0.122998	0	1	1.061636	-16.969741	220	1	1.118747	0.9;
	804	1	3.904329	1.4	0	0	1	1.028501	-23.148996	110	1	1.107832	0.9;
	805	1	0	0	0	0	1	1.068788	-1.068902	220	1	1.138217	0.9;
	806	2	0	0	0	20	1	1.068474	-18.145672	380	1	1.163599	0.9;
	807	2	452.5	113.2	0	0	1	1.044273	-27.200919	220	1	1.120973	0.9;
	808	1	0	0	0	0	1	1.077735	-35.161598	220	1	1.156641	0.9;
	809	2	24.92764	13.7	0	0	1	1.016409	-19.78518	220	1	1.1	0.9;
	810	1	0	0	0.259062	0	1	1.073504	-23.02874	380	1	1.142499	0.9;
	811	1	25.62842	7.9	0	0	1	1.037784	9.111298	110	1	1.117638	0.9;
	812	1	0	0	0	0	1	1.092821	-20.270457	380	1	1.161218	0.9;
	813	1	0	0	0	0	1	1.043609	-15.093189	220	1	1.130499	0.9;
	814	1	0	0	0	11.74494	1	1.010896	-21.971618	150	1	1.1	0.9;
	815	2	0	0	0	0	1	0.996099	-18.388802	110	1	1.1	0.9;
	816	1	82.69168	20.1	0	0	1	1.075891	-25.523591	220	1	1.165076	0.9;
	817	1	146.9	60.6	0	0	1	1.052998	-36.179675	220	1	1.135387	0.9;
	818	1	0	0	0	0	1	1.063944	-14.034728	380	1	1.115583	0.9;
	819	1	0	0	0	0	1	1.00395	-22.862459	150	1	1.1	0.9;
	820	1	0	0	0	0	1	1.075332	-9.043212	220	1	1.142042	0.9;
	821	1	125.9	-0.5	0	0	1	1.072965	-7.67217	220	1	1.143239	0.9;
	822	1	12.61399	4.5	0	0	1	1.000469	-18.781015	110	1	1.1	0.9;
	823	1	13.21465	2.2	0	0	1	1.010475	-36.627746	110	1	1.1	0.9;
	824	1	22.02442	13.2	0	0	1	1.038626	-21.38617	150	1	1.115133	0.9;
	825	2	0	0	0	0	1	1.047395	-2.297436	380	1	1.111996	0.9;
	826	1	0	0	0	-50.7692	1	1.067976	-17.521788	380	1	1.119103	0.9;
	827	1	24.52719	7.9	0	0	1	1.046077	-21.123218	110	1	1.1	0.9;
	828	2	0.800888	18.6	0	0	1	1.079591	-34.192611	220	1	1.128115	0.9;
	829	2	0	0	0	0	1	1.014182	-14.237512	110	1	1.1	0.9;
	830	1	110.1	29.4	0	0	1	0.965654	-25.969307	110	1	1.1	0.9;
	831	1	0	0	0	0	1	1.042167	9.63592	380	1	1.1	0.9;
	832	1	34.43818	8.5	0	0	1	1.075441	-19.762939	220	1	1.139903	0.9;
	833	1	0	0	0.117	0	1	1.015807	-17.563473	220	1	1.1	0.9;
	834	1	15.21687	7.6	0	0	1	0.967804	-13.115979	110	1	1.1	0.9;
	835	1	62.06882	14.7	0	0	1	1.05995	-22.058247	220	1	1.137556	0.9;
	836	1	10.0111	5	0	0	1	1.019985	-46.415581	220	1	1.1	0.9;
	837	1	8.00888	4.1	0	0	1	0.968579	-11.939655	110	1	1.1	0.9;
	838	2	0	0	0	0	1	1.067318	-39.809461	220	1	1.123903	0.9;
	839	1	112.9	19.3	0	0	1	1.056505	0.38933	150	1	1.118224	0.9;
	840	1	43.7485	12.4	0	0	1	1.071998	-41.398698	220	1	1.121808	0.9;
	841	2	1.101221	61.2	0	0	1	1.099553	-33.813071	380	1	1.145799	0.9;
	842	1	0	0	0	0	1	1.009885	-26.520065	150	1	1.1	0.9;
	843	1	77.58602	12.3	0	0	1	1.055383	-48.841081	220	1	1.113026	0.9;
	844	1	0	0	0	0	1	1.033372	-20.215155	220	1	1.1	0.9;
	845	1	2.502775	1.3	0	0	1	0.972717	-25.723178	110	1	1.1	0.9;
	846	1	0	0	0	0	1	1.059586	-23.331244	220	1	1.133666	0.9;
	847	1	0	0	0	0	1	1.002489	-20.013347	150	1	1.1	0.9;
	848	1	-12.5139	4.1	0	0	1	1.006199	-24.97228	150	1	1.1	0.9;
	849	1	0	0	0	0	1	1.054407	37.877043	380	1	1.1	0.9;
	850	1	30.33363	10.5	0	0	1	1.077517	-15.867506	220	1	1.140576	0.9;
	851	1	50.0555	16.3	0	0	1	1.009917	-24.553819	110	1	1.1	0.9;
	852	1	26.02886	9.8	0	0	1	0.97531	-20.273197	110	1	1.1	0.9;
	853	1	0	0	0	0	1	1.102702	-10.257462	380	1	1.1329	0.9;
	854	1	0	0	0	0	1	1.051281	-20.579653	220	1	1.1	0.9;
	855	1	30.33363	11	0	0	1	1.046009	-15.84881	110	1	1.107169	0.9;
	856	1	0.200222	0.1	0	0	1	1.021532	-28.047066	110	1	1.1	0.9;
	857	1	189.1	-11.9	0	0	1	1.036496	-45.673731	380	1	1.1	0.9;
	858	1	90.70056	5.8	0	0	1	1.045509	-44.666819	220	1	1.1	0.9;
	859	1	15.81754	4	0	0	1	1.02798	-50.516313	150	1	1.1	0.9;
	860	1	4.805328	2.2	0	0	1	0.937072	-25.43818	110	1	1.1	0.9;
	861	1	69.07658	22.6	0	0	1	1.06758	0.040681	150	1	1.131188	0.9;
	862	1	0.5	-271.9	0	0	1	1.074461	-14.959924	380	1	1.125581	0.9;
	863	1	0	0	0	0	1	1.069559	-15.015629	150	1	1.127407	0.9;
	864	2	118.8	-8.2	0	0	1	1.066541	-12.341042	380	1	1.163051	0.9;
	865	1	66.27348	13.6	0	0	1	1.050622	-20.596583	220	1	1.120055	0.9;
	866	1	244.2	152.7	0	0	1	1.064388	-11.012839	150	1	1.136369	0.9;
	867	1	19.42153	11.6	0	0	1	1.04499	-9.958324	110	1	1.118557	0.9;
	868	1	65.77292	16.1	0	0	1	1.042261	-42.076334	220	1	1.1	0.9;
	869	1	0	0	0	0	1	1.091142	-27.579975	380	1	1.137219	0.9;
	870	1	166.6	8.1	0	0	1	1.052	-12.817203	380	1	1.120605	0.9;
	871	1	0	0	0	0	1	1.10009	-8.858707	380	1	1.167814	0.9;
	872	1	32.93652	20.4	0	0	1	0.950004	-31.621675	150	1	1.1	0.84698;
	873	1	42.54717	17.1	0	0	1	1.03618	-41.755384	220	1	1.1	0.9;
	874	1	0	0	0	0	1	1.037247	-27.442276	220	1	1.1	0.9;
	875	1	88.49812	16.5	0	0	1	1.073586	-11.340651	380	1	1.133253	0.9;
	876	1	45.04995	9.9	0	0	1	1.012119	-36.962919	110	1	1.1	0.9;
	877	2	215.2	64.2	0	0	1	0.992788	14.756517	220	1	1.1	0.9;
	878	1	1.301443	0.4	0	0	1	0.987596	-17.491918	110	1	1.1	0.9;
	879	1	0.400444	1.1	0	0	1	1.086594	32.711972	380	1	1.119237	0.9;
	880	1	62.26904	-2.1	0	0	1	1.057447	-38.32951	220	1	1.106044	0.9;
	881	1	0	0	0.090703	0	1	1.093542	-14.627919	380	1	1.155009	0.9;
	882	2	0	0	0	0	1	1.101711	4.874542	380	1	1.141443	0.9;
	883	1	0	0	0	0	1	1.030031	-22.502842	150	1	1.1	0.9;
	884	2	6.707437	2.2	0	0	1	0.989091	-29.316426	110	1	1.1	0.9;
	885	1	33.23685	11.1	0	0	1	0.916942	-31.270092	110	1	1.1	0.9;
	886	2	0	0	0	0	1	1.084026	-44.998356	380	1	1.12102	0.9;
	887	2	165.7	7.9	0	0	1	1.091167	-9.896786	220	1	1.140581	0.9;
	888	2	0	0	0	0	1	1.016636	-21.968124	110	1	1.1	0.9;
	889	1	21.12342	7.7	0	0	1	0.954907	-18.641198	110	1	1.1	0.9;
	890	1	14.71632	4.6	0	0	1	1.057724	-23.986777	220	1	1.132341	0.9;
	891	1	0	0	0	0	1	1.018022	-40.097196	220	1	1.1	0.9;
	892	2	0	0	0	0	1	1.101921	44.71579	380	1	1.111781	0.9;
	893	2	0	0	0	0	1	1.097158	-36.856675	380	1	1.162448	0.9;
	894	1	17.51942	4.6	0	0	1	1.073119	-36.277558	220	1	1.134816	0.9;
	895	2	36.84085	5	0	0	1	1.072235	-47.702846	220	1	1.124532	0.9;
	896	1	0	0	0	0	1	1.019784	-34.821772	150	1	1.1	0.9;
	897	1	0	0	0	0	1	1.053768	-47.923512	220	1	1.111297	0.9;
	898	1	32.33585	10.6	0	0	1	0.960005	-28.524817	110	1	1.1	0.9;
	899	1	0	0	0	0	1	1.027793	-27.457202	150	1	1.1	0.9;
	900	2	18.12009	5.6	0	0	1	0.980456	-19.001786	110	1	1.1	0.9;
	901	2	38.34251	51.8	0	0	1	1.062909	-23.309107	220	1	1.116497	0.9;
	902	2	171	69.2	0	0	1	0.996763	3.707005	380	1	1.1	0.9;
	903	1	0	0	0	0	1	1.083461	0.052671	150	1	1.152917	0.9;
	904	1	0	0	0	0	1	1.073835	-26.993777	220	1	1.126657	0.9;
	905	1	0	0	0	0	1	0.979078	-11.994215	110	1	1.1	0.9;
	906	1	0	0	0	0	1	0.989308	-23.760215	150	1	1.1	0.9;
	907	1	22.32475	13.8	0	0	1	1.018707	-22.368862	150	1	1.1	0.9;
	908	1	20.0222	6.8	0	0	1	1.046305	15.044132	110	1	1.121927	0.9;
	909	1	0	0	0	0	1	1.038378	-29.382605	220	1	1.1	0.9;
	910	2	0	0	0	0	1	1.074682	-32.936649	220	1	1.177142	0.9;
	911	2	0	0	0	0	1	1.080409	-19.127652	220	1	1.143814	0.9;
	912	1	24.02664	8	0	0	1	0.952218	-29.159422	110	1	1.1	0.9;
	913	2	0	0	0	0	1	0.994409	11.216246	220	1	1.1	0.9;
	914	1	0	0	0	0	1	1.073994	-12.03159	220	1	1.154018	0.9;
	915	1	0	0	0	0	1	1.08826	-14.675473	380	1	1.13075	0.9;
	916	1	0	0	0	0	1	1.012214	-19.903974	150	1	1.1	0.9;
	917	1	0	0	0	0	1	0.973526	-27.912982	110	1	1.1	0.9;
	918	1	0.200222	0.4	0	0	1	1.055835	-9.16684	220	1	1.12821	0.9;
	919	1	21.32364	7.3	0	0	1	1.048795	17.547105	110	1	1.121388	0.9;
	920	2	0	0	0	0	1	1.064428	-2.424373	150	1	1.142309	0.9;
	921	1	0	0	0	0	1	1.00251	-26.107918	150	1	1.1	0.9;
	922	2	137.1	41	0	0	1	1.014597	-42.442545	220	1	1.1	0.9;
	923	1	0	0	0	0	1	0.981285	-23.776845	150	1	1.1	0.9;
	924	1	44.54939	0	0	0	1	1.077545	-5.217637	220	1	1.132583	0.9;
	925	1	0	0	0	0	1	1.085382	-34.962146	380	1	1.132635	0.9;
	926	1	26.92986	9.5	0	0	1	1.077988	-21.627611	220	1	1.124002	0.9;
	927	1	21.02331	7	0	0	1	0.943413	-28.027334	110	1	1.1	0.9;
	928	1	0	0	0	0	1	0.967616	-32.907594	110	1	1.1	0.9;
	929	1	8.209102	3.3	0	0	1	1.086238	4.965539	150	1	1.152801	0.9;
	930	2	227.9	36.1	0	0	1	1.010682	-42.454261	220	1	1.1	0.9;
	931	1	0	0	0	0	1	0.963374	-13.287601	110	1	1.1	0.9;
	932	1	0	0	0	0	1	1.03406	-47.034641	220	1	1.1	0.9;
	933	1	241.4	-2.5	0.219001	0	1	1.06672	-47.431082	380	1	1.121639	0.9;
	934	1	201.4	53	0	0	1	1.020422	-30.519274	220	1	1.1	0.9;
	935	2	0	0	0	0	1	1.036036	-17.284298	380	1	1.115091	0.9;
	936	1	13.81532	5	0	0	1	1.011853	-22.278151	110	1	1.11057	0.9;
	937	2	0	0	0	0	1	1.047364	66.194779	220	1	1.1	0.9;
	938	2	-34.5383	2.7	0	0	1	1.028955	-32.167854	220	1	1.1	0.9;
	939	1	84.59379	-18.5	0	0	1	1.086728	-37.673057	380	1	1.159601	0.9;
	940	1	27.02997	13	0	0	1	1.060117	-27.359002	220	1	1.122095	0.9;
	941	1	26.32919	5.3	0	0	1	1.072859	-23.310894	220	1	1.138043	0.9;
	942	1	0	0	0	0	1	1.05581	-27.142614	220	1	1.117972	0.9;
	943	2	392.6	63.3	0	0	1	1.057995	-31.402002	220	1	1.184816	0.9;
	944	2	0	0	0	0	1	1.065182	-27.607913	220	1	1.114403	0.9;
	945	1	0	0	0	0	1	1.087637	0.76719	380	1	1.136944	0.9;
	946	1	3.704107	0	0	0	1	1.022036	-16.045017	110	1	1.1	0.9;
	947	1	34.6384	0	0	0	1	1.049305	-18.303684	220	1	1.1	0.9;
	948	1	31.93541	8.8	0	0	1	1.078787	-8.375009	220	1	1.145367	0.9;
	949	1	12.11343	4.4	0	0	1	0.998391	-39.64231	110	1	1.1	0.9;
	950	1	153.3	27.6	0	0	1	1.069509	2.38695	380	1	1.108976	0.9;
	951	2	0	0	0	0	1	1.079053	-3.539546	380	1	1.141662	0.9;
	952	1	8.909878	5.1	0	0	1	0.935334	-27.894813	110	1	1.1	0.9;
	953	1	47.05217	19.3	0	0	1	1.065795	-3.059679	150	1	1.135856	0.9;
	954	1	1.902109	0.8	0	0	1	0.990766	-23.758622	110	1	1.1	0.9;
	955	2	28.2313	-1.3	0	0	1	0.970967	-24.764244	110	1	1.1	0.9;
	956	1	0	0	0	0	1	1.085256	-15.871687	380	1	1.157081	0.9;
	957	2	146.9083	24.3	0	0	1	1.076682	8.708948	220	1	1.159152	0.9;
	958	1	0	0	0	0	1	1.038957	-12.244221	220	1	1.129051	0.9;
	959	1	67.47482	32.6	0	0	1	1.056958	17.042819	220	1	1.13738	0.9;
	960	1	39.1434	13.2	0	0	1	0.95516	-23.841365	110	1	1.1	0.9;
	961	1	0	0	0	23.18786	1	1.030034	-24.47861	150	1	1.1	0.9;
	962	1	0	0	0	0	1	1.023768	-23.061171	150	1	1.1	0.9;
	963	1	0	0	0	0	1	1.078993	20.514548	220	1	1.163224	0.9;
	964	1	0	0	0	0	1	1.068363	-3.734237	380	1	1.127137	0.9;
	965	1	0	0	0	-42.2993	1	1.067545	-17.586073	380	1	1.118656	0.9;
	966	1	0	0	0	0	1	1.092306	-37.413602	380	1	1.142913	0.9;
	967	1	0	0	0	0	1	1.026574	-23.57534	150	1	1.1	0.9;
	968	1	84.89413	32.4	0	0	1	1.0635	-19.707728	220	1	1.1	0.9;
	969	1	50.0555	17	0	0	1	0.986976	-21.96064	110	1	1.1	0.9;
	970	1	215.2	-7.6	0	0	1	1.07013	-42.614333	220	1	1.1166	0.9;
	971	1	89.59934	20	0	0	1	1.052458	-49.271226	220	1	1.109979	0.9;
	972	1	0	0	0	0	1	1.066949	24.173625	380	1	1.115401	0.9;
	973	1	279.4	22.9	0	0	1	1.067128	-8.293514	220	1	1.129282	0.9;
	974	1	35.23907	11.6	0	0	1	1.002652	-29.856388	110	1	1.1	0.9;
	975	1	0	0	0	0	1	1.048366	-17.730474	220	1	1.108227	0.9;
	976	1	1.601776	0	0	0	1	1.022442	-15.661336	110	1	1.1	0.9;
	977	1	83.3	47.9	0	0	1	1.040513	-29.176141	150	1	1.1	0.9;
	978	1	0	0	0	0	1	1.06815	-20.375662	380	1	1.141001	0.9;
	979	1	21.92431	7.2	0	0	1	1.047135	-22.169959	110	1	1.150049	0.9;
	980	1	403.9	-0.1	0	0	1	1.076496	-39.155008	380	1	1.129104	0.9;
	981	2	27.63063	5.6	0	0	1	1.097121	-9.870688	110	1	1.173144	0.9;
	982	1	0	0	0	0	1	0.964642	-30.559805	150	1	1.1	0.9;
	983	1	0	0	0	0	1	1.043784	-12.042992	220	1	1.116685	0.9;
	984	1	0	0	0	0	1	1.087927	-3.436245	220	1	1.141574	0.9;
	985	1	0.200222	1.3	0.286541	0	1	1.036659	4.307241	380	1	1.110932	0.9;
	986	2	284	60.9	0	0	1	1.090124	-10.25809	380	1	1.15198	0.9;
	987	2	0	0	0	0	1	1.074737	-33.364893	380	1	1.153459	0.9;
	988	2	0	0	0	0	1	1.059409	-8.192606	220	1	1.116768	0.9;
	989	1	42.14673	14.2	0	0	1	1.044611	-21.804078	110	1	1.144781	0.9;
	990	1	162.1	43.2	0	0	1	1.009901	-7.309111	220	1	1.1	0.9;
	991	1	12.01332	5.6	0	0	1	1.062541	-43.003321	380	1	1.123903	0.9;
	992	1	0	0	0	0	1	0.96078	-28.337438	110	1	1.1	0.9;
	993	1	-7.60844	2.5	0	0	1	1.054438	-15.986693	150	1	1.10989	0.9;
	994	1	56.06216	18.7	0	0	1	1.0669	-20.396646	220	1	1.149392	0.9;
	995	2	0	0	0	0	1	1.03649	-10.38925	110	1	1.1	0.9;
	996	1	30.9343	-46.6	0	0	1	1.064631	-35.273636	220	1	1.12498	0.9;
	997	2	0	0	0	0	1	1.088579	-8.530321	380	1	1.143134	0.9;
	998	1	28.53163	11.3	0	0	1	1.082452	4.482732	150	1	1.149696	0.9;
	999	1	0	0	0	0	1	1.074706	-11.557029	380	1	1.1	0.9;
	1000	1	0	0	0	0	1	1.087016	-1.017499	220	1	1.152062	0.9;
	1001	2	46.25128	21	0	0	1	1.073	-12.478025	110	1	1.16098	0.9;
	1002	1	0	0	0	11.97037	1	1.005346	-23.226226	150	1	1.1	0.9;
	1003	2	0	0	0	0	1	1.079182	-4.01661	220	1	1.144308	0.9;
	1004	1	0	0	0	0	1	1.034565	-13.521388	220	1	1.1	0.9;
	1005	2	180.3753	76.9	0	0	1	1.014591	-6.203112	220	1	1.115893	0.9;
	1006	2	13.31475	8.1	0	0	1	1.105132	29.309268	380	1	1.194059	0.9;
	1007	1	5.105661	3.8	0	0	1	0.945903	-24.62856	110	1	1.1	0.9;
	1008	1	29.83308	7.3	0	0	1	1.076765	-25.444108	220	1	1.166585	0.9;
	1009	1	192.4	63.9	0	0	1	1.080354	3.737959	150	1	1.148218	0.9;
	1010	1	1.801998	0.6	0	0	1	0.958198	-18.793219	110	1	1.1	0.9;
	1011	1	0	0	0	0	1	1.043534	-22.391761	150	1	1.1	0.9;
	1012	1	-54.0599	13.1	0	0	1	1.047591	-30.997674	220	1	1.106175	0.9;
	1013	2	302.6	263.5	0	0	1	1.008854	12.959511	380	1	1.105836	0.9;
	1014	1	0	0	0	0	1	1.0133	-8.638766	220	1	1.112846	0.9;
	1015	1	10.9121	3.6	0	0	1	0.994221	-28.903513	110	1	1.1	0.9;
	1016	1	59.26571	5.1	0	0	1	1.090969	-34.906109	380	1	1.140234	0.9;
	1017	1	-40.545	9.3	0	0	1	1.081648	-34.935619	220	1	1.127791	0.9;
	1018	1	70.27792	58.9	0	0	1	1.025259	15.603569	380	1	1.1	0.9;
	1019	1	254	74.3	0	0	1	1.082229	-16.465736	380	1	1.150919	0.9;
	1020	2	0	0	0	0	1	1.078545	-13.274296	220	1	1.133229	0.9;
	1021	1	244.8	119.3	0	0	1	1.025885	-12.865223	220	1	1.1	0.9;
	1022	2	11.41265	36.1	0	0	1	1.075	-34.988125	220	1	1.11618	0.9;
	1023	2	0	0	0	0	1	1.090727	-21.053155	220	1	1.127973	0.9;
	1024	2	0	0	0	0	1	1.054154	15.193815	150	1	1.1	0.9;
	1025	1	0	0	0	0	1	1.019624	11.436964	220	1	1.1	0.9;
	1026	2	0	0	0	0	1	1.045455	-26.165766	220	1	1.105459	0.9;
	1027	1	0	0	0.100424	0	1	1.093184	-8.302072	220	1	1.147336	0.9;
	1028	1	-878.1	-312.7	0	20	1	1.084035	-6.977282	380	1	1.147437	0.9;
	1029	2	228.3	71.3	0	0	1	1.06054	-21.149573	220	1	1.14353	0.9;
	1030	2	0	0	0	0	1	1.085658	-21.02362	380	1	1.17932	0.9;
	1031	1	7.8	50.9	0	0	1	1.077447	-13.719185	380	1	1.112251	0.9;
	1032	1	115.2	319.2	0	0	1	1.09077	-39.494213	380	1	1.159162	0.9;
	1033	1	1.401554	0.6	0	0	1	0.964219	-21.451959	110	1	1.1	0.9;
	1034	1	16.21798	2	0	0	1	1.081119	-17.594285	220	1	1.145911	0.9;
	1035	1	0	0	0	0	1	1.059155	-23.290164	220	1	1.133323	0.9;
	1036	2	0	0	0	0	1	1.081658	8.036763	380	1	1.10774	0.9;
	1037	2	18.92098	9	0	0	1	1.0165	-19.808915	220	1	1.1	0.9;
	1038	1	0	0	0	0	1	1.040935	-45.757775	220	1	1.1	0.9;
	1039	1	0	0	0	0	1	1.083097	-8.229026	380	1	1.138152	0.9;
	1040	1	45.9	36.8	0	0	1	1.068354	-17.308	380	1	1.118632	0.9;
	1041	1	0	0	0.198999	0	1	1.05066	-17.738341	220	1	1.108585	0.9;
	1042	1	39.04329	-5	0	0	1	0.992303	37.804141	220	1	1.1	0.9;
	1043	2	-61.7688	-36.7	0	0	1	1.082591	-19.227352	220	1	1.118281	0.9;
	1044	1	32.83641	0	0	0	1	1.068784	-5.655699	220	1	1.124127	0.9;
	1045	1	20.0222	15	0	0	1	0.995273	10.382623	220	1	1.1	0.9;
	1046	1	141.4	123	0	0	1	1.034886	3.934959	380	1	1.108999	0.9;
	1047	1	0	0	0	-35.2657	1	1.020017	-24.344159	150	1	1.1	0.9;
	1048	1	0	0	0	0	1	1.067164	27.419435	380	1	1.131389	0.9;
	1049	1	0	0	0	0	1	1.077968	-19.60349	220	1	1.142322	0.9;
	1050	1	12.41376	3.9	0	0	1	1.074009	-6.621724	220	1	1.140099	0.9;
	1051	1	0	0	0	0	1	1.07952	-47.003678	380	1	1.136391	0.9;
	1052	1	65.67281	24	0	0	1	1.064415	-21.505816	220	1	1.141716	0.9;
	1053	1	24.12675	12.7	0	0	1	0.987383	-26.675135	110	1	1.1	0.9;
	1054	1	112.4	13	0	0	1	1.046067	-35.675451	220	1	1.152604	0.9;
	1055	1	0.100111	0.5	0	0	1	1.031286	4.040219	220	1	1.1	0.9;
	1056	1	170.2	60.8	0	0	1	1.040711	8.914671	380	1	1.116249	0.9;
	1057	1	0	0	0	0	1	1.062886	-19.033089	220	1	1.1	0.9;
	1058	1	0	0	0	0	1	1.047306	-3.612777	380	1	1.1	0.9;
	1059	1	26.4293	16.4	0	0	1	0.998013	-23.701566	150	1	1.1	0.9;
	1060	1	0	0	0	0	1	1.068762	18.520171	380	1	1.135749	0.9;
	1061	1	-39.8081	-10.9	0	0	1	1.02288	-22.098043	150	1	1.1	0.9;
	1062	1	0	0	0	0	1	1.031959	-25.555762	150	1	1.1	0.9;
	1063	1	0	0	0	0	1	1.090377	-38.215081	380	1	1.131438	0.9;
	1064	2	0	-0.2	0	0	1	1.013588	-46.627483	220	1	1.1	0.9;
	1065	1	79.88858	29	0	0	1	1.04222	-20.780378	110	1	1.1	0.9;
	1066	1	0	0	0	0	1	1.076452	-24.678106	220	1	1.165582	0.9;
	1067	2	129.1	5.8	0	0	1	1.027727	-37.676069	220	1	1.1	0.9;
	1068	2	0	0	0	0	1	1.022467	-34.199348	150	1	1.1	0.9;
	1069	2	-128.743	38	0	0	1	1.078727	-34.967579	220	1	1.130268	0.9;
	1070	2	92.70278	-11.6	0	0	1	1.015	-18.846841	110	1	1.1	0.9;
	1071	2	0	0	0	0	1	1.000338	57.089623	220	1	1.1	0.9;
	1072	1	35.83974	11.8	0	0	1	0.945259	-30.065159	110	1	1.1	0.9;
	1073	1	86.59601	23.4	0	0	1	1.080069	-3.779352	220	1	1.133597	0.9;
	1074	1	43.7485	16.8	0	0	1	1.038148	-29.874337	110	1	1.1	0.9;
	1075	1	0.100111	-19.3	0	0	1	1.067295	-18.37886	220	1	1.1	0.9;
	1076	1	10.81199	3.9	0	0	1	0.998031	-19.530608	110	1	1.1	0.9;
	1077	1	199.4	-19.9	0	0	1	1.046395	-32.617282	220	1	1.153015	0.9;
	1078	2	0	0	0	0	1	1.068421	-7.600622	380	1	1.130562	0.9;
	1079	1	150.4	15.3	0	0	1	1.056397	-39.907951	220	1	1.1059	0.9;
	1080	1	67.57492	21.2	0	0	1	1.074583	-5.000372	220	1	1.129379	0.9;
	1081	1	30.9343	6.6	0	0	1	1.079051	-13.763662	220	1	1.157541	0.9;
	1082	1	96.40689	18.1	0	0	1	1.042574	39.422858	220	1	1.1	0.9;
	1083	1	14.61621	5.6	0	0	1	1.006243	-16.153225	110	1	1.1	0.9;
	1084	1	13.21465	9.2	0	0	1	0.946667	-20.739671	110	1	1.1	0.9;
	1085	1	181.5	21.2	0	0	1	1.099864	-17.488937	380	1	1.165931	0.9;
	1086	1	129.6	13.3	0	0	1	1.07462	-14.521122	380	1	1.150458	0.9;
	1087	1	3.804218	2.5	0	0	1	1.002436	-15.395547	110	1	1.1	0.9;
	1088	1	0	0	0.191685	0	1	1.031505	-23.76333	220	1	1.1	0.9;
	1089	1	25.22797	10.6	0	0	1	0.975627	-24.912686	110	1	1.1	0.9;
	1090	1	0	0	0	0	1	1.063887	-13.368797	380	1	1.119303	0.9;
	1091	1	0	0	0.053002	0	1	1.065146	-15.381302	380	1	1.111359	0.9;
	1092	1	124.8	7.9	0	0	1	1.06889	-37.979089	380	1	1.134933	0.9;
	1093	1	114.1	23.5	0	0	1	1.025006	-19.164407	220	1	1.135097	0.9;
	1094	1	0	0	0	-25.7113	1	1.103384	-16.144236	380	1	1.163703	0.9;
	1095	1	0	0	0	0	1	1.020377	-21.036257	150	1	1.1	0.9;
	1096	2	0	0	0	0	1	1.063104	1.011281	150	1	1.137018	0.9;
	1097	1	27.13008	16.8	0	0	1	1.001476	-22.902653	150	1	1.1	0.9;
	1098	1	70.1778	28.7	0	0	1	1.079101	0.353621	150	1	1.148499	0.9;
	1099	1	183.0286	15.7	0	0	1	1.039895	-45.803964	220	1	1.1	0.9;
	1100	1	-15.8175	5.2	0	0	1	1.050466	-19.808717	150	1	1.1	0.9;
	1101	2	0	0	0.045352	-88.2763	1	1.09671	-8.146926	380	1	1.152267	0.9;
	1102	1	0	0	0	0	1	1.100879	-16.831907	380	1	1.18261	0.9;
	1103	1	105	412	0	20	1	1.038823	19.435403	380	1	1.1	0.9;
	1104	1	0	0	0	0	1	1.050233	-15.796787	150	1	1.105175	0.9;
	1105	1	0	0	0	0	1	1.082554	-10.992986	220	1	1.154172	0.9;
	1106	1	0	0	0	0	1	1.026278	-17.977972	150	1	1.1	0.9;
	1107	1	30.83419	13.2	0	0	1	0.979093	-36.422876	110	1	1.1	0.9;
	1108	1	1.401554	0.5	0	0	1	1.00416	-33.044166	110	1	1.1	0.9;
	1109	2	0	0	0	0	1	1.023727	-41.519434	220	1	1.1	0.9;
	1110	1	230.5	91	0	0	1	1.066853	-2.470678	150	1	1.144215	0.9;
	1111	1	23.02553	14.3	0	0	1	1.041394	-25.089083	150	1	1.1	0.9;
	1112	1	168.3	-57.1	0	0	1	1.07399	-35.19395	380	1	1.151933	0.9;
	1113	2	0	0	0	0	1	1.091909	-15.904391	220	1	1.12871	0.9;
	1114	1	0	0	0	0	1	1.037156	-21.490804	150	1	1.113958	0.9;
	1115	2	0	0	0	0	1	1.064464	9.242222	150	1	1.1	0.9;
	1116	1	53.15894	17.7	0	0	1	0.94541	-27.423788	110	1	1.1	0.9;
	1117	1	5.506105	2	0	0	1	1.03726	-21.668884	110	1	1.133639	0.9;
	1118	1	0	0	0	0	1	1.038399	-22.439623	150	1	1.1	0.9;
	1119	1	0	0	0	0	1	1.087869	-45.322628	380	1	1.149491	0.9;
	1120	1	0	0	0	0	1	1.064317	-21.529421	220	1	1.141363	0.9;
	1121	1	71.77958	17.9	0	0	1	1.079936	-15.533349	220	1	1.142399	0.9;
	1122	1	105.5	24.7	0	0	1	1.045474	-34.286609	220	1	1.108116	0.9;
	1123	2	0	0	0	0	1	1.100237	-5.518412	380	1	1.186987	0.9;
	1124	1	60.36693	35.4	0	0	1	0.963367	58.645088	220	1	1.1	0.9;
	1125	1	0	0	0	0	1	1.041038	-50.369746	220	1	1.1	0.9;
	1126	1	23.52608	4	0	0	1	1.065603	-37.542169	220	1	1.115206	0.9;
	1127	2	-30.1334	-22.3	0	0	1	1.079727	-34.55186	220	1	1.132832	0.9;
	1128	1	226.3	51.7	0	0	1	1.029132	-2.43524	220	1	1.107003	0.9;
	1129	2	0	0	0	0	1	1.086316	57.474165	380	1	1.1	0.9;
	1130	2	5.205771	2.2	0	0	1	1.014258	-22.123443	110	1	1.1	0.9;
	1131	1	2.00222	1	0	0	1	0.976978	-24.72408	110	1	1.1	0.9;
	1132	1	23.02553	8.3	0	0	1	1.032832	-20.701302	110	1	1.1	0.9;
	1133	1	0	0	0	0	1	1.074157	-10.279961	220	1	1.13572	0.9;
	1134	1	141.2566	30.6	0	0	1	1.07597	-10.43724	380	1	1.133384	0.9;
	1135	1	39.44373	24.9	0	0	1	1.015905	-31.528241	150	1	1.1	0.9;
	1136	2	0	0	0	0	1	1.088632	-8.539114	380	1	1.143457	0.9;
	1137	1	0	0	0	0	1	1.004772	-29.436706	110	1	1.1	0.9;
	1138	1	186.6	155.8	0	0	1	1.056233	23.158546	380	1	1.107147	0.9;
	1139	1	0	0	0	0	1	0.988497	-24.860737	110	1	1.1	0.9;
	1140	1	20.62287	8.2	0	0	1	1.067289	3.229482	150	1	1.13802	0.9;
	1141	1	1.902109	1.3	0	0	1	1.025806	-15.210579	110	1	1.1	0.9;
	1142	1	0	0	0	0	1	1.063121	-35.439749	220	1	1.123818	0.9;
	1143	2	0	0	0	0	1	1.014591	40.034083	220	1	1.1	0.9;
	1144	1	0	0	0	0	1	1.106746	-16.605214	380	1	1.162681	0.9;
	1145	1	50.85638	36.6	0	0	1	1.03711	14.041362	150	1	1.1	0.9;
	1146	1	279.5	25.8	0	0	1	1.057394	-40.2934	220	1	1.126613	0.9;
	1147	1	0	0	0	0	1	1.061892	-21.049539	220	1	1.138884	0.9;
	1148	1	0	0	0.219	0	1	1.015064	-6.393574	220	1	1.1	0.9;
	1149	1	224.9	-32.5	0	0	1	1.079701	-31.966953	380	1	1.154646	0.9;
	1150	1	173	-21.3	0	0	1	1.037443	-49.683096	220	1	1.162872	0.9;
	1151	1	15.41709	2.2	0	0	1	0.999946	-37.116054	110	1	1.1	0.9;
	1152	2	3.103441	1.5	0	0	1	1.008409	21.543852	220	1	1.1	0.9;
	1153	1	151.2	96.4	0.233362	292.6621	1	1.072422	-2.452171	380	1	1.1	0.9;
	1154	1	112.8	27.6	0	0	1	1.044499	-34.376189	220	1	1.107242	0.9;
	1155	1	1.801998	0.7	0	0	1	1.038896	-20.019694	110	1	1.1	0.9;
	1156	2	119.6	26.9	0	0	1	1.003864	-42.392439	220	1	1.1	0.9;
	1157	1	0	0	0	0	1	0.979121	-23.955722	150	1	1.1	0.9;
	1158	1	7.00777	2.3	0	0	1	0.94596	-31.279299	110	1	1.1	0.9;
	1159	1	149.9	53.2	0	0	1	0.959356	-18.460955	110	1	1.1	0.9;
	1160	1	52.45816	2.9	0	0	1	1.07171	-17.916683	220	1	1.135185	0.9;
	1161	1	0	0	0	0	1	1.030245	-21.622707	150	1	1.1	0.9;
	1162	2	0	0	0	0	1	1.064562	-2.422936	150	1	1.142434	0.9;
	1163	1	0	0	0	0	1	1.068476	-15.067859	380	1	1.119674	0.9;
	1164	1	0	0	0	0	1	1.020095	-46.407006	220	1	1.1	0.9;
	1165	1	0	0	0	-262.422	1	1.086307	-32.848111	380	1	1.165752	0.9;
	1166	1	0	0	0	0	1	0.986843	-17.234553	150	1	1.1	0.9;
	1167	1	202	34.1	0	0	1	1.071149	-42.592452	220	1	1.11973	0.9;
	1168	1	42.74739	15.2	0	0	1	0.997766	-20.153583	110	1	1.1	0.9;
	1169	1	19.02109	9.1	0	0	1	0.990866	-22.855841	110	1	1.1	0.9;
	1170	1	0	-0.2	0	0	1	1.063504	-11.344968	110	1	1.125449	0.9;
	1171	1	7.608436	2.5	0	0	1	0.944756	-24.708919	110	1	1.1	0.9;
	1172	2	0	0	0	0	1	1.080647	15.971884	380	1	1.135238	0.9;
	1173	1	9.610656	5.9	0	0	1	0.968237	-26.066436	150	1	1.1	0.9;
	1174	1	18.82087	6.2	0	0	1	0.926972	-31.411792	110	1	1.1	0.9;
	1175	1	22.32475	7.3	0	0	1	1.018693	-31.954636	110	1	1.1	0.9;
	1176	1	50.25572	19.3	0	0	1	1.01697	9.190559	220	1	1.1	0.9;
	1177	1	6.106771	0	0	0	1	1.038977	-44.338563	220	1	1.107346	0.9;
	1178	2	181.1	78.7	0	0	1	1.055364	-8.090508	220	1	1.116604	0.9;
	1179	1	0	0	0	0	1	1.006175	-23.119258	150	1	1.1	0.9;
	1180	1	19.02109	8	0	0	1	0.979708	-22.096714	110	1	1.1	0.9;
	1181	1	0	0	0	19.31922	1	1.017564	-29.802073	150	1	1.1	0.9;
	1182	1	30.53385	10	0	0	1	0.950476	-29.35272	110	1	1.1	0.9;
	1183	1	42.94762	26.6	0	0	1	0.9393	-29.246523	150	1	1.1	0.890626;
	1184	1	479	99.2	0	0	1	1.033456	-28.141178	220	1	1.138273	0.9;
	1185	2	0	0	0	0	1	0.968529	-19.494757	110	1	1.1	0.9;
	1186	1	0	0	0	0	1	1.082902	-16.316951	220	1	1.16641	0.9;
	1187	1	5.405994	1.8	0	0	1	1.046483	-30.78661	110	1	1.112207	0.9;
	1188	1	0	0	0	0	1	1.048099	-21.810708	220	1	1.136452	0.9;
	1189	1	8.509435	6.4	0	0	1	0.939249	-25.634299	110	1	1.1	0.9;
	1190	2	479.3	127.6	0	0	1	1.080632	-9.108559	380	1	1.122082	0.9;
	1191	1	71.37914	10.3	0	0	1	1.036179	-31.643376	220	1	1.1	0.9;
	1192	1	17.71965	3.8	0	0	1	0.944419	-23.372227	110	1	1.1	0.9;
	1193	1	0	0	0	0	1	0.958282	-28.037861	150	1	1.1	0.9;
	1194	1	37.94207	15.2	0	0	1	1.039341	-20.861329	380	1	1.142182	0.9;
	1195	2	18.72076	7.7	0	0	1	1.022455	-34.384361	110	1	1.1	0.9;
	1196	1	0	0	0	0	1	1.04056	-17.08067	220	1	1.113472	0.9;
	1197	2	278.1	95.2	0	0	1	1.085868	12.653441	380	1	1.161051	0.9;
	1198	1	13.81532	3.6	0	0	1	0.945112	-25.465665	110	1	1.1	0.9;
	1199	1	149.3	-2	0	0	1	1.074861	-0.506567	380	1	1.128106	0.9;
	1200	1	0.200222	0.8	0	0	1	1.024497	15.617788	220	1	1.1	0.9;
	1201	1	0	0	0	0	1	1.022692	-23.153062	150	1	1.1	0.9;
	1202	2	0	0	0	0	1	1.023136	-45.934477	220	1	1.1	0.9;
	1203	1	11.41265	5.1	0	0	1	1.024086	-16.266549	110	1	1.1	0.9;
	1204	1	0	0	0	15.2568	1	1.024147	-27.828053	150	1	1.1	0.9;
	1205	1	118.9	129.2	0	0	1	1.026565	3.62329	380	1	1.1	0.9;
	1206	1	37.14118	11.6	0	0	1	1.047224	-20.501954	110	1	1.107931	0.9;
	1207	2	0	0	0	0	1	1.049571	-21.485862	110	1	1.110305	0.9;
	1208	1	0	0	0	0	1	1.068088	-34.304332	220	1	1.172945	0.9;
	1209	1	135.7	31.6	0	0	1	1.046484	-34.268858	220	1	1.109054	0.9;
	1210	1	0	0	0	0	1	1.095567	-37.221237	380	1	1.161307	0.9;
	1211	1	100.6	15.3	0	0	1	1.04421	-1.72669	150	1	1.107875	0.9;
	1212	2	0	0	0	0	1	1.082965	16.353739	380	1	1.127866	0.9;
	1213	1	0	0	0	0	1	1.077041	-17.727646	150	1	1.1	0.9;
	1214	1	-9.61066	3.2	0	0	1	1.033846	-25.813239	150	1	1.1	0.9;
	1215	1	0	0	0	0	1	1.081787	-15.342426	220	1	1.144607	0.9;
	1216	1	0	0	0	0	1	1.101181	-19.095646	150	1	1.144659	0.9;
	1217	2	0	0	0	0	1	1.100789	-15.497949	380	1	1.145313	0.9;
	1218	1	29.03219	9.4	0	0	1	1.067513	8.315346	110	1	1.137273	0.9;
	1219	1	108.4	-2.4	0	0	1	1.085886	-42.051893	380	1	1.14472	0.9;
	1220	1	13.21465	6.6	0	0	1	0.952398	-37.898379	110	1	1.1	0.9;
	1221	2	0	0	0	0	1	1.033955	-13.517214	220	1	1.1	0.9;
	1222	2	31.13452	8.5	0	0	1	1.06053	0.390122	220	1	1.132137	0.9;
	1223	1	0	0	0	0	1	1.069655	-13.034805	380	1	1.125928	0.9;
	1224	1	0	0	0	0	1	1.07294	-41.879555	220	1	1.122143	0.9;
	1225	2	2.00222	6	0	0	1	1.013458	61.700688	150	1	1.1	0.9;
	1226	1	0	0	0	0	1	1.066738	-18.371595	220	1	1.1	0.9;
	1227	1	0	0	0	0	1	1.055801	-48.626867	220	1	1.110298	0.9;
	1228	1	30.53385	13.8	0	0	1	1.06725	-1.808695	150	1	1.143589	0.9;
	1229	1	0	0	0	0	1	0.952892	-12.866249	110	1	1.1	0.9;
	1230	1	28.93208	11.9	0	0	1	1.081184	0.090725	150	1	1.147492	0.9;
	1231	1	6.407104	2.3	0	0	1	1.015607	-17.748981	110	1	1.1	0.9;
	1232	2	291.9	91.9	0	0	1	1.035273	-6.948872	220	1	1.124913	0.9;
	1233	2	0	0	0	0	1	1.085148	-8.348088	380	1	1.14685	0.9;
	1234	1	538.3	90.8	0	0	1	1.082004	-40.333516	380	1	1.150938	0.9;
	1235	1	100.3	4.8	0	0	1	0.933329	-30.696806	110	1	1.1	0.9;
	1236	1	60.16671	30.6	0	0	1	1.027929	-4.445692	220	1	1.1	0.9;
	1237	1	0	0	0.296958	0	1	1.089319	-19.619044	380	1	1.15387	0.9;
	1238	2	211.8	87.9	0	0	1	0.951545	-20.521921	110	1	1.1	0.9;
	1239	1	0	0	0	0	1	1.078709	20.441211	220	1	1.161138	0.9;
	1240	1	0	0	0.260248	0	1	1.068575	-19.462303	380	1	1.115825	0.9;
	1241	1	40.84529	14.5	0	0	1	1.070711	-11.617571	220	1	1.132727	0.9;
	1242	2	0	0	0	0	1	1.068712	-37.038293	380	1	1.132032	0.9;
	1243	1	22.02442	7.7	0	0	1	1.030688	-36.27661	110	1	1.1	0.9;
	1244	1	0	0	0.154869	0	1	1.061118	-2.51717	380	1	1.118747	0.9;
	1245	1	0	0	0	0	1	0.99822	-23.421572	150	1	1.1	0.9;
	1246	1	0	0	0.264593	0	1	1.058004	-24.678366	380	1	1.129023	0.9;
	1247	1	46.55161	28.8	0	0	1	0.985991	-23.580828	150	1	1.1	0.9;
	1248	1	50.9565	16.8	0	0	1	0.945825	-30.274629	110	1	1.1	0.9;
	1249	1	0	0	0	0	1	0.954263	-25.880669	150	1	1.1	0.9;
	1250	1	0	0	0	0	1	1.038391	-22.424596	150	1	1.1	0.9;
	1251	1	10.11121	8.9	0	0	1	0.941456	-27.822407	110	1	1.1	0.9;
	1252	1	0	0	0	22.60474	1	1.030497	-26.504927	150	1	1.1	0.9;
	1253	1	18.01998	6	0	0	1	0.950026	-27.230988	110	1	1.1	0.9;
	1254	2	3.603996	2.2	0	0	1	0.987636	-23.245004	150	1	1.1	0.9;
	1255	1	145.2	21.4	0	0	1	1.009375	-22.219997	220	1	1.120971	0.9;
	1256	1	10.81199	3.5	0	0	1	1.037952	-31.307444	110	1	1.1	0.9;
	1257	1	0	0	0	0	1	1.04268	-18.757465	220	1	1.1	0.9;
	1258	1	0	0	0	0	1	1.088169	-16.449167	380	1	1.149947	0.9;
	1259	2	19.82198	3.1	0	0	1	1.014707	61.833986	150	1	1.1	0.9;
	1260	1	0	0	0	0	1	1.01978	-23.607485	150	1	1.1	0.9;
	1261	2	562.3	159.1	0	0	1	0.980591	-0.010459	220	1	1.1	0.9;
	1262	1	20.32253	6.7	0	0	1	1.000059	-29.871323	110	1	1.1	0.9;
	1263	1	0	0	0	16.62709	1	1.040555	-29.172708	150	1	1.1	0.9;
	1264	2	0	0	0	0	1	1.054273	-14.53172	220	1	1.122976	0.9;
	1265	2	0	0	0	0	1	1.076227	-33.847335	220	1	1.129157	0.9;
	1266	1	0	0	0	0	1	1.028774	13.664863	220	1	1.106242	0.9;
	1267	1	2.402664	0.7	0	0	1	1.046415	-41.551082	220	1	1.10623	0.9;
	1268	1	30.9343	7.6	0	0	1	1.076664	-25.432458	220	1	1.166639	0.9;
	1269	1	52.95872	15.2	0	0	1	1.02138	-31.456754	110	1	1.1	0.9;
	1270	1	0	-6	0	0	1	0.988479	-23.641661	150	1	1.1	0.9;
	1271	1	0	0	0	0	1	1.054816	-12.454437	220	1	1.138483	0.9;
	1272	1	196.7	44.2	0	0	1	1.052397	-17.819329	220	1	1.117299	0.9;
	1273	2	185.5	62.3	0	0	1	1.080763	-44.499809	380	1	1.149158	0.9;
	1274	1	39.24351	71.7	0	0	1	1.029056	3.094471	380	1	1.1	0.9;
	1275	1	11.91321	2.9	0	0	1	1.078078	-14.02878	220	1	1.138791	0.9;
	1276	2	0	0	0	0	1	1.037682	-17.619303	220	1	1.1	0.9;
	1277	2	0	0	0	0	1	1.037614	57.89664	220	1	1.1	0.9;
	1278	1	0	0	0	0	1	0.969449	-20.539803	110	1	1.1	0.9;
	1279	1	157.9	26	0	0	1	1.039598	-22.355349	220	1	1.1	0.9;
	1280	1	0	0	0	0	1	1.028537	-45.064836	150	1	1.1	0.9;
	1281	1	15.5172	5.2	0	0	1	0.936472	-28.090673	110	1	1.1	0.9;
	1282	2	138.2	153.5	0	0	1	1.065091	-43.903137	220	1	1.11356	0.9;
	1283	1	0	0	0	0	1	0.928893	-28.004198	110	1	1.1	0.9;
	1284	1	0	0	0	0	1	1.031297	-43.531848	220	1	1.1	0.9;
	1285	1	5.105661	0.9	0	0	1	0.980998	-21.937132	110	1	1.1	0.9;
	1286	1	0	0	0	0	1	0.995695	-20.48945	150	1	1.1	0.9;
	1287	1	144.7	20.3	0	0	1	1.052979	-26.449926	220	1	1.141184	0.9;
	1288	1	0	0	0	0	1	1.096041	-11.176882	220	1	1.152023	0.9;
	1289	1	7.808658	2.6	0	0	1	1.010679	-18.994193	110	1	1.1	0.9;
	1290	2	4.805328	1.5	0	0	1	0.983273	-8.757619	110	1	1.1	0.9;
	1291	1	40.54495	3.6	0	0	1	1.057563	-49.653534	220	1	1.1	0.9;
	1292	1	16.51831	3.3	0	0	1	1.008232	-39.56415	110	1	1.1	0.9;
	1293	1	42.94762	19.8	0	0	1	1.054231	-47.45836	220	1	1.1	0.9;
	1294	2	132.2	32.8	0	0	1	1.036182	-20.539581	220	1	1.12416	0.9;
	1295	1	0	0	0	0	1	1.064999	-35.026988	220	1	1.146223	0.9;
	1296	2	0	0	0	0	1	0.994613	-18.581575	110	1	1.1	0.9;
	1297	1	0.300333	0.2	0	0	1	1.074108	-7.025917	380	1	1.142049	0.9;
	1298	2	0	0	0	0	1	1.073684	27.284838	380	1	1.1	0.9;
	1299	1	0	0	0	0	1	0.981616	-27.983677	110	1	1.1	0.9;
	1300	1	22.82531	7.5	0	0	1	0.949529	-32.34433	110	1	1.1	0.9;
	1301	1	0	0	0	0	1	1.084368	-31.60295	380	1	1.136925	0.9;
	1302	1	20.82309	7.9	0	0	1	1.050447	-46.281085	220	1	1.1	0.9;
	1303	1	-8.30921	2.7	0	0	1	0.992609	-25.259383	150	1	1.1	0.9;
	1304	1	193.2	121.7	0	0	1	1.028232	-7.911823	220	1	1.1	0.9;
	1305	1	0	0	0	0	1	1.054853	-17.286978	220	1	1.116933	0.9;
	1306	1	0	0	0	0	1	1.064996	-23.362896	220	1	1.164848	0.9;
	1307	2	0	0	0	0	1	1.060455	-12.507277	220	1	1.144229	0.9;
	1308	1	204.6	53.2	0	0	1	1.008487	-34.190115	220	1	1.1	0.9;
	1309	1	0	0	0.310992	0	1	1.05459	-17.28688	220	1	1.112327	0.9;
	1310	1	110.1	2	0	0	1	1.065115	-43.884482	220	1	1.11399	0.9;
	1311	1	60.0666	20	0	0	1	0.938841	-27.705658	110	1	1.1	0.9;
	1312	2	298.7	85.3	0	0	1	1.07671	-6.193256	380	1	1.135591	0.9;
	1313	1	238.4	210.9	0	0	1	1.06943	-4.141656	220	1	1.128334	0.9;
	1314	1	10.9121	3	0	0	1	0.917123	-31.142264	110	1	1.1	0.9;
	1315	1	4.605106	1.7	0	0	1	1.046725	-15.766244	110	1	1.107664	0.9;
	1316	1	0	0	0	0	1	1.09917	-14.776255	220	1	1.194555	0.9;
	1317	1	0	0	0	0	1	1.041509	-25.081813	150	1	1.1	0.9;
	1318	1	39.54384	14.3	0	0	1	1.042688	-18.795553	110	1	1.128269	0.9;
	1319	1	20.52275	6.7	0	0	1	0.95807	-29.405476	110	1	1.1	0.9;
	1320	1	0	0	0	0	1	1.06901	4.686012	380	1	1.126556	0.9;
	1321	1	80.0888	26.3	0	0	1	0.958955	-31.63482	110	1	1.1	0.9;
	1322	1	20.62287	10.7	0	0	1	1.07742	-17.685245	150	1	1.1	0.9;
	1323	1	18.01998	4.8	0	0	1	0.963329	-27.248259	110	1	1.1	0.9;
	1324	1	0	0	0	0	1	0.931793	-33.314429	110	1	1.1	0.9;
	1325	1	20.82309	36	0	0	1	1.088018	-36.017311	220	1	1.140789	0.9;
	1326	2	4	172.7	0	0	1	1.098447	-34.636571	380	1	1.154471	0.9;
	1327	1	0	0	0	0	1	1.049262	-20.885126	220	1	1.118961	0.9;
	1328	1	0	0	0	0	1	1.039584	-31.524685	220	1	1.120365	0.9;
	1329	1	0	0	0	0	1	1.036204	-20.54008	220	1	1.124183	0.9;
	1330	1	109.6	36.4	0	0	1	1.072122	4.435848	150	1	1.142834	0.9;
	1331	1	22.7252	6.4	0	0	1	1.05778	6.387704	150	1	1.110037	0.9;
	1332	2	0	0	0	0	1	1.044907	-14.6501	380	1	1.115927	0.9;
	1333	1	42.94762	26.6	0	0	1	1.031512	-21.814447	150	1	1.108511	0.9;
	1334	1	0	0	0	0	1	1.032594	-25.572664	150	1	1.1	0.9;
	1335	1	0	0	0	0	1	1.017001	-19.73758	220	1	1.1	0.9;
	1336	1	0	0	0	0	1	1.091232	-30.618589	220	1	1.143336	0.9;
	1337	1	18.12009	-11.6	0	0	1	1.023141	-46.244898	220	1	1.1	0.9;
	1338	1	95.606	8.6	0	0	1	1.028966	-44.511405	220	1	1.1	0.9;
	1339	1	0	0	0	0	1	0.952937	-20.543815	110	1	1.1	0.9;
	1340	1	0	0	0.143001	0	1	0.97039	-25.91911	220	1	1.1	0.9;
	1341	2	1.501665	87.9	0	0	1	1.073316	-39.449718	380	1	1.130904	0.9;
	1342	1	30.63396	19	0	0	1	0.944372	-32.81358	150	1	1.1	0.831014;
	1343	1	0	0	0	0	1	1.08481	7.908254	150	1	1.139194	0.9;
	1344	2	0	0	0	0	1	1.060909	-14.76996	220	1	1.11349	0.9;
	1345	1	82.19112	23.4	0	0	1	1.082097	-39.366017	380	1	1.140718	0.9;
	1346	1	100.1	45.6	0	0	1	1.027925	-43.965179	220	1	1.1	0.9;
	1347	2	627.8	293.4	0	0	1	1.041864	3.033043	220	1	1.121341	0.9;
	1348	1	0	0	0	0	1	1.01711	-19.65202	150	1	1.1	0.9;
	1349	1	0	0	0	0	1	1.047912	-27.225683	220	1	1.172696	0.9;
	1350	2	0	0	0	0	1	1.055227	61.640938	220	1	1.108819	0.9;
	1351	1	144.9	87.7	0.210144	0	1	1.061571	-13.819717	380	1	1.128309	0.9;
	1352	1	79.58824	18.8	0	0	1	1.060521	-22.305182	220	1	1.138294	0.9;
	1353	2	-29.0322	-4	0	0	1	1.094409	-11.055518	220	1	1.14413	0.9;
	1354	1	0	0	0	0	1	1.094704	-38.343457	380	1	1.160292	0.9;
	1355	1	0	0	0	0	1	1.046886	-32.272663	220	1	1.105358	0.9;
	1356	1	23.02553	14.2	0	0	1	1.057574	-15.320487	150	1	1.1	0.9;
	1357	1	117.5	-3	0	0	1	1.059675	-43.033765	220	1	1.111341	0.9;
	1358	2	193.8	27.3	0	0	1	1.072636	-4.659792	220	1	1.128924	0.9;
	1359	1	1.101221	2.5	0.207794	20	1	1.067287	-4.112417	380	1	1.126442	0.9;
	1360	2	0	0	0	0	1	1.061162	-10.68696	380	1	1.113008	0.9;
	1361	1	18.01998	6	0	0	1	0.986505	-23.401632	110	1	1.1	0.9;
	1362	1	0.700777	0.2	0	0	1	0.990135	-29.552812	110	1	1.1	0.9;
	1363	1	0	0	0	0	1	1.040311	32.042834	220	1	1.1	0.9;
	1364	2	0	0	0	0	1	1.097789	-35.914605	380	1	1.169326	0.9;
	1365	1	6.407104	2.3	0	0	1	1.007552	-18.137118	110	1	1.1	0.9;
	1366	1	13.81532	5	0	0	1	1.018829	-16.415551	110	1	1.111702	0.9;
	1367	1	0	0	0	0	1	1.056878	-49.178223	220	1	1.115213	0.9;
	1368	1	0	0	0	0	1	1.035352	-25.853636	380	1	1.138509	0.9;
	1369	1	142.8	-6.4	0	0	1	1.029562	-40.04276	220	1	1.1	0.9;
	1370	1	0	0	0	0	1	1.041434	-14.319125	220	1	1.113544	0.9;
	1371	1	6.707437	3.3	0	0	1	1.002694	-23.611137	110	1	1.1	0.9;
	1372	1	0	0	0	0	1	1.082919	-14.840709	380	1	1.130724	0.9;
	1373	2	125.4	7.6	0	0	1	1.064546	-0.372288	220	1	1.16244	0.9;
	1374	2	0	0	0	0	1	1.065364	-16.411253	220	1	1.1	0.9;
	1375	1	16.4182	5.9	0	0	1	1.047214	-21.192944	110	1	1.108013	0.9;
	1376	1	0	0	0	0	1	1.01897	-19.024602	150	1	1.1	0.9;
	1377	1	0	0	0	0	1	1.002266	-22.324974	110	1	1.1	0.9;
	1378	1	339.1	89	0	0	1	1.060931	-42.00302	380	1	1.124784	0.9;
	1379	1	-35.7396	38.6	0	0	1	1.055232	-35.94238	220	1	1.107883	0.9;
	1380	1	0	0	0	0	1	1.027116	-16.57007	150	1	1.1	0.9;
	1381	1	22.02442	9	0	0	1	1.086457	0.692814	150	1	1.15334	0.9;
	1382	1	7.408214	2.7	0	0	1	1.024999	-21.09291	110	1	1.1	0.9;
	1383	1	0	0	0	0	1	0.958436	-25.540311	150	1	1.1	0.9;
	1384	1	-10.5117	3.5	0	0	1	1.048672	-20.010759	150	1	1.1	0.9;
	1385	1	173.6	72.9	0	0	1	1.016971	-42.41885	220	1	1.1	0.9;
	1386	1	0	0	0	0	1	1.02645	-23.775588	150	1	1.1	0.9;
	1387	1	72.1	71.8	0	0	1	1.064435	22.880718	380	1	1.117566	0.9;
	1388	1	57.86415	23.2	0	0	1	0.990484	-50.248133	220	1	1.1	0.9;
	1389	1	0	0	0	0	1	1.039454	15.72315	380	1	1.10519	0.9;
	1390	1	0	0	0	0	1	1.07981	18.773105	380	1	1.138581	0.9;
	1391	2	0	0	0	0	1	1.052282	-11.911518	220	1	1.124259	0.9;
	1392	1	39.64395	11.9	0	0	1	1.015967	-36.617099	110	1	1.1	0.9;
	1393	1	5.806438	2	0	0	1	0.962911	-26.12829	110	1	1.1	0.9;
	1394	1	0	8.4	0	0	1	1.038261	-21.049991	150	1	1.1	0.9;
	1395	1	0	0	0	65.8105	1	1.060919	-11.306723	150	1	1.113455	0.9;
	1396	1	2.502775	1.3	0	0	1	1.011424	-14.947239	110	1	1.1	0.9;
	1397	1	43.8	20.7	0	-42.3306	1	1.06716	-17.594597	380	1	1.118547	0.9;
	1398	1	0	0	0.28785	0	1	1.062352	-20.892642	380	1	1.140956	0.9;
	1399	1	20.32253	8.9	0	0	1	0.955922	-25.511825	110	1	1.1	0.9;
	1400	1	0	0	0	0	1	1.020044	-14.187352	110	1	1.1	0.9;
	1401	1	0	0	0	0	1	0.971835	-27.334067	150	1	1.1	0.9;
	1402	1	36.34029	15.2	0	0	1	1.042862	-8.265546	220	1	1.118107	0.9;
	1403	2	106.7	12.4	0	0	1	1.088761	-14.726452	220	1	1.165481	0.9;
	1404	1	0	0	0.296958	-84.3852	1	1.090288	-11.972403	380	1	1.15517	0.9;
	1405	1	125.4	45.8	0	0	1	1.087262	-18.579179	220	1	1.151065	0.9;
	1406	1	170.4	81.9	0	0	1	1.091811	-31.006362	380	1	1.134949	0.9;
	1407	1	116.7	64	0	0	1	1.02101	-45.060116	220	1	1.1	0.9;
	1408	1	0	0	0	0	1	1.090345	-20.137685	380	1	1.171877	0.9;
	1409	1	42.54717	8.2	0	0	1	1.078337	-22.236749	220	1	1.16324	0.9;
	1410	2	0	0	0	0	1	1.066526	-38.203373	380	1	1.13469	0.9;
	1411	1	0	0	0	0	1	1.045405	-14.232365	220	1	1.1	0.9;
	1412	1	7.608436	2.5	0	0	1	0.954516	-13.104112	110	1	1.1	0.9;
	1413	1	300.2	-3.1	0	0	1	1.080085	-48.18551	380	1	1.115488	0.9;
	1414	2	0	0	0	0	1	1.103737	-9.485996	380	1	1.158879	0.9;
	1415	1	0	0	0	0	1	1.080434	-19.414614	220	1	1.159022	0.9;
	1416	1	20.42264	8.1	0	0	1	1.071489	-15.934231	380	1	1.152932	0.9;
	1417	1	-21.624	-3.9	0	0	1	1.086074	-16.198131	220	1	1.131133	0.9;
	1418	1	0	0	0	0	1	1.056748	-15.753986	220	1	1.129308	0.9;
	1419	2	0	0	0	0	1	1.092237	-8.087624	380	1	1.128296	0.9;
	1420	1	0	0	0	0	1	1.095288	-11.339243	220	1	1.170055	0.9;
	1421	2	0	0	0.178992	0	1	1.064211	-16.157657	380	1	1.112869	0.9;
	1422	2	0	0	0	0	1	1.040733	-17.080269	220	1	1.113643	0.9;
	1423	1	175.2	82.4	0	0	1	1.065295	-38.962096	220	1	1.143097	0.9;
	1424	1	77.3858	24.1	0	0	1	1.052814	-24.688593	220	1	1.127955	0.9;
	1425	1	0	0	0.095843	0	1	1.053968	-13.412752	220	1	1.114573	0.9;
	1426	1	140.4	49.4	0	0	1	1.068942	6.077462	150	1	1.12422	0.9;
	1427	1	156.9	-4	0	0	1	1.06843	-23.445591	220	1	1.153324	0.9;
	1428	1	60.0666	27	0	0	1	1.024718	-20.485976	110	1	1.1	0.9;
	1429	1	69.07658	21.5	0	0	1	1.057421	-24.002728	220	1	1.132058	0.9;
	1430	1	0	0	0	11.06628	1	1.0415	-24.684348	150	1	1.1	0.9;
	1431	2	0	0	0	0	1	1.095842	-31.415193	380	1	1.165337	0.9;
	1432	1	329.4	12.9	0	0	1	1.085113	-28.859468	380	1	1.1433	0.9;
	1433	1	0	0	0	0	1	1.024563	-23.569742	150	1	1.1	0.9;
	1434	1	34.03774	4.4	0	0	1	1.006465	-36.913937	110	1	1.1	0.9;
	1435	1	0	0	0	0	1	1.040874	-29.526323	220	1	1.1	0.9;
	1436	1	0	0	0	0	1	1.045656	12.118274	380	1	1.106982	0.9;
	1437	1	32.03552	19.8	0	0	1	1.048467	-15.818185	150	1	1.105765	0.9;
	1438	1	0	0	0	0	1	1.042542	-12.136696	220	1	1.114054	0.9;
	1439	1	0	0	0	0	1	1.033087	-1.927681	380	1	1.1	0.9;
	1440	1	0	0	0	0	1	1.039866	-15.691519	150	1	1.1	0.9;
	1441	1	146.6	3.3	0	0	1	1.054052	-13.269986	220	1	1.115162	0.9;
	1442	2	0	0	0	0	1	1.084737	67.225872	380	1	1.1	0.9;
	1443	2	5.706326	1.9	0	0	1	0.98905	-19.178121	110	1	1.1	0.9;
	1444	1	0	0	0	0	1	1.022383	-13.62466	220	1	1.116153	0.9;
	1445	1	30.53385	10.4	0	0	1	1.058817	-4.74426	110	1	1.134348	0.9;
	1446	1	0	0	0	0	1	1.052549	-30.587652	380	1	1.120313	0.9;
	1447	1	78.5	137.8	0	0	1	0.951133	57.250853	220	1	1.1	0.9;
	1448	1	104.5	-56.4	0	0	1	1.071506	-7.296736	380	1	1.139813	0.9;
	1449	1	25.72853	9.3	0	0	1	1.025303	-16.833271	110	1	1.127294	0.9;
	1450	2	0	0	0	0	1	1.04174	13.591568	380	1	1.110044	0.9;
	1451	1	28.33141	44.8	0	0	1	1.083188	-3.299275	220	1	1.14459	0.9;
	1452	1	0	0	0	0	1	1.099824	61.294434	380	1	1.117193	0.9;
	1453	2	95.10545	21.7	0	0	1	1.046454	-27.759196	220	1	1.135464	0.9;
	1454	1	76.28458	45.7	0	0	1	1.077793	41.797847	380	1	1.115932	0.9;
	1455	1	51.95761	13.4	0	0	1	1.059031	9.321907	110	1	1.138556	0.9;
	1456	1	58.86526	36.4	0	0	1	0.976904	-22.764718	150	1	1.1	0.9;
	1457	1	0	0	0	0	1	1.039398	-44.330879	220	1	1.1	0.9;
	1458	1	38.74295	14.1	0	0	1	1.053401	-49.158103	220	1	1.110715	0.9;
	1459	1	63.5	-381.2	0.550858	0	1	1.070906	26.7089	380	1	1.1	0.9;
	1460	1	45.75072	4.2	0	0	1	1.059454	-54.618359	220	1	1.1	0.9;
	1461	1	137.2	16.1	0	0	1	1.041271	-23.126896	220	1	1.1	0.9;
	1462	2	0	0	0	0	1	1.0405	-6.236207	220	1	1.129808	0.9;
	1463	1	0	0	0	0	1	1.070451	-23.607654	220	1	1.171671	0.9;
	1464	2	0	0	0	0	1	1.097816	8.748499	380	1	1.129552	0.9;
	1465	1	73.7818	0	0	0	1	1.07079	-5.321793	380	1	1.124159	0.9;
	1466	1	25.12786	15.6	0	0	1	1.047683	-25.536543	150	1	1.1	0.9;
	1467	1	0	0	0	0	1	1.056415	-19.75713	220	1	1.125437	0.9;
	1468	1	164.8	42.5	0	0	1	1.051855	-41.961478	380	1	1.109143	0.9;
	1469	1	274.1	166.5	0	0	1	1.080835	-9.018311	150	1	1.147669	0.9;
	1470	1	149.8	63.1	0	0	1	1.075639	-16.662785	220	1	1.142364	0.9;
	1471	1	0	0	0	0	1	1.035773	11.357558	220	1	1.1	0.9;
	1472	1	50.0555	0	0	0	1	1.036105	-32.110208	150	1	1.1	0.9;
	1473	1	0	0	0	0	1	1.031779	-4.630219	380	1	1.1	0.9;
	1474	1	14.71632	5.3	0	0	1	1.070759	-0.379171	110	1	1.138204	0.9;
	1475	2	152.1	194.4	0	0	1	1.071632	-39.403361	380	1	1.135017	0.9;
	1476	1	20.32253	2.5	0	0	1	0.916618	-26.435399	110	1	1.1	0.9;
	1477	1	0	0	0	0	1	1.057436	-30.223855	380	1	1.123334	0.9;
	1478	2	0	0	0	0	1	1.030136	-5.556483	220	1	1.109634	0.9;
	1479	1	0	0	0	0	1	1.009528	39.734393	220	1	1.1	0.9;
	1480	1	164.2	-10.7	0	0	1	1.058801	-15.674027	220	1	1.112963	0.9;
	1481	1	73.18114	22.3	0	0	1	1.050155	-20.496786	220	1	1.138845	0.9;
	1482	1	123.5	-49.4	0	0	1	1.069166	-10.025969	220	1	1.13577	0.9;
	1483	1	3.303663	1.1	0	0	1	0.996214	-35.237396	110	1	1.1	0.9;
	1484	1	41.94651	8.5	0	0	1	1.050283	-11.276916	220	1	1.120065	0.9;
	1485	1	0	0	0	0	1	1.055835	33.894772	380	1	1.113088	0.9;
	1486	1	0	0	0	0	1	1.000632	-23.971326	150	1	1.1	0.9;
	1487	1	20.82309	7.9	0	0	1	1.055758	-47.792241	220	1	1.107812	0.9;
	1488	1	0	0	0	0	1	1.04345	14.702286	150	1	1.1	0.9;
	1489	1	5.305883	2.3	0	0	1	0.96198	-11.311463	110	1	1.1	0.9;
	1490	2	0	0	0	0	1	1.059861	9.042542	150	1	1.1	0.9;
	1491	1	0	0	0.122611	0	1	0.978776	67.068683	380	1	1.1	0.9;
	1492	1	-8.40932	2.7	0	0	1	1.030457	-19.149786	150	1	1.109908	0.9;
	1493	1	70.0777	11.1	0	0	1	1.054329	-49.289182	220	1	1.112286	0.9;
	1494	1	51.65727	8	0	0	1	1.057312	-48.491756	220	1	1.115238	0.9;
	1495	1	33.23685	11.5	0	0	1	1.066717	-17.234306	110	1	1.118164	0.9;
	1496	2	0	0	0	0	1	1.078	-0.521835	380	1	1.139948	0.9;
	1497	1	95.70611	25.9	0	0	1	1.052856	-24.721443	220	1	1.127814	0.9;
	1498	2	0	0	0	0	1	1.085557	3.011276	380	1	1.130872	0.9;
	1499	1	60.26682	9.4	0	0	1	1.049998	-49.48291	220	1	1.112114	0.9;
	1500	2	15.5172	1.2	0	0	1	1.041667	62.768437	150	1	1.1	0.9;
	1501	1	13.31476	4.8	0	0	1	1.06303	-16.871544	110	1	1.138206	0.9;
	1502	1	103.7	14.9	0	0	1	1.057643	-21.348618	220	1	1.1	0.9;
	1503	1	33.03663	15.7	0	0	1	1.006906	-21.87433	110	1	1.1	0.9;
	1504	1	0	0	0	0	1	1.082591	-15.408247	380	1	1.126748	0.9;
	1505	1	0	0	0	0	1	1.084971	-2.272792	220	1	1.145358	0.9;
	1506	1	22.02442	11.9	0	0	1	0.950378	-26.51067	110	1	1.1	0.9;
	1507	1	0	0	0.184	0	1	1.047104	-26.6495	380	1	1.105857	0.9;
	1508	1	0	0	0	0	1	1.09094	-22.493426	380	1	1.165943	0.9;
	1509	1	15.21687	7.6	0	0	1	0.965826	-9.442007	110	1	1.1	0.9;
	1510	1	0	0	0	0	1	1.075218	-14.342638	380	1	1.113589	0.9;
	1511	1	257.2	-26.8	0	0	1	1.04277	-1.098104	220	1	1.114529	0.9;
	1512	1	25.82864	5.7	0	0	1	1.000786	-22.42929	110	1	1.1	0.9;
	1513	1	-13.9154	4.6	0	0	1	1.03101	-24.242247	150	1	1.1	0.9;
	1514	1	22.02442	7.5	0	0	1	0.974973	-24.168823	110	1	1.1	0.9;
	1515	2	19.02109	4.7	0	0	1	1.076143	-19.793582	220	1	1.140842	0.9;
	1516	1	-592.205	-227.4	0.466992	0	1	1.075832	-8.262329	380	1	1.13857	0.9;
	1517	1	20.0222	6.6	0	0	1	0.951673	-29.260635	110	1	1.1	0.9;
	1518	1	34.03774	0	0	0	1	1.028872	-43.898627	220	1	1.1	0.9;
	1519	2	0	0	0	0	1	1.094364	45.053162	380	1	1.116776	0.9;
	1520	1	0	0	0.251998	0	1	1.075566	-15.12992	380	1	1.122679	0.9;
	1521	1	0	0	0.191685	0	1	1.099367	-6.506666	220	1	1.158006	0.9;
	1522	1	46.4515	19.1	0	0	1	1.087176	0.750438	150	1	1.154017	0.9;
	1523	1	3.303663	1.1	0	0	1	1.025533	-35.153359	110	1	1.1	0.9;
	1524	2	0	0	0	0	1	1.095247	-11.351247	220	1	1.170039	0.9;
	1525	2	4.705216	89.2	0	0	1	1.047019	61.886246	150	1	1.1	0.9;
	1526	1	16.91876	5.6	0	0	1	0.930913	-35.380515	110	1	1.1	0.9;
	1527	1	55.26127	-11.1	0	0	1	1.055982	-10.080928	220	1	1.127476	0.9;
	1528	1	0	0	0.154873	0	1	1.046773	-3.662338	380	1	1.1	0.9;
	1529	1	9.210212	3.3	0	0	1	1.053187	-15.319766	110	1	1.112544	0.9;
	1530	1	0	0	0	-14.7231	1	1.01622	-26.8587	150	1	1.183944	0.9;
	1531	1	0	0	0	0	1	1.056804	-32.477662	380	1	1.120031	0.9;
	1532	2	0	0	0	0	1	1.084579	-2.626137	380	1	1.148389	0.9;
	1533	1	7.408214	2.7	0	0	1	0.965198	-18.447	110	1	1.1	0.9;
	1534	2	20.22242	7.3	0	0	1	1.047354	-17.140158	110	1	1.130901	0.9;
	1535	1	0	0	0	0	1	1.023401	-17.848929	220	1	1.1	0.9;
	1536	2	415.5	52.5	0	0	1	1.062921	-34.261268	380	1	1.123418	0.9;
	1537	2	0	0	0	0	1	1.0234	-33.988231	150	1	1.1	0.9;
	1538	1	0	0	0.401381	0	1	1.046561	-20.33932	380	1	1.132541	0.9;
	1539	1	0	0	0	0	1	0.967916	-28.328391	150	1	1.1	0.9;
	1540	1	72.18002	31.9	0	0	1	1.066092	-38.895352	220	1	1.143843	0.9;
	1541	1	0	0	0	0	1	1.057499	9.97086	380	1	1.111317	0.9;
	1542	1	118.2	-33	0	0	1	1.029081	-19.315872	220	1	1.1	0.9;
	1543	1	83.19224	25.8	0	0	1	1.04874	-10.64346	220	1	1.119023	0.9;
	1544	2	29.23241	2.5	0	0	1	0.941739	-27.202264	110	1	1.1	0.9;
	1545	1	0	0	0.095843	0	1	1.028288	-18.659623	220	1	1.108718	0.9;
	1546	1	12.41376	4.4	0	0	1	0.970679	-20.185607	110	1	1.1	0.9;
	1547	2	207.9	9.2	0	0	1	1.087842	-45.684724	380	1	1.129694	0.9;
	1548	1	14.71632	5.3	0	0	1	1.084086	1.543403	110	1	1.148259	0.9;
	1549	1	4.304773	1.8	0	0	1	0.996669	-22.820275	110	1	1.1	0.9;
	1550	1	5.00555	2	0	0	1	0.98418	-24.283648	110	1	1.1	0.9;
	1551	1	0	0	0	0	1	0.956503	-29.30269	150	1	1.1	0.890689;
	1552	1	47.45261	18.7	0	0	1	1.079603	3.219078	150	1	1.148856	0.9;
	1553	1	68.2757	39	0.226015	0	1	1.085753	-39.641906	380	1	1.147151	0.9;
	1554	2	62.56937	8	0	0	1	1.048673	-13.159743	220	1	1.118404	0.9;
	1555	1	15.21687	2.5	0	0	1	0.923823	-28.733328	110	1	1.1	0.9;
	1556	1	0	0	0	0	1	1.042957	19.982571	380	1	1.1	0.9;
	1557	1	92.60267	36.3	0	0	1	1.003933	11.690271	220	1	1.1	0.9;
	1558	2	0	0	0	0	1	1.079342	-12.712394	380	1	1.150723	0.9;
	1559	2	0	0	0	0	1	1.064406	-13.906042	380	1	1.12585	0.9;
	1560	1	138.2	40.3	0	0	1	1.007272	-19.733683	110	1	1.1	0.9;
	1561	1	25.12786	-61.6	0	0	1	1.053044	-29.433914	220	1	1.146819	0.9;
	1562	1	2.502775	0.6	0	0	1	0.992644	-23.294223	110	1	1.1	0.9;
	1563	1	4.905439	1.4	0	0	1	0.971022	-27.787759	110	1	1.1	0.9;
	1564	1	53.55938	17.2	0	0	1	0.96728	-25.712909	110	1	1.1	0.9;
	1565	2	48.2535	6.4	0	0	1	1.019636	-29.537982	110	1	1.1	0.9;
	1566	1	0	0	0	0	1	1.067258	-37.453485	380	1	1.135704	0.9;
	1567	2	0	0	0	0	1	0.981001	-18.754176	110	1	1.1	0.9;
	1568	1	-14.3159	4.7	0	0	1	0.985174	-24.312737	150	1	1.1	0.9;
	1569	2	157.6333	115.5	0	0	1	1.065273	24.821342	220	1	1.114842	0.9;
	1570	2	0	0	0	0	1	1.052079	13.039106	380	1	1.112275	0.9;
	1571	2	0	0	0	0	1	1.096591	2.319811	220	1	1.163253	0.9;
	1572	2	0	0	0	0	1	1.066343	7.733385	150	1	1.115362	0.9;
	1573	1	0	0	0	0	1	1.08371	-16.510568	380	1	1.152623	0.9;
	1574	1	8.00888	4.9	0	0	1	0.997561	-24.755577	150	1	1.1	0.9;
	1575	1	0	0	0	0	1	1.032502	-25.294349	150	1	1.1	0.9;
	1576	2	38.04218	29	0	0	1	1.0735	-21.712517	220	1	1.118914	0.9;
	1577	1	0	0	0	0	1	0.992061	-20.283226	150	1	1.1	0.9;
	1578	1	34.43818	14.4	0	0	1	1.092603	-44.228184	380	1	1.134529	0.9;
	1579	1	68.37582	42.4	0	0	1	1.005417	-20.211066	150	1	1.1	0.9;
	1580	2	0	0	0	0	1	1.038717	-13.917056	380	1	1.132388	0.9;
	1581	1	19.92209	6.6	0	0	1	0.960676	-27.131501	110	1	1.1	0.9;
	1582	1	0	0	0.148479	0	1	1.077155	-19.737808	380	1	1.147614	0.9;
	1583	1	16.01776	5.3	0	0	1	0.966738	-32.096715	110	1	1.1	0.9;
	1584	1	0	0	0	0	1	1.029058	-49.638507	220	1	1.1	0.9;
	1585	1	0	0	0	0	1	1.064498	-15.151712	150	1	1.117758	0.9;
	1586	2	0	0	0	0	1	1.011364	-19.407991	110	1	1.1	0.9;
	1587	1	9.710767	3.5	0	0	1	1.056451	-17.590331	110	1	1.135925	0.9;
	1588	1	0	0	0	0	1	1.06681	-8.556398	220	1	1.12901	0.9;
	1589	1	10.11121	2.5	0	0	1	0.976099	-22.811038	110	1	1.1	0.9;
	1590	1	162.5	48.4	0	0	1	1.005024	-41.221466	220	1	1.1	0.9;
	1591	1	0	0	0	0	1	1.034308	-32.485081	220	1	1.134557	0.9;
	1592	2	0	-0.1	0	0	1	1.085591	-13.416257	220	1	1.170163	0.9;
	1593	1	2.00222	0.3	0	0	1	0.990662	-27.32064	110	1	1.1	0.9;
	1594	1	0	0	0	11.0253	1	1.04331	-21.001063	150	1	1.1	0.9;
	1595	1	0	0	0	0	1	1.05981	-22.957794	220	1	1.133772	0.9;
	1596	2	0	0	0	0	1	1.073026	10.864149	380	1	1.125198	0.9;
	1597	2	0	0	0	0	1	1.068105	-34.666452	380	1	1.130775	0.9;
	1598	1	39.24351	24.3	0	0	1	1.031855	-25.496848	150	1	1.1	0.9;
	1599	1	58.46482	6.6	0	0	1	1.071578	-18.271598	220	1	1.139318	0.9;
	1600	1	11.51276	2.3	0	0	1	1.017485	-14.54073	110	1	1.1	0.9;
	1601	1	85.79512	13.4	0	0	1	1.050788	-48.326364	220	1	1.110751	0.9;
	1602	1	0	0	0.119999	0	1	0.987557	-22.443821	220	1	1.1	0.9;
	1603	2	0	0	0	0	1	1.063	-16.376579	380	1	1.143822	0.9;
	1604	2	0	0	0	0	1	1.021069	-21.58337	110	1	1.1	0.9;
	1605	2	0	0	0.133294	0	1	1.024682	-20.77908	220	1	1.1	0.9;
	1606	1	80.0888	20	0	0	1	1.044027	16.269152	220	1	1.1	0.9;
	1607	1	0	0	0	0	1	1.073339	-4.602966	380	1	1.140462	0.9;
	1608	1	0	0	0	0	1	1.092533	-30.405019	220	1	1.144231	0.9;
	1609	1	328.9	30.4	0	0	1	1.063069	-41.416065	380	1	1.133392	0.9;
	1610	1	11.11232	6.9	0	0	1	0.990985	-23.000704	150	1	1.1	0.9;
	1611	1	45.75072	20.1	0	0	1	1.016454	-16.843255	110	1	1.1	0.9;
	1612	1	0	0	0	0	1	1.0264	-30.060736	380	1	1.134226	0.9;
	1613	1	25.42819	14.7	0	0	1	1.043533	-18.556644	220	1	1.117444	0.9;
	1614	1	122.2	68.9	0	0	1	1.093721	-7.959084	380	1	1.157223	0.9;
	1615	1	26.02886	8.7	0	0	1	0.996829	-22.713864	110	1	1.1	0.9;
	1616	1	333.5	-1.4	0	0	1	1.034685	-45.002232	220	1	1.152079	0.9;
	1617	1	73.98203	22.5	0	0	1	1.069988	-10.720539	220	1	1.131764	0.9;
	1618	1	35.13896	1.8	0	0	1	1.068313	-42.4901	220	1	1.119922	0.9;
	1619	1	0	0	0	0	1	1.033813	-37.331954	220	1	1.1	0.9;
	1620	1	22.42486	29.5	0	0	1	0.973529	-25.885712	150	1	1.1	0.9;
	1621	1	0	0	0	0	1	1.03126	29.815778	220	1	1.1	0.9;
	1622	1	-32.0355	10.3	0	0	1	1.08635	-41.190867	380	1	1.146141	0.9;
	1623	1	0	0	0	0	1	1.046017	-28.731866	220	1	1.113379	0.9;
	1624	2	0	0	0	0	1	1.084526	68.851453	380	1	1.1	0.9;
	1625	1	0	0	0	0	1	1.06828	11.312509	110	1	1.146394	0.9;
	1626	2	0	0	0	0	1	1.043385	14.686877	150	1	1.1	0.9;
	1627	1	5.506105	2	0	0	1	1.019217	-20.833485	110	1	1.1	0.9;
	1628	1	-1.8	-4.8	0	0	1	1.077063	-17.717886	150	1	1.1	0.9;
	1629	1	16.51831	5.5	0	0	1	1.040079	-29.568847	110	1	1.1	0.9;
	1630	1	28.63174	10	0	0	1	1.038312	-15.1678	110	1	1.127277	0.9;
	1631	1	0	-0.8	0	0	1	1.051053	5.125141	110	1	1.139539	0.9;
	1632	1	0	0	0	12.57819	1	1.000793	-24.4208	150	1	1.1	0.9;
	1633	2	0	0	0	0	1	0.967727	-12.726152	110	1	1.1	0.9;
	1634	2	18.4	10.1	0	-52.5877	1	1.098053	-12.085788	380	1	1.134093	0.9;
	1635	1	32.93652	10.1	0	0	1	1.025129	8.06856	110	1	1.106663	0.9;
	1636	2	0	0	0	0	1	1.043079	12.550941	380	1	1.110579	0.9;
	1637	1	45.75072	27.5	0	0	1	1.078536	-16.544345	220	1	1.118758	0.9;
	1638	1	0	0	0	0	1	1.030573	-2.352541	220	1	1.108488	0.9;
	1639	1	-19.3214	6.4	0	0	1	1.053766	-16.104425	150	1	1.109113	0.9;
	1640	1	0	0	0	0	1	1.032172	17.072875	380	1	1.1	0.9;
	1641	2	166.1	64.9	0	0	1	1.017091	-40.134796	220	1	1.1	0.9;
	1642	2	0	0	0	0	1	1.092	-0.717685	380	1	1.152964	0.9;
	1643	1	71.17892	23.8	0	0	1	0.975064	-25.139677	110	1	1.1	0.9;
	1644	1	13.31476	7.9	0	0	1	0.985744	-23.427539	110	1	1.1	0.9;
	1645	2	12.31365	0.2	0	0	1	0.979909	-22.070432	110	1	1.1	0.9;
	1646	2	0	0	0	0	1	1.037611	-7.797658	380	1	1.1	0.9;
	1647	1	0	0	0	0	1	1.026137	-17.992618	150	1	1.1	0.9;
	1648	1	8.509435	3.2	0	0	1	1.021233	-42.245154	220	1	1.1	0.9;
	1649	1	0	0	0	0	1	1.062173	-21.051612	220	1	1.139185	0.9;
	1650	1	23.02553	10.5	0	0	1	1.061057	-35.346347	220	1	1.142215	0.9;
	1651	1	25.42819	8.3	0	0	1	0.915108	-29.723929	110	1	1.1	0.9;
	1652	1	0	0	0	0	1	1.093858	35.188922	380	1	1.127727	0.9;
	1653	2	166.2	55.1	0	0	1	1.075182	-0.680521	220	1	1.144216	0.9;
	1654	1	0	0	0	-573.554	1	0.976383	-17.441685	750	1	1.1	0.9;
	1655	1	52.45816	11.3	0	0	1	1.070179	12.147287	110	1	1.147543	0.9;
	1656	1	99.51033	1.5	0	0	1	1.083448	34.161267	380	1	1.129362	0.9;
	1657	2	116.8	30.4	0	0	1	1.036709	-7.100736	220	1	1.125426	0.9;
	1658	2	34.13785	-2.8	0	0	1	1.088474	-26.226652	380	1	1.166874	0.9;
	1659	1	63.66304	9.7	0	0	1	1.081173	-2.392743	220	1	1.138544	0.9;
	1660	1	335.3	-101.3	0	0	1	1.086222	-21.30892	380	1	1.137635	0.9;
	1661	2	109.5	33.7	0	0	1	1.069693	-11.898239	380	1	1.166047	0.9;
	1662	1	-29.733	9.8	0	0	1	1.045113	-19.908863	150	1	1.1	0.9;
	1663	2	0	0	0	0	1	1.093262	-20.244147	380	1	1.163991	0.9;
	1664	1	0	0	0	0	1	1.030585	-12.523381	220	1	1.123097	0.9;
	1665	1	14.61621	2	0	0	1	1.099942	-7.615854	220	1	1.156779	0.9;
	1666	2	0	0	0	0	1	1.065737	-40.613091	380	1	1.137863	0.9;
	1667	2	0	0	0.328599	20	1	1.030737	17.054978	380	1	1.1	0.9;
	1668	2	0	0	0	0	1	1.041029	-19.93418	150	1	1.120587	0.9;
	1669	1	0	0	0	0	1	1.054603	-0.478483	380	1	1.112492	0.9;
	1670	1	56.4626	22	0	0	1	1.083988	-41.42952	220	1	1.155782	0.9;
	1671	1	2.502775	0.6	0	0	1	1.022935	-15.279912	110	1	1.1	0.9;
	1672	1	5.00555	13	0	0	1	1.072283	-23.045491	220	1	1.1	0.9;
	1673	1	17.81976	6.7	0	0	1	1.002143	-18.066246	110	1	1.1	0.9;
	1674	1	0	0	0	0	1	1.034858	-24.796436	150	1	1.1	0.9;
	1675	1	34.13785	21.1	0	0	1	1.029966	-24.478599	150	1	1.1	0.9;
	1676	1	0	0	0	0	1	1.072555	-21.340499	220	1	1.155432	0.9;
	1677	1	8.209102	2.5	0	0	1	0.992675	-18.790669	110	1	1.1	0.9;
	1678	2	102.2	45.4	0	0	1	1.021406	1.478548	220	1	1.121476	0.9;
	1679	1	62.36915	25.5	0	0	1	1.077577	-0.413992	150	1	1.144493	0.9;
	1680	2	202.4	66.4	0	0	1	1.041069	-0.488226	220	1	1.114549	0.9;
	1681	1	0	0	0	0	1	1.07381	-21.043675	220	1	1.156	0.9;
	1682	1	0.4	-170.2	0	0	1	1.074727	-15.229126	380	1	1.126138	0.9;
	1683	1	0	0	0	0	1	1.030631	-24.292102	150	1	1.1	0.9;
	1684	1	56.26238	8.5	0	0	1	1.056515	-48.621876	220	1	1.114258	0.9;
	1685	1	0	0	0	0	1	1.086185	-2.726652	220	1	1.146183	0.9;
	1686	2	30.53385	15.1	0	0	1	1.029	-15.337851	110	1	1.132265	0.9;
	1687	2	0	0	0	0	1	1.095264	-11.350693	220	1	1.170055	0.9;
	1688	2	0	0	0	0	1	0.977909	-22.784621	110	1	1.1	0.9;
	1689	1	22.02442	21.8	0	0	1	1.056931	-29.698285	220	1	1.11107	0.9;
	1690	1	5.105661	2.4	0	0	1	0.996166	-4.990037	110	1	1.1	0.9;
	1691	2	7.808658	5.8	0	0	1	1.081318	-46.663151	220	1	1.127902	0.9;
	1692	2	0	0	0	0	1	1.060364	-12.51623	220	1	1.144179	0.9;
	1693	2	295.8	90.9	0	0	1	1.047409	-25.385501	220	1	1.127081	0.9;
	1694	1	25.72853	8.9	0	0	1	0.997595	-16.488296	110	1	1.1	0.9;
	1695	1	0.200222	0.5	0	0	1	1.003749	60.303766	220	1	1.1	0.9;
	1696	1	0	0	0	0	1	0.9871	8.466334	220	1	1.1	0.9;
	1697	1	5.506105	2	0	0	1	1.092155	2.610009	110	1	1.154379	0.9;
	1698	1	0	0	0	0	1	1.054553	-10.118408	220	1	1.120942	0.9;
	1699	1	10.51165	4.7	0	0	1	0.989806	-17.16555	110	1	1.1	0.9;
	1700	1	369.5	16.7	0	0	1	1.004281	-27.221113	220	1	1.1	0.9;
	1701	1	88.79845	-3.2	0	0	1	1.050543	-25.509827	220	1	1.128427	0.9;
	1702	1	0	0	0	0	1	1.064995	20.495402	380	1	1.126296	0.9;
	1703	1	177.4	41.7	0	0	1	1.00778	-36.543341	150	1	1.1	0.9;
	1704	1	0	0	0	0	1	1.087569	28.190954	220	1	1.148426	0.9;
	1705	1	-19.6218	28.6	0	0	1	1.023823	-20.053123	150	1	1.120748	0.9;
	1706	1	55.26127	23.2	0	0	1	1.018627	-22.368885	220	1	1.1	0.9;
	1707	1	209.2	23.5	0	0	1	1.050561	-27.820616	220	1	1.143419	0.9;
	1708	1	0	0	0	0	1	1.077723	-35.157515	220	1	1.156576	0.9;
	1709	1	11.71299	3.8	0	0	1	1.072233	-14.31271	220	1	1.128262	0.9;
	1710	2	0	0	0	0	1	1.053042	-40.359058	220	1	1.11821	0.9;
	1711	1	0	0	0	0	1	1.067321	1.062909	150	1	1.129745	0.9;
	1712	2	95.1	35	0	0	1	1.013818	-47.209009	220	1	1.1	0.9;
	1713	1	94.30456	13.8	0	0	1	1.051014	-49.386974	220	1	1.113156	0.9;
	1714	1	83.29234	30.1	0	0	1	1.043117	-13.99344	110	1	1.158069	0.9;
	1715	1	223.9	72.6	0	0	1	1.040291	-27.194634	110	1	1.2	0.8;
	1716	1	0	0	0	12.60168	1	0.999869	-23.487073	150	1	1.1	0.9;
	1717	1	0	0	0	0	1	1.075466	-13.686052	380	1	1.146017	0.9;
	1718	2	1.201332	0.8	0	0	1	1.073867	-14.805858	150	1	1.118669	0.9;
	1719	1	0	0	0	0	1	0.994977	-22.771881	110	1	1.1	0.9;
	1720	1	0	0	0	0	1	1.034783	-24.416259	150	1	1.1	0.9;
	1721	1	50.25572	16.4	0	0	1	1.061422	0.471738	150	1	1.123912	0.9;
	1722	1	64.27126	20.1	0	0	1	1.052519	-36.57683	220	1	1.134243	0.9;
	1723	1	27.3303	5	0	0	1	1.079582	-15.435364	220	1	1.142619	0.9;
	1724	2	289.4	96.2	0	0	1	1.082211	-9.673947	380	1	1.126166	0.9;
	1725	1	0	0	0	0	1	1.089113	-23.923314	380	1	1.144333	0.9;
	1726	1	2.202442	1.1	0	0	1	1.006857	-36.046673	110	1	1.1	0.9;
	1727	2	195.1	41.4	0	0	1	1.084227	-3.40484	220	1	1.144289	0.9;
	1728	1	57.16338	0	0	0	1	1.047176	6.149572	110	1	1.11921	0.9;
	1729	2	0	0	0	0	1	1.042605	48.838336	380	1	1.114936	0.9;
	1730	1	28.63174	17.8	0	0	1	0.97218	-24.849475	150	1	1.1	0.9;
	1731	1	0	0	0	0	1	1.068348	-3.731118	380	1	1.127108	0.9;
	1732	1	22.22464	5.9	0	0	1	1.065912	-43.35851	220	1	1.114767	0.9;
	1733	1	20.82309	0	0	0	1	1.044736	7.888855	220	1	1.118103	0.9;
	1734	1	0	0	0	0	1	1.049808	-5.537084	380	1	1.134796	0.9;
	1735	1	53.95983	4.9	0	0	1	1.058082	-54.860986	220	1	1.1	0.9;
	1736	2	0.500555	21.5	0	0	1	1.096079	-31.987342	380	1	1.163913	0.9;
	1737	1	0	0	0.227005	0	1	1.104359	-42.674769	380	1	1.162511	0.9;
	1738	2	0	0	0	0	1	1.078947	59.161968	380	1	1.1	0.9;
	1739	1	0.200222	1.4	0	0	1	1.011728	10.817881	220	1	1.1	0.9;
	1740	1	26.4293	7.7	0	0	1	1.026726	-34.128746	110	1	1.1	0.9;
	1741	1	0	0	0	0	1	1.033179	-25.317358	150	1	1.1	0.9;
	1742	1	16.01776	5.3	0	0	1	0.978014	-33.950446	110	1	1.1	0.9;
	1743	2	17.5	502.7	0	0	1	1.064237	-8.852507	380	1	1.1	0.9;
	1744	2	0	0	0	0	1	1.021227	61.023584	220	1	1.1	0.9;
	1745	2	2.903219	1	0	0	1	0.983988	-18.556808	110	1	1.1	0.9;
	1746	2	68.37582	6.2	0	0	1	1.068591	-35.937395	220	1	1.116721	0.9;
	1747	1	0	0	0	0	1	1.077948	-6.328511	220	1	1.136568	0.9;
	1748	1	39.54384	14.3	0	0	1	1.048902	-19.345834	110	1	1.105254	0.9;
	1749	1	52.05772	14.3	0	0	1	1.00367	-41.00055	220	1	1.1	0.9;
	1750	1	91.1	54.8	0	0	1	1.087999	8.208077	150	1	1.143019	0.9;
	1751	2	0	0	0	0	1	1.054545	-20.312039	220	1	1.139092	0.9;
	1752	1	33.63729	3.9	0	0	1	1.088236	-24.217319	220	1	1.127687	0.9;
	1753	1	10.51165	3.5	0	0	1	0.947132	-34.996936	110	1	1.1	0.9;
	1754	1	197.5	-31	0	0	1	1.060906	-23.981122	220	1	1.1351	0.9;
	1755	1	0	0	0	0	1	1.070959	0.755863	380	1	1.139511	0.9;
	1756	2	0	0	0	0	1	1.099158	-15.352801	380	1	1.187622	0.9;
	1757	1	0	0	0	0	1	1.072554	-35.594915	380	1	1.122746	0.9;
	1758	1	49.25461	4.9	0	0	1	1.023678	-12.804055	110	1	1.1	0.9;
	1759	1	0	0	0.117	0	1	1.060627	-18.109041	220	1	1.164263	0.9;
	1760	2	195	47.3	0	0	1	1.049394	-11.299289	380	1	1.120755	0.9;
	1761	1	108.3	62	0	0	1	1.049997	-22.038572	220	1	1.1	0.9;
	1762	1	14.91654	4.9	0	0	1	1.054318	-30.124888	110	1	1.119567	0.9;
	1763	2	64.37138	38.3	0	0	1	1.044864	-3.944317	220	1	1.1	0.9;
	1764	1	0	0	0	0	1	1.028079	-35.257724	110	1	1.1	0.9;
	1765	1	14.8	6.6	0	0	1	1.097471	-15.695981	380	1	1.156291	0.9;
	1766	1	24.32697	15	0	0	1	1.032508	-22.742899	150	1	1.1	0.9;
	1767	1	51.85749	13.9	0	0	1	0.983314	-31.375817	110	1	1.2	0.8;
	1768	1	0	0	0	0	1	1.0898	-22.257603	380	1	1.139696	0.9;
	1769	1	0	0	0	0	1	1.035467	-24.80166	150	1	1.1	0.9;
	1770	1	0	0	0	0	1	1.072732	3.688233	150	1	1.144017	0.9;
	1771	1	0	0	0	0	1	1.022946	31.435394	380	1	1.1	0.9;
	1772	2	0	0	0	0	1	1.060729	-20.692146	220	1	1.124181	0.9;
	1773	1	0	0	0	0	1	1.092304	-17.09038	220	1	1.165837	0.9;
	1774	1	44.74961	4.1	0	0	1	1.059866	-53.686415	220	1	1.1	0.9;
	1775	1	4.404884	99.7	0	0	1	1.046441	6.357376	380	1	1.106398	0.9;
	1776	1	0	0	0	0	1	0.958607	58.546113	220	1	1.1	0.9;
	1777	1	0	0	0	0	1	0.967426	-30.526625	150	1	1.1	0.9;
	1778	1	19.42153	-0.8	0	0	1	1.049537	40.078758	220	1	1.1	0.9;
	1779	2	21.8242	10.3	0	0	1	1.08729	-36.112519	380	1	1.138096	0.9;
	1780	1	159.8	45	0	0	1	1.022603	-8.031697	220	1	1.1	0.9;
	1781	1	0	0	0	0	1	1.076097	0.508008	380	1	1.128593	0.9;
	1782	1	0	0	0	0	1	1.041048	-18.611914	220	1	1.1	0.9;
	1783	1	115.1	10.4	0	0	1	1.077915	-27.451114	220	1	1.13672	0.9;
	1784	1	0	0	0.062995	0	1	1.007112	65.354664	380	1	1.1	0.9;
	1785	1	21.62398	13.4	0	0	1	1.063065	-20.119821	150	1	1.1	0.9;
	1786	1	0	0	0	0	1	1.03372	17.09215	380	1	1.1	0.9;
	1787	1	0	0	0	0	1	1.050215	23.932183	380	1	1.127826	0.9;
	1788	1	76.28458	2.9	0	0	1	1.055667	-43.869008	220	1	1.110218	0.9;
	1789	2	0	0	0	0	1	1.097227	-10.50203	220	1	1.17224	0.9;
	1790	1	0	0	0	0	1	0.988245	-24.023199	110	1	1.1	0.9;
	1791	1	3.203552	1.5	0	0	1	0.959522	-13.514562	110	1	1.1	0.9;
	1792	1	0	0	0	0	1	0.958276	58.546671	220	1	1.1	0.9;
	1793	1	29.43263	10.7	0	0	1	1.015303	-16.661348	110	1	1.107751	0.9;
	1794	1	0	0	0	0	1	1.015728	-19.908581	220	1	1.1	0.9;
	1795	1	248.7	8.1	0	0	1	1.035931	-49.301023	220	1	1.162561	0.9;
	1796	1	-20.6	-9.4	0	0	1	1.04013	-24.325246	150	1	1.1	0.9;
	1797	1	46.05106	16.7	0	0	1	1.023967	-15.687738	110	1	1.127041	0.9;
	1798	1	118.4	137	0	0	1	1.074494	-11.191814	150	1	1.145532	0.9;
	1799	1	0	0	0	0	1	1.055113	-48.876893	220	1	1.11273	0.9;
	1800	2	9.210212	6.6	0	0	1	0.993364	-16.623204	110	1	1.1	0.9;
	1801	1	88.99868	-1.8	0	0	1	1.05383	-19.248567	220	1	1.106952	0.9;
	1802	1	1.8	4.9	0	0	1	1.104363	-17.10477	380	1	1.155675	0.9;
	1803	1	210.1	66.2	0	0	1	1.076746	-25.418477	220	1	1.166708	0.9;
	1804	1	241.9	38.8	0	0	1	0.982554	-40.074438	220	1	1.1	0.9;
	1805	1	0	0	0	0	1	1.042932	7.902757	380	1	1.1	0.9;
	1806	1	0	0	0	0	1	1.098202	-15.862108	380	1	1.186342	0.9;
	1807	1	0	0	0.152776	0	1	1.073844	-19.031574	380	1	1.144838	0.9;
	1808	1	0	0	0	0	1	1.074683	-32.93662	220	1	1.177143	0.9;
	1809	1	177.3	36.4	0	0	1	1.071021	-20.378888	220	1	1.135943	0.9;
	1810	1	12.41376	3.9	0	0	1	1.037876	-16.608064	110	1	1.1	0.9;
	1811	1	0	0	0	0	1	1.073834	3.181348	150	1	1.143672	0.9;
	1812	1	147.4	-55.9	0	0	1	1.033851	-47.129047	220	1	1.1	0.9;
	1813	1	0	0	0	0	1	1.002476	-23.210688	150	1	1.1	0.9;
	1814	1	70.1778	44.7	0	0	1	1.014166	-23.33878	150	1	1.1	0.9;
	1815	1	50.25572	5.5	0	0	1	0.999473	-22.34578	110	1	1.1	0.9;
	1816	1	69.77736	20.3	0	0	1	1.065779	10.012692	110	1	1.144788	0.9;
	1817	1	61.8686	-12.4	0	0	1	1.044198	-30.551886	220	1	1.106698	0.9;
	1818	2	0	0	0	0	1	1.066136	-38.885212	220	1	1.144109	0.9;
	1819	1	0.200222	0.2	0	0	1	1.043305	-1.107537	220	1	1.115107	0.9;
	1820	1	89.5	33.9	0	0	1	0.991041	10.981826	220	1	1.1	0.9;
	1821	1	0	0	0.432997	0	1	1.079797	-6.955997	380	1	1.146598	0.9;
	1822	1	15.21687	5.1	0	0	1	0.97808	-20.227752	110	1	1.1	0.9;
	1823	1	0.900999	0.5	0	0	1	1.029212	-27.128934	110	1	1.1	0.9;
	1824	1	25.72853	7.3	0	0	1	1.06863	-19.162605	220	1	1.136935	0.9;
	1825	2	0	0	0	0	1	1.101763	-35.742904	380	1	1.166712	0.9;
	1826	1	54	29.1	0	0	1	1.067083	-17.657536	380	1	1.1181	0.9;
	1827	1	0	0	0	0	1	1.040366	25.791028	380	1	1.130044	0.9;
	1828	1	139.1	60.4	0	0	1	1.008904	10.807994	220	1	1.1	0.9;
	1829	1	0	0	0	0	1	1.07179	-28.844243	110	1	1.132241	0.9;
	1830	1	54.6606	0	0	0	1	1.048243	-22.080269	220	1	1.1	0.9;
	1831	1	-40.4448	1.5	0	0	1	1.041765	-34.406196	220	1	1.1	0.9;
	1832	1	11.41265	4.4	0	0	1	0.98767	-26.644879	110	1	1.1	0.9;
	1833	1	154.8	-47.1	0	0	1	1.049431	-46.817678	220	1	1.110846	0.9;
	1834	1	0	0	0	12.53399	1	0.977314	-27.162759	150	1	1.1	0.9;
	1835	1	0	0	0	0	1	1.016526	-22.074766	110	1	1.1	0.9;
	1836	1	41.44595	4.9	0	0	1	1.073315	-17.787637	220	1	1.140737	0.9;
	1837	1	0	0	0	0	1	1.088437	-3.846111	220	1	1.149284	0.9;
	1838	1	0	0	0	0	1	1.099225	-10.212346	380	1	1.158631	0.9;
	1839	1	25.92875	8	0	0	1	0.991652	-21.546592	110	1	1.1	0.9;
	1840	1	119	-315.3	0	0	1	1.052005	8.518162	380	1	1.127258	0.9;
	1841	2	0	0	0	0	1	1.08841	-24.158387	220	1	1.127582	0.9;
	1842	2	0	0	0	0	1	1.014636	-7.733836	110	1	1.1	0.9;
	1843	1	72.07992	21.3	0	0	1	1.074542	-20.055567	220	1	1.139424	0.9;
	1844	1	0	0	0	0	1	1.061547	-22.965204	220	1	1.135621	0.9;
	1845	1	3.904329	1	0	0	1	1.060512	-23.817039	220	1	1.13467	0.9;
	1846	1	8.108991	2.9	0	0	1	1.005657	-19.218335	110	1	1.1	0.9;
	1847	1	36.4404	12	0	0	1	1.019251	-17.069368	110	1	1.1	0.9;
	1848	1	8.909878	1.8	0	0	1	0.990699	-24.268463	110	1	1.1	0.9;
	1849	1	89.09879	2.3	0	0	1	1.01875	-46.460937	220	1	1.1	0.9;
	1850	1	16.01776	4	0	0	1	1.05905	-20.850531	220	1	1.127518	0.9;
	1851	1	0	0	0.191685	0	1	1.024392	-15.561308	220	1	1.1	0.9;
	1852	1	90.0999	12.7	0	0	1	1.016081	10.893786	220	1	1.1	0.9;
	1853	2	0	0	0	0	1	1.096267	14.431763	150	1	1.145069	0.9;
	1854	1	0	0	0.147005	0	1	1.10175	-35.756914	380	1	1.166732	0.9;
	1855	1	56.86304	35.2	0	0	1	0.98802	-24.241778	150	1	1.1	0.9;
	1856	1	0	0	0	0	1	1.042065	-3.617474	220	1	1.1	0.9;
	1857	1	153.6423	29.2	0	0	1	1.054974	-27.092065	220	1	1.117528	0.9;
	1858	2	0	0	0	0	1	1.079974	-0.278582	380	1	1.141684	0.9;
	1859	1	6.306993	3.8	0	0	1	1.013608	-22.455538	110	1	1.1	0.9;
	1860	1	28.83197	9.6	0	0	1	0.945655	-28.41367	110	1	1.1	0.9;
	1861	1	0	0	0	0	1	1.069177	18.467062	380	1	1.13597	0.9;
	1862	1	297.7	258.6	0	0	1	1.098522	-16.865013	380	1	1.166467	0.9;
	1863	1	178.4	69.1	0	0	1	1.042885	-42.910435	220	1	1.1	0.9;
	1864	1	0	0	0.26688	0	1	1.093576	-11.856906	380	1	1.126236	0.9;
	1865	1	14.01554	6.5	0	0	1	1.065397	-6.852361	220	1	1.133026	0.9;
	1866	1	3.203552	1.7	0	0	1	1.036168	-20.540571	220	1	1.124147	0.9;
	1867	1	0	0	0	0	1	1.081534	-11.890374	220	1	1.147351	0.9;
	1868	1	0	0	0	0	1	0.9814	-25.637355	150	1	1.1	0.9;
	1869	2	0	0	0	0	1	1.084289	67.103658	380	1	1.1	0.9;
	1870	2	0	0	0	0	1	1.000074	-23.122851	150	1	1.1	0.9;
	1871	1	19.22131	6.9	0	0	1	1.032763	-15.342031	110	1	1.105935	0.9;
	1872	2	0	0	0	0	1	1.090211	-1.285702	380	1	1.141607	0.9;
	1873	2	0	0	0	0	1	1.083842	-16.472114	380	1	1.152673	0.9;
	1874	1	0	0	0	22.29212	1	1.037689	-25.230679	150	1	1.1	0.9;
	1875	1	0	0	0	0	1	1.083208	-25.839315	220	1	1.122986	0.9;
	1876	2	0	0	0	0	1	1.085182	-21.647196	220	1	1.133501	0.9;
	1877	1	52.65838	21.5	0	0	1	1.066918	-3.07571	150	1	1.137166	0.9;
	1878	1	0	0	0	0	1	1.034434	-8.797656	380	1	1.115305	0.9;
	1879	1	120.2	20.5	0	0	1	1.050123	-30.412107	220	1	1.143375	0.9;
	1880	1	0	0	0	0	1	1.086366	-41.190921	380	1	1.146157	0.9;
	1881	1	0	0	0	0	1	1.097878	-13.908707	380	1	1.145563	0.9;
	1882	1	115.1	23.5	0	0	1	1.04939	-52.897575	220	1	1.1	0.9;
	1883	1	0	0	0	0	1	0.967869	-13.111072	110	1	1.1	0.9;
	1884	2	0	0	0	0	1	1.007601	-17.4955	110	1	1.1	0.9;
	1885	2	108.7	10.3	0	0	1	1.089181	-11.033622	220	1	1.120533	0.9;
	1886	1	0	0	0	0	1	1.066789	-8.545704	220	1	1.128999	0.9;
	1887	1	0	0	0	35.51478	1	1.017212	-23.760297	150	1	1.1	0.9;
	1888	1	50.15561	18	0	0	1	1.073278	-17.257855	110	1	1.123565	0.9;
	1889	1	1.401554	-2	0	0	1	0.939711	65.740428	150	1	1.1	0.9;
	1890	1	0	0	0	0	1	1.058933	53.897537	220	1	1.1	0.9;
	1891	1	101.2	2.1	0	0	1	1.075445	-16.830962	380	1	1.155867	0.9;
	1892	1	86.89635	30.3	0	0	1	0.993079	-19.805792	110	1	1.1	0.9;
	1893	1	0	0	0	0	1	1.071557	3.091619	380	1	1.127589	0.9;
	1894	1	0	0	0	0	1	1.074334	-0.914673	380	1	1.129326	0.9;
	1895	1	72.18002	44.9	0	0	1	1.077614	10.043754	150	1	1.130993	0.9;
	1896	1	77.18558	18.6	0	0	1	1.070572	-13.939346	220	1	1.138015	0.9;
	1897	1	0	0	0	0	1	1.051476	-11.476172	220	1	1.107013	0.9;
	1898	1	163.7	94.1	0	0	1	1.051695	-7.495835	380	1	1.120709	0.9;
	1899	1	30.9343	2.8	0	0	1	1.027874	-44.240637	220	1	1.1	0.9;
	1900	1	121.4	20.2	0	0	1	1.081319	-17.542843	220	1	1.146095	0.9;
	1901	1	0	0	0	0	1	1.057337	19.943176	380	1	1.1	0.9;
	1902	2	0	0	0	0	1	1.081368	1.008666	380	1	1.132413	0.9;
	1903	2	0	0	0	0	1	1.053682	-16.123186	220	1	1.117174	0.9;
	1904	1	0	0	0	0	1	1.055253	-11.490549	220	1	1.112277	0.9;
	1905	2	208.5	60.9	0	0	1	1.053687	-24.606301	220	1	1.128563	0.9;
	1906	2	0	0	0	0	1	1.083955	-33.227563	220	1	1.133905	0.9;
	1907	2	110.6	53.9	0	0	1	0.975669	-17.901954	110	1	1.1	0.9;
	1908	2	0	0	0	0	1	1.058741	-7.624336	380	1	1.112648	0.9;
	1909	1	65.87304	31.4	0	0	1	0.967386	43.102065	220	1	1.1	0.9;
	1910	1	0	0	0	0	1	1.041508	-24.696325	150	1	1.1	0.9;
	1911	2	3.3	313.7	0	0	1	1.092184	-41.851672	380	1	1.131187	0.9;
	1912	1	54.26016	19.5	0	0	1	1.051172	-20.589914	220	1	1.1	0.9;
	1913	2	0	0	0	0	1	1.049526	2.132312	380	1	1.107161	0.9;
	1914	1	-3.70411	-8	0	0	1	1.05982	-24.885754	380	1	1.121688	0.9;
	1915	2	0.2	0.2	0	0	1	1.082446	-39.609662	380	1	1.152779	0.9;
	1916	1	0	0	0	0	1	1.078344	-5.042039	380	1	1.13691	0.9;
	1917	1	0	0	0	0	1	1.015842	29.767499	220	1	1.1	0.9;
	1918	1	0	0	0	0	1	1.054829	-27.078703	220	1	1.117242	0.9;
	1919	1	0	0	0	0	1	1.081874	67.200911	380	1	1.1	0.9;
	1920	1	0	0	0	0	1	0.965669	-26.14323	110	1	1.1	0.9;
	1921	2	0	0	0	0	1	1.054316	-0.111079	380	1	1.119837	0.9;
	1922	1	0	0	0.095843	0	1	1.055381	-19.264026	220	1	1.124671	0.9;
	1923	1	529.9	23.3	0	0	1	1.059489	-40.863954	380	1	1.126748	0.9;
	1924	1	0	0	0.127001	0	1	1.026031	-11.849162	220	1	1.123674	0.9;
	1925	1	12.81421	4.4	0	0	1	1.021709	-15.170832	110	1	1.1	0.9;
	1926	1	152.5	-0.1	0	0	1	1.011014	7.733915	220	1	1.1	0.9;
	1927	2	0	0	0	0	1	1.095395	-13.212546	380	1	1.17053	0.9;
	1928	1	5.706326	2	0	0	1	0.955445	-18.925975	110	1	1.1	0.9;
	1929	2	15.81754	4.6	0	0	1	0.964327	-19.364282	110	1	1.1	0.9;
	1930	1	53.35916	7.1	0	0	1	1.019666	-42.352732	220	1	1.1	0.9;
	1931	2	0	0	0	0	1	1.064571	-2.422808	150	1	1.142443	0.9;
	1932	1	0	0	0	0	1	0.978584	-33.230311	220	1	1.1	0.9;
	1933	2	0	0	0	0	1	1.088007	8.208129	150	1	1.143085	0.9;
	1934	1	133.2	24.8	0	0	1	1.091714	-24.951288	380	1	1.163658	0.9;
	1935	1	19.82198	6	0	0	1	0.949197	-28.666154	110	1	1.1	0.9;
	1936	2	8.108991	2.2	0	0	1	1.02989	63.604604	150	1	1.1	0.9;
	1937	1	0	0	0	0	1	1.026461	-23.792213	150	1	1.1	0.9;
	1938	1	92.30234	14.6	0	0	1	1.053932	-47.973212	220	1	1.105769	0.9;
	1939	1	30.33363	11	0	0	1	1.024925	-22.694434	110	1	1.127827	0.9;
	1940	1	79.08768	18.7	0	0	1	1.060279	-22.031555	220	1	1.137848	0.9;
	1941	1	0	0	0	0	1	1.075123	-7.795322	380	1	1.132039	0.9;
	1942	1	34.33807	12.5	0	0	1	1.028507	-23.232421	110	1	1.1	0.9;
	1943	1	16.21798	4	0	0	1	1.036962	-21.794217	150	1	1.113543	0.9;
	1944	1	107.3	14.2	0	0	1	1.077153	-14.825353	220	1	1.140053	0.9;
	1945	1	0	0	0	0	1	1.087274	53.394472	380	1	1.113953	0.9;
	1946	1	0	0	0	0	1	1.086926	-2.729912	220	1	1.146694	0.9;
	1947	1	78.68724	56.7	0	0	1	1.054626	-27.011402	220	1	1.117299	0.9;
	1948	1	137.8	64.5	0	0	1	1.049938	-10.596635	380	1	1.11822	0.9;
	1949	1	0	0	0	0	1	1.042967	-43.075987	220	1	1.1	0.9;
	1950	1	0	0	0	0	1	1.027755	-48.342081	150	1	1.1	0.9;
	1951	1	0	0	0	0	1	1.091649	-7.357152	220	1	1.16003	0.9;
	1952	1	129.5	38.3	0	0	1	1.008764	-39.230306	220	1	1.1	0.9;
	1953	1	0	0	0	0	1	1.073646	-19.168167	220	1	1.110335	0.9;
	1954	1	130.9	37.8	0	0	1	1.025149	-12.980709	220	1	1.1	0.9;
	1955	1	10.81199	1.3	0	0	1	0.977577	-28.892772	110	1	1.1	0.9;
	1956	1	6.106771	0	0	0	1	1.03055	-43.683953	220	1	1.1	0.9;
	1957	1	0	0	0	0	1	1.019284	-23.467665	150	1	1.1	0.9;
	1958	1	17.61954	5.8	0	0	1	0.90687	-31.7183	110	1	1.1	0.9;
	1959	2	0	0	0	0	1	1.099815	20.562167	380	1	1.137041	0.9;
	1960	1	0	0	0.191685	0	1	1.026756	-18.838366	220	1	1.107663	0.9;
	1961	2	708.5	228.6	0	0	1	1.043842	2.775577	380	1	1.1	0.9;
	1962	1	0	0	0.294	0	1	1.062351	9.000362	380	1	1.118138	0.9;
	1963	1	63.67059	20.2	0	0	1	1.065503	-6.84301	220	1	1.133126	0.9;
	1964	1	6.00666	2.2	0	0	1	0.979887	-8.445763	110	1	1.1	0.9;
	1965	1	0	0	0	0	1	1.065796	4.098938	380	1	1.111796	0.9;
	1966	1	0	0	0	0	1	1.000644	-20.374672	150	1	1.1	0.9;
	1967	1	51.95761	32.2	0	0	1	0.914372	-34.680605	150	1	1.1	0.795;
	1968	1	304.2	114.1	0	0	1	1.056021	-42.8691	380	1	1.122178	0.9;
	1969	2	52.15783	2.1	0	0	1	1.041988	-28.387202	220	1	1.130133	0.9;
	1970	1	12.21354	4.4	0	0	1	1.01964	-16.293028	110	1	1.1	0.9;
	1971	2	6.607326	2.4	0	0	1	1.034909	-17.044746	110	1	1.110837	0.9;
	1972	1	15.81754	9.8	0	0	1	1.020632	-23.361959	150	1	1.1	0.9;
	1973	1	20.9232	7.7	0	0	1	1.066551	-13.744298	220	1	1.118803	0.9;
	1974	2	29.93319	9.9	0	0	1	1.046818	-20.97316	110	1	1.108472	0.9;
	1975	2	20.62287	6.2	0	0	1	1.071915	-12.868628	150	1	1.145971	0.9;
	1976	1	0	0	0	0	1	1.073006	-21.8543	220	1	1.121341	0.9;
	1977	1	0	0	0.095843	0	1	1.013955	-23.461466	220	1	1.1	0.9;
	1978	1	114.2	34.4	0	0	1	1.051015	-45.445346	220	1	1.115511	0.9;
	1979	1	396.7	62.8	0	0	1	1.092258	-25.682858	120	1	1.142849	0.9;
	1980	1	0	0	0	0	1	1.069589	-12.689404	380	1	1.129639	0.9;
	1981	1	0	0	0.167999	0	1	1.040114	-18.636043	220	1	1.1	0.9;
	1982	2	-1397.75	-894.7	0.28548	20	1	1.066132	29.576315	380	1	1.124841	0.9;
	1983	2	171.1	49.1	0	0	1	1.038045	-30.374923	220	1	1.116075	0.9;
	1984	1	40.64506	13.2	0	0	1	1.056201	-26.969857	110	1	1.11988	0.9;
	1985	1	0	0	0	0	1	1.071295	-18.022579	150	1	1.1	0.9;
	1986	1	0	0	0	0	1	1.063118	-1.730318	150	1	1.139023	0.9;
	1987	1	0	0	0	0	1	1.044827	-15.453298	380	1	1.11051	0.9;
	1988	1	0	0	0	0	1	0.959599	-25.439559	150	1	1.1	0.9;
	1989	2	0	0	0	0	1	1.076727	-11.542915	220	1	1.132797	0.9;
	1990	2	156.5	31.2	0	0	1	1.096947	-14.681699	380	1	1.151957	0.9;
	1991	1	25.22797	11.9	0	0	1	1.084426	-45.12964	380	1	1.149818	0.9;
	1992	1	0	0	0	0	1	1.076102	-5.913744	220	1	1.135646	0.9;
	1993	2	0	0	0	0	1	1.088696	21.614515	380	1	1.133084	0.9;
	1994	1	0	0	0	0	1	1.064925	-23.116167	220	1	1.122344	0.9;
	1995	2	0	0	0.164823	0	1	1.086015	34.180815	380	1	1.1	0.9;
	1996	2	0	0	0	0	1	1.034636	-43.122784	220	1	1.1	0.9;
	1997	1	0	0	0	0	1	1.02572	-24.774376	150	1	1.1	0.9;
	1998	1	55.46149	18.5	0	0	1	0.934652	-28.440844	110	1	1.1	0.9;
	1999	1	0	0	0	0	1	1.04983	40.168868	220	1	1.1	0.9;
	2000	2	79.2879	40.1	0	0	1	1.080421	-5.808673	380	1	1.144171	0.9;
	2001	1	34.13785	11.4	0	0	1	0.957152	-26.880891	110	1	1.1	0.9;
	2002	1	297.2	93.2	0	0	1	1.0268	-47.663661	380	1	1.1	0.9;
	2003	1	137.3	-2.8	0	0	1	1.056425	-25.124087	220	1	1.113303	0.9;
	2004	1	29.03219	9.6	0	0	1	0.940714	-28.150605	110	1	1.1	0.9;
	2005	1	213	39.6	0	0	1	1.085579	-45.023251	380	1	1.129004	0.9;
	2006	1	0	0	0.260248	0	1	1.067753	-19.463791	380	1	1.114949	0.9;
	2007	1	0	0	0	0	1	1.058277	-43.759067	380	1	1.12617	0.9;
	2008	1	11.01221	3.6	0	0	1	1.000461	-32.232981	110	1	1.1	0.9;
	2009	2	0	0	0	0	1	1.0354	-30.485151	150	1	1.1	0.9;
	2010	2	17.51942	10.4	0	0	1	0.996182	-25.89864	110	1	1.1	0.9;
	2011	1	74.98314	26.9	0	0	1	1.046244	-13.088552	220	1	1.131839	0.9;
	2012	1	6.00666	4.1	0	0	1	0.949007	-24.357908	110	1	1.1	0.9;
	2013	1	0	-7.5	0	0	1	1.03844	-29.258672	150	1	1.1	0.9;
	2014	1	44.54939	16.1	0	0	1	0.938078	-18.334591	110	1	1.1	0.9;
	2015	1	60.0666	45	0	0	1	1.006136	12.990225	220	1	1.1	0.9;
	2016	1	0	0	0	0	1	0.957679	-25.680604	150	1	1.1	0.9;
	2017	2	54.36027	9.3	0	0	1	0.979656	61.937948	150	1	1.1	0.9;
	2018	1	0	0	0	0	1	1.080437	8.707094	380	1	1.157742	0.9;
	2019	1	0	0	0	0	1	0.977956	-26.817106	150	1	1.1	0.9;
	2020	1	0	0	0	0	1	0.986561	-22.383793	110	1	1.1	0.9;
	2021	1	20.32253	6.7	0	0	1	0.953886	-32.191617	110	1	1.1	0.9;
	2022	1	0	0	0	0	1	1.053638	-20.980592	150	1	1.114214	0.9;
	2023	2	24.12675	2.5	0	0	1	0.971857	-22.945461	110	1	1.1	0.9;
	2024	1	0	0	0	0	1	1.097593	57.978487	380	1	1.156784	0.9;
	2025	1	142.8	19.5	0	0	1	1.046385	1.114025	220	1	1.11276	0.9;
	2026	1	78.78735	27.4	0	0	1	1.011746	-47.239744	220	1	1.1	0.9;
	2027	2	88.69834	25.5	0	0	1	1.077	-40.082337	220	1	1.122837	0.9;
	2028	1	145.7	83.6	0.148479	0	1	1.083228	-12.278916	380	1	1.137559	0.9;
	2029	1	0	0	0	31.03723	1	1.040584	-29.176159	150	1	1.1	0.9;
	2030	1	143.1	18.9	0	0	1	1.040171	17.504257	110	1	1.105875	0.9;
	2031	1	-18.8	-5.9	0	0	1	1.01055	-21.646935	150	1	1.1	0.9;
	2032	1	0	0	0	0	1	1.072702	-21.341674	220	1	1.155589	0.9;
	2033	1	0	0	0	0	1	1.080026	1.169577	220	1	1.152774	0.9;
	2034	1	19.52164	11	0	0	1	1.063378	-31.383815	220	1	1.126573	0.9;
	2035	1	269.9	12.4	0	0	1	1.027489	-51.103445	150	1	1.1	0.9;
	2036	1	63.67059	39.5	0	0	1	1.002881	-23.202064	150	1	1.1	0.9;
	2037	1	193.1	69.9	0	0	1	1.061707	-40.976934	380	1	1.123581	0.9;
	2038	2	43.7485	16.7	0	0	1	1.046182	-37.986176	220	1	1.1	0.9;
	2039	2	0	0	0	0	1	1.067455	60.879306	220	1	1.113093	0.9;
	2040	1	0	0	0	0	1	0.976346	-12.024133	110	1	1.1	0.9;
	2041	2	1.301443	1.3	0	0	1	0.991489	-18.913899	110	1	1.1	0.9;
	2042	1	66.87415	21.8	0	0	1	1.051625	2.689291	150	1	1.109132	0.9;
	2043	1	0	0	0	0	1	1.093887	-17.969958	220	1	1.166948	0.9;
	2044	1	0	0	0	0	1	1.054298	-15.936125	150	1	1.109786	0.9;
	2045	1	7.608436	2.5	0	0	1	1.031969	-10.73282	110	1	1.1	0.9;
	2046	2	61.76849	15.6	0	0	1	0.979297	63.077688	150	1	1.1	0.9;
	2047	1	110.3	28.7	0	0	1	1.034561	-42.462607	220	1	1.1	0.9;
	2048	1	1.801998	0.7	0	0	1	1.058269	-20.827296	110	1	1.162201	0.9;
	2049	2	162.9	93.6	0	0	1	0.994045	-5.741575	220	1	1.1	0.9;
	2050	1	25.32808	12.7	0.191685	0	1	1.024626	-20.379787	220	1	1.1	0.9;
	2051	1	100.7	-4.2	0	0	1	1.03886	-47.399903	220	1	1.170118	0.9;
	2052	2	0	0	0	0	1	1.079368	-17.752186	380	1	1.16745	0.9;
	2053	1	0	-8	0	0	1	0.998778	-26.993978	150	1	1.1	0.9;
	2054	1	23.02553	8.3	0	0	1	0.970382	-17.400565	110	1	1.1	0.9;
	2055	1	97	18.7	0	0	1	1.044298	-22.510885	220	1	1.1	0.9;
	2056	2	0	0	0	0	1	1.041035	-11.394598	110	1	1.1	0.9;
	2057	1	5.706326	1.6	0	0	1	0.959068	-24.830209	110	1	1.1	0.9;
	2058	1	305.6	207.3	0	0	1	1.02696	6.048213	380	1	1.1	0.9;
	2059	1	0	0	0	0	1	1.089597	-36.552623	380	1	1.134623	0.9;
	2060	1	0	0	0	0	1	0.992698	-25.319579	150	1	1.1	0.9;
	2061	1	1.902109	0.7	0	0	1	1.005897	-18.266768	110	1	1.1	0.9;
	2062	2	0	0	0.092763	0	1	1.073816	1.973752	380	1	1.124434	0.9;
	2063	1	0	0	0.155924	0	1	1.02407	-5.509624	380	1	1.1	0.9;
	2064	1	0	0	0	0	1	1.044971	-18.038581	220	1	1.1	0.9;
	2065	1	0	0	0	0	1	1.069085	-14.55019	150	1	1.121015	0.9;
	2066	1	0	0	0	0	1	1.062472	9.523157	150	1	1.1	0.9;
	2067	1	0	0	0	0	1	1.087314	-11.084706	380	1	1.13193	0.9;
	2068	1	141.3	32.4	0	0	1	1.076366	-25.457525	220	1	1.166362	0.9;
	2069	2	342.1	61.4	0	0	1	0.968259	-9.027976	220	1	1.1	0.9;
	2070	1	91.90189	5.5	0	0	1	1.003708	-19.881503	110	1	1.1	0.9;
	2071	2	0	0	0	0	1	1.078364	-24.035135	220	1	1.138773	0.9;
	2072	1	0	0	0	0	1	1.078029	51.759076	380	1	1.119208	0.9;
	2073	1	73.18114	22.2	0	0	1	0.98352	-18.280498	110	1	1.1	0.9;
	2074	1	11.61288	5.8	0	0	1	0.983143	-21.252542	110	1	1.1	0.9;
	2075	1	0	0	0	0	1	0.989364	-34.33832	110	1	1.1	0.9;
	2076	1	1.401554	0.5	0	0	1	1.046997	-19.231915	110	1	1.11119	0.9;
	2077	2	100.7	12.8	0	0	1	1.083909	-9.304882	220	1	1.128367	0.9;
	2078	1	0	0	0	0	1	1.071398	63.309558	150	1	1.1	0.9;
	2079	2	209.3	60.8	0	0	1	1.063189	-9.056401	220	1	1.131362	0.9;
	2080	1	170.9	39	0	0	1	1.077787	-7.974825	220	1	1.143922	0.9;
	2081	1	0	0	0	0	1	1.080257	-12.6298	220	1	1.131486	0.9;
	2082	2	0	0	0	0	1	1.078591	-9.946636	220	1	1.154491	0.9;
	2083	1	13.04129	7.4	0	0	1	0.986518	-23.775783	150	1	1.1	0.9;
	2084	1	0	0	0	0	1	1.010168	-23.823969	150	1	1.1	0.9;
	2085	1	71.8797	51.3	0	0	1	1.090114	-12.470077	380	1	1.134823	0.9;
	2086	1	15.01665	5	0	0	1	0.980303	-23.06167	110	1	1.1	0.9;
	2087	1	30.83419	4.4	0	0	1	1.000053	-40.154204	110	1	1.1	0.9;
	2088	1	93.20334	13.6	0	0	1	1.050724	-49.392903	220	1	1.112831	0.9;
	2089	1	402.1	110.2	0	0	1	1.074377	-11.652678	150	1	1.134162	0.9;
	2090	1	0	0	0	0	1	1.053657	-20.14237	220	1	1.110117	0.9;
	2091	1	82.39135	12.3	0	0	1	1.05301	-49.246877	220	1	1.111446	0.9;
	2092	1	628.5	91.8	0	0	1	1.025641	-47.975712	380	1	1.1	0.9;
	2093	1	0	0	0	0	1	1.076191	-14.977423	220	1	1.138194	0.9;
	2094	1	88.69834	-0.1	0	0	1	1.072575	-25.115941	220	1	1.163697	0.9;
	2095	1	21.22353	8.3	0	0	1	1.069023	-0.86792	150	1	1.143949	0.9;
	2096	2	0	0	0	0	1	1.087726	44.129511	380	1	1.116015	0.9;
	2097	1	45.75072	3.3	0	0	1	1.01152	-29.161469	110	1	1.1	0.9;
	2098	1	32.03552	6.6	0	0	1	1.07072	-22.807456	220	1	1.155383	0.9;
	2099	2	0	0	0.248006	302.3515	1	1.092555	23.624753	380	1	1.130054	0.9;
	2100	2	0	0	0	0	1	1.040318	-19.032253	220	1	1.1	0.9;
	2101	1	8.409324	3	0	0	1	0.989029	-25.087295	110	1	1.1	0.9;
	2102	1	0	0	0	0	1	1.017705	-23.44175	150	1	1.1	0.9;
	2103	1	0	0	0	0	1	1.052938	-20.339422	220	1	1.122237	0.9;
	2104	1	110.6	20.4	0	0	1	1.044027	-34.32269	220	1	1.106764	0.9;
	2105	1	264	40.1	0	0	1	1.056943	-33.422296	380	1	1.119188	0.9;
	2106	1	0	0	0	0	1	1.048267	-33.866552	220	1	1.110653	0.9;
	2107	1	12.51387	3.3	0	0	1	0.974048	-27.37002	110	1	1.1	0.9;
	2108	1	39.64395	20.9	0	0	1	1.062292	-8.66485	220	1	1.137549	0.9;
	2109	2	0	0	0	0	1	1.057636	-42.134291	220	1	1.111155	0.9;
	2110	1	8.409324	2.8	0	0	1	0.9536	-34.078413	110	1	1.1	0.9;
	2111	1	0	0	0	0	1	0.951192	63.781	150	1	1.1	0.9;
	2112	1	0	0	0	0	1	1.015454	-19.579353	220	1	1.1	0.9;
	2113	1	24.22686	6.6	0	0	1	1.018623	-35.267457	110	1	1.1	0.9;
	2114	1	114.5	25.2	0	0	1	1.057752	-23.982239	220	1	1.132354	0.9;
	2115	1	0	0	0	0	1	1.062899	-13.937862	220	1	1.13032	0.9;
	2116	1	0	0	0	0	1	1.052785	10.506157	380	1	1.111564	0.9;
	2117	2	0	0	0	0	1	1.04909	-0.355926	150	1	1.124	0.9;
	2118	2	0	0	0	0	1	1.045999	14.03449	110	1	1.122429	0.9;
	2119	1	0	0	0	0	1	1.057704	-30.200666	380	1	1.123901	0.9;
	2120	1	27.93097	6.4	0	0	1	0.955566	-25.634344	110	1	1.1	0.9;
	2121	1	0	0	0	0	1	1.065328	-15.210557	150	1	1.123426	0.9;
	2122	2	0	0	0	0	1	1.086467	5.012024	150	1	1.15359	0.9;
	2123	1	103.4	27	0	0	1	1.026124	-29.245979	220	1	1.132348	0.9;
	2124	1	12.7141	4.9	0	0	1	0.939317	-28.219577	110	1	1.1	0.9;
	2125	1	24.32697	15	0	0	1	1.028044	-17.562426	150	1	1.1	0.9;
	2126	1	96.90745	13.9	0	0	1	1.041598	-12.350501	220	1	1.113128	0.9;
	2127	1	0	0	0	0	1	1.089761	-25.250235	220	1	1.182235	0.9;
	2128	1	0	0	0	0	1	1.095501	56.949995	380	1	1.115269	0.9;
	2129	1	33.03663	5.6	0	0	1	1.054334	-37.933355	220	1	1.106239	0.9;
	2130	1	0	0	0	0	1	1.10272	-17.259279	380	1	1.176031	0.9;
	2131	1	0	0	0	0	1	0.993392	-22.071097	220	1	1.1	0.9;
	2132	1	0	0	0	0	1	1.009679	-22.254751	150	1	1.1	0.9;
	2133	1	33.23685	11.1	0	0	1	0.899791	-32.618439	110	1	1.1	0.894909;
	2134	1	0	0	0.119999	0	1	0.993519	-21.944974	220	1	1.1	0.9;
	2135	1	0	0	0	0	1	1.039169	14.18306	150	1	1.1	0.9;
	2136	2	84.89413	6.1	0	0	1	1.008533	62.437332	150	1	1.1	0.9;
	2137	1	4.905439	1.4	0	0	1	0.938577	-25.369756	110	1	1.1	0.9;
	2138	1	75.78402	46.9	0	0	1	1.031963	-24.998566	150	1	1.1	0.9;
	2139	1	54.46038	11.9	0	0	1	1.055231	-46.495758	220	1	1.1	0.9;
	2140	1	0	0	0	11.03517	1	1.042876	-23.441963	150	1	1.1	0.9;
	2141	2	225.9	50	0	0	1	1.053447	-2.781342	380	1	1.126175	0.9;
	2142	1	52.95872	8.2	0	0	1	1.054552	-49.268703	220	1	1.112497	0.9;
	2143	2	0	0	0	0	1	1.076342	-9.586923	380	1	1.13957	0.9;
	2144	1	215.5	31.3	0	0	1	1.02057	-20.659148	220	1	1.130982	0.9;
	2145	2	23.32586	6.9	0	0	1	1.054551	-28.77905	220	1	1.151783	0.9;
	2146	2	0	0	0	0	1	1.100553	-13.119044	380	1	1.153901	0.9;
	2147	1	0	0	0	0	1	1.018153	-19.904768	220	1	1.1	0.9;
	2148	1	0	0	0	0	1	1.015472	10.409059	220	1	1.1	0.9;
	2149	1	37.94207	15.4	0	0	1	1.035189	-21.383815	380	1	1.143145	0.9;
	2150	1	0	0	0	0	1	1.026987	-46.297924	220	1	1.1	0.9;
	2151	1	-15.9176	5.2	0	0	1	1.041498	-21.661425	150	1	1.1	0.9;
	2152	1	178.5	65.8	0	0	1	1.096733	-24.680638	380	1	1.147285	0.9;
	2153	2	0	0	0	0	1	1.092955	-13.002888	220	1	1.13468	0.9;
	2154	1	0	0	0	0	1	1.054499	-40.556652	380	1	1.113311	0.9;
	2155	1	0	0	0	0	1	1.038889	-16.634555	380	1	1.117171	0.9;
	2156	1	0	0	0	0	1	0.968059	-24.775992	150	1	1.1	0.9;
	2157	1	17.11898	3.8	0	0	1	0.972378	-23.069382	110	1	1.1	0.9;
	2158	1	101.7	17.6	0	0	1	1.060449	-14.254126	220	1	1.120938	0.9;
	2159	1	-41.2457	13.5	0	0	1	1.036006	-18.6289	150	1	1.1	0.9;
	2160	1	1.401554	0.5	0	0	1	1.042905	-19.845059	110	1	1.1	0.9;
	2161	1	0	0	0	0	1	1.078039	-26.418161	220	1	1.132765	0.9;
	2162	1	32.93652	13	0	0	1	1.080446	3.749758	150	1	1.149146	0.9;
	2163	1	0	0	0	0	1	1.016684	-26.907965	150	1	1.1	0.9;
	2164	1	52.25794	44.2	0.191685	0	1	1.073295	-12.653118	220	1	1.129806	0.9;
	2165	1	0	0	0	0	1	1.073759	-34.560246	220	1	1.131523	0.9;
	2166	1	122.1	16.7	0	0	1	1.06667	-7.830651	380	1	1.130693	0.9;
	2167	1	24.12675	8.7	0.079606	0	1	1.060681	-16.705205	220	1	1.112927	0.9;
	2168	1	0.100111	1	0	0	1	0.995346	10.385073	220	1	1.1	0.9;
	2169	1	57.56382	24	0	0	1	1.054119	-36.431516	220	1	1.135522	0.9;
	2170	1	0	0	0	0	1	1.067505	-26.004209	220	1	1.108908	0.9;
	2171	1	0	0	0	0	1	0.99091	10.896018	220	1	1.1	0.9;
	2172	1	20.12231	12.5	0	0	1	1.044929	-22.909941	150	1	1.1	0.9;
	2173	1	173.7	23	0	0	1	1.046294	-23.139248	220	1	1.1	0.9;
	2174	1	20.82309	0	0	0	1	1.044754	8.006832	220	1	1.118123	0.9;
	2175	2	200.5	85	0	0	1	1.094947	56.244595	380	1	1.114943	0.9;
	2176	1	190.7	13.3	0	0	1	1.008125	-46.806456	220	1	1.1	0.9;
	2177	1	0	0	0	0	1	1.03077	-24.170177	150	1	1.1	0.9;
	2178	2	75.18336	50.1	0.26925	0	1	1.079237	-17.780998	380	1	1.165988	0.9;
	2179	1	69.37692	43	0	0	1	0.970908	-27.421821	150	1	1.1	0.9;
	2180	1	0	0	0	0	1	0.981817	-25.683243	150	1	1.1	0.9;
	2181	2	0	0	0	0	1	1.054682	-14.495161	220	1	1.122144	0.9;
	2182	1	37.24129	5.3	0	0	1	1.065728	-43.361953	220	1	1.115882	0.9;
	2183	1	0	0	0	0	1	0.915008	-31.362086	110	1	1.1	0.9;
	2184	1	0	0	0	0	1	1.059155	-23.290164	220	1	1.133323	0.9;
	2185	1	117.1299	52.4	0	0	1	1.015477	-35.606541	150	1	1.1	0.9;
	2186	1	64.8	99.1	0	0	1	1.043402	11.109021	380	1	1.1	0.9;
	2187	1	3.904329	-35.8	0	0	1	1.046707	-32.895073	220	1	1.105078	0.9;
	2188	1	0	0	0	0	1	0.985667	62.461818	150	1	1.1	0.9;
	2189	2	0	0	0	0	1	1.060091	-38.456982	220	1	1.1	0.9;
	2190	2	93.9	51.9	0	0	1	0.992136	10.925026	220	1	1.1	0.9;
	2191	1	5.506105	2	0	0	1	1.026886	-21.611337	110	1	1.12027	0.9;
	2192	1	14.31587	5.6	0	0	1	1.091958	-44.836341	380	1	1.14423	0.9;
	2193	2	0	0	0	0	1	1.006357	-14.652376	110	1	1.1	0.9;
	2194	1	15.9	116.3	0	0	1	1.070919	29.222478	220	1	1.108308	0.9;
	2195	1	0	0	0.253999	0	1	1.001136	-9.854836	220	1	1.1	0.9;
	2196	1	3.403774	123.8	0	0	1	1.061332	-16.6139	380	1	1.125568	0.9;
	2197	2	373.8	55.6	0	0	1	1.076909	-22.485013	220	1	1.16229	0.9;
	2198	1	39.34362	24.3	0	0	1	0.976226	-23.965635	150	1	1.1	0.9;
	2199	1	523.3	137.2	0	0	1	1.057714	-44.216629	380	1	1.125706	0.9;
	2200	1	44.54939	16.8	0	0	1	1.058363	-21.414583	220	1	1.135638	0.9;
	2201	2	197.4	52.4	0	0	1	1.073773	-9.652838	220	1	1.144532	0.9;
	2202	2	119.4	23.6	0	0	1	1.083545	-12.439283	220	1	1.137689	0.9;
	2203	1	0	0	0.296959	0	1	1.08198	-17.22128	380	1	1.152062	0.9;
	2204	1	0	0	0	0	1	1.071211	0.340339	150	1	1.135543	0.9;
	2205	2	99.4	35.8	0.265353	0	1	1.047289	13.686595	380	1	1.108665	0.9;
	2206	1	0	0	0	0	1	1.072182	-11.456445	220	1	1.134118	0.9;
	2207	1	0	0	0	0	1	1.026409	-24.557145	150	1	1.1	0.9;
	2208	1	291.2	35.4	0	0	1	1.035083	-32.148732	220	1	1.13434	0.9;
	2209	1	0	0	0	0	1	1.016587	8.208956	220	1	1.1	0.9;
	2210	1	0	0	0	0	1	1.022213	-21.234259	150	1	1.1	0.9;
	2211	1	0	0	0	0	1	1.080554	-17.689079	380	1	1.164319	0.9;
	2212	1	0	-8.6	0	0	1	1.088162	-37.799003	380	1	1.148816	0.9;
	2213	1	0	0	0	0	1	1.05291	-20.339111	220	1	1.122207	0.9;
	2214	1	0	0	0	0	1	1.042459	24.753832	220	1	1.1	0.9;
	2215	1	0	0	0	0	1	1.031298	4.49255	380	1	1.1	0.9;
	2216	1	0	0	0.222985	0	1	1.020348	-5.108166	380	1	1.1	0.9;
	2217	1	45.6	142	0	0	1	1.00834	9.675065	220	1	1.1	0.9;
	2218	1	0	0	0	0	1	1.066908	8.106528	380	1	1.126928	0.9;
	2219	1	0	0	0	0	1	1.028176	-24.446012	150	1	1.1	0.9;
	2220	2	0	0	0	0	1	1.104211	5.705528	380	1	1.147155	0.9;
	2221	1	34.23796	16.5	0	0	1	1.079137	20.553752	220	1	1.163682	0.9;
	2222	1	301.9	-103.2	0	0	1	1.088357	-8.448565	150	1	1.143522	0.9;
	2223	1	0	0	0	0	1	1.06079	-14.538737	150	1	1.110945	0.9;
	2224	2	189.1	38.8	0	0	1	1.049729	-1.189361	220	1	1.121341	0.9;
	2225	1	0	0	0	0	1	0.988074	-23.675673	150	1	1.1	0.9;
	2226	1	262.5	18.9	0	0	1	1.096173	-17.274095	380	1	1.182068	0.9;
	2227	1	30.73408	9.6	0	0	1	1.072019	-23.402566	220	1	1.136582	0.9;
	2228	1	132	26	0	0	1	1.040296	-12.941236	380	1	1.122615	0.9;
	2229	1	0	0	0	0	1	1.091262	-38.562846	380	1	1.153671	0.9;
	2230	2	6.807548	23.8	0	0	1	0.992364	-35.832036	220	1	1.1	0.9;
	2231	1	3.303663	1.9	0	0	1	1.061356	-4.386181	110	1	1.136538	0.9;
	2232	2	4.605106	0	0	0	1	1.013914	-22.616618	110	1	1.1	0.9;
	2233	1	0	0	0	0	1	1.09269	45.83635	380	1	1.1	0.9;
	2234	1	1.501665	0.4	0	0	1	0.972971	-28.821802	110	1	1.1	0.9;
	2235	2	324.4	93	0	0	1	1.048703	1.983533	380	1	1.106368	0.9;
	2236	1	0	0.5	0	0	1	0.924151	-28.491213	110	1	1.1	0.9;
	2237	1	0	0	0	0	1	1.067557	-13.866047	220	1	1.135555	0.9;
	2238	2	0	0	0	0	1	1.076591	-5.851285	220	1	1.136074	0.9;
	2239	1	0	0	0	0	1	1.008359	-23.241991	150	1	1.1	0.9;
	2240	1	0	0	0	0	1	1.056047	-18.952426	150	1	1.1	0.9;
	2241	1	103.5	21.7	0	0	1	1.049813	-43.07985	220	1	1.105252	0.9;
	2242	1	76.58492	14	0	0	1	0.983459	6.557129	220	1	1.1	0.9;
	2243	1	40.44484	12.8	0	0	1	0.906226	-31.809081	110	1	1.1	0.9;
	2244	1	35.03885	16	0	0	1	1.017724	-21.402252	220	1	1.1	0.9;
	2245	1	0	0	0	0	1	1.073321	-7.553008	220	1	1.143575	0.9;
	2246	1	0	0	0	0	1	0.987583	-17.285401	150	1	1.1	0.9;
	2247	1	50.9565	6.8	0	0	1	1.06807	10.768466	110	1	1.147536	0.9;
	2248	2	0	0	0	0	1	1.081245	-2.535791	380	1	1.13594	0.9;
	2249	1	0	0	0	0	1	1.006764	-32.247252	220	1	1.1	0.9;
	2250	2	0	0	0	0	1	1.057	-16.980392	220	1	1.1098	0.9;
	2251	2	0	0	0	0	1	1.060382	7.031571	150	1	1.112077	0.9;
	2252	1	0	0	0	0	1	1.054018	-41.348618	220	1	1.113232	0.9;
	2253	1	79.08768	-36.9	0	0	1	1.027776	12.426607	380	1	1.1	0.9;
	2254	1	0	0	0	0	1	1.07702	-18.39735	220	1	1.116241	0.9;
	2255	1	0	0	0	0	1	0.981983	-21.613426	110	1	1.1	0.9;
	2256	2	0	0	0	0	1	1.08	1.16987	220	1	1.152746	0.9;
	2257	1	0	0	0	0	1	0.971801	-18.027773	220	1	1.1	0.9;
	2258	1	74.18224	22.6	0	0	1	1.074673	-9.407892	220	1	1.144128	0.9;
	2259	1	0	0	0	0	1	1.058132	0.522482	150	1	1.119723	0.9;
	2260	1	0	0	0	0	1	1.056936	-21.92114	220	1	1.1	0.9;
	2261	1	50.55605	4.6	0	0	1	1.056622	-54.885101	220	1	1.1	0.9;
	2262	1	0	0	0	0	1	1.023141	-51.331437	220	1	1.1	0.9;
	2263	1	87.69723	22.2	0	0	1	1.074546	-14.841909	220	1	1.12069	0.9;
	2264	1	0	0	0	0	1	1.090587	-20.138786	380	1	1.172136	0.9;
	2265	2	0	0	0	0	1	1.091591	-20.935857	220	1	1.128713	0.9;
	2266	1	0	0	0	0	1	1.093457	-33.851876	380	1	1.141028	0.9;
	2267	1	0	0	0	0	1	1.069577	-20.208957	220	1	1.151985	0.9;
	2268	1	119.2	32.1	0	0	1	1.045017	-41.320233	220	1	1.1	0.9;
	2269	1	48.75406	16.2	0	0	1	0.930409	-30.781398	110	1	1.1	0.9;
	2270	2	0	0	0	0	1	0.969104	-17.853133	110	1	1.1	0.9;
	2271	1	0	0	0	0	1	1.103085	-37.690579	380	1	1.15036	0.9;
	2272	2	133.7	10.4	0	0	1	1.081045	-12.556413	220	1	1.119268	0.9;
	2273	2	142.7	34.8	0	0	1	1.092711	-11.685387	220	1	1.148988	0.9;
	2274	1	174.1	-519.7	0	0	1	1.071675	17.273174	380	1	1.141144	0.9;
	2275	1	0	0	0	0	1	1.091875	-30.038378	380	1	1.163561	0.9;
	2276	1	0	0	0	0	1	1.088424	36.303029	380	1	1.129694	0.9;
	2277	2	0	0	0	0	1	0.981636	-21.690715	110	1	1.1	0.9;
	2278	1	0	0	0	0	1	1.062392	-11.155914	220	1	1.1244	0.9;
	2279	2	42.24684	12.2	0	0	1	1.022273	-38.662487	220	1	1.1	0.9;
	2280	2	0	0	0	0	1	1.062122	-10.748993	380	1	1.123296	0.9;
	2281	2	89.79956	5.9	0	0	1	0.997533	66.207202	150	1	1.1	0.9;
	2282	1	0	0	0.047488	0	1	1.02642	-5.405739	380	1	1.1	0.9;
	2283	1	59.7	26.3	0	0	1	1.07116	-16.894665	380	1	1.116379	0.9;
	2284	1	0	0	0	0	1	1.078703	20.439743	220	1	1.161146	0.9;
	2285	1	18.32031	14.1	0	0	1	1.051659	-21.501573	150	1	1.1	0.9;
	2286	2	275.1	63	0	0	1	1.040521	-19.190396	380	1	1.150101	0.9;
	2287	1	59.76627	16.7	0	0	1	1.074655	-4.617535	220	1	1.128286	0.9;
	2288	1	-10.4086	-3.3	0	0	1	1.054099	-15.821508	150	1	1.109604	0.9;
	2289	1	0	0	0	0	1	0.985991	-23.580828	150	1	1.1	0.9;
	2290	1	77.3858	20.7	0	0	1	1.014477	-28.194121	110	1	1.1	0.9;
	2291	1	1.301443	0	0	0	1	0.991598	-18.795605	110	1	1.1	0.9;
	2292	1	0	0	0.077386	0	1	1.07955	46.363175	380	1	1.1	0.9;
	2293	2	0	0	0.196238	0	1	1.060318	-17.527667	220	1	1.127079	0.9;
	2294	1	51.55716	2.1	0	0	1	1.081562	-3.877074	220	1	1.142591	0.9;
	2295	1	15.11676	9.4	0	0	1	0.966203	-24.767487	150	1	1.1	0.9;
	2296	1	70.27792	16.2	0	0	1	1.011593	-22.564707	110	1	1.1	0.9;
	2297	1	118.8	39.5	0	0	1	1.021201	-39.923649	220	1	1.1	0.9;
	2298	2	177.4	85.5	0	0	1	0.994318	-5.766864	220	1	1.1	0.9;
	2299	1	278.3481	90.4	0	0	1	1.035779	-30.835079	220	1	1.1	0.9;
	2300	2	0	0	0	0	1	1.050067	-14.969538	150	1	1.1	0.9;
	2301	1	72.68058	45	0	0	1	1.020342	-23.553817	150	1	1.1	0.9;
	2302	1	38.44262	13.2	0	0	1	1.006474	-33.97354	110	1	1.1	0.9;
	2303	1	27.3303	5	0	0	1	1.06761	62.066032	150	1	1.1	0.9;
	2304	1	0	0	0	0	1	1.063756	19.876535	380	1	1.126878	0.9;
	2305	1	0	0	0	0	1	1.020065	-19.450214	110	1	1.1	0.9;
	2306	1	-4.87602	5.5	0	0	1	1.040517	-29.17268	150	1	1.1	0.9;
	2307	1	0	0	0	0	1	1.086074	-4.406955	220	1	1.139941	0.9;
	2308	2	0	0	0	0	1	1.078227	-6.3027	220	1	1.136773	0.9;
	2309	1	205.3	27.8	0	0	1	1.056093	-55.236043	220	1	1.1	0.9;
	2310	1	28.53163	11.3	0	0	1	1.078237	3.509968	150	1	1.146365	0.9;
	2311	2	92.50256	-23.8	0	0	1	1.09233	14.73935	220	1	1.17156	0.9;
	2312	1	32.53607	10.6	0	0	1	0.901131	-33.225217	110	1	1.1	0.9;
	2313	2	0	0	0	0	1	1.077179	9.472667	110	1	1.146118	0.9;
	2314	1	0	0	0	0	1	1.090785	-17.620574	220	1	1.12808	0.9;
	2315	1	0	-0.2	0	0	1	1.036883	-11.870091	110	1	1.1	0.9;
	2316	2	160.6	39.6	0	0	1	1.059808	-16.263455	220	1	1.133381	0.9;
	2317	1	17.01887	5.9	0	0	1	1.031301	-20.04868	110	1	1.1	0.9;
	2318	1	6.507215	1.4	0	0	1	0.958173	-20.304896	110	1	1.1	0.9;
	2319	1	1.801998	0.7	0	0	1	1.02276	-23.603719	110	1	1.122935	0.9;
	2320	1	-88.5982	-24.2	0	0	1	1.038682	-47.107197	220	1	1.1	0.9;
	2321	1	52.8586	13	0	0	1	1.081181	-4.685517	220	1	1.135227	0.9;
	2322	1	0	0	0	0	1	1.056555	-48.655971	220	1	1.114504	0.9;
	2323	1	0	0	0	0	1	0.971763	-10.064971	110	1	1.1	0.9;
	2324	1	0	0	0	0	1	1.074045	-40.819919	380	1	1.142419	0.9;
	2325	1	25.42819	7.1	0	0	1	0.999011	-23.600239	110	1	1.1	0.9;
	2326	2	78.58714	21.4	0	0	1	0.968364	-25.426775	110	1	1.1	0.9;
	2327	2	42.24684	10.6	0	0	1	1.055933	62.747132	150	1	1.1	0.9;
	2328	2	0	0	0	0	1	1.069898	-39.485725	220	1	1.128888	0.9;
	2329	1	0	0	0	0	1	1.014355	-21.995068	150	1	1.1	0.9;
	2330	2	0	0	0	0	1	1.098133	15.030212	150	1	1.137378	0.9;
	2331	2	255.7	67.4	0	0	1	1.015822	-2.555921	220	1	1.106134	0.9;
	2332	1	100.1	44.5	0	0	1	1.012744	-47.53743	220	1	1.1	0.9;
	2333	1	62.96982	17.4	0	0	1	1.080736	-3.707203	220	1	1.134236	0.9;
	2334	1	27.05028	22.7	0	0	1	1.044648	-22.312796	150	1	1.1	0.9;
	2335	1	0	0	0	0	1	1.068228	-32.197744	380	1	1.128998	0.9;
	2336	1	0	0	0	0	1	1.013944	-47.414081	220	1	1.1	0.9;
	2337	1	0	0	0	0	1	1.074097	-27.071588	110	1	1.139614	0.9;
	2338	1	0	0	0	0	1	0.957306	-29.199935	150	1	1.1	0.892694;
	2339	1	-15.7174	5.2	0	0	1	0.977884	-25.293762	150	1	1.1	0.9;
	2340	1	2.302553	0.8	0	0	1	1.050851	-18.185736	110	1	1.132961	0.9;
	2341	1	174.6	48.9	0	0	1	1.04901	-13.111468	380	1	1.117566	0.9;
	2342	1	0	0	0.119668	0	1	1.082732	-22.618066	380	1	1.15148	0.9;
	2343	1	0	0	0	0	1	1.067206	-14.596449	150	1	1.119202	0.9;
	2344	1	39.04329	24.2	0	0	1	1.032594	-25.572664	150	1	1.1	0.9;
	2345	2	235.1	54.6	0	0	1	1.027061	-13.176669	220	1	1.1	0.9;
	2346	2	0	0	0	0	1	1.055579	-14.053673	380	1	1.123961	0.9;
	2347	1	0	0	0	0	1	1.018458	4.127009	220	1	1.1	0.9;
	2348	2	254.2	84.6	0	0	1	1.063811	-15.015251	220	1	1.136494	0.9;
	2349	1	28.13119	14.2	0	0	1	0.976584	-20.308463	110	1	1.1	0.9;
	2350	1	-9.51054	3.1	0	0	1	1.042248	-20.373521	150	1	1.1	0.9;
	2351	2	0	0	0	0	1	1.096818	-26.364211	220	1	1.155714	0.9;
	2352	2	0	0	0	0	1	1.048921	14.430459	380	1	1.113244	0.9;
	2353	1	0	0	0	0	1	1.044172	-20.625952	150	1	1.1	0.9;
	2354	1	14.71632	5.4	0	0	1	1.073925	-10.106196	220	1	1.144374	0.9;
	2355	1	25.62842	12.7	0	0	1	1.072909	-14.34941	220	1	1.136215	0.9;
	2356	2	16.31809	3	0	0	1	1.084818	-12.332319	220	1	1.138824	0.9;
	2357	1	0	0	0	0	1	0.970522	-33.469219	220	1	1.1	0.9;
	2358	1	-23.1256	-8.4	0	0	1	1.051537	-21.258241	110	1	1.112303	0.9;
	2359	1	0	0	0	0	1	1.027092	-27.308585	150	1	1.1	0.9;
	2360	2	234.3	105.9	0	0	1	0.978909	14.712342	220	1	1.1	0.9;
	2361	1	115.3	-8.8	0	0	1	1.081696	-9.262047	220	1	1.13406	0.9;
	2362	2	0	0	0	0	1	1.035727	13.320365	220	1	1.1	0.9;
	2363	1	0	0	0	0	1	1.098549	-17.869836	380	1	1.144768	0.9;
	2364	1	226.8002	53.2	0	0	1	0.961377	20.810354	220	1	1.1	0.9;
	2365	1	14.91654	4.4	0	0	1	0.955456	-24.186037	110	1	1.1	0.9;
	2366	1	499.8002	110.3	0	0	1	1.015288	10.043328	220	1	1.1	0.9;
	2367	1	0	0	0	0	1	1.066435	-14.560224	380	1	1.117858	0.9;
	2368	1	0	0	0	-16.763	1	1.030682	-25.99429	150	1	1.159711	0.9;
	2369	1	0	0	0	0	1	0.973725	-33.01629	220	1	1.1	0.9;
	2370	1	0	0	0	0	1	1.028285	-6.864028	220	1	1.1	0.9;
	2371	1	0	0	0	0	1	1.049288	-42.336723	380	1	1.108307	0.9;
	2372	1	0	0	0	0	1	1.081445	3.955879	150	1	1.149481	0.9;
	2373	1	14.73807	14.7	0	0	1	1.023768	-23.061171	150	1	1.1	0.9;
	2374	1	0	0	0	0	1	1.054723	4.976814	150	1	1.109184	0.9;
	2375	2	0	0	0	0	1	1.031455	-26.976837	220	1	1.1	0.9;
	2376	1	0	0	0	0	1	1.086761	-45.147293	380	1	1.142854	0.9;
	2377	2	0	0	0	0	1	1.084136	-22.4956	220	1	1.122697	0.9;
	2378	1	10.61177	3.5	0	0	1	1.007946	-29.499638	110	1	1.1	0.9;
	2379	1	5.506105	2	0	0	1	1.024131	-16.146729	110	1	1.121942	0.9;
	2380	2	0	0	0	0	1	1.103421	5.677433	380	1	1.147005	0.9;
	2381	1	0	0	0	0	1	0.992518	-25.013232	150	1	1.1	0.9;
	2382	1	203.1	65.2	0	0	1	1.055989	-23.352663	220	1	1.130165	0.9;
	2383	1	166.7	87.6	0	0	1	1.059152	-11.813021	380	1	1.126011	0.9;
	2384	1	29.93319	-4.2	0	0	1	0.991371	37.899365	220	1	1.1	0.9;
	2385	1	655.8	72.8	0	0	1	1.071091	-8.839585	380	1	1.133347	0.9;
	2386	1	3.704107	1.3	0	0	1	0.99888	-19.443847	110	1	1.1	0.9;
	2387	1	0	0	0	0	1	1.051905	-35.442194	380	1	1.117641	0.9;
	2388	1	42.94762	26.6	0	0	1	0.937534	-33.294541	150	1	1.1	0.821661;
	2389	1	98.3	25.9	0	0	1	1.01153	-45.80468	220	1	1.1	0.9;
	2390	1	33.03663	12.1	0	0	1	1.005288	-35.676965	110	1	1.1	0.9;
	2391	1	21.72409	7.2	0	0	1	0.934116	-27.941655	110	1	1.1	0.9;
	2392	1	45.04995	-4.5	0	0	1	1.026373	-22.118533	110	1	1.1	0.9;
	2393	1	0	0	0	0	1	1.04862	-11.454829	220	1	1.118559	0.9;
	2394	2	87.8	492.1	0	0	1	1.079184	-33.296642	380	1	1.131239	0.9;
	2395	1	2.00222	0.5	0	0	1	0.995621	-26.339422	110	1	1.1	0.9;
	2396	1	0	0	0	0	1	1.075653	-23.932013	220	1	1.163898	0.9;
	2397	1	0	0	0	0	1	1.005622	-21.842862	110	1	1.1	0.9;
	2398	1	123.1	3.9	0	0	1	1.097822	-10.109143	380	1	1.143432	0.9;
	2399	1	70.1778	30.5	0	0	1	1.03006	3.943332	220	1	1.1	0.9;
	2400	1	46.65172	15.1	0	0	1	1.069955	8.085073	110	1	1.139405	0.9;
	2401	2	0	0	0	0	1	1.099211	-31.349972	380	1	1.145824	0.9;
	2402	2	0	0	0	0	1	1.078591	-35.269155	220	1	1.157551	0.9;
	2403	1	0	0	0	0	1	0.986958	-23.357532	110	1	1.1	0.9;
	2404	1	93.1	43.7	0	0	1	1.032502	-25.294349	150	1	1.1	0.9;
	2405	2	71.47926	24	0	0	1	1.060818	-21.979124	220	1	1.138338	0.9;
	2406	1	207.1	50.4	0	0	1	1.054649	-17.597823	220	1	1.119321	0.9;
	2407	1	0	0	0	0	1	0.995482	10.384398	220	1	1.1	0.9;
	2408	2	0	0	0	0	1	1.051348	-23.740727	380	1	1.122029	0.9;
	2409	1	0	0	0	0	1	0.999561	54.111164	220	1	1.1	0.9;
	2410	2	-10.812	35.4	0	0	1	1.048773	-44.801132	220	1	1.108449	0.9;
	2411	2	0	0	0	0	1	1.032533	-28.038967	150	1	1.1	0.9;
	2412	1	45.45039	28.1	0	0	1	1.024125	-25.664124	150	1	1.1	0.9;
	2413	2	62.66948	7.7	0	0	1	1.041933	60.744998	150	1	1.1	0.9;
	2414	1	150	104.3	0	0	1	1.044496	-21.184983	220	1	1.1	0.9;
	2415	2	7.207992	4.4	0	0	1	1.0764	-17.693338	150	1	1.1	0.9;
	2416	1	0	0	0	0	1	1.000498	11.102933	220	1	1.1	0.9;
	2417	1	70.67836	10.7	0	0	1	1.057501	-48.543217	220	1	1.114947	0.9;
	2418	1	252.9	95.6	0	0	1	1.017585	-40.196255	220	1	1.1	0.9;
	2419	1	223.5	48	0	0	1	1.030445	-46.540949	380	1	1.1	0.9;
	2420	1	20.9232	1.1	0	0	1	1.010346	-38.793036	110	1	1.1	0.9;
	2421	1	16.81865	5.5	0	0	1	1.005682	-29.681714	110	1	1.1	0.9;
	2422	2	7.908769	2.2	0	0	1	1.028639	-20.274014	110	1	1.1	0.9;
	2423	1	13.21465	4.4	0	0	1	0.932986	-35.426189	110	1	1.1	0.9;
	2424	1	0	0	0	0	1	1.030726	-24.276189	150	1	1.1	0.9;
	2425	1	0	0	0	0	1	1.062587	-39.952465	220	1	1.131439	0.9;
	2426	2	139.2	29	0	0	1	1.077784	-14.304801	220	1	1.146681	0.9;
	2427	1	0	0	0.238001	0	1	1.019993	-15.020372	220	1	1.108306	0.9;
	2428	1	10.61177	3.8	0	0	1	1.061033	-20.537394	110	1	1.121169	0.9;
	2429	1	27.93097	10.1	0	0	1	1.04124	-17.703726	110	1	1.125412	0.9;
	2430	1	0	0	0	0	1	1.079494	-35.172571	380	1	1.154696	0.9;
	2431	1	0	0	0	0	1	1.084389	39.260819	380	1	1.1	0.9;
	2432	1	0	0	0	0	1	1.030342	-27.363915	220	1	1.1	0.9;
	2433	1	394.9	45.9	0	0	1	1.049205	-35.250303	380	1	1.113489	0.9;
	2434	2	0	0	0	0	1	1.022955	-42.119894	220	1	1.1	0.9;
	2435	2	107.4	37.3	0	0	1	1.060818	-38.492625	220	1	1.1	0.9;
	2436	1	2.00222	1	0	0	1	0.973835	-27.369538	110	1	1.1	0.9;
	2437	1	0	0	0	0	1	0.9393	-29.246523	150	1	1.1	0.890626;
	2438	1	0	0	0	0	1	1.078199	-4.168971	220	1	1.139434	0.9;
	2439	1	0	0	0	0	1	1.070077	-18.776249	150	1	1.1	0.9;
	2440	2	13.91543	2.1	0	0	1	1.002973	-23.335056	110	1	1.1	0.9;
	2441	1	285.2	65.3	0	0	1	1.028894	-49.623367	220	1	1.1	0.9;
	2442	1	681.5	57.7	0	0	1	1.057948	-43.217344	380	1	1.127463	0.9;
	2443	1	16.21798	8.9	0	0	1	1.07857	-19.996963	220	1	1.141462	0.9;
	2444	1	0	0	0	0	1	0.998154	-23.43324	150	1	1.1	0.9;
	2445	1	0	0	0	0	1	0.958441	-25.538445	150	1	1.1	0.9;
	2446	1	0	-1.2	0	0	1	1.054888	20.017056	110	1	1.111135	0.9;
	2447	1	23.12564	8.8	0	0	1	1.03943	-29.43462	110	1	1.1	0.9;
	2448	1	0	0	0	0	1	1.070988	-35.466171	380	1	1.121727	0.9;
	2449	1	0	0	0	18.60544	1	1.036914	-29.382505	150	1	1.1	0.9;
	2450	1	151.7	27.6	0	0	1	1.053711	-24.448086	220	1	1.128319	0.9;
	2451	2	0	0	0	0	1	1.083674	7.639697	150	1	1.138236	0.9;
	2452	1	56.26238	18.5	0	0	1	0.934839	-34.45066	110	1	1.1	0.9;
	2453	1	117.4	12.1	0	0	1	1.012363	10.759849	220	1	1.1	0.9;
	2454	1	118	37.8	0	0	1	1.014396	-14.609682	110	1	1.1	0.9;
	2455	1	0	0	0.090703	0	1	1.080538	-14.284529	380	1	1.142016	0.9;
	2456	1	47.75294	18.8	0	0	1	1.062791	-1.720063	150	1	1.138705	0.9;
	2457	1	189.7	52	0	0	1	1.056412	-44.268937	380	1	1.125026	0.9;
	2458	2	349.6	96.7	0	20	1	1.064579	-31.085511	380	1	1.125032	0.9;
	2459	1	104.8	22	0	0	1	1.014512	7.65489	220	1	1.1	0.9;
	2460	1	11.01221	4	0	0	1	0.983202	-15.359246	110	1	1.1	0.9;
	2461	1	0	8.4	0	0	1	1.03829	-21.045165	150	1	1.1	0.9;
	2462	1	90.50034	27.8	0	0	1	1.074796	-15.265311	220	1	1.119296	0.9;
	2463	1	0	0	0.319997	0	1	1.061363	-21.532162	380	1	1.114768	0.9;
	2464	1	32.03552	11	0	0	1	0.999225	-18.929738	110	1	1.1	0.9;
	2465	2	0	0	0	0	1	1.073773	-13.96315	220	1	1.129672	0.9;
	2466	1	13.01443	14.2	0	0	1	1.048128	15.41589	150	1	1.1	0.9;
	2467	1	0	0	0	0	1	0.988355	-22.293577	380	1	1.1	0.9;
	2468	1	0	0	0	0	1	1.044563	7.422561	380	1	1.1	0.9;
	2469	1	3.804218	0.6	0	0	1	0.973274	-9.656202	110	1	1.1	0.9;
	2470	1	151.2	81.3	0	0	1	1.063721	9.7985	380	1	1.12327	0.9;
	2471	2	43.34806	26.9	0	0	1	1.0334	-25.161622	150	1	1.1	0.9;
	2472	1	0	0	0	0	1	1.038064	17.870194	380	1	1.1	0.9;
	2473	1	32.53607	11.7	0	0	1	0.984835	-17.034133	110	1	1.1	0.9;
	2474	2	0	0	0	0	1	1.07923	20.516073	220	1	1.163559	0.9;
	2475	2	26.62953	-1.3	0	0	1	1.030636	-25.767285	110	1	1.2	0.8;
	2476	1	0.1	152.7	0	0	1	1.085634	-39.733214	380	1	1.14892	0.9;
	2477	1	44.04884	14.6	0	0	1	0.922727	-31.153799	110	1	1.1	0.9;
	2478	2	86.59601	21.5	0	0	1	1.073441	-9.88268	380	1	1.128485	0.9;
	2479	2	34.13785	12.3	0	0	1	0.991722	-13.904976	110	1	1.1	0.9;
	2480	1	-61.6684	-0.9	0	0	1	1.028122	-43.707214	150	1	1.1	0.9;
	2481	2	0	0	0	0	1	1.017318	49.59778	220	1	1.1	0.9;
	2482	1	0	0	0.13851	0	1	1.07018	-18.694932	380	1	1.158917	0.9;
	2483	1	0	0	0	0	1	1.004579	-29.362917	220	1	1.1	0.9;
	2484	1	0	0	0	0	1	0.98674	-25.519193	150	1	1.1	0.9;
	2485	1	36.24018	5.7	0	0	1	1.050786	-49.373617	220	1	1.112889	0.9;
	2486	1	0	0	0	0	1	1.047017	-21.576427	150	1	1.106928	0.9;
	2487	2	0	0	0	0	1	1.054136	55.115493	220	1	1.1	0.9;
	2488	2	0	0	0	0	1	0.996493	-18.314691	110	1	1.1	0.9;
	2489	1	0	0	0	0	1	1.075843	-1.946106	150	1	1.145276	0.9;
	2490	2	0	0	0	0	1	1.019444	-21.678873	110	1	1.1	0.9;
	2491	1	65.4726	26.7	0	0	1	1.072296	-2.374486	150	1	1.142241	0.9;
	2492	1	55.5616	26.5	0	0	1	1.04932	-29.774715	220	1	1.143355	0.9;
	2493	1	34.53829	8.6	0	0	1	1.048493	-20.963168	220	1	1.118229	0.9;
	2494	1	12.21354	1	0	0	1	0.952126	-24.333039	110	1	1.1	0.9;
	2495	2	0	0	0	0	1	1.006417	-25.965277	150	1	1.1	0.9;
	2496	1	0	0	0	0	1	1.058723	-12.614185	220	1	1.113327	0.9;
	2497	1	0	0	0	0	1	1.036077	-0.498321	380	1	1.105225	0.9;
	2498	1	159.3	13.5	0	0	1	1.053704	-17.305692	380	1	1.125293	0.9;
	2499	1	0	0	0	0	1	1.042863	-23.437738	150	1	1.1	0.9;
	2500	2	263.3	21.9	0	0	1	1.081947	14.796398	380	1	1.154095	0.9;
	2501	1	90.40023	13.8	0	0	1	1.084921	-3.18017	220	1	1.144616	0.9;
	2502	1	64.87193	25.6	0	0	1	1.070369	3.43959	150	1	1.140714	0.9;
	2503	1	8.509435	2.8	0	0	1	1.018031	-28.467367	110	1	1.1	0.9;
	2504	1	16.51831	6	0	0	1	1.015778	-22.991165	110	1	1.115486	0.9;
	2505	2	0.900999	49.1	0	0	1	1.088579	-44.700546	380	1	1.144262	0.9;
	2506	1	0	0	0	0	1	1.064189	-35.188938	380	1	1.125801	0.9;
	2507	1	0	0	0	0	1	0.985628	-25.185929	110	1	1.1	0.9;
	2508	1	98.6	28.3	0	0	1	1.078971	20.270825	220	1	1.157272	0.9;
	2509	1	85.39468	14.3	0	0	1	1.050609	-9.704287	220	1	1.121718	0.9;
	2510	2	150.8	29.5	0	0	1	1.056239	6.801808	220	1	1.127073	0.9;
	2511	1	103.1	3.1	0	0	1	1.090409	-27.837881	380	1	1.136833	0.9;
	2512	1	4.705216	1.8	0	0	1	0.976393	-12.024571	110	1	1.1	0.9;
	2513	1	1.501665	0.3	0	0	1	0.996092	-22.684788	110	1	1.1	0.9;
	2514	2	10.11121	3.7	0	0	1	1.03873	-20.402988	110	1	1.1	0.9;
	2515	1	26.62953	7.6	0	0	1	1.066954	-13.302516	110	1	1.131625	0.9;
	2516	1	0	0	0	0	1	0.97218	-24.849475	150	1	1.1	0.9;
	2517	1	35.33918	21.9	0	0	1	1.006595	-25.950033	150	1	1.1	0.9;
	2518	1	2.903219	0.8	0	0	1	0.95906	-23.453285	110	1	1.1	0.9;
	2519	1	-9.41043	3.1	0	0	1	1.051746	-21.251126	150	1	1.1	0.9;
	2520	2	0	0	0	0	1	1.095266	-11.350467	220	1	1.170056	0.9;
	2521	1	0	0	0	0	1	1.029944	-24.359302	150	1	1.1	0.9;
	2522	1	0.300333	1.3	0	0	1	0.994193	11.442067	220	1	1.1	0.9;
	2523	1	189.4	25.8	0	0	1	1.03695	-45.25056	220	1	1.1	0.9;
	2524	2	0	0	0	0	1	1.102737	5.853766	380	1	1.140644	0.9;
	2525	1	0	0	0	0	1	1.005479	10.823786	220	1	1.1	0.9;
	2526	1	129.5	26	0	0	1	0.964005	21.394699	220	1	1.1	0.9;
	2527	1	0	0	0	0	1	0.998351	-27.112844	150	1	1.1	0.9;
	2528	1	0	0	0	0	1	1.029141	-16.062744	110	1	1.13483	0.9;
	2529	1	3.00333	0.1	0	0	1	0.948498	-24.196018	110	1	1.1	0.9;
	2530	1	0	0	0	0	1	1.081324	-17.542881	220	1	1.146101	0.9;
	2531	1	0	0	0	0	1	1.047876	-31.946451	220	1	1.154443	0.9;
	2532	1	100.7	17.5	0	0	1	1.083134	7.604784	150	1	1.137649	0.9;
	2533	2	274.1	69.3	0	0	1	1.078545	-4.14217	220	1	1.143729	0.9;
	2534	1	8.909878	2.7	0	0	1	0.996764	-23.821388	110	1	1.1	0.9;
	2535	2	0	0	0	0	1	1.074091	-12.009763	220	1	1.154056	0.9;
	2536	2	1	128	0	0	1	1.083868	-44.127421	380	1	1.120364	0.9;
	2537	1	0	0	0	0	1	1.072943	-23.292872	220	1	1.138116	0.9;
	2538	1	24.62731	8.1	0	0	1	0.965744	-26.191022	110	1	1.1	0.9;
	2539	1	69.07658	68	0	0	1	1.042877	-34.343944	220	1	1.105591	0.9;
	2540	1	31.33474	9.6	0	0	1	1.028928	8.403044	110	1	1.110378	0.9;
	2541	1	390.8	79.5	0	0	1	1.062335	-40.064695	380	1	1.131087	0.9;
	2542	1	0	0	0	0	1	1.075858	0.279328	220	1	1.151122	0.9;
	2543	1	327.4	125.3	0	0	1	1.084842	-14.215372	380	1	1.139333	0.9;
	2544	1	27.53052	11	0	0	1	0.952492	-38.906525	110	1	1.1	0.9;
	2545	1	6.907659	2.3	0	0	1	0.964307	-31.504183	110	1	1.1	0.9;
	2546	1	97	18.6	0	0	1	1.043376	14.686882	150	1	1.1	0.9;
	2547	1	32.93652	11.8	0	0	1	1.043646	-16.106382	110	1	1.105176	0.9;
	2548	1	0	0	0	0	1	1.089301	-14.512865	380	1	1.131393	0.9;
	2549	1	25.32808	14.2	0	0	1	1.060871	-21.957111	220	1	1.138391	0.9;
	2550	2	0	0	0	0	1	1.0915	-0.846843	380	1	1.152817	0.9;
	2551	2	0	0	0	0	1	1.024868	-42.149931	150	1	1.1	0.9;
	2552	1	22.02442	2.2	0	0	1	1.040858	-36.505145	110	1	1.1	0.9;
	2553	1	0	0	0	0	1	1.03278	-21.824671	150	1	1.1	0.9;
	2554	1	81.19002	31.4	0	0	1	1.00331	-47.836763	220	1	1.1	0.9;
	2555	2	0	0	0	0	1	1.053318	-44.251365	220	1	1.106118	0.9;
	2556	1	19.32142	6.4	0	0	1	1.058492	-29.577247	110	1	1.123181	0.9;
	2557	1	0	0	0	0	1	1.025323	13.046319	220	1	1.1	0.9;
	2558	1	0	0	0	0	1	1.026882	-24.48424	150	1	1.1	0.9;
	2559	1	8.909878	4.9	0	0	1	1.03903	0.250242	220	1	1.113288	0.9;
	2560	1	166.7	87.6	0	0	1	1.026462	-18.064619	220	1	1.1	0.9;
	2561	1	69.27681	23.8	0	0	1	0.975926	-34.624345	110	1	1.1	0.9;
	2562	1	46.55161	9.1	0	0	1	0.945136	-24.513567	110	1	1.1	0.9;
	2563	1	7.408214	2.7	0	0	1	1.048109	-21.465139	110	1	1.149932	0.9;
	2564	1	48.15339	14.8	0	0	1	1.02367	7.929674	110	1	1.105473	0.9;
	2565	1	3.804218	1.3	0	0	1	0.962896	-13.323297	110	1	1.1	0.9;
	2566	2	75.08324	4	0	0	1	1.084184	-3.438488	380	1	1.145003	0.9;
	2567	2	36.94096	8.4	0	0	1	0.969603	62.599067	150	1	1.1	0.9;
	2568	1	0	0	0	0	1	0.975873	-30.338776	150	1	1.1	0.9;
	2569	1	-9.31032	3	0	0	1	1.032479	-23.299073	150	1	1.1	0.9;
	2570	2	0	0	0	0	1	1.005	66.562332	150	1	1.1	0.9;
	2571	1	330.9	51.9	0	0	1	1.098019	-21.245866	120	1	1.148379	0.9;
	2572	2	0	0	0	0	1	1.059773	-28.201077	220	1	1.127171	0.9;
	2573	2	0	0	0	0	1	1.040364	-6.246332	220	1	1.129689	0.9;
	2574	1	25.12786	15.6	0	0	1	1.042845	-23.441941	150	1	1.1	0.9;
	2575	1	179	50.5	0	0	1	1.020534	-42.204037	220	1	1.1	0.9;
	2576	1	21.22353	5.3	0	0	1	0.985096	-23.404839	110	1	1.1	0.9;
	2577	1	42.24684	4.4	0.002004	0	1	0.962094	-19.013018	220	1	1.1	0.9;
	2578	1	40.14451	8.8	0	0	1	1.085139	-3.965624	220	1	1.138129	0.9;
	2579	2	454.1	140	0	0	1	1.044584	-7.943453	380	1	1.111332	0.9;
	2580	1	0	0	0	0	1	1.074589	-38.891749	380	1	1.139125	0.9;
	2581	1	0	0	0	0	1	1.072116	-17.809949	220	1	1.135568	0.9;
	2582	1	0	0	0	0	1	1.010941	-22.445338	150	1	1.1	0.9;
	2583	2	0.200222	77.4	0	0	1	1.096817	59.130678	220	1	1.108243	0.9;
	2584	2	0	0	0	0	1	1.024	-32.230453	150	1	1.1	0.9;
	2585	2	0	0	0	0	1	1.085298	-10.471932	380	1	1.153831	0.9;
	2586	1	0	0	0.25906	0	1	1.084605	-14.759209	380	1	1.151401	0.9;
	2587	1	7.107881	4.4	0	0	1	1.046549	-20.83721	150	1	1.1	0.9;
	2588	1	4.204662	1.1	0	0	1	0.986082	-17.403914	110	1	1.1	0.9;
	2589	1	192.3	84.6	0	0	1	1.039048	-19.30151	220	1	1.1	0.9;
	2590	1	-57	-24.9	0	0	1	0.976695	-26.902591	150	1	1.1	0.9;
	2591	1	0	0	0	0	1	0.966203	-24.767487	150	1	1.1	0.9;
	2592	1	17.71965	5.9	0	0	1	0.945102	-27.145607	110	1	1.1	0.9;
	2593	2	157.2	14.5	0	0	1	1.061909	-7.161958	220	1	1.120064	0.9;
	2594	2	39.24351	14.5	0	0	1	1.054352	-10.13625	220	1	1.120753	0.9;
	2595	1	30.0333	10	0	0	1	0.947734	-27.413084	110	1	1.1	0.9;
	2596	1	46.35139	16.8	0	0	1	1.058733	-17.312153	220	1	1.114511	0.9;
	2597	1	0	0	0	0	1	1.038853	-6.87386	220	1	1.127842	0.9;
	2598	2	235.3	75	0	0	1	1.072592	-22.107209	220	1	1.129474	0.9;
	2599	1	0	0	0	0	1	0.986659	-5.381227	110	1	1.1	0.9;
	2600	1	18.8	8.5	0	0	1	1.096855	-14.901823	380	1	1.160677	0.9;
	2601	1	0	0	0	0	1	1.087637	0.76719	380	1	1.136946	0.9;
	2602	2	87.59712	-0.6	0	0	1	1.043873	-45.520158	220	1	1.1	0.9;
	2603	2	0	0	0	0	1	1.065683	-10.23004	220	1	1.14158	0.9;
	2604	2	33.93763	21	0	0	1	1.010387	-27.444724	150	1	1.193765	0.9;
	2605	1	69.27681	23.1	0	0	1	1.033634	-35.120743	110	1	1.1	0.9;
	2606	1	24.12675	15	0	0	1	1.033124	-22.980448	150	1	1.1	0.9;
	2607	1	29.93319	0	0	0	1	1.071544	-13.995808	220	1	1.134893	0.9;
	2608	2	43.54828	12	0	0	1	1.058773	-10.628368	220	1	1.13673	0.9;
	2609	1	13.81532	5	0	0	1	1.059039	-4.925584	110	1	1.134585	0.9;
	2610	1	0	0	0	0	1	1.022186	-46.418083	220	1	1.1	0.9;
	2611	2	10.41154	2	0	0	1	0.959599	-24.248667	110	1	1.1	0.9;
	2612	1	205.9	38.3	0	0	1	1.058045	-11.45014	220	1	1.124477	0.9;
	2613	1	282.5	17.4	0	0	1	1.047706	-36.557601	380	1	1.114677	0.9;
	2614	1	0	0	0	0	1	1.083282	-23.736003	380	1	1.147287	0.9;
	2615	2	0	0	0	0	1	1.068864	-13.258918	220	1	1.120411	0.9;
	2616	1	57.66393	61	0	0	1	1.036982	-38.471532	220	1	1.1	0.9;
	2617	1	0	0	0	0	1	1.070275	-20.006583	220	1	1.152465	0.9;
	2618	1	0	0	0	0	1	1.087013	-1.01745	220	1	1.152058	0.9;
	2619	1	102.8	30.8	0	0	1	1.078821	-7.112744	380	1	1.145701	0.9;
	2620	1	254.9	71.4	0	0	1	1.060946	1.629209	380	1	1.118058	0.9;
	2621	1	10.11121	3.7	0	0	1	1.045905	-13.879132	110	1	1.157198	0.9;
	2622	1	0	0	0	0	1	1.018581	-29.434591	220	1	1.1	0.9;
	2623	1	0	0	0	0	1	1.019871	-34.869863	150	1	1.1	0.9;
	2624	1	151	45.7	0	0	1	1.049092	-12.921404	220	1	1.1132	0.9;
	2625	2	227	5	0	0	1	1.044318	-40.294865	220	1	1.1	0.9;
	2626	2	0	0	0	0	1	1.094545	-16.607931	110	1	1.135977	0.9;
	2627	1	56.76294	13.4	0	0	1	1.060166	-22.038997	220	1	1.137197	0.9;
	2628	1	0	0	0	0	1	1.08682	-40.55165	380	1	1.143714	0.9;
	2629	1	0	0	0	0	1	1.079603	3.219078	150	1	1.148856	0.9;
	2630	2	0	0	0	0	1	1.061894	-14.517943	220	1	1.129972	0.9;
	2631	2	0	0	0	0	1	1.085	-23.59459	220	1	1.143104	0.9;
	2632	1	133.7	59	0	0	1	0.989901	8.953609	220	1	1.1	0.9;
	2633	1	0	0	0.429913	0	1	1.095437	-10.964339	380	1	1.156108	0.9;
	2634	1	0	0	0	0	1	1.060654	-35.376624	220	1	1.141841	0.9;
	2635	1	255.8	144.8	0	0	1	1.054045	-9.435987	380	1	1.121889	0.9;
	2636	1	241	32.3	0	0	1	1.06531	-43.859277	220	1	1.114287	0.9;
	2637	1	203.8	61	0	0	1	1.074095	-10.56835	150	1	1.14543	0.9;
	2638	1	0	0	0	0	1	1.047044	-34.065276	220	1	1.109506	0.9;
	2639	1	0	0	0	0	1	1.086112	-3.23242	220	1	1.147382	0.9;
	2640	1	10.9121	3.6	0	0	1	1.003638	-29.701567	110	1	1.1	0.9;
	2641	1	46.25128	32.2	0	0	1	1.047531	40.020553	220	1	1.1	0.9;
	2642	1	51.95761	9.4	0	0	1	1.08437	-37.522987	380	1	1.128757	0.9;
	2643	1	0	0	0	0	1	1.023446	-23.350939	110	1	1.121201	0.9;
	2644	1	0	0	0	0	1	0.997344	-23.712148	110	1	1.1	0.9;
	2645	1	63.47037	30.2	0	0	1	1.094969	-34.411761	380	1	1.140048	0.9;
	2646	1	31.63507	7.5	0	0	1	1.070861	-42.37947	220	1	1.117225	0.9;
	2647	1	0	0	0	0	1	1.02914	-27.383698	380	1	1.126527	0.9;
	2648	1	0	0	0	0	1	1.036348	-17.406243	380	1	1.113852	0.9;
	2649	2	0	0	0	0	1	1.066553	-35.500883	380	1	1.128215	0.9;
	2650	1	37.24129	7.1	0	0	1	1.041199	-36.286309	220	1	1.1	0.9;
	2651	1	34.03774	12.3	0	0	1	1.07426	-19.582152	110	1	1.132654	0.9;
	2652	2	0	0	0	0	1	1.062342	-39.446995	380	1	1.129452	0.9;
	2653	1	19.92209	7.4	0	0	1	1.069672	-31.342825	220	1	1.156637	0.9;
	2654	1	19.03607	7.9	0	0	1	1.01711	-19.65202	150	1	1.1	0.9;
	2655	1	61.96871	14.3	0	0	1	1.075384	-17.534465	220	1	1.115929	0.9;
	2656	1	0	0	0	0	1	1.082444	16.019049	380	1	1.127705	0.9;
	2657	1	0	0	0	0	1	1.076574	-25.454026	220	1	1.166553	0.9;
	2658	1	0	0	0	0	1	0.997221	-23.481773	150	1	1.1	0.9;
	2659	2	0	0	0	0	1	1.060455	-12.501967	220	1	1.144242	0.9;
	2660	1	230.7	19.3	0	0	1	1.050685	-31.363096	220	1	1.162522	0.9;
	2661	1	17.7	-333.3	0.144009	0	1	1.079946	17.859832	380	1	1.132291	0.9;
	2662	1	70.0777	10	0	0	1	1.093009	-29.011561	380	1	1.134982	0.9;
	2663	2	0	0	0	0	1	1.080136	-6.175668	220	1	1.138639	0.9;
	2664	1	187.2	-44.2	0	0	1	1.082118	-9.810925	380	1	1.127285	0.9;
	2665	1	0	0	0	0	1	1.050362	-18.434224	220	1	1.111131	0.9;
	2666	1	16.71854	6.1	0	0	1	0.938353	-17.973868	110	1	1.1	0.9;
	2667	1	0	0	0	0	1	1.051329	-7.568721	380	1	1.117835	0.9;
	2668	1	0	0	0.261	0	1	1.004448	-21.42796	220	1	1.1	0.9;
	2669	1	103	28.7	0	0	1	1.0593	-13.982117	220	1	1.126257	0.9;
	2670	1	0	0	0	0	1	1.092016	-31.19555	380	1	1.138042	0.9;
	2671	2	17.5	502.7	0	0	1	1.064316	-8.855721	380	1	1.1	0.9;
	2672	1	0	0	0	0	1	1.016321	-24.394135	150	1	1.1	0.9;
	2673	1	6.607326	3.3	0	0	1	1.014059	-34.666398	110	1	1.1	0.9;
	2674	1	0	0	0	0	1	1.016558	8.202589	220	1	1.1	0.9;
	2675	1	41.94651	26	0	0	1	1.03201	-25.666017	150	1	1.1	0.9;
	2676	2	226.4	-13	0	0	1	1.014409	-42.051018	220	1	1.1	0.9;
	2677	2	2.402664	2.3	0	0	1	1.031212	-20.893656	110	1	1.1	0.9;
	2678	2	352.7	119.5	0	0	1	1.037439	0.906331	380	1	1.12651	0.9;
	2679	1	0	0	0	0	1	1.055346	-4.537223	220	1	1.144351	0.9;
	2680	1	16.21798	5.7	0	0	1	1.025859	-16.598019	110	1	1.1	0.9;
	2681	1	104.6	40.7	0	0	1	1.050279	-48.968345	220	1	1.1	0.9;
	2682	1	0	0	0	22.7313	1	1.031907	-25.496827	150	1	1.1	0.9;
	2683	1	4.805328	1.2	0	0	1	1.027298	-51.122233	150	1	1.1	0.9;
	2684	1	188.2	23.4	0	0	1	1.042041	-18.126406	220	1	1.1	0.9;
	2685	2	0	0	0	0	1	1.088658	-13.971565	380	1	1.181189	0.9;
	2686	1	81.19002	24.4	0	0	1	1.029672	-35.306555	220	1	1.1	0.9;
	2687	1	16.71854	5.5	0	0	1	0.946514	-34.954698	110	1	1.1	0.9;
	2688	2	0	0	0	0	1	1.010038	-17.487074	110	1	1.1	0.9;
	2689	1	0	0	0	0	1	0.998342	-16.784976	150	1	1.1	0.9;
	2690	1	159.9	53.2	0	0	1	1.073099	-23.184092	220	1	1.138213	0.9;
	2691	1	93.5	42.5	0	0	1	1.095428	-38.669676	380	1	1.152534	0.9;
	2692	1	0	0	0	0	1	1.074968	-23.273952	380	1	1.140333	0.9;
	2693	1	0	0	0	0	1	1.007507	-23.430479	150	1	1.1	0.9;
	2694	2	0	0	0	0	1	1.061073	15.8976	150	1	1.1	0.9;
	2695	1	4.404884	1.5	0	0	1	0.9632	-21.518806	110	1	1.1	0.9;
	2696	1	0	0	0	0	1	1.061365	-15.912754	220	1	1.1133	0.9;
	2697	2	134.3	24.2	0	0	1	1.081031	-8.81714	220	1	1.15125	0.9;
	2698	1	9.310323	3	0	0	1	0.957086	-29.325733	110	1	1.1	0.9;
	2699	1	0	0	0	0	1	1.026598	-17.952346	150	1	1.1	0.9;
	2700	1	0	0	0	0	1	1.032009	-6.372907	380	1	1.1	0.9;
	2701	2	108.3	26.1	0	0	1	1.045299	-14.335082	220	1	1.1	0.9;
	2702	2	0	0	0	0	1	1.081632	8.027599	380	1	1.107745	0.9;
	2703	1	65.4726	26.7	0	0	1	1.075578	-1.097104	150	1	1.144169	0.9;
	2704	2	53.85971	15.4	0	0	1	1.063636	-29.452452	110	1	1.124171	0.9;
	2705	1	0	0	0	0	1	0.957181	-27.388703	110	1	1.1	0.9;
	2706	1	5.806438	1.3	0	0	1	0.944665	-27.32933	110	1	1.1	0.9;
	2707	1	0	0	0	0	1	1.064938	-43.926837	220	1	1.113659	0.9;
	2708	1	0	0	0	0	1	1.100091	-8.858751	380	1	1.167824	0.9;
	2709	1	0	0	0	0	1	1.099531	-14.779109	220	1	1.194946	0.9;
	2710	2	0	0	0	0	1	0.992818	-21.421752	110	1	1.1	0.9;
	2711	1	26.82975	13	0	0	1	1.000697	-19.851554	220	1	1.1	0.9;
	2712	2	17.81976	0	0	0	1	0.966927	-24.57451	110	1	1.1	0.9;
	2713	2	137	22.8	0	0	1	1.092182	-8.170966	220	1	1.151933	0.9;
	2714	1	0	0	0	0	1	1.084874	-28.643421	380	1	1.142401	0.9;
	2715	2	0	0	0	0	1	1.098455	-11.580514	220	1	1.146121	0.9;
	2716	1	27.02997	9	0	0	1	0.953862	-29.467479	110	1	1.1	0.9;
	2717	2	629	184.3	0	0	1	1.052026	-8.038678	380	1	1.113499	0.9;
	2718	1	0	0	0	0	1	1.046368	9.07934	380	1	1.122007	0.9;
	2719	1	0	0	0	47.40056	1	0.937633	-33.294515	150	1	1.1	0.821661;
	2720	2	0	0	0	0	1	1.102526	-24.20092	380	1	1.159554	0.9;
	2721	1	17.71965	7.3	0	0	1	0.948563	-26.26888	110	1	1.1	0.9;
	2722	1	5.806438	1.9	0	0	1	0.995302	-30.131562	110	1	1.1	0.9;
	2723	1	0	0	0	0	1	1.083369	-3.783616	220	1	1.136437	0.9;
	2724	1	31.8353	5.5	0	0	1	0.963271	-13.280004	110	1	1.1	0.9;
	2725	2	193.9	221.1	0	0	1	1.083553	12.393441	380	1	1.163456	0.9;
	2726	1	17.01887	6.2	0	0	1	1.044528	-13.687294	110	1	1.106903	0.9;
	2727	1	139.6	68.4	0	0	1	1.049277	12.845433	150	1	1.1	0.9;
	2728	2	149.1	45	0	0	1	1.037701	2.208062	220	1	1.129641	0.9;
	2729	2	339.7	106.8	0	0	1	1.054955	5.750006	220	1	1.129113	0.9;
	2730	1	0	0	0	0	1	1.056551	-15.353344	150	1	1.10886	0.9;
	2731	1	0	0	0	0	1	1.0667	-7.830804	380	1	1.130725	0.9;
	2732	1	87.59712	27.4	0	0	1	1.043094	-5.889712	110	1	1.118573	0.9;
	2733	1	108.8	33.2	0.374986	0	1	1.070437	-47.421461	380	1	1.125124	0.9;
	2734	1	0	0	0	0	1	1.059381	-41.812151	380	1	1.126103	0.9;
	2735	2	146.9	2	0	0	1	1.057067	61.130764	150	1	1.1	0.9;
	2736	1	0	0	0	0	1	1.035945	-29.496065	150	1	1.1	0.9;
	2737	1	33.33696	11	0	0	1	0.899865	-33.250859	110	1	1.1	0.9;
	2738	1	-1398.9	-380.6	0	0	1	1.093523	-5.172233	380	1	1.153341	0.9;
	2739	1	2.202442	1.1	0	0	1	1.052224	-4.912474	220	1	1.141167	0.9;
	2740	1	12.81421	3.9	0	0	1	1.010502	-16.763633	110	1	1.1	0.9;
	2741	1	0	0	0	0	1	0.99831	-4.853194	220	1	1.1	0.9;
	2742	1	30.23352	10	0	0	1	0.99212	-21.950828	110	1	1.1	0.9;
	2743	1	0	0	0.296959	0	1	1.073029	-21.858929	380	1	1.142885	0.9;
	2744	2	0	0	0	0	1	1.039091	-12.214521	220	1	1.129856	0.9;
	2745	1	0	0	0	0	1	1.060889	-21.74615	220	1	1.138288	0.9;
	2746	1	0	0	0	0	1	1.085399	-39.425705	220	1	1.156142	0.9;
	2747	1	74.38247	8.8	0	0	1	1.044973	-11.54051	220	1	1.116515	0.9;
	2748	1	149	19.8	0	0	1	1.072247	-14.320456	380	1	1.148603	0.9;
	2749	1	0	0	0	0	1	1.029194	7.551285	380	1	1.1	0.9;
	2750	1	0	0	0	0	1	1.057205	-34.498318	220	1	1.135683	0.9;
	2751	1	0	0	0.169272	0	1	1.090745	-18.260083	380	1	1.149867	0.9;
	2752	1	0	0	0	0	1	0.960377	-27.224344	110	1	1.1	0.9;
	2753	1	0	0	0	0	1	1.056576	-48.669733	220	1	1.11451	0.9;
	2754	2	0	0	0	0	1	1.079816	-0.359435	380	1	1.141482	0.9;
	2755	2	28.73186	9.9	0	0	1	1.005545	-19.708393	110	1	1.1	0.9;
	2756	1	78.58714	25.6	0	0	1	0.941623	-27.818482	110	1	1.1	0.9;
	2757	1	6.607326	2.4	0	0	1	0.994537	-26.28118	110	1	1.1	0.9;
	2758	1	41.14562	5.7	0	0	1	0.969281	-23.072183	110	1	1.1	0.9;
	2759	1	4.605106	1.4	0	0	1	0.987078	-17.17728	110	1	1.1	0.9;
	2760	2	0	0	0	0	1	1.055658	-41.847447	380	1	1.111847	0.9;
	2761	1	116.9296	33.6	0	0	1	1.05737	-24.023488	220	1	1.131999	0.9;
	2762	1	221	146.8	0	0	1	1.015939	-6.171251	220	1	1.116705	0.9;
	2763	2	0	0	0	0	1	1.035045	13.09578	220	1	1.1	0.9;
	2764	2	96.40689	4.7	0	0	1	1.085773	-11.471439	220	1	1.121453	0.9;
	2765	1	0	0	0	0	1	1.030002	-24.352705	150	1	1.1	0.9;
	2766	1	0	0	0	0	1	1.092042	-15.905297	220	1	1.128847	0.9;
	2767	1	0	0	0	0	1	1.065028	-22.310743	380	1	1.121448	0.9;
	2768	1	26.92986	8	0	0	1	0.999762	-22.427486	110	1	1.1	0.9;
	2769	1	29.33252	5.6	0	0	1	1.020876	-20.587136	220	1	1.131314	0.9;
	2770	1	0	0	0	0	1	1.03981	-29.288278	220	1	1.1	0.9;
	2771	1	0	0	0	0	1	1.058082	2.045053	150	1	1.117785	0.9;
	2772	1	95.90633	-15.4	0	0	1	1.082534	-26.75273	220	1	1.122359	0.9;
	2773	1	0	0	0	0	1	1.046458	-22.327997	150	1	1.1	0.9;
	2774	2	0	0	0.191685	0	1	1.08369	-12.513783	220	1	1.14437	0.9;
	2775	1	0	0	0	0	1	1.039331	-24.598821	220	1	1.123711	0.9;
	2776	2	0	0	0	0	1	1.034455	-15.397715	110	1	1.132014	0.9;
	2777	1	0	0	0	0	1	1.065839	3.620543	380	1	1.131918	0.9;
	2778	1	10.31143	4.6	0	0	1	0.978671	-21.474884	110	1	1.1	0.9;
	2779	2	181.9	-68.8	0	0	1	1.032045	-46.969915	220	1	1.1	0.9;
	2780	1	22.5	9.5	0	0	1	1.103373	-16.268351	380	1	1.161956	0.9;
	2781	2	0	0	0.143721	0	1	1.105203	44.315547	380	1	1.119734	0.9;
	2782	2	0	0	0	0	1	0.967787	-19.52933	110	1	1.1	0.9;
	2783	1	0	0	0	0	1	1.041415	-20.285092	150	1	1.1	0.9;
	2784	1	493.4	53.5	0	0	1	1.076332	-43.49106	380	1	1.116876	0.9;
	2785	1	8.709657	1.6	0	0	1	1.050547	-7.283379	110	1	1.128366	0.9;
	2786	1	71.27902	6.9	0	0	1	1.037593	-46.236655	220	1	1.1	0.9;
	2787	1	55.86193	8.8	0	0	1	1.05238	-49.729887	220	1	1.112719	0.9;
	2788	1	119.7	3.7	0	0	1	1.04994	-49.502471	220	1	1.112059	0.9;
	2789	1	272.3	58.5	0	0	1	1.068861	-38.560337	380	1	1.134623	0.9;
	2790	1	566.1	54.2	0	0	1	1.035525	-45.305993	380	1	1.1	0.9;
	2791	1	89.59934	22.7	0	0	1	1.04307	-8.249229	220	1	1.118301	0.9;
	2792	1	0	0	0	0	1	1.070268	-20.006492	220	1	1.152458	0.9;
	2793	1	0	0	0	0	1	1.07104	-6.852238	220	1	1.130843	0.9;
	2794	1	92.90301	27.7	0	0	1	1.050072	-29.747047	220	1	1.144057	0.9;
	2795	1	100.7	-23.3	0	0	1	1.066659	-21.508622	220	1	1.1	0.9;
	2796	1	0	0	0	0	1	1.034547	-24.434307	150	1	1.1	0.9;
	2797	1	164	42.3	0	0	1	1.044151	-43.560599	220	1	1.1	0.9;
	2798	1	11.91321	3	0	0	1	1.019837	-15.300959	110	1	1.1	0.9;
	2799	2	0	0	0	0	1	1.059636	-26.448279	220	1	1.121975	0.9;
	2800	1	10.11121	3.3	0	0	1	0.97411	-18.433926	110	1	1.1	0.9;
	2801	1	0	0	0.301008	0	1	1.10113	-43.269272	380	1	1.156847	0.9;
	2802	1	0	0	0	0	1	1.069248	-15.072586	150	1	1.127509	0.9;
	2803	1	0	0	0	0	1	1.054851	-14.422775	380	1	1.116379	0.9;
	2804	2	575.6	132.7	0	0	1	1.075195	-26.800459	220	1	1.183602	0.9;
	2805	2	426.4	122.7	0	0	1	1.050273	1.81267	220	1	1.125224	0.9;
	2806	1	64.6717	24.5	0	0	1	1.074034	-26.856325	220	1	1.182585	0.9;
	2807	1	0	0	0	0	1	0.972538	-24.834318	150	1	1.1	0.9;
	2808	1	0	0	0	0	1	1.040877	-34.646783	220	1	1.1	0.9;
	2809	2	0	0	0	0	1	1.073227	-11.87513	220	1	1.126565	0.9;
	2810	1	15.21687	4.2	0	0	1	1.009293	-29.857443	110	1	1.2	0.8;
	2811	2	0	0	0	0	1	1.030455	-5.519567	220	1	1.109676	0.9;
	2812	1	28.63174	12.1	0	0	1	1.00646	-35.547763	110	1	1.1	0.9;
	2813	1	11.21243	3.7	0	0	1	1.070037	-27.372165	110	1	1.137685	0.9;
	2814	1	4.104551	1.5	0	0	1	1.036238	-15.421489	110	1	1.125603	0.9;
	2815	1	69.77736	15.3	0	0	1	1.04813	-49.199981	220	1	1.1	0.9;
	2816	1	128.1061	64.9	0	0	1	1.081518	0.98606	380	1	1.131714	0.9;
	2817	1	0	0	0	0	1	1.024472	-27.86673	220	1	1.1	0.9;
	2818	1	380.9	85.1	0	0	1	1.079144	-24.956358	380	1	1.139477	0.9;
	2819	1	0	0	0	0	1	0.982255	-23.806162	150	1	1.1	0.9;
	2820	1	0	8.4	0	0	1	1.040264	-20.896707	150	1	1.1	0.9;
	2821	1	125.2	105.4	0	0	1	1.053046	12.027943	380	1	1.126492	0.9;
	2822	1	0	0	0.059999	0	1	1.037644	-18.423267	220	1	1.1	0.9;
	2823	1	0	0	0.114	0	1	1.006279	-21.440196	220	1	1.1	0.9;
	2824	1	2.302553	0.8	0	0	1	0.961079	-32.322499	110	1	1.1	0.9;
	2825	1	30.63396	11.6	0	0	1	1.095921	-41.060743	380	1	1.135616	0.9;
	2826	1	0	0	0	0	1	1.067392	-0.76	220	1	1.138468	0.9;
	2827	1	0	0	0	0	1	0.975647	-23.94277	150	1	1.1	0.9;
	2828	1	0	0	0	0	1	1.061796	-38.628046	220	1	1.1	0.9;
	2829	2	0	0	0	0	1	1.080657	15.971905	380	1	1.135247	0.9;
	2830	1	0	0	0	0	1	1.066847	-37.417919	220	1	1.124595	0.9;
	2831	1	0	0	0	0	1	1.074742	2.456158	220	1	1.154924	0.9;
	2832	2	36.4	356.1	0	0	1	1.093184	-34.279704	380	1	1.155472	0.9;
	2833	2	0	0	0	0	1	1.071636	-10.916577	220	1	1.150764	0.9;
	2834	1	0	0	0	-84.0008	1	1.097059	-10.968097	380	1	1.159843	0.9;
	2835	1	57.86415	70.5	0.54138	0	1	1.036011	-0.497965	380	1	1.105155	0.9;
	2836	1	0	0	0	0	1	1.048163	-15.484434	150	1	1.1	0.9;
	2837	1	0	0	0	0	1	0.989739	38.018589	220	1	1.1	0.9;
	2838	1	99.9	20.1	0	0	1	1.066189	7.732993	150	1	1.115215	0.9;
	2839	1	0	0	0.191685	0	1	1.049231	-18.03817	220	1	1.120864	0.9;
	2840	2	247.6	75.8	0	0	1	1.036673	-6.89217	220	1	1.127764	0.9;
	2841	1	13.21465	4.4	0	0	1	0.933176	-27.925794	110	1	1.1	0.9;
	2842	1	115.5	71.5	0	0	1	1.088685	-22.237488	380	1	1.156404	0.9;
	2843	1	0	0	0	0	1	1.076037	-21.327851	220	1	1.121646	0.9;
	2844	1	0	0	0	0	1	1.05504	-36.158993	220	1	1.108715	0.9;
	2845	2	107.1	22.2	0	0	1	1.050103	-16.728876	220	1	1.11497	0.9;
	2846	2	0.7	270.7	0	0	1	1.070526	54.903209	380	1	1.1	0.9;
	2847	1	2.502775	0.5	0	0	1	0.995223	-16.764522	110	1	1.1	0.9;
	2848	1	109.5	22.8	0	0	1	0.919422	-26.134721	110	1	1.1	0.9;
	2849	2	0	0	0	0	1	1.070632	-18.079898	380	1	1.121334	0.9;
	2850	1	3.503885	1.4	0	0	1	0.964357	-29.885459	110	1	1.1	0.9;
	2851	2	0	0	0	0	1	0.992533	-16.445176	110	1	1.1	0.9;
	2852	2	56.06216	17.1	0	0	1	1.089773	-8.701437	220	1	1.15555	0.9;
	2853	2	278.3	174.6	0	0	1	0.983864	-14.904843	220	1	1.1	0.9;
	2854	1	0	0	0	0	1	1.09583	-33.926353	380	1	1.139224	0.9;
	2855	1	7.608436	2.5	0	0	1	0.982117	-24.507986	110	1	1.1	0.9;
	2856	1	0	0	0	0	1	1.089761	-25.250235	220	1	1.182236	0.9;
	2857	1	0	0	0	0	1	1.018979	-19.016221	150	1	1.1	0.9;
	2858	1	20.72298	6.3	0	0	1	1.052452	-20.164167	110	1	1.108194	0.9;
	2859	1	100.5	37	0	0	1	1.081925	-18.869476	380	1	1.145442	0.9;
	2860	1	19.82198	7.2	0	0	1	1.017967	-21.454816	110	1	1.1	0.9;
	2861	1	0	0	0	0	1	1.02912	-38.849966	380	1	1.117164	0.9;
	2862	1	0	0	0	0	1	1.086511	-17.124378	380	1	1.157425	0.9;
	2863	2	149.2	-74.6	0	0	1	1.104579	21.853809	380	1	1.177152	0.9;
	2864	1	0	0	0	0	1	1.075899	2.86124	150	1	1.145142	0.9;
	2865	1	80.0888	-1.8	0	0	1	1.042922	19.863301	380	1	1.1	0.9;
	2866	2	40.34473	7.8	0	0	1	1.041863	-44.098617	220	1	1.109386	0.9;
	2867	1	8.309213	3	0	0	1	1.022756	-19.14096	110	1	1.111481	0.9;
	2868	2	0	0	0	0	1	1.073273	-21.618659	220	1	1.116846	0.9;
	2869	1	68.67614	19.1	0	0	1	1.005744	-16.196641	110	1	1.1	0.9;
	2870	1	0	0	0	0	1	0.965084	-27.581006	150	1	1.1	0.9;
	2871	1	0	0	0	0	1	1.031304	4.041372	220	1	1.1	0.9;
	2872	2	0	0	0	0	1	1.079773	-34.243346	220	1	1.129192	0.9;
	2873	2	205	24.6	0	0	1	1.061267	63.934137	150	1	1.1	0.9;
	2874	1	136.7	19.3	0	0	1	1.090057	-22.031154	380	1	1.158051	0.9;
	2875	1	10.61177	51.2	0.211044	0	1	1.030073	13.067022	380	1	1.1	0.9;
	2876	2	60.46704	16.2	0	0	1	1.005569	-21.855857	110	1	1.1	0.9;
	2877	1	8.309213	2.5	0	0	1	1.001467	-19.115983	110	1	1.1	0.9;
	2878	2	0	-8.2	0	0	1	1.0335	-43.36439	220	1	1.1	0.9;
	2879	1	-16.5	3.6	0	0	1	1.035503	-24.795186	150	1	1.1	0.9;
	2880	2	196.8	36.4	0	0	1	1.07408	-16.856615	220	1	1.147847	0.9;
	2881	1	11.11232	3.6	0	0	1	0.998479	-30.140155	110	1	1.1	0.9;
	2882	1	0	0	0	0	1	1.070814	2.3646	150	1	1.144356	0.9;
	2883	1	39.64395	9.8	0	0	1	0.987377	-40.63887	110	1	1.1	0.9;
	2884	2	11.01221	4	0	0	1	0.983142	-18.289118	110	1	1.1	0.9;
	2885	1	0	0	0	0	1	1.040737	-20.316032	150	1	1.1	0.9;
	2886	1	0	0	0.224995	0	1	1.038236	-17.596212	220	1	1.109575	0.9;
	2887	1	292.8	114.2	0	0	1	1.053124	-42.574248	380	1	1.122001	0.9;
	2888	1	0	0	0	0	1	1.016981	-23.75469	150	1	1.1	0.9;
	2889	1	132.9	40.5	0	0	1	1.009887	7.38318	220	1	1.1	0.9;
	2890	1	39.34362	11.2	0	0	1	1.055546	6.091122	150	1	1.108105	0.9;
	2891	1	180.2	46.4	0	0	1	1.030206	-41.356831	220	1	1.1	0.9;
	2892	1	23.6262	7.8	0	0	1	1.055868	-29.87625	110	1	1.120854	0.9;
	2893	1	0	0	0	0	1	1.062192	-36.47934	380	1	1.125672	0.9;
	2894	1	0	0	0	0	1	1.026681	-25.760867	150	1	1.1	0.9;
	2895	1	31.93541	12.6	0	0	1	1.073834	3.181348	150	1	1.143672	0.9;
	2896	2	0	0	0	0	1	1.076455	-11.515071	220	1	1.155869	0.9;
	2897	1	14.11565	7.7	0	0	1	1.070373	-20.300599	220	1	1.135359	0.9;
	2898	1	0	0	0	0	1	1.02288	-22.098043	150	1	1.1	0.9;
	2899	1	-20.9232	2.5	0	0	1	1.037744	-29.629195	220	1	1.1	0.9;
	2900	1	107	34.7	0	0	1	1.043618	-12.299307	220	1	1.1	0.9;
	2901	1	75.5838	30.3	0	0	1	1.083289	-41.425378	220	1	1.155099	0.9;
	2902	2	0	0	0	0	1	1.007818	48.693314	220	1	1.1	0.9;
	2903	1	0	0	0	0	1	1.062981	11.749103	380	1	1.115328	0.9;
	2904	1	127.8	15.2	0	0	1	1.042791	-12.186136	220	1	1.114231	0.9;
	2905	1	0	0	0.38766	0	1	1.091495	-13.918236	380	1	1.162479	0.9;
	2906	1	0	0	0	-132.645	1	1.092832	-39.219982	380	1	1.163268	0.9;
	2907	1	104.5	31.8	0	0	1	1.052964	-36.357715	220	1	1.1349	0.9;
	2908	1	398	11.5	0	0	1	1.067691	-32.518832	380	1	1.128463	0.9;
	2909	1	15.91765	12	0	0	1	0.968446	-12.842848	220	1	1.1	0.9;
	2910	2	0	0	0	0	1	0.983182	-33.384459	220	1	1.1	0.9;
	2911	1	86.19556	19.9	0	0	1	1.06048	-22.022673	220	1	1.13803	0.9;
	2912	1	0	0	0	0	1	1.030401	-24.301847	150	1	1.1	0.9;
	2913	1	0.200222	0.9	0	0	1	1.069406	25.410679	220	1	1.125695	0.9;
	2914	1	27.02997	9	0	0	1	1.024037	-16.561097	110	1	1.1	0.9;
	2915	1	8.509435	3.2	0	0	1	1.063111	-9.524379	220	1	1.130599	0.9;
	2916	2	0	0	0	0	1	1.077342	-1.39469	380	1	1.127864	0.9;
	2917	1	146.7	20.7	0	0	1	1.096427	8.533967	150	1	1.150254	0.9;
	2918	1	98.00867	16.6	0	0	1	0.928457	-23.185415	110	1	1.1	0.9;
	2919	1	0	0	0	0	1	1.081465	-25.064029	380	1	1.139748	0.9;
	2920	1	2.302553	0.8	0	0	1	0.988449	-37.145049	110	1	1.1	0.9;
	2921	1	0	0	0	0	1	1.043088	7.259541	380	1	1.1	0.9;
	2922	1	0	0	0	0	1	1.031073	-40.297721	220	1	1.1	0.9;
	2923	1	0	0	0.191685	0	1	1.017209	-22.29518	220	1	1.1	0.9;
	2924	1	0	0	0	0	1	1.040196	-20.652844	150	1	1.1	0.9;
	2925	1	63.06993	25.8	0	0	1	1.071987	-2.443752	150	1	1.142004	0.9;
	2926	1	73.58158	30.1	0	0	1	1.070963	-2.55651	150	1	1.141083	0.9;
	2927	2	165.2	64.5	0	0	1	1.054318	-10.304506	220	1	1.120161	0.9;
	2928	2	341.2	46.6	0	0	1	1.057045	9.752306	220	1	1.121282	0.9;
	2929	1	76.38469	24.8	0	0	1	1.049979	-11.615394	220	1	1.107532	0.9;
	2930	1	9.00999	1.1	0	0	1	0.989834	-36.205638	110	1	1.1	0.9;
	2931	1	0	0	0	0	1	1.012901	-21.912101	150	1	1.1	0.9;
	2932	1	0	0	0	0	1	1.069318	22.313237	110	1	1.116234	0.9;
	2933	1	81.2	89.3	0	0	1	1.024544	12.674949	380	1	1.1	0.9;
	2934	2	0	0	0	0	1	1.088158	-26.898486	380	1	1.144139	0.9;
	2935	1	0	0	0	0	1	1.083431	-15.588848	220	1	1.165622	0.9;
	2936	2	0	0	0	0	1	1.098211	-15.581378	380	1	1.187947	0.9;
	2937	1	251.3	42.6	0	0	1	1.04127	-14.316994	220	1	1.11337	0.9;
	2938	1	133.9	32.4	0	0	1	0.997635	-26.477314	110	1	1.1	0.9;
	2939	2	374.1	61.4	0	0	1	1.020864	-42.048303	220	1	1.1	0.9;
	2940	2	-38.8431	-17	0	0	1	1.0705	-22.319689	220	1	1.121035	0.9;
	2941	2	3.203552	1.3	0	0	1	1.029379	-16.029925	110	1	1.1	0.9;
	2942	2	0	0	0	0	1	1.026591	-26.893398	220	1	1.1	0.9;
	2943	1	16.4182	5.9	0	0	1	1.047606	-21.218837	110	1	1.108829	0.9;
	2944	1	0	0	0	0	1	1.078628	-35.26126	220	1	1.157587	0.9;
	2945	2	0	0	0	0	1	1.092885	-11.251163	380	1	1.161936	0.9;
	2946	2	110.7	26.9	0	0	1	1.087545	-1.032971	220	1	1.152368	0.9;
	2947	1	132	27	0	0	1	1.054574	-16.871844	220	1	1.119817	0.9;
	2948	1	296.1	82.3	0	0	1	1.08624	-9.134961	150	1	1.151806	0.9;
	2949	1	0	0	0	0	1	1.036049	-17.284365	380	1	1.115104	0.9;
	2950	1	0	0	0	18.54267	1	1.043739	-20.653319	150	1	1.1	0.9;
	2951	1	0	0	0	0	1	1.027394	-21.20604	220	1	1.1	0.9;
	2952	1	20.9	-171.5	0	0	1	1.064365	8.117001	220	1	1.131278	0.9;
	2953	1	113	56.5	0	0	1	0.996651	-27.472636	150	1	1.1	0.9;
	2954	1	84.19334	17.3	0	0	1	1.095794	-11.226288	220	1	1.168438	0.9;
	2955	1	0	0	0	0	1	1.034796	0.11921	380	1	1.1	0.9;
	2956	1	26.32919	7.6	0	0	1	1.020164	-15.279162	110	1	1.1	0.9;
	2957	2	142.4	87.8	0	0	1	1.065364	24.835544	220	1	1.115276	0.9;
	2958	1	0	0	0	0	1	0.994519	-16.682469	150	1	1.1	0.9;
	2959	1	0	0	0	0	1	1.053981	-14.837508	380	1	1.111941	0.9;
	2960	1	0	0	0	0	1	1.062173	-21.051562	220	1	1.139184	0.9;
	2961	1	0	0	0	0	1	1.07952	-47.003678	380	1	1.136391	0.9;
	2962	1	0	0	0	0	1	1.010308	9.405488	220	1	1.1	0.9;
	2963	1	-11.713	3.8	0	0	1	1.003753	-24.689226	150	1	1.1	0.9;
	2964	1	0	0	0	0	1	1.082548	-17.322142	380	1	1.163932	0.9;
	2965	1	86.99646	29.6	0	0	1	1.001623	9.757293	220	1	1.1	0.9;
	2966	1	0	0	0.234904	0	1	1.104831	-20.196172	380	1	1.2	0.8;
	2967	2	9.210212	3.5	0	0	1	1.056273	-21.158846	110	1	1.151611	0.9;
	2968	2	0	0	0	0	1	1.069263	18.478192	380	1	1.135987	0.9;
	2969	1	0	0	0	0	1	1.072857	-46.821614	380	1	1.128481	0.9;
	2970	1	0	0	0	0	1	1.02902	-49.549768	220	1	1.1	0.9;
	2971	1	0	0	0	0	1	1.061676	21.764821	380	1	1.130047	0.9;
	2972	1	0	0	0	0	1	1.080021	-35.918561	380	1	1.143889	0.9;
	2973	1	0	0	0	0	1	1.059599	3.907864	380	1	1.116372	0.9;
	2974	1	-30.0333	9.9	0	0	1	0.987583	-17.285401	150	1	1.1	0.9;
	2975	1	0	0	0	0	1	0.952053	-24.0663	110	1	1.1	0.9;
	2976	1	12.11343	5.5	0	0	1	1.043903	-30.841539	110	1	1.10574	0.9;
	2977	1	0	0	0	0	1	1.070603	-39.006717	380	1	1.141351	0.9;
	2978	1	0	0	0	0	1	1.0913	-7.348617	220	1	1.159748	0.9;
	2979	1	0	0	0.268997	0	1	1.07182	-15.293222	380	1	1.120261	0.9;
	2980	2	66.67392	33.3	0	0	1	1.071091	-24.982263	220	1	1.11164	0.9;
	2981	1	0	0	0	0	1	1.083585	-6.006357	150	1	1.140716	0.9;
	2982	1	0	0	0	0	1	1.075349	-1.951692	150	1	1.144781	0.9;
	2983	1	10.61177	3.8	0	0	1	1.008822	-11.819252	110	1	1.1	0.9;
	2984	1	0	0	0	92.67897	1	1.066468	-51.893841	220	1	1.1	0.9;
	2985	2	0	0	0	0	1	1.058895	-29.704244	380	1	1.124642	0.9;
	2986	2	0	0	0	0	1	1.086684	-25.586738	380	1	1.143619	0.9;
	2987	1	0	0	0	0	1	1.003861	-15.722515	220	1	1.1	0.9;
	2988	2	0	0	0	0	1	1.015739	-19.90835	220	1	1.1	0.9;
	2989	1	11.01221	6	0	0	1	1.003496	-23.891826	150	1	1.1	0.9;
	2990	2	0	0	0	0	1	1.097053	8.622068	380	1	1.129297	0.9;
	2991	2	0.600666	28.7	0	0	1	1.042136	-40.513228	220	1	1.10816	0.9;
	2992	1	33.13674	11.6	0	0	1	1.045123	-11.331823	220	1	1.117982	0.9;
	2993	1	5.506105	2	0	0	1	1.064348	-1.36035	110	1	1.133657	0.9;
	2994	1	0	0	0	-22.9504	1	1.016345	-26.874598	150	1	1.187252	0.9;
	2995	1	0	0	0	0	1	1.012289	-21.9118	150	1	1.1	0.9;
	2996	1	0	0	0	0	1	1.003159	-23.571916	150	1	1.1	0.9;
	2997	2	22.12453	13.7	0	0	1	1.036267	-19.773625	150	1	1.116124	0.9;
	2998	1	332.9	99.6	0	0	1	1.05505	6.406052	150	1	1.106948	0.9;
	2999	1	12.01332	3.9	0	0	1	0.969295	-25.734721	110	1	1.1	0.9;
	3000	1	72.07992	29.5	0	0	1	1.075094	-1.729751	150	1	1.144306	0.9;
	3001	1	0	0	0	0	1	1.064543	-9.244203	220	1	1.135684	0.9;
	3002	2	0	0	0	0	1	1.046091	8.406196	220	1	1.119098	0.9;
	3003	1	0	0	0	0	1	1.026633	-23.601729	150	1	1.1	0.9;
	3004	1	0	0	0	0	1	1.080601	-16.830247	380	1	1.150695	0.9;
	3005	1	32.63618	11.8	0	0	1	1.020547	-22.845816	110	1	1.120601	0.9;
	3006	2	0	0	0	0	1	1	-5.043151	110	1	1.1	0.9;
	3007	1	19.82198	7.3	0	0	1	1.003783	-32.916916	110	1	1.1	0.9;
	3008	1	-10.1112	3.3	0	0	1	1.058575	-15.251834	150	1	1.1	0.9;
	3009	1	0	0	0	0	1	1.085527	-13.73514	380	1	1.152684	0.9;
	3010	2	0	0	0	0	1	1.068652	2.4194	150	1	1.14018	0.9;
	3011	1	0	0	0	0	1	1.036213	-21.55027	150	1	1.112726	0.9;
	3012	1	50.55605	16.6	0	0	1	1.019148	-32.434261	110	1	1.1	0.9;
	3013	1	0	0	0	0	1	0.968237	-26.066436	150	1	1.1	0.9;
	3014	1	120.4	39.5	0	0	1	1.054959	-14.736713	380	1	1.11836	0.9;
	3015	1	0	0	0	0	1	1.021075	-21.33261	150	1	1.1	0.9;
	3016	1	0	0	0	0	1	1.009742	13.572274	220	1	1.1	0.9;
	3017	2	78.18668	27	0	0	1	0.942133	65.870463	150	1	1.1	0.9;
	3018	2	70.47814	16.4	0	0	1	1.059273	-15.204714	220	1	1.128209	0.9;
	3019	1	0	0	0	0	1	1.069291	-19.082239	220	1	1.137538	0.9;
	3020	2	219.2189	70.1	0	0	1	1.073273	7.751892	220	1	1.158853	0.9;
	3021	1	0	0	0	0	1	1.040299	-11.625128	220	1	1.1	0.9;
	3022	1	14.71632	4.6	0	0	1	0.984444	-19.644972	110	1	1.1	0.9;
	3023	1	2.302553	0.8	0	0	1	0.968236	-17.871157	110	1	1.1	0.9;
	3024	2	70.47814	33.9	0	0	1	1.030364	65.884655	220	1	1.1	0.9;
	3025	1	0	0	0	0	1	1.063558	-9.477405	220	1	1.131008	0.9;
	3026	2	0	0	0	0	1	1.03285	39.128509	220	1	1.1	0.9;
	3027	2	48.35361	62.8	0	0	1	1.082318	11.673814	220	1	1.162517	0.9;
	3028	2	120.5	37.5	0	0	1	1.025909	-34.584864	220	1	1.1	0.9;
	3029	2	0	0	0	0	1	1.040526	25.906379	380	1	1.131747	0.9;
	3030	1	236.2	116.4	0	0	1	1.030231	-32.668653	220	1	1.1	0.9;
	3031	1	14.71632	6.7	0	0	1	0.957473	-22.872858	110	1	1.1	0.9;
	3032	1	0	0	0	0	1	1.026441	-23.760392	150	1	1.1	0.9;
	3033	2	379.5	113.3	0	0	1	1.003182	-9.399957	220	1	1.1	0.9;
	3034	1	46.85194	16.6	0	0	1	1.006631	-39.037707	150	1	1.1	0.9;
	3035	1	76.18446	13.2	0	0	1	1.02589	-36.295819	220	1	1.1	0.9;
	3036	1	379.9	73	0	0	1	1.049278	-42.33604	380	1	1.108765	0.9;
	3037	1	83.29234	24.1	0	0	1	1.0521	-15.562147	220	1	1.105694	0.9;
	3038	2	0	0	0	0	1	1.074632	-11.35296	380	1	1.134292	0.9;
	3039	1	38.04218	20	0	0	1	1.061833	-20.337599	220	1	1.124762	0.9;
	3040	1	0	0	0	0	1	1.057573	-30.223551	380	1	1.123335	0.9;
	3041	2	0	0	0	0	1	1.087237	66.661369	380	1	1.1	0.9;
	3042	1	1.301443	0.4	0	0	1	0.972237	-24.777487	110	1	1.1	0.9;
	3043	1	129.5	51.8	0	0	1	1.088193	-27.374381	380	1	1.167366	0.9;
	3044	1	0	0	0	0	1	1.051033	9.65765	380	1	1.11152	0.9;
	3045	2	25.12786	9.4	0	0	1	1.081773	-2.798707	220	1	1.146495	0.9;
	3046	1	0	0	0	14.75693	1	1.041428	-25.089107	150	1	1.1	0.9;
	3047	1	0	0	0	0	1	1.028381	-2.994963	220	1	1.1	0.9;
	3048	1	0	0	0.193831	0	1	1.085969	-14.458368	380	1	1.146725	0.9;
	3049	1	36.54051	27.4	0	0	1	1.013357	14.456595	220	1	1.1	0.9;
	3050	1	0	0	0	0	1	1.019781	-23.85086	110	1	1.1	0.9;
	3051	1	0	0	0	0	1	1.002717	-23.606678	110	1	1.1	0.9;
	3052	1	0	0	0	0	1	1.087105	-25.530327	380	1	1.144982	0.9;
	3053	1	0	0	0	0	1	1.062052	5.03699	380	1	1.127432	0.9;
	3054	2	0	0	0	0	1	1.055591	-15.852267	220	1	1.118723	0.9;
	3055	1	155.8	38.4	0	0	1	1.067642	1.578991	150	1	1.129653	0.9;
	3056	2	0	0	0	0	1	1.074	-12.014604	220	1	1.153995	0.9;
	3057	1	0	0	0.095843	0	1	1.048153	-18.119406	220	1	1.119934	0.9;
	3058	1	3.804218	1.3	0	0	1	0.969778	-22.645306	110	1	1.1	0.9;
	3059	1	57.46371	15.3	0	0	1	1.068901	-47.707222	220	1	1.123324	0.9;
	3060	2	201.6	127.8	0	0	1	1.086521	-6.970111	380	1	1.14742	0.9;
	3061	1	3.403774	1.4	0	0	1	0.949452	-20.611045	110	1	1.1	0.9;
	3062	1	0	0	0	0	1	1.082724	16.347499	380	1	1.127614	0.9;
	3063	2	0	0	0	0	1	1.036818	-31.317998	220	1	1.1	0.9;
	3064	1	0	0	0.130002	0	1	0.993199	-23.856781	220	1	1.1	0.9;
	3065	1	0.500555	0.5	0	0	1	1.024599	-15.800574	110	1	1.1	0.9;
	3066	1	156.2	24.5	0	0	1	1.063094	8.595146	150	1	1.106152	0.9;
	3067	2	322.1	131.9	0	0	1	1.004682	3.072873	220	1	1.1	0.9;
	3068	1	0	0	0	0	1	1.040333	-15.75639	150	1	1.1	0.9;
	3069	1	77.48591	21.4	0	0	1	1.066992	-9.713682	220	1	1.133864	0.9;
	3070	1	0	0	0	0	1	1.032648	-41.169306	220	1	1.1	0.9;
	3071	1	50.0555	19.7	0	0	1	0.998568	-21.430918	110	1	1.1	0.9;
	3072	1	8.309213	3	0	0	1	0.961372	-18.873153	110	1	1.1	0.9;
	3073	1	115	-4.5	0	0	1	1.072029	-42.461354	220	1	1.118822	0.9;
	3074	1	47.15228	15.2	0	0	1	1.005816	-32.10697	110	1	1.1	0.9;
	3075	1	0	0	0	0	1	1.02928	-43.852293	220	1	1.1	0.9;
	3076	1	0	0	0	0	1	0.958516	-29.073595	150	1	1.1	0.894865;
	3077	1	4.605106	1.7	0	0	1	1.017948	-22.940404	110	1	1.117869	0.9;
	3078	2	0	0	0	0	1	1.068652	2.4194	150	1	1.14018	0.9;
	3079	2	0	0	0	0	1	1.018318	-3.449897	220	1	1.1	0.9;
	3080	2	91.80178	18.6	0	0	1	1.044682	-11.480405	220	1	1.116954	0.9;
	3081	1	302.5	65.1	0	0	1	1.073766	-20.184055	220	1	1.177699	0.9;
	3082	1	0	0	0	0	1	0.957716	-25.67641	150	1	1.1	0.9;
	3083	2	40.44484	13.3	0	0	1	1.038648	-10.170538	110	1	1.1	0.9;
	3084	1	27.13008	9	0	0	1	0.977859	-25.427294	110	1	1.1	0.9;
	3085	1	0	0	0	0	1	1.003788	-24.739094	150	1	1.1	0.9;
	3086	2	0.2	108.3	0	0	1	1.074053	54.85819	380	1	1.1	0.9;
	3087	1	24.12675	9.8	0	0	1	1.078084	0.269717	150	1	1.147545	0.9;
	3088	2	0	0	0	0	1	0.978189	-28.791355	110	1	1.1	0.9;
	3089	1	50.0555	16	0	0	1	0.954497	-26.787892	110	1	1.1	0.9;
	3090	1	0	0	0	0	1	1.058125	-23.9344	220	1	1.1327	0.9;
	3091	1	23.22575	6.3	0	0	1	1.080147	-12.764753	220	1	1.134885	0.9;
	3092	2	207.0643	45.5	0	0	1	1.063538	0.577203	220	1	1.135679	0.9;
	3093	2	0	0	0	0	1	1.038064	-27.443733	220	1	1.1	0.9;
	3094	1	0	0	0	0	1	1.006418	-23.483663	150	1	1.1	0.9;
	3095	1	108.7	26.3	0	0	1	1.053678	-21.066389	220	1	1.1	0.9;
	3096	1	0	0	0	0	1	1.023853	-27.475478	150	1	1.1	0.9;
	3097	1	0	0	0	0	1	1.062375	-35.579827	380	1	1.136368	0.9;
	3098	1	52.05772	17	0	0	1	0.977118	-25.641839	110	1	1.1	0.9;
	3099	2	179.5	85.2	0	0	1	0.994091	-5.770157	220	1	1.1	0.9;
	3100	1	0	0	0	0	1	1.074784	2.456169	220	1	1.154924	0.9;
	3101	1	19.42153	7	0	0	1	1.023463	-44.10462	150	1	1.1	0.9;
	3102	1	78.08658	27	0	0	1	1.033301	-35.951008	110	1	1.1	0.9;
	3103	1	14.71632	5.3	0	0	1	1.027439	-21.240965	110	1	1.1	0.9;
	3104	2	0	0	0	0	1	1.003318	-32.861751	220	1	1.1	0.9;
	3105	1	0	0	0	0	1	0.979939	-26.475468	150	1	1.1	0.9;
	3106	1	0	0	0	0	1	1.020153	-21.031406	150	1	1.1	0.9;
	3107	1	0	0	0	0	1	1.016722	-29.403456	110	1	1.1	0.9;
	3108	1	6.707437	2.4	0	0	1	1.037993	-19.777788	110	1	1.1	0.9;
	3109	1	11.51276	4.2	0	0	1	1.039123	-15.478106	110	1	1.127607	0.9;
	3110	2	0	0	0	0	1	1.096545	-10.66646	220	1	1.171546	0.9;
	3111	1	173.1	4.6	0	0	1	1.043342	-17.012245	220	1	1.1	0.9;
	3112	1	83.59268	21.6	0	0	1	1.054874	-49.207278	220	1	1.111722	0.9;
	3113	1	146.9	26.6	0	0	1	1.052562	-49.019879	220	1	1.1113	0.9;
	3114	1	0	0	0	0	1	1.076969	-40.463479	380	1	1.128037	0.9;
	3115	1	175.4	28.8	0	0	1	1.02256	-51.68238	150	1	1.1	0.9;
	3116	1	125.1	44	0	0	1	0.993546	-15.351667	220	1	1.1	0.9;
	3117	2	0	0	0	0	1	1.047318	66.194703	220	1	1.1	0.9;
	3118	2	0	0	0	0	1	1.07932	-15.374402	220	1	1.120766	0.9;
	3119	1	0	0	0.120982	0	1	1.063758	-20.964184	380	1	1.116905	0.9;
	3120	1	4.204662	2	0	0	1	0.968204	-22.097585	110	1	1.1	0.9;
	3121	1	0	0	0	0	1	1.054091	-24.395115	220	1	1.128668	0.9;
	3122	2	314.5	83.9	0	0	1	1.052864	-12.884699	220	1	1.138245	0.9;
	3123	1	0	0	0	0	1	1.035448	-17.053031	110	1	1.111024	0.9;
	3124	2	140.1	10	0	0	1	1.055182	3.270548	220	1	1.129283	0.9;
	3125	1	0	0	0	0	1	1.087042	-12.466533	380	1	1.140855	0.9;
	3126	2	0	0	0.191685	0	1	1.0775	-13.253798	220	1	1.134136	0.9;
	3127	1	54.96094	12.2	0	0	1	1.055582	-48.671578	220	1	1.112132	0.9;
	3128	1	0	0	0	0	1	1.056115	5.504791	150	1	1.109787	0.9;
	3129	1	5.105661	1.3	0	0	1	0.949447	-24.354367	110	1	1.1	0.9;
	3130	2	0	0	0	0	1	1.044212	-6.1826	380	1	1.109704	0.9;
	3131	2	285.4	81.4	0	0	1	1.074294	-6.041978	380	1	1.118188	0.9;
	3132	1	120.1	30.5	0	0	1	1.042023	-36.688391	220	1	1.1	0.9;
	3133	1	-15.8	-2.8	0	0	1	1.054667	-20.81845	150	1	1.115303	0.9;
	3134	1	0	0	0	0	1	1.076794	-18.889467	380	1	1.115199	0.9;
	3135	1	78.98758	16.2	0	0	1	1.051404	-20.54431	220	1	1.1208	0.9;
	3136	1	20.82309	9.5	0	0	1	0.964291	-18.600522	110	1	1.1	0.9;
	3137	2	16.11787	4.7	0	0	1	0.934331	-25.674079	110	1	1.1	0.9;
	3138	1	3.904329	0.9	0	0	1	0.996729	-22.815655	110	1	1.1	0.9;
	3139	1	0	0	0	0	1	1.069999	-4.424369	380	1	1.122618	0.9;
	3140	1	0	0	0	0	1	1.079827	-12.289138	220	1	1.133885	0.9;
	3141	1	0	0	0	0	1	1.075448	-1.939198	150	1	1.144872	0.9;
	3142	1	23.12564	8.8	0	0	1	1.031482	-35.955999	110	1	1.1	0.9;
	3143	1	0	0	0	0	1	1.032947	-18.144229	150	1	1.1	0.9;
	3144	1	119.1	60.5	0	0	1	1.03054	-24.947004	150	1	1.1	0.9;
	3145	1	209.5	95	0	0	1	1.076029	9.805544	150	1	1.129578	0.9;
	3146	1	0	0	0	0	1	1.003816	-23.479426	150	1	1.1	0.9;
	3147	1	119.1	22.3	0	0	1	1.013932	-47.882634	220	1	1.1	0.9;
	3148	1	7.708547	2.2	0	0	1	1.03219	-30.472952	110	1	1.1	0.9;
	3149	1	1.501665	0.5	0	0	1	0.954703	-29.384894	110	1	1.1	0.9;
	3150	1	50.35583	-36.2	0	0	1	1.0668	-43.228151	220	1	1.117486	0.9;
	3151	1	0	0	0	0	1	1.030966	-27.498677	110	1	1.1	0.9;
	3152	1	123.1	51.8	0	0	1	0.981248	-22.885849	110	1	1.1	0.9;
	3153	1	18.72076	5.5	0	0	1	1.064122	-29.423453	110	1	1.124955	0.9;
	3154	1	18.01998	-15.7	0	0	1	1.055968	-19.423448	110	1	1.1154	0.9;
	3155	1	0	0	0	11.93467	1	1.002663	-26.096258	150	1	1.1	0.9;
	3156	1	23.92653	8.7	0	0	1	1.040234	-22.333869	110	1	1.1	0.9;
	3157	1	0	0	0	0	1	1.094347	34.690759	380	1	1.121672	0.9;
	3158	1	22.12453	7.9	0	0	1	1.023272	-15.231689	110	1	1.112099	0.9;
	3159	1	-3.40377	-0.5	0	0	1	1.031114	-40.292116	220	1	1.1	0.9;
	3160	1	124.9	104	0	0	1	1.000588	12.069677	220	1	1.1	0.9;
	3161	2	0	0	0	0	1	1.075727	-16.544665	220	1	1.142193	0.9;
	3162	1	142.5	23.6	0	0	1	1.070664	-19.654371	220	1	1.136304	0.9;
	3163	1	22.02442	13.7	0	0	1	0.961833	-30.882426	150	1	1.1	0.9;
	3164	1	0	0	0	0	1	1.083595	1.94371	380	1	1.129626	0.9;
	3165	1	0	0	0	0	1	1.06295	11.729464	380	1	1.115303	0.9;
	3166	1	0	0	0	0	1	1.03757	-15.536186	380	1	1.106054	0.9;
	3167	1	101.5	-18.3	0	0	1	1.075149	-19.823175	220	1	1.139193	0.9;
	3168	1	16.31809	7.6	0	0	1	1.049407	-35.985153	220	1	1.1	0.9;
	3169	1	55.66171	4.5	0	0	1	1.041017	-12.479683	220	1	1.112625	0.9;
	3170	1	0	0	0	0	1	1.055812	-48.631672	220	1	1.110316	0.9;
	3171	1	12.51387	4.1	0	0	1	1.041869	-32.577565	110	1	1.110701	0.9;
	3172	1	15.91765	5.2	0	0	1	0.955046	-34.785328	110	1	1.1	0.9;
	3173	2	473.4	113.1	0	0	1	1.058682	-8.152941	220	1	1.116006	0.9;
	3174	1	226.3	61.1	0	0	1	1.073299	-10.993906	150	1	1.1372	0.9;
	3175	1	0	0	0.241452	0	1	1.070895	-18.518174	380	1	1.121813	0.9;
	3176	1	36.34029	3.3	0	0	1	1.049129	-51.902872	220	1	1.1	0.9;
	3177	1	99.3	23.3	0	0	1	1.07358	-20.271266	220	1	1.152673	0.9;
	3178	1	100.5	32	0	0	1	1.073532	2.918109	150	1	1.1436	0.9;
	3179	1	138.5	19.3	0	0	1	1.069351	-9.820513	220	1	1.1301	0.9;
	3180	1	180.4	71.6	0	0	1	1.078839	-3.464642	380	1	1.142522	0.9;
	3181	1	94.20444	21.6	0	0	1	1.057646	-47.815172	220	1	1.112071	0.9;
	3182	1	98.2	19.1	0	0	1	1.068949	-9.760004	220	1	1.136284	0.9;
	3183	1	0	0	0	0	1	1.096026	-11.176767	220	1	1.152009	0.9;
	3184	1	8.909878	2.5	0	0	1	0.976557	-22.841304	110	1	1.1	0.9;
	3185	1	0	0	0	0	1	1.016398	-7.421913	220	1	1.1	0.9;
	3186	1	0	0	0	0	1	1.064651	-10.981418	150	1	1.138598	0.9;
	3187	2	0	0	0	0	1	1.055728	-35.644444	220	1	1.137513	0.9;
	3188	2	0	0	0.445544	0	1	1.075658	-20.923204	380	1	1.106135	0.9;
	3189	1	23.6262	16.5	0	0	1	1.067366	9.729743	150	1	1.1	0.9;
	3190	1	0	0	0	0	1	1.079692	-18.576331	380	1	1.153762	0.9;
	3191	2	0	0	0.095843	0	1	1.099364	1.523774	220	1	1.189273	0.9;
	3192	2	68.87637	0.2	0	0	1	1.046462	-48.378524	220	1	1.1	0.9;
	3193	2	0	0	0	0	1	1.078842	-3.620859	380	1	1.139595	0.9;
	3194	2	0	0	0	0	1	1.094331	45.036874	380	1	1.116989	0.9;
	3195	2	123.8	41.4	0	0	1	1.027682	0.999078	220	1	1.124704	0.9;
	3196	1	0	0	0	0	1	1.013112	-16.44844	110	1	1.1	0.9;
	3197	1	0	0	0	0	1	1.06287	6.595481	380	1	1.141992	0.9;
	3198	2	36.34029	45	0	0	1	1.038667	-7.750821	330	1	1.115607	0.9;
	3199	1	0	0	0	0	1	1.051378	-23.727266	380	1	1.121789	0.9;
	3200	1	0	0	0	0	1	1.065831	-6.830068	220	1	1.133431	0.9;
	3201	1	0	0	0	0	1	1.061351	8.11278	380	1	1.12482	0.9;
	3202	1	29.33252	9.6	0	0	1	0.958475	-29.378311	110	1	1.1	0.9;
	3203	1	29.93319	10.8	0	0	1	1.044371	-22.142775	110	1	1.106031	0.9;
	3204	1	0	0	0.038315	0	1	1.057919	65.60903	220	1	1.1	0.9;
	3205	1	33.33696	13.2	0	0	1	1.070342	-5.01357	220	1	1.136137	0.9;
	3206	1	0	0	0	0	1	0.973168	-22.371002	110	1	1.1	0.9;
	3207	1	0	0	0	0	1	1.020382	-23.331467	150	1	1.1	0.9;
	3208	1	0	0	0	0	1	1.072647	-18.619454	220	1	1.113131	0.9;
	3209	1	118.5	21	0	0	1	1.062107	0.616986	150	1	1.124536	0.9;
	3210	1	7.808658	2.8	0	0	1	1.030649	-21.731398	110	1	1.125476	0.9;
	3211	1	0	0	0	0	1	0.956692	-29.273887	150	1	1.1	0.891251;
	3212	1	159.1	-28.7	0	0	1	1.079236	-11.418151	220	1	1.132117	0.9;
	3213	1	8.609546	3.1	0	0	1	1.058499	-20.182301	110	1	1.116051	0.9;
	3214	1	0	0	0	0	1	1.06802	-23.157995	220	1	1.152961	0.9;
	3215	2	0	0	0	0	1	1.024741	-20.37732	220	1	1.1	0.9;
	3216	1	96.5	-43	0	0	1	1.076733	-9.582766	220	1	1.146003	0.9;
	3217	1	0	0	0	0	1	1.028778	13.663149	220	1	1.106248	0.9;
	3218	1	97	-52.1	0	0	1	1.055127	-10.341895	220	1	1.125532	0.9;
	3219	1	6.00666	3	0	0	1	1.060572	-35.38254	220	1	1.141766	0.9;
	3220	1	0	0	0	0	1	1.071282	-22.717958	220	1	1.155901	0.9;
	3221	1	0	0	0	0	1	0.992509	10.978088	220	1	1.1	0.9;
	3222	1	0	0	0	0	1	1.038476	-16.695508	380	1	1.115493	0.9;
	3223	1	0	0	0	0	1	1.068766	-39.399619	380	1	1.115823	0.9;
	3224	1	20.12231	7.6	0	0	1	1.090518	-7.456859	220	1	1.158973	0.9;
	3225	1	0	0	0	0	1	1.067243	-18.974382	380	1	1.140034	0.9;
	3226	2	0	0	0	0	1	1.091455	2.355425	110	1	1.155191	0.9;
	3227	1	10.0111	3	0	0	1	0.915007	-32.339722	110	1	1.1	0.9;
	3228	1	0	0	0	102.8747	1	1.029975	-38.966407	220	1	1.1	0.9;
	3229	2	0	0	0	0	1	1.065273	-19.51989	220	1	1.106153	0.9;
	3230	2	387.1	138.9	0	0	1	1.018318	4.128085	220	1	1.1	0.9;
	3231	1	0	0	0	0	1	1.048932	21.902365	380	1	1.108798	0.9;
	3232	1	0	0	0	0	1	0.993385	-16.719773	150	1	1.1	0.9;
	3233	1	362.2	-10.5	0	0	1	1.014888	-20.219618	220	1	1.1	0.9;
	3234	1	0	0	0	0	1	1.053504	-19.29452	150	1	1.1	0.9;
	3235	1	31.23463	12.9	0	0	1	1.042276	-43.14176	220	1	1.1	0.9;
	3236	1	87.19668	31.8	0	0	1	1.049154	-30.125574	220	1	1.144513	0.9;
	3237	1	0	0	0	0	1	1.096506	2.311152	220	1	1.163194	0.9;
	3238	1	4.104551	0.9	0	0	1	0.996965	-25.87292	110	1	1.1	0.9;
	3239	1	47.95317	17.2	0	0	1	1.052248	-34.762187	220	1	1.13114	0.9;
	3240	1	60.66726	37.6	0	0	1	0.970726	-24.528316	150	1	1.1	0.9;
	3241	1	6.306993	1.3	0	0	1	0.917824	-29.167224	110	1	1.1	0.9;
	3242	1	392.8	70.6	0	0	1	1.05965	-34.074774	380	1	1.120995	0.9;
	3243	2	19.72187	9.4	0	0	1	1.057682	-28.400536	220	1	1.125902	0.9;
	3244	1	5.205771	1.5	0	0	1	0.948251	-27.493913	110	1	1.1	0.9;
	3245	1	0	0	0	42.58596	1	1.016549	-30.024046	150	1	1.1	0.9;
	3246	1	215.3	37.7	0	0	1	1.074999	-19.957051	220	1	1.111651	0.9;
	3247	2	0	0	0.095844	0	1	1.0415	-18.45953	220	1	1.11594	0.9;
	3248	2	0	0	0	0	1	1.020227	-46.248334	220	1	1.1	0.9;
	3249	1	8.309213	3	0	0	1	1.0033	-12.417895	110	1	1.1	0.9;
	3250	1	0	0	0	0	1	1.026574	-23.580712	150	1	1.1	0.9;
	3251	1	0	0	0	0	1	1.030959	-27.2231	220	1	1.1	0.9;
	3252	1	24.42708	8.7	0	0	1	0.975812	-21.737974	110	1	1.1	0.9;
	3253	1	1.401554	0.5	0	0	1	1.045635	-14.894211	110	1	1.136405	0.9;
	3254	2	0	0	0	0	1	1.070314	-5.309326	380	1	1.137724	0.9;
	3255	1	0	0	0	0	1	1.070098	32.407615	380	1	1.1	0.9;
	3256	2	279.6	9.2	0	0	1	1.051337	-38.895968	220	1	1.185092	0.9;
	3257	2	0	0	0	0	1	1.076136	-32.717456	220	1	1.177876	0.9;
	3258	1	-10.6118	3.5	0	0	1	1.024328	-24.649029	150	1	1.1	0.9;
	3259	2	353	20.7	0	0	1	1.087053	-44.75495	380	1	1.135179	0.9;
	3260	2	0	0	0	0	1	1.069474	31.389322	380	1	1.133695	0.9;
	3261	2	115	34.8	0	0	1	1.0525	-40.473365	220	1	1.1	0.9;
	3262	2	0	0	0	0	1	1.024136	-40.425792	220	1	1.1	0.9;
	3263	1	0	0	0	0	1	1.039906	-18.727312	150	1	1.1	0.9;
	3264	1	0	0	0	0	1	1.094732	-36.201886	380	1	1.154606	0.9;
	3265	1	0	0	0	0	1	0.987907	-25.17741	150	1	1.1	0.9;
	3266	1	37.54162	18.3	0	0	1	1.039581	-6.324915	220	1	1.128981	0.9;
	3267	1	102.7	77.5	0	0	1	1.008716	-45.404537	220	1	1.1	0.9;
	3268	1	84.09324	52.1	0	0	1	1.032685	-25.351507	150	1	1.1	0.9;
	3269	1	31.43485	6.6	0	0	1	1.072893	-18.81232	220	1	1.136616	0.9;
	3270	1	14.01554	7.6	0	0	1	1.023126	-14.154816	110	1	1.1	0.9;
	3271	1	86.69612	34.6	0	0	1	1.007265	-48.506211	220	1	1.1	0.9;
	3272	1	0	0	0	0	1	1.015728	-19.908581	220	1	1.1	0.9;
	3273	1	4.404884	2.2	0	0	1	1.012068	-35.076991	110	1	1.1	0.9;
	3274	2	129.2	13.4	0	0	1	1.015545	-3.731409	220	1	1.1	0.9;
	3275	1	-11.0122	3.6	0	0	1	1.052139	-15.034527	150	1	1.1	0.9;
	3276	1	0	0	0	0	1	1.07049	-20.321105	220	1	1.135493	0.9;
	3277	1	86.59601	14.9	0	0	1	1.078536	-17.955881	220	1	1.143478	0.9;
	3278	1	0	0	0	0	1	1.047282	-21.55009	150	1	1.107215	0.9;
	3279	1	124.3	39.4	0	0	1	1.062439	-19.517429	380	1	1.115775	0.9;
	3280	2	0	0	0	0	1	1.075	-34.734836	220	1	1.133935	0.9;
	3281	1	194.3	27.9	0	0	1	0.967356	43.104772	220	1	1.1	0.9;
	3282	2	112	-12.1	0	0	1	1.033934	12.727123	220	1	1.1	0.9;
	3283	1	0	0	0	0	1	1.018901	-40.864908	380	1	1.139451	0.9;
	3284	2	0	0	0	0	1	1.09526	-11.351647	220	1	1.170051	0.9;
	3285	1	185.8	31.8	0	0	1	1.07219	-21.588298	220	1	1.155358	0.9;
	3286	2	217.8	60.5	0	0	1	1.051455	-17.711089	220	1	1.1	0.9;
	3287	1	152.5	14.4	0	0	1	1.024717	-27.274397	380	1	1.136037	0.9;
	3288	1	9.210212	3.3	0	0	1	1.012295	-17.05189	110	1	1.11013	0.9;
	3289	1	0	0	0	0	1	1.061344	-35.422202	380	1	1.136154	0.9;
	3290	2	222.4	81.7	0	0	1	1.041342	-0.67683	220	1	1.112178	0.9;
	3291	1	81.79068	51.1	0	0	1	1.030759	31.449083	220	1	1.1	0.9;
	3292	2	11.31254	4	0	0	1	1.012	-14.173778	110	1	1.1	0.9;
	3293	1	0	0	0	18.08379	1	1.030431	-24.498799	150	1	1.1	0.9;
	3294	1	35.13896	16.1	0	0	1	1.07468	1.207178	150	1	1.147701	0.9;
	3295	2	58.96538	27.1	0	0	1	1.0555	-47.601482	220	1	1.106999	0.9;
	3296	1	0	0	0	0	1	1.096315	-27.144461	380	1	1.155315	0.9;
	3297	1	0	0	0	0	1	1.057169	-29.960743	220	1	1.111616	0.9;
	3298	1	0	0	0	0	1	0.958127	-25.581572	150	1	1.1	0.9;
	3299	1	13.6151	8.4	0	0	1	1.022842	-19.179077	150	1	1.1	0.9;
	3300	1	0	0	0	0	1	1.02117	-22.117502	150	1	1.1	0.9;
	3301	2	24.62731	16.7	0	0	1	1.017182	-40.228324	220	1	1.1	0.9;
	3302	1	21.52386	7.6	0	0	1	0.948873	-27.043488	110	1	1.1	0.9;
	3303	1	0	0	0	0	1	1.097002	56.286031	380	1	1.116585	0.9;
	3304	1	-12.3	-10.6	0	0	1	1.039284	-19.924334	150	1	1.118818	0.9;
	3305	1	0	0	0	91.8129	1	1.072863	-50.091603	220	1	1.1	0.9;
	3306	2	0	0	0	0	1	1.098046	-26.524751	220	1	1.156258	0.9;
	3307	1	0	0	0	0	1	1.028074	16.478914	380	1	1.1	0.9;
	3308	1	23.92653	5.8	0	0	1	1.011915	-21.816908	110	1	1.1	0.9;
	3309	1	1.701887	0.3	0	0	1	0.939333	-25.674459	110	1	1.1	0.9;
	3310	1	0	0	0	0	1	1.039533	-22.817752	220	1	1.115642	0.9;
	3311	1	30.33363	11	0	0	1	1.074035	-19.563833	110	1	1.132409	0.9;
	3312	1	1.902109	1.4	0	0	1	0.969248	-20.984706	110	1	1.1	0.9;
	3313	1	24.22686	6.4	0	0	1	1.047248	-18.103372	220	1	1.1	0.9;
	3314	2	0	-5.5	0	0	1	1.054091	-38.789163	220	1	1.1	0.9;
	3315	1	42.54717	17.6	0	0	1	1.069365	-31.365263	220	1	1.15638	0.9;
	3316	1	32.23574	11.7	0	0	1	1.048224	-21.337576	110	1	1.10901	0.9;
	3317	1	-668.3	-230.2	0	0	1	1.084886	34.296505	380	1	1.138473	0.9;
	3318	1	0	0	0	0	1	1.086337	5.515906	380	1	1.121943	0.9;
	3319	1	0	0	0	0	1	1.018897	-38.460848	220	1	1.156226	0.9;
	3320	1	0	0	0	0	1	1.069032	-39.230696	220	1	1.11507	0.9;
	3321	1	0	0	0	0	1	1.082974	-22.278497	380	1	1.148751	0.9;
	3322	2	12.11343	7.5	0	0	1	1.053869	-20.881549	150	1	1.114602	0.9;
	3323	1	20.72298	5	0	0	1	1.076162	-25.51135	220	1	1.166186	0.9;
	3324	1	114.8	121.5	0	0	1	1.026483	3.626935	380	1	1.1	0.9;
	3325	2	138.4	39.5	0	0	1	1.070591	-11.063933	220	1	1.149952	0.9;
	3326	1	131	1.3	0	0	1	1.077592	-25.2018	220	1	1.167293	0.9;
	3327	1	0	0	0	0	1	1.020131	-21.03116	150	1	1.1	0.9;
	3328	2	0	0	0.141558	0	1	1.064474	1.143328	380	1	1.128862	0.9;
	3329	2	0	0	0	0	1	1.082816	4.41849	380	1	1.152125	0.9;
	3330	1	0	0	0	0	1	1.072364	-39.650398	380	1	1.130655	0.9;
	3331	2	0	0	0	0	1	1.063021	7.201363	150	1	1.112291	0.9;
	3332	1	295.4	42	0	0	1	1.014264	-39.380835	220	1	1.1	0.9;
	3333	1	0	0	0	0	1	1.021539	-22.194056	150	1	1.1	0.9;
	3334	1	0	0	0	0	1	1.050158	-20.692501	150	1	1.1	0.9;
	3335	1	0	0	0	0	1	1.043045	9.727364	380	1	1.1	0.9;
	3336	1	0	0	0	0	1	1.061744	-20.591143	110	1	1.166045	0.9;
	3337	1	28.03108	6.5	0	0	1	1.073953	-20.870822	220	1	1.153554	0.9;
	3338	1	22.82531	9.7	0	0	1	0.915012	-26.52649	110	1	1.1	0.9;
	3339	1	82.99202	25.9	0	0	1	1.041334	-12.869304	220	1	1.113525	0.9;
	3340	2	132.3	135.6	0	0	1	1.036227	-35.228121	220	1	1.1	0.9;
	3341	2	161.2	41.8	0	0	1	1.071885	-13.915092	220	1	1.135213	0.9;
	3342	1	0	0	0	0	1	0.985062	62.471679	150	1	1.1	0.9;
	3343	2	169.2	35.7	0	0	1	1.093923	-7.142414	380	1	1.151362	0.9;
	3344	1	5.506105	2.2	0	0	1	0.999976	-32.104144	110	1	1.1	0.9;
	3345	2	5.305883	3.3	0	0	1	1.056492	-20.818879	150	1	1.1	0.9;
	3346	2	64.87193	28.3	0	0	1	1.009926	-8.447309	110	1	1.1	0.9;
	3347	1	0	0	0	0	1	1.073697	-13.973278	220	1	1.129602	0.9;
	3348	1	0	0	0	0	1	0.956995	-29.24601	150	1	1.1	0.891768;
	3349	1	2.502775	1.3	0	0	1	0.949194	-24.320657	110	1	1.1	0.9;
	3350	1	28.93208	10.5	0	0	1	1.012908	-21.158847	110	1	1.10664	0.9;
	3351	1	182.3	65.9	0	0	1	1.090832	-21.256486	380	1	1.157175	0.9;
	3352	2	62.66948	6.6	0	0	1	0.928433	-23.21615	110	1	1.1	0.9;
	3353	1	0	0	0	0	1	1.038395	-22.432084	150	1	1.1	0.9;
	3354	1	0	0	0	0	1	1.086411	-45.482002	380	1	1.146866	0.9;
	3355	1	0	0	0.108	0	1	0.97316	-24.417613	220	1	1.1	0.9;
	3356	2	0	0	0	0	1	1.071727	-40.704987	220	1	1.124624	0.9;
	3357	1	-17.6195	5.8	0	0	1	1.013653	-22.158841	150	1	1.1	0.9;
	3358	2	0	0	0	0	1	1.076067	11.390954	150	1	1.10789	0.9;
	3359	2	223.4	143.1	0	0	1	1.009909	-38.308132	220	1	1.1	0.9;
	3360	1	0	0	0	0	1	1.005901	-21.573799	150	1	1.1	0.9;
	3361	2	0	0	0	0	1	1.084737	-29.745117	380	1	1.171887	0.9;
	3362	2	0	0	0	0	1	1.058124	-19.746396	380	1	1.153257	0.9;
	3363	2	74.5827	12.3	0	0	1	1.073636	-23.112933	220	1	1.138689	0.9;
	3364	1	16.61843	-10.3	0	0	1	1.065624	-40.081415	220	1	1.123865	0.9;
	3365	2	0	0	0	0	1	0.982218	-19.819994	110	1	1.1	0.9;
	3366	1	0	0	0	0	1	1.058573	-23.87324	220	1	1.133091	0.9;
	3367	1	0	0	0	0	1	1.082158	-19.271456	220	1	1.118064	0.9;
	3368	1	0	0	0	0	1	1.042601	-15.698722	150	1	1.1	0.9;
	3369	1	72.68058	45	0	0	1	1.020285	-23.565546	150	1	1.1	0.9;
	3370	1	120.7	24.3	0	0	1	1.046839	-34.274081	220	1	1.109403	0.9;
	3371	1	2.502775	0	0	0	1	1.023485	-15.971256	110	1	1.1	0.9;
	3372	1	77.48591	17.3	0	0	1	1.044146	-15.012966	220	1	1.130988	0.9;
	3373	1	0	0	0	0	1	1.028239	-4.422104	220	1	1.1	0.9;
	3374	1	0	0	0	0	1	1.084155	-25.937614	220	1	1.14085	0.9;
	3375	2	0	0	0	0	1	1.0198	-34.894617	150	1	1.1	0.9;
	3376	2	0	0	0	0	1	1.066214	2.276817	150	1	1.138867	0.9;
	3377	1	24.02664	8	0	0	1	1.009112	-18.286031	110	1	1.1	0.9;
	3378	1	-2835.7	-818.9	0	20	1	1.094135	-7.913485	380	1	1.159247	0.9;
	3379	1	68.77625	13	0	0	1	1.064765	-44.015029	220	1	1.113301	0.9;
	3380	1	85.1	-122.4	0	0	1	1.053754	-12.574131	380	1	1.122559	0.9;
	3381	1	0	0	0	0	1	1.004304	-21.727199	220	1	1.1	0.9;
	3382	1	179.4	58.3	0	0	1	1.056327	-24.093965	220	1	1.130986	0.9;
	3383	1	4.404884	2.2	0	0	1	1.021214	-34.194667	110	1	1.1	0.9;
	3384	2	104	51.2	0	0	1	1.002727	-0.385926	220	1	1.105398	0.9;
	3385	1	0	0	0	0	1	1.050814	-44.51312	220	1	1.1	0.9;
	3386	1	65.1	26.4	0	0	1	0.969993	-25.610197	110	1	1.1	0.9;
	3387	2	0	0	0	0	1	1.050894	44.187988	380	1	1.121022	0.9;
	3388	2	0	0	0	0	1	1.031955	-7.849073	220	1	1.1	0.9;
	3389	1	50.9565	17	0	0	1	0.965981	-25.901439	110	1	1.1	0.9;
	3390	1	16.51831	7.1	0	0	1	1.00079	-35.428405	110	1	1.1	0.9;
	3391	1	165.3	89.2	0	0	1	1.046543	8.029156	380	1	1.121614	0.9;
	3392	1	67.77514	10.4	0	0	1	1.05887	-48.132635	220	1	1.116714	0.9;
	3393	1	0	0	0.296958	0	1	1.105204	-8.715155	380	1	1.15437	0.9;
	3394	1	2.302553	0.8	0	0	1	0.974734	-17.792305	110	1	1.1	0.9;
	3395	1	0	0	0.095843	0	1	1.057325	-12.180266	220	1	1.117417	0.9;
	3396	2	0	0	0	0	1	1.04408	-9.779599	110	1	1.1	0.9;
	3397	1	0	0	0	0	1	0.989052	-19.182315	110	1	1.1	0.9;
	3398	1	51.35694	-0.7	0	0	1	1.070654	-42.501841	220	1	1.116883	0.9;
	3399	1	0	0	0	0	1	0.953361	-25.943321	150	1	1.1	0.9;
	3400	1	45.15006	4.1	0	0	1	1.056969	-52.85065	220	1	1.1	0.9;
	3401	2	48.15339	13.2	0	0	1	1.080953	-3.911881	220	1	1.142014	0.9;
	3402	1	18.42042	6	0	0	1	0.942133	-29.637286	110	1	1.1	0.9;
	3403	2	0.300333	1.3	0	0	1	1.067545	60.894272	220	1	1.113256	0.9;
	3404	1	0	0	0	0	1	1.078472	-20.197481	220	1	1.141342	0.9;
	3405	2	0	0	0	0	1	1.080816	-0.257425	380	1	1.141857	0.9;
	3406	1	122.3	45.9	0	0	1	1.060307	-20.732583	220	1	1.1	0.9;
	3407	2	71.37914	10.9	0	0	1	1.025067	64.668813	150	1	1.1	0.9;
	3408	2	120.4	18.2	0	0	1	1.071874	-42.196835	220	1	1.121591	0.9;
	3409	1	5.506105	2	0	0	1	0.971131	-27.359587	110	1	1.1	0.9;
	3410	2	34.13785	11.8	0	0	1	1.037364	-21.159226	110	1	1.1	0.9;
	3411	1	39.44373	5.9	0	0	1	1.052926	-48.502116	220	1	1.11104	0.9;
	3412	1	194.6	48.3	0	0	1	0.987809	-24.980661	110	1	1.1	0.9;
	3413	1	115.1	16.4	0	0	1	1.06697	-16.961048	220	1	1.149203	0.9;
	3414	2	69.27681	44.1	0.24	0	1	1.000136	-21.299784	220	1	1.1	0.9;
	3415	1	2.502775	0	0	0	1	1.017567	-14.51231	110	1	1.1	0.9;
	3416	1	7.608436	0.5	0	0	1	0.968531	-20.7919	110	1	1.1	0.9;
	3417	2	0	0	0	20	1	1.048421	10.271012	380	1	1.107835	0.9;
	3418	1	-24.9276	11.8	0	0	1	1.039143	-31.479248	220	1	1.120242	0.9;
	3419	1	35.5394	11.7	0	0	1	0.980764	-25.222208	110	1	1.1	0.9;
	3420	1	4.00444	1.3	0	0	1	0.9541	-32.133171	110	1	1.1	0.9;
	3421	1	83.49257	17.4	0	0	1	1.09286	-17.961845	220	1	1.165857	0.9;
	3422	2	0	0	0	0	1	1.0155	49.268373	220	1	1.1	0.9;
	3423	1	163.5	44.9	0	0	1	1.072448	-20.178414	220	1	1.176422	0.9;
	3424	1	71.0788	16.2	0	0	1	1.050277	-16.387558	220	1	1.1	0.9;
	3425	2	484.3	168	0	0	1	1.039631	-4.835891	380	1	1.119617	0.9;
	3426	1	119.1	61.4	0	0	1	1.043068	-20.691062	150	1	1.1	0.9;
	3427	1	37.84196	23.4	0	0	1	1.00239	-20.015909	150	1	1.1	0.9;
	3428	1	18.01998	6.5	0	0	1	1.048768	-19.06868	110	1	1.111683	0.9;
	3429	1	33.03663	21.6	0	0	1	1.075142	-45.486564	220	1	1.13248	0.9;
	3430	1	19.82198	10.9	0	0	1	1.048363	-20.971457	220	1	1.118107	0.9;
	3431	2	0	0	0	0	1	1.053636	-42.627113	220	1	1.108456	0.9;
	3432	1	27.13008	8.9	0	0	1	0.953822	-27.617023	110	1	1.1	0.9;
	3433	1	34.13785	13.5	0	0	1	1.062556	-1.777142	150	1	1.138504	0.9;
	3434	1	2.202442	-48.8	0	0	1	1.053025	6.314511	380	1	1.11326	0.9;
	3435	1	0	0	0	0	1	1.094654	-29.275071	380	1	1.140775	0.9;
	3436	1	0	0	0	0	1	0.967053	-28.063171	150	1	1.1	0.9;
	3437	2	63.57048	22.9	0	0	1	1.015158	-22.144172	220	1	1.1	0.9;
	3438	1	0	0	0	0	1	1.066825	9.62827	150	1	1.1	0.9;
	3439	1	77.18558	25.4	0	0	1	0.997258	9.421188	220	1	1.1	0.9;
	3440	1	0	0	0	0	1	1.049761	-21.032225	220	1	1.107608	0.9;
	3441	1	0	0	0	0	1	1.03857	-21.641596	150	1	1.115106	0.9;
	3442	1	0	0	0	0	1	1.009539	39.734249	220	1	1.1	0.9;
	3443	1	52.45816	19	0	0	1	1.037787	-21.193353	110	1	1.1	0.9;
	3444	2	162.4	14.4	0	0	1	1.007545	-3.521072	220	1	1.1	0.9;
	3445	1	46.35139	86.9	0	0	1	1.049961	11.753126	380	1	1.110601	0.9;
	3446	1	0	0	0	64.52167	1	1.055236	-20.072994	220	1	1.143215	0.9;
	3447	2	86.59601	24	0	0	1	0.992894	14.756874	220	1	1.1	0.9;
	3448	1	23.92653	8.7	0	0	1	1.057436	-21.201925	110	1	1.115457	0.9;
	3449	1	0	0	0	0	1	1.036453	17.687708	380	1	1.1	0.9;
	3450	1	0	0	0	0	1	1.031647	-23.623828	220	1	1.1	0.9;
	3451	2	0	0	0	0	1	1.0126	-35.915539	150	1	1.1	0.9;
	3452	2	105.4	92.5	0	0	1	1.071955	-40.04127	220	1	1.133517	0.9;
	3453	2	210.9	73.8	0	0	1	1.050529	-19.896605	220	1	1.137314	0.9;
	3454	1	165.3	7.4	0	0	1	1.025117	-27.367363	380	1	1.136342	0.9;
	3455	1	0	0	0	0	1	0.989247	-20.660067	150	1	1.1	0.9;
	3456	1	2.803108	0.6	0	0	1	0.999498	-19.366314	110	1	1.1	0.9;
	3457	2	272.5	127.2	0	0	1	1.019947	19.290059	380	1	1.1	0.9;
	3458	1	0	0	0	0	1	1.085555	-37.047725	380	1	1.129406	0.9;
	3459	1	0	0	0	0	1	0.960802	-34.44008	110	1	1.1	0.9;
	3460	1	0	0	0.191685	0	1	1.061813	-22.032886	220	1	1.130665	0.9;
	3461	1	10.11121	5.6	0	0	1	0.97089	-24.429547	110	1	1.1	0.9;
	3462	1	8.909878	1.3	0	0	1	0.927191	-29.169688	110	1	1.1	0.9;
	3463	1	5.00555	2	0	0	1	0.989702	-22.779086	110	1	1.1	0.9;
	3464	1	258.3	166	0	0	1	0.994237	-17.494445	220	1	1.1	0.9;
	3465	2	0	0	0	0	1	1.103974	11.408652	380	1	1.131353	0.9;
	3466	2	306.5	57.7	0	0	1	1.081364	-33.273848	220	1	1.176226	0.9;
	3467	2	0	0	0	0	1	1.043	-4.699071	110	1	1.1	0.9;
	3468	1	159.9	10.6	0	0	1	1.080883	-35.505992	380	1	1.155946	0.9;
	3469	1	0	0	0	0	1	1.03048	-24.302047	150	1	1.1	0.9;
	3470	2	227.6	56.8	0	0	1	1.054809	-48.898878	220	1	1.109866	0.9;
	3471	1	130.5	24.7	0	0	1	1.047489	-21.033898	220	1	1.117111	0.9;
	3472	1	16.31809	2.8	0	0	1	1.002346	-22.314062	110	1	1.1	0.9;
	3473	1	0	0	0	0	1	1.012909	-47.526125	220	1	1.1	0.9;
	3474	1	0	0	0	0	1	1.10001	-14.656263	220	1	1.180629	0.9;
	3475	1	44.44928	10.9	0	0	1	1.053461	-24.620604	220	1	1.128354	0.9;
	3476	1	0	0	0	0	1	1.066871	-3.237849	150	1	1.137858	0.9;
	3477	1	5.305883	1.8	0	0	1	0.994	-29.58117	110	1	1.1	0.9;
	3478	2	105.6	70	0	0	1	1.0305	-35.113464	220	1	1.1	0.9;
	3479	1	227.6	37.8	0	0	1	1.053046	-42.203761	380	1	1.110368	0.9;
	3480	1	0	0	0	0	1	0.987295	-24.257215	150	1	1.1	0.9;
	3481	2	0	0	0	0	1	1.088395	-8.550356	380	1	1.143262	0.9;
	3482	1	0	0	0	0	1	1.054287	-18.000889	220	1	1.131401	0.9;
	3483	1	18.82087	6.8	0	0	1	0.973805	-18.109586	110	1	1.1	0.9;
	3484	1	0	0	0	0	1	1.024855	-25.563044	150	1	1.1	0.9;
	3485	1	0	0	0	0	1	1.043436	-13.242384	380	1	1.115672	0.9;
	3486	1	74.98314	46.4	0	0	1	1.038995	-22.498736	150	1	1.1	0.9;
	3487	1	130.1	19.3	0	0	1	1.004844	-27.483216	220	1	1.1	0.9;
	3488	1	14.21576	6.1	0	0	1	1.000063	-22.438522	110	1	1.1	0.9;
	3489	1	160	69.2	0	0	1	1.059604	-21.254364	220	1	1.136743	0.9;
	3490	1	180.4	119.2	0	20	1	1.049698	-9.825999	380	1	1.118548	0.9;
	3491	1	44.14895	10.6	0	0	1	1.00106	-40.003272	110	1	1.1	0.9;
	3492	1	36.74074	22.7	0	0	1	0.958127	-25.581572	150	1	1.1	0.9;
	3493	1	89.09879	27.5	0	0	1	1.002652	-33.538123	110	1	1.1	0.9;
	3494	1	28.83197	9.6	0	0	1	0.943726	-27.354013	110	1	1.1	0.9;
	3495	2	44.34917	12.8	0	0	1	1.079818	-39.927872	220	1	1.125612	0.9;
	3496	2	0	0	0	0	1	1.075727	-16.550563	220	1	1.142313	0.9;
	3497	1	6.407104	2.3	0	0	1	0.988637	-17.843057	110	1	1.1	0.9;
	3498	1	123.9	32.6	0	0	1	1.046583	-39.196522	220	1	1.1	0.9;
	3499	1	12.91432	4.7	0	0	1	1.075208	-0.570671	110	1	1.143764	0.9;
	3500	1	0	0	0	0	1	1.062438	9.534242	150	1	1.1	0.9;
	3501	1	6.807548	2.3	0	0	1	0.983813	-26.129652	110	1	1.1	0.9;
	3502	2	141	17.1	0	0	1	1.066946	4.670151	220	1	1.143104	0.9;
	3503	1	0	0	0	0	1	1.010255	-26.498574	150	1	1.1	0.9;
	3504	2	0	0	0	0	1	1.075726	-16.542288	220	1	1.142185	0.9;
	3505	2	73.08102	15.7	0	0	1	1.0124	63.174664	150	1	1.1	0.9;
	3506	1	0	0	0	0	1	1.066632	32.563222	380	1	1.140217	0.9;
	3507	2	0	0	0	0	1	1.094182	-8.289104	220	1	1.148165	0.9;
	3508	1	85.5	118.1	0	0	1	1.004584	12.444403	220	1	1.1	0.9;
	3509	2	0	0	0	0	1	1.077512	-25.28947	220	1	1.167332	0.9;
	3510	1	28.03108	8.8	0	0	1	1.005824	-29.641371	110	1	1.1	0.9;
	3511	1	0	0	0	0	1	1.098182	-21.885136	220	1	1.189616	0.9;
	3512	1	0	0	0	0	1	1.069582	-18.874338	150	1	1.1	0.9;
	3513	1	0	0	0	0	1	1.065889	-0.213565	220	1	1.138447	0.9;
	3514	1	0	0	0	0	1	1.039458	61.163548	220	1	1.1	0.9;
	3515	2	0	0	0	0	1	1.038027	-19.760861	110	1	1.1	0.9;
	3516	1	0	0	0	0	1	1.067228	3.438002	380	1	1.110861	0.9;
	3517	1	-18.3203	6	0	0	1	1.007889	-22.61275	150	1	1.1	0.9;
	3518	1	0	0	0	0	1	1.021563	-16.890943	110	1	1.1	0.9;
	3519	1	0	0	0	0	1	1.070725	-13.937309	220	1	1.140353	0.9;
	3520	2	210.3	23	0	0	1	1.041649	-27.4675	220	1	1.119307	0.9;
	3521	1	0	0	0	0	1	1.008251	-22.247869	150	1	1.1	0.9;
	3522	2	0	0	0	0	1	1.065098	0.990072	150	1	1.127249	0.9;
	3523	2	0	0	0	0	1	1.081368	0.17522	380	1	1.142925	0.9;
	3524	1	50.0555	23	0	0	1	1.002549	-47.901959	220	1	1.1	0.9;
	3525	1	58.96538	18.1	0	0	1	0.959887	-13.728823	110	1	1.1	0.9;
	3526	1	51.25683	-17.2	0	0	1	1.092506	-17.29639	380	1	1.160444	0.9;
	3527	1	11.61288	4.3	0	0	1	1.003066	-19.206954	110	1	1.1	0.9;
	3528	2	102.1	27.8	0	0	1	1.047227	-9.482902	220	1	1.119541	0.9;
	3529	1	31.73519	3.8	0	0	1	1.035667	-10.833757	110	1	1.1	0.9;
	3530	1	0	0	0	0	1	1.056877	17.014348	220	1	1.136886	0.9;
	3531	2	-1.09167	1.5	0	0	1	1.062867	-18.428352	150	1	1.1	0.9;
	3532	2	0	0	0	0	1	1.095247	-11.350929	220	1	1.170039	0.9;
	3533	1	0	0	0	0	1	0.958479	-22.978341	110	1	1.1	0.9;
	3534	1	56.26238	20.1	0	0	1	0.963782	-19.199129	110	1	1.1	0.9;
	3535	1	3.203552	1.2	0	0	1	0.985006	-18.109717	110	1	1.1	0.9;
	3536	1	293	-45.9	0	0	1	1.083045	-25.861494	380	1	1.141963	0.9;
	3537	1	0	0	0.237627	0	1	1.067114	-9.018516	380	1	1.1	0.9;
	3538	1	94.30456	-27.5	0	0	1	1.011927	10.672739	220	1	1.1	0.9;
	3539	1	-49.0544	1	0	0	1	1.04007	-18.610582	220	1	1.1	0.9;
	3540	2	0	0	0	0	1	1.057273	-19.458444	220	1	1.2	0.8;
	3541	1	-10.4115	3.4	0	0	1	1.030714	-20.287002	150	1	1.1	0.9;
	3542	1	4.605106	1.7	0	0	1	1.02397	-18.257279	110	1	1.110045	0.9;
	3543	1	0	0	0	0	1	1.02157	12.034734	220	1	1.1	0.9;
	3544	1	19.62176	-3.1	0	0	1	1.08117	-35.095568	220	1	1.160504	0.9;
	3545	1	8.909878	5.1	0	0	1	1.066771	-13.193712	110	1	1.131444	0.9;
	3546	1	0	0	0	0	1	1.067746	-14.352994	220	1	1.137924	0.9;
	3547	1	-272.202	6.6	0	0	1	1.086199	-28.443153	380	1	1.163664	0.9;
	3548	1	33.03663	12.1	0	0	1	0.971465	-35.488211	110	1	1.1	0.9;
	3549	1	23.42597	8.4	0	0	1	1.007494	-19.779663	110	1	1.1	0.9;
	3550	1	121.5	29.6	0	0	1	1.047416	-1.613868	150	1	1.110971	0.9;
	3551	1	5.105661	2.3	0	0	1	0.935968	-25.540106	110	1	1.1	0.9;
	3552	2	0	0	0	0	1	1.040182	-18.120507	220	1	1.106152	0.9;
	3553	1	0	0	0	0	1	1.038501	-25.219629	150	1	1.1	0.9;
	3554	1	20.12231	8.6	0	0	1	0.96984	-22.890323	110	1	1.1	0.9;
	3555	1	0	0	0	0	1	1.06179	-23.375912	220	1	1.136042	0.9;
	3556	1	7.608436	2.5	0	0	1	0.9431	-24.373006	110	1	1.1	0.9;
	3557	1	195.3	87.6	0	0	1	0.99535	10.385031	220	1	1.1	0.9;
	3558	1	0	0	0	0	1	1.078701	-12.893344	380	1	1.148918	0.9;
	3559	2	5.305883	1.9	0	0	1	0.950818	-24.828633	110	1	1.1	0.9;
	3560	1	0	0	0	0	1	1.08469	-9.064404	220	1	1.129553	0.9;
	3561	2	0	0	0	0	1	1.060158	20.472246	380	1	1.1	0.9;
	3562	1	-3.3	-0.7	0	0	1	1.073887	-14.208011	150	1	1.130274	0.9;
	3563	1	0	0	0	40.80943	1	1.078663	-20.066917	220	1	1.141433	0.9;
	3564	1	143.9	85.2	0	0	1	1.009743	-7.307578	220	1	1.1	0.9;
	3565	1	16.31809	5.3	0	0	1	0.935773	-27.870196	110	1	1.1	0.9;
	3566	1	0	0	0.100424	0	1	1.093076	-8.303431	220	1	1.147249	0.9;
	3567	2	66.67392	16.9	0	0	1	1.075045	-13.824356	220	1	1.147021	0.9;
	3568	1	20.32253	6.7	0	0	1	0.962106	-27.631111	110	1	1.1	0.9;
	3569	2	0	0	0	56.71428	1	1.07292	-2.600542	380	1	1.135446	0.9;
	3570	1	0	0	0	0	1	1.07657	-5.732215	380	1	1.135441	0.9;
	3571	2	0	0	0	0	1	1.050316	2.280779	380	1	1.1082	0.9;
	3572	1	39.34362	10.3	0	0	1	1.010233	-15.887108	110	1	1.1	0.9;
	3573	2	0	0	0	0	1	1.061184	13.567723	380	1	1.11978	0.9;
	3574	1	11.11232	4	0	0	1	1.011415	-16.851129	110	1	1.1	0.9;
	3575	1	227.4	40.3	0	0	1	1.013293	-24.476413	110	1	1.1	0.9;
	3576	1	89.299	25.1	0	0	1	1.076148	-12.518107	220	1	1.142269	0.9;
	3577	1	29.73297	0	0	0	1	1.006941	-38.602553	110	1	1.1	0.9;
	3578	1	35.63951	22	0	0	1	0.956949	-29.228607	150	1	1.1	0.892108;
	3579	1	105.2	24.2	0	0	1	1.080759	-4.733418	220	1	1.134831	0.9;
	3580	1	48.95428	11.6	0	0	1	1.068029	-19.233796	220	1	1.136371	0.9;
	3581	2	123.9	46.5	0	0	1	1.013909	-5.513233	220	1	1.1	0.9;
	3582	1	0	0	0	0	1	1.057866	62.703284	150	1	1.1	0.9;
	3583	1	0	0	0	0	1	1.050552	-30.688132	220	1	1.108052	0.9;
	3584	1	37.94207	23.5	0	0	1	1.002231	-23.130116	150	1	1.1	0.9;
	3585	1	236.2	45.1	0	0	1	1.001058	7.062267	220	1	1.1	0.9;
	3586	1	0	0	0	0	1	1.076857	-4.443736	220	1	1.135978	0.9;
	3587	1	0	0	0	0	1	1.064768	-29.39372	110	1	1.129418	0.9;
	3588	1	83.19224	11.4	0	0	1	1.008643	-39.174216	220	1	1.1	0.9;
	3589	2	101.2	45.2	0	0	1	1.0215	1.491485	220	1	1.122274	0.9;
	3590	2	62.4	123.4	0	0	1	1.048909	-7.958413	220	1	1.119765	0.9;
	3591	1	4.805328	1.3	0	0	1	0.954148	-27.099792	110	1	1.1	0.9;
	3592	1	1.101221	0	0	0	1	1.031627	-31.733573	220	1	1.1	0.9;
	3593	1	0	0	0	0	1	1.031928	-11.500315	220	1	1.1	0.9;
	3594	1	0	0	0	0	1	1.048321	-17.731256	220	1	1.108188	0.9;
	3595	1	71.8797	44.5	0	0	1	0.9751	-23.973836	150	1	1.1	0.9;
	3596	1	20.0222	15	0	0	1	0.995271	10.381773	220	1	1.1	0.9;
	3597	2	74.28236	-11.3	0	0	1	1.074727	-34.666409	220	1	1.132018	0.9;
	3598	2	0	0	0	0	1	1.085273	-10.7949	220	1	1.139244	0.9;
	3599	1	0.400444	0.9	0	0	1	1.099322	35.801664	380	1	1.129731	0.9;
	3600	2	90.4	53.4	0	0	1	1.048044	-12.578724	220	1	1.131664	0.9;
	3601	1	5.105661	3.8	0	0	1	0.94802	-27.275943	110	1	1.1	0.9;
	3602	2	174.4	45.5	0	0	1	1.059995	7.895493	220	1	1.143166	0.9;
	3603	1	310.8	162.6	0	0	1	1.022115	-51.464733	220	1	1.1	0.9;
	3604	1	0	0	0	0	1	0.978395	-14.355986	110	1	1.1	0.9;
	3605	1	0	0	0	0	1	1.059793	-22.957662	220	1	1.133754	0.9;
	3606	1	0	0	0	0	1	1.024049	-40.441555	220	1	1.1	0.9;
	3607	2	0	0	0	0	1	1.074637	-33.896706	220	1	1.155072	0.9;
	3608	1	0	0	0	0	1	1.08481	7.909053	150	1	1.139195	0.9;
	3609	1	42.34695	7.2	0	0	1	1.030555	-43.671454	220	1	1.1	0.9;
	3610	2	22.62509	13.5	0	0	1	1.075322	-22.490307	220	1	1.159621	0.9;
	3611	2	127.8	37.6	0	0	1	1.054868	-43.862044	380	1	1.125562	0.9;
	3612	1	1.301443	0	0	0	1	0.992684	-22.798285	110	1	1.1	0.9;
	3613	1	4.204662	0.9	0	0	1	0.974334	-27.319338	110	1	1.1	0.9;
	3614	1	220.7	-14.8	0	0	1	1.076577	-38.910371	380	1	1.128044	0.9;
	3615	1	3.904329	1.4	0	0	1	1.071183	-17.193598	110	1	1.121792	0.9;
	3616	2	0	0	0	0	1	1.083103	16.348731	380	1	1.127979	0.9;
	3617	1	0	0	0	0	1	1.067194	-9.699079	220	1	1.134519	0.9;
	3618	1	5.105661	1.1	0	0	1	0.946776	-26.482993	110	1	1.1	0.9;
	3619	2	218.5	93.7	0	0	1	1.063682	-8.552007	220	1	1.138831	0.9;
	3620	1	419.5	114.1	0	0	1	1.091642	-17.694083	380	1	1.173596	0.9;
	3621	2	0	0	0	0	1	1.081274	-26.218023	220	1	1.131776	0.9;
	3622	2	0	0	0	0	1	1.067558	3.228612	150	1	1.138274	0.9;
	3623	1	148	26.1	0	0	1	1.009752	-23.361745	220	1	1.1	0.9;
	3624	1	159.5	42	0	0	1	1.063096	-14.469385	220	1	1.131373	0.9;
	3625	1	29.83308	14.4	0	0	1	1.084268	-2.303352	220	1	1.14492	0.9;
	3626	1	0	0	0	0	1	1.094357	-28.548815	380	1	1.150492	0.9;
	3627	1	0	0	0.357007	0	1	1.074232	-15.124951	380	1	1.12102	0.9;
	3628	2	0	0	0	0	1	1.0436	14.72913	150	1	1.1	0.9;
	3629	1	33.33696	20.7	0	0	1	1.018979	-19.016221	150	1	1.1	0.9;
	3630	1	77.18558	26.5	0	0	1	1.07378	0.0421	220	1	1.148624	0.9;
	3631	2	14.61621	3.7	0	0	1	0.991963	-23.10158	110	1	1.1	0.9;
	3632	2	0	0	0	0	1	1.039364	-29.376474	220	1	1.1	0.9;
	3633	1	22.7252	5.8	0	0	1	1.014199	-25.89686	110	1	1.1	0.9;
	3634	1	7.00777	1.8	0	0	1	1.092266	-44.323316	380	1	1.149387	0.9;
	3635	2	280.1	1.7	0	0	1	1.067682	4.421984	220	1	1.143482	0.9;
	3636	1	34.6384	21.4	0	0	1	0.992061	-20.283226	150	1	1.1	0.9;
	3637	1	0	0	0	0	1	1.020688	-21.042891	150	1	1.1	0.9;
	3638	1	36.34029	10.4	0	0	1	1.085047	-3.381952	220	1	1.149107	0.9;
	3639	1	-7.90877	2.6	0	0	1	1.016343	-24.391481	150	1	1.1	0.9;
	3640	2	148.9	23.5	0	0	1	1.083955	-5.197578	220	1	1.140876	0.9;
	3641	1	26.12897	9.7	0	0	1	1.052116	-11.919945	220	1	1.124105	0.9;
	3642	1	0	0	0	0	1	1.038403	-22.44711	150	1	1.1	0.9;
	3643	1	110.8	36.9	0	0	1	1.096946	-6.233338	220	1	1.164151	0.9;
	3644	1	0	0	0	0	1	1.051766	-14.811315	380	1	1.116967	0.9;
	3645	2	98.9	21.3	0	0	1	1.032109	19.029613	220	1	1.1	0.9;
	3646	1	0	0	0	0	1	1.06829	-11.615576	380	1	1.121323	0.9;
	3647	1	0	0	0	0	1	1.030629	-22.512233	150	1	1.1	0.9;
	3648	2	182.1	32.9	0	0	1	1.032834	-32.131615	220	1	1.123105	0.9;
	3649	2	0	0	0	0	1	1.06448	9.242403	150	1	1.1	0.9;
	3650	1	0	0	0.715884	0	1	1.094571	-17.099839	220	1	1.204951	0.9;
	3651	2	22.92542	8.6	0	0	1	1.061727	-4.32901	110	1	1.136863	0.9;
	3652	1	19.32142	6	0	0	1	1.011136	-36.374143	110	1	1.1	0.9;
	3653	1	144.8	39.7	0	0	1	1.053759	-49.171378	220	1	1.111003	0.9;
	3654	2	0	0	0	0	1	1.020364	64.057163	220	1	1.1	0.9;
	3655	1	0	0	0	0	1	1.027163	-50.923902	150	1	1.1	0.9;
	3656	1	0	0	0	63.87749	1	1.075765	-12.897718	150	1	1.129951	0.9;
	3657	1	0	0	0	0	1	0.963264	-26.444045	110	1	1.1	0.9;
	3658	2	0.300333	10.5	0	0	1	1.067909	-38.757609	220	1	1.115141	0.9;
	3659	1	0	0	0	0	1	1.042305	-35.721932	150	1	1.1	0.9;
	3660	1	0	0	0	0	1	1.046549	-20.83721	150	1	1.1	0.9;
	3661	1	394.9	-17.3	0	0	1	1.074898	-37.789987	380	1	1.144447	0.9;
	3662	1	0	0	0	0	1	1.083574	-45.423888	380	1	1.146228	0.9;
	3663	2	0	0	0	0	1	1.043276	7.261274	380	1	1.1	0.9;
	3664	1	0	0	0	0	1	1.08898	-27.054429	380	1	1.152441	0.9;
	3665	1	-1270.6	-476.3	0	0	1	1.068538	2.613738	380	1	1.134996	0.9;
	3666	2	0	0	0	0	1	1.081342	0.984154	380	1	1.132387	0.9;
	3667	1	0	0	0	0	1	1.085184	-3.371793	220	1	1.149234	0.9;
	3668	2	1.701887	-5.9	0	0	1	1.090552	-37.139536	380	1	1.136871	0.9;
	3669	1	0	0	0	0	1	1.092309	-24.093282	380	1	1.14898	0.9;
	3670	2	93.70389	28.5	0	0	1	1.030273	-23.158216	110	1	1.1	0.9;
	3671	1	20.9232	10.3	0	0	1	0.940967	-24.81399	110	1	1.1	0.9;
	3672	1	0	0	0	0	1	1.030691	-21.834225	150	1	1.107752	0.9;
	3673	1	0	0	0	0	1	1.065585	-10.238579	220	1	1.141507	0.9;
	3674	1	265.7	113.5	0	0	1	1.048489	-10.805049	220	1	1.116808	0.9;
	3675	1	54.96094	31.7	0	0	1	0.963242	58.646711	220	1	1.1	0.9;
	3676	1	24.02664	7.9	0	0	1	0.983852	-21.579956	110	1	1.1	0.9;
	3677	1	58.56493	15.9	0	0	1	1.065966	-19.71325	220	1	1.128819	0.9;
	3678	1	0	0	0	0	1	1.088645	-27.736805	380	1	1.169901	0.9;
	3679	1	352.6	193.5	0	0	1	1.083005	-20.922373	380	1	1.171797	0.9;
	3680	1	15.5172	4.1	0	0	1	0.949911	-31.576791	110	1	1.1	0.9;
	3681	1	0	0	0	0	1	1.064322	-21.529542	220	1	1.141386	0.9;
	3682	1	37.24129	40.5	0	0	1	1.090801	-19.787733	380	1	1.155098	0.9;
	3683	1	0	0	0	0	1	1.07213	30.569243	380	1	1.12757	0.9;
	3684	1	1.902109	0.7	0	0	1	1.050963	-10.176983	110	1	1.129647	0.9;
	3685	1	0	0	0	0	1	0.949039	-24.354578	110	1	1.1	0.9;
	3686	2	0	0	0	0	1	1.060105	-0.462704	380	1	1.123858	0.9;
	3687	1	0	0	0	0	1	1.06324	-15.145554	220	1	1.117907	0.9;
	3688	1	0	0	0	0	1	1.097452	-3.100422	220	1	1.154837	0.9;
	3689	1	0	0	0	0	1	1.063148	-9.400294	220	1	1.130492	0.9;
	3690	2	93.60378	-21	0	0	1	1.092336	14.75044	220	1	1.171462	0.9;
	3691	2	0	0	0	0	1	1.0845	50.910687	380	1	1.114764	0.9;
	3692	1	66.07326	12.5	0	0	1	1.071965	-42.498125	220	1	1.119697	0.9;
	3693	1	0	0	0	0	1	1.081344	-17.543032	220	1	1.146122	0.9;
	3694	1	347.7	96.6	0	0	1	1.082174	-15.882183	380	1	1.150267	0.9;
	3695	2	123.1	23.5	0	0	1	1.075227	-9.059483	220	1	1.141576	0.9;
	3696	1	74.78291	-31.4	0	0	1	1.052651	5.452255	110	1	1.121426	0.9;
	3697	1	0	0	0	0	1	1.028819	-43.806995	220	1	1.1	0.9;
	3698	1	0	0	0	0	1	1.04025	25.763647	380	1	1.130639	0.9;
	3699	1	0	0	0	0	1	1.066186	-14.05191	220	1	1.148337	0.9;
	3700	1	6.306993	0	0	0	1	1.039396	-44.332749	220	1	1.1	0.9;
	3701	2	74.08214	17.1	0	0	1	1.092661	-10.335431	110	1	1.168945	0.9;
	3702	1	131.3456	63.6	0	0	1	1.070884	-1.886866	220	1	1.137634	0.9;
	3703	1	14.71632	5.3	0	0	1	1.041278	-21.517762	110	1	1.146066	0.9;
	3704	1	0	0	0	0	1	1.052396	-15.260266	150	1	1.1	0.9;
	3705	1	18.2202	2.3	0	0	1	0.970924	-28.074146	110	1	1.1	0.9;
	3706	2	3.704107	1.1	0	0	1	1.025679	-23.473784	110	1	1.12657	0.9;
	3707	1	35.43929	22	0	0	1	0.987765	-20.844105	150	1	1.1	0.9;
	3708	1	4.204662	2	0	0	1	0.979002	-12.010178	110	1	1.1	0.9;
	3709	1	0	0	0	0	1	1.033241	-20.273278	150	1	1.112726	0.9;
	3710	1	0	0	0	0	1	1.026472	-23.291768	110	1	1.1	0.9;
	3711	2	0	-0.1	0	0	1	1.085591	-13.332113	220	1	1.1701	0.9;
	3712	2	0	0	0	0	1	1.025864	5.324589	220	1	1.1	0.9;
	3713	1	0	0	0	0	1	1.077886	14.517397	380	1	1.124928	0.9;
	3714	1	0	0	0.16999	0	1	1.059404	-27.484015	380	1	1.131278	0.9;
	3715	1	0	0	0	0	1	1.027678	-20.622708	150	1	1.105538	0.9;
	3716	1	0	0	0	0	1	1.061194	-48.158996	220	1	1.118506	0.9;
	3717	1	0	0	0	0	1	1.08284	10.639086	150	1	1.135689	0.9;
	3718	1	113	57.3	0	0	1	1.011539	-24.859836	150	1	1.1	0.9;
	3719	2	135.9	4	0	0	1	1.064455	-8.499647	220	1	1.139569	0.9;
	3720	2	0	0	0	0	1	1.009818	39.753919	220	1	1.1	0.9;
	3721	1	0	0	0	0	1	1.081946	15.961901	380	1	1.12696	0.9;
	3722	1	0	0	0	0	1	1.073517	-18.270087	220	1	1.108941	0.9;
	3723	1	195.2	92.2	0	0	1	1.001476	-19.779203	220	1	1.1	0.9;
	3724	1	0	0	0	0	1	1.084051	-18.258702	380	1	1.155187	0.9;
	3725	1	0	0	0	0	1	1.024501	15.617797	220	1	1.1	0.9;
	3726	1	40.24462	24.9	0	0	1	1.045559	-20.095596	150	1	1.1	0.9;
	3727	2	225.8	123	0	0	1	1.022909	-39.727421	220	1	1.120307	0.9;
	3728	1	65.27237	6	0	0	1	1.062241	-52.926721	220	1	1.1	0.9;
	3729	1	47.65283	7.1	0	0	1	1.039981	-44.216049	220	1	1.105095	0.9;
	3730	1	61.66837	28.3	0	0	1	0.981535	-39.866648	220	1	1.1	0.9;
	3731	1	173	11.3	0	0	1	1.058735	-15.483625	220	1	1.122947	0.9;
	3732	1	0	0	0	0	1	1.096487	-10.684474	220	1	1.171493	0.9;
	3733	1	128	72.9	0	0	1	1.063115	20.054679	380	1	1.132548	0.9;
	3734	1	5.305883	2	0	0	1	0.961019	-13.60166	110	1	1.1	0.9;
	3735	1	81.0899	12.5	0	0	1	1.074832	-20.610051	220	1	1.154232	0.9;
	3736	1	40.34473	10.4	0	0	1	1.048112	6.86009	110	1	1.122092	0.9;
	3737	1	0.500555	12.4	0	0	1	1.066602	-0.139882	220	1	1.138952	0.9;
	3738	1	0	0	0	39.30328	1	0.974397	53.78488	220	1	1.1	0.9;
	3739	1	0	0	0	0	1	0.988848	-17.086098	150	1	1.1	0.9;
	3740	1	13.01443	3.8	0	0	1	1.077565	0.676145	220	1	1.151136	0.9;
	3741	1	0	0	0.00501	0	1	1.074107	-11.681952	380	1	1.137936	0.9;
	3742	1	64.77181	6.4	0	0	1	1.014201	-40.702792	220	1	1.1	0.9;
	3743	1	0	0	0	0	1	1.057548	-15.161686	380	1	1.12363	0.9;
	3744	1	9.910988	5.4	0	0	1	1.059024	-23.342759	220	1	1.133098	0.9;
	3745	1	0	0	0.08753	0	1	1.052521	-7.619846	220	1	1.111949	0.9;
	3746	1	2.202442	0.5	0	0	1	1.023883	-46.27732	150	1	1.1	0.9;
	3747	1	0	0	0	0	1	1.056585	31.262238	380	1	1.114337	0.9;
	3748	2	0	0	0	0	1	1.070909	-13.878817	220	1	1.140415	0.9;
	3749	1	0	0	0	0	1	1.088154	44.187718	380	1	1.115964	0.9;
	3750	2	0	0	0	0	1	1.054954	-17.922363	150	1	1.10961	0.9;
	3751	1	0	0	0.277056	0	1	1.024935	15.108963	380	1	1.1	0.9;
	3752	2	0	0	0	0	1	0.951818	40.009831	220	1	1.1	0.9;
	3753	1	0	0	0	0	1	0.998005	-23.696172	150	1	1.1	0.9;
	3754	2	232.1	-6.9	0	0	1	1.029351	-41.60243	220	1	1.1	0.9;
	3755	2	0	0	0	0	1	1.060789	-8.129438	380	1	1.119882	0.9;
	3756	1	0	0	0	0	1	1.082716	-15.99577	220	1	1.166556	0.9;
	3757	1	39.1434	7.7	0	0	1	1.06857	-19.177059	220	1	1.136874	0.9;
	3758	2	0	0	0	0	1	1.062773	-14.1926	220	1	1.119617	0.9;
	3759	2	0	0	0	74.47162	1	1.040337	61.753812	150	1	1.1	0.9;
	3760	1	0	0	0	0	1	1.059858	9.042514	150	1	1.1	0.9;
	3761	2	39.84418	17.6	0	0	1	0.996727	-21.542915	110	1	1.1	0.9;
	3762	1	25.42819	7.9	0	0	1	1.023686	7.940798	110	1	1.105419	0.9;
	3763	1	145.8	42.4	0	0	1	1.039069	-45.937266	220	1	1.1	0.9;
	3764	1	0	0	0	0	1	1.074683	-32.93662	220	1	1.177143	0.9;
	3765	1	13.31476	4.4	0	0	1	0.956189	-26.43846	110	1	1.1	0.9;
	3766	1	13.6414	9	0	0	1	1.010255	-26.498574	150	1	1.1	0.9;
	3767	1	137.8	22.9	0	0	1	1.05456	-30.014381	220	1	1.117622	0.9;
	3768	1	0	0	0	0	1	1.026799	5.962987	380	1	1.1	0.9;
	3769	1	0	0	0	0	1	1.034312	-19.471128	220	1	1.1	0.9;
	3770	2	0	0	0	0	1	1.040945	-20.306635	380	1	1.139683	0.9;
	3771	1	0	0	0	0	1	1.077334	-18.7962	380	1	1.150871	0.9;
	3772	1	11.11232	4.1	0	0	1	1.08734	-1.055089	220	1	1.152178	0.9;
	3773	1	310.2	4.5	0	0	1	1.050678	-20.385913	220	1	1.13925	0.9;
	3774	2	128.3	37.3	0	0	1	1.042791	61.914452	150	1	1.1	0.9;
	3775	1	246.9	20.3	0	0	1	1.000441	10.623975	220	1	1.1	0.9;
	3776	1	0	0	0	0	1	1.057698	-30.202046	380	1	1.123914	0.9;
	3777	2	0	0	0	0	1	1.025867	-46.195649	220	1	1.1	0.9;
	3778	1	6.707437	1.3	0	0	1	1.03468	-11.144437	110	1	1.1	0.9;
	3779	1	1.902109	9.2	0	0	1	1.020246	-19.541099	220	1	1.1	0.9;
	3780	1	91.4	-88.2	0	0	1	1.050539	-1.068817	220	1	1.118678	0.9;
	3781	1	71.67947	14.8	0	0	1	1.012941	-47.518358	220	1	1.1	0.9;
	3782	1	0	0	0	0	1	1.049847	57.88708	220	1	1.1	0.9;
	3783	1	110.7	94.3	0	0	1	1.001635	10.627699	220	1	1.1	0.9;
	3784	1	15.01665	7	0	0	1	0.978635	-22.5794	110	1	1.1	0.9;
	3785	1	0.2	109.4	0	0	1	1.072302	59.240391	380	1	1.1	0.9;
	3786	1	0	0	0	0	1	1.076576	-25.454076	220	1	1.166575	0.9;
	3787	1	121.4	40	0	0	1	1.054879	-14.762284	380	1	1.118078	0.9;
	3788	1	175.2	70.7	0	0	1	1.091052	-20.9094	380	1	1.16265	0.9;
	3789	1	0	0	0	11.43648	1	1.028721	-23.034342	150	1	1.1	0.9;
	3790	1	29.03219	10.1	0	0	1	1.03492	-15.23592	110	1	1.125031	0.9;
	3791	1	0	0	0	0	1	1.095122	-14.520712	380	1	1.155219	0.9;
	3792	2	27.02997	69.9	0	0	1	1.055273	-0.189652	220	1	1.126644	0.9;
	3793	1	0	0	0	0	1	1.047501	-14.859142	380	1	1.108864	0.9;
	3794	1	325.2	38.1	0	0	1	1.085132	-40.714639	380	1	1.129196	0.9;
	3795	1	0	0	0	0	1	1.010428	-19.574142	150	1	1.1	0.9;
	3796	1	50.85638	-23	0	0	1	1.068189	-18.836288	220	1	1.109923	0.9;
	3797	1	200.2	90.6	0	0	1	1.013096	-47.479426	220	1	1.1	0.9;
	3798	1	62.06882	20	0	0	1	0.953204	-27.664137	110	1	1.1	0.9;
	3799	1	0	0	0	0	1	1.008252	-28.370538	220	1	1.1	0.9;
	3800	1	7.908769	3.8	0	0	1	1.013808	-22.16545	110	1	1.1	0.9;
	3801	2	0	0	0	0	1	0.966545	-31.741052	110	1	1.1	0.9;
	3802	1	0	0	0	0	1	0.983587	-23.595676	150	1	1.1	0.9;
	3803	2	0	0	0	0	1	1.101868	45.311653	380	1	1.111814	0.9;
	3804	1	144.5	40.9	0	0	1	1.047328	-12.551433	220	1	1.1	0.9;
	3805	1	20.22242	5.7	0	0	1	1.062628	-40.401018	220	1	1.108744	0.9;
	3806	2	0	0	0	0	1	1.096211	-37.018911	380	1	1.161828	0.9;
	3807	2	0	0	0	0	1	1.007364	-23.796828	110	1	1.1	0.9;
	3808	1	0	0	0	0	1	1.066762	-16.192717	220	1	1.120867	0.9;
	3809	1	195.6	-1	0	0	1	1.045876	-41.672624	220	1	1.107796	0.9;
	3810	2	29.3	90.5	0	0	1	1.063474	-39.969767	380	1	1.123944	0.9;
	3811	2	43.14784	22.8	0	0	1	0.985	-5.365432	110	1	1.1	0.9;
	3812	1	26.72964	9.6	0	0	1	1.028911	-16.650387	110	1	1.129941	0.9;
	3813	1	87.49701	21.3	0	0	1	1.05328	-24.641276	220	1	1.128381	0.9;
	3814	1	0	0	0	0	1	1.061693	-22.104906	220	1	1.139338	0.9;
	3815	1	0	0	0.095843	0	1	1.064474	-14.293668	220	1	1.129111	0.9;
	3816	1	59.1656	19.3	0	0	1	1.055946	-0.080088	150	1	1.118789	0.9;
	3817	2	205.3	32.6	0	0	1	1.030601	-25.385535	220	1	1.128956	0.9;
	3818	1	0	0	0	0	1	1.076708	-9.123437	380	1	1.152083	0.9;
	3819	1	12.7141	2.5	0	0	1	0.959653	-26.60083	110	1	1.1	0.9;
	3820	2	0	0	0	0	1	1.082763	16.845003	380	1	1.130593	0.9;
	3821	1	0	0	0	25.22114	1	0.985059	-23.231775	150	1	1.1	0.9;
	3822	2	1110.6	290.8	0	0	1	1.047368	5.606969	380	1	1.1	0.9;
	3823	2	0	0	0	0	1	1.092148	-7.575205	380	1	1.146036	0.9;
	3824	1	9.00999	4.9	0	0	1	1.079006	0.891541	220	1	1.163058	0.9;
	3825	2	0	0	0	0	1	0.993211	66.727836	380	1	1.1	0.9;
	3826	1	318.3	76.8	0.147737	0	1	1.063701	-19.150169	380	1	1.116333	0.9;
	3827	1	53.85971	18.7	0	0	1	1.040959	-35.030321	110	1	1.1	0.9;
	3828	1	0	0	0	0	1	1.019307	-29.651567	150	1	1.1	0.9;
	3829	2	52.45816	5.2	0	0	1	1.080318	-35.793108	220	1	1.124917	0.9;
	3830	1	5.806438	1.7	0	0	1	1.04835	-12.376678	220	1	1.133697	0.9;
	3831	1	0	0	0	0	1	1.07144	-0.428931	380	1	1.141151	0.9;
	3832	2	67.07436	34	0	0	1	1.089	-19.865745	220	1	1.1	0.9;
	3833	1	0	0	0	0	1	1.080086	-17.643828	380	1	1.162481	0.9;
	3834	1	0	0	0	0	1	1.082879	16.336764	380	1	1.128069	0.9;
	3835	2	-2.29168	0.3	0	0	1	1.046867	-19.02746	150	1	1.1	0.9;
	3836	2	0	0	0	0	1	1.089227	-21.241577	220	1	1.126635	0.9;
	3837	1	32.33585	14.1	0	0	1	1.008831	-22.630018	110	1	1.1	0.9;
	3838	1	-806.3	-50.1	0	0	1	1.101507	62.348958	380	1	1.118276	0.9;
	3839	1	0	0	0	0	1	1.084225	-21.835902	220	1	1.169005	0.9;
	3840	2	0	0	0	0	1	1.045861	4.674371	110	1	1.109367	0.9;
	3841	1	0	0	0.120317	0	1	1.004933	66.711146	380	1	1.1	0.9;
	3842	1	0	0	0	0	1	1.078	28.771281	220	1	1.116689	0.9;
	3843	1	8.409324	2.5	0	0	1	0.988984	-27.634709	110	1	1.1	0.9;
	3844	1	27.63063	9.2	0	0	1	0.946964	-27.282744	110	1	1.1	0.9;
	3845	2	0	0	0	0	1	1.027489	-20.035917	150	1	1.1	0.9;
	3846	1	0	0	0	0	1	1.026423	-15.435794	110	1	1.1	0.9;
	3847	2	0	0	0	0	1	1.07619	-34.635339	220	1	1.124612	0.9;
	3848	1	0	0	0	0	1	1.07159	-14.934968	380	1	1.122825	0.9;
	3849	2	0	0	0	0	1	0.963793	-19.836699	110	1	1.1	0.9;
	3850	2	119.3	106.3	0	0	1	1.078267	64.156453	150	1	1.1	0.9;
	3851	1	0	0	0	0	1	0.940228	-31.944613	150	1	1.1	0.843794;
	3852	1	57.76405	16	0	0	1	1.00453	-40.623486	220	1	1.1	0.9;
	3853	3	0	0	0	0	1	1.083974	0	380	1	1.148229	0.9;
	3854	1	11.01221	3.1	0	0	1	1.029887	-23.280801	110	1	1.1	0.9;
	3855	1	0	0	0	0	1	1.043215	-20.959314	150	1	1.1	0.9;
	3856	1	0	0	0	0	1	1.093749	-16.346568	220	1	1.168096	0.9;
	3857	1	0	0	0	0	1	1.028617	-18.630928	220	1	1.138346	0.9;
	3858	1	19.52164	6.4	0	0	1	1.01908	-23.486433	150	1	1.1	0.9;
	3859	1	0	0	0	0	1	1.030819	-43.578188	220	1	1.1	0.9;
	3860	1	190.7	52	0	0	1	1.047842	-12.85216	220	1	1.133203	0.9;
	3861	1	44.34917	14.8	0	0	1	0.947071	-27.269458	110	1	1.1	0.9;
	3862	1	0	0	0	0	1	1.073698	-39.71224	380	1	1.132605	0.9;
	3863	2	271.9	100.3	0	0	1	1.048641	-23.694458	110	1	1.13066	0.9;
	3864	1	188.4777	128.3	0	0	1	1.053918	-27.023931	220	1	1.116508	0.9;
	3865	2	286.3	94.1	0	0	1	0.997658	12.496136	380	1	1.1	0.9;
	3866	1	24.92764	9.6	0	0	1	1.073626	-9.667542	220	1	1.144393	0.9;
	3867	1	24.22686	5.5	0	0	1	1.031095	-36.363067	110	1	1.1	0.9;
	3868	1	33.53718	20.8	0	0	1	0.942113	-33.245557	150	1	1.1	0.8222;
	3869	1	0	0	0	0	1	1.077131	20.531055	220	1	1.154971	0.9;
	3870	1	57.06327	15.6	0	0	1	1.049036	-49.861851	220	1	1.111758	0.9;
	3871	1	-10.7119	3.5	0	0	1	1.053234	-18.165553	150	1	1.108243	0.9;
	3872	1	47.15228	15.7	0	0	1	1.062895	-16.635169	220	1	1.12702	0.9;
	3873	1	0	0	0	0	1	1.042028	-9.955603	110	1	1.1	0.9;
	3874	1	105	18	0	0	1	1.050687	-49.415994	220	1	1.113253	0.9;
	3875	1	44.54939	27.6	0	0	1	0.99488	-23.99917	150	1	1.1	0.9;
	3876	1	0	0	0	0	1	1.03293	19.736789	220	1	1.1	0.9;
	3877	2	204.2	87.7	0	0	1	1.054759	9.623756	380	1	1.137409	0.9;
	3878	1	9.510545	5.9	0	0	1	0.982366	-23.700211	150	1	1.1	0.9;
	3879	2	0	0	0	0	1	1.0184	-37.213663	150	1	1.1	0.9;
	3880	1	0	0	0	0	1	1.04993	-15.217456	150	1	1.1	0.9;
	3881	1	0	0	0	0	1	1.034048	3.91066	380	1	1.105841	0.9;
	3882	1	0	0	0	0	1	1.005417	-20.211066	150	1	1.1	0.9;
	3883	2	0	0	0	0	1	0.988349	-19.311698	110	1	1.1	0.9;
	3884	1	0	0	0	0	1	0.975075	-23.978347	150	1	1.1	0.9;
	3885	1	70.77847	19.6	0	0	1	1.078943	-8.276975	220	1	1.145522	0.9;
	3886	1	303.7	247.8	0	0	1	1.025091	30.35014	220	1	1.1	0.9;
	3887	1	24.92764	8.2	0	0	1	0.94067	-28.449907	110	1	1.1	0.9;
	3888	1	128.7	25.6	0	0	1	1.033525	-26.973656	110	1	1.2	0.8;
	3889	1	66.97426	45.1	0	0	1	0.930821	-13.291965	110	1	1.1	0.9;
	3890	1	184.7	71.5	0	0	1	1.062967	1.01125	150	1	1.136889	0.9;
	3891	1	31.93541	10	0	0	1	1.004677	-15.749941	110	1	1.1	0.9;
	3892	2	65.37248	17.5	0	0	1	1.054952	-48.882528	220	1	1.112483	0.9;
	3893	1	5.706326	2.1	0	0	1	1.04198	-22.669645	110	1	1.145378	0.9;
	3894	1	0	0	0	0	1	1.077365	-14.191463	380	1	1.143135	0.9;
	3895	1	0	0	0	0	1	1.018405	-35.062022	150	1	1.1	0.9;
	3896	2	9.610656	62.4	0	0	1	1.034591	-30.301036	220	1	1.1	0.9;
	3897	1	11.91321	4.3	0	0	1	1.058925	-4.08419	110	1	1.134432	0.9;
	3898	1	0	0	0	0	1	1.083085	-11.591823	220	1	1.155633	0.9;
	3899	1	22.22464	6	0	0	1	1.018331	-15.459762	110	1	1.1	0.9;
	3900	1	0	0	0	0	1	1.060523	-32.534175	220	1	1.136168	0.9;
	3901	1	214.4	52.9	0	0	1	1.079739	-15.228107	220	1	1.12116	0.9;
	3902	1	0	0	0	0	1	1.095499	45.558148	380	1	1.107041	0.9;
	3903	2	47.85305	8.8	0	0	1	0.969689	67.720595	150	1	1.1	0.9;
	3904	1	0.500555	0.3	0	0	1	1.032928	-18.145154	150	1	1.1	0.9;
	3905	1	0	0	0	0	1	1.077202	-4.693793	220	1	1.13606	0.9;
	3906	1	0	0	0	11.1585	1	1.041436	-21.688367	150	1	1.1	0.9;
	3907	1	-1.90211	-28.1	0	0	1	1.096125	-36.684481	380	1	1.162513	0.9;
	3908	2	0.300333	16.8	0	0	1	1.084409	-32.170484	220	1	1.129887	0.9;
	3909	1	80.58935	7.3	0	0	1	1.051092	-49.564262	220	1	1.1	0.9;
	3910	1	63.87082	55.1	0	0	1	1.060371	19.875207	380	1	1.123466	0.9;
	3911	2	0	0	0	0	1	1.076143	4.796241	150	1	1.14638	0.9;
	3912	2	0	0	0	0	1	1.086974	57.283285	380	1	1.1	0.9;
	3913	1	26.62953	9.7	0	0	1	1.038292	-22.173538	110	1	1.138199	0.9;
	3914	1	0	0	0	20	1	1.063747	4.511367	380	1	1.129216	0.9;
	3915	1	0	0	0	0	1	1.019193	-18.995192	150	1	1.1	0.9;
	3916	1	81.49035	25.1	0	0	1	1.052473	-15.786395	220	1	1.122107	0.9;
	3917	1	0	0	0	0	1	1.078673	-20.067288	220	1	1.141475	0.9;
	3918	1	30.53385	9	0	0	1	1.052692	17.711595	110	1	1.126022	0.9;
	3919	1	2.402664	0.9	0	0	1	1.070063	-20.544144	110	1	1.127638	0.9;
	3920	1	0	0	0	0	1	1.0264	-30.060736	380	1	1.134227	0.9;
	3921	1	8.209102	2.7	0	0	1	0.958878	-31.644373	110	1	1.1	0.9;
	3922	1	60.0666	-4.7	0	0	1	1.011221	-19.454291	110	1	1.1	0.9;
	3923	1	0	0	0	0	1	1.06015	-25.367188	220	1	1.1	0.9;
	3924	1	8.909878	2.5	0	0	1	0.997249	-16.439426	110	1	1.1	0.9;
	3925	1	11.41265	-3.7	0	0	1	1.010453	-19.644006	110	1	1.1	0.9;
	3926	1	133.4	29.4	0	0	1	1.0736	-20.106255	220	1	1.138541	0.9;
	3927	1	26.12897	2.7	0	0	1	1.079743	-13.997453	220	1	1.151938	0.9;
	3928	1	-113.326	45.2	0	0	1	1.092396	-34.693565	380	1	1.141024	0.9;
	3929	1	-9.00999	21.8	0	0	1	1.043686	-20.653311	150	1	1.1	0.9;
	3930	1	0	0	0	0	1	1.017438	-36.584025	220	1	1.1	0.9;
	3931	1	51.05661	16.5	0	0	1	1.067732	8.321653	110	1	1.13747	0.9;
	3932	1	50.35583	24.3	0	0	1	1.07703	20.478832	220	1	1.154867	0.9;
	3933	1	0	0	0	0	1	1.026471	-24.426502	150	1	1.1	0.9;
	3934	2	50.0555	-30.4	0	0	1	1.045273	-19.771313	110	1	1.111696	0.9;
	3935	2	0	0	0	0	1	1.069342	-21.373989	380	1	1.123743	0.9;
	3936	1	0	0	0	0	1	0.998005	-9.696315	220	1	1.1	0.9;
	3937	1	0	0	0	0	1	1.091581	-20.460227	380	1	1.157751	0.9;
	3938	1	10.81199	3.6	0	0	1	0.935797	-34.76331	110	1	1.1	0.9;
	3939	1	0	0	0.059489	0	1	1.014445	-24.357863	220	1	1.1	0.9;
	3940	1	0	0	0	0	1	1.030658	-24.273923	150	1	1.1	0.9;
	3941	1	96.2	37.3	0	0	1	0.98395	-24.710057	110	1	1.1	0.9;
	3942	2	0	0	0	0	1	1.092763	-25.255419	380	1	1.147458	0.9;
	3943	2	0	0	0	0	1	1.054591	-10.111801	220	1	1.12134	0.9;
	3944	1	0	0	0	0	1	1.097207	36.015856	380	1	1.129752	0.9;
	3945	2	0	0	0	0	1	1.014227	-26.725579	220	1	1.1	0.9;
	3946	2	0	0	0.51412	485.2212	1	1.095447	55.752356	380	1	1.115237	0.9;
	3947	1	0	0	0.14848	0	1	1.077003	-18.805732	380	1	1.151257	0.9;
	3948	1	55.46149	8.4	0	0	1	1.049832	-49.655239	220	1	1.112488	0.9;
	3949	2	0	0	0	0	1	1.044921	7.483766	380	1	1.1	0.9;
	3950	2	92.40245	22	0	0	1	1.066318	-43.246925	220	1	1.114954	0.9;
	3951	1	0	0	0	0	1	1.051462	-15.07259	380	1	1.113968	0.9;
	3952	1	0	0	0	0	1	1.079226	20.57903	220	1	1.163926	0.9;
	3953	1	18.62065	6.8	0	0	1	0.975311	-17.714619	110	1	1.1	0.9;
	3954	1	14.31587	7.2	0	0	1	1.075911	-18.410025	220	1	1.126496	0.9;
	3955	1	43.54828	14.3	0	0	1	0.973339	-25.651445	110	1	1.1	0.9;
	3956	1	68.47592	-10.6	0	0	1	1.034792	-46.679354	150	1	1.1	0.9;
	3957	1	0	0	0	0	1	1.06523	20.598022	380	1	1.127154	0.9;
	3958	1	57	39.9	0	0	1	1.07755	-15.532525	380	1	1.141746	0.9;
	3959	1	0	0	0	0	1	1.08109	-41.139967	220	1	1.152543	0.9;
	3960	1	24.12675	29.4	0	0	1	1.088647	45.996413	380	1	1.1	0.9;
	3961	2	0	0	0	0	1	1.055909	61.549473	220	1	1.10929	0.9;
	3962	1	0	0	0	0	1	0.948388	-27.271074	110	1	1.1	0.9;
	3963	1	163.6343	23	0	0	1	1.072445	-28.863104	220	1	1.133187	0.9;
	3964	1	165.8	44	0	0	1	1.072007	-25.850951	220	1	1.163962	0.9;
	3965	2	0	0	0	0	1	1.026367	-27.007514	220	1	1.1	0.9;
	3966	1	65.07214	15	0	0	1	1.075498	-25.553377	220	1	1.165086	0.9;
	3967	1	22.12453	6.5	0	0	1	0.991352	-29.877064	110	1	1.1	0.9;
	3968	1	235	58.9	0	0	1	1.040209	-15.658925	220	1	1.127458	0.9;
	3969	1	5.105661	2.3	0	0	1	0.984768	-7.189316	110	1	1.1	0.9;
	3970	1	20.82309	8.2	0	0	1	1.034831	-48.015438	220	1	1.1	0.9;
	3971	1	0	0	0.053001	0	1	0.985551	-23.569968	220	1	1.1	0.9;
	3972	1	43.7485	17.3	0	0	1	1.063029	3.26198	150	1	1.133965	0.9;
	3973	1	0	0	0	0	1	1.066496	-36.264284	380	1	1.138024	0.9;
	3974	1	1.301443	0	0	0	1	0.991	-20.008916	110	1	1.1	0.9;
	3975	1	-13.7152	4.5	0	0	1	1.040288	-20.641387	150	1	1.1	0.9;
	3976	1	156.9	122.5	0	0	1	1.06719	-7.793532	380	1	1.135457	0.9;
	3977	1	23.22575	7.6	0	0	1	1.00425	-29.766196	110	1	1.1	0.9;
	3978	1	90.60045	15.8	0	0	1	1.070102	-24.427404	220	1	1.159807	0.9;
	3979	1	136.1	18.8	0	0	1	1.047906	-13.010346	220	1	1.1	0.9;
	3980	1	72.28014	6.6	0	0	1	1.051636	-49.808087	220	1	1.1	0.9;
	3981	1	2.803108	1	0	0	1	1.054306	-14.156331	110	1	1.138226	0.9;
	3982	1	28.63174	5.6	0	0	1	1.034041	-36.562107	110	1	1.1	0.9;
	3983	1	0	0	0	0	1	1.035034	-21.258987	150	1	1.1	0.9;
	3984	2	0	0	0	0	1	1.029	-21.657253	220	1	1.1	0.9;
	3985	2	0	0	0	0	1	1.043667	63.713755	150	1	1.1	0.9;
	3986	1	0	0	0	0	1	0.979105	-11.840289	110	1	1.1	0.9;
	3987	1	0	0	0	0	1	1.077542	1.833654	380	1	1.126131	0.9;
	3988	2	0	0	0	0	1	1.068963	0.904206	150	1	1.131074	0.9;
	3989	1	0	0	0	0	1	1.051954	-41.961877	380	1	1.109246	0.9;
	3990	1	35.23907	12.7	0	0	1	1.029912	-31.623105	110	1	1.1	0.9;
	3991	1	17.81976	5.9	0	0	1	0.970894	-25.209083	110	1	1.1	0.9;
	3992	1	33.73741	11.8	0	0	1	1.042345	-29.631758	110	1	1.1	0.9;
	3993	1	60.9676	19.7	0	0	1	1.016991	-32.765201	110	1	1.1	0.9;
	3994	1	10.9121	3.6	0	0	1	1.009187	-31.867895	110	1	1.1	0.9;
	3995	1	0	0	0.255999	0	1	1.068268	-19.329558	220	1	1.174534	0.9;
	3996	1	0	0	0	0	1	1.071153	13.160519	380	1	1.128921	0.9;
	3997	1	0	0	0	0	1	1.090753	-35.344485	380	1	1.146526	0.9;
	3998	1	0	0	0	0	1	1.007584	-15.550276	110	1	1.1	0.9;
	3999	1	62.06882	20.7	0	0	1	0.986813	-22.778239	220	1	1.1	0.9;
	4000	2	37.04107	15.2	0	0	1	1.099526	-40.339835	380	1	1.15474	0.9;
	4001	1	0	0	0.148479	0	1	1.087392	-20.406711	380	1	1.155786	0.9;
	4002	2	169.9	41.9	0	0	1	1.076864	-19.639272	220	1	1.140588	0.9;
	4003	2	0	0	0.045351	0	1	1.077053	-15.846428	380	1	1.143095	0.9;
	4004	1	193	17.6	0	0	1	1.041008	-43.15916	220	1	1.1	0.9;
	4005	2	0	0	0	0	1	1.093526	-23.497068	380	1	1.149478	0.9;
	4006	2	41.04551	12.5	0	0	1	1.022818	-21.595331	110	1	1.1	0.9;
	4007	2	0	0	0	0	1	1.097789	8.752249	380	1	1.129517	0.9;
	4008	1	0	0	0	0	1	1.042849	-18.928465	220	1	1.1	0.9;
	4009	1	0	0	0	0	1	1.100012	-14.656288	220	1	1.180652	0.9;
	4010	1	35.23907	12.3	0	0	1	1.014307	-29.510158	110	1	1.1	0.9;
	4011	1	0	0	0	0	1	1.001476	-22.902653	150	1	1.1	0.9;
	4012	1	-15.0166	4.9	0	0	1	0.994519	-16.682469	150	1	1.1	0.9;
	4013	1	29.1323	9.6	0	0	1	0.980571	-23.778608	110	1	1.1	0.9;
	4014	1	0	0	0	-131.662	1	1.090622	-36.977069	380	1	1.151566	0.9;
	4015	1	11.01221	3.3	0	0	1	1.014537	-36.845749	110	1	1.1	0.9;
	4016	2	0	0	0	0	1	1.066048	15.355651	150	1	1.105841	0.9;
	4017	1	45.04995	13.2	0	0	1	1.040412	-36.39555	220	1	1.148855	0.9;
	4018	1	0	0	0	0	1	1.040845	-7.843844	220	1	1.110059	0.9;
	4019	1	0	0	0	0	1	1.077458	9.793654	150	1	1.13107	0.9;
	4020	1	303.6	47.5	0	0	1	1.074397	-33.503691	380	1	1.152819	0.9;
	4021	1	30.0333	9.7	0	0	1	1.045331	5.236048	110	1	1.111911	0.9;
	4022	1	117.4	12.5	0	0	1	1.047135	-12.812109	220	1	1.1	0.9;
	4023	1	0	0	0	0	1	1.010248	-33.109723	220	1	1.1	0.9;
	4024	1	0	0	0	0	1	1.070451	-23.607654	220	1	1.171671	0.9;
	4025	2	0	0	0	0	1	1.082684	4.402257	380	1	1.152041	0.9;
	4026	1	32.83641	9.3	0	0	1	0.94362	-24.166616	110	1	1.1	0.9;
	4027	1	77.18558	47.8	0	0	1	0.994049	-24.64024	150	1	1.1	0.9;
	4028	1	0	0	0	64.3288	1	1.074532	-11.651248	150	1	1.134541	0.9;
	4029	1	0	0	0	0	1	1.064715	-14.797502	220	1	1.128712	0.9;
	4030	1	0	0	0	0	1	1.002109	10.648094	220	1	1.1	0.9;
	4031	2	496.5	62.4	0	0	1	1.042198	-46.216765	220	1	1.163803	0.9;
	4032	1	296.8	77.7	0	0	1	1.067982	-11.022866	150	1	1.134853	0.9;
	4033	2	0	0	0	0	1	1.049818	-39.007952	220	1	1.1	0.9;
	4034	1	0	0	0	0	1	1.041122	-18.611066	220	1	1.1	0.9;
	4035	1	19.52164	5.6	0	0	1	1.060975	6.976965	150	1	1.110343	0.9;
	4036	1	0	0	0	0	1	1.076575	-25.454051	220	1	1.166568	0.9;
	4037	1	142.5	523.5	0	0	1	1.098796	-17.3085	380	1	1.166438	0.9;
	4038	1	60.0666	45	0	0	1	1.069009	25.391957	220	1	1.125329	0.9;
	4039	1	0	0	0	0	1	1.046203	-8.01519	380	1	1.1	0.9;
	4040	1	0	0	0	0	1	1.099656	-17.304305	380	1	1.173637	0.9;
	4041	1	0	0	0.163	0	1	0.99178	-22.452823	220	1	1.1	0.9;
	4042	1	5.506105	2	0	0	1	1.058854	-20.795987	110	1	1.118728	0.9;
	4043	1	13.81532	4.5	0	0	1	1.002972	-29.839446	110	1	1.1	0.9;
	4044	2	141.2	33.5	0	0	1	0.973067	62.692821	150	1	1.1	0.9;
	4045	1	1.301443	0.4	0	0	1	0.994095	-5.900606	110	1	1.1	0.9;
	4046	2	31.53496	12.5	0	0	1	1.086053	-45.526518	380	1	1.125979	0.9;
	4047	1	5.205771	1	0	0	1	1.004302	-24.117059	110	1	1.1	0.9;
	4048	1	16.51831	4.8	0	0	1	0.980177	-19.981623	110	1	1.1	0.9;
	4049	2	0.4	0.2	0	0	1	1.053342	-35.247991	380	1	1.118111	0.9;
	4050	1	0	0	0	0	1	1.073811	-35.366732	380	1	1.124564	0.9;
	4051	2	27.43041	13.7	0	0	1	1.082211	-19.10634	380	1	1.159517	0.9;
	4052	1	182.182	36.4	0	0	1	0.997408	-49.35051	220	1	1.1	0.9;
	4053	1	140.2	11.2	0	0	1	1.055155	-49.489684	220	1	1.114086	0.9;
	4054	1	8.909878	8.9	0	0	1	0.943691	-23.456261	110	1	1.1	0.9;
	4055	1	52.6	110.6	0	0	1	1.070499	-9.766049	220	1	1.127769	0.9;
	4056	1	0	0	0	0	1	1.054086	-20.329333	220	1	1.113317	0.9;
	4057	1	0	0	0	0	1	1.010258	-22.6726	150	1	1.1	0.9;
	4058	1	0	0	0	0	1	1.095749	-17.5494	380	1	1.145224	0.9;
	4059	1	0	0	0	0	1	1.039735	-24.332115	150	1	1.1	0.9;
	4060	1	22.42486	4.8	0	0	1	0.999744	-28.699449	110	1	1.1	0.9;
	4061	1	0	0	0.087538	0	1	1.039099	-9.655578	220	1	1.107897	0.9;
	4062	2	0	0	0	0	1	1.047	-22.224689	150	1	1.1	0.9;
	4063	1	15.11676	5	0	0	1	0.961345	-26.850621	110	1	1.1	0.9;
	4064	1	0	0	0	0	1	0.994217	11.442369	220	1	1.1	0.9;
	4065	1	0	0	0	0	1	1.016185	-42.158651	220	1	1.1	0.9;
	4066	1	0	0	0	0	1	1.01762	-33.51645	220	1	1.1	0.9;
	4067	1	37.44151	-28.3	0	0	1	1.014701	8.411271	220	1	1.1	0.9;
	4068	1	0	0	0	0	1	1.046308	3.543641	380	1	1.115942	0.9;
	4069	1	0	0	0	0	1	1.036268	-21.544554	150	1	1.113167	0.9;
	4070	1	0	0	0	0	1	1.039434	19.43199	380	1	1.1	0.9;
	4071	1	49.45483	30.6	0	0	1	1.002386	-26.000495	150	1	1.1	0.9;
	4072	1	24.92764	15.4	0	0	1	0.967426	-30.526625	150	1	1.1	0.9;
	4073	2	0	0	0	0	1	1.068652	2.4194	150	1	1.14018	0.9;
	4074	2	0	0	0	0	1	1.100421	58.422931	380	1	1.118732	0.9;
	4075	1	0	0	0	0	1	1.030546	-24.479822	150	1	1.1	0.9;
	4076	2	0	0	0	0	1	1.046091	8.293152	220	1	1.119086	0.9;
	4077	1	0	0	0	0	1	1.038758	-29.676596	220	1	1.1	0.9;
	4078	2	0	0	0	0	1	1.051318	-6.118834	220	1	1.139354	0.9;
	4079	1	0	0	0	17.25139	1	1.020672	-23.361982	150	1	1.1	0.9;
	4080	1	69.77736	25	0	0	1	1.04702	-12.992685	220	1	1.13252	0.9;
	4081	1	0	0	0	0	1	1.073878	13.755568	380	1	1.12251	0.9;
	4082	2	1.101221	0.4	0	0	1	1.001	-23.029485	110	1	1.1	0.9;
	4083	1	0	0	0	0	1	1.082659	-40.0038	380	1	1.15237	0.9;
	4084	1	71.8797	-2.2	0	0	1	1.042812	-18.944555	220	1	1.1	0.9;
	4085	1	0	0	0	0	1	1.06927	-42.639377	220	1	1.118451	0.9;
	4086	1	0	0	0	0	1	1.023391	-33.01377	220	1	1.1	0.9;
	4087	1	0	0	0	0	1	1.039984	-25.124667	220	1	1.138618	0.9;
	4088	1	0	0	0	0	1	1.078228	39.435629	380	1	1.120708	0.9;
	4089	1	15.31698	18.2	0	0	1	1.050617	-36.710174	220	1	1.1	0.9;
	4090	1	0	0	0	0	1	1.069052	-1.85991	380	1	1.127973	0.9;
	4091	1	336.3	104.4	0	0	1	1.066296	-11.384689	150	1	1.130939	0.9;
	4092	1	0	0	0	0	1	1.005964	12.583351	220	1	1.1	0.9;
	4093	1	0	0	0	0	1	0.944452	-11.809313	110	1	1.1	0.9;
	4094	2	0	0	0.536314	0	1	1.103553	46.223869	380	1	1.12031	0.9;
	4095	1	80.0888	20	0	0	1	1.007153	46.705412	220	1	1.1	0.9;
	4096	1	199.3	53.7	0	0	1	1.019862	14.993557	380	1	1.1	0.9;
	4097	1	65.6	-382	0.222924	0	1	1.06427	36.367054	380	1	1.126274	0.9;
	4098	2	0	0	0	0	1	1.053091	-42.381896	220	1	1.108872	0.9;
	4099	1	20.72298	6.8	0	0	1	0.932838	-33.0364	110	1	1.1	0.9;
	4100	1	170.1	-9.3	0	0	1	1.050857	-41.715254	220	1	1.185099	0.9;
	4101	1	95.50589	17.1	0	0	1	1.040479	-15.625301	220	1	1.127701	0.9;
	4102	2	219.6	62.5	0	0	1	1.073364	-6.535664	220	1	1.139312	0.9;
	4103	1	0	0	0	0	1	1.073413	-38.210279	380	1	1.138338	0.9;
	4104	2	62.36915	21.9	0	0	1	1.015363	61.626304	150	1	1.1	0.9;
	4105	1	77.98647	30.8	0	0	1	1.067525	1.756894	150	1	1.138234	0.9;
	4106	1	75.38358	21.8	0	0	1	1.029634	-35.314381	220	1	1.1	0.9;
	4107	1	81.8908	19.9	0	0	1	1.052525	-24.762366	220	1	1.12751	0.9;
	4108	1	0	0	0	0	1	1.058016	-43.869719	380	1	1.126584	0.9;
	4109	1	92.10212	29.5	0	0	1	1.058092	-9.147235	220	1	1.118796	0.9;
	4110	1	0	0	0	0	1	1.019764	-23.851115	110	1	1.1	0.9;
	4111	2	0	0	0	0	1	1.048682	-44.410211	220	1	1.1	0.9;
	4112	1	12.91432	4	0	0	1	1.02573	-14.659571	110	1	1.125043	0.9;
	4113	1	0	0	0	0	1	1.063149	-9.400285	220	1	1.130493	0.9;
	4114	1	349.6	22	0	0	1	1.083148	-40.524383	380	1	1.151764	0.9;
	4115	1	91.10101	36.1	0	0	1	1.00394	11.679232	220	1	1.1	0.9;
	4116	1	0	0	0	0	1	1.096775	-8.474033	220	1	1.157881	0.9;
	4117	2	41.04551	31.8	0	0	1	1.0565	-40.946308	220	1	1.11521	0.9;
	4118	1	34.03774	21.1	0	0	1	1.026443	-23.796336	150	1	1.1	0.9;
	4119	1	3.203552	8.3	0	0	1	0.975964	-25.309434	150	1	1.1	0.9;
	4120	2	58.2646	12.2	0	0	1	1.082712	-14.587335	220	1	1.160743	0.9;
	4121	1	15.31698	9.5	0	0	1	1.035034	-21.258987	150	1	1.1	0.9;
	4122	1	5.105661	1.9	0	0	1	0.962405	-12.134583	110	1	1.1	0.9;
	4123	2	405.5	247.2	0	0	1	1.079289	13.798602	380	1	1.15036	0.9;
	4124	1	3.203552	0.4	0	0	1	1.02249	-16.061245	110	1	1.1	0.9;
	4125	1	0	0	0.159842	0	1	1.064697	10.205544	380	1	1.125776	0.9;
	4126	1	14.91654	5.4	0	0	1	1.037323	-18.972928	110	1	1.123197	0.9;
	4127	1	7.708547	1.3	0	0	1	0.989624	-27.669548	110	1	1.1	0.9;
	4128	2	0	0	0	0	1	1.017432	-21.981306	110	1	1.1	0.9;
	4129	2	0	0	0	0	1	1.030702	-33.955807	110	1	1.1	0.9;
	4130	1	0	0	0	0	1	1.048802	-22.276481	220	1	1.1	0.9;
	4131	1	0	0	0	0	1	1.072106	-9.190276	380	1	1.129632	0.9;
	4132	1	0	0	0	0	1	1.072857	-39.076192	380	1	1.121655	0.9;
	4133	1	0	0	0	0	1	1.086411	-41.19112	380	1	1.146206	0.9;
	4134	1	0	0	0	0	1	1.071563	-13.757737	220	1	1.135478	0.9;
	4135	1	21.12342	13.1	0	0	1	1.043285	-21.001055	150	1	1.1	0.9;
	4136	1	2.803108	0.5	0	0	1	0.999423	-23.719985	110	1	1.1	0.9;
	4137	1	33.03663	10.9	0	0	1	0.948388	-28.03577	110	1	1.1	0.9;
	4138	1	20.0222	9	0	0	1	0.969693	-25.755992	110	1	1.1	0.9;
	4139	1	0	0	0	0	1	1.096039	-11.176902	220	1	1.152022	0.9;
	4140	1	0	0	0	0	1	1.07864	-10.074395	380	1	1.132704	0.9;
	4141	1	0	0	0	0	1	0.994202	-23.538525	150	1	1.1	0.9;
	4142	1	179.8	76.5	0	0	1	1.053675	-12.931186	220	1	1.111821	0.9;
	4143	1	119	21.9	0	0	1	1.075032	-9.588126	220	1	1.145706	0.9;
	4144	1	48.45372	16	0	0	1	0.946555	-27.924428	110	1	1.1	0.9;
	4145	1	138.2	21.3	0	0	1	1.026623	-44.440042	220	1	1.1	0.9;
	4146	1	93	66	0	0	1	1.007228	14.501841	220	1	1.1	0.9;
	4147	1	6.407104	2.3	0	0	1	1.057325	-21.009721	110	1	1.115132	0.9;
	4148	1	0	0	0	0	1	0.994045	-19.21757	110	1	1.1	0.9;
	4149	1	13.91543	6.3	0	0	1	0.910416	-29.478081	110	1	1.1	0.9;
	4150	1	101	20.6	0	0	1	1.072404	-42.308435	380	1	1.135782	0.9;
	4151	1	130.5	25.7	0	0	1	1.078281	-15.796259	220	1	1.141299	0.9;
	4152	2	0	0	0	0	1	0.966331	-19.670432	110	1	1.1	0.9;
	4153	2	0	0	0	0	1	1.083955	-15.561807	220	1	1.168203	0.9;
	4154	1	0	0	0	0	1	1.067604	9.787942	150	1	1.1	0.9;
	4155	2	77.58602	-6.1	0	0	1	1.082409	-26.854166	220	1	1.189016	0.9;
	4156	1	0	0	0	0	1	1.008261	-26.57636	150	1	1.1	0.9;
	4157	1	0	0	0.139597	0	1	1.029131	-21.49625	220	1	1.1	0.9;
	4158	1	11.91321	4.3	0	0	1	1.024882	-18.153663	110	1	1.118176	0.9;
	4159	1	73.28124	9.4	0	0	1	1.051976	33.85445	380	1	1.1	0.9;
	4160	1	47.95317	16.3	0	0	1	1.054688	-35.876203	220	1	1.13666	0.9;
	4161	2	0	0	0	0	1	1.039896	-45.77026	220	1	1.1	0.9;
	4162	1	8.509435	2.8	0	0	1	0.97618	-32.14869	110	1	1.1	0.9;
	4163	1	311.4	43	0	0	1	1.054413	-42.452836	380	1	1.123028	0.9;
	4164	1	2.202442	0.6	0	0	1	1.027136	-50.927306	150	1	1.1	0.9;
	4165	1	50.65616	56.5	0	0	1	1.043594	11.026839	380	1	1.1	0.9;
	4166	1	19.52164	6.4	0	0	1	1.019552	-23.406185	150	1	1.1	0.9;
	4167	1	11.01221	3.9	0	0	1	1.069928	-11.904726	220	1	1.14239	0.9;
	4168	2	95.40578	-22.5	0	0	1	1.092728	11.517878	220	1	1.168873	0.9;
	4169	1	0	0	0	0	1	1.053958	-28.394206	110	1	1.1	0.9;
	4170	1	0	0	0	0	1	1.071628	-12.597343	110	1	1.135518	0.9;
	4171	1	0	0	0	0	1	1.000933	-23.445073	110	1	1.1	0.9;
	4172	2	672.7	107.6	0	0	1	1.075	-11.523575	150	1	1.138555	0.9;
	4173	1	12.21354	7.9	0	0	1	0.976442	-6.492301	110	1	1.1	0.9;
	4174	1	0	0	0	0	1	1.086541	-45.4787	380	1	1.149574	0.9;
	4175	2	6.407104	4	0	0	1	1.028661	-23.307923	150	1	1.1	0.9;
	4176	1	-6.7	-7.5	0	0	1	1.051996	-20.135056	150	1	1.105973	0.9;
	4177	1	0	0	0	0	1	1.009796	-30.615157	110	1	1.1	0.9;
	4178	1	183.8	2.2	0	0	1	1.028337	-44.139756	220	1	1.1	0.9;
	4179	1	20.32253	-24.8	0	0	1	1.026389	9.933989	220	1	1.1	0.9;
	4180	2	0	0	0	0	1	1.045696	-5.875555	380	1	1.110874	0.9;
	4181	2	107.9	52.9	0	0	1	1.076509	11.129908	220	1	1.157016	0.9;
	4182	1	4.304773	1.4	0	0	1	0.977538	-30.445331	110	1	1.1	0.9;
	4183	1	0	0	0	0	1	0.989972	37.961196	220	1	1.1	0.9;
	4184	1	0	0	0	0	1	1.073572	-6.947927	380	1	1.127669	0.9;
	4185	1	0	0	0	0	1	1.066632	-16.528473	110	1	1.141571	0.9;
	4186	1	19.1212	7.4	0	0	1	1.085121	-45.332854	380	1	1.124154	0.9;
	4187	1	0	0	0	0	1	1.041676	63.002162	150	1	1.1	0.9;
	4188	2	37.94207	18.9	0	0	1	1.086526	-18.760126	380	1	1.16128	0.9;
	4189	1	70.27792	10.3	0	0	1	1.076936	-47.446677	220	1	1.12578	0.9;
	4190	1	0	0	0	0	1	1.052382	-19.853903	150	1	1.105893	0.9;
	4191	1	10.11121	3.7	0	0	1	1.06125	-20.538706	110	1	1.121438	0.9;
	4192	1	0	0	0	0	1	1.094706	-38.343507	380	1	1.160295	0.9;
	4193	1	10.11121	0	0	0	1	1.013763	-17.084142	110	1	1.1	0.9;
	4194	1	0	0	0	0	1	1.063724	-39.49242	380	1	1.132009	0.9;
	4195	1	16.51831	5.5	0	0	1	1.049235	-28.78343	110	1	1.1	0.9;
	4196	1	74.18224	14.6	0	0	1	1.039383	4.338386	110	1	1.106039	0.9;
	4197	1	42.54717	17.6	0	0	1	1.09292	-24.589657	220	1	1.18541	0.9;
	4198	1	30.83419	16.5	0	0	1	1.043164	14.675638	150	1	1.1	0.9;
	4199	1	64.57159	10	0	0	1	0.980215	-22.065702	110	1	1.1	0.9;
	4200	1	1.902109	0.6	0	0	1	0.967889	-34.590758	110	1	1.1	0.9;
	4201	1	74.6828	46.3	0	0	1	1.035945	-29.496065	150	1	1.1	0.9;
	4202	2	0	0	0	0	1	1.101947	38.288093	380	1	1.119982	0.9;
	4203	1	26.02886	16.1	0	0	1	0.979163	-23.955021	150	1	1.1	0.9;
	4204	1	40.0444	15	0	0	1	1.03064	29.734469	220	1	1.1	0.9;
	4205	1	0	0	0	0	1	1.050235	-16.00925	380	1	1.116125	0.9;
	4206	2	0	0	0	0	1	1.053999	-14.76471	220	1	1.108193	0.9;
	4207	1	162.1	73	0	0	1	0.97763	-25.388779	110	1	1.1	0.9;
	4208	1	350	129.4	0	0	1	1.067102	-18.559882	220	1	1.108446	0.9;
	4209	1	0	0	0	0	1	1.011057	-44.66348	220	1	1.1	0.9;
	4210	1	137.8	64.5	0	0	1	1.049939	-10.542285	380	1	1.118244	0.9;
	4211	1	16.11787	4.3	0	0	1	1.036901	-19.655021	110	1	1.106442	0.9;
	4212	1	0	0	0	0	1	1.000669	-29.584385	110	1	1.1	0.9;
	4213	1	75.5838	-7.5	0	0	1	1.055046	-11.959949	380	1	1.122961	0.9;
	4214	2	280.7	53.7	0	0	1	1.094263	-5.58809	380	1	1.145965	0.9;
	4215	1	17.91987	7.2	0	0	1	0.985688	-23.096643	110	1	1.1	0.9;
	4216	1	1.701887	0.6	0	0	1	0.989871	-23.037821	110	1	1.1	0.9;
	4217	1	0	0	0	0	1	1.074434	-17.92228	150	1	1.1	0.9;
	4218	1	3.704107	1.3	0	0	1	1.045717	-20.751565	110	1	1.10873	0.9;
	4219	1	18.2202	8.6	0	0	1	1.071019	-40.264213	380	1	1.134713	0.9;
	4220	1	0	0	0.278139	0	1	1.06827	2.55233	380	1	1.133901	0.9;
	4221	1	0	0	0	0	1	1.077906	-5.35564	220	1	1.136067	0.9;
	4222	2	0	0	0	0	1	1.070105	32.261058	380	1	1.137262	0.9;
	4223	1	41.74628	13.2	0	0	1	1.059328	-29.819851	110	1	1.120405	0.9;
	4224	1	5.00555	2	0	0	1	0.981737	-22.797891	110	1	1.1	0.9;
	4225	1	0	0	0	0	1	1.033826	-49.22739	220	1	1.1	0.9;
	4226	1	0	0	0	0	1	1.051614	-49.224479	220	1	1.114106	0.9;
	4227	1	0	0	0	0	1	1.089157	-40.725654	380	1	1.162358	0.9;
	4228	1	0	0	0	0	1	0.960691	-13.481854	110	1	1.1	0.9;
	4229	1	4.605106	1.7	0	0	1	1.063635	-20.07853	110	1	1.120735	0.9;
	4230	1	93.90411	-0.4	0	0	1	1.027531	-19.1411	220	1	1.13744	0.9;
	4231	2	68.37582	-24.4	0	0	1	1.085864	-34.505422	220	1	1.12365	0.9;
	4232	1	51.05661	7.4	0	0	1	1.027362	-43.454938	220	1	1.1	0.9;
	4233	1	27.13008	16.8	0	0	1	1.067156	-14.601085	150	1	1.118907	0.9;
	4234	1	16.51831	3.2	0	0	1	0.941654	-25.0239	110	1	1.1	0.9;
	4235	2	100.6	15.7	0	0	1	1.059682	-26.196637	220	1	1.126985	0.9;
	4236	2	87.9	75	0	0	1	1.090237	44.220786	380	1	1.1	0.9;
	4237	2	204.2	41.1	0	0	1	1.048864	-45.608306	220	1	1.111385	0.9;
	4238	1	0	0	0	0	1	1.01346	-14.793221	110	1	1.1	0.9;
	4239	1	390.5	62.5	0	0	1	1.069223	-11.967137	150	1	1.132492	0.9;
	4240	2	0	0	0	0	1	1.018455	-21.944802	220	1	1.1	0.9;
	4241	1	28.03108	10.2	0	0	1	1.033472	-22.89406	110	1	1.13133	0.9;
	4242	1	11.41265	5.1	0	0	1	1.012655	-19.192307	110	1	1.1	0.9;
	4243	2	0.300333	97.4	0.302736	0	1	1.088079	50.472051	380	1	1.1	0.9;
	4244	1	44.34917	14.8	0	0	1	0.9467	-27.010076	110	1	1.1	0.9;
	4245	1	91.2	56.1	0	0	1	1.049435	-30.184971	220	1	1.113258	0.9;
	4246	1	0	0	0	0	1	1.006513	-25.963756	150	1	1.1	0.9;
	4247	1	124	29.2	0	0	1	1.038694	-12.596835	220	1	1.109321	0.9;
	4248	1	0.200222	0.8	0	0	1	1.069407	25.410656	220	1	1.125697	0.9;
	4249	2	0	0	0	0	1	1.098238	-7.55811	380	1	1.15254	0.9;
	4250	1	0	0	0	0	1	1.038353	-25.050669	220	1	1.1	0.9;
	4251	1	25.12786	8.2	0	0	1	1.057572	-13.412354	110	1	1.110652	0.9;
	4252	1	102.7	17.8	0	0	1	1.078615	-19.618381	220	1	1.142847	0.9;
	4253	1	344.7	-31.1	0	0	1	1.084447	-26.093145	380	1	1.140905	0.9;
	4254	1	84.09324	1	0	0	1	1.081155	-17.641661	380	1	1.130142	0.9;
	4255	1	2.502775	1.3	0	0	1	0.916759	-29.652145	110	1	1.1	0.9;
	4256	1	38.84307	8.5	0	0	1	1.086485	-3.753567	220	1	1.139408	0.9;
	4257	1	0	0	0	0	1	1.008487	-17.950507	110	1	1.1	0.9;
	4258	1	46.15117	11.2	0	0	1	1.053569	-24.626352	220	1	1.128464	0.9;
	4259	1	16.01776	5.1	0	0	1	1.048578	5.684158	110	1	1.116397	0.9;
	4260	1	0	0	0	0	1	1.062107	0.616986	150	1	1.124536	0.9;
	4261	1	109.7	15	0	0	1	1.050479	-30.528475	220	1	1.148006	0.9;
	4262	2	0.800888	45.3	0	0	1	1.086447	-40.259436	380	1	1.131074	0.9;
	4263	1	0	8.3	0	0	1	1.006885	-27.664636	150	1	1.193762	0.9;
	4264	1	0	0	0	0	1	1.059561	-0.784545	150	1	1.123457	0.9;
	4265	2	0	0	0	0	1	1.047338	-18.569332	110	1	1.131127	0.9;
	4266	1	10.81199	3.9	0	0	1	0.996842	-20.261029	110	1	1.1	0.9;
	4267	2	124.3	42	0	0	1	1.027636	0.997562	220	1	1.124754	0.9;
	4268	1	15.81754	7.2	0	0	1	0.969912	-24.810045	110	1	1.1	0.9;
	4269	1	0	0	0.137257	0	1	1.099212	-10.183285	380	1	1.157729	0.9;
	4270	1	25.32808	8.8	0	0	1	1.031284	-30.255669	110	1	1.1	0.9;
	4271	1	0	0	0	0	1	0.979671	-26.689733	150	1	1.1	0.9;
	4272	1	162	83.6	0	0	1	1.032868	-23.644151	220	1	1.1	0.9;
	4273	2	0	0	0	0	1	1.090737	-9.839942	380	1	1.154871	0.9;
	4274	1	136.8	4.8	0	0	1	1.07824	-13.267959	220	1	1.13362	0.9;
	4275	1	19.72187	7.1	0	0	1	0.964291	-19.057794	110	1	1.1	0.9;
	4276	1	0	0	0.363006	0	1	1.094193	-31.741966	380	1	1.164814	0.9;
	4277	1	19.62176	7.4	0	0	1	1.018363	-15.400151	110	1	1.1	0.9;
	4278	1	0	0	0	0	1	1.08754	-41.885151	380	1	1.158669	0.9;
	4279	1	0	0	0	0	1	1.032608	-24.958809	150	1	1.1	0.9;
	4280	2	0	0	0	0	1	0.968182	-13.027538	110	1	1.1	0.9;
	4281	1	0	0	0	0	1	1.038405	5.179927	380	1	1.113066	0.9;
	4282	2	198.1	78	0	0	1	1.058228	-8.598386	220	1	1.120698	0.9;
	4283	1	0	0	0	0	1	1.045373	-11.497824	220	1	1.116883	0.9;
	4284	1	0	0	0.086002	0	1	1.056725	-30.707986	220	1	1.112013	0.9;
	4285	1	14.21576	5.1	0	0	1	1.004281	-19.833901	110	1	1.1	0.9;
	4286	1	0	0	0	0	1	1.059431	4.674702	380	1	1.117572	0.9;
	4287	1	0	0	0	0	1	1.072408	-45.257425	220	1	1.129318	0.9;
	4288	1	392	177.7	0.191685	0	1	1.032358	-15.190106	220	1	1.1	0.9;
	4289	1	0	0	0	0	1	1.079914	-38.536331	380	1	1.142017	0.9;
	4290	1	97	57.4	0	0	1	1.071826	-10.535682	380	1	1.139759	0.9;
	4291	2	0	0	0	0	1	1.057064	-36.152006	220	1	1.137791	0.9;
	4292	1	0	0	0	0	1	1.062746	63.163944	150	1	1.1	0.9;
	4293	1	0	0	0	0	1	1.029898	-49.788336	150	1	1.1	0.9;
	4294	1	0	0	0	0	1	1.040693	4.127604	380	1	1.114207	0.9;
	4295	1	0	0	0	0	1	1.031963	-24.998566	150	1	1.1	0.9;
	4296	1	4.905439	1.6	0	0	1	1.062082	-9.808739	220	1	1.129644	0.9;
	4297	2	55.36138	1.3	0	0	1	0.962984	-27.688024	110	1	1.1	0.9;
	4298	1	2.00222	0.6	0	0	1	0.931299	-27.969929	110	1	1.1	0.9;
	4299	1	3.103441	0.1	0	0	1	1.070482	-20.321034	220	1	1.135484	0.9;
	4300	1	0	0	0	0	1	1.052137	-9.681277	220	1	1.127291	0.9;
	4301	1	0	0	0	0	1	1.000766	-22.844685	150	1	1.1	0.9;
	4302	1	0	0	0	0	1	1.054515	-16.045688	150	1	1.109957	0.9;
	4303	1	17.7	-334.5	0.144009	0	1	1.081015	17.598362	380	1	1.133193	0.9;
	4304	1	0	0	0	0	1	1.082671	16.232982	380	1	1.127643	0.9;
	4305	1	0	0	0	0	1	1.047412	-9.451486	380	1	1.1	0.9;
	4306	1	0	0	0	0	1	1.039981	-41.264288	380	1	1.120231	0.9;
	4307	1	0	0	0.090702	0	1	1.095863	-10.91828	380	1	1.154366	0.9;
	4308	1	277.9	143.1	0	0	1	1.060398	1.706701	380	1	1.114891	0.9;
	4309	2	106	27.3	0	0	1	1.078886	-10.124128	220	1	1.152674	0.9;
	4310	2	0.400444	0.3	0	0	1	0.992	-7.358111	110	1	1.1	0.9;
	4311	2	0	0	0	0	1	1.070183	12.147358	110	1	1.147559	0.9;
	4312	2	16.4182	9.7	0	0	1	0.979616	-11.99626	110	1	1.1	0.9;
	4313	1	0	0	0	0	1	1.076091	0.506627	380	1	1.128564	0.9;
	4314	2	14.41598	20.7	0	0	1	1.010753	-25.094185	150	1	1.162093	0.9;
	4315	1	2.202442	0	0	0	1	0.991428	-37.216439	220	1	1.1	0.9;
	4316	1	63.67059	21.3	0	0	1	1.070783	-20.644255	220	1	1.107934	0.9;
	4317	1	0	0	0.191685	0	1	1.025565	-23.738907	220	1	1.1	0.9;
	4318	1	93.30344	27.8	0	0	1	1.045099	-12.282161	220	1	1.117712	0.9;
	4319	1	0	0	0	0	1	1.081531	-11.890355	220	1	1.147348	0.9;
	4320	1	137.7	32.6	0	0	1	1.02218	-41.645191	220	1	1.1	0.9;
	4321	1	0	0	0	0	1	1.032685	-25.351507	150	1	1.1	0.9;
	4322	1	101.1	19.8	0	0	1	1.027045	-43.838747	220	1	1.1	0.9;
	4323	1	0	0	0	0	1	1.011428	60.623275	220	1	1.1	0.9;
	4324	1	59.1656	14	0	0	1	1.005229	-21.055889	220	1	1.1	0.9;
	4325	2	28.93208	9.4	0	0	1	0.907012	-32.943513	110	1	1.1	0.9;
	4326	1	147	56.3	0	0	1	1.064239	1.868594	150	1	1.136787	0.9;
	4327	2	0	0	0	0	1	1.043368	13.643526	380	1	1.108391	0.9;
	4328	1	45.04995	19.2	0	0	1	1.003426	-15.781133	110	1	1.1	0.9;
	4329	1	65.07214	30	0	0	1	1.019758	-16.952213	110	1	1.1	0.9;
	4330	1	173.4	31	0	0	1	1.042697	-22.998136	220	1	1.1	0.9;
	4331	1	0	0	0	0	1	1.06526	-26.019192	150	1	1.12928	0.9;
	4332	1	31.93541	10.5	0	0	1	0.921341	-28.145224	110	1	1.1	0.9;
	4333	1	14.91654	5.1	0	0	1	0.986537	-25.013876	110	1	1.1	0.9;
	4334	1	33.03663	10.1	0	0	1	1.039575	-11.714717	220	1	1.1	0.9;
	4335	2	0	0	0	0	1	1.068366	48.04628	380	1	1.120061	0.9;
	4336	1	0	0	0	0	1	0.947442	-34.591343	110	1	1.1	0.9;
	4337	1	51.85749	19.2	0	0	1	1.058026	-35.635612	220	1	1.139566	0.9;
	4338	1	0	0	0	0	1	1.018034	-32.649591	150	1	1.1	0.9;
	4339	2	87.49701	25.9	0	0	1	1.0615	-43.639945	220	1	1.113389	0.9;
	4340	2	0	0	0	0	1	1.064636	-38.40314	220	1	1.105864	0.9;
	4341	1	63.27015	28.9	0	0	1	1.073546	0.807311	150	1	1.147141	0.9;
	4342	1	153.2	31.5	0	0	1	1.055795	-22.305694	220	1	1.1	0.9;
	4343	2	0	0	0	0	1	1.103711	24.36166	380	1	1.140522	0.9;
	4344	1	0	0	0	0	1	0.991964	-25.104251	150	1	1.1	0.9;
	4345	1	0	0	0	0	1	1.059289	31.671955	380	1	1.136205	0.9;
	4346	1	2.502775	1.5	0	0	1	0.966936	-25.731264	110	1	1.1	0.9;
	4347	1	40.14451	24.9	0	0	1	1.030389	-24.498815	150	1	1.1	0.9;
	4348	1	21.02331	14.8	0	0	1	1.006609	-23.355737	150	1	1.1	0.9;
	4349	2	0	0	0	0	1	1.046182	8.611679	220	1	1.119123	0.9;
	4350	1	69.77736	22.4	0	0	1	0.949674	-29.544673	110	1	1.1	0.9;
	4351	1	0	0	0	0	1	1.066755	-0.140506	220	1	1.139096	0.9;
	4352	1	89.9	40.2	0	0	1	1.047289	20.724263	380	1	1.1	0.9;
	4353	1	71.9798	27.1	0	0	1	1.048876	-14.073472	380	1	1.117115	0.9;
	4354	2	30.53385	10.9	0.117	0	1	1.024545	-4.932416	220	1	1.1	0.9;
	4355	1	146.9	-5.1	0	0	1	1.08818	2.779791	220	1	1.150829	0.9;
	4356	1	0	0	0	0	1	0.954046	-25.902941	150	1	1.1	0.9;
	4357	1	9.110101	3	0	0	1	1.011489	-29.162385	110	1	1.1	0.9;
	4358	1	0	0	0	-15.677	1	1.022725	-23.284679	150	1	1.1	0.9;
	4359	1	336.9	81	0	0	1	1.03694	-2.805018	220	1	1.107343	0.9;
	4360	1	11.41265	3.8	0	0	1	1.078611	-12.67781	110	1	1.143111	0.9;
	4361	1	0	0	0	0	1	1.100344	-39.798051	380	1	1.160875	0.9;
	4362	1	-9.6	-2.5	0	0	1	1.048945	-17.63162	150	1	1.122971	0.9;
	4363	1	0	0	0	0	1	1.040725	25.76517	380	1	1.130696	0.9;
	4364	1	0	0	0	0	1	0.979163	-23.955021	150	1	1.1	0.9;
	4365	1	0	0	0	0	1	1.046872	7.405599	380	1	1.1	0.9;
	4366	1	0	0	0	0	1	1.018248	-23.608019	150	1	1.1	0.9;
	4367	1	0	0	0.282417	0	1	1.070689	-21.45181	380	1	1.145765	0.9;
	4368	1	20.12231	4.9	0	0	1	0.996281	-24.746663	110	1	1.1	0.9;
	4369	1	0	0	0	0	1	0.972783	-26.493774	110	1	1.1	0.9;
	4370	1	118.6	43.4	0	0	1	1.064592	-18.590147	220	1	1.1	0.9;
	4371	2	371.6	100.4	0	0	1	1.060364	6.841014	220	1	1.148391	0.9;
	4372	1	210.8	-0.7	0	0	1	1.073593	-42.303446	220	1	1.122208	0.9;
	4373	1	39.84418	6.4	0	0	1	1.081436	-37.859784	380	1	1.139729	0.9;
	4374	1	7.00777	1	0	0	1	1.05943	-27.29641	150	1	1.128496	0.9;
	4375	1	151	7.1	0	0	1	1.019533	-41.713093	220	1	1.1	0.9;
	4376	1	0	0	0.073	0	1	1.037732	-24.672947	220	1	1.2	0.8;
	4377	1	-875.4	-172.2	0	0	1	1.09007	-5.057312	380	1	1.148771	0.9;
	4378	1	141.3567	53.5	0	0	1	1.015533	-47.147028	220	1	1.1	0.9;
	4379	1	41.24573	13.5	0	0	1	0.941942	-25.941252	110	1	1.1	0.9;
	4380	1	0	0	0	0	1	0.993748	-21.817459	110	1	1.1	0.9;
	4381	2	68.07548	13.6	0	0	1	1.076364	-23.178405	220	1	1.16521	0.9;
	4382	1	270.5	77.8	0	0	1	1.097805	-21.480439	380	1	1.14712	0.9;
	4383	2	0	0	0	0	1	1.081342	0.192816	380	1	1.142942	0.9;
	4384	1	0	0	0	0	1	1.096716	-44.758513	120	1	1.173262	0.9;
	4385	1	0	0	0	0	1	1.003822	-23.875344	150	1	1.1	0.9;
	4386	1	78.48702	48.6	0	0	1	0.979605	-24.633729	150	1	1.1	0.9;
	4387	2	0	0	0	0	1	1.066409	-14.026379	220	1	1.148504	0.9;
	4388	2	0	0	0	0	1	1.099237	-10.241415	380	1	1.159529	0.9;
	4389	1	0	0	0	0	1	1.02929	7.56658	380	1	1.1	0.9;
	4390	1	103.6	16	0	0	1	1.070451	-23.607654	220	1	1.171669	0.9;
	4391	1	0	0	0	0	1	0.996058	59.965117	220	1	1.1	0.9;
	4392	1	0	0	0	11.65603	1	1.018733	-22.368868	150	1	1.1	0.9;
	4393	1	171.9	39.1	0	0	1	1.039839	-5.318729	220	1	1.124446	0.9;
	4394	1	0	0	0.126994	0	1	1.08867	-10.255363	380	1	1.146559	0.9;
	4395	2	136.5	46.1	0	0	1	1.066339	-14.340888	220	1	1.122024	0.9;
	4396	1	19.22131	6.4	0	0	1	1.083908	-41.437345	220	1	1.155723	0.9;
	4397	1	5.105661	2.5	0	0	1	0.913653	-29.37562	110	1	1.1	0.9;
	4398	1	34.13785	21.1	0	0	1	1.006878	-27.666196	150	1	1.193903	0.9;
	4399	1	0	0	0.114011	0	1	1.08582	-28.312985	380	1	1.142974	0.9;
	4400	2	14.61621	1.3	0	0	1	1.036552	-10.371872	110	1	1.1	0.9;
	4401	1	19.82198	6.5	0	0	1	0.966067	-25.801762	110	1	1.1	0.9;
	4402	1	121.6	49.3	0	0	1	0.994178	7.354863	220	1	1.1	0.9;
	4403	1	151.1	66.9	0	0	1	1.047127	-14.854499	380	1	1.106534	0.9;
	4404	1	25.22797	15.6	0	0	1	0.968059	-24.775992	150	1	1.1	0.9;
	4405	2	0	0	0	0	1	1.080177	-26.077615	220	1	1.120804	0.9;
	4406	2	0	0	0	0	1	1.080053	-38.296895	380	1	1.150684	0.9;
	4407	2	173.202	-7	0	0	1	0.923591	36.005668	220	1	1.1	0.9;
	4408	2	45.04995	50.5	0	0	1	1.023933	62.246724	150	1	1.1	0.9;
	4409	2	0	0	0	0	1	1.0762	11.426402	150	1	1.10794	0.9;
	4410	2	347.3	105.6	0	0	1	1.062789	13.975286	380	1	1.130625	0.9;
	4411	2	58.06438	4.1	0	0	1	1.021136	-20.441527	220	1	1.111942	0.9;
	4412	1	0.200222	0.5	0.656836	0	1	1.078653	-6.58426	380	1	1.133834	0.9;
	4413	2	0	0	0	0	1	1.067091	-13.061327	220	1	1.118904	0.9;
	4414	1	2.302553	0.3	0	0	1	0.954094	-23.46733	110	1	1.1	0.9;
	4415	1	45.5505	-59.9	0	0	1	1.079166	-19.349807	220	1	1.143016	0.9;
	4416	1	0	0	0	0	1	1.017297	-24.487789	150	1	1.1	0.9;
	4417	1	0	0	0	0	1	0.985698	8.182729	220	1	1.1	0.9;
	4418	1	127.9	-34	0	0	1	1.061419	-54.168785	220	1	1.1	0.9;
	4419	1	45.95095	7.5	0	0	1	1.058645	-40.945979	220	1	1.108255	0.9;
	4420	1	78.68724	19.6	0	0	1	1.01292	-27.735152	110	1	1.1	0.9;
	4421	1	0	0	0	0	1	1.050164	23.850522	380	1	1.127199	0.9;
	4422	1	261.2	105.6	0	0	1	1.093863	56.369655	380	1	1.114475	0.9;
	4423	1	0	0	0	0	1	1.071205	-2.590788	380	1	1.12859	0.9;
	4424	1	0	0	0.086	0	1	1.004056	-21.217422	220	1	1.1	0.9;
	4425	2	12.01332	3.8	0	0	1	0.982253	-19.814483	110	1	1.1	0.9;
	4426	1	0	0	0	0	1	1.066342	9.614521	150	1	1.1	0.9;
	4427	2	0	0	0	0	1	1.054868	-12.447754	220	1	1.138681	0.9;
	4428	1	0	0	0	0	1	1.03595	-46.107851	380	1	1.1	0.9;
	4429	1	169	56.4	0	0	1	1.058737	21.347659	380	1	1.127229	0.9;
	4430	1	65.77292	27.3	0	0	1	1.075838	-4.346949	220	1	1.137193	0.9;
	4431	1	2.202442	0.5	0	0	1	1.031506	-43.461893	150	1	1.1	0.9;
	4432	1	0	0	0	0	1	1.083942	-15.5845	220	1	1.167984	0.9;
	4433	2	29.43263	0.7	0	0	1	1.078545	-18.836987	110	1	1.1	0.9;
	4434	1	15.41709	3.9	0	0	1	1.037211	3.990298	110	1	1.1	0.9;
	4435	1	262.6	31.8	0	0	1	1.009809	-46.035611	220	1	1.1	0.9;
	4436	2	0	0	0	0	1	1.056306	-38.450266	220	1	1.189623	0.9;
	4437	1	0	0	0	0	1	0.993293	11.080085	220	1	1.1	0.9;
	4438	1	230.9	42.6	0	0	1	1.050406	-21.616362	220	1	1.1	0.9;
	4439	1	0.900999	0.3	0	0	1	1.057299	-20.043859	110	1	1.115227	0.9;
	4440	1	5.105661	1.7	0	0	1	0.997044	-29.475195	110	1	1.1	0.9;
	4441	1	0	0	0	0	1	1.036081	-0.498374	380	1	1.10523	0.9;
	4442	2	135.9	24.7	0	0	1	1.071818	-14.320246	220	1	1.127893	0.9;
	4443	1	5.105661	0	0	0	1	0.941711	-27.396493	110	1	1.1	0.9;
	4444	1	0	0	0	0	1	0.963971	-25.25361	150	1	1.1	0.9;
	4445	1	0	0	0	0	1	1.075826	0.245464	380	1	1.128812	0.9;
	4446	1	166.4	50.4	0	0	1	1.076195	-25.467685	220	1	1.164079	0.9;
	4447	1	0	0	0	0	1	1.099789	61.333768	380	1	1.117151	0.9;
	4448	1	53.55938	33.2	0	0	1	1.027424	-21.729128	150	1	1.1	0.9;
	4449	1	74.98314	19.8	0	0	1	1.040619	-12.407295	220	1	1.111131	0.9;
	4450	1	185	43.6	0	0	1	1.061738	-43.893605	220	1	1.111866	0.9;
	4451	2	0	0	0	0	1	1.084818	-2.229715	220	1	1.145207	0.9;
	4452	1	0	0	0	0	1	1.064341	5.336883	380	1	1.1	0.9;
	4453	1	21.62398	7.8	0	0	1	1.065196	-12.842837	110	1	1.157664	0.9;
	4454	1	-7.60844	2.5	0	0	1	1.060759	-14.417897	150	1	1.110927	0.9;
	4455	1	329.7	32.4	0	0	1	1.055538	-53.154312	220	1	1.1	0.9;
	4456	2	32.23574	8.9	0	0	1	1.043409	-42.033559	220	1	1.1	0.9;
	4457	1	0	0	0.119999	0	1	1.002514	-23.821989	220	1	1.1	0.9;
	4458	1	0	0	0	0	1	1.005859	-24.704675	110	1	1.1	0.9;
	4459	1	1.902109	0	0	0	1	0.988886	-27.73563	110	1	1.1	0.9;
	4460	1	0	0	0	0	1	1.10389	-9.233376	380	1	1.147608	0.9;
	4461	1	96	33.3	0	0	1	1.052691	-36.236867	220	1	1.135169	0.9;
	4462	1	17.11898	5.6	0	0	1	1.002974	-29.839566	110	1	1.1	0.9;
	4463	1	175.8	70.7	0	0	1	1.049919	-41.783587	220	1	1.1	0.9;
	4464	2	125.6	29.2	0	0	1	1.085075	-13.051716	220	1	1.158808	0.9;
	4465	1	50.0555	9.9	0	0	1	1.078205	-20.248238	220	1	1.141112	0.9;
	4466	1	220.6	83.4	0	0	1	1.065279	-19.06845	380	1	1.118445	0.9;
	4467	1	0	0	0	0	1	1.072999	-42.384618	220	1	1.121555	0.9;
	4468	1	0	0	0	0	1	1.051365	-20.177973	150	1	1.105289	0.9;
	4469	1	0	0	0	0	1	1.073931	-19.173675	220	1	1.110921	0.9;
	4470	1	-24.0266	7.9	0	0	1	1.054668	-19.075202	150	1	1.1	0.9;
	4471	2	53.65949	7.9	0	0	1	1.060948	-21.95364	220	1	1.13846	0.9;
	4472	1	0	0	0	0	1	0.958901	-25.606606	150	1	1.1	0.9;
	4473	1	0	0	0	0	1	1.039965	-10.180746	110	1	1.1	0.9;
	4474	2	13.11454	4.9	0	0	1	1.099818	-10.968893	220	1	1.172469	0.9;
	4475	1	15.81754	9.8	0	0	1	1.024299	-23.111209	150	1	1.1	0.9;
	4476	1	0	0	0	0	1	1.062251	7.638396	380	1	1.1215	0.9;
	4477	1	0	0	0	0	1	1.072414	-15.25219	380	1	1.122416	0.9;
	4478	1	0	0	0	0	1	1.08334	-16.091468	380	1	1.146967	0.9;
	4479	2	78.18668	12.5	0	0	1	1.094008	-15.962661	110	1	1.176871	0.9;
	4480	1	0	0	0	0	1	1.072713	-21.341778	220	1	1.155602	0.9;
	4481	1	3.804218	1.1	0	0	1	1.08221	-37.878541	380	1	1.140551	0.9;
	4482	1	40.0444	10	0	0	1	1.056218	-44.357876	380	1	1.124468	0.9;
	4483	1	37	8.7	0.066781	0	1	1.087634	0.767244	380	1	1.136932	0.9;
	4484	1	2.502775	1.3	0	0	1	0.980598	-21.829479	110	1	1.1	0.9;
	4485	2	0	0	0	0	1	1.072578	-39.303479	380	1	1.139648	0.9;
	4486	1	0	0	0	0	1	0.987765	-20.844105	150	1	1.1	0.9;
	4487	1	-19.2213	6.3	0	0	1	1.01794	-22.770842	150	1	1.1	0.9;
	4488	1	22.42486	7.3	0	0	1	0.957465	-34.604332	110	1	1.1	0.9;
	4489	1	0	0	0	0	1	1.036207	-26.663395	220	1	1.1	0.9;
	4490	1	312.2	28	0	0	1	1.050578	-30.435543	220	1	1.148011	0.9;
	4491	1	11.61288	2.5	0	0	1	1.032027	-30.788814	110	1	1.1	0.9;
	4492	2	0	0	0.112004	0	1	1.080737	-7.655879	380	1	1.134633	0.9;
	4493	1	56.76294	31.1	0	0	1	1.074646	-4.535258	220	1	1.128254	0.9;
	4494	1	0	0	0	0	1	0.957307	-29.216707	150	1	1.1	0.892303;
	4495	1	59.66615	25.6	0	0	1	1.000647	10.159957	220	1	1.1	0.9;
	4496	1	109.9	33.6	0	0	1	1.077931	-12.319855	220	1	1.143997	0.9;
	4497	1	0	0	0	0	1	1.099869	-17.488973	380	1	1.16595	0.9;
	4498	2	36.84085	14	0	0	1	0.983182	-18.272465	110	1	1.1	0.9;
	4499	2	0.100111	35.9	0	0	1	1.087605	-36.861058	380	1	1.140696	0.9;
	4500	1	329.7	220.4	0	0	1	1.04564	-12.310538	380	1	1.111459	0.9;
	4501	1	88.2979	2.8	0	0	1	1.075191	-35.539772	220	1	1.11715	0.9;
	4502	1	10.21132	-3.2	0	0	1	1.035978	-32.271315	220	1	1.115757	0.9;
	4503	2	0	0	0	0	1	1.058727	6.239312	110	1	1.146971	0.9;
	4504	1	218.7	-9.1	0	0	1	1.034657	-39.455268	220	1	1.144134	0.9;
	4505	1	71.9798	34.1	0	0	1	1.088869	-2.863272	220	1	1.142513	0.9;
	4506	2	67.3747	31.4	0	0	1	1.059182	-16.316058	220	1	1.111638	0.9;
	4507	1	66.77403	11.1	0	0	1	1.053757	-47.038258	220	1	1.1	0.9;
	4508	2	52.7585	5.9	0	0	1	1.054591	-38.321649	220	1	1.106217	0.9;
	4509	1	0	0	0	0	1	1.067246	-0.419665	220	1	1.139094	0.9;
	4510	2	1.801998	91.2	0	0	1	1.088895	-34.997542	380	1	1.140083	0.9;
	4511	1	0	0	0	0	1	1.051241	-41.375725	220	1	1.1	0.9;
	4512	1	12.91432	7.1	0	0	1	0.949173	-27.31455	110	1	1.1	0.9;
	4513	1	62.96982	68	0	0	1	0.952525	-17.591369	220	1	1.1	0.9;
	4514	1	29.1323	18	0	0	1	0.987295	-24.257215	150	1	1.1	0.9;
	4515	1	0	0	0	0	1	0.998124	-23.701706	150	1	1.1	0.9;
	4516	1	0	0	0	0	1	1.053374	9.835532	380	1	1.112612	0.9;
	4517	1	0	0	0	0	1	0.990985	-23.000704	150	1	1.1	0.9;
	4518	1	110.7	78.7	0	0	1	1.032942	-37.502485	220	1	1.1	0.9;
	4519	2	292	81.8	0	0	1	1.045955	-37.806223	220	1	1.1	0.9;
	4520	1	0	0	0	0	1	1.067878	9.943282	380	1	1.121805	0.9;
	4521	2	0	0	0	0	1	1.092211	-8.103558	380	1	1.128334	0.9;
	4522	1	0	0	0	0	1	1.071349	-17.101581	380	1	1.152481	0.9;
	4523	1	130	128	0	0	1	1.021631	-13.268766	220	1	1.1	0.9;
	4524	1	0	0	0	10.72713	1	1.057605	-15.320527	150	1	1.1	0.9;
	4525	2	89.49923	26.9	0	0	1	1.060616	-13.068609	220	1	1.125939	0.9;
	4526	1	0	0	0	0	1	0.919681	-29.100332	150	1	1.1	0.868813;
	4527	1	23.42597	7.5	0	0	1	0.969994	-24.860406	110	1	1.1	0.9;
	4528	1	0	0	0.191685	0	1	1.054452	-25.05058	220	1	1.1261	0.9;
	4529	1	0	0	0	0	1	1.030827	-24.328356	150	1	1.1	0.9;
	4530	2	0	0	0	0	1	1.093262	-20.244147	380	1	1.163991	0.9;
	4531	1	0	0	0	0	1	1.01869	-24.764279	150	1	1.1	0.9;
	4532	1	4.805328	1.6	0	0	1	0.98737	-36.979224	110	1	1.1	0.9;
	4533	1	10.11121	1.6	0	0	1	0.969373	-24.065598	110	1	1.1	0.9;
	4534	1	32.5	30.5	0	-25.6712	1	1.086419	-15.479273	380	1	1.144112	0.9;
	4535	2	0	0	0	0	1	1.042295	-20.455497	110	1	1.1	0.9;
	4536	2	10.11121	5.7	0	0	1	1.016203	-21.877494	110	1	1.1	0.9;
	4537	1	11.41265	3.8	0	0	1	1.02456	-16.232778	110	1	1.1	0.9;
	4538	2	0	0	0	0	1	0.978572	-20.060492	110	1	1.1	0.9;
	4539	1	0	0	0.095843	0	1	1.013508	-21.823066	220	1	1.1	0.9;
	4540	1	501.1	10.7	0	0	1	1.055878	-43.384709	380	1	1.124216	0.9;
	4541	1	58.36471	42.7	0	0	1	0.943149	-13.573815	110	1	1.1	0.9;
	4542	2	0	0	0	0	1	1.040962	25.90764	380	1	1.13215	0.9;
	4543	2	0	0	0	0	1	1.074816	-26.772268	380	1	1.131421	0.9;
	4544	1	80.28902	28.4	0	0	1	1.066561	-23.468244	220	1	1.109819	0.9;
	4545	2	0	0	0	0	1	0.996273	-28.146012	110	1	1.1	0.9;
	4546	1	0	0	0	0	1	1.001315	-21.173616	220	1	1.1	0.9;
	4547	1	0	0	0	0	1	0.961174	-21.556367	220	1	1.1	0.9;
	4548	1	0	0	0	0	1	1.001314	-21.173562	220	1	1.1	0.9;
	4549	1	82.49146	24.5	0	0	1	1.073307	10.735459	110	1	1.15581	0.9;
	4550	1	0	0	0	0	1	1.068938	-12.076088	220	1	1.152559	0.9;
	4551	2	42.94762	0	0	0	1	1.076083	-9.472581	220	1	1.146694	0.9;
	4552	2	0	0	0	0	1	1.065591	-14.768536	220	1	1.148338	0.9;
	4553	2	0	0	0	0	1	1.097553	56.225682	380	1	1.117837	0.9;
	4554	1	-58.7652	18.7	0	0	1	1.01839	-33.288526	150	1	1.1	0.9;
	4555	1	0	0	0	0	1	1.053272	-14.274504	380	1	1.107373	0.9;
	4556	1	0	0	0	0	1	1.042848	9.633576	380	1	1.105123	0.9;
	4557	1	0	0	0.296958	0	1	1.100288	-10.065255	380	1	1.152066	0.9;
	4558	1	0	0	0	0	1	1.034985	13.119172	380	1	1.1	0.9;
	4559	1	5.105661	2.5	0	0	1	0.940141	-24.955574	110	1	1.1	0.9;
	4560	1	0	0	0	0	1	1.025672	-24.44021	150	1	1.1	0.9;
	4561	1	70.0777	35	0	0	1	1.083341	-15.597036	220	1	1.164953	0.9;
	4562	1	27.13008	9.8	0	0	1	1.010608	-17.153665	110	1	1.10889	0.9;
	4563	1	0	0	0	0	1	1.068705	-4.662019	380	1	1.131121	0.9;
	4564	1	0.100111	-9.6	0	0	1	1.070612	-18.15395	220	1	1.1	0.9;
	4565	1	9.510545	2.8	0	0	1	0.997088	-22.788396	110	1	1.1	0.9;
	4566	2	0	0	0	0	1	1.060455	-12.5075	220	1	1.144236	0.9;
	4567	2	0	0	0	0	1	1.045136	-22.772074	220	1	1.2	0.8;
	4568	1	0	0	0	0	1	0.964441	53.98231	220	1	1.1	0.9;
	4569	1	38.64284	23.9	0	0	1	0.999834	-23.487068	150	1	1.1	0.9;
	4570	1	9.510545	3.1	0	0	1	1.02044	-17.759741	110	1	1.118241	0.9;
	4571	2	162.6	36.4	0	0	1	1.026136	-43.747876	220	1	1.1	0.9;
	4572	1	0	0	0	0	1	1.060413	-5.735082	380	1	1.134391	0.9;
	4573	2	130.1	40.5	0	0	1	1.063	-14.108078	110	1	1.119773	0.9;
	4574	2	8.909878	3.2	0	0	1	1.037937	-11.667277	110	1	1.1	0.9;
	4575	1	30.33363	18.8	0	0	1	1.037624	-25.230646	150	1	1.1	0.9;
	4576	1	174.7	35	0	0	1	1.058828	6.781396	150	1	1.110564	0.9;
	4577	1	14.31587	4.4	0	0	1	1.024728	-31.130941	110	1	1.1	0.9;
	4578	1	32.23574	4.8	0	0	1	1.058657	-47.061638	220	1	1.111186	0.9;
	4579	1	0	0	0	0	1	1.056358	-40.996321	380	1	1.110841	0.9;
	4580	2	0	0	0	0	1	1.014	-16.846087	110	1	1.1	0.9;
	4581	1	10.31143	4.3	0	0	1	0.947039	-27.426829	110	1	1.1	0.9;
	4582	1	0	0	0	0	1	1.075845	-10.542346	380	1	1.133346	0.9;
	4583	1	20.32253	10.1	0	0	1	0.94672	-27.1454	110	1	1.1	0.9;
	4584	1	0	8.8	0	0	1	1.023955	-22.466353	150	1	1.151662	0.9;
	4585	1	24.32697	10	0	0	1	1.08974	-45.143135	380	1	1.131327	0.9;
	4586	2	0	0	0	0	1	1.053318	-14.913589	220	1	1.121331	0.9;
	4587	1	0	0	0	0	1	1.09376	-36.343942	380	1	1.145747	0.9;
	4588	1	11.41265	2.5	0	0	1	1.002344	-15.881011	110	1	1.1	0.9;
	4589	1	0	0	0.067054	0	1	1.040662	65.88704	220	1	1.1	0.9;
	4590	1	139.8	-25.8	0	0	1	1.059779	-12.210974	220	1	1.121399	0.9;
	4591	1	4.605106	1.3	0	0	1	0.986966	-25.381029	110	1	1.1	0.9;
	4592	1	0	0	0	0	1	1.089107	54.287863	380	1	1.113993	0.9;
	4593	1	136.6	29	0	0	1	1.075825	-19.261997	220	1	1.157177	0.9;
	4594	1	0	0	0.191685	0	1	1.045799	-17.469101	220	1	1.110566	0.9;
	4595	2	21.32364	7.9	0	0	1	1.033409	-15.59446	220	1	1.123837	0.9;
	4596	1	0	0	0	0	1	1.001538	-20.313715	220	1	1.1	0.9;
	4597	1	43.44817	26.9	0	0	1	1.017705	-23.44175	150	1	1.1	0.9;
	4598	1	55.16116	19.8	0	0	1	1.084738	37.869513	380	1	1.10513	0.9;
	4599	1	0	0	0.249	0	1	1.015359	-17.407	220	1	1.1	0.9;
	4600	1	0	0	0	0	1	0.976015	-25.068242	110	1	1.1	0.9;
	4601	1	9.510545	0.5	0	0	1	1.035345	-24.815021	150	1	1.1	0.9;
	4602	2	0	0	0	0	1	1.058868	-29.713317	380	1	1.12459	0.9;
	4603	2	0	0	0	0	1	1.095816	-36.759679	380	1	1.160502	0.9;
	4604	1	205.8	66.8	0	0	1	1.045552	9.71944	380	1	1.1	0.9;
	4605	1	160.6	51.2	0	0	1	1.098776	-22.783276	380	1	1.144503	0.9;
	4606	1	0	0	0	0	1	0.985443	-25.578358	150	1	1.1	0.9;
	4607	1	22.32475	7.3	0	0	1	0.932214	-32.391678	110	1	1.1	0.9;
	4608	2	0	0	0	0	1	1.103421	25.837074	380	1	1.140553	0.9;
	4609	1	35.63951	22	0	0	1	0.939995	-29.215258	150	1	1.1	0.891359;
	4610	2	0	0	0	0	1	1.041632	-7.700924	380	1	1.129228	0.9;
	4611	1	3.804218	1.3	0	0	1	0.977454	-22.53634	110	1	1.1	0.9;
	4612	2	0	0	0	0	1	1.080718	-26.093129	220	1	1.122129	0.9;
	4613	1	50.0555	25	0	0	1	0.94683	-31.833015	110	1	1.1	0.9;
	4614	1	0	0	0	0	1	0.976502	-9.172641	110	1	1.1	0.9;
	4615	1	0	0	0	-14.5987	1	1.010426	-27.437893	150	1	1.193337	0.9;
	4616	1	5.105661	2.5	0	0	1	0.926511	-29.214484	110	1	1.1	0.9;
	4617	2	148.3	12.1	0	0	1	1.061864	-7.162268	220	1	1.119911	0.9;
	4618	1	19.1212	2.8	0	0	1	0.940098	-25.620147	110	1	1.1	0.9;
	4619	2	0	0	0	0	1	1.069	-18.766423	220	1	1.11445	0.9;
	4620	1	92.40245	6.3	0	0	1	1.051469	-31.449458	220	1	1.136558	0.9;
	4621	1	0	0	0	0	1	1.039943	-25.059294	150	1	1.1	0.9;
	4622	1	3.704107	1.3	0	0	1	0.941342	-18.315441	110	1	1.1	0.9;
	4623	1	43.14784	11.1	0	0	1	1.049925	-21.326045	110	1	1.110639	0.9;
	4624	1	80.38913	10.8	0	0	1	1.04144	-12.29925	220	1	1.113002	0.9;
	4625	1	0	0	0	0	1	1.045476	-13.967777	750	1	1.117712	0.9;
	4626	2	185.6	62.3	0	0	1	1.070682	-6.883058	220	1	1.130531	0.9;
	4627	1	7.408214	2.7	0	0	1	1.033767	-13.764787	110	1	1.118393	0.9;
	4628	2	0	0	0	0	1	1.06504	-11.133413	110	1	1.126912	0.9;
	4629	2	90.0999	21	0	0	1	1.059732	-20.417722	220	1	1.11188	0.9;
	4630	1	155.2486	23.3	0	0	1	1.031187	-32.682592	220	1	1.1	0.9;
	4631	2	71.57936	57.6	0	0	1	1.090684	-35.319448	380	1	1.145392	0.9;
	4632	2	0	0	0	0	1	1.103658	-8.449869	380	1	1.142187	0.9;
	4633	1	70.77847	28	0	0	1	1.060618	-14.558014	220	1	1.129053	0.9;
	4634	1	0	0	0	0	1	1.050384	-30.723795	220	1	1.107951	0.9;
	4635	1	0	0	0	0	1	0.951708	-23.940019	110	1	1.1	0.9;
	4636	2	0	0	0	0	1	1.085077	-35.099296	380	1	1.135847	0.9;
	4637	2	0	0	0	0	1	1.086212	-9.132152	150	1	1.151573	0.9;
	4638	1	0	0	0	0	1	1.090709	1.488574	220	1	1.158312	0.9;
	4639	1	117.7	115	0	0	1	1.071547	-7.297108	380	1	1.139702	0.9;
	4640	1	0	0	0.090403	64.52145	1	1.054999	65.866145	380	1	1.1	0.9;
	4641	1	99.4	-347.1	0	0	1	1.08398	34.189139	380	1	1.133373	0.9;
	4642	1	10.81199	1.9	0	0	1	0.970217	-10.333919	110	1	1.1	0.9;
	4643	1	0	0	0	0	1	1.06119	-18.485624	220	1	1.114989	0.9;
	4644	1	39.94429	13	0	0	1	1.062295	-30.692825	220	1	1.188432	0.9;
	4645	1	122.9	19.3	0	0	1	1.084805	-16.310524	380	1	1.151997	0.9;
	4646	2	0	0	0	0	1	1.076205	10.475753	110	1	1.154646	0.9;
	4647	1	213.1	-12.3	0	0	1	1.086327	-25.719984	380	1	1.144297	0.9;
	4648	2	94.705	15.6	0	0	1	1.068545	-9.50534	220	1	1.12851	0.9;
	4649	1	139.9	1.1	0	0	1	1.034865	-48.036195	220	1	1.1	0.9;
	4650	1	41.34584	25.6	0	0	1	1.030629	-22.512233	150	1	1.1	0.9;
	4651	1	0	0	0	0	1	1.086043	-17.156453	220	1	1.168549	0.9;
	4652	2	2.502775	-0.1	0	0	1	0.99876	64.434016	150	1	1.1	0.9;
	4653	1	124.9	77	0	0	1	1.028106	7.590633	380	1	1.1	0.9;
	4654	1	46.25128	13.1	0	0	1	1.066165	-42.327939	220	1	1.113392	0.9;
	4655	1	-5.5061	1.8	0	0	1	0.997369	-22.465399	150	1	1.1	0.9;
	4656	2	2.402664	1	0	0	1	1.079727	-23.869736	220	1	1.139629	0.9;
	4657	2	24.72742	9	0	0	1	1.048658	-38.89454	220	1	1.1	0.9;
	4658	2	4.104551	0.9	0	0	1	0.956255	-26.392192	110	1	1.1	0.9;
	4659	1	0	0	0	0	1	1.057574	-30.223524	380	1	1.123334	0.9;
	4660	1	0	0	0	0	1	1.09066	28.271356	380	1	1.175507	0.9;
	4661	1	17.81976	7.3	0	0	1	0.966355	-23.226656	110	1	1.1	0.9;
	4662	1	66.4737	22.1	0	0	1	0.968649	-25.640513	110	1	1.1	0.9;
	4663	1	572.3	159.7	0	0	1	1.056172	-31.71744	380	1	1.119137	0.9;
	4664	1	0	0	0	0	1	1.076647	-25.454886	220	1	1.166632	0.9;
	4665	1	12.31365	3	0	0	1	0.934696	-25.719967	110	1	1.1	0.9;
	4666	1	0	0	0	0	1	1.0207	-21.043057	150	1	1.1	0.9;
	4667	1	0	0	0	0	1	0.961018	-22.397335	110	1	1.1	0.9;
	4668	1	52.95872	32.8	0	0	1	1.038421	-29.258697	150	1	1.1	0.9;
	4669	1	16.91876	0	0	0	1	1.081626	-15.392458	220	1	1.144453	0.9;
	4670	2	0	0	0	0	1	1.090316	-24.189717	380	1	1.147556	0.9;
	4671	1	24.62731	2.1	0	0	1	1.068106	4.616363	220	1	1.143953	0.9;
	4672	1	26.34628	30	0	0	1	1.021887	-24.085723	150	1	1.1	0.9;
	4673	1	18.92098	6.6	0	0	1	0.98599	-25.575448	110	1	1.1	0.9;
	4674	1	0	0	0	0	1	1.002507	56.960561	220	1	1.1	0.9;
	4675	1	0	0	0	0	1	0.975669	-24.74531	150	1	1.1	0.9;
	4676	2	0	0	0	0	1	1.082895	-2.884645	380	1	1.144165	0.9;
	4677	2	0.6	0.4	0	0	1	1.042136	-18.612008	220	1	1.1	0.9;
	4678	1	229.9	48.9	0	0	1	1.010513	-33.792788	220	1	1.108389	0.9;
	4679	1	0	0	0	0	1	1.082158	1.047433	380	1	1.132754	0.9;
	4680	2	198.2	62.3	0	0	1	1.069794	-6.708376	220	1	1.124979	0.9;
	4681	1	0.200222	1.6	0	0	1	1.014024	13.924802	220	1	1.1	0.9;
	4682	1	0	0	0	0	1	1.029384	-23.369523	150	1	1.1	0.9;
	4683	1	0	0	0	-33.7219	1	1.041908	-25.420228	150	1	1.1	0.9;
	4684	2	0	0	0	0	1	1.06941	25.410658	220	1	1.125711	0.9;
	4685	1	18.52053	6.1	0	0	1	0.920097	-31.459301	110	1	1.1	0.9;
	4686	2	82.39135	12.8	0	0	1	0.968333	67.547457	150	1	1.1	0.9;
	4687	1	411.3	168	0	0	1	1.067018	-10.886588	150	1	1.135826	0.9;
	4688	1	0	0	0	19.58753	1	1.062308	-20.249085	150	1	1.1	0.9;
	4689	1	0	0	0	0	1	1.005754	-19.847241	110	1	1.1	0.9;
	4690	1	34.03774	11	0	0	1	0.986302	-35.119387	110	1	1.1	0.9;
	4691	1	21.12342	7.7	0	0	1	1.058642	-4.454718	110	1	1.134202	0.9;
	4692	1	0	0	0.086002	0	1	1.02952	-9.384755	220	1	1.110963	0.9;
	4693	1	0	0	0	0	1	1.048129	-31.924771	220	1	1.1	0.9;
	4694	1	55.26127	0	0	0	1	1.042472	-44.049702	220	1	1.111954	0.9;
	4695	1	-11.8131	-12	0	0	1	1.023038	-29.597008	220	1	1.114624	0.9;
	4696	2	0	0	0	0	1	1.094105	36.653001	380	1	1.114941	0.9;
	4697	1	0	0	0	0	1	1.044394	7.37431	380	1	1.1	0.9;
	4698	1	0	0	0	0	1	0.967196	-28.04469	150	1	1.1	0.9;
	4699	1	12.01332	0.7	0	0	1	1.055635	61.518383	220	1	1.109074	0.9;
	4700	1	0	0	0	0	1	1.0134	-31.245433	110	1	1.1	0.9;
	4701	1	0	0	0	0	1	1.060262	13.426192	380	1	1.119276	0.9;
	4702	1	0	0	0	0	1	1.053927	-48.624156	220	1	1.116994	0.9;
	4703	2	106.8	28.8	0	0	1	1.047	-42.547385	220	1	1.105529	0.9;
	4704	1	3.00333	1.7	0	0	1	1.081002	-17.369417	110	1	1.128031	0.9;
	4705	1	0	0	0	0	1	0.966002	-29.140275	110	1	1.1	0.9;
	4706	1	0	0	0	0	1	1.075055	-20.009507	220	1	1.139911	0.9;
	4707	1	0	0	0.067002	0	1	1.041295	-11.828016	220	1	1.1	0.9;
	4708	1	1.902109	0.7	0	0	1	1.030302	-23.049432	110	1	1.1	0.9;
	4709	1	0	0	0.457708	0	1	1.082112	-17.269677	380	1	1.1313	0.9;
	4710	1	90.2	5.8	0	0	1	1.080124	36.567759	380	1	1.126789	0.9;
	4711	2	189	10.9	0	0	1	1.051405	-30.188682	220	1	1.148969	0.9;
	4712	1	45.04995	15.3	0	0	1	1.052372	-36.135814	220	1	1.134659	0.9;
	4713	1	0	0	0	0	1	1.081768	-7.412341	220	1	1.148107	0.9;
	4714	1	0	0	0	0	1	1.050061	-22.805199	220	1	1.1	0.9;
	4715	1	0	0.2	0	0	1	1.028239	-4.422104	220	1	1.1	0.9;
	4716	1	0	0	0.095843	0	1	1.049627	-11.095204	220	1	1.112908	0.9;
	4717	1	36.34029	22.5	0	0	1	0.995695	-20.48945	150	1	1.1	0.9;
	4718	1	107.9	17.6	0	0	1	1.058834	0.30698	150	1	1.120538	0.9;
	4719	1	24.32697	5.8	0	0	1	1.070214	-42.473416	220	1	1.116913	0.9;
	4720	1	0	0	0	0	1	0.973925	-25.040553	110	1	1.1	0.9;
	4721	1	67.97536	9.2	0	0	1	1.052688	-45.508297	220	1	1.117102	0.9;
	4722	2	0	0	0	0	1	1.075727	-16.552721	220	1	1.142318	0.9;
	4723	2	11.31254	4.6	0	0	1	1.041091	-9.961801	110	1	1.1	0.9;
	4724	1	102.7	39	0	0	1	1.084166	37.671212	380	1	1.1	0.9;
	4725	1	33.23685	11.1	0	0	1	0.916305	-30.954277	110	1	1.1	0.9;
	4726	1	7.00777	2.3	0	0	1	1.038135	-26.184523	110	1	1.1	0.9;
	4727	1	35.13896	12.6	0	0	1	1.003341	-16.394774	110	1	1.1	0.9;
	4728	1	7.408214	2.7	0	0	1	1.005743	-18.393766	110	1	1.1	0.9;
	4729	1	76.08436	17.1	0	0	1	1.029982	-21.168518	220	1	1.1	0.9;
	4730	1	35.93985	26	0	0	1	1.043554	-21.729684	220	1	1.2	0.8;
	4731	1	85.09435	13.4	0	0	1	1.056494	-48.67117	220	1	1.114444	0.9;
	4732	1	221.9	86.3	0	0	1	1.064708	3.812117	150	1	1.135654	0.9;
	4733	1	0	0	0	0	1	1.0354	-42.694636	220	1	1.119073	0.9;
	4734	1	0	0	0	0	1	1.048251	-18.174211	220	1	1.1	0.9;
	4735	1	11.71299	3.9	0	0	1	0.938334	-35.31118	110	1	1.1	0.9;
	4736	1	41.64617	13.6	0	0	1	0.977679	-28.272654	110	1	1.1	0.9;
	4737	2	222.7	64.1	0	0	1	1.030177	3.029891	380	1	1.121719	0.9;
	4738	2	360	84.1	0	0	1	1.069389	-4.900056	220	1	1.126989	0.9;
	4739	1	0	0	0	0	1	1.087783	-19.574419	380	1	1.14791	0.9;
	4740	1	0	0	0	0	1	1.046834	-11.033265	220	1	1.118496	0.9;
	4741	1	246.8	36.3	0	0	1	1.049622	-25.523965	220	1	1.158137	0.9;
	4742	1	0	0	0	0	1	1.089791	-2.79084	220	1	1.148792	0.9;
	4743	1	0	0	0	0	1	1.066716	-5.866607	220	1	1.122297	0.9;
	4744	1	0	0	0	0	1	1.040273	-18.73954	150	1	1.1	0.9;
	4745	1	12.11343	0	0	0	1	1.03883	-44.012045	220	1	1.1	0.9;
	4746	1	68.67614	21	0	0	1	0.976451	-26.066799	110	1	1.1	0.9;
	4747	1	23.02553	8.3	0	0	1	1.044651	-21.20515	110	1	1.10642	0.9;
	4748	1	0	0	0	0	1	0.949041	-24.332914	110	1	1.1	0.9;
	4749	1	4.605106	1.7	0	0	1	1.007964	-17.835064	110	1	1.1	0.9;
	4750	1	27.73075	17.2	0	0	1	0.977287	-27.162776	150	1	1.1	0.9;
	4751	2	75.5838	26.1	0	0	1	1.086636	-14.643227	220	1	1.136451	0.9;
	4752	2	14.5	3.1	0	0	1	1.104553	-13.906648	380	1	1.118531	0.9;
	4753	1	1.301443	0.4	0	0	1	1.029192	-13.403481	110	1	1.1	0.9;
	4754	1	44.94984	18.7	0	0	1	1.044709	-3.96431	220	1	1.1	0.9;
	4755	1	0	0	0	0	1	1.046747	-31.689706	220	1	1.1	0.9;
	4756	2	8.209102	0.5	0	0	1	1.021818	-21.141219	110	1	1.1	0.9;
	4757	1	53.45927	15.4	0	0	1	1.00488	10.738319	220	1	1.1	0.9;
	4758	1	0	0	0	0	1	1.082534	-10.991796	220	1	1.15415	0.9;
	4759	1	264.3	111.6	0	0	1	1.051435	31.059389	380	1	1.1	0.9;
	4760	1	14.21576	4.7	0	0	1	0.970786	-32.495668	110	1	1.1	0.9;
	4761	2	0	0	0	0	1	1.086227	-13.690711	220	1	1.127398	0.9;
	4762	2	0	0	0	0	1	1.066364	-13.811698	220	1	1.118726	0.9;
	4763	1	30.0333	10	0	0	1	0.970652	-26.656231	110	1	1.1	0.9;
	4764	2	344.4	87.2	0	0	1	1.05	-43.418123	380	1	1.122847	0.9;
	4765	1	0	0	0	0	1	1.049162	-3.27079	380	1	1.135109	0.9;
	4766	1	0	0	0	0	1	1.08518	-21.647686	220	1	1.133498	0.9;
	4767	1	0	0	0	0	1	0.987562	-17.297195	150	1	1.1	0.9;
	4768	1	0	0	0	0	1	1.082877	10.646155	150	1	1.135945	0.9;
	4769	1	0	0	0	0	1	1.07815	0.274164	150	1	1.147608	0.9;
	4770	1	4.00444	0.1	0	0	1	0.997172	-23.491067	150	1	1.1	0.9;
	4771	2	0	0	0	0	1	0.944964	-25.465574	110	1	1.1	0.9;
	4772	1	101.2	48.9	0	0	1	0.998351	-27.112844	150	1	1.1	0.9;
	4773	2	2.302553	0.8	0	0	1	0.988545	-16.29827	110	1	1.1	0.9;
	4774	1	0	0	0	0	1	1.005788	-16.189925	110	1	1.1	0.9;
	4775	1	69.1767	38.1	0	0	1	1.081223	-41.51288	220	1	1.152965	0.9;
	4776	1	0	0	0	11.07044	1	1.045586	-20.095589	150	1	1.1	0.9;
	4777	1	43.94873	8.5	0	0	1	1.083661	7.639683	150	1	1.138148	0.9;
	4778	1	91.0009	36	0	0	1	1.049267	-16.242423	220	1	1.120298	0.9;
	4779	1	0	0	0	0	1	1.076102	-5.913744	220	1	1.135646	0.9;
	4780	1	0	0	0	0	1	1.071954	26.976004	380	1	1.1	0.9;
	4781	2	0	-0.2	0	0	1	1.039671	-11.504496	110	1	1.1	0.9;
	4782	1	77	-239.9	0	0	1	1.041236	8.979767	220	1	1.1	0.9;
	4783	2	0	0	0	0	1	1.102816	-10.818244	380	1	1.141795	0.9;
	4784	1	0	0	0	0	1	1.086821	-40.551615	380	1	1.14372	0.9;
	4785	1	18.12009	11.2	0	0	1	0.999827	-23.664813	150	1	1.1	0.9;
	4786	1	2.202442	0.5	0	0	1	1.050378	-30.724477	220	1	1.107946	0.9;
	4787	1	72.18002	44.7	0	0	1	0.973252	-24.235149	150	1	1.1	0.9;
	4788	1	66.97426	41.5	0	0	1	1.031959	-25.555762	150	1	1.1	0.9;
	4789	1	40.44484	13.3	0	0	1	0.963767	-26.782877	110	1	1.1	0.9;
	4790	1	0	0	0	0	1	1.078236	-20.068303	220	1	1.157339	0.9;
	4791	1	69.1767	19.7	0	0	1	1.062271	7.179293	150	1	1.11156	0.9;
	4792	1	0	0	0	0	1	1.094046	34.601912	380	1	1.127227	0.9;
	4793	2	46.05106	1	0	0	1	1.053533	63.973823	150	1	1.1	0.9;
	4794	1	2.202442	0.5	0	0	1	1.028919	-42.99778	150	1	1.1	0.9;
	4795	2	60.26682	18.3	0	0	1	1.021892	-17.227422	110	1	1.124297	0.9;
	4796	1	143.2	5.9	0	0	1	1.049149	-39.438556	220	1	1.1	0.9;
	4797	2	0	0	0	0	1	1.073909	-23.063208	220	1	1.139405	0.9;
	4798	1	158.8	105.3	0	0	1	1.043149	14.673862	150	1	1.1	0.9;
	4799	2	199.2	22.8	0	0	1	1.063773	-36.704728	220	1	1.118288	0.9;
	4800	1	19.82198	12.3	0	0	1	1.010428	-19.574142	150	1	1.1	0.9;
	4801	1	13.91543	6.9	0	0	1	0.974878	-19.542096	110	1	1.1	0.9;
	4802	1	0	0	0	0	1	1.036501	-17.961905	220	1	1.1	0.9;
	4803	2	0	0	0	0	1	1.0435	-23.922628	220	1	1.1	0.9;
	4804	1	17.21909	5.1	0	0	1	1.073769	-21.045717	220	1	1.155778	0.9;
	4805	1	0	0	0	0	1	1.05486	-14.422877	380	1	1.116445	0.9;
	4806	1	0	0	0	0	1	1.096637	-34.29802	380	1	1.159105	0.9;
	4807	2	-40.3447	18.3	0	0	1	1.030273	-31.891158	220	1	1.1	0.9;
	4808	1	0	0	0	0	1	1.025055	-45.602752	220	1	1.1	0.9;
	4809	1	0	0	0	0	1	1.050525	16.067937	220	1	1.130402	0.9;
	4810	1	16.01776	4.7	0	0	1	1.046035	13.884545	110	1	1.122591	0.9;
	4811	1	55.26127	11	0	0	1	1.073243	-23.159959	220	1	1.138325	0.9;
	4812	1	0	0	0	0	1	1.04067	-19.935673	150	1	1.120242	0.9;
	4813	2	0	0	0	0	1	1.056658	27.937802	380	1	1.129243	0.9;
	4814	1	0	0	0	0	1	1.047315	-11.106511	220	1	1.118327	0.9;
	4815	2	0	0	0	0	1	1.067455	60.881157	220	1	1.113176	0.9;
	4816	1	267.1	57.4	0	0	1	1.060992	-33.523982	380	1	1.122419	0.9;
	4817	1	40.0444	20	0	0	1	1.019886	-46.661502	220	1	1.1	0.9;
	4818	1	0	0	0	0	1	1.03039	-23.440764	150	1	1.1	0.9;
	4819	2	180.2	24.7	0	0	1	1.072773	-4.667877	220	1	1.129325	0.9;
	4820	2	0	0	0	0	1	0.997323	-23.673346	110	1	1.1	0.9;
	4821	1	0	0	0	0	1	1.04141	-21.688395	150	1	1.1	0.9;
	4822	1	0	0	0	0	1	1.032163	-17.256489	380	1	1.108847	0.9;
	4823	2	0	0	0	0	1	1.087	-15.831137	220	1	1.132262	0.9;
	4824	2	186.8	38.3	0	0	1	1.049752	-1.181911	220	1	1.121581	0.9;
	4825	1	0	0	0	0	1	1.021569	12.035196	220	1	1.1	0.9;
	4826	1	59.86637	19.9	0	0	1	0.961571	-26.17277	110	1	1.1	0.9;
	4827	1	15.21687	2.5	0	0	1	1.009387	-17.49788	110	1	1.1	0.9;
	4828	1	0	0	0	0	1	1.089097	-23.922758	380	1	1.14432	0.9;
	4829	1	0	0	0	0	1	1.030837	17.054427	380	1	1.1	0.9;
	4830	1	0	0	0	0	1	1.084658	39.435245	380	1	1.1	0.9;
	4831	1	293.2	87.4	0	0	1	1.026737	-41.933265	220	1	1.11849	0.9;
	4832	1	46.95206	54.6	0	0	1	1.033259	-25.194893	150	1	1.1	0.9;
	4833	1	21.12342	7.7	0	0	1	1.046145	-21.100661	110	1	1.1	0.9;
	4834	1	18.12009	5.9	0	0	1	1.066895	2.975479	150	1	1.12654	0.9;
	4835	2	51.05661	59.6	0	0	1	0.95776	-15.22782	110	1	1.1	0.9;
	4836	2	0	0	0	0	1	1.047364	57.449027	220	1	1.107736	0.9;
	4837	1	0	0	0	0	1	1.042046	-3.617375	220	1	1.1	0.9;
	4838	1	46.55161	-18.9	0	0	1	1.060378	-19.108647	220	1	1.1	0.9;
	4839	1	42.44706	-10.2	0	0	1	1.022847	-21.872746	220	1	1.10884	0.9;
	4840	1	0	0	0.249309	0	1	1.053123	-20.442133	380	1	1.136852	0.9;
	4841	1	288.7	117.3	0	0	1	1.085003	-39.915583	380	1	1.153331	0.9;
	4842	2	75.68391	18.4	0	0	1	1.060455	-39.111329	220	1	1.1	0.9;
	4843	1	-12.614	4.1	0	0	1	1.006939	-21.709092	150	1	1.1	0.9;
	4844	1	0	0	0	0	1	1.023579	-32.305575	150	1	1.1	0.9;
	4845	1	366.6	57.4	0	0	1	1.075864	-46.22102	380	1	1.112614	0.9;
	4846	1	83.59268	-7.2	0	0	1	1.02961	-42.704703	220	1	1.1	0.9;
	4847	1	17.51942	6.6	0	0	1	1.042526	-12.74454	220	1	1.114644	0.9;
	4848	1	1.601776	0.8	0	0	1	0.974612	-25.003966	110	1	1.1	0.9;
	4849	1	0	0	0	0	1	1.031785	-12.187347	110	1	1.1	0.9;
	4850	1	0	0	0.095843	0	1	1.006573	-25.159987	220	1	1.1	0.9;
	4851	1	-16.6184	8.7	0	0	1	1.011796	-28.59804	220	1	1.1	0.9;
	4852	2	220.8	13.2	0	0	1	1.052182	-10.84262	220	1	1.132757	0.9;
	4853	1	64.27126	17.5	0	0	1	1.06789	-20.36213	220	1	1.150325	0.9;
	4854	1	32.13563	13.1	0	0	1	1.08071	0.061128	150	1	1.148351	0.9;
	4855	1	0	0	0	0	1	1.053792	-5.091263	380	1	1.121376	0.9;
	4856	1	21.72409	7.2	0	0	1	0.953872	-27.609472	110	1	1.1	0.9;
	4857	1	0	0	0	0	1	1.06784	22.273412	110	1	1.115765	0.9;
	4858	1	8.709657	2.5	0	0	1	0.971881	-27.629697	110	1	1.1	0.9;
	4859	2	0	0	0	0	1	1.052136	-38.766772	220	1	1.1	0.9;
	4860	1	0	0	0	0	1	0.985502	-24.258087	380	1	1.1	0.9;
	4861	1	0	0	0	0	1	1.079975	-12.777536	220	1	1.134109	0.9;
	4862	1	0	0	0	0	1	1.013738	-28.91817	110	1	1.1	0.9;
	4863	1	70.1778	19.5	0	0	1	1.081969	-8.528901	220	1	1.152135	0.9;
	4864	1	229	77.1	0	0	1	1.094905	-37.289029	380	1	1.160834	0.9;
	4865	1	0	0	0	25.27635	1	0.988075	-24.241803	150	1	1.1	0.9;
	4866	1	0	0	0	0	1	1.02444	-25.67911	150	1	1.1	0.9;
	4867	1	0	0	0	0	1	0.961889	-29.652528	150	1	1.1	0.87723;
	4868	1	173.6	105.6	0	0	1	1.043815	-13.729485	380	1	1.106905	0.9;
	4869	1	0	0	0	0	1	0.987688	-38.98641	220	1	1.1	0.9;
	4870	1	152.7336	-17.2	0	0	1	1.077029	-22.99678	220	1	1.1619	0.9;
	4871	1	79.78846	49.4	0	0	1	1.030584	-24.499978	150	1	1.1	0.9;
	4872	1	110.6	38.1	0	0	1	1.062704	-2.539532	150	1	1.140636	0.9;
	4873	2	11.11232	3	0	0	1	1.095182	-11.36468	220	1	1.169978	0.9;
	4874	2	0	0	0	0	1	1.092636	2.719886	110	1	1.154429	0.9;
	4875	2	118.5	18.3	0	0	1	1.058909	7.002481	220	1	1.143095	0.9;
	4876	1	13.91543	4.1	0	0	1	0.964731	-9.392736	110	1	1.1	0.9;
	4877	1	-13.3148	20.6	0	0	1	1.023866	-22.456237	150	1	1.149894	0.9;
	4878	1	21.02331	8	0	0	1	1.087438	-36.671155	380	1	1.13145	0.9;
	4879	1	-8.70966	2.9	0	0	1	1.016275	-24.275268	150	1	1.1	0.9;
	4880	2	16.31809	6.3	0	0	1	1.022265	61.683559	150	1	1.1	0.9;
	4881	1	408.3	84.5	0	0	1	1.054522	-34.987516	380	1	1.116606	0.9;
	4882	1	0	0	0	0	1	1.051864	65.950176	220	1	1.1	0.9;
	4883	1	13.21465	4.4	0	0	1	0.998202	-40.055334	110	1	1.1	0.9;
	4884	2	0	0	0	0	1	1.101474	-6.061233	380	1	1.187472	0.9;
	4885	1	127.3	90.5	0	0	1	1.028972	4.138976	380	1	1.1	0.9;
	4886	1	0	0	0	0	1	1.03201	-25.666017	150	1	1.1	0.9;
	4887	1	0.800888	0.5	0	0	1	1.003203	-24.362845	150	1	1.1	0.9;
	4888	1	0	0	0	0	1	1.058946	-23.324369	220	1	1.132985	0.9;
	4889	1	48.05328	28.8	0	0	1	1.018169	-24.787152	150	1	1.1	0.9;
	4890	1	0	0	0	0	1	1.087668	-19.637637	380	1	1.147741	0.9;
	4891	2	370	159	0	0	1	0.990794	-6.358469	220	1	1.1	0.9;
	4892	1	37.94207	51.4	0	0	1	1.050389	10.637937	380	1	1.106993	0.9;
	4893	1	0	0	0	0	1	1.06617	3.92527	380	1	1.111569	0.9;
	4894	1	0	0	0	0	1	1.026274	-12.986511	220	1	1.1	0.9;
	4895	1	8.909878	3.8	0	0	1	0.990677	-24.273303	110	1	1.1	0.9;
	4896	1	57.06327	11.9	0	0	1	1.063267	-43.871415	220	1	1.112824	0.9;
	4897	1	0	0	0	0	1	1.043641	-20.98115	150	1	1.1	0.9;
	4898	1	0	0	0	0	1	1.065072	-14.814066	220	1	1.147903	0.9;
	4899	1	0	0	0	0	1	1.067938	-34.303317	220	1	1.172781	0.9;
	4900	1	0	0	0	0	1	1.00045	-17.105601	150	1	1.1	0.9;
	4901	2	186.2	3.7	0	0	1	0.976364	-25.47293	110	1	1.1	0.9;
	4902	2	0	0	0	0	1	1.054289	-39.692653	380	1	1.126862	0.9;
	4903	2	276.9	61.1	0	0	1	1.076242	-13.307428	220	1	1.143001	0.9;
	4904	2	0	0	0	0	1	1.094237	55.306032	380	1	1.117992	0.9;
	4905	1	54.76072	33.9	0	0	1	1.000758	-24.420775	150	1	1.1	0.9;
	4906	2	0	0	0	0	1	1.021316	14.202934	380	1	1.1	0.9;
	4907	1	5.205771	1.7	0	0	1	1.02604	-27.370287	110	1	1.1	0.9;
	4908	1	0	0	0	0	1	0.99488	-23.99917	150	1	1.1	0.9;
	4909	1	0	0	0	0	1	1.09547	-17.547933	380	1	1.144952	0.9;
	4910	1	6.607326	2.4	0	0	1	0.996858	-22.784044	110	1	1.1	0.9;
	4911	1	11.91321	4.3	0	0	1	1.021698	-21.446636	110	1	1.1	0.9;
	4912	2	0.900999	27.4	0	0	1	1.060864	-35.231323	220	1	1.109521	0.9;
	4913	1	69.77736	26	0	0	1	1.052011	-20.447909	220	1	1.121382	0.9;
	4914	2	0	0	0	0	1	1.092763	9.585602	380	1	1.126797	0.9;
	4915	1	0	0	0	0	1	1.092164	-17.089345	220	1	1.165688	0.9;
	4916	1	0	0	0	11.0254	1	1.047716	-25.536527	150	1	1.1	0.9;
	4917	1	0	0	0	0	1	1.073939	1.972031	380	1	1.125008	0.9;
	4918	1	32.7363	5.3	0	0	1	1.091409	-11.958431	220	1	1.166598	0.9;
	4919	1	0	0	0	0	1	1.07824	-20.248633	220	1	1.141149	0.9;
	4920	1	349.3	144.3	0	0	1	1.022737	-14.448106	220	1	1.1	0.9;
	4921	2	75.5838	8.6	0	0	1	1.047569	-22.417358	380	1	1.129016	0.9;
	4922	1	0	0	0	0	1	1.054859	-14.422786	380	1	1.116387	0.9;
	4923	1	28.63174	7.3	0	0	1	1.019296	2.766325	110	1	1.1	0.9;
	4924	1	0	0	0	0	1	1.076892	-19.367186	110	1	1.135171	0.9;
	4925	2	0.500555	17.9	0	0	1	1.092737	-31.886777	380	1	1.164296	0.9;
	4926	1	152.8	38	0	0	1	1.060005	2.206369	150	1	1.119714	0.9;
	4927	1	53.35916	44.2	0	0	1	1.063202	27.238018	380	1	1.125682	0.9;
	4928	2	0	0	0	0	1	1.039955	13.009838	220	1	1.1	0.9;
	4929	1	0	0	0	0	1	1.030718	-25.984664	150	1	1.15644	0.9;
	4930	1	103.3	24.1	0	0	1	1.031814	-43.155048	220	1	1.1	0.9;
	4931	2	0	0	0	0	1	0.981264	-19.880686	110	1	1.1	0.9;
	4932	1	0	0	0	0	1	1.060752	-14.796115	380	1	1.112686	0.9;
	4933	1	104.7	-12.8	0	0	1	1.081761	-15.459274	220	1	1.128005	0.9;
	4934	1	0	0	0	0	1	1.08474	57.231504	380	1	1.1	0.9;
	4935	1	116.4	-8.8	0	0	1	1.081727	-9.225019	220	1	1.134125	0.9;
	4936	2	162.8	45.7	0	0	1	1.081788	-3.806726	220	1	1.142805	0.9;
	4937	2	0	0	0	0	1	1.042938	14.663837	150	1	1.1	0.9;
	4938	1	2.803108	1	0	0	1	1.033976	-22.768329	110	1	1.112925	0.9;
	4939	2	0	0	0	20	1	1.050579	10.648082	380	1	1.108001	0.9;
	4940	1	45.45039	7.1	0	0	1	1.019726	-23.879442	110	1	1.1	0.9;
	4941	1	9.610656	1.8	0	0	1	0.989307	-17.364248	110	1	1.1	0.9;
	4942	2	3.804218	5.1	0	0	1	0.951174	-26.909034	110	1	1.1	0.9;
	4943	1	0	0	0	0	1	1.061409	-26.838087	150	1	1.128591	0.9;
	4944	1	0	0	0	0	1	1.089135	-0.220896	150	1	1.164961	0.9;
	4945	1	0	0	0	0	1	1.065434	-40.084315	220	1	1.123758	0.9;
	4946	1	82.09102	39.8	0	0	1	1.031065	-43.537885	150	1	1.1	0.9;
	4947	2	66.57381	2.3	0	0	1	0.988442	63.228004	150	1	1.1	0.9;
	4948	1	0	0	0	0	1	1.019852	-18.597204	150	1	1.1	0.9;
	4949	1	0	0	0	0	1	1.081763	-7.412323	220	1	1.148101	0.9;
	4950	1	0	0	0	0	1	0.944069	-18.310009	110	1	1.1	0.9;
	4951	2	6.00666	2.2	0	0	1	1.061273	-6.009379	110	1	1.136817	0.9;
	4952	1	0	0	0	0	1	1.076531	-20.35319	220	1	1.155741	0.9;
	4953	1	211.3	69.3	0	0	1	1.010906	-47.175229	220	1	1.1	0.9;
	4954	2	0	0	0	0	1	0.995489	-17.918991	380	1	1.1	0.9;
	4955	2	0	0	0	0	1	1.067632	-17.816115	380	1	1.150235	0.9;
	4956	2	217.1	52.6	0	0	1	1.054592	9.631356	380	1	1.135672	0.9;
	4957	1	88.39801	-5.9	0	0	1	1.04718	-37.939527	220	1	1.1	0.9;
	4958	1	25.22797	3.3	0	0	1	1.089663	-34.962151	380	1	1.139783	0.9;
	4959	2	0	0	0	0	1	1.021727	-42.753892	220	1	1.1	0.9;
	4960	1	1.00111	0.3	0	0	1	1.024373	-23.536247	110	1	1.124999	0.9;
	4961	2	148.8	24.9	0	0	1	1.012138	-8.852713	220	1	1.1	0.9;
	4962	1	106.6	-18.8	0	0	1	1.044249	-37.512326	220	1	1.17813	0.9;
	4963	1	3.303663	1.1	0	0	1	1.051385	-28.604445	110	1	1.1	0.9;
	4964	2	0	0	0	0	1	1.041263	-45.097116	380	1	1.109216	0.9;
	4965	1	1.301443	0	0	0	1	1.032697	-15.908548	110	1	1.1	0.9;
	4966	1	0	0	0	0	1	1.022946	31.456988	380	1	1.1	0.9;
	4967	1	8.909878	2.5	0	0	1	0.956588	-23.592797	110	1	1.1	0.9;
	4968	1	1.00111	1	0	0	1	1.046831	-19.647359	110	1	1.10936	0.9;
	4969	1	0	0	0	0	1	1.051463	-44.12176	380	1	1.11841	0.9;
	4970	1	0	0	0	0	1	1.077392	-25.285642	220	1	1.1672	0.9;
	4971	2	25.32808	70.8	0	0	1	1.018737	30.799273	380	1	1.1	0.9;
	4972	1	30.9343	5.8	0	0	1	1.071861	-25.870434	220	1	1.163836	0.9;
	4973	2	0	0	0	0	1	1.100684	-8.723889	380	1	1.171454	0.9;
	4974	1	144	56.6	0	0	1	1.052406	-36.555686	220	1	1.134112	0.9;
	4975	1	0	0	0	-22.8665	1	1.007473	-27.596761	150	1	1.192761	0.9;
	4976	1	0	0	0	0	1	1.022262	-1.872114	220	1	1.1	0.9;
	4977	2	0	0	0	0	1	1.082763	4.401725	380	1	1.152109	0.9;
	4978	2	0	0	0.041323	0	1	1.031727	-13.531246	220	1	1.107903	0.9;
	4979	2	148.2	31.2	0	0	1	1.035545	-29.482161	220	1	1.12856	0.9;
	4980	1	11.8131	2.4	0	0	1	0.967355	-12.901202	110	1	1.1	0.9;
	4981	2	0	0	0	20	1	1.079815	-14.655749	380	1	1.158587	0.9;
	4982	1	0	0	0.471696	0	1	1.05562	3.031422	380	1	1.11184	0.9;
	4983	1	0	0	0	0	1	1.014599	7.654459	220	1	1.1	0.9;
	4984	1	0	0	0	0	1	1.035171	-44.706388	220	1	1.10933	0.9;
	4985	1	0	0	0	0	1	1.018058	-33.211081	220	1	1.1	0.9;
	4986	2	0	0	0	79.23356	1	1.063786	54.079777	380	1	1.1	0.9;
	4987	1	1.62675	8.8	0	0	1	1.03063	-24.285357	150	1	1.1	0.9;
	4988	1	0	0	0.239002	0	1	1.016639	-17.037746	220	1	1.1	0.9;
	4989	1	574.3	94.6	0	0	1	1.042213	-33.792438	380	1	1.10619	0.9;
	4990	1	20.12231	7.9	0	0	1	0.991493	-20.407539	110	1	1.1	0.9;
	4991	1	111.5	24.8	0	0	1	1.054646	-48.98724	220	1	1.109987	0.9;
	4992	1	0	0	0	0	1	0.990979	-33.803375	220	1	1.1	0.9;
	4993	1	0	0	0	0	1	1.07706	-19.521274	220	1	1.141028	0.9;
	4994	1	342.7	82.2	0	0	1	1.100084	-11.49184	380	1	1.142121	0.9;
	4995	2	17.61954	5.5	0	0	1	1.021909	-34.118747	110	1	1.1	0.9;
	4996	2	164.7	33.9	0	0	1	1.093887	-7.200508	380	1	1.151751	0.9;
	4997	1	29.53274	25.1	0	0	1	0.92637	-29.465867	150	1	1.1	0.864633;
	4998	1	17.61954	7.7	0	0	1	1.012557	-35.136778	110	1	1.1	0.9;
	4999	1	0	0	0	0	1	0.976226	-23.965635	150	1	1.1	0.9;
	5000	1	0	0	0.287458	0	1	1.026492	-4.778401	380	1	1.1	0.9;
	5001	1	499	191.1	0	0	1	1.058661	-43.962347	380	1	1.126091	0.9;
	5002	1	67.67503	6.2	0	0	1	1.058873	-54.041713	220	1	1.1	0.9;
	5003	2	73.18114	13.1	0	0	1	1.038318	-37.281212	220	1	1.1	0.9;
	5004	2	456	123.8	0	0	1	1.042773	-3.644362	220	1	1.115518	0.9;
	5005	1	0	0	0	0	1	1.012716	54.217215	220	1	1.1	0.9;
	5006	1	22.12453	8	0	0	1	1.022567	-15.118502	110	1	1.122743	0.9;
	5007	1	18.2202	6	0	0	1	0.960809	-29.36477	110	1	1.1	0.9;
	5008	1	34.23796	-19.6	0	0	1	0.999475	-38.73639	150	1	1.1	0.9;
	5009	1	4.404884	2.2	0	0	1	1.047313	-32.872392	110	1	1.1	0.9;
	5010	1	0	0	0	0	1	1.059198	-23.383655	220	1	1.133261	0.9;
	5011	1	30.43374	10.1	0	0	1	1.042055	-19.357263	110	1	1.148851	0.9;
	5012	1	0	0	0	0	1	1.079127	0.900468	220	1	1.163168	0.9;
	5013	1	150.2	25.5	0	0	1	1.069888	2.301633	380	1	1.110092	0.9;
	5014	1	14.51609	5.3	0	0	1	0.974359	-17.671904	110	1	1.1	0.9;
	5015	2	239.5	61.6	0	0	1	1.078794	-13.377441	220	1	1.138656	0.9;
	5016	2	0	0	0	0	1	1.070866	15.596049	150	1	1.110515	0.9;
	5017	1	0	0	0	0	1	1.068356	-15.137081	380	1	1.120502	0.9;
	5018	1	93.6	77.1	0	20	1	1.047422	20.739742	380	1	1.105899	0.9;
	5019	1	0	0	0	0	1	1.081809	-3.806916	220	1	1.142827	0.9;
	5020	2	154.9	18.7	0	0	1	1.021364	5.965106	220	1	1.1	0.9;
	5021	1	0	0	0	0	1	1.055375	61.495633	220	1	1.108872	0.9;
	5022	1	0	0	0	0	1	1.08235	-41.084062	220	1	1.153552	0.9;
	5023	2	0	0	0	0	1	1.084273	-12.350951	220	1	1.13833	0.9;
	5024	1	0	0	0	0	1	1.05236	-28.525907	110	1	1.122248	0.9;
	5025	1	68.07548	42.2	0	0	1	1.030337	-25.114039	150	1	1.1	0.9;
	5026	2	0	0	0.095843	0	1	1.069062	-20.703581	220	1	1.140006	0.9;
	5027	1	12.21354	4.4	0	0	1	0.947832	-18.693605	110	1	1.1	0.9;
	5028	1	0	0	0	0	1	0.997558	-16.556263	110	1	1.1	0.9;
	5029	1	11.71299	2.2	0	0	1	1.027357	-31.05218	110	1	1.1	0.9;
	5030	1	6.707437	3	0	0	1	0.985585	-20.677355	110	1	1.1	0.9;
	5031	1	0	0	0	0	1	1.053645	-12.549097	380	1	1.121692	0.9;
	5032	1	47.65283	29.5	0	0	1	0.961864	-29.660638	150	1	1.1	0.876388;
	5033	1	33.03663	8.1	0	0	1	1.00243	-39.868186	110	1	1.1	0.9;
	5034	2	0	0	0	0	1	1.079182	-23.957885	220	1	1.139285	0.9;
	5035	1	10.0111	3	0	0	1	0.966752	-26.950565	110	1	1.1	0.9;
	5036	2	0	0	0	0	1	0.994643	-25.42766	150	1	1.1	0.9;
	5037	1	30.0333	12.2	0	0	1	1.07847	0.167244	150	1	1.147883	0.9;
	5038	1	0	0	0	0	1	1.044181	-24.964292	220	1	1.140825	0.9;
	5039	1	15.11676	5	0	0	1	0.93541	-34.037955	110	1	1.1	0.9;
	5040	1	23.92653	8.7	0	0	1	1.04051	-20.687933	110	1	1.1	0.9;
	5041	1	2.22375	7.8	0	0	1	0.989936	-25.381475	150	1	1.1	0.9;
	5042	1	50.25572	16.5	0	0	1	0.912694	-28.341696	110	1	1.1	0.9;
	5043	1	0	0	0	0	1	1.033411	9.14053	380	1	1.108203	0.9;
	5044	2	0	0	0	0	1	1.049026	12.691023	380	1	1.109642	0.9;
	5045	1	21.22353	6.7	0	0	1	0.973821	-27.371839	110	1	1.1	0.9;
	5046	1	0	0	0	0	1	1.041647	13.58319	380	1	1.109816	0.9;
	5047	1	0	0	0	0	1	1.023019	-38.798322	220	1	1.1	0.9;
	5048	1	0	0	0	0	1	1.02325	-21.644839	150	1	1.1	0.9;
	5049	1	31.13452	12	0	0	1	0.986222	-21.061048	110	1	1.1	0.9;
	5050	1	0	0	0.221009	0	1	1.100943	-41.681361	380	1	1.160498	0.9;
	5051	1	21.62398	7.8	0	0	1	1.03022	-14.985172	110	1	1.13924	0.9;
	5052	2	0	0	0	0	1	1.072818	-11.68441	220	1	1.132596	0.9;
	5053	1	5.105661	1.8	0	0	1	1.055226	-5.635792	110	1	1.131646	0.9;
	5054	1	0	0	0	0	1	1.060166	2.069044	380	1	1.115103	0.9;
	5055	1	0	0	0	0	1	1.052872	-17.751422	220	1	1.117748	0.9;
	5056	1	0	0	0	0	1	1.040438	-29.183477	150	1	1.1	0.9;
	5057	1	0	0	0	0	1	0.977676	-11.998185	110	1	1.1	0.9;
	5058	1	23.52608	8.5	0	0	1	1.005654	-22.603518	110	1	1.1	0.9;
	5059	1	0	0	0	0	1	1.072705	-21.341834	220	1	1.155608	0.9;
	5060	2	0	-0.1	0	0	1	1.085591	-13.374423	220	1	1.170136	0.9;
	5061	1	0	0	0	0	1	1.084934	-23.600508	220	1	1.143063	0.9;
	5062	1	99.7	6	0	0	1	1.045969	-12.045469	220	1	1.116073	0.9;
	5063	1	0	0	0	0	1	1.063738	-38.618842	380	1	1.128926	0.9;
	5064	1	220	68.3	0	0	1	1.063826	-21.424801	220	1	1.1	0.9;
	5065	1	0	0	0	0	1	1.088644	-27.736759	380	1	1.169894	0.9;
	5066	1	26.62953	8.6	0	0	1	0.922125	-30.37203	110	1	1.1	0.9;
	5067	1	16.21798	5.3	0	0	1	0.935217	-31.814997	110	1	1.1	0.9;
	5068	1	130.7	89.1	0	0	1	1.027526	-4.458232	220	1	1.1	0.9;
	5069	1	3.203552	1	0	0	1	1.022735	-15.026146	110	1	1.1	0.9;
	5070	1	1.902109	0.7	0	0	1	1.020412	-16.099983	110	1	1.122615	0.9;
	5071	2	0	0	0	0	1	1.050954	16.948067	220	1	1.1	0.9;
	5072	1	0	0	0	0	1	1.024268	-33.75346	110	1	1.1	0.9;
	5073	1	0	0	0	0	1	1.074792	2.456204	220	1	1.154924	0.9;
	5074	1	0	0	0	0	1	1.073472	-10.96851	150	1	1.138678	0.9;
	5075	1	0	0	0	0	1	0.981817	-25.683296	150	1	1.1	0.9;
	5076	1	0	0	0	0	1	1.0264	-30.060688	380	1	1.13422	0.9;
	5077	1	118.6	41.5	0	0	1	1.061639	-2.598967	150	1	1.13956	0.9;
	5078	1	20.82309	7.9	0	0	1	1.042954	-46.936243	220	1	1.1	0.9;
	5079	1	19.52164	7	0	0	1	1.006074	-19.84962	110	1	1.1	0.9;
	5080	2	275.3	55.2	0	0	1	1.078456	-27.703272	220	1	1.126987	0.9;
	5081	1	168.5	54.8	0	0	1	1.074597	2.141231	150	1	1.145029	0.9;
	5082	1	172.8	11.3	0	0	1	1.058722	-15.489879	220	1	1.122524	0.9;
	5083	1	0	0	0	0	1	1.066986	-8.35975	220	1	1.129167	0.9;
	5084	1	0	0	0.064829	0	1	1.089323	68.732225	380	1	1.1	0.9;
	5085	1	0	0	0	0	1	1.027633	-19.398775	220	1	1.1	0.9;
	5086	1	0	0	0	0	1	1.043365	-30.281415	380	1	1.106749	0.9;
	5087	1	0	0	0	0	1	1.025887	-23.238837	150	1	1.1	0.9;
	5088	1	34.13785	12.4	0	0	1	1.055788	-20.770248	110	1	1.113251	0.9;
	5089	1	449.4	14.6	0	0	1	1.086575	-45.850231	380	1	1.12846	0.9;
	5090	2	30.33363	-3.2	0	0	1	1.042074	60.842738	150	1	1.1	0.9;
	5091	1	99.5	9.2	0	0	1	1.077771	-22.928705	220	1	1.163046	0.9;
	5092	2	0	0	0	0	1	1.101816	58.124549	380	1	1.160079	0.9;
	5093	1	0	0	0	0	1	0.996214	-24.120702	110	1	1.1	0.9;
	5094	1	225.2	35.2	0	0	1	1.073155	-34.531887	220	1	1.173707	0.9;
	5095	1	38.04218	3.4	0	0	1	1.029672	-44.691559	220	1	1.1	0.9;
	5096	1	0	0	0	0	1	1.066954	-3.076842	150	1	1.137201	0.9;
	5097	1	0	0	0.191685	0	1	1.024571	-22.341365	220	1	1.1	0.9;
	5098	2	301.6	122.3	0	0	1	1.05	11.662187	380	1	1.1	0.9;
	5099	1	6.507215	3.7	0	0	1	0.936303	-25.846412	110	1	1.1	0.9;
	5100	1	158.3	-20.5	0	0	1	1.060615	-15.034148	220	1	1.128768	0.9;
	5101	1	513.4	199.8	0	0	1	1.078285	-33.032181	380	1	1.157757	0.9;
	5102	1	152.7	6	0	0	1	1.07111	-24.929929	120	1	1.133557	0.9;
	5103	1	0	0	0	0	1	1.092905	-26.273766	380	1	1.156912	0.9;
	5104	1	13.71521	4.7	0	0	1	0.948294	-24.111507	110	1	1.1	0.9;
	5105	1	0	0	0	0	1	1.034318	-32.48517	220	1	1.134568	0.9;
	5106	2	5.405994	0.9	0	0	1	0.948178	-24.463149	110	1	1.1	0.9;
	5107	1	0	0	0	0	1	0.939484	-29.984236	110	1	1.1	0.9;
	5108	1	0	0	0	0	1	1.08242	-4.285377	220	1	1.137731	0.9;
	5109	1	105.5	31.1	0	0	1	1.008572	-39.150101	220	1	1.1	0.9;
	5110	2	27.93097	6.7	0	0	1	0.9545	-25.785317	110	1	1.1	0.9;
	5111	1	29.1323	13.5	0	0	1	0.987314	-51.222018	220	1	1.1	0.9;
	5112	1	21.12342	13.1	0	0	1	1.027479	-25.796733	150	1	1.1	0.9;
	5113	1	0	0	0	0	1	0.980798	-28.746266	150	1	1.1	0.9;
	5114	1	0	0	0	0	1	1.033124	-22.980448	150	1	1.1	0.9;
	5115	2	0	0	0	0	1	1.050197	-21.21685	110	1	1.154136	0.9;
	5116	1	0	0	0	0	1	1.027798	-24.471899	150	1	1.1	0.9;
	5117	1	0	0	0	0	1	1.027854	-11.101795	110	1	1.1	0.9;
	5118	2	172.9	-29.6	0	0	1	1.04324	9.33319	220	1	1.123977	0.9;
	5119	1	0	0	0	0	1	1.046587	-28.5271	220	1	1.113697	0.9;
	5120	1	72.78069	45.1	0	0	1	1.024103	-27.828033	150	1	1.1	0.9;
	5121	1	0	0	0	0	1	1.030938	-51.273949	220	1	1.1	0.9;
	5122	1	45.04995	40	0	0	1	0.994677	9.895038	220	1	1.1	0.9;
	5123	2	26.4293	21.7	0	0	1	1.004818	-47.686658	220	1	1.1	0.9;
	5124	1	20.0222	8.5	0	0	1	0.95373	-19.019817	110	1	1.1	0.9;
	5125	1	6.306993	2.5	0	0	1	0.968883	-25.519409	110	1	1.1	0.9;
	5126	1	0	0	0	0	1	1.044202	-12.962089	380	1	1.116476	0.9;
	5127	1	0	0	0	0	1	1.073731	15.027848	380	1	1.129437	0.9;
	5128	1	0	0	0	0	1	0.979605	-24.633729	150	1	1.1	0.9;
	5129	1	45.04995	5	0	0	1	1.079676	-23.06798	380	1	1.132097	0.9;
	5130	2	104.1	9.8	0	0	1	1.011642	-7.301264	220	1	1.112722	0.9;
	5131	2	5.506105	2.2	0	0	1	0.998327	-34.518067	110	1	1.1	0.9;
	5132	2	0	0	0	0	1	1.055681	2.027819	150	1	1.115174	0.9;
	5133	2	198.7089	60.3	0	0	1	1.054409	-0.594917	220	1	1.12606	0.9;
	5134	1	0	0	0	0	1	1.019214	-21.266957	220	1	1.1	0.9;
	5135	2	0	0	0	0	1	1.038053	-22.736158	380	1	1.129734	0.9;
	5136	1	-27.4304	9	0	0	1	1.041037	-20.248675	150	1	1.1	0.9;
	5137	1	30.0333	9.9	0	0	1	0.957467	-31.668028	110	1	1.1	0.9;
	5138	1	0	0	0	0	1	0.970726	-24.528316	150	1	1.1	0.9;
	5139	1	185.7	18.4	0	0	1	1.030655	-38.060656	220	1	1.1	0.9;
	5140	1	0	0	0	11.74184	1	1.010122	-23.832912	150	1	1.1	0.9;
	5141	2	0	0	0	0	1	1.081	1.148113	220	1	1.164467	0.9;
	5142	1	0	0	0	0	1	1.065707	33.567766	380	1	1.1	0.9;
	5143	1	124.8	34.7	0	0	1	1.052665	-10.96252	220	1	1.122286	0.9;
	5144	2	89.69945	7.9	0	0	1	1.049228	-39.992266	220	1	1.1	0.9;
	5145	2	-218.342	-16.8	0	0	1	1.013182	-30.312424	220	1	1.1	0.9;
	5146	2	0	0	0	0	1	1.036545	-42.689597	220	1	1.107536	0.9;
	5147	1	25.02775	1.4	0	0	1	0.985989	-24.54982	110	1	1.1	0.9;
	5148	1	139.7	9.7	0	0	1	0.995926	-37.31835	220	1	1.1	0.9;
	5149	1	67.27459	1.5	0	0	1	1.061976	0.847725	220	1	1.138139	0.9;
	5150	1	0	0	0	0	1	1.04207	7.12458	380	1	1.1	0.9;
	5151	1	54.96094	17.6	0	0	1	1.018035	-36.440764	110	1	1.1	0.9;
	5152	1	6.206882	2.2	0	0	1	1.041729	-20.442445	110	1	1.1	0.9;
	5153	2	43.44817	7	0	0	1	1.068591	-19.036287	220	1	1.130767	0.9;
	5154	1	25.12786	8.8	0	0	1	0.94804	-27.202569	110	1	1.1	0.9;
	5155	2	77.88636	11.2	0	0	1	1.0404	63.390913	150	1	1.1	0.9;
	5156	1	153.5	46	0	0	1	1.049013	-12.938199	220	1	1.112394	0.9;
	5157	1	198.1	33.9	0	0	1	1.053553	-48.770605	220	1	1.116574	0.9;
	5158	1	0	0	0	0	1	1.069096	2.602341	380	1	1.109427	0.9;
	5159	1	0	0	0	0	1	1.083534	-19.155472	380	1	1.158993	0.9;
	5160	1	34.73852	16.3	0	0	1	1.075022	-39.231893	380	1	1.139915	0.9;
	5161	1	52.25794	19.6	0	0	1	1.063529	-9.480673	220	1	1.123737	0.9;
	5162	1	19.1212	11.9	0	0	1	0.98826	-24.112721	150	1	1.1	0.9;
	5163	1	14.11565	5.2	0	0	1	1.077391	-7.518845	220	1	1.143061	0.9;
	5164	1	0	0	0	0	1	1.060318	-17.527667	220	1	1.127079	0.9;
	5165	1	195.1	-14.2	0	0	1	1.046688	-44.655582	220	1	1.163382	0.9;
	5166	2	0	0	0	0	1	1.087632	-26.850351	380	1	1.144853	0.9;
	5167	1	0	0	0	0	1	1.073103	-15.458439	380	1	1.121526	0.9;
	5168	1	40.24462	14.4	0	0	1	1.054677	-34.229073	220	1	1.132608	0.9;
	5169	1	142.8	20.3	0	0	1	1.046025	-26.691916	220	1	1.130325	0.9;
	5170	1	45.35028	28.1	0	0	1	1.062261	-20.249097	150	1	1.1	0.9;
	5171	1	131.6	33.8	0	0	1	1.045946	-42.371403	220	1	1.1	0.9;
	5172	1	11.8131	4.2	0	0	1	0.996908	-30.458105	110	1	1.2	0.8;
	5173	1	0	0	0	0	1	1.05047	-6.18502	220	1	1.13874	0.9;
	5174	1	31.93541	4.4	0	0	1	1.011277	-37.0678	110	1	1.1	0.9;
	5175	1	0	0	0	0	1	1.017749	-27.199363	110	1	1.1	0.9;
	5176	1	29.03219	18	0	0	1	1.061731	-18.470734	150	1	1.1	0.9;
	5177	1	119.6	10.6	0	0	1	1.013167	-47.625393	220	1	1.1	0.9;
	5178	2	0	0	0	0	1	1.09675	-44.75309	120	1	1.173402	0.9;
	5179	1	32.23574	11.7	0	0	1	1.044542	-21.852502	110	1	1.105519	0.9;
	5180	1	0	0	0	0	1	1.030488	-26.508284	150	1	1.1	0.9;
	5181	2	0	0	0	0	1	1.088605	-8.5472	380	1	1.143502	0.9;
	5182	1	0.400444	0.3	0	0	1	0.987063	-25.893547	110	1	1.1	0.9;
	5183	1	0	0	0	0	1	1.03458	10.575531	380	1	1.1	0.9;
	5184	2	190.5	48.3	0	0	1	1.075474	-4.567192	220	1	1.129123	0.9;
	5185	1	20.72298	-2.5	0	0	1	0.977641	-22.809325	110	1	1.1	0.9;
	5186	2	0	0	0	0	1	1.048317	-33.124522	220	1	1.111191	0.9;
	5187	2	99.51033	-9.1	0	0	1	1.048313	62.141167	150	1	1.1	0.9;
	5188	1	0	0	0	0	1	1.067198	23.060115	380	1	1.120305	0.9;
	5189	2	0	0	0	0	1	1.098289	62.830076	380	1	1.114039	0.9;
	5190	1	0	0	0	0	1	0.940762	-34.413098	220	1	1.1	0.9;
	5191	1	0	0	0	0	1	1.053425	56.316615	220	1	1.1	0.9;
	5192	1	150.1	35	0	0	1	1.058599	-21.119309	220	1	1.122307	0.9;
	5193	1	337.6	-15.4	0	0	1	1.012204	-25.153549	220	1	1.1	0.9;
	5194	2	7.708547	2.2	0	0	1	1.075636	-30.848461	110	1	1.108398	0.9;
	5195	1	141.2	45.4	0	0	1	1.057156	-9.168029	220	1	1.112692	0.9;
	5196	1	20.12231	8.6	0	0	1	0.970027	-22.858793	110	1	1.1	0.9;
	5197	1	19.42153	6.3	0	0	1	0.974839	-27.251443	110	1	1.1	0.9;
	5198	2	0.500555	0	0	0	1	1.064318	-38.442621	220	1	1.105698	0.9;
	5199	1	42.44706	14	0	0	1	0.969067	-29.024191	110	1	1.1	0.9;
	5200	1	4.605106	0.9	0	0	1	0.993204	-25.271184	110	1	1.1	0.9;
	5201	1	0	0	0	0	1	1.028916	-24.443996	150	1	1.1	0.9;
	5202	1	0	0	0	0	1	1.051748	-21.283409	110	1	1.112186	0.9;
	5203	1	0	0	0	0	1	1.01583	-19.909253	220	1	1.1	0.9;
	5204	2	167.7	60.2	0	0	1	1.037352	2.446505	220	1	1.11288	0.9;
	5205	1	0	0	0	0	1	1.048838	-15.521327	220	1	1.117683	0.9;
	5206	2	181.8	18.1	0	0	1	1.08384	-26.056992	220	1	1.176447	0.9;
	5207	1	15.31698	3.5	0	0	1	1.018542	-12.021067	110	1	1.1	0.9;
	5208	2	31.73519	5.4	0	0	1	0.979545	-22.109928	110	1	1.1	0.9;
	5209	1	0	0	0	0	1	1.061615	-9.755699	220	1	1.129104	0.9;
	5210	1	0	0	0	0	1	1.033281	-16.836889	150	1	1.1	0.9;
	5211	1	0	0	0	0	1	1.053128	3.80764	150	1	1.109204	0.9;
	5212	1	38.64284	13.8	0	0	1	1.011777	-22.531672	110	1	1.1	0.9;
	5213	2	0	0	0	0	1	1.079682	20.341577	220	1	1.157871	0.9;
	5214	1	0	0	0	0	1	1.062645	-22.991952	220	1	1.136795	0.9;
	5215	1	13.71521	4.5	0	0	1	0.957394	-31.68357	110	1	1.1	0.9;
	5216	1	71.9798	28.8	0	0	1	1.056754	-33.389872	220	1	1.133469	0.9;
	5217	1	371.2	73.5	0	0	1	1.054758	-44.334477	380	1	1.123408	0.9;
	5218	1	0	0	0	0	1	1.065874	-3.054446	150	1	1.135931	0.9;
	5219	1	-928	-491.4	0	0	1	1.062781	13.637912	380	1	1.129291	0.9;
	5220	2	174	47.7	0	0	1	1.024727	-35.880095	110	1	1.1	0.9;
	5221	1	52.95872	50.3	0	0	1	1.024061	-25.852448	110	1	1.2	0.8;
	5222	1	0	0	0	0	1	1.078182	-4.168825	220	1	1.139416	0.9;
	5223	1	222.6	62.6	0	0	1	1.03961	11.947993	380	1	1.1	0.9;
	5224	1	0	0	0	0	1	1.025485	-23.559049	150	1	1.1	0.9;
	5225	1	0	0	0	0	1	1.00757	-27.670908	150	1	1.1	0.9;
	5226	2	0	0	0	0	1	1.098447	-10.167082	380	1	1.153507	0.9;
	5227	1	0	0	0	0	1	1.079449	5.584069	150	1	1.148811	0.9;
	5228	1	0	0	0	0	1	1.034262	-19.908208	150	1	1.11407	0.9;
	5229	1	0	0	0	0	1	1.057292	0.508753	380	1	1.121597	0.9;
	5230	1	0	0	0	0	1	1.066701	-1.300022	220	1	1.134411	0.9;
	5231	1	99.4	4.3	0	0	1	1.022245	-24.832235	220	1	1.125919	0.9;
	5232	1	176.4	17.2	0	0	1	1.000607	-28.391153	220	1	1.1	0.9;
	5233	1	0	0	0	0	1	1.06495	-43.927039	220	1	1.113656	0.9;
	5234	1	0	0	0	0	1	1.071201	-2.650765	380	1	1.128669	0.9;
	5235	1	0	0	0	0	1	1.033527	-25.154105	150	1	1.1	0.9;
	5236	1	0	0	0	0	1	1.020822	-20.979814	150	1	1.1	0.9;
	5237	1	190.5	70.9	0	0	1	1.051016	-43.03966	380	1	1.120009	0.9;
	5238	2	0	0	0	0	1	1.084502	4.007361	150	1	1.152978	0.9;
	5239	1	0	0	0	0	1	1.088675	-2.785482	220	1	1.147928	0.9;
	5240	1	0	0	0	0	1	1.071168	32.825632	380	1	1.140646	0.9;
	5241	1	30.0333	10	0	0	1	0.967978	-26.879847	110	1	1.1	0.9;
	5242	1	222.7	80.5	0	0	1	1.077081	-8.979341	150	1	1.123998	0.9;
	5243	1	0	0	0	0	1	1.030141	-25.359954	150	1	1.1	0.9;
	5244	1	0	0	0	0	1	1.053018	-28.279883	220	1	1.1	0.9;
	5245	1	86.19556	-30.7	0	0	1	1.042988	-3.956599	220	1	1.1	0.9;
	5246	1	0	0	0	0	1	1.055964	-48.589282	220	1	1.111432	0.9;
	5247	2	0	0	0	20	1	1.067816	-10.846116	380	1	1.118889	0.9;
	5248	1	0	0	0	-25.1413	1	1.030753	-25.975063	150	1	1.15317	0.9;
	5249	1	69.07658	-49.6	0	0	1	1.075693	8.620576	110	1	1.144337	0.9;
	5250	1	0	0	0	0	1	1.013947	54.226407	220	1	1.1	0.9;
	5251	2	0	0	0	0	1	1.004273	-39.755826	110	1	1.1	0.9;
	5252	2	43.24795	19.5	0	0	1	1.039818	-39.680795	220	1	1.1	0.9;
	5253	1	0	0	0	0	1	1.054144	-15.632721	220	1	1.12368	0.9;
	5254	1	4.404884	1.1	0	0	1	1.046819	-32.283292	220	1	1.105296	0.9;
	5255	1	0	0	0	0	1	0.989124	-24.046929	110	1	1.1	0.9;
	5256	1	40.9454	13.4	0	0	1	1.071092	0.343159	150	1	1.134509	0.9;
	5257	1	6.106771	0	0	0	1	1.030557	-43.676057	220	1	1.1	0.9;
	5258	1	0	0	0	0	1	0.980294	64.238657	150	1	1.1	0.9;
	5259	1	124.6	15.9	0	0	1	1.041043	-22.069678	220	1	1.1	0.9;
	5260	1	0	0	0	0	1	0.957976	-28.060539	150	1	1.1	0.9;
	5261	1	334.1	103.5	0	0	1	1.02471	-51.361638	220	1	1.1	0.9;
	5262	1	5.03807	6.1	0	0	1	1.041592	-21.215923	150	1	1.117897	0.9;
	5263	1	0	0	0	0	1	1.050972	-31.365034	220	1	1.162838	0.9;
	5264	1	0	0	0	-25.7076	1	1.103459	-16.255802	380	1	1.163282	0.9;
	5265	1	0	0	0	0	1	1.034248	-25.406123	150	1	1.1	0.9;
	5266	1	0	0	0	0	1	1.075061	-13.824476	220	1	1.147038	0.9;
	5267	1	36.24018	12.8	0	0	1	0.970123	-17.73643	110	1	1.1	0.9;
	5268	1	167.6	-9	0	0	1	1.029315	-50.016274	150	1	1.1	0.9;
	5269	1	0	0	0	0	1	1.004611	12.444126	220	1	1.1	0.9;
	5270	1	0	0	0	0	1	1.043364	-21.670608	220	1	1.1	0.9;
	5271	1	-7.90877	2.6	0	0	1	1.062111	-15.142547	150	1	1.105546	0.9;
	5272	2	0	0	0	0	1	0.994227	-16.787644	220	1	1.1	0.9;
	5273	2	32.03552	14.9	0	0	1	0.99	-24.172807	110	1	1.1	0.9;
	5274	2	0	0	0	0	1	1.025818	11.683296	220	1	1.1	0.9;
	5275	1	0	0	0	0	1	0.939253	-25.672555	110	1	1.1	0.9;
	5276	2	190.6	51.6	0	0	1	1.053727	-11.951964	220	1	1.122997	0.9;
	5277	1	5.205771	1.3	0	0	1	1.081707	-46.636699	380	1	1.138653	0.9;
	5278	1	13.01443	2.4	0	0	1	0.98995	-23.171232	110	1	1.1	0.9;
	5279	1	0	0	0	22.68425	1	1.030833	-24.134994	150	1	1.1	0.9;
	5280	1	343.1	109.5	0	0	1	1.071488	-15.757207	380	1	1.139314	0.9;
	5281	1	0	0	0	0	1	0.961293	58.446793	220	1	1.1	0.9;
	5282	2	96.80733	19.4	0	0	1	1.065365	-39.933325	220	1	1.10867	0.9;
	5283	2	145.7	-38	0	0	1	1.103763	19.112191	380	1	1.170854	0.9;
	5284	2	0.900999	28.3	0	0	1	1.082474	-35.321601	380	1	1.132501	0.9;
	5285	2	0	0	0	0	1	1.089868	-31.223352	380	1	1.126808	0.9;
	5286	2	-91.6016	-4.9	0.029998	0	1	1.002682	-33.045353	220	1	1.1	0.9;
	5287	1	83.5	41.5	0	0	1	1.030445	-26.504949	150	1	1.1	0.9;
	5288	2	0	0	0	0	1	1.081591	-33.319262	220	1	1.132552	0.9;
	5289	1	155.5	38.3	0	0	1	1.059444	-0.781765	150	1	1.122434	0.9;
	5290	1	53.85971	17.7	0	0	1	0.933829	-27.860892	110	1	1.1	0.9;
	5291	2	183.1	33.9	0	0	1	1.086947	-43.378128	380	1	1.132659	0.9;
	5292	1	0	0	0	0	1	0.923046	-28.924124	150	1	1.1	0.872357;
	5293	1	-46.1512	15.1	0	0	1	1.04489	-15.267517	150	1	1.1	0.9;
	5294	1	0.900999	0.3	0	0	1	1.060839	-29.460457	110	1	1.125422	0.9;
	5295	1	0	0	0	0	1	1.04309	14.665291	150	1	1.1	0.9;
	5296	2	277.3	65.5	0	0	1	1.042389	-12.134736	220	1	1.113892	0.9;
	5297	1	0	0	0	0	1	1.046315	-9.832747	380	1	1.129749	0.9;
	5298	2	0	0	0	0	1	1.042909	-44.786728	220	1	1.1	0.9;
	5299	1	0	0	0	0	1	1.092306	-24.093277	380	1	1.14898	0.9;
	5300	1	0	0	0	0	1	1.003453	-24.232525	110	1	1.1	0.9;
	5301	1	0	0	0	0	1	1.065589	13.076699	380	1	1.115669	0.9;
	5302	2	0	0	0	0	1	1.089227	-15.879341	220	1	1.1261	0.9;
	5303	1	0	0	0	0	1	1.072125	-20.630428	220	1	1.111463	0.9;
	5304	1	0	0	0.294	0	1	1.036108	-0.782672	380	1	1.1	0.9;
	5305	1	20.52275	7.6	0	0	1	0.972582	-24.772902	110	1	1.1	0.9;
	5306	1	0	0	0	0	1	1.068911	-18.536889	380	1	1.121842	0.9;
	5307	1	0	0	0	0	1	1.040337	61.753812	150	1	1.1	0.9;
	5308	1	122.5	38.9	0	0	1	1.069029	-7.023066	220	1	1.128961	0.9;
	5309	1	18.42042	1.2	0	0	1	1.063317	-31.395123	220	1	1.126514	0.9;
	5310	2	0	0	0	0	1	1.092368	-8.075315	380	1	1.12834	0.9;
	5311	2	105.9	29.5	0	0	1	1.087409	-11.150147	220	1	1.152779	0.9;
	5312	1	78.38691	21.2	0	0	1	1.071928	-20.314126	220	1	1.175983	0.9;
	5313	1	0	0	0	0	1	1.034403	-24.995026	150	1	1.1	0.9;
	5314	2	2.903219	1.8	0	0	1	0.996478	-25.753619	150	1	1.1	0.9;
	5315	1	18.52053	11.5	0	0	1	1.024563	-23.569742	150	1	1.1	0.9;
	5316	1	0	0	0	0	1	0.960386	-25.429701	150	1	1.1	0.9;
	5317	2	0	0	0	0	1	1.084553	50.904537	380	1	1.115043	0.9;
	5318	2	187	108.2	0	0	1	1.054955	5.758469	220	1	1.129025	0.9;
	5319	1	20.72298	6.8	0	0	1	0.939517	-27.096282	110	1	1.1	0.9;
	5320	1	10.0111	-30	0	0	1	1.083482	-41.20359	220	1	1.154804	0.9;
	5321	1	55.26127	34.2	0	0	1	1.029874	-24.505648	150	1	1.1	0.9;
	5322	2	0	0	0	0	1	1.087263	-26.964113	380	1	1.143509	0.9;
	5323	1	145.7	122.2	0	0	1	1.062767	-9.945994	220	1	1.130019	0.9;
	5324	1	0	0	0	0	1	1.040116	-29.192576	150	1	1.1	0.9;
	5325	1	28.83197	-3.8	0	0	1	1.019083	-34.799984	110	1	1.1	0.9;
	5326	1	177.5	72.5	0	0	1	1.064852	9.49322	150	1	1.1	0.9;
	5327	1	0	0	0	0	1	1.062112	-41.498515	380	1	1.126682	0.9;
	5328	2	39.84418	13.8	0	0	1	0.9792	64.87197	150	1	1.1	0.9;
	5329	1	2.00222	1	0	0	1	0.961703	-24.666757	110	1	1.1	0.9;
	5330	1	0	0	0	0	1	1.013244	-22.629287	150	1	1.1	0.9;
	5331	1	0	0	0	0	1	1.086009	-45.528106	380	1	1.148768	0.9;
	5332	1	0	0	0	0	1	1.043755	-18.442865	220	1	1.1	0.9;
	5333	1	0	0	0	0	1	0.966103	-25.11584	150	1	1.1	0.9;
	5334	1	159.9773	76.4	0	0	1	1.038727	25.751839	220	1	1.1	0.9;
	5335	2	0	0	0	0	1	1.067909	-10.001935	220	1	1.143208	0.9;
	5336	2	0	0	0	0	1	1.097763	8.751083	380	1	1.129835	0.9;
	5337	2	25.2	18.5	0	0	1	1.104658	-11.877837	380	1	1.185585	0.9;
	5338	1	0	0	0	0	1	1.087051	-35.841413	380	1	1.133253	0.9;
	5339	2	0	0	0	0	1	1.067136	-19.13722	220	1	1.129829	0.9;
	5340	1	16.4182	5.3	0	0	1	0.960675	-20.129332	110	1	1.1	0.9;
	5341	1	6.907659	-11.1	0	0	1	1.028892	16.248611	110	1	1.1	0.9;
	5342	1	0	0	0	0	1	1.083759	15.065877	380	1	1.129485	0.9;
	5343	1	0	0	0	0	1	1.071432	-41.626287	220	1	1.12278	0.9;
	5344	1	14.01554	5.5	0	0	1	1.043458	-12.279921	220	1	1.113732	0.9;
	5345	1	49.85527	14.1	0	0	1	1.065021	-43.914517	220	1	1.113662	0.9;
	5346	1	26.02886	10.6	0	0	1	1.063365	11.010672	110	1	1.142432	0.9;
	5347	2	0	0	0	0	1	1.045455	-3.852874	220	1	1.1	0.9;
	5348	1	0	0	0	0	1	1.002737	-27.984949	150	1	1.1	0.9;
	5349	2	0	0	0	208.8414	1	1.0245	15.617812	220	1	1.1	0.9;
	5350	1	0	0	0	0	1	1.082411	16.159047	380	1	1.127152	0.9;
	5351	1	-85.695	-2	0	0	1	1.092334	-13.202235	220	1	1.140173	0.9;
	5352	1	0	0	0	0	1	1.032264	-21.776765	150	1	1.109205	0.9;
	5353	1	7.207992	3.3	0	0	1	0.979196	-21.3216	110	1	1.1	0.9;
	5354	1	0	0	0	0	1	1.005349	-23.450164	150	1	1.1	0.9;
	5355	1	0	-7.2	0	0	1	1.010566	-21.646959	150	1	1.1	0.9;
	5356	1	196.3	-10.4	0	0	1	1.058254	-12.587404	220	1	1.116429	0.9;
	5357	1	5.305883	2.3	0	0	1	0.938814	-24.951004	110	1	1.1	0.9;
	5358	1	186.1	14.8	0	0	1	1.083328	-8.373115	220	1	1.153396	0.9;
	5359	1	10.51165	1.3	0	0	1	0.989263	-27.684044	110	1	1.1	0.9;
	5360	1	181.9	92.6	0	0	1	1.02502	-18.249224	220	1	1.1	0.9;
	5361	1	42.54717	14	0	0	1	0.940579	-33.679373	110	1	1.1	0.9;
	5362	1	0	0	0	0	1	1.04617	-22.229073	220	1	1.1	0.9;
	5363	1	0	0	0	0	1	1.055662	-19.329187	220	1	1.107818	0.9;
	5364	1	80.9898	30.1	0	0	1	1.055861	-27.174209	220	1	1.118414	0.9;
	5365	1	94.1	39	0	0	1	1.054684	1.623613	380	1	1.111944	0.9;
	5366	1	0	0	0	0	1	0.98066	-23.884148	150	1	1.1	0.9;
	5367	2	0	0	0	0	1	1.075388	12.779146	110	1	1.152225	0.9;
	5368	1	61.66837	17.4	0	0	1	1.046918	-21.640492	110	1	1.10747	0.9;
	5369	2	75.08324	21	0	0	1	1.052776	-25.634557	220	1	1.131615	0.9;
	5370	1	0	0	0	0	1	1.085404	-45.401785	380	1	1.142103	0.9;
	5371	2	26.4293	6.6	0	0	1	1.007818	-33.751404	110	1	1.1	0.9;
	5372	1	74.08214	10.7	0	0	1	1.055739	-49.281017	220	1	1.113359	0.9;
	5373	1	29.03219	18	0	0	1	1.002636	-26.096245	150	1	1.1	0.9;
	5374	1	0	0	0	11.08316	1	1.044962	-22.909957	150	1	1.1	0.9;
	5375	1	0	0	0	0	1	1.038526	-47.392008	220	1	1.166943	0.9;
	5376	1	0	0	0	0	1	1.018029	-32.680373	150	1	1.1	0.9;
	5377	1	0	0	0	0	1	0.999702	-21.934678	150	1	1.1	0.9;
	5378	1	0	0	0	0	1	1.025625	64.659771	220	1	1.1	0.9;
	5379	2	5.205771	1.7	0	0	1	1.039954	-26.359635	220	1	1.1	0.9;
	5380	1	53.05883	17.4	0	0	1	0.968366	-25.979549	110	1	1.1	0.9;
	5381	1	78.48702	18.5	0	0	1	1.044387	-43.573767	220	1	1.1	0.9;
	5382	1	0	0	0	0	1	1.072534	-21.989444	220	1	1.146509	0.9;
	5383	1	0	0	0	0	1	1.042763	-15.701583	150	1	1.1	0.9;
	5384	1	0	0	0	0	1	1.038839	-25.408015	150	1	1.126041	0.9;
	5385	1	0	0	0	0	1	1.093191	-30.278888	380	1	1.133731	0.9;
	5386	1	0	0	0	0	1	1.03518	-20.117592	150	1	1.112253	0.9;
	5387	2	114.4	34.2	0	0	1	1.055636	-13.503388	220	1	1.117502	0.9;
	5388	2	0	0	0	0	1	1.087711	53.903087	380	1	1.119671	0.9;
	5389	1	5.205771	1.7	0	0	1	1.008753	-28.358028	110	1	1.1	0.9;
	5390	1	15.11676	7.5	0	0	1	1.007449	12.777621	220	1	1.1	0.9;
	5391	1	71.8797	21.7	0	0	1	1.073376	-25.148663	220	1	1.164143	0.9;
	5392	1	0	0	0	0	1	1.075188	-45.425084	220	1	1.132429	0.9;
	5393	1	0	0	0	0	1	1.048353	-37.98018	220	1	1.1	0.9;
	5394	1	0	0	0	0	1	1.048518	-15.80688	150	1	1.105699	0.9;
	5395	1	57.06327	13.3	0	0	1	1.06458	-44.072293	220	1	1.113179	0.9;
	5396	1	160.7	40.3	0	0	1	1.055269	2.023613	150	1	1.114782	0.9;
	5397	1	24.92764	22.9	0	0	1	1.005313	-23.226242	150	1	1.1	0.9;
	5398	1	0	0	0	0	1	0.982186	-23.807483	150	1	1.1	0.9;
	5399	1	40.34473	15.9	0	0	1	1.073562	3.158101	150	1	1.143422	0.9;
	5400	1	-7.96759	6.3	0	0	1	1.032635	-26.851724	150	1	1.1	0.9;
	5401	1	106.7	33.6	0	0	1	1.06489	-14.11963	110	1	1.122378	0.9;
	5402	2	80.0888	23.4	0	0	1	1.031818	-10.743343	110	1	1.1	0.9;
	5403	1	0	0	0	0	1	1.017238	-23.748086	150	1	1.1	0.9;
	5404	2	208.5	36.6	0	0	1	1.069805	-15.727219	220	1	1.144946	0.9;
	5405	1	209.8	102.6	0	0	1	1.011129	-20.287747	220	1	1.1	0.9;
	5406	1	160	50	0	0	1	1.043875	-34.3488	220	1	1.10663	0.9;
	5407	1	0	0	0	0	1	1.038689	-29.674826	220	1	1.1	0.9;
	5408	1	113.5	11.5	0	0	1	1.004319	-42.659386	220	1	1.1	0.9;
	5409	1	64.27126	22.9	0	0	1	1.05066	-19.481806	110	1	1.108412	0.9;
	5410	1	0	0	0	0	1	1.094748	45.096924	380	1	1.117233	0.9;
	5411	2	43.24795	15.2	0	0	1	0.986545	-19.43431	110	1	1.1	0.9;
	5412	1	144.5	67.8	0	0	1	0.994571	10.38781	220	1	1.1	0.9;
	5413	1	62.56937	22.7	0	0	1	1.029851	-23.204077	110	1	1.1	0.9;
	5414	1	0	0	0	0	1	1.069033	-18.970343	150	1	1.1	0.9;
	5415	1	0	0	0	0	1	1.074474	-17.808032	150	1	1.1	0.9;
	5416	1	99.5	28.1	0	0	1	1.027232	13.660125	380	1	1.1	0.9;
	5417	2	24.22686	9.3	0	0	1	1.060182	-37.199521	220	1	1.107677	0.9;
	5418	2	0	0	0	0	1	1.100527	4.470687	380	1	1.14025	0.9;
	5419	1	148.2	65.3	0	0	1	1.046885	-14.837735	380	1	1.105819	0.9;
	5420	2	0	0	0	0	1	1.094273	-11.646662	220	1	1.143567	0.9;
	5421	1	0	0	0	0	1	1.067604	9.787942	150	1	1.1	0.9;
	5422	2	0	0	0	0	1	0.966818	-29.040951	110	1	1.1	0.9;
	5423	1	0	0	0	0	1	1.007624	-22.114318	220	1	1.1	0.9;
	5424	2	0	0	0	0	1	0.979763	67.633441	380	1	1.1	0.9;
	5425	1	3.503885	0.9	0	0	1	0.997903	-24.722002	110	1	1.1	0.9;
	5426	1	0	0	0	0	1	1.078993	20.514616	220	1	1.163338	0.9;
	5427	2	0	0	0	0	1	1.077532	-17.610691	150	1	1.1	0.9;
	5428	1	17.11898	1.3	0	0	1	0.987885	-23.526922	110	1	1.1	0.9;
	5429	1	0	0	0	0	1	1.074177	-33.734227	380	1	1.151386	0.9;
	5430	2	157.7	39.7	0	0	1	1.087026	-14.363254	380	1	1.161432	0.9;
	5431	1	27.53052	12.1	0	0	1	1.043169	-30.760662	110	1	1.1	0.9;
	5432	2	37.94207	15.6	0	0	1	1.086553	-18.645224	380	1	1.161522	0.9;
	5433	2	198.6	63	0	0	1	1.089921	33.59766	380	1	1.10704	0.9;
	5434	2	49.85527	8.2	0	0	1	1.081318	-13.350588	220	1	1.148922	0.9;
	5435	2	0	0	0	0	1	1.076947	7.418721	380	1	1.124898	0.9;
	5436	1	83.1	26.1	0	0	1	1.006911	-22.38241	220	1	1.1	0.9;
	5437	1	0	0	0	0	1	1.071912	-9.450336	220	1	1.139262	0.9;
	5438	2	241.4	116.9	0	0	1	1.055318	-45.59038	220	1	1.1	0.9;
	5439	1	408.9	77.3	0	0	1	1.041988	-23.021134	220	1	1.1	0.9;
	5440	1	0	0	0	0	1	1.04683	-11.033199	220	1	1.118491	0.9;
	5441	2	117.5	36.4	0	0	1	1.069183	-13.841014	220	1	1.124349	0.9;
	5442	1	18.52053	6.1	0	0	1	0.970718	-26.651895	110	1	1.1	0.9;
	5443	1	0	0	0.138277	0	1	1.057243	0.508939	380	1	1.121545	0.9;
	5444	1	25.82864	12.5	0	0	1	1.075324	-16.687618	220	1	1.142067	0.9;
	5445	1	36.03996	-8	0	0	1	1.060093	-21.39214	220	1	1.126439	0.9;
	5446	2	0	0	0	0	1	1.086447	15.302727	380	1	1.132887	0.9;
	5447	1	4.605106	1.7	0	0	1	1.020788	-21.380238	110	1	1.1	0.9;
	5448	1	7.00777	2.3	0	0	1	0.953733	-32.247596	110	1	1.1	0.9;
	5449	1	0	0	0	0	1	1.101192	-43.269517	380	1	1.156912	0.9;
	5450	1	0	0	0	0	1	1.100652	-13.514071	380	1	1.14581	0.9;
	5451	1	0	0	0.250997	0	1	1.082411	-17.766259	220	1	1.191503	0.9;
	5452	2	196.2	50.1	0	0	1	1.09119	-8.229532	380	1	1.127949	0.9;
	5453	1	20.82309	7.9	0	0	1	1.045416	-46.866537	220	1	1.1	0.9;
	5454	1	0	0	0	0	1	1.098256	-21.885663	220	1	1.189695	0.9;
	5455	1	11.41265	0.7	0	0	1	1.020976	-14.482205	110	1	1.1	0.9;
	5456	1	-8.80977	2.9	0	0	1	0.99828	-22.259085	150	1	1.1	0.9;
	5457	1	105.8	51.8	0	0	1	1.03686	-29.382527	150	1	1.1	0.9;
	5458	2	0.300333	2	0	0	1	1.017958	-1.215397	220	1	1.1	0.9;
	5459	1	0	0	0	0	1	1.064108	-2.429522	150	1	1.142474	0.9;
	5460	1	92.10212	45	0	0	1	0.994486	-49.605106	220	1	1.1	0.9;
	5461	1	0	0	0	0	1	1.051282	9.664234	380	1	1.111698	0.9;
	5462	1	27.53052	17	0	0	1	0.99822	-23.421572	150	1	1.1	0.9;
	5463	1	0	0	0	0	1	1.064457	-14.461155	380	1	1.116733	0.9;
	5464	1	0.100111	-19.3	0	0	1	1.067234	-18.379332	220	1	1.1	0.9;
	5465	1	0	0	0	0	1	1.021539	-34.037449	220	1	1.1	0.9;
	5466	1	9.00999	3	0	0	1	1.01016	-22.01165	110	1	1.1	0.9;
	5467	1	0	0	0	0	1	1.026575	-23.564669	150	1	1.1	0.9;
	5468	2	0	0	0	0	1	1.066762	10.142309	110	1	1.145854	0.9;
	5469	1	0	0	0	0	1	1.019757	-23.38694	150	1	1.1	0.9;
	5470	1	61.26793	11	0	0	1	1.060132	19.875737	380	1	1.120805	0.9;
	5471	1	0	0	0	66.25125	1	1.045734	4.676754	110	1	1.109245	0.9;
	5472	1	23.02553	7.6	0	0	1	0.972038	-34.824014	110	1	1.1	0.9;
	5473	1	0	0	0	0	1	1.053332	-14.826721	220	1	1.107581	0.9;
	5474	1	0	0	0	0	1	1.044125	-23.321334	150	1	1.1	0.9;
	5475	2	23.02553	9.2	0	0	1	1.018091	-20.223937	110	1	1.1	0.9;
	5476	1	0	0	0	0	1	0.994399	-21.760484	110	1	1.1	0.9;
	5477	1	-18.5205	6.1	0	0	1	1.003955	-21.150202	150	1	1.1	0.9;
	5478	2	0	0	0	0	1	1.052289	-0.671313	380	1	1.137191	0.9;
	5479	1	21.62398	9.8	0	0	1	0.97111	-22.792228	110	1	1.1	0.9;
	5480	1	43.54828	-20.5	0	0	1	1.050117	-41.505473	220	1	1.112016	0.9;
	5481	1	165.9	154.6	0	0	1	1.080407	-46.273385	380	1	1.144552	0.9;
	5482	1	-57.4637	-30.3	0	0	1	1.061088	-40.1389	220	1	1.119049	0.9;
	5483	1	0	0	0	0	1	1.038567	-47.392266	220	1	1.166989	0.9;
	5484	1	0	0	0	0	1	0.967167	-28.002039	150	1	1.1	0.9;
	5485	1	49.75517	15	0	0	1	0.916053	-29.563462	110	1	1.1	0.9;
	5486	2	0	0	0	0	1	1.078087	0.269769	150	1	1.147562	0.9;
	5487	2	60.56715	39	0.090998	0	1	1.049727	-8.944492	220	1	1.146198	0.9;
	5488	1	79.48813	22.7	0	0	1	1.063018	7.20133	150	1	1.112274	0.9;
	5489	1	6.607326	1.9	0	0	1	0.974726	-25.973328	110	1	1.1	0.9;
	5490	1	0	0	0	0	1	1.012192	-7.348013	220	1	1.1	0.9;
	5491	1	5.606215	2	0	0	1	0.955902	-14.775444	110	1	1.1	0.9;
	5492	1	0	0	0	0	1	1.082684	16.349883	150	1	1.127585	0.9;
	5493	1	412.1	79.3	0	0	1	1.078937	-25.786209	220	1	1.171157	0.9;
	5494	1	4.104551	1.4	0	0	1	0.995906	-28.182808	110	1	1.1	0.9;
	5495	1	19.32142	7	0	0	1	1.059284	-2.475132	110	1	1.130258	0.9;
	5496	1	0	0	0	0	1	1.047004	-21.590936	150	1	1.10688	0.9;
	5497	1	47.65283	19.4	0	0	1	1.074657	-0.444655	150	1	1.142288	0.9;
	5498	1	0	0	0	0	1	1.048974	5.661309	110	1	1.116813	0.9;
	5499	1	35.13896	14.3	0	0	1	1.014238	-34.649929	110	1	1.1	0.9;
	5500	1	0	0	0	0	1	1.09753	-34.313294	380	1	1.160046	0.9;
	5501	1	15.01665	9	0	0	1	0.963866	-25.258412	150	1	1.1	0.9;
	5502	1	13.31476	2.8	0	0	1	0.997954	-22.492518	110	1	1.1	0.9;
	5503	2	61.8686	24.3	0	0	1	1.000917	-23.444175	110	1	1.1	0.9;
	5504	1	0	0	0	0	1	1.103538	-13.743719	380	1	1.18471	0.9;
	5505	1	25.72853	9.3	0	0	1	1.004978	-17.256884	110	1	1.1	0.9;
	5506	1	0	0	0	0	1	1.053778	-11.696454	380	1	1.121649	0.9;
	5507	1	0	0	0	0	1	1.051479	-33.926783	220	1	1.113903	0.9;
	5508	1	0	0	0	0	1	1.062399	11.210977	380	1	1.114981	0.9;
	5509	1	0	0	0.236732	0	1	1.021429	-6.494634	380	1	1.1	0.9;
	5510	2	0	0	0	0	1	1.058392	-21.711891	220	1	1.122284	0.9;
	5511	1	163.5	35.6	0	0	1	1.046379	-15.449733	220	1	1.1	0.9;
	5512	1	47.6	41.3	0	0	1	1.091453	-14.44708	380	1	1.16284	0.9;
	5513	1	2.702997	0.9	0	0	1	0.962661	-31.45214	110	1	1.1	0.9;
	5514	1	0	0	0	0	1	1.083002	-12.82591	220	1	1.136293	0.9;
	5515	1	1.801998	0.7	0	0	1	1.028706	-15.341709	110	1	1.131885	0.9;
	5516	1	102.7	-102.3	0	0	1	1.080322	-9.759686	380	1	1.111568	0.9;
	5517	1	0	0	0	0	1	0.992271	-20.268219	150	1	1.1	0.9;
	5518	1	160.2	89.9	0	0	1	1.042575	-15.702618	150	1	1.1	0.9;
	5519	1	0	0	0.396003	0	1	1.086715	-38.336899	380	1	1.142066	0.9;
	5520	1	5.806438	1.4	0	0	1	0.999285	-22.541653	110	1	1.1	0.9;
	5521	1	6.00666	1.5	0	0	1	0.982483	-24.025538	110	1	1.1	0.9;
	5522	1	111.5	147.5	0	0	1	1.03114	12.478573	380	1	1.1	0.9;
	5523	1	0	0	0.074818	0	1	0.992487	66.347787	380	1	1.1	0.9;
	5524	2	353	123.4	0	0	1	1.066053	5.732402	380	1	1.113146	0.9;
	5525	1	26.22908	8.6	0	0	1	1.0049	-29.725638	110	1	1.1	0.9;
	5526	2	0	5.9	0	0	1	1.0825	-34.619493	220	1	1.129776	0.9;
	5527	1	0	0	0	0	1	1.044339	-8.888837	220	1	1.115925	0.9;
	5528	2	1.801998	72	0	0	1	1.094658	-35.860203	380	1	1.131543	0.9;
	5529	1	21.02331	12.4	0	0	1	1.035956	-36.414903	220	1	1.1	0.9;
	5530	1	0	0	0	0	1	1.014582	-22.626233	150	1	1.1	0.9;
	5531	1	0	0	0	0	1	1.08464	53.191273	380	1	1.119683	0.9;
	5532	1	0	0	0	0	1	1.022842	-19.179077	150	1	1.1	0.9;
	5533	1	0	0	0	0	1	1.03082	17.054492	380	1	1.1	0.9;
	5534	1	0	0	0.152776	0	1	1.080055	-18.867896	380	1	1.151308	0.9;
	5535	1	0.300333	2	0	0	1	1.006741	-0.784501	220	1	1.1	0.9;
	5536	2	521.7	154.9	0	0	1	1.044636	2.192575	220	1	1.122509	0.9;
	5537	1	103	40.2	0	0	1	1.068594	-10.09756	220	1	1.139231	0.9;
	5538	1	0	0	0	0	1	0.993738	-23.426386	150	1	1.1	0.9;
	5539	2	0	0	0	0	1	1.040363	-19.037852	220	1	1.1	0.9;
	5540	2	107.4	10.5	0	0	1	1.056517	-11.734603	220	1	1.135949	0.9;
	5541	2	0	0	0	0	1	0.978091	-23.36938	110	1	1.1	0.9;
	5542	1	1.301443	3.8	0	0	1	0.951334	-26.861221	110	1	1.1	0.9;
	5543	1	128.6	27.8	0	0	1	1.043336	-34.217279	220	1	1.106061	0.9;
	5544	1	0	0	0	0	1	1.093855	35.179654	380	1	1.130345	0.9;
	5545	2	291.7	104.6	0	0	1	1.020195	-8.914341	380	1	1.106673	0.9;
	5546	1	78.58714	16	0	0	1	1.054953	-35.309127	220	1	1.160343	0.9;
	5547	2	56.66282	18.7	0	0	1	1.070914	-8.629408	220	1	1.137526	0.9;
	5548	1	0	0	0	0	1	1.030193	-24.334844	150	1	1.1	0.9;
	5549	1	0	0	0	0	1	1.050232	-10.468122	220	1	1.120425	0.9;
	5550	1	0	0	0	0	1	1.040677	-12.912468	220	1	1.120855	0.9;
	5551	1	0	0	0.133294	0	1	1.03169	-13.531282	220	1	1.107681	0.9;
	5552	1	0	0	0	0	1	1.032493	-23.307444	150	1	1.1	0.9;
	5553	2	82.69168	32.9	0	0	1	1.07896	20.273749	220	1	1.157267	0.9;
	5554	2	251.2	82.6	0	0	1	1.063956	-14.98634	220	1	1.128453	0.9;
	5555	2	0	0	0	0	1	1.047909	21.260442	220	1	1.105436	0.9;
	5556	1	16.12131	5.8	0	0	1	1.069534	-15.058313	150	1	1.127814	0.9;
	5557	1	3.704107	1.4	0	0	1	1.018329	-15.922496	110	1	1.1	0.9;
	5558	1	0	0	0	0	1	1.052033	-20.563178	150	1	1.1	0.9;
	5559	1	-0.10011	68.5	0	0	1	1.067639	-42.923665	220	1	1.131449	0.9;
	5560	1	27.23019	8.9	0	0	1	1.005081	-32.316496	110	1	1.1	0.9;
	5561	2	0	0	0	0	1	1.088763	-13.874493	380	1	1.181195	0.9;
	5562	2	0	0	0	0	1	1.032727	39.12418	220	1	1.1	0.9;
	5563	2	0	0	0	0	1	1.099466	17.411888	150	1	1.163763	0.9;
	5564	1	60.66726	5.9	0	0	1	1.046803	-45.218904	220	1	1.1	0.9;
	5565	1	109	28.3	0	0	1	1.043633	-20.867765	110	1	1.1	0.9;
	5566	1	42.54717	6.5	0	0	1	1.053359	-48.363389	220	1	1.111341	0.9;
	5567	1	2.202442	0.5	0	0	1	1.007418	-51.035744	150	1	1.1	0.9;
	5568	1	0	0	0	0	1	1.08663	45.263414	380	1	1.120788	0.9;
	5569	2	162	4.8	0	0	1	1.048091	-40.131078	220	1	1.1	0.9;
	5570	1	76.88525	35.1	0	0	1	1.088952	-0.217408	150	1	1.164764	0.9;
	5571	1	11.51276	3.8	0	0	1	0.943051	-33.563738	110	1	1.1	0.9;
	5572	1	44.04884	16	0	0	1	0.977431	-23.917326	110	1	1.1	0.9;
	5573	2	0	0	0	0	1	0.985818	-24.913546	110	1	1.1	0.9;
	5574	2	15.81754	22	0	0	1	1.038773	-29.114269	220	1	1.1	0.9;
	5575	1	0	0	0	0	1	1.092606	-11.682996	220	1	1.148868	0.9;
	5576	1	0	0	0	0	1	1.013775	-22.290041	150	1	1.1	0.9;
	5577	1	8.709657	5.4	0	0	1	0.973131	-27.816607	150	1	1.1	0.9;
	5578	1	39.04329	24.8	0	0	1	1.042936	14.663812	150	1	1.1	0.9;
	5579	1	15.31698	5	0	0	1	1.001695	-19.038118	110	1	1.1	0.9;
	5580	1	60.66726	33.2	0	0	1	1.064798	-22.746035	220	1	1.149844	0.9;
	5581	1	0	0	0	0	1	0.94979	55.935489	220	1	1.1	0.9;
	5582	2	0.7	0	0	0	1	1.069816	-17.233074	380	1	1.129975	0.9;
	5583	1	0	0	0	0	1	1.03834	-25.051439	220	1	1.1	0.9;
	5584	1	67.97536	54.6	0	0	1	1.014786	29.780394	220	1	1.1	0.9;
	5585	1	95.9	26.2	0	0	1	1.053863	-36.176784	220	1	1.136281	0.9;
	5586	1	160.2	88.7	0	0	1	1.039748	-15.692148	150	1	1.1	0.9;
	5587	1	0	0	0.250997	0	1	1.036274	-11.739288	220	1	1.13312	0.9;
	5588	1	0	0	0.143075	0	1	1.030426	-21.783103	220	1	1.1	0.9;
	5589	1	0	0	0	0	1	1.037493	9.020762	380	1	1.112655	0.9;
	5590	1	383.6	-37.9	0	0	1	1.055469	-43.54985	380	1	1.123982	0.9;
	5591	1	0	0	0	23.04696	1	1.020394	-23.553821	150	1	1.1	0.9;
	5592	1	0	0	0	0	1	1.074048	-20.851833	220	1	1.153634	0.9;
	5593	1	0	0	0.219948	0	1	1.042454	13.487558	380	1	1.107653	0.9;
	5594	1	0	0	0	0	1	1.000341	-23.805127	150	1	1.1	0.9;
	5595	2	89.09879	8.3	0	0	1	1.074803	-5.681992	220	1	1.132159	0.9;
	5596	1	152.6	61.8	0	0	1	1.05978	9.032617	150	1	1.1	0.9;
	5597	2	0	0	0	0	1	1.087545	-12.147224	220	1	1.140639	0.9;
	5598	2	124.8	-14.7	0	0	1	1.028591	9.773039	220	1	1.120562	0.9;
	5599	1	28.93208	0.2	0	0	1	0.965996	21.692667	220	1	1.1	0.9;
	5600	1	66.97426	14.1	0	0	1	1.003978	-21.347285	220	1	1.1	0.9;
	5601	2	0	0	0	0	1	0.984977	38.283168	220	1	1.1	0.9;
	5602	1	289.1	84.2	0	0	1	1.030999	-43.443289	220	1	1.1	0.9;
	5603	1	16.71854	5.5	0	0	1	0.961309	-32.071063	110	1	1.1	0.9;
	5604	1	18.42042	5.3	0	0	1	0.99346	-22.872313	110	1	1.1	0.9;
	5605	1	91.0009	28.4	0	0	1	1.072196	-23.386527	220	1	1.137426	0.9;
	5606	2	4.304773	1.5	0	0	1	1.047364	-32.592401	220	1	1.12292	0.9;
	5607	1	0	0	0	0	1	1.071873	13.345648	380	1	1.13019	0.9;
	5608	2	0	0	0	0	1	1.038909	-25.035541	220	1	1.1	0.9;
	5609	1	59.36582	12.7	0	0	1	1.054943	-48.805042	220	1	1.111645	0.9;
	5610	1	16.71854	5.5	0	0	1	0.937076	-35.110129	110	1	1.1	0.9;
	5611	1	0	0	0	0	1	1.00277	-21.521905	150	1	1.1	0.9;
	5612	1	63.06993	13.4	0	0	1	1.053832	-48.993828	220	1	1.110912	0.9;
	5613	1	0	0	0.396	0	1	1.082558	-37.272077	380	1	1.138792	0.9;
	5614	1	209.9	38	0	0	1	1.017807	-46.385761	150	1	1.1	0.9;
	5615	1	131.9	50.6	0	0	1	1.069196	-39.361633	380	1	1.134351	0.9;
	5616	2	206.2	72.2	0	0	1	1.086483	-11.013389	380	1	1.151191	0.9;
	5617	2	0	0	0	0	1	1.057064	-36.152006	220	1	1.137791	0.9;
	5618	1	50.55605	5.9	0	0	1	1.01135	-39.082784	220	1	1.1	0.9;
	5619	1	165.4	130.5	0	0	1	1.048696	-10.481131	380	1	1.117457	0.9;
	5620	1	0	0	0	0	1	1.082866	-15.357375	220	1	1.145744	0.9;
	5621	1	0	0	0	0	1	1.049772	-28.989033	150	1	1.119947	0.9;
	5622	1	17.51942	6.1	0	0	1	1.055003	-13.239009	110	1	1.122601	0.9;
	5623	1	126.7	46.1	0	0	1	1.033368	-13.059743	220	1	1.1	0.9;
	5624	2	19.62176	8.7	0	0	1	1.070872	-11.906164	380	1	1.168012	0.9;
	5625	1	0	0	0	0	1	0.987616	-17.278852	150	1	1.1	0.9;
	5626	2	0	0	0	0	1	1.0855	-28.63908	380	1	1.143321	0.9;
	5627	1	83.7929	33.1	0	0	1	1.080731	2.939529	150	1	1.151041	0.9;
	5628	1	14.01554	5.6	0	0	1	1.077066	-26.516567	220	1	1.185416	0.9;
	5629	2	0	0	0	0	1	1.068591	-19.041391	220	1	1.1	0.9;
	5630	1	0	0	0	0	1	0.994821	-18.120521	110	1	1.1	0.9;
	5631	2	0	0	0	0	1	1.072851	0.067024	380	1	1.137035	0.9;
	5632	2	0	0	0	0	1	1.099211	-15.427181	380	1	1.187603	0.9;
	5633	1	18.72076	7.9	0	0	1	1.002423	-23.254579	110	1	1.1	0.9;
	5634	1	69.47704	19.7	0	0	1	1.064996	-43.94067	220	1	1.11392	0.9;
	5635	1	253.2	78.1	0	0	1	1.017567	-24.058204	220	1	1.1	0.9;
	5636	1	117	19.2	0	0	1	1.019099	-4.698393	220	1	1.1	0.9;
	5637	1	0	0	0	0	1	1.002461	-20.008582	150	1	1.1	0.9;
	5638	1	7.608436	2.5	0	0	1	0.994749	-30.777664	110	1	1.2	0.8;
	5639	1	0	0	0	0	1	1.031712	-24.876266	150	1	1.1	0.9;
	5640	1	13.71521	4.5	0	0	1	1.009069	-32.228006	110	1	1.1	0.9;
	5641	1	0	0	0	0	1	1.082394	-42.003919	380	1	1.145863	0.9;
	5642	1	88.79845	35.1	0	0	1	1.067651	3.742862	150	1	1.137212	0.9;
	5643	1	0	0	0	0	1	1.042533	-12.744628	220	1	1.114652	0.9;
	5644	1	231.9	27.6	0	0	1	1.068592	-22.024249	220	1	1.161249	0.9;
	5645	1	0	0	0	0	1	1.006207	60.405804	220	1	1.1	0.9;
	5646	2	0	0	0.108316	45.08263	1	1.063403	54.132585	380	1	1.1	0.9;
	5647	1	0	0	0	0	1	1.003861	-25.992135	150	1	1.1	0.9;
	5648	1	0	0	0	0	1	1.068544	-12.436724	380	1	1.127161	0.9;
	5649	1	0	0	0	0	1	1.064121	-17.025128	220	1	1.1	0.9;
	5650	1	0	0	0	0	1	0.972134	-24.848285	150	1	1.1	0.9;
	5651	1	88	70.3	0	0	1	1.052631	19.936817	380	1	1.112675	0.9;
	5652	1	26.4293	5.5	0	0	1	1.004128	-38.547321	110	1	1.1	0.9;
	5653	1	0	0	0	0	1	0.990131	-25.365867	150	1	1.1	0.9;
	5654	1	0	0	0	0	1	1.077753	-7.475284	220	1	1.143393	0.9;
	5655	1	49.25461	12	0	0	1	1.05351	-24.629846	220	1	1.128405	0.9;
	5656	2	0	0	0	0	1	1.086474	1.563116	380	1	1.135464	0.9;
	5657	1	227.6	71	0	0	1	1.037424	-16.885309	220	1	1.1	0.9;
	5658	1	16.51831	0	0	0	1	1.036973	-21.791816	150	1	1.113615	0.9;
	5659	1	2.502775	0	0	0	1	1.038586	-10.425594	110	1	1.1	0.9;
	5660	1	0	0	0	0	1	1.075529	1.305281	380	1	1.14614	0.9;
	5661	2	0	0	0	0	1	1.071363	-40.279815	220	1	1.124849	0.9;
	5662	1	0	0	0	0	1	1.075591	-33.570278	220	1	1.17927	0.9;
	5663	1	0	0	0	0	1	1.032712	-25.347593	150	1	1.1	0.9;
	5664	2	0	0	0	0	1	1.015364	-18.606987	110	1	1.1	0.9;
	5665	2	90.7	36.9	0	0	1	1.053974	-40.153574	220	1	1.119249	0.9;
	5666	1	0	0	0	0	1	1.031465	-44.836838	220	1	1.1	0.9;
	5667	1	0	0	0.296958	0	1	1.104118	-9.182671	380	1	1.15552	0.9;
	5668	1	0	0	0	0	1	1.05959	-34.322172	380	1	1.136658	0.9;
	5669	1	0	0	0	0	1	0.997042	-23.625232	110	1	1.1	0.9;
	5670	2	0	0	0	0	1	1.082017	-19.140171	380	1	1.158207	0.9;
	5671	2	0	0	0	0	1	1.054079	-39.690856	380	1	1.126627	0.9;
	5672	1	0	0	0	0	1	0.972387	-20.796309	110	1	1.1	0.9;
	5673	2	391.3	118.9	0	0	1	1.048972	-2.225213	380	1	1.114382	0.9;
	5674	1	164.7	44.5	0	0	1	1.011903	-32.914373	220	1	1.1	0.9;
	5675	1	128.8	29.3	0	0	1	1.053746	-24.589696	220	1	1.128807	0.9;
	5676	1	0	0	0	123.9802	1	1.032107	-49.880605	220	1	1.1	0.9;
	5677	1	0	0	0	-16.4625	1	1.010392	-27.447382	150	1	1.194529	0.9;
	5678	1	0	0	0.177284	0	1	1.083407	-17.404382	380	1	1.156914	0.9;
	5679	1	0	0	0	0	1	1.106467	-16.572509	380	1	1.162776	0.9;
	5680	1	11.61288	2.5	0	0	1	0.988362	-17.78243	110	1	1.1	0.9;
	5681	1	178	33.8	0	0	1	1.048214	-23.326949	220	1	1.1	0.9;
	5682	1	47.55272	29.4	0	0	1	1.033686	-21.485099	150	1	1.1	0.9;
	5683	1	75.88414	11.2	0	0	1	1.051541	-49.725979	220	1	1.112484	0.9;
	5684	1	12.81421	4.7	0	0	1	0.98922	-24.114922	110	1	1.1	0.9;
	5685	1	0	0	0	0	1	1.062698	11.250878	380	1	1.115247	0.9;
	5686	1	0	0	0	0	1	0.990982	-24.260269	110	1	1.1	0.9;
	5687	2	61.36804	13.9	0	0	1	0.99851	-22.44434	110	1	1.1	0.9;
	5688	1	91.30122	18.1	0	0	1	1.073039	-23.189883	220	1	1.138135	0.9;
	5689	2	266.1	33.7	0	0	1	1.079629	-9.374264	220	1	1.128256	0.9;
	5690	1	3.704107	2.6	0	0	1	1.076862	-19.321458	220	1	1.175393	0.9;
	5691	1	125.9	-21.2	0	0	1	1.077167	39.182513	380	1	1.118067	0.9;
	5692	1	12.11343	4.4	0	0	1	1.01824	-37.214561	110	1	1.1	0.9;
	5693	1	441.2	200.7	0	20	1	1.05092	-12.781007	380	1	1.119324	0.9;
	5694	1	122.8	37.2	0	0	1	1.066788	-3.235896	150	1	1.137132	0.9;
	5695	1	65.07214	40.3	0	0	1	1.010808	-27.34916	150	1	1.1	0.9;
	5696	1	15.21687	4.9	0	0	1	0.975791	-25.884266	110	1	1.1	0.9;
	5697	2	77.48591	24.1	0	0	1	0.979455	-22.124893	110	1	1.1	0.9;
	5698	2	0	0	0	0	1	1.037955	-29.573567	220	1	1.1	0.9;
	5699	2	0	0	0	0	1	1.07943	20.61835	220	1	1.164345	0.9;
	5700	1	0	0	0	0	1	1.073943	13.768887	380	1	1.122585	0.9;
	5701	2	0	0	0	0	1	1.094237	55.26009	380	1	1.118504	0.9;
	5702	1	51.95761	22.8	0	0	1	0.966357	-23.228813	110	1	1.1	0.9;
	5703	1	0	0	0	0	1	1.047021	39.861532	220	1	1.1	0.9;
	5704	1	0	0	0	0	1	1.0535	0.387504	220	1	1.122478	0.9;
	5705	1	25.22797	10.3	0	0	1	1.072926	-2.196351	150	1	1.142556	0.9;
	5706	1	47.65283	20.4	0	0	1	1.055498	-12.314331	220	1	1.140326	0.9;
	5707	2	0	0	0	0	1	1.042182	-23.269032	220	1	1.133992	0.9;
	5708	1	186.5	49.9	0	0	1	1.057962	-44.353688	220	1	1.121755	0.9;
	5709	1	0	0	0	0	1	1.024798	-4.720058	380	1	1.1	0.9;
	5710	1	0	0	0	0	1	1.090326	6.998132	150	1	1.156602	0.9;
	5711	1	15.01665	6	0	0	1	1.053186	-22.11449	220	1	1.1	0.9;
	5712	1	20.0222	8	0	0	1	1.051153	-22.694482	220	1	1.1	0.9;
	5713	1	0	0	0.312993	0	1	1.084675	-43.832909	380	1	1.1414	0.9;
	5714	1	0	0	0	0	1	1.057836	-40.76727	220	1	1.11958	0.9;
	5715	2	0	0	0	0	1	1.04657	-22.327199	150	1	1.1	0.9;
	5716	1	134.0486	13.3	0	0	1	1.056632	-27.090292	220	1	1.119074	0.9;
	5717	1	49.95539	15.6	0	0	1	1.012924	-47.760563	220	1	1.1	0.9;
	5718	1	1.701887	0.6	0	0	1	1.0371	-20.040738	110	1	1.1	0.9;
	5719	1	31.33474	6.4	0	0	1	0.905869	-32.770463	110	1	1.1	0.9;
	5720	1	0	0	0	0	1	1.07083	-23.5476	220	1	1.173055	0.9;
	5721	1	132.4	-10.4	0	0	1	1.049286	-15.665896	220	1	1.113462	0.9;
	5722	1	0	0	0	0	1	1.055202	-24.38554	220	1	1.129932	0.9;
	5723	1	0	-7.5	0	0	1	1.032846	-25.473853	150	1	1.1	0.9;
	5724	1	0	0	0	0	1	0.97153	-26.060266	150	1	1.1	0.9;
	5725	1	0	0	0	0	1	0.987704	-38.981186	220	1	1.1	0.9;
	5726	1	0	0	0	0	1	1.065884	-12.098673	220	1	1.124847	0.9;
	5727	1	48.95428	14.9	0	0	1	1.06613	-22.654098	220	1	1.151082	0.9;
	5728	1	30.0333	15	0	0	1	1.039174	-19.849128	110	1	1.1	0.9;
	5729	1	95.70611	19.6	0	0	1	1.012443	-38.860558	220	1	1.1	0.9;
	5730	1	474.3	203.4	0	0	1	1.082936	-40.335175	380	1	1.151816	0.9;
	5731	1	0	0	0.095843	0	1	1.041674	-18.455056	220	1	1.116917	0.9;
	5732	1	0	0	0	0	1	0.961438	-25.178362	150	1	1.1	0.9;
	5733	1	0	0	0.148479	0	1	1.067591	-13.09912	380	1	1.128611	0.9;
	5734	2	0	0	0	0	1	1.068652	2.4194	150	1	1.14018	0.9;
	5735	1	128.8	70	0	0	1	1.001218	11.782514	220	1	1.1	0.9;
	5736	1	0	0	0	0	1	0.975789	-26.967136	150	1	1.1	0.9;
	5737	1	116.1	30.2	0	0	1	1.052033	1.962158	150	1	1.111082	0.9;
	5738	1	-5.00555	0	0	0	1	1.020036	-46.648939	220	1	1.1	0.9;
	5739	1	0	0	0	0	1	0.982366	-23.700211	150	1	1.1	0.9;
	5740	2	0	0	0	0	1	1.027532	-40.104457	150	1	1.1	0.9;
	5741	1	0	0	0	0	1	1.035274	-24.810791	150	1	1.1	0.9;
	5742	2	0	0	0	0	1	1.089773	-0.773888	220	1	1.154073	0.9;
	5743	2	0	0	0	0	1	1.096	-11.219086	220	1	1.170597	0.9;
	5744	1	42.74739	4	0	0	1	1.019418	-36.710513	220	1	1.1	0.9;
	5745	2	0	0	0	0	1	1.014727	-47.173148	220	1	1.1	0.9;
	5746	2	0	0	0	0	1	1.043413	-22.959619	220	1	1.1	0.9;
	5747	1	245.5095	46.8	0	0	1	1.011752	10.817904	220	1	1.1	0.9;
	5748	1	0	0	0	0	1	1.086362	-8.942949	380	1	1.178413	0.9;
	5749	2	0	0	0	0	1	0.998	-23.282478	110	1	1.1	0.9;
	5750	1	7.207992	2.4	0	0	1	0.954658	-26.736884	110	1	1.1	0.9;
	5751	1	0	0	0	0	1	1.086411	-41.19112	380	1	1.146205	0.9;
	5752	1	22.7252	5.3	0	0	1	0.994519	-22.83765	110	1	1.1	0.9;
	5753	1	0	0	0	0	1	1.072667	12.661962	110	1	1.149656	0.9;
	5754	1	143.9	123.3	0.2259	0	1	1.023621	5.281931	380	1	1.1	0.9;
	5755	1	0	0	0	0	1	1.021466	-39.914182	220	1	1.1	0.9;
	5756	1	0	0	0	0	1	1.029145	7.525305	380	1	1.1	0.9;
	5757	1	0	0	0	0	1	0.945982	-33.207225	110	1	1.1	0.9;
	5758	1	22.22464	13.8	0	0	1	1.008251	-22.247869	150	1	1.1	0.9;
	5759	1	32.03552	15	0	0	1	0.980371	-25.247316	110	1	1.1	0.9;
	5760	1	55.46149	14.1	0	0	1	1.062065	-43.925498	220	1	1.111963	0.9;
	5761	1	35.03885	20	0	0	1	0.981723	-23.44267	220	1	1.1	0.9;
	5762	2	21.72409	7.5	0	0	1	1.063409	-15.60069	220	1	1.146948	0.9;
	5763	1	0	0	0	0	1	0.948832	-29.405654	110	1	1.1	0.9;
	5764	1	0.600666	0	0	0	1	0.979687	-19.948005	110	1	1.1	0.9;
	5765	1	0	0	0	0	1	1.075797	4.793244	150	1	1.146054	0.9;
	5766	1	0	0	0	0	1	1.053729	-20.965708	150	1	1.11431	0.9;
	5767	1	44.94984	12.8	0	0	1	1.057041	5.957605	150	1	1.110051	0.9;
	5768	1	430.8	107.9	0	0	1	1.003632	-23.378834	220	1	1.1	0.9;
	5769	2	9.00999	6.4	0	0	1	0.954182	-29.44919	110	1	1.1	0.9;
	5770	1	74.6828	10.8	0	0	1	1.055053	-49.210492	220	1	1.113033	0.9;
	5771	1	0	0	0	0	1	1.072528	-14.468789	150	1	1.129475	0.9;
	5772	1	8.108991	1.4	0	0	1	0.971546	-27.955302	110	1	1.1	0.9;
	5773	1	24.02664	8	0	0	1	1.098766	-36.088649	120	1	1.177284	0.9;
	5774	2	13.41487	4.9	0	0	1	0.977652	-12.003639	110	1	1.1	0.9;
	5775	1	0	0	0	0	1	1.018279	-22.190256	150	1	1.1	0.9;
	5776	1	0	0	0	0	1	1.037476	-24.936872	220	1	1.1	0.9;
	5777	2	0	0	0	0	1	1.098447	8.810252	380	1	1.12969	0.9;
	5778	1	7.408214	2.7	0	0	1	0.996383	-10.998077	110	1	1.1	0.9;
	5779	2	259.3	91.5	0	0	1	1.008409	-7.954941	220	1	1.109562	0.9;
	5780	1	0	0	0	0	1	1.065226	19.880823	380	1	1.128344	0.9;
	5781	1	0	0	0	0	1	1.075198	-0.680671	220	1	1.144234	0.9;
	5782	1	0	0	0	0	1	1.037429	-29.288299	150	1	1.1	0.9;
	5783	2	325	75.8	0	0	1	1.011955	-2.014009	220	1	1.1	0.9;
	5784	2	0	0	0	0	1	1.079342	59.368644	380	1	1.1	0.9;
	5785	2	0	0	0	0	1	1.065535	-23.175093	220	1	1.123366	0.9;
	5786	2	251	100.5	0	0	1	1.072273	-0.161958	220	1	1.1441	0.9;
	5787	1	11.61288	7.2	0	0	1	1.019307	-29.651567	150	1	1.1	0.9;
	5788	1	11.31254	3.7	0	0	1	0.986684	-32.330576	110	1	1.1	0.9;
	5789	1	0	0	0	0	1	1.027176	-20.591632	150	1	1.106195	0.9;
	5790	1	246.2	119.1	0	0	1	1.029253	-51.424782	220	1	1.1	0.9;
	5791	2	97.60822	18.2	0	0	1	1.075959	-9.450845	220	1	1.146578	0.9;
	5792	1	15.61732	5.3	0	0	1	1.042477	17.342064	110	1	1.111997	0.9;
	5793	1	159.9	47.4	0	0	1	1.077974	-1.499889	220	1	1.144364	0.9;
	5794	2	6.00666	3.7	0	0	1	0.99733	-20.487354	150	1	1.1	0.9;
	5795	1	0	0	0	0	1	1.06272	-30.53261	220	1	1.111369	0.9;
	5796	1	0	0	0	0	1	1.027889	-24.465849	150	1	1.1	0.9;
	5797	1	0	0	0	0	1	1.101218	-43.269742	380	1	1.156939	0.9;
	5798	2	93.00311	15.6	0	0	1	1.064842	62.849891	150	1	1.1	0.9;
	5799	1	80.0888	19	0	0	1	0.970943	-20.197249	110	1	1.1	0.9;
	5800	1	93.70389	29.9	0	0	1	1.034495	3.851565	110	1	1.1	0.9;
	5801	1	53.85971	-15.3	0	0	1	1.062031	-41.724824	220	1	1.109652	0.9;
	5802	1	154.7	48.2	0	0	1	1.07822	6.970501	150	1	1.13301	0.9;
	5803	1	4.00444	0.1	0	0	1	0.919573	-29.126124	150	1	1.1	0.868697;
	5804	1	0	0	0	0	1	1.030601	-38.034428	220	1	1.1	0.9;
	5805	1	0	0	0	0	1	1.049716	-11.431567	750	1	1.108733	0.9;
	5806	1	44.04884	18	0	0	1	1.078463	-0.231901	150	1	1.146984	0.9;
	5807	1	7.908769	2.6	0	0	1	0.991108	-28.621634	110	1	1.1	0.9;
	5808	1	0	0	0	0	1	0.964719	-30.558951	150	1	1.1	0.9;
	5809	1	76.4848	19.8	0	0	1	1.060272	-15.022399	220	1	1.13125	0.9;
	5810	1	2.502775	1.3	0	0	1	0.992838	-18.310234	110	1	1.1	0.9;
	5811	1	0.300333	-34.1	0	0	1	1.067196	-0.143455	220	1	1.139511	0.9;
	5812	1	126.9	80.5	0	0	1	1.102372	-17.243849	380	1	1.173294	0.9;
	5813	2	15.31698	5.5	0	0	1	0.938651	-25.707202	110	1	1.1	0.9;
	5814	1	115.2	38.9	0	0	1	1.048206	-21.666726	220	1	1.1	0.9;
	5815	1	0	0	0	0	1	0.958421	-29.660795	150	1	1.1	0.884308;
	5816	2	0	0	0.265355	0	1	1.066684	1.65306	380	1	1.119723	0.9;
	5817	2	-119.633	-2.7	0	0	1	0.996409	-34.255155	220	1	1.1	0.9;
	5818	1	67.67503	16.4	0	0	1	1.047739	-9.444683	220	1	1.12001	0.9;
	5819	1	0	0	0	11.34805	1	1.032668	-26.85174	150	1	1.1	0.9;
	5820	1	43.7485	2.2	0	0	1	1.026608	9.79188	220	1	1.1	0.9;
	5821	2	0	0	0	20	1	1.068842	17.040124	380	1	1.122862	0.9;
	5822	1	0	0	0	0	1	0.958791	-29.071349	150	1	1.1	0.894911;
	5823	1	174.3	11.8	0	0	1	1.076082	-2.019354	220	1	1.14314	0.9;
	5824	2	17.71965	5.1	0	0	1	0.952368	-25.971974	110	1	1.1	0.9;
	5825	1	0	0	0	0	1	1.06721	-3.055136	150	1	1.137439	0.9;
	5826	2	0	0	0	0	1	1.083153	16.348945	380	1	1.128028	0.9;
	5827	1	0	0	0	0	1	1.033318	-25.434967	150	1	1.1	0.9;
	5828	1	10.21132	3.8	0	0	1	1.086915	1.477847	220	1	1.157082	0.9;
	5829	1	83.19224	22.9	0	0	1	1.038958	-12.623077	380	1	1.120516	0.9;
	5830	1	93.1	43.7	0	0	1	1.032295	-25.224724	150	1	1.1	0.9;
	5831	2	0	0	0	0	1	1.074727	-17.478054	220	1	1.141697	0.9;
	5832	1	0	0	0	0	1	1.032963	-28.360713	220	1	1.13851	0.9;
	5833	1	1.401554	0.5	0	0	1	0.994738	-29.030251	110	1	1.1	0.9;
	5834	2	0	0	0	0	1	1.033591	67.189089	220	1	1.1	0.9;
	5835	1	0	0	0	0	1	1.049125	-29.884519	150	1	1.1	0.9;
	5836	2	69.77736	35.9	0	0	1	1.062955	-8.295414	220	1	1.154909	0.9;
	5837	1	28.63174	11	0	0	1	1.008145	-31.626375	110	1	1.1	0.9;
	5838	2	91.70167	17	0	0	1	1.038864	-43.935745	220	1	1.1	0.9;
	5839	1	0	0	0	0	1	1.02585	-23.240517	150	1	1.1	0.9;
	5840	1	0.100111	-19.1	0	0	1	0.984638	7.853647	220	1	1.1	0.9;
	5841	1	0	0	0	0	1	1.069406	4.526731	150	1	1.139443	0.9;
	5842	1	81.79068	10.8	0	0	1	1.031909	38.690583	220	1	1.1	0.9;
	5843	1	3.904329	1.1	0	0	1	0.950208	-23.686499	110	1	1.1	0.9;
	5844	1	163.3	70.3	0	0	1	1.079855	10.325486	150	1	1.133017	0.9;
	5845	1	43.94873	15.5	0	0	1	1.052421	-45.359057	220	1	1.114565	0.9;
	5846	1	193.8	18	0	0	1	1.088221	-28.504615	380	1	1.174719	0.9;
	5847	1	1	-2.3	0	0	1	1.074288	-15.033789	150	1	1.12089	0.9;
	5848	1	77.18558	23	0	0	1	1.07351	-10.406022	380	1	1.138533	0.9;
	5849	1	54.96094	13.2	0	0	1	1.060084	-29.771965	110	1	1.121098	0.9;
	5850	2	12.51387	4.2	0	0	1	0.94391	-27.963865	110	1	1.1	0.9;
	5851	1	0	0	0.148479	0	1	1.076723	-20.749304	380	1	1.147316	0.9;
	5852	1	5.506105	2.1	0	0	1	1.059015	-10.103849	220	1	1.126783	0.9;
	5853	1	132.1	13.2	0	0	1	1.06197	-30.509053	220	1	1.132011	0.9;
	5854	1	0	0	0	0	1	1.036869	-4.376082	220	1	1.108518	0.9;
	5855	1	8.309213	3	0	0	1	1.06264	-14.052799	110	1	1.120719	0.9;
	5856	1	0	0	0	0	1	1.055393	-49.17299	220	1	1.113337	0.9;
	5857	1	20.52275	13.3	0	0	1	1.06156	-21.070554	220	1	1.138573	0.9;
	5858	1	159.4	28.4	0	0	1	1.069563	-42.539063	220	1	1.116535	0.9;
	5859	2	0	0	0	0	1	1.074241	21.819732	380	1	1.119401	0.9;
	5860	2	0	0	0	0	1	1.089263	57.155542	380	1	1.1	0.9;
	5861	2	0	0	0	0	1	1.024533	-34.312593	150	1	1.1	0.9;
	5862	1	169.8	-19.6	0	0	1	1.083122	-29.908244	380	1	1.143609	0.9;
	5863	1	22.22464	8	0	0	1	0.996957	-21.649002	220	1	1.1	0.9;
	5864	2	5.706326	1.8	0	0	1	0.973755	-19.289463	110	1	1.1	0.9;
	5865	1	9.410433	13.8	0	0	1	1.087009	-26.234847	380	1	1.160901	0.9;
	5866	1	3.203552	-0.4	0	0	1	0.995916	-26.366452	110	1	1.1	0.9;
	5867	2	24.02664	0.1	0	0	1	1.094112	45.731275	380	1	1.105846	0.9;
	5868	1	172.5	48.7	0	0	1	1.056673	8.818928	380	1	1.1	0.9;
	5869	1	0	0	0	0	1	1.050785	57.535591	220	1	1.1	0.9;
	5870	1	0	0	0.209494	0	1	1.016674	14.555252	380	1	1.1	0.9;
	5871	1	34.13785	11.5	0	0	1	0.973501	-26.007539	110	1	1.1	0.9;
	5872	1	0	0	0	0	1	1.076553	-22.778637	220	1	1.161167	0.9;
	5873	2	153.9	11.5	0	0	1	1.073318	-7.552976	220	1	1.143571	0.9;
	5874	1	0	0	0	0	1	1.087891	-8.703125	380	1	1.141562	0.9;
	5875	1	0	0	0	0	1	1.050526	16.066956	220	1	1.130411	0.9;
	5876	1	141.2	34.8	0	0	1	1.077528	-13.726173	220	1	1.134619	0.9;
	5877	1	0	0	0	0	1	1.036978	-25.085194	150	1	1.1	0.9;
	5878	2	0	0	0	0	1	1.055674	2.027724	150	1	1.115167	0.9;
	5879	1	0	0	0.095843	0	1	1.066753	-13.398469	220	1	1.127124	0.9;
	5880	1	0	0	0	0	1	1.054478	-40.557794	380	1	1.113302	0.9;
	5881	2	3.704107	1.5	0	0	1	1.023182	-14.855462	110	1	1.1	0.9;
	5882	2	2.1	122.3	0	0	1	1.089921	-41.334228	380	1	1.133462	0.9;
	5883	1	0	0	0.589275	0	1	1.072853	-8.310527	380	1	1.134995	0.9;
	5884	1	31.73519	17.4	0	0	1	1.077938	-3.993865	220	1	1.131155	0.9;
	5885	1	18.62065	3.6	0	0	1	0.988561	-20.052075	110	1	1.1	0.9;
	5886	2	0	0	0	0	1	1.041142	25.766352	380	1	1.131082	0.9;
	5887	1	84.59379	12.9	0	0	1	1.057111	-47.954706	220	1	1.110935	0.9;
	5888	1	0	0	0	0	1	1.086685	-23.373311	380	1	1.15378	0.9;
	5889	2	0	0	0.060511	0	1	1.032237	66.125334	380	1	1.1	0.9;
	5890	1	9.510545	4.6	0	0	1	0.984596	-25.618529	150	1	1.1	0.9;
	5891	1	0	0	0	0	1	1.087236	-37.608004	380	1	1.141482	0.9;
	5892	1	49.65505	12.3	0	0	1	1.049382	-20.83599	220	1	1.119038	0.9;
	5893	1	0	0	0	0	1	1.087088	-36.918893	380	1	1.130868	0.9;
	5894	2	0	0	0.225995	0	1	1.071789	33.183569	380	1	1.142899	0.9;
	5895	1	0	0	0	0	1	1.056964	-41.040463	220	1	1.11895	0.9;
	5896	1	25.12786	9	0	0	1	1.01602	-17.273256	110	1	1.1	0.9;
	5897	1	25.82864	9.4	0	0	1	1.022696	-23.404539	110	1	1.120267	0.9;
	5898	2	4.605106	2.6	0	0	1	1.024091	-23.300389	110	1	1.121254	0.9;
	5899	1	13.21465	3.5	0	0	1	1.079191	-12.979237	110	1	1.144287	0.9;
	5900	1	49.25461	30.9	0.104003	0	1	1.024049	-40.441555	220	1	1.1	0.9;
	5901	1	0	0	0	0	1	1.024515	-20.770843	220	1	1.1	0.9;
	5902	2	41.34584	7.3	0	0	1	1.059239	61.761821	150	1	1.1	0.9;
	5903	1	0	0	0	0	1	1.067761	2.89451	380	1	1.134141	0.9;
	5904	1	0	0	0	0	1	1.03971	-15.695364	150	1	1.1	0.9;
	5905	1	88.19778	26.7	0	0	1	1.049528	-26.186212	220	1	1.128734	0.9;
	5906	1	20.0222	9.6	0	0	1	0.97439	-24.833898	150	1	1.1	0.9;
	5907	1	0	0	0	0	1	1.031013	-49.289345	150	1	1.1	0.9;
	5908	1	13.91543	4.3	0	0	1	0.951163	-13.633756	110	1	1.1	0.9;
	5909	2	0	0	0	0	1	1.086417	4.967047	150	1	1.15297	0.9;
	5910	1	20.32253	8.6	0	0	1	0.953921	-30.29828	110	1	1.1	0.9;
	5911	2	0	0	0	0	1	1.087182	-3.635396	220	1	1.139964	0.9;
	5912	2	0	0	0	0	1	0.980909	-24.198521	110	1	1.1	0.9;
	5913	2	0	0	0	0	1	1.068152	9.782387	150	1	1.1	0.9;
	5914	1	20.32253	-22	0	0	1	1.084396	-23.396692	220	1	1.120757	0.9;
	5915	1	0	0	0	0	1	1.080873	-4.716217	220	1	1.134941	0.9;
	5916	1	30.43374	12	0	0	1	0.975858	-20.255317	110	1	1.1	0.9;
	5917	1	24.82753	9	0	0	1	1.049556	-4.477922	110	1	1.125598	0.9;
	5918	1	113.4	61.9	0	0	1	1.027077	13.7058	380	1	1.1	0.9;
	5919	1	8.409324	2.6	0	0	1	1.030505	8.632899	110	1	1.111587	0.9;
	5920	1	100.1	48.9	0	0	1	1.027075	-20.600248	150	1	1.105527	0.9;
	5921	1	69.07658	24	0	0	1	1.055358	-35.62604	220	1	1.136731	0.9;
	5922	1	446.5	105.6	0	0	1	1.044447	-36.871401	380	1	1.112375	0.9;
	5923	1	42.8475	14.3	0	0	1	1.026475	-36.050383	110	1	1.1	0.9;
	5924	1	0	0	0	0	1	1.04845	-25.548613	150	1	1.1	0.9;
	5925	2	136.6	24.1	0	0	1	1.040241	-31.266163	220	1	1.134965	0.9;
	5926	2	38.2424	23.7	0	0	1	1.007427	-27.603015	150	1	1.192741	0.9;
	5927	1	0	0	0	0	1	1.046867	-19.02746	150	1	1.1	0.9;
	5928	1	20.72298	7.5	0	0	1	1.036215	-19.868752	110	1	1.1	0.9;
	5929	1	349.2	61.3	0.238392	0	1	1.100919	57.840858	380	1	1.11975	0.9;
	5930	1	68.47592	22.4	0	0	1	1.063891	0.75733	150	1	1.126409	0.9;
	5931	1	0.900999	0.3	0	0	1	1.018009	-32.682929	150	1	1.1	0.9;
	5932	1	27.43041	-36.4	0	0	1	1.031697	-7.88345	220	1	1.1	0.9;
	5933	1	208.2404	11.9	0	0	1	1.087188	-2.777686	220	1	1.146766	0.9;
	5934	1	118.4	7.1	0	0	1	1.064529	-44.086637	220	1	1.113115	0.9;
	5935	1	62.16893	0.7	0	0	1	1.015545	-39.165468	220	1	1.1	0.9;
	5936	1	1.101221	0	0	0	1	1.023892	-14.330241	110	1	1.1	0.9;
	5937	1	26.32919	8.7	0	0	1	0.953015	-31.147533	110	1	1.1	0.9;
	5938	1	7.608436	1.3	0	0	1	0.942568	-27.781795	110	1	1.1	0.9;
	5939	1	0	0	0	0	1	0.949519	-30.776944	150	1	1.1	0.864418;
	5940	1	3.103441	-5.2	0	0	1	1.027329	-14.644898	110	1	1.1	0.9;
	5941	2	253.4	57.4	0	0	1	1.053545	-20.204923	220	1	1.122795	0.9;
	5942	1	262.1	13.9	0	0	1	1.032209	3.928209	380	1	1.1	0.9;
	5943	2	0	0	0	0	1	1.018516	-42.885414	380	1	1.142411	0.9;
	5944	1	0	0	0	0	1	1.048385	-12.37184	220	1	1.133729	0.9;
	5945	1	3.704107	1.3	0	0	1	1.057777	-20.696203	110	1	1.115186	0.9;
	5946	1	0	0	0	0	1	1.033974	5.282982	380	1	1.107744	0.9;
	5947	1	105.5	35.3	0	0	1	1.076348	-0.353141	150	1	1.144335	0.9;
	5948	1	488.7	118.4	0	0	1	1.07277	-31.674175	380	1	1.132614	0.9;
	5949	1	216	56.2	0	0	1	1.029934	-45.003029	220	1	1.1	0.9;
	5950	1	0	0	0	0	1	0.9464	-27.90981	150	1	1.1	0.9;
	5951	1	0	0	0	0	1	1.072568	-23.431129	220	1	1.137798	0.9;
	5952	1	328.9	-9.3	0	0	1	1.081286	-39.281876	380	1	1.131758	0.9;
	5953	1	0	0	0	0	1	1.038673	-6.925344	220	1	1.129928	0.9;
	5954	1	43.14784	14.4	0	0	1	1.054552	-34.527629	220	1	1.132961	0.9;
	5955	1	43.14784	14.8	0	0	1	1.055765	-27.087956	220	1	1.1183	0.9;
	5956	1	0	0	0	0	1	1.082332	-16.235945	380	1	1.137181	0.9;
	5957	2	205.4	37.7	0	0	1	1.080647	7.793171	380	1	1.107534	0.9;
	5958	1	0	0	0	0	1	1.075468	-19.759654	220	1	1.140141	0.9;
	5959	1	0	0	0	0	1	1.002186	-22.335738	110	1	1.1	0.9;
	5960	1	23.92653	7.9	0	0	1	0.933918	-29.143855	110	1	1.1	0.9;
	5961	1	114.8	33.1	0	0	1	1.01001	-19.585828	110	1	1.1	0.9;
	5962	1	0	0	0	0	1	1.079308	-16.365113	380	1	1.136385	0.9;
	5963	1	0	0	0	0	1	1.064363	1.88895	380	1	1.141739	0.9;
	5964	1	51.75739	21	0	0	1	1.019818	-30.289894	110	1	1.1	0.9;
	5965	1	15.1414	11.3	0	0	1	1.017721	-26.785907	150	1	1.1	0.9;
	5966	1	8.909878	2.4	0	0	1	0.990852	-25.023866	110	1	1.1	0.9;
	5967	1	26.62953	9.4	0	0	1	0.999106	-22.353072	110	1	1.1	0.9;
	5968	1	153.8	23.4	0	0	1	1.083585	-2.832034	380	1	1.144574	0.9;
	5969	1	0	0	0	0	1	1.081261	10.439151	150	1	1.134393	0.9;
	5970	1	0	0	0	0	1	1.027424	-21.729128	150	1	1.1	0.9;
	5971	2	0	0	0.113564	0	1	1.095569	57.196065	380	1	1.112409	0.9;
	5972	1	24.72742	6.4	0	0	1	1.03143	-22.823498	110	1	1.1	0.9;
	5973	1	0	0	0	0	1	1.068601	-8.940777	220	1	1.135766	0.9;
	5974	1	124.8	45.6	0	0	1	1.032226	16.675612	380	1	1.1	0.9;
	5975	1	0	0	0	0	1	1.076322	-17.416168	380	1	1.138417	0.9;
	5976	2	9.610656	20.2	0	0	1	1.057273	0.385796	220	1	1.128352	0.9;
	5977	1	1.301443	0	0	0	1	0.949704	-27.156203	110	1	1.1	0.9;
	5978	1	0	0	0	0	1	0.965656	-27.541312	150	1	1.1	0.9;
	5979	2	117.7305	51.5	0	0	1	1.059909	-26.674879	220	1	1.121655	0.9;
	5980	1	8.809768	2.9	0	0	1	0.994758	-29.613872	110	1	1.1	0.9;
	5981	1	109.1	15	0	0	1	1.03768	-49.745802	220	1	1.1	0.9;
	5982	1	52.35805	7.4	0	0	1	1.065344	-43.914276	220	1	1.114275	0.9;
	5983	1	16.51831	2.1	0	0	1	1.025271	-31.165653	110	1	1.1	0.9;
	5984	1	69.97758	23.1	0	0	1	1.061826	-4.201612	110	1	1.13516	0.9;
	5985	1	25.82864	7.2	0	0	1	1.077163	9.472921	110	1	1.146102	0.9;
	5986	2	91.70167	-0.5	0	0	1	1.019773	15.722766	220	1	1.1	0.9;
	5987	1	0	0	0	0	1	1.036743	-27.395713	220	1	1.1	0.9;
	5988	1	174	49.6	0	0	1	1.022805	-17.463685	220	1	1.1	0.9;
	5989	1	15.21687	7.2	0	0	1	1.067813	-36.042723	220	1	1.120506	0.9;
	5990	1	137.4	81.7	0	0	1	1.051816	-15.081333	380	1	1.109578	0.9;
	5991	1	0	0	0	0	1	1.014528	-24.673955	150	1	1.1	0.9;
	5992	1	0	0	0	0	1	1.086034	-37.480475	380	1	1.130486	0.9;
	5993	2	65.77292	5.2	0	0	1	1.013	62.491973	150	1	1.1	0.9;
	5994	1	483.8	105.2	0	0	1	1.038566	-23.423552	120	1	1.1	0.9;
	5995	1	14.01554	6	0	0	1	1.022897	-21.327496	110	1	1.1	0.9;
	5996	1	0	0	0	0	1	1.09907	60.570057	380	1	1.116979	0.9;
	5997	2	44.84972	17.4	0	0	1	1.050203	-31.221003	220	1	1.13839	0.9;
	5998	1	0	0	0	0	1	1.085914	-2.507452	220	1	1.146086	0.9;
	5999	1	13.31476	3.5	0	0	1	0.995327	-16.051557	110	1	1.1	0.9;
	6000	1	0	0	0	0	1	1.077932	41.758959	380	1	1.115599	0.9;
	6001	1	0	0	0	0	1	1.057361	-14.887035	220	1	1.110697	0.9;
	6002	1	0	0	0	0	1	1.043607	-15.093164	220	1	1.130497	0.9;
	6003	1	0	0	0	0	1	0.995816	-22.6443	150	1	1.1	0.9;
	6004	2	0.4	191.7	0	0	1	1.070318	48.334417	220	1	1.1	0.9;
	6005	1	0	0	0	0	1	1.046862	-34.257461	220	1	1.109422	0.9;
	6006	2	0	0	0	0	1	1.081926	4.176302	380	1	1.151585	0.9;
	6007	1	18.82087	1.7	0	0	1	1.064212	-51.695776	220	1	1.1	0.9;
	6008	2	0	0	0	0	1	1.010636	-22.814195	110	1	1.1	0.9;
	6009	2	0	0	0	0	1	1.024545	-17.662628	110	1	1.111218	0.9;
	6010	1	35.03885	12.9	0	0	1	0.943035	-28.85938	110	1	1.1	0.9;
	6011	2	63.06993	16.3	0	0	1	1.096201	-11.296894	220	1	1.168992	0.9;
	6012	1	134.1	24.8	0	0	1	1.076419	-22.873563	220	1	1.161776	0.9;
	6013	1	57.56382	21.1	0	0	1	1.051427	-35.986504	220	1	1.133126	0.9;
	6014	2	13.8	461.3	0	0	1	1.100184	-36.835036	380	1	1.142509	0.9;
	6015	1	0	0	0	0	1	1.063503	-22.635004	220	1	1.1187	0.9;
	6016	1	0	0	0	0	1	1.097928	18.233552	220	1	1.174095	0.9;
	6017	1	23.42597	14.5	0	0	1	0.99237	-23.526985	150	1	1.1	0.9;
	6018	1	0	0	0	0	1	1.088472	-0.354998	220	1	1.148843	0.9;
	6019	1	0	0	0	0	1	1.087798	-24.889604	380	1	1.146691	0.9;
	6020	2	72.78069	13.3	0	0	1	1.061794	-13.970465	220	1	1.133161	0.9;
	6021	1	176.7	41.2	0	0	1	1.099042	-2.830485	220	1	1.156273	0.9;
	6022	2	0	0	0	0	1	1.005364	-39.657197	110	1	1.1	0.9;
	6023	1	4.805328	1	0	0	1	0.997261	-16.599221	110	1	1.1	0.9;
	6024	1	0	0	0	14.06254	1	0.926407	-29.465863	150	1	1.1	0.864633;
	6025	1	0	0	0	0	1	0.979261	-23.950358	150	1	1.1	0.9;
	6026	1	27.63063	17.1	0	0	1	0.985004	-23.231746	150	1	1.1	0.9;
	6027	2	13.91543	3.8	0	0	1	0.992868	-23.666406	110	1	1.1	0.9;
	6028	2	0	0	0	0	1	1.095255	-11.352283	220	1	1.170047	0.9;
	6029	1	0	0	0	0	1	1.054765	-36.049798	220	1	1.136747	0.9;
	6030	1	15.31698	5	0	0	1	0.976271	-28.877187	110	1	1.1	0.9;
	6031	1	4.605106	1.7	0	0	1	1.023128	-17.130944	110	1	1.125277	0.9;
	6032	2	0	0	0	0	1	1.062421	-0.730386	220	1	1.130532	0.9;
	6033	1	0	0	0.063758	0	1	1.03662	65.964673	380	1	1.1	0.9;
	6034	2	30.83419	80	0	0	1	1.072091	-42.140952	220	1	1.11772	0.9;
	6035	1	117.5	40	0	0	1	1.04931	-15.6741	220	1	1.113572	0.9;
	6036	2	0	0	0	-18.4703	1	1.081364	-33.664365	220	1	1.132436	0.9;
	6037	1	0	0	0	0	1	1.070912	26.947116	380	1	1.114501	0.9;
	6038	1	29.73297	11	0	0	1	0.923661	-34.977629	110	1	1.1	0.9;
	6039	1	199.6	27.1	0	0	1	1.085649	-41.781945	380	1	1.131603	0.9;
	6040	1	435.2	1.3	0	0	1	1.04832	-45.58922	220	1	1.1	0.9;
	6041	2	0	0	0	0	1	0.993273	11.063676	220	1	1.1	0.9;
	6042	2	0	0	0	0	1	1.068652	2.4194	150	1	1.14018	0.9;
	6043	1	74.5827	16.6	0	0	1	1.038859	-44.171832	220	1	1.1	0.9;
	6044	1	0	0	0	0	1	1.042707	-45.732589	220	1	1.164514	0.9;
	6045	1	0	0	0.095843	0	1	1.051072	-18.544256	220	1	1.118268	0.9;
	6046	1	0	0	0.215849	0	1	1.063637	-23.173622	380	1	1.121715	0.9;
	6047	1	0	0	0	0	1	1.057483	-2.587818	380	1	1.128236	0.9;
	6048	1	261.7	104.6	0	0	1	1.093868	56.369567	380	1	1.114533	0.9;
	6049	1	132.9	13.6	0	0	1	1.028209	-48.562922	220	1	1.1	0.9;
	6050	1	0	0	0	0	1	1.063028	-22.975395	220	1	1.127676	0.9;
	6051	1	0	0	0	0	1	1.091	-15.940599	220	1	1.179862	0.9;
	6052	1	0	0	0	0	1	0.962593	-21.559922	110	1	1.1	0.9;
	6053	1	20.0222	6	0	0	1	0.975625	-25.689073	110	1	1.1	0.9;
	6054	1	37.94207	7.5	0	0	1	1.060896	-22.255199	220	1	1.138635	0.9;
	6055	1	27.93097	17.3	0	0	1	0.999672	-21.935927	150	1	1.1	0.9;
	6056	1	0	0	0	0	1	1.05306	-48.841622	220	1	1.110658	0.9;
	6057	2	-8.2091	-2.4	0	0	1	1.060182	-43.538599	220	1	1.110287	0.9;
	6058	2	0	0	0	-116.149	1	1.09621	-36.709521	380	1	1.163121	0.9;
	6059	1	0	0	0	0	1	1.027994	-30.888148	110	1	1.1	0.9;
	6060	1	66.67392	27.8	0	0	1	1.062327	-19.761968	220	1	1.1	0.9;
	6061	1	0	0	0	0	1	1.055201	-24.385563	220	1	1.129931	0.9;
	6062	1	70.67836	11.9	0	0	1	1.026703	-21.407059	220	1	1.1	0.9;
	6063	1	0	0	0	0	1	1.020236	-17.246564	150	1	1.1	0.9;
	6064	1	0	0	0	0	1	1.082942	16.343348	380	1	1.128108	0.9;
	6065	1	0	0	0	0	1	1.10196	-24.275897	380	1	1.15937	0.9;
	6066	2	2.803108	1.3	0	0	1	0.943281	-24.86378	110	1	1.1	0.9;
	6067	1	0	0	0	18.58871	1	1.037347	-29.423816	150	1	1.1	0.9;
	6068	1	42.54717	12.5	0	0	1	1.080564	-33.35747	220	1	1.175497	0.9;
	6069	2	30.43374	7.8	0	0	1	1.08393	-14.704322	220	1	1.14476	0.9;
	6070	1	0	0	0	0	1	1.099764	-16.504415	380	1	1.156848	0.9;
	6071	1	0	0	0	0	1	1.039667	-39.179092	220	1	1.1	0.9;
	6072	1	45.95095	16.7	0	0	1	1.048484	-20.418873	110	1	1.105051	0.9;
	6073	1	0	0	0	0	1	0.989976	-24.007308	110	1	1.1	0.9;
	6074	1	309	74.5	0	0	1	1.044687	-36.46346	220	1	1.150912	0.9;
	6075	1	0	0	0	0	1	1.054196	-18.35252	150	1	1.108653	0.9;
	6076	1	0	0	0	0	1	1.03054	-24.947004	150	1	1.1	0.9;
	6077	1	11.91321	1.8	0	0	1	1.06171	-7.190667	220	1	1.119767	0.9;
	6078	1	0	0	0	0	1	1.002231	-23.130116	150	1	1.1	0.9;
	6079	2	0	0	0	0	1	1.105368	-8.714115	380	1	1.147639	0.9;
	6080	1	0	0	0.175312	0	1	1.036985	-4.734042	380	1	1.1	0.9;
	6081	1	89.2	42.7	0	0	1	1.020382	-23.331467	150	1	1.1	0.9;
	6082	1	0.100111	4.4	0	0	1	0.978369	-11.9913	110	1	1.1	0.9;
	6083	1	119.5	44.5	0	0	1	1.065299	9.496505	150	1	1.1	0.9;
	6084	2	264.9	46.8	0	0	1	1.054613	1.746885	220	1	1.129389	0.9;
	6085	1	33.23685	11.1	0	0	1	0.932706	-28.535798	110	1	1.1	0.9;
	6086	1	0	0	0	0	1	1.091239	-20.218299	380	1	1.157342	0.9;
	6087	1	12.11343	5.4	0	0	1	0.960569	-21.666921	110	1	1.1	0.9;
	6088	1	7.608436	6.3	0	0	1	0.937567	-28.026039	110	1	1.1	0.9;
	6089	1	72.48036	44.9	0	0	1	1.037304	-29.42381	150	1	1.1	0.9;
	6090	1	0	0	0	0	1	1.075197	-0.680619	220	1	1.144232	0.9;
	6091	2	0	0	0	0	1	1.045409	59.25741	220	1	1.1	0.9;
	6092	1	0	0	0	0	1	1.075828	-3.806387	220	1	1.135615	0.9;
	6093	1	-20.8231	-8.8	0	0	1	1.043545	-42.082383	220	1	1.1	0.9;
	6094	1	139.8	-23.6	0	0	1	1.059143	-18.845121	220	1	1.113664	0.9;
	6095	1	61.06771	11.1	0	0	1	1.050581	-13.061337	220	1	1.108262	0.9;
	6096	1	-1.60178	-0.9	0	0	1	0.988761	-29.661967	110	1	1.1	0.9;
	6097	1	32.53607	11.7	0	0	1	0.975779	-21.700551	110	1	1.1	0.9;
	6098	1	18.01998	5.9	0	0	1	0.941107	-31.117418	110	1	1.1	0.9;
	6099	1	0	0	0	0	1	1.059886	-8.819501	220	1	1.120388	0.9;
	6100	1	151.5596	-314.2	0	0	1	1.072116	16.080958	380	1	1.139858	0.9;
	6101	1	487.2	116.7	0	0	1	1.057279	-42.230886	380	1	1.127855	0.9;
	6102	1	18.12009	6	0	0	1	0.951031	-27.309906	110	1	1.1	0.9;
	6103	1	0	0	0	0	1	0.981805	-25.683128	150	1	1.1	0.9;
	6104	1	0	0	0.296959	0	1	1.078594	-15.019484	380	1	1.142612	0.9;
	6105	1	0	0	0	0	1	1.078673	-20.067238	220	1	1.141476	0.9;
	6106	1	-6.70744	-2.4	0	0	1	1.049342	-14.361741	110	1	1.143421	0.9;
	6107	2	39.24351	11.1	0	0	1	0.959147	-20.264117	110	1	1.1	0.9;
	6108	2	279.5	104.7	0	0	1	1.087909	-3.435881	220	1	1.141473	0.9;
	6109	1	0	0	0	0	1	1.077344	-1.394687	380	1	1.127873	0.9;
	6110	2	0	0	0	0	1	1.079105	-33.580962	380	1	1.157871	0.9;
	6111	1	0	0	0	0	1	1.069231	-9.742623	220	1	1.136559	0.9;
	6112	2	0	0	0	0	1	1.026168	-23.994746	380	1	1.138388	0.9;
	6113	2	0	0	0	0	1	1.057182	-13.903916	110	1	1.138993	0.9;
	6114	1	0	0	0	0	1	1.105619	-13.114636	380	1	1.183143	0.9;
	6115	1	137.4	23.9	0	0	1	1.07556	-20.425849	220	1	1.154846	0.9;
	6116	1	0	0	0	0	1	1.10042	-18.790943	380	1	1.165381	0.9;
	6117	2	248.5	97.9	0	0	1	1.075038	-9.508839	220	1	1.145278	0.9;
	6118	1	25.62842	51.6	0	0	1	1.049914	-35.838739	220	1	1.1	0.9;
	6119	1	0	0	0	0	1	1.063143	1.007039	150	1	1.138444	0.9;
	6120	1	0	0	0	0	1	1.050952	-31.314431	220	1	1.137649	0.9;
	6121	1	5.906549	1.7	0	0	1	1.051179	-44.679871	220	1	1.11408	0.9;
	6122	1	267.1	57.4	0	0	1	1.048235	-34.103675	380	1	1.111609	0.9;
	6123	1	6.807548	2.5	0	0	1	1.024163	-15.493635	110	1	1.1	0.9;
	6124	2	0	0	0	0	1	1.070632	33.324996	380	1	1.1	0.9;
	6125	1	0	0	0	0	1	1.085381	-13.686767	380	1	1.152565	0.9;
	6126	1	18.42042	6	0	0	1	0.956632	-29.130177	110	1	1.1	0.9;
	6127	2	0	0	0	0	1	1.060727	-43.480472	220	1	1.110576	0.9;
	6128	1	0	0	0	0	1	1.051721	-0.958968	380	1	1.137	0.9;
	6129	1	53.65949	19.5	0	0	1	1.035538	-23.061156	110	1	1.1	0.9;
	6130	1	0	0	0	0	1	1.069992	-15.035183	380	1	1.121683	0.9;
	6131	1	4.204662	1.2	0	0	1	1.076469	-38.296951	380	1	1.136847	0.9;
	6132	2	25.72853	9.3	0	0	1	1.017453	-19.982289	110	1	1.107279	0.9;
	6133	1	207.2	28.6	0	0	1	1.041344	-17.213487	220	1	1.124723	0.9;
	6134	1	0	0	0	0	1	1.077093	-22.997539	220	1	1.161969	0.9;
	6135	2	0	0	0	0	1	0.988455	-24.586053	110	1	1.1	0.9;
	6136	1	12.61399	7.8	0	0	1	1.025855	-23.24051	150	1	1.1	0.9;
	6137	1	0	0	0	0	1	1.051039	9.645424	380	1	1.111581	0.9;
	6138	1	0	0	0	0	1	0.959856	-29.494059	150	1	1.1	0.887446;
	6139	1	27.53052	9.9	0	0	1	1.068604	-41.187665	380	1	1.133458	0.9;
	6140	1	99	23.2	0	0	1	1.027551	-40.400441	220	1	1.1	0.9;
	6141	1	60.26682	21.3	0	0	1	0.99894	-22.299386	110	1	1.1	0.9;
	6142	1	0	0	0	0	1	1.068515	-39.425881	380	1	1.114232	0.9;
	6143	1	61.36804	23.2	0	0	1	1.076461	-22.862519	220	1	1.161933	0.9;
	6144	1	0	0	0	0	1	1.073824	-23.07646	220	1	1.138904	0.9;
	6145	1	34.13785	21.1	0	0	1	0.987472	-25.207427	150	1	1.1	0.9;
	6146	1	283.2	101.8	0.21599	0	1	1.057096	-15.673053	380	1	1.108895	0.9;
	6147	1	139.9	-26.2	0	0	1	1.003628	-29.120001	220	1	1.1	0.9;
	6148	1	-1377.93	-320	0	20	1	1.083907	-6.534223	380	1	1.148785	0.9;
	6149	2	0	0	0	0	1	1.068079	-34.685433	380	1	1.130792	0.9;
	6150	1	74.6828	46.3	0	0	1	1.030141	-25.359954	150	1	1.1	0.9;
	6151	2	0	0	0	0	1	1.077474	-15.03894	380	1	1.123862	0.9;
	6152	1	140.1	25.5	0	0	1	1.054894	-49.226841	220	1	1.112397	0.9;
	6153	1	11.91321	4.4	0	0	1	1.001384	-18.125567	110	1	1.1	0.9;
	6154	1	0	0	0	0	1	1.092494	-1.989532	220	1	1.145855	0.9;
	6155	2	0	0	0	0	1	1.032778	-15.915349	110	1	1.1	0.9;
	6156	1	0	0	0.095843	0	1	1.031108	-13.813238	220	1	1.10649	0.9;
	6157	1	15.21687	3.9	0	0	1	0.980423	-24.240767	110	1	1.1	0.9;
	6158	1	0	0	0	0	1	1.028981	-24.438798	150	1	1.1	0.9;
	6159	1	154.9	25	0	0	1	1.098209	-21.68584	380	1	1.146707	0.9;
	6160	1	46.35139	9.1	0	0	1	1.05749	-24.073933	220	1	1.132168	0.9;
	6161	1	43.34806	16.2	0	0	1	0.93588	-25.880533	110	1	1.1	0.9;
	6162	2	0	0	0	0	1	1.092773	-30.360135	220	1	1.144357	0.9;
	6163	1	79.48813	17	0	0	1	1.089613	27.619926	220	1	1.175541	0.9;
	6164	1	-73.982	-29.8	0	0	1	1.031019	-44.81095	220	1	1.1	0.9;
	6165	1	42.24684	5.9	0	0	1	1.06228	-45.954433	220	1	1.114735	0.9;
	6166	2	0	0	0	0	1	1.086974	1.804658	380	1	1.135946	0.9;
	6167	1	253.4	81.7	0	0	1	0.988843	-21.431804	220	1	1.1	0.9;
	6168	2	0	0	0	0	1	1.063195	-11.384981	110	1	1.125153	0.9;
	6169	1	4.204662	0.7	0	0	1	0.961765	-13.582751	110	1	1.1	0.9;
	6170	1	0	0	0	0	1	1.074043	-6.617969	220	1	1.140129	0.9;
	6171	1	55.76183	19.9	0	0	1	1.02534	-11.788186	110	1	1.1	0.9;
	6172	1	0	0	0	15.18218	1	1.013522	-27.244383	150	1	1.1	0.9;
	6173	2	7.308103	2.5	0	0	1	0.986089	-24.902563	110	1	1.1	0.9;
	6174	2	140.4	18.1	0	0	1	1.052874	-20.338728	220	1	1.122169	0.9;
	6175	2	30.73408	6.3	0	0	1	1.072818	-18.704207	220	1	1.131051	0.9;
	6176	2	270.3	86.8	0	0	1	0.9765	6.846003	220	1	1.1	0.9;
	6177	1	76.78513	30.7	0	0	1	1.056814	-35.770686	220	1	1.138638	0.9;
	6178	1	20.0222	10	0	0	1	1.059937	-35.375246	220	1	1.141641	0.9;
	6179	2	0.5	159.8	0	0	1	1.079868	46.300064	380	1	1.1	0.9;
	6180	1	0	0	0	0	1	1.057454	-27.547532	380	1	1.126528	0.9;
	6181	1	0	0	0	0	1	1.048086	-25.500506	150	1	1.1	0.9;
	6182	1	3.403774	1.1	0	0	1	0.983494	-17.071732	110	1	1.1	0.9;
	6183	2	59.46593	21.4	0	0	1	1.075455	-24.368691	220	1	1.136862	0.9;
	6184	1	106	24.6	0	0	1	1.062929	-22.61292	220	1	1.109005	0.9;
	6185	1	0	0	0	0	1	1.035265	-23.333976	220	1	1.112017	0.9;
	6186	1	0	0	0	0	1	0.987316	-25.219959	150	1	1.1	0.9;
	6187	2	0	0	0	0	1	1.078273	-6.286875	220	1	1.136933	0.9;
	6188	2	0	0	0	0	1	1.093237	-20.241901	380	1	1.163895	0.9;
	6189	1	36.84085	14.4	0	0	1	1.080237	2.913046	150	1	1.150586	0.9;
	6190	1	68.77625	28.2	0	0	1	1.063691	-3.19073	150	1	1.133875	0.9;
	6191	1	211.1	93.1	0	0	1	0.953481	-18.655188	110	1	1.1	0.9;
	6192	1	326.9	-82.3	0	0	1	1.081447	-9.210569	150	1	1.133266	0.9;
	6193	1	55.46149	18.5	0	0	1	0.918738	-30.074142	110	1	1.1	0.9;
	6194	1	58.16449	16.5	0	0	1	1.05393	-13.405807	220	1	1.124547	0.9;
	6195	2	196.4	42.3	0	0	1	0.923923	33.397245	220	1	1.1	0.9;
	6196	1	0	0	0	0	1	0.995041	9.043046	220	1	1.1	0.9;
	6197	2	127.1	44.7	0	0	1	1.058864	-9.105654	220	1	1.119526	0.9;
	6198	2	0	0	0	0	1	1.041152	25.766409	380	1	1.131091	0.9;
	6199	1	0	0	0.198044	0	1	1.013401	-7.172885	380	1	1.1	0.9;
	6200	1	3.804218	2.8	0	0	1	0.981372	-24.251981	110	1	1.1	0.9;
	6201	2	0	0	0	0	1	1.029304	13.664973	220	1	1.106738	0.9;
	6202	1	184.8822	-21.1	0	0	1	1.064999	-43.928154	220	1	1.113832	0.9;
	6203	1	47.75294	11	0	0	1	1.091573	-7.368307	220	1	1.159958	0.9;
	6204	2	0	0	0	0	1	1.082267	4.269472	150	1	1.149106	0.9;
	6205	2	89.49923	21.8	0	0	1	1.040728	62.167298	150	1	1.1	0.9;
	6206	1	122.8	20.8	0	0	1	1.05604	-48.746295	220	1	1.114016	0.9;
	6207	1	98	21.7	0	0	1	1.080678	-7.852401	220	1	1.147033	0.9;
	6208	2	29.03219	11.7	0	0	1	0.973091	-23.833649	110	1	1.1	0.9;
	6209	1	0	0	0	0	1	1.046585	-27.76032	220	1	1.135605	0.9;
	6210	2	0	0	0	0	1	1.052455	-10.823202	220	1	1.133845	0.9;
	6211	1	44.34917	19.3	0	0	1	1.078412	-4.506437	220	1	1.132576	0.9;
	6212	2	0.300333	0.5	0.530634	0	1	1.036579	3.883812	380	1	1.107142	0.9;
	6213	1	50.45594	36	0	0	1	0.959606	-22.494479	110	1	1.1	0.9;
	6214	1	2.502775	0	0	0	1	0.980347	-21.877488	110	1	1.1	0.9;
	6215	2	372.4	121.1	0	0	1	1.015	-18.792091	220	1	1.105651	0.9;
	6216	1	71.0788	44	0	0	1	0.975061	-24.159228	150	1	1.1	0.9;
	6217	1	0	0	0	0	1	1.095931	-11.242293	220	1	1.168605	0.9;
	6218	1	-16.8186	5.5	0	0	1	0.946027	-27.486371	150	1	1.1	0.9;
	6219	1	16.51831	5.5	0	0	1	1.045531	-35.177286	110	1	1.1	0.9;
	6220	2	0	0	0	0	1	1.086447	-8.592839	380	1	1.132348	0.9;
	6221	1	0	0	0	10.99016	1	1.04497	-20.907482	150	1	1.1	0.9;
	6222	1	0	8.5	0	0	1	1.010707	-25.084076	150	1	1.160839	0.9;
	6223	1	0	0	0	0	1	1.005096	-18.336722	220	1	1.1	0.9;
	6224	2	118.4	18.6	0	0	1	1.058864	6.987348	220	1	1.142218	0.9;
	6225	1	-42.2468	2.8	0	0	1	1.024096	-25.846291	110	1	1.2	0.8;
	6226	1	0	0	0	0	1	1.060964	-27.358247	220	1	1.122864	0.9;
	6227	2	0	0	0	0	1	1.054789	24.887974	380	1	1.115927	0.9;
	6228	1	0	0	0	0	1	1.029322	-21.484523	220	1	1.1	0.9;
	6229	1	6.407104	2.1	0	0	1	1.035599	-31.299201	110	1	1.1	0.9;
	6230	1	1.201332	0.4	0	0	1	1.054042	-21.107295	110	1	1.157016	0.9;
	6231	1	54.9	-96.8	0	0	1	1.05659	-48.427341	220	1	1.110646	0.9;
	6232	1	0	0	0	0	1	1.045079	66.081108	220	1	1.1	0.9;
	6233	1	0	0	0	0	1	1.065951	-18.662121	220	1	1.1	0.9;
	6234	1	0	0	0	0	1	1.010953	-22.101037	150	1	1.1	0.9;
	6235	1	0	0	0	0	1	0.994049	-24.64024	150	1	1.1	0.9;
	6236	2	0	0	0	0	1	1.0875	-13.992504	380	1	1.180223	0.9;
	6237	1	0	0	0	0	1	1.026604	-23.597996	150	1	1.1	0.9;
	6238	1	1.801998	0.6	0	0	1	0.98638	-34.264429	110	1	1.1	0.9;
	6239	2	0	0	0	0	1	1.093026	-37.597184	380	1	1.141092	0.9;
	6240	1	25.22797	8.3	0	0	1	0.970511	-28.863462	110	1	1.1	0.9;
	6241	1	199.6	24.4	0	0	1	1.043894	-41.480492	220	1	1.105778	0.9;
	6242	1	0	0	0	0	1	1.075884	-4.451664	220	1	1.129438	0.9;
	6243	1	11.41265	5.1	0	0	1	0.918147	-29.40424	110	1	1.1	0.9;
	6244	1	67.47482	6.1	0	0	1	1.051191	-49.012108	220	1	1.1	0.9;
	6245	1	0	0	0	0	1	1.093723	-36.345894	380	1	1.145736	0.9;
	6246	2	0	0	0	0	1	1.040622	25.765263	380	1	1.130982	0.9;
	6247	1	25.72853	15.9	0	0	1	0.983818	-25.515252	150	1	1.1	0.9;
	6248	1	119.3	-2.8	0	0	1	0.994148	11.205777	220	1	1.1	0.9;
	6249	1	0	0	0	0	1	1.081589	-2.852752	220	1	1.146424	0.9;
	6250	2	109	20.4	0	0	1	1.086045	-12.355378	220	1	1.139313	0.9;
	6251	2	0	0	0	0	1	1.081051	0.960035	380	1	1.132132	0.9;
	6252	2	156.7	36.8	0	0	1	1.076	-14.115553	220	1	1.1482	0.9;
	6253	1	-53.0588	-100	0	0	1	1.098701	-26.515796	220	1	1.156884	0.9;
	6254	1	-1235.46	-43	0	0	1	1.092189	55.845924	380	1	1.113721	0.9;
	6255	1	182.5	47	0	0	1	1.063904	-41.034613	380	1	1.127557	0.9;
	6256	1	8.309213	2.7	0	0	1	0.982119	-28.764297	110	1	1.1	0.9;
	6257	1	176.8	39.4	0	0	1	1.071421	-15.937999	380	1	1.149787	0.9;
	6258	1	31.03441	10	0	0	1	0.975228	-27.287982	110	1	1.1	0.9;
	6259	1	8.00888	2.6	0	0	1	1.014743	-28.455318	110	1	1.1	0.9;
	6260	2	0	0	0	0	1	1.100105	48.797681	380	1	1.115099	0.9;
	6261	2	0	0	0	0	1	1.000005	-31.20713	380	1	1.121882	0.9;
	6262	1	0.300333	-9	0	0	1	1.063401	0.831232	220	1	1.13951	0.9;
	6263	1	0	0	0	0	1	1.069651	-50.753656	220	1	1.1	0.9;
	6264	1	0	0	0	0	1	0.998941	-27.323126	150	1	1.1	0.9;
	6265	1	115.2	55.1	0	0	1	1.042609	12.84083	380	1	1.106336	0.9;
	6266	2	214.8	42.7	0	0	1	1.088264	-32.758278	380	1	1.13585	0.9;
	6267	1	0	0	0	0	1	1.035769	-39.464583	220	1	1.145359	0.9;
	6268	1	122.5	8.7	0	0	1	1.061237	-22.100636	220	1	1.138858	0.9;
	6269	2	0	0	0	0	1	1.052026	-35.416196	380	1	1.117805	0.9;
	6270	1	18.72076	7.7	0	0	1	1.001862	-35.322511	110	1	1.1	0.9;
	6271	1	0	0	0	0	1	1.009874	-22.688716	150	1	1.1	0.9;
	6272	1	18.41154	41.8	0	0	1	0.945991	-31.812812	110	1	1.1	0.9;
	6273	2	0	0	0	0	1	1.096447	55.364659	380	1	1.106328	0.9;
	6274	1	11.71299	3.7	0	0	1	1.043106	9.307185	110	1	1.123774	0.9;
	6275	1	0	0	0	0	1	1.040833	-19.936292	150	1	1.120403	0.9;
	6276	1	0	0	0	0	1	1.101371	-38.529556	380	1	1.153451	0.9;
	6277	1	6.707437	1	0	0	1	0.954492	-23.37845	110	1	1.1	0.9;
	6278	1	21.62398	9.8	0	0	1	0.971366	-22.740743	110	1	1.1	0.9;
	6279	2	0	0	0	0	1	1.050657	-1.399408	380	1	1.136473	0.9;
	6280	1	113.6	44.7	0	0	1	1.055623	-8.074091	220	1	1.116809	0.9;
	6281	1	33.33696	8.6	0	0	1	1.07276	10.330624	110	1	1.151393	0.9;
	6282	2	0	0	0	0	1	1.0475	-31.568458	220	1	1.1	0.9;
	6283	2	0	0	0	0	1	1.039853	-29.777016	110	1	1.1	0.9;
	6284	1	124.5	101.4	0	0	1	1.035432	-11.883466	220	1	1.1	0.9;
	6285	1	0	0	0	0	1	0.994648	-23.35805	150	1	1.1	0.9;
	6286	2	0.1	0.1	0	0	1	1.076684	-12.404791	380	1	1.141056	0.9;
	6287	1	117	20.4	0	0	1	1.055493	-48.713966	220	1	1.11203	0.9;
	6288	1	0	0	0	0	1	1.049103	-21.285338	110	1	1.110009	0.9;
	6289	2	76.38469	6.8	0	0	1	1.058267	63.062459	150	1	1.1	0.9;
	6290	1	0	0	0	0	1	0.987205	-17.426307	150	1	1.1	0.9;
	6291	2	160.4	63.9	0	0	1	1.032044	-35.062416	220	1	1.1	0.9;
	6292	2	197.7433	31.6	0	0	1	1.084727	-6.293475	220	1	1.127886	0.9;
	6293	1	3.203552	1.1	0	0	1	1.04701	-29.321258	110	1	1.11133	0.9;
	6294	1	135	25.7	0	0	1	1.067822	-34.734715	380	1	1.129705	0.9;
	6295	1	0	0	0	0	1	1.069514	-26.567806	220	1	1.166836	0.9;
	6296	1	21.62398	7.8	0	0	1	1.007633	-19.412386	110	1	1.1	0.9;
	6297	1	0	0	0	0	1	1.007693	-22.365392	150	1	1.1	0.9;
	6298	1	0	0	0	0	1	1.01936	-22.34446	150	1	1.1	0.9;
	6299	1	0.200222	1.8	0	0	1	1.05217	56.873664	220	1	1.1	0.9;
	6300	2	0	0	0	0	1	1.0795	-3.758793	220	1	1.132587	0.9;
	6301	2	0	0	0	0	1	1.010227	13.127002	220	1	1.1	0.9;
	6302	1	35.43929	5.3	0	0	1	1.051972	-49.131983	220	1	1.111148	0.9;
	6303	1	180.9	33.2	0	0	1	1.059317	-24.461569	220	1	1.160516	0.9;
	6304	1	61.76849	10.2	0	0	1	1.045134	-45.684949	220	1	1.1	0.9;
	6305	1	18.12009	10	0	0	1	1.069232	-19.083733	220	1	1.137208	0.9;
	6306	1	0	0	0	13.90073	1	1.017703	-18.376038	150	1	1.1	0.9;
	6307	1	86.4959	29	0	0	1	1.071593	26.944148	380	1	1.1	0.9;
	6308	2	28.53163	17.7	0	0	1	1.016274	-26.865265	150	1	1.184769	0.9;
	6309	1	20.0222	9	0	0	1	1.060401	-35.397056	220	1	1.141606	0.9;
	6310	1	0	0	0	0	1	1.076564	-13.516974	220	1	1.1302	0.9;
	6311	1	389.8	-68.1	0	0	1	1.032366	-51.30807	220	1	1.1	0.9;
	6312	1	56.56271	18.4	0	0	1	1.074754	0.052937	220	1	1.144831	0.9;
	6313	1	0	0	0.555664	0	1	1.084948	-21.280627	380	1	1.113679	0.9;
	6314	1	0	0	0	0	1	1.082516	-16.246016	220	1	1.164786	0.9;
	6315	1	120.1	2.5	0	0	1	1.035614	-41.0876	220	1	1.1	0.9;
	6316	1	125	107.2	0.216876	0	1	1.027718	9.28457	380	1	1.1	0.9;
	6317	2	0	0	0	0	1	1.047273	-18.139366	220	1	1.118541	0.9;
	6318	1	0	0	0	0	1	1.072264	-23.048182	220	1	1.1	0.9;
	6319	2	0	0	0	0	1	1.086045	-23.492834	220	1	1.14373	0.9;
	6320	1	0	0	0	0	1	1.081234	-42.537364	380	1	1.133961	0.9;
	6321	1	0	0	0	0	1	1.03074	-12.507289	220	1	1.123255	0.9;
	6322	2	95.606	7.6	0	0	1	1.085918	-4.552233	220	1	1.146563	0.9;
	6323	1	92.1	43.8	0	0	1	1.015854	9.200173	220	1	1.1	0.9;
	6324	1	0	0	0	0	1	0.987718	-16.757718	110	1	1.1	0.9;
	6325	1	0.600666	0.1	0	0	1	0.993291	-22.846945	110	1	1.1	0.9;
	6326	1	36.03996	22.3	0	0	1	0.956692	-29.273887	150	1	1.1	0.891251;
	6327	1	53.05883	75.8	0	0	1	1.054602	-0.478484	380	1	1.112484	0.9;
	6328	1	0	0	0	0	1	1.04282	59.90484	220	1	1.12881	0.9;
	6329	1	195.5	67.9	0	0	1	1.007794	7.207557	220	1	1.1	0.9;
	6330	2	89.09879	20.2	0	0	1	1.053818	-20.475751	220	1	1.138739	0.9;
	6331	1	0	0	0	0	1	1.016608	14.370089	380	1	1.1	0.9;
	6332	1	0	0	0	27.06707	1	0.942169	-33.245566	150	1	1.1	0.8222;
	6333	1	125.2	1.2	0	0	1	1.056467	-49.256756	220	1	1.114159	0.9;
	6334	1	0	0	0	0	1	1.01197	54.192839	220	1	1.1	0.9;
	6335	1	112.1	36.3	0	0	1	1.073208	2.290341	150	1	1.143538	0.9;
	6336	1	0	0	0	0	1	0.954112	-27.589869	110	1	1.1	0.9;
	6337	1	0	0	0	0	1	1.077312	-22.776803	220	1	1.162723	0.9;
	6338	1	0	0	0	0	1	1.059879	-22.960138	220	1	1.133847	0.9;
	6339	2	4.304773	2.2	0	0	1	0.99434	-24.147533	110	1	1.1	0.9;
	6340	1	89.69945	17.5	0	0	1	1.039552	-35.759806	220	1	1.1	0.9;
	6341	2	5.405994	67.5	0	0	1	1.084474	35.430297	380	1	1.112066	0.9;
	6342	1	0	0	0	0	1	1.060278	-37.225419	220	1	1.108837	0.9;
	6343	1	0	0	0	0	1	1.011983	3.688734	220	1	1.1	0.9;
	6344	1	0	0	0	0	1	1.054314	-15.855879	220	1	1.107245	0.9;
	6345	1	190.3	84.7	0	0	1	1.009557	-6.776649	220	1	1.1	0.9;
	6346	1	30.0333	8.6	0	0	1	1.060377	7.03155	150	1	1.112042	0.9;
	6347	2	91.9	41.9	0.266883	0	1	1.101579	-10.395554	380	1	1.129818	0.9;
	6348	1	2.302553	0.8	0	0	1	1.025834	-16.804169	110	1	1.127678	0.9;
	6349	1	46.35139	28.7	0	0	1	1.017672	-18.376018	150	1	1.1	0.9;
	6350	1	51.15672	12	0	0	1	1.028204	-38.133456	220	1	1.1	0.9;
	6351	1	2.903219	0	0	0	1	1.036493	-11.89682	110	1	1.1	0.9;
	6352	1	0	0	0	0	1	1.082724	16.347499	380	1	1.127614	0.9;
	6353	1	10.0111	2	0	0	1	1.058359	-27.618851	150	1	1.129168	0.9;
	6354	1	0	0	0	0	1	0.952939	54.709651	220	1	1.1	0.9;
	6355	1	9.310323	0.8	0	0	1	1.004515	-24.876005	110	1	1.1	0.9;
	6356	2	0	0	0	0	1	1.073182	-25.360554	220	1	1.114263	0.9;
	6357	1	0	0	0.095843	0	1	1.027663	-23.877185	220	1	1.1	0.9;
	6358	1	84.59379	21.7	0	0	1	1.041863	4.586964	110	1	1.108246	0.9;
	6359	1	-15.1168	5	0	0	1	1.054549	-18.221883	150	1	1.108994	0.9;
	6360	1	0	0	0	0	1	1.081878	-3.594659	220	1	1.135327	0.9;
	6361	1	0	0	0	0	1	1.053194	-18.336218	150	1	1.108187	0.9;
	6362	1	190	72.7	0	0	1	1.053794	-36.464733	220	1	1.135451	0.9;
	6363	1	0	0	0.136054	0	1	1.068978	-12.883753	380	1	1.129497	0.9;
	6364	1	53.25905	29.5	0	0	1	1.067604	9.787942	150	1	1.1	0.9;
	6365	1	52.55827	12.8	0	0	1	1.053173	-24.65335	220	1	1.127973	0.9;
	6366	1	0	0	0	0	1	0.996651	-27.472636	150	1	1.1	0.9;
	6367	1	226.8	53.3	0	0	1	1.052226	-16.169514	220	1	1.121632	0.9;
	6368	1	14.91654	5.4	0	0	1	1.048236	-19.459921	110	1	1.109924	0.9;
	6369	1	0	0	0	15.27401	1	1.029908	-24.505692	150	1	1.1	0.9;
	6370	1	0	0	0	0	1	1.032639	-24.956846	150	1	1.1	0.9;
	6371	2	235.8	23.3	0	0	1	0.999818	9.906003	220	1	1.1	0.9;
	6372	2	285.7	31.9	0	0	1	1.078864	-15.523704	220	1	1.120325	0.9;
	6373	2	0	0	0	0	1	1.030868	15.99602	380	1	1.1	0.9;
	6374	2	1.7	145.9	0	0	1	1.091553	-33.351982	380	1	1.150741	0.9;
	6375	1	0	0	0	0	1	1.073322	-7.553	220	1	1.143575	0.9;
	6376	1	0	0	0	0	1	1.056617	-32.522846	380	1	1.120094	0.9;
	6377	2	16.71854	6.4	0	0	1	1.044091	-14.586824	110	1	1.116468	0.9;
	6378	1	0	0	0	0	1	1.026456	-23.784051	150	1	1.1	0.9;
	6379	1	0	0	0.368139	296.694	1	1.068128	24.8825	380	1	1.113948	0.9;
	6380	2	0	0	0	0	1	1.072864	-11.677848	220	1	1.132599	0.9;
	6381	1	0	0	0	0	1	1.032512	-50.871762	220	1	1.1	0.9;
	6382	1	19.52164	6.4	0	0	1	0.982894	-29.717833	110	1	1.1	0.9;
	6383	1	324.9	37.3	0	0	1	1.075675	-10.068429	380	1	1.139374	0.9;
	6384	1	21.8242	6.3	0	0	1	1.0041	-24.176487	110	1	1.1	0.9;
	6385	1	-175.795	-19.6	0	0	1	1.033556	-37.877381	220	1	1.1	0.9;
	6386	2	225	92.8	0	0	1	1.019	-32.16571	220	1	1.1	0.9;
	6387	1	6.907659	1.9	0	0	1	1.038819	-10.276338	110	1	1.1	0.9;
	6388	1	120.6	35.1	0	0	1	1.072805	-31.841984	380	1	1.132943	0.9;
	6389	1	1.902109	0	0	0	1	1.027257	-15.260431	110	1	1.1	0.9;
	6390	1	0	0	0	0	1	1.039017	-21.371631	150	1	1.115499	0.9;
	6391	1	339.9	70.5	0	0	1	1.074284	-20.797503	220	1	1.153816	0.9;
	6392	1	12.91432	5.1	0	0	1	1.004928	-19.836462	110	1	1.1	0.9;
	6393	2	0	0	0	0	1	1.055409	-13.537061	220	1	1.109477	0.9;
	6394	1	120	26	0	0	1	1.07828	-20.216472	220	1	1.141162	0.9;
	6395	1	0	0	0	0	1	1.054461	-2.935294	380	1	1.109201	0.9;
	6396	1	111	36.6	0	0	1	1.049001	-11.740495	220	1	1.1	0.9;
	6397	1	0	0	0	0	1	1.089432	-37.120585	380	1	1.147081	0.9;
	6398	1	129.7	27.1	0	0	1	1.029894	-23.578135	110	1	1.1	0.9;
	6399	1	125.2	8.5	0	0	1	1.049259	-12.552286	220	1	1.114135	0.9;
	6400	2	41.94651	14.8	0	0	1	1.052421	-4.774741	380	1	1.116734	0.9;
	6401	1	0	0	0	0	1	1.03057	-24.302841	150	1	1.1	0.9;
	6402	2	0	0	0	0	1	1.014636	-25.16249	220	1	1.110395	0.9;
	6403	1	28.93208	8.3	0	0	1	1.065955	7.709858	150	1	1.11499	0.9;
	6404	1	0	0	0	0	1	1.03535	-11.369884	110	1	1.1	0.9;
	6405	1	-20.1223	6.6	0	0	1	1.068174	-18.299794	150	1	1.1	0.9;
	6406	2	0	0	0	0	1	1.102368	-13.737786	380	1	1.147482	0.9;
	6407	1	0	0	0	0	1	1.03317	-42.812728	220	1	1.1	0.9;
	6408	2	0	0	0	0	1	1.037545	-17.653059	220	1	1.1	0.9;
	6409	1	41.24573	48.4	0	0	1	1.085382	-39.425543	220	1	1.156124	0.9;
	6410	1	6.306993	5.1	0	0	1	0.950601	-26.897435	110	1	1.1	0.9;
	6411	1	2.803108	1	0	0	1	1.023958	-18.178164	110	1	1.117564	0.9;
	6412	2	260.8	43.9	0	0	1	1.043125	-7.630374	220	1	1.132224	0.9;
	6413	1	142.9	25.8	0	0	1	1.046649	-13.060636	220	1	1.132212	0.9;
	6414	1	80.1889	23.6	0	0	1	1.045005	-34.389209	220	1	1.107729	0.9;
	6415	1	62.66948	14.6	0	0	1	1.031993	-31.371938	110	1	1.1	0.9;
	6416	1	0	0	0	0	1	1.082159	6.989116	150	1	1.149172	0.9;
	6417	1	15.71743	9.7	0	0	1	0.978165	-24.260489	150	1	1.1	0.9;
	6418	2	62.16893	16.9	0	0	1	1.0245	-38.932567	220	1	1.1	0.9;
	6419	1	0	0	0	0	1	1.03884	-12.57322	220	1	1.109459	0.9;
	6420	1	18.82087	6.8	0	0	1	1.030835	-22.917884	110	1	1.1	0.9;
	6421	1	-13.8153	4.6	0	0	1	0.981876	-23.418564	150	1	1.1	0.9;
	6422	1	0	0	0	0	1	1.079519	-47.003717	380	1	1.136391	0.9;
	6423	1	0	0	0	0	1	1.039782	-15.690408	150	1	1.1	0.9;
	6424	1	0	0	0	0	1	1.088338	-33.115167	380	1	1.138573	0.9;
	6425	1	152.2	53.2	0	0	1	1.074205	-6.195953	380	1	1.124996	0.9;
	6426	2	0	0	0	0	1	1.036591	11.960313	220	1	1.1	0.9;
	6427	1	10.81199	3.6	0	0	1	0.933849	-34.308822	110	1	1.1	0.9;
	6428	1	20.22242	7.3	0	0	1	1.017845	-19.970386	110	1	1.107625	0.9;
	6429	1	0	0	0	0	1	1.040675	-25.193133	150	1	1.1	0.9;
	6430	2	220.6	70.7	0	0	1	1.058342	12.596166	380	1	1.13873	0.9;
	6431	1	82.89191	41.3	0	0	1	1.095332	9.125398	150	1	1.149585	0.9;
	6432	1	0	0	0	0	1	0.999681	-21.93368	150	1	1.1	0.9;
	6433	1	44.6495	8.8	0	0	1	1.058674	-23.947784	220	1	1.133268	0.9;
	6434	2	0	0	0	0	1	1.094605	36.287029	380	1	1.12965	0.9;
	6435	1	0	0	0	0	1	1.009268	-22.948292	150	1	1.1	0.9;
	6436	1	0	0	0	0	1	1.096434	-36.625065	380	1	1.162446	0.9;
	6437	1	109.5	33.6	0	0	1	1.014339	8.381863	220	1	1.1	0.9;
	6438	1	0	0	0	0	1	1.085954	-3.230672	220	1	1.147216	0.9;
	6439	1	102	26.6	0	0	1	1.024153	37.846255	220	1	1.1	0.9;
	6440	1	35.93985	22.3	0	0	1	0.971835	-27.334067	150	1	1.1	0.9;
	6441	1	4.605106	1.7	0	0	1	1.044719	-21.786113	110	1	1.145048	0.9;
	6442	1	0	0	0.168227	0	1	1.068341	2.568749	380	1	1.133927	0.9;
	6443	1	0	0	0	0	1	1.053328	-26.347794	220	1	1.11595	0.9;
	6444	2	0	0	0	0	1	1.02	-25.041055	220	1	1.1	0.9;
	6445	1	0	0	0	0	1	1.086052	-26.949954	380	1	1.149934	0.9;
	6446	1	0	0	0	0	1	1.035446	-20.101947	150	1	1.114749	0.9;
	6447	2	316	102.9	0	0	1	1.007727	-4.671612	220	1	1.1	0.9;
	6448	1	282.5	56.2	0	0	1	1.043512	-39.423441	380	1	1.110929	0.9;
	6449	1	0	0	0	0	1	0.963277	-13.510739	110	1	1.1	0.9;
	6450	1	0	0	0	-48.701	1	1.067629	-37.496046	220	1	1.12221	0.9;
	6451	1	0	0	0	0	1	1.070271	-34.585119	110	1	1.12291	0.9;
	6452	1	0.100111	0	0	0	1	1.062041	-20.988119	110	1	1.119912	0.9;
	6453	2	301.6	87.4	0	0	1	1.077368	5.472904	380	1	1.107623	0.9;
	6454	2	23.32586	4	0	0	1	1.074864	-18.40872	220	1	1.132589	0.9;
	6455	1	117.5	91	0.373248	0	1	1.031799	2.599251	380	1	1.1	0.9;
	6456	1	0	-0.6	0	0	1	1.061211	21.183032	110	1	1.112921	0.9;
	6457	1	0	0	0	0	1	1.084207	-15.954934	380	1	1.143257	0.9;
	6458	1	149.5267	34.9	0	0	1	1.059528	-15.386704	220	1	1.128491	0.9;
	6459	1	0	0	0	0	1	1.07519	-9.143774	220	1	1.154865	0.9;
	6460	1	29.43263	10.7	0	0	1	1.038743	-21.34799	110	1	1.144523	0.9;
	6461	1	0	0	0	0	1	1.095636	-13.769323	380	1	1.164439	0.9;
	6462	1	38.34251	12.5	0	0	1	1.056409	-35.931589	220	1	1.137995	0.9;
	6463	1	0	0	0	0	1	1.078627	-26.350347	220	1	1.186722	0.9;
	6464	1	0.100111	2.2	0	0	1	1.007849	12.817054	220	1	1.1	0.9;
	6465	1	5.706326	3	0	0	1	0.963904	-19.825086	110	1	1.1	0.9;
	6466	1	231.2	103.5	0	0	1	1.023852	-16.403982	220	1	1.1	0.9;
	6467	1	33.33696	5	0	0	1	1.081662	-3.825748	220	1	1.142686	0.9;
	6468	1	19.32142	7	0	0	1	1.023463	-19.578524	110	1	1.112221	0.9;
	6469	2	15.01665	5.6	0	0	1	1.072578	-19.095663	110	1	1.115058	0.9;
	6470	1	222	90.5	0	0	1	1.06056	-14.373026	220	1	1.116911	0.9;
	6471	1	102.3	19	0	0	1	1.081051	-12.900996	220	1	1.135129	0.9;
	6472	1	0	0	0.314024	0	1	1.076419	-9.823569	380	1	1.140061	0.9;
	6473	1	29.23241	9.4	0	0	1	1.040873	4.455223	110	1	1.109237	0.9;
	6474	2	244	97.8	0	0	1	1.087026	-41.330686	380	1	1.134046	0.9;
	6475	1	150.2	51.6	0	0	1	1.067445	-2.877794	150	1	1.137516	0.9;
	6476	1	99.4	22.2	0	0	1	1.052837	-49.088959	220	1	1.110304	0.9;
	6477	1	64.2	26.4	0	0	1	1.002192	-22.473139	110	1	1.1	0.9;
	6478	1	32.33585	16.8	0	0	1	0.958677	-24.853407	110	1	1.1	0.9;
	6479	1	0	0	0	0	1	0.984963	-25.556308	150	1	1.1	0.9;
	6480	2	14.71632	1.6	0	0	1	1.039663	-12.953042	220	1	1.11265	0.9;
	6481	1	46.75183	24.4	0	0	1	1.061896	-52.852314	220	1	1.1	0.9;
	6482	1	0	0	0	0	1	1.041091	25.762276	380	1	1.131161	0.9;
	6483	1	0	0	0	0	1	1.080008	-22.122891	220	1	1.165387	0.9;
	6484	1	6.00666	2	0	0	1	1.060706	-13.125439	110	1	1.11307	0.9;
	6485	1	26.02886	-14.1	0	0	1	1.073863	-19.463842	110	1	1.116262	0.9;
	6486	1	0	0	0	0	1	0.949534	-24.621387	110	1	1.1	0.9;
	6487	1	89.79956	19.7	0	0	1	1.072216	-24.991216	120	1	1.134687	0.9;
	6488	2	22.22464	5.6	0	0	1	1.075	-23.26968	380	1	1.140318	0.9;
	6489	1	73.58158	26.8	0	0	1	1.068403	-1.424278	220	1	1.137461	0.9;
	6490	2	0	0	0	0	1	0.955879	-26.369848	110	1	1.1	0.9;
	6491	1	0	0	0	0	1	1.082583	6.947018	150	1	1.149934	0.9;
	6492	2	0	0	0	0	1	1.082711	4.406445	380	1	1.152044	0.9;
	6493	1	0	0	0	0	1	1.06449	4.604321	380	1	1.112584	0.9;
	6494	1	122.8	20.4	0	0	1	0.993031	11.180845	220	1	1.1	0.9;
	6495	1	0	0	0	0	1	1.046813	-13.063111	220	1	1.132387	0.9;
	6496	1	0	0	0	0	1	1.060405	-23.719799	220	1	1.134562	0.9;
	6497	2	0	0	0	0	1	1.052168	-21.951725	380	1	1.127462	0.9;
	6498	1	0	0	0	0	1	1.070943	-33.530036	220	1	1.174196	0.9;
	6499	2	34.73852	9.1	0	0	1	0.963182	-22.295935	110	1	1.1	0.9;
	6500	1	64.47148	40	0	0	1	1.051773	-34.95763	220	1	1.131034	0.9;
	6501	1	-20.8231	6.8	0	0	1	1.031383	-16.111646	150	1	1.1	0.9;
	6502	2	0	0	0	0	1	1.104553	43.071515	380	1	1.11859	0.9;
	6503	1	-1083.9	-433.7	0	0	1	1.084483	-5.902296	380	1	1.151888	0.9;
	6504	1	0.100111	-9.6	0	0	1	1.070617	-18.153815	220	1	1.1	0.9;
	6505	1	0	0	0	0	1	1.026574	-23.570041	150	1	1.1	0.9;
	6506	2	151.1	31.8	0	0	1	1.042571	3.016788	220	1	1.134025	0.9;
	6507	1	0	0	0	0	1	1.067886	24.236579	380	1	1.116343	0.9;
	6508	1	10.21132	3.3	0	0	1	1.041318	-30.682819	110	1	1.106896	0.9;
	6509	1	17.61954	7.7	0	0	1	0.970543	-33.427373	110	1	1.1	0.9;
	6510	1	0	0	0	0	1	1.062878	-13.937703	220	1	1.130297	0.9;
	6511	1	0	0	0.195349	0	1	1.012958	-7.260364	380	1	1.1	0.9;
	6512	1	1.00111	0.2	0	0	1	1.053969	-20.664877	110	1	1.115495	0.9;
	6513	1	290.8	30.5	0	0	1	1.042657	-40.385792	380	1	1.109876	0.9;
	6514	2	0	0	0	0	1	1.101368	-6.115232	380	1	1.187387	0.9;
	6515	1	-18.5205	6.1	0	0	1	1.004132	-21.313701	150	1	1.1	0.9;
	6516	2	19.62176	5.3	0	0	1	0.967222	-19.816326	110	1	1.1	0.9;
	6517	1	7.207992	2.4	0	0	1	0.986616	-36.455833	110	1	1.1	0.9;
	6518	1	0	0	0	0	1	1.064958	-14.763438	220	1	1.129323	0.9;
	6519	1	0	0	0	0	1	0.986643	-24.276248	150	1	1.1	0.9;
	6520	1	55.46149	15.7	0	0	1	1.065181	-43.894138	220	1	1.114152	0.9;
	6521	1	0	0	0	21.91948	1	1.050826	-20.668445	150	1	1.1	0.9;
	6522	1	0	0	0	0	1	1.038628	-42.632862	220	1	1.1	0.9;
	6523	1	456.6	227.7	0	0	1	0.967657	53.590673	220	1	1.1	0.9;
	6524	1	99.9	19.9	0	0	1	1.063702	0.930173	150	1	1.125924	0.9;
	6525	1	72.68058	1.1	0	0	1	1.04923	-49.917687	220	1	1.111954	0.9;
	6526	1	0	0	0.080003	0	1	1.064053	-41.124704	380	1	1.13464	0.9;
	6527	1	83.19224	21.8	0	0	1	1.031448	-2.293725	220	1	1.109386	0.9;
	6528	1	6.306993	1.3	0	0	1	0.994536	-22.619568	110	1	1.1	0.9;
	6529	2	21.92431	14.1	0	0	1	0.983636	-24.837459	110	1	1.1	0.9;
	6530	1	8.108991	3.8	0	0	1	0.962399	-21.030095	110	1	1.1	0.9;
	6531	1	18.62065	6.1	0	0	1	1.078501	-3.935434	220	1	1.131642	0.9;
	6532	1	43.24795	15.7	0	0	1	1.04643	-21.671367	110	1	1.106602	0.9;
	6533	2	0	0	0	0	1	1.02852	-20.106955	220	1	1.115006	0.9;
	6534	1	48.05328	23	0	0	1	0.97477	-24.706521	110	1	1.1	0.9;
	6535	2	0	0	0	0	1	1.094263	36.299529	380	1	1.129517	0.9;
	6536	1	46.55161	15.5	0	0	1	0.908812	-31.131009	110	1	1.1	0.9;
	6537	2	0	0	0	0	1	1.080091	-6.180893	220	1	1.138591	0.9;
	6538	1	0	0	0	0	1	1.036157	-30.28049	110	1	1.1	0.9;
	6539	1	6.607326	1.6	0	0	1	0.994179	-16.827128	110	1	1.1	0.9;
	6540	2	0.700777	1.5	0	0	1	1.062947	-42.049394	380	1	1.128375	0.9;
	6541	2	0	0	0	0	1	1.092711	9.587763	380	1	1.126738	0.9;
	6542	1	19.22131	3.8	0	0	1	1.038222	-10.209517	110	1	1.1	0.9;
	6543	1	0	0	0	0	1	0.923128	-30.310974	110	1	1.1	0.9;
	6544	1	0	0	0	0	1	1.053254	-8.06051	220	1	1.119307	0.9;
	6545	1	89.69945	40.5	0	0	1	1.054514	-20.632637	220	1	1.1	0.9;
	6546	1	0	0.2	0	0	1	1.037175	-2.80582	220	1	1.107595	0.9;
	6547	1	17.71965	6.3	0	0	1	0.910729	-29.454934	110	1	1.1	0.9;
	6548	1	0	0	0	0	1	1.078039	-23.194243	220	1	1.167013	0.9;
	6549	1	0	0	0	0	1	1.002627	-23.356594	110	1	1.1	0.9;
	6550	1	0	0	0.191685	0	1	1.030528	-18.705006	220	1	1.1	0.9;
	6551	1	144	31.3	0	0	1	1.047565	-15.409626	220	1	1.12053	0.9;
	6552	2	3.1	6.8	0	0	1	1.103895	-12.562017	380	1	1.184848	0.9;
	6553	1	0	0	0	0	1	1.030312	-24.549974	150	1	1.1	0.9;
	6554	2	0	0	0	0	1	1.065416	9.498234	150	1	1.1	0.9;
	6555	1	0	0	0	0	1	1.044682	-11.480405	220	1	1.116954	0.9;
	6556	1	7.608436	2.5	0	0	1	0.988346	-27.616885	110	1	1.1	0.9;
	6557	1	0	0	0	-25.4492	1	1.041547	-24.708334	150	1	1.1	0.9;
	6558	1	388.7	97.7	0	0	1	1.044996	-39.746904	380	1	1.111893	0.9;
	6559	1	0	0	0	0	1	1.058036	33.159322	380	1	1.1	0.9;
	6560	1	0	0	0	0	1	0.982156	-39.812704	220	1	1.1	0.9;
	6561	1	0	0	0	23.0549	1	1.024611	-23.90432	150	1	1.1	0.9;
	6562	1	200.4	115.2	0	0	1	1.056532	-27.332784	220	1	1.1	0.9;
	6563	1	0	0	0	0	1	1.052224	56.873514	220	1	1.1	0.9;
	6564	1	11.21243	3.7	0	0	1	0.955782	-34.938343	110	1	1.1	0.9;
	6565	1	3.403774	1.2	0	0	1	1.030216	-23.049638	110	1	1.1	0.9;
	6566	1	0	0	0	0	1	0.969928	-28.747934	110	1	1.1	0.9;
	6567	1	0	0	0	0	1	0.998103	10.514609	220	1	1.1	0.9;
	6568	1	7.408214	2.7	0	0	1	1.018562	-21.540622	110	1	1.1	0.9;
	6569	1	0	0	0	0	1	1.04175	-21.833152	150	1	1.1	0.9;
	6570	1	233.2	39.4	0	0	1	1.077044	-30.674868	380	1	1.150367	0.9;
	6571	1	37.94207	9.8	0	0	1	1.018224	-38.267476	220	1	1.1	0.9;
	6572	2	20.0222	6.8	0	0	1	1.044455	-20.404642	110	1	1.1	0.9;
	6573	1	0	0	0	0	1	1.001996	-23.702615	150	1	1.1	0.9;
	6574	1	220.8	68.6	0	0	1	0.981264	-40.200668	220	1	1.1	0.9;
	6575	1	7.508325	3.1	0	0	1	1.090239	-2.826944	220	1	1.14914	0.9;
	6576	1	0	0	0	0	1	1.043756	-18.442847	220	1	1.1	0.9;
	6577	1	4.605106	1.5	0	0	1	0.972979	-35.84018	110	1	1.1	0.9;
	6578	2	422	107.1	0	0	1	1.018513	-37.690182	220	1	1.13185	0.9;
	6579	2	0	0	0	0	1	1.076553	-42.643792	380	1	1.132582	0.9;
	6580	1	10.61177	3.5	0	0	1	1.031908	-31.47972	110	1	1.1	0.9;
	6581	1	82.39135	0	0	0	1	1.07312	-7.620202	220	1	1.143385	0.9;
	6582	2	0	0	0	0	1	1.058317	16.617411	150	1	1.1	0.9;
	6583	1	96.7	23.7	0	0	1	1.026787	32.026056	380	1	1.1	0.9;
	6584	1	0	0	0	0	1	1.090517	-28.744771	380	1	1.15928	0.9;
	6585	1	0	0	0	0	1	1.049586	-10.978673	220	1	1.121383	0.9;
	6586	1	7.608436	3.8	0	0	1	1.024596	-16.266443	110	1	1.1	0.9;
	6587	1	0	0	0	0	1	1.070213	-5.54753	220	1	1.125738	0.9;
	6588	2	207.7	63	0	0	1	1.055947	-7.77533	380	1	1.123159	0.9;
	6589	1	406.3	92.5	0	0	1	1.062509	-17.116173	380	1	1.132632	0.9;
	6590	1	0	0	0	0	1	1.035313	-24.808034	150	1	1.1	0.9;
	6591	2	0	0	0	0	1	1.104158	23.683351	380	1	1.14069	0.9;
	6592	1	0	0	0.095843	0	1	1.02524	-20.413349	220	1	1.1	0.9;
	6593	1	18.62065	6.7	0	0	1	0.992107	-20.381648	110	1	1.1	0.9;
	6594	1	219.6	36.4	0	0	1	1.063174	-23.339564	220	1	1.162862	0.9;
	6595	2	55.26127	16.5	0	0	1	1.079255	-4.446303	220	1	1.134024	0.9;
	6596	1	79.38802	31.3	0	0	1	1.077889	3.477923	150	1	1.146058	0.9;
	6597	1	0	0	0	0	1	1.051548	-19.561747	220	1	1.1	0.9;
	6598	2	0	0	0	0	1	1.087318	-12.011419	220	1	1.138865	0.9;
	6599	2	715.8	24.1	0	0	1	1.075613	-12.897724	150	1	1.129951	0.9;
	6600	1	7.408214	2.7	0	0	1	1.005859	-18.787501	110	1	1.1	0.9;
	6601	2	124	24.4	0	0	1	1.102974	-5.117309	380	1	1.150093	0.9;
	6602	1	101.1	6.6	0	0	1	1.033298	-41.751036	220	1	1.1	0.9;
	6603	1	0	0	0	0	1	1.141889	-1.859861	220	1	1.20454	0.9;
	6604	1	0	0	0	0	1	1.092623	-24.681524	380	1	1.154765	0.9;
	6605	1	11.91321	4.4	0	0	1	1.00158	-18.110757	110	1	1.1	0.9;
	6606	1	0	0	0	0	1	1.088312	-36.997976	380	1	1.132241	0.9;
	6607	1	0	0	0	0	1	1.099068	-5.804519	220	1	1.165936	0.9;
	6608	1	0	0	0	-16.3585	1	0.979515	-25.47388	150	1	1.1	0.9;
	6609	1	0	0	0	0	1	1.08295	-10.345392	380	1	1.140601	0.9;
	6610	1	11.71299	3.8	0	0	1	1.001078	-29.590634	110	1	1.1	0.9;
	6611	2	114.7	4.1	0	0	1	1.086591	-7.929386	220	1	1.131118	0.9;
	6612	1	2.702997	0.8	0	0	1	0.988153	-23.419976	110	1	1.1	0.9;
	6613	1	0	0	0	0	1	1.080717	2.034615	150	1	1.151859	0.9;
	6614	1	36.84085	13.3	0	0	1	1.032678	-12.831957	110	1	1.1	0.9;
	6615	1	0	0	0	0	1	1.080706	2.034053	150	1	1.151846	0.9;
	6616	2	217.1	22.6	0	0	1	1.05427	11.051321	220	1	1.14231	0.9;
	6617	1	0	0	0	0	1	0.948733	55.084968	220	1	1.1	0.9;
	6618	1	0	0	0	0	1	1.029351	-23.363762	150	1	1.1	0.9;
	6619	1	36.03996	12	0	0	1	0.95471	-27.552372	110	1	1.1	0.9;
	6620	2	2.502775	94.6	0	0	1	1.086105	-40.254533	380	1	1.129853	0.9;
	6621	1	0.400444	0.1	0	0	1	1.00717	-23.107833	110	1	1.1	0.9;
	6622	1	82.89191	32.6	0	0	1	1.082306	2.982109	150	1	1.152807	0.9;
	6623	1	0	0	0	0	1	1.0437	-30.204767	220	1	1.106214	0.9;
	6624	1	0	0	0	0	1	0.977764	-28.563389	150	1	1.1	0.9;
	6625	2	0	0	0	0	1	0.953641	-18.634138	110	1	1.1	0.9;
	6626	1	0	0	0	0	1	1.095931	-11.242293	220	1	1.168605	0.9;
	6627	1	256	69.9	0	0	1	1.085342	-42.36031	380	1	1.1321	0.9;
	6628	1	102.6	12	0	0	1	1.037915	-45.860113	380	1	1.105461	0.9;
	6629	1	56.66282	15.6	0	0	1	1.0256	-43.925456	220	1	1.1	0.9;
	6630	1	0	0	0	0	1	1.097575	-18.631623	380	1	1.178442	0.9;
	6631	1	0	0	0.13851	0	1	1.072885	-20.143996	380	1	1.156933	0.9;
	6632	1	29.53274	2.4	0	0	1	1.022345	-31.503088	110	1	1.1	0.9;
	6633	1	0	0	0	0	1	1.071041	-6.852285	220	1	1.130845	0.9;
	6634	1	210.9	48.3	0	0	1	1.095109	-3.498908	220	1	1.152644	0.9;
	6635	1	0	0	0	0	1	1.04675	20.566718	380	1	1.105515	0.9;
	6636	1	0	0	0	0	1	1.010766	-19.540915	150	1	1.1	0.9;
	6637	1	0	0	0	0	1	1.061762	-35.285625	220	1	1.142821	0.9;
	6638	1	201.4	47.5	0	0	1	0.965703	21.580102	220	1	1.1	0.9;
	6639	1	36.94096	62.6	0	0	1	0.998712	-23.907111	150	1	1.1	0.9;
	6640	1	18.2202	6.3	0	0	1	0.972428	-25.603048	110	1	1.1	0.9;
	6641	1	0	0	0	0	1	1.033389	-4.161221	220	1	1.1	0.9;
	6642	1	0	0	0	0	1	1.048392	-12.371918	220	1	1.133736	0.9;
	6643	2	34.93874	8.6	0	0	1	1.059227	-16.884821	220	1	1.112397	0.9;
	6644	2	0	0	0	0	1	1.019289	-13.828213	380	1	1.124702	0.9;
	6645	2	0	0	0	0	1	1.054727	-30.041553	220	1	1.109874	0.9;
	6646	2	0	0	0	0	1	1.102079	-14.470544	380	1	1.183119	0.9;
	6647	2	0	0	0	0	1	1.079273	-33.619526	110	1	1.125763	0.9;
	6648	1	0	0	0	0	1	1.045036	-0.982761	220	1	1.112846	0.9;
	6649	1	36.54051	22.6	0	0	1	1.017132	-23.760304	150	1	1.1	0.9;
	6650	1	26.4293	9.9	0	0	1	1.011492	-35.010805	110	1	1.1	0.9;
	6651	1	25.12786	14	0	0	1	1.028194	-4.424981	220	1	1.1	0.9;
	6652	1	45.75072	15.8	0	0	1	0.994253	-13.583593	110	1	1.1	0.9;
	6653	1	0	0	0	0	1	1.071805	-13.815613	220	1	1.135486	0.9;
	6654	1	0	0	0	0	1	1.024384	-23.126331	150	1	1.1	0.9;
	6655	1	23.82642	8.6	0	0	1	1.038186	-22.490885	110	1	1.132224	0.9;
	6656	1	37.44151	0	0	0	1	1.085702	-3.360929	220	1	1.146983	0.9;
	6657	1	169	70.8	0	0	1	1.044474	-14.960099	220	1	1.107729	0.9;
	6658	1	0	0	0	0	1	1.085127	-39.155943	380	1	1.143572	0.9;
	6659	1	31.03441	8.9	0	0	1	0.995818	-26.374075	110	1	1.1	0.9;
	6660	1	0	0	0	0	1	0.975272	-27.839009	110	1	1.1	0.9;
	6661	1	0	0	0	0	1	1.068527	15.370749	380	1	1.137544	0.9;
	6662	1	101.3	17.6	0	0	1	1.039294	-44.348121	220	1	1.1	0.9;
	6663	1	16.51831	6.3	0	0	1	1.095825	-12.573332	110	1	1.161158	0.9;
	6664	2	77.28568	13.8	0	0	1	1.035572	-47.00775	220	1	1.1	0.9;
	6665	1	61.66837	16.3	0	0	1	1.081568	-3.828843	220	1	1.142596	0.9;
	6666	1	0	0	0	0	1	1.06286	1.49577	380	1	1.11937	0.9;
	6667	1	127.3	26.7	0	0	1	1.018185	-21.166702	220	1	1.1	0.9;
	6668	1	28.53163	11.5	0	0	1	0.968367	-25.37697	110	1	1.1	0.9;
	6669	1	12.7141	4.2	0	0	1	0.97937	-25.729316	110	1	1.1	0.9;
	6670	1	0	0	0	0	1	1.051209	-22.829461	220	1	1.1	0.9;
	6671	1	-40.9454	-28.4	0	0	1	1.033941	-44.083559	150	1	1.1	0.9;
	6672	2	0	0	0	0	1	1.092395	-24.093032	380	1	1.14898	0.9;
	6673	1	88.19778	21	0	0	1	1.013739	14.475313	220	1	1.1	0.9;
	6674	1	0	0	0	0	1	1.095176	-14.771049	220	1	1.162555	0.9;
	6675	1	223.3053	61.3	0	0	1	1.077374	-2.423541	220	1	1.140665	0.9;
	6676	1	0	0	0.261498	0	1	1.082013	-19.140241	380	1	1.158174	0.9;
	6677	1	6.306993	1.3	0	0	1	0.955502	-23.249911	110	1	1.1	0.9;
	6678	1	70.47814	22.5	0	0	1	1.052581	-36.516164	220	1	1.107236	0.9;
	6679	1	1.501665	0.4	0	0	1	1.055447	-17.747716	220	1	1.113054	0.9;
	6680	1	0	0	0	0	1	1.03908	-24.339985	150	1	1.1	0.9;
	6681	1	0	0	0	0	1	1.039473	15.732709	380	1	1.105269	0.9;
	6682	1	10.0111	4.3	0	0	1	1.022106	-15.973324	110	1	1.1	0.9;
	6683	1	0	0	0	0	1	1.073882	-22.728551	220	1	1.160704	0.9;
	6684	2	0	0	0	0	1	1.063236	-1.729047	150	1	1.139134	0.9;
	6685	1	53.05883	19.2	0	0	1	1.053018	-20.467315	110	1	1.111498	0.9;
	6686	2	0	0	0	0	1	1.020818	4.408546	220	1	1.1	0.9;
	6687	1	0	0	0	0	1	1.059317	-15.205138	220	1	1.128255	0.9;
	6688	2	132.7	11.7	0	0	1	1.049818	-13.948205	380	1	1.117585	0.9;
	6689	1	0	0	0	0	1	0.958951	-25.600778	150	1	1.1	0.9;
	6690	1	0	0	0	0	1	0.976904	-22.764718	150	1	1.1	0.9;
	6691	2	0	0	0	0	1	1.017773	-31.330217	220	1	1.1	0.9;
	6692	1	6.106771	2.2	0	0	1	0.978544	-18.146287	110	1	1.1	0.9;
	6693	1	1.101221	0	0	0	1	1.008916	-36.064813	110	1	1.1	0.9;
	6694	1	0	0	0	0	1	0.997572	-24.754594	150	1	1.1	0.9;
	6695	1	0	0	0	0	1	1.041091	25.762276	380	1	1.131161	0.9;
	6696	1	0	0	0	0	1	1.03314	-25.180792	150	1	1.1	0.9;
	6697	1	16.61843	5.5	0	0	1	1.079262	-31.346731	110	1	1.147554	0.9;
	6698	1	-23.8264	7.8	0	0	1	1.055215	-14.522849	150	1	1.106484	0.9;
	6699	1	142	27.8	0	0	1	1.050667	-49.755104	220	1	1.1	0.9;
	6700	2	41.04551	-46	0	0	1	1.067636	-21.597753	220	1	1.1	0.9;
	6701	1	55.66171	22.4	0	0	1	1.057297	6.505594	110	1	1.125142	0.9;
	6702	1	0	0	0	0	1	1.071008	-40.022594	380	1	1.131323	0.9;
	6703	1	0	0	0	0	1	1.050606	-18.311935	220	1	1.116245	0.9;
	6704	1	41.44595	17.3	0	0	1	0.951068	-23.734961	110	1	1.1	0.9;
	6705	1	100.3	17.2	0	0	1	1.031545	-46.259071	220	1	1.1	0.9;
	6706	1	168.9	113	0	0	1	1.033732	17.127387	380	1	1.1	0.9;
	6707	1	30.0333	18.6	0	0	1	1.008807	-20.023367	150	1	1.1	0.9;
	6708	2	0	0	0	0	1	1.014955	11.147918	220	1	1.1	0.9;
	6709	1	0	0	0	0	1	0.9751	-23.973836	150	1	1.1	0.9;
	6710	1	59.66615	6	0	0	1	1.016245	-47.76275	220	1	1.1	0.9;
	6711	1	19.02109	5.4	0	0	1	1.069676	8.609245	110	1	1.139241	0.9;
	6712	1	0	-0.8	0	0	1	1.046511	-21.588091	110	1	1.1	0.9;
	6713	1	94.50478	-27.8	0	0	1	1.062462	-43.252427	220	1	1.114232	0.9;
	6714	2	0	0	0	0	1	1.072349	-2.440198	150	1	1.142345	0.9;
	6715	1	0	0	0	0	1	0.94171	-32.292003	150	1	1.1	0.838677;
	6716	1	107.2	45.6	0	0	1	1.037437	-16.879842	220	1	1.1	0.9;
	6717	1	0	0	0	0	1	1.019821	-34.891149	150	1	1.1	0.9;
	6718	1	0	0	0	0	1	1.028354	65.17963	220	1	1.1	0.9;
	6719	1	33.63729	13.7	0	0	1	1.052953	-48.975647	220	1	1.110392	0.9;
	6720	1	23.42597	8.5	0	0	1	0.982671	-24.385099	110	1	1.1	0.9;
	6721	1	16.21798	5	0	0	1	1.004182	-21.992654	110	1	1.1	0.9;
	6722	1	0	0	0.155865	0	1	1.078863	67.176565	380	1	1.1	0.9;
	6723	1	106.4	19	0	0	1	1.038844	-41.687266	220	1	1.1	0.9;
	6724	1	0	0	0.090703	0	1	1.075504	-15.951953	380	1	1.143398	0.9;
	6725	1	18.01998	12.2	0	0	1	0.962369	-21.031232	110	1	1.1	0.9;
	6726	1	415.5	103.4	0	0	1	1.056364	-32.545422	380	1	1.119416	0.9;
	6727	1	8.108991	2.7	0	0	1	1.057005	-29.620525	110	1	1.12173	0.9;
	6728	1	65.67281	25.6	0	0	1	1.074467	-10.053005	220	1	1.144877	0.9;
	6729	1	0	0	0	0	1	1.061338	-36.461977	220	1	1.120516	0.9;
	6730	1	123.1	20.9	0	0	1	1.054192	-11.134229	380	1	1.122281	0.9;
	6731	2	0	0	0	0	1	1.087526	0.753832	150	1	1.154348	0.9;
	6732	1	1.801998	0.6	0	0	1	1.026235	-14.935499	110	1	1.1	0.9;
	6733	2	0	0	0	0	1	1.0855	-15.40087	380	1	1.123946	0.9;
	6734	1	0	0	0	-130.727	1	1.08545	-45.541927	380	1	1.149073	0.9;
	6735	1	60.36693	54.7	0	0	1	0.988159	10.739767	220	1	1.1	0.9;
	6736	2	0	0	0	-87.5013	1	1.039182	-29.715625	220	1	1.1	0.9;
	6737	2	0	0	0	0	1	1.041909	-17.2289	220	1	1.112643	0.9;
	6738	1	116.8	59	0	0	1	1.023853	-27.475478	150	1	1.1	0.9;
	6739	1	0	0	0	0	1	0.988409	-24.210444	150	1	1.1	0.9;
	6740	1	0	0	0	0	1	1.08842	-44.931621	380	1	1.130626	0.9;
	6741	1	18.32031	5.3	0	0	1	1.054837	-24.707017	220	1	1.160254	0.9;
	6742	1	121.3	20.7	0	0	1	1.045106	-12.69957	220	1	1.117351	0.9;
	6743	1	12.7141	2.5	0	0	1	0.946366	-27.476428	110	1	1.1	0.9;
	6744	1	0	0	0	0	1	1.086583	32.712269	380	1	1.119226	0.9;
	6745	2	15.21687	5.1	0	0	1	1.027532	-16.203875	110	1	1.1	0.9;
	6746	1	0	0	0	0	1	1.041554	14.683138	150	1	1.1	0.9;
	6747	1	0	0	0	0	1	1.073057	15.563948	380	1	1.129289	0.9;
	6748	1	111.9	118.3	0	0	1	1.023803	12.111666	380	1	1.1	0.9;
	6749	1	0	0	0	0	1	1.05189	60.720146	220	1	1.105759	0.9;
	6750	2	0	0	0	0	1	1.043773	-41.705027	220	1	1.1	0.9;
	6751	1	0	0	0	0	1	1.101488	-13.991306	380	1	1.147767	0.9;
	6752	2	16.01776	3.9	0	0	1	1.009791	61.711396	150	1	1.1	0.9;
	6753	1	108.8	28.1	0	0	1	1.062562	9.501026	150	1	1.1	0.9;
	6754	1	227.4892	61.6	0	0	1	0.998059	10.359685	220	1	1.1	0.9;
	6755	1	49.45483	16.5	0	0	1	1.035502	-35.143982	110	1	1.1	0.9;
	6756	1	0	0	0	0	1	0.96045	-25.421932	150	1	1.1	0.9;
	6757	1	0	0	0	0	1	1.080204	-6.871882	220	1	1.136243	0.9;
	6758	1	0	0	0.147237	0	1	1.105109	-9.675955	380	1	1.162966	0.9;
	6759	1	9.00999	3.3	0	0	1	1.047266	-14.289962	110	1	1.141434	0.9;
	6760	1	96.2	23.3	0	0	1	1.033111	-2.093061	220	1	1.110515	0.9;
	6761	1	0	0	0	0	1	1.096501	-20.015151	380	1	1.143695	0.9;
	6762	1	-7.40821	-1.6	0	0	1	1.075527	-18.99976	220	1	1.124953	0.9;
	6763	1	0	0	0	0	1	1.078761	-41.834953	380	1	1.120116	0.9;
	6764	1	10.11121	3.7	0	0	1	0.998989	-18.953416	110	1	1.1	0.9;
	6765	1	0	0	0	0	1	1.053578	-16.603797	150	1	1.109073	0.9;
	6766	1	178.7	31.3	0	0	1	1.05506	-48.889867	220	1	1.112628	0.9;
	6767	1	0	0	0	0	1	1.072679	-11.702141	220	1	1.132448	0.9;
	6768	2	45.45039	7.4	0	0	1	1.0432	61.671732	150	1	1.1	0.9;
	6769	2	244	30.2	0	0	1	1.030409	-33.128913	220	1	1.163534	0.9;
	6770	2	0	0	0	0	1	1.068287	-21.359046	220	1	1.122097	0.9;
	6771	1	15.71743	5.2	0	0	1	0.943018	-28.897304	110	1	1.1	0.9;
	6772	1	1.501665	0.4	0	0	1	0.982044	-17.832275	110	1	1.1	0.9;
	6773	1	13.71521	8.5	0	0	1	1.015513	-26.374751	150	1	1.178355	0.9;
	6774	2	0	0	0	0	1	1.049596	-5.649828	380	1	1.134518	0.9;
	6775	1	0	0	0	0	1	1.07549	-15.487219	380	1	1.121502	0.9;
	6776	2	253.9	94.2	0	0	1	1.071462	-14.443967	220	1	1.127527	0.9;
	6777	2	0	0	0	0	1	1.022636	-13.585203	220	1	1.116471	0.9;
	6778	1	84.79401	12.5	0	0	1	1.050213	-49.44782	220	1	1.111465	0.9;
	6779	1	59.66615	27.1	0	0	1	1.073727	4.551247	150	1	1.144306	0.9;
	6780	1	106.8	22.2	0	0	1	1.008887	-21.685173	110	1	1.1	0.9;
	6781	1	47.95317	29.7	0	0	1	0.965656	-27.541312	150	1	1.1	0.9;
	6782	1	0	0	0	0	1	1.010652	-48.380312	220	1	1.1	0.9;
	6783	2	0	0	0	0	1	1.048135	15.415947	150	1	1.1	0.9;
	6784	1	0	0	0	0	1	1.030677	-24.28051	150	1	1.1	0.9;
	6785	1	-1316.59	-129.9	0.205959	0	1	1.09072	55.461606	380	1	1.118195	0.9;
	6786	1	0.400444	0.1	0	0	1	1.023354	-20.876203	110	1	1.117431	0.9;
	6787	1	44.34917	10.8	0	0	1	1.05318	-24.661371	220	1	1.128295	0.9;
	6788	1	4.404884	2.7	0	0	1	0.953366	-23.598083	110	1	1.1	0.9;
	6789	1	38.04218	18	0	0	1	1.056247	-27.116514	220	1	1.11871	0.9;
	6790	1	0.200222	0.7	0	0	1	1.024498	15.617773	220	1	1.1	0.9;
	6791	1	70.0777	35	0	0	1	1.039456	-45.895051	220	1	1.1	0.9;
	6792	2	198.6	282.2	0	0	1	1.085289	46.029963	380	1	1.1	0.9;
	6793	1	0	0	0	0	1	1.068929	-20.202165	220	1	1.15129	0.9;
	6794	1	0	0	0	0	1	1.078699	18.806409	380	1	1.136331	0.9;
	6795	1	0	0	0	0	1	1.067575	4.616673	220	1	1.143423	0.9;
	6796	2	0	0	0	0	1	1.057061	-36.152298	220	1	1.137772	0.9;
	6797	2	113.4	19.7	0	0	1	1.061091	-14.099184	220	1	1.124289	0.9;
	6798	1	10.11121	3.7	0	0	1	1.017843	-18.607511	110	1	1.1	0.9;
	6799	1	4.404884	2.2	0	0	1	1.029282	-35.178392	110	1	1.1	0.9;
	6800	2	100.6	49.6	0	0	1	0.986636	-12.446407	110	1	1.1	0.9;
	6801	1	136.2	4.8	0	0	1	1.078257	-13.263757	220	1	1.133872	0.9;
	6802	1	39.84418	24.6	0	0	1	1.001454	-26.883551	150	1	1.1	0.9;
	6803	1	128.9	73.7	0	0	1	1.065018	20.6395	380	1	1.13505	0.9;
	6804	1	0	0	0	0	1	1.021659	-38.912603	220	1	1.1	0.9;
	6805	2	0	0	0	0	1	1.055091	-38.833387	220	1	1.1	0.9;
	6806	1	0	0	0	0	1	1.07103	0.755545	380	1	1.139585	0.9;
	6807	1	0	0	0.148479	0	1	1.081508	-18.716585	380	1	1.144888	0.9;
	6808	1	4.00444	0.8	0	0	1	1.059594	-22.956161	220	1	1.133542	0.9;
	6809	2	2.8	264.7	0	0	1	1.085053	-35.935996	380	1	1.135421	0.9;
	6810	1	87.99757	36.2	0	0	1	1.093541	-43.286912	380	1	1.150611	0.9;
	6811	1	46.35139	14.9	0	0	1	1.071949	-42.366013	220	1	1.120313	0.9;
	6812	1	0.100111	0	0	0	1	1.052437	1.364967	220	1	1.126885	0.9;
	6813	1	0	0	0	0	1	1.059552	4.040351	380	1	1.117089	0.9;
	6814	1	0	0	0	0	1	1.057586	-14.3481	380	1	1.119469	0.9;
	6815	1	1.4	100	0	0	1	1.031217	25.2997	220	1	1.1	0.9;
	6816	1	0	0	0.095843	0	1	1.078295	-12.756161	220	1	1.133481	0.9;
	6817	1	3.804218	0.6	0	0	1	0.973483	-17.962011	110	1	1.1	0.9;
	6818	1	9.00999	4	0	0	1	1.017082	-19.098009	110	1	1.1	0.9;
	6819	2	0	0	0	0	1	0.966052	-17.89012	110	1	1.1	0.9;
	6820	2	120.9	41.4	0	0	1	1.077045	-10.046571	220	1	1.147372	0.9;
	6821	2	0	0	0	0	1	1.032711	-8.844462	380	1	1.114922	0.9;
	6822	1	0	0	0.148	0	1	0.982769	-25.079209	220	1	1.1	0.9;
	6823	2	0	0	0	0	1	1.040974	-10.806154	380	1	1.107204	0.9;
	6824	1	0	0	0	0	1	1.090648	-0.669768	220	1	1.154767	0.9;
	6825	1	0	0	0	0	1	1.026626	-7.590834	220	1	1.1	0.9;
	6826	2	0	0	0	0	1	1.050629	15.255975	150	1	1.1	0.9;
	6827	1	0	0	0	0	1	1.039581	6.347264	380	1	1.1	0.9;
	6828	1	211	59.3	0	0	1	0.988826	-22.928882	110	1	1.1	0.9;
	6829	1	0	0	0	0	1	1.051401	-22.940212	220	1	1.1	0.9;
	6830	1	0	0	0	0	1	1.059893	-18.990432	380	1	1.149872	0.9;
	6831	1	26.22908	8.4	0	0	1	0.966059	-30.910296	110	1	1.1	0.9;
	6832	1	13.81532	5	0	0	1	1.094221	-17.138301	220	1	1.204637	0.9;
	6833	1	43.14784	15.4	0	0	1	1.03296	-15.379405	110	1	1.106119	0.9;
	6834	1	20.62287	6.7	0	0	1	0.964941	-26.462817	110	1	1.1	0.9;
	6835	1	0	0	0	0	1	0.984088	-26.370497	150	1	1.1	0.9;
	6836	1	21.8242	6.3	0	0	1	1.004211	-24.167276	110	1	1.1	0.9;
	6837	1	0	0	0	0	1	1.036214	-14.944421	380	1	1.107109	0.9;
	6838	1	130.5	17.2	0	0	1	1.016542	-47.353779	220	1	1.1	0.9;
	6839	1	40.84529	15.7	0	0	1	1.052202	-12.044753	220	1	1.134253	0.9;
	6840	1	0	0	0	0	1	1.013032	-47.569974	220	1	1.1	0.9;
	6841	2	355.5	92.9	0	0	1	1.065664	-15.745233	380	1	1.154052	0.9;
	6842	1	0	0	0	0	1	1.027168	-20.593467	150	1	1.106203	0.9;
	6843	1	72.48036	44.9	0	0	1	1.023321	-27.887179	150	1	1.1	0.9;
	6844	2	43.64839	10.8	0	0	1	0.957659	66.516623	150	1	1.1	0.9;
	6845	1	0	0	0.265633	487.2386	1	1.099323	35.801706	380	1	1.129739	0.9;
	6846	1	0	0	0	0	1	0.956949	-29.228607	150	1	1.1	0.892108;
	6847	1	0	0	0	0	1	1.077938	-16.582514	380	1	1.132826	0.9;
	6848	1	281.3	40.9	0	0	1	1.048775	-23.580018	220	1	1.119481	0.9;
	6849	2	26.82975	11.9	0	0	1	1.007689	-15.538118	110	1	1.1	0.9;
	6850	1	0	0	0	0	1	1.03333	-34.760191	220	1	1.1	0.9;
	6851	2	227.4	73.1	0	0	1	1.044003	-3.613721	220	1	1.114806	0.9;
	6852	1	0	0	0	0	1	1.039703	-25.070337	150	1	1.1	0.9;
	6853	1	0	0	0	0	1	1.066153	10.834958	380	1	1.146568	0.9;
	6854	1	2.903219	1	0	0	1	1.03743	-20.556045	110	1	1.1	0.9;
	6855	1	2.102331	0.8	0	0	1	1.003575	-22.097961	110	1	1.1	0.9;
	6856	1	0	0	0	0	1	0.991828	-17.297411	220	1	1.107223	0.9;
	6857	1	34.43818	12.1	0	0	1	0.999227	-18.848247	110	1	1.1	0.9;
	6858	2	9.00999	-2.7	0	0	1	1.06807	62.143387	150	1	1.1	0.9;
	6859	1	34.6384	12.4	0	0	1	1.012804	-16.862609	110	1	1.111103	0.9;
	6860	1	0	0	0	0	1	1.034479	-5.111942	380	1	1.1	0.9;
	6861	1	218	18.6	0	0	1	1.07427	-12.803559	380	1	1.139181	0.9;
	6862	1	531.7	-12.8	0	0	1	1.077379	-12.869674	150	1	1.140258	0.9;
	6863	1	0	0	0	0	1	1.013067	-46.705296	220	1	1.1	0.9;
	6864	1	0	0	0	0	1	1.077535	9.509833	110	1	1.146448	0.9;
	6865	1	0	0	0	0	1	0.973529	-25.885712	150	1	1.1	0.9;
	6866	2	0	0	0	0	1	1.0385	-26.275309	220	1	1.1	0.9;
	6867	1	0	0	0	0	1	1.007611	-50.995608	150	1	1.1	0.9;
	6868	1	0	0	0	0	1	1.080578	-13.784328	220	1	1.152687	0.9;
	6869	1	0	0	0	0	1	1.089567	-10.058833	380	1	1.140194	0.9;
	6870	1	126.6	28.7	0	0	1	1.053218	-49.060948	220	1	1.110488	0.9;
	6871	1	0	0	0	0	1	1.062888	-35.299027	220	1	1.123527	0.9;
	6872	1	29.03219	6.1	0	0	1	1.063459	7.277026	110	1	1.131761	0.9;
	6873	1	0	0	0	0	1	1.042404	-12.134897	220	1	1.113908	0.9;
	6874	1	0	0	0	0	1	1.054699	-10.692241	220	1	1.124169	0.9;
	6875	1	45.95095	15.4	0	0	1	1.002746	-23.619608	110	1	1.1	0.9;
	6876	1	136.9	11.4	0	0	1	1.041024	-25.7205	220	1	1.137994	0.9;
	6877	1	0	0	0	0	1	1.057062	-11.614007	220	1	1.136457	0.9;
	6878	1	0	0	0	0	1	1.067991	-23.157733	220	1	1.15293	0.9;
	6879	1	5.00555	1.8	0	0	1	1.078592	-0.794236	220	1	1.146587	0.9;
	6880	1	127.6	45.3	0	0	1	1.069395	-1.799866	150	1	1.146635	0.9;
	6881	1	27.93097	8.9	0	0	1	0.958534	-22.956879	110	1	1.1	0.9;
	6882	2	0	0	0	0	1	1.06958	-2.768638	150	1	1.13953	0.9;
	6883	1	0	0	0	0	1	1.029279	-25.709268	150	1	1.1	0.9;
	6884	1	10.11121	4.4	0	0	1	0.973751	-20.358038	110	1	1.1	0.9;
	6885	1	9.710767	3.2	0	0	1	0.932012	-34.988721	110	1	1.1	0.9;
	6886	1	0	0	0	0	1	0.960193	-26.464873	110	1	1.1	0.9;
	6887	1	11.21243	3.7	0	0	1	1.002893	-29.83287	110	1	1.1	0.9;
	6888	1	159.4	-7.3	0	0	1	1.08813	2.747973	220	1	1.150167	0.9;
	6889	1	58.76516	16.1	0	0	1	1.04379	-41.066667	220	1	1.1	0.9;
	6890	1	116.5	33.4	0	0	1	1.046618	-4.828803	110	1	1.122851	0.9;
	6891	1	30.43374	10	0	0	1	0.953616	-27.63711	110	1	1.1	0.9;
	6892	1	59.06549	23.4	0	0	1	1.063911	-1.637652	150	1	1.139718	0.9;
	6893	1	18.01998	4	0	0	1	1.086952	-26.306588	220	1	1.149219	0.9;
	6894	1	0.500555	0	0	0	1	0.935134	-28.732453	110	1	1.1	0.9;
	6895	1	0	0	0	0	1	1.074055	-6.618266	220	1	1.140143	0.9;
	6896	1	119	31.7	0	0	1	1.034298	-41.81864	220	1	1.1	0.9;
	6897	1	0	0	0	0	1	1.001299	60.196423	220	1	1.1	0.9;
	6898	1	119.7	43.4	0	0	1	1.064548	-18.593507	220	1	1.1	0.9;
	6899	1	5.506105	2.2	0	0	1	0.989276	-37.326531	110	1	1.1	0.9;
	6900	1	86.99646	30.8	0	0	1	0.949883	-27.129655	110	1	1.1	0.9;
	6901	1	0	0	0	12.05509	1	1.026477	-23.796331	150	1	1.1	0.9;
	6902	1	2.803108	0.5	0	0	1	0.995579	-26.448436	110	1	1.1	0.9;
	6903	2	13.11454	2.8	0	0	1	1.062273	-12.965436	110	1	1.114326	0.9;
	6904	1	8.309213	3	0	0	1	0.963533	-16.53255	110	1	1.1	0.9;
	6905	1	24.02664	-56.8	0	0	1	1.048571	-1.02191	220	1	1.116768	0.9;
	6906	1	26.82131	24.2	0	0	1	0.977764	-28.563389	150	1	1.1	0.9;
	6907	1	115.4	31	0	0	1	1.072007	-42.438226	220	1	1.118674	0.9;
	6908	1	-20.1223	6.6	0	0	1	1.053783	-16.107268	150	1	1.109229	0.9;
	6909	1	6.306993	2.5	0	0	1	0.916307	-29.64919	110	1	1.1	0.9;
	6910	1	70.47814	17.6	0	0	1	1.018766	-48.702548	220	1	1.1	0.9;
	6911	2	0	0	0	0	1	1.097184	-14.664789	380	1	1.157636	0.9;
	6912	1	0	0	0.242637	0	1	1.07295	-15.214222	380	1	1.124431	0.9;
	6913	1	0	0	0	0	1	1.055557	-27.181328	220	1	1.118034	0.9;
	6914	2	0	0	0	0	1	1.067895	-34.7104	380	1	1.130031	0.9;
	6915	1	136.9518	87.7	0	0	1	1.011068	11.513101	220	1	1.1	0.9;
	6916	1	0	0	0	0	1	1.016222	-26.56258	150	1	1.1	0.9;
	6917	1	0	0	0	0	1	1.083363	-2.640205	220	1	1.145037	0.9;
	6918	1	32.7363	10.5	0	0	1	1.072265	-19.68828	220	1	1.154204	0.9;
	6919	2	312.1	159.9	0	0	1	1.077468	5.784078	380	1	1.151442	0.9;
	6920	1	78.88747	45.6	0	0	1	1.038611	-12.58932	220	1	1.108438	0.9;
	6921	2	27.43041	15.7	0	0	1	1.016591	-19.786578	220	1	1.1	0.9;
	6922	2	0	0	0	0	1	1.098999	4.884085	150	1	1.151639	0.9;
	6923	1	0	0	0	0	1	1.049085	-50.113908	220	1	1.1	0.9;
	6924	2	10.31143	3.8	0	0	1	1.014045	-40.61911	220	1	1.1	0.9;
	6925	1	0	0	0	0	1	1.044602	39.775528	220	1	1.1	0.9;
	6926	1	37.94207	3.4	0	0	1	1.067443	-51.189386	220	1	1.1	0.9;
	6927	1	91.60156	20.3	0	0	1	1.064072	-34.385493	220	1	1.143791	0.9;
	6928	1	0	0	0	0	1	1.091218	-7.346361	220	1	1.15938	0.9;
	6929	1	-12.614	11.6	0	0	1	1.040065	-18.730751	150	1	1.1	0.9;
	6930	1	0	0	0.294	0	1	1.062174	2.482756	380	1	1.117542	0.9;
	6931	1	429.5	163.1	0	0	1	1.064726	-22.358935	220	1	1.110749	0.9;
	6932	2	110.1	44.6	0	0	1	1.044773	-34.177538	220	1	1.107245	0.9;
	6933	1	68.77625	35	0	0	1	1.080309	-5.811728	380	1	1.144099	0.9;
	6934	2	0	0	0	0	1	1.0195	-5.813566	220	1	1.119839	0.9;
	6935	1	90.30012	5.7	0	0	1	1.079473	36.437439	380	1	1.123538	0.9;
	6936	1	124.1	-30.3	0	0	1	1.036142	-41.944096	220	1	1.1	0.9;
	6937	1	0	0	0	0	1	1.092782	-10.565668	380	1	1.136178	0.9;
	6938	1	0	0	0	0	1	1.078196	-19.440076	220	1	1.181609	0.9;
	6939	1	15.11676	5.4	0	0	1	1.000261	-20.153353	110	1	1.1	0.9;
	6940	1	29.1323	11.3	0	0	1	0.963672	-13.148621	110	1	1.1	0.9;
	6941	1	439	254.8	0	0	1	1.081393	46.114388	380	1	1.1	0.9;
	6942	1	61.96871	38.4	0	0	1	1.03994	-21.302399	150	1	1.116385	0.9;
	6943	1	255.3	78.7	0	0	1	0.983962	7.860721	220	1	1.1	0.9;
	6944	1	33.93763	11.4	0	0	1	1.060853	-21.749719	220	1	1.137984	0.9;
	6945	1	117	15.2	0	0	1	1.013352	9.907876	220	1	1.1	0.9;
	6946	2	0.2	0.2	0	0	1	1.027864	-30.996231	220	1	1.1	0.9;
	6947	2	0	0	0	0	1	1.078921	-15.237345	380	1	1.148057	0.9;
	6948	2	142.3	19.2	0	0	1	1.031091	-36.660872	220	1	1.1	0.9;
	6949	1	0	0	0.095843	0	1	1.057785	-10.619402	220	1	1.121115	0.9;
	6950	2	0	0	0	0	1	1.054727	-30.048028	220	1	1.10984	0.9;
	6951	2	0	0	0	0	1	1.066194	7.73301	150	1	1.115258	0.9;
	6952	1	37.14118	2.7	0	0	1	1.051336	-12.090049	220	1	1.120823	0.9;
	6953	1	256.8	-61.8	0	0	1	1.085574	-21.094946	380	1	1.175347	0.9;
	6954	1	16.51831	5.5	0	0	1	1.036651	-36.635298	110	1	1.1	0.9;
	6955	1	93.4	27.8	0	0	1	1.063578	9.234591	150	1	1.1	0.9;
	6956	2	44.04884	17.5	0	0	1	1.067664	-0.692773	220	1	1.139058	0.9;
	6957	1	67.07436	25.2	0	0	1	1.064554	-44.100246	220	1	1.113212	0.9;
	6958	2	63.06993	7.1	0	0	1	0.940267	67.623133	150	1	1.1	0.9;
	6959	2	0	0	0	0	1	1.094227	-8.289764	220	1	1.148662	0.9;
	6960	1	0	0	0	0	1	1.10027	-13.976466	380	1	1.155651	0.9;
	6961	1	20.82309	7.9	0	0	1	1.09304	-38.751734	380	1	1.133932	0.9;
	6962	1	0	0	0.236673	0	1	1.084042	-23.753324	380	1	1.2	0.8;
	6963	1	66.07326	21.6	0	0	1	1.057901	0.28342	150	1	1.120746	0.9;
	6964	2	0	0	0	0	1	1.089579	38.036908	380	1	1.114996	0.9;
	6965	1	0	0	0	0	1	1.043993	-8.95686	220	1	1.115607	0.9;
	6966	1	0	0	0	0	1	1.075755	-25.280549	220	1	1.125959	0.9;
	6967	2	0	0	0	0	1	1.09413	45.725227	380	1	1.10595	0.9;
	6968	2	36.94096	15.2	0	0	1	1.063409	-17.664682	220	1	1.1	0.9;
	6969	1	0	0	0	0	1	1.039563	-15.831313	150	1	1.1	0.9;
	6970	1	0	0	0	0	1	1.036822	61.14157	220	1	1.1	0.9;
	6971	1	49.45483	6.5	0	0	1	1.053024	-39.848025	220	1	1.1	0.9;
	6972	1	0	0	0	0	1	1.08782	-24.540019	380	1	1.13964	0.9;
	6973	2	90.2	30.3	0	0	1	0.975909	-25.143696	110	1	1.1	0.9;
	6974	1	12.7141	5.1	0	0	1	1.08243	-45.311822	380	1	1.138273	0.9;
	6975	1	0	0	0	0	1	1.069203	4.568016	380	1	1.127038	0.9;
	6976	1	0	0	0	0	1	1.07546	-3.735455	380	1	1.136646	0.9;
	6977	1	0	0	0.133168	0	1	1.034304	-23.255188	220	1	1.1	0.9;
	6978	1	25.12786	15.6	0	0	1	1.026159	-24.472203	150	1	1.1	0.9;
	6979	2	0	0	0	0	1	1.070785	-13.933045	220	1	1.140409	0.9;
	6980	1	0	0	0	0	1	1.026272	-27.007979	220	1	1.1	0.9;
	6981	1	10.9121	2.5	0	0	1	0.973798	-27.289807	110	1	1.1	0.9;
	6982	2	0	0	0	0	1	1.087182	-3.617282	220	1	1.139969	0.9;
	6983	1	0	0	0	0	1	1.062168	-18.452467	150	1	1.1	0.9;
	6984	1	0	0	0	0	1	1.060768	9.443089	380	1	1.115649	0.9;
	6985	1	66.17337	27.4	0	0	1	1.067331	-9.73067	220	1	1.13478	0.9;
	6986	1	11.91321	3.5	0	0	1	1.054659	4.97652	150	1	1.10912	0.9;
	6987	1	0	0	0	0	1	1.081749	-15.342099	220	1	1.144567	0.9;
	6988	1	51.95761	6.5	0	0	1	0.979545	-22.120236	110	1	1.1	0.9;
	6989	1	0	0	0	0	1	0.973252	-24.235149	150	1	1.1	0.9;
	6990	1	0	0	0	0	1	0.975459	-27.222766	110	1	1.1	0.9;
	6991	1	0	0	0	0	1	1.079776	-6.576424	220	1	1.136257	0.9;
	6992	2	131.3	19.1	0	0	1	0.997045	20.081505	220	1	1.1	0.9;
	6993	1	0.200222	88.5	0	0	1	0.941618	68.041102	150	1	1.1	0.9;
	6994	1	2.202442	0.5	0	0	1	1.033791	-37.338992	220	1	1.1	0.9;
	6995	1	60.0666	36	0	0	1	0.950564	-26.072262	150	1	1.1	0.9;
	6996	1	126.2	-12.5	0	0	1	1.077674	-1.771034	220	1	1.14462	0.9;
	6997	1	0	0	0	0	1	1.000264	-22.280101	150	1	1.1	0.9;
	6998	1	15.11676	4.9	0	0	1	0.990589	-33.223626	110	1	1.1	0.9;
	6999	1	0	0	0	0	1	1.024302	11.796701	220	1	1.1	0.9;
	7000	1	85.9	93.8	0	0	1	1.046525	10.175765	380	1	1.106007	0.9;
	7001	1	0	0	0	0	1	1.067086	-8.325421	220	1	1.129248	0.9;
	7002	2	0	0	0	0	1	1.053864	-36.117302	220	1	1.136451	0.9;
	7003	1	0	0	0	0	1	1.045001	-17.9902	220	1	1.1	0.9;
	7004	1	70.77847	43.8	0	0	1	0.940228	-31.944613	150	1	1.1	0.843794;
	7005	1	38.94318	11	0	0	1	1.022742	-29.884466	110	1	1.2	0.8;
	7006	1	17.01887	5.5	0	0	1	0.985698	-27.551281	110	1	1.1	0.9;
	7007	1	220.2	83.2	0	0	1	1.035914	-27.498447	220	1	1.1	0.9;
	7008	1	39.64395	4.4	0	0	1	1.034167	-23.426603	110	1	1.119659	0.9;
	7009	1	65.97314	64.1	0	0	1	1.046085	9.750951	380	1	1.106827	0.9;
	7010	1	0	0	0	0	1	1.067117	-9.698373	220	1	1.134438	0.9;
	7011	1	46.05106	22	0	0	1	1.060477	-31.367303	220	1	1.123698	0.9;
	7012	1	20.82309	6.7	0	0	1	1.043514	5.148487	110	1	1.110921	0.9;
	7013	1	0	0	0	0	1	0.999367	-23.854285	110	1	1.1	0.9;
	7014	1	71.57936	48.2	0.289482	0	1	1.050863	-8.558965	380	1	1.120949	0.9;
	7015	1	137	33.2	0	0	1	1.044503	-23.637682	220	1	1.1	0.9;
	7016	1	65.5727	44.6	0	0	1	0.9464	-27.90981	150	1	1.1	0.9;
	7017	1	0	0	0	0	1	1.043343	-15.729372	150	1	1.1	0.9;
	7018	1	0	0	0	0	1	1.06956	-1.802965	150	1	1.147822	0.9;
	7019	2	35.73963	11.1	0	0	1	1.073817	-17.186738	110	1	1.123865	0.9;
	7020	1	31.8353	9.2	0	0	1	1.045312	-21.493095	220	1	1.131383	0.9;
	7021	1	82.29124	29.6	0	0	1	0.970346	-25.284915	110	1	1.1	0.9;
	7022	1	0	0	0	0	1	1.006122	-25.019137	150	1	1.1	0.9;
	7023	2	0	0	0	0	1	0.980864	43.59736	220	1	1.1	0.9;
	7024	2	39.04329	12	0	0	1	0.982867	61.927297	150	1	1.1	0.9;
	7025	2	0	0	0	0	1	1.018455	-21.983016	220	1	1.1	0.9;
	7026	1	8.309213	3	0	0	1	1.01927	-20.86309	110	1	1.10816	0.9;
	7027	1	24.22686	4.7	0	0	1	0.989144	-24.869914	110	1	1.1	0.9;
	7028	1	0	0	0	0	1	1.025112	-15.842941	110	1	1.1	0.9;
	7029	1	122.7	-29.8	0	0	1	1.038411	-4.981088	220	1	1.1	0.9;
	7030	1	0	0	0	0	1	1.078721	-16.397852	220	1	1.139158	0.9;
	7031	1	0	0	0	0	1	1.078443	-0.216923	150	1	1.146517	0.9;
	7032	1	524.5	126	0	0	1	1.090243	-17.400872	220	1	1.163545	0.9;
	7033	1	0	0	0	0	1	1.026347	-45.414167	220	1	1.1	0.9;
	7034	2	0	0	0	0	1	1.089737	-37.888117	380	1	1.134269	0.9;
	7035	2	8.609546	5.4	0	0	1	0.986835	-23.27976	150	1	1.1	0.9;
	7036	1	17.51942	5.7	0	0	1	1.074406	-13.200076	110	1	1.139368	0.9;
	7037	2	0	0	0	0	1	1.102	-10.984823	380	1	1.143513	0.9;
	7038	2	0	0	0	0	1	1.105763	-11.639118	380	1	1.186745	0.9;
	7039	1	10.61177	1	0	0	1	0.986667	-17.641829	110	1	1.1	0.9;
	7040	2	0	0	0	0	1	1.033864	-37.31668	220	1	1.1	0.9;
	7041	1	0	0	0.261317	20	1	1.034202	-3.975882	380	1	1.1	0.9;
	7042	1	0	0	0	0	1	1.044734	-22.312041	150	1	1.1	0.9;
	7043	1	0	0	0	0	1	1.074782	-17.572839	380	1	1.142357	0.9;
	7044	1	11.01221	3.6	0	0	1	0.967959	-26.250037	110	1	1.1	0.9;
	7045	1	78.88747	25.6	0	0	1	1.042076	-12.670272	220	1	1.114472	0.9;
	7046	1	0	0	0	0	1	1.067289	-10.004155	220	1	1.134096	0.9;
	7047	1	11.91321	3.9	0	0	1	1.036916	-28.99692	110	1	1.1	0.9;
	7048	1	47.15228	15.5	0	0	1	1.030014	-41.833994	220	1	1.12098	0.9;
	7049	1	0	0	0	0	1	1.094708	45.080976	380	1	1.117366	0.9;
	7050	1	223.6	51.8	0	0	1	1.047276	-25.522367	220	1	1.127314	0.9;
	7051	1	0	0	0	0	1	1.043068	-20.691062	150	1	1.1	0.9;
	7052	1	27.23019	10.1	0	0	1	1.072246	-23.050874	220	1	1.1	0.9;
	7053	1	151.5	7.9	0	0	1	1.068988	-27.441829	220	1	1.165848	0.9;
	7054	2	0	0	0	0	1	1.102158	5.615684	380	1	1.140157	0.9;
	7055	1	0	0	0	0	1	1.077222	-15.058076	380	1	1.123693	0.9;
	7056	1	153	21.5	0	0	1	1.072576	-15.858225	380	1	1.150859	0.9;
	7057	2	0	0	0	0	1	1.0968	8.66518	150	1	1.157061	0.9;
	7058	1	108.5	35.3	0	0	1	1.037031	9.681796	380	1	1.1	0.9;
	7059	1	1.401554	0.5	0	0	1	1.048643	-21.637845	110	1	1.147271	0.9;
	7060	1	26.12897	8.5	0	0	1	0.966934	-26.980067	110	1	1.1	0.9;
	7061	1	11.01221	5.5	0	0	1	0.991751	-35.000203	110	1	1.1	0.9;
	7062	2	0	0	0	0	1	1.078	-21.481135	220	1	1.123105	0.9;
	7063	1	113.4	11.9	0	0	1	1.019095	-4.701969	220	1	1.1	0.9;
	7064	1	30.0333	9.6	0	0	1	1.035439	-46.06124	220	1	1.1	0.9;
	7065	1	0	0	0.096002	0	1	1.040516	-17.079916	220	1	1.113458	0.9;
	7066	1	0	0	0	0	1	1.039662	-12.933214	220	1	1.112647	0.9;
	7067	1	0	0	0	0	1	1.025724	-33.703825	150	1	1.1	0.9;
	7068	1	7.808658	2.8	0	0	1	1.006991	-18.053104	110	1	1.1	0.9;
	7069	1	44.34917	5.1	0	0	1	1.069105	-42.06113	220	1	1.117588	0.9;
	7070	1	12.91432	4.7	0	0	1	1.065646	-20.057669	110	1	1.122465	0.9;
	7071	1	0	0	0	0	1	1.061558	-22.965362	220	1	1.135667	0.9;
	7072	2	0	0	0	0	1	1.031909	-10.73565	110	1	1.1	0.9;
	7073	1	86.4959	29	0	0	1	1.071638	26.939087	380	1	1.1	0.9;
	7074	2	9.610656	3.9	0	0	1	0.996273	-25.767058	110	1	1.1	0.9;
	7075	1	0	0	0	0	1	1.004863	-23.271618	150	1	1.1	0.9;
	7076	1	0	0	0.373992	0	1	1.019315	13.622221	380	1	1.1	0.9;
	7077	1	91.10101	33.5	0	0	1	1.053829	-35.41466	220	1	1.134235	0.9;
	7078	1	0	0	0	0	1	1.054748	-49.100485	220	1	1.111865	0.9;
	7079	1	0	0	0	0	1	0.963356	-13.277647	110	1	1.1	0.9;
	7080	1	0	0	0	0	1	1.100967	-13.436012	380	1	1.145632	0.9;
	7081	2	0	0	0	0	1	1.092	-0.661068	380	1	1.153052	0.9;
	7082	1	169.5	27.7	0	0	1	1.056609	-10.642723	220	1	1.119556	0.9;
	7083	1	0	0	0	0	1	1.089309	-45.27716	380	1	1.144835	0.9;
	7084	1	0	0	0	0	1	1.051911	60.74	220	1	1.105786	0.9;
	7085	1	22.82531	3.8	0	0	1	1.020409	-14.595189	110	1	1.1	0.9;
	7086	1	1.801998	0.5	0	0	1	0.961349	-18.477656	110	1	1.1	0.9;
	7087	1	149.1	-18.3	0	0	1	1.047743	-21.190113	220	1	1.1	0.9;
	7088	1	40.24462	16	0	0	1	1.073504	30.283226	380	1	1.131975	0.9;
	7089	1	14.81643	0	0	0	1	1.090865	-21.723802	220	1	1.141037	0.9;
	7090	1	-41.6462	13.7	0	0	1	1.053235	-15.373644	150	1	1.108789	0.9;
	7091	2	0	0	0	0	1	1.064681	-36.33216	220	1	1.127016	0.9;
	7092	1	127.2	-10.7	0	0	1	1.054494	-49.44811	220	1	1.1	0.9;
	7093	1	179	22.9	0	0	1	1.085415	-46.137996	380	1	1.127076	0.9;
	7094	1	0	0	0	0	1	0.914372	-34.680605	150	1	1.1	0.795;
	7095	1	34.33807	12.2	0	0	1	1.018431	-18.314488	110	1	1.113949	0.9;
	7096	1	9.210212	3.4	0	0	1	1.076097	-5.913718	220	1	1.135641	0.9;
	7097	1	0	0	0	0	1	0.989976	-25.376837	150	1	1.1	0.9;
	7098	1	32.43596	20.1	0	0	1	1.048163	-15.484434	150	1	1.1	0.9;
	7099	1	137.2	41.9	0.110996	0	1	1.072595	-38.502302	380	1	1.133627	0.9;
	7100	1	0	0	0.095843	0	1	1.053455	-24.692752	220	1	1.12604	0.9;
	7101	1	-3.20355	-7.9	0	0	1	1.082921	-5.922173	220	1	1.136666	0.9;
	7102	1	0	0	0.152776	0	1	1.060454	-20.768926	380	1	1.137379	0.9;
	7103	2	0	0	0	0	1	1.068978	-10.939093	380	1	1.125443	0.9;
	7104	1	0	0	0	0	1	0.973661	-25.897989	150	1	1.1	0.9;
	7105	2	193.3	42	0	0	1	1.039591	-43.461851	220	1	1.116878	0.9;
	7106	1	134.7	17.5	0	0	1	1.037828	-48.439945	220	1	1.1	0.9;
	7107	2	110.3	10.7	0	0	1	1.075227	-29.933773	220	1	1.128726	0.9;
	7108	2	0	0	0	0	1	0.995	-16.258164	110	1	1.1	0.9;
	7109	1	0	0	0	0	1	0.986864	-24.173699	150	1	1.1	0.9;
	7110	1	10.11121	4.3	0	0	1	0.958053	-24.989793	110	1	1.1	0.9;
	7111	2	0	0	0	0	1	0.976402	-27.769314	110	1	1.1	0.9;
	7112	1	0	0	0	0	1	1.090578	-2.797451	220	1	1.149462	0.9;
	7113	1	26.52941	7.1	0	0	1	1.060524	-31.365126	220	1	1.124032	0.9;
	7114	2	0	0	0	0	1	1.072474	-11.780965	380	1	1.126224	0.9;
	7115	2	0	0	0.083	0	1	1.049818	-17.386768	220	1	1.1	0.9;
	7116	1	0	0	0	0	1	1.095383	9.125973	150	1	1.149968	0.9;
	7117	1	38.64284	6	0	0	1	1.051639	-48.862652	220	1	1.110783	0.9;
	7118	1	0	0	0	11.48854	1	1.026192	-24.47222	150	1	1.1	0.9;
	7119	1	0	0	0	0	1	1.021552	-8.652584	220	1	1.1193	0.9;
	7120	1	0	0	0	0	1	0.972392	-26.068567	110	1	1.1	0.9;
	7121	1	0	0	0	0	1	1.066859	29.878114	380	1	1.126469	0.9;
	7122	1	0	0	0	0	1	1.043136	-18.538696	220	1	1.1	0.9;
	7123	2	0	0	0	0	1	1.059798	-0.581511	150	1	1.122561	0.9;
	7124	1	40.64506	16	0	0	1	0.9939	-23.027755	110	1	1.1	0.9;
	7125	2	316.5	100.8	0	0	1	0.94391	31.499903	220	1	1.1	0.9;
	7126	1	0	0	0	0	1	1.04366	-34.215288	220	1	1.106371	0.9;
	7127	1	0	0	0	0	1	1.082486	16.216207	380	1	1.12751	0.9;
	7128	1	0	0	0	0	1	1.075838	0.279493	220	1	1.151101	0.9;
	7129	1	0	0	0	0	1	1.026445	-23.767126	150	1	1.1	0.9;
	7130	1	0	0	0	0	1	1.006458	10.014592	220	1	1.1	0.9;
	7131	2	0	0	0	0	1	1.091763	-36.40366	380	1	1.143242	0.9;
	7132	1	178.9	9.1	0	0	1	1.049973	-37.023618	220	1	1.109632	0.9;
	7133	1	0	0	0	0	1	1.078148	-20.065595	220	1	1.157244	0.9;
	7134	1	0	0	0	0	1	0.978855	-26.750908	150	1	1.1	0.9;
	7135	1	11.41265	0.3	0	0	1	0.989876	-26.619696	110	1	1.1	0.9;
	7136	1	84.99424	19.6	0	0	1	1.076304	-25.487035	220	1	1.166309	0.9;
	7137	1	2.202442	0.8	0	0	1	0.974161	-18.155217	110	1	1.1	0.9;
	7138	1	88.49812	10.8	0	0	1	1.050823	-19.781828	220	1	1.1	0.9;
	7139	1	0	0	0	0	1	1.033904	-25.118203	150	1	1.1	0.9;
	7140	1	101.9	20	0	0	1	1.07208	-42.176185	220	1	1.118001	0.9;
	7141	1	0	0	0	-91.049	1	1.075487	-17.275574	380	1	1.147298	0.9;
	7142	1	0	0	0	0	1	0.987472	-25.207427	150	1	1.1	0.9;
	7143	1	17.01887	2	0	0	1	1.076682	-17.420027	220	1	1.115223	0.9;
	7144	2	13.31476	2.4	0	0	1	1.064234	-14.260181	220	1	1.135016	0.9;
	7145	1	0	0	0	0	1	1.028516	16.675292	380	1	1.1	0.9;
	7146	1	0	0	0	0	1	1.007278	-41.454021	220	1	1.1	0.9;
	7147	1	0	0	0.037402	0	1	1.07998	67.259738	380	1	1.1	0.9;
	7148	2	65.77292	6.4	0	0	1	1.038913	-45.949342	220	1	1.1	0.9;
	7149	1	0.300333	-97.2	0	0	1	1.074286	0.702798	220	1	1.149733	0.9;
	7150	1	73.58158	48.6	0	0	1	1.007391	-50.933435	150	1	1.1	0.9;
	7151	2	0	0	0	0	1	1.047868	20.808588	380	1	1.107779	0.9;
	7152	1	0	0	0	0	1	1.083129	-14.793888	220	1	1.162185	0.9;
	7153	1	26.62953	9.7	0	0	1	1.025963	-15.483788	110	1	1.115678	0.9;
	7154	1	47.55272	17.8	0	0	1	1.032701	-28.213048	220	1	1.137587	0.9;
	7155	1	35.83974	14.1	0	0	1	1.08353	3.086707	150	1	1.153961	0.9;
	7156	1	142.4	20	0	0	1	1.078274	-20.096469	220	1	1.141065	0.9;
	7157	1	162.5	42.7	0	0	1	1.049702	-18.078204	220	1	1.115104	0.9;
	7158	2	51.75739	6.2	0	0	1	0.986899	63.936157	150	1	1.1	0.9;
	7159	1	0	0	0.095843	0	1	1.067061	-14.106802	220	1	1.131156	0.9;
	7160	2	125.3	7.7	0	0	1	1.064722	-0.360925	220	1	1.163979	0.9;
	7161	1	427.9	145.8	0	0	1	1.064521	-35.795365	380	1	1.12685	0.9;
	7162	1	36.94096	11	0	0	1	1.051053	-10.982984	220	1	1.13171	0.9;
	7163	1	0	0	0	0	1	1.048829	-44.680892	220	1	1.1	0.9;
	7164	1	338.7	52.5	0	0	1	1.078899	-18.848241	220	1	1.180037	0.9;
	7165	2	0	0	0	0	1	1.093273	-11.105072	220	1	1.14338	0.9;
	7166	1	3.203552	0.6	0	0	1	0.993622	-24.438347	110	1	1.1	0.9;
	7167	1	0	0	0	0	1	1.062965	23.326229	380	1	1.1	0.9;
	7168	1	31.73519	6.3	0	0	1	0.978063	-22.052468	110	1	1.1	0.9;
	7169	1	26.92986	9.6	0	0	1	0.996028	-22.428711	110	1	1.1	0.9;
	7170	1	191.7	6.8	0	0	1	1.008653	-38.44521	220	1	1.1	0.9;
	7171	1	0	0	0	0	1	1.03984	-11.497765	110	1	1.1	0.9;
	7172	1	0	0	0	0	1	1.050358	-41.174225	220	1	1.184189	0.9;
	7173	1	30.34628	30.5	0	0	1	1.030031	-22.502842	150	1	1.1	0.9;
	7174	1	347.1	87.2	0	0	1	1.059253	-17.144593	380	1	1.149797	0.9;
	7175	1	0	0	0	0	1	1.044792	-22.264221	150	1	1.1	0.9;
	7176	1	0	0	0	0	1	1.015262	-18.684439	110	1	1.1	0.9;
	7177	1	0	0	0	0	1	1.06991	1.005972	380	1	1.124129	0.9;
	7178	1	0	8.3	0	0	1	1.007429	-27.604642	150	1	1.193087	0.9;
	7179	1	0	0	0	0	1	1.006595	-25.950033	150	1	1.1	0.9;
	7180	2	0	0	0	0	1	1.100789	-8.698038	380	1	1.17141	0.9;
	7181	1	-596.4	-284.9	0.321115	0	1	1.083769	-6.709414	380	1	1.150313	0.9;
	7182	2	0	0	0	0	1	1.043032	14.666805	150	1	1.1	0.9;
	7183	1	35.13896	7.6	0	0	1	1.018857	-33.907754	110	1	1.1	0.9;
	7184	1	0	0	0	0	1	1.075024	-15.566842	380	1	1.122313	0.9;
	7185	1	0	0	0	0	1	1.088024	-43.623718	380	1	1.132289	0.9;
	7186	1	0	0	0	0	1	1.095453	45.663249	380	1	1.106935	0.9;
	7187	1	160.3	89.7	0	0	1	1.052007	-7.599801	380	1	1.121553	0.9;
	7188	1	38.14229	23.6	0	0	1	1.056465	-15.360872	150	1	1.108268	0.9;
	7189	1	0	0	0.129001	0	1	1.09824	-15.421677	380	1	1.187965	0.9;
	7190	1	0	0	0	0	1	1.0626	-21.149034	220	1	1.139654	0.9;
	7191	1	106.1	-6.8	0	0	1	1.040701	5.335823	380	1	1.115677	0.9;
	7192	1	36.84085	14.1	0	0	1	1.091541	-38.815441	380	1	1.163968	0.9;
	7193	1	2.202442	0.5	0	0	1	1.057157	-29.962628	220	1	1.111605	0.9;
	7194	2	302.1	216	0	0	1	1.083474	-39.891533	380	1	1.125126	0.9;
	7195	1	17.51942	9.1	0	0	1	0.969177	-24.677143	110	1	1.1	0.9;
	7196	2	0	0	0	0	1	1.039859	-19.227714	220	1	1.112536	0.9;
	7197	1	0	0	0	0	1	1.007912	-18.859427	110	1	1.1	0.9;
	7198	2	0	0	0	0	1	0.999118	-16.135873	110	1	1.1	0.9;
	7199	2	0	0	0.546086	0	1	1.073026	47.072896	380	1	1.1	0.9;
	7200	1	4.504995	1.1	0	0	1	1.09384	-43.293679	380	1	1.150928	0.9;
	7201	1	115.8	29.3	0	0	1	1.044709	-13.062855	220	1	1.1	0.9;
	7202	1	56.06216	17	0	0	1	1.064359	-7.164324	220	1	1.119929	0.9;
	7203	1	31.03441	10.3	0	0	1	0.939324	-28.10132	110	1	1.1	0.9;
	7204	1	0	0	0	15.5502	1	1.001498	-26.883562	150	1	1.1	0.9;
	7205	2	0	0	0	0	1	1.099132	-15.350846	380	1	1.187598	0.9;
	7206	1	0	0	0	0	1	1.052471	56.405779	220	1	1.1	0.9;
	7207	1	100.9	26.7	0	0	1	1.042449	18.72672	380	1	1.1	0.9;
	7208	1	38.44262	2	0	0	1	1.059649	-15.0708	220	1	1.112233	0.9;
	7209	1	0	0	0	0	1	1.043302	-1.107156	220	1	1.115103	0.9;
	7210	1	17.91987	5.8	0	0	1	1.067596	-5.775818	220	1	1.123066	0.9;
	7211	1	0	0	0.191685	0	1	1.064938	-20.948845	220	1	1.135427	0.9;
	7212	1	0	0	0	0	1	1.057495	-30.223747	380	1	1.123335	0.9;
	7213	1	56.56271	3.5	0	0	1	0.975574	-23.053534	110	1	1.1	0.9;
	7214	1	156	60.7	0	0	1	1.048107	-0.310559	150	1	1.123027	0.9;
	7215	1	60.0666	0	0	0	1	1.030024	-25.175733	150	1	1.1	0.9;
	7216	2	488.7	184.7	0	0	1	1.056132	-1.591374	380	1	1.134025	0.9;
	7217	1	0	0	0	0	1	1.038283	4.775676	380	1	1.109663	0.9;
	7218	1	111.6	8.9	0	0	1	1.074809	-22.773441	220	1	1.162792	0.9;
	7219	1	0	0	0	0	1	0.986855	-9.591664	220	1	1.1	0.9;
	7220	1	0	0	0	0	1	1.055821	-9.166194	220	1	1.128193	0.9;
	7221	1	0	0	0	0	1	1.078153	0.274482	150	1	1.147611	0.9;
	7222	1	0	0	0	-25.7039	1	1.103538	-16.242814	380	1	1.163261	0.9;
	7223	2	0	0	0	0	1	1.086026	1.571045	380	1	1.135057	0.9;
	7224	1	50.55605	8.1	0	0	1	1.077626	-8.011946	220	1	1.143783	0.9;
	7225	1	0	0	0	15.16794	1	1.030372	-25.114039	150	1	1.1	0.9;
	7226	1	0	0	0	0	1	1.097543	-34.310796	380	1	1.16006	0.9;
	7227	1	124.5	59.2	0	0	1	1.073701	-5.745494	380	1	1.142979	0.9;
	7228	1	0	0	0	0	1	1.045444	-11.499606	220	1	1.116958	0.9;
	7229	1	4.805328	1.6	0	0	1	0.998642	-29.559814	110	1	1.1	0.9;
	7230	1	0	0	0.275009	0	1	1.075997	-38.291036	380	1	1.136889	0.9;
	7231	1	30.0333	10	0	0	1	0.977604	-24.596546	110	1	1.1	0.9;
	7232	1	0	0	0	0	1	1.040028	7.268137	380	1	1.114029	0.9;
	7233	1	237.9	68.5	0	0	1	1.050024	-18.065682	220	1	1.1	0.9;
	7234	1	115.7	10.7	0	0	1	1.050682	-49.319411	220	1	1.112759	0.9;
	7235	1	234	13.1	0	0	1	1.069289	-19.082173	220	1	1.137521	0.9;
	7236	1	0	0	0	0	1	0.965598	-21.358434	110	1	1.1	0.9;
	7237	1	0	0	0	0	1	0.940568	-29.185387	150	1	1.1	0.891966;
	7238	1	0	0	0.118998	0	1	1.002778	-21.989824	220	1	1.1	0.9;
	7239	2	253.7	85	0	0	1	1.104789	-18.769461	380	1	1.145175	0.9;
	7240	1	233.5	151	0	0	1	1.071598	-39.403259	380	1	1.132749	0.9;
	7241	1	3.00333	1.5	0	0	1	0.978805	-22.629217	110	1	1.1	0.9;
	7242	1	112	19.7	0	0	1	1.061627	-22.104434	220	1	1.139268	0.9;
	7243	2	254.5	67.3	0	0	1	1.04406	-3.619901	220	1	1.115141	0.9;
	7244	1	0	-8.1	0	0	1	1.05168	-21.501579	150	1	1.1	0.9;
	7245	2	0	0	0	0	1	1.07169	-2.020785	380	1	1.135965	0.9;
	7246	1	0	0	0	0	1	1.086488	-3.75361	220	1	1.139412	0.9;
	7247	1	12.61399	7.8	0	0	1	1.044944	-20.907491	150	1	1.1	0.9;
	7248	1	0	0	0	0	1	1.037705	-19.714048	110	1	1.1	0.9;
	7249	1	400.4	23.6	0.534155	0	1	1.079497	-24.475019	380	1	1.131596	0.9;
	7250	1	18.2202	6.3	0	0	1	0.966154	-19.655089	110	1	1.1	0.9;
	7251	1	11.91321	3.3	0	0	1	1.054548	-15.939933	220	1	1.126503	0.9;
	7252	1	0	0	0	0	1	1.066983	-2.473762	150	1	1.145362	0.9;
	7253	1	25.22797	47.6	0	0	1	1.02704	12.019811	220	1	1.1	0.9;
	7254	2	0	0	0	0	1	1.076053	3.801855	380	1	1.122721	0.9;
	7255	1	0	0	0	0	1	1.065278	-43.867345	220	1	1.114382	0.9;
	7256	1	0	0	0	0	1	1.046839	-15.762775	150	1	1.1	0.9;
	7257	1	0	0	0	0	1	0.951192	63.781	150	1	1.1	0.9;
	7258	1	0	0	0	0	1	0.968094	-24.771768	150	1	1.1	0.9;
	7259	1	0	0	0	0	1	1.058592	-27.8238	110	1	1.1	0.9;
	7260	1	0	0	0	0	1	0.991895	-25.112032	150	1	1.1	0.9;
	7261	1	0	0	0	0	1	1.07823	-13.04085	220	1	1.144878	0.9;
	7262	1	0	0	0	0	1	1.085055	32.549692	380	1	1.122598	0.9;
	7263	1	7.908769	2.7	0	0	1	1.027148	-50.925176	150	1	1.1	0.9;
	7264	1	0	0	0	0	1	1.082395	-42.004033	380	1	1.145864	0.9;
	7265	1	3.203552	1.5	0	0	1	0.994862	-24.324268	110	1	1.1	0.9;
	7266	1	0	0	0	0	1	0.958132	-29.694244	150	1	1.1	0.883683;
	7267	1	36.84085	14.1	0	0	1	1.091334	-36.545337	380	1	1.151411	0.9;
	7268	1	40.0444	15.1	0	0	1	0.943563	-26.277956	110	1	1.1	0.9;
	7269	1	149.6277	59	0	0	1	1.05294	-36.242686	220	1	1.135228	0.9;
	7270	1	28.83197	10.4	0	0	1	0.978599	-23.767561	110	1	1.1	0.9;
	7271	1	23.32586	9.1	0	0	1	1.069206	3.317991	150	1	1.139762	0.9;
	7272	1	0	0	0	0	1	1.068149	9.782361	150	1	1.1	0.9;
	7273	1	9.410433	2.3	0	0	1	0.987372	-17.021451	110	1	1.1	0.9;
	7274	1	0	0	0	0	1	1.016272	29.760965	220	1	1.1	0.9;
	7275	1	1.801998	0.7	0	0	1	0.98583	-18.022117	110	1	1.1	0.9;
	7276	1	14.21576	4.7	0	0	1	1.010369	-27.104249	220	1	1.1	0.9;
	7277	1	209.1	1.7	0	0	1	1.062241	-10.341724	220	1	1.121586	0.9;
	7278	1	0	0	0	0	1	1.017139	-26.263379	150	1	1.179396	0.9;
	7279	1	0	0	0	0	1	0.997525	-25.844665	110	1	1.1	0.9;
	7280	1	0	0	0	0	1	0.96649	-22.247036	110	1	1.1	0.9;
	7281	2	52.55827	8	0	0	1	1.024771	62.246746	150	1	1.1	0.9;
	7282	1	0	0	0	0	1	1.075795	0.231624	380	1	1.12878	0.9;
	7283	1	0	0	0	0	1	1.036696	-28.275684	220	1	1.1	0.9;
	7284	1	8.309213	3	0	0	1	1.05064	-4.743452	110	1	1.126163	0.9;
	7285	1	0	0	0	0	1	1.073496	-4.632062	380	1	1.140384	0.9;
	7286	1	12.91432	4.2	0	0	1	0.931588	-35.432278	110	1	1.1	0.9;
	7287	1	0	0	0	0	1	0.943347	65.005995	150	1	1.1	0.9;
	7288	1	0	0	0	0	1	1.030319	11.808679	220	1	1.1	0.9;
	7289	1	83.99313	6	0	0	1	1.083362	-45.468661	380	1	1.148225	0.9;
	7290	1	0	0	0	0	1	1.081068	-15.386625	220	1	1.143389	0.9;
	7291	2	37.84196	12.3	0	0	1	1.059864	-44.810705	220	1	1.115737	0.9;
	7292	1	0	0	0	0	1	1.084771	7.89377	150	1	1.139161	0.9;
	7293	1	12.11343	5.1	0	0	1	0.915847	-26.450034	110	1	1.1	0.9;
	7294	1	70.37804	27.5	0	0	1	0.989613	-25.883988	110	1	1.1	0.9;
	7295	1	93.804	15.9	0	0	1	1.100181	-14.669544	110	1	1.182775	0.9;
	7296	1	0	0	0	0	1	1.04045	-19.934266	150	1	1.120019	0.9;
	7297	2	123.8	40.2	0	0	1	1.066182	-10.712501	220	1	1.142672	0.9;
	7298	2	0	0	0	0	1	1.072727	22.538622	110	1	1.117717	0.9;
	7299	2	3.303663	1.1	0	0	1	1.003909	-35.336982	110	1	1.1	0.9;
	7300	1	106.4	24.2	0	0	1	1.043654	-21.334026	110	1	1.105474	0.9;
	7301	1	0	0	0	11.992	1	0.999861	-23.664782	150	1	1.1	0.9;
	7302	1	0	0	0	0	1	1.099781	-17.1543	380	1	1.168112	0.9;
	7303	1	42.34695	13.5	0	0	1	0.967256	-20.737696	110	1	1.1	0.9;
	7304	1	19.02109	6	0	0	1	0.978362	-23.354553	110	1	1.1	0.9;
	7305	1	53.55938	18.8	0	0	1	1.06625	-19.327155	220	1	1.1	0.9;
	7306	1	0	0	0	11.367	1	1.027512	-25.796708	150	1	1.1	0.9;
	7307	1	28.63174	8.8	0	0	1	1.009701	-30.79746	110	1	1.1	0.9;
	7308	1	14.31587	2.2	0	0	1	1.010357	-36.439182	110	1	1.1	0.9;
	7309	1	69.1767	9.5	0	0	1	1.082409	-8.528567	220	1	1.15255	0.9;
	7310	1	10.71188	17.7	0	0	1	1.067222	9.766752	150	1	1.1	0.9;
	7311	1	170.8	70.4	0	0	1	1.008924	-19.483087	220	1	1.1	0.9;
	7312	1	0	0	0	0	1	1.0468	-13.06294	220	1	1.132373	0.9;
	7313	1	-19.3214	6.4	0	0	1	0.986843	-17.234553	150	1	1.1	0.9;
	7314	1	25.32808	8.9	0	0	1	1.004639	-19.753918	110	1	1.1	0.9;
	7315	1	135.2	70.6	0	0	1	1.016428	-30.02403	150	1	1.1	0.9;
	7316	2	0	0	0	0	1	1.099842	-19.478075	380	1	1.144866	0.9;
	7317	2	199.1	23.3	0	0	1	1.003182	-43.017392	220	1	1.1	0.9;
	7318	1	66.77403	19.8	0	0	1	1.041587	-11.432247	110	1	1.117961	0.9;
	7319	1	41.74628	7.4	0	0	1	1.054729	-48.853168	220	1	1.111518	0.9;
	7320	1	174.2	54.3	0	0	1	1.043601	-45.249468	380	1	1.110514	0.9;
	7321	1	3.804218	1.3	0	0	1	0.97364	-25.769326	110	1	1.1	0.9;
	7322	1	0	0	0	0	1	0.989795	-17.165993	110	1	1.1	0.9;
	7323	1	0	0	0	0	1	1.086039	-2.972324	220	1	1.145678	0.9;
	7324	1	68.47592	42.7	0	0	1	1.044052	-3.980561	220	1	1.1	0.9;
	7325	1	33.13674	20.5	0	0	1	1.008188	-21.944848	150	1	1.1	0.9;
	7326	1	6.206882	1.5	0	0	1	0.99369	-23.398187	110	1	1.1	0.9;
	7327	1	0	0	0	0	1	1.020962	-20.978968	150	1	1.1	0.9;
	7328	1	0	0	0	0	1	1.0394	-44.329828	220	1	1.1	0.9;
	7329	1	0	0	0	0	1	1.038995	-22.498736	150	1	1.1	0.9;
	7330	1	35.93985	22.3	0	0	1	1.010096	-23.832967	150	1	1.1	0.9;
	7331	2	0	0	0	0	1	1.016359	-26.868392	150	1	1.185836	0.9;
	7332	1	35.13896	11	0	0	1	1.017816	-36.440402	110	1	1.1	0.9;
	7333	1	0	0	0	0	1	1.066041	11.922022	380	1	1.11714	0.9;
	7334	2	70.57825	60.4	0	0	1	1.026682	-34.588015	220	1	1.1	0.9;
	7335	1	0	0	0.411998	0	1	1.079488	-47.003549	380	1	1.136358	0.9;
	7336	1	39.34362	24.4	0	0	1	0.96616	-25.110965	150	1	1.1	0.9;
	7337	1	73.18114	23.4	0	0	1	0.980678	-18.403818	110	1	1.1	0.9;
	7338	1	0	0	0	0	1	0.992287	-16.750701	150	1	1.1	0.9;
	7339	1	41.84639	15.2	0	0	1	1.054939	-3.541927	110	1	1.128503	0.9;
	7340	1	0	0	0	11.90083	1	1.008222	-21.944875	150	1	1.1	0.9;
	7341	2	30.0333	8.3	0	0	1	0.981917	62.517424	150	1	1.1	0.9;
	7342	2	2.702997	-0.1	0	0	1	1.043636	-20.348152	110	1	1.1	0.9;
	7343	1	7.608436	0	0	0	1	0.91536	-26.565359	110	1	1.1	0.9;
	7344	1	206.4	21	0	0	1	1.025583	-45.498818	150	1	1.1	0.9;
	7345	1	0	0	0	0	1	1.06814	5.260044	380	1	1.126374	0.9;
	7346	2	0	0	0	0	1	1.069409	-20.36507	220	1	1.110316	0.9;
	7347	1	82.1	29.1	0	0	1	0.975062	-25.660261	110	1	1.1	0.9;
	7348	1	39.34362	-10.7	0	0	1	1.045868	-4.172344	150	1	1.1	0.9;
	7349	1	127.2	40.8	0	0	1	1.072915	-24.30099	220	1	1.161422	0.9;
	7350	1	68.17558	16.1	0	0	1	1.060304	-22.324592	220	1	1.138094	0.9;
	7351	1	35.83974	13	0	0	1	1.04197	-20.022602	110	1	1.1	0.9;
	7352	2	0	0	0	0	1	1.052974	-20.392315	380	1	1.13826	0.9;
	7353	1	0	0	0	0	1	1.033245	-25.20708	150	1	1.1	0.9;
	7354	1	87.89746	29.3	0	0	1	1.075831	-4.470419	220	1	1.137173	0.9;
	7355	1	107	43	0	0	1	1.033921	-5.408596	220	1	1.108675	0.9;
	7356	1	0	0	0	0	1	1.030167	9.800946	380	1	1.1	0.9;
	7357	1	19.02109	6.3	0	0	1	0.960477	-13.579467	110	1	1.1	0.9;
	7358	2	73.98203	-7.4	0	0	1	1.068816	-37.996206	380	1	1.13622	0.9;
	7359	1	111.1	35.6	0	0	1	0.982896	-30.124397	110	1	1.1	0.9;
	7360	2	0	0	0	0	1	1.021818	-16.696237	110	1	1.1	0.9;
	7361	1	0	0	0	0	1	1.058247	-43.767572	380	1	1.126039	0.9;
	7362	2	0	0	0	0	1	1.018818	14.10792	220	1	1.1	0.9;
	7363	1	80.48924	7.3	0	0	1	1.0565	-54.901983	220	1	1.1	0.9;
	7364	1	0	0	0	0	1	1.084619	-2.261476	220	1	1.145123	0.9;
	7365	1	0	0	0	0	1	1.084236	-12.691052	380	1	1.128556	0.9;
	7366	1	8.909878	2.5	0	0	1	0.943406	-25.101453	110	1	1.1	0.9;
	7367	1	0	0	0	0	1	1.086363	-8.942888	380	1	1.178413	0.9;
	7368	1	0	0	0	0	1	1.013493	61.701025	150	1	1.1	0.9;
	7369	1	34.43818	27	0	0	1	0.919681	-29.100332	150	1	1.1	0.868813;
	7370	1	0	0	0	0	1	1.03045	-26.48896	150	1	1.1	0.9;
	7371	1	12.61399	6.3	0	0	1	1.082824	-2.478966	220	1	1.144137	0.9;
	7372	2	0	0	0	0	1	1.018773	14.125259	220	1	1.1	0.9;
	7373	1	0	0	0.045353	0	1	1.07591	-19.564358	380	1	1.149884	0.9;
	7374	1	0	0	0	0	1	1.061345	-39.084359	220	1	1.123312	0.9;
	7375	1	242.4	46.7	0	0	1	1.078434	-31.462863	380	1	1.152776	0.9;
	7376	1	18.72076	6.6	0	0	1	1.01355	-31.844116	110	1	1.1	0.9;
	7377	1	0	0	0	0	1	1.085043	2.138944	150	1	1.157391	0.9;
	7378	1	10.0111	2	0	0	1	1.050293	-28.923613	150	1	1.120508	0.9;
	7379	1	22.32475	4	0	0	1	1.03448	-30.394998	110	1	1.1	0.9;
	7380	1	0	0	0	0	1	1.056508	-3.235638	110	1	1.129239	0.9;
	7381	2	171	43.6	0	0	1	1.048327	-33.406028	220	1	1.1	0.9;
	7382	1	0	0	0	12.27504	1	0.988287	-24.112735	150	1	1.1	0.9;
	7383	1	303.5	56.2	0	0	1	1.044596	-43.768766	220	1	1.11464	0.9;
	7384	1	3.804218	1.9	0	0	1	1.022873	-15.336259	110	1	1.1	0.9;
	7385	2	112.2	28.4	0	0	1	1.024915	-42.062156	220	1	1.1	0.9;
	7386	1	0	0	0	0	1	1.055031	-24.416808	220	1	1.129785	0.9;
	7387	1	0	0	0	0	1	0.975882	-30.33119	150	1	1.1	0.9;
	7388	1	21.52386	5.1	0	0	1	1.007229	-19.866171	110	1	1.1	0.9;
	7389	1	0	0	0	0	1	1.076978	-17.688167	150	1	1.1	0.9;
	7390	1	125.3	34.6	0	0	1	1.08421	7.867258	150	1	1.139425	0.9;
	7391	2	0	0	0	0	1	1.092421	9.512841	380	1	1.126526	0.9;
	7392	2	0	0	0	0	1	1.068658	-37.065184	380	1	1.131999	0.9;
	7393	2	0	0	0	0	1	1.091182	-0.520256	220	1	1.155126	0.9;
	7394	2	2.1	396.2	0	0	1	1.086079	-36.434836	380	1	1.129521	0.9;
	7395	1	0	0	0	0	1	1.011539	-24.859836	150	1	1.1	0.9;
	7396	1	-34.0377	-2	0	0	1	1.082249	-42.003331	380	1	1.145711	0.9;
	7397	1	0	0	0	0	1	1.031597	-22.320669	150	1	1.1	0.9;
	7398	1	76.58492	16.2	0	0	1	1.05464	-49.049606	220	1	1.110285	0.9;
	7399	1	1.801998	0.6	0	0	1	1.056095	5.507076	150	1	1.109766	0.9;
	7400	1	0	0	0	0	1	1.087922	-3.436366	220	1	1.141486	0.9;
	7401	1	0	0	0.095843	0	1	1.057017	-22.355401	220	1	1.124936	0.9;
	7402	1	0	0	0	0	1	1.053136	3.799813	150	1	1.10922	0.9;
	7403	1	20.62287	7.5	0	0	1	1.011946	-17.894096	110	1	1.112058	0.9;
	7404	1	0	0	0	0	1	1.02933	-23.360155	150	1	1.1	0.9;
	7405	2	0	0	0	0	1	1.094034	-29.027059	220	1	1.145695	0.9;
	7406	1	2.8	14.6	0	0	1	1.095042	-14.731238	380	1	1.155602	0.9;
	7407	1	0	0	0	0	1	1.036956	-24.239868	220	1	1.1	0.9;
	7408	1	22.12453	8	0	0	1	1.025264	-23.321501	110	1	1.1	0.9;
	7409	1	0	0	0	0	1	1.058858	-20.840179	110	1	1.119228	0.9;
	7410	1	13.11454	3.6	0	0	1	1.061668	-7.191828	220	1	1.119836	0.9;
	7411	1	10.81199	3.6	0	0	1	0.938336	-34.936179	110	1	1.1	0.9;
	7412	1	0	0	0	0	1	1.00239	-20.015909	150	1	1.1	0.9;
	7413	1	0	0	0.165589	0	1	1.027598	-4.990479	380	1	1.1	0.9;
	7414	1	291.6	47	0	0	1	1.06727	-29.900429	220	1	1.192597	0.9;
	7415	1	0	0	0.181698	0	1	1.028341	-4.857575	380	1	1.1	0.9;
	7416	1	73.58158	-13.1	0	0	1	1.05914	-18.830154	220	1	1.113902	0.9;
	7417	1	0	0	0	0	1	1.101209	-43.269618	380	1	1.15693	0.9;
	7418	1	0	0	0	0	1	1.033584	-25.139205	150	1	1.1	0.9;
	7419	2	0	0	0	0	1	1.052053	-35.414382	380	1	1.117803	0.9;
	7420	1	99.6	22.9	0	0	1	1.053	-49.113901	220	1	1.110387	0.9;
	7421	2	54.96094	62.6	0	0	1	1.029	3.15799	380	1	1.1	0.9;
	7422	1	120.2	30.7	0	0	1	0.969584	45.39224	220	1	1.1	0.9;
	7423	1	0	0	0	0	1	1.058643	-16.954948	220	1	1.112262	0.9;
	7424	1	0	0	0	0	1	1.050246	-10.468262	220	1	1.12044	0.9;
	7425	1	0	0	0	0	1	1.045191	-14.970076	220	1	1.132063	0.9;
	7426	1	2.00222	1	0	0	1	0.946135	-26.77586	110	1	1.1	0.9;
	7427	1	0	0	0.564347	0	1	1.078405	42.899591	380	1	1.114438	0.9;
	7428	2	0	0	0	0	1	1.065427	9.498358	150	1	1.1	0.9;
	7429	2	0	0	0	0	1	1.046293	-15.946321	380	1	1.1	0.9;
	7430	1	-134.349	-6.5	0	0	1	1.023046	-34.057581	150	1	1.1	0.9;
	7431	1	165.9	49.6	0	0	1	1.032707	25.138987	220	1	1.1	0.9;
	7432	1	169.6	98.3	0	0	1	1.043895	-13.725475	380	1	1.107513	0.9;
	7433	1	0	0	0	0	1	1.049795	-5.537043	380	1	1.134782	0.9;
	7434	2	0.200222	87.4	0	0	1	0.951192	63.781	150	1	1.1	0.9;
	7435	1	0	0	0	0	1	1.086666	0.705652	150	1	1.153537	0.9;
	7436	1	4.00444	1.3	0	0	1	0.979585	-33.930043	110	1	1.1	0.9;
	7437	1	0	0	0.06131	0	1	0.988674	66.037364	380	1	1.1	0.9;
	7438	2	49.1545	12.8	0	0	1	1.020364	-38.191099	220	1	1.1	0.9;
	7439	1	3.804218	3.8	0	0	1	0.946857	-27.169564	110	1	1.1	0.9;
	7440	1	83.39246	16.5	0	0	1	1.068815	-19.1578	220	1	1.137107	0.9;
	7441	2	7.5	173.3	0	0	1	1.086553	-34.061549	380	1	1.148336	0.9;
	7442	1	0	0	0	0	1	1.042113	15.943032	380	1	1.110707	0.9;
	7443	2	0	0	0	0	1	0.976364	-23.167276	110	1	1.1	0.9;
	7444	1	207.9	25.4	0	0	1	1.034428	-42.057824	220	1	1.1	0.9;
	7445	1	58.96538	11.1	0	0	1	1.045028	-11.552767	220	1	1.11656	0.9;
	7446	2	0	0	0	0	1	1.075773	-34.056762	220	1	1.156154	0.9;
	7447	1	0	0	0	0	1	1.032464	-4.605038	220	1	1.113562	0.9;
	7448	1	63.77071	39.5	0	0	1	0.981805	-25.683128	150	1	1.1	0.9;
	7449	1	0	0	0	0	1	1.031778	-16.150622	110	1	1.13181	0.9;
	7450	1	0	0	0	0	1	1.070809	-14.452301	150	1	1.123387	0.9;
	7451	1	0	0	0	0	1	1.040092	-18.623313	220	1	1.1	0.9;
	7452	2	7.207992	2.2	0	0	1	0.995385	-22.744437	110	1	1.1	0.9;
	7453	2	0	0	0	0	1	1.023316	-10.201438	380	1	1.108278	0.9;
	7454	2	0	0	0	0	1	1.089237	15.535813	380	1	1.133194	0.9;
	7455	1	92.70278	-5.6	0.284	0	1	1.079148	-37.791928	380	1	1.137118	0.9;
	7456	2	0	0	0	0	1	1.071136	-11.023303	220	1	1.15037	0.9;
	7457	1	0	0	0.482331	0	1	1.083203	-30.001501	380	1	1.124491	0.9;
	7458	1	0	0	0	64.04321	1	1.077234	-8.979349	150	1	1.123998	0.9;
	7459	1	0	0	0	0	1	1.057628	10.061056	380	1	1.111508	0.9;
	7460	1	0	0	0	0	1	0.992206	-25.088203	150	1	1.1	0.9;
	7461	1	1.801998	0.7	0	0	1	1.042001	-16.273815	110	1	1.1	0.9;
	7462	1	0	0	0	0	1	1.006568	-20.152812	150	1	1.1	0.9;
	7463	1	0	0	0	11.1728	1	1.040772	-25.025005	150	1	1.1	0.9;
	7464	1	1.902109	1.3	0	0	1	1.061377	-12.921398	110	1	1.125126	0.9;
	7465	1	118.1	20.2	0	0	1	1.052842	-49.163713	220	1	1.111215	0.9;
	7466	1	370.2	107.7	0	0	1	1.082402	-39.625228	380	1	1.152282	0.9;
	7467	1	0	0	0	24.63838	1	1.077479	-17.685252	150	1	1.1	0.9;
	7468	1	16.51831	6.6	0	0	1	1.018958	-35.920482	110	1	1.1	0.9;
	7469	2	36.84085	14.1	0	0	1	1.097921	-35.204136	380	1	1.146088	0.9;
	7470	1	5.105661	2.5	0	0	1	1.066356	-13.279193	110	1	1.13096	0.9;
	7471	2	38.54273	49.6	0	0	1	1.034909	-42.172747	220	1	1.1	0.9;
	7472	1	44.74961	27.7	0	0	1	0.958749	-25.493894	150	1	1.1	0.9;
	7473	1	0	0	0	0	1	1.074468	1.596591	380	1	1.128233	0.9;
	7474	1	0	0	0	0	1	0.95113	-23.731422	110	1	1.1	0.9;
	7475	1	216.8	30.7	0	0	1	1.092549	-16.633678	380	1	1.152131	0.9;
	7476	1	50.75628	17.5	0	0	1	1.055573	-27.184742	220	1	1.118155	0.9;
	7477	2	0.4	182.7	0	0	1	1.085237	-36.747971	380	1	1.149987	0.9;
	7478	1	0	0	0	0	1	1.042314	9.505492	380	1	1.1	0.9;
	7479	1	196.1	32.1	0	0	1	1.006694	-26.231003	220	1	1.1	0.9;
	7480	1	9.610656	2.5	0	0	1	0.944551	-24.127927	110	1	1.1	0.9;
	7481	2	3.00333	0	0	0	1	1.059789	-40.87026	380	1	1.131066	0.9;
	7482	2	0	0	0	0	1	1.046427	-22.327748	150	1	1.1	0.9;
	7483	1	39.04329	13.1	0	0	1	1.071949	-13.13778	220	1	1.138518	0.9;
	7484	2	0	0	0	0	1	1.082251	19.554224	380	1	1.138224	0.9;
	7485	1	70.0777	43.4	0	0	1	0.985443	-25.578358	150	1	1.1	0.9;
	7486	1	0	0	0	0	1	1.095833	-18.484352	380	1	1.177688	0.9;
	7487	1	0	0	0	0	1	1.085791	-45.353619	380	1	1.142068	0.9;
	7488	1	69.37692	22.5	0	0	1	0.986791	-24.0981	110	1	1.1	0.9;
	7489	1	255.9	128.3	0	0	1	1.063255	-37.627627	380	1	1.117634	0.9;
	7490	1	0	0	0	0	1	1.042129	15.939173	380	1	1.110755	0.9;
	7491	1	40.44484	20.5	0	0	1	1.075706	-23.225412	220	1	1.164603	0.9;
	7492	1	316.7	46.1	0	0	1	1.034901	-46.113476	380	1	1.1	0.9;
	7493	2	0	0	0	0	1	1.090553	-9.884281	380	1	1.154755	0.9;
	7494	1	0	0	0	0	1	1.090434	-21.116068	380	1	1.166202	0.9;
	7495	2	12.11343	4.7	0	0	1	1.015636	-16.61729	110	1	1.107573	0.9;
	7496	1	-632.101	-447	0	0	1	1.043922	18.322967	380	1	1.105877	0.9;
	7497	1	15.61732	2.8	0	0	1	0.954854	-30.624521	110	1	1.1	0.9;
	7498	1	7.408214	2.7	0	0	1	1.069086	-20.015739	110	1	1.128134	0.9;
	7499	1	120.6	35.7	0	0	1	1.054451	-32.963655	380	1	1.116509	0.9;
	7500	1	15.61732	5.7	0	0	1	1.037477	-20.06872	110	1	1.1	0.9;
	7501	1	40.0444	20	0	0	1	1.012957	-47.504066	220	1	1.1	0.9;
	7502	1	73.68169	15.4	0	0	1	1.040387	-22.163206	220	1	1.1	0.9;
	7503	1	57.66393	6.1	0	0	1	1.028624	16.621691	110	1	1.1	0.9;
	7504	2	0	0	0	0	1	1.025867	-28.489218	150	1	1.1	0.9;
	7505	1	0	0	0	0	1	1.091193	-8.22951	380	1	1.127963	0.9;
	7506	1	14.31587	5.5	0	0	1	0.970742	-37.942295	110	1	1.1	0.9;
	7507	1	9.410433	5.6	0	0	1	1.03423	-25.001431	150	1	1.1	0.9;
	7508	1	0	0	0	0	1	1.107788	-9.080277	380	1	1.160322	0.9;
	7509	1	0	0	0	0	1	0.95972	-30.260309	110	1	1.1	0.9;
	7510	2	0	0	0	0	1	1.041081	25.766699	380	1	1.131026	0.9;
	7511	1	156.6	16.9	0	0	1	1.051735	-49.264396	220	1	1.1	0.9;
	7512	1	0	0	0	0	1	1.015569	-12.199598	110	1	1.1	0.9;
	7513	1	0	0	0	0	1	1.055733	-21.451281	220	1	1.106164	0.9;
	7514	1	16.01776	5.1	0	0	1	1.069485	8.050012	110	1	1.139057	0.9;
	7515	2	0	0	0	0	1	1.056507	5.933749	110	1	1.14478	0.9;
	7516	1	0	0	0	0	1	0.973638	-25.718869	110	1	1.1	0.9;
	7517	1	0	0	0	0	1	1.083886	-7.151518	220	1	1.150256	0.9;
	7518	1	141.5	70.8	0	0	1	1.052698	-15.383663	380	1	1.118524	0.9;
	7519	1	112.7	14.5	0	0	1	1.058645	-23.660755	220	1	1.132778	0.9;
	7520	1	14.21576	6.3	0	0	1	1.053297	-11.986612	220	1	1.122594	0.9;
	7521	1	5.00555	2	0	0	1	1.077359	-18.278899	110	1	1.119549	0.9;
	7522	1	0	0	0	0	1	1.04038	-29.187544	150	1	1.1	0.9;
	7523	1	0	0	0	0	1	1.041174	25.746391	380	1	1.131352	0.9;
	7524	1	0	0	0	0	1	1.002881	-23.202064	150	1	1.1	0.9;
	7525	1	0	0	0	0	1	1.087345	28.19803	220	1	1.147479	0.9;
	7526	1	59.66615	9.2	0	0	1	1.059983	-47.024751	220	1	1.113433	0.9;
	7527	1	13.91543	3.8	0	0	1	0.978627	-25.958771	110	1	1.1	0.9;
	7528	1	0	0	0	0	1	1.02638	-17.695222	150	1	1.1	0.9;
	7529	2	0	0	0	0	1	1.099184	-15.417402	380	1	1.187585	0.9;
	7530	1	99.3	23.6	0	0	1	0.98737	-24.009306	110	1	1.1	0.9;
	7531	1	0	0	0	0	1	1.078866	-3.883265	220	1	1.13198	0.9;
	7532	1	33.43707	11	0	0	1	0.978674	-21.972797	110	1	1.1	0.9;
	7533	1	0	0	0	0	1	1.059236	-8.216139	220	1	1.116588	0.9;
	7534	2	43.44817	15.9	0	0	1	1.073455	-13.981342	220	1	1.129371	0.9;
	7535	2	0	0	0	20	1	1.087079	-11.571665	380	1	1.16904	0.9;
	7536	2	0	0	0	0	1	1.032678	-15.886812	110	1	1.1	0.9;
	7537	1	13.91543	8.9	0	0	1	0.984714	-17.422043	110	1	1.1	0.9;
	7538	1	19.32142	7	0	0	1	1.005985	-22.550643	110	1	1.1	0.9;
	7539	1	0	0	0	0	1	1.051126	-25.843931	220	1	1.160998	0.9;
	7540	1	0	0	0	0	1	1.082643	16.352349	380	1	1.127554	0.9;
	7541	1	18.01998	6.5	0	0	1	1.054067	-21.066519	110	1	1.157522	0.9;
	7542	1	313.7526	96.8	0	0	1	1.006731	-48.524321	220	1	1.1	0.9;
	7543	1	74.78291	26.4	0	0	1	1.014007	-36.810376	110	1	1.1	0.9;
	7544	1	0	0	0.276	20	1	1.077548	-8.832521	380	1	1.112087	0.9;
	7545	1	0	0	0	0	1	1.029332	-21.484298	220	1	1.1	0.9;
	7546	1	41.44595	56.2	0	0	1	1.0212	-33.805391	220	1	1.1	0.9;
	7547	1	0	-8	0	0	1	1.01083	-27.349194	150	1	1.1	0.9;
	7548	1	82.39135	14.4	0	0	1	1.073234	-23.210416	220	1	1.138356	0.9;
	7549	1	0	0	0.14255	0	1	1.07856	0.918793	380	1	1.128411	0.9;
	7550	2	205.8	113.2	0	0	1	1.037295	2.458746	220	1	1.112396	0.9;
	7551	1	0	0	0	0	1	1.06035	-12.521323	220	1	1.144159	0.9;
	7552	1	0	0	0	0	1	1.089388	54.328218	380	1	1.114129	0.9;
	7553	1	174.7	31	0.231652	0	1	1.100982	57.873214	380	1	1.120225	0.9;
	7554	1	0	0	0	0	1	1.059373	-13.671668	380	1	1.111807	0.9;
	7555	2	658.8	156.2	0	0	1	1.060754	-11.335388	150	1	1.112964	0.9;
	7556	2	257.8	102	0.23336	503.4284	1	1.076169	0.558076	380	1	1.12829	0.9;
	7557	1	149.9	0	0	0	1	1.074478	-38.749191	380	1	1.122571	0.9;
	7558	2	0	0	0	0	1	1.018404	-40.011149	220	1	1.1	0.9;
	7559	1	0	0	0	0	1	0.984427	-24.998416	110	1	1.1	0.9;
	7560	2	29.74906	27.1	0	0	1	1.041467	-24.684337	150	1	1.1	0.9;
	7561	1	97	-539.3	0	0	1	1.071714	17.280251	380	1	1.141517	0.9;
	7562	1	44.04884	14.4	0	0	1	0.937106	-26.705923	110	1	1.1	0.9;
	7563	1	0	0	0	0	1	1.088875	-16.361627	380	1	1.164024	0.9;
	7564	1	19.2	9.3	0	0	1	1.076875	-15.711893	380	1	1.125262	0.9;
	7565	1	80.28902	-4.1	0	0	1	1.025917	9.790606	220	1	1.1	0.9;
	7566	2	0	0	0	0	1	1.043683	-17.615243	380	1	1.112316	0.9;
	7567	1	2.502775	1.3	0	0	1	0.97596	-20.171897	110	1	1.1	0.9;
	7568	1	0.100111	-0.9	0	0	1	1.042665	39.43861	220	1	1.1	0.9;
	7569	2	19.82198	6.4	0	0	1	1.02724	-14.662565	110	1	1.1	0.9;
	7570	1	67.27459	15.9	0	0	1	1.059217	-35.548537	220	1	1.107843	0.9;
	7571	1	155.1	39.9	0	0	1	1.068386	-50.905834	220	1	1.1	0.9;
	7572	1	32.83641	20.3	0	0	1	1.052178	-19.924975	150	1	1.105648	0.9;
	7573	1	0	0	0	0	1	1.035005	3.901435	380	1	1.106391	0.9;
	7574	1	0	0	0	0	1	1.067355	1.023788	150	1	1.129802	0.9;
	7575	1	50.9565	17	0	0	1	0.972834	-25.109377	110	1	1.1	0.9;
	7576	1	0	0	0.152776	0	1	1.058032	-18.846545	380	1	1.143925	0.9;
	7577	1	2.803108	1	0	0	1	1.031115	-15.471309	110	1	1.124627	0.9;
	7578	1	126.5403	29.1	0	0	1	1.075661	-25.540736	220	1	1.165729	0.9;
	7579	2	0	0	0	0	1	1.072818	-8.962601	220	1	1.131065	0.9;
	7580	1	6.306993	2.5	0	0	1	0.957827	-24.97398	110	1	1.1	0.9;
	7581	1	12.01332	4	0	0	1	0.917387	-29.529842	110	1	1.1	0.9;
	7582	1	77.98647	19.3	0	0	1	1.075018	-6.226738	220	1	1.140548	0.9;
	7583	1	0	0	0	0	1	1.032482	-25.568709	150	1	1.1	0.9;
	7584	1	0	0	0	0	1	1.058465	-8.600663	220	1	1.120948	0.9;
	7585	1	81.19002	23.4	0	0	1	1.051788	33.745546	380	1	1.1	0.9;
	7586	2	0	0	0	0	1	1.101632	-25.12639	380	1	1.157225	0.9;
	7587	1	14.71632	5.3	0	0	1	1.009862	-22.463264	110	1	1.107187	0.9;
	7588	1	216.3	-6.9	0	0	1	1.037847	-47.617195	220	1	1.166435	0.9;
	7589	1	153.4	86.8	0	0	1	1.083492	-38.093723	380	1	1.156212	0.9;
	7590	2	0	0	0	0	1	1.053316	-34.653485	380	1	1.116157	0.9;
	7591	1	0	0	0	0	1	1.102702	-10.257034	380	1	1.132904	0.9;
	7592	2	60.86749	5	0	0	1	0.982818	-21.606809	110	1	1.1	0.9;
	7593	2	0	0	0	0	1	1.074136	-34.708658	220	1	1.129282	0.9;
	7594	1	0	0	0	0	1	0.978297	-27.062887	150	1	1.1	0.9;
	7595	1	0	0	0	0	1	1.04598	14.035313	110	1	1.122412	0.9;
	7596	1	0	0	0	0	1	1.101023	-40.683725	380	1	1.16116	0.9;
	7597	1	0	0	0	0	1	0.978973	-25.674713	110	1	1.1	0.9;
	7598	2	0	0	0	0	1	0.994318	-17.310977	220	1	1.1	0.9;
	7599	1	0	0	0	0	1	1.062528	-15.31383	150	1	1.120496	0.9;
	7600	1	6.73241	14.7	0	0	1	1.04074	-25.025055	150	1	1.1	0.9;
	7601	2	229.3	31.3	0	0	1	1.04305	9.325779	220	1	1.122066	0.9;
	7602	1	96.8	76.1	0.210661	0	1	1.051191	32.290884	380	1	1.107635	0.9;
	7603	1	72.18002	49.6	0	0	1	1.047583	15.045171	150	1	1.1	0.9;
	7604	1	100.8	32.3	0	0	1	1.079078	5.560329	150	1	1.148466	0.9;
	7605	1	0	0	0	0	1	1.037013	14.559277	380	1	1.1	0.9;
	7606	1	70.87859	43.9	0	0	1	1.001996	-23.702615	150	1	1.1	0.9;
	7607	2	117.8	35.7	0	0	1	1.056173	-20.578179	220	1	1.1	0.9;
	7608	1	131.7	32.9	0	0	1	1.092365	-36.974748	380	1	1.155298	0.9;
	7609	1	0	0	0	0	1	1.046565	-25.128289	220	1	1.128151	0.9;
	7610	2	18.42042	7	0	0	1	1.060909	-29.429313	110	1	1.125444	0.9;
	7611	1	79.38802	24.7	0	0	1	1.094682	-38.343434	380	1	1.160269	0.9;
	7612	1	15.81754	9.5	0	0	1	0.989113	-23.768427	150	1	1.1	0.9;
	7613	2	28.03108	58.8	0	0	1	1.076711	5.772751	380	1	1.151294	0.9;
	7614	1	0	0	0	0	1	0.985142	-20.54627	110	1	1.1	0.9;
	7615	1	0	0	0	0	1	0.99592	-26.370981	110	1	1.1	0.9;
	7616	1	0	0	0	0	1	1.095494	57.00054	380	1	1.115263	0.9;
	7617	1	40.74517	12.5	0	0	1	1.047959	-12.304077	220	1	1.12057	0.9;
	7618	1	0	0	0	0	1	1.061317	-23.986739	220	1	1.135537	0.9;
	7619	1	39.94429	13.3	0	0	1	0.908866	-31.88013	110	1	1.1	0.9;
	7620	1	64.57159	23.1	0	0	1	1.024481	-19.443432	110	1	1.113688	0.9;
	7621	1	0	0	0	0	1	1.038831	-25.42002	150	1	1.130152	0.9;
	7622	1	0	0	0	0	1	1.049601	-21.379361	110	1	1.1	0.9;
	7623	1	39.64395	9.6	0	0	1	1.05339	-24.639847	220	1	1.128293	0.9;
	7624	1	33.83752	11.1	0	0	1	1.046555	-30.525491	110	1	1.112194	0.9;
	7625	2	0	0	0	0	1	1.095579	-37.054832	380	1	1.16155	0.9;
	7626	2	0	0	0	0	1	1.091001	28.621944	220	1	1.204972	0.9;
	7627	2	0	0	0	0	1	1.099211	54.075078	380	1	1.110233	0.9;
	7628	1	0	0	0	0	1	1.104595	-18.848715	380	1	1.176777	0.9;
	7629	1	266	62.8	0	0	1	1.028701	-43.805283	220	1	1.1	0.9;
	7630	1	18.52053	6.1	0	0	1	0.985028	-25.139552	110	1	1.1	0.9;
	7631	1	99.1	-98.6	0	0	1	1.057575	-54.651461	220	1	1.1	0.9;
	7632	1	0	0	0	0	1	1.081133	-8.966174	150	1	1.150498	0.9;
	7633	1	223.3	-7.6	0	0	1	1.067078	26.534507	380	1	1.11209	0.9;
	7634	1	19.72187	6.5	0	0	1	0.983248	-24.343244	110	1	1.1	0.9;
	7635	1	0	0	0	0	1	0.963675	-18.286304	110	1	1.1	0.9;
	7636	1	11.21243	4.4	0	0	1	0.992107	-20.354002	110	1	1.1	0.9;
	7637	1	176.2	-46.9	0	0	1	1.042144	-39.856233	220	1	1.1	0.9;
	7638	2	0	0	0	0	1	1.044237	22.308586	380	1	1.109879	0.9;
	7639	1	0	0	0	0	1	1.065797	-12.109524	220	1	1.124758	0.9;
	7640	1	0	0	0	-33.6445	1	1.064764	19.989184	380	1	1.128612	0.9;
	7641	1	88.49812	27.4	0	0	1	0.960132	-26.469884	110	1	1.1	0.9;
	7642	1	0	0	0	0	1	0.937168	38.013814	220	1	1.1	0.9;
	7643	2	0	0	0	0	1	1.053252	9.782193	380	1	1.112565	0.9;
	7644	2	374.4	85.5	0	0	1	1.093026	-29.758002	380	1	1.131888	0.9;
	7645	1	0	0	0	0	1	0.958631	-25.509455	150	1	1.1	0.9;
	7646	1	8.809768	5.5	0	0	1	0.983324	-24.610896	150	1	1.1	0.9;
	7647	1	35.03885	12	0	0	1	0.950821	-29.314368	110	1	1.1	0.9;
	7648	1	125.5	22.5	0	0	1	1.040688	-43.719323	220	1	1.117052	0.9;
	7649	1	380.8	62.8	0	0	1	1.0694	-21.852468	220	1	1.144316	0.9;
	7650	1	31.43485	12.4	0	0	1	1.080422	2.020897	150	1	1.151584	0.9;
	7651	1	332.8	92.2	0	0	1	1.020084	-46.650173	220	1	1.1	0.9;
	7652	1	0	0	0	0	1	1.005994	-22.840332	150	1	1.1	0.9;
	7653	1	-7.6	28.6	0	0	1	0.999227	-16.806682	150	1	1.1	0.9;
	7654	1	33.63729	4.5	0	0	1	1.040382	-22.909294	110	1	1.143926	0.9;
	7655	1	17.01887	5.6	0	0	1	0.958293	-29.384389	110	1	1.1	0.9;
	7656	1	286.1	89.3	0	0	1	1.074238	-19.410514	220	1	1.155965	0.9;
	7657	2	179.4	-18.6	0	0	1	1.04043	9.060658	220	1	1.131856	0.9;
	7658	1	19.82198	5.8	0	0	1	1.016903	-34.249377	110	1	1.1	0.9;
	7659	2	0	0	0	0	1	1.033364	19.78063	220	1	1.1	0.9;
	7660	1	0	0	0	0	1	1.000692	-23.282087	150	1	1.1	0.9;
	7661	1	35.83974	10.9	0	0	1	1.010385	-16.735239	110	1	1.1	0.9;
	7662	1	143.3	71.4	0	0	1	1.052603	-15.382834	380	1	1.117289	0.9;
	7663	1	13.91543	10.1	0	0	1	0.963324	-13.515136	110	1	1.1	0.9;
	7664	2	38.04218	3.2	0	0	1	1.041729	61.691954	150	1	1.1	0.9;
	7665	1	0	0	0	0	1	1.038232	-21.067465	150	1	1.1	0.9;
	7666	2	0	0	0	0	1	1.092553	-9.409611	380	1	1.153366	0.9;
	7667	1	285.2	83.5	0	0	1	0.984937	-39.350744	220	1	1.1	0.9;
	7668	1	0	0	0	0	1	1.044122	-23.317811	150	1	1.1	0.9;
	7669	2	0	0	0	0	1	1.033407	-16.835714	150	1	1.1	0.9;
	7670	1	24.12675	10.8	0	0	1	1.074164	4.307358	150	1	1.144914	0.9;
	7671	2	0	0	0	0	1	1.027909	-16.225632	110	1	1.1	0.9;
	7672	1	0	0	0	0	1	1.087595	66.700021	380	1	1.1	0.9;
	7673	1	0	0	0	0	1	1.075929	-4.465259	220	1	1.1295	0.9;
	7674	1	216.2	75.3	0	0	1	1.03522	-21.64483	220	1	1.1	0.9;
	7675	1	21.42375	11.8	0	0	1	1.061946	-14.159811	380	1	1.12346	0.9;
	7676	1	10.81199	3.9	0	0	1	1.034189	-15.215318	110	1	1.107288	0.9;
	7677	1	0	0	0	0	1	0.981922	-23.577363	220	1	1.1	0.9;
	7678	1	0	0	0	0	1	1.025192	-21.432963	110	1	1.1	0.9;
	7679	2	299.5	48.7	0	0	1	1.0795	-8.240382	220	1	1.146038	0.9;
	7680	2	55.96205	30	0	0	1	1.065447	-17.147078	380	1	1.152008	0.9;
	7681	1	0	0	0	0	1	1.038563	-47.307162	220	1	1.167043	0.9;
	7682	1	17.11898	6.3	0	0	1	0.969371	-17.765795	110	1	1.1	0.9;
	7683	1	0	0	0	0	1	1.083334	-9.560669	220	1	1.154216	0.9;
	7684	1	0	0	0	0	1	1.08444	-8.385942	220	1	1.154573	0.9;
	7685	1	161.2788	54	0	0	1	0.992362	11.001852	220	1	1.1	0.9;
	7686	1	0	0	0	0	1	1.066734	10.142689	110	1	1.145826	0.9;
	7687	2	18.01998	5.9	0	0	1	0.998636	-24.288343	110	1	1.1	0.9;
	7688	1	0	0	0	0	1	0.991609	-16.600289	110	1	1.1	0.9;
	7689	2	0	0	0	0	1	1.082959	16.353661	380	1	1.12786	0.9;
	7690	1	0	0	0	0	1	1.061944	-35.266445	220	1	1.142971	0.9;
	7691	1	149	99.8	0	0	1	1.008193	-7.010549	220	1	1.109546	0.9;
	7692	2	0	0	0	0	1	1.060683	-18.70645	380	1	1.150357	0.9;
	7693	2	51.05661	11.3	0	0	1	1.051092	-4.634811	380	1	1.108779	0.9;
	7694	1	66.67392	0	0	0	1	1.050017	-21.866158	220	1	1.1	0.9;
	7695	1	0	0	0	0	1	1.038849	-25.396036	150	1	1.121931	0.9;
	7696	2	0	0	0	0	1	1.063342	-1.709091	150	1	1.139225	0.9;
	7697	1	98.6	27.1	0	0	1	1.072042	-2.547997	150	1	1.142171	0.9;
	7698	1	0	0	0	0	1	0.983217	-32.700469	220	1	1.1	0.9;
	7699	2	244	38	0	0	1	1.085263	-9.128122	380	1	1.177357	0.9;
	7700	1	9.510545	3.4	0	0	1	0.998665	-16.233295	110	1	1.1	0.9;
	7701	1	0	0	0.128449	0	1	1.040495	-24.067803	220	1	1.2	0.8;
	7702	1	38.84307	14.2	0	0	1	1.077013	-14.061133	220	1	1.15271	0.9;
	7703	2	51.75739	-47	0	0	1	1.060776	53.476393	220	1	1.107023	0.9;
	7704	2	0	0	0	0	1	0.998578	-16.322326	110	1	1.1	0.9;
	7705	1	0	0	0	0	1	1.052269	-33.116498	380	1	1.119468	0.9;
	7706	1	10.11121	0	0	0	1	0.930845	-28.345129	110	1	1.1	0.9;
	7707	1	0	0	0	0	1	0.985505	-25.571139	150	1	1.1	0.9;
	7708	1	97.6	33.9	0	0	1	1.03098	-4.916672	220	1	1.1	0.9;
	7709	1	6.407104	2.2	0	0	1	1.102807	-34.618691	380	1	1.157468	0.9;
	7710	2	0	0	0	0	1	1.011318	-17.158024	110	1	1.109412	0.9;
	7711	1	0	0	0	0	1	0.992019	-20.288129	150	1	1.1	0.9;
	7712	1	37.44151	9.1	0	0	1	1.076422	-25.465226	220	1	1.166424	0.9;
	7713	2	0	0	0	0	1	1.058578	15.102452	150	1	1.1	0.9;
	7714	1	0	0	0	0	1	1.074981	-28.168673	220	1	1.135233	0.9;
	7715	1	0	0	0	0	1	1.019667	11.45253	220	1	1.1	0.9;
	7716	1	93.6	-43.5	0	0	1	1.072978	-23.295778	220	1	1.138149	0.9;
	7717	1	140.7	10.4	0	0	1	1.026879	-43.344842	220	1	1.1	0.9;
	7718	1	13.51498	5.6	0	0	1	0.975933	-21.575455	110	1	1.1	0.9;
	7719	1	30.33363	11	0	0	1	1.024666	-16.825005	110	1	1.127054	0.9;
	7720	2	101.7	59.8	0	0	1	1.073909	-12.030538	220	1	1.153928	0.9;
	7721	1	104.3	12.3	0	0	1	1.017643	9.255273	220	1	1.1	0.9;
	7722	1	268.2	-257.9	0	0	1	1.049345	9.179725	380	1	1.125066	0.9;
	7723	1	13.71521	8.5	0	0	1	1.047017	-21.576427	150	1	1.106928	0.9;
	7724	1	0	0	0	0	1	1.066723	-9.277771	220	1	1.137729	0.9;
	7725	1	53.85971	33.4	0	0	1	1.062602	-31.408036	220	1	1.126054	0.9;
	7726	2	83.99313	19.4	0	0	1	1.062692	-14.201439	220	1	1.118943	0.9;
	7727	1	0	0	0	0	1	1.085557	-36.891738	220	1	1.147889	0.9;
	7728	1	0	-3.7	0	0	1	1.06246	21.122539	110	1	1.114185	0.9;
	7729	1	0	0	0	0	1	1.07926	-19.157083	380	1	1.135906	0.9;
	7730	1	28.13119	10.2	0	0	1	0.978386	-18.246427	110	1	1.1	0.9;
	7731	1	0	0	0	0	1	1.013109	54.199303	220	1	1.1	0.9;
	7732	1	76.98536	-0.7	0	0	1	1.033076	-42.791022	220	1	1.1	0.9;
	7733	1	0	0	0	0	1	1.0342	-42.450979	220	1	1.1	0.9;
	7734	2	0	0	0	0	1	1.037317	-17.632362	220	1	1.1	0.9;
	7735	2	825.87	363.57	0	0	1	1.097605	-15.940599	380	1	1.186976	0.9;
	7736	1	12.31365	4.1	0	0	1	0.987353	-16.973183	110	1	1.1	0.9;
	7737	1	295.2	39.3	0	0	1	1.048142	-39.121414	220	1	1.1	0.9;
	7738	1	72.68058	24.3	0	0	1	1.074112	-20.820286	220	1	1.153657	0.9;
	7739	1	36.4404	22.6	0	0	1	1.024544	-23.904322	150	1	1.1	0.9;
	7740	2	213.3	42.6	0	0	1	1.067722	8.383125	220	1	1.147914	0.9;
	7741	1	0	0	0	0	1	1.032295	-25.224724	150	1	1.1	0.9;
	7742	1	0	0	0	0	1	1.093161	-30.294538	380	1	1.133747	0.9;
	7743	1	0	0	0	0	1	1.030584	-24.499978	150	1	1.1	0.9;
	7744	1	0	0	0	0	1	1.035773	11.357558	220	1	1.1	0.9;
	7745	1	87.69723	26	0	0	1	1.074933	-20.007611	220	1	1.139784	0.9;
	7746	1	0	0	0	0	1	0.985687	8.227515	220	1	1.1	0.9;
	7747	1	9.910988	3.3	0	0	1	1.016979	-35.487098	110	1	1.1	0.9;
	7748	1	5.00555	1.8	0	0	1	1.087215	-11.175905	220	1	1.152605	0.9;
	7749	1	21.22353	9.9	0	0	1	1.028102	-47.701115	220	1	1.1	0.9;
	7750	1	0	0	0	0	1	1.008807	-20.023367	150	1	1.1	0.9;
	7751	1	0	0	0	0	1	1.007328	-19.452709	110	1	1.1	0.9;
	7752	1	0	0	0	0	1	1.032584	-23.222153	150	1	1.1	0.9;
	7753	1	0	0	0.157206	0	1	0.972194	-16.681909	220	1	1.1	0.9;
	7754	1	0	0	0	0	1	1.015992	10.094606	220	1	1.1	0.9;
	7755	2	0	0	0.095323	0	1	1.102263	62.220893	380	1	1.118539	0.9;
	7756	1	0	0	0	0	1	1.040048	-21.315741	150	1	1.116528	0.9;
	7757	1	22.02442	7.7	0	0	1	1.034305	-36.594127	110	1	1.1	0.9;
	7758	1	128.1	42.8	0	0	1	1.074058	-36.482065	380	1	1.12108	0.9;
	7759	1	0	0	0	0	1	0.956986	-34.801477	110	1	1.1	0.9;
	7760	1	0	0	0	0	1	0.961833	-30.882426	150	1	1.1	0.9;
	7761	1	0	-7.5	0	0	1	1.004116	-23.318365	150	1	1.1	0.9;
	7762	1	0	0	0	0	1	1.032176	-17.256556	380	1	1.108861	0.9;
	7763	1	0	0	0	-260.038	1	1.079872	-38.907709	380	1	1.163062	0.9;
	7764	1	0	0	0.415286	0	1	1.071035	-18.616061	380	1	1.141534	0.9;
	7765	2	179.6	3	0	0	1	1.050591	-18.311743	220	1	1.11622	0.9;
	7766	2	0	0	0	0	1	1.081016	5.911702	150	1	1.150039	0.9;
	7767	1	38.04218	15	0	0	1	0.979309	-23.183979	110	1	1.1	0.9;
	7768	1	15.41709	4.4	0	0	1	1.022399	-34.351959	110	1	1.1	0.9;
	7769	1	175.7	56.2	0	0	1	1.064869	-6.053825	220	1	1.120535	0.9;
	7770	2	0	0	0	0	1	1.068079	-34.68332	380	1	1.130804	0.9;
	7771	1	0	0	0	0	1	0.957449	-27.292495	110	1	1.1	0.9;
	7772	1	126.8	21.4	0	0	1	1.056119	-49.060158	220	1	1.113959	0.9;
	7773	2	5.305883	3.2	0	0	1	0.99403	-22.839948	110	1	1.1	0.9;
	7774	1	0	0	0	0	1	1.056489	-44.266306	380	1	1.12519	0.9;
	7775	1	0	0	0.392784	0	1	1.033277	-4.811887	380	1	1.1	0.9;
	7776	1	89.59934	13.1	0	0	1	1.054125	-49.281455	220	1	1.112086	0.9;
	7777	1	10.9121	3.6	0	0	1	1.084448	-27.138586	110	1	1.15104	0.9;
	7778	1	0	0	0	0	1	1.036381	-22.198555	220	1	1.126429	0.9;
	7779	1	9.910988	3.3	0	0	1	1.036267	-31.415666	110	1	1.1	0.9;
	7780	1	0	0	0	0	1	1.090937	-21.656182	380	1	1.153671	0.9;
	7781	1	0	0	0	0	1	1.048604	7.335342	380	1	1.10508	0.9;
	7782	1	27.83086	-2.6	0	0	1	1.058014	-47.036607	220	1	1.109927	0.9;
	7783	1	0	0	0	0	1	1.074551	-9.013996	220	1	1.154476	0.9;
	7784	1	0	0	0	20	1	1.020021	14.051093	380	1	1.1	0.9;
	7785	2	13.6151	5.2	0	0	1	1.012091	-22.37398	110	1	1.1	0.9;
	7786	1	27.53052	17	0	0	1	1.026278	-17.977972	150	1	1.1	0.9;
	7787	1	317.9	55.4	0	0	1	1.036855	-46.062487	380	1	1.1	0.9;
	7788	1	0	0	0	0	1	0.984884	-25.553026	150	1	1.1	0.9;
	7789	1	-33.7374	11.1	0	0	1	1.068557	-17.690625	150	1	1.1	0.9;
	7790	2	12.31365	7.4	0	0	1	1.070867	-14.367371	150	1	1.122305	0.9;
	7791	1	0	0	0	0	1	1.083394	-3.783849	220	1	1.136463	0.9;
	7792	1	171.3	13.9	0	0	1	1.083389	46.039212	380	1	1.1	0.9;
	7793	1	3.804218	3.8	0	0	1	0.950523	-26.91671	110	1	1.1	0.9;
	7794	1	0	0	0	0	1	0.919592	-29.123533	150	1	1.1	0.868718;
	7795	1	0	0	0	0	1	1.065199	-18.958934	220	1	1.1	0.9;
	7796	1	26.92986	9.6	0	0	1	1.006253	-19.193975	110	1	1.1	0.9;
	7797	1	0	0	0	0	1	1.073105	-34.385266	220	1	1.153614	0.9;
	7798	1	0	0	0	0	1	1.07952	-47.003678	380	1	1.136391	0.9;
	7799	1	0	0	0	0	1	0.998159	10.927225	220	1	1.1	0.9;
	7800	2	0	0	0	0	1	1.050133	-14.964503	150	1	1.1	0.9;
	7801	1	0	0	0	0	1	1.04283	12.552693	380	1	1.110196	0.9;
	7802	2	0	0	0	0	1	1.061184	-2.463511	380	1	1.119297	0.9;
	7803	1	0	0	0	19.20831	1	1.020329	-23.565523	150	1	1.1	0.9;
	7804	1	0	0	0	0	1	0.986269	-27.842525	150	1	1.1	0.9;
	7805	2	0	0	0	0	1	1.084263	-34.852165	380	1	1.158657	0.9;
	7806	1	382.4	86.3	0	0	1	1.053621	-21.252085	220	1	1.1	0.9;
	7807	2	471.7872	140.6	0	0	1	1.071227	7.598945	220	1	1.156472	0.9;
	7808	2	77.98647	18.5	0	0	1	1.0848	63.181359	150	1	1.1	0.9;
	7809	1	0	0	0	0	1	0.98565	-30.123712	220	1	1.1	0.9;
	7810	1	101.8	20.2	0	0	1	1.0614	1.716269	150	1	1.121928	0.9;
	7811	1	0	0	0	0	1	1.085775	45.465418	380	1	1.117216	0.9;
	7812	1	0	0	0	0	1	1.054127	-16.034795	150	1	1.109503	0.9;
	7813	1	0	0	0	0	1	1.03055	-24.303096	150	1	1.1	0.9;
	7814	1	0	0	0	0	1	0.937072	-27.775559	110	1	1.1	0.9;
	7815	1	0	0	0	0	1	0.985647	-22.902526	220	1	1.1	0.9;
	7816	1	0	0	0	0	1	1.052079	-15.811523	380	1	1.1	0.9;
	7817	1	76.68502	23.3	0	0	1	1.068006	-9.009397	220	1	1.135212	0.9;
	7818	2	0	0	0	0	1	1.048247	15.429631	150	1	1.1	0.9;
	7819	2	-8.4	-17.5	0	0	1	0.955	-31.167894	110	1	1.1	0.9;
	7820	2	-26.0289	-2.9	0	0	1	1.076633	-20.34272	380	1	1.139296	0.9;
	7821	1	25.02775	18.8	0	0	1	0.99527	10.383534	220	1	1.1	0.9;
	7822	2	0	0	0	0	1	1.087632	-37.204024	380	1	1.13045	0.9;
	7823	1	136	36.1	0	0	1	1.045188	-17.016093	220	1	1.115225	0.9;
	7824	1	0	0	0	0	1	1.051587	-25.547649	220	1	1.160296	0.9;
	7825	1	277.1	37	0	0	1	1.078744	-22.355862	220	1	1.163772	0.9;
	7826	1	98.60933	50.3	0	0	1	0.992136	-39.127026	150	1	1.1	0.9;
	7827	2	0	0	0	0	1	1.101579	-42.912	380	1	1.156839	0.9;
	7828	1	0	0	0	0	1	0.942789	-29.028903	150	1	1.1	0.894323;
	7829	1	0	0	0	0	1	1.07704	-17.728436	150	1	1.1	0.9;
	7830	1	0	0	0.095843	0	1	1.030547	-12.451076	220	1	1.1	0.9;
	7831	1	74.18224	27.3	0	0	1	1.05471	-27.052061	220	1	1.117334	0.9;
	7832	1	282	21.5	0	0	1	1.036206	-44.684746	380	1	1.1	0.9;
	7833	1	31.43485	10.6	0	0	1	1.017205	-21.177503	220	1	1.127997	0.9;
	7834	1	-20.3225	6.7	0	0	1	1.031023	-24.244365	150	1	1.1	0.9;
	7835	1	62.26904	23.8	0	0	1	1.023911	11.766114	220	1	1.1	0.9;
	7836	1	130.2	88.8	0	0	1	1.027528	-4.457385	220	1	1.1	0.9;
	7837	2	0	0	0	0	1	1.104947	-12.711993	380	1	1.186527	0.9;
	7838	1	0	0	0	0	1	1.091011	-24.281821	380	1	1.157365	0.9;
	7839	1	163.2	31.9	0	0	1	1.081643	7.248169	150	1	1.136245	0.9;
	7840	1	29.1323	9.5	0	0	1	0.990753	-24.285592	110	1	1.1	0.9;
	7841	1	0	0	0	0	1	1.103186	-15.328867	380	1	1.184447	0.9;
	7842	2	49.95539	13.2	0	0	1	0.975256	-20.511296	110	1	1.1	0.9;
	7843	1	-2.30255	-0.8	0	0	1	1.063238	-19.994161	110	1	1.12163	0.9;
	7844	1	0	0	0	0	1	1.07209	2.783715	380	1	1.127699	0.9;
	7845	1	0	0	0	0	1	1.031512	-21.814447	150	1	1.108511	0.9;
	7846	1	4.404884	2.2	0	0	1	1.035094	-36.397918	110	1	1.1	0.9;
	7847	2	176.4	50.6	0	0	1	1.061374	-15.646952	220	1	1.11801	0.9;
	7848	1	0	0	0	0	1	1.069616	5.164497	220	1	1.145506	0.9;
	7849	1	44.54939	7.5	0	0	1	1.075277	-15.130769	220	1	1.137331	0.9;
	7850	1	-16.4182	5.4	0	0	1	1.024598	-23.214106	150	1	1.1	0.9;
	7851	1	5.506105	1.3	0	0	1	1.052118	-9.684101	220	1	1.127274	0.9;
	7852	1	38.54273	12	0	0	1	1.052583	-49.288859	220	1	1.110099	0.9;
	7853	1	0	0	0	0	1	1.0065	-18.297165	110	1	1.1	0.9;
	7854	1	129.8	67.2	0	0	1	1.037429	-29.288299	150	1	1.1	0.9;
	7855	1	226.7	39.3	0	0	1	1.065098	-43.731442	220	1	1.115032	0.9;
	7856	2	0	0	0	0	1	1.041657	25.747878	380	1	1.131799	0.9;
	7857	2	0	0	0	0	1	1.055789	-14.272739	380	1	1.12439	0.9;
	7858	1	34.53829	15.4	0	0	1	1.085379	-42.095104	380	1	1.158187	0.9;
	7859	1	45.45039	14.9	0	0	1	0.936675	-26.392245	110	1	1.1	0.9;
	7860	1	1.301443	0.5	0	0	1	0.979065	-17.456048	110	1	1.1	0.9;
	7861	1	9.910988	3.3	0	0	1	0.953332	-35.121241	110	1	1.1	0.9;
	7862	1	69.87748	19.3	0	0	1	1.040882	9.340861	110	1	1.120389	0.9;
	7863	1	74.98314	10.5	0	0	1	1.044281	-14.977518	220	1	1.131085	0.9;
	7864	2	192.1	33.4	0	0	1	1.050455	-6.184885	220	1	1.138724	0.9;
	7865	1	62.7696	5.7	0	0	1	1.054378	-46.390576	220	1	1.1	0.9;
	7866	1	0	0	0	0	1	1.054294	-20.799581	380	1	1.125779	0.9;
	7867	1	0	0	0	0	1	0.985669	-28.277429	150	1	1.1	0.9;
	7868	1	0	0	0	0	1	1.056863	-21.916708	220	1	1.1	0.9;
	7869	2	0	0	0	0	1	1.078409	-21.324325	220	1	1.121204	0.9;
	7870	2	334.5	135.2	0	0	1	1.026955	-20.45425	220	1	1.1	0.9;
	7871	1	0	0	0	0	1	1.064169	-9.488505	220	1	1.131655	0.9;
	7872	1	0	0	0	0	1	1.033686	-21.485099	150	1	1.1	0.9;
	7873	1	0	0	0	0	1	0.994151	-49.60977	220	1	1.1	0.9;
	7874	2	143.8	23.7	0	0	1	1.073818	-23.076418	220	1	1.138856	0.9;
	7875	1	0	0	0	0	1	0.964715	-30.548346	150	1	1.1	0.9;
	7876	2	230.1	62.8	0	0	1	1.033702	-2.709709	380	1	1.11914	0.9;
	7877	1	0	0	0	0	1	1.041174	-25.056668	150	1	1.1	0.9;
	7878	2	0	0	0.38558	0	1	1.041474	16.522751	380	1	1.106564	0.9;
	7879	1	0	0	0	0	1	1.03994	-21.302399	150	1	1.116385	0.9;
	7880	1	42.64728	18.2	0	0	1	1.074883	-19.69327	220	1	1.138965	0.9;
	7881	1	0	0	0	0	1	1.071714	-15.115319	380	1	1.121547	0.9;
	7882	2	315.2	82.6	0	0	1	1.033472	-33.351433	220	1	1.140307	0.9;
	7883	1	0	0	0.095843	0	1	1.068941	-13.056266	220	1	1.128648	0.9;
	7884	2	0	0	0	0	1	1.100605	62.012482	380	1	1.117383	0.9;
	7885	1	0	0	0	-16.4305	1	0.974223	-25.809337	150	1	1.1	0.9;
	7886	1	0	0	0	0	1	1.062038	-11.039225	220	1	1.124208	0.9;
	7887	1	0	0	0	0	1	1.050039	-28.956716	150	1	1.120231	0.9;
	7888	1	20.0222	15	0	0	1	0.997892	10.50336	220	1	1.1	0.9;
	7889	1	89.299	10	0	0	1	1.015465	11.072444	220	1	1.1	0.9;
	7890	1	0	0	0	0	1	1.032769	-25.345353	150	1	1.1	0.9;
	7891	1	82.59157	23.4	0	0	1	1.012719	-39.896648	220	1	1.1	0.9;
	7892	2	12.11343	-4	0	0	1	1.049227	-34.258772	220	1	1.114505	0.9;
	7893	1	0	0	0	0	1	1.032622	-25.38479	150	1	1.1	0.9;
	7894	1	180.7	71.8	0	0	1	1.065912	-7.510331	220	1	1.126006	0.9;
	7895	1	0	0	0	0	1	0.992578	-23.037494	110	1	1.1	0.9;
	7896	1	40.34473	15.9	0	0	1	1.063139	3.422897	150	1	1.133987	0.9;
	7897	1	101.9	24.3	0	0	1	1.096436	-11.291024	220	1	1.153394	0.9;
	7898	1	0	0	0	0	1	1.074177	-33.734182	380	1	1.151379	0.9;
	7899	1	3.704107	1.3	0	0	1	1.031708	-18.753941	110	1	1.118603	0.9;
	7900	1	124.5	-6	0	0	1	1.060386	-43.461523	220	1	1.112131	0.9;
	7901	1	0	0	0	0	1	1.042928	-12.188219	220	1	1.114377	0.9;
	7902	2	4.00444	2.5	0	0	1	0.99443	-25.426334	150	1	1.1	0.9;
	7903	1	0	0	0	0	1	1.030829	17.054451	380	1	1.1	0.9;
	7904	1	0	0	0	0	1	1.099799	-19.585548	380	1	1.154722	0.9;
	7905	1	0	0	0	0	1	1.071995	26.970992	380	1	1.1	0.9;
	7906	1	23.52608	8.5	0	0	1	1.049708	-14.572188	110	1	1.143847	0.9;
	7907	1	15.31698	5	0	0	1	1.056818	-30.009091	110	1	1.121942	0.9;
	7908	1	91.80178	8.3	0	0	1	1.034608	-46.904379	220	1	1.1	0.9;
	7909	1	3.503885	2.7	0	0	1	0.949867	-24.250246	110	1	1.1	0.9;
	7910	2	0	0	0	0	1	1.086	-15.920372	380	1	1.149894	0.9;
	7911	1	0	0	0	0	1	1.074159	-24.168054	220	1	1.162574	0.9;
	7912	1	0	0	0	0	1	1.090118	-16.317473	220	1	1.177744	0.9;
	7913	1	127.9	3.7	0	0	1	1.07014	-20.103597	380	1	1.143215	0.9;
	7914	1	126.3	109.2	0	0	1	1.032163	8.878969	380	1	1.1	0.9;
	7915	1	100.4	1.4	0	0	1	1.029708	-20.505209	220	1	1.1	0.9;
	7916	1	0	0	0	13.29987	1	0.97316	-27.816597	150	1	1.1	0.9;
	7917	1	443.8	172.1	0	0	1	1.045088	4.09307	380	1	1.1	0.9;
	7918	2	12.11343	-1.7	0	0	1	1.061501	-34.538616	220	1	1.115969	0.9;
	7919	1	0	0	0	0	1	1.052147	-19.967	150	1	1.105742	0.9;
	7920	1	191.2	16.4	0	0	1	1.063818	-36.693187	380	1	1.135463	0.9;
	7921	1	66.27348	19.3	0	0	1	1.078495	-20.133188	220	1	1.141312	0.9;
	7922	1	0	0	0.124999	0	1	0.976602	-24.39641	220	1	1.1	0.9;
	7923	1	13.41487	3.4	0	0	1	0.983082	-25.559044	110	1	1.1	0.9;
	7924	1	0	0	0	0	1	0.998013	-23.701566	150	1	1.1	0.9;
	7925	2	61.96871	38.4	0	0	1	1.065867	-21.376259	150	1	1.1	0.9;
	7926	1	0	0	0	0	1	1.070822	4.709742	150	1	1.140736	0.9;
	7927	1	105.8	53.6	0	0	1	1.01751	-29.802069	150	1	1.1	0.9;
	7928	1	2.502775	0.6	0	0	1	0.986168	-24.510276	110	1	1.1	0.9;
	7929	1	0	0	0	0	1	1.094706	-38.343507	380	1	1.160295	0.9;
	7930	1	0	0	0	0	1	1.078208	-8.950081	380	1	1.131692	0.9;
	7931	1	11.61288	7.6	0	0	1	1.025674	-35.332123	220	1	1.134993	0.9;
	7932	1	0	0	0	0	1	0.96916	-30.439998	150	1	1.1	0.9;
	7933	1	0	0	0	0	1	0.997961	-23.67409	150	1	1.1	0.9;
	7934	1	0	0	0	0	1	1.010227	-17.557332	110	1	1.107153	0.9;
	7935	1	0	0	0	0	1	1.07562	-17.781983	150	1	1.1	0.9;
	7936	1	279.7	4.6	0	0	1	1.033185	-49.206309	220	1	1.162446	0.9;
	7937	1	5.105661	2.5	0	0	1	0.96026	-26.585304	110	1	1.1	0.9;
	7938	1	0	0	0	0	1	1.014166	-23.33878	150	1	1.1	0.9;
	7939	1	0	0	0	0	1	1.079857	-15.596473	220	1	1.137165	0.9;
	7940	1	61.98314	39.2	0	0	1	1.028695	-23.034352	150	1	1.1	0.9;
	7941	2	0	0	0	0	1	1.069492	-10.02076	220	1	1.136078	0.9;
	7942	1	30.0333	15	0	0	1	0.980846	-21.936546	110	1	1.1	0.9;
	7943	1	19.02109	3.8	0	0	1	0.91795	-26.310644	110	1	1.1	0.9;
	7944	1	54.6606	18.6	0	0	1	1.055155	-54.394187	220	1	1.1	0.9;
	7945	1	43.04773	25	0	0	1	0.997571	-23.697211	150	1	1.1	0.9;
	7946	2	62.96982	57.7	0	0	1	1.089318	11.826055	220	1	1.165321	0.9;
	7947	1	0	0	0.213168	0	1	1.019908	14.419382	380	1	1.1	0.9;
	7948	1	92.10212	38	0	0	1	0.953979	-19.885643	220	1	1.1	0.9;
	7949	2	18.62065	1.4	0	0	1	0.952927	63.146319	150	1	1.1	0.9;
	7950	2	58.76516	19.5	0	0	1	1.074316	50.229719	380	1	1.120779	0.9;
	7951	1	0	0	0	0	1	1.081457	1.127226	380	1	1.132577	0.9;
	7952	2	201.8	71.2	0	0	1	1.000364	-1.227439	220	1	1.1	0.9;
	7953	1	0	0	0	0	1	1.044495	-14.963367	220	1	1.131311	0.9;
	7954	1	17.51942	6.5	0	0	1	0.983	-21.131124	110	1	1.1	0.9;
	7955	1	25.72853	9.3	0	0	1	1.038067	-19.620026	110	1	1.1	0.9;
	7956	1	0	0	0	0	1	1.056465	-15.360872	150	1	1.108268	0.9;
	7957	2	0	0	0	0	1	1.025733	-33.701191	150	1	1.1	0.9;
	7958	1	0	0	0	0	1	1.060377	7.03155	150	1	1.112042	0.9;
	7959	1	0	0	0	0	1	1.030176	9.798721	380	1	1.1	0.9;
	7960	1	0	0	0	0	1	1.065395	-31.185137	220	1	1.128468	0.9;
	7961	1	45.25017	19.1	0	0	1	1.008244	7.364908	220	1	1.1	0.9;
	7962	1	0	0	0	0	1	1.02166	-34.02022	220	1	1.1	0.9;
	7963	1	7.608436	6.3	0	0	1	0.970825	-24.835647	110	1	1.1	0.9;
	7964	1	35.33918	37.8	0	0	1	1.041703	-1.023587	220	1	1.113138	0.9;
	7965	1	292.4	-228.7	0	0	1	1.079587	15.977341	380	1	1.13422	0.9;
	7966	1	1.801998	0.8	0	0	1	1.000852	-16.223379	110	1	1.1	0.9;
	7967	1	8.809768	2.2	0	0	1	1.002374	-36.772954	110	1	1.1	0.9;
	7968	1	26.22908	9.5	0	0	1	1.040943	-22.019814	110	1	1.141186	0.9;
	7969	1	11.51276	3.8	0	0	1	0.965797	-28.169504	110	1	1.1	0.9;
	7970	1	100.5	19.3	0	0	1	1.065165	9.501807	150	1	1.1	0.9;
	7971	1	0	0	0	0	1	1.085509	-2.640478	220	1	1.146194	0.9;
	7972	1	128.8	54.5	0	0	1	0.995444	-25.923321	220	1	1.1	0.9;
	7973	1	22.82531	8	0	0	1	1.090603	-16.358197	220	1	1.187286	0.9;
	7974	2	148.2	48.9	0	0	1	1.069628	-5.775412	220	1	1.124347	0.9;
	7975	1	-29.733	23.6	0	0	1	1.024867	-42.158637	150	1	1.1	0.9;
	7976	1	0	0	0	0	1	1.071238	-21.462865	220	1	1.163684	0.9;
	7977	1	107.9	35.9	0	0	1	1.065801	2.272319	150	1	1.13848	0.9;
	7978	1	3.804218	-12.7	0	0	1	1.036497	-11.81928	110	1	1.1	0.9;
	7979	2	0	0	0	0	1	1.094409	-0.147898	220	1	1.157539	0.9;
	7980	1	0	0	0	0	1	1.024545	-27.438145	150	1	1.1	0.9;
	7981	1	80.58935	5.2	0	0	1	1.038163	-4.602183	220	1	1.107571	0.9;
	7982	1	24.02664	5.6	0	0	1	1.011276	-27.557302	110	1	1.1	0.9;
	7983	1	9.910988	3.3	0	0	1	1.001546	-39.259042	110	1	1.1	0.9;
	7984	1	31.03441	8.9	0	0	1	1.067962	-13.226561	110	1	1.13275	0.9;
	7985	1	0	0	0	0	1	1.087148	66.625518	380	1	1.1	0.9;
	7986	2	0	0	0	0	1	1.042342	-22.080817	110	1	1.14238	0.9;
	7987	1	22.02442	4.4	0	0	1	1.051726	-9.519561	220	1	1.122766	0.9;
	7988	1	35.23907	11.4	0	0	1	1.031233	-31.097442	110	1	1.1	0.9;
	7989	1	0	0	0	-262.278	1	1.089345	-39.544751	380	1	1.166258	0.9;
	7990	2	373	131.1	0	0	1	1.078864	-3.883273	220	1	1.131979	0.9;
	7991	2	226.9	29.2	0	0	1	1.074093	-25.095199	220	1	1.164781	0.9;
	7992	1	101.2	-2.9	0	0	1	1.034579	-32.362074	220	1	1.133884	0.9;
	7993	1	0.300333	1.5	0	0	1	1.0736	0.708263	220	1	1.14909	0.9;
	7994	1	47.3525	15.5	0	0	1	1.036382	-36.471747	110	1	1.1	0.9;
	7995	1	32.23574	10.1	0	0	1	0.978502	-21.030806	110	1	1.1	0.9;
	7996	1	0	0	0	0	1	0.977343	-27.137047	150	1	1.1	0.9;
	7997	1	89.89967	29.3	0	0	1	1.067649	3.054146	150	1	1.127241	0.9;
	7998	2	12.51387	3.5	0	0	1	0.935847	-25.027893	110	1	1.1	0.9;
	7999	1	82.19112	15.6	0	0	1	1.025375	9.794922	220	1	1.1	0.9;
	8000	2	37.3414	12.1	0	0	1	1.080636	0.673375	220	1	1.149097	0.9;
	8001	2	47.65283	14	0	0	1	1.05607	-48.454129	220	1	1.112632	0.9;
	8002	2	0	0	0.155331	0	1	1.045818	-17.262589	220	1	1.119981	0.9;
	8003	1	8.309213	3	0	0	1	1.050056	-21.075801	110	1	1.15084	0.9;
	8004	2	289.7	128.5	0	0	1	1.080474	3.89002	380	1	1.151756	0.9;
	8005	1	0	0	0	0	1	0.99153	-25.177518	150	1	1.1	0.9;
	8006	1	0	0	0	0	1	1.047984	-3.535117	380	1	1.108788	0.9;
	8007	1	75.3	7.1	0	0	1	1.017321	-21.973398	110	1	1.1	0.9;
	8008	1	0	0	0	0	1	1.003204	-24.362746	150	1	1.1	0.9;
	8009	1	0	0	0	0	1	1.054681	4.984542	150	1	1.109136	0.9;
	8010	1	91.60156	27.1	0	0	1	1.02052	-20.651688	220	1	1.130977	0.9;
	8011	1	21.32364	7.2	0	0	1	1.0715	-19.566487	110	1	1.127826	0.9;
	8012	1	26.32919	10.3	0	0	1	0.911886	-29.617218	110	1	1.1	0.9;
	8013	1	0	0	0	0	1	1.062625	-1.767803	150	1	1.138568	0.9;
	8014	1	0	0	0	0	1	1.037675	-29.396728	150	1	1.1	0.9;
	8015	1	0	0	0	0	1	1.060812	-21.1531	220	1	1.143823	0.9;
	8016	1	113.7	22.3	0	0	1	1.062666	-21.890087	220	1	1.140125	0.9;
	8017	1	54.56049	16.7	0	0	1	0.969425	-20.557007	110	1	1.1	0.9;
	8018	1	13.31476	4.4	0	0	1	1.063941	-32.0004	110	1	1.132583	0.9;
	8019	1	20.0222	12.4	0	0	1	0.964642	-30.559805	150	1	1.1	0.9;
	8020	1	86.7	41.5	0	0	1	1.013479	-27.244362	150	1	1.1	0.9;
	8021	1	1.501665	0.3	0	0	1	0.981495	-25.800237	110	1	1.1	0.9;
	8022	1	8.309213	2.7	0	0	1	0.948422	-32.347301	110	1	1.1	0.9;
	8023	1	0	0	0	0	1	0.995949	-22.378624	150	1	1.1	0.9;
	8024	1	208.7	63	0	0	1	1.052087	-10.852874	220	1	1.131861	0.9;
	8025	1	1.301443	0	0	0	1	0.950547	-26.909288	110	1	1.1	0.9;
	8026	1	13	8.8	0	20	1	1.084076	-17.818149	380	1	1.132105	0.9;
	8027	1	20.82309	0	0	0	1	1.044537	8.12219	220	1	1.117919	0.9;
	8028	1	0	0	0.160798	0	1	1.054534	-3.186033	380	1	1.108991	0.9;
	8029	1	0	0	0	0	1	0.935328	-28.779899	150	1	1.1	0.879727;
	8030	1	4.404884	2.2	0	0	1	1.007001	-35.27348	110	1	1.1	0.9;
	8031	1	7.408214	2.4	0	0	1	1.031978	-27.670676	110	1	1.1	0.9;
	8032	2	0	0	0	0	1	1.094263	64.204498	380	1	1.107153	0.9;
	8033	1	6.00666	2.2	0	0	1	1.02125	-23.626294	110	1	1.120767	0.9;
	8034	1	410.3	91.6	0	0	1	1.049564	-35.900593	220	1	1.155746	0.9;
	8035	1	69.1767	35.1	0	0	1	1.025821	-42.016464	220	1	1.117695	0.9;
	8036	1	88.4	51.1	0.154992	0	1	1.068084	-12.27681	380	1	1.122733	0.9;
	8037	2	0	0	0	0	1	1.044737	-22.311979	150	1	1.1	0.9;
	8038	1	0	0	0	0	1	1.05306	-44.281193	220	1	1.105963	0.9;
	8039	2	207.4	44.3	0	0	1	1.068682	-39.334757	220	1	1.114923	0.9;
	8040	1	0	0	0	0	1	1.023521	-17.638061	220	1	1.1	0.9;
	8041	1	57.16338	19.7	0	0	1	1.069108	-9.42227	220	1	1.136295	0.9;
	8042	1	0	0	0	0	1	1.029005	-49.632939	220	1	1.1	0.9;
	8043	1	50.9565	15.7	0	0	1	1.053324	9.692441	110	1	1.135707	0.9;
	8044	1	55.5616	34.4	0	0	1	0.999928	-23.123736	150	1	1.1	0.9;
	8045	1	0	0	0	0	1	1.050466	-31.190276	220	1	1.138624	0.9;
	8046	1	50.9565	16.7	0	0	1	0.961147	-26.670409	110	1	1.1	0.9;
	8047	1	0	0	0	0	1	1.067086	-8.324229	220	1	1.129247	0.9;
	8048	1	131.5	12.9	0	0	1	1.064667	-44.102968	220	1	1.113365	0.9;
	8049	1	0	0	0	0	1	1.060015	2.362681	380	1	1.11529	0.9;
	8050	2	0	0	0	0	1	1.095868	-38.146922	380	1	1.161346	0.9;
	8051	1	0	0	0	0	1	1.092439	-43.647235	380	1	1.146714	0.9;
	8052	2	426.6	32.4	0	0	1	1.059636	-47.870938	220	1	1.115133	0.9;
	8053	2	0	0	0	0	1	1.085789	-15.942866	380	1	1.149683	0.9;
	8054	1	0	0	0	0	1	1.077931	-26.448027	220	1	1.186095	0.9;
	8055	1	285.2	61.4	0	0	1	1.047472	-20.467454	220	1	1.136151	0.9;
	8056	2	0	0	0	0	1	1.095	15.496741	150	1	1.141752	0.9;
	8057	1	0	0	0	0	1	1.064729	-38.431376	380	1	1.129696	0.9;
	8058	1	0	0	0	0	1	1.078881	-19.654388	220	1	1.142284	0.9;
	8059	1	58.2646	16.9	0	0	1	1.042939	17.311515	110	1	1.111787	0.9;
	8060	1	0	0	0	0	1	1.072496	-21.590631	220	1	1.155686	0.9;
	8061	1	9.210212	3.4	0	0	1	1.057372	-14.890969	110	1	1.116015	0.9;
	8062	1	0	0	0	0	1	1.078443	-0.216923	150	1	1.146517	0.9;
	8063	2	0	0	0	0	1	1.069696	8.608812	110	1	1.139261	0.9;
	8064	1	0	0	0	0	1	1.095527	-28.581964	380	1	1.148012	0.9;
	8065	1	35.63951	10.7	0	0	1	0.91545	-26.556458	110	1	1.1	0.9;
	8066	1	0	0	0.155011	0	1	1.105762	-9.663358	380	1	1.163925	0.9;
	8067	1	0	0	0	0	1	1.011466	-3.800408	380	1	1.1	0.9;
	8068	1	0	0	0	0	1	1.072842	1.035004	380	1	1.121338	0.9;
	8069	1	101.6	4.5	0	0	1	1.095752	-40.374829	220	1	1.152942	0.9;
	8070	2	0	0	0	0	1	1.081905	5.962573	150	1	1.150865	0.9;
	8071	1	22.12453	8	0	0	1	1.011218	-18.984452	110	1	1.1	0.9;
	8072	1	18.32031	6.6	0	0	1	1.080389	-17.70468	110	1	1.122413	0.9;
	8073	1	0	0	0	0	1	1.010388	-19.589756	150	1	1.1	0.9;
	8074	1	278.3	33.6	0	0	1	1.044626	-14.868073	220	1	1.131348	0.9;
	8075	2	0	0	0	0	1	1.093	-24.810567	380	1	1.156554	0.9;
	8076	2	0	0	0	0	1	1.064532	-2.42322	150	1	1.142407	0.9;
	8077	2	48.95428	21.4	0	0	1	1.060545	-20.586915	110	1	1.117873	0.9;
	8078	1	0	0	0	0	1	1.082684	16.349883	380	1	1.127585	0.9;
	8079	1	0	0	0.224857	0	1	1.074535	-15.644916	380	1	1.123085	0.9;
	8080	1	3.103441	1.6	0	0	1	1.076886	-22.486738	220	1	1.162268	0.9;
	8081	1	36.03996	22.3	0	0	1	0.971733	-24.413773	150	1	1.1	0.9;
	8082	1	5.506105	2.2	0	0	1	1.037351	-29.982488	110	1	1.1	0.9;
	8083	2	0	0	0	0	1	1.071053	-9.61605	380	1	1.119646	0.9;
	8084	1	24.62731	15.3	0	0	1	1.038232	-21.067465	150	1	1.1	0.9;
	8085	1	0	0	0.107995	0	1	1.063665	-14.756731	380	1	1.114238	0.9;
	8086	1	91.0009	21	0	0	1	1.060341	-22.031311	220	1	1.137916	0.9;
	8087	1	29.23241	-49.9	0	0	1	1.062144	-24.695257	220	1	1.126899	0.9;
	8088	1	0	0	0	0	1	1.023286	-21.143715	150	1	1.1	0.9;
	8089	1	33.53718	12.2	0	0	1	1.05662	-3.573828	110	1	1.130552	0.9;
	8090	1	-6.80755	2.2	0	0	1	1.042268	-16.539218	150	1	1.1	0.9;
	8091	1	0	0	0	0	1	1.077564	-19.549069	220	1	1.141331	0.9;
	8092	2	0	0	0	0	1	0.956818	-29.259213	110	1	1.1	0.9;
	8093	1	0	0	0	0	1	0.998475	10.35515	220	1	1.1	0.9;
	8094	1	0	0	0	0	1	1.075364	-9.355022	220	1	1.144789	0.9;
	8095	1	59.66615	13.1	0	0	1	1.010281	2.059626	110	1	1.1	0.9;
	8096	1	10.11121	2.5	0	0	1	1.009649	-19.200664	110	1	1.1	0.9;
	8097	2	99.10989	-12.2	0	0	1	1.053731	62.38817	150	1	1.1	0.9;
	8098	2	0	0	0	0	1	1.014909	-36.682899	110	1	1.1	0.9;
	8099	1	0	0	0	0	1	1.039718	-4.363515	380	1	1.1	0.9;
	8100	1	3.403774	1.1	0	0	1	0.992026	-16.446788	110	1	1.1	0.9;
	8101	1	7.408214	2.7	0	0	1	1.030875	-22.924726	110	1	1.13219	0.9;
	8102	2	88.2979	8.9	0	0	1	1.074792	-5.68719	220	1	1.131985	0.9;
	8103	1	6.407104	2.3	0	0	1	1.011733	-17.159783	110	1	1.109652	0.9;
	8104	1	67.07436	35.5	0	0	1	1.031156	-51.268939	220	1	1.1	0.9;
	8105	2	0	0	0	0	1	1.0132	-30.7442	150	1	1.1	0.9;
	8106	1	93.9	21.6	0	0	1	1.029489	-2.444312	220	1	1.107483	0.9;
	8107	1	139.4	21.2	0	0	1	1.041968	-17.942946	220	1	1.112405	0.9;
	8108	2	0	0	0	0	1	1.08598	-8.94169	380	1	1.178	0.9;
	8109	1	0	-7.5	0	0	1	1.005429	-23.167425	150	1	1.1	0.9;
	8110	1	0	0	0	0	1	1.013028	-46.709563	220	1	1.1	0.9;
	8111	2	0	0	0	0	1	0.98531	38.285983	220	1	1.1	0.9;
	8112	1	0	0	0	0	1	1.084586	-12.689974	380	1	1.128848	0.9;
	8113	1	0	0	0	0	1	1.056554	-48.659741	220	1	1.114506	0.9;
	8114	2	0	0	0	0	1	1.000447	57.085217	220	1	1.1	0.9;
	8115	1	186.8	22.9	0	0	1	1.051129	-49.822438	220	1	1.1	0.9;
	8116	1	15.81754	3.7	0	0	1	1.002844	-24.28335	110	1	1.1	0.9;
	8117	1	4.204662	1.4	0	0	1	0.989132	-28.609077	110	1	1.1	0.9;
	8118	1	19.92209	6.6	0	0	1	0.945036	-27.143723	110	1	1.1	0.9;
	8119	1	0	0	0	0	1	0.940679	-23.49502	220	1	1.1	0.9;
	8120	2	0	0	0	0	1	1.096545	60.64197	220	1	1.105462	0.9;
	8121	2	1.601776	88.5	0	0	1	1.087868	-37.44758	380	1	1.128252	0.9;
	8122	1	3.704107	0.2	0	0	1	1.044678	-11.48146	220	1	1.11695	0.9;
	8123	1	0	0	0	0	1	1.064686	4.528	380	1	1.112212	0.9;
	8124	1	20.32253	6.3	0	0	1	1.001838	-23.378143	110	1	1.1	0.9;
	8125	2	300.1013	73.7	0	0	1	1.051227	-14.017195	220	1	1.105577	0.9;
	8126	1	0	0	0	0	1	1.001813	-48.215202	220	1	1.1	0.9;
	8127	1	142.2	33.7	0	0	1	1.024892	24.648782	220	1	1.1	0.9;
	8128	1	516.7	144.3	0	0	1	1.035583	-36.288156	220	1	1.1	0.9;
	8129	1	51.55716	18.7	0	0	1	1.047183	-21.614604	110	1	1.10804	0.9;
	8130	1	0	0	0	0	1	1.011186	7.885466	220	1	1.1	0.9;
	8131	2	0	0	0	0	1	1.099658	20.464932	380	1	1.136907	0.9;
	8132	1	0	0	0	0	1	1.029104	-49.564321	220	1	1.1	0.9;
	8133	1	28.63174	17.7	0	0	1	1.038627	-24.436385	150	1	1.1	0.9;
	8134	1	0	0	0	0	1	0.989566	-24.003447	150	1	1.1	0.9;
	8135	1	0	0	0	0	1	1.046134	20.389885	380	1	1.10516	0.9;
	8136	2	100.7	20.5	0	0	1	1.081545	-12.826258	220	1	1.135528	0.9;
	8137	1	44.04884	21	0	0	1	1.013844	-47.890567	220	1	1.1	0.9;
	8138	1	0	0	0	0	1	1.035293	-36.30173	220	1	1.1	0.9;
	8139	2	926.5	150.4	0	0	1	1.071689	-12.898043	150	1	1.143967	0.9;
	8140	1	45.04995	12.7	0	0	1	1.029273	-31.122251	110	1	1.1	0.9;
	8141	2	0	0	0	0	1	1.104632	43.061304	380	1	1.11917	0.9;
	8142	1	23.42597	8.4	0	0	1	1.00713	-19.855801	110	1	1.1	0.9;
	8143	1	261.5	95	0	0	1	1.059399	-10.106515	380	1	1.126872	0.9;
	8144	1	0	0	0	0	1	0.997589	-26.478235	110	1	1.1	0.9;
	8145	1	-17.3192	5.7	0	0	1	1.01087	-21.971601	150	1	1.1	0.9;
	8146	2	298.3991	122.5	0	0	1	1.050409	1.041234	220	1	1.125021	0.9;
	8147	2	0	0	0	0	1	0.982545	-28.584165	110	1	1.1	0.9;
	8148	2	274.3	269.4	0	0	1	1.096921	-32.610613	380	1	1.13559	0.9;
	8149	1	0	0	0	0	1	1.087024	-30.413379	380	1	1.139047	0.9;
	8150	1	0	0	0	0	1	1.074101	-23.078964	220	1	1.139155	0.9;
	8151	2	0	0	0	0	1	1.061364	-15.912717	220	1	1.113299	0.9;
	8152	1	81.29012	26.5	0	0	1	1.063173	0.743397	150	1	1.125605	0.9;
	8153	2	60.56715	30.7	0	0	1	1.059364	-3.580923	110	1	1.134746	0.9;
	8154	1	0	0	0	0	1	0.991526	-25.179296	150	1	1.1	0.9;
	8155	1	63.97093	12.8	0	0	1	1.046298	-11.104949	220	1	1.117991	0.9;
	8156	2	0	0	0	0	1	1.044682	-38.334545	220	1	1.1	0.9;
	8157	1	0	0	0	0	1	1.018874	14.12435	220	1	1.1	0.9;
	8158	1	4.104551	1.5	0	0	1	1.037368	-18.535832	110	1	1.122679	0.9;
	8159	1	18.52053	11.1	0	0	1	1.024776	-24.481037	150	1	1.1	0.9;
	8160	1	0	0	0	11.28922	1	1.031048	-24.244347	150	1	1.1	0.9;
	8161	1	92.9	48.4	0	0	1	1.064739	-19.569367	220	1	1.1	0.9;
	8162	2	44.94984	21.2	0	0	1	1.094211	-37.531797	380	1	1.155094	0.9;
	8163	1	88.49812	24.5	0	0	1	1.053885	-24.414641	220	1	1.127868	0.9;
	8164	1	2.00222	1	0	0	1	1.014227	-21.513386	110	1	1.1	0.9;
	8165	2	0	0	0	0	1	1.054081	2.726228	150	1	1.111473	0.9;
	8166	1	0	0	0	0	1	1.08043	-3.800409	220	1	1.133528	0.9;
	8167	2	79.38802	16.8	0	0	1	1.07709	-34.232316	220	1	1.129768	0.9;
	8168	1	69.27681	31.3	0	0	1	1.073591	4.470344	150	1	1.144254	0.9;
	8169	2	0	0	0	0	1	1.071605	-13.681846	380	1	1.128264	0.9;
	8170	1	0	0	0	0	1	0.986882	-39.01979	220	1	1.1	0.9;
	8171	1	41.14562	10.3	0	0	1	1.0507	-20.628728	220	1	1.120234	0.9;
	8172	2	0	0	0	0	1	1.066579	-38.174065	380	1	1.134739	0.9;
	8173	2	79.2879	46.6	0	0	1	1.052226	-37.430032	220	1	1.1	0.9;
	8174	2	120.6	13.2	0	0	1	1.069094	-9.737902	220	1	1.136414	0.9;
	8175	1	0	0	0	0	1	1.063716	41.03181	380	1	1.121694	0.9;
	8176	1	102.6	14.7	0	0	1	1.058372	-21.152208	220	1	1.1	0.9;
	8177	1	0	0	0.059549	0	1	1.047705	-19.018944	220	1	1.117252	0.9;
	8178	1	0	0	0	0	1	1.058271	-43.759826	380	1	1.126136	0.9;
	8179	1	73.68169	24	0	0	1	1.058203	0.803908	150	1	1.119528	0.9;
	8180	1	27.13008	16.8	0	0	1	1.009885	-26.520065	150	1	1.1	0.9;
	8181	1	145.3	3.8	0	0	1	1.074189	-33.366814	380	1	1.151268	0.9;
	8182	1	0	0	0.269993	0	1	1.083008	-15.730141	380	1	1.135028	0.9;
	8183	1	88.39802	25.2	0	0	1	1.062003	-22.003471	220	1	1.139598	0.9;
	8184	1	25.62842	6.5	0	0	1	1.079411	1.100004	220	1	1.152197	0.9;
	8185	2	0	0	0	0	1	1.057409	-43.805072	220	1	1.108518	0.9;
	8186	1	70.77847	19.6	0	0	1	1.096386	-6.272965	220	1	1.163634	0.9;
	8187	2	20.52275	5.6	0	0	1	0.993061	-23.211833	110	1	1.1	0.9;
	8188	1	0	0	0.275618	0	1	1.04871	-20.663115	380	1	1.13308	0.9;
	8189	1	5.105661	1.3	0	0	1	0.994078	-23.711618	110	1	1.1	0.9;
	8190	1	55.46149	16.9	0	0	1	1.038574	-6.218962	220	1	1.125181	0.9;
	8191	1	91.40134	18.1	0	0	1	1.07238	-23.457163	220	1	1.137624	0.9;
	8192	1	6.407104	4.3	0	0	1	0.937092	-25.78259	110	1	1.1	0.9;
	8193	2	154.4	23.3	0	0	1	1.06336	-7.235824	220	1	1.131117	0.9;
	8194	1	80.48924	30	0	0	1	1.014056	-18.538633	220	1	1.1	0.9;
	8195	1	0	0	0.14273	0	1	1.093078	-0.707594	380	1	1.143278	0.9;
	8196	1	0	0	0.177458	0	1	1.028243	-4.853285	380	1	1.1	0.9;
	8197	1	0.200222	0.1	0	0	1	1.067656	-19.81109	110	1	1.126034	0.9;
	8198	2	0	0	0	0	1	1.085211	5.318795	380	1	1.154968	0.9;
	8199	1	81.99091	32.2	0	0	1	1.060651	-48.074084	220	1	1.11789	0.9;
	8200	1	0	0	0	-18.3313	1	1.043028	14.666794	150	1	1.1	0.9;
	8201	1	132.2	49.1	0	0	1	1.07445	-12.743261	380	1	1.139919	0.9;
	8202	1	0	0	0	0	1	1.101192	-43.269517	380	1	1.156912	0.9;
	8203	1	76.88525	16.4	0	0	1	1.049195	-26.22954	220	1	1.128434	0.9;
	8204	2	33.73741	7.4	0	0	1	1.011273	-14.597546	110	1	1.1	0.9;
	8205	2	0	0	0	0	1	1.024332	-21.820854	110	1	1.1	0.9;
	8206	2	0	0	0.199988	0	1	1.122614	58.578636	380	1	1.180138	0.9;
	8207	1	0	0	0	0	1	1.067096	-14.605988	150	1	1.118571	0.9;
	8208	2	45.04995	-2.2	0	0	1	1.050818	-38.327103	220	1	1.105348	0.9;
	8209	2	151	30.9	0	0	1	1.04259	3.013761	220	1	1.134304	0.9;
	8210	1	0	0	0	0	1	0.966304	-25.099908	150	1	1.1	0.9;
	8211	1	0	0	0	0	1	1.078763	-18.798191	380	1	1.140493	0.9;
	8212	2	44.6495	-1.6	0	0	1	1.064364	-36.381225	220	1	1.12647	0.9;
	8213	1	0	0	0	0	1	1.034491	-28.952818	220	1	1.1	0.9;
	8214	1	60.56715	18.9	0	0	1	1.071869	-23.412372	220	1	1.135882	0.9;
	8215	2	0	0	0	0	1	1.095079	45.243438	380	1	1.116693	0.9;
	8216	2	0	0	0	0	1	1.005474	66.939571	380	1	1.1	0.9;
	8217	2	0	0	0	0	1	1.082909	-11.177499	220	1	1.137779	0.9;
	8218	1	36.84085	14.1	0	0	1	1.057766	-36.59422	220	1	1.105931	0.9;
	8219	1	0	0	0	26.62507	1	0.950061	-31.621698	150	1	1.1	0.84698;
	8220	1	0	0	0	0	1	1.088469	-0.354946	220	1	1.148841	0.9;
	8221	1	0	0	0	0	1	1.049868	57.886868	220	1	1.1	0.9;
	8222	2	27.93097	11.1	0	0	1	1.082184	-18.818858	380	1	1.15956	0.9;
	8223	2	100.9	2.9	0	0	1	1.033909	-45.163717	220	1	1.1	0.9;
	8224	2	21.22353	8.9	0	0	1	1.013411	-22.130474	110	1	1.1	0.9;
	8225	1	72.38026	23	0	0	1	1.063822	-43.892268	220	1	1.112999	0.9;
	8226	2	42.44706	-1.8	0	0	1	0.952933	63.823203	150	1	1.1	0.9;
	8227	1	0	0	0	0	1	1.079275	-26.357005	220	1	1.187432	0.9;
	8228	2	492.4	162	0	0	1	1.040684	1.603051	380	1	1.1	0.9;
	8229	1	86.1	40.6	0	0	1	1.032264	-21.776765	150	1	1.109205	0.9;
	8230	2	0	0	0	0	1	1.077674	10.049499	150	1	1.131408	0.9;
	8231	1	0	0	0	0	1	1.032217	-4.652475	380	1	1.1	0.9;
	8232	1	0	0	0	0	1	1.081054	0.960033	380	1	1.132136	0.9;
	8233	2	25.62842	4.6	0	0	1	1.04416	-19.575934	220	1	1.1	0.9;
	8234	2	0	0	0	0	1	1.093636	0.519822	220	1	1.152622	0.9;
	8235	1	0	0	0	0	1	1.045899	-44.79094	220	1	1.115188	0.9;
	8236	1	0	0	0	0	1	1.008849	-20.013564	150	1	1.1	0.9;
	8237	1	0	0	0	0	1	1.081385	3.916054	150	1	1.149486	0.9;
	8238	1	0	0	0	0	1	0.966122	-33.947664	220	1	1.1	0.9;
	8239	1	242.4	5.1	0	0	1	1.080007	-41.441262	220	1	1.151532	0.9;
	8240	1	0	0	0	0	1	0.967309	-27.980684	150	1	1.1	0.9;
	8241	1	213.6002	50.1	0	0	1	0.961361	20.813525	220	1	1.1	0.9;
	8242	1	24.32697	9.9	0	0	1	0.956998	-25.096292	110	1	1.1	0.9;
	8243	1	0	0	0	0	1	0.99237	-23.526985	150	1	1.1	0.9;
	8244	1	188.6	164.9	0	0	1	1.066257	-11.958341	380	1	1.132494	0.9;
	8245	1	0	0	0	0	1	1.056455	-15.363437	150	1	1.108198	0.9;
	8246	1	88.09768	36	0	0	1	1.0834	0.054101	150	1	1.152392	0.9;
	8247	1	0	0	0	0	1	1.062113	21.417942	380	1	1.130677	0.9;
	8248	1	125	9.6	0	0	1	1.049284	-12.544159	220	1	1.114157	0.9;
	8249	1	55.9	102.9	0	0	1	1.063629	36.415633	380	1	1.121299	0.9;
	8250	1	0	0	0	0	1	0.970616	53.860989	220	1	1.1	0.9;
	8251	1	0	0	0	0	1	1.075332	-9.043212	220	1	1.14204	0.9;
	8252	1	13.94129	9.5	0	0	1	0.988461	-23.641662	150	1	1.1	0.9;
	8253	1	3.804218	2.5	0	0	1	1.029457	-12.369751	110	1	1.1	0.9;
	8254	1	187.4	72.8	0	0	1	1.07285	4.058076	150	1	1.143824	0.9;
	8255	1	119	97.8	0	0	1	0.98992	10.850414	220	1	1.1	0.9;
	8256	2	0	0	0	0	1	1.055409	-35.999143	220	1	1.108513	0.9;
	8257	1	0	0	0	0	1	1.002812	-26.075246	150	1	1.1	0.9;
	8258	2	0	0	0	0	1	1.098421	62.813841	380	1	1.114128	0.9;
	8259	2	140.8	44.9	0	0	1	1.052447	-22.814783	380	1	1.116957	0.9;
	8260	1	11.31254	13.8	0	0	1	0.995816	-22.6443	150	1	1.1	0.9;
	8261	2	0	-0.3	0	0	1	1.100091	-12.432246	110	1	1.165432	0.9;
	8262	1	0	0	0	0	1	1.029746	-46.548332	220	1	1.1	0.9;
	8263	1	87.3969	35.5	0	0	1	1.082951	-16.206345	380	1	1.150684	0.9;
	8264	2	59.26571	13.3	0	0	1	1.058727	-15.293683	220	1	1.111296	0.9;
	8265	1	0	0	0	0	1	1.088472	-0.354998	220	1	1.148843	0.9;
	8266	2	-61.7	-30.3	0	0	1	1.095105	-14.491761	380	1	1.154822	0.9;
	8267	1	35.13896	12.7	0	0	1	1.020789	-15.470646	110	1	1.121563	0.9;
	8268	1	0	0	0	0	1	1.06035	-12.521323	220	1	1.144158	0.9;
	8269	2	0	0	0	0	1	1.060455	-35.317994	220	1	1.142176	0.9;
	8270	1	25.02775	12	0	0	1	1.050516	-28.312129	220	1	1.112784	0.9;
	8271	1	152.7	81	0	0	1	1.054259	-12.368299	380	1	1.123212	0.9;
	8272	1	36.74074	14.1	0	0	1	1.082202	-47.458259	380	1	1.119834	0.9;
	8273	2	0	0	0	0	1	1.068143	3.051821	150	1	1.127711	0.9;
	8274	1	0	0	0	0	1	1.077513	-19.53412	220	1	1.14174	0.9;
	8275	1	0	0	0	0	1	1.058624	2.100167	150	1	1.118282	0.9;
	8276	1	349.6	96.3	0	0	1	1.060169	-42.331484	380	1	1.128951	0.9;
	8277	1	50.45594	-10.8	0	0	1	1.095592	-21.484045	380	1	1.2	0.8;
	8278	1	22.02442	7.1	0	0	1	1.010927	-17.766758	110	1	1.1	0.9;
	8279	2	0	0	0	0	1	1.041182	-17.102385	220	1	1.112945	0.9;
	8280	1	0	0	0	0	1	1.050444	-32.049573	380	1	1.123508	0.9;
	8281	1	38.84307	9.6	0	0	1	1.026687	16.067632	110	1	1.1	0.9;
	8282	1	50.55605	8.7	0	0	1	1.081366	-3.662194	220	1	1.134837	0.9;
	8283	2	255.8	17.7	0	0	1	1.050377	-30.360076	220	1	1.143582	0.9;
	8284	1	0	0	0	0	1	1.061598	-22.96586	220	1	1.135675	0.9;
	8285	2	0	0	0	0	1	1.089921	-27.715846	380	1	1.181268	0.9;
	8286	1	0	0	0	0	1	1.060836	-40.06031	380	1	1.113463	0.9;
	8287	1	21.72409	13.4	0	0	1	1.020236	-17.246564	150	1	1.1	0.9;
	8288	1	169.0395	33.9	0	0	1	1.002942	-49.292834	220	1	1.1	0.9;
	8289	1	0	0	0	0	1	1.033673	-49.283158	220	1	1.1	0.9;
	8290	2	253.4	111.3	0	0	1	1.041591	-29.35046	220	1	1.109579	0.9;
	8291	1	431.5	111.5	0	0	1	1.035947	-46.107892	380	1	1.1	0.9;
	8292	2	158.4	31	0	0	1	1.052273	-4.904432	220	1	1.141204	0.9;
	8293	1	0	0	0	19.10083	1	1.023365	-27.887148	150	1	1.1	0.9;
	8294	1	0	0	0	0	1	1.028044	-17.562426	150	1	1.1	0.9;
	8295	1	89.6	96.2	0	0	1	1.045777	9.749847	380	1	1.105093	0.9;
	8296	1	0	0	0.237999	0	1	1.042538	-10.035334	220	1	1.178121	0.9;
	8297	1	66.57381	23.6	0	0	1	1.075223	-17.126714	110	1	1.125085	0.9;
	8298	1	35.43929	11.4	0	0	1	1.041688	4.767588	110	1	1.107069	0.9;
	8299	1	0	0	0	0	1	1.026573	-23.587433	150	1	1.1	0.9;
	8300	1	15.21687	5.1	0	0	1	1.090398	-12.643457	110	1	1.155498	0.9;
	8301	1	0	0	0	0	1	0.978165	-24.260489	150	1	1.1	0.9;
	8302	2	0	0	0	0	1	1.079368	-12.725638	380	1	1.151031	0.9;
	8303	1	0	0	0	0	1	1.068566	15.373576	380	1	1.137569	0.9;
	8304	1	2.202442	1.1	0	0	1	1.021591	-34.200998	110	1	1.1	0.9;
	8305	1	83.19224	37.3	0	0	1	1.03978	18.071349	380	1	1.1	0.9;
	8306	1	0	0	0	0	1	1.048081	-19.630212	220	1	1.106963	0.9;
	8307	1	0	8.5	0	0	1	1.04852	-20.896839	150	1	1.1	0.9;
	8308	1	0	0	0	22.88222	1	1.024191	-25.664082	150	1	1.1	0.9;
	8309	2	60.36693	10.5	0	0	1	1.087296	-13.811918	220	1	1.146764	0.9;
	8310	1	0	0	0	0	1	1.090956	-39.526529	380	1	1.154013	0.9;
	8311	1	311.4	310.5	0.196284	0	1	1.055714	0.620613	380	1	1.112481	0.9;
	8312	2	1.601776	5.3	0	0	1	1.062091	-19.238092	220	1	1.1	0.9;
	8313	1	94.9	40.3	0	0	1	1.05684	-55.008171	220	1	1.1	0.9;
	8314	1	0	0	0	19.91451	1	1.043936	-23.359098	150	1	1.1	0.9;
	8315	1	7.708547	2.4	0	0	1	1.037159	-18.357961	220	1	1.1	0.9;
	8316	1	244.6	96.3	0	0	1	1.023849	-16.414638	220	1	1.1	0.9;
	8317	1	9.910988	2	0	0	1	1.012705	-22.246851	110	1	1.1	0.9;
	8318	2	0	0	0	0	1	1.093316	-23.769098	380	1	1.149513	0.9;
	8319	2	0	0	0	0	1	1.054273	-14.551096	220	1	1.122964	0.9;
	8320	2	66.67392	11.1	0	0	1	1.081298	-16.854834	220	1	1.120699	0.9;
	8321	1	43.14784	26.7	0	0	1	1.03078	-24.13502	150	1	1.1	0.9;
	8322	1	6.00666	3	0	0	1	1.037232	-20.045549	110	1	1.1	0.9;
	8323	1	0	0	0	0	1	0.970908	-27.421821	150	1	1.1	0.9;
	8324	1	66.07326	15.6	0	0	1	1.068256	-19.219378	220	1	1.136584	0.9;
	8325	2	489.8	162	0	0	1	1.048053	-30.655625	380	1	1.132466	0.9;
	8326	1	20.32253	6.7	0	0	1	1.003446	-29.805364	110	1	1.1	0.9;
	8327	1	0	0	0	0	1	1.060584	-12.523938	220	1	1.14441	0.9;
	8328	1	11.41265	3.7	0	0	1	1.013882	-31.626891	110	1	1.1	0.9;
	8329	1	4.605106	0.8	0	0	1	1.040546	-16.246448	110	1	1.108934	0.9;
	8330	2	0	0	0	0	1	1.002513	-23.21089	150	1	1.1	0.9;
	8331	2	0	0	0	0	1	1.094545	-8.235816	220	1	1.156161	0.9;
	8332	2	8.00888	3	0	0	1	1.05987	-8.820727	220	1	1.120373	0.9;
	8333	1	156.3	34.9	0	0	1	1.097858	-25.438895	380	1	1.148216	0.9;
	8334	1	50.0555	44	0	0	1	1.057413	-19.408473	220	1	1.108632	0.9;
	8335	1	66.97426	18.6	0	0	1	1.097086	-6.218882	220	1	1.16428	0.9;
	8336	1	-28.3314	9.3	0	0	1	1.00693	-23.441528	150	1	1.1	0.9;
	8337	2	0	0	0	0	1	1.054455	-26.939908	220	1	1.117053	0.9;
	8338	1	62.56937	-61	0	0	1	1.058564	-23.873119	220	1	1.13308	0.9;
	8339	1	0	0	0	0	1	1.032749	12.769564	220	1	1.1	0.9;
	8340	1	0	0	0	0	1	1.075648	-23.931953	220	1	1.163893	0.9;
	8341	1	17.91987	23.3	0	0	1	1.074698	-10.969203	150	1	1.145656	0.9;
	8342	1	0	0	0	0	1	1.016285	-21.508683	110	1	1.1	0.9;
	8343	1	0	0	0.239999	0	1	1.061978	-14.473086	380	1	1.1	0.9;
	8344	2	35.63951	6.4	0	0	1	1.026364	-23.315703	110	1	1.1	0.9;
	8345	2	19.82198	2.2	0	0	1	1.019769	-37.65866	110	1	1.1	0.9;
	8346	2	0	0	0	0	1	1.050545	-44.538257	220	1	1.1	0.9;
	8347	1	282.2	268.3	0	0	1	0.987658	11.111357	220	1	1.1	0.9;
	8348	1	14.41598	3.8	0	0	1	0.906258	-32.501794	110	1	1.1	0.9;
	8349	2	0	0	0	0	1	1.036136	11.403902	220	1	1.1	0.9;
	8350	1	137.9	25.6	0	0	1	1.008321	-39.126182	220	1	1.1	0.9;
	8351	1	15.41709	6.6	0	0	1	1.021896	-34.467865	110	1	1.1	0.9;
	8352	1	0	0	0	0	1	1.072181	-11.842489	380	1	1.125901	0.9;
	8353	1	11.71299	3.8	0	0	1	0.999188	-22.496658	110	1	1.1	0.9;
	8354	1	23.42597	7.3	0	0	1	1.041686	9.135988	110	1	1.123229	0.9;
	8355	1	226.1	88.8	0	0	1	1.070636	2.368796	150	1	1.142793	0.9;
	8356	1	0	0	0	0	1	1.073607	0.708259	220	1	1.149096	0.9;
	8357	2	16.21798	4.1	0	0	1	0.988967	-23.281738	110	1	1.1	0.9;
	8358	1	12.11343	4.4	0	0	1	0.950757	-18.035931	110	1	1.1	0.9;
	8359	1	217.7	164.6	0	0	1	1.059573	6.001438	380	1	1.124409	0.9;
	8360	1	14.81643	0.6	0	0	1	1.046554	-21.578615	110	1	1.1	0.9;
	8361	1	0	0	0	0	1	1.071313	-4.366309	380	1	1.133304	0.9;
	8362	1	1.301443	0	0	0	1	0.93279	-29.207007	110	1	1.1	0.9;
	8363	2	23.52608	9.6	0	0	1	0.926182	33.151862	220	1	1.1	0.9;
	8364	1	103.6	4.7	0	0	1	1.065035	-43.922407	220	1	1.113613	0.9;
	8365	1	0	0	0	0	1	1.021215	-8.650177	220	1	1.118932	0.9;
	8366	1	0	0	0.059549	0	1	1.043021	-19.618724	220	1	1.114143	0.9;
	8367	1	39.34362	13.3	0	0	1	0.936897	-28.577632	110	1	1.1	0.9;
	8368	1	0	0	0	0	1	1.031955	12.609782	380	1	1.1	0.9;
	8369	2	0	0	0	0	1	1.084862	-16.115269	380	1	1.146672	0.9;
	8370	1	133.5	24.1	0	0	1	1.026656	3.68985	110	1	1.1	0.9;
	8371	1	0	0	0	0	1	1.039053	-4.159838	220	1	1.105493	0.9;
	8372	1	0	0	0	10.83751	1	1.052202	-19.925009	150	1	1.105648	0.9;
	8373	1	0	0	0.095843	0	1	1.031138	-13.821291	220	1	1.106732	0.9;
	8374	1	0	0	0	85.96878	1	1.033054	66.109598	380	1	1.1	0.9;
	8375	1	38.34251	12.6	0	0	1	1.057549	-29.735084	110	1	1.122387	0.9;
	8376	2	415.8	126.6	0	0	1	1.006409	-6.553767	220	1	1.1	0.9;
	8377	2	255.7	65.6	0	0	1	1.076407	-18.154082	220	1	1.141485	0.9;
	8378	1	31.93541	8.8	0	0	1	1.049423	-28.756936	110	1	1.1	0.9;
	8379	1	157	44.2	0	0	1	1.082508	-2.447694	220	1	1.143768	0.9;
	8380	2	2.00222	0	0	0	1	1.057568	0.748584	220	1	1.12632	0.9;
	8381	2	5.606215	2.2	0	0	1	1.083789	-20.896509	380	1	1.144211	0.9;
	8382	1	246.5	-1.9	0	0	1	1.08234	-37.978446	380	1	1.154807	0.9;
	8383	1	0	0	0	0	1	1.042532	-18.820599	110	1	1.128176	0.9;
	8384	2	0	0	0	0	1	1.104105	16.485935	380	1	1.169088	0.9;
	8385	2	0	0	0	0	1	1.086053	1.575069	380	1	1.135073	0.9;
	8386	1	0	0	0	0	1	1.026495	-24.551994	150	1	1.1	0.9;
	8387	1	0	0	0	0	1	1.051742	-9.519694	220	1	1.122783	0.9;
];

%% generator data
%	bus	Pg	Qg	Qmax	Qmin	Vg	mBase	status	Pmax	Pmin	Pc1	Pc2	Qc1min	Qc1max	Qc2min	Qc2max	ramp_agc	ramp_10	ramp_30	ramp_q	apf
mpc.gen = [
	5	1044.4	140.4	524.5259	-211.567	1.060105	100	1	1200	399.999996	0	0	0	0	0	0	0	0	0	0	0;
	10	31.25	13.32	Inf	-Inf	1.045776	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	14	467.5	104	212.965	-93.4811	1.040474	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	20	319.6	86.8	179.4347	-66.2055	1.012955	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	27	8	0.8	0.8	0.8	1.020864	100	1	8	8	0	0	0	0	0	0	0	0	0	0	0;
	28	5.3	0.1	0.1	0.1	1.025773	100	1	5.3	5.3	0	0	0	0	0	0	0	0	0	0	0;
	34	34.3	1.2	17.7615	-7.14375	1.023133	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	35	8.01	3.26	Inf	-Inf	0.952528	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	42	309.2	59.2	218.5186	-75.9716	1.051136	100	1	480	0	0	0	0	0	0	0	0	0	0	0	0;
	43	7.3	1.8	1.8	1.8	1.026	100	1	7.3	7.3	0	0	0	0	0	0	0	0	0	0	0;
	51	4.87	2.96	Inf	-Inf	1.030763	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	70	180.5	28.3	139.9369	-43.7779	1.022545	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	71	0.02	0.52	Inf	-Inf	1.048726	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	79	1.95	1.11	Inf	-Inf	1.031659	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	85	19.73	8.33	Inf	-Inf	1.068345	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	96	570	10.2	253.8527	-113.787	1.103763	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	109	24.9	10.2	18.84201	-6.06272	1.066727	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	110	861.3	112.3	440.32	-172.589	1.066	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	111	4.79	1.72	Inf	-Inf	1.0287	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	121	5.05	3.39	Inf	-Inf	1.029075	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	125	27.7	5.9	18.59004	-6.32799	1.002727	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	126	158.7	14.4	78.14512	-32.3995	1.080818	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	131	144	4.8	70.12466	-29.5005	1.095727	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	134	79.6	9.6	77.09313	-22.5412	1.0975	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	144	4.64	2.37	Inf	-Inf	1.073994	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	147	292.2	63.3	127.6701	-56.5257	1.030091	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	152	2.2	0.94	Inf	-Inf	1.0096	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	153	14.02	5.75	Inf	-Inf	1.0196	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	156	0.69	5.46	Inf	-Inf	1.008891	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	159	31.2	3.1	52.63864	-51.7172	1.089842	100	1	158.0808	0	0	0	0	0	0	0	0	0	0	0	0;
	169	19.85	6.42	Inf	-Inf	0.998323	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	170	99.9	20.7	54.23686	-20.6272	0.993364	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	172	135.8	13	91.61313	-30.9805	1.087684	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	174	268.6	61.3	130.1877	-53.534	1.075867	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	175	2.64	3.05	Inf	-Inf	1.092354	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	177	292.2	-21.9	123.8859	-59.5392	1.105553	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	181	5.2	0.3	0.3	0.3	1.057864	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	186	0.42	0	Inf	-Inf	1.03615	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	188	25.4	5.3	18.70203	-6.16587	0.972727	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	191	123.91	52.96	Inf	-Inf	1.040354	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	193	164.8	64.9	191.9957	-52.889	1.056395	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	197	11.16	5.88	Inf	-Inf	1.049023	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	198	1171	381.4	937.433	-288.159	1.079342	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	199	668.7	-45.9	402.9197	-146.951	1.0985	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	200	296	62	183.9209	-65.2995	1.076684	100	1	400	133.33333	0	0	0	0	0	0	0	0	0	0	0;
	201	28.9	2.9	12.50527	-5.62181	0.944667	100	1	29.60204	9.867347	0	0	0	0	0	0	0	0	0	0	0;
	209	2.13	0.78	Inf	-Inf	1.011746	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	210	0.13	1	Inf	-Inf	1.04248	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	217	39	15.8	38.52133	-11.2425	1.036909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	224	439	127.6	378.0823	-112.594	1.057909	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	226	52.2	9.6	37.47434	-12.3498	1.0115	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	232	423.2	37.6	220.6567	-85.886	1.102342	100	1	500	166.666667	0	0	0	0	0	0	0	0	0	0	0;
	235	0.68	14.79	Inf	-Inf	1.033422	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	238	45.4	9.6	38.09511	-11.7155	1.028136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	243	0.13	1	Inf	-Inf	1.079235	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	253	49.2	10	37.70184	-12.0883	0.979909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	257	1295.6	65.4	924.6445	-300.349	1.104	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	263	52.7	-0.5	26.01981	-10.8118	1.090773	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	272	1204	144.4	721.2503	-258.712	1.078658	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	274	14.85	4.95	Inf	-Inf	0.97336	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	281	95.1	21.4	42.62279	-18.7213	1.014045	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	282	-2.1	286.38	99999	-99999	1.092632	100	1	-2.1	-2.1	0	0	0	0	0	0	0	0	0	0	0;
	284	0.16	2.91	Inf	-Inf	1.075726	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	292	94	22.2	95.31674	-27.2949	1.087711	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	295	213.6	26.5	188.9056	-56.1422	0.948091	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	303	-1.81	1.5	Inf	-Inf	1.030147	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	304	83.7	12.6	55.54403	-19.1586	1.076818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	308	12.84	3.2	Inf	-Inf	1.060331	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	313	160.4	20.8	110.5715	-36.9041	1.102211	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	316	0.15	1.93	Inf	-Inf	1.029313	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	317	1250.4	256.2	926.3492	-297.678	1.092132	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	320	4.86	2.15	Inf	-Inf	0.975854	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	325	238.3	26.9	186.3895	-58.5035	1.095395	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	326	0.03	0.68	Inf	-Inf	0.97953	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	334	318.8	103.6	282.9458	-84.3021	1.092053	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	338	0.18	1	Inf	-Inf	1.064448	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	339	106.95	44.5	Inf	-Inf	0.999614	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	343	42.2	17.4	38.39833	-11.4385	1.090682	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	348	0.19	4.38	Inf	-Inf	1.086247	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	350	69	6	56.72846	-17.8435	1.0236	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	355	2404.4	187.9	1224.256	-490.165	1.082	100	1	2800	933.333331	0	0	0	0	0	0	0	0	0	0	0;
	361	30.67	11.83	Inf	-Inf	0.997732	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	362	6.57	4.31	Inf	-Inf	1.021904	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	365	9.81	3.56	Inf	-Inf	1.012912	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	367	30.4	-0.7	22.21964	-17.3452	0.949182	100	1	61.22614	0	0	0	0	0	0	0	0	0	0	0	0;
	369	3.31	1.62	Inf	-Inf	1.034669	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	370	35.9	4.4	17.65988	-7.27293	1.024636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	371	231	24.5	124.1169	-49.8432	1.014864	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	381	636.3	103.6	304.3814	-125.481	1.070579	100	1	700	233.333331	0	0	0	0	0	0	0	0	0	0	0;
	388	142.98	57.87	Inf	-Inf	1.059877	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	390	436	105.5	217.1751	-89.2166	1.047842	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	394	74	7	45.10484	-16.1804	1.021182	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	395	308.6	-28.2	175.156	-69.2068	1.098682	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	396	83.1	9.9	55.72715	-19.0053	1.085545	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	398	4.5	0.6	0.6	0.6	1.0075	100	1	4.5	4.5	0	0	0	0	0	0	0	0	0	0	0;
	402	5.37	2.33	Inf	-Inf	1.045887	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	403	39.9	3.6	16.84024	-8.04071	1.001545	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	408	47.6	11.8	37.86198	-11.9408	1.012818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	409	260.1	75.6	257.2252	-74.3989	1.077842	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	410	6.88	3.44	Inf	-Inf	1.057826	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	411	42.5	6.1	44.79799	-39.8056	0.996818	100	1	129.18183	0	0	0	0	0	0	0	0	0	0	0	0;
	427	16.34	5.12	Inf	-Inf	1.045996	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	438	229.5	-11.1	132.4063	-50.8474	0.989364	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	441	3.8	2.48	Inf	-Inf	1.011891	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	442	8.5	1.1	1.1	1.1	1.031	100	1	8.5	8.5	0	0	0	0	0	0	0	0	0	0	0;
	444	105.4	38	94.56444	-28.1008	1.099	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	451	135.89	60.88	Inf	-Inf	1.032798	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	459	19.83	5.95	Inf	-Inf	0.960088	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	462	101.4	5.8	94.3813	-28.0441	1.078553	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	466	129.6	25.6	72.50375	-27.2399	1.075364	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	467	539.4	133.2	261.7932	-106.67	1.075818	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	469	192.6	28	85.62814	-37.2118	1.031395	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	503	0.18	3.57	Inf	-Inf	1.04818	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	509	6.9	3	18.46584	-19.3502	1.040727	100	1	56.816328	0	0	0	0	0	0	0	0	0	0	0	0;
	522	0.13	1	Inf	-Inf	1.078858	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	523	55.7	4.4	25.87924	-11.0132	0.997091	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	527	33.6	7.3	17.88329	-7.03048	1.057636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	531	249	66.8	185.5624	-59.3858	1.0595	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	534	400.8	78.8	380.3121	-110.078	1.094921	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	557	38.2	-3.3	17.0903	-7.78858	1.096591	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	558	379.2	55.2	217.8958	-81.1598	1.079273	100	1	480	159.999996	0	0	0	0	0	0	0	0	0	0	0;
	559	9.73	6.29	Inf	-Inf	0.973935	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	560	671.2	61.6	351.176	-139.111	1.082368	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	569	152.4	5.7	78.788	-31.6917	0.992455	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	576	109.8	21.6	52.99446	-21.8745	1.056636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	577	20.59	8.08	Inf	-Inf	1.020676	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	578	133.7	56	135.5178	-38.9046	1.050773	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	580	1.23	-0.1	Inf	-Inf	1.081123	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	581	38.7	4	90.12852	-93.0288	1.085368	100	1	276.09134	0	0	0	0	0	0	0	0	0	0	0	0;
	584	0.59	0.83	Inf	-Inf	1.05456	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	586	47.2	5.6	37.9476	-11.8693	1.054773	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	589	17.53	3.75	Inf	-Inf	0.950047	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	593	0.15	1.93	Inf	-Inf	1.029293	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	603	5.61	2.79	Inf	-Inf	1.07087	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	608	67	11.9	71.32494	-62.9243	1.063974	100	1	205.80612	0	0	0	0	0	0	0	0	0	0	0	0;
	618	37.4	12.2	38.76588	-11.0648	1.084227	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	622	20.7	5.2	34.74669	-34.262	1.053045	100	1	104.27135	0	0	0	0	0	0	0	0	0	0	0	0;
	627	274.8	-4.8	182.1713	-62.7871	1.09875	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	631	0.61	1.21	Inf	-Inf	1.073545	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	632	634.8	61.4	357.3998	-132.748	1.096316	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	633	274.8	8.2	129.9381	-53.9068	1.016727	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	643	16	1.4	22.22661	-20.8341	1.085395	100	1	65.29948	0	0	0	0	0	0	0	0	0	0	0	0;
	645	451.2	139.2	377.0113	-113.626	1.055184	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	647	99.3	9.3	53.98251	-20.7936	0.991455	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	648	453.4	122.2	376.3604	-114.083	1.092237	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	651	17.19	5.2	Inf	-Inf	1.039331	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	652	0.54	0.99	Inf	-Inf	1.043796	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	659	419.2	174.6	379.6141	-111.091	1.097447	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	665	96.77	44.54	Inf	-Inf	1.04125	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	666	14.71	4.51	Inf	-Inf	1.033004	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	668	102	22	107.8562	-95.1534	1.016545	100	1	311.391	34.599	0	0	0	0	0	0	0	0	0	0	0;
	669	467	46	212.9697	-93.4506	1.079789	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	677	49.2	10.4	37.77862	-12.0423	0.989364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	679	215.4	61.2	105.6865	-43.844	0.995136	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	681	208.5	30	136.9403	-47.2439	1.023545	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	686	116.7	11.4	51.29568	-23.4138	1.066816	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	692	37.3	6.7	27.88098	-8.97053	1.081455	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	693	361.7	56.6	174.172	-71.125	1.000409	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	694	170.7	3.9	76.76137	-33.8382	1.062591	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	698	0.43	0	Inf	-Inf	1.065508	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	702	14.87	4.17	Inf	-Inf	0.988124	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	703	55.14	27.14	Inf	-Inf	1.081084	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	704	30.8	3.3	18.16351	-6.7311	1.063773	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	705	16.5	2	12.36838	-9.36122	1.078	100	1	33.58376	0	0	0	0	0	0	0	0	0	0	0	0;
	706	37.8	8.7	17.34023	-7.58033	1.068842	100	1	45	0	0	0	0	0	0	0	0	0	0	0	0;
	707	50.8	3.2	37.43079	-12.3168	1.024136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	711	26	14.72	Inf	-Inf	1.076152	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	716	0.15	1	Inf	-Inf	1.077179	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	717	621.8	175.4	567.628	-167.477	1.103342	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	720	58.4	9.8	36.75392	-13.0447	1.073864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	721	31.6	14.2	39.0043	-10.7449	1.014545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	728	111.2	22.4	74.03435	-25.5438	1.038955	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	732	1820	154	864.4675	-360.968	1.085842	100	1	2000	666.666665	0	0	0	0	0	0	0	0	0	0	0;
	736	-17.83	9.5	Inf	-Inf	1.059217	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	739	-3.15	1.79	Inf	-Inf	1.03953	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	746	1274.2	277.8	926.6834	-298.308	1.084658	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	752	85.2	10.2	55.58738	-19.18	0.999909	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	753	507	76.4	264.7092	-102.832	1.013227	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	757	44	4.2	27.19461	-9.66223	1.066136	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	764	97.5	15.9	54.35089	-20.4699	1.091289	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	767	66.8	3.9	49.63444	-38.1223	1.045867	100	1	136.7347	22.789117	0	0	0	0	0	0	0	0	0	0	0;
	768	8.7	1.1	1.1	1.1	1.087553	100	1	8.7	8.7	0	0	0	0	0	0	0	0	0	0	0;
	770	94	32.5	96.63999	-27.8097	1.065727	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	776	855.6	159.3	440.6747	-172.154	1.081211	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	778	55.8	15.9	57.97606	-16.677	1.054545	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	787	11.81	4.86	Inf	-Inf	1.057058	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	790	343.5	24	228.1617	-78.185	1.081263	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	801	2.85	7.24	Inf	-Inf	1.032767	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	803	199.4	58.9	138.6001	-45.2708	1.061636	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	806	1702.8	411.6	1097.413	-373.912	1.068474	100	1	2400	799.999998	0	0	0	0	0	0	0	0	0	0	0;
	807	56.7	22.2	58.05788	-16.6605	1.044273	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	809	289.6	55.2	141.9267	-57.7392	1.016409	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	815	4.94	3.67	Inf	-Inf	0.996099	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	825	1249.4	419	772.113	-269.937	1.047395	100	1	1700	566.666665	0	0	0	0	0	0	0	0	0	0	0;
	828	231	20	134.4455	-49.7625	1.079591	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	829	32	6.4	18.22456	-6.72812	1.014182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	838	27.4	9.3	28.69322	-8.15604	1.067318	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	841	680.7	26.7	406.6568	-144.751	1.099553	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	864	10.1	3.95	Inf	-Inf	1.066541	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	877	24.29	8.74	Inf	-Inf	0.992788	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	882	1221.2	346.2	934.725	-291.521	1.101711	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	884	90.6	12	54.3648	-20.2118	0.989091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	886	21.3	1.8	36.01688	-35.3457	1.084026	100	1	107.8283	0	0	0	0	0	0	0	0	0	0	0	0;
	887	70.12	21.9	Inf	-Inf	1.091167	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	888	14.9	1.6	1.6	1.6	1.016636	100	1	14.9	14.9	0	0	0	0	0	0	0	0	0	0	0;
	892	1034.8	111.2	948.357	-277.27	1.101921	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	893	1029.2	138.4	738.935	-240.835	1.097158	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	895	20.86	4.14	Inf	-Inf	1.072235	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	900	39.8	2.3	16.8293	-8.04593	0.980455	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	901	53.3	6.7	25.96651	-10.8717	1.062909	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	902	34.1	3.9	17.84731	-7.07292	0.996763	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	910	92.1	29.1	55.01811	-19.8133	1.074682	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	911	171	28	89.55361	-35.1511	1.080409	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	913	86.4	22.7	43.98878	-17.3978	0.994409	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	920	0.2	1	Inf	-Inf	1.064428	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	922	1.04	0.6	Inf	-Inf	1.014597	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	930	48.6	10.5	58.33813	-16.2312	1.010682	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	935	5.91	2.33	Inf	-Inf	1.036036	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	937	138.6	11	90.64089	-31.7295	1.047364	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	938	130.4	11.6	71.66254	-27.8871	1.028955	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	943	10.1	5.4	Inf	-Inf	1.057995	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	944	202.5	10.5	137.3939	-46.4522	1.065182	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	951	829.5	71.1	387.8762	-163.395	1.079053	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	955	0.43	0.73	Inf	-Inf	0.970967	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	957	484.82	93.74	173.501	-71.7856	1.076682	100	1	518.72	252.053333	0	0	0	0	0	0	0	0	0	0	0;
	981	18.55	1.66	Inf	-Inf	1.097121	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	986	6.14	1.95	Inf	-Inf	1.090124	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	987	364.4	7.6	171.6218	-73.4993	1.074737	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	988	91.8	21	54.91012	-19.8779	1.059409	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	995	24.85	3.56	Inf	-Inf	1.03649	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	997	837	135.6	550.6898	-185.152	1.088579	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	1001	139.6	46.2	91.41771	-31.2436	1.073	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1003	58.2	7.4	36.80028	-13.0061	1.079182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1005	56.22	14.15	38.05292	-11.5858	1.014591	100	1	94.82	14.82	0	0	0	0	0	0	0	0	0	0	0;
	1006	419	4.4	376.6637	-112.659	1.105132	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	1013	75.26	27.37	Inf	-Inf	1.008854	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1020	298.7	48.1	126.491	-57.3311	1.078545	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1022	301.2	28.2	158.2033	-62.8062	1.075	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	1023	75.2	10.2	68.00069	-56.0836	1.090727	100	1	190.2778	0	0	0	0	0	0	0	0	0	0	0	0;
	1024	9	10.19	Inf	-Inf	1.054154	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1026	129.6	27.6	114.5684	-34.7375	1.045455	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1029	37.83	16.27	Inf	-Inf	1.06054	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1030	667.5	11.1	407.6605	-143.61	1.085658	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1036	930.3	134.3	432.4411	-180.696	1.081658	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	1037	194.4	44.4	107.8861	-41.504	1.0165	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1043	111.7	28.6	153.0727	-144.243	1.082591	100	1	452.5252	37.710433	0	0	0	0	0	0	0	0	0	0	0;
	1064	9.51	3.12	Inf	-Inf	1.013588	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1067	4.4	0.2	0.2	0.2	1.027727	100	1	4.4	4.4	0	0	0	0	0	0	0	0	0	0	0;
	1068	26.8	2.3	18.61375	-6.27599	1.022467	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1069	180.8	22	108.3462	-39.116	1.078727	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1070	99.4	19.8	94.12515	-28.0909	1.015	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1071	2.35	2.06	Inf	-Inf	1.000338	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1078	329.2	100	178.017	-67.5522	1.068421	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1096	0.19	1	Inf	-Inf	1.063104	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1101	0.59	2.55	Inf	-Inf	1.09671	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1109	107.4	33.3	84.07069	-26.4829	1.023727	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	1113	282.8	28	118.8385	-55.6368	1.091909	100	1	290.43	0	0	0	0	0	0	0	0	0	0	0	0;
	1115	0.18	1	Inf	-Inf	1.064464	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1123	1504.8	284.8	906.6108	-319.442	1.100237	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	1127	16	1.1	22.29117	-20.9811	1.079727	100	1	65.63264	0	0	0	0	0	0	0	0	0	0	0	0;
	1129	81.7	2	44.29841	-17.0198	1.086316	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	1130	9.78	3.51	Inf	-Inf	1.014258	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1136	837	134.7	550.6898	-185.152	1.088632	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	1143	71	8.6	34.913	-14.828	1.014591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1152	158.4	48.4	191.3772	-52.8968	1.008409	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1156	19.1	2.7	32.54085	-32.1088	1.003864	100	1	97.7041	0	0	0	0	0	0	0	0	0	0	0	0;
	1162	0.2	1	Inf	-Inf	1.064562	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1172	0.19	78.58	Inf	-Inf	1.080647	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1178	57	9.2	36.92151	-12.8801	1.055364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1185	19.83	6.01	Inf	-Inf	0.968529	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1190	69.2	10	35.48049	-14.3412	1.080632	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1195	36.9	4.4	17.51154	-7.41836	1.022455	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1197	43.2	12.8	38.25805	-11.5515	1.085868	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1202	279.6	46.8	160.6208	-60.2991	1.023136	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	1207	-37.69	19.49	Inf	-Inf	1.049571	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1212	0.17	1.82	Inf	-Inf	1.082965	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1217	184.1	44.5	190.6658	-54.1992	1.100789	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1221	67.2	26.7	57.02725	-17.6161	1.033955	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1222	16.82	7.63	Inf	-Inf	1.06053	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1225	0.09	2.05	Inf	-Inf	1.013458	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1232	45.4	13.6	38.08381	-11.7299	1.035273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1233	3.29	1.32	Inf	-Inf	1.085148	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1238	128.2	17.8	91.50876	-30.7596	0.951545	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1242	314	52	179.34	-66.1002	1.068711	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	1254	0.69	3.65	Inf	-Inf	0.987636	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1259	5.07	3.17	Inf	-Inf	1.014707	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1261	363.9	67.3	174.1079	-71.2403	0.980591	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1264	299.2	47	126.7997	-57.0953	1.054273	100	1	301.01	100.336667	0	0	0	0	0	0	0	0	0	0	0;
	1265	26.2	4.2	18.74091	-6.17365	1.076227	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1273	10.9	0.9	11.63997	-10.2023	1.080763	100	1	33.33672	0	0	0	0	0	0	0	0	0	0	0	0;
	1276	200.2	13.3	136.4827	-46.6725	1.037682	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1277	22.07	1.97	Inf	-Inf	1.037614	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1282	766.4	63.2	338.6569	-151.759	1.065091	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	1290	44.8	6.6	38.08405	-11.7045	0.983273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1294	31.2	4.6	18.09436	-6.79388	1.036182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1296	4.89	2.35	Inf	-Inf	0.994613	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1298	416.5	69	220.3241	-86.2224	1.073684	100	1	500	0	0	0	0	0	0	0	0	0	0	0	0;
	1307	49.4	6.8	37.72001	-12.0877	1.060455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1312	126.74	44.57	Inf	-Inf	1.07671	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1326	1592	80.5	895.2274	-330.781	1.098447	100	1	2000	666.666665	0	0	0	0	0	0	0	0	0	0	0;
	1332	11.03	5.11	Inf	-Inf	1.044907	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1341	785.4	87.6	394.4526	-156.998	1.073316	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1344	62.6	15.4	36.39672	-13.4413	1.060909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1347	925.6	186.3	539.9899	-195.853	1.041864	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	1350	288.3	7.2	127.7235	-56.3612	1.055227	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1353	360	25	176.6219	-72.7521	1.094409	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1358	455.4	118.8	376.3053	-114.176	1.072636	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	1360	1.85	1.63	Inf	-Inf	1.061162	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1364	5.2	0.2	0.2	0.2	1.097789	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	1373	26.2	5.3	18.77376	-6.1563	1.064091	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1374	149.2	35	90.08666	-32.4922	1.065364	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1391	10.86	5.93	Inf	-Inf	1.052282	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1403	33.12	9.16	Inf	-Inf	1.088761	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1410	341.2	44	175.2955	-69.9556	1.066526	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	1414	350	3.1	174.2149	-70.7485	1.103737	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1419	453.4	122	376.3604	-114.083	1.092237	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	1421	993.3	173.8	478.4516	-196.844	1.064211	100	1	1100	366.666663	0	0	0	0	0	0	0	0	0	0	0;
	1422	0.5	1	Inf	-Inf	1.040733	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1431	121.4	13.4	93.1621	-29.4852	1.095842	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1442	312	18.8	177.9814	-67.0068	1.084737	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	1443	9.85	3.73	Inf	-Inf	0.98905	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1450	0.16	11.88	Inf	-Inf	1.04174	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1453	23.34	8.13	Inf	-Inf	1.046454	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1462	50.6	13.2	37.66067	-12.1682	1.0405	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1464	358.2	56.4	174.6007	-70.7064	1.097816	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1475	15.5	2.4	21.27174	-20.0725	1.071632	100	1	62.70708	0	0	0	0	0	0	0	0	0	0	0	0;
	1478	0.22	0.4	Inf	-Inf	1.030136	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1490	0.18	1.06	Inf	-Inf	1.059861	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1496	50.4	14.8	37.73852	-12.1144	1.078	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1498	33.1	14.35	Inf	-Inf	1.085557	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1500	0.18	2.5	Inf	-Inf	1.041667	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1515	15.35	6.56	Inf	-Inf	1.076143	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1519	0.75	5.36	Inf	-Inf	1.094364	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1524	28.13	0.83	Inf	-Inf	1.095247	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1525	0.03	0.99	Inf	-Inf	1.047019	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1532	1250.4	281.6	928.8886	-296.103	1.084579	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	1534	-2.06	1.63	Inf	-Inf	1.047354	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1536	1014.8	137.8	530.3008	-205.674	1.062921	100	1	1200	399.999998	0	0	0	0	0	0	0	0	0	0	0;
	1537	62.7	8.4	57.40759	-17.2186	1.0234	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1544	0.49	1.27	Inf	-Inf	0.941739	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1547	18.9	2.1	31.94427	-31.2368	1.087842	100	1	95.45455	0	0	0	0	0	0	0	0	0	0	0	0;
	1554	0.41	0	Inf	-Inf	1.048673	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1558	860.8	118.8	439.3657	-173.297	1.079342	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	1559	15.38	4.94	Inf	-Inf	1.064406	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1565	73.6	26.8	35.11255	-14.7618	1.019636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1567	19.9	4.3	19.24783	-5.64568	0.981	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1569	426.8	112	378.6989	-111.847	1.065273	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	1570	994.5	270.3	534.2367	-201.816	1.052079	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	1571	28.1	1.7	18.52145	-6.38562	1.096591	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1572	0.2	1	Inf	-Inf	1.066343	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1576	86.7	24.3	55.28327	-19.4424	1.0735	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1580	17.18	4.35	Inf	-Inf	1.038717	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1586	200	35	84.74093	-38.128	1.011364	100	1	202.58	67.526666	0	0	0	0	0	0	0	0	0	0	0;
	1592	79.6	2.8	55.21395	-18.4759	1.085591	100	1	125	0	0	0	0	0	0	0	0	0	0	0	0;
	1596	835.5	187.8	387.6747	-163.732	1.073026	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1597	326.8	47.6	176.6853	-68.4505	1.068105	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	1603	40	4.6	16.99073	-7.92455	1.063	100	1	41.06	0	0	0	0	0	0	0	0	0	0	0	0;
	1604	9.78	3.29	Inf	-Inf	1.021069	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1605	178.8	20	87.4773	-35.3658	1.024682	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1624	595	-83.34	99999	-99999	1.084526	100	1	624.75	0	0	0	0	0	0	0	0	0	0	0	0;
	1626	0.18	3.85	Inf	-Inf	1.043385	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1633	108.3	6	51.65595	-22.8464	0.967727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1634	1142.4	182.4	515.0866	-220.893	1.098053	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	1636	430.2	116.6	272.2458	-94.9153	1.043079	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1641	75.8	6	34.63629	-15.1999	1.017091	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1642	1250.4	259.8	926.3492	-297.678	1.092	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	1645	54.6	8.2	36.93058	-12.7888	0.979909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1646	0.63	0.83	Inf	-Inf	1.037611	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1653	165.6	24	111.2286	-38.158	1.075182	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1657	27.88	10.61	Inf	-Inf	1.036709	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1658	345.6	-3.6	174.2199	-70.909	1.088474	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	1661	0.59	0	Inf	-Inf	1.069693	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1663	246	-5.5	132.4556	-51.7136	1.093263	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1666	658.2	80.2	356.0514	-134.469	1.065737	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	1667	64.8	17.8	36.03834	-13.7799	1.030737	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1668	0.65	1.81	Inf	-Inf	1.041029	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1678	57.97	23.21	Inf	-Inf	1.021406	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1680	89.13	38.73	Inf	-Inf	1.041069	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1686	59.7	32.7	58.01401	-16.804	1.029	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1687	28.14	0.76	Inf	-Inf	1.095264	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1688	40.6	9.2	38.41651	-11.3627	0.977909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1691	10.7	1.2	11.51823	-10.0851	1.081318	100	1	32.9694	0	0	0	0	0	0	0	0	0	0	0	0;
	1692	15.6	0.9	11.76153	-8.88877	1.060364	100	1	31.91836	0	0	0	0	0	0	0	0	0	0	0	0;
	1693	54	21.3	58.16419	-16.5195	1.047409	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1710	2.23	1.27	Inf	-Inf	1.053042	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1712	43.4	9.1	27.30786	-9.5627	1.013818	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	1718	100.2	43.2	117.1385	-32.3714	1.073867	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	1724	25.07	9.51	Inf	-Inf	1.082211	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1727	768.78	150	396.0222	-155.827	1.084227	100	1	912.78	312.779997	0	0	0	0	0	0	0	0	0	0	0;
	1729	1526	296.8	686.7284	-294.622	1.042605	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	1736	246.5	12.5	133.1073	-51.2529	1.096079	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1738	478.2	-4.6	263.4934	-103.767	1.078947	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	1743	330.6	152	283.4205	-84.4476	1.064237	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1744	76.2	6.2	44.8563	-16.4272	1.021227	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	1745	14.85	4.92	Inf	-Inf	0.983988	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1746	37.2	8	38.67061	-11.1014	1.068591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1751	115.5	15.6	160.3622	-150.947	1.054545	100	1	472.16328	39.346941	0	0	0	0	0	0	0	0	0	0	0;
	1756	411.8	5.6	274.1207	-93.0109	1.099158	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1760	2.89	1.8	Inf	-Inf	1.049394	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1763	413.5	113	221.3513	-85.3618	1.044864	100	1	500	0	0	0	0	0	0	0	0	0	0	0	0;
	1772	11.44	3.29	Inf	-Inf	1.060729	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1779	4.8	0.4	0.4	0.4	1.087289	100	1	4.8	4.8	0	0	0	0	0	0	0	0	0	0	0;
	1789	31	-2.5	18.14115	-6.75433	1.097227	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1800	99.9	20.7	54.23686	-20.6272	0.993364	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1818	44	4.2	27.19461	-9.66223	1.066136	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	1825	141.6	7.2	80.51183	-30.0723	1.101763	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	1841	216	83.5	239.7539	-66.8644	1.088409	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	1842	42.8	4.2	38.09945	-11.6185	1.014636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1853	576	126.6	257.0038	-110.847	1.096267	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1858	855.6	155.8	441.8502	-171.271	1.079974	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	1869	234	15.6	133.486	-50.2551	1.084289	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	1870	0.65	1.44	Inf	-Inf	1.000074	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1872	836.2	234.8	962.2016	-263.654	1.090211	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	1873	71	9.2	98.71382	-92.6957	1.083842	100	1	290.28568	0	0	0	0	0	0	0	0	0	0	0	0;
	1876	1.8	0.3	0.3	0.3	1.085182	100	1	1.8	1.8	0	0	0	0	0	0	0	0	0	0	0;
	1884	0.61	1.33	Inf	-Inf	1.007601	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1885	156.3	22.7	67.03379	-30.8274	1.089182	100	1	159.9082	53.302733	0	0	0	0	0	0	0	0	0	0	0;
	1902	176	30.5	88.39104	-36.1858	1.081368	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	1903	64.2	7.4	36.34278	-13.5403	1.053682	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1905	8.64	3.97	Inf	-Inf	1.053687	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1906	33.9	2.7	17.87846	-7.04432	1.083955	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1907	0.51	0.58	Inf	-Inf	0.975669	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1908	5.64	2.84	Inf	-Inf	1.058741	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1911	1066.8	92.4	522.7406	-212.369	1.092184	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	1913	178	29.4	86.24416	-36.2867	1.049526	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1915	152.3	16.4	160.6927	-140.967	1.082447	100	1	462.7272	51.414133	0	0	0	0	0	0	0	0	0	0	0;
	1921	1269.4	343.9	661.9019	-257.621	1.054316	100	1	1500	499.999998	0	0	0	0	0	0	0	0	0	0	0;
	1927	379.2	33.5	172.0666	-73.2919	1.095395	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	1929	4.82	1.45	Inf	-Inf	0.964327	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1931	0.2	1	Inf	-Inf	1.064571	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1933	0.19	3.76	Inf	-Inf	1.088007	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1936	0.04	0.97	Inf	-Inf	1.02989	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1959	498.4	113.4	266.1681	-101.497	1.099816	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	1961	211.9	49.5	137.4139	-46.4711	1.043842	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	1969	16.5	11.49	Inf	-Inf	1.041988	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1971	39.7	7.4	17.2416	-7.71534	1.034909	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	1974	125	42.2	92.3537	-30.1097	1.046818	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	1975	12.02	7.68	Inf	-Inf	1.071915	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1982	45.8	16.6	38.20466	-11.6772	1.066132	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	1983	49.8	27.9	58.35417	-16.2626	1.038045	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1989	99.3	3.6	53.94042	-20.8278	1.076727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	1990	668.4	-16.8	405.2462	-145.323	1.096947	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	1993	14.27	6.6	Inf	-Inf	1.088696	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1995	19.79	21.11	Inf	-Inf	1.086015	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	1996	7.9	3.5	3.5	3.5	1.034636	100	1	7.9	7.9	0	0	0	0	0	0	0	0	0	0	0;
	2000	26	1.9	19.50505	-14.7367	1.080421	100	1	52.9137	0	0	0	0	0	0	0	0	0	0	0	0;
	2009	204.3	14.1	137.2936	-46.5789	1.0354	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	2010	31.6	10.8	38.95567	-10.7555	0.996182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2017	0.04	1.32	Inf	-Inf	0.979656	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2023	9.87	3.63	Inf	-Inf	0.971857	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2027	5.2	1.1	1.1	1.1	1.077	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	2038	5.3	0.5	0.5	0.5	1.046182	100	1	5.3	5.3	0	0	0	0	0	0	0	0	0	0	0;
	2039	100	16.4	94.84088	-27.7595	1.067455	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2041	4.81	1.82	Inf	-Inf	0.991489	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2046	14.31	4.06	Inf	-Inf	0.979297	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2049	54.4	8.4	26.07309	-10.8235	0.994045	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2052	706.8	164.6	351.2981	-139.521	1.079368	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2056	14.83	3.56	Inf	-Inf	1.041035	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2062	362	105.5	228.1609	-78.8258	1.073816	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	2069	73.54	31.68	Inf	-Inf	0.968259	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2071	38.9	4.9	41.66955	-36.497	1.078364	100	1	119.29593	0	0	0	0	0	0	0	0	0	0	0	0;
	2077	41.2	11.2	38.47569	-11.352	1.083909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2079	0.42	0	Inf	-Inf	1.063189	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2082	110.5	32	95.68823	-28.9294	1.078591	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2096	27.25	9.38	Inf	-Inf	1.087726	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2099	1.8	12.01	Inf	-Inf	1.092555	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2100	164.5	18.9	141.3924	-42.3685	1.040318	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2109	82.2	17.8	54.86086	-18.7998	1.057636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2117	0.2	1.33	Inf	-Inf	1.04909	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2118	0.15	1.4	Inf	-Inf	1.045999	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2122	431.2	112	273.4914	-94.1036	1.086467	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2136	49	5.3	47.28096	-13.9058	1.008533	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	2141	41.8	14.6	38.31299	-11.4668	1.053447	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2143	78.2	3.6	34.35174	-15.4997	1.076342	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2145	6.43	3.05	Inf	-Inf	1.054551	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2146	662.2	-29.8	351.4238	-138.017	1.100553	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2153	129.4	35.8	92.71762	-30.0438	1.092955	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	2175	400.8	78.4	380.3121	-110.078	1.094947	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2178	548.6	178.8	368.841	-121.978	1.079237	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2181	179.5	33.5	88.69579	-36.0649	1.054682	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2189	67.3	13.5	134.543	-136.293	1.060091	100	1	409.09092	0	0	0	0	0	0	0	0	0	0	0	0;
	2190	145.2	48.4	90.67669	-31.9333	0.992136	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2193	0.54	1	Inf	-Inf	1.006357	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2197	53.4	10.6	37.38561	-12.4525	1.076909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2201	96.9	14.1	54.30003	-20.4859	1.073773	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2202	76.2	9.6	105.3378	-99.0244	1.083545	100	1	309.92896	0	0	0	0	0	0	0	0	0	0	0	0;
	2205	470	109	214.104	-92.6526	1.047289	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	2220	1367	312.6	921.6151	-304.728	1.104211	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	2224	90.74	39.27	Inf	-Inf	1.049729	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2230	44.6	5.6	38.01075	-11.7382	0.992364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2232	9.83	3.77	Inf	-Inf	1.013914	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2235	9.75	5.49	Inf	-Inf	1.048703	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2238	46.8	9.6	38.01763	-11.8145	1.076591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2248	2.65	1.17	Inf	-Inf	1.081245	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2250	31.8	17.2	39.1164	-10.7096	1.057	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2251	0.21	1.98	Inf	-Inf	1.060382	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2256	286.5	31.2	162.3973	-61.7859	1.08	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	2265	76.8	9.6	34.56102	-15.292	1.091591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2270	14.9	4.81	Inf	-Inf	0.969104	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2272	41.2	11.4	38.46016	-11.36	1.081045	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2273	26.83	5.39	Inf	-Inf	1.092711	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2277	27.7	3.7	18.42517	-6.43617	0.981636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2279	46.1	11.8	78.13037	-77.0658	1.022273	100	1	234.51775	0	0	0	0	0	0	0	0	0	0	0	0;
	2280	5.29	1.91	Inf	-Inf	1.062122	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2281	41.8	2.5	27.23773	-9.55191	0.997533	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2286	36.65	16.14	Inf	-Inf	1.040521	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2293	179	11.2	86.22723	-36.3258	1.060318	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	2298	55	8.1	25.99267	-10.9039	0.994318	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2300	123.2	27.6	73.33611	-26.4443	1.050067	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2308	54.3	18.6	58.22931	-16.4918	1.078227	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2311	66.73	12.2	Inf	-Inf	1.09233	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2313	0.15	1	Inf	-Inf	1.077179	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2316	1.33	1.29	Inf	-Inf	1.059808	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2326	159.5	41.5	91.24639	-33.5163	0.968364	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2327	24.4	2.4	18.86354	-6.03025	1.055933	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2328	28.96	8.62	Inf	-Inf	1.069898	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2330	411.2	115.2	276.1909	-91.6307	1.098133	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2331	103.52	43.41	Inf	-Inf	1.015822	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2345	39	5.4	17.0554	-7.84143	1.028545	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2346	567	-47.4	250.6568	-116.19	1.055579	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	2348	0.43	0	Inf	-Inf	1.063811	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2351	127.2	1.5	81.7908	-28.6771	1.096818	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2352	994.2	273	533.8653	-202.071	1.048921	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	2356	92.8	25.2	76.08315	-23.5873	1.084818	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2360	37.8	3.6	17.14383	-7.73222	0.978909	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2362	291.9	62.4	128.018	-56.2324	1.035727	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2375	201	20.4	107.6217	-41.9226	1.031455	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	2377	53.4	20.1	58.26545	-16.4419	1.084136	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2380	1367	318.8	921.522	-304.789	1.103421	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	2394	1984.8	108.2	848.7641	-377.669	1.079184	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	2401	156.9	45.3	141.8499	-42.116	1.099211	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	2402	116.6	2.6	50.85036	-22.9062	1.078591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2405	113.4	16.8	51.94106	-22.8025	1.060818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2408	0.72	0	Inf	-Inf	1.051348	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2410	297.5	7.5	124.6845	-59.3932	1.048773	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2411	144.9	3.6	79.68832	-30.775	1.032533	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2413	34.8	3.7	27.99116	-8.80931	1.041933	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2415	378.8	38.9	171.7157	-73.5513	1.0764	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2422	0.5	0.4	Inf	-Inf	1.028639	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2426	15.96	5.35	Inf	-Inf	1.077784	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2434	157.2	25.5	78.09426	-32.3759	1.022955	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2435	34.5	6.4	69.00791	-69.9001	1.060818	100	1	209.8182	0	0	0	0	0	0	0	0	0	0	0	0;
	2440	19.81	5.69	Inf	-Inf	1.002973	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2451	0.2	4.97	Inf	-Inf	1.083674	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2458	1118.3	189.9	665.7662	-242.998	1.064579	100	1	1480	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	2465	77.6	8.4	70.15367	-58.105	1.073773	100	1	196.71715	0	0	0	0	0	0	0	0	0	0	0	0;
	2471	148.8	27.6	69.22198	-30.3499	1.0334	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2474	0.13	1	Inf	-Inf	1.07923	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2475	65.1	15.9	57.40889	-17.323	1.030636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2478	1.01	0.78	Inf	-Inf	1.073441	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2479	-2.05	1.55	Inf	-Inf	0.991722	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2481	204.3	11.1	137.0248	-46.7535	1.017318	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2487	69.2	4.2	35.584	-14.2628	1.054136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2488	9.82	3.4	Inf	-Inf	0.996493	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2490	11.77	3.81	Inf	-Inf	1.019444	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2495	0.63	1.17	Inf	-Inf	1.006417	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2500	872.2	142.1	748.514	-231.559	1.081947	100	1	1600	466.666665	0	0	0	0	0	0	0	0	0	0	0;
	2505	576.3	76.2	416.6955	-134.5	1.088579	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	2510	93.59	36.62	Inf	-Inf	1.056239	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2514	-6.65	4.21	Inf	-Inf	1.03873	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2520	28.13	0.81	Inf	-Inf	1.095266	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2524	1221.2	333.8	934.8061	-291.472	1.102737	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	2533	179	15.5	88.32609	-36.3317	1.078545	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2535	66.8	14	77.91347	-21.6558	1.074091	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2536	1049.2	89.2	525.344	-209.845	1.083868	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	2550	1250.4	258.8	926.2575	-297.735	1.0915	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	2551	71.8	-1.2	34.79788	-14.9432	1.024867	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2555	86.4	17.8	54.6213	-19.1096	1.053318	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2566	831.6	171	551.0249	-184.738	1.084184	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	2567	5.66	3.06	Inf	-Inf	0.969603	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2570	29.2	3.6	28.38039	-8.37922	1.005	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	2572	80.7	21.3	55.98089	-18.7627	1.059773	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2573	31.6	6.1	18.18763	-6.74162	1.040364	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2579	0.87	0.84	Inf	-Inf	1.044584	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2583	138.6	2.4	91.19925	-31.3544	1.096818	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2584	126	3.6	72.02056	-27.4713	1.024	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2585	12.33	2.65	Inf	-Inf	1.085298	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2593	47.6	12	37.93587	-11.8917	1.061909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2594	1.51	0.66	Inf	-Inf	1.054352	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2598	115.2	19.8	110.2864	-93.1338	1.072591	100	1	311.38048	0	0	0	0	0	0	0	0	0	0	0	0;
	2602	28.28	19.72	Inf	-Inf	1.043873	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2603	4.5	0.7	0.7	0.7	1.065682	100	1	4.5	4.5	0	0	0	0	0	0	0	0	0	0	0;
	2604	0.58	1	Inf	-Inf	1.010387	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2608	52.5	9.1	72.95267	-68.7068	1.058773	100	1	214.85716	0	0	0	0	0	0	0	0	0	0	0	0;
	2611	17.88	8.45	Inf	-Inf	0.959599	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2615	60.9	25.8	57.68599	-17.019	1.068864	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2625	3.6	0.3	0.3	0.3	1.044318	100	1	3.6	3.6	0	0	0	0	0	0	0	0	0	0	0;
	2626	99.6	2.5	94.67771	-27.8706	1.094545	100	1	200	66.666667	0	0	0	0	0	0	0	0	0	0	0;
	2630	0.51	20.69	Inf	-Inf	1.061894	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2631	20.9	5.4	19.18509	-5.71506	1.085	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2649	658.5	174	563.9284	-170.684	1.066553	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	2652	962.1	151.2	536.8406	-198.88	1.062342	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	2659	53.2	6.8	37.24111	-12.5371	1.060455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2663	3.4	0.5	0.5	0.5	1.080136	100	1	3.4	3.4	0	0	0	0	0	0	0	0	0	0	0;
	2671	330.6	151.6	283.4205	-84.4476	1.064316	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2676	4.3	0.2	0.2	0.2	1.014409	100	1	4.3	4.3	0	0	0	0	0	0	0	0	0	0	0;
	2677	23.78	15.53	Inf	-Inf	1.031212	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2678	137	59.22	Inf	-Inf	1.037439	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2685	362.8	-4	172.5338	-72.7462	1.088658	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	2688	0.45	0.03	Inf	-Inf	1.010038	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2694	29.96	21.31	Inf	-Inf	1.061073	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2697	21.79	7.69	Inf	-Inf	1.081031	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2701	10.48	5.51	Inf	-Inf	1.045299	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2702	930.1	134.6	432.4675	-180.67	1.081632	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	2704	30.2	4.7	18.34657	-6.58203	1.063636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2710	79.4	15	34.04062	-15.7819	0.992818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2712	0.47	0.8	Inf	-Inf	0.966927	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2713	183.6	32.3	184.4563	-132.672	1.092182	100	1	493.34672	0	0	0	0	0	0	0	0	0	0	0	0;
	2715	296.8	7.2	139.6427	-59.7518	1.098455	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	2717	87	18.9	55.22231	-19.4824	1.052026	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2720	200.1	50.2	189.579	-55.3286	1.102526	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2725	2528.4	452.4	1434.053	-527.42	1.083553	100	1	3200	1066.66666	0	0	0	0	0	0	0	0	0	0	0;
	2728	61.37	21.83	Inf	-Inf	1.037701	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2729	516.6	134	370.5762	-119.771	1.054955	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2735	43.9	6.8	47.71976	-13.5005	1.057067	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	2744	162.5	12	89.61202	-34.8173	1.039091	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	2754	855.7	157.7	441.8502	-171.271	1.079816	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	2755	49.6	7	37.49499	-12.2378	1.005545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2760	676.2	205.2	564.6201	-170.921	1.055658	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	2763	292.2	63.3	127.9105	-56.335	1.035045	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	2764	37.1	4.9	17.42935	-7.48901	1.085773	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2774	0.54	0.77	Inf	-Inf	1.08369	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2776	29.7	8.7	22.35191	-16.7674	1.034455	100	1	60.4264	0	0	0	0	0	0	0	0	0	0	0	0;
	2779	392.3	-5.5	167.7199	-77.1137	1.032045	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2781	17.8	3.68	Inf	-Inf	1.105203	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2782	14.85	4.99	Inf	-Inf	0.967787	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2799	340.5	56.5	228.1885	-78.0551	1.059636	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	2804	7.63	1.97	Inf	-Inf	1.075195	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2805	344.07	126.7	188.93	-56.3231	1.050273	100	1	525.57	258.903333	0	0	0	0	0	0	0	0	0	0	0;
	2809	203.4	19.2	106.961	-42.4946	1.073227	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	2811	47	21.6	38.05564	-11.7985	1.030455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2829	0.19	73.31	Inf	-Inf	1.080657	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2832	16.1	0.7	22.39079	-20.9936	1.093184	100	1	65.79592	0	0	0	0	0	0	0	0	0	0	0	0;
	2833	38.6	5.1	17.33017	-7.61288	1.071636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	2840	5.44	2.16	Inf	-Inf	1.036673	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2845	2.32	1.28	Inf	-Inf	1.050103	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2846	581.4	-33	250.267	-116.974	1.070526	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	2849	357.4	40.8	278.1559	-88.4691	1.070632	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	2851	14.9	4.6	Inf	-Inf	0.992533	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2852	141.6	32.4	113.9748	-35.5577	1.089773	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	2853	142	9.6	70.25158	-29.3318	0.983864	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	2863	233.4	12.1	187.3535	-57.7612	1.104579	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2866	9	1.5	1.5	1.5	1.041864	100	1	9	9	0	0	0	0	0	0	0	0	0	0	0;
	2868	94.5	24.9	54.68458	-20.1297	1.073273	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2872	9	0.6	0.6	0.6	1.079773	100	1	9	9	0	0	0	0	0	0	0	0	0	0	0;
	2873	181.3	13.4	138.8055	-44.4952	1.061267	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	2876	0.49	0.66	Inf	-Inf	1.005569	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2878	33.9	7.6	67.92443	-69.0188	1.0335	100	1	206.86296	0	0	0	0	0	0	0	0	0	0	0	0;
	2880	11.89	3.6	Inf	-Inf	1.07408	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2884	-2.83	2.15	Inf	-Inf	0.983142	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2896	98.7	5.7	53.70363	-20.9803	1.076455	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2902	77.1	3.1	44.68199	-16.5811	1.007818	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	2910	144	33.6	113.5482	-35.8641	0.983182	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	2916	416.7	90.9	243.6108	-88.463	1.077342	100	1	540	0	0	0	0	0	0	0	0	0	0	0	0;
	2927	57.8	8	36.79117	-13.0042	1.054318	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2928	304.7	55.4	180.5489	-64.6315	1.057045	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2934	360	35.2	173.9584	-71.5641	1.088158	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	2936	323.2	-0.4	177.6058	-67.6604	1.098211	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	2939	387.1	11.9	178.4736	-79.1125	1.020864	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	2940	69.4	17.4	56.13718	-17.5304	1.0705	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2941	9.78	2.97	Inf	-Inf	1.029379	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2942	359.3	32.4	173.8463	-71.3061	1.026591	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	2945	5.06	1.84	Inf	-Inf	1.092885	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2946	68.4	6	62.23372	-51.372	1.087545	100	1	174.2386	0	0	0	0	0	0	0	0	0	0	0	0;
	2957	428	111.2	378.6379	-111.908	1.065364	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2967	39.6	15.6	38.64547	-11.2044	1.056273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	2968	37.8	8.7	17.35786	-7.56344	1.069263	100	1	45	0	0	0	0	0	0	0	0	0	0	0	0;
	2980	113.4	15	52.30851	-22.5164	1.071091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	2985	130.2	25.8	81.96635	-28.662	1.058895	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2986	654.8	86.4	356.436	-134.086	1.086684	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	2988	0.57	0.91	Inf	-Inf	1.015739	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	2990	176.7	23.1	76.53817	-34.1961	1.097053	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	2991	327	30	229.0024	-77.0209	1.042136	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	2997	228.6	42	103.5849	-45.8927	1.036267	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3002	861.3	120.9	439.1991	-173.427	1.046091	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	3006	29.8	3.8	22.58182	-16.939	1	100	1	61.06122	0	0	0	0	0	0	0	0	0	0	0	0;
	3010	0.2	1.64	Inf	-Inf	1.068652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3017	52.6	7.3	46.86475	-14.2632	0.942133	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	3018	44	7.6	38.13373	-11.6413	1.059273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3020	660.4	180.9	564.4374	-170.467	1.073273	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	3024	114.6	11.8	93.36139	-29.109	1.030364	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3026	10.59	5.25	Inf	-Inf	1.03285	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3027	738.4	101.8	345.7083	-144.761	1.082318	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	3028	75	1.4	34.41229	-15.3417	1.025909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3029	-600	325.55	99999	-99999	1.040526	100	1	630	-630	0	0	0	0	0	0	0	0	0	0	0;
	3033	91.8	20.7	55.00357	-19.8098	1.003182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3038	56.8	10.8	37.10117	-12.7605	1.074632	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3041	136.8	12.8	91.47759	-31.1124	1.087237	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3045	44.2	9	38.20483	-11.6107	1.081773	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3054	59.2	7.8	36.83316	-13.0254	1.055591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3056	44.5	1.9	47.19351	-41.4382	1.074	100	1	135.27273	0	0	0	0	0	0	0	0	0	0	0	0;
	3060	57.81	20.84	Inf	-Inf	1.086521	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3063	61.8	11.2	36.3129	-13.4715	1.036818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3067	347.7	66.6	175.6632	-69.5685	1.004682	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3078	0.2	1.64	Inf	-Inf	1.068652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3079	78.6	21.3	56.34953	-18.438	1.018318	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3080	94.2	22.5	54.58719	-20.1874	1.044682	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3083	24.87	3.71	Inf	-Inf	1.038648	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3086	249.3	-6.3	131.7502	-52.0409	1.074053	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3088	0.75	0.93	Inf	-Inf	0.978189	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3092	66.13	26.79	Inf	-Inf	1.063538	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3093	0.76	0.86	Inf	-Inf	1.038064	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3099	55.8	8.3	25.57268	-11.2551	0.994091	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	3104	171.3	3.6	76.42851	-34.1246	1.003318	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3110	53.7	-2.4	58.19449	-16.4966	1.096545	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3117	127.6	12	92.2584	-30.2648	1.047318	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3118	-39.05	11.82	Inf	-Inf	1.07932	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3122	835.5	195.6	550.9653	-184.892	1.052864	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	3124	52.6	13.2	37.39729	-12.4151	1.055182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3126	159.2	8	89.44973	-33.2807	1.0775	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	3130	7.63	4.3	Inf	-Inf	1.044212	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3131	13.97	6.85	Inf	-Inf	1.074294	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3137	0.49	-0.1	Inf	-Inf	0.934331	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3161	239.6	-14.4	185.188	-59.5457	1.075727	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	3173	102.6	19.5	53.75122	-21.0757	1.058682	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3187	109.8	2.6	51.57644	-22.1262	1.055727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3188	96.9	35.1	54.08884	-20.6392	1.075658	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3191	522.8	5.8	366.737	-122.452	1.099364	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	3192	3.24	1.18	Inf	-Inf	1.046462	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3193	829.8	73.2	387.8762	-163.395	1.078842	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3194	0.64	3.53	Inf	-Inf	1.094331	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3195	38.4	12.8	38.53735	-11.2184	1.027682	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3198	845	-12	433.8317	-177.9	1.038667	100	1	1000	333.33333	0	0	0	0	0	0	0	0	0	0	0;
	3215	0.48	-0.1	Inf	-Inf	1.024741	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3226	140	16	91.05973	-31.4976	1.091455	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	3229	341.1	51.3	157.1492	-67.4015	1.065273	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	3230	350	63.8	175.5053	-69.7598	1.018318	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3243	35.6	15.6	38.82578	-10.9681	1.057682	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3247	298.9	20	124.1502	-59.2122	1.0415	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3248	233.6	27.6	100.9133	-46.4341	1.020227	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3254	38.7	12.99	Inf	-Inf	1.070314	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3256	1.88	1.85	Inf	-Inf	1.051337	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3257	539	133	261.6953	-107.358	1.076136	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	3259	31.2	2.8	52.83077	-52.0446	1.087053	100	1	158.88325	0	0	0	0	0	0	0	0	0	0	0	0;
	3260	36.3	5.8	17.701	-7.25467	1.069474	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3261	26.6	5.1	19.79709	-15.1203	1.0525	100	1	53.9798	0	0	0	0	0	0	0	0	0	0	0	0;
	3262	50.8	3.2	37.43079	-12.3168	1.024136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3274	43.6	15.2	38.11837	-11.6405	1.015545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3280	158.7	16.8	78.60296	-32.0414	1.075	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3282	16.9	7.26	Inf	-Inf	1.033934	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3284	28.13	0.82	Inf	-Inf	1.09526	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3286	92.4	22.2	54.59248	-20.1214	1.051455	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3290	113.97	50.37	Inf	-Inf	1.041342	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3292	65.8	13.4	36.22158	-13.6796	1.012	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3295	261	14	130.0067	-54.0425	1.0555	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3301	79.6	20.4	76.85015	-22.6417	1.017182	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3306	412	-9	217.4353	-88.2017	1.098045	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	3314	16.4	3.1	22.35149	-21.217	1.054091	100	1	66.09044	0	0	0	0	0	0	0	0	0	0	0	0;
	3322	0.61	1.34	Inf	-Inf	1.053869	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3325	112.2	43.2	116.0739	-33.3221	1.070591	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3328	1097.4	514.2	943.644	-282.012	1.064474	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	3329	861.2	106.9	441.0395	-172.049	1.082816	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	3331	0.2	0.96	Inf	-Inf	1.063021	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3340	556.8	30.6	257.6069	-110.427	1.036227	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	3341	6.15	2.24	Inf	-Inf	1.071885	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3343	74.2	25.27	Inf	-Inf	1.093923	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3345	0.62	1.41	Inf	-Inf	1.056492	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3346	14.96	4.49	Inf	-Inf	1.009926	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3352	9.88	3.96	Inf	-Inf	0.928433	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3356	5.2	0.5	0.5	0.5	1.071727	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	3358	239.6	65	133.8361	-49.8906	1.076067	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3359	387.8	27.3	180.1052	-77.8448	1.009909	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3361	662.8	13.6	354.58	-135.702	1.084737	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	3362	32.71	13.76	Inf	-Inf	1.058124	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3363	62	12.2	36.33525	-13.4711	1.073636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3365	4.86	2.12	Inf	-Inf	0.982218	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3375	40.4	5.6	38.3637	-11.3757	1.0198	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3376	0.19	0.87	Inf	-Inf	1.066214	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3384	28.6	5.8	18.39755	-6.48899	1.002727	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3387	1146.3	206.1	514.1476	-221.73	1.050895	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	3388	94.9	21.4	42.93563	-18.4696	1.031955	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	3396	24.87	3.7	Inf	-Inf	1.04408	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3401	5.67	2.84	Inf	-Inf	1.080953	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3403	34.4	7.4	38.89954	-10.8975	1.067545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3405	854.7	204	548.5682	-187.155	1.080816	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	3407	29.9	3.7	28.37993	-8.40395	1.025067	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	3408	26.5	7.25	Inf	-Inf	1.071874	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3410	44.8	20	38.05757	-11.7222	1.037364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3414	239.2	105.6	148.0244	-51.7016	1.000136	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	3417	572.4	125.4	255.7925	-112.326	1.048421	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	3422	147.6	5.8	90.31483	-32.2665	1.0155	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3425	146.08	63.89	Inf	-Inf	1.039631	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3431	162.4	35.6	110.3415	-37.1203	1.053636	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3437	8.03	7.84	Inf	-Inf	1.015158	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3444	35.5	5.5	17.76234	-7.17904	1.007545	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3447	6.03	3.15	Inf	-Inf	0.992894	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3451	90.6	5.7	54.70099	-19.9681	1.0126	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3452	190.8	8.4	108.0619	-41.254	1.071955	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	3453	13.2	6.15	Inf	-Inf	1.050529	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3457	68.3	9.9	35.41076	-14.3663	1.019947	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3465	615	151.2	467.0212	-146.118	1.103974	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	3466	563.5	156.1	313.012	-116.735	1.081364	100	1	700	233.333331	0	0	0	0	0	0	0	0	0	0	0;
	3467	89.6	-12.2	94.80645	-27.3595	1.043	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	3470	1.6	1.01	Inf	-Inf	1.054809	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3478	397.6	5.6	165.7633	-79.1699	1.0305	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	3481	837.3	80.1	388.1704	-163.4	1.088395	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3495	21.4	4.4	36.28851	-35.4239	1.079818	100	1	108.33335	0	0	0	0	0	0	0	0	0	0	0	0;
	3496	143.4	-9.3	79.52042	-30.8462	1.075727	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3502	17.91	7.9	Inf	-Inf	1.066946	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3504	239.6	-14.4	185.188	-59.5457	1.075727	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	3505	31.6	3.6	28.22675	-8.55108	1.0124	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	3507	287	14.1	182.7313	-62.5427	1.094182	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3509	38.73	15.32	Inf	-Inf	1.077512	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3515	0.58	2.28	Inf	-Inf	1.038027	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3520	9.24	3.54	Inf	-Inf	1.041649	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3522	8.43	7.98	Inf	-Inf	1.065098	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3523	855.6	157.5	440.6871	-172.142	1.081368	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	3528	77	13.4	34.27487	-15.5221	1.047227	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3531	139.6	32.8	71.01919	-28.6777	1.062867	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3532	28.13	0.82	Inf	-Inf	1.095247	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3540	70.2	19.2	56.83658	-17.8426	1.057273	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3552	99.6	22.2	54.18915	-20.6545	1.040182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3559	10	2	2	2	0.950818	100	1	10	10	0	0	0	0	0	0	0	0	0	0	0;
	3561	362	57.8	173.8408	-71.3995	1.060158	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3567	45.4	16.2	38.02338	-11.766	1.075045	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3569	11.8	5.47	Inf	-Inf	1.07292	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3571	349.3	59.5	175.5273	-69.7258	1.050316	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3573	571.2	127.2	256.1129	-112.058	1.061184	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	3581	47.2	14	38.01499	-11.8286	1.013909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3589	57.4	6.8	25.38834	-11.4507	1.0215	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	3590	723	-7	450.57	-161.535	1.048909	100	1	1000	333.33333	0	0	0	0	0	0	0	0	0	0	0;
	3597	50.6	4.4	26.31069	-10.5231	1.074727	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	3598	124	15	94.11957	-30.35	1.085273	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3600	63.3	6.4	126.1635	-127.559	1.048045	100	1	382.43214	0	0	0	0	0	0	0	0	0	0	0	0;
	3602	66.09	31.27	Inf	-Inf	1.059995	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3607	79.2	6.6	55.30057	-18.398	1.074636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3610	3.21	1.7	Inf	-Inf	1.075322	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3611	334.8	36.4	177.0911	-68.4099	1.054868	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	3616	0.18	2.27	Inf	-Inf	1.083103	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3619	84	19.2	55.6678	-19.087	1.063682	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3621	-2.51	1.64	Inf	-Inf	1.081274	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3622	0.18	1.42	Inf	-Inf	1.067558	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3628	229.9	70.7	134.9794	-48.7516	1.0436	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3631	0.47	-0.1	Inf	-Inf	0.991963	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3632	44.6	5.8	38.05465	-11.7155	1.039364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3635	69.21	20.71	37.62842	-12.2185	1.067682	100	1	97.81	17.81	0	0	0	0	0	0	0	0	0	0	0;
	3640	27.7	5.4	18.61941	-6.31	1.083955	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3645	9.97	4.55	Inf	-Inf	1.032109	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3648	10.92	4.41	Inf	-Inf	1.032834	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3649	0.18	1	Inf	-Inf	1.06448	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3651	89.4	12.9	54.88151	-19.8023	1.061727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3654	46	6.6	37.87363	-11.8723	1.020364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3658	109.5	3	52.24768	-22.4385	1.067909	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3663	0.18	1	Inf	-Inf	1.043276	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3666	89.5	40	97.07457	-27.4377	1.081342	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3668	401	51	223.0745	-83.7009	1.090553	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	3670	30	8.1	18.44	-6.50996	1.030273	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3686	323.1	63.9	178.6809	-66.5659	1.060105	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3690	66.03	12.38	Inf	-Inf	1.092336	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3691	68.2	7.2	35.734	-14.119	1.0845	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3695	46.2	8.4	38.05318	-11.7714	1.075227	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3701	17.55	1.7	Inf	-Inf	1.092661	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3706	-3.59	1.83	Inf	-Inf	1.025679	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3711	79.6	2.8	55.21395	-18.4759	1.085591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3712	162.8	33	88.50638	-34.0826	1.025864	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	3719	502	86	264.9409	-103.736	1.064455	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	3720	188	21.4	85.82804	-36.9183	1.009818	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3727	391	40.5	223.9141	-82.7537	1.022909	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	3748	41	7.4	38.3743	-11.404	1.070909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3750	0.54	1	Inf	-Inf	1.054954	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3752	216	19.5	134.1368	-49.1225	0.951818	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	3754	1.68	0.97	Inf	-Inf	1.029351	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3755	59.6	13.2	36.88517	-13.0037	1.060789	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3758	67.5	19.2	57.03059	-17.6274	1.062773	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3759	0.02	0.51	Inf	-Inf	1.040337	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3761	269.1	82.2	184.1854	-60.9595	0.996727	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3770	0.62	0	Inf	-Inf	1.040945	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3774	0.12	2.27	Inf	-Inf	1.042791	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3777	13.26	7.13	Inf	-Inf	1.025867	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3792	219	69.3	188.0693	-56.8225	1.055273	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	3801	26.2	5.6	18.58363	-6.27517	0.966545	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3803	1530.2	61.6	904.6709	-321.606	1.101868	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	3806	514.6	41.2	264.003	-103.612	1.096211	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	3807	95.1	10.8	54.59263	-20.2121	1.007364	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	3810	740	71.2	344.9026	-146.059	1.063474	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	3811	40.6	8.6	38.27121	-11.4298	0.985	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3817	4.65	1.88	Inf	-Inf	1.030601	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3820	767.4	192.3	396.3844	-154.993	1.082763	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	3822	621.7	162	465.4215	-147.305	1.047368	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	3823	3.2	1.05	Inf	-Inf	1.092148	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3825	260.1	5.8	131.2471	-52.4515	0.993211	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3829	4.3	0.4	0.4	0.4	1.080318	100	1	4.3	4.3	0	0	0	0	0	0	0	0	0	0	0;
	3832	491	120.2	373.4631	-118.655	1.089	100	1	800	200	0	0	0	0	0	0	0	0	0	0	0;
	3835	139.6	33.6	71.60573	-28.236	1.046867	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	3836	16	6.9	19.54586	-5.36425	1.089227	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3840	0.17	8.06	Inf	-Inf	1.045861	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3845	0.66	2.79	Inf	-Inf	1.027489	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3847	20.56	3.88	Inf	-Inf	1.07619	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3849	9.87	3.76	Inf	-Inf	0.963793	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3850	46.2	5.5	47.62189	-13.6344	1.078267	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	3853	2641.24	435.57	99999	-99999	1.083974	100	1	4188.9475	1333.33333	0	0	0	0	0	0	0	0	0	0	0;
	3863	2.68	3.06	Inf	-Inf	1.048641	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3865	641	88.8	354.4813	-135.04	0.997658	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	3877	103.78	29.33	Inf	-Inf	1.054759	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3879	28	1.5	18.41111	-6.45366	1.0184	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3883	11.81	4.02	Inf	-Inf	0.988349	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3892	29.88	5.97	Inf	-Inf	1.054952	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3896	525	27.6	261.5961	-106.385	1.034591	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	3903	301.09	38.39	Inf	-Inf	0.969689	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3908	251.3	33.6	196.0804	-61.8802	1.084409	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	3911	0.18	1	Inf	-Inf	1.076143	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3912	33.5	4.8	18.02218	-6.92517	1.086974	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3934	158.8	26.1	141.7507	-41.9698	1.045273	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3935	212.8	66.4	149.9819	-49.4706	1.069342	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	3942	1176	173.2	723.4602	-256.992	1.092763	100	1	1600	533.33333	0	0	0	0	0	0	0	0	0	0	0;
	3943	60.2	7.6	36.73438	-13.1213	1.054591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3945	152.4	6.6	79.098	-31.4618	1.014227	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	3946	73.3	8.9	45.09404	-16.1662	1.095447	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	3949	490.5	320.7	576.779	-158.19	1.044921	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	3950	39.9	2.9	16.93089	-7.96917	1.066318	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	3961	167	7.8	87.36773	-35.065	1.055909	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	3965	0.73	1.93	Inf	-Inf	1.026367	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	3984	59.2	11	36.93053	-12.951	1.029	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	3985	273.6	2.7	129.2863	-54.3776	1.043667	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	3988	67.31	44.04	Inf	-Inf	1.068963	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4000	152.4	11.1	142	-41.8689	1.099526	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	4002	40.8	13.8	38.48228	-11.3333	1.076864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4003	147.6	29.5	142.8157	-41.016	1.077053	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4005	236	30	134.3927	-49.9888	1.093526	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4006	31.9	9.6	18.09992	-6.81448	1.022818	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4007	358.2	56.5	174.6007	-70.7064	1.097789	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4016	5.15	16.03	Inf	-Inf	1.066048	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4025	861.3	108.3	441.0395	-172.049	1.082684	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	4031	11	125.78	Inf	-Inf	1.042198	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4033	14.7	2.7	20.48971	-19.2362	1.049818	100	1	60.24488	0	0	0	0	0	0	0	0	0	0	0	0;
	4044	90.6	10	94.73242	-27.4281	0.973067	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	4046	20.3	2	34.61421	-33.9213	1.086053	100	1	103.5533	0	0	0	0	0	0	0	0	0	0	0	0;
	4049	253	76.2	269.0087	-234.65	1.053342	100	1	772.5381	85.837567	0	0	0	0	0	0	0	0	0	0	0;
	4051	289.2	65.6	182.8731	-62.5172	1.082211	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4062	50.1	46.5	58.4114	-16.2632	1.047	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4073	0.2	1.64	Inf	-Inf	1.068652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4074	410.5	55.5	220.6235	-85.8069	1.100421	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4076	861.3	121.3	439.1991	-173.427	1.046091	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	4078	35.2	14	38.85147	-10.9486	1.051318	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4082	39.6	4.9	16.90316	-7.98099	1.001	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4094	156.8	7.4	89.57813	-33.1114	1.103553	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4098	115	14.4	51.15334	-22.6177	1.053091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4102	47.8	15	37.85738	-11.9536	1.073364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4104	6.39	3.43	Inf	-Inf	1.015363	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4111	364.8	49.2	173.0676	-72.3924	1.048682	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	4117	47.5	7.7	26.83413	-10.0356	1.0565	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	4120	9.78	3.5	Inf	-Inf	1.082712	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4123	1935.6	328.2	1074.337	-397.656	1.079289	100	1	2400	799.999998	0	0	0	0	0	0	0	0	0	0	0;
	4128	14.82	5.03	Inf	-Inf	1.017432	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4129	15.88	4.9	Inf	-Inf	1.030702	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4152	9.87	3.77	Inf	-Inf	0.966331	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4153	318	-11	122.6399	-61.6555	1.083955	100	1	325	0	0	0	0	0	0	0	0	0	0	0	0;
	4155	47.2	8.2	37.94159	-11.874	1.082409	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4161	31.26	13.2	Inf	-Inf	1.039896	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4168	59.13	22.68	Inf	-Inf	1.092728	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4172	127	56.2	92.96222	-29.8058	1.075	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	4175	0.64	1.32	Inf	-Inf	1.028661	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4180	1.73	1.71	Inf	-Inf	1.045696	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4181	23.31	10.08	Inf	-Inf	1.076509	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4188	466.8	142.8	375.6952	-114.931	1.086526	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	4202	115.2	20.6	93.85476	-28.8521	1.101947	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4206	68	14.4	35.62251	-14.1944	1.054	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4214	44.4	11.8	38.2839	-11.575	1.094263	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4222	78.4	15.2	45.00697	-16.3986	1.070105	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4231	170	30.5	140.9785	-43.3118	1.085864	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4235	21.7	6.2	19.14344	-5.7673	1.059682	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4236	1597.2	112.8	897.5751	-328.796	1.090237	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	4237	88.2	7.2	55.07343	-19.6265	1.048864	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4240	491	-45.5	198.8395	-105.442	1.018455	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4243	450.5	-47.5	214.7051	-91.5765	1.088079	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4249	763.6	22.4	343.2496	-147.429	1.098237	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	4262	541.8	65.7	417.981	-132.474	1.086447	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4265	-29.38	14.09	Inf	-Inf	1.047338	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4267	38.4	12.2	38.53424	-11.2199	1.027636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4273	93.2	5	98.40225	-86.291	1.090737	100	1	281.84922	0	0	0	0	0	0	0	0	0	0	0	0;
	4280	79.5	9.3	56.02602	-18.6845	0.968182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4282	2.7	1.16	Inf	-Inf	1.058228	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4291	2.91	0.6	Inf	-Inf	1.057064	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4297	7.81	2.71	Inf	-Inf	0.962984	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4309	4.11	1.58	Inf	-Inf	1.078886	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4310	50	6.6	37.44187	-12.2776	0.992	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4311	0.15	0.9	Inf	-Inf	1.070183	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4312	5.92	2.38	Inf	-Inf	0.979616	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4314	0.65	1.97	Inf	-Inf	1.010753	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4325	17.96	15.71	Inf	-Inf	0.907012	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4327	570.6	126.6	257.4896	-110.952	1.043368	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	4335	16.84	7.96	Inf	-Inf	1.068366	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4339	165.5	10	89.53827	-34.9773	1.0615	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4340	16	3.6	22.15173	-20.9025	1.064636	100	1	65.29948	0	0	0	0	0	0	0	0	0	0	0	0;
	4343	805.5	156.6	389.8258	-161.157	1.103711	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	4349	861.3	120.1	439.1991	-173.427	1.046182	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	4354	418	103.6	379.5125	-111.07	1.024545	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	4371	188.1	67.3	190.2855	-54.5208	1.060364	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4381	42	12.2	38.37408	-11.4354	1.076364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4383	855.6	157.8	440.6871	-172.142	1.081342	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	4387	10.3	1.8	1.8	1.8	1.066409	100	1	10.3	10.3	0	0	0	0	0	0	0	0	0	0	0;
	4388	728.4	107.1	558.524	-176.387	1.099237	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	4395	6.1	1.77	Inf	-Inf	1.066339	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4400	24.9	3.98	Inf	-Inf	1.036552	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4405	5.01	1.81	Inf	-Inf	1.080177	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4406	6.8	0.7	0.7	0.7	1.080053	100	1	6.8	6.8	0	0	0	0	0	0	0	0	0	0	0;
	4407	399.8	32.6	377.4793	-111.422	0.923591	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	4408	38.8	2.8	27.53551	-9.24479	1.023933	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	4409	239.6	64.6	133.8361	-49.8906	1.0762	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4410	173.6	63.4	192.0998	-53.1536	1.062789	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4411	219	19	187.3059	-57.2401	1.021136	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4413	88.8	33.6	76.52478	-23.1868	1.067091	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4425	17.89	7.56	Inf	-Inf	0.982253	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4427	28.21	6	Inf	-Inf	1.054868	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4433	87.7	28.2	43.87105	-17.5297	1.078545	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4436	0.58	53.03	Inf	-Inf	1.056306	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4442	95.6	20.4	75.61747	-23.9643	1.071818	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4451	357	50	176.3036	-72.865	1.084818	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	4456	50.8	15.2	37.58176	-12.232	1.043409	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4464	0.44	0	Inf	-Inf	1.085075	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4471	38.75	16.63	Inf	-Inf	1.060948	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4474	54.8	9	37.26511	-12.5747	1.099818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4479	6.43	1.13	Inf	-Inf	1.094008	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4485	20.2	2.5	34.39271	-33.6434	1.072579	100	1	102.7919	0	0	0	0	0	0	0	0	0	0	0	0;
	4492	1201	193.2	933.8733	-291.319	1.080737	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	4498	28	7.6	18.63635	-6.30424	0.983182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4499	385.5	13.5	224.402	-82.2158	1.087605	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4503	60	2.4	57.62968	-16.9993	1.058727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4506	54.8	15.6	37.22543	-12.6066	1.059182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4508	93.6	18.6	54.45642	-20.2598	1.054591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4510	962.8	85.6	533.3919	-201.124	1.088895	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	4519	375.2	32.2	182.2005	-75.8345	1.045955	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	4521	453.4	124.6	375.7286	-114.446	1.092211	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	4525	0.42	0	Inf	-Inf	1.060616	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4530	246	-5.5	132.4556	-51.7136	1.093263	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4535	22.8	16.4	Inf	-Inf	1.042295	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4536	14.77	4.59	Inf	-Inf	1.016203	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4538	4.89	2.4	Inf	-Inf	0.978572	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4542	0.12	1.74	Inf	-Inf	1.040962	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4543	360.4	35.6	173.7644	-71.7238	1.074816	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	4545	130.8	15.2	71.89431	-27.7171	0.996273	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4551	25.08	8.59	Inf	-Inf	1.076083	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4552	54.3	10	74.73169	-70.4102	1.065591	100	1	220.1212	0	0	0	0	0	0	0	0	0	0	0	0;
	4553	406	38.5	220.8797	-85.4495	1.097553	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4566	49.4	6.8	37.72001	-12.0877	1.060455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4567	35.2	11.6	38.87417	-10.9304	1.045136	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4571	7.5	1.5	1.5	1.5	1.026136	100	1	7.5	7.5	0	0	0	0	0	0	0	0	0	0	0;
	4573	119.8	32	93.57975	-29.1811	1.063	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	4574	21.83	4.73	Inf	-Inf	1.037937	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4580	75.2	14	35.0585	-14.8524	1.014	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4586	179.5	33	88.65206	-36.0981	1.053318	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4595	104.8	14.8	74.54863	-24.9693	1.033409	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4602	130.2	25.8	81.96635	-28.662	1.058868	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	4603	13.3	1.1	14.22468	-12.4344	1.095816	100	1	40.68366	0	0	0	0	0	0	0	0	0	0	0	0;
	4608	1008.3	184.5	532.9598	-203.194	1.103421	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	4610	542	33	258.9085	-109.582	1.041632	100	1	600	0	0	0	0	0	0	0	0	0	0	0	0;
	4612	19.07	6.06	Inf	-Inf	1.080718	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4617	55.7	6.6	25.91258	-10.9885	1.061864	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	4619	47.8	20.6	37.88687	-11.9287	1.069	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4626	242.4	18.6	101.3244	-48.1175	1.070682	100	1	249.06	0	0	0	0	0	0	0	0	0	0	0	0;
	4628	14.81	0.52	Inf	-Inf	1.06504	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4629	2.72	111.15	Inf	-Inf	1.059732	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4631	348	19.2	151.8106	-69.2526	1.090684	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	4632	416.4	42.9	221.6174	-84.9588	1.103658	100	1	500	166.666667	0	0	0	0	0	0	0	0	0	0	0;
	4636	2.27	1.81	Inf	-Inf	1.085077	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4637	21.1	6.2	6.2	6.2	1.086467	100	1	21.1	21.1	0	0	0	0	0	0	0	0	0	0	0;
	4646	55.83	31.99	Inf	-Inf	1.076205	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4648	42.2	10.6	38.2824	-11.4991	1.068545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4652	0.03	0.52	Inf	-Inf	0.99876	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4656	78	15.9	56.34794	-18.4204	1.079727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4657	1.33	0.62	Inf	-Inf	1.048658	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4658	20.01	7.01	Inf	-Inf	0.956255	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4670	308.8	42.8	179.8595	-65.5608	1.090316	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	4676	1713.6	180.4	882.8552	-343.242	1.082895	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	4677	422.6	143.8	450.4023	-391.11	1.042136	100	1	1290.5178	143.390867	0	0	0	0	0	0	0	0	0	0	0;
	4680	58.43	21.16	Inf	-Inf	1.069794	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4684	50.6	17.2	37.75053	-12.114	1.069409	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4686	28.3	4.2	28.42471	-8.31984	0.968333	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	4696	629.3	65.1	301.8902	-127.188	1.094105	100	1	700	233.333331	0	0	0	0	0	0	0	0	0	0	0;
	4703	19.2	5	32.8857	-32.2851	1.047	100	1	98.4694	0	0	0	0	0	0	0	0	0	0	0	0;
	4711	9.2	5.23	Inf	-Inf	1.051405	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4722	143.4	-9.3	79.52042	-30.8462	1.075727	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	4723	49.8	2	37.47262	-12.2604	1.041091	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4737	63.91	24.65	Inf	-Inf	1.030177	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4738	12.39	9.73	Inf	-Inf	1.069389	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4751	254.4	28.8	145.2776	-54.1518	1.086636	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	4752	249.6	48.8	145.5332	-53.8084	1.104553	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	4756	129	22.8	91.9066	-30.5371	1.021818	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	4761	403.7	39.6	193.5558	-80.8261	1.086227	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	4762	106.8	32.8	75.02463	-24.7419	1.066364	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4764	-1000	268.95	99999	-99999	1.05	100	1	1050	-1050	0	0	0	0	0	0	0	0	0	0	0;
	4771	4.88	2.79	Inf	-Inf	0.944964	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4773	139.6	28.8	71.43607	-28.3603	0.988545	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	4781	17.8	3.58	Inf	-Inf	1.039671	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4783	418.2	41.5	221.3348	-85.2247	1.102816	100	1	500	166.666667	0	0	0	0	0	0	0	0	0	0	0;
	4793	40.7	6.9	47.9121	-13.2946	1.053533	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	4795	-6.2	4.16	Inf	-Inf	1.021892	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4797	119.4	14.7	51.30161	-23.4931	1.073909	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	4799	5.3	0.7	0.7	0.7	1.063773	100	1	5.3	5.3	0	0	0	0	0	0	0	0	0	0	0;
	4803	151	20.5	91.72492	-32.914	1.0435	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4807	292.2	24.6	159.6655	-61.422	1.030273	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	4813	152.3	64.6	542.7318	-582.679	1.056658	100	1	1689.1358	51.185933	0	0	0	0	0	0	0	0	0	0	0;
	4815	102.6	16.2	94.65528	-27.9451	1.067455	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	4819	256	56.9	185.6771	-59.5635	1.072773	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4820	0.47	-0.1	Inf	-Inf	0.997323	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4823	44.9	6.9	132.5755	-139.968	1.087	100	1	409.09095	0	0	0	0	0	0	0	0	0	0	0	0;
	4824	89.74	38.63	Inf	-Inf	1.049752	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4835	9.86	3.08	Inf	-Inf	0.95776	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4836	41	6.6	38.41524	-11.3816	1.047364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4842	20.9	4.4	35.24747	-34.5986	1.060455	100	1	105.52765	0	0	0	0	0	0	0	0	0	0	0	0;
	4852	61.8	12.2	36.37445	-13.4287	1.052182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4859	10.8	0.9	11.36311	-10.0847	1.052136	100	1	32.74371	0	0	0	0	0	0	0	0	0	0	0	0;
	4873	30.7	-0.5	18.13765	-6.74603	1.095182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	4874	200	24.7	137.6496	-45.9109	1.092636	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	4875	42	14	38.34252	-11.4514	1.058909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4880	3.12	3.05	Inf	-Inf	1.022265	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4884	1250.6	308.4	926.4741	-297.626	1.101474	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	4891	126.55	54.72	Inf	-Inf	0.990794	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4901	55.2	16.2	36.81412	-12.8885	0.976364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4902	333.2	39.6	177.2869	-68.2141	1.054289	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	4903	35.28	8.32	Inf	-Inf	1.076242	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4904	91.8	7.4	42.68333	-18.5732	1.094237	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	4906	439	110.5	219.4251	-87.6007	1.021316	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	4912	220.4	12.8	102.6016	-44.6926	1.060864	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	4914	359.1	50.9	174.7512	-70.6137	1.092763	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4921	18.62	6.66	Inf	-Inf	1.047569	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4925	181.6	13.2	108.2667	-39.2001	1.092737	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	4928	292.5	63.9	127.1233	-56.9614	1.039955	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	4931	4.86	2.12	Inf	-Inf	0.981264	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4936	24.79	11.15	Inf	-Inf	1.081788	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4937	0.17	0.84	Inf	-Inf	1.042938	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4939	865.8	185.7	382.8671	-169.372	1.050579	100	1	900	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	4942	0.47	2.15	Inf	-Inf	0.951174	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4947	20.99	4.25	Inf	-Inf	0.988442	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4951	40	10.6	38.49277	-11.2965	1.061273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4954	15.89	6.42	Inf	-Inf	0.995489	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4955	43.4	11.2	38.26298	-11.5479	1.067632	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4956	102.68	29.02	Inf	-Inf	1.054592	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4959	18.5	3.7	31.08572	-30.7638	1.021727	100	1	93.46735	0	0	0	0	0	0	0	0	0	0	0	0;
	4961	38.8	18.07	Inf	-Inf	1.012138	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4964	519.6	75	263.395	-104.856	1.041263	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	4971	204.4	54.2	188.3489	-56.1134	1.018737	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	4973	1368.6	246.4	920.3133	-305.642	1.100684	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	4977	861.3	107.4	441.0395	-172.049	1.082763	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	4978	265.8	12.3	129.2974	-54.4309	1.031727	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	4979	51.8	13.8	37.49169	-12.3251	1.035545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4981	3424.8	452	1765.358	-686.674	1.079816	100	1	4000	1333.33333	0	0	0	0	0	0	0	0	0	0	0;
	4986	-195.74	32.87	Inf	-Inf	1.063786	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	4995	47.2	13.4	37.97503	-11.8597	1.021909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	4996	72.11	25.53	Inf	-Inf	1.093887	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5003	148.8	11.2	69.66893	-30.0037	1.038318	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5004	260	64.6	185.0586	-60.0879	1.042773	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5015	3.17	2.25	Inf	-Inf	1.078794	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5016	8.71	19.54	Inf	-Inf	1.070866	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5020	59	20.2	36.944	-12.9418	1.021364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5023	173	25.5	88.87103	-35.7095	1.084273	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5026	0.53	0.91	Inf	-Inf	1.069062	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5034	21.5	6.2	19.15591	-5.7534	1.079182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5036	0.94	6.39	Inf	-Inf	0.994643	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5044	571.8	124.2	255.9258	-112.203	1.049026	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	5052	57.6	9	36.80262	-12.9898	1.072818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5060	79.6	2.8	55.21395	-18.4759	1.085591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5071	0.19	14.8	Inf	-Inf	1.050954	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5080	37.1	5.3	17.51672	-7.42136	1.078455	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5090	7.02	3.47	Inf	-Inf	1.042074	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5092	33.7	-2.8	17.99491	-6.95173	1.101816	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5098	136.5	33.9	80.7264	-29.7409	1.05	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5106	9.84	4.46	Inf	-Inf	0.948178	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5110	0.47	0.83	Inf	-Inf	0.9545	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5115	-1.38	1.15	Inf	-Inf	1.050197	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5118	108.57	27.54	Inf	-Inf	1.04324	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5123	104.6	17.4	52.51436	-21.2352	1.004818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5130	7.51	2.52	Inf	-Inf	1.011642	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5131	5.65	3.71	Inf	-Inf	0.998327	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5132	0.21	1.04	Inf	-Inf	1.055681	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5133	234.4	66.2	187.0576	-57.9599	1.054409	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5135	229.8	49.2	103.1765	-46.2511	1.038053	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5141	128	21.2	72.7078	-27.0531	1.081	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5144	241.8	58.2	164.82	-56.0667	1.049227	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5145	76.2	1.8	34.46826	-15.3437	1.013182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5146	16.6	3.7	22.96798	-21.7009	1.036545	100	1	67.75512	0	0	0	0	0	0	0	0	0	0	0	0;
	5153	107.8	7	52.13783	-21.6313	1.068591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5155	41.8	2.4	27.24957	-9.54503	1.0404	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	5166	654.8	86.8	356.5338	-134.014	1.087632	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5178	33.4	2.3	28.17304	-8.65299	1.09675	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	5181	837	135.3	550.6898	-185.152	1.088605	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5184	7.43	3.65	Inf	-Inf	1.075474	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5186	74.8	14.6	55.4801	-18.128	1.048318	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5187	10	3.41	Inf	-Inf	1.048313	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5189	84.5	5.1	44.00077	-17.3317	1.098289	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	5194	122.7	22.8	82.65072	-27.9427	1.075636	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5198	19	4.2	31.98434	-31.5355	1.064318	100	1	95.9799	0	0	0	0	0	0	0	0	0	0	0	0;
	5204	130.56	56.18	Inf	-Inf	1.037352	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5206	20.5	8.06	Inf	-Inf	1.08384	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5208	45.6	9.2	38.00595	-11.7795	0.979545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5213	84.5	15.4	35.6861	-16.4299	1.079682	100	1	85.13568	28.37856	0	0	0	0	0	0	0	0	0	0	0;
	5220	302.1	28.5	123.764	-59.9473	1.024727	100	1	306	102	0	0	0	0	0	0	0	0	0	0	0;
	5226	448.4	41	376.646	-113.744	1.098447	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5238	38.01	22.73	Inf	-Inf	1.084502	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5247	1652.7	291.3	779.2916	-324.794	1.067816	100	1	1800	599.999997	0	0	0	0	0	0	0	0	0	0	0;
	5251	31.7	4.6	23.3014	-17.979	1.004273	100	1	63.8392	0	0	0	0	0	0	0	0	0	0	0	0;
	5252	106.8	31.6	114.6652	-32.5647	1.039818	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5272	354.4	-28.4	169.1888	-75.03	0.994227	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	5273	34.7	5	17.41918	-7.41332	0.99	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5274	292.5	66.3	126.9759	-57.0775	1.025818	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	5276	371.8	82.5	196.9838	-77.174	1.053727	100	1	440	146.666663	0	0	0	0	0	0	0	0	0	0	0;
	5282	22.6	5.2	45.82579	-46.3016	1.065364	100	1	138.85716	0	0	0	0	0	0	0	0	0	0	0	0;
	5283	257.4	19.5	185.2292	-59.899	1.103763	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5284	405.5	53	220.7483	-85.5432	1.082474	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	5285	935.4	153.3	539.2236	-196.321	1.089868	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5286	171	3.9	76.50068	-34.0491	1.002682	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5288	13.2	1	13.93053	-12.2653	1.081591	100	1	39.97989	0	0	0	0	0	0	0	0	0	0	0	0;
	5291	3.7	0.2	0.2	0.2	1.086947	100	1	3.7	3.7	0	0	0	0	0	0	0	0	0	0	0;
	5296	33.21	16.58	Inf	-Inf	1.042389	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5298	101.4	16	53.01273	-20.7603	1.042909	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5302	696.8	92.8	351.2482	-139.878	1.089227	100	1	800	0	0	0	0	0	0	0	0	0	0	0	0;
	5310	453.4	125.8	376.3911	-114.065	1.092368	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5311	411.6	30	214.1296	-84.9969	1.087409	100	1	480	159.999996	0	0	0	0	0	0	0	0	0	0	0;
	5314	0.7	2.28	Inf	-Inf	0.996478	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5317	68.2	7.2	35.734	-14.119	1.084553	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5318	625.2	120.4	358.2046	-131.844	1.054955	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5322	514.8	97.6	369.7199	-120.262	1.087263	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5328	56	8.2	46.52619	-14.5939	0.9792	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	5335	60	13.4	36.71858	-13.1328	1.067909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5336	106.8	17.4	52.96332	-21.8053	1.097763	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5337	1056.6	131.1	526.7683	-209.308	1.104658	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5339	107.8	7	52.11096	-21.6518	1.067136	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5347	182.6	43.4	86.2963	-36.3924	1.045455	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5349	55.6	19.4	37.31919	-12.5723	1.0245	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5367	57.59	30.52	Inf	-Inf	1.075388	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5369	1.25	4.59	Inf	-Inf	1.052776	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5371	16.2	7.6	19.48889	-5.3929	1.007818	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5379	160.4	11.2	68.32973	-31.4162	1.039955	100	1	164.92	0	0	0	0	0	0	0	0	0	0	0	0;
	5387	39.5	3.3	17.21198	-7.73312	1.055636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5388	51.2	10.6	37.52717	-12.2824	1.087711	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5402	47.8	2.4	37.77729	-11.9912	1.031818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5404	0.44	0	Inf	-Inf	1.069805	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5411	52.6	8	37.04554	-12.6281	0.986545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5417	151.8	37.2	172.4674	-48.3432	1.060182	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	5418	1221.2	350	934.545	-291.637	1.100526	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	5420	102.8	15.6	75.26134	-24.4359	1.094273	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5422	60.4	9.2	35.89595	-13.7154	0.966818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5424	310.7	10.3	179.909	-65.2683	0.979763	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5427	269.1	29.7	129.7179	-54.2207	1.077533	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	5430	55.2	9.4	37.27217	-12.591	1.087026	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5432	598.6	131.4	362.2147	-128.135	1.086553	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5433	1698.2	244.4	884.3854	-341.633	1.089921	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	5434	25.82	6.52	Inf	-Inf	1.081318	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5435	492	134.2	373.5819	-117.057	1.076947	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5438	1096	48	517.5506	-217.251	1.055318	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	5441	30.2	4.1	60.5129	-61.0826	1.069182	100	1	183.2727	0	0	0	0	0	0	0	0	0	0	0	0;
	5446	617.7	204.3	411.6599	-139.212	1.086447	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	5452	4.24	11.53	Inf	-Inf	1.09119	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5458	0.26	2.05	Inf	-Inf	1.017958	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5468	0.16	1.92	Inf	-Inf	1.066762	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5475	72.9	22.2	56.85985	-17.9276	1.018091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5478	820.2	119.4	390.3955	-161.157	1.052289	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	5486	0.2	0.98	Inf	-Inf	1.078087	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5487	697.2	165.6	351.6586	-138.952	1.049727	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	5503	0.51	0.64	Inf	-Inf	1.000917	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5510	14.58	5.96	Inf	-Inf	1.058392	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5524	74.8	11	34.67426	-15.1397	1.066053	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5526	60.4	7.6	36.58297	-13.2341	1.0825	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5528	776.7	95.7	395.5863	-155.881	1.094658	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	5536	382.19	126.6	185.6273	-59.5527	1.044636	100	1	527.29	260.623333	0	0	0	0	0	0	0	0	0	0	0;
	5539	164.5	18.8	141.3924	-42.3685	1.040364	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	5540	14.28	5.51	Inf	-Inf	1.056517	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5541	84.9	20.4	55.72582	-19.0656	0.978091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5545	114.95	52.9	Inf	-Inf	1.020195	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5547	11.38	3.43	Inf	-Inf	1.070914	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5553	0.14	2.32	Inf	-Inf	1.07896	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5554	29.95	10.04	Inf	-Inf	1.063956	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5555	58.6	17.4	36.73408	-13.0721	1.047909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5561	356	-1.2	173.4863	-71.8009	1.088763	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	5562	49.4	11.2	37.68306	-12.1168	1.032727	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5563	602.4	102.6	415.2591	-136.327	1.099467	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	5569	40.1	2.3	16.9082	-7.99332	1.048091	100	1	41.16	0	0	0	0	0	0	0	0	0	0	0	0;
	5573	9.9	1.7	1.7	1.7	0.985818	100	1	9.9	9.9	0	0	0	0	0	0	0	0	0	0	0;
	5574	206.4	9.2	104.7581	-42.5984	1.038773	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	5582	95	18.3	249.649	-261.817	1.069816	100	1	769.93936	32.080807	0	0	0	0	0	0	0	0	0	0	0;
	5595	21.82	8.63	Inf	-Inf	1.074803	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5597	26.3	3.1	19.85663	-14.9319	1.087545	100	1	53.7551	0	0	0	0	0	0	0	0	0	0	0	0;
	5598	39	10	38.47819	-11.2711	1.028591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5601	6.3	3.18	Inf	-Inf	0.984977	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5606	70.2	8.8	35.42857	-14.4123	1.047364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5608	55.8	16.2	57.93207	-16.6968	1.038909	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5616	64.68	21.72	Inf	-Inf	1.086483	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5617	2.91	0.63	Inf	-Inf	1.057064	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5624	24.44	9.58	Inf	-Inf	1.070872	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5626	310.4	43.2	178.9671	-66.249	1.0855	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	5629	767.5	36	447.1289	-164.561	1.068591	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	5631	7.62	2.52	Inf	-Inf	1.072851	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5632	411.8	5.2	274.1207	-93.0109	1.099211	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	5646	11.79	22.64	Inf	-Inf	1.063403	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5656	879.6	244.8	545.677	-189.968	1.086474	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5661	5.3	0.5	0.5	0.5	1.071364	100	1	5.3	5.3	0	0	0	0	0	0	0	0	0	0	0;
	5664	81.3	10.5	55.51535	-19.0839	1.015364	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5665	2.29	2.48	Inf	-Inf	1.053974	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5670	0.6	1.58	Inf	-Inf	1.082017	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5671	333.2	40.4	177.2869	-68.2141	1.054079	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	5673	10.77	5.53	Inf	-Inf	1.048972	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5687	14.81	4.65	Inf	-Inf	0.99851	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5689	153.93	44.76	Inf	-Inf	1.079629	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5697	44.6	8.8	38.09102	-11.6929	0.979455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5698	117.9	2.4	51.24997	-23.4864	1.037955	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5699	0.28	5.54	Inf	-Inf	1.07943	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5701	91.9	7.4	42.68333	-18.5732	1.094237	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	5707	55.5	24.3	58.11852	-16.5976	1.042182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5715	0.57	1	Inf	-Inf	1.04657	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5734	0.2	1.64	Inf	-Inf	1.068652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5740	223.4	-1.7	134.0765	-49.1543	1.027533	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	5742	48.8	8.8	37.75939	-12.0502	1.089773	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5743	96	3	96.55635	-27.9482	1.096	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5745	66.6	18.2	56.24379	-17.3592	1.014727	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5746	-92.76	46.04	Inf	-Inf	1.043413	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5749	29.5	4.8	21.92169	-16.8577	0.998	100	1	59.9797	0	0	0	0	0	0	0	0	0	0	0	0;
	5762	43.8	16.8	38.28945	-11.556	1.063409	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5769	49.8	33.8	37.69186	-12.1244	0.954182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5774	0.48	0	Inf	-Inf	0.977652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5777	188.9	51.5	139.0741	-44.6174	1.098447	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	5779	114	8.4	51.11374	-23.4556	1.008409	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5783	124	28	72.49993	-27.059	1.011955	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	5784	318.4	-1.2	176.6996	-68.1513	1.079342	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	5785	18.69	7.9	Inf	-Inf	1.065535	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5786	223.2	32	148.6542	-50.7336	1.072273	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	5791	22.18	7.71	Inf	-Inf	1.075959	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5794	0.6	0.89	Inf	-Inf	0.99733	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5798	18.68	4.52	Inf	-Inf	1.064842	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5813	4.88	2.77	Inf	-Inf	0.938651	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5816	958.2	223.5	537.0295	-198.633	1.066684	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5817	191.2	2.6	83.57609	-38.7814	0.996409	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	5821	1156.8	244.2	507.5835	-228.203	1.068842	100	1	1200	399.999996	0	0	0	0	0	0	0	0	0	0	0;
	5824	0.47	0.84	Inf	-Inf	0.952368	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5826	0.18	2.28	Inf	-Inf	1.083153	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5831	171	28	88.83223	-35.6876	1.074727	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	5834	291.2	16	180.9344	-64.1998	1.033591	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	5836	837.3	281.1	551.3255	-184.728	1.062955	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5838	5.2	0.8	0.8	0.8	1.038864	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	5850	31.9	17.4	18.35941	-6.62741	0.943909	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	5859	1.5	41.98	Inf	-Inf	1.074241	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5860	91.2	1.4	43.08973	-18.2391	1.089263	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	5861	77.4	12.9	86.16914	-24.2529	1.024533	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	5864	9.88	3.96	Inf	-Inf	0.973755	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5867	22.2	5.79	Inf	-Inf	1.094112	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5873	872.4	135.6	438.7154	-174.146	1.073318	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	5878	0.21	1.04	Inf	-Inf	1.055674	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5881	90	21.3	55.3415	-19.5146	1.023182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	5882	1170.9	48	508.021	-227.301	1.089921	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	5886	0.12	1.74	Inf	-Inf	1.041142	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5889	304.8	15.3	180.0197	-65.0036	1.032237	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5894	701.6	100	348.8315	-141.855	1.071789	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	5898	49.6	13.8	37.5588	-12.1921	1.024091	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5902	0.04	1.34	Inf	-Inf	1.059239	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5909	0.18	1	Inf	-Inf	1.086417	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5911	228.2	42	126.7283	-47.8284	1.087182	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	5912	14.73	9.85	Inf	-Inf	0.980909	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5913	0.19	1.94	Inf	-Inf	1.068152	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5925	15.91	6	Inf	-Inf	1.040241	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5926	0.65	2.03	Inf	-Inf	1.007427	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5941	49.2	12.6	37.57981	-12.1648	1.053545	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	5943	37.81	21.64	Inf	-Inf	1.018516	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5957	26.78	23.26	Inf	-Inf	1.080647	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5971	22.27	8.9	Inf	-Inf	1.095569	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5976	219	68.3	188.4472	-56.6116	1.057273	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	5979	606.6	104.1	414.3165	-137.058	1.059909	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	5986	14.12	6.04	Inf	-Inf	1.019773	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	5993	43.7	5.8	47.5639	-13.5606	1.013	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	5997	3.71	2.15	Inf	-Inf	1.050203	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6004	290.7	84.3	125.5963	-58.1225	1.070318	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6006	37.33	11.88	Inf	-Inf	1.081926	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6008	31.9	11.7	18.32016	-6.65635	1.010636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6009	99.5	5.4	42.14038	-19.2346	1.024545	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	6011	0.44	0	Inf	-Inf	1.096201	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6014	2027.7	155.4	1378.193	-459.093	1.100184	100	1	3000	999.999999	0	0	0	0	0	0	0	0	0	0	0;
	6020	11.13	3.06	Inf	-Inf	1.061794	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6022	37.8	13.6	38.53067	-11.1889	1.005364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6027	19.84	6.04	Inf	-Inf	0.992868	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6028	26.03	0.84	Inf	-Inf	1.095255	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6032	9.32	10.58	Inf	-Inf	1.062421	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6034	773.6	56	337.0381	-153.263	1.072091	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	6036	249	19	131.8789	-52.2396	1.081364	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6041	79.9	23.4	44.49301	-16.8156	0.993273	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	6042	0.2	1.64	Inf	-Inf	1.068652	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6057	33.2	9.3	28.15462	-8.65873	1.060182	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6058	21.3	1.7	36.22044	-35.5049	1.096211	100	1	108.37565	0	0	0	0	0	0	0	0	0	0	0	0;
	6066	9.9	5.32	Inf	-Inf	0.943281	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6069	20.1	4.5	Inf	-Inf	1.08393	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6079	989.7	34.2	533.9777	-201.865	1.105368	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	6084	218.82	93.95	Inf	-Inf	1.054613	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6091	32.8	0.7	18.02568	-6.89546	1.045409	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6107	14.86	5.19	Inf	-Inf	0.959147	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6108	662.2	106.1	311.5164	-132.093	1.087909	100	1	720	0	0	0	0	0	0	0	0	0	0	0	0;
	6110	670.8	42.9	405.8782	-144.96	1.079105	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6112	4.91	2.53	Inf	-Inf	1.026168	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6113	59.4	19.2	36.92949	-12.9664	1.057182	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6117	4.38	2.61	Inf	-Inf	1.075038	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6124	68.2	12.6	35.86946	-14.0181	1.070632	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6127	32.2	9.4	28.24344	-8.57106	1.060727	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6132	-8.62	5.68	Inf	-Inf	1.017453	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6135	9.9	1.6	1.6	1.6	0.988455	100	1	9.9	9.9	0	0	0	0	0	0	0	0	0	0	0;
	6149	276.8	53.2	183.4019	-62.0244	1.068079	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	6151	14.5	5.6	10.67677	-8.06034	1.077474	100	1	29.18592	4.86432	0	0	0	0	0	0	0	0	0	0	0;
	6155	6.77	2.13	Inf	-Inf	1.032778	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6162	32.6	1.8	28.21062	-8.60515	1.092773	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6166	949.7	181.4	429.6092	-183.481	1.086974	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	6168	6.84	0.61	Inf	-Inf	1.063195	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6173	7.81	2.67	Inf	-Inf	0.986089	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6174	4.56	2.13	Inf	-Inf	1.052874	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6175	74.8	27.2	77.143	-22.4628	1.072818	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6176	0.16	29.24	Inf	-Inf	0.9765	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6179	114.4	-12.8	93.15287	-29.2258	1.079868	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6183	76	24.8	77.35327	-22.2595	1.075455	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6187	15.9	2	11.98894	-9.05853	1.078273	100	1	32.53062	0	0	0	0	0	0	0	0	0	0	0	0;
	6188	251	-6	131.8117	-52.3576	1.093237	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6195	150.2	13.83	Inf	-Inf	0.923923	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6197	69.6	7.6	35.62461	-14.2456	1.058864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6198	0.12	1.74	Inf	-Inf	1.041152	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6201	0.15	1.93	Inf	-Inf	1.029304	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6204	578	107	255.2267	-112.308	1.082267	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	6205	10.76	4	Inf	-Inf	1.040728	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6208	89.7	24.9	54.68224	-19.958	0.973091	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6210	102.9	17.1	53.23502	-21.4697	1.052455	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6212	298	108	181.6255	-63.956	1.036579	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6215	27.5	3.9	18.55563	-6.34345	1.015	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6220	2463.6	632.4	1865.83	-585.945	1.086447	100	1	4000	1333.33333	0	0	0	0	0	0	0	0	0	0	0;
	6224	42	14	38.34252	-11.4514	1.058864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6227	64.6	16	36.07788	-13.7451	1.054789	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6236	362.8	-3.6	172.4173	-72.8485	1.0875	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	6239	19.8	2.5	33.56642	-32.9514	1.093026	100	1	100.50505	0	0	0	0	0	0	0	0	0	0	0	0;
	6246	0.12	1.31	Inf	-Inf	1.040622	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6250	242	56	189.3052	-59.8709	1.086045	100	1	400	133.33333	0	0	0	0	0	0	0	0	0	0	0;
	6251	12.49	5.83	Inf	-Inf	1.081051	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6252	21.09	9.94	Inf	-Inf	1.076	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6260	420.5	10	218.8106	-87.4814	1.100105	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	6261	37.84	19.23	Inf	-Inf	1.000005	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6266	735.3	87.4	327.0448	-146.461	1.088263	100	1	760	0	0	0	0	0	0	0	0	0	0	0	0;
	6269	116.8	32	104.7978	-87.0942	1.052026	100	1	295.83335	0	0	0	0	0	0	0	0	0	0	0	0;
	6273	173.4	0.4	87.27873	-35.3472	1.096447	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6279	37.34	13.33	Inf	-Inf	1.050657	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6282	131.2	22.4	56.01159	-25.3869	1.0475	100	1	132.8384	44.279467	0	0	0	0	0	0	0	0	0	0	0;
	6283	10.36	4.66	Inf	-Inf	1.039853	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6286	98.4	22.9	72.87702	-55.0102	1.076684	100	1	199.23232	33.205387	0	0	0	0	0	0	0	0	0	0	0;
	6289	33.7	3.2	28.11731	-8.69483	1.058267	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6291	826.2	24.6	388.7	-162.642	1.032045	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6292	354.36	101.97	188.1474	-56.9601	1.084727	100	1	530.46	263.793333	0	0	0	0	0	0	0	0	0	0	0;
	6300	227.5	42	126.0931	-48.2881	1.0795	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6301	319.2	91.2	179.3856	-66.228	1.010227	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6308	0.67	3.48	Inf	-Inf	1.016274	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6317	269	31.4	184.5344	-60.7314	1.047273	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6319	39.8	10.4	38.53143	-11.2704	1.086045	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6322	12.25	5	Inf	-Inf	1.085918	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6330	76.5	18.3	56.21861	-18.444	1.053818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6339	24.68	9.24	Inf	-Inf	0.99434	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6341	157.6	23.4	89.69457	-33.0531	1.084474	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6347	1872.8	207.4	862.0493	-363.971	1.101579	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	6356	862	92	401.302	-186.614	1.073182	100	1	960	0	0	0	0	0	0	0	0	0	0	0	0;
	6371	36.2	3.6	17.58638	-7.3383	0.999818	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6372	363	41	175.5077	-73.6979	1.078864	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6373	438.5	108	219.3174	-87.6839	1.030868	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	6374	1154.4	6.8	507.7266	-226.726	1.091553	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6377	199.6	63	190.4426	-54.8494	1.044091	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6380	65.8	8.2	36.16321	-13.7172	1.072864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6386	4.02	1.8	Inf	-Inf	1.019	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6393	55.5	27	58.1769	-16.5599	1.055409	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6400	490.8	108	267.3579	-100.985	1.052421	100	1	600	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	6402	50.4	17.4	37.59435	-12.2008	1.014636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6406	423.2	37.5	220.6567	-85.886	1.102368	100	1	500	166.666667	0	0	0	0	0	0	0	0	0	0	0;
	6408	199.5	13.8	136.871	-46.6784	1.037545	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	6412	2.01	1.08	Inf	-Inf	1.043125	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6418	12.8	0.6	13.40967	-11.8881	1.0245	100	1	38.62311	0	0	0	0	0	0	0	0	0	0	0	0;
	6426	292.8	65.1	126.9295	-57.1241	1.036591	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6430	41.8	12.6	38.34582	-11.4419	1.058342	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6434	119.2	34	93.13635	-29.4212	1.094605	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6444	234.2	11.2	133.0191	-50.2862	1.02	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6447	325.5	69.7	177.4357	-67.5455	1.007727	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6453	94.5	18.3	54.37656	-20.3491	1.077368	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6454	89.2	12.4	95.05883	-83.3387	1.074864	100	1	272.28426	0	0	0	0	0	0	0	0	0	0	0	0;
	6469	19.75	1.37	Inf	-Inf	1.072578	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6474	1120.4	43.2	516.035	-219.174	1.087026	100	1	1200	400	0	0	0	0	0	0	0	0	0	0	0;
	6480	6.24	3.42	Inf	-Inf	1.039663	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6488	35.7	8.8	17.6075	-7.30711	1.075	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6490	8.01	3.24	Inf	-Inf	0.955879	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6492	861.3	108.2	441.0395	-172.049	1.082711	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	6497	18.6	6.72	Inf	-Inf	1.052168	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6499	89.4	16.5	55.10733	-19.653	0.963182	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6502	56.4	3	37.06226	-12.7731	1.104553	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6506	73.19	26.77	Inf	-Inf	1.042571	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6514	1250.6	310.6	926.4741	-297.626	1.101368	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	6516	14.87	5.46	Inf	-Inf	0.967222	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6529	199	35.5	86.17981	-38.6108	0.983636	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6533	241.32	48.72	Inf	-Inf	1.02852	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6535	119.2	34.8	93.13635	-29.4212	1.094263	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6537	7.3	1.1	1.1	1.1	1.080091	100	1	7.3	7.3	0	0	0	0	0	0	0	0	0	0	0;
	6540	67	11.6	70.22175	-62.0219	1.062947	100	1	202.70352	0	0	0	0	0	0	0	0	0	0	0	0;
	6541	359.1	51.2	174.7512	-70.6137	1.092711	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6552	528	88.6	370.3469	-120.319	1.103895	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	6554	0.18	1	Inf	-Inf	1.065416	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6572	39.6	8	17.25551	-7.70147	1.044455	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6578	29.75	15.08	Inf	-Inf	1.018513	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6579	21.3	1.3	35.76807	-35.2298	1.076553	100	1	107.28645	0	0	0	0	0	0	0	0	0	0	0	0;
	6582	32	13.7	13.73162	-6.27692	1.058333	100	1	32.70408	10.90136	0	0	0	0	0	0	0	0	0	0	0;
	6588	168	28	89.83811	-34.8411	1.055947	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6591	766.8	155.7	394.8591	-156.123	1.104158	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6595	0.33	0	Inf	-Inf	1.079255	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6598	354.7	14.4	174.0095	-71.0502	1.087318	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6599	34.1	10	14.78255	-6.64399	1.075667	100	1	34.91837	11.639457	0	0	0	0	0	0	0	0	0	0	0;
	6601	2464.4	367.6	1862.294	-588.165	1.102974	100	1	4000	1333.33333	0	0	0	0	0	0	0	0	0	0	0;
	6611	50	10.6	37.71116	-12.1241	1.086591	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6616	119.69	27.08	Inf	-Inf	1.05427	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6620	858.9	29.7	384.3376	-167.039	1.086105	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6625	9.86	2.8	Inf	-Inf	0.953641	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6643	94	32.4	76.10756	-23.6156	1.059227	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6644	53	19.4	37.45443	-12.3922	1.019289	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6645	62.8	2.4	36.2567	-13.554	1.054727	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6646	399	41	167.1121	-77.8096	1.102079	100	1	401.01	133.67	0	0	0	0	0	0	0	0	0	0	0;
	6647	32.1	6.5	18.11857	-6.80864	1.079273	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	6664	1.73	0.76	Inf	-Inf	1.035572	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6672	116.8	9.2	51.00006	-22.788	1.092395	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6684	0.2	0.83	Inf	-Inf	1.063236	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6686	80.4	21.9	56.1777	-18.6229	1.020818	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6688	27.03	11.14	Inf	-Inf	1.049818	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6691	100.6	19.7	42.53724	-19.539	1.017773	100	1	101.397	33.799	0	0	0	0	0	0	0	0	0	0	0;
	6700	32.3	13.7	34.07797	-29.8701	1.067636	100	1	98.09091	10.89899	0	0	0	0	0	0	0	0	0	0	0;
	6708	242.4	71.7	134.0057	-50.1547	1.014955	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	6714	0.2	1	Inf	-Inf	1.072349	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6731	0.2	1	Inf	-Inf	1.087526	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6733	856.4	304.8	752.6096	-226.418	1.0855	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	6736	735.6	40.5	399.876	-151.376	1.039182	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6737	49.6	17.4	37.64625	-12.1398	1.041909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6745	9.78	2.96	Inf	-Inf	1.027532	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6750	13.3	2.7	14.28313	-12.5521	1.043773	100	1	40.95918	0	0	0	0	0	0	0	0	0	0	0	0;
	6752	0.06	1.72	Inf	-Inf	1.009791	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6768	37.9	4	27.69362	-9.11151	1.0432	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6769	41.47	18.47	75.85474	-77.0613	1.030409	100	1	233.84236	3.37	0	0	0	0	0	0	0	0	0	0	0;
	6770	11.14	3.38	Inf	-Inf	1.068287	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6774	18.95	6.93	Inf	-Inf	1.049596	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6776	2.43	3.34	Inf	-Inf	1.071462	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6777	76.5	9.9	56.26624	-18.4124	1.022636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6783	0.18	3.54	Inf	-Inf	1.048135	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6792	196.4	-20.4	83.00204	-39.4157	1.085289	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	6796	6.79	1.44	Inf	-Inf	1.057061	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6797	189.2	59.4	214.2317	-59.7698	1.061091	100	1	440	146.666663	0	0	0	0	0	0	0	0	0	0	0;
	6800	148.4	7.8	90.3117	-32.2944	0.986636	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	6805	5.2	0.4	0.4	0.4	1.055091	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	6809	717.9	30.9	402.1496	-149.167	1.085053	100	1	900	300	0	0	0	0	0	0	0	0	0	0	0;
	6819	14.89	4.46	Inf	-Inf	0.966052	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6820	22.93	9.6	Inf	-Inf	1.077045	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6821	702.4	113.6	350.3641	-140.731	1.032711	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	6823	508	144	371.1129	-119.669	1.040974	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	6826	8.66	6.01	Inf	-Inf	1.050629	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6841	9.08	3.86	Inf	-Inf	1.065664	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6844	14	3.62	Inf	-Inf	0.957659	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6849	5.91	2.55	Inf	-Inf	1.007689	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6851	125.92	53.57	Inf	-Inf	1.044003	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6858	0.03	1.09	Inf	-Inf	1.06807	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6866	100.8	9.9	53.40224	-21.2625	1.0385	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6882	18.27	12.99	Inf	-Inf	1.06958	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6903	179.2	44.1	131.8909	-42.6196	1.062273	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6911	375.8	79.2	276.9884	-89.8843	1.097184	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	6914	115.8	12	51.08731	-22.6943	1.067895	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	6919	147.69	58.76	Inf	-Inf	1.077468	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6921	289.4	77	180.6115	-64.062	1.016591	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6922	243	54.5	133.775	-50.0428	1.099	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	6924	43.4	4.8	45.84971	-40.5002	1.014045	100	1	131.81817	0	0	0	0	0	0	0	0	0	0	0	0;
	6932	37.4	7.7	27.79209	-9.02801	1.044773	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6934	769.2	68.3	378.2056	-162.352	1.0195	100	1	880	199.999998	0	0	0	0	0	0	0	0	0	0	0;
	6946	101.8	19.6	108.4714	-95.1824	1.027864	100	1	312.3366	34.704067	0	0	0	0	0	0	0	0	0	0	0;
	6947	378	77.2	172.4099	-72.9867	1.078921	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6948	102.8	8.4	74.7031	-24.7791	1.031091	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	6950	168.6	1.2	76.75097	-33.7898	1.054727	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	6951	0.21	2.37	Inf	-Inf	1.066194	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6956	1.39	39.72	Inf	-Inf	1.067664	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6958	32.2	3.9	28.04922	-8.67368	0.940267	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	6959	285.1	16.5	182.9465	-62.3341	1.094227	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	6964	46.6	9.6	38.03439	-11.8053	1.089579	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6967	22.2	5.79	Inf	-Inf	1.09413	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6968	367	62	176.1811	-73.3057	1.063409	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	6973	50	35	37.99934	-11.9442	0.975909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	6979	14.96	4.7	Inf	-Inf	1.070785	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	6982	228.2	42	126.7283	-47.8284	1.087182	100	1	280	0	0	0	0	0	0	0	0	0	0	0	0;
	6992	37.4	9.6	38.67011	-11.1061	0.997045	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7002	348	11.6	173.9112	-71.2165	1.053864	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7019	60	14.8	36.89859	-13.0017	1.073818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7023	24.99	4.94	Inf	-Inf	0.980864	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7024	20.4	2.5	19.09781	-5.74242	0.982867	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7025	229.6	-18.4	97.38247	-49.104	1.018455	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7034	6.3	0.7	0.7	0.7	1.089737	100	1	6.3	6.3	0	0	0	0	0	0	0	0	0	0	0;
	7035	0.58	-0.1	Inf	-Inf	0.986835	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7037	382.8	37.4	195.3043	-78.8309	1.102	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	7038	1425.1	173.8	700.1584	-281.594	1.105763	100	1	1600	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	7040	106.8	2.1	52.74746	-21.9661	1.033864	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7054	1221.2	335.2	934.725	-291.521	1.102158	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	7057	345.2	95.8	281.917	-85.8167	1.0968	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7062	212.4	41.4	105.8524	-43.6262	1.078	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7072	47.6	2.4	37.77729	-11.9912	1.031909	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7074	86.1	12.9	55.01944	-19.5943	0.996273	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7081	1250.4	259.6	926.3492	-297.678	1.092	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	7091	25.6	2.1	18.89849	-14.4403	1.064682	100	1	51.53768	0	0	0	0	0	0	0	0	0	0	0	0;
	7103	17.17	7.1	Inf	-Inf	1.068978	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7105	208.8	40.2	106.7723	-42.8119	1.039591	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7107	133.8	18.6	81.41106	-29.1694	1.075227	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	7108	14.9	1.1	1.1	1.1	0.995	100	1	14.9	14.9	0	0	0	0	0	0	0	0	0	0	0;
	7111	9.78	3.66	Inf	-Inf	0.976402	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7114	165	27.5	90.08493	-34.5619	1.072474	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7115	390.4	94.4	170.1382	-75.4637	1.049818	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	7123	3.53	4.62	Inf	-Inf	1.059798	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7125	321.3	-2	175.2932	-68.9541	0.943909	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	7131	397	64.6	380.358	-109.911	1.091763	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7144	34.25	8.56	Inf	-Inf	1.064234	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7148	1.56	0.81	Inf	-Inf	1.038913	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7151	436	105.5	217.1751	-89.2166	1.047868	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7158	9.88	3.18	Inf	-Inf	0.986899	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7160	26.3	5.2	18.76393	-6.16613	1.064273	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7165	216.8	29.6	149.2959	-50.0868	1.093273	100	1	320	0	0	0	0	0	0	0	0	0	0	0	0;
	7180	1368.6	256.4	920.408	-305.58	1.100789	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	7182	0.18	1.31	Inf	-Inf	1.043032	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7194	1403.6	134	698.0721	-282.166	1.083474	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	7196	0.67	2.78	Inf	-Inf	1.039859	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7198	14.91	5	Inf	-Inf	0.999118	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7199	207.9	-30.3	136.0283	-47.5511	1.073026	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	7205	411.8	5.8	274.1207	-93.0109	1.099132	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7216	313	106.8	282.7794	-84.1676	1.056132	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7223	419	124.4	274.1952	-93.2181	1.086026	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7239	1890.8	51.7	1487.057	-470.652	1.104789	100	1	3200	1066.66666	0	0	0	0	0	0	0	0	0	0	0;
	7243	127.52	54.47	Inf	-Inf	1.04406	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7245	3.55	1.43	Inf	-Inf	1.07169	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7254	811.8	128.7	391.5692	-160.947	1.076053	100	1	900	299.999997	0	0	0	0	0	0	0	0	0	0	0;
	7281	0.08	1.57	Inf	-Inf	1.024771	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7291	33.2	9.4	28.20379	-8.63105	1.059864	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	7297	96.8	28.4	75.67197	-23.9621	1.066182	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	7298	337.6	123.8	281.6325	-85.708	1.072727	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7299	10.7	2.3	2.3	2.3	1.003909	100	1	10.7	10.7	0	0	0	0	0	0	0	0	0	0	0;
	7316	399.4	50.2	276.9394	-90.7452	1.099842	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7317	10.2	1.4	10.72726	-9.53466	1.003182	100	1	30.93468	0	0	0	0	0	0	0	0	0	0	0	0;
	7331	0.67	3.26	Inf	-Inf	1.016359	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7334	392.7	4.9	177.2228	-80.2727	1.026682	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	7341	8.15	3.57	Inf	-Inf	0.981917	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7342	49.6	20.2	37.66225	-12.1378	1.043636	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7346	76.3	25.5	151.599	-153.823	1.069409	100	1	461.30652	0	0	0	0	0	0	0	0	0	0	0	0;
	7352	992	211.68	99999	-99999	1.052974	100	1	1041.6	0	0	0	0	0	0	0	0	0	0	0	0;
	7358	1036.4	119.6	525.0919	-210.367	1.068816	100	1	1200	399.999998	0	0	0	0	0	0	0	0	0	0	0;
	7360	75.8	14.2	34.53229	-15.2813	1.021818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7362	57.4	19	36.82699	-12.9573	1.018818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7372	56.8	19	36.87177	-12.9125	1.018773	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7381	1.07	0.72	Inf	-Inf	1.048327	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7385	0.84	0.96	Inf	-Inf	1.024915	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7391	178	26.4	87.42633	-35.3852	1.092421	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	7392	314	52.4	179.34	-66.1002	1.068658	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7393	153	19	91.66575	-33.0104	1.091182	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7394	700.7	51.8	291.7106	-137.405	1.086079	100	1	709.87	236.623331	0	0	0	0	0	0	0	0	0	0	0;
	7405	12.15	0.84	Inf	-Inf	1.094034	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7419	116.8	31.8	104.1908	-86.7384	1.052053	100	1	294.34675	0	0	0	0	0	0	0	0	0	0	0	0;
	7421	441	123	219.0976	-87.9121	1.029	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7428	0.18	1	Inf	-Inf	1.065427	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7429	15.25	6.42	Inf	-Inf	1.046293	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7434	0.02	-0.1	Inf	-Inf	0.951192	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7438	50.9	4.5	70.1748	-66.4543	1.020364	100	1	207.26904	0	0	0	0	0	0	0	0	0	0	0	0;
	7441	1329.2	61.6	709.4384	-271.313	1.086553	100	1	1600	533.333332	0	0	0	0	0	0	0	0	0	0	0;
	7443	49.8	15.8	37.47627	-12.2563	0.976364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7446	103.8	4.2	52.21295	-21.4456	1.075773	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7452	4.78	1.71	Inf	-Inf	0.995385	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7453	66.8	12.6	35.75759	-14.053	1.023316	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7454	401	131.8	276.8136	-90.8809	1.089237	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7456	17	7.2	19.49097	-5.42359	1.071136	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7469	21.2	1.6	36.33778	-35.5854	1.097921	100	1	108.67345	0	0	0	0	0	0	0	0	0	0	0	0;
	7471	396.9	40.6	177.815	-79.9253	1.034909	100	1	420	0	0	0	0	0	0	0	0	0	0	0	0;
	7477	253.8	35.4	164.2984	-56.8499	1.085237	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	7481	323.6	30.4	178.0869	-67.3184	1.059789	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7482	0.57	1	Inf	-Inf	1.046427	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7484	0.24	23.15	Inf	-Inf	1.082251	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7493	203.4	12	106.8417	-42.5638	1.090553	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7495	25	9.3	18.94217	-6.00423	1.015636	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7504	97.2	2.7	53.87593	-20.8004	1.025867	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7510	0.12	1.31	Inf	-Inf	1.041081	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7515	29.84	2.96	Inf	-Inf	1.056507	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7529	411.8	5.4	274.1207	-93.0109	1.099184	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7534	34.7	4.4	17.7956	-7.1323	1.073455	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7535	1702.2	355.8	1096.325	-374.644	1.087079	100	1	2400	799.999998	0	0	0	0	0	0	0	0	0	0	0;
	7536	9.75	2.74	Inf	-Inf	1.032678	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7550	116.36	49.54	Inf	-Inf	1.037295	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7555	20.88	16.63	Inf	-Inf	1.060754	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7556	0.21	405.11	Inf	-Inf	1.076169	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7558	2.3	3.34	Inf	-Inf	1.018404	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7560	124.8	36.4	72.47199	-27.1094	1.041467	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	7566	2.09	1.43	Inf	-Inf	1.043683	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7569	4.96	3.41	Inf	-Inf	1.02724	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7579	407	20.9	192.7011	-81.5855	1.072818	100	1	440	0	0	0	0	0	0	0	0	0	0	0	0;
	7586	242	43	185.6584	-59.0841	1.101632	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	7590	459.2	58.4	203.2176	-91.523	1.053316	100	1	480	0	0	0	0	0	0	0	0	0	0	0	0;
	7592	224.1	26.7	136.0398	-48.1043	0.982818	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	7593	33.8	2.7	17.78071	-7.11316	1.074136	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7598	331	-23	225.3638	-79.4871	0.994318	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7601	114.27	29.09	Inf	-Inf	1.04305	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7607	0.67	1.84	Inf	-Inf	1.056173	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7610	80.2	5.6	34.16326	-15.7098	1.060909	100	1	82.32	0	0	0	0	0	0	0	0	0	0	0	0;
	7613	443	134.8	377.6349	-112.991	1.076711	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7625	379.2	21.2	171.0144	-74.4335	1.095579	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7626	0.24	25.61	Inf	-Inf	1.091001	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7627	425	2.5	218.0701	-88.1844	1.099211	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7638	50	17.6	37.60763	-12.1862	1.044237	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7643	0.19	17.71	Inf	-Inf	1.053252	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7644	375.6	116	380.548	-109.053	1.093026	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7657	68.32	25.75	Inf	-Inf	1.04043	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7659	57.2	18.2	37.15246	-12.732	1.033364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7664	13.82	4.26	Inf	-Inf	1.041729	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7666	517.1	75.8	474.4186	-138.493	1.092553	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	7669	0.59	0.86	Inf	-Inf	1.033407	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7671	194.6	34	84.6123	-38.076	1.027909	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	7679	164.5	17	90.21334	-34.4517	1.0795	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7680	648.4	135.4	356.6585	-133.716	1.065447	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7687	79.5	26.7	55.64274	-18.9316	0.998636	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7689	0.17	1.82	Inf	-Inf	1.082959	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7692	78.8	11	34.15589	-15.6732	1.060684	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7693	6.61	3.62	Inf	-Inf	1.051092	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7696	0.19	1	Inf	-Inf	1.063342	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7699	8.03	3.48	Inf	-Inf	1.085263	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7703	-184.68	51.55	Inf	-Inf	1.060776	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7704	14.89	4.43	Inf	-Inf	0.998578	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7710	0.59	2.63	Inf	-Inf	1.011318	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7713	8.02	15.14	Inf	-Inf	1.058578	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7720	134	13	93.02833	-31.4293	1.073909	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	7726	8.3	1.1	1.1	1.1	1.062682	100	1	8.3	8.3	0	0	0	0	0	0	0	0	0	0	0;
	7734	194.8	1.8	83.35362	-39.0826	1.037318	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	7735	-2.3	299.03	99999	-99999	1.097605	100	1	-2.3	-2.3	0	0	0	0	0	0	0	0	0	0	0;
	7740	25.66	10.66	Inf	-Inf	1.067722	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7755	2486.5	237.3	1851.265	-596.857	1.102263	100	1	4000	1266.66666	0	0	0	0	0	0	0	0	0	0	0;
	7765	114.6	12	51.91257	-22.8621	1.050591	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7766	0.19	0.92	Inf	-Inf	1.081016	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7770	296.8	51.6	181.2367	-64.1896	1.068079	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7773	4.92	3.44	Inf	-Inf	0.99403	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7785	48.4	7	37.74531	-12.0404	1.012091	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7790	354.8	97.8	278.8385	-87.9664	1.070867	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7800	205.2	38.4	106.8707	-42.6208	1.050133	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7802	427	141.6	273.94	-93.6538	1.061184	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	7805	285	34	233.6468	-72.5884	1.084263	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7807	289.7	70	224.7033	-70.313	1.071227	100	1	480	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	7808	55.8	6.3	46.83548	-14.4116	1.0848	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	7818	70.45	35.14	Inf	-Inf	1.048247	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7819	100.8	0.9	52.65974	-21.8056	0.955	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	7820	157.2	34.4	110.42	-36.8838	1.076632	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7822	20.3	2.2	34.47588	-33.719	1.087632	100	1	103.0303	0	0	0	0	0	0	0	0	0	0	0	0;
	7827	298.2	-10.8	124.56	-59.1927	1.101579	100	1	300	99.999999	0	0	0	0	0	0	0	0	0	0	0;
	7837	264.3	15.9	131.333	-52.8395	1.104947	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	7842	24.78	9.83	Inf	-Inf	0.975256	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7847	13.91	5.49	Inf	-Inf	1.061374	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7856	0.12	1.74	Inf	-Inf	1.041657	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7857	461	-40.5	207.95	-97.1517	1.055789	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	7864	53.8	12.6	37.40093	-12.4562	1.050455	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7869	47	12.8	38.01728	-11.8298	1.078409	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7870	75.4	13.6	34.92256	-14.9646	1.026955	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7874	106.52	23.19	54.32125	-20.3883	1.073818	100	1	132.02	12.02	0	0	0	0	0	0	0	0	0	0	0;
	7876	61.97	26.49	Inf	-Inf	1.033702	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7878	0.15	152.82	Inf	-Inf	1.041474	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7882	10.16	5.44	Inf	-Inf	1.033472	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7884	62.2	4.8	36.39263	-13.4382	1.100605	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7892	142.8	60.6	173.2019	-47.6709	1.049227	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	7902	0.96	16.26	Inf	-Inf	0.99443	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7910	1232.6	301.6	932.6775	-293.145	1.086	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	7918	33	3.3	34.95247	-30.7993	1.0615	100	1	100.36365	0	0	0	0	0	0	0	0	0	0	0	0;
	7925	417.8	371	379.6843	-111.007	1.065867	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7941	23.09	46.44	Inf	-Inf	1.069492	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7946	738.2	98.6	346.4941	-144.145	1.089318	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	7949	0.26	2.35	Inf	-Inf	0.952927	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7950	386	42	168.924	-76.2989	1.074316	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	7952	330.4	59.4	178.0439	-67.2532	1.000364	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	7957	21.7	2.6	19.07478	-5.80385	1.025733	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	7974	4.4	2.45	Inf	-Inf	1.069628	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7979	166.2	23.4	111.7	-37.8503	1.094409	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	7986	-31.21	13.87	Inf	-Inf	1.042342	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7990	69	11	35.84041	-14.0654	1.078864	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	7991	23.3	9.39	Inf	-Inf	1.074093	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	7998	0.49	-0.1	Inf	-Inf	0.935847	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8000	82	15.6	76.87015	-22.7457	1.080636	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8001	9.78	2.41	Inf	-Inf	1.05607	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8002	118.8	21.2	92.95259	-29.5106	1.045818	100	1	200	66.666666	0	0	0	0	0	0	0	0	0	0	0;
	8004	229.98	97.55	192.7245	-52.4967	1.080474	100	1	467.78	201.113333	0	0	0	0	0	0	0	0	0	0	0;
	8032	76.9	6.1	44.89052	-16.4296	1.094263	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	8037	0.57	1	Inf	-Inf	1.044737	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8039	250.2	19.2	163.9089	-56.9781	1.068682	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8050	334.4	81.8	384.3012	-105.833	1.095868	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	8052	192.8	16	106.621	-40.7537	1.059636	100	1	240	0	0	0	0	0	0	0	0	0	0	0	0;
	8053	1232.6	306.8	932.6775	-293.145	1.085789	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	8056	443.8	124.4	272.7733	-95.0191	1.095	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8063	0.16	1.44	Inf	-Inf	1.069696	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8070	8.15	5.23	Inf	-Inf	1.081905	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8075	299.8	42.7	179.7341	-65.0445	1.093	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	8076	0.2	1	Inf	-Inf	1.064532	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8077	75	26.1	56.50421	-18.2121	1.060545	100	1	120	0	0	0	0	0	0	0	0	0	0	0	0;
	8083	17	2.3	28.99114	-28.4124	1.071053	100	1	86.7347	0	0	0	0	0	0	0	0	0	0	0	0;
	8092	50.4	10.2	37.19023	-12.4552	0.956818	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8097	6.01	2.92	Inf	-Inf	1.053731	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8098	38.8	4.5	17.26669	-7.66826	1.014909	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8102	21.72	8.34	Inf	-Inf	1.074792	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8105	37.2	6	38.63549	-11.1152	1.0132	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8108	0.56	0	Inf	-Inf	1.08598	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8111	10.07	3.91	Inf	-Inf	0.98531	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8114	2.35	2.06	Inf	-Inf	1.000447	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8120	88.4	-4.6	42.89557	-18.294	1.096545	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	8121	758.4	88.8	339.054	-151.205	1.087868	100	1	800	266.666664	0	0	0	0	0	0	0	0	0	0	0;
	8125	44.4	17.2	38.23025	-11.6107	1.051227	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8131	498.4	114.6	266.1681	-101.497	1.099658	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
	8136	38.9	5.1	53.52049	-50.3809	1.081545	100	1	157.57576	0	0	0	0	0	0	0	0	0	0	0	0;
	8139	0.59	2.43	Inf	-Inf	1.071689	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8141	203.1	9.9	136.9636	-46.7499	1.104632	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	8146	203	69.2	189.1979	-55.6225	1.050409	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	8147	25.3	5.2	18.69955	-6.16757	0.982545	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8148	1899	90	858.9986	-367.117	1.096921	100	1	2000	666.666666	0	0	0	0	0	0	0	0	0	0	0;
	8151	110	42	95.64927	-28.9496	1.061364	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8153	627.8	78	359.9627	-130.661	1.059364	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	8156	19.3	3.3	32.65772	-32.1828	1.044682	100	1	97.9798	0	0	0	0	0	0	0	0	0	0	0	0;
	8162	23	1.2	31.68029	-29.9156	1.094211	100	1	93.72588	0	0	0	0	0	0	0	0	0	0	0	0;
	8165	8.49	17.09	Inf	-Inf	1.054081	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8167	33.2	8.6	38.98063	-10.818	1.077091	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8169	78.2	13.2	107.5901	-101.48	1.071605	100	1	317.09088	0	0	0	0	0	0	0	0	0	0	0	0;
	8172	341.2	43.6	175.2955	-69.9556	1.066579	100	1	400	133.333332	0	0	0	0	0	0	0	0	0	0	0;
	8173	418	40	220.4901	-86.144	1.052227	100	1	500	166.666665	0	0	0	0	0	0	0	0	0	0	0;
	8174	1.58	0.66	Inf	-Inf	1.069094	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8185	31.2	9.8	28.36122	-8.46913	1.057409	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	8187	4.89	2.33	Inf	-Inf	0.993061	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8193	2.22	1.67	Inf	-Inf	1.06336	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8198	458	130.8	376.2517	-114.309	1.085211	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	8204	59.6	5	36.3536	-13.361	1.011273	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8205	19.83	13.51	Inf	-Inf	1.024332	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8206	22.37	8.9	Inf	-Inf	1.122614	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8208	5.2	0.8	0.8	0.8	1.050818	100	1	5.2	5.2	0	0	0	0	0	0	0	0	0	0	0;
	8209	72.99	26.65	Inf	-Inf	1.04259	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8212	34.1	3.3	17.82291	-7.08859	1.064364	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8215	340.4	24.8	176.1399	-69.3011	1.095079	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	8216	238	20.6	186.9582	-58.1393	1.005474	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	8217	144	15.5	92.44556	-32.1609	1.082909	100	1	200	0	0	0	0	0	0	0	0	0	0	0	0;
	8222	334.3	59.6	177.8749	-67.5089	1.082184	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	8223	10.9	1.1	11.60485	-10.232	1.033909	100	1	33.33672	0	0	0	0	0	0	0	0	0	0	0	0;
	8224	9.87	5.63	Inf	-Inf	1.013411	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8226	46.7	5.5	47.34312	-13.7808	0.952933	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	8228	73.6	11.8	34.79433	-15.0072	1.040684	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8230	41.57	23.4	Inf	-Inf	1.077674	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8233	2.37	1.59	Inf	-Inf	1.04416	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8234	369.6	70.4	198.075	-76.3385	1.093636	100	1	440	146.666663	0	0	0	0	0	0	0	0	0	0	0;
	8256	40.1	2	17.03131	-7.89505	1.055409	100	1	41.16	0	0	0	0	0	0	0	0	0	0	0	0;
	8258	84.5	4.9	44.00077	-17.3317	1.098421	100	1	100	0	0	0	0	0	0	0	0	0	0	0	0;
	8259	839.6	150.1	442.9899	-169.918	1.052447	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	8261	99.4	-12.7	40.58133	-20.483	1.100091	100	1	100	33.333333	0	0	0	0	0	0	0	0	0	0	0;
	8264	25.2	4.7	18.64871	-14.1561	1.058727	100	1	50.69346	0	0	0	0	0	0	0	0	0	0	0	0;
	8266	376.5	42.8	172.5303	-72.8475	1.095105	100	1	400	133.333333	0	0	0	0	0	0	0	0	0	0	0;
	8269	123.4	1	49.65318	-24.0691	1.060455	100	1	125.78	0	0	0	0	0	0	0	0	0	0	0	0;
	8273	0.22	3.34	Inf	-Inf	1.068143	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8279	39.7	7	17.27831	-7.68619	1.041182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8283	10.29	8.53	Inf	-Inf	1.050377	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8285	663	17.6	355.138	-135.291	1.089921	100	1	800	266.666666	0	0	0	0	0	0	0	0	0	0	0;
	8290	296.1	133	335.959	-93.2291	1.041591	100	1	700	233.333331	0	0	0	0	0	0	0	0	0	0	0;
	8292	121.2	24.8	73.18618	-26.4808	1.052273	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8302	860.8	118.3	439.3782	-173.285	1.079368	100	1	1000	333.333333	0	0	0	0	0	0	0	0	0	0	0;
	8309	5.41	1.08	Inf	-Inf	1.087296	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8312	368	61	175.9554	-73.5101	1.062091	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	8318	119.1	20.4	82.75521	-27.7465	1.093316	100	1	180	0	0	0	0	0	0	0	0	0	0	0	0;
	8319	299.2	47	126.7997	-57.0953	1.054273	100	1	301.01	100.336667	0	0	0	0	0	0	0	0	0	0	0;
	8320	-59.54	18.83	Inf	-Inf	1.081298	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8325	158.4	96.3	175.0039	-49.1586	1.048053	100	1	360	0	0	0	0	0	0	0	0	0	0	0	0;
	8330	0.68	2.44	Inf	-Inf	1.002513	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8331	72	22.4	77.68714	-21.9685	1.094545	100	1	160	0	0	0	0	0	0	0	0	0	0	0	0;
	8332	5.58	1.65	Inf	-Inf	1.05987	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8337	47.1	6.3	26.85162	-10.0095	1.054455	100	1	60	0	0	0	0	0	0	0	0	0	0	0	0;
	8344	65.8	21.4	36.14994	-13.7259	1.026364	100	1	80	0	0	0	0	0	0	0	0	0	0	0	0;
	8345	1.11	1.56	Inf	-Inf	1.019769	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8346	46.4	6.8	34.6463	-26.4201	1.050545	100	1	95.0204	0	0	0	0	0	0	0	0	0	0	0	0;
	8349	292.5	64.8	126.9042	-57.1442	1.036136	100	1	300	0	0	0	0	0	0	0	0	0	0	0	0;
	8357	4.78	1.72	Inf	-Inf	0.988967	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8363	28.3	1	18.19123	-6.61422	0.926182	100	1	40	0	0	0	0	0	0	0	0	0	0	0	0;
	8369	5.15	2.04	Inf	-Inf	1.084862	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8376	289.1	44.3	127.957	-55.878	1.006409	100	1	300	100	0	0	0	0	0	0	0	0	0	0	0;
	8377	13.62	5.62	Inf	-Inf	1.076407	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8380	123.89	52.61	Inf	-Inf	1.057568	100	1	Inf	-Inf	0	0	0	0	0	0	0	0	0	0	0;
	8381	359	89	176.3275	-72.9451	1.083789	100	1	400	0	0	0	0	0	0	0	0	0	0	0	0;
	8384	829.8	157.8	552.0842	-183.993	1.104105	100	1	1200	399.999999	0	0	0	0	0	0	0	0	0	0	0;
	8385	419	124.2	274.1952	-93.2181	1.086053	100	1	600	200	0	0	0	0	0	0	0	0	0	0	0;
];

%% branch data
%	fbus	tbus	r	x	b	rateA	rateB	rateC	ratio	angle	status	angmin	angmax
mpc.branch = [
	4679	2816	0.0006	0.00616	0.19426	837.866207	837.866207	837.866207	0	0	1	-360	360;
	4679	7951	0.0005	0.0074	0.16192	939.310725	939.310725	939.310725	0	0	1	-360	360;
	1976	2843	0.00107	0.00626	0.0176	489.131391	489.131391	489.131391	0	0	1	-360	360;
	1976	2940	0.000841	0.0055	0.00899	488.494654	488.494654	488.494654	0	0	1	-360	360;
	3197	645	0.00143	0.01563	0.4664	1660.15923	1660.15923	1660.15923	0	0	1	-360	360;
	3197	4123	0.00146	0.01692	0.38774	1742.65082	1742.65082	1742.65082	0	0	1	-360	360;
	384	4925	7e-05	0.00057	0.01238	1713.38817	1713.38817	1713.38817	0	0	1	-360	360;
	384	4276	6e-05	0.00056	0.01238	1530.30638	1530.30638	1530.30638	0	0	1	-360	360;
	5407	6736	0.003349	0.01656	0.02751	231.433719	231.433719	231.433719	0	0	1	-360	360;
	5407	2899	0.003409	0.016211	0.02803	286.353101	286.353101	286.353101	0	0	1	-360	360;
	4077	6736	0.003479	0.01586	0.02855	235.506394	235.506394	235.506394	0	0	1	-360	360;
	4077	2899	0.003099	0.017969	0.02543	282.911066	282.911066	282.911066	0	0	1	-360	360;
	480	8352	0.00127	0.01842	0.46004	830.576213	830.576213	830.576213	0	0	1	-360	360;
	480	756	0.00126	0.01847	0.48802	1552.79242	1552.79242	1552.79242	0	0	1	-360	360;
	1510	105	0.0013	0.01808	0.46926	1220.32104	1220.32104	1220.32104	0	0	1	-360	360;
	1510	3134	0.00131	0.01765	0.50568	1757.02217	1757.02217	1757.02217	0	0	1	-360	360;
	4625	1654	0.00065	0.01311	4.8838	2817.51882	2817.51882	2817.51882	0	0	1	-360	360;
	4625	5805	0.00067	0.00974	5.004478	2634.65323	2634.65323	2634.65323	0	0	1	-360	360;
	4733	7105	0.00236	0.01418	0.02922	566.739703	566.739703	566.739703	0	0	1	-360	360;
	4733	7048	0.002169	0.015909	0.02764	524.512373	524.512373	524.512373	0	0	1	-360	360;
	319	1163	0.00027	0.00254	0.08831	1378.41176	1378.41176	1378.41176	0	0	1	-360	360;
	319	6912	0.00027	0.00259	0.08492	1131.56622	1131.56622	1131.56622	0	0	1	-360	360;
	2367	1163	0.00033	0.00302	0.10365	1462.49417	1462.49417	1462.49417	0	0	1	-360	360;
	2367	818	0.00029	0.00312	0.08619	1630.6853	1630.6853	1630.6853	0	0	1	-360	360;
	1685	5933	0.000289	0.00137	0.22216	544.409342	544.409342	544.409342	0	0	1	-360	360;
	1685	6917	0.000219	0.00263	0.28226	281.672649	281.672649	281.672649	0	0	1	-360	360;
	1946	5933	0.000289	0.001341	0.22651	465.512314	465.512314	465.512314	0	0	1	-360	360;
	1946	7971	0.000229	0.00257	0.2878	289.387031	289.387031	289.387031	0	0	1	-360	360;
	5998	5933	0.000841	0.00689	0.00957	522.453554	522.453554	522.453554	0	0	1	-360	360;
	5998	1505	0.000979	0.00586	0.01124	503.907031	503.907031	503.907031	0	0	1	-360	360;
	5239	5933	0.00014	0.0012	0.0041	268.472275	268.472275	268.472275	0	0	1	-360	360;
	5239	7112	0.000211	0.001539	0.00302	280.884255	280.884255	280.884255	0	0	1	-360	360;
	4742	5933	0.000401	0.004901	0.01141	292.571856	292.571856	292.571856	0	0	1	-360	360;
	4742	7112	0.000219	0.00151	0.00617	544.34342	544.34342	544.34342	0	0	1	-360	360;
	7323	5933	0.000729	0.004539	0.00827	509.486382	509.486382	509.486382	0	0	1	-360	360;
	7323	2501	0.00062	0.00482	0.00701	512.761136	512.761136	512.761136	0	0	1	-360	360;
	2700	5509	0.00052	0.00407	0.12738	1486.45584	1486.45584	1486.45584	0	0	1	-360	360;
	2700	3130	0.00034	0.00502	0.10364	1862.3631	1862.3631	1862.3631	0	0	1	-360	360;
	6860	5509	0.00065	0.00698	0.21868	1729.12787	1729.12787	1729.12787	0	0	1	-360	360;
	6860	8006	0.00059	0.00806	0.19224	1708.37876	1708.37876	1708.37876	0	0	1	-360	360;
	6120	8045	0.00063	0.00464	0.01791	502.488759	502.488759	502.488759	0	0	1	-360	360;
	6120	4620	0.000599	0.005099	0.00753	540.018489	540.018489	540.018489	0	0	1	-360	360;
	4509	4351	0.00064	0.00307	0.00977	328.235579	328.235579	328.235579	0	0	1	-360	360;
	4509	6956	0.000659	0.003	0.00994	279.764726	279.764726	279.764726	0	0	1	-360	360;
	7816	6146	0.00053	0.00528	0.13049	1106.54092	1106.54092	1106.54092	0	0	1	-360	360;
	7816	7429	0.00056	0.0053	0.13554	1298.46535	1298.46535	1298.46535	0	0	1	-360	360;
	2826	4351	0.0018	0.007919	0.0274	283.888462	283.888462	283.888462	0	0	1	-360	360;
	2826	6489	0.001651	0.00862	0.02553	282.802085	282.802085	282.802085	0	0	1	-360	360;
	4932	6146	0.00045	0.00545	0.10855	1383.10326	1383.10326	1383.10326	0	0	1	-360	360;
	4932	818	0.00052	0.00476	0.12312	1549.81811	1549.81811	1549.81811	0	0	1	-360	360;
	2075	6238	0.01286	0.037421	0.00426	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2075	5131	0.032397	0.123	0.01071	57.157677	57.157677	57.157677	0	0	1	-360	360;
	428	6238	0.032851	0.069727	0.00686	57.157677	57.157677	57.157677	0	0	1	-360	360;
	428	7299	0.059942	0.112802	0.01253	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7184	6775	0.00011	0.00109	0.0275	1117.5233	1117.5233	1117.5233	0	0	1	-360	360;
	7184	8079	0.00011	0.00107	0.02724	1195.98947	1195.98947	1195.98947	0	0	1	-360	360;
	5649	6968	0.000461	0.00395	0.00707	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5649	1374	0.000599	0.003789	0.00905	609.489453	609.489453	609.489453	0	0	1	-360	360;
	6457	5956	0.00173	0.01542	0.41037	991.605046	991.605046	991.605046	0	0	1	-360	360;
	6457	3958	0.00166	0.02083	0.38623	1139.93957	1139.93957	1139.93957	0	0	1	-360	360;
	3151	8031	0.006314	0.018488	0.0014	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3151	1823	0.015727	0.037669	0.00402	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6856	6402	0.00374	0.02095	0.02838	813.890259	813.890259	813.890259	0	0	1	-360	360;
	6856	3936	0.00511	0.019461	0.03356	805.730048	805.730048	805.730048	0	0	1	-360	360;
	975	2598	0.011281	0.050771	0.08633	470.645455	470.645455	470.645455	0	0	1	-360	360;
	975	2345	0.0115	0.049719	0.088	449.934942	449.934942	449.934942	0	0	1	-360	360;
	3594	2598	0.011719	0.048659	0.08965	452.083309	452.083309	452.083309	0	0	1	-360	360;
	3594	2345	0.01193	0.047599	0.09132	454.348277	454.348277	454.348277	0	0	1	-360	360;
	117	2691	0.00064	0.00715	0.21543	1506.89297	1506.89297	1506.89297	0	0	1	-360	360;
	117	4000	0.00089	0.00712	0.21232	1309.97954	1309.97954	1309.97954	0	0	1	-360	360;
	4587	628	0.00048	0.00555	0.12144	1001.31056	1001.31056	1001.31056	0	0	1	-360	360;
	4587	7131	0.00047	0.00566	0.11886	1106.93978	1106.93978	1106.93978	0	0	1	-360	360;
	6245	628	0.00046	0.00524	0.11627	1061.99403	1061.99403	1061.99403	0	0	1	-360	360;
	6245	7131	0.00052	0.00514	0.13178	1066.33341	1066.33341	1066.33341	0	0	1	-360	360;
	1324	4099	0.00438	0.010686	0.00092	72.65579	72.65579	72.65579	0	0	1	-360	360;
	1324	6038	0.031347	0.062653	0.00639	72.65579	72.65579	72.65579	0	0	1	-360	360;
	5891	4499	0.00029	0.00298	0.08127	1628.17106	1628.17106	1628.17106	0	0	1	-360	360;
	5891	5519	0.0003	0.00291	0.08283	1543.31858	1543.31858	1543.31858	0	0	1	-360	360;
	1258	8182	0.00155	0.01529	0.48969	1213.99791	1213.99791	1213.99791	0	0	1	-360	360;
	1258	2862	0.00148	0.0152	0.35671	1311.67567	1311.67567	1311.67567	0	0	1	-360	360;
	7887	7378	0.000169	0.00096	0.00546	322.117283	322.117283	322.117283	0	0	1	-360	360;
	7887	5621	0.000182	0.000942	0.00558	322.117283	322.117283	322.117283	0	0	1	-360	360;
	5142	381	0.00016	0.00184	0.05863	2676.56513	2676.56513	2676.56513	0	0	1	-360	360;
	5142	6559	0.00022	0.00287	0.072	2676.56513	2676.56513	2676.56513	0	0	1	-360	360;
	3506	5894	0.00015	0.00187	0.05744	2521.34013	2521.34013	2521.34013	0	0	1	-360	360;
	3506	4345	0.00022	0.00266	0.07043	2521.34013	2521.34013	2521.34013	0	0	1	-360	360;
	610	8066	0.00138	0.0174	0.4384	1321.99	1321.99	1321.99	0	0	1	-360	360;
	610	2834	0.00272	0.0288	0.87792	1111.99789	1111.99789	1111.99789	0	0	1	-360	360;
	7365	4709	0.00065	0.01024	0.22418	1854.81642	1854.81642	1854.81642	0	0	1	-360	360;
	7365	5452	0.00073	0.01004	0.25405	1897.31506	1897.31506	1897.31506	0	0	1	-360	360;
	8112	4709	0.00075	0.00983	0.25905	1927.10557	1927.10557	1927.10557	0	0	1	-360	360;
	8112	5452	0.00076	0.00963	0.26401	1937.40542	1937.40542	1937.40542	0	0	1	-360	360;
	6482	6695	0.00042	0.02666	0	1205.77988	1205.77988	1205.77988	0	0	1	-360	360;
	3646	8036	0.00029	0.0026	0.11024	2185.86218	2185.86218	2185.86218	0	0	1	-360	360;
	3646	5247	0.00022	0.00303	0.09849	1209.37815	1209.37815	1209.37815	0	0	1	-360	360;
	6344	3424	0.001	0.006159	0.01409	267.888504	267.888504	267.888504	0	0	1	-360	360;
	6344	8264	0.00113	0.00657	0.01345	543.343542	543.343542	543.343542	0	0	1	-360	360;
	1102	7486	0.0013	0.01277	0.3118	1318.15149	1318.15149	1318.15149	0	0	1	-360	360;
	1102	6646	0.00163	0.01808	0.39172	1409.43965	1409.43965	1409.43965	0	0	1	-360	360;
	7841	95	0.00047	0.00406	0.10714	1325.91472	1325.91472	1325.91472	0	0	1	-360	360;
	7841	5337	0.00235	0.02343	0.53841	1334.22292	1334.22292	1334.22292	0	0	1	-360	360;
	86	5816	0.00036	0.00445	0.18766	1579.1887	1579.1887	1579.1887	0	0	1	-360	360;
	86	1498	0.00017	0.0025	0	1911.92861	1911.92861	1911.92861	0	0	1	-360	360;
	3987	5816	0.00036	0.00436	0.19134	1593.21911	1593.21911	1593.21911	0	0	1	-360	360;
	3987	3164	0.00017	0.00255	0	1817.45026	1817.45026	1817.45026	0	0	1	-360	360;
	1725	2919	0.00096	0.00995	0.22139	1380.46923	1380.46923	1380.46923	0	0	1	-360	360;
	1725	6159	0.00152	0.01949	0.37613	1402.56233	1402.56233	1402.56233	0	0	1	-360	360;
	4828	2919	0.00094	0.01014	0.21687	1268.95325	1268.95325	1268.95325	0	0	1	-360	360;
	4828	6159	0.00149	0.01986	0.36829	1430.78506	1430.78506	1430.78506	0	0	1	-360	360;
	2614	2919	0.00107	0.01243	0.24707	1178.38449	1178.38449	1178.38449	0	0	1	-360	360;
	2614	2342	0.00102	0.01034	0.24764	1234.99251	1234.99251	1234.99251	0	0	1	-360	360;
	3381	4411	0.0047	0.02061	0.03819	447.966813	447.966813	447.966813	0	0	1	-360	360;
	3381	3355	0.00739	0.040531	0.05525	404.795239	404.795239	404.795239	0	0	1	-360	360;
	5504	7837	0.00049	0.00611	0.19795	1947.28074	1947.28074	1947.28074	0	0	1	-360	360;
	5504	6646	0.00035	0.00431	0.0827	1567.79298	1567.79298	1567.79298	0	0	1	-360	360;
	4882	937	0.00261	0.01344	0.0252	437.660522	437.660522	437.660522	0	0	1	-360	360;
	4882	3204	0.003521	0.019591	0.03391	429.711413	429.711413	429.711413	0	0	1	-360	360;
	6232	3117	0.0025	0.01419	0.03183	372.206218	372.206218	372.206218	0	0	1	-360	360;
	6232	4589	0.00287	0.0237	0.04153	385.660754	385.660754	385.660754	0	0	1	-360	360;
	6070	6116	0.00088	0.00956	0.22766	1569.53387	1569.53387	1569.53387	0	0	1	-360	360;
	6070	261	0.00095	0.00877	0.19307	1400.70692	1400.70692	1400.70692	0	0	1	-360	360;
	1875	682	0.00332	0.020651	0.02957	470.772726	470.772726	470.772726	0	0	1	-360	360;
	1875	2772	0.003969	0.018781	0.03027	471.867867	471.867867	471.867867	0	0	1	-360	360;
	2622	2483	0.002849	0.014909	0.02173	445.358899	445.358899	445.358899	0	0	1	-360	360;
	2622	1435	0.00439	0.02451	0.0346	451.864585	451.864585	451.864585	0	0	1	-360	360;
	1389	7878	0.0002	0.0018	0.07654	2095.83247	2095.83247	2095.83247	0	0	1	-360	360;
	1389	7605	0.00025	0.00261	0.06775	2102.64923	2102.64923	2102.64923	0	0	1	-360	360;
	6681	7878	0.0002	0.00176	0.07802	2134.4235	2134.4235	2134.4235	0	0	1	-360	360;
	6681	123	0.00025	0.00255	0.06908	1937.40542	1937.40542	1937.40542	0	0	1	-360	360;
	928	5603	0.012488	0.047413	0.00424	67.329076	67.329076	67.329076	0	0	1	-360	360;
	928	6509	0.010479	0.027711	0.00207	85.815012	85.815012	85.815012	0	0	1	-360	360;
	2363	6761	0.00044	0.00428	0.12975	1833.26566	1833.26566	1833.26566	0	0	1	-360	360;
	2363	1217	0.00036	0.00475	0.10092	1994.3919	1994.3919	1994.3919	0	0	1	-360	360;
	4306	4764	0.00058	0.00534	0.13981	1877.33866	1877.33866	1877.33866	0	0	1	-360	360;
	4306	2861	0.00054	0.00589	0.13209	1858.83724	1858.83724	1858.83724	0	0	1	-360	360;
	2670	1301	0.00206	0.02057	0.51797	1174.66906	1174.66906	1174.66906	0	0	1	-360	360;
	2670	1406	0.00077	0.00872	0.20069	961.484786	961.484786	961.484786	0	0	1	-360	360;
	3435	8333	0.00161	0.01963	0.40525	1358.87042	1358.87042	1358.87042	0	0	1	-360	360;
	3435	1406	0.00075	0.00888	0.19651	1403.07505	1403.07505	1403.07505	0	0	1	-360	360;
	1771	6583	0.00016	0.00152	0.04776	1657.35934	1657.35934	1657.35934	0	0	1	-360	360;
	1771	4971	0.00019	0.00163	0.04733	1717.03185	1717.03185	1717.03185	0	0	1	-360	360;
	4966	6583	0.00015	0.0016	0.04863	1564.93846	1564.93846	1564.93846	0	0	1	-360	360;
	4966	4971	0.00015	0.00183	0.03978	1665.4352	1665.4352	1665.4352	0	0	1	-360	360;
	3318	2990	0.00121	0.01426	0.55448	1444.7332	1444.7332	1444.7332	0	0	1	-360	360;
	3318	5013	0.00116	0.01456	0.60295	1466.36887	1466.36887	1466.36887	0	0	1	-360	360;
	407	128	0.00068	0.00727	0.22697	1800.68973	1800.68973	1800.68973	0	0	1	-360	360;
	407	1459	0.0007	0.00734	0.22711	1898.6742	1898.6742	1898.6742	0	0	1	-360	360;
	7167	128	0.00076	0.00712	0.25723	1607.73459	1607.73459	1607.73459	0	0	1	-360	360;
	7167	1459	0.0008	0.00719	0.25739	1927.51496	1927.51496	1927.51496	0	0	1	-360	360;
	1919	1442	0.0002	0.0022	0.062275	1244.20966	1244.20966	1244.20966	0	0	1	-360	360;
	1919	6722	0.0002	0.0022	0.062275	1261.68234	1261.68234	1261.68234	0	0	1	-360	360;
	245	1869	0.000205	0.00215	0.06346	1418.85688	1418.85688	1418.85688	0	0	1	-360	360;
	245	6722	0.000205	0.00215	0.06346	1453.65613	1453.65613	1453.65613	0	0	1	-360	360;
	5385	7644	0.00053	0.00589	0.16224	1317.58283	1317.58283	1317.58283	0	0	1	-360	360;
	5385	740	0.00046	0.00675	0.14064	1137.86565	1137.86565	1137.86565	0	0	1	-360	360;
	7742	7644	0.00054	0.00577	0.1653	1330.82013	1330.82013	1330.82013	0	0	1	-360	360;
	7742	740	0.00045	0.00625	0.13758	1096.47012	1096.47012	1096.47012	0	0	1	-360	360;
	1508	1408	0.00275	0.02752	0.56046	1411.80843	1411.80843	1411.80843	0	0	1	-360	360;
	1508	6962	0.00184	0.01515	0.34348	1346.9758	1346.9758	1346.9758	0	0	1	-360	360;
	2112	3779	0.00438	0.02432	0.03353	333.346238	333.346238	333.346238	0	0	1	-360	360;
	2112	543	0.0043	0.01949	0.03241	339.998629	339.998629	339.998629	0	0	1	-360	360;
	7591	6347	0.00033	0.00269	0.0779	1242.21208	1242.21208	1242.21208	0	0	1	-360	360;
	7591	4460	0.00243	0.01904	0.5944	1144.88118	1144.88118	1144.88118	0	0	1	-360	360;
	853	6347	0.00034	0.00263	0.07939	1144.71598	1144.71598	1144.71598	0	0	1	-360	360;
	853	4460	0.00248	0.01864	0.60583	1111.11461	1111.11461	1111.11461	0	0	1	-360	360;
	6125	5616	0.00122	0.01249	0.30518	1013.27297	1013.27297	1013.27297	0	0	1	-360	360;
	6125	3694	0.00106	0.01031	0.38362	1081.29119	1081.29119	1081.29119	0	0	1	-360	360;
	3009	5616	0.00119	0.01273	0.29895	999.643389	999.643389	999.643389	0	0	1	-360	360;
	3009	3694	0.00108	0.01009	0.39072	1111.15674	1111.15674	1111.15674	0	0	1	-360	360;
	1257	4008	0.001039	0.013711	0.03649	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1257	4677	0.00114	0.011409	0.04302	265.692506	265.692506	265.692506	0	0	1	-360	360;
	7122	5332	0.001021	0.013969	0.03573	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7122	4677	0.001169	0.011159	0.04381	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7257	7434	0	0.000271	0	116.419016	116.419016	116.419016	0	0	1	-360	360;
	2111	7434	0	0.000271	0	107.991376	107.991376	107.991376	0	0	1	-360	360;
	763	136	0.009372	0.035702	0.00318	57.157677	57.157677	57.157677	0	0	1	-360	360;
	763	3273	0.029157	0.103686	0.0096	62.300153	62.300153	62.300153	0	0	1	-360	360;
	479	5017	0.00065	0.00673	0.27895	1520.73579	1520.73579	1520.73579	0	0	1	-360	360;
	479	3793	0.00066	0.00658	0.28412	1707.55538	1707.55538	1707.55538	0	0	1	-360	360;
	3208	2254	0.000969	0.006021	0.02439	569.472221	569.472221	569.472221	0	0	1	-360	360;
	3208	3796	0.00099	0.005901	0.02496	392.38364	392.38364	392.38364	0	0	1	-360	360;
	1057	7795	0.00088	0.006599	0.0224	301.459494	301.459494	301.459494	0	0	1	-360	360;
	1057	4838	0.00086	0.006729	0.02201	497.991593	497.991593	497.991593	0	0	1	-360	360;
	7881	2979	9e-05	0.00077	0.02844	1730.48898	1730.48898	1730.48898	0	0	1	-360	360;
	7881	3848	9e-05	0.00078	0.02787	1594.3288	1594.3288	1594.3288	0	0	1	-360	360;
	4477	2979	8e-05	0.00082	0.02558	1504.1273	1504.1273	1504.1273	0	0	1	-360	360;
	4477	6912	8e-05	0.00076	0.02572	1261.72907	1261.72907	1261.72907	0	0	1	-360	360;
	6130	5017	9e-05	0.00074	0.02909	1625.281	1625.281	1625.281	0	0	1	-360	360;
	6130	3848	9e-05	0.00073	0.02972	1271.96969	1271.96969	1271.96969	0	0	1	-360	360;
	3722	2254	0.00062	0.003781	0.00889	259.689045	259.689045	259.689045	0	0	1	-360	360;
	3722	709	0.000599	0.0035	0.00876	564.635157	564.635157	564.635157	0	0	1	-360	360;
	6233	7795	0.00064	0.00274	0.00858	380.223535	380.223535	380.223535	0	0	1	-360	360;
	6233	1226	0.000651	0.00268	0.0087	613.291582	613.291582	613.291582	0	0	1	-360	360;
	6180	1246	0.00055	0.00685	0.14191	1806.74827	1806.74827	1806.74827	0	0	1	-360	360;
	6180	337	0.00061	0.00639	0.15911	1857.40636	1857.40636	1857.40636	0	0	1	-360	360;
	5463	5017	0.00042	0.00523	0.10137	1525.08636	1525.08636	1525.08636	0	0	1	-360	360;
	5463	7554	0.00045	0.00605	0.10902	1499.50688	1499.50688	1499.50688	0	0	1	-360	360;
	5167	2979	0.0003	0.00228	0.06581	1478.00482	1478.00482	1478.00482	0	0	1	-360	360;
	5167	8079	0.00026	0.00264	0.05674	1225.58779	1225.58779	1225.58779	0	0	1	-360	360;
	3047	8376	0.00536	0.03568	0.04968	567.016727	567.016727	567.016727	0	0	1	-360	360;
	3047	5704	0.006091	0.03549	0.05313	585.517143	585.517143	585.517143	0	0	1	-360	360;
	4976	8376	0.006419	0.033271	0.05979	608.629421	608.629421	608.629421	0	0	1	-360	360;
	4976	1347	0.005789	0.035901	0.0553	621.061978	621.061978	621.061978	0	0	1	-360	360;
	6318	1672	0	0.000146	0	341.060619	341.060619	341.060619	0	0	1	-360	360;
	6318	7052	0	0.000146	0	340.650227	340.650227	340.650227	0	0	1	-360	360;
	7219	8376	0.008401	0.04011	0.05941	424.979901	424.979901	424.979901	0	0	1	-360	360;
	7219	2909	0.007831	0.041531	0.05618	436.344752	436.344752	436.344752	0	0	1	-360	360;
	6850	604	0.00575	0.046461	0.0553	576.885572	576.885572	576.885572	0	0	1	-360	360;
	6850	1109	0.006711	0.04426	0.0558	609.453635	609.453635	609.453635	0	0	1	-360	360;
	3289	7920	0.00036	0.00485	0.12447	2133.01565	2133.01565	2133.01565	0	0	1	-360	360;
	3289	5668	0.00042	0.00415	0.1458	2180.83435	2180.83435	2180.83435	0	0	1	-360	360;
	6376	1531	9e-05	0.00069	0.42124	1343.48547	1343.48547	1343.48547	0	0	1	-360	360;
	6376	6726	4e-05	0.00036	0.43761	1442.88898	1442.88898	1442.88898	0	0	1	-360	360;
	925	2266	0.00055	0.00569	0.13419	1343.84616	1343.84616	1343.84616	0	0	1	-360	360;
	925	7758	0.0007	0.00767	0.1666	1214.18023	1214.18023	1214.18023	0	0	1	-360	360;
	5516	3537	0.0013	0.01299	0.31812	1209.50715	1209.50715	1209.50715	0	0	1	-360	360;
	2664	7544	0.00092	0.01031	0.37753	1646.17621	1646.17621	1646.17621	0	0	1	-360	360;
	999	3537	0.00158	0.01496	0.49876	1458.81428	1458.81428	1458.81428	0	0	1	-360	360;
	999	1031	0.00123	0.01309	0.36164	1924.44653	1924.44653	1924.44653	0	0	1	-360	360;
	3688	6021	0.000599	0.00274	0.15573	544.88718	544.88718	544.88718	0	0	1	-360	360;
	3688	6634	0.0005	0.00405	0.195	351.46484	351.46484	351.46484	0	0	1	-360	360;
	2307	7101	0.003349	0.02445	0.039	571.116838	571.116838	571.116838	0	0	1	-360	360;
	2307	4505	0.003281	0.02493	0.03821	569.773251	569.773251	569.773251	0	0	1	-360	360;
	3626	3296	0.00087	0.01097	0.2087	1269.102	1269.102	1269.102	0	0	1	-360	360;
	3626	8149	0.00161	0.01464	0.36978	1353.0554	1353.0554	1353.0554	0	0	1	-360	360;
	6352	3062	0	0.00046	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3937	3351	0.00118	0.01223	0.24339	1017.70054	1017.70054	1017.70054	0	0	1	-360	360;
	3937	6086	0.00045	0.00364	0.12321	1273.75138	1273.75138	1273.75138	0	0	1	-360	360;
	2161	1783	0.00714	0.03393	0.05753	357.797911	357.797911	357.797911	0	0	1	-360	360;
	2161	6966	0.00655	0.036849	0.05109	341.786521	341.786521	341.786521	0	0	1	-360	360;
	4572	7216	0.00088	0.00722	0.2163	1820.74643	1820.74643	1820.74643	0	0	1	-360	360;
	4572	5848	0.00071	0.00835	0.18603	1933.17464	1933.17464	1933.17464	0	0	1	-360	360;
	5963	7216	0.00084	0.00705	0.28116	1846.57865	1846.57865	1846.57865	0	0	1	-360	360;
	5963	7613	0.00102	0.00799	0.25557	1642.66154	1642.66154	1642.66154	0	0	1	-360	360;
	5795	5080	0.003781	0.029969	0.04141	544.659692	544.659692	544.659692	0	0	1	-360	360;
	5795	7381	0.00386	0.02969	0.03574	542.174477	542.174477	542.174477	0	0	1	-360	360;
	6071	7637	0.00076	0.00463	0.00841	586.97695	586.97695	586.97695	0	0	1	-360	360;
	6071	2616	0.000711	0.004841	0.0079	586.847392	586.847392	586.847392	0	0	1	-360	360;
	4734	8334	0.00631	0.034469	0.05046	402.417098	402.417098	402.417098	0	0	1	-360	360;
	4734	6716	0.006169	0.03512	0.04938	424.509302	424.509302	424.509302	0	0	1	-360	360;
	2808	4693	0.002961	0.017651	0.03977	622.602187	622.602187	622.602187	0	0	1	-360	360;
	2808	2616	0.00299	0.02462	0.033	604.767467	604.767467	604.767467	0	0	1	-360	360;
	230	8334	0.002789	0.011659	0.04297	496.342022	496.342022	496.342022	0	0	1	-360	360;
	230	6094	0.00249	0.013211	0.0383	492.616866	492.616866	492.616866	0	0	1	-360	360;
	3808	8334	0.002669	0.016031	0.04088	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3808	4274	0.00261	0.014841	0.03996	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5363	8334	0.00074	0.004599	0.00929	386.547841	386.547841	386.547841	0	0	1	-360	360;
	5363	1801	0.000719	0.00468	0.00865	438.262964	438.262964	438.262964	0	0	1	-360	360;
	1838	4269	3.5e-05	0.000355	0.003435	1270.81524	1270.81524	1270.81524	0	0	1	-360	360;
	1838	4388	3.5e-05	0.000355	0.003435	1270.38545	1270.38545	1270.38545	0	0	1	-360	360;
	1305	224	0.00511	0.02949	0.03885	432.507948	432.507948	432.507948	0	0	1	-360	360;
	1305	5205	0.007719	0.044271	0.05873	412.076365	412.076365	412.076365	0	0	1	-360	360;
	2665	224	0.004281	0.022469	0.03267	384.422718	384.422718	384.422718	0	0	1	-360	360;
	2665	2822	0.005961	0.03382	0.04539	463.516368	463.516368	463.516368	0	0	1	-360	360;
	4452	5868	0.00073	0.01034	0.34561	1687.44471	1687.44471	1687.44471	0	0	1	-360	360;
	4452	950	0.00076	0.00889	0.43866	1867.31261	1867.31261	1867.31261	0	0	1	-360	360;
	2357	2369	0.005159	0.025789	0.03929	413.516738	413.516738	413.516738	0	0	1	-360	360;
	2357	8238	0.00549	0.027711	0.03804	363.919879	363.919879	363.919879	0	0	1	-360	360;
	1932	2369	0.00614	0.01764	0.04668	401.137528	401.137528	401.137528	0	0	1	-360	360;
	1932	2910	0.003229	0.01788	0.02558	354.248419	354.248419	354.248419	0	0	1	-360	360;
	7596	4361	0.00053	0.00587	0.17659	1807.65919	1807.65919	1807.65919	0	0	1	-360	360;
	7596	5050	0.00055	0.00666	0.18184	1743.41562	1743.41562	1743.41562	0	0	1	-360	360;
	5230	6032	0.00326	0.01388	0.05107	224.967281	224.967281	224.967281	0	0	1	-360	360;
	5230	3702	0.003081	0.01488	0.04795	236.602678	236.602678	236.602678	0	0	1	-360	360;
	3097	1566	0.00051	0.00614	0.17963	2097.54176	2097.54176	2097.54176	0	0	1	-360	360;
	3097	5668	0.00043	0.00406	0.1485	2188.61272	2188.61272	2188.61272	0	0	1	-360	360;
	7283	5379	0.00761	0.046031	0.05785	443.171284	443.171284	443.171284	0	0	1	-360	360;
	7283	8213	0.00351	0.01632	0.02771	429.333791	429.333791	429.333791	0	0	1	-360	360;
	4579	1468	0.00031	0.00495	0.12257	1769.36829	1769.36829	1769.36829	0	0	1	-360	360;
	4579	8286	0.00043	0.00485	0.1425	1958.39871	1958.39871	1958.39871	0	0	1	-360	360;
	7939	131	0.00426	0.027099	0.04049	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7939	5339	0.00437	0.024969	0.04258	599.835522	599.835522	599.835522	0	0	1	-360	360;
	5514	7165	0.005	0.03932	0.05477	432.956826	432.956826	432.956826	0	0	1	-360	360;
	5514	6776	0.006039	0.036789	0.02852	463.000425	463.000425	463.000425	0	0	1	-360	360;
	7494	8277	0.0009	0.00893	0.21405	1228.69637	1228.69637	1228.69637	0	0	1	-360	360;
	7494	6631	0.00286	0.02156	0.27942	1072.60849	1072.60849	1072.60849	0	0	1	-360	360;
	615	131	0.00418	0.022729	0.03947	647.787002	647.787002	647.787002	0	0	1	-360	360;
	615	5153	0.00405	0.02231	0.03933	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6609	4394	0.00013	0.00675	0.15658	1666.22106	1666.22106	1666.22106	0	0	1	-360	360;
	6609	1134	0.00013	0.00688	0.15324	1709.39302	1709.39302	1709.39302	0	0	1	-360	360;
	8280	3714	0.00078	0.01201	0.33298	2019.31517	2019.31517	2019.31517	0	0	1	-360	360;
	8280	5922	0.00075	0.01254	0.31987	1789.71459	1789.71459	1789.71459	0	0	1	-360	360;
	2116	4701	0.00071	0.00722	0.28953	2172.33528	2172.33528	2172.33528	0	0	1	-360	360;
	2116	3949	0.00084	0.0074	0.28098	1906.49403	1906.49403	1906.49403	0	0	1	-360	360;
	3771	7373	0.00105	0.0081	0.20937	1077.70806	1077.70806	1077.70806	0	0	1	-360	360;
	3771	7141	0.00203	0.01565	0.40346	1151.52287	1151.52287	1151.52287	0	0	1	-360	360;
	7705	1446	0.00035	0.00581	0.14223	1901.67286	1901.67286	1901.67286	0	0	1	-360	360;
	7705	2387	0.00045	0.00534	0.176	2061.70719	2061.70719	2061.70719	0	0	1	-360	360;
	6670	6829	8.1e-05	0.001331	0.03287	280.915882	280.915882	280.915882	0	0	1	-360	360;
	6670	5712	0.00025	0.001591	0.0426	347.458087	347.458087	347.458087	0	0	1	-360	360;
	6310	2081	0.00401	0.026219	0.04089	420.669069	420.669069	420.669069	0	0	1	-360	360;
	6310	6776	0.004669	0.027521	0.04056	423.104367	423.104367	423.104367	0	0	1	-360	360;
	279	7780	0.00115	0.01265	0.25963	1093.69063	1093.69063	1093.69063	0	0	1	-360	360;
	279	1237	0.00112	0.0129	0.25734	1094.59431	1094.59431	1094.59431	0	0	1	-360	360;
	7043	5306	0.00123	0.01589	0.35725	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7043	7141	0.00054	0.00484	0.15614	1196.36331	1196.36331	1196.36331	0	0	1	-360	360;
	3440	7407	0.00401	0.03961	0.10773	316.917597	316.917597	316.917597	0	0	1	-360	360;
	3440	3540	0.002909	0.01981	0.03145	519.73704	519.73704	519.73704	0	0	1	-360	360;
	909	5574	0.002211	0.014419	0.01811	439.349722	439.349722	439.349722	0	0	1	-360	360;
	909	2899	0.002169	0.013349	0.01773	473.003018	473.003018	473.003018	0	0	1	-360	360;
	3321	7780	0.00055	0.00508	0.12874	1224.32145	1224.32145	1224.32145	0	0	1	-360	360;
	3321	810	0.00046	0.00595	0.10852	1456.20658	1456.20658	1456.20658	0	0	1	-360	360;
	6015	6770	0.00201	0.00988	0.01587	568.869398	568.869398	568.869398	0	0	1	-360	360;
	6015	2003	0.003401	0.019099	0.03377	573.588717	573.588717	573.588717	0	0	1	-360	360;
	1805	616	0.00028	0.00308	0.11808	1692.51269	1692.51269	1692.51269	0	0	1	-360	360;
	1805	5150	0.00038	0.00307	0.13025	1860.94999	1860.94999	1860.94999	0	0	1	-360	360;
	2211	3833	0.00044	0.00577	0.09831	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2211	2178	0.00102	0.01295	0.24007	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5086	1507	0.00085	0.01005	0.21741	1713.66724	1713.66724	1713.66724	0	0	1	-360	360;
	5086	4989	0.00073	0.0097	0.20189	1850.71399	1850.71399	1850.71399	0	0	1	-360	360;
	6937	1724	0.00139	0.0126	0.35664	1350.3799	1350.3799	1350.3799	0	0	1	-360	360;
	6937	4994	0.00117	0.01389	0.31297	1225.1422	1225.1422	1225.1422	0	0	1	-360	360;
	5962	1372	0.00076	0.00908	0.18213	1391.0725	1391.0725	1391.0725	0	0	1	-360	360;
	5962	7764	0.00137	0.01338	0.30919	1315.00737	1315.00737	1315.00737	0	0	1	-360	360;
	2987	2345	0.006	0.0402	0.05688	448.698812	448.698812	448.698812	0	0	1	-360	360;
	2987	625	0.005719	0.03906	0.05652	450.172337	450.172337	450.172337	0	0	1	-360	360;
	2067	1724	0.0007	0.01065	0.26892	887.038783	887.038783	887.038783	0	0	1	-360	360;
	2067	2085	0.00066	0.01062	0.25535	869.052059	869.052059	869.052059	0	0	1	-360	360;
	3724	2862	0.00153	0.01695	0.36722	1153.59153	1153.59153	1153.59153	0	0	1	-360	360;
	3724	5534	0.00083	0.00925	0.19859	1320.38404	1320.38404	1320.38404	0	0	1	-360	360;
	5500	4806	0.00033	0.00101	1.61324	855.017043	855.017043	855.017043	0	0	1	-360	360;
	7226	4806	0.00028	0.00104	1.58996	767.876077	767.876077	767.876077	0	0	1	-360	360;
	956	547	0.00139	0.0125	0.30823	1314.89548	1314.89548	1314.89548	0	0	1	-360	360;
	956	7563	0.00119	0.01236	0.26778	1147.8996	1147.8996	1147.8996	0	0	1	-360	360;
	2147	7915	0.010099	0.05513	0.07809	392.132146	392.132146	392.132146	0	0	1	-360	360;
	2147	543	0.00464	0.027021	0.03615	378.772873	378.772873	378.772873	0	0	1	-360	360;
	805	3513	0.002211	0.012031	0.03387	605.873659	605.873659	605.873659	0	0	1	-360	360;
	805	3702	0.002479	0.0115	0.03858	587.716189	587.716189	587.716189	0	0	1	-360	360;
	5024	7777	0.030281	0.058314	0.00629	109.200884	109.200884	109.200884	0	0	1	-360	360;
	5024	7005	0.028174	0.054256	0.00584	121.431865	121.431865	121.431865	0	0	1	-360	360;
	2337	7777	0.01114	0.026231	0.00232	89.412897	89.412897	89.412897	0	0	1	-360	360;
	2337	1984	0.018157	0.045248	0.00384	87.812101	87.812101	87.812101	0	0	1	-360	360;
	5776	318	0.00369	0.017159	0.02813	342.439643	342.439643	342.439643	0	0	1	-360	360;
	5776	4376	0.004909	0.03018	0.03738	314.563082	314.563082	314.563082	0	0	1	-360	360;
	7838	6604	0.00408	0.0056	0.09831	1264.8212	1264.8212	1264.8212	0	0	1	-360	360;
	7838	6962	0.00301	0.00925	0.21504	1303.95259	1303.95259	1303.95259	0	0	1	-360	360;
	4250	318	0.001599	0.00489	0.00882	242.108487	242.108487	242.108487	0	0	1	-360	360;
	4250	5608	0.00031	0.00175	0.00444	405.194199	405.194199	405.194199	0	0	1	-360	360;
	5583	318	0.00163	0.004789	0.00899	504.683613	504.683613	504.683613	0	0	1	-360	360;
	5583	5608	0.00031	0.001789	0.00435	401.800176	401.800176	401.800176	0	0	1	-360	360;
	2770	2942	0.00988	0.063229	0.07517	401.905728	401.905728	401.905728	0	0	1	-360	360;
	2770	1435	0.00114	0.00638	0.00903	418.532896	418.532896	418.532896	0	0	1	-360	360;
	617	6424	0.00065	0.00755	0.15871	1102.25683	1102.25683	1102.25683	0	0	1	-360	360;
	617	6266	0.00052	0.00503	0.13206	998.491576	998.491576	998.491576	0	0	1	-360	360;
	512	6424	0.00063	0.0077	0.15547	1108.21862	1108.21862	1108.21862	0	0	1	-360	360;
	512	6266	0.00053	0.00493	0.1346	938.817091	938.817091	938.817091	0	0	1	-360	360;
	2154	3036	0.00055	0.00683	0.26639	1647.37542	1647.37542	1647.37542	0	0	1	-360	360;
	2154	7489	0.00118	0.01134	0.3066	1386.45274	1386.45274	1386.45274	0	0	1	-360	360;
	5880	3036	0.00056	0.00668	0.27133	1961.10053	1961.10053	1961.10053	0	0	1	-360	360;
	5880	7489	0.0012	0.0111	0.3125	1355.47158	1355.47158	1355.47158	0	0	1	-360	360;
	6019	8075	0.0098	0.01047	0.23401	1029.97361	1029.97361	1029.97361	0	0	1	-360	360;
	6019	2919	0.00134	0.01215	0.30439	1091.39753	1091.39753	1091.39753	0	0	1	-360	360;
	5888	8075	0.00092	0.01198	0.21894	1297.59063	1297.59063	1297.59063	0	0	1	-360	360;
	5888	2342	0.00057	0.00616	0.13745	1285.87701	1285.87701	1285.87701	0	0	1	-360	360;
	7508	6758	0.00108	0.01773	0.44114	1578.03754	1578.03754	1578.03754	0	0	1	-360	360;
	7508	3393	0.00064	0.01047	0.28743	767.876077	767.876077	767.876077	0	0	1	-360	360;
	323	8217	0.002211	0.013341	0.02565	571.922761	571.922761	571.922761	0	0	1	-360	360;
	323	85	0.001901	0.014229	0.02177	601.141765	601.141765	601.141765	0	0	1	-360	360;
	6185	673	0.00193	0.011771	0.01584	413.796049	413.796049	413.796049	0	0	1	-360	360;
	6185	7701	0.00251	0.014909	0.01965	455.503243	455.503243	455.503243	0	0	1	-360	360;
	4582	554	6e-05	0.00084	0.02052	1452.69058	1452.69058	1452.69058	0	0	1	-360	360;
	4582	1134	6e-05	0.00092	0.01978	1452.87619	1452.87619	1452.87619	0	0	1	-360	360;
	7639	5356	0.00138	0.010349	0.01637	537.048957	537.048957	537.048957	0	0	1	-360	360;
	7639	6767	0.00156	0.009151	0.01696	544.635686	544.635686	544.635686	0	0	1	-360	360;
	5726	5356	0.001349	0.010539	0.01603	561.397746	561.397746	561.397746	0	0	1	-360	360;
	5726	6767	0.001591	0.00895	0.01728	524.851509	524.851509	524.851509	0	0	1	-360	360;
	5307	3759	0	0.000271	0	146.658804	146.658804	146.658804	0	0	1	-360	360;
	3973	4103	0.00057	0.0071	0.1846	1638.80329	1638.80329	1638.80329	0	0	1	-360	360;
	3973	5668	0.00057	0.00697	0.1885	1619.64435	1619.64435	1619.64435	0	0	1	-360	360;
	7642	3752	0.000969	0.00743	0.02614	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7642	4407	0.00099	0.007271	0.02662	647.787002	647.787002	647.787002	0	0	1	-360	360;
	150	3752	0.00088	0.00824	0.02372	647.787002	647.787002	647.787002	0	0	1	-360	360;
	150	4407	0.00086	0.008401	0.02323	647.787002	647.787002	647.787002	0	0	1	-360	360;
	566	3450	0.00293	0.01664	0.02232	409.423105	409.423105	409.423105	0	0	1	-360	360;
	566	7238	0.008719	0.05249	0.06623	386.602331	386.602331	386.602331	0	0	1	-360	360;
	3004	1573	0.00057	0.00627	0.05878	874.696604	874.696604	874.696604	0	0	1	-360	360;
	3004	7141	0.00088	0.00878	0.09612	800.412513	800.412513	800.412513	0	0	1	-360	360;
	8211	200	0.00033	0.00421	0.05828	1914.5429	1914.5429	1914.5429	0	0	1	-360	360;
	8211	7820	0.00084	0.00353	0.06713	2086.5291	2086.5291	2086.5291	0	0	1	-360	360;
	382	6827	0.0001	0.00119	0.0375	1810.69142	1810.69142	1810.69142	0	0	1	-360	360;
	382	5150	0.0001	0.0013	0.01685	1632.07735	1632.07735	1632.07735	0	0	1	-360	360;
	4365	7781	9e-05	0.00121	0.03588	1523.92599	1523.92599	1523.92599	0	0	1	-360	360;
	4365	3949	0.0001	0.00134	0.0165	1271.10747	1271.10747	1271.10747	0	0	1	-360	360;
	5975	2751	0.00063	0.01067	0.22597	1543.18432	1543.18432	1543.18432	0	0	1	-360	360;
	5975	2196	0.00073	0.00959	0.25241	1532.53761	1532.53761	1532.53761	0	0	1	-360	360;
	6597	3539	0.002021	0.030271	0.08068	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6597	4629	0.00218	0.02786	0.08704	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4478	2751	0.00086	0.00781	0.24161	1580.39382	1580.39382	1580.39382	0	0	1	-360	360;
	4478	3894	0.00092	0.00666	0.27844	1519.16933	1519.16933	1519.16933	0	0	1	-360	360;
	5085	3539	0.003969	0.0272	0.13136	221.966503	221.966503	221.966503	0	0	1	-360	360;
	5085	5405	0.00311	0.029729	0.12083	227.430776	227.430776	227.430776	0	0	1	-360	360;
	4739	2751	0.00055	0.00837	0.18311	1313.25398	1313.25398	1313.25398	0	0	1	-360	360;
	4739	8381	0.00053	0.0084	0.18127	1419.70527	1419.70527	1419.70527	0	0	1	-360	360;
	4890	2751	0.00053	0.00853	0.17922	1320.16223	1320.16223	1320.16223	0	0	1	-360	360;
	4890	8381	0.00052	0.00778	0.17733	1437.72425	1437.72425	1437.72425	0	0	1	-360	360;
	7003	3539	0.0017	0.008159	0.02643	567.770065	567.770065	567.770065	0	0	1	-360	360;
	7003	7115	0.001729	0.00799	0.0269	267.699884	267.699884	267.699884	0	0	1	-360	360;
	2064	3539	0.00176	0.00781	0.02742	595.212228	595.212228	595.212228	0	0	1	-360	360;
	2064	7115	0.00157	0.008849	0.02441	263.180617	263.180617	263.180617	0	0	1	-360	360;
	713	3539	0.000219	0.002229	0.0078	258.815295	258.815295	258.815295	0	0	1	-360	360;
	713	4677	0.000229	0.00218	0.00795	221.922682	221.922682	221.922682	0	0	1	-360	360;
	1782	3539	0.000229	0.00213	0.0081	219.596364	219.596364	219.596364	0	0	1	-360	360;
	1782	4677	0.000211	0.002419	0.00721	219.560164	219.560164	219.560164	0	0	1	-360	360;
	4034	3539	0.00019	0.00255	0.00636	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4034	4677	0.000211	0.0025	0.0072	252.769156	252.769156	252.769156	0	0	1	-360	360;
	7451	3539	0	0.000103	0	496.099673	496.099673	496.099673	0	0	1	-360	360;
	7451	1981	0	0.000103	0	496.051661	496.051661	496.051661	0	0	1	-360	360;
	1768	7249	0.00067	0.01058	0.45838	1340.17681	1340.17681	1340.17681	0	0	1	-360	360;
	1768	6761	0.00065	0.01078	0.44866	1185.13148	1185.13148	1185.13148	0	0	1	-360	360;
	3374	240	0.009281	0.05299	0.07436	343.244423	343.244423	343.244423	0	0	1	-360	360;
	3374	6966	0.009419	0.057219	0.07358	338.387164	338.387164	338.387164	0	0	1	-360	360;
	4714	2173	0.00055	0.00505	0.03957	285.353604	285.353604	285.353604	0	0	1	-360	360;
	4714	5712	9.9e-05	0.00164	0.00403	283.417483	283.417483	283.417483	0	0	1	-360	360;
	6972	7249	0.00064	0.00998	0.43888	1170.50213	1170.50213	1170.50213	0	0	1	-360	360;
	6972	2152	0.00216	0.02943	0.53664	1025.63226	1025.63226	1025.63226	0	0	1	-360	360;
	4520	1962	0.00039	0.00474	0.16913	1856.2552	1856.2552	1856.2552	0	0	1	-360	360;
	4520	1596	0.00045	0.00467	0.16216	1461.53322	1461.53322	1461.53322	0	0	1	-360	360;
	6984	1962	0.00052	0.00567	0.21225	1672.48298	1672.48298	1672.48298	0	0	1	-360	360;
	6984	1541	0.00054	0.00667	0.18021	1344.16735	1344.16735	1344.16735	0	0	1	-360	360;
	4066	4023	0.008909	0.042349	0.07091	365.827421	365.827421	365.827421	0	0	1	-360	360;
	4066	7546	0.00713	0.03293	0.05806	322.540388	322.540388	322.540388	0	0	1	-360	360;
	8064	1301	0.0021	0.02015	0.52812	1448.82641	1448.82641	1448.82641	0	0	1	-360	360;
	8064	8333	0.00177	0.02084	0.45884	1458.34171	1458.34171	1458.34171	0	0	1	-360	360;
	2260	6700	0.001461	0.00781	0.01124	267.602716	267.602716	267.602716	0	0	1	-360	360;
	2260	5362	0.001651	0.007651	0.01275	313.916057	313.916057	313.916057	0	0	1	-360	360;
	7868	6700	0.00168	0.0075	0.01299	281.573576	281.573576	281.573576	0	0	1	-360	360;
	7868	5362	0.001719	0.007341	0.01325	369.814741	369.814741	369.814741	0	0	1	-360	360;
	7196	6317	0.00626	0.020771	0.03716	396.862896	396.862896	396.862896	0	0	1	-360	360;
	7196	6317	0.00638	0.02032	0.03785	408.637378	408.637378	408.637378	0	0	1	-360	360;
	4211	5728	0.047983	0.076364	0.00627	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4211	1428	0.010868	0.039083	0.00351	99.38272	99.38272	99.38272	0	0	1	-360	360;
	4968	3154	0.021851	0.032364	0.00251	61.937011	61.937011	61.937011	0	0	1	-360	360;
	3934	3154	0.075008	0.121636	0.01102	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3154	4704	0.049273	0.084942	0.00708	69.341216	69.341216	69.341216	0	0	1	-360	360;
	6485	6469	0.058521	0.110017	0.00836	58.277205	58.277205	58.277205	0	0	1	-360	360;
	6818	6009	0.012992	0.052719	0.00484	69.288441	69.288441	69.288441	0	0	1	-360	360;
	6009	6818	0.012727	0.053752	0.00474	80.213369	80.213369	80.213369	0	0	1	-360	360;
	4839	2244	0.013341	0.054169	0.07854	338.181396	338.181396	338.181396	0	0	1	-360	360;
	4411	4839	0.014961	0.083521	0.12171	419.841425	419.841425	419.841425	0	0	1	-360	360;
	6477	1503	0.023653	0.033603	0.00338	76.065246	76.065246	76.065246	0	0	1	-360	360;
	6477	1503	0.024099	0.032876	0.00345	63.301174	63.301174	63.301174	0	0	1	-360	360;
	1503	6818	0.024934	0.109579	0.00929	71.459289	71.459289	71.459289	0	0	1	-360	360;
	1503	6818	0.024397	0.111653	0.00909	90.671699	90.671699	90.671699	0	0	1	-360	360;
	1503	5466	0.03414	0.051058	0.00503	71.674774	71.674774	71.674774	0	0	1	-360	360;
	5728	4968	0.018893	0.029083	0.0027	61.170146	61.170146	61.170146	0	0	1	-360	360;
	2244	5134	0.000581	0.003081	0.00591	489.163399	489.163399	489.163399	0	0	1	-360	360;
	2244	5134	0.0005	0.002789	0.00559	496.193031	496.193031	496.193031	0	0	1	-360	360;
	2244	5901	0.009281	0.03489	0.05529	383.028833	383.028833	383.028833	0	0	1	-360	360;
	5436	2244	0.006081	0.034	0.04918	451.724359	451.724359	451.724359	0	0	1	-360	360;
	5436	2244	0.00595	0.034651	0.04816	419.472568	419.472568	419.472568	0	0	1	-360	360;
	1550	6477	0.05019	0.076339	0.00593	74.536087	74.536087	74.536087	0	0	1	-360	360;
	6477	3463	0.05724	0.101868	0.00818	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7486	6630	0.0013	0.01496	0.32143	934.259857	934.259857	934.259857	0	0	1	-360	360;
	1408	7486	0.00108	0.01115	0.23482	1419.03853	1419.03853	1419.03853	0	0	1	-360	360;
	7486	95	0.00223	0.02267	0.51652	1352.09183	1352.09183	1352.09183	0	0	1	-360	360;
	7196	5901	0.008229	0.03325	0.04859	400.38038	400.38038	400.38038	0	0	1	-360	360;
	7196	5423	0.018159	0.06526	0.1092	427.747475	427.747475	427.747475	0	0	1	-360	360;
	3934	8322	0.02257	0.034777	0.00266	60.636483	60.636483	60.636483	0	0	1	-360	360;
	3934	2392	0.032339	0.103017	0.0099	94.752948	94.752948	94.752948	0	0	1	-360	360;
	3934	2392	0.031678	0.105041	0.0097	95.596595	95.596595	95.596595	0	0	1	-360	360;
	3934	754	0.028331	0.050182	0.0043	72.070114	72.070114	72.070114	0	0	1	-360	360;
	3934	5728	0.047314	0.063678	0.00716	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4411	6533	0.0045	0.02618	0.03665	425.528614	425.528614	425.528614	0	0	1	-360	360;
	4411	5436	0.01913	0.10681	0.15565	390.058846	390.058846	390.058846	0	0	1	-360	360;
	7521	6485	0.059273	0.08086	0.00848	82.159016	82.159016	82.159016	0	0	1	-360	360;
	7241	3784	0.008041	0.014298	0.00115	59.253839	59.253839	59.253839	0	0	1	-360	360;
	4224	7241	0.047802	0.073562	0.00683	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3386	1550	0.049636	0.058818	0.00593	69.60357	69.60357	69.60357	0	0	1	-360	360;
	8322	1428	0.04738	0.075992	0.00599	58.910893	58.910893	58.910893	0	0	1	-360	360;
	5466	8007	0.035603	0.045331	0.00524	70.79264	70.79264	70.79264	0	0	1	-360	360;
	2626	4704	0.024529	0.031678	0.00356	74.794821	74.794821	74.794821	0	0	1	-360	360;
	8072	2626	0.036851	0.047521	0.00535	71.423661	71.423661	71.423661	0	0	1	-360	360;
	5759	7559	0.004058	0.011041	0.00123	79.96721	79.96721	79.96721	0	0	1	-360	360;
	4207	5759	0.012818	0.022554	0.00188	62.586132	62.586132	62.586132	0	0	1	-360	360;
	4138	1506	0.079256	0.118479	0.01163	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4207	4138	0.012554	0.022967	0.00184	72.881563	72.881563	72.881563	0	0	1	-360	360;
	7559	7347	0.007917	0.029628	0.0024	72.776202	72.776202	72.776202	0	0	1	-360	360;
	7559	7347	0.007744	0.027438	0.00235	67.41767	67.41767	67.41767	0	0	1	-360	360;
	1408	2264	0.00039	0.00461	0.09592	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6533	248	0.014039	0.071669	0.1423	449.407567	449.407567	449.407567	0	0	1	-360	360;
	6533	248	0.01251	0.081219	0.12677	488.900474	488.900474	488.900474	0	0	1	-360	360;
	248	5423	0.0045	0.016831	0.01788	344.542664	344.542664	344.542664	0	0	1	-360	360;
	5436	248	0.00261	0.009401	0.01039	392.462517	392.462517	392.462517	0	0	1	-360	360;
	6534	2392	0.08024	0.135628	0.01173	72.226536	72.226536	72.226536	0	0	1	-360	360;
	3941	6534	0.043587	0.057314	0.00386	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6533	70	0.00326	0.01895	0.02654	493.152243	493.152243	493.152243	0	0	1	-360	360;
	673	70	0.005409	0.02675	0.04408	400.963388	400.963388	400.963388	0	0	1	-360	360;
	3784	3071	0.083537	0.116554	0.01217	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2305	3071	0.012463	0.054793	0.00464	137.445922	137.445922	137.445922	0	0	1	-360	360;
	754	8007	0.044992	0.07586	0.00662	66.346726	66.346726	66.346726	0	0	1	-360	360;
	6490	1506	0.032777	0.050628	0.00421	60.572467	60.572467	60.572467	0	0	1	-360	360;
	3463	4224	0.060314	0.092818	0.00863	57.988178	57.988178	57.988178	0	0	1	-360	360;
	8072	7521	0.02462	0.034967	0.00352	65.810968	65.810968	65.810968	0	0	1	-360	360;
	7347	3386	0.026603	0.03038	0.00317	57.157677	57.157677	57.157677	0	0	1	-360	360;
	35	4658	0.01595	0.022281	0.00194	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6490	4658	0.018083	0.021843	0.0022	57.157677	57.157677	57.157677	0	0	1	-360	360;
	70	5436	0.0132	0.06293	0.10683	401.089897	401.089897	401.089897	0	0	1	-360	360;
	70	5436	0.01345	0.06162	0.10888	417.45414	417.45414	417.45414	0	0	1	-360	360;
	4207	3386	0.010612	0.023017	0.00241	68.076508	68.076508	68.076508	0	0	1	-360	360;
	904	6848	0.01601	0.06706	0.12503	431.471869	431.471869	431.471869	0	0	1	-360	360;
	904	5080	0.00362	0.01457	0.02827	471.790895	471.790895	471.790895	0	0	1	-360	360;
	7346	4342	0.010281	0.04007	0.08012	436.585195	436.585195	436.585195	0	0	1	-360	360;
	7346	8334	0.01549	0.06756	0.12161	391.990014	391.990014	391.990014	0	0	1	-360	360;
	1660	4709	0.00073	0.01141	0.26279	1904.73669	1904.73669	1904.73669	0	0	1	-360	360;
	1660	7249	0.00108	0.01698	0.39106	1492.38275	1492.38275	1492.38275	0	0	1	-360	360;
	2598	466	0.0022	0.027409	0.07783	387.495515	387.495515	387.495515	0	0	1	-360	360;
	2598	466	0.002151	0.02794	0.07621	397.599468	397.599468	397.599468	0	0	1	-360	360;
	2598	6893	0.00445	0.02582	0.06201	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2598	240	0.005099	0.02426	0.07118	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2598	1576	0.019539	0.100021	0.15273	307.813141	307.813141	307.813141	0	0	1	-360	360;
	2598	1576	0.01914	0.101979	0.14962	288.610448	288.610448	288.610448	0	0	1	-360	360;
	4709	2006	0.00158	0.02551	0.55935	1535.2368	1535.2368	1535.2368	0	0	1	-360	360;
	4709	4254	0.00021	0.00296	0.07527	1393.59925	1393.59925	1393.59925	0	0	1	-360	360;
	4709	5129	0.00157	0.02091	0.55027	1606.18392	1606.18392	1606.18392	0	0	1	-360	360;
	1240	4254	0.00162	0.02163	0.57081	1504.22273	1504.22273	1504.22273	0	0	1	-360	360;
	6931	4342	0.008521	0.03463	0.06921	360.914528	360.914528	360.914528	0	0	1	-360	360;
	6931	8334	0.00463	0.078531	0.18354	247.567045	247.567045	247.567045	0	0	1	-360	360;
	6931	8334	0.004531	0.072711	0.17955	224.315683	224.315683	224.315683	0	0	1	-360	360;
	6931	99	0.00575	0.02876	0.047	390.045129	390.045129	390.045129	0	0	1	-360	360;
	6931	2868	0.00411	0.01826	0.03166	446.15339	446.15339	446.15339	0	0	1	-360	360;
	6931	1576	0.008091	0.03594	0.06232	397.810952	397.810952	397.810952	0	0	1	-360	360;
	3482	1850	0.011031	0.060581	0.08626	434.235634	434.235634	434.235634	0	0	1	-360	360;
	3482	5445	0.01326	0.07163	0.10328	428.06451	428.06451	428.06451	0	0	1	-360	360;
	4612	1672	0.007031	0.069341	0.24929	229.457969	229.457969	229.457969	0	0	1	-360	360;
	4612	5080	0.00238	0.031081	0.08439	237.061464	237.061464	237.061464	0	0	1	-360	360;
	4342	8334	0.010831	0.057479	0.08699	461.511277	461.511277	461.511277	0	0	1	-360	360;
	4342	99	0.001849	0.01024	0.01488	397.874206	397.874206	397.874206	0	0	1	-360	360;
	6848	6133	0.017581	0.07664	0.13729	504.656559	504.656559	504.656559	0	0	1	-360	360;
	6848	6133	0.017909	0.07501	0.13988	500.978653	500.978653	500.978653	0	0	1	-360	360;
	6848	5445	0.01312	0.05724	0.1025	399.633519	399.633519	399.633519	0	0	1	-360	360;
	6848	5445	0.01337	0.056021	0.10444	403.813651	403.813651	403.813651	0	0	1	-360	360;
	6848	500	0.007961	0.0347	0.06213	421.980266	421.980266	421.980266	0	0	1	-360	360;
	4619	8334	0.01181	0.044729	0.09514	392.866813	392.866813	392.866813	0	0	1	-360	360;
	4619	109	0.00457	0.018729	0.03678	314.688067	314.688067	314.688067	0	0	1	-360	360;
	7644	7457	0.0012	0.0141	0.42013	1310.53438	1310.53438	1310.53438	0	0	1	-360	360;
	7644	7457	0.00107	0.01598	0.3743	1297.30432	1297.30432	1297.30432	0	0	1	-360	360;
	7644	2662	0.00046	0.00666	0.28083	900.703243	900.703243	900.703243	0	0	1	-360	360;
	7644	2511	0.001	0.0181	0.60041	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1043	3367	9.1e-05	0.00139	0.00362	217.283003	217.283003	217.283003	0	0	1	-360	360;
	1043	3367	0.00013	0.00089	0.00677	250.361293	250.361293	250.361293	0	0	1	-360	360;
	6893	240	0.000281	0.001289	0.00395	575.360986	575.360986	575.360986	0	0	1	-360	360;
	6313	3188	0.00119	0.01964	0.42295	1120.87739	1120.87739	1120.87739	0	0	1	-360	360;
	6313	3188	0.00117	0.02001	0.41395	1189.85853	1189.85853	1189.85853	0	0	1	-360	360;
	1672	3832	0.00238	0.032341	0.08398	275.255367	275.255367	275.255367	0	0	1	-360	360;
	1672	3832	0.002331	0.02994	0.08215	348.489211	348.489211	348.489211	0	0	1	-360	360;
	1672	4405	0.005479	0.07155	0.19422	292.212144	292.212144	292.212144	0	0	1	-360	360;
	1783	240	0.008091	0.036841	0.0648	391.645544	391.645544	391.645544	0	0	1	-360	360;
	5080	4405	0.003081	0.03707	0.109	247.542277	247.542277	247.542277	0	0	1	-360	360;
	5080	500	0.01086	0.05936	0.08485	442.525783	442.525783	442.525783	0	0	1	-360	360;
	2662	869	0.00063	0.00975	0.38002	848.036393	848.036393	848.036393	0	0	1	-360	360;
	8334	7806	0.00143	0.025771	0.09611	273.521965	273.521965	273.521965	0	0	1	-360	360;
	8334	7806	0.001401	0.026271	0.0941	269.916459	269.916459	269.916459	0	0	1	-360	360;
	8334	109	0.00631	0.03037	0.0508	408.417511	408.417511	408.417511	0	0	1	-360	360;
	1850	5445	0.00325	0.01388	0.02498	414.445741	414.445741	414.445741	0	0	1	-360	360;
	7806	3367	0.015461	0.0828	0.11986	423.340237	423.340237	423.340237	0	0	1	-360	360;
	7806	3367	0.017521	0.08114	0.13584	425.536616	425.536616	425.536616	0	0	1	-360	360;
	5129	7249	0.00036	0.00554	0.1258	1545.82493	1545.82493	1545.82493	0	0	1	-360	360;
	2511	7249	0.0018	0.02089	0.63352	1651.14481	1651.14481	1651.14481	0	0	1	-360	360;
	869	7249	0.00183	0.02045	0.6457	1639.23769	1639.23769	1639.23769	0	0	1	-360	360;
	3188	413	0.00052	0.01489	0.36998	934.470474	934.470474	934.470474	0	0	1	-360	360;
	3188	413	0.00051	0.01518	0.36228	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5303	5445	0.020669	0.110081	0.16146	363.832618	363.832618	363.832618	0	0	1	-360	360;
	5303	5445	0.02024	0.1122	0.1581	371.156803	371.156803	371.156803	0	0	1	-360	360;
	271	240	0.01089	0.05145	0.08506	595.18746	595.18746	595.18746	0	0	1	-360	360;
	271	240	0.01111	0.050401	0.08673	565.213974	565.213974	565.213974	0	0	1	-360	360;
	2868	1576	0.00493	0.026719	0.03795	249.659397	249.659397	249.659397	0	0	1	-360	360;
	3539	6700	0.005169	0.067461	0.18294	246.435704	246.435704	246.435704	0	0	1	-360	360;
	3539	6700	0.00506	0.068729	0.17905	245.365331	245.365331	245.365331	0	0	1	-360	360;
	1742	6238	0.011397	0.04438	0.00394	64.529493	64.529493	64.529493	0	0	1	-360	360;
	1762	7624	0.022504	0.082851	0.00776	74.8053	74.8053	74.8053	0	0	1	-360	360;
	1762	3587	0.015793	0.061587	0.00526	69.202133	69.202133	69.202133	0	0	1	-360	360;
	5190	2369	0.012659	0.05405	0.10094	391.549138	391.549138	391.549138	0	0	1	-360	360;
	661	3652	0.016149	0.057248	0.00549	65.114216	65.114216	65.114216	0	0	1	-360	360;
	136	3652	0.038612	0.122471	0.01301	76.924706	76.924706	76.924706	0	0	1	-360	360;
	3652	6517	0.044273	0.108959	0.00923	58.120212	58.120212	58.120212	0	0	1	-360	360;
	8031	5389	0.038876	0.139587	0.01341	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8031	5807	0.10805	0.191992	0.02242	75.058889	75.058889	75.058889	0	0	1	-360	360;
	2813	8031	0.035802	0.084231	0.00743	79.449742	79.449742	79.449742	0	0	1	-360	360;
	3074	4700	0.00614	0.02205	0.00212	137.18776	137.18776	137.18776	0	0	1	-360	360;
	3074	5560	0.003587	0.01395	0.00124	61.674086	61.674086	61.674086	0	0	1	-360	360;
	4700	7047	0.019554	0.05981	0.00675	96.392611	96.392611	96.392611	0	0	1	-360	360;
	5560	3012	0.038066	0.142281	0.01313	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2452	3938	0.007421	0.02181	0.00257	58.228812	58.228812	58.228812	0	0	1	-360	360;
	3938	7411	0.007562	0.021339	0.00261	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2452	6885	0.007388	0.026537	0.00255	85.408049	85.408049	85.408049	0	0	1	-360	360;
	6885	1526	0.00724	0.02705	0.0025	58.759806	58.759806	58.759806	0	0	1	-360	360;
	63	6427	0.004893	0.018273	0.00081	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6427	5039	0.004777	0.018612	0.00079	70.148283	70.148283	70.148283	0	0	1	-360	360;
	2452	5757	0.015289	0.041463	0.00528	69.577087	69.577087	69.577087	0	0	1	-360	360;
	2452	5757	0.01143	0.042091	0.00394	69.377797	69.377797	69.377797	0	0	1	-360	360;
	63	2452	0.006793	0.014182	0.00141	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5039	2452	0.012967	0.048471	0.00447	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5571	5757	0.009281	0.029554	0.0032	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5610	5757	0.041248	0.116479	0.01423	78.907316	78.907316	78.907316	0	0	1	-360	360;
	3459	5757	0.071479	0.137678	0.01483	78.617717	78.617717	78.617717	0	0	1	-360	360;
	2110	5757	0.020612	0.075893	0.00711	57.187399	57.187399	57.187399	0	0	1	-360	360;
	5757	465	0.05214	0.100421	0.01081	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4336	5757	0.047942	0.179157	0.01654	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6584	2275	0.0016	0.01402	0.37663	1148.98493	1148.98493	1148.98493	0	0	1	-360	360;
	5361	5571	0.009149	0.027992	0.13951	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7819	5361	0.031107	0.116281	0.01073	83.559379	83.559379	83.559379	0	0	1	-360	360;
	7819	5039	0.026463	0.091273	0.00855	64.756599	64.756599	64.756599	0	0	1	-360	360;
	367	7819	0.023562	0.075017	0.00813	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3680	7819	0.015702	0.050017	0.00542	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3993	3012	0.004802	0.017231	0.00166	78.514261	78.514261	78.514261	0	0	1	-360	360;
	3993	1108	0.065579	0.208826	0.02263	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3993	3171	0.049917	0.112843	0.01036	58.159841	58.159841	58.159841	0	0	1	-360	360;
	3993	120	0.047793	0.095603	0.00988	67.898747	67.898747	67.898747	0	0	1	-360	360;
	6697	3012	0.043421	0.168942	0.01498	83.343514	83.343514	83.343514	0	0	1	-360	360;
	7047	3012	0.029364	0.093017	0.01064	91.217555	91.217555	91.217555	0	0	1	-360	360;
	2640	3477	0.023521	0.086562	0.00811	57.157677	57.157677	57.157677	0	0	1	-360	360;
	884	3477	0.017909	0.058463	0.00617	59.859711	59.859711	59.859711	0	0	1	-360	360;
	3420	483	0.008198	0.030248	0.00279	64.080995	64.080995	64.080995	0	0	1	-360	360;
	483	2824	0.015579	0.053744	0.00538	57.157677	57.157677	57.157677	0	0	1	-360	360;
	483	5448	0.030471	0.09738	0.01017	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2545	483	0.017157	0.066	0.00573	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1158	4099	0.017628	0.065868	0.00608	72.883468	72.883468	72.883468	0	0	1	-360	360;
	1158	4705	0.022686	0.072273	0.00782	86.674663	86.674663	86.674663	0	0	1	-360	360;
	2021	3921	0.010322	0.03719	0.00344	61.28046	61.28046	61.28046	0	0	1	-360	360;
	3921	6831	0.010802	0.038281	0.00368	71.135586	71.135586	71.135586	0	0	1	-360	360;
	4532	6517	0.019298	0.073041	0.00656	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2920	4532	0.011678	0.035273	0.00398	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4607	4099	0.019719	0.057256	0.00672	57.157677	57.157677	57.157677	0	0	1	-360	360;
	155	661	0.026421	0.086471	0.009	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7411	1753	0.014281	0.045471	0.00493	63.000144	63.000144	63.000144	0	0	1	-360	360;
	884	6382	0.004529	0.013686	0.00154	114.101584	114.101584	114.101584	0	0	1	-360	360;
	6382	6831	0.014603	0.058463	0.00498	81.179333	81.179333	81.179333	0	0	1	-360	360;
	6098	5067	0.007702	0.020876	0.00265	98.486678	98.486678	98.486678	0	0	1	-360	360;
	6098	3149	0.010281	0.03957	0.00343	128.306409	128.306409	128.306409	0	0	1	-360	360;
	5067	4607	0.00686	0.023661	0.00236	72.529472	72.529472	72.529472	0	0	1	-360	360;
	3420	4613	0.02457	0.09557	0.00847	73.239561	73.239561	73.239561	0	0	1	-360	360;
	15	1108	0.03024	0.085372	0.01043	69.082864	69.082864	69.082864	0	0	1	-360	360;
	15	6697	0.04157	0.14924	0.01433	111.668	111.668	111.668	0	0	1	-360	360;
	5472	15	0.028132	0.097008	0.0097	61.036968	61.036968	61.036968	0	0	1	-360	360;
	15	7861	0.038661	0.123091	0.01334	74.617251	74.617251	74.617251	0	0	1	-360	360;
	1187	7907	0.019364	0.077413	0.00659	59.753778	59.753778	59.753778	0	0	1	-360	360;
	7907	3587	0.009562	0.036182	0.00325	80.66701	80.66701	80.66701	0	0	1	-360	360;
	6584	5103	0.00147	0.01643	0.34292	1442.72838	1442.72838	1442.72838	0	0	1	-360	360;
	636	5640	0.005719	0.016826	0.00197	74.640495	74.640495	74.640495	0	0	1	-360	360;
	5640	8328	0.007843	0.027041	0.0027	77.078461	77.078461	77.078461	0	0	1	-360	360;
	5640	1175	0.016826	0.045628	0.00581	59.369869	59.369869	59.369869	0	0	1	-360	360;
	5640	2008	0.02238	0.062777	0.00756	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1108	636	0.012959	0.041248	0.00447	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8328	3994	0.004702	0.017562	0.00162	70.942203	70.942203	70.942203	0	0	1	-360	360;
	6998	8328	0.023331	0.080446	0.00816	72.875847	72.875847	72.875847	0	0	1	-360	360;
	1362	884	0.00505	0.018653	0.00172	58.805723	58.805723	58.805723	0	0	1	-360	360;
	3967	1362	0.008562	0.025686	0.00265	58.152601	58.152601	58.152601	0	0	1	-360	360;
	3172	2687	0.026124	0.083149	0.00901	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2687	4336	0.02038	0.04605	0.00422	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7007	5987	0.000169	0.00106	0.00342	305.351551	305.351551	305.351551	0	0	1	-360	360;
	874	7007	0.00043	0.003901	0.00876	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7229	4212	0.005298	0.021	0.00183	76.280539	76.280539	76.280539	0	0	1	-360	360;
	4212	6610	0.001322	0.004289	0.00045	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5833	1015	0.005471	0.016554	0.00186	61.963685	61.963685	61.963685	0	0	1	-360	360;
	5833	4440	0.022198	0.064653	0.00753	74.807777	74.807777	74.807777	0	0	1	-360	360;
	4357	2503	0.027653	0.099256	0.00954	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2378	4357	0.044099	0.152107	0.01521	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2722	2881	0.017603	0.047711	0.0607	69.096582	69.096582	69.096582	0	0	1	-360	360;
	2722	3967	0.024198	0.06562	0.00835	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7586	3296	0.00165	0.01428	0.38483	1309.6623	1309.6623	1309.6623	0	0	1	-360	360;
	6604	7586	0.00131	0.01467	0.3032	1167.85954	1167.85954	1167.85954	0	0	1	-360	360;
	5389	8117	0.040802	0.129909	0.01407	78.285631	78.285631	78.285631	0	0	1	-360	360;
	856	2503	0.005058	0.019686	0.00175	82.329727	82.329727	82.329727	0	0	1	-360	360;
	2503	1262	0.075521	0.145463	0.01567	57.157677	57.157677	57.157677	0	0	1	-360	360;
	856	4726	0.024033	0.088488	0.00829	74.64964	74.64964	74.64964	0	0	1	-360	360;
	6866	4489	0.001599	0.007401	0.01272	439.166817	439.166817	439.166817	0	0	1	-360	360;
	1175	7624	0.025289	0.082934	0.00861	92.115121	92.115121	92.115121	0	0	1	-360	360;
	8018	6697	0.012421	0.035058	0.00429	100.712969	100.712969	100.712969	0	0	1	-360	360;
	2942	2483	0.00855	0.03774	0.06864	470.086834	470.086834	470.086834	0	0	1	-360	360;
	5610	4735	0.010686	0.03019	0.00369	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3801	5603	0.010331	0.03438	0.00345	80.441999	80.441999	80.441999	0	0	1	-360	360;
	5603	5937	0.01581	0.050339	0.00546	67.980673	67.980673	67.980673	0	0	1	-360	360;
	7229	4440	0.005661	0.022744	0.00192	61.049543	61.049543	61.049543	0	0	1	-360	360;
	5294	7610	0.000636	0.002512	0.00023	60.372796	60.372796	60.372796	0	0	1	-360	360;
	3587	5294	0.005281	0.017521	0.00175	77.983648	77.983648	77.983648	0	0	1	-360	360;
	5294	2556	0.002826	0.00781	0.04179	74.89237	74.89237	74.89237	0	0	1	-360	360;
	1015	4545	0.018066	0.056876	0.00615	65.016476	65.016476	65.016476	0	0	1	-360	360;
	4545	4705	0.038537	0.144033	0.01329	79.822791	79.822791	79.822791	0	0	1	-360	360;
	4545	1137	0.049669	0.141083	0.01734	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4545	6610	0.058628	0.17286	0.02053	73.927548	73.927548	73.927548	0	0	1	-360	360;
	4545	5494	0.00043	0.001273	0.00014	65.511843	65.511843	65.511843	0	0	1	-360	360;
	4545	6256	0.025917	0.093058	0.00894	78.440719	78.440719	78.440719	0	0	1	-360	360;
	1583	3801	0.02295	0.065	0.00766	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5422	4705	0.001504	0.00486	0.00051	71.038609	71.038609	71.038609	0	0	1	-360	360;
	5422	4705	0.002182	0.006678	0.00075	61.359147	61.359147	61.359147	0	0	1	-360	360;
	3093	874	0.002979	0.02218	0.06106	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3093	874	0.003021	0.022531	0.06192	215.929128	215.929128	215.929128	0	0	1	-360	360;
	367	3680	0.020843	0.058843	0.00719	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4613	367	0.009331	0.027413	0.00321	57.157677	57.157677	57.157677	0	0	1	-360	360;
	367	6272	0.009182	0.024769	0.00314	58.225763	58.225763	58.225763	0	0	1	-360	360;
	4803	318	0.00275	0.01488	0.02254	477.954397	477.954397	477.954397	0	0	1	-360	360;
	6065	2720	0.0001	0.00082	0.02422	1172.27461	1172.27461	1172.27461	0	0	1	-360	360;
	5379	874	0.01476	0.08713	0.1124	360.433642	360.433642	360.433642	0	0	1	-360	360;
	5379	4489	0.00212	0.011469	0.0168	409.642591	409.642591	409.642591	0	0	1	-360	360;
	6610	5980	0.013521	0.049959	0.00461	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5980	884	0.017793	0.07124	0.00607	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7809	2369	0.01481	0.086021	0.1188	427.409483	427.409483	427.409483	0	0	1	-360	360;
	4489	7809	0.01686	0.07207	0.13497	471.902924	471.902924	471.902924	0	0	1	-360	360;
	6831	5137	0.012331	0.034413	0.00419	75.05203	75.05203	75.05203	0	0	1	-360	360;
	5215	5137	0.000777	0.00276	0.00026	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2930	136	0.028421	0.08657	0.00957	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1753	4200	0.023521	0.081132	0.00811	60.678208	60.678208	60.678208	0	0	1	-360	360;
	3251	2375	0.000901	0.005	0.01242	446.593504	446.593504	446.593504	0	0	1	-360	360;
	2375	2817	0.00263	0.01551	0.02542	430.340147	430.340147	430.340147	0	0	1	-360	360;
	6508	2892	0.017463	0.071669	0.00662	76.347604	76.347604	76.347604	0	0	1	-360	360;
	6508	7988	0.016826	0.065512	0.00581	81.993068	81.993068	81.993068	0	0	1	-360	360;
	2881	3510	0.019587	0.076612	0.00654	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1256	1187	0.022537	0.071769	0.00778	66.284424	66.284424	66.284424	0	0	1	-360	360;
	2920	155	0.013256	0.04005	0.0045	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6577	2930	0.046397	0.13586	0.01564	67.644586	67.644586	67.644586	0	0	1	-360	360;
	6577	518	0.026471	0.105893	0.00901	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7759	6564	0.025504	0.08119	0.00879	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4488	7759	0.027603	0.10738	0.00952	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6517	3459	0.084843	0.156942	0.0176	60.516452	60.516452	60.516452	0	0	1	-360	360;
	4760	2824	0.030521	0.115653	0.01065	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1526	7286	0.004	0.010851	0.00138	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7286	4735	0.016017	0.043421	0.00552	61.223493	61.223493	61.223493	0	0	1	-360	360;
	4760	1742	0.027091	0.087488	0.00947	63.818261	63.818261	63.818261	0	0	1	-360	360;
	6727	6293	0.02243	0.047793	0.00532	61.138709	61.138709	61.138709	0	0	1	-360	360;
	7047	6293	0.016851	0.051537	0.00581	59.512001	59.512001	59.512001	0	0	1	-360	360;
	7610	7624	0.01638	0.053901	0.00557	82.018599	82.018599	82.018599	0	0	1	-360	360;
	7610	7624	0.012298	0.045521	0.00417	88.183435	88.183435	88.183435	0	0	1	-360	360;
	7624	6229	0.025017	0.048653	0.00525	69.526217	69.526217	69.526217	0	0	1	-360	360;
	3149	8092	0.00186	0.007174	0.00063	79.67742	79.67742	79.67742	0	0	1	-360	360;
	2698	8092	0.001397	0.005372	0.00047	62.076476	62.076476	62.076476	0	0	1	-360	360;
	2008	3994	0.010777	0.043479	0.00362	61.967114	61.967114	61.967114	0	0	1	-360	360;
	5788	2008	0.016413	0.047628	0.00558	62.20108	62.20108	62.20108	0	0	1	-360	360;
	1319	5007	0.004	0.011281	0.00138	60.014798	60.014798	60.014798	0	0	1	-360	360;
	4705	5007	0.004041	0.015099	0.00139	79.899192	79.899192	79.899192	0	0	1	-360	360;
	3149	7655	0.004636	0.017579	0.00158	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7655	4705	0.006529	0.025091	0.00222	86.834133	86.834133	86.834133	0	0	1	-360	360;
	5937	3149	0.01781	0.052339	0.00614	85.656875	85.656875	85.656875	0	0	1	-360	360;
	1319	2698	0.003744	0.010149	0.00129	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6030	4705	0.006248	0.023702	0.00212	72.839647	72.839647	72.839647	0	0	1	-360	360;
	6030	5807	0.035107	0.106207	0.01201	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6030	8147	0.014769	0.045182	0.0051	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6030	3088	0.004322	0.012198	0.00149	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6030	6256	0.015678	0.054083	0.00541	57.157677	57.157677	57.157677	0	0	1	-360	360;
	30	8018	0.011331	0.023636	0.00236	83.343514	83.343514	83.343514	0	0	1	-360	360;
	874	5987	0.000591	0.002969	0.00473	403.533578	403.533578	403.533578	0	0	1	-360	360;
	874	5987	0.000281	0.0025	0.00587	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4489	5987	0.00781	0.039159	0.06232	373.547899	373.547899	373.547899	0	0	1	-360	360;
	5987	1026	0.00419	0.02431	0.03336	431.73289	431.73289	431.73289	0	0	1	-360	360;
	318	5987	0.014479	0.06989	0.11605	389.805447	389.805447	389.805447	0	0	1	-360	360;
	318	5987	0.013419	0.07594	0.10737	402.963526	402.963526	402.963526	0	0	1	-360	360;
	3296	3664	0.00208	0.02183	0.48327	1080.92458	1080.92458	1080.92458	0	0	1	-360	360;
	6564	3459	0.028083	0.103397	0.00968	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5448	2021	0.005793	0.017802	0.00193	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8117	8147	0.016124	0.049322	0.00555	57.157677	57.157677	57.157677	0	0	1	-360	360;
	30	3171	0.018471	0.041802	0.00384	67.397474	67.397474	67.397474	0	0	1	-360	360;
	1262	974	0.031719	0.066198	0.00658	57.157677	57.157677	57.157677	0	0	1	-360	360;
	188	1583	0.017554	0.056124	0.00587	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5525	1137	0.011207	0.037917	0.11382	71.077095	71.077095	71.077095	0	0	1	-360	360;
	2369	3251	0.02261	0.105521	0.32503	240.347268	240.347268	240.347268	0	0	1	-360	360;
	3459	4488	0.002397	0.013603	0.00114	64.949792	64.949792	64.949792	0	0	1	-360	360;
	4488	465	0.021901	0.055818	0.00454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3172	3459	0.007719	0.030066	0.00267	73.404365	73.404365	73.404365	0	0	1	-360	360;
	3459	2110	0.017711	0.05	0.00611	57.157677	57.157677	57.157677	0	0	1	-360	360;
	518	3459	0.035322	0.070802	0.00733	73.218984	73.218984	73.218984	0	0	1	-360	360;
	4200	5472	0.017281	0.050777	0.00596	68.152718	68.152718	68.152718	0	0	1	-360	360;
	4182	4760	0.04324	0.168223	0.01492	72.276644	72.276644	72.276644	0	0	1	-360	360;
	5494	4182	0.022678	0.082017	0.00788	81.094168	81.094168	81.094168	0	0	1	-360	360;
	4182	2545	0.026	0.079537	0.00896	61.620548	61.620548	61.620548	0	0	1	-360	360;
	4907	6259	0.027347	0.080372	0.00943	61.706094	61.706094	61.706094	0	0	1	-360	360;
	4726	4907	0.023364	0.074372	0.00806	67.956095	67.956095	67.956095	0	0	1	-360	360;
	7777	2813	0.011421	0.02757	0.00237	86.084224	86.084224	86.084224	0	0	1	-360	360;
	4489	3251	0.011039	0.047169	0.08828	362.83655	362.83655	362.83655	0	0	1	-360	360;
	1026	3251	0.015159	0.0732	0.12146	399.369832	399.369832	399.369832	0	0	1	-360	360;
	2432	3251	0.00119	0.005729	0.00946	392.050982	392.050982	392.050982	0	0	1	-360	360;
	6980	3251	0.00182	0.060349	0.09625	316.709162	316.709162	316.709162	0	0	1	-360	360;
	6610	3510	0.00795	0.032132	0.00267	57.157677	57.157677	57.157677	0	0	1	-360	360;
	773	6610	0.006372	0.021934	0.00202	73.057799	73.057799	73.057799	0	0	1	-360	360;
	2378	6610	0.015992	0.036041	0.00977	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6610	3967	0.022843	0.087157	0.00773	62.673773	62.673773	62.673773	0	0	1	-360	360;
	6887	2640	0.00495	0.014942	0.00169	57.157677	57.157677	57.157677	0	0	1	-360	360;
	974	6887	0.000901	0.00238	0.0134	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3510	6887	0.008942	0.027339	0.00301	58.623581	58.623581	58.623581	0	0	1	-360	360;
	3510	2640	0.00476	0.017587	0.0016	57.754403	57.754403	57.754403	0	0	1	-360	360;
	6259	2640	0.037339	0.128802	0.01289	57.157677	57.157677	57.157677	0	0	1	-360	360;
	974	4462	0.00081	0.00219	0.01251	81.155899	81.155899	81.155899	0	0	1	-360	360;
	974	4043	0.005132	0.018438	0.00177	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3977	3510	0.005934	0.023107	0.00207	74.070824	74.070824	74.070824	0	0	1	-360	360;
	3977	3510	0.005802	0.021397	0.00203	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3977	205	0.000463	0.001421	0.00675	71.2419	71.2419	71.2419	0	0	1	-360	360;
	3977	5525	0.000909	0.002521	0.0069	74.387668	74.387668	74.387668	0	0	1	-360	360;
	3510	4043	0.009347	0.025364	0.00322	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8326	4462	0.00062	0.002017	0.00979	76.910989	76.910989	76.910989	0	0	1	-360	360;
	205	8326	0.000917	0.002364	0.01364	58.602242	58.602242	58.602242	0	0	1	-360	360;
	8326	2378	0.009248	0.016446	0.00193	72.799637	72.799637	72.799637	0	0	1	-360	360;
	5525	2421	0.00295	0.008	0.00103	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2378	5525	0.003339	0.012612	0.00118	73.415987	73.415987	73.415987	0	0	1	-360	360;
	2378	2421	0.00319	0.012504	0.00112	65.225102	65.225102	65.225102	0	0	1	-360	360;
	773	2378	0.007826	0.030521	0.0027	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6445	3664	0.00063	0.00702	0.14912	1002.36101	1002.36101	1002.36101	0	0	1	-360	360;
	3664	5103	0.00184	0.02018	0.42835	1097.32246	1097.32246	1097.32246	0	0	1	-360	360;
	1321	5513	0.003818	0.010661	0.0013	85.476829	85.476829	85.476829	0	0	1	-360	360;
	6831	1321	0.008678	0.025198	0.00296	83.875461	83.875461	83.875461	0	0	1	-360	360;
	5215	1321	0.008843	0.024661	0.003	57.157677	57.157677	57.157677	0	0	1	-360	360;
	120	6229	0.023347	0.041909	0.00491	63.368049	63.368049	63.368049	0	0	1	-360	360;
	7861	4735	0.018091	0.055347	0.00624	60.726793	60.726793	60.726793	0	0	1	-360	360;
	6580	1256	0.03443	0.093347	0.01187	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6580	7988	0.041033	0.111248	0.01415	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7436	6998	0.015843	0.05686	0.00546	62.290055	62.290055	62.290055	0	0	1	-360	360;
	4200	7436	0.020397	0.064959	0.00704	79.002579	79.002579	79.002579	0	0	1	-360	360;
	188	4760	0.017893	0.054959	0.00598	57.157677	57.157677	57.157677	0	0	1	-360	360;
	884	5513	0.029496	0.09295	0.01005	71.399845	71.399845	71.399845	0	0	1	-360	360;
	8375	7610	0.011579	0.046397	0.00395	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6980	2942	0.00099	0.0042	0.00784	397.419231	397.419231	397.419231	0	0	1	-360	360;
	6980	2942	0.00076	0.00449	0.00607	395.07005	395.07005	395.07005	0	0	1	-360	360;
	8022	1300	0.006587	0.016868	0.00139	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1300	5513	0.016876	0.064909	0.00574	63.011385	63.011385	63.011385	0	0	1	-360	360;
	8022	4613	0.036612	0.115372	0.01248	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2892	3587	0.009421	0.030826	0.0032	91.936789	91.936789	91.936789	0	0	1	-360	360;
	8375	3587	0.008041	0.030438	0.00274	82.392981	82.392981	82.392981	0	0	1	-360	360;
	4613	6272	0.000744	0.002182	0.00025	72.869179	72.869179	72.869179	0	0	1	-360	360;
	6980	3965	0.000901	0.003979	0.00712	218.266496	218.266496	218.266496	0	0	1	-360	360;
	6980	2817	0.01212	0.05176	0.09653	385.759065	385.759065	385.759065	0	0	1	-360	360;
	6065	5103	0.00256	0.02765	0.62005	1327.50488	1327.50488	1327.50488	0	0	1	-360	360;
	2556	6727	0.00224	0.007256	0.03306	78.828057	78.828057	78.828057	0	0	1	-360	360;
	8075	5103	0.0009	0.00943	0.20859	1327.08101	1327.08101	1327.08101	0	0	1	-360	360;
	2892	8375	0.007744	0.024273	0.00265	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6727	8375	0.006372	0.017992	0.0022	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7988	7047	0.048554	0.142669	0.01675	83.027813	83.027813	83.027813	0	0	1	-360	360;
	3088	8147	0.010802	0.030488	0.00373	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5788	4162	0.015496	0.047397	0.01394	58.897747	58.897747	58.897747	0	0	1	-360	360;
	6831	5513	0.010207	0.030826	0.00348	63.697277	63.697277	63.697277	0	0	1	-360	360;
	5513	4162	0.026504	0.077876	0.00914	79.850227	79.850227	79.850227	0	0	1	-360	360;
	2816	596	0.00073	0.00905	0.18786	1213.79914	1213.79914	1213.79914	0	0	1	-360	360;
	8195	4483	0.00192	0.02296	0.54388	904.693784	904.693784	904.693784	0	0	1	-360	360;
	1659	5933	0.01431	0.05899	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5933	1313	0.02607	0.07168	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5509	6199	0.00054	0.00465	0.13926	1612.27406	1612.27406	1612.27406	0	0	1	-360	360;
	5509	2063	0.00126	0.01124	0.32378	1613.43837	1613.43837	1613.43837	0	0	1	-360	360;
	5509	6511	0.00048	0.00527	0.12407	1550.58488	1550.58488	1550.58488	0	0	1	-360	360;
	5509	5709	0.00082	0.00938	0.20371	1605.24141	1605.24141	1605.24141	0	0	1	-360	360;
	1798	8341	0.00028	0.003502	0	178.878846	178.878846	178.878846	0	0	1	-360	360;
	1798	3174	0.691467	4.414711	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1798	866	0.02004	0.0916	0	122.174274	122.174274	122.174274	0	0	1	-360	360;
	1798	207	0.841911	4.768444	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1798	6192	0.026	0.08948	0	110.329385	110.329385	110.329385	0	0	1	-360	360;
	1798	8139	0.002631	0.011889	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	1798	5242	0.7232	1.944	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8341	866	0.156169	0.546756	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8341	6192	0.168809	0.580622	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8341	8139	0.019222	0.068209	0	118.133226	118.133226	118.133226	0	0	1	-360	360;
	7775	1473	0.00039	0.00386	0.07949	1445.70533	1445.70533	1445.70533	0	0	1	-360	360;
	7775	1058	0.00096	0.01138	0.24684	1478.73145	1478.73145	1478.73145	0	0	1	-360	360;
	7775	8231	0.00037	0.00325	0.09724	1441.53905	1441.53905	1441.53905	0	0	1	-360	360;
	7775	8099	0.00029	0.00339	0.07351	1587.36723	1587.36723	1587.36723	0	0	1	-360	360;
	3174	2089	0.009458	0.158871	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2089	4028	0	0.000222	0	136.284686	136.284686	136.284686	0	0	1	-360	360;
	6192	2089	0.312951	7.130667	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	2089	0.247649	1.075911	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4032	2089	0.06056	0.18612	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7555	2089	2.142489	4.347733	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4687	2089	0.07132	1.068889	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	246	2089	0.02268	0.08652	0	109.652326	109.652326	109.652326	0	0	1	-360	360;
	2222	2089	2.007778	4.494222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	2089	0.006178	0.028938	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	2089	0.027711	0.260951	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3174	5074	0	0.000222	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	4172	3174	0.00692	0.337378	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	207	3174	0.025991	0.10048	0	145.747139	145.747139	145.747139	0	0	1	-360	360;
	6192	3174	0.083711	0.8648	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	3174	0.447378	1.862089	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4032	3174	0.062169	0.153191	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7555	3174	2.492533	4.725778	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4687	3174	0.020271	0.08244	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	246	3174	0.11416	1.777556	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5242	3174	0.574978	2.992578	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2222	3174	1.887067	4.964444	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	3174	0.250778	0.817422	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	3174	0.007049	0.029609	0	132.075542	132.075542	132.075542	0	0	1	-360	360;
	5074	4172	0.003049	0.020218	0.00658	118.157908	118.157908	118.157908	0	0	1	-360	360;
	4028	4172	0.002911	0.020262	0.00658	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4483	2601	4e-05	0.00041	0.01232	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4483	596	0.00087	0.01087	0.21786	1313.02823	1313.02823	1313.02823	0	0	1	-360	360;
	4483	945	4e-05	0.00046	0.01098	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4483	596	0.00095	0.00924	0.25272	1493.44241	1493.44241	1493.44241	0	0	1	-360	360;
	1659	4738	0.528017	2.636157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1659	1313	0.032031	0.128659	0	249.138881	249.138881	249.138881	0	0	1	-360	360;
	1473	7041	0.00044	0.0032	0.09137	1654.21587	1654.21587	1654.21587	0	0	1	-360	360;
	1473	2282	0.00046	0.00489	0.11819	1742.56526	1742.56526	1742.56526	0	0	1	-360	360;
	8231	7041	0.0003	0.00352	0.07776	1581.43836	1581.43836	1581.43836	0	0	1	-360	360;
	8231	78	0.00051	0.00454	0.13025	1654.5318	1654.5318	1654.5318	0	0	1	-360	360;
	1199	596	0.00068	0.00746	0.17522	1625.48371	1625.48371	1625.48371	0	0	1	-360	360;
	1199	7802	0.00093	0.01022	0.2401	1761.39445	1761.39445	1761.39445	0	0	1	-360	360;
	7802	1244	4e-05	0.0004	0.00941	1507.92302	1507.92302	1507.92302	0	0	1	-360	360;
	7802	8099	0.00078	0.00698	0.20002	1770.56223	1770.56223	1770.56223	0	0	1	-360	360;
	1058	1528	4e-05	0.00043	0.0094	1592.28581	1592.28581	1592.28581	0	0	1	-360	360;
	1058	6395	0.00025	0.00309	0.06458	1690.57107	1690.57107	1690.57107	0	0	1	-360	360;
	825	7041	0.00047	0.00538	0.12192	1639.19754	1639.19754	1639.19754	0	0	1	-360	360;
	1921	7041	0.00046	0.00548	0.11938	1776.18637	1776.18637	1776.18637	0	0	1	-360	360;
	1921	5443	0.00016	0.00173	0.03322	1866.42012	1866.42012	1866.42012	0	0	1	-360	360;
	1921	5709	0.00166	0.01471	0.4064	1764.38785	1764.38785	1764.38785	0	0	1	-360	360;
	825	7041	0.00051	0.00512	0.10635	1809.14931	1809.14931	1809.14931	0	0	1	-360	360;
	6080	8099	0.00021	0.00248	0.05377	1534.29231	1534.29231	1534.29231	0	0	1	-360	360;
	207	4172	0.004391	0.021689	0	288.312154	288.312154	288.312154	0	0	1	-360	360;
	2216	5709	0.00024	0.00289	0.06298	1660.04866	1660.04866	1660.04866	0	0	1	-360	360;
	866	3186	0	0.000222	0	337.883968	337.883968	337.883968	0	0	1	-360	360;
	866	8139	0.01788	0.100471	0	125.356398	125.356398	125.356398	0	0	1	-360	360;
	3186	2637	0.006458	0.039982	0.01196	125.741433	125.741433	125.741433	0	0	1	-360	360;
	207	6192	0.106151	0.993333	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	207	5242	0.699778	3.232311	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8139	6192	0.257058	0.43592	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5242	6192	0.002769	0.011449	0	136.345221	136.345221	136.345221	0	0	1	-360	360;
	596	7549	0.00034	0.00305	0.08424	1603.19513	1603.19513	1603.19513	0	0	1	-360	360;
	596	6395	0.00123	0.01255	0.32089	1845.21358	1845.21358	1845.21358	0	0	1	-360	360;
	596	8385	6e-05	0.00065	0.01652	1445.69282	1445.69282	1445.69282	0	0	1	-360	360;
	596	7223	6e-05	0.00063	0.01685	1515.6586	1515.6586	1515.6586	0	0	1	-360	360;
	596	5656	6e-05	0.00055	0.01526	1767.88015	1767.88015	1767.88015	0	0	1	-360	360;
	596	4917	0.00079	0.00968	0.20826	1374.64631	1374.64631	1374.64631	0	0	1	-360	360;
	596	6166	7e-05	0.00079	0.0173	2092.7489	2092.7489	2092.7489	0	0	1	-360	360;
	718	2981	0	0.000222	0	381.428505	381.428505	381.428505	0	0	1	-360	360;
	6862	718	2.064178	7.545333	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6599	718	0.153711	0.482489	0	115.998647	115.998647	115.998647	0	0	1	-360	360;
	7555	718	0.015809	0.12368	0	154.358723	154.358723	154.358723	0	0	1	-360	360;
	2222	718	0.057978	0.363422	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	718	0.6932	2.196889	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	718	0.560356	2.395867	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4917	2062	0.00024	0.00213	0.05875	775.769622	775.769622	775.769622	0	0	1	-360	360;
	4917	6930	0.00093	0.00884	0.23566	1305.99361	1305.99361	1305.99361	0	0	1	-360	360;
	8139	1975	0	0.000222	0	315.913597	315.913597	315.913597	0	0	1	-360	360;
	3328	5443	0.00018	0.00204	0.0455	1802.92161	1802.92161	1802.92161	0	0	1	-360	360;
	5229	5443	0.00018	0.00208	0.04456	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2948	4637	0	0.000222	0	114.164145	114.164145	114.164145	0	0	1	-360	360;
	6862	2948	0.02068	0.093391	0	173.042528	173.042528	173.042528	0	0	1	-360	360;
	4032	2948	0.777778	7.408	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2948	6599	0.00884	0.07292	0	123.083341	123.083341	123.083341	0	0	1	-360	360;
	246	2948	0.131671	8.295556	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2948	1469	0.082462	0.568222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4637	1469	0.0042	0.02624	0.01124	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7415	7413	0.00031	0.00317	0.0695	1256.43731	1256.43731	1256.43731	0	0	1	-360	360;
	7415	7041	0.00043	0.00498	0.10234	1494.85947	1494.85947	1494.85947	0	0	1	-360	360;
	8196	7041	0.00042	0.00508	0.1002	1732.63004	1732.63004	1732.63004	0	0	1	-360	360;
	6862	4032	0.110671	0.773067	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	6599	0.00912	0.037609	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	7555	0.098049	0.598711	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	4687	0.019031	1.160356	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	246	0.024071	0.915733	0	104.516189	104.516189	104.516189	0	0	1	-360	360;
	2222	6862	1.484267	4.479111	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	6862	0.011578	0.063271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	6862	0.654889	1.839111	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6862	1469	0.022049	0.10076	0	135.609706	135.609706	135.609706	0	0	1	-360	360;
	4032	6599	0.459867	2.604711	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4032	7555	2.320222	6.088	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4032	4687	0.006369	0.048089	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4032	246	0.006631	0.0432	0	145.521106	145.521106	145.521106	0	0	1	-360	360;
	2222	4032	2.567244	6.767111	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	4032	0.309889	1.214178	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	4032	0.009191	0.043289	0	127.213762	127.213762	127.213762	0	0	1	-360	360;
	4032	1469	0.691022	5.239556	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4738	1313	0.002669	0.01718	0	225.874182	225.874182	225.874182	0	0	1	-360	360;
	7413	5000	0.00074	0.00694	0.19186	1260.94255	1260.94255	1260.94255	0	0	1	-360	360;
	7413	7041	0.00068	0.008	0.1602	1549.46861	1549.46861	1549.46861	0	0	1	-360	360;
	6599	3656	0	0.000222	0	159.490962	159.490962	159.490962	0	0	1	-360	360;
	7555	6599	0.00524	0.030631	0	168.883787	168.883787	168.883787	0	0	1	-360	360;
	4687	6599	0.33932	3.814667	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	246	6599	0.22696	2.614933	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2222	6599	0.190671	0.719378	0	106.727412	106.727412	106.727412	0	0	1	-360	360;
	4239	6599	1.050222	2.105556	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	6599	1.431644	4.294578	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1469	6599	0.022618	0.161089	0	132.945119	132.945119	132.945119	0	0	1	-360	360;
	5000	1439	0.00088	0.00895	0.22766	1817.88993	1817.88993	1817.88993	0	0	1	-360	360;
	7555	1395	0	0.000222	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	4687	7555	1.827778	7.768	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2222	7555	0.020849	0.153382	0	120.989551	120.989551	120.989551	0	0	1	-360	360;
	4239	7555	0.156231	0.656267	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	7555	0.176498	0.891556	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7041	2063	0.00067	0.00777	0.17154	1639.08565	1639.08565	1639.08565	0	0	1	-360	360;
	4687	246	0.004578	0.061889	0	108.460069	108.460069	108.460069	0	0	1	-360	360;
	4687	2637	0.01348	0.05932	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4239	4687	0.066391	2.802622	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	4687	0.194258	2.220444	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4687	1469	0.008151	0.072822	0	119.61387	119.61387	119.61387	0	0	1	-360	360;
	246	2637	0.016502	0.055382	0	110.780151	110.780151	110.780151	0	0	1	-360	360;
	4239	246	0.0806	4.633778	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4091	246	0.831556	3.729644	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	246	1469	0.007951	0.074151	0	126.14985	126.14985	126.14985	0	0	1	-360	360;
	1439	5304	0.0004	0.00362	0.10368	1616.27974	1616.27974	1616.27974	0	0	1	-360	360;
	6395	8028	0.00015	0.00177	0.0384	1581.38109	1581.38109	1581.38109	0	0	1	-360	360;
	5242	7458	0	0.000222	0	136.505782	136.505782	136.505782	0	0	1	-360	360;
	288	5656	2e-05	0.00022	0.00551	1701.55674	1701.55674	1701.55674	0	0	1	-360	360;
	4239	2222	0.02864	0.172902	0	109.491505	109.491505	109.491505	0	0	1	-360	360;
	2222	4091	0.017591	0.1404	0	111.040738	111.040738	111.040738	0	0	1	-360	360;
	4239	4091	0.00752	0.031711	0	104.584519	104.584519	104.584519	0	0	1	-360	360;
	1469	7632	0	0.000222	0	549.63939	549.63939	549.63939	0	0	1	-360	360;
	7955	2076	0.032587	0.058388	0.00653	57.157677	57.157677	57.157677	0	0	1	-360	360;
	576	2076	0.029752	0.070686	0.00591	73.633949	73.633949	73.633949	0	0	1	-360	360;
	3995	3650	0.00699	0.04244	0.06863	488.072068	488.072068	488.072068	0	0	1	-360	360;
	3650	6832	0.0008	0.006081	0.00853	337.195617	337.195617	337.195617	0	0	1	-360	360;
	3650	5451	0.004669	0.041919	0.11404	299.841932	299.841932	299.841932	0	0	1	-360	360;
	3650	5451	0.00476	0.04106	0.11627	305.991717	305.991717	305.991717	0	0	1	-360	360;
	6347	1864	0.00146	0.0163	0.33408	1327.94915	1327.94915	1327.94915	0	0	1	-360	360;
	6347	1864	0.00098	0.01485	0.35427	1236.69983	1236.69983	1236.69983	0	0	1	-360	360;
	6347	1864	0.00143	0.01661	0.32712	1133.3874	1133.3874	1133.3874	0	0	1	-360	360;
	1372	6347	0.00209	0.0241	0.75637	1925.56806	1925.56806	1925.56806	0	0	1	-360	360;
	1372	6347	0.00273	0.02677	0.70404	1460.85398	1460.85398	1460.85398	0	0	1	-360	360;
	1372	6347	0.00279	0.02623	0.71784	1281.88384	1281.88384	1281.88384	0	0	1	-360	360;
	5011	739	0.007479	0.032207	0.00402	83.045151	83.045151	83.045151	0	0	1	-360	360;
	5011	1939	0.04319	0.16186	0.01732	73.860483	73.860483	73.860483	0	0	1	-360	360;
	1474	2993	0.007256	0.029702	0.00418	100.348303	100.348303	100.348303	0	0	1	-360	360;
	2993	5495	0.006488	0.036041	0.00367	109.528398	109.528398	109.528398	0	0	1	-360	360;
	3448	4147	0.028653	0.066364	0.00692	66.256417	66.256417	66.256417	0	0	1	-360	360;
	6452	3448	0.010463	0.027446	0.0025	66.435511	66.435511	66.435511	0	0	1	-360	360;
	1309	2886	0.012409	0.071919	0.1254	418.371711	418.371711	418.371711	0	0	1	-360	360;
	1309	2886	0.012659	0.07045	0.12786	453.921881	453.921881	453.921881	0	0	1	-360	360;
	2886	8279	0.003419	0.02344	0.03412	426.544878	426.544878	426.544878	0	0	1	-360	360;
	2886	6737	0.002469	0.01486	0.02003	329.654614	329.654614	329.654614	0	0	1	-360	360;
	2886	4988	0.00682	0.03368	0.04313	467.578755	467.578755	467.578755	0	0	1	-360	360;
	2886	833	0.02036	0.09743	0.12263	341.634482	341.634482	341.634482	0	0	1	-360	360;
	7955	3428	0.048	0.080331	0.00958	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7955	3108	0.00262	0.008083	0.00087	77.992983	77.992983	77.992983	0	0	1	-360	360;
	7955	5928	0.004107	0.023083	0.00244	67.940282	67.940282	67.940282	0	0	1	-360	360;
	7955	7248	0.003207	0.015719	0.00189	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7955	649	0.004727	0.02143	0.00257	70.712239	70.712239	70.712239	0	0	1	-360	360;
	7955	3515	0.004041	0.012702	0.00151	64.123102	64.123102	64.123102	0	0	1	-360	360;
	7955	2317	0.017802	0.037777	0.0043	61.134898	61.134898	61.134898	0	0	1	-360	360;
	2651	4924	0.002934	0.014273	0.00164	71.068712	71.068712	71.068712	0	0	1	-360	360;
	3311	2651	0.006231	0.032174	0.00338	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5491	2666	0.026876	0.085248	0.00679	77.810269	77.810269	77.810269	0	0	1	-360	360;
	5491	5778	0.050636	0.099223	0.01231	94.807057	94.807057	94.807057	0	0	1	-360	360;
	889	3023	0.045347	0.092	0.01084	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5267	3023	0.005231	0.013504	0.00128	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1365	7853	0.024752	0.057446	0.00473	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1365	7197	0.015083	0.065182	0.00823	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1365	8103	0.03986	0.070686	0.00773	65.268351	65.268351	65.268351	0	0	1	-360	360;
	4728	1365	0.036231	0.064248	0.00703	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1365	4257	0.004537	0.022041	0.00247	64.711826	64.711826	64.711826	0	0	1	-360	360;
	8267	6859	0.052603	0.157802	0.01363	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8103	6859	0.005562	0.024083	0.00249	71.136158	71.136158	71.136158	0	0	1	-360	360;
	8103	6859	0.005446	0.024554	0.00244	58.6815	58.6815	58.6815	0	0	1	-360	360;
	3288	6859	0.009107	0.051661	0.00535	72.870703	72.870703	72.870703	0	0	1	-360	360;
	6859	690	0.00419	0.009868	0.00117	68.626174	68.626174	68.626174	0	0	1	-360	360;
	4112	6859	0.024281	0.13338	0.01325	65.972534	65.972534	65.972534	0	0	1	-360	360;
	4112	6859	0.02376	0.123496	0.01296	69.192607	69.192607	69.192607	0	0	1	-360	360;
	4562	6859	0.005331	0.020752	0.00196	61.595208	61.595208	61.595208	0	0	1	-360	360;
	4158	6411	0.001107	0.004603	0.00063	59.551821	59.551821	59.551821	0	0	1	-360	360;
	6411	7095	0.005628	0.030702	0.00301	65.382666	65.382666	65.382666	0	0	1	-360	360;
	649	7351	0.024628	0.130893	0.01447	59.970025	59.970025	59.970025	0	0	1	-360	360;
	649	7351	0.024132	0.133463	0.01418	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5409	7351	0.01419	0.07062	0.00827	76.579093	76.579093	76.579093	0	0	1	-360	360;
	5409	7351	0.014471	0.069182	0.00843	76.559278	76.559278	76.559278	0	0	1	-360	360;
	6132	7620	0.008372	0.045231	0.00455	61.627407	61.627407	61.627407	0	0	1	-360	360;
	6296	6132	0.024628	0.097983	0.01339	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6132	6428	0.002471	0.009331	0.00134	57.157677	57.157677	57.157677	0	0	1	-360	360;
	303	1714	0.025702	0.144587	0.01399	67.419004	67.419004	67.419004	0	0	1	-360	360;
	3812	303	0.003421	0.019752	0.00201	59.581543	59.581543	59.581543	0	0	1	-360	360;
	3213	38	0.004562	0.017174	0.00175	57.157677	57.157677	57.157677	0	0	1	-360	360;
	38	827	0.018868	0.083661	0.01108	59.437125	59.437125	59.437125	0	0	1	-360	360;
	38	4439	0.01019	0.027917	0.00334	58.634631	58.634631	58.634631	0	0	1	-360	360;
	8011	3213	0.019471	0.064058	0.00777	62.815525	62.815525	62.815525	0	0	1	-360	360;
	5680	1533	0.053446	0.118207	0.01283	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1533	5027	0.049521	0.123438	0.01198	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2427	4599	0.00914	0.042039	0.05368	358.472371	358.472371	358.472371	0	0	1	-360	360;
	4599	833	0.004711	0.021591	0.02835	290.324798	290.324798	290.324798	0	0	1	-360	360;
	1318	8383	0	0.000893	0	75.600173	75.600173	75.600173	0	0	1	-360	360;
	4265	1318	0.00405	0.021132	0.0022	64.47767	64.47767	64.47767	0	0	1	-360	360;
	4265	1318	0.004587	0.020711	0.00249	76.221477	76.221477	76.221477	0	0	1	-360	360;
	8383	3350	0.044182	0.091041	0.00842	93.379258	93.379258	93.379258	0	0	1	-360	360;
	8383	4126	0.022653	0.047421	0.00551	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8383	6786	0.058876	0.108463	0.01145	78.845014	78.845014	78.845014	0	0	1	-360	360;
	412	3249	0.013174	0.026992	0.0032	108.170141	108.170141	108.170141	0	0	1	-360	360;
	6652	412	0.001273	0.002818	0.00031	104.963214	104.963214	104.963214	0	0	1	-360	360;
	1548	1474	0.009471	0.048769	0.00511	116.055614	116.055614	116.055614	0	0	1	-360	360;
	5413	3156	0.017388	0.04138	0.00666	85.578569	85.578569	85.578569	0	0	1	-360	360;
	3156	8077	0.017339	0.054661	0.00372	112.484593	112.484593	112.484593	0	0	1	-360	360;
	4924	4191	0.025033	0.137942	0.0147	58.256057	58.256057	58.256057	0	0	1	-360	360;
	4924	4191	0.024521	0.140653	0.0144	57.784125	57.784125	57.784125	0	0	1	-360	360;
	4191	6512	0.021198	0.056041	0.00381	62.871539	62.871539	62.871539	0	0	1	-360	360;
	4191	7409	0.007174	0.035	0.00421	73.7757	73.7757	73.7757	0	0	1	-360	360;
	2428	4191	0.010826	0.029992	0.00272	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2428	4191	0.012876	0.028058	0.00323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2191	7026	0.034248	0.143124	0.01898	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5475	7026	0.015421	0.066364	0.00843	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4728	7853	0.014512	0.035901	0.00349	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1020	1309	0.00863	0.054419	0.08751	539.911033	539.911033	539.911033	0	0	1	-360	360;
	1020	1309	0.008461	0.055479	0.08573	515.908619	515.908619	515.908619	0	0	1	-360	360;
	434	3897	0.013256	0.073248	0.00724	74.693842	74.693842	74.693842	0	0	1	-360	360;
	8153	3897	0.004033	0.018174	0.00221	93.218645	93.218645	93.218645	0	0	1	-360	360;
	576	3428	0.016008	0.027066	0.00319	78.301825	78.301825	78.301825	0	0	1	-360	360;
	4749	3288	0.013537	0.070083	0.0073	70.0494	70.0494	70.0494	0	0	1	-360	360;
	7068	4749	0.004744	0.024876	0.00261	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3108	2514	0.014364	0.037438	0.00315	64.491959	64.491959	64.491959	0	0	1	-360	360;
	7541	2048	0.010364	0.020314	0.0025	63.511324	63.511324	63.511324	0	0	1	-360	360;
	2048	3336	0.007289	0.018207	0.00175	65.363423	65.363423	65.363423	0	0	1	-360	360;
	7577	1366	0.011744	0.052446	0.00645	75.682099	75.682099	75.682099	0	0	1	-360	360;
	6759	7577	0.015438	0.062876	0.00836	78.015275	78.015275	78.015275	0	0	1	-360	360;
	8182	5956	0.00157	0.01612	0.38916	1026.09562	1026.09562	1026.09562	0	0	1	-360	360;
	8182	5956	0.00154	0.01643	0.38122	1052.24376	1052.24376	1052.24376	0	0	1	-360	360;
	3311	4924	0.001967	0.008438	0.00471	97.802691	97.802691	97.802691	0	0	1	-360	360;
	4924	7498	0.01162	0.033653	0.0053	83.411722	83.411722	83.411722	0	0	1	-360	360;
	1041	803	0.006901	0.04661	0.0697	433.127537	433.127537	433.127537	0	0	1	-360	360;
	1041	803	0.00782	0.04568	0.07899	448.13257	448.13257	448.13257	0	0	1	-360	360;
	2596	803	0.005401	0.0343	0.05455	399.238369	399.238369	399.238369	0	0	1	-360	360;
	602	803	0.011419	0.07275	0.10245	382.609677	382.609677	382.609677	0	0	1	-360	360;
	3995	5451	0.01044	0.05919	0.1062	492.606196	492.606196	492.606196	0	0	1	-360	360;
	8101	979	0.050314	0.112132	0.01002	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8101	979	0.057017	0.109893	0.01135	78.915699	78.915699	78.915699	0	0	1	-360	360;
	979	3893	0.007	0.028231	0.00401	89.076619	89.076619	89.076619	0	0	1	-360	360;
	979	7654	0.010719	0.040868	0.00585	72.536712	72.536712	72.536712	0	0	1	-360	360;
	979	4241	0.04443	0.10976	0.01116	57.157677	57.157677	57.157677	0	0	1	-360	360;
	979	513	0.021372	0.121033	0.01035	76.547656	76.547656	76.547656	0	0	1	-360	360;
	979	799	0.020579	0.095331	0.01	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2643	8101	0.02386	0.06986	0.02731	72.134512	72.134512	72.134512	0	0	1	-360	360;
	8033	2643	0.062901	0.105628	0.00955	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2643	1939	0.035686	0.149314	0.0209	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5898	2643	0.00181	0.006876	0.00061	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5897	2643	0.002628	0.01462	0.00316	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5897	2643	0.002983	0.014331	0.00358	57.157677	57.157677	57.157677	0	0	1	-360	360;
	380	2643	0.006157	0.03114	0.00335	75.774885	75.774885	75.774885	0	0	1	-360	360;
	1846	6764	0.022	0.043066	0.00519	68.828703	68.828703	68.828703	0	0	1	-360	360;
	1289	1846	0.013438	0.033479	0.00296	58.908988	58.908988	58.908988	0	0	1	-360	360;
	1846	7751	0.013488	0.028893	0.00221	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3109	8329	0.114058	0.230033	0.02527	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8329	855	0.062901	0.133777	0.01366	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8329	1810	0.041868	0.106554	0.01034	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3103	7538	0.052347	0.07786	0.00802	75.05203	75.05203	75.05203	0	0	1	-360	360;
	7538	5058	0.003099	0.006174	0.00072	60.190844	60.190844	60.190844	0	0	1	-360	360;
	804	6655	0.04357	0.108678	0.01068	57.157677	57.157677	57.157677	0	0	1	-360	360;
	804	7408	0.020893	0.043289	0.00509	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4147	8077	0.035289	0.072058	0.0085	71.915217	71.915217	71.915217	0	0	1	-360	360;
	1939	8101	0.050579	0.074298	0.00768	57.875387	57.875387	57.875387	0	0	1	-360	360;
	3706	8101	0.042934	0.087917	0.01045	57.157677	57.157677	57.157677	0	0	1	-360	360;
	739	6460	0.036711	0.086397	0.00943	61.320661	61.320661	61.320661	0	0	1	-360	360;
	6460	7541	0.053124	0.100066	0.00693	61.807263	61.807263	61.807263	0	0	1	-360	360;
	6460	3703	0.022959	0.041727	0.00334	65.934238	65.934238	65.934238	0	0	1	-360	360;
	3615	1495	0.00476	0.023868	0.00255	60.090056	60.090056	60.090056	0	0	1	-360	360;
	7019	3615	0.005529	0.011182	0.00133	64.364688	64.364688	64.364688	0	0	1	-360	360;
	6869	659	0.00183	0.02821	0.56605	1693.06688	1693.06688	1693.06688	0	0	1	-360	360;
	6758	6869	0.00198	0.02074	0.40897	1249.10651	1249.10651	1249.10651	0	0	1	-360	360;
	8061	5401	0.022917	0.094893	0.01245	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8061	5401	0.023347	0.092868	0.01269	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5401	2726	0.030364	0.073256	0.01043	64.086711	64.086711	64.086711	0	0	1	-360	360;
	5401	5855	0.004562	0.020132	0.00248	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5401	5855	0.010562	0.020686	0.00255	66.493049	66.493049	66.493049	0	0	1	-360	360;
	855	5401	0.01838	0.07314	0.00999	85.956953	85.956953	85.956953	0	0	1	-360	360;
	5401	4573	0.005628	0.029702	0.00306	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5401	4573	0.005777	0.029273	0.00314	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2547	5401	0.014529	0.236802	0.009	60.706406	60.706406	60.706406	0	0	1	-360	360;
	2340	1587	0.005818	0.032182	0.00332	73.964701	73.964701	73.964701	0	0	1	-360	360;
	2340	4265	0.003636	0.021843	0.00837	72.404487	72.404487	72.404487	0	0	1	-360	360;
	3125	1864	0.00318	0.02624	0.70048	1150.80743	1150.80743	1150.80743	0	0	1	-360	360;
	3125	6758	0.00442	0.03649	0.9038	1264.05245	1264.05245	1264.05245	0	0	1	-360	360;
	4185	1501	0.005099	0.027868	0.00297	62.528974	62.528974	62.528974	0	0	1	-360	360;
	4185	1587	0.00995	0.049397	0.0058	71.112533	71.112533	71.112533	0	0	1	-360	360;
	4185	1587	0.01014	0.048397	0.00592	70.927151	70.927151	70.927151	0	0	1	-360	360;
	4185	1534	0.007917	0.04757	0.00466	75.498241	75.498241	75.498241	0	0	1	-360	360;
	4185	82	0.004008	0.022182	0.00237	96.642581	96.642581	96.642581	0	0	1	-360	360;
	4185	185	0.006603	0.039124	0.00364	72.03239	72.03239	72.03239	0	0	1	-360	360;
	7541	5115	0.00657	0.028281	0.00355	74.431298	74.431298	74.431298	0	0	1	-360	360;
	6655	7541	0.071471	0.124661	0.01173	63.170093	63.170093	63.170093	0	0	1	-360	360;
	6655	7541	0.069983	0.127058	0.01149	59.465323	59.465323	59.465323	0	0	1	-360	360;
	7541	3336	0.00743	0.018182	0.00733	107.606566	107.606566	107.606566	0	0	1	-360	360;
	8011	7070	0.004653	0.024339	0.00785	71.035561	71.035561	71.035561	0	0	1	-360	360;
	8077	8011	0.007157	0.03462	0.00419	112.710747	112.710747	112.710747	0	0	1	-360	360;
	8077	8011	0.007298	0.033909	0.00427	102.873149	102.873149	102.873149	0	0	1	-360	360;
	6151	7055	0.00024	0.00281	0.06108	1225.57397	1225.57397	1225.57397	0	0	1	-360	360;
	4911	6568	0.010711	0.056603	0.00585	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3410	4911	0.064273	0.068058	0.00746	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7403	5070	0.013917	0.06676	0.00745	80.675012	80.675012	80.675012	0	0	1	-360	360;
	7403	7796	0.034298	0.076091	0.00831	63.496082	63.496082	63.496082	0	0	1	-360	360;
	4158	2867	0.00676	0.044942	0.00464	74.986109	74.986109	74.986109	0	0	1	-360	360;
	4158	3812	0.025579	0.109438	0.01403	60.808337	60.808337	60.808337	0	0	1	-360	360;
	4158	3812	0.026091	0.107198	0.0143	73.806946	73.806946	73.806946	0	0	1	-360	360;
	3005	3077	0.00319	0.01543	0.00178	58.037905	58.037905	58.037905	0	0	1	-360	360;
	3077	2504	0.003669	0.014628	0.00197	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2609	1445	0.001653	0.007331	0.00092	71.196555	71.196555	71.196555	0	0	1	-360	360;
	434	2609	0.009661	0.04095	0.00386	72.418776	72.418776	72.418776	0	0	1	-360	360;
	1372	8182	0.00089	0.00981	0.23011	1178.86496	1178.86496	1178.86496	0	0	1	-360	360;
	1372	8182	0.00087	0.01	0.22532	1320.96982	1320.96982	1320.96982	0	0	1	-360	360;
	8182	7055	0.00114	0.01867	0.4455	1459.47707	1459.47707	1459.47707	0	0	1	-360	360;
	4453	2621	0.034661	0.089231	0.00845	64.032221	64.032221	64.032221	0	0	1	-360	360;
	1001	4453	0.009463	0.018603	0.00519	88.538384	88.538384	88.538384	0	0	1	-360	360;
	2563	6441	0.032777	0.061248	0.00629	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6230	2563	0.029182	0.043207	0.00506	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6129	5413	0.012826	0.022587	0.00274	60.806432	60.806432	60.806432	0	0	1	-360	360;
	1797	8267	0.01519	0.065017	0.0082	57.157677	57.157677	57.157677	0	0	1	-360	360;
	351	8267	0.002107	0.009	0.00115	74.953719	74.953719	74.953719	0	0	1	-360	360;
	8267	4112	0.010099	0.046587	0.0054	68.109469	68.109469	68.109469	0	0	1	-360	360;
	185	1501	0.001934	0.009008	0.0009	70.789592	70.789592	70.789592	0	0	1	-360	360;
	5680	6764	0.041537	0.116843	0.01002	73.667672	73.667672	73.667672	0	0	1	-360	360;
	6296	7197	0.013091	0.050041	0.00712	65.642162	65.642162	65.642162	0	0	1	-360	360;
	989	6441	0.003058	0.005579	0.00063	61.764395	61.764395	61.764395	0	0	1	-360	360;
	1529	8061	0.008554	0.035397	0.00464	61.671799	61.671799	61.671799	0	0	1	-360	360;
	8358	4950	0.020314	0.102727	0.01104	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4622	4950	0.005008	0.025364	0.00273	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5027	4950	0.042182	0.161132	0.02292	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7065	1196	0.00014	0.000581	0.00124	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3123	1971	0.001339	0.007438	0.00074	72.494224	72.494224	72.494224	0	0	1	-360	360;
	2231	1445	0.004223	0.009347	0.00101	107.166452	107.166452	107.166452	0	0	1	-360	360;
	1445	5053	0.016314	0.034562	0.0041	97.242355	97.242355	97.242355	0	0	1	-360	360;
	8153	1445	0.009331	0.045339	0.0051	88.512282	88.512282	88.512282	0	0	1	-360	360;
	2231	3651	0.00038	0.002769	0.00022	74.70013	74.70013	74.70013	0	0	1	-360	360;
	8153	2231	0.007066	0.043107	0.00389	72.122318	72.122318	72.122318	0	0	1	-360	360;
	1686	1797	0.003124	0.017231	0.00167	97.148807	97.148807	97.148807	0	0	1	-360	360;
	2319	4960	0.020512	0.048562	0.00494	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3706	4960	0.016934	0.03314	0.00417	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8197	3311	0.011083	0.030769	0.00334	72.303318	72.303318	72.303318	0	0	1	-360	360;
	434	5984	0.024314	0.099174	0.01303	63.452833	63.452833	63.452833	0	0	1	-360	360;
	434	5984	0.024769	0.097058	0.01327	63.940769	63.940769	63.940769	0	0	1	-360	360;
	434	5984	0.025231	0.09495	0.00135	79.040874	79.040874	79.040874	0	0	1	-360	360;
	434	4691	0.012917	0.049612	0.00692	73.219555	73.219555	73.219555	0	0	1	-360	360;
	434	2983	0.048058	0.135157	0.01169	115.961685	115.961685	115.961685	0	0	1	-360	360;
	434	3651	0.009372	0.048876	0.00555	80.044372	80.044372	80.044372	0	0	1	-360	360;
	434	4951	0.00538	0.031752	0.00316	58.677118	58.677118	58.677118	0	0	1	-360	360;
	434	4951	0.005273	0.029397	0.00309	62.303773	62.303773	62.303773	0	0	1	-360	360;
	434	8089	0.044562	0.09762	0.01137	78.793381	78.793381	78.793381	0	0	1	-360	360;
	5984	3499	0.017876	0.097091	0.00101	113.091417	113.091417	113.091417	0	0	1	-360	360;
	5984	3226	0.037207	0.166008	0.02178	137.167755	137.167755	137.167755	0	0	1	-360	360;
	6758	8066	0.00018	0.00154	0.03899	1008.03847	1008.03847	1008.03847	0	0	1	-360	360;
	6758	8066	0.00023	0.00194	0.05083	914.652037	914.652037	914.652037	0	0	1	-360	360;
	1148	2195	0.005669	0.040521	0.05699	523.567367	523.567367	523.567367	0	0	1	-360	360;
	1148	4354	0.001581	0.012159	0.04179	381.224556	381.224556	381.224556	0	0	1	-360	360;
	1148	4354	0.001409	0.013781	0.03723	403.623125	403.623125	403.623125	0	0	1	-360	360;
	1148	3185	0.00111	0.0075	0.01105	449.06767	449.06767	449.06767	0	0	1	-360	360;
	1148	502	0.0115	0.07662	0.11668	476.890121	476.890121	476.890121	0	0	1	-360	360;
	2014	2666	0.002661	0.01286	0.00142	79.415638	79.415638	79.415638	0	0	1	-360	360;
	1587	4265	0.010793	0.046942	0.01357	94.238529	94.238529	94.238529	0	0	1	-360	360;
	889	8358	0.023826	0.100992	0.01295	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8358	6904	0.014603	0.07986	0.0079	81.124462	81.124462	81.124462	0	0	1	-360	360;
	8358	4622	0.026132	0.099818	0.01495	72.646454	72.646454	72.646454	0	0	1	-360	360;
	7275	3535	0.002289	0.012554	0.00153	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7275	5680	0.00743	0.030719	0.00376	78.068051	78.068051	78.068051	0	0	1	-360	360;
	6854	3103	0.014653	0.029174	0.00366	69.935275	69.935275	69.935275	0	0	1	-360	360;
	5409	6854	0.016388	0.04286	0.00414	91.494008	91.494008	91.494008	0	0	1	-360	360;
	4126	8158	0.056248	0.122041	0.01373	57.157677	57.157677	57.157677	0	0	1	-360	360;
	309	8158	0.024901	0.049099	0.00639	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4218	6512	0.026554	0.061851	0.00654	69.398184	69.398184	69.398184	0	0	1	-360	360;
	3109	6113	0.073678	0.305793	0.02916	74.048342	74.048342	74.048342	0	0	1	-360	360;
	2814	3109	0.029083	0.060017	0.0056	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3109	57	0.021893	0.134521	0.01286	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1206	5040	0.018983	0.03914	0.00356	62.235184	62.235184	62.235184	0	0	1	-360	360;
	5040	5152	0.024529	0.041612	0.00471	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6420	4938	0.039421	0.085686	0.00949	69.514785	69.514785	69.514785	0	0	1	-360	360;
	6655	4938	0.039421	0.085769	0.00961	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6565	6420	0.00876	0.016901	0.00211	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3670	6565	0.006967	0.018248	0.00167	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7587	2504	0.025099	0.107901	0.00893	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3350	7587	0.02605	0.103438	0.00928	64.121958	64.121958	64.121958	0	0	1	-360	360;
	4728	2061	0.012719	0.033719	0.00305	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2061	7068	0.022273	0.043661	0.00532	71.110437	71.110437	71.110437	0	0	1	-360	360;
	7300	4747	0.00124	0.006653	0.00071	74.339465	74.339465	74.339465	0	0	1	-360	360;
	1974	4747	0.001587	0.007636	0.00092	91.79942	91.79942	91.79942	0	0	1	-360	360;
	5565	3443	0.006207	0.032331	0.00298	62.528593	62.528593	62.528593	0	0	1	-360	360;
	5565	3443	0.006083	0.032942	0.00292	81.025198	81.025198	81.025198	0	0	1	-360	360;
	3443	3410	0.001314	0.005446	0.00071	71.737457	71.737457	71.737457	0	0	1	-360	360;
	3893	391	0.003851	0.01662	0.00218	66.515913	66.515913	66.515913	0	0	1	-360	360;
	2732	7284	0.005124	0.031934	0.00301	89.833387	89.833387	89.833387	0	0	1	-360	360;
	8153	7284	0.005008	0.029562	0.00294	128.393099	128.393099	128.393099	0	0	1	-360	360;
	533	7409	0.048331	0.094372	0.01169	76.112496	76.112496	76.112496	0	0	1	-360	360;
	533	2358	0.015678	0.073826	0.00867	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2429	784	0.005579	0.019174	0.0013	90.521756	90.521756	90.521756	0	0	1	-360	360;
	309	2429	0.029397	0.077132	0.00735	74.573049	74.573049	74.573049	0	0	1	-360	360;
	1534	784	0.002901	0.006802	0.00074	77.579162	77.579162	77.579162	0	0	1	-360	360;
	1534	82	0.004769	0.020826	0.00279	95.95364	95.95364	95.95364	0	0	1	-360	360;
	1534	7899	0.031388	0.06305	0.00587	81.630879	81.630879	81.630879	0	0	1	-360	360;
	1534	3542	0.032587	0.068752	0.00783	85.689646	85.689646	85.689646	0	0	1	-360	360;
	1534	1231	0.042802	0.083983	0.00227	72.741527	72.741527	72.741527	0	0	1	-360	360;
	3499	3226	0.01214	0.069124	0.00681	131.74311	131.74311	131.74311	0	0	1	-360	360;
	2621	1714	0.008388	0.017033	0.0021	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8103	7934	0.008521	0.046653	0.00463	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8103	7710	0.003331	0.013818	0.00135	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2427	2195	0.016729	0.090401	0.16879	510.62725	510.62725	510.62725	0	0	1	-360	360;
	2427	5487	0.01526	0.081479	0.15891	512.371702	512.371702	512.371702	0	0	1	-360	360;
	2427	7753	0.00905	0.068219	0.09844	471.92693	471.92693	471.92693	0	0	1	-360	360;
	2427	1759	0.01205	0.08575	0.12153	493.960071	493.960071	493.960071	0	0	1	-360	360;
	2427	6321	0.007841	0.04113	0.07873	505.127919	505.127919	505.127919	0	0	1	-360	360;
	74	217	0.002413	0.010686	0.00111	57.157677	57.157677	57.157677	0	0	1	-360	360;
	799	74	0.005041	0.022769	0.00252	58.116973	58.116973	58.116973	0	0	1	-360	360;
	2167	31	0.00126	0.007531	0.01503	427.376331	427.376331	427.376331	0	0	1	-360	360;
	7019	1888	0.00176	0.007603	0.00107	70.112274	70.112274	70.112274	0	0	1	-360	360;
	7019	1888	0.00157	0.00862	0.00095	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8033	2319	0.02743	0.059157	0.00679	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1529	1315	0.028992	0.060893	0.00748	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3210	2191	0.012942	0.051851	0.00705	57.157677	57.157677	57.157677	0	0	1	-360	360;
	989	3210	0.071157	0.104149	0.00841	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1065	6072	0.002017	0.010686	0.00448	141.170126	141.170126	141.170126	0	0	1	-360	360;
	2858	6072	0.00195	0.009364	0.00112	102.824183	102.824183	102.824183	0	0	1	-360	360;
	2858	6072	0.001992	0.009174	0.00114	107.510732	107.510732	107.510732	0	0	1	-360	360;
	1206	519	0.027017	0.074521	0.00708	108.714473	108.714473	108.714473	0	0	1	-360	360;
	519	6420	0.025107	0.072339	0.00595	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3953	3394	0.002942	0.006248	0.00038	79.095555	79.095555	79.095555	0	0	1	-360	360;
	3394	3483	0.006983	0.028678	0.00385	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1714	5051	0.049636	0.077364	0.00984	60.776139	60.776139	60.776139	0	0	1	-360	360;
	2776	5051	0.071752	0.121198	0.01436	57.157677	57.157677	57.157677	0	0	1	-360	360;
	351	5006	0.00219	0.009174	0.0012	71.006791	71.006791	71.006791	0	0	1	-360	360;
	5515	351	0.018719	0.067983	0.00626	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8153	4691	0.006339	0.025397	0.0034	70.418639	70.418639	70.418639	0	0	1	-360	360;
	6568	18	0.011876	0.064967	0.00646	64.925595	64.925595	64.925595	0	0	1	-360	360;
	6113	3981	0.004603	0.024603	0.00251	71.164165	71.164165	71.164165	0	0	1	-360	360;
	1630	6113	0.037934	0.102868	0.01202	74.650403	74.650403	74.650403	0	0	1	-360	360;
	2726	6614	0.020818	0.067711	0.007	63.532853	63.532853	63.532853	0	0	1	-360	360;
	2514	1065	0.020223	0.048471	0.00379	72.643787	72.643787	72.643787	0	0	1	-360	360;
	1065	3410	0.00705	0.034314	0.00379	61.187483	61.187483	61.187483	0	0	1	-360	360;
	1065	6572	0.015388	0.074008	0.0083	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1065	6572	0.015066	0.075463	0.00813	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1206	29	0.060331	0.128752	0.01196	57.157677	57.157677	57.157677	0	0	1	-360	360;
	29	6368	0.037752	0.085843	0.00761	57.157677	57.157677	57.157677	0	0	1	-360	360;
	29	576	0.060471	0.109992	0.01196	61.976069	61.976069	61.976069	0	0	1	-360	360;
	5152	29	0.021959	0.050058	0.00446	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6420	1942	0.005347	0.026198	0.0031	67.869597	67.869597	67.869597	0	0	1	-360	360;
	3670	1942	0.005058	0.015769	0.00133	57.157677	57.157677	57.157677	0	0	1	-360	360;
	889	1010	0.010256	0.053223	0.00577	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3253	3981	0.01476	0.082157	0.00867	65.117645	65.117645	65.117645	0	0	1	-360	360;
	4218	1132	0.047843	0.112033	0.01135	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7153	3790	0.013909	0.036793	0.00341	63.195433	63.195433	63.195433	0	0	1	-360	360;
	3790	2814	0.030289	0.064818	0.00488	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3790	1001	0.026132	0.123132	0.01535	96.373368	96.373368	96.373368	0	0	1	-360	360;
	3790	1001	0.026628	0.120562	0.01565	104.075556	104.075556	104.075556	0	0	1	-360	360;
	7495	3790	0.016727	0.063959	0.00905	81.794922	81.794922	81.794922	0	0	1	-360	360;
	6031	4795	0.002413	0.010248	0.00142	76.080297	76.080297	76.080297	0	0	1	-360	360;
	6031	3812	0.009	0.040843	0.00582	61.259502	61.259502	61.259502	0	0	1	-360	360;
	391	7654	0.008421	0.018488	0.00215	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7719	7449	0.014917	0.034322	0.00306	73.223175	73.223175	73.223175	0	0	1	-360	360;
	7719	2528	0.00743	0.035942	0.00404	75.170728	75.170728	75.170728	0	0	1	-360	360;
	4795	7719	0.003851	0.016	0.0021	85.911227	85.911227	85.911227	0	0	1	-360	360;
	1449	7719	0.00162	0.008008	0.00225	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3158	7153	0.014372	0.045793	0.00401	58.747994	58.747994	58.747994	0	0	1	-360	360;
	4627	3158	0.038678	0.089967	0.01002	64.620183	64.620183	64.620183	0	0	1	-360	360;
	7070	4229	0.009504	0.021727	0.00765	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4229	4439	0.018364	0.096521	0.01002	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4573	5855	0.00195	0.010669	0.00106	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5855	5622	0.021182	0.046826	0.00513	70.474272	70.474272	70.474272	0	0	1	-360	360;
	6652	6904	0.031752	0.131826	0.01732	75.464899	75.464899	75.464899	0	0	1	-360	360;
	5006	4112	0.002917	0.01238	0.00151	108.045918	108.045918	108.045918	0	0	1	-360	360;
	18	1627	0.016157	0.035826	0.00307	83.580908	83.580908	83.580908	0	0	1	-360	360;
	18	2860	0.003727	0.020529	0.00201	57.157677	57.157677	57.157677	0	0	1	-360	360;
	18	5447	0.019992	0.079231	0.01163	63.573626	63.573626	63.573626	0	0	1	-360	360;
	2867	1627	0.023653	0.094587	0.01281	67.112829	67.112829	67.112829	0	0	1	-360	360;
	2860	1382	0.009322	0.049058	0.00512	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5409	1382	0.026727	0.146752	0.01447	72.424111	72.424111	72.424111	0	0	1	-360	360;
	4692	502	0.000531	0.002919	0.00513	335.986541	335.986541	335.986541	0	0	1	-360	360;
	517	5622	0.021992	0.059397	0.00535	80.966516	80.966516	80.966516	0	0	1	-360	360;
	517	4627	0.044413	0.10176	0.01167	72.780203	72.780203	72.780203	0	0	1	-360	360;
	3684	517	0.043223	0.088562	0.01045	71.328589	71.328589	71.328589	0	0	1	-360	360;
	3249	2983	0.006992	0.014066	0.0017	139.376899	139.376899	139.376899	0	0	1	-360	360;
	1548	4874	0.007256	0.028802	0.00407	158.663995	158.663995	158.663995	0	0	1	-360	360;
	1548	4874	0.006471	0.032636	0.00363	123.854398	123.854398	123.854398	0	0	1	-360	360;
	1548	8089	0.043198	0.156397	0.0216	99.809307	99.809307	99.809307	0	0	1	-360	360;
	7095	4570	0.016248	0.04014	0.00371	60.373558	60.373558	60.373558	0	0	1	-360	360;
	1309	1041	0.00206	0.01393	0.01952	461.880516	461.880516	461.880516	0	0	1	-360	360;
	1309	1041	0.002021	0.01419	0.01911	499.402625	499.402625	499.402625	0	0	1	-360	360;
	1309	31	0.00324	0.02076	0.03178	449.475013	449.475013	449.475013	0	0	1	-360	360;
	1309	31	0.00318	0.021159	0.03113	457.922537	457.922537	457.922537	0	0	1	-360	360;
	1206	6420	0.031603	0.070331	0.0078	71.665057	71.665057	71.665057	0	0	1	-360	360;
	1974	1206	0.003132	0.014554	0.00179	90.98378	90.98378	90.98378	0	0	1	-360	360;
	1375	1206	0.006669	0.032091	0.0039	75.184636	75.184636	75.184636	0	0	1	-360	360;
	4257	7934	0.008934	0.046298	0.00485	70.877615	70.877615	70.877615	0	0	1	-360	360;
	827	4833	0.002942	0.012008	0.00157	59.985648	59.985648	59.985648	0	0	1	-360	360;
	5115	3703	0.010942	0.062207	0.00637	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6532	5368	0.002529	0.011777	0.0021	58.79029	58.79029	58.79029	0	0	1	-360	360;
	6532	458	0.015107	0.042769	0.00546	76.878409	76.878409	76.878409	0	0	1	-360	360;
	7843	8197	0.00705	0.022099	0.00213	61.540718	61.540718	61.540718	0	0	1	-360	360;
	5505	3497	0.038231	0.105893	0.00988	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3574	5505	0.005901	0.021289	0.00191	88.846274	88.846274	88.846274	0	0	1	-360	360;
	103	4728	0.013636	0.022686	0.00266	64.832809	64.832809	64.832809	0	0	1	-360	360;
	6368	576	0.023826	0.059364	0.00486	62.390652	62.390652	62.390652	0	0	1	-360	360;
	6420	3203	0.016091	0.045967	0.00392	89.773181	89.773181	89.773181	0	0	1	-360	360;
	3203	3919	0.017033	0.053008	0.00523	125.92141	125.92141	125.92141	0	0	1	-360	360;
	6692	3497	0.017744	0.09686	0.00967	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5368	4623	0.003636	0.018818	0.00197	71.079762	71.079762	71.079762	0	0	1	-360	360;
	5368	4623	0.003562	0.019174	0.00178	70.250595	70.250595	70.250595	0	0	1	-360	360;
	5945	5088	0.008256	0.021909	0.00163	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5088	7070	0.026917	0.061636	0.00643	61.902716	61.902716	61.902716	0	0	1	-360	360;
	5202	4042	0.022702	0.057149	0.00679	57.213691	57.213691	57.213691	0	0	1	-360	360;
	4042	7498	0.023983	0.070124	0.00701	67.697933	67.697933	67.697933	0	0	1	-360	360;
	3288	2379	0.035231	0.078314	0.00994	75.763072	75.763072	75.763072	0	0	1	-360	360;
	2379	2776	0.02462	0.053603	0.00603	64.854339	64.854339	64.854339	0	0	1	-360	360;
	3535	4498	0.008091	0.03214	0.00444	57.157677	57.157677	57.157677	0	0	1	-360	360;
	377	3072	0.016793	0.081388	0.00913	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3072	1010	0.008636	0.047388	0.00467	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4708	6420	0.009364	0.018207	0.00227	59.146573	59.146573	59.146573	0	0	1	-360	360;
	6420	7408	0.023331	0.063727	0.00567	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6420	450	0.014983	0.033579	0.00263	68.147002	68.147002	68.147002	0	0	1	-360	360;
	2858	5565	0.004636	0.019678	0.00249	135.292411	135.292411	135.292411	0	0	1	-360	360;
	2858	5565	0.004727	0.019273	0.00253	115.155571	115.155571	115.155571	0	0	1	-360	360;
	2732	139	0.018446	0.089686	0.01008	88.802262	88.802262	88.802262	0	0	1	-360	360;
	8153	2732	0.011529	0.056587	0.00679	122.306378	122.306378	122.306378	0	0	1	-360	360;
	7337	5680	0.00862	0.021587	0.00246	81.898186	81.898186	81.898186	0	0	1	-360	360;
	7337	2884	0.005702	0.013017	0.00109	76.994629	76.994629	76.994629	0	0	1	-360	360;
	6468	583	0.048256	0.078099	0.00918	66.539919	66.539919	66.539919	0	0	1	-360	360;
	583	3913	0.065529	0.159802	0.01323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5475	583	0.02681	0.110702	0.01462	63.106076	63.106076	63.106076	0	0	1	-360	360;
	57	1630	0.04538	0.142983	0.01484	57.157677	57.157677	57.157677	0	0	1	-360	360;
	989	3005	0.024992	0.097777	0.01516	67.190754	67.190754	67.190754	0	0	1	-360	360;
	7968	3005	0.020256	0.109372	0.01102	66.176015	66.176015	66.176015	0	0	1	-360	360;
	3005	472	0.024636	0.095653	0.0134	73.941457	73.941457	73.941457	0	0	1	-360	360;
	2014	4622	0.003364	0.017669	0.00182	65.888702	65.888702	65.888702	0	0	1	-360	360;
	3336	8003	0.018463	0.039124	0.005	72.462597	72.462597	72.462597	0	0	1	-360	360;
	8003	1117	0.021636	0.060793	0.00546	62.890592	62.890592	62.890592	0	0	1	-360	360;
	6600	103	0.01786	0.039711	0.00348	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2195	3185	0.00588	0.03186	0.0588	274.905943	274.905943	274.905943	0	0	1	-360	360;
	5267	6652	0.045058	0.177512	0.02556	75.095089	75.095089	75.095089	0	0	1	-360	360;
	6652	2479	0.003686	0.017711	0.00202	69.176793	69.176793	69.176793	0	0	1	-360	360;
	6652	2479	0.003612	0.018058	0.00198	80.258333	80.258333	80.258333	0	0	1	-360	360;
	6652	2479	0.007876	0.021066	0.00185	66.344821	66.344821	66.344821	0	0	1	-360	360;
	2054	6652	0.03186	0.159347	0.01917	74.59515	74.59515	74.59515	0	0	1	-360	360;
	2776	3253	0.034901	0.071347	0.00845	71.907596	71.907596	71.907596	0	0	1	-360	360;
	1697	4874	0.006339	0.033281	0.00355	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1697	4874	0.006207	0.033917	0.00348	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3226	4874	0.006207	0.034983	0.00357	65.092496	65.092496	65.092496	0	0	1	-360	360;
	5917	8153	0.006289	0.024661	0.0034	85.903606	85.903606	85.903606	0	0	1	-360	360;
	5917	6890	0.002554	0.01376	0.00133	83.267494	83.267494	83.267494	0	0	1	-360	360;
	855	1315	0.003579	0.01486	0.00195	62.884114	62.884114	62.884114	0	0	1	-360	360;
	5070	1686	0.01495	0.02843	0.00223	87.696452	87.696452	87.696452	0	0	1	-360	360;
	4439	6685	0.011727	0.028934	0.00386	64.946744	64.946744	64.946744	0	0	1	-360	360;
	6685	7843	0.013843	0.048909	0.00445	63.163805	63.163805	63.163805	0	0	1	-360	360;
	1714	1686	0.032388	0.080711	0.00784	78.387371	78.387371	78.387371	0	0	1	-360	360;
	1714	1686	0.011719	0.074793	0.00718	87.972333	87.972333	87.972333	0	0	1	-360	360;
	1686	5515	0.00057	0.002612	0.00024	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2160	1155	0.012554	0.026314	0.00237	68.939017	68.939017	68.939017	0	0	1	-360	360;
	5409	2160	0.022554	0.049397	0.00465	57.157677	57.157677	57.157677	0	0	1	-360	360;
	855	2547	0.005438	0.021669	0.00296	65.31522	65.31522	65.31522	0	0	1	-360	360;
	1520	1372	0.00079	0.00721	0.16684	1149.67273	1149.67273	1149.67273	0	0	1	-360	360;
	1520	3627	0.00023	0.00228	0.05302	1120.03163	1120.03163	1120.03163	0	0	1	-360	360;
	7019	8297	0.002248	0.012083	0.00139	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7019	8297	0.002198	0.012314	0.00136	64.012597	64.012597	64.012597	0	0	1	-360	360;
	6655	7059	0.014727	0.064099	0.00876	82.875392	82.875392	82.875392	0	0	1	-360	360;
	217	6655	0.016364	0.040901	0.00395	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4241	6655	0.024471	0.063686	0.00546	67.912656	67.912656	67.912656	0	0	1	-360	360;
	7544	3537	0.00037	0.00429	0.08614	1196.95502	1196.95502	1196.95502	0	0	1	-360	360;
	8343	3537	0.00201	0.02375	0.4777	1301.0125	1301.0125	1301.0125	0	0	1	-360	360;
	3537	2671	9e-05	0.00093	0.01831	1421.58437	1421.58437	1421.58437	0	0	1	-360	360;
	3537	1743	9e-05	0.00095	0.01793	1437.98225	1437.98225	1437.98225	0	0	1	-360	360;
	5487	8296	0.004969	0.034979	0.05007	446.83471	446.83471	446.83471	0	0	1	-360	360;
	5487	8296	0.00468	0.033031	0.04667	434.264594	434.264594	434.264594	0	0	1	-360	360;
	8296	5836	0.00076	0.007909	0.01917	647.787002	647.787002	647.787002	0	0	1	-360	360;
	8296	4988	0.020539	0.080461	0.12379	494.392183	494.392183	494.392183	0	0	1	-360	360;
	1714	2528	0.024562	0.099231	0.01335	83.620919	83.620919	83.620919	0	0	1	-360	360;
	1714	2776	0.103397	0.231174	0.0206	72.37305	72.37305	72.37305	0	0	1	-360	360;
	1714	57	0.063843	0.141661	0.01882	68.102991	68.102991	68.102991	0	0	1	-360	360;
	7544	659	0.00059	0.00906	0.18397	1823.85632	1823.85632	1823.85632	0	0	1	-360	360;
	5487	5836	0.001979	0.02175	0.05121	225.666891	225.666891	225.666891	0	0	1	-360	360;
	5487	1924	0.005281	0.040039	0.05891	502.694526	502.694526	502.694526	0	0	1	-360	360;
	5487	502	0.012281	0.069219	0.12107	512.558417	512.558417	512.558417	0	0	1	-360	360;
	5487	6321	0.00668	0.04144	0.07014	535.256111	535.256111	535.256111	0	0	1	-360	360;
	5836	5587	0.00725	0.03801	0.07109	563.913827	563.913827	563.913827	0	0	1	-360	360;
	3627	1864	0.00225	0.02188	0.48605	1428.96519	1428.96519	1428.96519	0	0	1	-360	360;
	5451	1759	0.00525	0.03749	0.05254	528.586954	528.586954	528.586954	0	0	1	-360	360;
	5451	1759	0.00605	0.03069	0.06171	531.951255	531.951255	531.951255	0	0	1	-360	360;
	3336	2967	0.054793	0.13405	0.0107	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7059	3336	0.019099	0.076488	0.00907	87.050189	87.050189	87.050189	0	0	1	-360	360;
	3336	6230	0.035686	0.057678	0.00692	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6129	6398	0.00481	0.020579	0.00264	102.36997	102.36997	102.36997	0	0	1	-360	360;
	6129	6398	0.004909	0.020149	0.00269	62.519067	62.519067	62.519067	0	0	1	-360	360;
	6129	3919	0.010843	0.055512	0.00616	171.47303	171.47303	171.47303	0	0	1	-360	360;
	6129	3919	0.012289	0.054397	0.00698	171.47303	171.47303	171.47303	0	0	1	-360	360;
	450	6129	0.003562	0.018124	0.00409	61.812788	61.812788	61.812788	0	0	1	-360	360;
	7906	7449	0.038017	0.081934	0.00907	73.805612	73.805612	73.805612	0	0	1	-360	360;
	7906	1001	0.020149	0.083124	0.01091	68.172151	68.172151	68.172151	0	0	1	-360	360;
	7906	6106	0.009554	0.040959	0.00562	61.110321	61.110321	61.110321	0	0	1	-360	360;
	2547	7461	0.008331	0.043322	0.00454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7461	1810	0.02114	0.114479	0.01241	57.157677	57.157677	57.157677	0	0	1	-360	360;
	989	2967	0.017488	0.03838	0.00337	79.402301	79.402301	79.402301	0	0	1	-360	360;
	6398	3854	0.00381	0.019256	0.00207	83.920997	83.920997	83.920997	0	0	1	-360	360;
	3670	3854	0.001397	0.005802	0.00076	74.209907	74.209907	74.209907	0	0	1	-360	360;
	2073	5680	0.004719	0.024603	0.0025	81.041012	81.041012	81.041012	0	0	1	-360	360;
	2073	5680	0.00462	0.025058	0.00245	73.028458	73.028458	73.028458	0	0	1	-360	360;
	2884	5680	0.005413	0.027876	0.00294	71.093671	71.093671	71.093671	0	0	1	-360	360;
	5680	5267	0.032041	0.196413	0.01877	63.10036	63.10036	63.10036	0	0	1	-360	360;
	1800	5680	0.002198	0.022248	0.00362	128.199525	128.199525	128.199525	0	0	1	-360	360;
	5680	170	0.00262	0.018463	0.00426	152.788757	152.788757	152.788757	0	0	1	-360	360;
	5680	377	0.030347	0.112149	0.01322	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3483	6692	0.014512	0.05981	0.00795	59.468371	59.468371	59.468371	0	0	1	-360	360;
	1155	7500	0.005438	0.009124	0.00105	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8279	1196	0.005169	0.022219	0.04108	215.929128	215.929128	215.929128	0	0	1	-360	360;
	649	5928	0.007793	0.029769	0.00423	57.157677	57.157677	57.157677	0	0	1	-360	360;
	649	7248	0.006091	0.028653	0.00357	68.095179	68.095179	68.095179	0	0	1	-360	360;
	3812	4570	0.02876	0.060388	0.00667	88.349383	88.349383	88.349383	0	0	1	-360	360;
	1495	1748	0.02443	0.09576	0.01327	85.67326	85.67326	85.67326	0	0	1	-360	360;
	5409	1748	0.009083	0.03857	0.00493	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2776	690	0.048157	0.08462	0.00702	68.723151	68.723151	68.723151	0	0	1	-360	360;
	4498	2073	0.002512	0.013826	0.00133	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7500	5718	0.012298	0.026826	0.00232	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7620	7796	0.022744	0.133521	0.01347	61.933391	61.933391	61.933391	0	0	1	-360	360;
	4833	4439	0.022554	0.088876	0.01237	78.176841	78.176841	78.176841	0	0	1	-360	360;
	6468	7899	0.019231	0.033802	0.00386	72.759055	72.759055	72.759055	0	0	1	-360	360;
	8077	5945	0.010017	0.022711	0.00247	66.703771	66.703771	66.703771	0	0	1	-360	360;
	8077	458	0.00219	0.011843	0.00128	109.56898	109.56898	109.56898	0	0	1	-360	360;
	5202	5179	0.019281	0.064488	0.00546	70.648794	70.648794	70.648794	0	0	1	-360	360;
	2358	5179	0.02119	0.091727	0.01158	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3913	7986	0.003099	0.013893	0.0018	68.1752	68.1752	68.1752	0	0	1	-360	360;
	989	7986	0.004132	0.021959	0.00234	61.612736	61.612736	61.612736	0	0	1	-360	360;
	989	7986	0.004041	0.02238	0.00229	80.480105	80.480105	80.480105	0	0	1	-360	360;
	989	7968	0.002471	0.010901	0.0021	67.535224	67.535224	67.535224	0	0	1	-360	360;
	7055	8343	0.002	0.017	0.49082	1295.43904	1295.43904	1295.43904	0	0	1	-360	360;
	1091	8343	0.0016	0.01758	0.38197	1147.85615	1147.85615	1147.85615	0	0	1	-360	360;
	7620	299	0.027397	0.111917	0.01489	63.746242	63.746242	63.746242	0	0	1	-360	360;
	7620	6428	0.008752	0.044802	0.00478	80.124774	80.124774	80.124774	0	0	1	-360	360;
	649	5718	0.045727	0.07386	0.00879	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5267	7730	0.044281	0.119579	0.01069	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1231	7730	0.057612	0.11224	0.01489	65.261873	65.261873	65.261873	0	0	1	-360	360;
	3953	7730	0.063314	0.107273	0.00918	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3670	4708	0.006769	0.017909	0.00162	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6348	3812	0.003273	0.01157	0.0011	79.246832	79.246832	79.246832	0	0	1	-360	360;
	6348	489	0.000248	0.001331	0.00014	68.582734	68.582734	68.582734	0	0	1	-360	360;
	2479	2460	0.010603	0.045727	0.00579	101.117074	101.117074	101.117074	0	0	1	-360	360;
	2460	3953	0.016273	0.089678	0.0088	72.790301	72.790301	72.790301	0	0	1	-360	360;
	3542	6798	0.008397	0.022777	0.00204	72.484127	72.484127	72.484127	0	0	1	-360	360;
	6798	1289	0.013421	0.036314	0.00296	68.449747	68.449747	68.449747	0	0	1	-360	360;
	8153	3651	0.009628	0.050752	0.00524	63.688894	63.688894	63.688894	0	0	1	-360	360;
	5447	299	0.011777	0.046843	0.0064	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3410	299	0.014446	0.059967	0.00782	60.48311	60.48311	60.48311	0	0	1	-360	360;
	5267	2054	0.005174	0.029562	0.0036	71.351642	71.351642	71.351642	0	0	1	-360	360;
	7498	2428	0.069256	0.110017	0.01323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1366	1793	0.00524	0.022124	0.00291	61.27322	61.27322	61.27322	0	0	1	-360	360;
	4623	8129	0.005083	0.019421	0.00198	75.288663	75.288663	75.288663	0	0	1	-360	360;
	8129	1207	0.004099	0.013868	0.00154	60.799383	60.799383	60.799383	0	0	1	-360	360;
	8071	6600	0.04181	0.081397	0.00789	61.28427	61.28427	61.28427	0	0	1	-360	360;
	8071	6296	0.041446	0.067174	0.00799	63.029485	63.029485	63.029485	0	0	1	-360	360;
	31	2153	0.008841	0.0587	0.08921	484.907438	484.907438	484.907438	0	0	1	-360	360;
	1697	3226	0.011876	0.064	0.01206	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5027	2014	0.037579	0.182612	0.02042	62.655293	62.655293	62.655293	0	0	1	-360	360;
	139	6614	0.025752	0.129421	0.01394	104.206637	104.206637	104.206637	0	0	1	-360	360;
	1793	7495	0.00543	0.011421	0.00103	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7495	3574	0.003727	0.010066	0.00091	84.65433	84.65433	84.65433	0	0	1	-360	360;
	3919	6452	0.015868	0.053702	0.00512	70.622882	70.622882	70.622882	0	0	1	-360	360;
	1971	8071	0.026388	0.120603	0.01437	69.888025	69.888025	69.888025	0	0	1	-360	360;
	7300	1974	0.002372	0.010669	0.00138	119.689509	119.689509	119.689509	0	0	1	-360	360;
	4623	1207	0.001289	0.005041	0.00052	71.319634	71.319634	71.319634	0	0	1	-360	360;
	4623	3316	0.001438	0.006421	0.00079	83.225388	83.225388	83.225388	0	0	1	-360	360;
	4623	6288	0.001471	0.006289	0.0008	61.915101	61.915101	61.915101	0	0	1	-360	360;
	3627	1091	0.00105	0.01056	0.23998	1115.72516	1115.72516	1115.72516	0	0	1	-360	360;
	2596	31	0.01137	0.056789	0.11233	395.555129	395.555129	395.555129	0	0	1	-360	360;
	602	31	0.00519	0.028151	0.04155	492.20152	492.20152	492.20152	0	0	1	-360	360;
	1495	5409	0.030091	0.150033	0.01638	65.392764	65.392764	65.392764	0	0	1	-360	360;
	5475	83	0.011207	0.023851	0.00269	70.191151	70.191151	70.191151	0	0	1	-360	360;
	83	7751	0.073347	0.103488	0.01038	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4951	5778	0.064149	0.131041	0.01542	120.636611	120.636611	120.636611	0	0	1	-360	360;
	1422	1196	0.002341	0.009719	0.01369	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5058	254	0.051231	0.126554	0.01203	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1117	254	0.031967	0.079413	0.00783	75.20426	75.20426	75.20426	0	0	1	-360	360;
	1449	489	0.00024	0.001	0.00014	65.975391	65.975391	65.975391	0	0	1	-360	360;
	5053	2785	0.034008	0.069388	0.00823	87.072099	87.072099	87.072099	0	0	1	-360	360;
	472	936	0.003	0.01243	0.00172	73.796276	73.796276	73.796276	0	0	1	-360	360;
	936	6786	0.036281	0.073678	0.00711	75.888057	75.888057	75.888057	0	0	1	-360	360;
	1974	2943	0.003587	0.017231	0.00211	75.905204	75.905204	75.905204	0	0	1	-360	360;
	1664	6321	0.00012	0.00062	0.00111	308.706325	308.706325	308.706325	0	0	1	-360	360;
	6759	6106	0.006909	0.039231	0.00406	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5897	5898	0.003314	0.007017	0.00255	82.372595	82.372595	82.372595	0	0	1	-360	360;
	8089	8153	0.011289	0.063661	0.00613	69.206705	69.206705	69.206705	0	0	1	-360	360;
	8089	8153	0.01105	0.058942	0.00599	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6890	8153	0.007711	0.039413	0.00414	99.052348	99.052348	99.052348	0	0	1	-360	360;
	513	217	0.004017	0.013669	0.00132	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3515	1132	0.040521	0.08386	0.00967	75.404503	75.404503	75.404503	0	0	1	-360	360;
	2317	2422	0.024579	0.061231	0.00595	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2785	3684	0.067992	0.151463	0.01636	69.514785	69.514785	69.514785	0	0	1	-360	360;
	4562	7710	0.003579	0.011942	0.00144	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5897	380	0.00776	0.021587	0.00172	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8089	7339	0.003298	0.009463	0.00308	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7380	7339	0.005157	0.01376	0.00125	74.622014	74.622014	74.622014	0	0	1	-360	360;
	5495	7380	0.009934	0.034479	0.00427	74.629064	74.629064	74.629064	0	0	1	-360	360;
	3316	1375	0.002008	0.010802	0.00118	78.877213	78.877213	78.877213	0	0	1	-360	360;
	6288	2943	0.001967	0.011008	0.00116	57.157677	57.157677	57.157677	0	0	1	-360	360;
	592	5270	0.003219	0.01439	0.05131	254.361949	254.361949	254.361949	0	0	1	-360	360;
	592	1761	0.00239	0.01426	0.05989	240.649442	240.649442	240.649442	0	0	1	-360	360;
	7869	5914	0.00143	0.013591	0.0415	606.721117	606.721117	606.721117	0	0	1	-360	360;
	7869	6545	0.004669	0.05795	0.13416	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7869	1417	0.00412	0.043151	0.07423	628.09542	628.09542	628.09542	0	0	1	-360	360;
	7869	2843	0.000521	0.00612	0.02052	314.757799	314.757799	314.757799	0	0	1	-360	360;
	4943	4374	0.005849	0.02768	0.00824	278.91907	278.91907	278.91907	0	0	1	-360	360;
	4331	4943	0.012742	0.049311	0.01307	279.160691	279.160691	279.160691	0	0	1	-360	360;
	5814	6545	0.00206	0.01061	0.03339	368.383131	368.383131	368.383131	0	0	1	-360	360;
	5814	1830	0.001591	0.01	0.02584	229.462541	229.462541	229.462541	0	0	1	-360	360;
	1163	1421	0.00098	0.00874	0.26102	1561.41259	1561.41259	1561.41259	0	0	1	-360	360;
	1163	5017	0.00045	0.00548	0.11706	856.068155	856.068155	856.068155	0	0	1	-360	360;
	4330	5439	0.000719	0.003349	0.0115	262.990853	262.990853	262.990853	0	0	1	-360	360;
	4330	2055	0.000919	0.00687	0.0211	284.610173	284.610173	284.610173	0	0	1	-360	360;
	4330	5746	0.000219	0.0015	0.0058	288.601303	288.601303	288.601303	0	0	1	-360	360;
	2980	2377	0.00355	0.019669	0.02695	596.841603	596.841603	596.841603	0	0	1	-360	360;
	2980	682	0.00389	0.02201	0.0313	510.812059	510.812059	510.812059	0	0	1	-360	360;
	2980	6562	0.00251	0.01455	0.02093	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4022	4142	0.00825	0.036781	0.09258	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4022	3804	0.000581	0.00412	0.01069	600.961147	600.961147	600.961147	0	0	1	-360	360;
	6146	8085	0.00103	0.01236	0.26718	1493.19165	1493.19165	1493.19165	0	0	1	-360	360;
	6146	5017	0.0014	0.01768	0.36691	1373.40499	1373.40499	1373.40499	0	0	1	-360	360;
	4933	321	0.00088	0.01169	0.04491	287.849108	287.849108	287.849108	0	0	1	-360	360;
	3598	131	0.001711	0.01164	0.01869	629.653539	629.653539	629.653539	0	0	1	-360	360;
	3598	4055	0.00268	0.015031	0.02919	565.757734	565.757734	565.757734	0	0	1	-360	360;
	3598	8217	0.0018	0.01132	0.02123	574.654136	574.654136	574.654136	0	0	1	-360	360;
	1023	3836	0.00051	0.00268	0.11874	477.988692	477.988692	477.988692	0	0	1	-360	360;
	1023	2265	0.000651	0.003289	0.05377	423.314326	423.314326	423.314326	0	0	1	-360	360;
	3119	3826	0.0012	0.01137	0.31943	1754.40919	1754.40919	1754.40919	0	0	1	-360	360;
	3119	6046	0.00168	0.02275	0.45155	1747.4213	1747.4213	1747.4213	0	0	1	-360	360;
	4130	5439	0.002331	0.0225	0.06862	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4130	5439	0.002081	0.0255	0.06113	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4130	624	0.000781	0.004979	0.0118	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4130	1994	0.00514	0.06699	0.15273	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6775	3175	0.00088	0.00804	0.2287	1945.37202	1945.37202	1945.37202	0	0	1	-360	360;
	6775	2979	0.00032	0.00311	0.08594	1694.2029	1694.2029	1694.2029	0	0	1	-360	360;
	6775	198	0.00015	0.00197	0.0682	1027.06511	1027.06511	1027.06511	0	0	1	-360	360;
	6968	5270	0.004719	0.023831	0.07504	448.094084	448.094084	448.094084	0	0	1	-360	360;
	6968	2055	0.00669	0.029979	0.23504	461.993688	461.993688	461.993688	0	0	1	-360	360;
	8312	7607	0.001031	0.01225	0.08539	337.939428	337.939428	337.939428	0	0	1	-360	360;
	8312	7795	0.00105	0.00937	0.03855	555.216715	555.216715	555.216715	0	0	1	-360	360;
	8312	854	0.00201	0.01293	0.03199	406.708116	406.708116	406.708116	0	0	1	-360	360;
	1841	682	0.002	0.012229	0.02204	555.371803	555.371803	555.371803	0	0	1	-360	360;
	1841	1752	8.1e-05	0.000521	0.00167	401.483523	401.483523	401.483523	0	0	1	-360	360;
	1841	5914	0.00113	0.010969	0.03293	277.930346	277.930346	277.930346	0	0	1	-360	360;
	8085	7930	0.00195	0.02538	0.46869	1871.5197	1871.5197	1871.5197	0	0	1	-360	360;
	8085	2979	0.00164	0.01684	0.43143	1541.92259	1541.92259	1541.92259	0	0	1	-360	360;
	4208	5351	0.006289	0.06895	0.18349	327.946742	327.946742	327.946742	0	0	1	-360	360;
	4208	8161	0.001169	0.010229	0.02277	248.393164	248.393164	248.393164	0	0	1	-360	360;
	4208	5302	0.00805	0.060409	0.09432	233.075669	233.075669	233.075669	0	0	1	-360	360;
	4208	5302	0.00789	0.061581	0.09236	291.341823	291.341823	291.341823	0	0	1	-360	360;
	4208	6545	0.007159	0.04164	0.08886	540.540529	540.540529	540.540529	0	0	1	-360	360;
	4208	6545	0.007289	0.040729	0.09051	538.323192	538.323192	538.323192	0	0	1	-360	360;
	4208	6372	0.004531	0.02731	0.04807	313.792215	313.792215	313.792215	0	0	1	-360	360;
	4208	968	0.004521	0.033581	0.05003	535.224103	535.224103	535.224103	0	0	1	-360	360;
	7607	3095	0.0005	0.005081	0.01489	271.808759	271.808759	271.808759	0	0	1	-360	360;
	7607	736	0.00132	0.006521	0.0205	232.105131	232.105131	232.105131	0	0	1	-360	360;
	4823	5351	0.004401	0.030531	0.05633	321.780953	321.780953	321.780953	0	0	1	-360	360;
	4823	1417	0.000469	0.003409	0.00602	643.524564	643.524564	643.524564	0	0	1	-360	360;
	2767	3175	0.00239	0.02364	0.61786	1789.50331	1789.50331	1789.50331	0	0	1	-360	360;
	2767	3935	0.00034	0.0041	0.08806	1779.58784	1779.58784	1779.58784	0	0	1	-360	360;
	2767	6046	0.00027	0.00327	0.06986	1660.80162	1660.80162	1660.80162	0	0	1	-360	360;
	2767	6046	0.00026	0.00333	0.0684	1802.81302	1802.81302	1802.81302	0	0	1	-360	360;
	2767	2463	0.00059	0.00574	0.156	1570.31183	1570.31183	1570.31183	0	0	1	-360	360;
	926	7062	0.000789	0.003711	0.01196	286.46513	286.46513	286.46513	0	0	1	-360	360;
	926	6829	0.00419	0.04988	0.11703	219.752976	219.752976	219.752976	0	0	1	-360	360;
	926	4766	0.001651	0.010919	0.02403	219.571596	219.571596	219.571596	0	0	1	-360	360;
	2263	6372	0.007281	0.03162	0.08034	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2263	2081	0.006831	0.03505	0.0666	578.552289	578.552289	578.552289	0	0	1	-360	360;
	3175	5017	0.00108	0.01258	0.28231	1655.29331	1655.29331	1655.29331	0	0	1	-360	360;
	3175	3935	0.00174	0.0229	0.45216	1571.53473	1571.53473	1571.53473	0	0	1	-360	360;
	3175	2463	0.00223	0.02125	0.57705	1704.15523	1704.15523	1704.15523	0	0	1	-360	360;
	5270	5439	0.0023	0.014081	0.03657	369.777398	369.777398	369.777398	0	0	1	-360	360;
	5270	1761	0.00213	0.026831	0.0619	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5270	5259	0.001039	0.00549	0.0166	282.818851	282.818851	282.818851	0	0	1	-360	360;
	5270	7502	0.00301	0.016711	0.04538	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5270	1279	0.000849	0.005229	0.01351	355.522273	355.522273	355.522273	0	0	1	-360	360;
	5270	1461	0.002781	0.020281	0.0376	598.577672	598.577672	598.577672	0	0	1	-360	360;
	5270	854	0.002159	0.01389	0.03437	322.978978	322.978978	322.978978	0	0	1	-360	360;
	135	7378	0.003502	0.031631	0.00912	332.210809	332.210809	332.210809	0	0	1	-360	360;
	4374	6353	0.003929	0.024729	0.01023	271.951029	271.951029	271.951029	0	0	1	-360	360;
	6353	135	0.00076	0.005609	0.00197	262.590161	262.590161	262.590161	0	0	1	-360	360;
	2377	3836	0.002039	0.01249	0.01547	539.720126	539.720126	539.720126	0	0	1	-360	360;
	4492	7930	0.00036	0.00572	0.15238	1958.41779	1958.41779	1958.41779	0	0	1	-360	360;
	4492	1039	0.00028	0.00448	0.10868	1905.37117	1905.37117	1905.37117	0	0	1	-360	360;
	2715	5351	0.001091	0.01663	0.05227	474.976482	474.976482	474.976482	0	0	1	-360	360;
	2715	6372	0.011909	0.080909	0.1231	482.510245	482.510245	482.510245	0	0	1	-360	360;
	2715	131	0.000401	0.00232	0.00894	392.08299	392.08299	392.08299	0	0	1	-360	360;
	2715	131	0.000349	0.00263	0.00797	354.583363	354.583363	354.583363	0	0	1	-360	360;
	1353	5420	0.000531	0.0063	0.02412	307.666437	307.666437	307.666437	0	0	1	-360	360;
	1353	419	0.005531	0.02938	0.05367	534.532114	534.532114	534.532114	0	0	1	-360	360;
	1353	7165	0.00019	0.002521	0.00785	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5420	5351	0.00075	0.010909	0.03263	511.465943	511.465943	511.465943	0	0	1	-360	360;
	7579	3804	0.007	0.03088	0.07578	644.41508	644.41508	644.41508	0	0	1	-360	360;
	7579	4055	0.002349	0.009281	0.02529	569.47565	569.47565	569.47565	0	0	1	-360	360;
	5351	4751	0.003219	0.02125	0.04809	291.559784	291.559784	291.559784	0	0	1	-360	360;
	5351	4751	0.004461	0.027961	0.05305	531.340048	531.340048	531.340048	0	0	1	-360	360;
	4761	3901	0.00105	0.007901	0.02657	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6095	4142	0.002461	0.01019	0.02723	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6095	3804	0.00605	0.028409	0.06875	536.893488	536.893488	536.893488	0	0	1	-360	360;
	1912	854	7e-05	0.00039	0.00104	252.383151	252.383151	252.383151	0	0	1	-360	360;
	5439	1761	0.0047	0.02175	0.09382	503.639914	503.639914	503.639914	0	0	1	-360	360;
	6762	3954	0.00118	0.00719	0.01343	605.0041	605.0041	605.0041	0	0	1	-360	360;
	6762	206	0.002349	0.01887	0.02665	345.582934	345.582934	345.582934	0	0	1	-360	360;
	1914	1446	0.00104	0.01314	0.27809	1906.38938	1906.38938	1906.38938	0	0	1	-360	360;
	1914	6046	0.0004	0.004	0.10249	2006.47081	2006.47081	2006.47081	0	0	1	-360	360;
	8161	6060	0.001289	0.006229	0.01022	482.011449	482.011449	482.011449	0	0	1	-360	360;
	8161	3246	0.00555	0.035419	0.04153	451.028559	451.028559	451.028559	0	0	1	-360	360;
	8161	3246	0.00381	0.0328	0.04063	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4142	5356	0.00093	0.006841	0.02087	251.965519	251.965519	251.965519	0	0	1	-360	360;
	4142	5356	0.000909	0.006979	0.02045	248.63818	248.63818	248.63818	0	0	1	-360	360;
	2795	5064	0.001091	0.0082	0.01442	523.098293	523.098293	523.098293	0	0	1	-360	360;
	2795	4316	0.002081	0.020419	0.04497	508.625588	508.625588	508.625588	0	0	1	-360	360;
	3901	3118	0.00082	0.00276	0.00857	237.060321	237.060321	237.060321	0	0	1	-360	360;
	3901	6372	0.00162	0.007021	0.01655	251.328782	251.328782	251.328782	0	0	1	-360	360;
	7143	5302	0.002831	0.015021	0.02115	573.192042	573.192042	573.192042	0	0	1	-360	360;
	7143	119	0.006781	0.03026	0.05171	515.199483	515.199483	515.199483	0	0	1	-360	360;
	1421	2979	0.00041	0.00523	0.11604	1801.45453	1801.45453	1801.45453	0	0	1	-360	360;
	1421	3826	0.00062	0.0098	0.21409	1839.2281	1839.2281	1839.2281	0	0	1	-360	360;
	1421	3826	0.00083	0.00975	0.21448	1857.40504	1857.40504	1857.40504	0	0	1	-360	360;
	3246	7795	0.00131	0.01564	0.05189	275.592597	275.592597	275.592597	0	0	1	-360	360;
	3246	6545	0.002039	0.026831	0.07657	232.284987	232.284987	232.284987	0	0	1	-360	360;
	3246	4316	0.001031	0.00945	0.02885	339.193087	339.193087	339.193087	0	0	1	-360	360;
	3246	3406	0.001789	0.015211	0.04788	275.126952	275.126952	275.126952	0	0	1	-360	360;
	1761	6545	0.003021	0.03739	0.10346	232.8478	232.8478	232.8478	0	0	1	-360	360;
	1761	343	0.004211	0.04031	0.12272	305.323353	305.323353	305.323353	0	0	1	-360	360;
	1761	3229	0.004979	0.029409	0.08011	614.022057	614.022057	614.022057	0	0	1	-360	360;
	1761	3229	0.00488	0.029979	0.07848	353.703897	353.703897	353.703897	0	0	1	-360	360;
	1761	2173	0.002211	0.02399	0.06523	251.11044	251.11044	251.11044	0	0	1	-360	360;
	1761	2173	0.00231	0.023021	0.06769	230.077177	230.077177	230.077177	0	0	1	-360	360;
	419	6372	0.00719	0.046841	0.07983	560.755294	560.755294	560.755294	0	0	1	-360	360;
	1113	5302	0.00155	0.00968	0.02973	484.421598	484.421598	484.421598	0	0	1	-360	360;
	1113	5302	0.0015	0.01051	0.02831	270.95025	270.95025	270.95025	0	0	1	-360	360;
	1113	2766	0.00169	0.013401	0.01813	259.96264	259.96264	259.96264	0	0	1	-360	360;
	1113	2314	0.001659	0.012409	0.01774	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2462	6372	0.005091	0.0167	0.05619	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2462	2081	0.009919	0.04649	0.10071	568.75051	568.75051	568.75051	0	0	1	-360	360;
	1876	4766	9.1e-05	0.000531	0.00124	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5302	6545	0.00982	0.045331	0.10057	556.001681	556.001681	556.001681	0	0	1	-360	360;
	5302	8320	0.003039	0.02964	0.09625	216.573485	216.573485	216.573485	0	0	1	-360	360;
	1752	3621	0.004159	0.02051	0.0658	390.493245	390.493245	390.493245	0	0	1	-360	360;
	3118	6372	0.00087	0.00469	0.01383	245.892706	245.892706	245.892706	0	0	1	-360	360;
	5064	526	0.0017	0.021039	0.04746	417.060514	417.060514	417.060514	0	0	1	-360	360;
	2979	198	0.00019	0.0032	0.08677	911.329548	911.329548	911.329548	0	0	1	-360	360;
	2254	2655	0.006979	0.03975	0.10729	506.157519	506.157519	506.157519	0	0	1	-360	360;
	2254	526	0.000521	0.004081	0.0572	328.735518	328.735518	328.735518	0	0	1	-360	360;
	2254	968	0.00382	0.02214	0.0701	557.93361	557.93361	557.93361	0	0	1	-360	360;
	1246	3826	0.00204	0.02487	0.52209	1688.6584	1688.6584	1688.6584	0	0	1	-360	360;
	5914	3836	0.00312	0.018469	0.03241	573.969768	573.969768	573.969768	0	0	1	-360	360;
	2655	6372	0.005479	0.038841	0.06048	554.350205	554.350205	554.350205	0	0	1	-360	360;
	119	6545	0.000789	0.007729	0.03256	289.582129	289.582129	289.582129	0	0	1	-360	360;
	5244	6562	0.001789	0.01618	0.02179	586.898453	586.898453	586.898453	0	0	1	-360	360;
	5244	6562	0.00175	0.016479	0.02133	566.512215	566.512215	566.512215	0	0	1	-360	360;
	6562	3621	0.00375	0.025099	0.05933	306.419256	306.419256	306.419256	0	0	1	-360	360;
	1472	5835	0.010778	0.035609	0	196.218667	196.218667	196.218667	0	0	1	-360	360;
	5835	1472	0.0096	0.04036	0	195.501598	195.501598	195.501598	0	0	1	-360	360;
	7201	3804	0.001031	0.016919	0.04358	292.691506	292.691506	292.691506	0	0	1	-360	360;
	7201	3804	0.00101	0.01725	0.04268	229.717084	229.717084	229.717084	0	0	1	-360	360;
	6545	4438	0.00124	0.013169	0.03573	278.910028	278.910028	278.910028	0	0	1	-360	360;
	6545	4438	0.00119	0.01356	0.03384	606.962322	606.962322	606.962322	0	0	1	-360	360;
	6545	3406	0.00119	0.013281	0.03282	282.908779	282.908779	282.908779	0	0	1	-360	360;
	3954	6770	0.00306	0.02405	0.03112	351.538383	351.538383	351.538383	0	0	1	-360	360;
	3954	4751	0.002469	0.018081	0.02783	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7015	5681	0.0018	0.014651	0.04525	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7015	2173	0.001081	0.010219	0.03127	245.254826	245.254826	245.254826	0	0	1	-360	360;
	5681	6829	0.000669	0.003521	0.00934	319.23858	319.23858	319.23858	0	0	1	-360	360;
	6372	321	0.001039	0.0142	0.05059	371.401438	371.401438	371.401438	0	0	1	-360	360;
	6372	1637	0.002771	0.019919	0.05964	235.887445	235.887445	235.887445	0	0	1	-360	360;
	3095	4438	0.00295	0.013401	0.02644	221.182299	221.182299	221.182299	0	0	1	-360	360;
	4438	2055	0.00411	0.02488	0.15424	224.066857	224.066857	224.066857	0	0	1	-360	360;
	4438	7694	0.00061	0.007211	0.00216	215.929128	215.929128	215.929128	0	0	1	-360	360;
	968	736	0.002219	0.011081	0.03524	241.588352	241.588352	241.588352	0	0	1	-360	360;
	6488	2692	6e-05	0.00066	0.01637	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3836	2314	0.00411	0.026081	0.03724	647.787002	647.787002	647.787002	0	0	1	-360	360;
	8381	6046	0.00168	0.02222	0.41917	1557.84328	1557.84328	1557.84328	0	0	1	-360	360;
	8381	2692	0.00049	0.0061	0.13519	1846.04947	1846.04947	1846.04947	0	0	1	-360	360;
	2055	5746	0.00082	0.00575	0.01892	332.007224	332.007224	332.007224	0	0	1	-360	360;
	7165	2081	0.001409	0.01918	0.06003	268.796168	268.796168	268.796168	0	0	1	-360	360;
	5259	7502	0.001841	0.011159	0.0291	228.849049	228.849049	228.849049	0	0	1	-360	360;
	3714	2692	0.00109	0.01044	0.29924	2080.51203	2080.51203	2080.51203	0	0	1	-360	360;
	4394	1039	0.00088	0.01608	0.35869	1936.11736	1936.11736	1936.11736	0	0	1	-360	360;
	3804	3979	0.000591	0.007781	0.0252	282.067418	282.067418	282.067418	0	0	1	-360	360;
	3804	2809	0.001901	0.0247	0.0804	295.424786	295.424786	295.424786	0	0	1	-360	360;
	7062	6184	0.00382	0.018401	0.0563	278.729791	278.729791	278.729791	0	0	1	-360	360;
	6829	6184	0.004169	0.02775	0.06151	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6829	2173	0.001849	0.01889	0.05159	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2081	2809	0.00076	0.01036	0.03222	285.671019	285.671019	285.671019	0	0	1	-360	360;
	1830	624	0.00207	0.009849	0.03353	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7089	343	0.001849	0.013841	0.04527	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7089	343	0.00243	0.015771	0.03613	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6046	1507	0.00128	0.01149	0.31595	1705.40577	1705.40577	1705.40577	0	0	1	-360	360;
	1994	654	0.00057	0.005961	0.01659	300.24013	300.24013	300.24013	0	0	1	-360	360;
	1994	5785	0.00218	0.015099	0.02858	280.854152	280.854152	280.854152	0	0	1	-360	360;
	343	4766	0.00131	0.008771	0.02664	559.075239	559.075239	559.075239	0	0	1	-360	360;
	654	5785	0.00163	0.008729	0.0132	262.250471	262.250471	262.250471	0	0	1	-360	360;
	1279	4272	0.004831	0.02588	0.06123	537.49936	537.49936	537.49936	0	0	1	-360	360;
	206	2843	0.00062	0.00857	0.02449	270.033822	270.033822	270.033822	0	0	1	-360	360;
	3979	5356	0.002281	0.0362	0.10585	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1637	8320	0.00132	0.01036	0.02782	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4272	2173	0.00281	0.017099	0.04517	245.630543	245.630543	245.630543	0	0	1	-360	360;
	2173	1461	0.00424	0.0258	0.06757	215.929128	215.929128	215.929128	0	0	1	-360	360;
	645	7216	0.00036	0.00313	0.09213	1753.28371	1753.28371	1753.28371	0	0	1	-360	360;
	5524	8228	0.00112	0.0114	0.29718	1679.86775	1679.86775	1679.86775	0	0	1	-360	360;
	5524	5435	0.00046	0.007	0.14613	1069.03458	1069.03458	1069.03458	0	0	1	-360	360;
	5524	334	0.00113	0.01516	0.40614	1460.75526	1460.75526	1460.75526	0	0	1	-360	360;
	8228	1961	0.00138	0.01478	0.36973	1203.33343	1203.33343	1203.33343	0	0	1	-360	360;
	8228	469	0.00107	0.01447	0.34246	1457.16555	1457.16555	1457.16555	0	0	1	-360	360;
	5535	1261	0.012651	0.08939	0.12174	353.432207	353.432207	353.432207	0	0	1	-360	360;
	5535	5458	0.00957	0.05424	0.08491	333.415589	333.415589	333.415589	0	0	1	-360	360;
	1190	3131	0.0017	0.01823	0.50667	1014.44321	1014.44321	1014.44321	0	0	1	-360	360;
	1190	5452	0.00092	0.00892	0.29068	1293.34998	1293.34998	1293.34998	0	0	1	-360	360;
	1190	1724	0.00063	0.00829	0.20794	1263.98136	1263.98136	1263.98136	0	0	1	-360	360;
	1961	7917	0.00052	0.00757	0.18021	930.733991	930.733991	930.733991	0	0	1	-360	360;
	7917	5957	0.00189	0.02546	0.71464	1013.67314	1013.67314	1013.67314	0	0	1	-360	360;
	7917	3822	0.00044	0.00602	0.1561	1011.82037	1011.82037	1011.82037	0	0	1	-360	360;
	128	5098	0.00138	0.01477	0.50511	1791.03885	1791.03885	1791.03885	0	0	1	-360	360;
	5098	5868	0.00055	0.00597	0.18674	1650.4162	1650.4162	1650.4162	0	0	1	-360	360;
	334	3465	0.00047	0.00684	0.15672	1629.43806	1629.43806	1629.43806	0	0	1	-360	360;
	1261	3067	0.0038	0.026331	0.0334	483.546704	483.546704	483.546704	0	0	1	-360	360;
	1961	2235	0.001	0.00937	0.2598	836.812462	836.812462	836.812462	0	0	1	-360	360;
	1961	334	0.00203	0.01819	0.54116	1635.92902	1635.92902	1635.92902	0	0	1	-360	360;
	5458	5783	0.002419	0.015021	0.02221	426.437802	426.437802	426.437802	0	0	1	-360	360;
	5458	5783	0.00274	0.014719	0.02517	375.410477	375.410477	375.410477	0	0	1	-360	360;
	5458	3274	0.014599	0.077461	0.10083	333.570296	333.570296	333.570296	0	0	1	-360	360;
	5458	6447	0.012461	0.05613	0.08607	380.952866	380.952866	380.952866	0	0	1	-360	360;
	5458	3067	0.014581	0.083849	0.09574	357.852401	357.852401	357.852401	0	0	1	-360	360;
	6453	5957	0.0006	0.00665	0.21174	1576.88178	1576.88178	1576.88178	0	0	1	-360	360;
	6453	5957	0.00057	0.00856	0.24311	1321.73726	1321.73726	1321.73726	0	0	1	-360	360;
	6453	5452	0.00234	0.02648	0.89205	1998.47919	1998.47919	1998.47919	0	0	1	-360	360;
	2235	3571	0.00019	0.00167	0.05205	913.573282	913.573282	913.573282	0	0	1	-360	360;
	2235	1913	0.00019	0.00164	0.05303	838.310478	838.310478	838.310478	0	0	1	-360	360;
	5452	7505	3e-05	0.00047	0.01248	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5452	764	3e-05	0.00048	0.01222	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5673	3686	0.00097	0.01084	0.36793	982.583382	982.583382	982.583382	0	0	1	-360	360;
	5452	5310	8e-05	0.00073	0.01727	1103.74629	1103.74629	1103.74629	0	0	1	-360	360;
	5452	4521	8e-05	0.0006	0.02184	1030.1658	1030.1658	1030.1658	0	0	1	-360	360;
	5452	1419	7e-05	0.00067	0.02103	1029.54184	1029.54184	1029.54184	0	0	1	-360	360;
	5452	648	7e-05	0.00066	0.02449	1062.39026	1062.39026	1062.39026	0	0	1	-360	360;
	3274	507	0.00145	0.00707	0.00958	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3274	3079	0.00164	0.00693	0.01086	348.049097	348.049097	348.049097	0	0	1	-360	360;
	334	6541	8e-05	0.00084	0.02079	1144.4481	1144.4481	1144.4481	0	0	1	-360	360;
	334	4914	0.0001	0.00083	0.02602	1152.48052	1152.48052	1152.48052	0	0	1	-360	360;
	334	7391	0.0001	0.00081	0.027	1126.74308	1126.74308	1126.74308	0	0	1	-360	360;
	2990	5777	0.00019	0.00215	0.05848	767.876077	767.876077	767.876077	0	0	1	-360	360;
	128	3561	0.00037	0.00292	0.13501	1150.87062	1150.87062	1150.87062	0	0	1	-360	360;
	128	1901	0.00033	0.00332	0.12028	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3230	3712	0.002031	0.01382	0.01998	473.144388	473.144388	473.144388	0	0	1	-360	360;
	3230	2347	0.001979	0.014091	0.01956	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5957	2702	4e-05	0.00052	0.01872	1854.27474	1854.27474	1854.27474	0	0	1	-360	360;
	5957	1036	4e-05	0.00054	0.01872	1701.77262	1701.77262	1701.77262	0	0	1	-360	360;
	3230	6686	0.00136	0.0067	0.00955	303.692073	303.692073	303.692073	0	0	1	-360	360;
	2990	4007	0.0001	0.00078	0.02402	1143.62867	1143.62867	1143.62867	0	0	1	-360	360;
	2990	1464	0.00011	0.00076	0.02449	1098.62237	1098.62237	1098.62237	0	0	1	-360	360;
	3131	2990	0.00376	0.04872	1.06146	1461.24033	1461.24033	1461.24033	0	0	1	-360	360;
	2990	5336	0.0002	0.00258	0.07285	1186.48733	1186.48733	1186.48733	0	0	1	-360	360;
	2990	334	0.00089	0.0087	0.25229	1098.76717	1098.76717	1098.76717	0	0	1	-360	360;
	265	3230	0.01282	0.053159	0.08859	242.121061	242.121061	242.121061	0	0	1	-360	360;
	265	6343	0.002219	0.01038	0.02509	463.580004	463.580004	463.580004	0	0	1	-360	360;
	265	3230	0.01307	0.05205	0.09029	250.290799	250.290799	250.290799	0	0	1	-360	360;
	265	3067	0.00449	0.01863	0.05085	476.198514	476.198514	476.198514	0	0	1	-360	360;
	128	3822	0.00215	0.0248	0.72596	1754.79752	1754.79752	1754.79752	0	0	1	-360	360;
	2141	5673	0.00192	0.02113	0.49768	1082.01585	1082.01585	1082.01585	0	0	1	-360	360;
	2141	7216	0.00165	0.01704	0.58637	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5957	5868	0.00168	0.02469	0.56593	1078.12074	1078.12074	1078.12074	0	0	1	-360	360;
	5957	3822	0.00222	0.02324	0.95146	1290.22363	1290.22363	1290.22363	0	0	1	-360	360;
	5673	469	0.00173	0.01682	0.44602	1133.44795	1133.44795	1133.44795	0	0	1	-360	360;
	5673	4855	0.00174	0.01872	0.50096	1349.32352	1349.32352	1349.32352	0	0	1	-360	360;
	5673	2235	0.00194	0.02194	0.48735	1248.14952	1248.14952	1248.14952	0	0	1	-360	360;
	8376	3033	0.013831	0.056841	0.10506	373.062821	373.062821	373.062821	0	0	1	-360	360;
	8376	3033	0.014091	0.055599	0.10701	357.945377	357.945377	357.945377	0	0	1	-360	360;
	3274	182	0.009081	0.05263	0.08773	417.38898	417.38898	417.38898	0	0	1	-360	360;
	3274	4961	0.007401	0.05661	0.06824	436.771529	436.771529	436.771529	0	0	1	-360	360;
	7216	4855	0.00164	0.01627	0.40651	1412.96222	1412.96222	1412.96222	0	0	1	-360	360;
	6447	2345	0.022909	0.124591	0.15812	393.200613	393.200613	393.200613	0	0	1	-360	360;
	2717	4855	0.00074	0.00791	0.18963	1462.98122	1462.98122	1462.98122	0	0	1	-360	360;
	2717	1724	0.00136	0.01726	0.35698	1177.42091	1177.42091	1177.42091	0	0	1	-360	360;
	182	3230	0.011211	0.07026	0.1091	388.970945	388.970945	388.970945	0	0	1	-360	360;
	182	3230	0.01143	0.06882	0.11124	376.785691	376.785691	376.785691	0	0	1	-360	360;
	4855	3822	0.00544	0.06455	1.74389	909.884186	909.884186	909.884186	0	0	1	-360	360;
	3033	2345	0.016289	0.066719	0.115	388.317824	388.317824	388.317824	0	0	1	-360	360;
	3033	2345	0.016591	0.065271	0.11713	394.349102	394.349102	394.349102	0	0	1	-360	360;
	5452	1724	0.00077	0.0096	0.29761	1359.4687	1359.4687	1359.4687	0	0	1	-360	360;
	5452	1724	0.00093	0.00905	0.35052	1319.79299	1319.79299	1319.79299	0	0	1	-360	360;
	2345	4961	0.021729	0.15268	0.20206	351.264788	351.264788	351.264788	0	0	1	-360	360;
	441	6343	0.00064	0.002919	0.02343	218.455497	218.455497	218.455497	0	0	1	-360	360;
	6343	3067	0.002211	0.009169	0.02004	472.866221	472.866221	472.866221	0	0	1	-360	360;
	3738	8250	0.00014	0.001729	0.0046	350.454673	350.454673	350.454673	0	0	1	-360	360;
	4568	8250	0.0002	0.00281	0.00707	325.715688	325.715688	325.715688	0	0	1	-360	360;
	2409	5005	0.000711	0.004289	0.00743	639.1661	639.1661	639.1661	0	0	1	-360	360;
	2409	6334	0.0007	0.00438	0.00728	609.694078	609.694078	609.694078	0	0	1	-360	360;
	3738	2409	0.00032	0.00437	0.01102	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5005	5250	7e-05	0.000401	0.00066	628.208211	628.208211	628.208211	0	0	1	-360	360;
	6334	7731	7e-05	0.000401	0.00065	584.009323	584.009323	584.009323	0	0	1	-360	360;
	742	5250	0.00189	0.0083	0.01824	635.553354	635.553354	635.553354	0	0	1	-360	360;
	742	7731	0.00168	0.009401	0.01625	577.875542	577.875542	577.875542	0	0	1	-360	360;
	3738	6523	0.000151	0.00169	0.00521	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5657	6716	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5657	6716	0.001	0.120531	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	5657	4.583678	10.16446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5657	5156	8.430371	11.9126	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5657	2624	8.911157	11.64008	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	5657	7.455785	17.24773	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	5657	9.746074	17.22748	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	5657	0.796777	2.654959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	5657	0.960248	2.559091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	5657	5.215909	9.009917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	5657	5.097107	9.164256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5657	19.22748	20.93388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	5657	17.18905	24.75	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5657	2684	17.2064	22.05785	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	5657	0.227459	1.377252	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	5657	0.13382	0.643926	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	5657	0.211219	1.206674	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	5657	0.230165	1.120413	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	5657	29.19835	48.54959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	5657	0.13231	0.939504	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	5657	0.842851	2.338843	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	5657	0.820764	2.471488	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5657	0.553244	1.973719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5657	0.01405	0.0548	0.1144	431.034423	431.034423	431.034423	0	0	1	-360	360;
	5657	1441	1.071157	2.739256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5657	1441	0.00774	0.04075	0.05749	455.789794	455.789794	455.789794	0	0	1	-360	360;
	6035	5657	16.64566	21.19835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	5657	16.22913	16.64318	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	6716	10.87128	19.07149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6716	5156	15.62583	25.67355	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6716	2624	16.95331	27.19835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	6716	16.6593	34.37603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	6716	19.6814	37.93595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	6716	1.881012	4.967975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	6716	1.902851	5.765909	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	6716	10.14587	20.68595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	6716	10.93058	18.98595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	6716	36.07851	50.20248	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	6716	40.72727	46.32438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6716	2684	36.24587	46.68595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	6716	0.444855	2.8	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	6716	0.273946	1.391054	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	6716	0.514545	2.160744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	6716	0.450992	2.507851	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	6716	0.292665	1.871012	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	6716	1.795062	4.826033	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	6716	1.551219	5.755991	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	6716	1.089959	4.426653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6716	1441	0.00757	0.037729	0.05624	498.202314	498.202314	498.202314	0	0	1	-360	360;
	6716	1441	2.158264	6.029752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6035	6716	35.04132	44.83471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	6716	30.36983	39.79132	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4904	3946	0.00058	0.00554	0.18339	1323.95137	1323.95137	1323.95137	0	0	1	-360	360;
	4904	5388	0.00116	0.01169	0.35768	1394.35484	1394.35484	1394.35484	0	0	1	-360	360;
	5701	3946	0.00052	0.00627	0.16339	1471.83768	1471.83768	1471.83768	0	0	1	-360	360;
	5701	5388	0.00118	0.01144	0.36442	1454.46043	1454.46043	1454.46043	0	0	1	-360	360;
	4904	5701	0.001	0.223012	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4904	3946	0.05072	0.617493	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4904	5388	0.05373	0.600208	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	292	4904	0.381143	1.437258	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5701	3946	0.04991	0.630229	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5701	5388	0.05131	0.624993	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	292	5701	0.389938	1.410526	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4088	4710	0.00059	0.00463	0.20204	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4088	1454	0.00047	0.0038	0.15935	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	5691	6935	0.0006	0.00453	0.20578	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	5691	4710	0.03442	0.298892	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5691	6935	0.03377	0.305229	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5691	6000	0.00047	0.00372	0.16236	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4068	544	0.00036	0.00341	0.11612	1204.80841	1204.80841	1204.80841	0	0	1	-360	360;
	3665	4068	0.00164	0.01639	0.51589	1213.85508	1213.85508	1213.85508	0	0	1	-360	360;
	7063	5636	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	5636	0.001	0.13669	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	6996	2.103099	5.086777	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	6527	0.004781	0.02089	0.04591	498.020933	498.020933	498.020933	0	0	1	-360	360;
	7063	4355	46.67975	63.95455	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	6888	45.98967	55.45455	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	5823	1.337128	3.255992	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	990	0.09056	0.365041	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	3564	0.085081	0.477665	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7063	3564	0.008	0.040669	0.06203	427.846549	427.846549	427.846549	0	0	1	-360	360;
	5636	6996	2.446281	4.115289	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5636	87	0.00488	0.020469	0.04681	591.653972	591.653972	591.653972	0	0	1	-360	360;
	5636	4355	44.43595	63.26653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5636	6888	47.34298	50.52893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5636	5823	1.382355	2.978719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5636	990	0.07226	0.393864	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5636	990	0.007831	0.04144	0.06071	432.560914	432.560914	432.560914	0	0	1	-360	360;
	5636	3564	0.08989	0.428161	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3391	7232	0.00055	0.00435	0.17271	1827.44406	1827.44406	1827.44406	0	0	1	-360	360;
	2058	7232	0.00128	0.00706	0.25912	1915.94351	1915.94351	1915.94351	0	0	1	-360	360;
	5589	5043	0.00053	0.00407	0.16804	1314.27482	1314.27482	1314.27482	0	0	1	-360	360;
	6316	5043	0.00056	0.00478	0.18777	1741.10476	1741.10476	1741.10476	0	0	1	-360	360;
	7561	3391	0.208393	1.089612	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2274	3391	0.199571	1.1759	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1056	3391	0.02232	0.346918	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2821	3391	0.858172	1.866898	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1840	3391	0.00031	0.00216	0.0967	1575.18236	1575.18236	1575.18236	0	0	1	-360	360;
	7722	3391	0.001	0.274529	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	3391	4.431302	8.136427	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	7755	0.36115	2.404917	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	172	0.319377	2.230125	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	3041	0.325519	2.183657	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	8032	1.726316	5.458795	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	6254	2.428047	7.275623	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	8258	0.00056	0.00716	0.17465	1349.83361	1349.83361	1349.83361	0	0	1	-360	360;
	8032	8258	0.527043	3.451801	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	5189	0.515582	3.196122	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	5646	1.334072	5.856163	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8032	7672	0.0011	0.01318	0.3431	1562.60455	1562.60455	1562.60455	0	0	1	-360	360;
	8032	7985	0.00108	0.01344	0.3361	1605.38423	1605.38423	1605.38423	0	0	1	-360	360;
	8032	5189	0.00063	0.00702	0.19793	1614.21898	1614.21898	1614.21898	0	0	1	-360	360;
	3654	937	0.246612	1.335723	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	937	0.008229	0.04962	0.08326	479.190527	479.190527	479.190527	0	0	1	-360	360;
	3654	3117	0.219628	1.572521	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	3675	4.089256	11.14855	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	1124	0.00943	0.04726	0.09545	526.873367	526.873367	526.873367	0	0	1	-360	360;
	3654	1124	3.689876	10.15083	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	394	0.202581	1.497417	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	1744	0.1973	1.490124	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	251	0.0028	0.013271	0.02916	464.578739	464.578739	464.578739	0	0	1	-360	360;
	3654	3024	0.12369	1.125744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	1447	1.335496	4.761364	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	8114	1.15	4.888636	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	1071	1.052913	5.261983	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3654	3117	0.00932	0.04863	0.09436	490.801538	490.801538	490.801538	0	0	1	-360	360;
	3654	3675	0.00961	0.046281	0.09729	517.440064	517.440064	517.440064	0	0	1	-360	360;
	3654	5378	0.0025	0.015031	0.02598	427.215909	427.215909	427.215909	0	0	1	-360	360;
	4391	6897	0.0008	0.00519	0.0084	455.367589	455.367589	455.367589	0	0	1	-360	360;
	4391	1124	0.00612	0.03036	0.05912	506.487129	506.487129	506.487129	0	0	1	-360	360;
	6897	1695	0.00038	0.00243	0.00393	461.92548	461.92548	461.92548	0	0	1	-360	360;
	4323	5645	0.000909	0.005091	0.00952	455.841617	455.841617	455.841617	0	0	1	-360	360;
	4323	394	0.00187	0.00944	0.01839	451.683967	451.683967	451.683967	0	0	1	-360	360;
	5645	1695	0.00043	0.00238	0.00446	460.293437	460.293437	460.293437	0	0	1	-360	360;
	7884	3838	0.00013	0.00176	0.04234	1435.42786	1435.42786	1435.42786	0	0	1	-360	360;
	7884	3838	0.00013	0.00179	0.04144	1610.94058	1610.94058	1610.94058	0	0	1	-360	360;
	6379	972	0.00018	0.00189	0.06238	2093.85596	2093.85596	2093.85596	0	0	1	-360	360;
	6379	6507	0.0002	0.00185	0.0707	2045.50479	2045.50479	2045.50479	0	0	1	-360	360;
	972	1387	0.00039	0.00344	0.13109	1881.82481	1881.82481	1881.82481	0	0	1	-360	360;
	6507	5188	0.0004	0.00337	0.13366	1946.86675	1946.86675	1946.86675	0	0	1	-360	360;
	6379	7633	0.00027	0.00274	0.09232	2271.10298	2271.10298	2271.10298	0	0	1	-360	360;
	6379	1138	0.00056	0.00437	0.18862	2119.6138	2119.6138	2119.6138	0	0	1	-360	360;
	6379	6037	0.00026	0.00279	0.0904	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	5599	2526	0.001349	0.00763	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5599	2364	0.00374	0.02218	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5599	6638	0.000349	0.00186	0	232.989551	232.989551	232.989551	0	0	1	-360	360;
	5599	2364	0.004651	0.02025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5599	6638	0.00037	0.00182	0	232.791023	232.791023	232.791023	0	0	1	-360	360;
	6803	3733	0.001	0.209093	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6803	7561	0.00062	0.00513	0.19541	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	6803	2274	0.04427	0.493158	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6803	1387	2.319321	5.378532	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4429	6803	2.762188	4.809072	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	6803	24.69598	23.43698	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6803	7640	1.383864	4.675693	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7088	6803	0.470817	2.500277	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1048	6803	0.00099	0.00946	0.31191	2244.29797	2244.29797	2244.29797	0	0	1	-360	360;
	4927	6803	0.05862	0.781925	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3733	2274	0.04335	0.502652	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3733	2274	0.00063	0.00502	0.19924	1710.46717	1710.46717	1710.46717	0	0	1	-360	360;
	3733	1387	2.364751	5.268767	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4429	3733	2.814266	4.706787	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	3733	25.15305	22.92729	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3733	8247	0.00032	0.00223	0.09984	2289.24043	2289.24043	2289.24043	0	0	1	-360	360;
	3733	7640	1.353809	4.329363	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7088	3733	0.460582	2.315097	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4927	3733	0.0574	0.796953	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7274	1917	0.00439	0.016341	0.0518	242.387035	242.387035	242.387035	0	0	1	-360	360;
	229	5191	0.004039	0.019031	0.04007	543.11834	543.11834	543.11834	0	0	1	-360	360;
	229	7703	0.001331	0.007669	0.01326	521.358413	521.358413	521.358413	0	0	1	-360	360;
	5191	6563	0.001091	0.005159	0.01134	551.743434	551.743434	551.743434	0	0	1	-360	360;
	7703	1890	0.00131	0.00782	0.01299	463.519798	463.519798	463.519798	0	0	1	-360	360;
	1890	2487	0.00449	0.022021	0.04465	476.050285	476.050285	476.050285	0	0	1	-360	360;
	7561	2274	0	4e-05	0.00136	1450.81872	1450.81872	1450.81872	0	0	1	-360	360;
	7561	2274	0.001	0.229342	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1056	7561	0.159093	0.880956	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2821	7561	0.11009	0.785942	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1840	7561	0.00133	0.01275	0.42052	2134.25171	2134.25171	2134.25171	0	0	1	-360	360;
	7722	7561	0.08106	0.897645	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	7561	0.0009	0.00658	0.28274	1533.16946	1533.16946	1533.16946	0	0	1	-360	360;
	6100	7561	1.066967	3.005956	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1861	7561	0.00037	0.00335	0.11768	1702.71711	1702.71711	1702.71711	0	0	1	-360	360;
	1056	2274	0.14919	0.968975	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2821	2274	0.107472	0.83054	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2274	1387	6.591066	12.31994	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7722	2274	0.089861	0.931163	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7722	2274	0.00151	0.0125	0.47659	2133.06304	2133.06304	2133.06304	0	0	1	-360	360;
	4429	2274	7.50831	11.99515	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	2274	57.05332	52.93698	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	2274	1.218213	2.697091	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	2274	0.00091	0.00644	0.28798	1489.05894	1489.05894	1489.05894	0	0	1	-360	360;
	7640	2274	5.391066	9.229224	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7088	2274	1.899723	5.672438	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1060	2274	0.00034	0.00341	0.11528	1833.09716	1833.09716	1833.09716	0	0	1	-360	360;
	4927	2274	0.372853	1.916066	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	162	2641	0.001	0.398657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	162	1999	0.00157	0.00875	0.01382	407.457643	407.457643	407.457643	0	0	1	-360	360;
	1778	2641	0.013419	0.18632	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1778	1999	0.001771	0.00857	0.01567	358.04788	358.04788	358.04788	0	0	1	-360	360;
	3665	6442	3e-05	0.00035	0.0147	1334.28742	1334.28742	1334.28742	0	0	1	-360	360;
	3665	4220	4e-05	0.00047	0.01897	1306.73735	1306.73735	1306.73735	0	0	1	-360	360;
	7277	6996	6.406818	19.24731	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	6996	5.408471	18.64401	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	6996	4.777893	12.85021	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6996	973	7.316529	22.04545	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6996	973	0.00518	0.053229	0.20193	405.225445	405.225445	405.225445	0	0	1	-360	360;
	4355	6996	1.28	5.819421	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	6996	1.306178	5.47314	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6996	5823	0.001	0.244421	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	990	6996	0.326384	2.031736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3564	6996	0.34876	2.295661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6996	5823	0.000401	0.00324	0.01013	436.861457	436.861457	436.861457	0	0	1	-360	360;
	172	7755	1.103116	4.165166	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3041	7755	0.982756	4.720568	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	7755	17.77701	32.87327	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7755	6254	11.94321	24.78324	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8258	7755	0.13187	1.326939	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5189	7755	0.129169	1.352909	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	7755	15.47438	32.5367	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7884	7755	0.00019	0.00211	0.05553	1256.68084	1256.68084	1256.68084	0	0	1	-360	360;
	7884	7755	0.0002	0.00206	0.05662	1234.24877	1234.24877	1234.24877	0	0	1	-360	360;
	6970	3514	0.00013	0.001771	0.00463	349.25703	349.25703	349.25703	0	0	1	-360	360;
	6091	4836	0.008081	0.041651	0.08445	417.671339	417.671339	417.671339	0	0	1	-360	360;
	6091	4836	0.168091	0.83657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6091	4836	0.008229	0.040771	0.08604	442.754795	442.754795	442.754795	0	0	1	-360	360;
	4815	4836	2.293802	5.98719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2039	4836	2.825	5.516736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4836	3961	0.731178	1.985165	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7084	6091	0.001349	0.01345	0.04665	327.082137	327.082137	327.082137	0	0	1	-360	360;
	586	7084	0.00051	0.00643	0.01752	411.540607	411.540607	411.540607	0	0	1	-360	360;
	6749	6091	0.00137	0.013159	0.04751	301.981153	301.981153	301.981153	0	0	1	-360	360;
	586	6749	0.0005	0.00655	0.01716	393.981387	393.981387	393.981387	0	0	1	-360	360;
	8206	5971	0.001	0.425381	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5894	5971	4.659141	9.407202	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4335	5971	1.022853	3.286357	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5971	3912	1.18795	4.384626	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5971	5092	1.213781	4.303324	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4222	5971	8.581024	12.40582	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6785	5971	0.203857	1.400623	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5971	3912	0.00082	0.0099	0.25793	2151.72637	2151.72637	2151.72637	0	0	1	-360	360;
	6785	5971	0.00056	0.0066	0.17182	1802.40363	1802.40363	1802.40363	0	0	1	-360	360;
	4074	5971	0.00031	0.00634	0.21893	1023.40564	1023.40564	1023.40564	0	0	1	-360	360;
	5894	8206	4.02687	9.790858	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4335	8206	1.000554	3.350416	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8206	5092	0.00081	0.0101	0.25266	1525.7452	1525.7452	1525.7452	0	0	1	-360	360;
	8206	3912	1.418213	3.650623	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8206	5092	1.263504	4.137327	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4222	8206	8.415512	12.67105	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6785	8206	0.172188	1.486496	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4815	6091	0.008531	0.04201	0.08867	457.444318	457.444318	457.444318	0	0	1	-360	360;
	4815	6091	0.19088	1.317025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2039	6091	0.20351	1.242748	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6091	6563	0.003419	0.02226	0.0355	517.970487	517.970487	517.970487	0	0	1	-360	360;
	6091	3961	0.05611	0.44343	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2039	6091	0.00869	0.041091	0.09031	457.715626	457.715626	457.715626	0	0	1	-360	360;
	6091	3782	0.003349	0.02268	0.03475	508.932715	508.932715	508.932715	0	0	1	-360	360;
	6328	6091	0.00176	0.018651	0.06155	802.722411	802.722411	802.722411	0	0	1	-360	360;
	6806	1755	0.00019	0.00245	0.05615	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6806	1755	0.00019	0.00227	0.05493	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4934	3912	0.00018	0.00209	0.05447	2260.52999	2260.52999	2260.52999	0	0	1	-360	360;
	2024	5092	0.00017	0.00213	0.05336	1761.36022	1761.36022	1761.36022	0	0	1	-360	360;
	6124	381	0.00024	0.0025	0.07634	1383.0381	1383.0381	1383.0381	0	0	1	-360	360;
	5894	4335	0.897576	2.903878	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4222	5894	0.01432	0.301302	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5240	5894	0.00025	0.00245	0.07778	1409.9833	1409.9833	1409.9833	0	0	1	-360	360;
	5894	6785	3.381094	7.270776	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4281	2821	0.00142	0.01435	0.48502	1804.48808	1804.48808	1804.48808	0	0	1	-360	360;
	4281	985	0.00023	0.00181	0.07208	1842.41698	1842.41698	1842.41698	0	0	1	-360	360;
	7217	2821	0.00139	0.01463	0.47492	1846.21073	1846.21073	1846.21073	0	0	1	-360	360;
	7217	6212	0.00023	0.00178	0.07347	1850.01764	1850.01764	1850.01764	0	0	1	-360	360;
	1056	5589	0.00047	0.00384	0.15877	1237.15463	1237.15463	1237.15463	0	0	1	-360	360;
	1056	2718	0.0005	0.00403	0.16655	1665.07517	1665.07517	1665.07517	0	0	1	-360	360;
	7722	2718	0.00025	0.00236	0.07786	1253.71706	1253.71706	1253.71706	0	0	1	-360	360;
	1056	2821	0.838573	2.37223	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7722	1056	0.078193	0.509841	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	1056	5.654848	7.590028	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6661	2821	0.0009	0.0069	0.2846	1827.72839	1827.72839	1827.72839	0	0	1	-360	360;
	6100	6661	0.0002	0.00149	0.06146	2125.99551	2125.99551	2125.99551	0	0	1	-360	360;
	8303	2821	0.00092	0.00675	0.28997	1881.44109	1881.44109	1881.44109	0	0	1	-360	360;
	6100	8303	0.0002	0.00145	0.06262	2097.74316	2097.74316	2097.74316	0	0	1	-360	360;
	7722	2821	0.413608	1.666759	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	2821	0.03562	0.272507	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4447	564	0.0001	0.00144	0.03317	1762.45412	1762.45412	1762.45412	0	0	1	-360	360;
	7884	4447	9e-05	0.00129	0.02848	1991.92899	1991.92899	1991.92899	0	0	1	-360	360;
	1452	5996	0.0001	0.00133	0.03245	2107.54477	2107.54477	2107.54477	0	0	1	-360	360;
	7884	1452	9e-05	0.00132	0.02788	2147.72727	2147.72727	2147.72727	0	0	1	-360	360;
	564	7616	0.00063	0.00676	0.19823	1947.42093	1947.42093	1947.42093	0	0	1	-360	360;
	5996	2128	0.00065	0.00662	0.20205	1798.42493	1798.42493	1798.42493	0	0	1	-360	360;
	3946	7616	0.00017	0.00236	0.0546	1676.47155	1676.47155	1676.47155	0	0	1	-360	360;
	3946	2128	0.00017	0.00219	0.05341	1765.1204	1765.1204	1765.1204	0	0	1	-360	360;
	1128	8106	0.0017	0.007469	0.02515	275.897438	275.897438	275.897438	0	0	1	-360	360;
	8106	6527	0.001039	0.005599	0.01573	515.784777	515.784777	515.784777	0	0	1	-360	360;
	1128	6760	0.001729	0.00731	0.02563	496.593897	496.593897	496.593897	0	0	1	-360	360;
	6760	87	0.001021	0.0057	0.01539	510.503789	510.503789	510.503789	0	0	1	-360	360;
	1128	1638	0.00168	0.007159	0.02631	427.540564	427.540564	427.540564	0	0	1	-360	360;
	1638	6527	0.00112	0.00495	0.01749	432.637505	432.637505	432.637505	0	0	1	-360	360;
	8106	6760	0.001	0.364917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1128	8106	0.106669	0.821178	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8106	6527	0.0977	0.935248	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8106	87	0.10481	0.756446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1128	6760	0.094539	0.895083	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6760	6527	0.09563	0.921384	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6760	87	0.085479	0.810062	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3946	3303	0.0002	0.0021	0.06172	1505.6964	1505.6964	1505.6964	0	0	1	-360	360;
	3303	5929	0.00051	0.00613	0.17787	1699.04973	1699.04973	1699.04973	0	0	1	-360	360;
	6617	5581	0.00062	0.006331	0.02172	302.946737	302.946737	302.946737	0	0	1	-360	360;
	5581	1447	0.00075	0.00981	0.02957	377.47044	377.47044	377.47044	0	0	1	-360	360;
	3946	7811	0.00105	0.01369	0.32909	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7811	444	0.00132	0.01291	0.41076	1670.83951	1670.83951	1670.83951	0	0	1	-360	360;
	3946	5568	0.00103	0.01395	0.32209	1796.47607	1796.47607	1796.47607	0	0	1	-360	360;
	5568	6845	0.00134	0.01263	0.41836	2096.86186	2096.86186	2096.86186	0	0	1	-360	360;
	3946	5388	0.269647	1.156579	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	292	3946	0.06211	0.718075	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3946	534	0.00062	0.005	0.18404	1598.88537	1598.88537	1598.88537	0	0	1	-360	360;
	4553	3946	0.00031	0.00248	0.10243	1760.79222	1760.79222	1760.79222	0	0	1	-360	360;
	3946	2175	0.0006	0.0051	0.18029	1690.95347	1690.95347	1690.95347	0	0	1	-360	360;
	742	2487	0.007331	0.044461	0.07089	463.647069	463.647069	463.647069	0	0	1	-360	360;
	742	2487	0.00718	0.0453	0.06938	459.322519	459.322519	459.322519	0	0	1	-360	360;
	6617	336	0.0002	0.00211	0.00668	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4568	336	0.000659	0.002909	0.02552	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6037	6744	0.00088	0.0079	0.30106	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7633	7262	0.00089	0.00774	0.30697	2194.07692	2194.07692	2194.07692	0	0	1	-360	360;
	5188	3957	0.00089	0.00706	0.30347	2095.95555	2095.95555	2095.95555	0	0	1	-360	360;
	1387	1702	0.0009	0.0069	0.30909	2008.89291	2008.89291	2008.89291	0	0	1	-360	360;
	4429	1387	0.527957	1.890097	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	1387	6.434972	8.855956	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7640	1387	0.02292	0.469398	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7088	1387	0.03034	0.81482	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4927	1387	0.225173	1.343006	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3914	3053	0.0001	0.00122	0.04187	1349.56771	1349.56771	1349.56771	0	0	1	-360	360;
	8359	3053	0.00025	0.00221	0.11189	1903.08005	1903.08005	1903.08005	0	0	1	-360	360;
	81	950	2.696676	13.13573	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	81	5013	3.125	11.9903	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3665	81	0.00054	0.00664	0.22073	1690.43087	1690.43087	1690.43087	0	0	1	-360	360;
	2620	81	0.899169	5.002355	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4308	81	0.06172	1.437119	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	54	81	0.521247	2.927216	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	81	0.00397	0.614771	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7556	81	0.956579	5.569321	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2470	81	0.071011	1.246122	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	81	22.47576	40.81994	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	571	81	0.00141	0.00778	0.19423	2204.08849	2204.08849	2204.08849	0	0	1	-360	360;
	5780	1702	0.00021	0.00179	0.07124	1985.27809	1985.27809	1985.27809	0	0	1	-360	360;
	7640	3957	0.00021	0.00175	0.07261	2018.69649	2018.69649	2018.69649	0	0	1	-360	360;
	4641	4710	0.00051	0.00411	0.17181	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	1656	6935	0.00052	0.00402	0.17505	2297.40778	2297.40778	2297.40778	0	0	1	-360	360;
	4710	6935	0.00439	0.248677	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4815	2039	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4815	2039	0.001	0.289959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4815	3403	9.9e-05	0.000789	0.00289	279.113129	279.113129	279.113129	0	0	1	-360	360;
	4815	3961	0.13343	0.727438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2039	3961	0.11368	0.773802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4286	6813	0.0001	0.0011	0.03679	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4476	4286	0.00069	0.00512	0.22636	2237.04089	2237.04089	2237.04089	0	0	1	-360	360;
	7345	1320	0.0001	0.00102	0.03599	2288.35913	2288.35913	2288.35913	0	0	1	-360	360;
	2218	7345	0.0007	0.005	0.23055	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	5158	3516	0.00031	0.00399	0.09656	1596.80355	1596.80355	1596.80355	0	0	1	-360	360;
	950	5158	9e-05	0.00103	0.02544	1577.6354	1577.6354	1577.6354	0	0	1	-360	360;
	3516	4893	0.00021	0.00231	0.06482	1549.66475	1549.66475	1549.66475	0	0	1	-360	360;
	4893	1965	8e-05	0.00082	0.02308	1436.15844	1436.15844	1436.15844	0	0	1	-360	360;
	1965	8123	0.00016	0.00203	0.04892	1463.40377	1463.40377	1463.40377	0	0	1	-360	360;
	8123	6493	3e-05	0.00036	0.01167	1399.16613	1399.16613	1399.16613	0	0	1	-360	360;
	54	6493	0.00061	0.00766	0.18271	1572.08958	1572.08958	1572.08958	0	0	1	-360	360;
	950	5013	0	3.5e-05	0	1924.46561	1924.46561	1924.46561	0	0	1	-360	360;
	950	5013	0.01257	0.43678	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2620	950	22.73892	49	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4308	950	4.844044	19.18698	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	54	950	0.001	1.752701	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	950	0.908172	6.096537	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6425	950	4.358657	9.756232	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7556	950	25.29224	66.18213	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2470	950	1.071399	5.397922	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	950	0.155789	0.985526	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2620	5013	23.51524	49.62604	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4308	5013	5.179155	18.97368	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	54	5013	0.0144	1.921053	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	5013	0.997091	6.027978	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6425	5013	4.636704	9.781162	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7556	5013	29.33518	59.12049	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2470	5013	1.098269	5.695983	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	5013	0.00085	0.00853	0.24675	1796.20095	1796.20095	1796.20095	0	0	1	-360	360;
	1153	5013	0.15653	1.070014	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5531	5388	0.00015	0.00148	0.04891	2038.76306	2038.76306	2038.76306	0	0	1	-360	360;
	429	5388	0.00014	0.00167	0.04357	1373.44448	1373.44448	1373.44448	0	0	1	-360	360;
	292	5388	0.525672	1.551039	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5869	7206	0.00411	0.01861	0.04081	497.955774	497.955774	497.955774	0	0	1	-360	360;
	3782	5869	0.000969	0.005841	0.0101	484.078652	484.078652	484.078652	0	0	1	-360	360;
	7206	2487	0.004581	0.02156	0.04549	504.435549	504.435549	504.435549	0	0	1	-360	360;
	2385	5883	0.00026	0.00323	0.10874	968.449639	968.449639	968.449639	0	0	1	-360	360;
	2385	5883	0.00026	0.00329	0.10643	990.219578	990.219578	990.219578	0	0	1	-360	360;
	172	7672	6e-05	0.00065	0.01567	1202.17635	1202.17635	1202.17635	0	0	1	-360	360;
	7672	1442	0.00038	0.00377	0.12003	1327.66811	1327.66811	1327.66811	0	0	1	-360	360;
	3041	7985	6e-05	0.00066	0.01534	1374.84443	1374.84443	1374.84443	0	0	1	-360	360;
	7985	1869	0.00038	0.00369	0.12225	1495.04771	1495.04771	1495.04771	0	0	1	-360	360;
	172	3041	0.01122	0.71669	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	172	25.10803	39.16205	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	172	6254	32.07964	43.87258	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	172	8258	0.427673	3.90554	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	172	5189	0.418573	3.979224	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	172	23.26662	36.54363	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	3041	25.59141	38.34626	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3041	6254	32.6849	42.93906	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3041	8258	0.409481	3.684488	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3041	5189	0.464072	3.610803	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	3041	23.70568	35.76593	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	937	3117	0	0.000103	0	405.930771	405.930771	405.930771	0	0	1	-360	360;
	937	3117	0.001	0.19151	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	937	3675	17.56446	20.66736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	937	1124	13.87645	21.76446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	937	394	1.145248	3.637603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	937	1744	1.372004	3.186983	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	937	0.119159	0.998905	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	937	1447	6.052479	9.575826	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	937	4.474793	10.74483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	937	4.534711	10.47087	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3117	3675	18.49236	22.54545	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3117	1124	13.69401	22.82645	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3117	394	1.333636	3.866942	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3117	1744	1.179442	4.148554	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	3117	0.114281	1.035289	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3117	1447	5.617769	11.75372	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	3117	4.939669	11.00083	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	3117	4.903306	10.93058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6975	1320	2e-05	0.00021	0.0099	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	6975	1893	0.00028	0.00263	0.09558	2271.2129	2271.2129	2271.2129	0	0	1	-360	360;
	2973	6813	2e-05	0.00023	0.00882	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2973	8049	0.00027	0.00268	0.09359	2293.56533	2293.56533	2293.56533	0	0	1	-360	360;
	1128	6527	0.02601	0.665517	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1128	87	0.028591	0.528326	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7191	5942	0.01088	0.212832	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2215	5942	0.00032	0.00331	0.10347	1593.04798	1593.04798	1593.04798	0	0	1	-360	360;
	3881	5942	0.00031	0.00306	0.10122	1242.33253	1242.33253	1242.33253	0	0	1	-360	360;
	4335	215	0.00057	0.01002	0.40595	2281.15667	2281.15667	2281.15667	0	0	1	-360	360;
	4335	2072	0.00053	0.00558	0.16379	1971.28717	1971.28717	1971.28717	0	0	1	-360	360;
	4335	3912	22.58241	30.56787	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4335	5092	22.12119	31.16759	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4222	4335	1.627632	4.119529	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7950	4335	0.0003	0.00319	0.09362	1586.96904	1586.96904	1586.96904	0	0	1	-360	360;
	6785	4335	0.00232	0.519501	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4335	8175	0.00058	0.00981	0.41361	2237.27454	2237.27454	2237.27454	0	0	1	-360	360;
	8249	8175	0.00065	0.00643	0.21784	2216.29903	2216.29903	2216.29903	0	0	1	-360	360;
	4097	215	0.00063	0.00655	0.2134	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	1840	7722	0.001	0.215229	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1840	7191	0.00102	0.00911	0.32053	1501.3195	1501.3195	1501.3195	0	0	1	-360	360;
	7722	5946	0.00094	0.00901	0.29689	1756.94252	1756.94252	1756.94252	0	0	1	-360	360;
	6100	7722	3.359072	5.253809	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	985	4294	0.00019	0.00221	0.05932	1497.60408	1497.60408	1497.60408	0	0	1	-360	360;
	7191	985	0.00039	0.00355	0.12404	1569.92088	1569.92088	1569.92088	0	0	1	-360	360;
	985	1046	0.0001	0.00073	0.02981	2215.10838	2215.10838	2215.10838	0	0	1	-360	360;
	6212	544	0.00018	0.00226	0.05811	1552.3804	1552.3804	1552.3804	0	0	1	-360	360;
	7573	6212	0.00038	0.00368	0.1194	987.365054	987.365054	987.365054	0	0	1	-360	360;
	6212	6455	0.00047	0.00386	0.15773	1843.85247	1843.85247	1843.85247	0	0	1	-360	360;
	6527	87	0.001	0.309525	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2641	1999	0.000711	0.00393	0.01225	564.505981	564.505981	564.505981	0	0	1	-360	360;
	2641	6925	0.001219	0.00569	0.01778	231.391041	231.391041	231.391041	0	0	1	-360	360;
	7568	5703	0.00125	0.005581	0.01813	460.078906	460.078906	460.078906	0	0	1	-360	360;
	1999	5703	0.0007	0.00401	0.01199	279.163427	279.163427	279.163427	0	0	1	-360	360;
	6100	4429	0.0599	0.313109	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4429	2971	0.00051	0.00517	0.16114	1772.23532	1772.23532	1772.23532	0	0	1	-360	360;
	4429	7640	0.92403	2.090097	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4429	7088	0.05633	0.587403	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4429	4927	0.333109	1.178601	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4568	6523	0.0002	0.002599	0.00692	398.665649	398.665649	398.665649	0	0	1	-360	360;
	5903	3665	6e-05	0.00066	0.02297	1850.54286	1850.54286	1850.54286	0	0	1	-360	360;
	3665	4294	0.00203	0.01921	0.65626	1425.54464	1425.54464	1425.54464	0	0	1	-360	360;
	3665	7556	0.00132	0.01562	0.56572	1649.10248	1649.10248	1649.10248	0	0	1	-360	360;
	3665	1755	0.00039	0.00541	0.17125	2160.66181	2160.66181	2160.66181	0	0	1	-360	360;
	3665	7556	0.0013	0.01592	0.55009	1686.05464	1686.05464	1686.05464	0	0	1	-360	360;
	3665	6047	0.00084	0.00852	0.34862	2105.32407	2105.32407	2105.32407	0	0	1	-360	360;
	5903	2777	0.00016	0.0017	0.06742	1556.82245	1556.82245	1556.82245	0	0	1	-360	360;
	1652	4641	0.00092	0.00761	0.29003	1691.30099	1691.30099	1691.30099	0	0	1	-360	360;
	1652	444	0.00062	0.00468	0.19331	1730.48701	1730.48701	1730.48701	0	0	1	-360	360;
	5544	1656	0.00094	0.00746	0.29571	1581.23234	1581.23234	1581.23234	0	0	1	-360	360;
	5544	6845	0.00063	0.00458	0.19696	1650.86311	1650.86311	1650.86311	0	0	1	-360	360;
	4641	1656	0	3.5e-05	0	1550.15246	1550.15246	1550.15246	0	0	1	-360	360;
	4641	1656	0.001	0.227327	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4641	3317	3e-05	0.00034	0.01043	2021.06462	2021.06462	2021.06462	0	0	1	-360	360;
	7088	4641	0.00052	0.00501	0.16506	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	3683	1656	0.00059	0.00491	0.18707	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	219	1124	0.006229	0.029719	0.06026	504.648176	504.648176	504.648176	0	0	1	-360	360;
	219	394	0.00194	0.00924	0.01875	493.473088	493.473088	493.473088	0	0	1	-360	360;
	2837	2384	0.00189	0.00769	0.02943	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2837	8111	0.0037	0.017021	0.0577	488.18867	488.18867	488.18867	0	0	1	-360	360;
	4183	1042	0.00168	0.008719	0.02622	453.101096	453.101096	453.101096	0	0	1	-360	360;
	4183	5601	0.003771	0.016669	0.05881	531.314518	531.314518	531.314518	0	0	1	-360	360;
	5842	6925	0.004669	0.024729	0.03859	448.148574	448.148574	448.148574	0	0	1	-360	360;
	1082	7568	0	0.000103	0	286.915533	286.915533	286.915533	0	0	1	-360	360;
	1082	7568	0.00099	0.00607	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5842	1082	0.00476	0.02624	0.04071	397.032845	397.032845	397.032845	0	0	1	-360	360;
	1124	5281	0.00051	0.00286	0.00493	459.3591	459.3591	459.3591	0	0	1	-360	360;
	1124	5281	0.0005	0.002909	0.00483	488.033582	488.033582	488.033582	0	0	1	-360	360;
	3675	1124	0	0.000103	0	302.002492	302.002492	302.002492	0	0	1	-360	360;
	3675	1124	0.001	0.18874	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3675	394	0.704421	2.986983	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3675	1744	0.680207	2.970661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	3675	4.389256	10.44091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3675	1447	0.00861	0.432252	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	3675	0.006521	0.039849	0.0689	470.446165	470.446165	470.446165	0	0	1	-360	360;
	8114	3675	0.450599	2.525826	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	3675	0.51345	2.178099	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3675	1792	0.000159	0.002091	0.00547	369.55372	369.55372	369.55372	0	0	1	-360	360;
	1124	394	0.547603	3.122934	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1124	1744	0.63281	2.590496	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	1124	4.541736	8.227686	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1124	1447	0.00687	0.424525	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	1124	0.394752	2.341736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	1124	0.00638	0.04063	0.0675	458.385895	458.385895	458.385895	0	0	1	-360	360;
	1071	1124	0.459607	1.981798	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1124	1776	0.000159	0.00213	0.00535	341.245428	341.245428	341.245428	0	0	1	-360	360;
	2384	1042	0.001	0.319917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3026	2384	0.074539	0.567996	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3026	1042	0.054219	0.438326	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1519	3749	0.00049	0.0046	0.16791	1744.33313	1744.33313	1744.33313	0	0	1	-360	360;
	7427	3749	0.00063	0.00676	0.21595	1549.95632	1549.95632	1549.95632	0	0	1	-360	360;
	1519	2096	0.02777	0.440042	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3194	2096	0.02704	0.444668	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3194	2096	0.00048	0.00469	0.16449	1806.78842	1806.78842	1806.78842	0	0	1	-360	360;
	7427	2096	0.00062	0.00626	0.21126	1521.53746	1521.53746	1521.53746	0	0	1	-360	360;
	1519	3194	0	4e-05	0.00136	1429.47989	1429.47989	1429.47989	0	0	1	-360	360;
	1519	3194	0.001	0.247292	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1519	5410	3e-05	0.00035	0.0107	1529.44022	1529.44022	1529.44022	0	0	1	-360	360;
	8215	1519	0.00014	0.00118	0.02927	1270.35583	1270.35583	1270.35583	0	0	1	-360	360;
	3194	7049	3e-05	0.00034	0.0109	1527.34457	1527.34457	1527.34457	0	0	1	-360	360;
	4094	5410	0.00072	0.00903	0.22686	1538.28154	1538.28154	1538.28154	0	0	1	-360	360;
	4094	7049	0.00082	0.00885	0.25705	1688.30232	1688.30232	1688.30232	0	0	1	-360	360;
	6744	879	0.00012	0.00093	0.03979	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6299	6563	0.000831	0.004229	0.00859	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8221	3782	0.000841	0.00414	0.00875	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7950	5531	0.00049	0.00607	0.15204	1892.51891	1892.51891	1892.51891	0	0	1	-360	360;
	6785	2072	0.00048	0.00567	0.14764	1559.81453	1559.81453	1559.81453	0	0	1	-360	360;
	6785	429	0.00047	0.00578	0.14462	1353.29959	1353.29959	1353.29959	0	0	1	-360	360;
	4986	5646	6e-05	0.00069	0.0167	1623.45389	1623.45389	1623.45389	0	0	1	-360	360;
	4986	6254	0.198269	1.245429	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	8258	28.23615	48.46122	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	5189	32.00069	47.49169	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4986	5646	0.001	0.339079	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3912	5092	0.001	0.23876	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6785	3912	7.242382	14.44598	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6785	5092	7.08795	14.71814	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4590	5083	0.00493	0.059419	0.16378	293.646421	293.646421	293.646421	0	0	1	-360	360;
	4590	5083	0.01006	0.07863	0.12554	529.496904	529.496904	529.496904	0	0	1	-360	360;
	4590	5156	0.001	0.434587	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	2624	0.001	0.434132	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	4084	9.525206	20.12748	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	7277	0.490393	3.121901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	236	0.324483	3.440909	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	5360	7.110331	13.27211	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	2560	8.147314	13.64236	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	5082	1.873657	4.730165	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	3731	1.953223	4.544628	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	570	0.893533	3.267562	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	3111	0.947417	3.197521	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	2684	1.735227	5.478926	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	4590	49.21281	47.44215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	4590	21.70248	30.12603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	4274	9.99835	24.78719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	6801	10.26198	26.88017	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	973	0.625785	4.196281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	4590	0.531426	3.034091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	4590	3.245868	10.84669	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	4590	3.282025	10.98388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	1441	0.002919	0.01451	0.04697	559.589659	559.589659	559.589659	0	0	1	-360	360;
	4590	1441	0.001	0.67564	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	6035	0.227169	1.409215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	5721	0.192229	1.194917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4590	1441	0.002979	0.0142	0.04787	562.220436	562.220436	562.220436	0	0	1	-360	360;
	4699	5021	0.000151	0.0015	0.00497	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4699	3961	0.000151	0.001461	0.00507	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5883	1359	0.00136	0.0167	0.3953	1437.37475	1437.37475	1437.37475	0	0	1	-360	360;
	5156	2624	0	0.000103	0	328.943953	328.943953	328.943953	0	0	1	-360	360;
	5156	2624	0.001	0.256632	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	5156	6.78595	11.8655	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	5156	1.097169	5.085744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	5156	0.904029	5.181818	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	5156	4.170248	8.645247	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	5156	4.773967	8.884503	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	5156	0.874421	3.172727	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	5156	1.05436	2.929339	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5156	0.513368	2.495455	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	5156	0.604525	2.218802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2684	5156	1.308616	3.033264	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	5156	39.69008	38.26033	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	5156	17.11942	22.4938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	5156	20.47789	32.19835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	5156	21.35331	30.98554	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	5156	1.604091	6.167149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	5156	1.137851	5.037397	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	5156	6.625413	15.09483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	5156	6.196074	16.5874	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5156	59.24587	56.3595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5156	1441	0.086409	1.077231	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6035	5156	0.080841	0.916612	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	5156	0.07195	0.701488	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	2624	5.952479	12.55207	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	2624	1.208636	4.790289	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	2624	0.897293	4.88781	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	2624	4.399793	8.436157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	2624	4.466736	9.775413	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	2624	0.924855	3.415083	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	2624	1.008368	3.168595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	2624	0.519401	2.533471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	2624	0.611653	2.252893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2684	2624	1.180868	3.490496	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	2624	38.08471	41.4876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	2624	19.69132	22.41116	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	2624	21.63017	31.43802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	2624	20.91529	32.94215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	2624	1.695537	6.03657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	2624	1.302851	4.551033	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	2624	5.718182	16.36116	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	2624	6.683678	15.9062	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2624	1441	0.10661	0.987583	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6035	2624	0.079729	0.948512	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	2624	0.06449	0.807169	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	2624	0.002151	0.02656	0.0604	273.517392	273.517392	273.517392	0	0	1	-360	360;
	5883	4377	0.00094	0.01246	0.34666	1727.53376	1727.53376	1727.53376	0	0	1	-360	360;
	4377	4412	0.00109	0.00971	0.4561	1968.86836	1968.86836	1968.86836	0	0	1	-360	360;
	5883	2738	0.00099	0.01086	0.40729	1697.67085	1697.67085	1697.67085	0	0	1	-360	360;
	5883	2738	0.00092	0.0127	0.34839	1621.61823	1621.61823	1621.61823	0	0	1	-360	360;
	2738	4412	0.00097	0.01101	0.40635	1938.37887	1938.37887	1938.37887	0	0	1	-360	360;
	394	1744	0	0.000103	0	233.919316	233.919316	233.919316	0	0	1	-360	360;
	394	1744	0.001	0.236901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	394	1.033202	3.18595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1447	394	0.07336	0.952996	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	394	4.823554	9.755785	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	394	4.344835	10.74711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6970	394	0.000789	0.01039	0.0272	281.979396	281.979396	281.979396	0	0	1	-360	360;
	3024	1744	0.877293	3.230992	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1447	1744	0.08095	0.839008	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	1744	4.635744	9.702067	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	1744	4.775207	9.229752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5946	2215	0.00017	0.0018	0.0582	1954.98802	1954.98802	1954.98802	0	0	1	-360	360;
	7573	3881	0.00017	0.00184	0.05697	1271.16276	1271.16276	1271.16276	0	0	1	-360	360;
	7844	1893	6e-05	0.00055	0.01834	2271.41891	2271.41891	2271.41891	0	0	1	-360	360;
	5054	8049	5e-05	0.00051	0.01794	2228.52537	2228.52537	2228.52537	0	0	1	-360	360;
	6434	6535	1e-05	4e-05	0	1659.30492	1659.30492	1659.30492	0	0	1	-360	360;
	6434	6535	0.001	0.502417	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6434	444	0.11597	1.150069	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6434	3944	0.00027	0.00201	0.08964	1646.45002	1646.45002	1646.45002	0	0	1	-360	360;
	6434	292	12.69044	26.69945	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5317	6434	5.039543	12.54224	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3691	6434	4.861565	13.04917	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6535	444	0.131427	1.127078	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6535	2276	0.00034	0.00324	0.10679	1806.7022	1806.7022	1806.7022	0	0	1	-360	360;
	6535	292	14.34765	23.73407	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5317	6535	4.937535	12.78878	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3691	6535	4.758934	13.29571	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3024	251	0.00189	0.01011	0.01969	499.436539	499.436539	499.436539	0	0	1	-360	360;
	3024	6718	0.00193	0.009901	0.02007	512.024945	512.024945	512.024945	0	0	1	-360	360;
	290	6718	0.00095	0.004461	0.00979	497.968348	497.968348	497.968348	0	0	1	-360	360;
	290	5378	0.000841	0.00505	0.00872	466.497332	466.497332	466.497332	0	0	1	-360	360;
	5834	3024	0.00074	0.00837	0.02581	433.671297	433.671297	433.671297	0	0	1	-360	360;
	3024	1447	3.807645	6.613017	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	3024	0.512211	2.208058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	3024	0.432727	2.575207	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3024	290	0.00551	0.13675	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7427	6000	0.00018	0.00165	0.05914	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7427	1454	0.00017	0.00168	0.05849	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2777	3914	0.00022	0.00207	0.09688	1437.87826	1437.87826	1437.87826	0	0	1	-360	360;
	6576	5332	6e-05	0.000531	0.00129	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	4008	6e-05	0.000521	0.00132	513.633362	513.633362	513.633362	0	0	1	-360	360;
	4084	5360	2.538636	7.69814	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	2560	2.881818	7.882025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	5082	0.648822	3.687603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	3731	0.662004	3.615496	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	570	0.347417	1.771488	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	3111	0.302727	2.122521	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	2684	0.13874	1.055682	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	4084	32.21488	46.94215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	4084	16.64897	25.36157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	1441	31.40496	45.72727	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	6035	1.96531	5.971488	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4084	5721	1.628843	5.349793	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5332	2684	0.003831	0.04362	0.12635	227.828594	227.828594	227.828594	0	0	1	-360	360;
	4008	2684	0.003901	0.042711	0.12878	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2620	4308	0	3.5e-05	0	1468.28285	1468.28285	1468.28285	0	0	1	-360	360;
	2620	4308	0.03143	0.812673	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6666	2620	6e-05	0.00062	0.01407	1639.57533	1639.57533	1639.57533	0	0	1	-360	360;
	54	2620	5.540582	14.44598	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	2620	0.171863	2.799169	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6425	2620	31.73753	62.63988	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7556	2620	0.0037	0.670762	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2470	2620	4.69723	16.67659	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	2620	13.8608	37.26524	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5054	4308	6e-05	0.00063	0.01377	2051.04271	2051.04271	2051.04271	0	0	1	-360	360;
	54	4308	0.842036	5.10651	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	4308	0.001	0.767382	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7556	4308	0.0122	0.920499	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2470	4308	0.831094	4.707341	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	4308	13.1392	39.18767	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7177	6666	0.00022	0.0023	0.07588	1777.95951	1777.95951	1777.95951	0	0	1	-360	360;
	7473	7844	0.00022	0.00213	0.07423	2269.96763	2269.96763	2269.96763	0	0	1	-360	360;
	7552	6254	0.00034	0.003	0.11433	2027.43316	2027.43316	2027.43316	0	0	1	-360	360;
	292	7552	0.00022	0.00172	0.07388	2263.45165	2263.45165	2263.45165	0	0	1	-360	360;
	4592	6254	0.00034	0.00294	0.11657	2209.32562	2209.32562	2209.32562	0	0	1	-360	360;
	1945	4592	0.00022	0.00168	0.07525	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	8258	6254	25.47853	44.51731	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5189	6254	24.93629	45.35734	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	6254	0.138622	1.351247	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8114	1447	1.420021	4.04938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1071	1447	1.226446	4.635537	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1447	5281	0.000771	0.0083	0.03054	374.674286	374.674286	374.674286	0	0	1	-360	360;
	2526	8241	0.00449	0.023349	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	579	6439	0.001	0.313037	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	579	5842	0.025331	0.364938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	579	5842	0.004219	0.018211	0.0463	236.348136	236.348136	236.348136	0	0	1	-360	360;
	6439	5842	0.02581	0.357169	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6439	5842	0.00413	0.01686	0.0453	567.372629	567.372629	567.372629	0	0	1	-360	360;
	444	3157	0.00018	0.00157	0.05526	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	292	444	4.622922	11.33102	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5317	444	1.337465	6.217936	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3691	444	1.319252	5.765305	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6845	4792	0.00017	0.0016	0.05413	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	6845	3944	0.00017	0.00163	0.05298	1799.36876	1799.36876	1799.36876	0	0	1	-360	360;
	6845	3599	2e-05	0.00022	0.0076	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8111	5601	0	0.000103	0	548.098679	548.098679	548.098679	0	0	1	-360	360;
	8111	5601	0.01018	0.46657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5562	8111	0.00887	0.046021	0.13858	241.260648	241.260648	241.260648	0	0	1	-360	360;
	3442	8111	0.007479	0.02551	0.08524	556.942496	556.942496	556.942496	0	0	1	-360	360;
	1479	5601	0.00762	0.024979	0.08688	571.81721	571.81721	571.81721	0	0	1	-360	360;
	6744	3157	0.00034	0.00277	0.11449	2118.1125	2118.1125	2118.1125	0	0	1	-360	360;
	7262	4792	0.00035	0.00271	0.11665	1800.97735	1800.97735	1800.97735	0	0	1	-360	360;
	4097	2276	0.00146	0.01134	0.48903	1564.8654	1564.8654	1564.8654	0	0	1	-360	360;
	8114	1071	0	0.000103	0	219.695819	219.695819	219.695819	0	0	1	-360	360;
	8114	1071	0.001	0.250041	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8114	4674	0.00031	0.00232	0.3001	445.285356	445.285356	445.285356	0	0	1	-360	360;
	3026	5562	0	0.000103	0	266.147862	266.147862	266.147862	0	0	1	-360	360;
	3026	5562	0.001	0.319463	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3026	5842	0.002479	0.011479	0.02267	448.58564	448.58564	448.58564	0	0	1	-360	360;
	3026	1143	1.778616	5.164256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1999	3026	0.00644	0.026229	0.0492	427.561522	427.561522	427.561522	0	0	1	-360	360;
	3720	3026	0.342769	1.509876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5562	5842	0.002521	0.01124	0.02311	454.622253	454.622253	454.622253	0	0	1	-360	360;
	5562	1143	1.29219	4.678099	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3720	5562	0.236529	1.420661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	236	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	236	0.05505	0.67124	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	7277	4.467975	11.21446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	5082	34.72521	52.45248	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	3731	36.34711	50.63223	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	570	22.76033	33.1281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	3111	21.48554	36.70661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	2684	39.50826	49.23967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	7277	17.04917	42.27893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	7277	16.04855	40.82025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	973	0.053771	1.208946	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	973	0.003281	0.016341	0.09716	519.511458	519.511458	519.511458	0	0	1	-360	360;
	3212	7277	0.825227	5.014463	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	4355	4.278926	11.33285	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	6888	3.46157	11.98905	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	7277	4.532025	17.77417	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	7277	0.009169	0.03831	0.10447	520.941924	520.941924	520.941924	0	0	1	-360	360;
	2361	7277	4.582645	17.99917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	5823	4.507231	15.28285	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	1441	0.001	1.211157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	6035	7.098967	16.7219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7277	5721	6.249587	14.57107	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	236	3.945041	12.27665	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	5082	34.19215	51.58471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	3731	31.52686	50.73347	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	570	18.00289	35.73554	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	3111	21.65083	34.30165	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	2684	35.85331	52.29959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	236	18.05847	51.41322	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	236	21.28512	43.97521	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	973	0.036771	0.970744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	236	0.78843	6.091736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	4355	3.228512	12.74483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	6888	3.551653	11.03347	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	236	5.941736	18.63595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	236	4.915083	20.96901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	236	0.01039	0.037539	0.11815	500.974461	500.974461	500.974461	0	0	1	-360	360;
	236	5823	4.096281	15.0905	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	1441	0.001	1.352211	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	6035	6.635537	16.54979	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	236	5721	5.82438	14.40847	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3691	1945	0.00046	0.00466	0.15732	2258.41197	2258.41197	2258.41197	0	0	1	-360	360;
	5317	292	0.03625	0.46376	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5317	292	0.00045	0.00475	0.15405	2227.18071	2227.18071	2227.18071	0	0	1	-360	360;
	3691	292	0.04062	0.4091	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7277	1441	0.006271	0.0373	0.16824	285.120401	285.120401	285.120401	0	0	1	-360	360;
	586	5021	0.000331	0.00331	0.01147	215.929128	215.929128	215.929128	0	0	1	-360	360;
	586	3961	0.000341	0.00324	0.01168	269.816623	269.816623	269.816623	0	0	1	-360	360;
	586	1350	0.00013	0.00132	0.00437	399.056989	399.056989	399.056989	0	0	1	-360	360;
	5360	2560	0.001	0.1802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	5082	0.04812	0.726715	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	5082	0.005169	0.036401	0.08293	508.047533	508.047533	508.047533	0	0	1	-360	360;
	5360	3731	0.04006	0.78624	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5360	0.559318	2.376033	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	5360	0.574607	2.433264	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	2684	0.449463	2.310744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	5360	0.18062	1.392789	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	5360	0.092211	0.750041	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	5360	6.405372	12.55372	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	5360	5.734504	14.28161	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	5360	15.87066	21.82231	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	5360	45.78719	60.88223	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	5360	52.13017	54.55165	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5360	0.615	2.63781	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	1441	16.26798	23.5124	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	6035	0.818636	3.652686	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5360	5721	0.670434	2.968182	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	5082	0.041711	0.912376	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	3731	0.04611	0.822479	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	3731	0.005271	0.03561	0.08447	547.759925	547.759925	547.759925	0	0	1	-360	360;
	570	2560	0.605434	2.534091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	2560	0.542851	2.99876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	2684	0.520909	2.569628	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	2560	0.177789	1.458264	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	2560	0.09881	0.801281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	2560	6.124793	15.47397	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	2560	5.91157	16.17149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	2560	14.54876	25.34091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	2560	54.67355	60.11777	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	2560	0.802355	2.58657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	1441	19.94897	25.18388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	6035	1.027479	3.393802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2560	5721	0.843471	3.042975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6100	6794	0.00078	0.00782	0.26443	1927.29579	1927.29579	1927.29579	0	0	1	-360	360;
	6100	7640	7.575485	11.77562	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	7088	1.23705	3.353947	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	4927	3.429501	6.941136	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6100	1390	0.00076	0.00797	0.25892	1836.19258	1836.19258	1836.19258	0	0	1	-360	360;
	7484	1390	0.00025	0.00233	0.08577	1954.44568	1954.44568	1954.44568	0	0	1	-360	360;
	2099	6794	0.00179	0.01397	0.60249	1615.30563	1615.30563	1615.30563	0	0	1	-360	360;
	8359	54	0.173643	1.472438	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	54	7556	5.182133	18.94529	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	54	2470	0.188989	1.114889	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	54	1.860596	8.193906	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	220	287	0.00424	0.029211	0.05345	572.506531	572.506531	572.506531	0	0	1	-360	360;
	220	973	6.176859	17.47149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	220	10.52665	27.33264	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	4355	0.202	1.377087	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	6888	0.19876	1.343988	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	6888	0.00545	0.042341	0.06601	563.604795	563.604795	563.604795	0	0	1	-360	360;
	4935	220	7.300207	16.26798	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	220	6.816736	17.84773	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	5823	3.997727	8.729958	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	1441	15.84711	34.44008	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	990	30.61777	53.27893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	220	3564	36.28512	62.88017	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	54	2470	0.00082	0.01007	0.24211	1582.60004	1582.60004	1582.60004	0	0	1	-360	360;
	6757	6991	0.0005	0.004169	0.00631	573.593289	573.593289	573.593289	0	0	1	-360	360;
	4935	6757	0.00525	0.033151	0.06338	505.114201	505.114201	505.114201	0	0	1	-360	360;
	6991	4221	0.001969	0.01718	0.02496	510.262584	510.262584	510.262584	0	0	1	-360	360;
	4221	3905	0.001289	0.009229	0.01637	561.30134	561.30134	561.30134	0	0	1	-360	360;
	3905	3586	0.00043	0.00349	0.0055	561.38098	561.38098	561.38098	0	0	1	-360	360;
	3586	6092	0.000979	0.008901	0.01234	510.425674	510.425674	510.425674	0	0	1	-360	360;
	6092	287	0.000219	0.00155	0.00272	532.059473	532.059473	532.059473	0	0	1	-360	360;
	6383	6472	0.00025	0.00319	0.10266	909.210869	909.210869	909.210869	0	0	1	-360	360;
	6383	6472	0.00024	0.00295	0.10043	858.218427	858.218427	858.218427	0	0	1	-360	360;
	7640	663	0.00022	0.00171	0.07373	1465.3533	1465.3533	1465.3533	0	0	1	-360	360;
	3910	663	0.0005	0.00382	0.14859	1543.64833	1543.64833	1543.64833	0	0	1	-360	360;
	5780	2304	0.0002	0.00192	0.06748	1270.79813	1270.79813	1270.79813	0	0	1	-360	360;
	3910	2304	0.0005	0.00383	0.14829	1588.12941	1588.12941	1588.12941	0	0	1	-360	360;
	5082	3731	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	3731	0.001	0.268678	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5082	0.042159	0.956198	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5082	0.002341	0.022341	0.07898	301.176373	301.176373	301.176373	0	0	1	-360	360;
	3111	5082	0.04555	0.93969	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	2684	0.077909	1.08686	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4029	5082	0.0003	0.003289	0.01013	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7416	5082	1.527831	6.098554	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	5082	0.958926	2.68595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	5082	28.84298	48.1157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	5082	33.55372	46.46901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	5082	47.90702	63.74587	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	5082	28.68802	30.77066	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5082	4.526653	11.10372	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	1441	6.754959	11.13905	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	6035	0.105229	1.236012	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5082	5721	0.085781	1.020186	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	3731	0.044289	0.919483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	3731	0.047781	0.901963	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	2684	0.076341	1.006446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6518	3731	0.0003	0.00336	0.00993	621.673184	621.673184	621.673184	0	0	1	-360	360;
	7416	3731	1.621322	5.733678	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	3731	0.903202	2.852273	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	3731	30.65289	49.86157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	3731	32.78719	47.26653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	3731	44.17355	62.70041	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	3731	25.07438	35.58471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	3731	5.060744	9.753098	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	1441	6.015083	12.60103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	6035	0.099409	1.193017	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	6035	0.002021	0.01812	0.07377	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3731	5721	0.09157	0.961405	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4222	5240	0.00032	0.00387	0.10059	1353.71358	1353.71358	1353.71358	0	0	1	-360	360;
	4222	3260	0.00038	0.00378	0.1204	1468.15188	1468.15188	1468.15188	0	0	1	-360	360;
	3255	6124	0.00032	0.00394	0.09853	1504.45112	1504.45112	1504.45112	0	0	1	-360	360;
	3255	549	0.00038	0.0037	0.12263	1411.44183	1411.44183	1411.44183	0	0	1	-360	360;
	4222	6785	6.036911	10.78324	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5883	6472	0.0008	0.00945	0.23125	1501.55184	1501.55184	1501.55184	0	0	1	-360	360;
	6472	565	0.00079	0.00828	0.33875	1380.37972	1380.37972	1380.37972	0	0	1	-360	360;
	5883	6472	0.00078	0.00964	0.22644	1432.65232	1432.65232	1432.65232	0	0	1	-360	360;
	6472	565	0.00081	0.00811	0.34526	1318.11069	1318.11069	1318.11069	0	0	1	-360	360;
	570	3111	0	0.000103	0	433.63967	433.63967	433.63967	0	0	1	-360	360;
	570	3111	0.001	0.26281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	2684	0.001	0.461798	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	2684	0.00164	0.020729	0.05245	541.62119	541.62119	541.62119	0	0	1	-360	360;
	7416	570	7.774793	17.76901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	570	3.756405	9.209917	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	973	27.82645	45.00826	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	570	20.96901	31.09711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	570	18.12293	35.08471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	1441	3.360124	8.657025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	6035	0.14569	1.294938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	570	5721	0.125919	1.049607	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	2684	0.001	0.471508	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	2684	0.001599	0.02112	0.05134	232.711765	232.711765	232.711765	0	0	1	-360	360;
	7416	3111	8.624174	16.76653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	3111	4.365909	9.241321	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	973	29.57231	44.15909	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	3111	22.67562	29.81818	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	3111	22.21281	29.92975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	1441	3.288017	9.179959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	6035	0.15644	1.276653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3111	5721	0.124669	1.119298	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	2684	6.748554	16.31901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	2684	3.853099	7.976446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	2684	53.92562	59.04132	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	2684	37.32851	39.04752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	2684	16.76198	30.60331	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2684	1441	6.070868	12.11942	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6035	2684	0.29624	1.60312	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	2684	0.293884	1.293946	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8359	7556	0.15633	3.578947	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	2470	0.242957	1.518213	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	1153	8.069251	24.53047	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8359	3201	0.00029	0.00357	0.12575	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2215	4885	0.00012	0.00131	0.04268	1791.71677	1791.71677	1791.71677	0	0	1	-360	360;
	7416	6094	0	0.000103	0	278.891357	278.891357	278.891357	0	0	1	-360	360;
	7416	6094	0.001	0.168909	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	4274	0.837231	3.56405	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	6801	0.690682	3.987397	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	6801	0.00988	0.05469	0.15166	582.032049	582.032049	582.032049	0	0	1	-360	360;
	3212	7416	5.606818	11.56529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	7416	15.6624	27.83058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	7416	16.76116	26.42355	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	1801	0.039521	0.57	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	1441	25.64463	32.3843	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	6035	12.49401	21.31612	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7416	5721	10.90826	18.38058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	4274	0.367355	2.263636	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	6801	0.394566	2.152479	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	6094	2.95062	6.117354	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	6094	8.41281	14.42025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	6094	8.672314	14.29339	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	1801	0.0192	0.321384	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	1441	12.78037	18.27831	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	6035	6.830371	10.80145	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6094	5721	5.092149	10.9593	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4412	6425	0.00043	0.00522	0.1263	1222.37719	1222.37719	1222.37719	0	0	1	-360	360;
	1153	6425	0.00184	0.01667	0.53893	1595.77021	1595.77021	1595.77021	0	0	1	-360	360;
	1153	6425	0.248089	2.761427	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1894	4445	0.00037	0.00464	0.15876	1496.42594	1496.42594	1496.42594	0	0	1	-360	360;
	1894	5234	0.00059	0.00694	0.25699	1555.57783	1555.57783	1555.57783	0	0	1	-360	360;
	8258	7884	0.00027	0.00348	0.08358	1483.48943	1483.48943	1483.48943	0	0	1	-360	360;
	5189	7884	0.00027	0.00354	0.0818	1549.2343	1549.2343	1549.2343	0	0	1	-360	360;
	8258	5189	0.02114	0.839889	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	8258	30.32756	43.46053	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5646	5189	30.89958	42.53601	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4274	6801	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	6801	0.001	0.238595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	4274	0.256632	1.964752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	4274	0.63343	3.574793	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	4274	0.709752	3.284711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	4274	0.01324	0.090909	0.16004	512.380847	512.380847	512.380847	0	0	1	-360	360;
	1801	4274	0.000591	1.201426	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4274	1441	1.821963	7.541736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	6801	0.210826	2.191942	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	6801	0.746074	3.056198	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	6801	0.740041	3.16219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	6801	0.000841	1.29095	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6801	1441	2.154545	7.118182	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4445	1781	8e-05	0.00105	0.03658	1497.24669	1497.24669	1497.24669	0	0	1	-360	360;
	7473	7556	0.00025	0.00186	0.08844	2126.47335	2126.47335	2126.47335	0	0	1	-360	360;
	973	5083	9.1e-05	0.000581	0.00112	576.483181	576.483181	576.483181	0	0	1	-360	360;
	7177	7556	0.00023	0.00211	0.07879	1904.85187	1904.85187	1904.85187	0	0	1	-360	360;
	4313	7282	8e-05	0.00107	0.03581	1629.22086	1629.22086	1629.22086	0	0	1	-360	360;
	4313	7556	2e-05	0.0002	0.00472	1424.3224	1424.3224	1424.3224	0	0	1	-360	360;
	7001	1588	0.00063	0.00406	0.00826	555.114212	555.114212	555.114212	0	0	1	-360	360;
	7001	973	9.1e-05	0.00056	0.00111	543.27419	543.27419	543.27419	0	0	1	-360	360;
	8047	1886	0.00064	0.003969	0.00842	553.212767	553.212767	553.212767	0	0	1	-360	360;
	973	8047	9.1e-05	0.00055	0.00113	494.295015	494.295015	494.295015	0	0	1	-360	360;
	7556	1781	2e-05	0.0002	0.00535	1596.38363	1596.38363	1596.38363	0	0	1	-360	360;
	7282	4423	0.00091	0.01097	0.39624	1478.88744	1478.88744	1478.88744	0	0	1	-360	360;
	1588	2278	0.00711	0.045841	0.09356	516.302245	516.302245	516.302245	0	0	1	-360	360;
	1886	7886	0.00724	0.044849	0.09529	556.175821	556.175821	556.175821	0	0	1	-360	360;
	2470	7556	5.952838	17.29294	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1153	7556	20.08726	50.82479	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3212	973	4.890702	15.41012	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4355	973	5.639256	15.40868	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	973	4.76281	15.7	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	973	21.2624	53.66529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	973	25.16942	46.14463	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	5823	5.023554	17.83409	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	1441	0.483492	3.46095	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	100	973	0.000211	0.0012	0.00194	435.126912	435.126912	435.126912	0	0	1	-360	360;
	973	6035	9.888017	20.17066	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	973	5721	9.581818	15.8126	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1359	964	0.00015	0.00151	0.04535	1612.41096	1612.41096	1612.41096	0	0	1	-360	360;
	4423	1731	0.00044	0.00443	0.19022	1568.21553	1568.21553	1568.21553	0	0	1	-360	360;
	2278	4590	0.00301	0.01868	0.03962	502.367203	502.367203	502.367203	0	0	1	-360	360;
	7886	4590	0.00268	0.02118	0.0353	494.1083	494.1083	494.1083	0	0	1	-360	360;
	5234	964	0.00044	0.00433	0.19387	1493.89722	1493.89722	1493.89722	0	0	1	-360	360;
	1731	1359	0.00015	0.00148	0.0462	1450.98788	1450.98788	1450.98788	0	0	1	-360	360;
	3212	4355	36.67975	64.23967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	6888	34.68595	57.82851	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	3212	0.001	0.997107	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	3212	0.001	0.990558	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	3212	0.00389	0.03206	0.04613	489.006787	489.006787	489.006787	0	0	1	-360	360;
	3212	1801	2.394421	10.2281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	1441	0.001	0.942996	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	6035	7.162397	16.07149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3212	5721	6.970248	12.65785	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4125	5607	0.00041	0.00489	0.21356	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7965	5607	0.00035	0.00415	0.1885	2116.91658	2116.91658	2116.91658	0	0	1	-360	360;
	2470	3996	0.00041	0.00478	0.21767	2040.85146	2040.85146	2040.85146	0	0	1	-360	360;
	7965	3996	0.00036	0.00406	0.19206	1735.14955	1735.14955	1735.14955	0	0	1	-360	360;
	4125	6747	0.00069	0.00736	0.21961	1841.77855	1841.77855	1841.77855	0	0	1	-360	360;
	3201	4125	0.00035	0.00355	0.15003	2189.45518	2189.45518	2189.45518	0	0	1	-360	360;
	2218	4125	0.0003	0.0037	0.13435	2226.83714	2226.83714	2226.83714	0	0	1	-360	360;
	2470	5127	0.00079	0.00639	0.25232	2186.54142	2186.54142	2186.54142	0	0	1	-360	360;
	1153	2470	10.65028	23.35249	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4476	2470	0.00035	0.00375	0.14437	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	571	2470	0.00035	0.00368	0.1472	2030.2245	2030.2245	2030.2245	0	0	1	-360	360;
	4355	6888	0	0.000103	0	483.283398	483.283398	483.283398	0	0	1	-360	360;
	4355	6888	0.001	0.307397	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	4355	30.07851	44.4938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	4355	28.13843	48.89876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4355	5823	1.021095	4.339256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4355	1441	19.04773	44.46694	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4355	990	11.71198	25.83058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4355	3564	13.4062	28.03306	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	6888	25.77686	44.3843	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	6888	29.27686	39.66322	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	5823	0.90343	4.251446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	1441	15.67913	46.04752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	990	10.06777	25.78512	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6888	3564	11.7936	27.51653	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4303	5127	0.00037	0.00319	0.12089	1766.88169	1766.88169	1766.88169	0	0	1	-360	360;
	2661	6747	0.00037	0.0032	0.12142	1685.67882	1685.67882	1685.67882	0	0	1	-360	360;
	2971	8247	6e-05	0.00057	0.01778	2289.24043	2289.24043	2289.24043	0	0	1	-360	360;
	4927	2971	0.00081	0.00793	0.25734	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4935	2361	0	0.000103	0	534.079806	534.079806	534.079806	0	0	1	-360	360;
	4935	2361	0.001	0.218595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	1801	5.508884	16.92169	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	1441	0.417335	3.096901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	6035	35.56405	45.20868	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4935	5721	26.29545	47.22107	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	1801	5.134091	18.53492	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	1441	0.373512	3.536364	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	6035	31.99174	51.7624	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2361	5721	26.61157	47.82645	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5823	1441	22.02479	61.94007	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	990	5823	0.20395	1.135248	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3564	5823	0.207314	1.339256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	1441	15.82955	28.76446	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	6035	27.39876	40.44215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1801	5721	23.59711	35.69835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6035	1441	1.210248	3.369008	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5721	1441	0.914421	3.375826	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3442	1143	0.005729	0.02032	0.06514	418.132411	418.132411	418.132411	0	0	1	-360	360;
	1479	1143	0.005841	0.019909	0.06641	440.798859	440.798859	440.798859	0	0	1	-360	360;
	3720	1143	0.084781	0.493471	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5317	3691	0	4e-05	0.00136	1351.23948	1351.23948	1351.23948	0	0	1	-360	360;
	5317	3691	0.001	0.19356	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5317	7427	0.00192	0.01356	0.6072	2287.0974	2287.0974	2287.0974	0	0	1	-360	360;
	3691	7427	0.00172	0.01537	0.54096	2157.07079	2157.07079	2157.07079	0	0	1	-360	360;
	1755	3831	0.00035	0.00346	0.15707	1936.44777	1936.44777	1936.44777	0	0	1	-360	360;
	990	3564	0	0.000103	0	542.642026	542.642026	542.642026	0	0	1	-360	360;
	990	3564	0.001	0.159719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	990	5490	0.000341	0.00199	0.00559	594.646748	594.646748	594.646748	0	0	1	-360	360;
	5490	6825	0.00219	0.012901	0.05905	267.521171	267.521171	267.521171	0	0	1	-360	360;
	6825	4018	0.002531	0.013539	0.07283	250.025206	250.025206	250.025206	0	0	1	-360	360;
	4018	6544	0.002281	0.01263	0.06169	298.140919	298.140919	298.140919	0	0	1	-360	360;
	6544	100	0.00232	0.01393	0.06281	238.421817	238.421817	238.421817	0	0	1	-360	360;
	5780	1861	0.00043	0.00418	0.14721	1826.45613	1826.45613	1826.45613	0	0	1	-360	360;
	7088	7640	0.113102	0.936981	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7640	1060	0.00043	0.00419	0.14734	1995.81093	1995.81093	1995.81093	0	0	1	-360	360;
	4927	7640	0.124889	1.006302	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2364	8241	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2364	8241	0.003651	0.01288	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8241	6638	0.004021	0.01951	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7088	1048	0.00057	0.00402	0.18006	2104.26967	2104.26967	2104.26967	0	0	1	-360	360;
	3683	4927	0.00064	0.00448	0.20032	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7088	4927	0.00646	0.473407	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2968	1861	8e-05	0.00061	0.02325	900.438655	900.438655	900.438655	0	0	1	-360	360;
	706	1060	6e-05	0.00065	0.01998	1108.52402	1108.52402	1108.52402	0	0	1	-360	360;
	3720	3442	0.00012	0.000461	0.00152	407.259878	407.259878	407.259878	0	0	1	-360	360;
	3720	1479	0.00013	0.00045	0.00155	377.156073	377.156073	377.156073	0	0	1	-360	360;
	6035	5721	0	0.000103	0	297.538858	297.538858	297.538858	0	0	1	-360	360;
	6035	5721	0.001	0.249814	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7792	6941	0.0001	0.00075	0.77789	1245.8867	1245.8867	1245.8867	0	0	1	-360	360;
	7792	6792	0.0038	0.11597	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7792	6792	0.0001	0.00081	0.40407	1283.72543	1283.72543	1283.72543	0	0	1	-360	360;
	6179	6941	0.0001	0.00094	0.58389	1300.27798	1300.27798	1300.27798	0	0	1	-360	360;
	6179	6941	9e-05	0.00107	0.5335	1181.67078	1181.67078	1181.67078	0	0	1	-360	360;
	52	6179	7e-05	0.00085	0.39053	1150.20783	1150.20783	1150.20783	0	0	1	-360	360;
	52	6179	8e-05	0.00081	0.48117	1284.75284	1284.75284	1284.75284	0	0	1	-360	360;
	6941	6792	0.00018	0.00171	1.06574	1104.97117	1104.97117	1104.97117	0	0	1	-360	360;
	52	2292	0.00017	0.00191	0.05127	1415.82135	1415.82135	1415.82135	0	0	1	-360	360;
	52	2292	0.00016	0.00194	0.05023	1373.97168	1373.97168	1373.97168	0	0	1	-360	360;
	52	7199	0.0005	0.00317	0.1002	1526.82988	1526.82988	1526.82988	0	0	1	-360	360;
	52	7199	0.00048	0.00322	0.09807	1552.75885	1552.75885	1552.75885	0	0	1	-360	360;
	6792	3960	0.00427	0.19732	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6792	3960	0.0001	0.00122	0.95678	1164.8793	1164.8793	1164.8793	0	0	1	-360	360;
	6792	3960	0.0001	0.00119	0.97672	1266.23036	1266.23036	1266.23036	0	0	1	-360	360;
	3960	4094	0.0024	0.02048	0.84984	1552.19742	1552.19742	1552.19742	0	0	1	-360	360;
	3960	177	0.00042	0.00474	0.14676	1897.74287	1897.74287	1897.74287	0	0	1	-360	360;
	3281	1909	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3281	1909	0.003409	0.129831	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3281	7023	0.00605	0.024031	0.07065	468.452886	468.452886	468.452886	0	0	1	-360	360;
	1909	7023	0.005919	0.0245	0.06917	433.784088	433.784088	433.784088	0	0	1	-360	360;
	2612	5623	0.406364	1.478988	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2612	358	0.12912	0.95312	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2612	4523	0.23062	0.90281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1021	2612	0.05107	0.366901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2612	358	0.00207	0.0168	0.08303	315.543146	315.543146	315.543146	0	0	1	-360	360;
	2612	1482	0.00119	0.009659	0.04779	529.761353	529.761353	529.761353	0	0	1	-360	360;
	3216	2612	0.120151	0.700041	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2612	5323	0.032581	0.326756	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3380	5031	2e-05	0.0002	0.00601	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3380	870	0.00025	0.00344	0.12743	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3380	870	0.01671	0.288767	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3380	2341	0.01954	0.281191	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	3380	0.00356	0.222701	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	3380	0.00271	0.199571	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4213	5031	0.00016	0.00253	0.08027	944.139787	944.139787	944.139787	0	0	1	-360	360;
	5693	5031	0.00018	0.00214	0.0555	1277.57737	1277.57737	1277.57737	0	0	1	-360	360;
	5623	358	0.002091	0.0138	0.06906	281.19443	281.19443	281.19443	0	0	1	-360	360;
	5623	358	0.291467	1.202521	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4523	5623	0.204719	0.794277	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1021	5623	0.026401	0.214814	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1021	5623	0.00295	0.01256	0.04253	439.779166	439.779166	439.779166	0	0	1	-360	360;
	4838	7087	0.359525	1.423657	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3796	7087	0.32907	1.339607	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3796	7087	0.00469	0.03913	0.18279	334.283623	334.283623	334.283623	0	0	1	-360	360;
	2414	7087	0.000841	0.00544	0.03236	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2414	7087	0.014581	0.16557	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4370	7087	1.038884	1.650227	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6898	7087	0.863926	1.96155	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	48	358	0.005081	0.02426	0.0791	434.420444	434.420444	434.420444	0	0	1	-360	360;
	1480	48	0.470062	1.363636	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	48	6466	0.035969	0.35219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	48	8316	0.0033	0.01864	0.05088	483.406478	483.406478	483.406478	0	0	1	-360	360;
	48	8316	0.018021	0.1825	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4629	5711	0.002651	0.027659	0.08744	256.645589	256.645589	256.645589	0	0	1	-360	360;
	4643	4629	0.00226	0.04061	0.06212	532.284674	532.284674	532.284674	0	0	1	-360	360;
	4629	7513	0.000711	0.0098	0.02356	267.498308	267.498308	267.498308	0	0	1	-360	360;
	2090	4629	0.002211	0.037599	0.06077	255.407935	255.407935	255.407935	0	0	1	-360	360;
	3743	2196	0.0016	0.01356	0.51826	1323.36427	1323.36427	1323.36427	0	0	1	-360	360;
	3743	2196	0.00165	0.01328	0.52819	1299.52699	1299.52699	1299.52699	0	0	1	-360	360;
	5693	870	0.01336	0.25865	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	870	0.00015	0.00253	0.07483	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8036	1516	0.00225	0.02106	0.71123	1642.5865	1642.5865	1642.5865	0	0	1	-360	360;
	3848	1516	0.00426	0.03506	1.32273	1706.16728	1706.16728	1706.16728	0	0	1	-360	360;
	6148	1516	0.00043	0.00581	0.20816	2025.79758	2025.79758	2025.79758	0	0	1	-360	360;
	4523	3593	0.001151	0.009909	0.0435	435.584174	435.584174	435.584174	0	0	1	-360	360;
	1021	3593	0.000961	0.00751	0.02752	615.698682	615.698682	615.698682	0	0	1	-360	360;
	771	3593	0.00032	0.00324	0.01186	447.740468	447.740468	447.740468	0	0	1	-360	360;
	8036	1516	0.00245	0.02191	0.76733	1080.14596	1080.14596	1080.14596	0	0	1	-360	360;
	1028	1516	0.00037	0.00606	0.18033	930.758343	930.758343	930.758343	0	0	1	-360	360;
	6327	1516	0.0029	0.0289	0.89272	1763.74283	1763.74283	1763.74283	0	0	1	-360	360;
	6399	358	0.000151	0.00156	0.00506	275.154388	275.154388	275.154388	0	0	1	-360	360;
	8248	358	0.00014	0.00144	0.00495	291.890918	291.890918	291.890918	0	0	1	-360	360;
	4523	358	0.026841	0.406942	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1021	358	0.025909	0.304649	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2383	749	0	0.00012	0	2059.06196	2059.06196	2059.06196	0	0	1	-360	360;
	2383	749	0.001	0.18678	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7662	2383	0.989958	2.675	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7518	2383	0.994529	2.704017	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	2383	0.154938	0.902078	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3014	2383	0.141247	1.002008	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8271	2383	0.01709	0.303539	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2383	6730	0.00081	0.00826	0.24179	1561.01439	1561.01439	1561.01439	0	0	1	-360	360;
	2383	3743	0.00174	0.01199	0.54183	1487.91766	1487.91766	1487.91766	0	0	1	-360	360;
	2383	5619	0.00136	0.012	0.42329	1042.47967	1042.47967	1042.47967	0	0	1	-360	360;
	7662	749	1.135249	2.306648	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7518	749	1.077147	2.483726	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	749	0.140319	0.988089	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3014	749	0.15108	0.940582	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8271	749	0.00045	0.00442	0.15175	1387.82899	1387.82899	1387.82899	0	0	1	-360	360;
	8271	749	0.01578	0.332528	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4290	749	0.00032	0.00308	0.10147	2062.21728	2062.21728	2062.21728	0	0	1	-360	360;
	294	4213	0.00018	0.0025	0.08139	1961.13081	1961.13081	1961.13081	0	0	1	-360	360;
	2635	4213	0.03152	0.366053	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1480	3424	0.00251	0.13469	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1480	4838	0.00943	0.04787	0.12183	433.07076	433.07076	433.07076	0	0	1	-360	360;
	1480	6466	0.002781	0.02494	0.10699	282.302527	282.302527	282.302527	0	0	1	-360	360;
	1480	6466	0.22876	0.857376	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1480	8316	0.113521	0.454752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8036	6912	0.00159	0.01603	0.49748	1446.82423	1446.82423	1446.82423	0	0	1	-360	360;
	3424	3796	0.00924	0.04881	0.11935	501.222907	501.222907	501.222907	0	0	1	-360	360;
	3424	8316	0.00643	0.027271	0.09824	546.045575	546.045575	546.045575	0	0	1	-360	360;
	8036	3848	0.00183	0.01368	0.57847	1839.40647	1839.40647	1839.40647	0	0	1	-360	360;
	7432	4868	0	0.00012	0	984.488811	984.488811	984.488811	0	0	1	-360	360;
	7432	4868	0.001	0.191392	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2959	7432	0.00081	0.00981	0.35587	1554.98876	1554.98876	1554.98876	0	0	1	-360	360;
	7432	4500	0.090907	0.571413	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7432	4500	0.00058	0.00647	0.2001	1573.88706	1573.88706	1573.88706	0	0	1	-360	360;
	7432	8143	0.441572	1.595429	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7432	4403	0.0463	0.493158	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7432	5419	0.053	0.456427	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5419	4868	0.00048	0.00707	0.21018	1527.22018	1527.22018	1527.22018	0	0	1	-360	360;
	4868	4500	0.0007	0.00626	0.21496	1787.73084	1787.73084	1787.73084	0	0	1	-360	360;
	4868	4500	0.099273	0.491122	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4868	8143	0.411558	1.466205	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4868	4403	0.05226	0.428179	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4868	5419	0.04979	0.473137	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6503	1297	0.00018	0.00218	0.05434	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	3490	3976	0.344619	1.353116	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	2166	0.581787	1.764958	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	7187	0.075831	0.835526	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	7187	0.00095	0.0105	0.32475	1500.55273	1500.55273	1500.55273	0	0	1	-360	360;
	3490	1898	0.081281	0.782202	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	6730	0.06719	0.554377	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	5619	0.00409	0.271641	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3490	5619	0.00041	0.00351	0.13804	1478.26875	1478.26875	1478.26875	0	0	1	-360	360;
	7220	590	0.00138	0.008789	0.02153	551.481271	551.481271	551.481271	0	0	1	-360	360;
	7220	3001	0.002331	0.01832	0.04663	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3976	2166	0	0.00011	0	1905.66933	1905.66933	1905.66933	0	0	1	-360	360;
	3976	2166	0.001	0.311717	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1448	3976	0.00043	0.00408	0.14672	1851.35045	1851.35045	1851.35045	0	0	1	-360	360;
	3976	7187	0.083899	0.525478	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3976	1898	0.07124	0.605921	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3976	1297	0.00013	0.00147	0.03966	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	3976	6730	2.055194	4.798753	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8143	3976	0.00048	0.00587	0.16772	1771.95691	1771.95691	1771.95691	0	0	1	-360	360;
	5619	3976	0.81115	2.560526	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	918	7220	0.00095	0.004169	0.01466	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7355	7220	0.01012	0.04631	0.08989	498.22975	498.22975	498.22975	0	0	1	-360	360;
	2731	2166	0.00013	0.0014	0.04059	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1078	2166	0.00013	0.00143	0.03973	1444.25799	1444.25799	1444.25799	0	0	1	-360	360;
	2166	7187	0.091447	0.726385	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2166	7187	0.00093	0.00869	0.29632	998.478412	998.478412	998.478412	0	0	1	-360	360;
	2166	1898	0.089307	0.727078	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2166	6730	3.590651	5.504778	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2635	2166	0.00047	0.00544	0.16411	1609.8592	1609.8592	1609.8592	0	0	1	-360	360;
	5619	2166	1.2741	3.747645	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7227	3831	0.00142	0.01566	0.6303	1433.25718	1433.25718	1433.25718	0	0	1	-360	360;
	1898	6047	0.00082	0.00794	0.37094	2117.22001	2117.22001	2117.22001	0	0	1	-360	360;
	7662	7518	0	0.00011	0	1429.20411	1429.20411	1429.20411	0	0	1	-360	360;
	7662	7518	0.001	0.149882	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	7662	0.088033	0.626579	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	7662	0.001	0.00866	0.22327	1299.18803	1299.18803	1299.18803	0	0	1	-360	360;
	3014	7662	0.090852	0.619418	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8271	7662	0.87133	2.390028	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	7518	0.087632	0.633012	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3014	7518	0.097832	0.57669	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8271	7518	0.857064	2.463089	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7518	3743	0.0004	0.0039	0.09838	1107.03522	1107.03522	1107.03522	0	0	1	-360	360;
	4838	3796	0.001	0.486508	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2414	4838	0.003909	0.034711	0.15186	592.263273	592.263273	592.263273	0	0	1	-360	360;
	2414	4838	0.497789	2.055971	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4838	4370	0.357087	1.677376	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4838	6898	0.349194	1.694876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2414	3796	0.438409	2.013802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3796	4370	0.307107	1.517831	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3796	6898	0.354174	1.443988	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7817	2258	0.00274	0.144159	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7817	5973	0.00039	0.00175	0.006	224.526023	224.526023	224.526023	0	0	1	-360	360;
	2258	8094	0.000401	0.001711	0.00611	463.610107	463.610107	463.610107	0	0	1	-360	360;
	7227	4639	0.06537	0.54912	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7227	1448	0.00098	0.00852	0.31768	1105.01395	1105.01395	1105.01395	0	0	1	-360	360;
	7227	1448	0.076039	0.582729	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7227	1821	0.00185	0.01418	0.59674	1070.02382	1070.02382	1070.02382	0	0	1	-360	360;
	2619	7227	0.790374	2.170014	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1527	590	0.00188	0.0087	0.02894	503.76528	503.76528	503.76528	0	0	1	-360	360;
	1527	3218	0.042461	0.437066	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	375	3218	0.00068	0.008531	0.0272	490.127458	490.127458	490.127458	0	0	1	-360	360;
	3490	7187	0.00107	0.01108	0.32435	1955.39346	1955.39346	1955.39346	0	0	1	-360	360;
	3490	6730	0.00092	0.00947	0.27717	1722.87187	1722.87187	1722.87187	0	0	1	-360	360;
	3787	3014	0	0.00011	0	1685.70317	1685.70317	1685.70317	0	0	1	-360	360;
	3787	3014	0.001	0.196378	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3787	8271	0.102652	0.915512	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3014	8271	0.110602	0.871537	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5506	3014	0.00153	0.01371	0.31254	1320.4288	1320.4288	1320.4288	0	0	1	-360	360;
	8079	3014	0.00296	0.02271	0.63138	1081.6341	1081.6341	1081.6341	0	0	1	-360	360;
	4639	1448	0	0.0001	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4639	1448	0.001	0.229661	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4639	1821	0.00076	0.0066	0.2459	1099.3694	1099.3694	1099.3694	0	0	1	-360	360;
	2619	4639	0.134051	0.802355	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2619	1448	0.136413	0.960942	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7724	8094	0.0027	0.02195	0.07162	502.077604	502.077604	502.077604	0	0	1	-360	360;
	7724	3001	0.001211	0.004961	0.01837	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1954	6399	0.010091	0.211612	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1954	8248	0.008349	0.215682	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1021	1954	0.000401	0.001781	0.00535	256.547278	256.547278	256.547278	0	0	1	-360	360;
	6399	8248	0.00076	0.205461	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8143	6148	0.00086	0.0138	0.37901	1716.69025	1716.69025	1716.69025	0	0	1	-360	360;
	2635	1028	0.00101	0.01174	0.44381	1630.14033	1630.14033	1630.14033	0	0	1	-360	360;
	7187	1898	0	0.00011	0	1483.51839	1483.51839	1483.51839	0	0	1	-360	360;
	7187	1898	0.001	0.17286	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7187	6730	0.917036	2.770706	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5619	7187	0.299162	1.508587	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1898	6730	0.973546	2.445291	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5619	1898	0.335817	1.383795	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	709	4370	0.00114	0.00625	0.01594	556.396069	556.396069	556.396069	0	0	1	-360	360;
	4564	709	0.00069	0.00343	0.02608	313.778116	313.778116	313.778116	0	0	1	-360	360;
	6504	709	0.000669	0.00349	0.02555	323.7628	323.7628	323.7628	0	0	1	-360	360;
	1374	709	0.00264	0.01612	0.0693	546.524557	546.524557	546.524557	0	0	1	-360	360;
	2959	3848	0.001	0.01303	0.4362	1767.61622	1767.61622	1767.61622	0	0	1	-360	360;
	862	3848	0.00018	0.00112	0.06237	831.227152	831.227152	831.227152	0	0	1	-360	360;
	5464	1226	0.000771	0.0027	0.00545	426.880584	426.880584	426.880584	0	0	1	-360	360;
	1075	1226	0.000719	0.003039	0.00495	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1226	6898	0.00138	0.005409	0.0184	509.84076	509.84076	509.84076	0	0	1	-360	360;
	1682	6912	0.00017	0.0011	0.06071	1764.1305	1764.1305	1764.1305	0	0	1	-360	360;
	1021	4523	0.01424	0.207211	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6466	4523	0.00288	0.023031	0.11118	357.676736	357.676736	357.676736	0	0	1	-360	360;
	8271	5693	0.00059	0.00738	0.19892	1229.72313	1229.72313	1229.72313	0	0	1	-360	360;
	3793	4403	2e-05	0.0002	0.0083	1270.38742	1270.38742	1270.38742	0	0	1	-360	360;
	2959	5990	0.00034	0.00243	0.122	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	2341	0.00939	0.239481	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	2341	0.0003	0.00385	0.14929	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5712	7513	0.00082	0.0117	0.02754	379.840959	379.840959	379.840959	0	0	1	-360	360;
	5712	5711	0.000831	0.010711	0.03388	239.563827	239.563827	239.563827	0	0	1	-360	360;
	4210	1948	0	0.00011	0	1164.35013	1164.35013	1164.35013	0	0	1	-360	360;
	4210	1948	0.001	0.424958	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4210	2635	0.00035	0.00315	0.10812	2077.42517	2077.42517	2077.42517	0	0	1	-360	360;
	4210	2635	0.01362	0.396288	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4210	5693	0.088213	0.739543	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1948	2635	0.01388	0.388199	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1948	5693	0.0012	0.01148	0.36391	1868.41638	1868.41638	1868.41638	0	0	1	-360	360;
	1948	5693	0.086413	0.754086	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	294	8143	0.00049	0.00475	0.16449	1570.94632	1570.94632	1570.94632	0	0	1	-360	360;
	2630	5100	0.002669	0.01443	0.04143	432.824982	432.824982	432.824982	0	0	1	-360	360;
	3218	590	0.00338	0.01488	0.05173	572.347252	572.347252	572.347252	0	0	1	-360	360;
	3216	8094	0.00101	0.005841	0.01811	268.664324	268.664324	268.664324	0	0	1	-360	360;
	6148	1821	0.00034	0.00381	0.15065	1760.88107	1760.88107	1760.88107	0	0	1	-360	360;
	2619	1821	0.00017	0.00223	0.05319	1702.52031	1702.52031	1702.52031	0	0	1	-360	360;
	5323	5973	0.00161	0.013789	0.06062	553.814066	553.814066	553.814066	0	0	1	-360	360;
	1821	7181	0.00032	0.00234	0.11759	1781.359	1781.359	1781.359	0	0	1	-360	360;
	5619	6730	0.01246	0.249882	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4500	2635	0.00073	0.00838	0.25339	1518.51444	1518.51444	1518.51444	0	0	1	-360	360;
	4500	8143	0.01019	0.13124	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4500	8143	0.00078	0.00873	0.26973	1606.63477	1606.63477	1606.63477	0	0	1	-360	360;
	4500	4403	0.463179	1.173823	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4500	5419	0.428629	1.321537	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5693	2635	0.099591	0.647687	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7014	2635	0.00019	0.00245	0.07706	1554.71759	1554.71759	1554.71759	0	0	1	-360	360;
	8143	4403	1.591828	3.383587	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8143	5419	1.558587	3.254017	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2414	4370	1.410971	2.609298	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2414	6898	1.456178	2.474793	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4370	6898	0	0.000103	0	504.348669	504.348669	504.348669	0	0	1	-360	360;
	4370	6898	0.001	0.139341	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4403	5419	0	0.00012	0	1119.81508	1119.81508	1119.81508	0	0	1	-360	360;
	4403	5419	0.001	0.196801	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7046	1482	0.00025	0.00232	0.01016	257.22593	257.22593	257.22593	0	0	1	-360	360;
	5323	7046	0.000719	0.005461	0.02933	285.665304	285.665304	285.665304	0	0	1	-360	360;
	1482	7941	0.00012	0.00069	0	234.743911	234.743911	234.743911	0	0	1	-360	360;
	5932	1482	0.003099	0.11999	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6466	8316	0	0.000103	0	232.419117	232.419117	232.419117	0	0	1	-360	360;
	6466	8316	0.001	0.13145	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3216	5323	0.10799	0.554256	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7553	7755	0.00077	0.00927	0.40578	1762.08883	1762.08883	1762.08883	0	0	1	-360	360;
	5929	7755	0.00078	0.00908	0.41358	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4422	6254	0.00038	0.0045	0.20244	790.420693	790.420693	790.420693	0	0	1	-360	360;
	6254	5860	0.00147	0.0125	0.4929	1793.59127	1793.59127	1793.59127	0	0	1	-360	360;
	6048	6254	0.00039	0.00441	0.20633	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6254	1129	0.00202	0.01654	0.67923	1615.26745	1615.26745	1615.26745	0	0	1	-360	360;
	5929	7553	0	3.5e-05	0	1464.00469	1464.00469	1464.00469	0	0	1	-360	360;
	5929	7553	0.00147	0.112341	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5929	4422	0.00069	0.00564	0.18589	1642.26334	1642.26334	1642.26334	0	0	1	-360	360;
	7553	6048	0.00068	0.00574	0.18202	1600.79475	1600.79475	1600.79475	0	0	1	-360	360;
	4422	6048	0	3.5e-05	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4422	6048	0.00088	0.07777	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	534	4422	0.00062	0.00515	0.18247	1212.0576	1212.0576	1212.0576	0	0	1	-360	360;
	2175	6048	0.0007	0.00505	0.2068	1205.84044	1205.84044	1205.84044	0	0	1	-360	360;
	534	2175	0	3.5e-05	0	1095.72046	1095.72046	1095.72046	0	0	1	-360	360;
	534	2175	0.00076	0.06481	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2811	414	0.00011	0.00149	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5347	1763	9.9e-05	0.000979	0	280.003645	280.003645	280.003645	0	0	1	-360	360;
	4402	3585	0.01788	0.227831	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4402	3585	0.004669	0.015969	0.03215	424.005172	424.005172	424.005172	0	0	1	-360	360;
	6196	4402	0.011229	0.03868	0.07741	383.319575	383.319575	383.319575	0	0	1	-360	360;
	349	4402	0.01001	0.043841	0.06896	395.680113	395.680113	395.680113	0	0	1	-360	360;
	4402	430	0.002151	0.01032	0.0213	374.875862	374.875862	374.875862	0	0	1	-360	360;
	3876	5349	0.003281	0.017901	0.04863	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2214	3876	0.005271	0.026909	0.07367	474.902177	474.902177	474.902177	0	0	1	-360	360;
	3876	7659	0.00044	0.002169	0.00639	505.996716	505.996716	505.996716	0	0	1	-360	360;
	3876	7659	0.00045	0.00213	0.00651	505.418661	505.418661	505.418661	0	0	1	-360	360;
	2865	5651	0.452251	2.097438	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5651	1103	0.247659	1.380679	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	5651	0.436378	1.290097	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5018	5651	0.218089	1.220914	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4352	5651	0.0612	1.024169	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	5651	0.06168	0.959765	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	5651	0.0011	0.00838	0.33964	1622.54758	1622.54758	1622.54758	0	0	1	-360	360;
	3910	5651	0.05459	0.962258	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5018	5651	0.00126	0.01278	0.39143	1517.42844	1517.42844	1517.42844	0	0	1	-360	360;
	7744	1471	5e-05	0.000229	0.00081	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	7744	0.001331	0.009331	0.02305	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7744	8349	6e-05	0.00031	0.00084	473.467901	473.467901	473.467901	0	0	1	-360	360;
	4179	5820	0.00155	0.00805	0.02726	370.930458	370.930458	370.930458	0	0	1	-360	360;
	5820	140	0.351674	1.414483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	5820	0.239545	1.416095	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4754	1763	0.00011	0.00088	0.10721	399.839287	399.839287	399.839287	0	0	1	-360	360;
	1478	414	0.00012	0.00064	0.13346	303.802959	303.802959	303.802959	0	0	1	-360	360;
	7447	414	0.002349	0.01088	0.01892	449.418237	449.418237	449.418237	0	0	1	-360	360;
	7447	414	0.0023	0.011099	0.01854	433.166785	433.166785	433.166785	0	0	1	-360	360;
	414	771	0.004771	0.03713	0.16488	347.748067	347.748067	347.748067	0	0	1	-360	360;
	7324	1763	0.00056	0.0013	0.12734	365.842282	365.842282	365.842282	0	0	1	-360	360;
	5245	1763	0.001091	0.007409	0.02704	453.522539	453.522539	453.522539	0	0	1	-360	360;
	5245	1763	0.00168	0.007479	0.02676	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1304	1763	0.003151	0.01876	0.05102	526.960627	526.960627	526.960627	0	0	1	-360	360;
	1055	2399	0.005719	0.255764	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2871	1055	0.000599	0.00294	0.00951	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2871	2399	0.00062	0.00288	0.00969	568.595041	568.595041	568.595041	0	0	1	-360	360;
	5149	4351	0.001169	0.012591	0.03659	351.916386	351.916386	351.916386	0	0	1	-360	360;
	5149	4351	0.00131	0.01119	0.04134	268.996982	268.996982	268.996982	0	0	1	-360	360;
	5811	4351	0.000159	0.00137	0.00499	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3737	4351	0.000159	0.001341	0.00509	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7721	1176	0.00051	0.0025	0.04207	311.17668	311.17668	311.17668	0	0	1	-360	360;
	4067	6437	0.00031	0.001289	0.11495	469.307965	469.307965	469.307965	0	0	1	-360	360;
	7721	6437	0.05163	0.55124	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6945	6437	0.201229	1.043864	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7029	5245	1.027128	2.385331	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7029	148	0.288306	1.382335	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6345	7029	0.05949	0.340661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7029	2370	0.007211	0.030409	0.10181	248.800508	248.800508	248.800508	0	0	1	-360	360;
	7836	7029	0.899153	2.660744	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5068	7029	0.809979	2.92686	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7029	4837	0.00251	0.011211	0.0404	423.976593	423.976593	423.976593	0	0	1	-360	360;
	7355	4359	0.08695	0.512231	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7355	7981	0.006289	0.693946	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1511	7355	0.140669	0.495269	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7964	7355	0.00882	0.041831	0.07707	493.394972	493.394972	493.394972	0	0	1	-360	360;
	6327	1669	2e-05	0.00025	0.00757	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4684	4038	0.000159	0.00074	0.00233	215.929128	215.929128	215.929128	0	0	1	-360	360;
	913	6041	0.004469	0.139591	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	913	2190	0.001039	0.006211	0.01297	447.01228	447.01228	447.01228	0	0	1	-360	360;
	6735	8255	0.00813	0.186659	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6735	2190	0.00101	0.00625	0.01246	407.658076	407.658076	407.658076	0	0	1	-360	360;
	5390	516	8.1e-05	0.000781	0.00238	456.728704	456.728704	456.728704	0	0	1	-360	360;
	5390	6301	0.0003	0.002031	0.0041	494.527837	494.527837	494.527837	0	0	1	-360	360;
	6464	516	9.1e-05	0.000719	0.00278	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6464	7362	0.00086	0.005581	0.02151	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4070	1103	0.00062	0.00605	0.19445	874.472165	874.472165	874.472165	0	0	1	-360	360;
	4064	516	0.00181	0.009599	0.03449	301.696127	301.696127	301.696127	0	0	1	-360	360;
	5735	516	0.00099	0.00481	0.01427	647.787002	647.787002	647.787002	0	0	1	-360	360;
	516	4092	0.00038	0.00218	0.00626	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3324	1274	0.00031	0.00342	0.1086	1785.21198	1785.21198	1785.21198	0	0	1	-360	360;
	6455	1274	0.00083	0.00929	0.26519	1441.82536	1441.82536	1441.82536	0	0	1	-360	360;
	2955	1274	0.00074	0.00793	0.30117	2107.86925	2107.86925	2107.86925	0	0	1	-360	360;
	2835	2955	0.00016	0.00166	0.05189	1849.69776	1849.69776	1849.69776	0	0	1	-360	360;
	1274	7421	0	0.00012	0	1135.45211	1135.45211	1135.45211	0	0	1	-360	360;
	1274	7421	0.01471	0.443012	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1205	7421	0.00037	0.00337	0.11641	1868.46311	1868.46311	1868.46311	0	0	1	-360	360;
	6196	2632	0.001461	0.006711	0.02463	271.809521	271.809521	271.809521	0	0	1	-360	360;
	1696	2632	0.001479	0.00656	0.02509	604.26829	604.26829	604.26829	0	0	1	-360	360;
	6546	4359	0.00068	0.0113	0.04379	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4441	2835	0.00019	0.00196	0.0695	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2497	2835	0.00019	0.00198	0.06439	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1205	4885	0.00886	0.294079	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1205	4885	0.00028	0.00275	0.09336	1731.53746	1731.53746	1731.53746	0	0	1	-360	360;
	3324	4885	0.01321	0.296717	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7565	7999	0.001	0.169901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7565	140	0.03111	0.255	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	7565	0.00057	0.002849	0.00853	238.87603	238.87603	238.87603	0	0	1	-360	360;
	7999	140	0.03143	0.234731	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	7999	0.000581	0.002789	0.00869	554.238938	554.238938	554.238938	0	0	1	-360	360;
	7799	3557	0.00157	0.007	0.02151	581.132387	581.132387	581.132387	0	0	1	-360	360;
	7799	619	0.00131	0.00636	0.01933	606.515349	606.515349	606.515349	0	0	1	-360	360;
	6754	3557	0.001669	0.00724	0.025	532.760988	532.760988	532.760988	0	0	1	-360	360;
	3557	3775	0.00663	0.249339	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3557	1045	0.000169	0.00032	0.03116	268.496281	268.496281	268.496281	0	0	1	-360	360;
	3557	3596	0.000151	0.00038	0.04218	304.012156	304.012156	304.012156	0	0	1	-360	360;
	3557	2407	0.000841	0.009581	0.02766	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3557	488	0.0355	0.360558	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3557	619	0.027539	0.325145	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3557	3508	0.025271	0.339442	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3557	7821	0.000169	0.000219	0.02927	278.067905	278.067905	278.067905	0	0	1	-360	360;
	2209	140	0.000401	0.0018	0.0058	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	2209	0.00205	0.01162	0.03066	315.696328	315.696328	315.696328	0	0	1	-360	360;
	2459	2209	0.000651	0.0045	0.00984	333.36872	333.36872	333.36872	0	0	1	-360	360;
	2674	140	0.00038	0.00187	0.00555	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	2674	0.002091	0.01139	0.03126	445.657643	445.657643	445.657643	0	0	1	-360	360;
	2459	2674	0.00064	0.004169	0.00963	375.187943	375.187943	375.187943	0	0	1	-360	360;
	4179	140	0.03551	0.421116	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	801	0.090159	0.624236	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	7253	0.025919	0.457603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	2366	0.140581	0.992087	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4179	7288	0.00082	0.007331	0.01985	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7447	745	0.008469	0.038599	0.12888	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7447	5854	0.003729	0.012081	0.03914	459.638029	459.638029	459.638029	0	0	1	-360	360;
	6345	5245	0.071281	0.33186	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6345	5245	0.011919	0.055651	0.08325	430.260127	430.260127	430.260127	0	0	1	-360	360;
	5245	8371	0.0038	0.01182	0.03987	460.743078	460.743078	460.743078	0	0	1	-360	360;
	2835	7014	0.00233	0.0191	0.71999	2055.78687	2055.78687	2055.78687	0	0	1	-360	360;
	2835	5754	0.00258	0.02125	0.79958	1821.4362	1821.4362	1821.4362	0	0	1	-360	360;
	2835	7014	0.00246	0.01795	0.76234	2122.71449	2122.71449	2122.71449	0	0	1	-360	360;
	2835	8311	0.00231	0.02199	0.71653	1664.02143	1664.02143	1664.02143	0	0	1	-360	360;
	6455	2835	0.00105	0.0114	0.36542	1936.96575	1936.96575	1936.96575	0	0	1	-360	360;
	2835	788	0.00268	0.02035	0.83233	1865.01228	1865.01228	1865.01228	0	0	1	-360	360;
	4064	1820	0.00124	0.005711	0.021	276.063576	276.063576	276.063576	0	0	1	-360	360;
	4064	8347	0.0002	0.0022	0.00566	581.02836	581.02836	581.02836	0	0	1	-360	360;
	4064	2522	0.00019	0.00225	0.00555	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4064	6248	0.001169	0.005159	0.01741	557.180272	557.180272	557.180272	0	0	1	-360	360;
	4064	7685	0.001271	0.005591	0.02136	269.095293	269.095293	269.095293	0	0	1	-360	360;
	4064	3016	0.00182	0.01218	0.0343	491.285473	491.285473	491.285473	0	0	1	-360	360;
	4064	6494	0.00107	0.005789	0.01549	230.18692	230.18692	230.18692	0	0	1	-360	360;
	5870	69	0.00026	0.00275	0.0846	1632.57033	1632.57033	1632.57033	0	0	1	-360	360;
	5870	723	0.0001	0.00093	0.03377	1929.73105	1929.73105	1929.73105	0	0	1	-360	360;
	8305	7496	0.00019	0.00166	0.06341	1896.43639	1896.43639	1896.43639	0	0	1	-360	360;
	8305	7496	0.00019	0.00163	0.06463	1928.91293	1928.91293	1928.91293	0	0	1	-360	360;
	3434	4982	0.00126	0.00953	0.38902	1829.84971	1829.84971	1829.84971	0	0	1	-360	360;
	3434	7009	0.00131	0.00994	0.40489	1957.497	1957.497	1957.497	0	0	1	-360	360;
	1775	4982	0.00128	0.00933	0.39636	1799.5932	1799.5932	1799.5932	0	0	1	-360	360;
	1775	8295	0.00134	0.00972	0.41253	1778.35902	1778.35902	1778.35902	0	0	1	-360	360;
	3434	1775	0.001	0.180201	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3434	7009	1.365928	3.472299	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3434	8295	0.186177	1.029224	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	616	3434	0.98608	2.33608	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3434	7000	0.487888	1.618975	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3434	4892	0.850277	2.290235	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1775	7009	1.29162	3.486565	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1775	8295	0.201801	0.8991	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	616	1775	0.842729	2.35651	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1775	7000	0.415547	1.630194	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1775	4892	0.830956	1.995914	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	430	8130	0.005729	0.021281	0.05386	400.850978	400.850978	400.850978	0	0	1	-360	360;
	430	8130	0.005081	0.029331	0.08408	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6329	8130	0.00576	0.035781	0.08156	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3585	8130	0.002021	0.00976	0.02545	544.025623	544.025623	544.025623	0	0	1	-360	360;
	2889	8130	0.00563	0.03313	0.07978	466.93173	466.93173	466.93173	0	0	1	-360	360;
	7914	6316	0.001	0.260568	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6265	7914	0.127098	1.298546	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5223	7914	0.04182	0.986219	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5223	7914	0.00129	0.01393	0.42962	1929.34799	1929.34799	1929.34799	0	0	1	-360	360;
	2058	7914	0.00082	0.00915	0.27462	1893.71877	1893.71877	1893.71877	0	0	1	-360	360;
	297	7914	0.00119	0.01287	0.39523	1788.99322	1788.99322	1788.99322	0	0	1	-360	360;
	6265	6316	0.118677	1.418975	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6265	6316	0.00148	0.01585	0.4955	1922.22254	1922.22254	1922.22254	0	0	1	-360	360;
	5223	6316	0.04574	0.913504	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7058	6316	0.00116	0.01192	0.38664	1456.05586	1456.05586	1456.05586	0	0	1	-360	360;
	6455	1046	0.06698	0.909972	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6455	1046	0.00042	0.00331	0.14062	2042.44425	2042.44425	2042.44425	0	0	1	-360	360;
	1542	3769	0.002591	0.01299	0.03353	252.057733	252.057733	252.057733	0	0	1	-360	360;
	3769	2090	0.01169	0.067849	0.20339	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6641	3373	0.00074	0.00368	0.00663	457.225594	457.225594	457.225594	0	0	1	-360	360;
	6641	148	0.00064	0.0042	0.01042	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6641	4837	0.0018	0.01139	0.02956	598.290359	598.290359	598.290359	0	0	1	-360	360;
	762	3373	0.00082	0.003401	0.00774	476.980431	476.980431	476.980431	0	0	1	-360	360;
	762	148	0.00069	0.00381	0.01117	228.566309	228.566309	228.566309	0	0	1	-360	360;
	762	4837	0.00189	0.01094	0.03095	603.654417	603.654417	603.654417	0	0	1	-360	360;
	3373	4715	0.000151	0.000711	0.00246	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3373	1236	0.00013	0.000789	0.00233	295.407639	295.407639	295.407639	0	0	1	-360	360;
	3373	6651	5e-05	0.00024	0.00074	266.272847	266.272847	266.272847	0	0	1	-360	360;
	3373	783	5e-05	0.00025	0.00073	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7836	3373	0.00012	0.000711	0.0016	310.823827	310.823827	310.823827	0	0	1	-360	360;
	5068	3373	0.00011	0.000719	0.00157	270.086788	270.086788	270.086788	0	0	1	-360	360;
	594	3783	0.003091	0.015719	0.04755	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	7835	0.018159	0.39124	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	5747	0.169419	0.883946	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	3160	0.656054	1.65126	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	4757	0.452913	1.939091	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	7889	0.055969	0.74907	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	3783	0.549483	1.897438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	3783	0.483574	1.959731	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	6915	0.279731	1.818967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3783	6915	0.006521	0.02749	0.08096	399.311531	399.311531	399.311531	0	0	1	-360	360;
	6748	5522	0.07687	1.255956	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6748	5522	0.00068	0.00577	0.2297	1681.04129	1681.04129	1681.04129	0	0	1	-360	360;
	4067	6323	0.00812	0.12314	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	13	4067	0.00157	0.007031	0.02061	523.375317	523.375317	523.375317	0	0	1	-360	360;
	1926	4067	0.001599	0.00688	0.02099	513.789593	513.789593	513.789593	0	0	1	-360	360;
	4067	7721	0.00113	0.00724	0.01355	632.552957	632.552957	632.552957	0	0	1	-360	360;
	4067	7721	0.001271	0.007091	0.01535	608.508247	608.508247	608.508247	0	0	1	-360	360;
	594	7889	0.000919	0.005211	0.01443	587.639217	587.639217	587.639217	0	0	1	-360	360;
	6754	5735	0.00199	0.010219	0.02314	528.120547	528.120547	528.120547	0	0	1	-360	360;
	6754	7746	0.005409	0.02475	0.04868	528.760332	528.760332	528.760332	0	0	1	-360	360;
	6754	4417	0.00482	0.02805	0.04337	574.148481	574.148481	574.148481	0	0	1	-360	360;
	6754	2965	0.028479	0.302273	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6754	5747	0.11506	0.606198	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6754	5735	0.025271	0.321983	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6754	8093	0.00194	0.010091	0.08242	248.209116	248.209116	248.209116	0	0	1	-360	360;
	6754	2965	0.00314	0.019719	0.05134	276.605431	276.605431	276.605431	0	0	1	-360	360;
	3751	6373	0.00042	0.00384	0.14129	1913.83733	1913.83733	1913.83733	0	0	1	-360	360;
	3751	723	0.00031	0.00229	0.0998	1769.04184	1769.04184	1769.04184	0	0	1	-360	360;
	3751	1018	0.00024	0.00235	0.07144	1669.80682	1669.80682	1669.80682	0	0	1	-360	360;
	3751	2205	0.00103	0.01072	0.33055	1697.53592	1697.53592	1697.53592	0	0	1	-360	360;
	2217	7721	0.00275	0.01311	0.03965	499.932286	499.932286	499.932286	0	0	1	-360	360;
	2217	7754	0.001771	0.01225	0.02388	481.551139	481.551139	481.551139	0	0	1	-360	360;
	281	5349	0.001	0.385207	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	281	5349	0.003349	0.022711	0.03426	448.573827	448.573827	448.573827	0	0	1	-360	360;
	281	4681	0.00032	0.001531	0.00462	215.929128	215.929128	215.929128	0	0	1	-360	360;
	281	5555	0.359525	1.472438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	281	4684	1.632355	4.693388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	281	7372	0.08494	0.651364	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	281	7362	0.052581	0.671529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7659	281	0.12794	1.157789	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	619	281	0.002461	0.014531	0.02375	638.151742	638.151742	638.151742	0	0	1	-360	360;
	6943	6176	0.002659	0.013271	0.04286	601.345247	601.345247	601.345247	0	0	1	-360	360;
	6943	6176	0.002711	0.01299	0.04368	275.906202	275.906202	275.906202	0	0	1	-360	360;
	4248	4684	3.9e-05	0.0002	0.00074	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2913	4684	5e-05	0.00031	0.00081	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3231	6227	0.00089	0.00698	0.28591	2083.73711	2083.73711	2083.73711	0	0	1	-360	360;
	3231	5018	0.00028	0.00271	0.08013	2014.12346	2014.12346	2014.12346	0	0	1	-360	360;
	1556	2865	0.00025	0.00273	0.0818	1372.034	1372.034	1372.034	0	0	1	-360	360;
	1556	1103	0.00027	0.00238	0.08344	1637.14863	1637.14863	1637.14863	0	0	1	-360	360;
	1556	5018	0.00027	0.00277	0.07842	1955.22167	1955.22167	1955.22167	0	0	1	-360	360;
	2865	1103	0.00808	0.60428	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	2865	0.01086	0.386371	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2865	5018	0.03494	0.767659	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2865	4352	0.0265	0.810665	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	2865	0.524993	2.841898	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3910	2865	0.475568	2.62313	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7836	148	0.02113	0.408595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5068	148	0.022469	0.384442	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8311	5816	0.00045	0.0038	0.15224	2058.55385	2058.55385	2058.55385	0	0	1	-360	360;
	8311	5816	0.00047	0.00374	0.16119	1867.0704	1867.0704	1867.0704	0	0	1	-360	360;
	7835	6999	0.00019	0.00086	0.00262	409.649069	409.649069	409.649069	0	0	1	-360	360;
	5747	7835	0.542397	1.40281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3160	7835	1.299236	2.713017	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7835	4757	1.066591	3.481612	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7835	7889	0.276529	1.159153	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	7835	1.192211	3.154339	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	7835	1.210785	2.840289	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6915	7835	0.804979	2.837397	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	801	8339	9.1e-05	0.00055	0.0014	337.05196	337.05196	337.05196	0	0	1	-360	360;
	7253	801	0.009909	0.377748	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	801	2366	0.914793	3.247727	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2763	8339	0.000349	0.002159	0.00556	647.101872	647.101872	647.101872	0	0	1	-360	360;
	7253	6999	0.00132	0.00624	0.01907	547.84871	547.84871	547.84871	0	0	1	-360	360;
	7253	8339	0.00088	0.005159	0.01252	517.37719	517.37719	517.37719	0	0	1	-360	360;
	5274	2366	0.001349	0.01051	0.05416	482.816991	482.816991	482.816991	0	0	1	-360	360;
	481	1852	0.000539	0.003099	0.00825	618.436916	618.436916	618.436916	0	0	1	-360	360;
	481	7288	0.00207	0.01182	0.03149	578.415873	578.415873	578.415873	0	0	1	-360	360;
	6323	7721	0.000539	0.003159	0.00825	245.69494	245.69494	245.69494	0	0	1	-360	360;
	6323	7754	0.004081	0.0247	0.06557	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6905	3780	0.000531	0.003521	0.0059	499.537517	499.537517	499.537517	0	0	1	-360	360;
	7209	1819	0.000539	0.00345	0.00601	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1511	7209	0.005169	0.02325	0.03636	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6943	4417	0.000669	0.004099	0.00602	578.955442	578.955442	578.955442	0	0	1	-360	360;
	6943	5840	0.000599	0.00344	0.00519	356.047361	356.047361	356.047361	0	0	1	-360	360;
	349	6943	0.00139	0.008919	0.01245	262.996569	262.996569	262.996569	0	0	1	-360	360;
	6943	7746	0.00069	0.00419	0.00628	574.813796	574.813796	574.813796	0	0	1	-360	360;
	1696	6943	0.001539	0.008099	0.0138	275.686717	275.686717	275.686717	0	0	1	-360	360;
	5747	2453	0.00062	0.00356	0.01007	496.583989	496.583989	496.583989	0	0	1	-360	360;
	2453	7889	0.001091	0.006479	0.01286	444.628424	444.628424	444.628424	0	0	1	-360	360;
	8135	6635	6e-05	0.00052	0.02087	1972.10266	1972.10266	1972.10266	0	0	1	-360	360;
	8135	7207	0.00023	0.00485	0.17201	1899.05463	1899.05463	1899.05463	0	0	1	-360	360;
	5018	6635	7e-05	0.00051	0.02127	1915.43605	1915.43605	1915.43605	0	0	1	-360	360;
	6790	5349	5e-05	0.000229	0.00082	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1200	5349	8.1e-05	0.0003	0.00107	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3725	5349	6e-05	0.000341	0.00095	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5555	5349	0.339112	2.39938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4684	5349	0.982252	4.492975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5349	7372	0.299979	2.042583	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5349	7362	0.216508	1.7125	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	73	5349	0.002841	0.01551	0.04214	277.856422	277.856422	277.856422	0	0	1	-360	360;
	7659	5349	0.00326	1.128223	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4146	5349	0.00232	0.016961	0.04206	348.258294	348.258294	348.258294	0	0	1	-360	360;
	2015	5349	0.003979	0.02513	0.03704	503.364414	503.364414	503.364414	0	0	1	-360	360;
	6673	5349	0.002281	0.017289	0.04118	597.452809	597.452809	597.452809	0	0	1	-360	360;
	1138	1103	0.111143	0.652313	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	14	1103	0.00045	0.00399	0.15258	1718.49564	1718.49564	1718.49564	0	0	1	-360	360;
	5018	1103	0.01009	0.588213	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4352	1103	0.0029	0.60196	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	1103	0.468573	2.303532	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3910	1103	0.401032	2.262812	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5018	1103	0.00054	0.00451	0.18353	1783.44017	1783.44017	1783.44017	0	0	1	-360	360;
	14	1103	0.0004	0.00439	0.13731	1663.35009	1663.35009	1663.35009	0	0	1	-360	360;
	2148	2366	0.00132	0.0077	0.01898	224.091244	224.091244	224.091244	0	0	1	-360	360;
	2148	7889	0.002289	0.013909	0.03278	461.61378	461.61378	461.61378	0	0	1	-360	360;
	4558	2875	0.00015	0.00153	0.04776	1459.29936	1459.29936	1459.29936	0	0	1	-360	360;
	4558	5593	0.00022	0.00174	0.07202	1793.92957	1793.92957	1793.92957	0	0	1	-360	360;
	4558	8368	0.00031	0.00276	0.10567	1522.60436	1522.60436	1522.60436	0	0	1	-360	360;
	7253	2366	0.498802	2.231818	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7253	7889	0.00243	0.01461	0.03115	536.812705	536.812705	536.812705	0	0	1	-360	360;
	7253	7288	0.0013	0.005919	0.02135	561.478148	561.478148	561.478148	0	0	1	-360	360;
	4684	5555	2.422314	7.164876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4684	5555	0.005271	0.03874	0.06348	586.345929	586.345929	586.345929	0	0	1	-360	360;
	5555	7372	0.157781	1.324793	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5555	7372	0.00695	0.055591	0.08519	586.267433	586.267433	586.267433	0	0	1	-360	360;
	5555	7362	0.211777	1.52969	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7659	5555	0.1823	1.686529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1982	7638	2.525346	8.326177	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1982	7638	0.00086	0.00871	0.24623	2296.19015	2296.19015	2296.19015	0	0	1	-360	360;
	7638	1667	0.13018	1.288158	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7638	3260	5.25734	15.96122	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7638	6227	0.196787	1.606648	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7878	1667	0.00124	0.01282	0.35391	1660.98459	1660.98459	1660.98459	0	0	1	-360	360;
	13	1926	0	0.000103	0	267.580234	267.580234	267.580234	0	0	1	-360	360;
	13	1926	0.227149	1.244607	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	13	1852	0.024581	0.402459	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2459	13	0.04388	0.657934	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6329	13	0.12525	0.756384	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2889	13	0.18987	0.937128	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	13	7961	0.00163	0.007031	0.02291	232.306326	232.306326	232.306326	0	0	1	-360	360;
	1926	1852	0.456736	1.546157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2459	1926	0.03062	0.528388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6329	1926	0.192099	0.773678	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2889	1926	0.027081	0.45345	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2889	1926	0.002211	0.01299	0.03229	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1917	5584	0.00344	0.015909	0.02714	231.070196	231.070196	231.070196	0	0	1	-360	360;
	1138	5018	0.123539	0.685277	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	4352	0.096593	0.575748	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	5470	0.225512	1.104709	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	3910	0.17135	1.111427	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1138	5018	0.00117	0.00907	0.39123	1819.68083	1819.68083	1819.68083	0	0	1	-360	360;
	3378	1614	3e-05	0.00039	0.01156	1795.73956	1795.73956	1795.73956	0	0	1	-360	360;
	565	3378	0.00079	0.0065	0.26159	1765.65945	1765.65945	1765.65945	0	0	1	-360	360;
	4290	3378	0.00091	0.00755	0.30391	2123.01922	2123.01922	2123.01922	0	0	1	-360	360;
	565	3378	0.00081	0.00636	0.26652	1875.40756	1875.40756	1875.40756	0	0	1	-360	360;
	4290	3378	0.00086	0.00818	0.28587	1872.91833	1872.91833	1872.91833	0	0	1	-360	360;
	2214	4684	0.003979	0.02364	0.05889	600.473783	600.473783	600.473783	0	0	1	-360	360;
	4684	7372	6.941322	16.0093	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4684	7362	6.73657	15.66095	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7659	4684	0.185229	2.220661	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1982	1667	5.786496	13.8054	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1982	3260	0.04375	0.648878	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1982	549	0.0008	0.00847	0.26865	1736.62058	1736.62058	1736.62058	0	0	1	-360	360;
	1982	6227	0.19205	1.80741	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1982	6227	0.00157	0.01084	0.63402	2034.44738	2034.44738	2034.44738	0	0	1	-360	360;
	4097	1982	0.00155	0.01203	0.51873	2227.69343	2227.69343	2227.69343	0	0	1	-360	360;
	1982	3747	0.00062	0.00656	0.20806	1788.84777	1788.84777	1788.84777	0	0	1	-360	360;
	1982	1667	0.00203	0.0219	0.60893	2233.58479	2233.58479	2233.58479	0	0	1	-360	360;
	1982	7121	0.00017	0.00135	0.05208	1900.16695	1900.16695	1900.16695	0	0	1	-360	360;
	4437	7685	0.000341	0.001729	0.01133	434.513039	434.513039	434.513039	0	0	1	-360	360;
	3160	4437	0.00101	0.00926	0.03427	568.650293	568.650293	568.650293	0	0	1	-360	360;
	2190	4437	0.00055	0.00262	0.00926	240.127021	240.127021	240.127021	0	0	1	-360	360;
	1820	7685	0.018781	0.265517	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4495	1820	0.0298	0.392211	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1820	2171	0.00057	0.002841	0.00933	537.231862	537.231862	537.231862	0	0	1	-360	360;
	2190	1820	0.00943	0.27814	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4495	7685	0.027961	0.338967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7685	3221	0.000581	0.002781	0.00951	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2190	7685	0.010831	0.239876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4495	5122	0.00261	0.01243	0.0358	387.818266	387.818266	387.818266	0	0	1	-360	360;
	2416	3775	0.00214	0.01187	0.03186	541.776659	541.776659	541.776659	0	0	1	-360	360;
	488	3775	0.04931	0.648058	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	488	3775	0.001711	0.010781	0.03681	581.665096	581.665096	581.665096	0	0	1	-360	360;
	3775	4030	0.00068	0.003271	0.01067	255.798512	255.798512	255.798512	0	0	1	-360	360;
	619	3775	0.07599	0.788884	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3508	3775	0.05989	0.967314	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6567	3775	0.00036	0.002719	0.01126	241.357435	241.357435	241.357435	0	0	1	-360	360;
	750	1018	0.00022	0.00201	0.07516	1715.34494	1715.34494	1715.34494	0	0	1	-360	360;
	5223	750	0.04617	0.493712	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	750	6706	0.00031	0.00302	0.0995	1832.50085	1832.50085	1832.50085	0	0	1	-360	360;
	6265	5223	0.001	0.254681	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6265	2205	0.00034	0.00288	0.11445	1800.89706	1800.89706	1800.89706	0	0	1	-360	360;
	4604	8295	8e-05	0.00085	0.02483	1291.82103	1291.82103	1291.82103	0	0	1	-360	360;
	4604	8295	8e-05	0.00087	0.02431	1317.76383	1317.76383	1317.76383	0	0	1	-360	360;
	4604	8295	7e-05	0.00089	0.02379	1324.73724	1324.73724	1324.73724	0	0	1	-360	360;
	3291	5584	0.808079	2.998967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	269	5584	0.11606	0.766157	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3886	5584	0.044669	0.42969	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3886	5584	0.0045	0.023841	0.04053	438.327361	438.327361	438.327361	0	0	1	-360	360;
	3687	375	0.009969	0.07739	0.22162	256.654353	256.654353	256.654353	0	0	1	-360	360;
	1374	3687	0.01431	0.10638	0.13415	390.982133	390.982133	390.982133	0	0	1	-360	360;
	3687	4643	0.00774	0.06957	0.11908	496.624	496.624	496.624	0	0	1	-360	360;
	114	375	0.016669	0.26188	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7981	375	0.009969	0.061409	0.1339	599.201072	599.201072	599.201072	0	0	1	-360	360;
	114	375	0.001729	0.00995	0.02672	409.078254	409.078254	409.078254	0	0	1	-360	360;
	7014	5506	0.0015	0.01398	0.30616	1258.07157	1258.07157	1258.07157	0	0	1	-360	360;
	4290	7014	0.00243	0.01942	0.75784	1435.4476	1435.4476	1435.4476	0	0	1	-360	360;
	7708	745	0.00455	0.02705	0.06801	503.845301	503.845301	503.845301	0	0	1	-360	360;
	4837	7708	0.004331	0.01676	0.06138	606.195647	606.195647	606.195647	0	0	1	-360	360;
	7708	1780	0.24907	0.891963	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7708	745	0.033831	0.228099	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	269	3291	0.344318	0.959132	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3291	3886	0.087211	0.561364	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3291	1363	0.003909	0.019531	0.06125	216.567007	216.567007	216.567007	0	0	1	-360	360;
	5974	723	0.01509	0.413102	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5974	1786	9e-05	0.001	0.03216	2075.34137	2075.34137	2075.34137	0	0	1	-360	360;
	5974	5918	0.00094	0.00943	0.03022	1741.44306	1741.44306	1741.44306	0	0	1	-360	360;
	5974	4096	0.03361	0.58162	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3768	2058	0.0001	0.00107	0.13718	1602.09992	1602.09992	1602.09992	0	0	1	-360	360;
	3768	788	0.00084	0.00882	0.28008	1609.9158	1609.9158	1609.9158	0	0	1	-360	360;
	2214	269	0.00501	0.021599	0.04073	582.486643	582.486643	582.486643	0	0	1	-360	360;
	269	3886	0.04149	0.219607	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	269	1621	0.000781	0.00744	0.02392	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5405	7311	0.011789	0.277645	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5405	7311	0.0052	0.02574	0.07842	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7311	5988	0.031669	0.507603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7311	5988	0.00438	0.025461	0.06618	317.340564	317.340564	317.340564	0	0	1	-360	360;
	1542	7311	0.27186	0.993388	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	69	723	0.0037	0.327618	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	723	4096	0.01026	0.269931	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2871	536	0.000281	0.0023	0.00532	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6905	6648	0.000539	0.005021	0.01143	279.068165	279.068165	279.068165	0	0	1	-360	360;
	6905	4359	0.251054	1.054483	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7964	6905	0.00539	0.022409	0.0366	382.198523	382.198523	382.198523	0	0	1	-360	360;
	1780	771	0.18813	0.545888	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1780	1304	0.02382	0.200979	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1780	1304	0.00375	0.019289	0.06158	494.312162	494.312162	494.312162	0	0	1	-360	360;
	1780	2370	0.003151	0.01874	0.05036	323.754798	323.754798	323.754798	0	0	1	-360	360;
	1780	745	0.02419	0.20719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4324	5405	0.005479	0.022531	0.08146	571.222389	571.222389	571.222389	0	0	1	-360	360;
	4324	5600	0.008581	0.133581	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5269	3508	0.000849	0.00436	0.0121	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3445	4701	0.0004	0.00418	0.13271	1871.52364	1871.52364	1871.52364	0	0	1	-360	360;
	3445	297	0.00023	0.00253	0.0736	1914.10126	1914.10126	1914.10126	0	0	1	-360	360;
	4983	2459	0.000659	0.006909	0.02466	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3445	2186	0.14356	0.922161	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3445	4165	0.05061	0.949861	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2242	7685	0.001669	0.19101	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3538	2242	0.008901	0.19594	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3538	7685	0.008849	0.207376	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	528	2058	0.04068	0.314813	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	528	788	0.001	0.358151	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	528	788	0.00029	0.00339	0.0914	1487.85447	1487.85447	1487.85447	0	0	1	-360	360;
	528	5754	0.001	0.381247	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3324	1205	0	0.00011	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3324	1205	0.001	0.261032	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3557	2168	5e-05	0.000409	0.00116	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6567	3557	0.00036	0.00411	0.01252	559.074477	559.074477	559.074477	0	0	1	-360	360;
	4115	1557	0	0.000103	0	235.092954	235.092954	235.092954	0	0	1	-360	360;
	4115	1557	0.06963	0.647459	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	488	4115	0.00138	0.00794	0.02018	564.787197	564.787197	564.787197	0	0	1	-360	360;
	619	4115	0.04905	0.618285	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6673	4115	0.01481	0.65845	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4146	4115	0.057229	0.834525	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	73	1557	0.00276	0.034271	0.09028	277.382394	277.382394	277.382394	0	0	1	-360	360;
	619	1557	0.045901	0.629938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6673	1557	0.05236	0.804298	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4146	1557	0.010979	0.705971	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3324	2058	0.00111	0.00985	0.37649	1962.48863	1962.48863	1962.48863	0	0	1	-360	360;
	3260	1667	12.66967	25.61427	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1667	6227	0.7509	2.64626	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7372	7362	0.001	0.522603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7372	7362	0	0.000103	0	338.792983	338.792983	338.792983	0	0	1	-360	360;
	7659	7372	1.015517	3.518182	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3508	7372	0.00274	0.01674	0.05173	612.997411	612.997411	612.997411	0	0	1	-360	360;
	8157	7372	0.001031	0.006669	0.02988	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7659	7362	0.827397	3.941942	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3307	1667	0.00018	0.00147	0.04956	2021.69778	2021.69778	2021.69778	0	0	1	-360	360;
	4829	1667	0.00025	0.00253	0.07717	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3449	1667	0.00045	0.00372	0.13673	1644.71374	1644.71374	1644.71374	0	0	1	-360	360;
	7145	1667	0.00019	0.00142	0.06148	1764.71957	1764.71957	1764.71957	0	0	1	-360	360;
	5533	1667	0.00017	0.00161	0.05477	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4204	1621	0.00031	0.00387	0.03188	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	6708	0.00057	0.00261	0.00953	369.204297	369.204297	369.204297	0	0	1	-360	360;
	5747	1739	0.000591	0.002591	0.00987	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2416	619	0.001599	0.00775	0.02278	215.929128	215.929128	215.929128	0	0	1	-360	360;
	488	619	0.014031	0.614215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	488	3508	0.012341	0.556384	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	488	3508	0.001599	0.00738	0.02419	378.389154	378.389154	378.389154	0	0	1	-360	360;
	8244	1614	0.077687	1.036288	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8244	565	0.00057	0.00542	0.19206	2046.78824	2046.78824	2046.78824	0	0	1	-360	360;
	8244	565	0.001	0.513781	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8244	1351	0.01036	0.396828	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8244	1351	0.00067	0.00594	0.22858	1978.86348	1978.86348	1978.86348	0	0	1	-360	360;
	8244	4290	0.897853	2.161981	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1351	3894	0.0015	0.01442	0.50193	1538.7653	1538.7653	1538.7653	0	0	1	-360	360;
	4290	3894	0.00139	0.01387	0.42737	1870.99447	1870.99447	1870.99447	0	0	1	-360	360;
	565	1614	0.001	0.795637	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1351	1614	0.230298	1.27597	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4290	1614	0.05565	0.673331	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8305	14	0.00019	0.00214	0.06427	1767.60108	1767.60108	1767.60108	0	0	1	-360	360;
	8305	14	0.00019	0.00198	0.06287	1766.94093	1766.94093	1766.94093	0	0	1	-360	360;
	5018	4352	0.07358	0.966343	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	5018	0.06384	1.43054	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3910	5018	0.076281	1.166136	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	4352	0.210457	1.788158	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3910	4352	0.185533	1.718213	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5018	4352	2e-05	0.00024	0.00745	1578.66808	1578.66808	1578.66808	0	0	1	-360	360;
	297	4653	0.01408	0.37133	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7058	4653	0.01647	0.363712	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7058	4653	0.00068	0.00621	0.21206	1823.33307	1823.33307	1823.33307	0	0	1	-360	360;
	4653	788	0.00079	0.00807	0.24891	1841.23357	1841.23357	1841.23357	0	0	1	-360	360;
	2205	5223	0.00052	0.00514	0.17427	1875.58659	1875.58659	1875.58659	0	0	1	-360	360;
	788	2058	0.111198	0.57367	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5754	2058	0.126101	0.495623	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3886	1363	0.0032	0.016281	0.03007	489.52273	489.52273	489.52273	0	0	1	-360	360;
	3886	1363	0.003659	0.01407	0.03445	528.686027	528.686027	528.686027	0	0	1	-360	360;
	3886	1621	0.002281	0.010039	0.03839	246.920782	246.920782	246.920782	0	0	1	-360	360;
	3886	1621	0.001919	0.011669	0.03237	232.747202	232.747202	232.747202	0	0	1	-360	360;
	6648	4359	0.00236	0.02263	0.06913	282.083423	282.083423	282.083423	0	0	1	-360	360;
	2871	6648	0.009849	0.054531	0.14748	305.736031	305.736031	305.736031	0	0	1	-360	360;
	1304	771	0.02949	0.13405	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1304	771	0.005021	0.030341	0.08218	271.263093	271.263093	271.263093	0	0	1	-360	360;
	3260	7121	0.00076	0.00675	0.25223	2030.17448	2030.17448	2030.17448	0	0	1	-360	360;
	3260	6227	0.568047	3.572576	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	575	4030	0.003419	0.014531	0.05386	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1852	575	0.002979	0.016461	0.0453	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	2965	0.09382	0.44438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6945	2965	0.003789	0.02306	0.04756	390.05656	390.05656	390.05656	0	0	1	-360	360;
	2472	3449	0.00014	0.00108	0.04496	1839.48216	1839.48216	1839.48216	0	0	1	-360	360;
	2472	8305	0.00015	0.00119	0.04852	1837.3003	1837.3003	1837.3003	0	0	1	-360	360;
	1786	8305	0.00031	0.00232	0.09283	2020.11421	2020.11421	2020.11421	0	0	1	-360	360;
	1640	1786	0.00025	0.00193	0.08514	1430.41845	1430.41845	1430.41845	0	0	1	-360	360;
	8305	7207	0.137791	1.05367	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8305	6706	0.00054	0.00395	0.16687	1905.88521	1905.88521	1905.88521	0	0	1	-360	360;
	5020	2459	0.00113	0.012581	0.04248	310.196616	310.196616	310.196616	0	0	1	-360	360;
	5020	2871	0.00238	0.02086	0.08955	275.430269	275.430269	275.430269	0	0	1	-360	360;
	5020	6329	0.127331	1.369545	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5020	2025	0.699669	2.452893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8194	5988	0.00331	0.025159	0.11776	535.549521	535.549521	535.549521	0	0	1	-360	360;
	565	1351	0.01975	0.825554	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4290	565	0.507639	2.270152	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5405	5988	0.20336	0.874897	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5405	5600	0.00732	0.030659	0.11167	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1542	5405	0.075401	0.292748	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1542	5405	0.00314	0.02136	0.09397	248.153864	248.153864	248.153864	0	0	1	-360	360;
	8311	6327	0.00044	0.00325	0.13816	1895.01867	1895.01867	1895.01867	0	0	1	-360	360;
	47	2190	0.00076	0.00364	0.01273	267.807721	267.807721	267.807721	0	0	1	-360	360;
	47	5412	0.00075	0.00474	0.01175	243.186099	243.186099	243.186099	0	0	1	-360	360;
	5747	3160	3.557851	5.439876	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	4757	4.011157	6.327273	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	7889	0.006	0.327025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	5747	3.78905	6.419628	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	5747	3.833471	5.760331	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3538	5747	0.000841	0.00362	0.01002	379.733122	379.733122	379.733122	0	0	1	-360	360;
	5747	1828	0.03	0.279855	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	1828	0.00107	0.004849	0.01757	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	6915	3.092975	5.176033	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	6915	0.02688	0.291116	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5747	6915	0.00188	0.0132	0.04087	563.160489	563.160489	563.160489	0	0	1	-360	360;
	5932	1304	0.001469	0.00724	0.02184	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5932	1304	0.0015	0.007081	0.02224	270.418303	270.418303	270.418303	0	0	1	-360	360;
	7288	4928	0.0017	0.00806	0.02325	606.342352	606.342352	606.342352	0	0	1	-360	360;
	1852	2362	0.00369	0.016091	0.05386	534.873155	534.873155	534.873155	0	0	1	-360	360;
	7721	6426	0.00301	0.0177	0.047	605.308179	605.308179	605.308179	0	0	1	-360	360;
	5593	4327	5e-05	0.00053	0.01592	1434.70254	1434.70254	1434.70254	0	0	1	-360	360;
	4701	3573	6e-05	0.0005	0.01713	1501.91252	1501.91252	1501.91252	0	0	1	-360	360;
	4701	5219	4e-05	0.00047	0.01454	2204.94148	2204.94148	2204.94148	0	0	1	-360	360;
	7889	147	0.00174	0.016521	0.06438	413.115491	413.115491	413.115491	0	0	1	-360	360;
	1436	5044	0.00019	0.00196	0.0649	1831.93811	1831.93811	1831.93811	0	0	1	-360	360;
	2205	5821	0.00074	0.00584	0.24125	2053.2509	2053.2509	2053.2509	0	0	1	-360	360;
	1436	1570	0.00019	0.00183	0.06224	2262.57825	2262.57825	2262.57825	0	0	1	-360	360;
	5593	2352	0.00018	0.00186	0.06097	2196.73399	2196.73399	2196.73399	0	0	1	-360	360;
	5470	3910	0	0.00011	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	3910	0.001	0.413968	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5470	5018	0.00215	0.02177	0.66355	1523.33692	1523.33692	1523.33692	0	0	1	-360	360;
	6345	745	0.008659	0.0422	0.05975	389.347805	389.347805	389.347805	0	0	1	-360	360;
	69	6331	0.00029	0.00252	0.09063	1434.35173	1434.35173	1434.35173	0	0	1	-360	360;
	69	4096	0.0343	0.612763	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2366	7754	0.00036	0.00195	0.00558	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2366	7754	0.000409	0.00213	0.00543	282.046461	282.046461	282.046461	0	0	1	-360	360;
	4495	2525	0.00118	0.008719	0.01341	477.77416	477.77416	477.77416	0	0	1	-360	360;
	3439	4495	0.0053	0.023409	0.08261	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	4495	0.005401	0.022909	0.08416	612.986741	612.986741	612.986741	0	0	1	-360	360;
	5412	4495	0.000901	0.00412	0.01477	615.381267	615.381267	615.381267	0	0	1	-360	360;
	4495	2190	0.05139	0.435785	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3160	4757	0.01661	0.339194	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3160	7889	2.441736	4.517975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	3160	0.03336	0.374174	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	3160	0.033219	0.37814	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3160	4092	0.003151	0.017599	0.05201	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6915	3160	0.02231	0.393079	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7947	3307	0.0005	0.00519	0.16674	2058.23858	2058.23858	2058.23858	0	0	1	-360	360;
	7947	7784	0.00024	0.00227	0.07779	1790.73279	1790.73279	1790.73279	0	0	1	-360	360;
	7947	6331	0.00049	0.00529	0.16319	1455.28644	1455.28644	1455.28644	0	0	1	-360	360;
	7947	4906	0.00036	0.00305	0.11465	1455.23247	1455.23247	1455.23247	0	0	1	-360	360;
	3513	5149	0.00264	0.014781	0.04058	274.400288	274.400288	274.400288	0	0	1	-360	360;
	6262	5149	0.002539	0.01357	0.03928	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2025	5149	0.0053	0.021599	0.07962	553.361377	553.361377	553.361377	0	0	1	-360	360;
	6032	5149	0.003091	0.01568	0.04932	262.244755	262.244755	262.244755	0	0	1	-360	360;
	8356	5149	0.00274	0.0127	0.04242	233.005555	233.005555	233.005555	0	0	1	-360	360;
	2025	5149	0.04224	0.3325	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5365	4982	0.00086	0.00873	0.26836	1789.55992	1789.55992	1789.55992	0	0	1	-360	360;
	8311	4982	0.00206	0.01561	0.63618	1689.12505	1689.12505	1689.12505	0	0	1	-360	360;
	3335	8295	0.00032	0.00271	0.10771	1536.72231	1536.72231	1536.72231	0	0	1	-360	360;
	7058	3335	0.00051	0.00524	0.17072	1774.10652	1774.10652	1774.10652	0	0	1	-360	360;
	228	7009	0.00032	0.00265	0.10974	1838.07892	1838.07892	1838.07892	0	0	1	-360	360;
	297	228	0.0005	0.00534	0.16716	1672.19602	1672.19602	1672.19602	0	0	1	-360	360;
	7009	8295	0	0.00011	0	1583.85585	1583.85585	1583.85585	0	0	1	-360	360;
	7009	8295	0.163629	1.021814	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	616	7009	0.001	0.350312	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7000	7009	0.0571	0.699515	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3417	7009	0.00022	0.00214	0.06822	1416.28734	1416.28734	1416.28734	0	0	1	-360	360;
	4892	7009	0.05584	0.597071	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4165	8295	0.00044	0.00329	0.13718	1962.39451	1962.39451	1962.39451	0	0	1	-360	360;
	616	8295	0.0684	0.765166	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	616	8295	0.0005	0.00377	0.16816	1745.10122	1745.10122	1745.10122	0	0	1	-360	360;
	7000	8295	0.02916	0.635783	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4939	8295	0.00021	0.00218	0.06677	1773.73202	1773.73202	1773.73202	0	0	1	-360	360;
	4892	8295	0.076648	0.893421	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6196	7130	0.003581	0.01713	0.06044	261.258214	261.258214	261.258214	0	0	1	-360	360;
	349	108	0.003651	0.01676	0.06158	322.18944	322.18944	322.18944	0	0	1	-360	360;
	5918	5416	0	0.0001	0	894.36366	894.36366	894.36366	0	0	1	-360	360;
	5918	5416	0.001	0.363179	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5918	2205	0.03201	0.610942	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5416	2205	0.03493	0.703809	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5416	2875	0.00038	0.00378	0.12455	1724.9096	1724.9096	1724.9096	0	0	1	-360	360;
	1542	5988	0.662562	1.966736	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1542	5988	0.007461	0.03539	0.11275	555.232719	555.232719	555.232719	0	0	1	-360	360;
	4290	1351	1.057548	2.384834	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1852	7130	0.00313	0.015591	0.0514	304.446554	304.446554	304.446554	0	0	1	-360	360;
	1852	108	0.00325	0.01495	0.05337	347.021783	347.021783	347.021783	0	0	1	-360	360;
	2459	1852	0.174211	0.885269	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6329	1852	0.05432	0.507293	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2889	1852	0.315744	1.143306	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6945	7721	0.106531	0.728988	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4701	2205	0.00046	0.00355	0.1514	1913.4365	1913.4365	1913.4365	0	0	1	-360	360;
	5593	2205	0.00046	0.00356	0.15188	1505.73589	1505.73589	1505.73589	0	0	1	-360	360;
	2525	4757	0.00056	0.00351	0.00895	468.275316	468.275316	468.275316	0	0	1	-360	360;
	6915	2525	0.001271	0.00699	0.01306	489.220175	489.220175	489.220175	0	0	1	-360	360;
	4757	7889	2.229752	5.528926	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	4757	0.186169	1.134959	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	4757	0.191419	0.99374	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6915	4757	0.001	0.342025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8368	5522	8e-05	0.00071	0.02592	1872.33123	1872.33123	1872.33123	0	0	1	-360	360;
	1436	5522	0.00047	0.00487	0.15853	1314.63418	1314.63418	1314.63418	0	0	1	-360	360;
	3439	7889	2.630992	4.595455	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	7889	2.602893	4.194835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3538	7889	0.002849	0.0122	0.0368	412.659754	412.659754	412.659754	0	0	1	-360	360;
	6915	7889	1.630455	4.689463	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2875	5522	0.136641	1.051662	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	609	5522	0.00035	0.00282	0.11661	1541.72447	1541.72447	1541.72447	0	0	1	-360	360;
	5365	8311	0.00092	0.0092	0.2838	1835.68381	1835.68381	1835.68381	0	0	1	-360	360;
	2186	4165	0	0.0001	0	1025.96925	1025.96925	1025.96925	0	0	1	-360	360;
	2186	4165	0.02074	0.459238	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2186	1436	0.00025	0.00224	0.08553	1886.11087	1886.11087	1886.11087	0	0	1	-360	360;
	6945	7754	0.000651	0.0037	0.00982	246.66624	246.66624	246.66624	0	0	1	-360	360;
	3439	4782	0.001	0.238182	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3439	6915	0.168469	0.943719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4782	6915	0.1577	0.919711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5854	4359	0.0033	0.023531	0.06839	316.83872	316.83872	316.83872	0	0	1	-360	360;
	8371	4359	0.00336	0.023031	0.06968	561.400033	561.400033	561.400033	0	0	1	-360	360;
	7981	4359	0.050961	0.353843	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1511	4359	0.07906	0.386591	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1511	4359	0.00224	0.02005	0.06754	338.456896	338.456896	338.456896	0	0	1	-360	360;
	7964	4359	0.001979	0.018841	0.072	286.876665	286.876665	286.876665	0	0	1	-360	360;
	3016	20	0.0003	0.002669	0.01028	537.965766	537.965766	537.965766	0	0	1	-360	360;
	1667	1640	0.00017	0.00166	0.05075	1429.84715	1429.84715	1429.84715	0	0	1	-360	360;
	7145	4096	0.00057	0.00619	0.18422	1764.71957	1764.71957	1764.71957	0	0	1	-360	360;
	5533	7903	8e-05	0.00082	0.02265	767.876077	767.876077	767.876077	0	0	1	-360	360;
	619	3508	0.001	0.452087	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	619	6673	0.028289	0.776901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	619	4146	0.027469	0.75064	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3508	2015	0.001409	0.00714	0.01369	485.093391	485.093391	485.093391	0	0	1	-360	360;
	4837	6032	0.00776	0.04488	0.12714	264.957077	264.957077	264.957077	0	0	1	-360	360;
	609	2253	0.00028	0.0023	0.09163	1479.70029	1479.70029	1479.70029	0	0	1	-360	360;
	2933	609	0.00015	0.00155	0.04675	1605.16769	1605.16769	1605.16769	0	0	1	-360	360;
	4146	475	0.000229	0.00113	0.00304	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6673	3049	0.00024	0.001099	0.0031	530.440006	530.440006	530.440006	0	0	1	-360	360;
	5854	7981	0.000651	0.005031	0.01147	226.683916	226.683916	226.683916	0	0	1	-360	360;
	8371	7981	0.000581	0.0057	0.01024	568.625525	568.625525	568.625525	0	0	1	-360	360;
	1511	7981	0.395351	1.062603	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7058	1436	0.00073	0.00574	0.22316	1960.49369	1960.49369	1960.49369	0	0	1	-360	360;
	5183	5593	0.00074	0.00562	0.22745	2126.93539	2126.93539	2126.93539	0	0	1	-360	360;
	723	4096	0.00021	0.00217	0.06651	1588.30646	1588.30646	1588.30646	0	0	1	-360	360;
	7836	5068	0.001	0.253636	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7076	2933	0.00078	0.00691	0.2642	1646.61193	1646.61193	1646.61193	0	0	1	-360	360;
	5183	5754	0.00135	0.01007	0.42526	2035.03579	2035.03579	2035.03579	0	0	1	-360	360;
	2459	6329	0.048271	0.517335	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2459	6329	0.001901	0.01187	0.027	295.61455	295.61455	295.61455	0	0	1	-360	360;
	2459	2889	0.02499	0.413099	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2459	2889	0.00186	0.01099	0.02641	543.571791	543.571791	543.571791	0	0	1	-360	360;
	297	7058	0.00351	0.296032	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2171	8255	0.000169	0.000789	0.00279	352.870157	352.870157	352.870157	0	0	1	-360	360;
	2190	2171	0.000211	0.001031	0.00355	295.356959	295.356959	295.356959	0	0	1	-360	360;
	3221	6041	0.00038	0.001909	0.00632	224.91279	224.91279	224.91279	0	0	1	-360	360;
	2190	3221	0.000219	0.001	0.00362	231.369321	231.369321	231.369321	0	0	1	-360	360;
	616	7000	0.0114	0.491198	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	616	4892	0.0111	0.44741	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6329	2889	0.065849	0.525558	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6329	7961	0.00095	0.0053	0.01475	527.720824	527.720824	527.720824	0	0	1	-360	360;
	5412	2190	0.00139	0.008281	0.02203	245.998638	245.998638	245.998638	0	0	1	-360	360;
	6248	6494	0.001	0.26688	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4837	1856	0.00036	0.00286	0.01313	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6915	1828	0.02231	0.26281	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6915	1828	0.001669	0.01064	0.02196	579.160828	579.160828	579.160828	0	0	1	-360	360;
	5018	7151	3e-05	0.00031	0.00971	2028.0762	2028.0762	2028.0762	0	0	1	-360	360;
	5018	390	3e-05	0.0003	0.01101	1960.51409	1960.51409	1960.51409	0	0	1	-360	360;
	7207	6706	0.0004	0.00637	2.1986	1663.39616	1663.39616	1663.39616	0	0	1	-360	360;
	7076	4906	0.00018	0.0019	0.0597	1727.1507	1727.1507	1727.1507	0	0	1	-360	360;
	7076	7784	0.00032	0.00267	0.10002	1729.70838	1729.70838	1729.70838	0	0	1	-360	360;
	3417	7000	0.00023	0.00202	0.07866	1644.26749	1644.26749	1644.26749	0	0	1	-360	360;
	7000	4892	0.001	0.382043	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7149	8356	0.00011	0.00075	0.0014	543.535972	543.535972	543.535972	0	0	1	-360	360;
	7993	8356	0.00013	0.000531	0.00208	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6673	4146	0.010909	0.62314	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1363	1621	0.005969	0.037081	0.10316	247.249248	247.249248	247.249248	0	0	1	-360	360;
	4097	8249	0	0.00011	0	1515.58357	1515.58357	1515.58357	0	0	1	-360	360;
	4097	8249	0.001	0.39696	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4097	7602	0.192888	1.242729	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8249	1485	0.00078	0.00629	0.26024	1904.21278	1904.21278	1904.21278	0	0	1	-360	360;
	8249	7602	0.215693	1.63338	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1485	7602	0.00048	0.00397	0.15883	2120.64254	2120.64254	2120.64254	0	0	1	-360	360;
	3747	7602	0.00049	0.00388	0.16183	2038.26942	2038.26942	2038.26942	0	0	1	-360	360;
	6567	7888	0.000211	0.00113	0.00284	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4939	4892	3e-05	0.00035	0.01109	1323.0885	1323.0885	1323.0885	0	0	1	-360	360;
	2871	2025	0.007409	0.042281	0.13096	268.151048	268.151048	268.151048	0	0	1	-360	360;
	3388	5932	0.00012	0.0007	0	274.895654	274.895654	274.895654	0	0	1	-360	360;
	788	5754	0	0.00011	0	879.526324	879.526324	879.526324	0	0	1	-360	360;
	788	5754	0.001	0.284017	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7965	7484	1.997645	4.178116	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7965	2829	0.00015	0.00152	0.04716	1353.70437	1353.70437	1353.70437	0	0	1	-360	360;
	7965	1172	0.00014	0.00141	0.04613	1327.27123	1327.27123	1327.27123	0	0	1	-360	360;
	7965	4303	0.00103	0.00914	0.34598	1661.76124	1661.76124	1661.76124	0	0	1	-360	360;
	7965	2661	0.00105	0.00896	0.35277	1540.35875	1540.35875	1540.35875	0	0	1	-360	360;
	2099	7965	0.334702	1.492936	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2099	7965	0.00142	0.01428	0.47948	1733.48436	1733.48436	1733.48436	0	0	1	-360	360;
	455	7965	0.340893	1.460526	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	455	7965	0.00139	0.01455	0.46949	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	355	4236	0.0004	0.00376	0.14492	1977.00478	1977.00478	1977.00478	0	0	1	-360	360;
	355	4830	0.00101	0.01129	0.37192	1726.50305	1726.50305	1726.50305	0	0	1	-360	360;
	355	4159	0.00076	0.0089	0.27101	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	355	4236	0.0004	0.00368	0.14771	2102.34252	2102.34252	2102.34252	0	0	1	-360	360;
	355	2431	0.00099	0.01151	0.36417	1815.95357	1815.95357	1815.95357	0	0	1	-360	360;
	355	849	0.00075	0.00824	0.26512	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7199	4243	0.00156	0.01568	0.52998	1775.97246	1775.97246	1775.97246	0	0	1	-360	360;
	7199	4243	0.00152	0.01599	0.51894	1741.76359	1741.76359	1741.76359	0	0	1	-360	360;
	438	6354	0.00539	0.01612	0.06314	600.763381	600.763381	600.763381	0	0	1	-360	360;
	438	6354	0.0048	0.018271	0.05625	583.994081	583.994081	583.994081	0	0	1	-360	360;
	7422	782	0.008729	0.034159	0.10083	437.515722	437.515722	437.515722	0	0	1	-360	360;
	7422	782	0.00855	0.03482	0.09873	423.284223	423.284223	423.284223	0	0	1	-360	360;
	3785	1738	0.00146	0.00651	0	1159.96336	1159.96336	1159.96336	0	0	1	-360	360;
	462	1738	0.134252	0.771745	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	462	1738	0.00072	0.00587	0.24108	1823.27778	1823.27778	1823.27778	0	0	1	-360	360;
	462	1738	0.00071	0.006	0.23653	1897.31901	1897.31901	1897.31901	0	0	1	-360	360;
	5784	1738	0.00013	0.00132	0.04328	1842.8086	1842.8086	1842.8086	0	0	1	-360	360;
	3086	1738	0.322853	1.362742	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2583	8120	0.161841	1.112975	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2583	8120	0.00356	0.022409	0.08577	600.594576	600.594576	600.594576	0	0	1	-360	360;
	395	8120	0.211508	1.021529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	395	8120	0.00838	0.038479	0.12634	226.665626	226.665626	226.665626	0	0	1	-360	360;
	438	8120	0.912748	1.922025	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5334	7431	0.005211	0.016901	0.06075	474.439581	474.439581	474.439581	0	0	1	-360	360;
	5334	7431	0.219566	0.646591	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5334	7431	0.00531	0.016539	0.06189	482.331151	482.331151	482.331151	0	0	1	-360	360;
	4759	6583	0.140519	0.391918	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1459	4759	0.00131	0.01484	0.47656	1834.75051	1834.75051	1834.75051	0	0	1	-360	360;
	4759	4159	0.05123	0.434093	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4759	4159	0.0006	0.00673	0.22193	1886.84935	1886.84935	1886.84935	0	0	1	-360	360;
	4759	7585	0.05191	0.445222	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4724	4759	0.736704	3.325	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4598	4759	2.622299	8.981301	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6967	5867	0	3.5e-05	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6967	5867	0.001	0.291627	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2233	6967	0.00016	0.00138	0.05229	1420.75507	1420.75507	1420.75507	0	0	1	-360	360;
	3902	6967	0.00017	0.00135	0.05331	1450.28625	1450.28625	1450.28625	0	0	1	-360	360;
	432	5867	0.00017	0.00133	0.05434	1254.63324	1254.63324	1254.63324	0	0	1	-360	360;
	7186	5867	0.00017	0.0013	0.05536	1345.87598	1345.87598	1345.87598	0	0	1	-360	360;
	7199	2233	0.00189	0.01446	0.59875	1341.3951	1341.3951	1341.3951	0	0	1	-360	360;
	3902	892	0.00074	0.00698	0.23645	1405.02919	1405.02919	1405.02919	0	0	1	-360	360;
	7199	432	0.00193	0.01416	0.61026	1352.28007	1352.28007	1352.28007	0	0	1	-360	360;
	7186	3803	0.00073	0.00711	0.23142	1365.21	1365.21	1365.21	0	0	1	-360	360;
	2099	7484	0.136731	0.910873	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2099	7484	0.00128	0.0134	0.44392	1876.3968	1876.3968	1876.3968	0	0	1	-360	360;
	455	7484	0.13394	0.92874	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5433	1995	0.271641	0.707964	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5433	1995	0.00037	0.0037	0.12426	1665.67807	1665.67807	1665.67807	0	0	1	-360	360;
	5433	1995	0.00038	0.00363	0.12685	1882.25328	1882.25328	1882.25328	0	0	1	-360	360;
	4724	1995	0.00156	0.01473	0.54676	1841.52778	1841.52778	1841.52778	0	0	1	-360	360;
	4598	1995	0.00159	0.01442	0.55727	1690.21828	1690.21828	1690.21828	0	0	1	-360	360;
	7525	3842	0.01049	0.04336	0.12284	443.873942	443.873942	443.873942	0	0	1	-360	360;
	1704	3842	0.010271	0.04418	0.12023	457.509478	457.509478	457.509478	0	0	1	-360	360;
	5334	3842	0.00881	0.037021	0.0865	470.866845	470.866845	470.866845	0	0	1	-360	360;
	2194	3842	0.004409	0.01682	0.05151	483.884316	483.884316	483.884316	0	0	1	-360	360;
	2194	3842	0.00432	0.017151	0.05046	541.125443	541.125443	541.125443	0	0	1	-360	360;
	6815	3842	0.01025	0.042341	0.11984	563.279758	563.279758	563.279758	0	0	1	-360	360;
	3422	2481	0.05386	0.203789	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3422	2481	0.005	0.017849	0.04792	366.690502	366.690502	366.690502	0	0	1	-360	360;
	3422	2481	0.004909	0.01824	0.04704	343.79428	343.79428	343.79428	0	0	1	-360	360;
	2481	2902	0.442645	0.98938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2481	2902	0.00789	0.025831	0.09286	527.989084	527.989084	527.989084	0	0	1	-360	360;
	2481	2902	0.00774	0.02638	0.0911	476.576135	476.576135	476.576135	0	0	1	-360	360;
	5860	462	0.00141	0.01209	0.47667	1774.50867	1774.50867	1774.50867	0	0	1	-360	360;
	5860	7627	0.104079	0.42973	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	892	4236	0.00139	0.01016	0.44742	1374.32183	1374.32183	1374.32183	0	0	1	-360	360;
	3803	4236	0.00142	0.00994	0.45571	1415.50938	1415.50938	1415.50938	0	0	1	-360	360;
	1993	347	9e-05	0.00064	0.01813	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1993	4303	0.195942	0.820914	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1993	2661	0.199778	0.804155	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	455	1993	0.02836	0.37876	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	347	4303	0.00109	0.00814	0.36229	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	455	347	0.00054	0.00424	0.18131	2110.64545	2110.64545	2110.64545	0	0	1	-360	360;
	2846	462	0.00263	0.02613	0.95605	1620.34728	1620.34728	1620.34728	0	0	1	-360	360;
	462	6273	0.00166	0.01885	0.60522	1577.82495	1577.82495	1577.82495	0	0	1	-360	360;
	462	3086	0.847576	2.035042	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1129	462	0.00094	0.00794	0.31303	1651.20141	1651.20141	1651.20141	0	0	1	-360	360;
	2583	1277	0.014081	0.055831	0.16451	469.442857	469.442857	469.442857	0	0	1	-360	360;
	2583	395	2.006198	3.572314	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2583	395	0.0122	0.05982	0.21601	242.566891	242.566891	242.566891	0	0	1	-360	360;
	2583	438	0.30564	1.254566	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2583	438	0.0235	0.076229	0.27448	565.155673	565.155673	565.155673	0	0	1	-360	360;
	2583	2902	0.029969	0.12889	0.33931	484.675378	484.675378	484.675378	0	0	1	-360	360;
	2583	2902	0.02932	0.119341	0.33193	510.429484	510.429484	510.429484	0	0	1	-360	360;
	6004	4095	0.00699	0.02886	0.08199	639.034256	639.034256	639.034256	0	0	1	-360	360;
	4095	782	0.004909	0.014719	0.05766	558.910244	558.910244	558.910244	0	0	1	-360	360;
	8127	6815	0.005591	0.01814	0.06525	472.099546	472.099546	472.099546	0	0	1	-360	360;
	8127	6815	0.00569	0.01776	0.06648	528.333174	528.333174	528.333174	0	0	1	-360	360;
	6583	4159	0.03624	0.415741	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6583	7585	0.03359	0.482777	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6583	7585	0.00088	0.01001	0.32139	1722.41181	1722.41181	1722.41181	0	0	1	-360	360;
	6583	849	0.001	0.00981	0.36424	2189.7382	2189.7382	2189.7382	0	0	1	-360	360;
	4724	6583	0.706579	3.541136	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4598	6583	2.102562	10.3982	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	438	1277	0.00962	0.03813	0.1124	535.249634	535.249634	535.249634	0	0	1	-360	360;
	6964	4696	0.04235	0.255499	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6964	6341	0.03604	0.287562	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6964	6341	0.00069	0.00676	0.25112	1972.63117	1972.63117	1972.63117	0	0	1	-360	360;
	6964	4202	0.02303	0.242479	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4094	6964	0.002	0.02189	0.72223	2071.81682	2071.81682	2071.81682	0	0	1	-360	360;
	4830	4598	0.00063	0.00569	0.22889	1854.0786	1854.0786	1854.0786	0	0	1	-360	360;
	2431	4724	0.00064	0.00557	0.2332	1917.16575	1917.16575	1917.16575	0	0	1	-360	360;
	4696	6341	0.001	0.193123	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4696	4202	0.01778	0.156988	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4696	4202	0.00104	0.00843	0.33955	1622.31327	1622.31327	1622.31327	0	0	1	-360	360;
	2099	4696	0.00096	0.01465	0.50375	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4094	4696	0.003	0.02517	1.09679	1966.46864	1966.46864	1966.46864	0	0	1	-360	360;
	6341	4202	0.462929	1.374931	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	96	6341	0.00244	0.01729	0.77702	2087.2406	2087.2406	2087.2406	0	0	1	-360	360;
	455	6341	0.00094	0.01356	0.4928	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7525	6163	0.01526	0.05118	0.17718	411.302069	411.302069	411.302069	0	0	1	-360	360;
	1704	6163	0.01555	0.05011	0.18059	430.163721	430.163721	430.163721	0	0	1	-360	360;
	395	438	0.26188	0.583802	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2846	7199	0.0036	0.04	1.31756	1574.50773	1574.50773	1574.50773	0	0	1	-360	360;
	2846	448	0.00363	0.03446	1.33225	1651.96819	1651.96819	1651.96819	0	0	1	-360	360;
	2846	3086	0.00099	0.01049	0.35944	1218.07533	1218.07533	1218.07533	0	0	1	-360	360;
	8141	4202	0.00332	0.03036	1.06338	1674.14028	1674.14028	1674.14028	0	0	1	-360	360;
	7905	1298	0.00027	0.00309	0.09942	1767.5945	1767.5945	1767.5945	0	0	1	-360	360;
	4780	1298	0.00031	0.00303	0.11268	1431.60055	1431.60055	1431.60055	0	0	1	-360	360;
	3422	2902	0.027099	0.21312	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3422	2902	0.00357	0.011021	0.03973	566.526314	566.526314	566.526314	0	0	1	-360	360;
	3422	2902	0.00351	0.01126	0.03899	507.61504	507.61504	507.61504	0	0	1	-360	360;
	7199	448	0.00037	0.00418	0.13766	1255.03539	1255.03539	1255.03539	0	0	1	-360	360;
	3752	6004	0.04124	0.192349	0	246.715395	246.715395	246.715395	0	0	1	-360	360;
	3752	6004	0.015669	0.064711	0.18353	636.937332	636.937332	636.937332	0	0	1	-360	360;
	3752	6004	0.016	0.06349	0.18743	632.552957	632.552957	632.552957	0	0	1	-360	360;
	6004	782	0.012281	0.03676	0.14412	620.421431	620.421431	620.421431	0	0	1	-360	360;
	5334	173	0.02043	0.19226	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5334	173	0.00287	0.01095	0.03367	508.141653	508.141653	508.141653	0	0	1	-360	360;
	5334	173	0.00281	0.011169	0.03299	496.105008	496.105008	496.105008	0	0	1	-360	360;
	7627	6273	0.00054	0.00596	0.1861	1602.9674	1602.9674	1602.9674	0	0	1	-360	360;
	6260	7627	0.00134	0.01244	0.48288	1896.33964	1896.33964	1896.33964	0	0	1	-360	360;
	892	3803	0.001	0.265693	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8141	892	0.15358	0.524508	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6502	892	0.613823	2.118352	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6502	892	0.00081	0.0077	0.27595	1361.28922	1361.28922	1361.28922	0	0	1	-360	360;
	892	4724	0.00228	0.02083	0.79746	1745.0874	1745.0874	1745.0874	0	0	1	-360	360;
	8141	3803	0.168047	0.592382	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8141	3803	0.0008	0.00785	0.27032	1575.06455	1575.06455	1575.06455	0	0	1	-360	360;
	6502	3803	0.698199	2.371468	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3803	4598	0.00232	0.02039	0.81251	1887.28704	1887.28704	1887.28704	0	0	1	-360	360;
	2099	5433	0.00177	0.01939	0.61823	2142.27821	2142.27821	2142.27821	0	0	1	-360	360;
	455	5433	0.00173	0.01976	0.60535	2179.99386	2179.99386	2179.99386	0	0	1	-360	360;
	8141	6502	0	3.5e-05	0	1618.88678	1618.88678	1618.88678	0	0	1	-360	360;
	8141	6502	0.001	0.213047	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8141	4598	0.00194	0.01469	0.60971	1573.43095	1573.43095	1573.43095	0	0	1	-360	360;
	96	8141	0.00123	0.00838	0.39207	1526.1717	1526.1717	1526.1717	0	0	1	-360	360;
	8141	2781	0.00121	0.0116	0.37729	1689.58445	1689.58445	1689.58445	0	0	1	-360	360;
	8141	4094	0.00346	0.03846	1.26564	1713.14596	1713.14596	1713.14596	0	0	1	-360	360;
	6502	4724	0.00198	0.01438	0.62144	1681.41118	1681.41118	1681.41118	0	0	1	-360	360;
	6502	4094	0.00338	0.0392	1.23927	1692.28497	1692.28497	1692.28497	0	0	1	-360	360;
	6502	177	0.00248	0.02446	0.79576	900.560418	900.560418	900.560418	0	0	1	-360	360;
	4303	2661	0.001	0.26367	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	455	4303	0.580512	1.382895	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5859	4303	0.00094	0.00807	0.28556	2253.73297	2253.73297	2253.73297	0	0	1	-360	360;
	1459	2661	0.00232	0.01638	0.73547	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2099	2661	0.00217	0.0131	0.57337	1883.21949	1883.21949	1883.21949	0	0	1	-360	360;
	455	2661	0.568421	1.409488	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1606	5071	0.005331	0.017341	0.06161	280.496344	280.496344	280.496344	0	0	1	-360	360;
	1459	7585	0.00199	0.02218	0.73055	1775.259	1775.259	1775.259	0	0	1	-360	360;
	1459	4724	0.00163	0.01513	0.51173	2166.84014	2166.84014	2166.84014	0	0	1	-360	360;
	1459	4598	0.00159	0.01541	0.50084	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	1459	5859	0.00136	0.00959	0.44952	2210.45571	2210.45571	2210.45571	0	0	1	-360	360;
	1459	7905	0.00042	0.00496	0.15691	1596.35467	1596.35467	1596.35467	0	0	1	-360	360;
	1459	4780	0.00041	0.00505	0.15358	960.180275	960.180275	960.180275	0	0	1	-360	360;
	5334	6815	0.016591	0.06199	0.17718	361.2087	361.2087	361.2087	0	0	1	-360	360;
	4159	7585	0	3.5e-05	0	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4159	7585	0.001	0.404868	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4724	4159	0.130547	2.317729	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4598	4159	0.431627	6.284488	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4724	7585	0.127403	2.421884	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4598	7585	0.397583	6.979917	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3086	6260	0.00356	0.03374	1.30463	1665.53524	1665.53524	1665.53524	0	0	1	-360	360;
	4724	4598	0.00164	0.14108	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4094	6260	0.00058	0.00658	0.21154	1940.73515	1940.73515	1940.73515	0	0	1	-360	360;
	4094	6260	0.00065	0.00645	0.23974	1914.51658	1914.51658	1914.51658	0	0	1	-360	360;
	2781	177	0.079079	0.405817	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2781	177	0.00139	0.01268	0.44775	1495.54727	1495.54727	1495.54727	0	0	1	-360	360;
	2099	455	0	3.5e-05	0	1001.89436	1001.89436	1001.89436	0	0	1	-360	360;
	2099	455	0.001	0.270111	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7626	6163	0.185091	0.839215	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7626	6163	0.004651	0.039669	0.11195	513.522857	513.522857	513.522857	0	0	1	-360	360;
	7626	6163	0.00476	0.042841	0.11444	513.586112	513.586112	513.586112	0	0	1	-360	360;
	4094	177	0.0018	0.02033	0.63949	1187.1047	1187.1047	1187.1047	0	0	1	-360	360;
	7073	6307	0.001	0.335748	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7905	7073	0.0001	0.00068	0.02901	1595.06003	1595.06003	1595.06003	0	0	1	-360	360;
	4780	6307	0.0001	0.00067	0.02955	1598.25023	1598.25023	1598.25023	0	0	1	-360	360;
	4458	8344	0.026107	0.07605	0.01003	75.253607	75.253607	75.253607	0	0	1	-360	360;
	7294	4458	0.019727	0.062124	0.0076	88.838081	88.838081	88.838081	0	0	1	-360	360;
	6355	8344	0.026603	0.074438	0.01022	85.798626	85.798626	85.798626	0	0	1	-360	360;
	7294	6355	0.020107	0.060851	0.00775	89.054899	89.054899	89.054899	0	0	1	-360	360;
	7379	6059	0.014017	0.040471	0.00536	67.068437	67.068437	67.068437	0	0	1	-360	360;
	6059	1269	0.012488	0.045289	0.00479	63.434161	63.434161	63.434161	0	0	1	-360	360;
	6538	5029	0.014281	0.039612	0.00546	75.12062	75.12062	75.12062	0	0	1	-360	360;
	5029	6632	0.010826	0.032686	0.00415	63.020721	63.020721	63.020721	0	0	1	-360	360;
	4060	6096	0.012554	0.049893	0.00487	73.452568	73.452568	73.452568	0	0	1	-360	360;
	6096	7359	0.00705	0.023174	0.00281	75.217216	75.217216	75.217216	0	0	1	-360	360;
	3992	4491	0.018818	0.054132	0.00479	75.272278	75.272278	75.272278	0	0	1	-360	360;
	1269	4491	0.032769	0.043661	0.00361	64.95989	64.95989	64.95989	0	0	1	-360	360;
	5175	4060	0.015322	0.049653	0.00588	110.17066	110.17066	110.17066	0	0	1	-360	360;
	2290	601	0.002413	0.006917	0.00069	97.308468	97.308468	97.308468	0	0	1	-360	360;
	601	4862	0.010537	0.025917	0.00292	70.822743	70.822743	70.822743	0	0	1	-360	360;
	3992	7379	0.009124	0.032314	0.0035	69.828962	69.828962	69.828962	0	0	1	-360	360;
	3992	6538	0.008934	0.032934	0.00343	75.170918	75.170918	75.170918	0	0	1	-360	360;
	8344	3710	0	0.000413	0	171.47303	171.47303	171.47303	0	0	1	-360	360;
	7622	3710	0.015562	0.044669	0.00562	171.47303	171.47303	171.47303	0	0	1	-360	360;
	5972	3710	0.013	0.049983	0.005	64.935884	64.935884	64.935884	0	0	1	-360	360;
	8344	4940	0.014347	0.041182	0.00519	96.365938	96.365938	96.365938	0	0	1	-360	360;
	8344	4940	0.014628	0.040322	0.00529	87.886977	87.886977	87.886977	0	0	1	-360	360;
	1565	626	0.017446	0.029752	0.00367	59.759685	59.759685	59.759685	0	0	1	-360	360;
	1565	5964	0.031579	0.070992	0.00668	58.214903	58.214903	58.214903	0	0	1	-360	360;
	3107	1565	0.007843	0.018793	0.0021	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4010	1565	0.007992	0.018413	0.00214	61.528715	61.528715	61.528715	0	0	1	-360	360;
	5362	6977	0.001271	0.016781	0.0439	567.726244	567.726244	567.726244	0	0	1	-360	360;
	5362	6977	0.00124	0.015539	0.04295	611.444627	611.444627	611.444627	0	0	1	-360	360;
	626	3992	0.033777	0.06419	0.00707	70.299179	70.299179	70.299179	0	0	1	-360	360;
	3992	1074	0.014347	0.028397	0.00316	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5964	3992	0.023331	0.052182	0.00489	95.2466	95.2466	95.2466	0	0	1	-360	360;
	3992	6283	0.009446	0.022636	0.00633	57.993703	57.993703	57.993703	0	0	1	-360	360;
	6632	1269	0.006992	0.025488	0.00272	71.084907	71.084907	71.084907	0	0	1	-360	360;
	6283	1074	0.003339	0.008851	0.00525	64.615229	64.615229	64.615229	0	0	1	-360	360;
	7622	4433	0.016471	0.06081	0.00633	161.816812	161.816812	161.816812	0	0	1	-360	360;
	6712	5972	0.020149	0.058727	0.00777	100.887681	100.887681	100.887681	0	0	1	-360	360;
	4940	555	0.013992	0.054	0.00539	98.415231	98.415231	98.415231	0	0	1	-360	360;
	4940	3633	0.013686	0.05	0.00527	119.656929	119.656929	119.656929	0	0	1	-360	360;
	4940	4110	0.000223	0.000826	6e-05	93.968745	93.968745	93.968745	0	0	1	-360	360;
	4940	3050	0.000248	0.00081	7e-05	101.674362	101.674362	101.674362	0	0	1	-360	360;
	8360	4433	0.036512	0.098843	0.01359	120.381498	120.381498	120.381498	0	0	1	-360	360;
	555	187	0.012876	0.037438	0.00495	93.319624	93.319624	93.319624	0	0	1	-360	360;
	3633	187	0.013124	0.036636	0.00504	90.08831	90.08831	90.08831	0	0	1	-360	360;
	2097	3107	0.013909	0.033041	0.00368	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4862	4010	0.014182	0.032372	0.00375	79.206631	79.206631	79.206631	0	0	1	-360	360;
	4677	7545	0.00474	0.027521	0.04875	305.277246	305.277246	305.277246	0	0	1	-360	360;
	4677	6228	0.00538	0.026969	0.05525	321.763044	321.763044	321.763044	0	0	1	-360	360;
	4420	7359	0.020372	0.06276	0.00772	112.689789	112.689789	112.689789	0	0	1	-360	360;
	8360	6712	0	0.000413	0	165.30343	165.30343	165.30343	0	0	1	-360	360;
	5588	5362	0.0053	0.02831	0.03929	424.51921	424.51921	424.51921	0	0	1	-360	360;
	5588	5362	0.00601	0.02774	0.04453	447.944712	447.944712	447.944712	0	0	1	-360	360;
	7545	5588	0.00113	0.006229	0.01189	295.312757	295.312757	295.312757	0	0	1	-360	360;
	6228	5588	0.001151	0.00611	0.01213	230.844233	230.844233	230.844233	0	0	1	-360	360;
	2290	2097	0.014364	0.029529	0.00388	75.306192	75.306192	75.306192	0	0	1	-360	360;
	4420	2290	0.009017	0.019727	0.00237	98.519258	98.519258	98.519258	0	0	1	-360	360;
	4420	2290	0.00919	0.019314	0.00242	99.908761	99.908761	99.908761	0	0	1	-360	360;
	6228	4157	9.1e-05	0.000219	0	287.78471	287.78471	287.78471	0	0	1	-360	360;
	7545	4157	9.1e-05	0.000229	0	253.434852	253.434852	253.434852	0	0	1	-360	360;
	5175	4420	0.00205	0.00562	0.00078	303.954998	303.954998	303.954998	0	0	1	-360	360;
	187	5175	0.016	0.049818	0.00614	57.157677	57.157677	57.157677	0	0	1	-360	360;
	187	5175	0.014331	0.054347	0.00545	59.042165	59.042165	59.042165	0	0	1	-360	360;
	7982	4420	0.00738	0.021149	0.00206	58.163842	58.163842	58.163842	0	0	1	-360	360;
	187	7982	0.011256	0.027769	0.0026	74.291452	74.291452	74.291452	0	0	1	-360	360;
	453	4420	0.007207	0.019587	0.00202	67.871502	67.871502	67.871502	0	0	1	-360	360;
	187	453	0.01195	0.024826	0.00292	68.898816	68.898816	68.898816	0	0	1	-360	360;
	5155	2873	0.00796	0.04504	0.08493	127.332495	127.332495	127.332495	0	0	1	-360	360;
	5155	6289	0.00844	0.068649	0.02157	144.176862	144.176862	144.176862	0	0	1	-360	360;
	5155	5328	0.021551	0.0842	0.21639	261.827885	261.827885	261.827885	0	0	1	-360	360;
	2017	7949	0.002809	0.01632	0.00585	311.769145	311.769145	311.769145	0	0	1	-360	360;
	2017	326	0.00028	0.00084	0.02673	251.862964	251.862964	251.862964	0	0	1	-360	360;
	2017	326	0.00032	0.000822	0.03029	252.241763	252.241763	252.241763	0	0	1	-360	360;
	2017	2567	0.007711	0.054271	0.02431	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2017	4104	0.002298	0.017129	0.18446	311.769145	311.769145	311.769145	0	0	1	-360	360;
	7281	8097	0.004711	0.026049	0.01193	201.457428	201.457428	201.457428	0	0	1	-360	360;
	7281	4947	0.008462	0.03792	0.01733	195.893647	195.893647	195.893647	0	0	1	-360	360;
	7147	1624	0.00095	0.01117	0.24986	1107.98694	1107.98694	1107.98694	0	0	1	-360	360;
	7147	4640	0.00142	0.01429	0.37777	1182.81337	1182.81337	1182.81337	0	0	1	-360	360;
	5258	7158	0.000569	0.005382	0.0929	261.340746	261.340746	261.340746	0	0	1	-360	360;
	3017	1889	0.00984	0.04424	0.02041	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3017	3903	0.0166	0.105449	0.07307	105.444482	105.444482	105.444482	0	0	1	-360	360;
	767	6768	0.007209	0.034969	0.01431	105.128816	105.128816	105.128816	0	0	1	-360	360;
	767	1525	0.023951	0.07004	0.02867	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8097	2078	0.002258	0.01644	0.00875	252.722667	252.722667	252.722667	0	0	1	-360	360;
	4408	1259	0.00948	0.02892	0.04356	107.211434	107.211434	107.211434	0	0	1	-360	360;
	4408	2327	0.01224	0.04384	0.06914	162.80273	162.80273	162.80273	0	0	1	-360	360;
	2136	2327	0.016	0.062369	0.01861	155.274804	155.274804	155.274804	0	0	1	-360	360;
	2136	2046	0.020422	0.082729	0.02999	129.87731	129.87731	129.87731	0	0	1	-360	360;
	2136	2046	0.020009	0.08436	0.02938	103.959422	103.959422	103.959422	0	0	1	-360	360;
	4044	2046	0.010702	0.030338	0.01312	141.720381	141.720381	141.720381	0	0	1	-360	360;
	4044	2046	0.010902	0.02968	0.01336	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5090	2413	0.00848	0.040809	0.01717	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5090	3759	0.012862	0.05772	0.02574	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5084	1624	0.00047	0.00546	0.18642	892.452308	892.452308	892.452308	0	0	1	-360	360;
	6993	6958	0.002871	0.02108	0.02396	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6993	3903	0.004849	0.02916	0.01015	149.723755	149.723755	149.723755	0	0	1	-360	360;
	6993	3903	0.004751	0.029711	0.00994	142.019679	142.019679	142.019679	0	0	1	-360	360;
	4686	3903	0.001569	0.005489	0.00176	118.542943	118.542943	118.542943	0	0	1	-360	360;
	6033	5889	0.0003	0.00316	0.05829	1226.89559	1226.89559	1226.89559	0	0	1	-360	360;
	6033	4640	0.00114	0.01207	0.35564	1152.57333	1152.57333	1152.57333	0	0	1	-360	360;
	6033	5523	0.00147	0.01724	0.82394	1114.65825	1114.65825	1114.65825	0	0	1	-360	360;
	6033	8374	0.00039	0.00552	0.12156	1269.41792	1269.41792	1269.41792	0	0	1	-360	360;
	7949	2567	0.004991	0.036511	0.01915	113.169082	113.169082	113.169082	0	0	1	-360	360;
	7949	7341	0.007302	0.041071	0.05157	135.568656	135.568656	135.568656	0	0	1	-360	360;
	1491	5424	0.00022	0.00305	0.06498	904.005329	904.005329	904.005329	0	0	1	-360	360;
	1491	260	0.00074	0.00834	0.21328	961.139242	961.139242	961.139242	0	0	1	-360	360;
	1491	3841	0.00126	0.0174	0.3602	882.877773	882.877773	882.877773	0	0	1	-360	360;
	3850	2873	0.00216	0.012578	0.10273	233.851801	233.851801	233.851801	0	0	1	-360	360;
	3850	585	0.001329	0.0106	0.00365	222.928449	222.928449	222.928449	0	0	1	-360	360;
	3850	6289	0.006049	0.063742	0.02119	114.194803	114.194803	114.194803	0	0	1	-360	360;
	4880	7664	0.006769	0.02648	0.03167	136.100742	136.100742	136.100742	0	0	1	-360	360;
	4880	7368	0.016022	0.047791	0.01956	111.000728	111.000728	111.000728	0	0	1	-360	360;
	4880	2567	0.02884	0.096209	0.06197	260.402061	260.402061	260.402061	0	0	1	-360	360;
	6752	7368	0.00656	0.024578	0.0081	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6752	1259	0.00388	0.017342	0.01777	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6752	7024	0.01772	0.064978	0.04576	237.501838	237.501838	237.501838	0	0	1	-360	360;
	2873	589	0.01312	0.052809	0.07034	311.769145	311.769145	311.769145	0	0	1	-360	360;
	2873	585	0.001738	0.002889	0.11805	178.690486	178.690486	178.690486	0	0	1	-360	360;
	589	7287	0.001169	0.003489	0.13077	309.867094	309.867094	309.867094	0	0	1	-360	360;
	3774	6858	0.00368	0.035022	0.01377	134.125944	134.125944	134.125944	0	0	1	-360	360;
	3774	6205	0.001391	0.010338	0.00535	114.041516	114.041516	114.041516	0	0	1	-360	360;
	3774	4187	0.004222	0.030831	0.01604	143.022537	143.022537	143.022537	0	0	1	-360	360;
	2735	5187	0.004271	0.036089	0.06696	132.435116	132.435116	132.435116	0	0	1	-360	360;
	2735	6858	0.00596	0.04024	0.03238	150.587875	150.587875	150.587875	0	0	1	-360	360;
	3505	4652	0.011529	0.046698	0.01283	151.232977	151.232977	151.232977	0	0	1	-360	360;
	3505	2327	0.02252	0.087338	0.0258	120.255855	120.255855	120.255855	0	0	1	-360	360;
	5993	3342	0.004071	0.035591	0.01535	163.298183	163.298183	163.298183	0	0	1	-360	360;
	5993	4187	0.004138	0.0386	0.09434	165.930294	165.930294	165.930294	0	0	1	-360	360;
	2570	2281	0.010911	0.027231	0.2464	107.45981	107.45981	107.45981	0	0	1	-360	360;
	71	5902	0.00256	0.024622	0.00972	111.260795	111.260795	111.260795	0	0	1	-360	360;
	71	4187	0.005129	0.040671	0.01966	128.463697	128.463697	128.463697	0	0	1	-360	360;
	7664	6768	0.005462	0.032671	0.0108	115.593607	115.593607	115.593607	0	0	1	-360	360;
	7664	1525	0.002871	0.014209	0.00608	111.430969	111.430969	111.430969	0	0	1	-360	360;
	7664	1525	0.002929	0.013911	0.0062	111.979163	111.979163	111.979163	0	0	1	-360	360;
	7664	180	0.003422	0.011009	0.00374	237.420518	237.420518	237.420518	0	0	1	-360	360;
	5902	6858	0.00212	0.016769	0.00796	135.197391	135.197391	135.197391	0	0	1	-360	360;
	1784	4640	0.00125	0.01387	0.42569	1729.68864	1729.68864	1729.68864	0	0	1	-360	360;
	7808	2327	0.004289	0.017071	0.06656	272.622372	272.622372	272.622372	0	0	1	-360	360;
	7808	4793	0.0084	0.024738	0.03888	256.484682	256.484682	256.484682	0	0	1	-360	360;
	2281	3903	0.01044	0.039671	0.01267	137.789751	137.789751	137.789751	0	0	1	-360	360;
	2281	3903	0.010218	0.040418	0.0124	168.315588	168.315588	168.315588	0	0	1	-360	360;
	2281	3407	0.008351	0.02432	0.03888	238.519244	238.519244	238.519244	0	0	1	-360	360;
	3342	2188	0.00188	0.006711	0.1829	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3342	7341	0.00132	0.00352	-0.21274	129.896796	129.896796	129.896796	0	0	1	-360	360;
	6722	3841	0.00113	0.01322	0.39074	1418.87465	1418.87465	1418.87465	0	0	1	-360	360;
	6722	1624	0.00098	0.01301	0.25792	1061.47012	1061.47012	1061.47012	0	0	1	-360	360;
	1889	4947	0.005271	0.05104	0.01852	165.312472	165.312472	165.312472	0	0	1	-360	360;
	1889	6844	0.017458	0.109631	0.03602	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1889	201	0.002831	0.01016	0.01323	163.125931	163.125931	163.125931	0	0	1	-360	360;
	1889	201	0.001978	0.010902	0.01288	176.14515	176.14515	176.14515	0	0	1	-360	360;
	260	3841	0.00059	0.00764	0.16549	805.880667	805.880667	805.880667	0	0	1	-360	360;
	1225	7368	0.000231	0.000769	0.00047	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5187	2078	0.002809	0.02468	0.05365	174.756998	174.756998	174.756998	0	0	1	-360	360;
	5187	7158	0.009942	0.050222	0.02026	217.422086	217.422086	217.422086	0	0	1	-360	360;
	3841	7437	0.00117	0.01756	0.38857	1559.02867	1559.02867	1559.02867	0	0	1	-360	360;
	3841	8216	9e-05	0.0017	0.03454	883.620199	883.620199	883.620199	0	0	1	-360	360;
	2078	5798	0.002569	0.02344	0.00926	112.412782	112.412782	112.412782	0	0	1	-360	360;
	2078	4292	0.00144	0.011369	0.00486	147.394839	147.394839	147.394839	0	0	1	-360	360;
	8226	7434	0.002129	0.019689	0.00821	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3903	6844	0.00756	0.048378	0.05422	111.710002	111.710002	111.710002	0	0	1	-360	360;
	3903	4652	0.031138	0.104969	0.0381	137.536179	137.536179	137.536179	0	0	1	-360	360;
	7437	5523	0.00043	0.00623	0.16698	1213.77676	1213.77676	1213.77676	0	0	1	-360	360;
	6858	2303	0.000769	0.005769	0.00292	117.819379	117.819379	117.819379	0	0	1	-360	360;
	666	4187	0.01464	0.043018	0.01701	115.576979	115.576979	115.576979	0	0	1	-360	360;
	666	121	0.006609	0.014329	0.38187	129.547874	129.547874	129.547874	0	0	1	-360	360;
	3825	5523	0.00021	0.00252	0.07942	859.773047	859.773047	859.773047	0	0	1	-360	360;
	3985	4187	0.002689	0.019791	0.00972	148.552802	148.552802	148.552802	0	0	1	-360	360;
	3985	4187	0.002738	0.01936	0.0099	157.00902	157.00902	157.00902	0	0	1	-360	360;
	3985	4187	0.001338	0.010018	0.02019	196.492764	196.492764	196.492764	0	0	1	-360	360;
	6205	4187	0.001698	0.012871	0.0062	151.276365	151.276365	151.276365	0	0	1	-360	360;
	1500	4187	0.00384	0.028342	0.01425	137.819629	137.819629	137.819629	0	0	1	-360	360;
	2567	7434	0.00616	0.033618	0.05121	159.817281	159.817281	159.817281	0	0	1	-360	360;
	2567	7024	0.01488	0.045698	0.01825	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5889	8374	4e-05	0.00051	0.01741	1226.631	1226.631	1226.631	0	0	1	-360	360;
	1525	2413	0.028751	0.090409	0.09316	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1525	3759	0.00092	0.004609	0.00187	247.786323	247.786323	247.786323	0	0	1	-360	360;
	3582	2327	0.005151	0.012329	0.29654	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5798	4292	0.002129	0.018	0.03246	104.629465	104.629465	104.629465	0	0	1	-360	360;
	201	5328	0.008929	0.041471	0.04082	136.775462	136.775462	136.775462	0	0	1	-360	360;
	201	1936	0.0222	0.091822	0.02525	165.504469	165.504469	165.504469	0	0	1	-360	360;
	3407	4793	0.005671	0.021169	0.02506	251.546518	251.546518	251.546518	0	0	1	-360	360;
	4292	1936	0.00956	0.035551	0.05292	278.66212	278.66212	278.66212	0	0	1	-360	360;
	2046	7434	0.003169	0.021911	0.00964	174.594099	174.594099	174.594099	0	0	1	-360	360;
	2046	7434	0.00324	0.021471	0.00983	244.144599	244.144599	244.144599	0	0	1	-360	360;
	2046	4187	0.01168	0.055391	0.08798	192.709185	192.709185	192.709185	0	0	1	-360	360;
	2046	4187	0.011902	0.054209	0.08964	211.788678	211.788678	211.788678	0	0	1	-360	360;
	4187	121	0.01908	0.062769	0.36975	111.347571	111.347571	111.347571	0	0	1	-360	360;
	3759	4104	0.001591	0.011098	0.00459	311.769145	311.769145	311.769145	0	0	1	-360	360;
	5949	1177	0.003289	0.019091	0.03572	462.056561	462.056561	462.056561	0	0	1	-360	360;
	5949	1518	0.004169	0.025349	0.04586	484.538962	484.538962	484.538962	0	0	1	-360	360;
	5949	3729	0.00439	0.023909	0.05804	476.201562	476.201562	476.201562	0	0	1	-360	360;
	4289	3862	0.00156	0.01956	0.53992	1340.05768	1340.05768	1340.05768	0	0	1	-360	360;
	3862	6702	0.00029	0.00391	0.12748	1854.05754	1854.05754	1854.05754	0	0	1	-360	360;
	3862	5327	0.00179	0.02522	0.80239	1653.72685	1653.72685	1653.72685	0	0	1	-360	360;
	1345	3862	0.00079	0.00756	0.26735	1866.43987	1866.43987	1866.43987	0	0	1	-360	360;
	3862	1341	0.00027	0.00281	0.06816	1665.43388	1665.43388	1665.43388	0	0	1	-360	360;
	3862	1341	0.00026	0.00286	0.06674	1571.21617	1571.21617	1571.21617	0	0	1	-360	360;
	2047	239	0.000281	0.00213	0.0052	359.676112	359.676112	359.676112	0	0	1	-360	360;
	2047	7733	0.000419	0.00299	0.00785	258.637344	258.637344	258.637344	0	0	1	-360	360;
	3408	4372	0.001659	0.009091	0.018	364.84469	364.84469	364.84469	0	0	1	-360	360;
	3408	5661	0.00439	0.03324	0.08166	396.716573	396.716573	396.716573	0	0	1	-360	360;
	5438	7865	0.000841	0.00612	0.01716	405.018154	405.018154	405.018154	0	0	1	-360	360;
	5438	181	0.003341	0.022841	0.058	551.997214	551.997214	551.997214	0	0	1	-360	360;
	5438	3192	0.00581	0.04232	0.12566	262.560266	262.560266	262.560266	0	0	1	-360	360;
	5438	1302	0.00495	0.02919	0.04457	444.914974	444.914974	444.914974	0	0	1	-360	360;
	5438	5078	0.01789	0.07795	0.16192	398.978492	398.978492	398.978492	0	0	1	-360	360;
	5438	4507	0.00186	0.014289	0.03878	386.682733	386.682733	386.682733	0	0	1	-360	360;
	5438	2139	0.002281	0.014401	0.04256	254.473216	254.473216	254.473216	0	0	1	-360	360;
	4106	2686	0.00031	0.00132	0.00239	267.777999	267.777999	267.777999	0	0	1	-360	360;
	4106	6291	0.001349	0.0077	0.01017	399.585888	399.585888	399.585888	0	0	1	-360	360;
	5171	5381	0.00586	0.02574	0.06732	485.221043	485.221043	485.221043	0	0	1	-360	360;
	5171	5482	0.00556	0.020211	0.06365	607.964868	607.964868	607.964868	0	0	1	-360	360;
	4719	2646	0.00026	0.001159	0.12727	568.505494	568.505494	568.505494	0	0	1	-360	360;
	2646	6034	0.000599	0.002419	0.31101	591.796866	591.796866	591.796866	0	0	1	-360	360;
	6056	6476	0.0005	0.00386	0.009	261.154568	261.154568	261.154568	0	0	1	-360	360;
	6056	5566	0.00107	0.008289	0.01652	536.373353	536.373353	536.373353	0	0	1	-360	360;
	6056	5566	0.001091	0.00811	0.01683	245.223961	245.223961	245.223961	0	0	1	-360	360;
	6719	6056	0.00038	0.00255	0.00726	548.622625	548.622625	548.622625	0	0	1	-360	360;
	4719	5858	0.000271	0.000979	0.14244	278.737031	278.737031	278.737031	0	0	1	-360	360;
	2142	1493	7e-05	0.00076	0.10408	266.750304	266.750304	266.750304	0	0	1	-360	360;
	2142	7772	0.000521	0.003901	0.0096	241.950732	241.950732	241.950732	0	0	1	-360	360;
	4289	7477	0.0016	0.01839	0.38509	1263.04675	1263.04675	1263.04675	0	0	1	-360	360;
	5252	1267	0.007651	0.058781	0.16238	481.88494	481.88494	481.88494	0	0	1	-360	360;
	5252	5665	0.00394	0.02182	0.05366	226.294482	226.294482	226.294482	0	0	1	-360	360;
	5252	7558	0.003461	0.03005	0.07832	554.41003	554.41003	554.41003	0	0	1	-360	360;
	5252	681	0.00337	0.01831	0.0459	450.192533	450.192533	450.192533	0	0	1	-360	360;
	7511	3909	0.000581	0.00194	0.1989	505.111915	505.111915	505.111915	0	0	1	-360	360;
	6699	3909	0.000539	0.001719	0.17952	234.490512	234.490512	234.490512	0	0	1	-360	360;
	6014	841	0.00249	0.03075	0.59862	1587.24679	1587.24679	1587.24679	0	0	1	-360	360;
	841	4014	0.0014	0.01409	0.33536	1706.79453	1706.79453	1706.79453	0	0	1	-360	360;
	7106	5981	0.008	0.050031	0.07101	399.584745	399.584745	399.584745	0	0	1	-360	360;
	7106	5078	0.002211	0.01562	0.01967	575.604096	575.604096	575.604096	0	0	1	-360	360;
	1291	3305	0.00281	0.021831	0.03108	492.299831	492.299831	492.299831	0	0	1	-360	360;
	1291	7092	0.000711	0.004919	0.00656	257.315096	257.315096	257.315096	0	0	1	-360	360;
	2650	5529	0.001841	0.00738	0.00847	364.989108	364.989108	364.989108	0	0	1	-360	360;
	2650	3168	0.002849	0.008979	0.01308	367.707909	367.707909	367.707909	0	0	1	-360	360;
	7571	3305	0.00232	0.014099	0.04337	250.808266	250.808266	250.808266	0	0	1	-360	360;
	7571	3305	0.00237	0.0138	0.04419	254.378716	254.378716	254.378716	0	0	1	-360	360;
	7571	6263	0.00081	0.004591	0.00913	491.232126	491.232126	491.232126	0	0	1	-360	360;
	7571	6007	0.002219	0.010469	0.03064	283.924281	283.924281	283.924281	0	0	1	-360	360;
	5683	1713	0.001081	0.00851	0.05702	253.810187	253.810187	253.810187	0	0	1	-360	360;
	1713	7234	0.00087	0.006969	0.02047	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1713	7234	0.00088	0.006831	0.02087	242.549363	242.549363	242.549363	0	0	1	-360	360;
	1713	4226	0.000349	0.00237	0.00773	268.813696	268.813696	268.813696	0	0	1	-360	360;
	644	5895	0.00232	0.00924	0.01751	382.371139	382.371139	382.371139	0	0	1	-360	360;
	644	3452	0.00812	0.03669	0.06171	411.64692	411.64692	411.64692	0	0	1	-360	360;
	1996	6093	0.003521	0.02825	0.06753	488.000812	488.000812	488.000812	0	0	1	-360	360;
	1996	4322	0.00264	0.017651	0.05145	224.960422	224.960422	224.960422	0	0	1	-360	360;
	6936	5895	0.00486	0.02149	0.0442	472.399433	472.399433	472.399433	0	0	1	-360	360;
	6936	6140	0.006831	0.034669	0.05963	437.514579	437.514579	437.514579	0	0	1	-360	360;
	6936	6418	0.017211	0.108151	0.15779	388.117391	388.117391	388.117391	0	0	1	-360	360;
	6936	7444	0.00063	0.00475	0.01195	437.976794	437.976794	437.976794	0	0	1	-360	360;
	6936	7444	0.00062	0.004841	0.0117	380.949818	380.949818	380.949818	0	0	1	-360	360;
	6014	2271	0.0009	0.01092	0.30354	1871.38016	1871.38016	1871.38016	0	0	1	-360	360;
	2271	6276	0.00124	0.01096	0.41979	1808.37924	1808.37924	1808.37924	0	0	1	-360	360;
	2815	2681	0.003211	0.00176	0.00454	310.027049	310.027049	310.027049	0	0	1	-360	360;
	2815	7092	0.001521	0.008031	0.02227	304.773496	304.773496	304.773496	0	0	1	-360	360;
	4925	1736	0.0005	0.00441	0.12541	1435.04019	1435.04019	1435.04019	0	0	1	-360	360;
	4925	1364	0.00247	0.02573	0.62683	1724.72267	1724.72267	1724.72267	0	0	1	-360	360;
	4925	1165	0.00042	0.00367	0.09617	1308.03988	1308.03988	1308.03988	0	0	1	-360	360;
	6736	3896	0.0023	0.01644	0.024	459.626598	459.626598	459.626598	0	0	1	-360	360;
	6736	3896	0.00226	0.01676	0.0235	469.735886	469.735886	469.735886	0	0	1	-360	360;
	6736	3896	0.0023	0.0162	0.02471	479.874133	479.874133	479.874133	0	0	1	-360	360;
	6736	3896	0.00225	0.015	0.02417	429.145933	429.145933	429.145933	0	0	1	-360	360;
	2129	5393	0.007281	0.035729	0.0942	439.96893	439.96893	439.96893	0	0	1	-360	360;
	2038	5393	0.001919	0.00945	0.02493	455.394262	455.394262	455.394262	0	0	1	-360	360;
	3478	1067	0.005211	0.02981	0.03943	520.448463	520.448463	520.448463	0	0	1	-360	360;
	3478	6291	0.000659	0.00586	0.00221	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3478	6291	0.000591	0.00664	0.00197	263.008763	263.008763	263.008763	0	0	1	-360	360;
	3478	2676	0.011969	0.07424	0.10468	542.634786	542.634786	542.634786	0	0	1	-360	360;
	5683	2787	0.000531	0.003919	0.03462	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3754	3070	0.001521	0.00718	0.01157	440.06343	440.06343	440.06343	0	0	1	-360	360;
	3754	3070	0.001211	0.00613	0.00912	478.677251	478.677251	478.677251	0	0	1	-360	360;
	3754	4375	0.00676	0.03725	0.05131	419.611652	419.611652	419.611652	0	0	1	-360	360;
	3754	4320	0.00399	0.018091	0.03041	386.883547	386.883547	386.883547	0	0	1	-360	360;
	6043	4745	0.000341	0.00339	0.00786	279.441976	279.441976	279.441976	0	0	1	-360	360;
	6043	3729	0.001271	0.0102	0.01646	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6043	3729	0.00125	0.010401	0.01613	331.928346	331.928346	331.928346	0	0	1	-360	360;
	4418	460	0.00261	0.016651	0.02966	306.664653	306.664653	306.664653	0	0	1	-360	360;
	8313	460	0.000219	0.002409	0.32232	528.080537	528.080537	528.080537	0	0	1	-360	360;
	6140	6315	0.00601	0.02814	0.04548	448.899626	448.899626	448.899626	0	0	1	-360	360;
	6315	6350	0.008969	0.03787	0.069	275.352916	275.352916	275.352916	0	0	1	-360	360;
	6315	6241	0.00232	0.01094	0.01765	443.904807	443.904807	443.904807	0	0	1	-360	360;
	6014	6276	0.00196	0.02382	0.66228	1652.93506	1652.93506	1652.93506	0	0	1	-360	360;
	6014	8051	0.00422	0.03598	1.10237	1220.86404	1220.86404	1220.86404	0	0	1	-360	360;
	6014	8051	0.0043	0.03522	1.12317	1283.7399	1283.7399	1283.7399	0	0	1	-360	360;
	7709	6014	0.00208	0.01691	0.55078	1561.76406	1561.76406	1561.76406	0	0	1	-360	360;
	6014	2476	0.00544	0.05439	1.29424	1279.06815	1279.06815	1279.06815	0	0	1	-360	360;
	6014	6809	0.00239	0.02598	0.60825	1348.96811	1348.96811	1348.96811	0	0	1	-360	360;
	6014	346	0.00396	0.03096	1.00098	1731.54865	1731.54865	1731.54865	0	0	1	-360	360;
	6014	346	0.00353	0.03509	0.89178	1550.82841	1550.82841	1550.82841	0	0	1	-360	360;
	3320	8039	0.000531	0.003729	0.005	521.784809	521.784809	521.784809	0	0	1	-360	360;
	3320	8039	0.00038	0.0025	0.0074	222.734321	222.734321	222.734321	0	0	1	-360	360;
	3320	8039	0.00039	0.00244	0.00755	239.275752	239.275752	239.275752	0	0	1	-360	360;
	3320	4419	0.00836	0.044521	0.07368	411.996725	411.996725	411.996725	0	0	1	-360	360;
	1475	966	0.00264	0.03029	0.63268	1383.34745	1383.34745	1383.34745	0	0	1	-360	360;
	1475	7989	0.00396	0.04589	0.96451	787.658973	787.658973	787.658973	0	0	1	-360	360;
	7749	6049	0.00411	0.021729	0.03745	410.585311	410.585311	410.585311	0	0	1	-360	360;
	7749	6705	0.00607	0.0285	0.05541	460.164261	460.164261	460.164261	0	0	1	-360	360;
	7363	2261	0.00075	0.0047	0.01387	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7631	2261	0.00114	0.007969	0.02092	560.785778	560.785778	560.785778	0	0	1	-360	360;
	2854	2645	0.00036	0.00324	0.08914	1326.84802	1326.84802	1326.84802	0	0	1	-360	360;
	2854	8148	0.00094	0.00874	0.23192	1231.90894	1231.90894	1231.90894	0	0	1	-360	360;
	2241	8069	0.00949	0.05145	0.07187	493.945591	493.945591	493.945591	0	0	1	-360	360;
	2241	4463	0.011719	0.0337	0.06488	415.536309	415.536309	415.536309	0	0	1	-360	360;
	2797	2241	0.0022	0.012331	0.0411	436.999779	436.999779	436.999779	0	0	1	-360	360;
	3810	2037	0.00038	0.00448	0.09815	1411.45763	1411.45763	1411.45763	0	0	1	-360	360;
	3810	5327	0.00098	0.01126	0.25795	1011.50049	1011.50049	1011.50049	0	0	1	-360	360;
	36	5370	0.00055	0.00621	0.13477	1794.75756	1794.75756	1794.75756	0	0	1	-360	360;
	36	1911	0.00163	0.01287	0.40301	1713.95157	1713.95157	1713.95157	0	0	1	-360	360;
	5708	5559	0.0018	0.01886	0.04523	281.768674	281.768674	281.768674	0	0	1	-360	360;
	5708	5559	0.0047	0.024969	0.03581	249.747039	249.747039	249.747039	0	0	1	-360	360;
	5708	2410	0.009711	0.059669	0.08431	406.681061	406.681061	406.681061	0	0	1	-360	360;
	5708	4721	0.005271	0.03451	0.09799	287.615905	287.615905	287.615905	0	0	1	-360	360;
	840	1069	0.016169	0.11936	0.16403	429.229764	429.229764	429.229764	0	0	1	-360	360;
	840	1224	0.00226	0.014789	0.02294	426.345207	426.345207	426.345207	0	0	1	-360	360;
	7865	1293	0.00151	0.01081	0.03097	600.226099	600.226099	600.226099	0	0	1	-360	360;
	643	4227	0.00228	0.0217	0.76722	1823.53645	1823.53645	1823.53645	0	0	1	-360	360;
	643	3264	0.00137	0.01593	0.4613	1728.82445	1728.82445	1728.82445	0	0	1	-360	360;
	7394	3458	0.00099	0.0121	0.24926	1117.00137	1117.00137	1117.00137	0	0	1	-360	360;
	7394	2059	0	0.01881	0.41004	1140.86168	1140.86168	1140.86168	0	0	1	-360	360;
	7394	7034	0.00143	0.0163	0.35353	1212.8507	1212.8507	1212.8507	0	0	1	-360	360;
	7394	6606	0.00057	0.00562	0.14915	1147.06634	1147.06634	1147.06634	0	0	1	-360	360;
	7394	6606	0.00055	0.00572	0.14604	1194.53357	1194.53357	1194.53357	0	0	1	-360	360;
	7394	5893	0.00057	0.00658	0.19029	1452.42665	1452.42665	1452.42665	0	0	1	-360	360;
	7394	4878	0.00075	0.0082	0.1938	1269.03881	1269.03881	1269.03881	0	0	1	-360	360;
	7394	732	0.00253	0.02153	0.62922	1318.16597	1318.16597	1318.16597	0	0	1	-360	360;
	181	7782	0.00236	0.011219	0.03658	545.863052	545.863052	545.863052	0	0	1	-360	360;
	2906	6397	0.00173	0.0154	0.43998	1445.28344	1445.28344	1445.28344	0	0	1	-360	360;
	2906	2505	0.00385	0.04172	1.00164	1188.49675	1188.49675	1188.49675	0	0	1	-360	360;
	2906	2505	0.00377	0.04251	0.98033	1182.04396	1182.04396	1182.04396	0	0	1	-360	360;
	2906	1364	0.00199	0.02121	0.5037	1595.98938	1595.98938	1595.98938	0	0	1	-360	360;
	2906	1032	0.00236	0.01933	0.6099	1191.66128	1191.66128	1191.66128	0	0	1	-360	360;
	2906	1032	0.00241	0.01891	0.62119	1148.42943	1148.42943	1148.42943	0	0	1	-360	360;
	2091	3113	0.000521	0.004729	0.15416	230.886149	230.886149	230.886149	0	0	1	-360	360;
	2091	7776	0.00094	0.005581	0.02629	247.599434	247.599434	247.599434	0	0	1	-360	360;
	4219	991	0.00134	0.01324	0.44968	1220.65737	1220.65737	1220.65737	0	0	1	-360	360;
	4219	2580	0.00073	0.00638	0.24442	1304.54495	1304.54495	1304.54495	0	0	1	-360	360;
	1326	2229	0.00248	0.02774	0.83587	1840.12191	1840.12191	1840.12191	0	0	1	-360	360;
	1326	431	0.00152	0.013	0.37644	1600.77106	1600.77106	1600.77106	0	0	1	-360	360;
	1326	431	0.00155	0.01272	0.38342	1692.18295	1692.18295	1692.18295	0	0	1	-360	360;
	1326	7441	0.00117	0.01301	0.28981	1370.095	1370.095	1370.095	0	0	1	-360	360;
	1326	2832	0.00197	0.02054	0.48554	1123.96622	1123.96622	1123.96622	0	0	1	-360	360;
	1326	8162	0.00157	0.01909	0.53074	1640.52969	1640.52969	1640.52969	0	0	1	-360	360;
	2787	4053	0.00107	0.008669	0.04933	241.059453	241.059453	241.059453	0	0	1	-360	360;
	4053	1367	0.000469	0.00314	0.13255	343.398368	343.398368	343.398368	0	0	1	-360	360;
	1079	640	0.000409	0.00305	0.00769	268.686806	268.686806	268.686806	0	0	1	-360	360;
	1126	640	0.00569	0.032979	0.10568	275.310238	275.310238	275.310238	0	0	1	-360	360;
	3181	8052	0.00087	0.006771	0.01515	529.81432	529.81432	529.81432	0	0	1	-360	360;
	3181	7782	0.002151	0.014039	0.04245	544.945481	544.945481	544.945481	0	0	1	-360	360;
	5770	6152	9.9e-05	0.00043	0.04608	555.226623	555.226623	555.226623	0	0	1	-360	360;
	5856	5770	0.000151	0.0005	0.0704	282.136008	282.136008	282.136008	0	0	1	-360	360;
	7105	7648	0.001531	0.00812	0.02192	238.365421	238.365421	238.365421	0	0	1	-360	360;
	7105	7648	0.00113	0.00857	0.02074	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3874	6525	0.00226	0.01868	0.09983	526.767434	526.767434	526.767434	0	0	1	-360	360;
	3874	4226	0.000401	0.00239	0.00835	354.091426	354.091426	354.091426	0	0	1	-360	360;
	3852	1749	0.001169	0.008349	0.01269	523.2751	523.2751	523.2751	0	0	1	-360	360;
	3852	5729	0.00324	0.02312	0.03518	540.661703	540.661703	540.661703	0	0	1	-360	360;
	563	3970	0.006031	0.034789	0.05445	498.025125	498.025125	498.025125	0	0	1	-360	360;
	563	733	0.0103	0.05269	0.09313	454.379904	454.379904	454.379904	0	0	1	-360	360;
	2691	6276	0.00129	0.01494	0.43238	1484.80776	1484.80776	1484.80776	0	0	1	-360	360;
	2691	6276	0.00126	0.01523	0.42338	1379.02716	1379.02716	1379.02716	0	0	1	-360	360;
	4511	4463	0.001031	0.00601	0.01921	279.115415	279.115415	279.115415	0	0	1	-360	360;
	4511	4463	0.00105	0.00588	0.01957	268.551152	268.551152	268.551152	0	0	1	-360	360;
	2394	5284	0.00197	0.01719	0.47473	1133.3874	1133.3874	1133.3874	0	0	1	-360	360;
	2394	1779	0.00114	0.01237	0.27421	1411.59716	1411.59716	1411.59716	0	0	1	-360	360;
	2394	3614	0.00314	0.02868	0.77451	1377.59167	1377.59167	1377.59167	0	0	1	-360	360;
	2394	980	0.00294	0.03326	0.73614	1239.36085	1239.36085	1239.36085	0	0	1	-360	360;
	2394	980	0.00288	0.0308	0.72013	1462.33423	1462.33423	1462.33423	0	0	1	-360	360;
	2394	732	0.00088	0.02821	0.22351	1098.76519	1098.76519	1098.76519	0	0	1	-360	360;
	3794	6039	0.00152	0.01702	0.34786	907.228433	907.228433	907.228433	0	0	1	-360	360;
	3794	6620	0.00022	0.00209	0.05191	1494.50142	1494.50142	1494.50142	0	0	1	-360	360;
	1641	2418	0.00087	0.005	0.01298	231.698931	231.698931	231.698931	0	0	1	-360	360;
	1641	7558	0.000711	0.00439	0.01507	396.13547	396.13547	396.13547	0	0	1	-360	360;
	1641	7558	0.000729	0.004289	0.01534	425.649026	425.649026	425.649026	0	0	1	-360	360;
	4237	5845	0.001031	0.00549	0.00924	275.158961	275.158961	275.158961	0	0	1	-360	360;
	4237	470	0.003219	0.01512	0.02444	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4237	539	0.007271	0.031419	0.05502	393.023806	393.023806	393.023806	0	0	1	-360	360;
	4237	2991	0.02145	0.085781	0.16269	450.429547	450.429547	450.429547	0	0	1	-360	360;
	5569	6805	0.002841	0.017771	0.05254	323.893501	323.893501	323.893501	0	0	1	-360	360;
	5569	2625	0.003331	0.017289	0.03022	415.082858	415.082858	415.082858	0	0	1	-360	360;
	4799	2129	0.01524	0.064091	0.19686	521.149597	521.149597	521.149597	0	0	1	-360	360;
	7737	4657	0.000831	0.0052	0.00706	530.941469	530.941469	530.941469	0	0	1	-360	360;
	4657	8208	0.00164	0.0103	0.01431	506.032153	506.032153	506.032153	0	0	1	-360	360;
	3609	3859	0.00139	0.00864	0.00902	396.870136	396.870136	396.870136	0	0	1	-360	360;
	3609	1284	0.00143	0.00764	0.01541	443.547381	443.547381	443.547381	0	0	1	-360	360;
	3609	5257	0.000229	0.00126	0.00264	260.694258	260.694258	260.694258	0	0	1	-360	360;
	3609	5257	0.000211	0.00143	0.00235	285.532698	285.532698	285.532698	0	0	1	-360	360;
	886	1413	0.0008	0.01072	0.30217	1553.3986	1553.3986	1553.3986	0	0	1	-360	360;
	886	4186	0.00067	0.00661	0.22487	1122.38988	1122.38988	1122.38988	0	0	1	-360	360;
	886	2536	0.0003	0.00427	0.11118	1394.01851	1394.01851	1394.01851	0	0	1	-360	360;
	886	2536	0.00068	0.00648	0.22919	1374.57457	1374.57457	1374.57457	0	0	1	-360	360;
	1458	7420	0.0002	0.001289	0.14894	271.333588	271.333588	271.333588	0	0	1	-360	360;
	1458	3653	0.00013	0.000961	0.10653	544.066396	544.066396	544.066396	0	0	1	-360	360;
	5237	2037	0.00118	0.01285	0.28304	1640.01697	1640.01697	1640.01697	0	0	1	-360	360;
	5237	4764	0.00062	0.00577	0.14292	1100.8582	1100.8582	1100.8582	0	0	1	-360	360;
	8364	1282	9.1e-05	0.000289	0.03828	334.969135	334.969135	334.969135	0	0	1	-360	360;
	8364	1282	9.1e-05	0.000271	0.03745	299.933384	299.933384	299.933384	0	0	1	-360	360;
	8364	5982	0.0003	0.002719	0.34533	487.052756	487.052756	487.052756	0	0	1	-360	360;
	8364	6202	0.00049	0.002091	0.17448	319.205809	319.205809	319.205809	0	0	1	-360	360;
	8364	12	0.000159	0.001461	0.13821	579.635618	579.635618	579.635618	0	0	1	-360	360;
	386	5564	0.00301	0.02281	0.05603	232.182104	232.182104	232.182104	0	0	1	-360	360;
	6040	5564	0.0027	0.019979	0.05034	250.625362	250.625362	250.625362	0	0	1	-360	360;
	1310	587	0.00019	0.000539	0.055	229.073488	229.073488	229.073488	0	0	1	-360	360;
	587	6202	0.000169	0.00061	0.049	517.135604	517.135604	517.135604	0	0	1	-360	360;
	7629	4930	0.00163	0.00939	0.01781	552.361118	552.361118	552.361118	0	0	1	-360	360;
	7629	4930	0.001159	0.009031	0.02067	528.054244	528.054244	528.054244	0	0	1	-360	360;
	7629	3697	9.9e-05	0.00037	0.00051	467.917509	467.917509	467.917509	0	0	1	-360	360;
	5095	8223	0.003789	0.01805	0.05793	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5095	1338	0.00069	0.00389	0.01067	225.31518	225.31518	225.31518	0	0	1	-360	360;
	7608	7192	0.00201	0.01845	0.41995	1284.09335	1284.09335	1284.09335	0	0	1	-360	360;
	7989	7192	0.00121	0.00902	0.25097	1294.08517	1294.08517	1294.08517	0	0	1	-360	360;
	4132	1779	0.00155	0.01573	0.37795	1303.37866	1303.37866	1303.37866	0	0	1	-360	360;
	4132	7557	0.00031	0.00274	0.0767	1345.9859	1345.9859	1345.9859	0	0	1	-360	360;
	4132	3614	0.00047	0.00563	0.1564	1214.88711	1214.88711	1214.88711	0	0	1	-360	360;
	4132	3223	0.00066	0.00721	0.15865	1161.86484	1161.86484	1161.86484	0	0	1	-360	360;
	6811	4372	0.00326	0.016711	0.04153	332.537647	332.537647	332.537647	0	0	1	-360	360;
	6811	2636	0.002479	0.01262	0.1157	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6811	2636	0.00243	0.01286	0.11329	628.504669	628.504669	628.504669	0	0	1	-360	360;
	6811	3950	0.004521	0.02306	0.04906	476.590615	476.590615	476.590615	0	0	1	-360	360;
	6811	1167	0.00211	0.00899	0.02802	280.898735	280.898735	280.898735	0	0	1	-360	360;
	6811	7069	0.001289	0.005651	0.01662	560.047682	560.047682	560.047682	0	0	1	-360	360;
	6811	3073	0.002479	0.009021	0.03307	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6811	5858	0.00151	0.010521	0.01515	425.594536	425.594536	425.594536	0	0	1	-360	360;
	4912	7570	0.00101	0.004151	0.01376	453.739738	453.739738	453.739738	0	0	1	-360	360;
	4912	722	0.015599	0.079	0.13566	388.324301	388.324301	388.324301	0	0	1	-360	360;
	4912	7438	0.013961	0.067591	0.12195	464.507482	464.507482	464.507482	0	0	1	-360	360;
	4912	1127	0.007099	0.043169	0.06179	522.093842	522.093842	522.093842	0	0	1	-360	360;
	2192	4585	0.00317	0.02987	0.7558	993.356461	993.356461	993.356461	0	0	1	-360	360;
	2192	2505	0.00081	0.0088	0.23188	1133.5322	1133.5322	1133.5322	0	0	1	-360	360;
	2192	2505	0.00098	0.00961	0.23348	994.489845	994.489845	994.489845	0	0	1	-360	360;
	2192	159	0.00194	0.02225	0.55539	927.67214	927.67214	927.67214	0	0	1	-360	360;
	8173	4957	0.001031	0.006349	0.01018	360.673323	360.673323	360.673323	0	0	1	-360	360;
	8173	401	0.00668	0.0297	0.09857	569.022962	569.022962	569.022962	0	0	1	-360	360;
	4375	2297	0.00439	0.02236	0.03336	545.187829	545.187829	545.187829	0	0	1	-360	360;
	2297	5804	0.00249	0.01349	0.01885	592.266321	592.266321	592.266321	0	0	1	-360	360;
	1899	4145	0.001969	0.01205	0.03173	499.681555	499.681555	499.681555	0	0	1	-360	360;
	1899	4178	0.000461	0.003031	0.01045	575.218854	575.218854	575.218854	0	0	1	-360	360;
	3267	242	0.00444	0.023479	0.03367	454.188617	454.188617	454.188617	0	0	1	-360	360;
	3267	4209	0.00175	0.01288	0.03245	235.205364	235.205364	235.205364	0	0	1	-360	360;
	3267	4209	0.001711	0.01312	0.03177	291.322771	291.322771	291.322771	0	0	1	-360	360;
	1413	524	0.0008	0.00784	0.18031	1496.47465	1496.47465	1496.47465	0	0	1	-360	360;
	1413	8272	0.00051	0.00612	0.11615	1412.40211	1412.40211	1412.40211	0	0	1	-360	360;
	2984	6481	0.00137	0.00905	0.02867	347.408931	347.408931	347.408931	0	0	1	-360	360;
	2984	4418	0.002771	0.021409	0.05787	383.130193	383.130193	383.130193	0	0	1	-360	360;
	2984	4418	0.00314	0.020979	0.06559	328.130028	328.130028	328.130028	0	0	1	-360	360;
	2984	3728	0.002	0.01118	0.04113	345.462903	345.462903	345.462903	0	0	1	-360	360;
	7511	634	0.00119	0.00875	0.02203	374.989797	374.989797	374.989797	0	0	1	-360	360;
	7511	6244	0.00031	0.00236	0.00669	291.562071	291.562071	291.562071	0	0	1	-360	360;
	7511	7092	0.000831	0.00539	0.01782	224.679206	224.679206	224.679206	0	0	1	-360	360;
	242	6664	0.004349	0.02394	0.03297	425.183763	425.183763	425.183763	0	0	1	-360	360;
	1092	7920	0.0007	0.00782	0.23545	1701.691	1701.691	1701.691	0	0	1	-360	360;
	1092	1566	0.00046	0.00432	0.15278	1870.55678	1870.55678	1870.55678	0	0	1	-360	360;
	1092	2789	0.00049	0.00617	0.16468	1697.4701	1697.4701	1697.4701	0	0	1	-360	360;
	1092	980	0.00127	0.01514	0.42707	1789.23214	1789.23214	1789.23214	0	0	1	-360	360;
	3228	5935	0.00232	0.01843	0.02524	295.997506	295.997506	295.997506	0	0	1	-360	360;
	3228	1369	0.00488	0.032151	0.05298	564.130264	564.130264	564.130264	0	0	1	-360	360;
	1282	3379	0.000229	0.00136	0.08924	537.758093	537.758093	537.758093	0	0	1	-360	360;
	1282	6202	0.0003	0.001531	0.16504	494.459248	494.459248	494.459248	0	0	1	-360	360;
	1282	5345	0.00026	0.00126	0.09112	410.118524	410.118524	410.118524	0	0	1	-360	360;
	1282	5395	0.00056	0.002091	0.18673	565.11452	565.11452	565.11452	0	0	1	-360	360;
	4014	2691	0.00103	0.0093	0.25567	1711.30503	1711.30503	1711.30503	0	0	1	-360	360;
	4014	2212	0.00183	0.01689	0.43755	1331.82122	1331.82122	1331.82122	0	0	1	-360	360;
	6402	4695	0.00324	0.01095	0.01848	865.03952	865.03952	865.03952	0	0	1	-360	360;
	214	7465	0.001419	0.008539	0.01475	280.336684	280.336684	280.336684	0	0	1	-360	360;
	214	3113	0.00068	0.00438	0.00621	501.543752	501.543752	501.543752	0	0	1	-360	360;
	2645	5338	0.00178	0.01651	0.53234	1310.37906	1310.37906	1310.37906	0	0	1	-360	360;
	2645	3928	0.00151	0.01827	0.50815	1261.86663	1261.86663	1261.86663	0	0	1	-360	360;
	2645	3928	0.00148	0.01862	0.49734	1320.93757	1320.93757	1320.93757	0	0	1	-360	360;
	2691	2212	0.00201	0.02546	0.67956	1598.2239	1598.2239	1598.2239	0	0	1	-360	360;
	2691	2212	0.00197	0.02357	0.66478	1550.67045	1550.67045	1550.67045	0	0	1	-360	360;
	3411	3113	0.001091	0.00775	0.02176	270.851177	270.851177	270.851177	0	0	1	-360	360;
	3411	897	0.000841	0.006669	0.01656	280.543976	280.543976	280.543976	0	0	1	-360	360;
	843	3392	0.00151	0.012081	0.08385	515.463932	515.463932	515.463932	0	0	1	-360	360;
	843	6766	0.00032	0.002521	0.01816	439.371823	439.371823	439.371823	0	0	1	-360	360;
	7334	3035	0.006091	0.03019	0.08771	520.546393	520.546393	520.546393	0	0	1	-360	360;
	7334	6948	0.003651	0.016461	0.02768	554.305622	554.305622	554.305622	0	0	1	-360	360;
	7334	3028	0.00076	0.00425	0.00591	396.585491	396.585491	396.585491	0	0	1	-360	360;
	7334	3028	0.00087	0.004031	0.00652	352.690682	352.690682	352.690682	0	0	1	-360	360;
	1494	3392	0.000651	0.004729	0.03372	592.728155	592.728155	592.728155	0	0	1	-360	360;
	1494	8113	0.00045	0.00337	0.05643	524.480746	524.480746	524.480746	0	0	1	-360	360;
	2636	435	0.001539	0.00839	0.09999	454.016762	454.016762	454.016762	0	0	1	-360	360;
	2636	435	0.001919	0.010479	0.1012	522.023347	522.023347	522.023347	0	0	1	-360	360;
	435	8225	0.001169	0.007031	0.00292	445.141699	445.141699	445.141699	0	0	1	-360	360;
	1732	435	0.00739	0.034081	0.09433	320.036882	320.036882	320.036882	0	0	1	-360	360;
	4450	435	0.00164	0.00964	0.01325	433.618331	433.618331	433.618331	0	0	1	-360	360;
	5343	435	0.00206	0.012	0.02872	458.67054	458.67054	458.67054	0	0	1	-360	360;
	4896	435	0.000961	0.00531	0.00828	469.937462	469.937462	469.937462	0	0	1	-360	360;
	3223	6142	6e-05	0.00029	0.00673	1012.39298	1012.39298	1012.39298	0	0	1	-360	360;
	1224	4372	0.001919	0.013151	0.0196	383.256321	383.256321	383.256321	0	0	1	-360	360;
	785	4372	0.00019	0.00107	0.00179	416.157423	416.157423	416.157423	0	0	1	-360	360;
	785	4372	0.00018	0.001091	0.00176	442.898451	442.898451	442.898451	0	0	1	-360	360;
	5160	2229	0.00147	0.01714	0.49579	1539.39979	1539.39979	1539.39979	0	0	1	-360	360;
	2476	2229	0.00143	0.01632	0.35893	1297.65645	1297.65645	1297.65645	0	0	1	-360	360;
	2476	2229	0.00142	0.01416	0.481	1601.29562	1601.29562	1601.29562	0	0	1	-360	360;
	2476	2229	0.00145	0.01386	0.49025	1748.92919	1748.92919	1748.92919	0	0	1	-360	360;
	2229	8162	0.00073	0.00732	0.25771	1625.28692	1625.28692	1625.28692	0	0	1	-360	360;
	1079	1126	0.009039	0.05989	0.08228	479.598633	479.598633	479.598633	0	0	1	-360	360;
	1079	3658	0.007919	0.04975	0.07193	461.392389	461.392389	461.392389	0	0	1	-360	360;
	2418	891	0.000521	0.005841	0.00779	460.199318	460.199318	460.199318	0	0	1	-360	360;
	2418	5755	0.000521	0.0028	0.00351	329.938878	329.938878	329.938878	0	0	1	-360	360;
	5952	980	0.0007	0.00744	0.16018	900.669676	900.669676	900.669676	0	0	1	-360	360;
	5952	3668	0.00159	0.0148	0.40712	938.482736	938.482736	938.482736	0	0	1	-360	360;
	3948	3870	0.000659	0.005151	0.01925	564.699936	564.699936	564.699936	0	0	1	-360	360;
	6525	3870	0.000729	0.004901	0.02146	433.452574	433.452574	433.452574	0	0	1	-360	360;
	4186	581	0.00064	0.00772	0.21455	1139.20834	1139.20834	1139.20834	0	0	1	-360	360;
	6870	5612	0.00018	0.00093	0.14315	609.151842	609.151842	609.151842	0	0	1	-360	360;
	6870	7420	0.00019	0.00118	0.14931	509.798844	509.798844	509.798844	0	0	1	-360	360;
	6719	6870	0.000409	0.002341	0.00783	529.028211	529.028211	529.028211	0	0	1	-360	360;
	7908	932	0.00087	0.006591	0.01868	384.244387	384.244387	384.244387	0	0	1	-360	360;
	7908	2786	0.00162	0.0098	0.03017	263.343707	263.343707	263.343707	0	0	1	-360	360;
	6809	6397	0.00096	0.00849	0.24256	1564.65149	1564.65149	1564.65149	0	0	1	-360	360;
	5609	6287	0.000341	0.001081	0.1269	499.523037	499.523037	499.523037	0	0	1	-360	360;
	5609	6287	0.0003	0.001219	0.11305	527.175159	527.175159	527.175159	0	0	1	-360	360;
	5609	7319	9.9e-05	0.00076	0.1008	536.780316	536.780316	536.780316	0	0	1	-360	360;
	5609	7319	0.00011	0.00074	0.10274	566.819343	566.819343	566.819343	0	0	1	-360	360;
	1460	4418	0.00151	0.009039	0.03189	233.43881	233.43881	233.43881	0	0	1	-360	360;
	1460	2309	0.00356	0.02331	0.06623	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7255	1310	0.000281	0.00169	0.00071	359.038613	359.038613	359.038613	0	0	1	-360	360;
	7255	5982	0.000169	0.001531	0.18247	496.990571	496.990571	496.990571	0	0	1	-360	360;
	7255	7855	0.00038	0.00326	0.01039	220.978056	220.978056	220.978056	0	0	1	-360	360;
	991	6579	0.00136	0.01738	0.33629	1481.64784	1481.64784	1481.64784	0	0	1	-360	360;
	5845	470	0.00186	0.0133	0.03487	239.331386	239.331386	239.331386	0	0	1	-360	360;
	5845	470	0.00182	0.01356	0.03416	237.353349	237.353349	237.353349	0	0	1	-360	360;
	1833	5845	0.002531	0.015919	0.02317	551.810499	551.810499	551.810499	0	0	1	-360	360;
	5845	6121	0.00289	0.015969	0.02583	432.357051	432.357051	432.357051	0	0	1	-360	360;
	1022	4501	0.001771	0.01107	0.01897	415.513065	415.513065	415.513065	0	0	1	-360	360;
	3829	4501	0.007969	0.04238	0.08525	347.066366	347.066366	347.066366	0	0	1	-360	360;
	2636	6520	0.00013	0.00062	0.06807	505.315015	505.315015	505.315015	0	0	1	-360	360;
	2636	259	0.0002	0.00201	0.19656	579.571601	579.571601	579.571601	0	0	1	-360	360;
	431	8310	0.00095	0.01024	0.2367	1351.87331	1351.87331	1351.87331	0	0	1	-360	360;
	1032	8310	0.00553	0.04909	1.38635	1177.01086	1177.01086	1177.01086	0	0	1	-360	360;
	2891	5895	0.009229	0.03693	0.07005	446.296666	446.296666	446.296666	0	0	1	-360	360;
	3452	5895	0.00451	0.028469	0.04079	446.228838	446.228838	446.228838	0	0	1	-360	360;
	5895	5480	0.003591	0.01906	0.02731	403.755731	403.755731	403.755731	0	0	1	-360	360;
	5895	6705	0.010159	0.05169	0.07711	504.306754	504.306754	504.306754	0	0	1	-360	360;
	65	1938	0.00039	0.003099	0.00829	272.166185	272.166185	272.166185	0	0	1	-360	360;
	65	6231	0.000729	0.0048	0.01515	237.396789	237.396789	237.396789	0	0	1	-360	360;
	6474	6039	0.00141	0.01296	0.32383	929.505828	929.505828	929.505828	0	0	1	-360	360;
	3340	6118	0.001969	0.019229	0.04229	232.39092	232.39092	232.39092	0	0	1	-360	360;
	3340	8128	0.001081	0.0105	0.02346	282.234319	282.234319	282.234319	0	0	1	-360	360;
	3340	8128	0.00105	0.010711	0.02297	277.90291	277.90291	277.90291	0	0	1	-360	360;
	5160	2580	0.00052	0.00449	0.17173	1431.61305	1431.61305	1431.61305	0	0	1	-360	360;
	3070	5804	0.01024	0.056349	0.07757	484.295089	484.295089	484.295089	0	0	1	-360	360;
	3070	1385	0.004599	0.023659	0.03518	436.221672	436.221672	436.221672	0	0	1	-360	360;
	3070	2939	0.010021	0.057409	0.07592	443.838123	443.838123	443.838123	0	0	1	-360	360;
	3070	2991	0.00275	0.01525	0.05088	464.671334	464.671334	464.671334	0	0	1	-360	360;
	628	7469	0.00107	0.01122	0.25633	1562.13725	1562.13725	1562.13725	0	0	1	-360	360;
	628	5338	0.00242	0.02336	0.60162	1231.35344	1231.35344	1231.35344	0	0	1	-360	360;
	628	3264	0.00165	0.01303	0.39323	844.215663	844.215663	844.215663	0	0	1	-360	360;
	628	3264	0.00155	0.01332	0.38611	861.21775	861.21775	861.21775	0	0	1	-360	360;
	6342	5417	0.00087	0.005469	0.00731	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6342	5417	0.002281	0.011909	0.01965	229.463303	229.463303	229.463303	0	0	1	-360	360;
	6342	880	0.00106	0.01076	0.14261	333.020439	333.020439	333.020439	0	0	1	-360	360;
	6342	4859	0.001409	0.01618	0.03782	391.637923	391.637923	391.637923	0	0	1	-360	360;
	6342	1746	0.00306	0.015521	0.02667	470.700326	470.700326	470.700326	0	0	1	-360	360;
	3498	7737	0.000969	0.006831	0.00834	440.21204	440.21204	440.21204	0	0	1	-360	360;
	3498	8208	0.002531	0.016961	0.02191	552.437709	552.437709	552.437709	0	0	1	-360	360;
	930	234	0.01207	0.077901	0.10557	380.141609	380.141609	380.141609	0	0	1	-360	360;
	930	6291	0.014091	0.080581	0.10654	509.480286	509.480286	509.480286	0	0	1	-360	360;
	930	2676	0.003289	0.01481	0.02492	228.918019	228.918019	228.918019	0	0	1	-360	360;
	930	152	0.003349	0.0145	0.0254	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7374	3452	0.005651	0.026479	0.04281	435.338396	435.338396	435.338396	0	0	1	-360	360;
	7374	6948	0.01193	0.064669	0.09039	462.495532	462.495532	462.495532	0	0	1	-360	360;
	6049	4649	0.004151	0.017281	0.03148	433.09248	433.09248	433.09248	0	0	1	-360	360;
	5002	6481	0.00213	0.014091	0.04461	293.088562	293.088562	293.088562	0	0	1	-360	360;
	3950	1732	0.00113	0.00651	0.01228	420.163033	420.163033	420.163033	0	0	1	-360	360;
	3950	1732	0.001159	0.00637	0.01252	433.125632	433.125632	433.125632	0	0	1	-360	360;
	3950	5858	0.00611	0.03288	0.06521	474.686503	474.686503	474.686503	0	0	1	-360	360;
	8121	768	0.00012	0.00156	0.05752	1755.04894	1755.04894	1755.04894	0	0	1	-360	360;
	8121	768	0.00013	0.00153	0.05863	1662.16273	1662.16273	1662.16273	0	0	1	-360	360;
	8121	6139	0.00285	0.03071	0.85138	1674.98933	1674.98933	1674.98933	0	0	1	-360	360;
	5417	2625	0.00676	0.03787	0.06371	505.78066	505.78066	505.78066	0	0	1	-360	360;
	5417	8218	0.004461	0.023521	0.03676	449.469679	449.469679	449.469679	0	0	1	-360	360;
	4845	560	0.00064	0.0078	0.24608	1522.33056	1522.33056	1522.33056	0	0	1	-360	360;
	4845	560	0.00065	0.00764	0.25072	1528.9117	1528.9117	1528.9117	0	0	1	-360	360;
	4845	524	0.00066	0.00602	0.1499	1495.42419	1495.42419	1495.42419	0	0	1	-360	360;
	4845	4046	0.00149	0.01224	0.42885	902.087394	902.087394	902.087394	0	0	1	-360	360;
	4845	738	0.00061	0.00582	0.13992	1025.75139	1025.75139	1025.75139	0	0	1	-360	360;
	3305	6263	0.001461	0.00744	0.02186	378.383438	378.383438	378.383438	0	0	1	-360	360;
	3305	7092	0.00386	0.02406	0.07282	232.859994	232.859994	232.859994	0	0	1	-360	360;
	3305	6926	0.00306	0.01555	0.04616	275.626891	275.626891	275.626891	0	0	1	-360	360;
	3305	6926	0.003	0.01586	0.04522	296.858301	296.858301	296.858301	0	0	1	-360	360;
	3301	3742	0.003271	0.020651	0.06106	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3742	6924	0.000849	0.00657	0.01589	376.103228	376.103228	376.103228	0	0	1	-360	360;
	1499	2788	0.000211	0.001169	0.00392	486.514712	486.514712	486.514712	0	0	1	-360	360;
	1499	7234	0.000469	0.00345	0.00849	510.64554	510.64554	510.64554	0	0	1	-360	360;
	4499	6809	0.00126	0.0145	0.28823	1398.23349	1398.23349	1398.23349	0	0	1	-360	360;
	1788	5381	0.008219	0.04925	0.0708	376.625649	376.625649	376.625649	0	0	1	-360	360;
	1788	1618	0.006771	0.04132	0.07479	455.994418	455.994418	455.994418	0	0	1	-360	360;
	6723	6896	0.000841	0.004539	0.02502	591.415053	591.415053	591.415053	0	0	1	-360	360;
	6723	3452	0.00701	0.03161	0.05314	493.349246	493.349246	493.349246	0	0	1	-360	360;
	6723	6241	0.000781	0.004271	0.00588	581.766075	581.766075	581.766075	0	0	1	-360	360;
	6723	873	0.00049	0.0032	0.00908	551.849366	551.849366	551.849366	0	0	1	-360	360;
	3170	3470	0.000331	0.00199	0.13889	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1227	3470	0.00032	0.002031	0.13594	628.950118	628.950118	628.950118	0	0	1	-360	360;
	3470	4991	9.1e-05	0.00057	0.08911	485.353268	485.353268	485.353268	0	0	1	-360	360;
	3692	785	0.001651	0.006789	0.02229	540.013154	540.013154	540.013154	0	0	1	-360	360;
	3692	3073	0.001461	0.007229	0.01944	315.337378	315.337378	315.337378	0	0	1	-360	360;
	7570	8218	0.00437	0.023979	0.03915	456.340794	456.340794	456.340794	0	0	1	-360	360;
	1710	427	0.00781	0.04595	0.05746	432.948443	432.948443	432.948443	0	0	1	-360	360;
	1710	5665	0.0015	0.006531	0.01533	471.637331	471.637331	471.637331	0	0	1	-360	360;
	5233	8225	0.001031	0.006469	0.00259	365.281375	365.281375	365.281375	0	0	1	-360	360;
	2536	2784	0.00183	0.01686	0.51302	1150.60405	1150.60405	1150.60405	0	0	1	-360	360;
	2536	560	0.0002	0.00294	0.08665	1520.32641	1520.32641	1520.32641	0	0	1	-360	360;
	2536	560	0.00021	0.00288	0.08835	1339.102	1339.102	1339.102	0	0	1	-360	360;
	2536	4585	0.00171	0.01889	0.40723	1286.7004	1286.7004	1286.7004	0	0	1	-360	360;
	8051	2505	0.00069	0.00565	0.18019	1213.32262	1213.32262	1213.32262	0	0	1	-360	360;
	8051	2505	0.0007	0.00553	0.18352	1405.19637	1405.19637	1405.19637	0	0	1	-360	360;
	6436	3907	0.00011	0.00115	0.03329	1146.78859	1146.78859	1146.78859	0	0	1	-360	360;
	6436	6058	0.00012	0.00126	0.0418	1330.07244	1330.07244	1330.07244	0	0	1	-360	360;
	6436	8148	0.00326	0.02919	0.8563	1413.55656	1413.55656	1413.55656	0	0	1	-360	360;
	6436	4603	0.00092	0.00876	0.22199	827.435381	827.435381	827.435381	0	0	1	-360	360;
	5246	6287	0.00087	0.005409	0.00757	458.188511	458.188511	458.188511	0	0	1	-360	360;
	5246	6231	0.00131	0.00701	0.01131	450.39106	450.39106	450.39106	0	0	1	-360	360;
	3908	6036	0.002531	0.018781	0.04731	278.941275	278.941275	278.941275	0	0	1	-360	360;
	3908	7918	0.00894	0.056909	0.1695	228.905444	228.905444	228.905444	0	0	1	-360	360;
	5003	4957	0.00187	0.010659	0.01539	556.269179	556.269179	556.269179	0	0	1	-360	360;
	6971	4957	0.00286	0.01538	0.03108	622.33164	622.33164	622.33164	0	0	1	-360	360;
	4957	4519	0.0008	0.00444	0.00987	426.256041	426.256041	426.256041	0	0	1	-360	360;
	922	7146	0.00545	0.03936	0.10673	272.499604	272.499604	272.499604	0	0	1	-360	360;
	1590	7146	0.00119	0.009031	0.02348	221.808366	221.808366	221.808366	0	0	1	-360	360;
	43	5146	0.00295	0.0162	0.02797	435.386027	435.386027	435.386027	0	0	1	-360	360;
	43	5602	0.001581	0.0087	0.01527	515.364478	515.364478	515.364478	0	0	1	-360	360;
	43	2279	0.007729	0.025401	0.06733	598.188238	598.188238	598.188238	0	0	1	-360	360;
	7078	3112	0.00013	0.001461	0.18234	469.188315	469.188315	469.188315	0	0	1	-360	360;
	7078	7852	0.000341	0.02219	0.15864	314.313493	314.313493	314.313493	0	0	1	-360	360;
	7078	3653	0.00018	0.001211	0.14627	603.616312	603.616312	603.616312	0	0	1	-360	360;
	7078	8199	0.003099	0.0147	0.06035	595.279293	595.279293	595.279293	0	0	1	-360	360;
	7078	8199	0.003159	0.014401	0.06154	538.539629	538.539629	538.539629	0	0	1	-360	360;
	6287	3892	0.00281	0.01305	0.02407	377.471964	377.471964	377.471964	0	0	1	-360	360;
	6287	8052	0.002281	0.01075	0.02042	513.443599	513.443599	513.443599	0	0	1	-360	360;
	6287	8052	0.00232	0.010521	0.0208	474.467779	474.467779	474.467779	0	0	1	-360	360;
	6287	3127	0.000151	0.000841	0.13408	308.465501	308.465501	308.465501	0	0	1	-360	360;
	6304	2410	0.002771	0.012091	0.02504	577.782566	577.782566	577.782566	0	0	1	-360	360;
	6304	5078	0.00613	0.03005	0.05518	471.803088	471.803088	471.803088	0	0	1	-360	360;
	3035	6948	0.003219	0.02486	0.0602	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5612	7319	0.000211	0.001331	0.16723	586.41528	586.41528	586.41528	0	0	1	-360	360;
	7709	1736	0.00185	0.01917	0.4907	1613.10731	1613.10731	1613.10731	0	0	1	-360	360;
	1167	785	0.00082	0.00381	0.01084	580.991398	580.991398	580.991398	0	0	1	-360	360;
	8212	7091	0.00075	0.004531	0.00685	387.170479	387.170479	387.170479	0	0	1	-360	360;
	8212	7091	0.000729	0.00419	0.00671	422.950422	422.950422	422.950422	0	0	1	-360	360;
	8212	27	0.02626	0.15576	0.22783	447.63949	447.63949	447.63949	0	0	1	-360	360;
	8212	6729	0.011469	0.03937	0.06728	305.627432	305.627432	305.627432	0	0	1	-360	360;
	6476	7420	0.00011	0.00061	0.00193	514.649245	514.649245	514.649245	0	0	1	-360	360;
	6476	971	0.000531	0.00313	0.00962	393.582427	393.582427	393.582427	0	0	1	-360	360;
	6476	5566	0.001211	0.00864	0.02267	328.911564	328.911564	328.911564	0	0	1	-360	360;
	3948	4226	0.00094	0.006289	0.08077	267.926228	267.926228	267.926228	0	0	1	-360	360;
	662	3662	0.001	0.00886	0.2856	845.941409	845.941409	845.941409	0	0	1	-360	360;
	6140	2891	0.001979	0.010469	0.01725	516.026364	516.026364	516.026364	0	0	1	-360	360;
	6140	2891	0.0017	0.00855	0.01466	562.010096	562.010096	562.010096	0	0	1	-360	360;
	2891	1930	0.00757	0.04624	0.06619	422.939753	422.939753	422.939753	0	0	1	-360	360;
	2891	6602	0.001719	0.00794	0.01285	509.458566	509.458566	509.458566	0	0	1	-360	360;
	2891	6602	0.00176	0.007781	0.0131	259.530909	259.530909	259.530909	0	0	1	-360	360;
	2872	828	0.00126	0.0053	0.01065	403.877286	403.877286	403.877286	0	0	1	-360	360;
	2872	1127	0.00444	0.022789	0.03743	434.649836	434.649836	434.649836	0	0	1	-360	360;
	431	2212	0.00163	0.0145	0.41102	1097.38302	1097.38302	1097.38302	0	0	1	-360	360;
	6140	234	0.003281	0.02124	0.0288	511.601597	511.601597	511.601597	0	0	1	-360	360;
	6140	6602	0.004979	0.025271	0.03769	486.379058	486.379058	486.379058	0	0	1	-360	360;
	8039	2182	0.0072	0.055289	0.1441	509.408267	509.408267	509.408267	0	0	1	-360	360;
	8039	4339	0.00787	0.04582	0.16058	284.109091	284.109091	284.109091	0	0	1	-360	360;
	8039	3658	0.003961	0.017711	0.05328	276.466728	276.466728	276.466728	0	0	1	-360	360;
	2182	6165	0.00756	0.04826	0.15137	463.376522	463.376522	463.376522	0	0	1	-360	360;
	5618	7438	0.003401	0.02086	0.1519	459.742437	459.742437	459.742437	0	0	1	-360	360;
	5618	3588	0.001341	0.00669	0.05955	428.780886	428.780886	428.780886	0	0	1	-360	360;
	6957	8048	6e-05	0.000581	0.06778	445.800918	445.800918	445.800918	0	0	1	-360	360;
	6957	5395	9.1e-05	0.00074	0.09438	530.204897	530.204897	530.204897	0	0	1	-360	360;
	6699	8115	0.000531	0.00175	0.17578	269.327353	269.327353	269.327353	0	0	1	-360	360;
	1016	4636	0.00116	0.01159	0.39356	1390.80857	1390.80857	1390.80857	0	0	1	-360	360;
	1016	3928	0.00038	0.00457	0.12703	1624.93216	1624.93216	1624.93216	0	0	1	-360	360;
	2788	7234	0.000729	0.00405	0.01351	555.896892	555.896892	555.896892	0	0	1	-360	360;
	239	6407	0.000331	0.001919	0.00615	322.723293	322.723293	322.723293	0	0	1	-360	360;
	1991	4278	0.00086	0.01029	0.28608	1923.80743	1923.80743	1923.80743	0	0	1	-360	360;
	6734	1991	0.00027	0.00332	0.08843	1718.47524	1718.47524	1718.47524	0	0	1	-360	360;
	4702	5157	0.00012	0.00156	0.19385	264.896109	264.896109	264.896109	0	0	1	-360	360;
	4702	5157	0.00012	0.001531	0.19765	356.637229	356.637229	356.637229	0	0	1	-360	360;
	7944	1882	0.00256	0.014211	0.05547	322.115516	322.115516	322.115516	0	0	1	-360	360;
	7944	7631	0.00049	0.00338	0.00537	352.431567	352.431567	352.431567	0	0	1	-360	360;
	5453	1302	0.007531	0.04444	0.06788	415.676155	415.676155	415.676155	0	0	1	-360	360;
	5453	5078	0.007409	0.03226	0.06686	295.330667	295.330667	295.330667	0	0	1	-360	360;
	2866	7383	0.001789	0.012841	0.03964	282.70225	282.70225	282.70225	0	0	1	-360	360;
	2866	3729	0.001831	0.01257	0.04039	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1487	3295	0.000169	0.00143	0.00362	371.853364	371.853364	371.853364	0	0	1	-360	360;
	1487	6231	0.00088	0.005099	0.01786	615.819095	615.819095	615.819095	0	0	1	-360	360;
	2784	6763	0.0004	0.00479	0.10823	1678.98974	1678.98974	1678.98974	0	0	1	-360	360;
	2784	6763	0.00052	0.00481	0.11479	1754.62639	1754.62639	1754.62639	0	0	1	-360	360;
	2784	7194	0.00135	0.01431	0.30856	1751.18082	1751.18082	1751.18082	0	0	1	-360	360;
	2784	738	0.00051	0.00549	0.11722	1510.16807	1510.16807	1510.16807	0	0	1	-360	360;
	634	6244	0.00105	0.007969	0.02253	492.134836	492.134836	492.134836	0	0	1	-360	360;
	634	858	0.00339	0.020531	0.06318	335.71066	335.71066	335.71066	0	0	1	-360	360;
	634	2681	0.00169	0.009331	0.02372	237.536254	237.536254	237.536254	0	0	1	-360	360;
	3261	3805	0.00374	0.01905	0.0327	455.363778	455.363778	455.363778	0	0	1	-360	360;
	3261	4796	0.00487	0.026341	0.04189	535.303743	535.303743	535.303743	0	0	1	-360	360;
	4578	7526	0.00057	0.004581	0.01064	278.960708	278.960708	278.960708	0	0	1	-360	360;
	4578	7782	0.000409	0.00301	0.00633	225.314037	225.314037	225.314037	0	0	1	-360	360;
	1328	3418	9.1e-05	0.00076	0.0009	437.858287	437.858287	437.858287	0	0	1	-360	360;
	1328	5606	0.003919	0.01682	0.02779	482.365827	482.365827	482.365827	0	0	1	-360	360;
	5284	966	0.00117	0.0131	0.27951	1535.62118	1535.62118	1535.62118	0	0	1	-360	360;
	4799	4508	0.008271	0.02787	0.07459	513.344144	513.344144	513.344144	0	0	1	-360	360;
	4799	1022	0.016599	0.0572	0.08632	433.58556	433.58556	433.58556	0	0	1	-360	360;
	4799	7132	0.00999	0.057	0.09047	378.704283	378.704283	378.704283	0	0	1	-360	360;
	4799	7132	0.01132	0.05586	0.10253	377.712026	377.712026	377.712026	0	0	1	-360	360;
	6620	4262	4e-05	0.00045	0.01152	950.048921	950.048921	950.048921	0	0	1	-360	360;
	4508	7737	0.003281	0.01038	0.0295	496.768799	496.768799	496.768799	0	0	1	-360	360;
	3930	62	0.003711	0.01725	0.06426	459.588493	459.588493	459.588493	0	0	1	-360	360;
	3930	5744	0.00069	0.00232	0.01192	463.189045	463.189045	463.189045	0	0	1	-360	360;
	7892	4502	0.004151	0.02436	0.03919	521.864068	521.864068	521.864068	0	0	1	-360	360;
	4695	4502	0.006781	0.028729	0.06373	488.481698	488.481698	488.481698	0	0	1	-360	360;
	3170	6231	0.000331	0.00151	0.106	452.161805	452.161805	452.161805	0	0	1	-360	360;
	1227	6231	0.000331	0.001479	0.108	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3354	5331	0.0001	0.00144	1.34286	1519.21079	1519.21079	1519.21079	0	0	1	-360	360;
	3354	5370	0.00027	0.00221	0.07134	1079.01652	1079.01652	1079.01652	0	0	1	-360	360;
	1119	4174	0.0001	0.00141	1.49259	1734.15043	1734.15043	1734.15043	0	0	1	-360	360;
	1119	2505	0.00054	0.00546	0.13771	1466.75588	1466.75588	1466.75588	0	0	1	-360	360;
	6734	5331	7e-05	0.00053	0.01684	1392.9075	1392.9075	1392.9075	0	0	1	-360	360;
	4174	6734	6e-05	0.0006	0.015	1494.71204	1494.71204	1494.71204	0	0	1	-360	360;
	8115	3980	0.000341	0.00245	0.00745	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8115	7092	0.000841	0.006719	0.01563	315.042063	315.042063	315.042063	0	0	1	-360	360;
	5634	6520	0.000401	0.00164	0.20126	481.461592	481.461592	481.461592	0	0	1	-360	360;
	470	6664	0.004409	0.032531	0.08197	215.929128	215.929128	215.929128	0	0	1	-360	360;
	386	7148	0.001961	0.013581	0.04292	383.536012	383.536012	383.536012	0	0	1	-360	360;
	1732	4450	0.006659	0.03437	0.08634	376.243074	376.243074	376.243074	0	0	1	-360	360;
	3700	1457	9.1e-05	0.0007	0.00099	254.127603	254.127603	254.127603	0	0	1	-360	360;
	3700	7328	0.00051	0.004211	0.00734	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6206	8113	0.00043	0.0027	0.00504	367.836323	367.836323	367.836323	0	0	1	-360	360;
	6206	2322	0.00038	0.00306	0.00449	365.429985	365.429985	365.429985	0	0	1	-360	360;
	7900	1357	0.001461	0.009479	0.01684	225.682514	225.682514	225.682514	0	0	1	-360	360;
	7900	6713	0.001841	0.011081	0.02137	520.764735	520.764735	520.764735	0	0	1	-360	360;
	880	6805	0.00094	0.007169	0.01761	567.745678	567.745678	567.745678	0	0	1	-360	360;
	5481	7289	0.00082	0.00846	0.22494	1416.7869	1416.7869	1416.7869	0	0	1	-360	360;
	5481	6734	0.00029	0.00325	0.07845	1717.59525	1717.59525	1717.59525	0	0	1	-360	360;
	7234	6778	0.000781	0.0062	0.01441	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7234	2485	8.1e-05	0.00187	0.15378	508.603487	508.603487	508.603487	0	0	1	-360	360;
	7234	2088	8.1e-05	0.001909	0.15058	435.898922	435.898922	435.898922	0	0	1	-360	360;
	6239	966	0.00113	0.00956	0.29493	1201.01664	1201.01664	1201.01664	0	0	1	-360	360;
	6239	966	0.00115	0.00935	0.30039	1220.18019	1220.18019	1220.18019	0	0	1	-360	360;
	8156	4033	0.00161	0.02036	0.02963	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8156	4033	0.00157	0.02075	0.029	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1684	3392	0.00101	0.007401	0.01325	541.916886	541.916886	541.916886	0	0	1	-360	360;
	1684	6766	0.00087	0.00711	0.09328	495.281557	495.281557	495.281557	0	0	1	-360	360;
	3259	5370	0.00135	0.01442	0.32839	1375.05044	1375.05044	1375.05044	0	0	1	-360	360;
	3259	1911	0.00176	0.01441	0.42876	1782.16988	1782.16988	1782.16988	0	0	1	-360	360;
	1833	6664	0.00682	0.03187	0.06032	428.810227	428.810227	428.810227	0	0	1	-360	360;
	7289	1273	0.00109	0.00987	0.29673	1207.63068	1207.63068	1207.63068	0	0	1	-360	360;
	4227	7289	0.00273	0.02829	0.78052	1308.32422	1308.32422	1308.32422	0	0	1	-360	360;
	1601	7117	0.001581	0.011289	0.02964	416.355569	416.355569	416.355569	0	0	1	-360	360;
	1601	3113	0.001669	0.013099	0.0349	259.23064	259.23064	259.23064	0	0	1	-360	360;
	6662	1457	9.9e-05	0.00069	0.00124	506.385388	506.385388	506.385388	0	0	1	-360	360;
	6662	7328	9.9e-05	0.000669	0.00127	517.884369	517.884369	517.884369	0	0	1	-360	360;
	4455	3400	0.000591	0.00382	0.01088	283.149985	283.149985	283.149985	0	0	1	-360	360;
	4455	6926	0.00524	0.02276	0.07926	338.924446	338.924446	338.924446	0	0	1	-360	360;
	3400	6007	0.002719	0.01143	0.04139	311.41598	311.41598	311.41598	0	0	1	-360	360;
	669	7763	0.00294	0.02744	0.82496	954.65091	954.65091	954.65091	0	0	1	-360	360;
	669	6702	0.0011	0.01339	0.26522	1023.17593	1023.17593	1023.17593	0	0	1	-360	360;
	669	5327	0.00193	0.01695	0.52495	1367.4669	1367.4669	1367.4669	0	0	1	-360	360;
	669	4406	0.00134	0.01054	0.37476	1043.95663	1043.95663	1043.95663	0	0	1	-360	360;
	5528	669	0.00245	0.02064	0.57895	1398.9792	1398.9792	1398.9792	0	0	1	-360	360;
	1267	6093	0.001841	0.00901	0.02549	470.56772	470.56772	470.56772	0	0	1	-360	360;
	1267	6093	0.00187	0.00882	0.02599	504.48318	504.48318	504.48318	0	0	1	-360	360;
	7557	732	0.00197	0.02312	0.49089	1303.5748	1303.5748	1303.5748	0	0	1	-360	360;
	2037	6702	0.00078	0.00887	0.197	1406.96884	1406.96884	1406.96884	0	0	1	-360	360;
	7717	4930	0.003771	0.0288	0.0405	456.402905	456.402905	456.402905	0	0	1	-360	360;
	7717	4232	0.000909	0.00737	0.0098	466.884861	466.884861	466.884861	0	0	1	-360	360;
	7194	6763	0.00086	0.00783	0.23042	1750.79381	1750.79381	1750.79381	0	0	1	-360	360;
	7194	6763	0.00095	0.00855	0.21268	1564.8134	1564.8134	1564.8134	0	0	1	-360	360;
	272	6763	0.00012	0.00157	0.04397	1866.86439	1866.86439	1866.86439	0	0	1	-360	360;
	272	6763	0.00013	0.00154	0.04984	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4654	5858	0.00089	0.004771	0.0079	535.698893	535.698893	535.698893	0	0	1	-360	360;
	4654	5801	0.002	0.00881	0.01778	558.068502	558.068502	558.068502	0	0	1	-360	360;
	5804	6418	0.007711	0.035331	0.23617	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6418	6291	0.005919	0.030419	0.05427	504.884808	504.884808	504.884808	0	0	1	-360	360;
	6418	2676	0.0053	0.031719	0.0613	602.012086	602.012086	602.012086	0	0	1	-360	360;
	3892	1799	0.00025	0.001289	0.00214	334.557599	334.557599	334.557599	0	0	1	-360	360;
	2676	1930	0.0062	0.040039	0.05425	370.312393	370.312393	370.312393	0	0	1	-360	360;
	4636	4958	0.0008	0.00707	0.27089	1538.72054	1538.72054	1538.72054	0	0	1	-360	360;
	4636	4510	0.0002	0.00207	0.0674	975.879826	975.879826	975.879826	0	0	1	-360	360;
	4636	4050	0.00036	0.00322	0.12332	1074.2105	1074.2105	1074.2105	0	0	1	-360	360;
	1017	5526	0.0023	0.0157	0.02417	394.094559	394.094559	394.094559	0	0	1	-360	360;
	1017	3829	0.003479	0.021159	0.0367	421.609122	421.609122	421.609122	0	0	1	-360	360;
	3614	3114	0.00107	0.01013	0.35846	1396.14706	1396.14706	1396.14706	0	0	1	-360	360;
	1310	6202	0.00025	0.000909	0.08572	526.054869	526.054869	526.054869	0	0	1	-360	360;
	1310	5345	0.00037	0.00188	0.13545	449.748227	449.748227	449.748227	0	0	1	-360	360;
	8350	3588	0.0003	0.00145	0.13724	421.279513	421.279513	421.279513	0	0	1	-360	360;
	8350	3359	0.001419	0.007539	0.02769	487.864776	487.864776	487.864776	0	0	1	-360	360;
	7763	241	0.0012	0.0122	0.35023	1279.27877	1279.27877	1279.27877	0	0	1	-360	360;
	7763	2324	0.00397	0.02936	1.15862	1256.17996	1256.17996	1256.17996	0	0	1	-360	360;
	7763	7989	0.00092	0.01112	0.21917	1190.69178	1190.69178	1190.69178	0	0	1	-360	360;
	7763	6734	0.00387	0.0452	0.93392	1467.51739	1467.51739	1467.51739	0	0	1	-360	360;
	7763	1165	0.00432	0.03156	1.26373	1100.75224	1100.75224	1100.75224	0	0	1	-360	360;
	2476	1032	0.00407	0.03937	1.01194	1252.18087	1252.18087	1252.18087	0	0	1	-360	360;
	2476	1032	0.00415	0.03856	1.03178	1168.85734	1168.85734	1168.85734	0	0	1	-360	360;
	5559	4649	0.00869	0.05343	0.0754	584.023803	584.023803	584.023803	0	0	1	-360	360;
	5559	1978	0.0068	0.036961	0.05163	462.41132	462.41132	462.41132	0	0	1	-360	360;
	5559	2410	0.01805	0.095469	0.1368	404.190891	404.190891	404.190891	0	0	1	-360	360;
	6896	6602	0.00045	0.00206	0.01384	534.902496	534.902496	534.902496	0	0	1	-360	360;
	427	6522	0.0095	0.037159	0.0715	436.258634	436.258634	436.258634	0	0	1	-360	360;
	427	4571	0.02106	0.0885	0.158	461.600824	461.600824	461.600824	0	0	1	-360	360;
	6702	5327	0.00157	0.02318	0.69807	1823.63649	1823.63649	1823.63649	0	0	1	-360	360;
	6702	3330	0.00032	0.00282	0.07561	1731.73887	1731.73887	1731.73887	0	0	1	-360	360;
	6702	4485	0.00165	0.01605	0.39059	920.139936	920.139936	920.139936	0	0	1	-360	360;
	4930	442	0.00107	0.00714	0.01154	413.129971	413.129971	413.129971	0	0	1	-360	360;
	4930	442	0.000961	0.00555	0.0143	400.434489	400.434489	400.434489	0	0	1	-360	360;
	4930	7732	0.00074	0.00461	0.01388	278.756846	278.756846	278.756846	0	0	1	-360	360;
	4930	6407	0.00076	0.00451	0.01414	275.574687	275.574687	275.574687	0	0	1	-360	360;
	4930	3697	0.001419	0.01093	0.0155	548.127639	548.127639	548.127639	0	0	1	-360	360;
	4930	4232	0.003729	0.024599	0.04002	485.837203	485.837203	485.837203	0	0	1	-360	360;
	2212	6374	0.00233	0.01992	0.57451	1701.78578	1701.78578	1701.78578	0	0	1	-360	360;
	2212	6374	0.00238	0.01949	0.58514	1655.2479	1655.2479	1655.2479	0	0	1	-360	360;
	2212	4361	0.00217	0.0253	0.73216	1775.04509	1775.04509	1775.04509	0	0	1	-360	360;
	2212	4361	0.00212	0.02579	0.71691	1873.0809	1873.0809	1873.0809	0	0	1	-360	360;
	2212	1032	0.00458	0.04138	1.1569	1106.3257	1106.3257	1106.3257	0	0	1	-360	360;
	3452	496	0.01555	0.089841	0.13668	396.967685	396.967685	396.967685	0	0	1	-360	360;
	5714	3452	0.00274	0.0168	0.02382	476.83868	476.83868	476.83868	0	0	1	-360	360;
	3452	6241	0.005831	0.032211	0.04427	442.122631	442.122631	442.122631	0	0	1	-360	360;
	3452	3809	0.006669	0.031341	0.058	478.727169	478.727169	478.727169	0	0	1	-360	360;
	7892	4695	0.01081	0.0538	0.10162	481.652118	481.652118	481.652118	0	0	1	-360	360;
	7892	7132	0.005581	0.02457	0.0523	510.418434	510.418434	510.418434	0	0	1	-360	360;
	7892	7132	0.00569	0.024039	0.05327	543.95208	543.95208	543.95208	0	0	1	-360	360;
	868	6750	0.002081	0.01294	0.01872	468.776399	468.776399	468.776399	0	0	1	-360	360;
	868	4456	0.001599	0.00919	0.01439	279.460266	279.460266	279.460266	0	0	1	-360	360;
	4764	5327	0.00083	0.00768	0.19377	1760.27554	1760.27554	1760.27554	0	0	1	-360	360;
	7648	7383	0.004979	0.024099	0.06968	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7648	7383	0.005081	0.023591	0.07102	477.75587	477.75587	477.75587	0	0	1	-360	360;
	3907	241	0.00412	0.03917	1.14425	1291.03517	1291.03517	1291.03517	0	0	1	-360	360;
	241	4406	0.00086	0.00798	0.24791	1120.92017	1120.92017	1120.92017	0	0	1	-360	360;
	2977	241	0.00087	0.00961	0.1919	1248.48519	1248.48519	1248.48519	0	0	1	-360	360;
	7477	241	0.00095	0.01034	0.27472	1252.38622	1252.38622	1252.38622	0	0	1	-360	360;
	7589	241	0.00211	0.0196	0.61015	1123.29356	1123.29356	1123.29356	0	0	1	-360	360;
	6522	4571	0.00963	0.05281	0.072	352.763843	352.763843	352.763843	0	0	1	-360	360;
	970	3398	0.00018	0.001349	0.00342	560.865799	560.865799	560.865799	0	0	1	-360	360;
	970	5858	0.000961	0.006021	0.01061	397.97366	397.97366	397.97366	0	0	1	-360	360;
	970	5858	0.000979	0.00589	0.01081	395.263624	395.263624	395.263624	0	0	1	-360	360;
	7148	2786	0.000461	0.002729	0.00843	622.201702	622.201702	622.201702	0	0	1	-360	360;
	895	3059	0.005039	0.01901	0.04421	399.944838	399.944838	399.944838	0	0	1	-360	360;
	895	4189	0.00812	0.028919	0.07492	429.533843	429.533843	429.533843	0	0	1	-360	360;
	6579	6320	0.00106	0.01005	0.26163	956.834749	956.834749	956.834749	0	0	1	-360	360;
	5327	6139	0.00074	0.0081	0.17958	1228.37189	1228.37189	1228.37189	0	0	1	-360	360;
	7383	5602	0.012961	0.03605	0.11124	259.560631	259.560631	259.560631	0	0	1	-360	360;
	7383	5602	0.00844	0.0378	0.1234	414.25331	414.25331	414.25331	0	0	1	-360	360;
	7383	4694	0.000591	0.004169	0.00667	557.673352	557.673352	557.673352	0	0	1	-360	360;
	6241	5480	0.003091	0.012289	0.0233	461.482317	461.482317	461.482317	0	0	1	-360	360;
	4145	4178	0.000969	0.0053	0.01476	555.910228	555.910228	555.910228	0	0	1	-360	360;
	2189	2435	0.00161	0.012919	0.033	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2189	2435	0.003219	0.01795	0.06745	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2189	3314	0.004331	0.025	0.09024	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2189	3314	0.004409	0.02445	0.09192	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4418	2309	0.00425	0.02014	0.05893	237.546542	237.546542	237.546542	0	0	1	-360	360;
	4418	6263	0.009531	0.060979	0.10839	510.70727	510.70727	510.70727	0	0	1	-360	360;
	4418	1735	0.002409	0.0135	0.045	234.363241	234.363241	234.363241	0	0	1	-360	360;
	2972	7441	0.0006	0.00662	0.1499	1693.56578	1693.56578	1693.56578	0	0	1	-360	360;
	2972	2580	0.00173	0.01701	0.43536	1704.89305	1704.89305	1704.89305	0	0	1	-360	360;
	5804	1067	0.004289	0.018531	0.03245	444.093428	444.093428	444.093428	0	0	1	-360	360;
	5804	6291	0.006919	0.039789	0.05258	423.086076	423.086076	423.086076	0	0	1	-360	360;
	5804	6291	0.007591	0.035539	0.05742	505.188506	505.188506	505.188506	0	0	1	-360	360;
	5804	4320	0.00961	0.05225	0.07296	435.949221	435.949221	435.949221	0	0	1	-360	360;
	3332	5935	0.000581	0.004599	0.00619	226.443092	226.443092	226.443092	0	0	1	-360	360;
	3332	5935	0.00057	0.00469	0.00605	266.938925	266.938925	266.938925	0	0	1	-360	360;
	3332	5935	0.00087	0.003159	0.00447	605.053637	605.053637	605.053637	0	0	1	-360	360;
	5838	4745	0.00018	0.001419	0.00423	233.727266	233.727266	233.727266	0	0	1	-360	360;
	5838	6093	0.00195	0.018841	0.04532	356.312954	356.312954	356.312954	0	0	1	-360	360;
	62	5529	0.00293	0.01293	0.05088	402.793577	402.793577	402.793577	0	0	1	-360	360;
	1385	2939	0.001969	0.010599	0.01484	453.574743	453.574743	453.574743	0	0	1	-360	360;
	6333	772	3.9e-05	0.000591	0.06673	275.229074	275.229074	275.229074	0	0	1	-360	360;
	6333	373	0.00364	0.02431	0.14332	292.454874	292.454874	292.454874	0	0	1	-360	360;
	6333	1367	0.00114	0.007211	0.09227	215.929128	215.929128	215.929128	0	0	1	-360	360;
	442	3859	0.001271	0.008659	0.01417	388.868443	388.868443	388.868443	0	0	1	-360	360;
	442	1284	0.00105	0.00668	0.01789	366.259152	366.259152	366.259152	0	0	1	-360	360;
	3907	6058	0.00011	0.00122	0.04059	1009.51081	1009.51081	1009.51081	0	0	1	-360	360;
	3907	3264	0.00117	0.0124	0.32387	1350.16139	1350.16139	1350.16139	0	0	1	-360	360;
	3907	7589	0.00168	0.01891	0.46934	1191.17028	1191.17028	1191.17028	0	0	1	-360	360;
	722	7091	0.00132	0.008281	0.01107	573.39057	573.39057	573.39057	0	0	1	-360	360;
	722	7593	0.004531	0.01924	0.06299	280.680392	280.680392	280.680392	0	0	1	-360	360;
	722	1127	0.007841	0.041909	0.11339	422.303016	422.303016	422.303016	0	0	1	-360	360;
	7194	7034	0.00113	0.01235	0.2836	1794.32448	1794.32448	1794.32448	0	0	1	-360	360;
	7194	7034	0.00128	0.0121	0.32141	1776.48387	1776.48387	1776.48387	0	0	1	-360	360;
	7194	7822	0.0015	0.01792	0.35784	1421.78314	1421.78314	1421.78314	0	0	1	-360	360;
	4178	1338	0.00111	0.005281	0.01707	282.694629	282.694629	282.694629	0	0	1	-360	360;
	4178	1338	0.00113	0.005169	0.01739	265.731755	265.731755	265.731755	0	0	1	-360	360;
	5482	2830	0.00563	0.03255	0.0444	494.024469	494.024469	494.024469	0	0	1	-360	360;
	1069	2830	0.005289	0.022539	0.04082	355.215527	355.215527	355.215527	0	0	1	-360	360;
	2830	5661	0.007159	0.05086	0.13337	408.47886	408.47886	408.47886	0	0	1	-360	360;
	2830	3356	0.009669	0.05649	0.13532	324.375531	324.375531	324.375531	0	0	1	-360	360;
	2830	5606	0.0092	0.052169	0.06528	554.467188	554.467188	554.467188	0	0	1	-360	360;
	7608	4631	0.00197	0.01594	0.38068	1324.85571	1324.85571	1324.85571	0	0	1	-360	360;
	7608	7989	0.00343	0.02573	0.71559	1267.15576	1267.15576	1267.15576	0	0	1	-360	360;
	7608	7477	0.00151	0.01485	0.36026	1191.92323	1191.92323	1191.92323	0	0	1	-360	360;
	7608	4603	0.00092	0.01042	0.2249	1037.43736	1037.43736	1037.43736	0	0	1	-360	360;
	7608	7267	0.00029	0.00336	0.07002	1548.00614	1548.00614	1548.00614	0	0	1	-360	360;
	1357	4085	0.001711	0.01255	0.03159	293.674237	293.674237	293.674237	0	0	1	-360	360;
	1357	1369	0.005	0.03962	0.05923	571.522657	571.522657	571.522657	0	0	1	-360	360;
	768	1063	0.0014	0.01687	0.34391	1255.65671	1255.65671	1255.65671	0	0	1	-360	360;
	768	1063	0.00067	0.00678	0.17127	1943.77067	1943.77067	1943.77067	0	0	1	-360	360;
	768	1063	0.00065	0.00692	0.16778	1700.86828	1700.86828	1700.86828	0	0	1	-360	360;
	5528	768	0.00137	0.01524	0.32277	1231.35871	1231.35871	1231.35871	0	0	1	-360	360;
	3495	622	0.00424	0.02376	0.06041	235.143252	235.143252	235.143252	0	0	1	-360	360;
	3495	5282	0.005591	0.031711	0.05037	483.788672	483.788672	483.788672	0	0	1	-360	360;
	3495	2027	0.00138	0.00649	0.01411	435.897398	435.897398	435.897398	0	0	1	-360	360;
	3495	2027	0.00107	0.00544	0.01489	527.999754	527.999754	527.999754	0	0	1	-360	360;
	4631	3997	3e-05	0.00041	0.01153	1777.88776	1777.88776	1777.88776	0	0	1	-360	360;
	4631	3997	3e-05	0.00041	0.01129	1629.61313	1629.61313	1629.61313	0	0	1	-360	360;
	4631	4510	0.00044	0.00342	0.09556	1334.65073	1334.65073	1334.65073	0	0	1	-360	360;
	6750	2268	0.003841	0.01737	0.03461	446.137767	446.137767	446.137767	0	0	1	-360	360;
	1590	1749	0.001771	0.01337	0.01926	505.359979	505.359979	505.359979	0	0	1	-360	360;
	437	2410	0.00139	0.007349	0.01217	534.679962	534.679962	534.679962	0	0	1	-360	360;
	437	5078	0.00625	0.03463	0.05639	542.392057	542.392057	542.392057	0	0	1	-360	360;
	4649	3970	0.00055	0.00255	0.00485	360.858895	360.858895	360.858895	0	0	1	-360	360;
	5981	3192	0.00489	0.043331	0.10534	258.732225	258.732225	258.732225	0	0	1	-360	360;
	7891	7438	0.00145	0.013099	0.03507	397.691683	397.691683	397.691683	0	0	1	-360	360;
	7891	1156	0.004039	0.02881	0.09804	330.849971	330.849971	330.849971	0	0	1	-360	360;
	2038	4519	0.002219	0.00988	0.03167	447.281302	447.281302	447.281302	0	0	1	-360	360;
	7776	7117	0.001919	0.0153	0.03557	416.696229	416.696229	416.696229	0	0	1	-360	360;
	7858	1273	0.00189	0.01319	0.32109	1410.03333	1410.03333	1410.03333	0	0	1	-360	360;
	4731	8113	0.00011	0.000659	0.00129	498.042272	498.042272	498.042272	0	0	1	-360	360;
	4731	2322	0.00011	0.00064	0.00132	467.091771	467.091771	467.091771	0	0	1	-360	360;
	4731	2753	0.000211	0.00111	0.00191	260.18403	260.18403	260.18403	0	0	1	-360	360;
	6202	5634	0.000469	0.001669	0.24078	406.630381	406.630381	406.630381	0	0	1	-360	360;
	5372	772	6e-05	0.000849	0.11043	279.394726	279.394726	279.394726	0	0	1	-360	360;
	6474	6627	0.00051	0.00597	0.13001	1593.06707	1593.06707	1593.06707	0	0	1	-360	360;
	6474	6320	0.00119	0.01087	0.29433	1218.45378	1218.45378	1218.45378	0	0	1	-360	360;
	5291	6474	0.00219	0.01922	0.55346	1402.92104	1402.92104	1402.92104	0	0	1	-360	360;
	5370	2505	0.00051	0.00465	0.12608	1583.9605	1583.9605	1583.9605	0	0	1	-360	360;
	8052	8001	0.00055	0.0067	0.01252	284.39869	284.39869	284.39869	0	0	1	-360	360;
	2268	4842	0.005099	0.027349	0.04815	564.960575	564.960575	564.960575	0	0	1	-360	360;
	2939	2575	0.000599	0.003961	0.01127	489.098239	489.098239	489.098239	0	0	1	-360	360;
	2939	2991	0.005461	0.032719	0.04907	488.871133	488.871133	488.871133	0	0	1	-360	360;
	7385	2625	0.00349	0.02256	0.0383	478.105675	478.105675	478.105675	0	0	1	-360	360;
	7385	1648	0.00132	0.00861	0.01366	401.775027	401.775027	401.775027	0	0	1	-360	360;
	5288	396	0.0045	0.01838	0.03264	356.165487	356.165487	356.165487	0	0	1	-360	360;
	5288	1265	0.004289	0.019719	0.03123	409.1095	409.1095	409.1095	0	0	1	-360	360;
	7465	2753	0.001409	0.00881	0.01461	496.965422	496.965422	496.965422	0	0	1	-360	360;
	1063	5992	0.00212	0.01857	0.51971	1423.37462	1423.37462	1423.37462	0	0	1	-360	360;
	1063	6961	0.00049	0.00574	0.11994	1628.94311	1628.94311	1628.94311	0	0	1	-360	360;
	1063	684	0.00184	0.021	0.45843	1566.55231	1566.55231	1566.55231	0	0	1	-360	360;
	2435	4340	0.00593	0.037409	0.12416	437.053126	437.053126	437.053126	0	0	1	-360	360;
	2027	3805	0.00514	0.024169	0.03245	439.603121	439.603121	439.603121	0	0	1	-360	360;
	3458	2642	0.00041	0.00492	0.10366	1258.3875	1258.3875	1258.3875	0	0	1	-360	360;
	3223	2642	0.00296	0.02677	0.7466	1349.70988	1349.70988	1349.70988	0	0	1	-360	360;
	4463	496	0.0072	0.03406	0.06334	398.857318	398.857318	398.857318	0	0	1	-360	360;
	3113	6302	0.00043	0.00276	0.00814	424.753556	424.753556	424.753556	0	0	1	-360	360;
	3113	8199	0.001789	0.013469	0.01352	551.683228	551.683228	551.683228	0	0	1	-360	360;
	4450	6713	0.00163	0.011091	0.01944	471.43766	471.43766	471.43766	0	0	1	-360	360;
	4450	5760	0.00036	0.002651	0.00698	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8048	12	0.00013	0.00132	0.11865	525.679533	525.679533	525.679533	0	0	1	-360	360;
	8048	259	0.000169	0.001659	0.18764	460.038133	460.038133	460.038133	0	0	1	-360	360;
	4033	7737	0.00136	0.012831	0.02471	267.426289	267.426289	267.426289	0	0	1	-360	360;
	4033	7737	0.00139	0.01256	0.02517	255.973034	255.973034	255.973034	0	0	1	-360	360;
	4033	4796	0.00126	0.009521	0.02337	225.265644	225.265644	225.265644	0	0	1	-360	360;
	4033	4796	0.00143	0.009331	0.02649	225.565531	225.565531	225.565531	0	0	1	-360	360;
	3314	4033	0.00164	0.012961	0.034	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3314	4033	0.001599	0.012	0.03326	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6291	2686	0.001271	0.00524	0.00955	438.762141	438.762141	438.762141	0	0	1	-360	360;
	7469	8148	0.00226	0.02465	0.6298	1528.46084	1528.46084	1528.46084	0	0	1	-360	360;
	5934	3379	0.00032	0.0015	0.11985	530.036472	530.036472	530.036472	0	0	1	-360	360;
	5934	5395	0.00032	0.00112	0.10237	432.587968	432.587968	432.587968	0	0	1	-360	360;
	2977	4485	0.00075	0.00817	0.16312	1059.38896	1059.38896	1059.38896	0	0	1	-360	360;
	7398	3112	0.00037	0.0027	0.32698	557.541508	557.541508	557.541508	0	0	1	-360	360;
	3112	6152	3.9e-05	0.000219	0.00076	353.272547	353.272547	353.272547	0	0	1	-360	360;
	1978	539	0.012781	0.05762	0.0969	348.456822	348.456822	348.456822	0	0	1	-360	360;
	7852	971	0.00036	0.001289	0.17186	353.167377	353.167377	353.167377	0	0	1	-360	360;
	5526	1069	0.006771	0.039349	0.07095	322.173436	322.173436	322.173436	0	0	1	-360	360;
	2825	6961	0.00292	0.02831	0.72728	1432.39036	1432.39036	1432.39036	0	0	1	-360	360;
	2825	1911	0.00134	0.01229	0.31832	1533.89017	1533.89017	1533.89017	0	0	1	-360	360;
	7398	4991	0.00011	0.00063	0.10373	584.791621	584.791621	584.791621	0	0	1	-360	360;
	6571	7438	0.0003	0.003331	0.0085	283.934569	283.934569	283.934569	0	0	1	-360	360;
	6571	3359	0.001289	0.01555	0.03654	222.195514	222.195514	222.195514	0	0	1	-360	360;
	6152	5372	0.00011	0.0017	0.16068	556.383113	556.383113	556.383113	0	0	1	-360	360;
	6152	7776	0.00043	0.00249	0.00828	534.30234	534.30234	534.30234	0	0	1	-360	360;
	1022	4231	0.002651	0.016771	0.02821	237.954267	237.954267	237.954267	0	0	1	-360	360;
	3168	1022	0.0082	0.025081	0.04994	403.271034	403.271034	403.271034	0	0	1	-360	360;
	3127	8001	0.00037	0.002781	0.007	594.054213	594.054213	594.054213	0	0	1	-360	360;
	4419	4463	0.010719	0.054719	0.09428	413.481681	413.481681	413.481681	0	0	1	-360	360;
	2797	4463	0.006469	0.04074	0.12044	406.865109	406.865109	406.865109	0	0	1	-360	360;
	5338	8148	0.00101	0.01274	0.24051	1312.95056	1312.95056	1312.95056	0	0	1	-360	360;
	7438	5408	9.1e-05	0.05076	0.06756	437.183446	437.183446	437.183446	0	0	1	-360	360;
	7438	3301	0.00588	0.03938	0.05157	473.029311	473.029311	473.029311	0	0	1	-360	360;
	7438	3359	0.003591	0.022099	0.03114	446.585883	446.585883	446.585883	0	0	1	-360	360;
	7438	6924	0.00555	0.032521	0.04324	268.765303	268.765303	268.765303	0	0	1	-360	360;
	6239	7034	0.00116	0.01347	0.30147	1526.62321	1526.62321	1526.62321	0	0	1	-360	360;
	6239	7034	0.00113	0.01247	0.29492	1571.75917	1571.75917	1571.75917	0	0	1	-360	360;
	4585	159	0.00107	0.00871	0.30588	984.906097	984.906097	984.906097	0	0	1	-360	360;
	4585	4046	0.00078	0.00609	0.22253	1384.28733	1384.28733	1384.28733	0	0	1	-360	360;
	2324	6139	0.00075	0.00604	0.21992	1146.03892	1146.03892	1146.03892	0	0	1	-360	360;
	4456	6093	0.00224	0.011091	0.02035	298.994093	298.994093	298.994093	0	0	1	-360	360;
	6778	6302	0.001961	0.013969	0.03661	418.120979	418.120979	418.120979	0	0	1	-360	360;
	1177	4694	0.001271	0.00761	0.01431	486.446504	486.446504	486.446504	0	0	1	-360	360;
	6058	7989	0.00374	0.03581	0.90723	1396.77694	1396.77694	1396.77694	0	0	1	-360	360;
	6058	7989	0.00385	0.03284	1.01893	1334.34007	1334.34007	1334.34007	0	0	1	-360	360;
	6058	7989	0.00392	0.03216	1.03852	1190.95834	1190.95834	1190.95834	0	0	1	-360	360;
	6058	8148	0.00307	0.03131	0.81414	1302.91991	1302.91991	1302.91991	0	0	1	-360	360;
	7091	27	0.027	0.131289	0.23246	453.439851	453.439851	453.439851	0	0	1	-360	360;
	5744	5003	0.005539	0.0195	0.09625	242.757417	242.757417	242.757417	0	0	1	-360	360;
	7737	8208	0.003099	0.017849	0.02707	506.488272	506.488272	506.488272	0	0	1	-360	360;
	7737	4796	0.002831	0.014099	0.02429	534.598417	534.598417	534.598417	0	0	1	-360	360;
	3668	2059	0	0.01368	0	1611.01364	1611.01364	1611.01364	0	0	1	-360	360;
	3829	4231	0.01082	0.079719	0.11605	377.953994	377.953994	377.953994	0	0	1	-360	360;
	662	2505	0.00102	0.00941	0.25147	1250.56109	1250.56109	1250.56109	0	0	1	-360	360;
	662	541	0.00183	0.01819	0.6177	1806.07824	1806.07824	1806.07824	0	0	1	-360	360;
	662	346	0.00028	0.00346	0.07481	1268.36352	1268.36352	1268.36352	0	0	1	-360	360;
	3392	2322	0.0012	0.00982	0.01289	237.7706	237.7706	237.7706	0	0	1	-360	360;
	2505	2376	0.00027	0.0032	0.09026	1838.64627	1838.64627	1838.64627	0	0	1	-360	360;
	2505	6974	0.00064	0.00679	0.15697	1158.39426	1158.39426	1158.39426	0	0	1	-360	360;
	2505	7487	0.00047	0.00442	0.15628	1735.90251	1735.90251	1735.90251	0	0	1	-360	360;
	373	3059	0.00493	0.017599	0.04325	499.905994	499.905994	499.905994	0	0	1	-360	360;
	373	3716	0.00207	0.0153	0.04117	277.720006	277.720006	277.720006	0	0	1	-360	360;
	6606	7034	0.00094	0.0092	0.31227	1909.95013	1909.95013	1909.95013	0	0	1	-360	360;
	6606	7034	0.00095	0.00901	0.31828	1717.48665	1717.48665	1717.48665	0	0	1	-360	360;
	858	2523	0.00101	0.00719	0.01887	340.716148	340.716148	340.716148	0	0	1	-360	360;
	3398	6034	0.000419	0.00331	0.00803	573.547563	573.547563	573.547563	0	0	1	-360	360;
	2309	1735	0.002099	0.015599	0.03927	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7989	7858	0.00163	0.01283	0.27543	1312.01002	1312.01002	1312.01002	0	0	1	-360	360;
	7989	6734	0.00445	0.03788	1.27074	1531.01195	1531.01195	1531.01195	0	0	1	-360	360;
	7989	4227	0.00092	0.00703	0.1816	1218.87699	1218.87699	1218.87699	0	0	1	-360	360;
	227	27	9.1e-05	0.000961	0.00046	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1590	227	0.00387	0.028599	0.07207	303.707315	303.707315	303.707315	0	0	1	-360	360;
	227	4807	0.00949	0.06625	0.18391	309.626945	309.626945	309.626945	0	0	1	-360	360;
	227	4065	0.00414	0.035331	0.08022	267.958237	267.958237	267.958237	0	0	1	-360	360;
	227	4807	0.009781	0.065581	0.18949	296.815623	296.815623	296.815623	0	0	1	-360	360;
	1265	27	0.02818	0.116099	0.20127	387.059593	387.059593	387.059593	0	0	1	-360	360;
	27	1952	0.00324	0.01912	0.08909	397.676059	397.676059	397.676059	0	0	1	-360	360;
	2734	6139	0.00135	0.01609	0.454	1616.36793	1616.36793	1616.36793	0	0	1	-360	360;
	5602	5146	0.00294	0.0182	0.02782	444.711493	444.711493	444.711493	0	0	1	-360	360;
	7317	5408	0.000711	0.00649	0.01737	458.21709	458.21709	458.21709	0	0	1	-360	360;
	398	7317	0.002651	0.01375	0.02404	296.32521	296.32521	296.32521	0	0	1	-360	360;
	1156	5408	0.00105	0.008469	0.02555	437.408266	437.408266	437.408266	0	0	1	-360	360;
	785	4085	0.002229	0.016461	0.02693	500.815182	500.815182	500.815182	0	0	1	-360	360;
	785	4085	0.002281	0.01613	0.02746	505.352358	505.352358	505.352358	0	0	1	-360	360;
	546	4085	0.003341	0.012651	0.01616	274.338557	274.338557	274.338557	0	0	1	-360	360;
	546	4085	0.003271	0.012901	0.01583	261.584775	261.584775	261.584775	0	0	1	-360	360;
	5882	7185	0.00101	0.01218	0.33876	1855.39101	1855.39101	1855.39101	0	0	1	-360	360;
	5882	6740	0.0009	0.0095	0.24781	1980.89462	1980.89462	1980.89462	0	0	1	-360	360;
	7185	4262	0.00098	0.01241	0.33155	2041.82162	2041.82162	2041.82162	0	0	1	-360	360;
	5002	1774	0.001219	0.00951	0.02586	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5002	7631	0.001789	0.01382	0.03345	225.44207	225.44207	225.44207	0	0	1	-360	360;
	5002	7631	0.00175	0.014081	0.03273	242.635862	242.635862	242.635862	0	0	1	-360	360;
	6907	3073	9.1e-05	0.00087	0.12315	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6907	3073	9.1e-05	0.00086	0.12543	217.164877	217.164877	217.164877	0	0	1	-360	360;
	6907	7140	0.000229	0.00236	0.32841	409.066441	409.066441	409.066441	0	0	1	-360	360;
	622	4703	0.00125	0.00557	0.01779	594.916532	594.916532	594.916532	0	0	1	-360	360;
	1978	4721	0.00757	0.04013	0.0575	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1547	159	0.00157	0.01604	0.39323	1411.21805	1411.21805	1411.21805	0	0	1	-360	360;
	1547	159	0.00154	0.01634	0.38504	1298.38045	1298.38045	1298.38045	0	0	1	-360	360;
	159	1911	0.00169	0.02383	0.71005	1563.19955	1563.19955	1563.19955	0	0	1	-360	360;
	1578	159	0.00068	0.0089	0.28329	1552.47912	1552.47912	1552.47912	0	0	1	-360	360;
	1691	4189	0.010289	0.036651	0.09501	486.001055	486.001055	486.001055	0	0	1	-360	360;
	1691	4189	0.00563	0.03395	0.10451	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2828	5198	0.003591	0.02763	0.06689	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2828	5198	0.00351	0.028159	0.06547	255.015452	255.015452	255.015452	0	0	1	-360	360;
	5198	4340	0.00143	0.00838	0.03016	318.870865	318.870865	318.870865	0	0	1	-360	360;
	6948	6241	0.01213	0.06156	0.09183	495.321948	495.321948	495.321948	0	0	1	-360	360;
	5602	4703	0.004789	0.022849	0.06926	560.253831	560.253831	560.253831	0	0	1	-360	360;
	477	7471	0.00019	0.00211	0.00538	469.661581	469.661581	469.661581	0	0	1	-360	360;
	477	7732	0.000331	0.002349	0.00731	285.04419	285.04419	285.04419	0	0	1	-360	360;
	477	7733	9.1e-05	0.000539	0.00112	398.927051	398.927051	398.927051	0	0	1	-360	360;
	5893	5992	0.00088	0.00776	0.29732	1394.7458	1394.7458	1394.7458	0	0	1	-360	360;
	4227	4278	0.00039	0.00368	0.13016	1979.35777	1979.35777	1979.35777	0	0	1	-360	360;
	6036	1906	0.00275	0.01268	0.02402	411.958239	411.958239	411.958239	0	0	1	-360	360;
	6036	8167	0.00205	0.00874	0.02813	284.521007	284.521007	284.521007	0	0	1	-360	360;
	6036	1127	0.00082	0.00519	0.01536	595.667965	595.667965	595.667965	0	0	1	-360	360;
	2410	733	0.00519	0.02355	0.04694	465.913942	465.913942	465.913942	0	0	1	-360	360;
	2991	6121	0.016711	0.08874	0.14919	430.909438	430.909438	430.909438	0	0	1	-360	360;
	6713	4004	0.005099	0.022659	0.03759	495.229734	495.229734	495.229734	0	0	1	-360	360;
	6713	1618	0.001469	0.01213	0.01502	601.622271	601.622271	601.622271	0	0	1	-360	360;
	6713	1618	0.001659	0.01189	0.01702	552.155731	552.155731	552.155731	0	0	1	-360	360;
	922	4065	0.000521	0.00374	0.01092	421.39078	421.39078	421.39078	0	0	1	-360	360;
	922	4571	0.006461	0.048349	0.12384	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7185	6627	0.00293	0.0257	0.74111	1316.59885	1316.59885	1316.59885	0	0	1	-360	360;
	7185	5089	0.00178	0.0157	0.44951	1415.92666	1415.92666	1415.92666	0	0	1	-360	360;
	5291	7185	0.00115	0.01332	0.28982	1089.13142	1089.13142	1089.13142	0	0	1	-360	360;
	7185	2005	0.00078	0.00752	0.19838	1687.74221	1687.74221	1687.74221	0	0	1	-360	360;
	67	6040	3.9e-05	0.00037	0.01462	375.978624	375.978624	375.978624	0	0	1	-360	360;
	7363	7631	0.0012	0.006729	0.02243	550.894071	550.894071	550.894071	0	0	1	-360	360;
	28	3028	0.000211	0.001659	0.00497	365.53706	365.53706	365.53706	0	0	1	-360	360;
	5887	7526	0.001659	0.011919	0.03316	513.100653	513.100653	513.100653	0	0	1	-360	360;
	5887	6231	0.002	0.0133	0.04009	444.65243	444.65243	444.65243	0	0	1	-360	360;
	3980	7092	0.00105	0.00837	0.02095	228.351777	228.351777	228.351777	0	0	1	-360	360;
	2625	4859	0.00138	0.014979	0.037	359.210086	359.210086	359.210086	0	0	1	-360	360;
	5233	6202	0	0.0003	0.03478	254.555523	254.555523	254.555523	0	0	1	-360	360;
	2707	6202	0	0.0003	0.03546	434.697849	434.697849	434.697849	0	0	1	-360	360;
	2707	4896	0.00164	0.00782	0.01408	403.082795	403.082795	403.082795	0	0	1	-360	360;
	3301	209	0.007099	0.04044	0.0622	522.240546	522.240546	522.240546	0	0	1	-360	360;
	8208	7132	0.00181	0.00914	0.01575	583.331433	583.331433	583.331433	0	0	1	-360	360;
	1345	6658	0.00027	0.00247	0.09082	1943.29678	1943.29678	1943.29678	0	0	1	-360	360;
	6658	4510	0.00266	0.03206	1.13637	1676.32017	1676.32017	1676.32017	0	0	1	-360	360;
	7558	891	0.000591	0.005081	0.00898	412.018826	412.018826	412.018826	0	0	1	-360	360;
	7558	681	0.002229	0.016099	0.04517	435.74231	435.74231	435.74231	0	0	1	-360	360;
	3668	4878	0.00205	0.02585	0.06907	1110.26885	1110.26885	1110.26885	0	0	1	-360	360;
	7093	581	0.00101	0.0091	0.2532	1335.85059	1335.85059	1335.85059	0	0	1	-360	360;
	7093	5089	0.001	0.00871	0.25243	1318.5543	1318.5543	1318.5543	0	0	1	-360	360;
	8256	1379	0.00112	0.00895	0.02078	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8256	1379	0.001099	0.008289	0.02033	220.59548	220.59548	220.59548	0	0	1	-360	360;
	8256	2844	0.001289	0.0095	0.02391	231.229095	231.229095	231.229095	0	0	1	-360	360;
	8256	2844	0.00126	0.00968	0.02342	231.911176	231.911176	231.911176	0	0	1	-360	360;
	3928	4958	0.00065	0.00723	0.21781	1551.52345	1551.52345	1551.52345	0	0	1	-360	360;
	1774	3728	0.001781	0.012669	0.03664	309.734401	309.734401	309.734401	0	0	1	-360	360;
	3614	980	0.00086	0.00956	0.21021	1066.62367	1066.62367	1066.62367	0	0	1	-360	360;
	3614	732	0.00248	0.02429	0.63341	1307.25336	1307.25336	1307.25336	0	0	1	-360	360;
	3614	4103	0.00151	0.01699	0.3774	1347.54512	1347.54512	1347.54512	0	0	1	-360	360;
	5343	3356	0.003159	0.01537	0.04444	255.341251	255.341251	255.341251	0	0	1	-360	360;
	7441	2832	0.00075	0.00683	0.18112	1377.87798	1377.87798	1377.87798	0	0	1	-360	360;
	7441	2580	0.00223	0.02524	0.55052	1639.50491	1639.50491	1639.50491	0	0	1	-360	360;
	581	5089	0.00191	0.0208	0.46218	993.197181	993.197181	993.197181	0	0	1	-360	360;
	1547	581	0.00074	0.00784	0.18584	935.02532	935.02532	935.02532	0	0	1	-360	360;
	1547	581	0.00073	0.00799	0.18188	1104.62299	1104.62299	1104.62299	0	0	1	-360	360;
	1293	3295	0.000789	0.00482	0.01619	529.719057	529.719057	529.719057	0	0	1	-360	360;
	1938	1293	0.000789	0.004409	0.01467	386.72503	386.72503	386.72503	0	0	1	-360	360;
	2575	2991	0.005461	0.03161	0.05048	547.037833	547.037833	547.037833	0	0	1	-360	360;
	1493	7776	6e-05	0.00087	0.09273	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2878	1284	0.00593	0.040961	0.1172	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2878	4571	0.003771	0.030841	0.0752	268.225734	268.225734	268.225734	0	0	1	-360	360;
	5856	7772	0.00038	0.001461	0.16307	548.742656	548.742656	548.742656	0	0	1	-360	360;
	7772	3716	0.00149	0.00938	0.02776	304.793311	304.793311	304.793311	0	0	1	-360	360;
	7772	3716	0.001521	0.00918	0.02829	350.298823	350.298823	350.298823	0	0	1	-360	360;
	1906	396	0.00407	0.015469	0.02869	357.367323	357.367323	357.367323	0	0	1	-360	360;
	4507	1293	0.00082	0.006219	0.01525	275.399023	275.399023	275.399023	0	0	1	-360	360;
	2417	1799	0.00161	0.00888	0.0144	445.381381	445.381381	445.381381	0	0	1	-360	360;
	2417	8199	0.00118	0.006461	0.01047	523.997573	523.997573	523.997573	0	0	1	-360	360;
	8001	5566	0.00082	0.009091	0.01791	215.929128	215.929128	215.929128	0	0	1	-360	360;
	681	4089	0.01893	0.0848	0.26385	460.390605	460.390605	460.390605	0	0	1	-360	360;
	681	4089	0.01313	0.082409	0.2672	451.384461	451.384461	451.384461	0	0	1	-360	360;
	681	5755	0.000281	0.00149	0.00187	243.959252	243.959252	243.959252	0	0	1	-360	360;
	4806	2832	0.00019	0.00199	0.05042	1351.76998	1351.76998	1351.76998	0	0	1	-360	360;
	2832	4806	0.0002	0.00188	0.05312	1398.93511	1398.93511	1398.93511	0	0	1	-360	360;
	152	2676	0.003159	0.025331	0.05892	233.139304	233.139304	233.139304	0	0	1	-360	360;
	6971	5801	0.00445	0.01886	0.03964	616.123554	616.123554	616.123554	0	0	1	-360	360;
	5282	4842	0.00461	0.02205	0.04378	525.746979	525.746979	525.746979	0	0	1	-360	360;
	6602	7444	0.00061	0.00494	0.01145	248.684668	248.684668	248.684668	0	0	1	-360	360;
	6602	873	0.00082	0.00462	0.02449	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3150	7855	0.000331	0.003271	0.00943	560.796829	560.796829	560.796829	0	0	1	-360	360;
	3150	1618	0.00056	0.004081	0.01565	604.571607	604.571607	604.571607	0	0	1	-360	360;
	5666	8223	0.00163	0.007521	0.02413	299.494794	299.494794	299.494794	0	0	1	-360	360;
	5666	1338	0.001659	0.00736	0.02459	228.445897	228.445897	228.445897	0	0	1	-360	360;
	5935	1369	0.00719	0.049211	0.07788	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6766	1799	6e-05	0.00039	0.00056	452.721569	452.721569	452.721569	0	0	1	-360	360;
	938	5465	0.00288	0.018	0.02583	547.767927	547.767927	547.767927	0	0	1	-360	360;
	938	7962	0.00288	0.0182	0.02628	506.247066	506.247066	506.247066	0	0	1	-360	360;
	938	4807	0.00111	0.008531	0.01377	455.463614	455.463614	455.463614	0	0	1	-360	360;
	938	3896	0.00307	0.022289	0.13473	519.69703	519.69703	519.69703	0	0	1	-360	360;
	2789	4103	0.00091	0.00868	0.30687	1547.69613	1547.69613	1547.69613	0	0	1	-360	360;
	5089	8272	0.00098	0.0118	0.22347	1498.15564	1498.15564	1498.15564	0	0	1	-360	360;
	5089	6740	0.00053	0.00591	0.14433	1678.34539	1678.34539	1678.34539	0	0	1	-360	360;
	5089	2005	0.00086	0.00991	0.21418	1370.21216	1370.21216	1370.21216	0	0	1	-360	360;
	6350	3028	0.00545	0.034099	0.05137	496.737172	496.737172	496.737172	0	0	1	-360	360;
	1882	3176	0.00095	0.006091	0.02027	557.962189	557.962189	557.962189	0	0	1	-360	360;
	3223	732	0.00305	0.0272	0.76729	1317.23333	1317.23333	1317.23333	0	0	1	-360	360;
	785	4467	0.000281	0.001841	0.00299	444.160874	444.160874	444.160874	0	0	1	-360	360;
	785	5858	0.004531	0.02451	0.03616	328.684457	328.684457	328.684457	0	0	1	-360	360;
	4004	1369	0.00618	0.02961	0.05724	551.631024	551.631024	551.631024	0	0	1	-360	360;
	1379	6118	0.00064	0.004461	0.0135	307.600896	307.600896	307.600896	0	0	1	-360	360;
	828	3847	0.00143	0.00513	0.01209	570.250327	570.250327	570.250327	0	0	1	-360	360;
	828	1127	0.0052	0.021841	0.04386	473.866099	473.866099	473.866099	0	0	1	-360	360;
	546	4467	0.00138	0.008669	0.01594	452.285647	452.285647	452.285647	0	0	1	-360	360;
	5858	546	0.00314	0.013169	0.015	373.931999	373.931999	373.931999	0	0	1	-360	360;
	3847	1746	0.002479	0.01331	0.02234	491.930211	491.930211	491.930211	0	0	1	-360	360;
	2088	2485	7e-05	0.00194	0.14737	369.585729	369.585729	369.585729	0	0	1	-360	360;
	8313	7631	0.001159	0.00931	0.02166	216.314751	216.314751	216.314751	0	0	1	-360	360;
	4896	5760	0.00062	0.003669	0.01131	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2279	4695	0.015969	0.056711	0.13475	635.211932	635.211932	635.211932	0	0	1	-360	360;
	4695	3418	0.0085	0.03719	0.0693	419.618511	419.618511	419.618511	0	0	1	-360	360;
	2523	8223	0.001531	0.00732	0.02343	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2523	8223	0.00205	0.01062	0.03022	221.707388	221.707388	221.707388	0	0	1	-360	360;
	4510	4958	0.00091	0.01056	0.22032	767.876077	767.876077	767.876077	0	0	1	-360	360;
	732	7822	0.00167	0.02069	0.39692	1558.8944	1558.8944	1558.8944	0	0	1	-360	360;
	5714	6241	0.002719	0.016219	0.0236	530.822962	530.822962	530.822962	0	0	1	-360	360;
	7477	3997	0.002	0.02376	0.67032	1736.84305	1736.84305	1736.84305	0	0	1	-360	360;
	7477	3997	0.00226	0.02328	0.7597	1664.45451	1664.45451	1664.45451	0	0	1	-360	360;
	6924	398	0.004151	0.02119	0.03874	544.293121	544.293121	544.293121	0	0	1	-360	360;
	8167	7593	0.001841	0.010909	0.02506	229.497598	229.497598	229.497598	0	0	1	-360	360;
	6241	3809	0.000669	0.00399	0.00581	481.73633	481.73633	481.73633	0	0	1	-360	360;
	3330	1341	8e-05	0.00098	0.0192	1637.6291	1637.6291	1637.6291	0	0	1	-360	360;
	7471	7733	0.0002	0.0025	0.01729	312.997343	312.997343	312.997343	0	0	1	-360	360;
	4985	4807	0.006219	0.031599	0.055	535.941241	535.941241	535.941241	0	0	1	-360	360;
	4985	1308	0.00438	0.023169	0.03876	545.127623	545.127623	545.127623	0	0	1	-360	360;
	5574	3896	0.00581	0.03088	0.06283	516.022934	516.022934	516.022934	0	0	1	-360	360;
	5574	3896	0.0065	0.02637	0.07223	438.647825	438.647825	438.647825	0	0	1	-360	360;
	4227	541	0.00198	0.02208	0.66516	1693.34332	1693.34332	1693.34332	0	0	1	-360	360;
	2828	4842	0.00093	0.006659	0.01743	328.529751	328.529751	328.529751	0	0	1	-360	360;
	2828	4842	0.000909	0.006789	0.01708	297.063688	297.063688	297.063688	0	0	1	-360	360;
	3668	980	0.00244	0.01919	0.6083	864.403996	864.403996	864.403996	0	0	1	-360	360;
	2139	7782	0.00194	0.01549	0.03601	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1518	3697	0.0002	0.001479	0.00229	519.71875	519.71875	519.71875	0	0	1	-360	360;
	4089	1379	0.001091	0.009461	0.02148	617.911065	617.911065	617.911065	0	0	1	-360	360;
	1379	5989	0.003081	0.00899	0.01672	499.655643	499.655643	499.655643	0	0	1	-360	360;
	5989	1325	0.00357	0.0138	0.01939	470.841315	470.841315	470.841315	0	0	1	-360	360;
	3176	6923	0.001229	0.00986	0.02292	493.872429	493.872429	493.872429	0	0	1	-360	360;
	1812	8262	0.001211	0.00913	0.02242	577.845058	577.845058	577.845058	0	0	1	-360	360;
	4510	7267	0.00116	0.01053	0.28682	1582.64611	1582.64611	1582.64611	0	0	1	-360	360;
	4226	5157	0.0017	0.009781	0.03441	248.831754	248.831754	248.831754	0	0	1	-360	360;
	4226	5157	0.001521	0.011091	0.03062	220.763143	220.763143	220.763143	0	0	1	-360	360;
	7069	401	0.0043	0.0233	0.06331	573.180611	573.180611	573.180611	0	0	1	-360	360;
	2448	4050	0.00019	0.0016	0.06168	991.811056	991.811056	991.811056	0	0	1	-360	360;
	4339	7526	0.00838	0.072479	0.17676	229.296022	229.296022	229.296022	0	0	1	-360	360;
	4339	897	0.00787	0.049211	0.15585	353.351043	353.351043	353.351043	0	0	1	-360	360;
	398	209	0.00557	0.03376	0.05029	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3859	6629	0.01213	0.06564	0.13833	340.04626	340.04626	340.04626	0	0	1	-360	360;
	6678	2844	0.003531	0.01086	0.0176	400.38762	400.38762	400.38762	0	0	1	-360	360;
	6678	6729	0.0158	0.057031	0.07883	338.373827	338.373827	338.373827	0	0	1	-360	360;
	684	1911	0.00289	0.0244	0.70742	1292.96231	1292.96231	1292.96231	0	0	1	-360	360;
	1457	3729	0.000521	0.00413	0.00749	553.25354	553.25354	553.25354	0	0	1	-360	360;
	7328	3729	0.000531	0.004039	0.00763	509.045506	509.045506	509.045506	0	0	1	-360	360;
	209	1648	0.00431	0.02226	0.04491	419.196306	419.196306	419.196306	0	0	1	-360	360;
	2753	3716	0.00163	0.00899	0.01424	558.003724	558.003724	558.003724	0	0	1	-360	360;
	4227	1165	0.00274	0.02428	0.78775	1430.67317	1430.67317	1430.67317	0	0	1	-360	360;
	5729	5465	0.0093	0.056979	0.08381	526.04458	526.04458	526.04458	0	0	1	-360	360;
	5729	7918	0.011281	0.080039	0.20991	253.337684	253.337684	253.337684	0	0	1	-360	360;
	5729	5109	0.00449	0.02376	0.07508	450.476797	450.476797	450.476797	0	0	1	-360	360;
	2376	7487	0.00014	0.00148	0.04807	1754.011	1754.011	1754.011	0	0	1	-360	360;
	6165	7526	0.0055	0.033659	0.10993	429.020567	429.020567	429.020567	0	0	1	-360	360;
	4065	4959	0.00249	0.02082	0.05329	471.233798	471.233798	471.233798	0	0	1	-360	360;
	1578	1911	0.00164	0.02252	0.68801	1606.12008	1606.12008	1606.12008	0	0	1	-360	360;
	5257	1956	0.000961	0.0047	0.00912	262.218463	262.218463	262.218463	0	0	1	-360	360;
	5257	1956	0.000979	0.004599	0.00929	270.097839	270.097839	270.097839	0	0	1	-360	360;
	6974	346	0.00057	0.00515	0.14926	1270.9528	1270.9528	1270.9528	0	0	1	-360	360;
	7526	6231	0.00143	0.011651	0.03006	382.701891	382.701891	382.701891	0	0	1	-360	360;
	1367	3716	0.00126	0.009211	0.02652	279.618022	279.618022	279.618022	0	0	1	-360	360;
	5465	630	9.1e-05	0.000211	0.00044	396.181577	396.181577	396.181577	0	0	1	-360	360;
	7962	630	0.000211	0.000711	0.0012	374.560733	374.560733	374.560733	0	0	1	-360	360;
	7962	5109	0.01095	0.07062	0.09816	464.957885	464.957885	464.957885	0	0	1	-360	360;
	1952	5109	0.00125	0.007281	0.01067	300.864292	300.864292	300.864292	0	0	1	-360	360;
	1952	5109	0.001	0.0045	0.0089	311.965837	311.965837	311.965837	0	0	1	-360	360;
	4959	4571	0.00282	0.026419	0.05997	303.581949	303.581949	303.581949	0	0	1	-360	360;
	6926	6007	0.003479	0.01456	0.05279	224.454005	224.454005	224.454005	0	0	1	-360	360;
	4807	3896	0.005521	0.03418	0.05866	460.549123	460.549123	460.549123	0	0	1	-360	360;
	4807	3896	0.00563	0.033461	0.05978	490.825544	490.825544	490.825544	0	0	1	-360	360;
	4571	4322	0.00095	0.0072	0.02091	425.181096	425.181096	425.181096	0	0	1	-360	360;
	4571	6629	0.00093	0.007159	0.01053	523.959468	523.959468	523.959468	0	0	1	-360	360;
	1308	3896	0.00988	0.058651	0.1092	586.366506	586.366506	586.366506	0	0	1	-360	360;
	6034	7140	3.9e-05	0.00044	0.06158	606.884588	606.884588	606.884588	0	0	1	-360	360;
	6034	7140	3.9e-05	0.00043	0.06272	544.680269	544.680269	544.680269	0	0	1	-360	360;
	6665	4936	0.00018	0.000781	0.00101	379.143635	379.143635	379.143635	0	0	1	-360	360;
	3927	6868	0.002961	0.01669	0.03408	434.588868	434.588868	434.588868	0	0	1	-360	360;
	5266	3567	0.00056	0.00432	0.00697	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6128	5478	0.0002	0.00246	0.05268	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6128	6279	0.0003	0.00377	0.07873	1447.11778	1447.11778	1447.11778	0	0	1	-360	360;
	4566	8268	0.00014	0.00057	0.00095	351.41759	351.41759	351.41759	0	0	1	-360	360;
	2659	8268	0.00013	0.000729	0.00089	359.459294	359.459294	359.459294	0	0	1	-360	360;
	1307	8268	0.00012	0.000581	0.00093	378.117084	378.117084	378.117084	0	0	1	-360	360;
	1692	8268	0.00011	0.00064	0.00091	280.599609	280.599609	280.599609	0	0	1	-360	360;
	7551	8268	0.00019	0.00056	0.00107	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5644	7976	0.00062	0.004909	0.00728	647.787002	647.787002	647.787002	0	0	1	-360	360;
	8352	7114	7e-05	0.00076	0.01698	1000.37002	1000.37002	1000.37002	0	0	1	-360	360;
	8352	8169	0.00066	0.00622	0.11119	1733.80357	1733.80357	1733.80357	0	0	1	-360	360;
	8352	1941	0.00054	0.00752	0.26021	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	8352	24	0.00105	0.00877	0.19663	1434.03844	1434.03844	1434.03844	0	0	1	-360	360;
	105	8083	8e-05	0.00091	0.0182	767.876077	767.876077	767.876077	0	0	1	-360	360;
	105	4184	0.00056	0.00719	0.26685	1080.22955	1080.22955	1080.22955	0	0	1	-360	360;
	2202	5023	0.000531	0.002039	0.0042	435.108622	435.108622	435.108622	0	0	1	-360	360;
	2202	4395	0.008659	0.03769	0.0546	403.161291	403.161291	403.161291	0	0	1	-360	360;
	2202	8136	0.003849	0.021099	0.03049	378.911575	378.911575	378.911575	0	0	1	-360	360;
	2202	7726	0.00688	0.035969	0.04966	414.189674	414.189674	414.189674	0	0	1	-360	360;
	2202	6250	0.006729	0.03713	0.04898	240.178081	240.178081	240.178081	0	0	1	-360	360;
	2202	3091	0.00111	0.006651	0.01236	526.441255	526.441255	526.441255	0	0	1	-360	360;
	2202	580	0.0023	0.008909	0.01481	360.931294	360.931294	360.931294	0	0	1	-360	360;
	2202	2356	0.000841	0.00313	0.00503	377.327927	377.327927	377.327927	0	0	1	-360	360;
	2202	4861	0.003521	0.017479	0.02439	409.168563	409.168563	409.168563	0	0	1	-360	360;
	2775	651	3.9e-05	0.00025	0	620.730082	620.730082	620.730082	0	0	1	-360	360;
	2761	2114	0.000159	0.000729	0.06691	457.491568	457.491568	457.491568	0	0	1	-360	360;
	7847	6776	0.005281	0.032099	0.05218	509.751594	509.751594	509.751594	0	0	1	-360	360;
	7847	6776	0.005651	0.029419	0.05541	485.345266	485.345266	485.345266	0	0	1	-360	360;
	816	3323	0.000271	0.00036	0.02718	318.400267	318.400267	318.400267	0	0	1	-360	360;
	7921	6394	0.00044	0.00143	0.19132	582.640587	582.640587	582.640587	0	0	1	-360	360;
	7921	3563	0.000159	0.00074	0.06733	564.796723	564.796723	564.796723	0	0	1	-360	360;
	2068	1803	0.00019	0.000591	0.14587	441.4516	441.4516	441.4516	0	0	1	-360	360;
	4664	2657	0.00019	0.00087	0.15682	299.879275	299.879275	299.879275	0	0	1	-360	360;
	6011	4474	0.002591	0.013659	0.0182	429.504502	429.504502	429.504502	0	0	1	-360	360;
	6011	6626	0.00305	0.01168	0.02236	280.555788	280.555788	280.555788	0	0	1	-360	360;
	371	1005	0.000159	0.000599	0.06412	538.327384	538.327384	538.327384	0	0	1	-360	360;
	371	1005	0.000159	0.00062	0.06278	503.836537	503.836537	503.836537	0	0	1	-360	360;
	2954	4474	0.00101	0.004229	0.00593	573.509839	573.509839	573.509839	0	0	1	-360	360;
	2954	3898	0.00319	0.018651	0.11399	566.230618	566.230618	566.230618	0	0	1	-360	360;
	2954	6626	0.000729	0.00337	0.00402	274.904037	274.904037	274.904037	0	0	1	-360	360;
	1044	1080	0.002961	0.012831	0.01603	390.3393	390.3393	390.3393	0	0	1	-360	360;
	1044	7210	0.000591	0.00339	0.0122	378.609783	378.609783	378.609783	0	0	1	-360	360;
	8080	2197	0.00024	0.001219	0.00175	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5511	2701	0.006219	0.036341	0.07005	485.907317	485.907317	485.907317	0	0	1	-360	360;
	5511	5473	0.001841	0.01193	0.02173	580.651119	580.651119	580.651119	0	0	1	-360	360;
	2992	6585	0.001409	0.01113	0.01514	518.798511	518.798511	518.798511	0	0	1	-360	360;
	2992	258	0.00151	0.00768	0.01116	482.193973	482.193973	482.193973	0	0	1	-360	360;
	2992	2791	0.00599	0.03313	0.04364	501.005707	501.005707	501.005707	0	0	1	-360	360;
	2992	3080	0.001771	0.010151	0.02023	470.454929	470.454929	470.454929	0	0	1	-360	360;
	4261	4490	0.00074	0.00311	0.31301	238.895844	238.895844	238.895844	0	0	1	-360	360;
	4261	4490	0.000729	0.003159	0.30635	299.358759	299.358759	299.358759	0	0	1	-360	360;
	2927	7082	0.005271	0.01906	0.03513	383.731111	383.731111	383.731111	0	0	1	-360	360;
	2927	4949	0.00906	0.053151	0.07992	436.824114	436.824114	436.824114	0	0	1	-360	360;
	2927	1698	0.00405	0.0153	0.02875	364.175945	364.175945	364.175945	0	0	1	-360	360;
	6174	4913	0.00055	0.003961	0.02467	507.257614	507.257614	507.257614	0	0	1	-360	360;
	6174	4913	0.00055	0.003729	0.01897	451.015603	451.015603	451.015603	0	0	1	-360	360;
	6174	1467	0.00107	0.007789	0.01233	526.277402	526.277402	526.277402	0	0	1	-360	360;
	6174	1467	0.00112	0.00775	0.01328	580.409914	580.409914	580.409914	0	0	1	-360	360;
	6174	1467	0.001031	0.00775	0.0122	284.016495	284.016495	284.016495	0	0	1	-360	360;
	6174	2213	0.00062	0.003219	0.00455	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6174	865	0.00068	0.002659	0.00532	520.876002	520.876002	520.876002	0	0	1	-360	360;
	8292	454	0.00193	0.012479	0.02215	593.787478	593.787478	593.787478	0	0	1	-360	360;
	8292	2739	0.001159	0.007229	0.01333	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8292	2679	0.00137	0.008729	0.0151	499.489505	499.489505	499.489505	0	0	1	-360	360;
	8292	2679	0.001401	0.00855	0.01539	545.544112	545.544112	545.544112	0	0	1	-360	360;
	8292	2679	0.001331	0.008521	0.0151	541.137636	541.137636	541.137636	0	0	1	-360	360;
	6413	2011	0.00031	0.00081	0.05914	392.237316	392.237316	392.237316	0	0	1	-360	360;
	6413	7312	0.000341	0.00119	0.1412	399.770317	399.770317	399.770317	0	0	1	-360	360;
	6413	4080	0.00031	0.001281	0.13155	495.071978	495.071978	495.071978	0	0	1	-360	360;
	6413	3860	0.00094	0.0035	0.15159	543.601132	543.601132	543.601132	0	0	1	-360	360;
	7738	6391	7e-05	0.000651	0.06234	217.31844	217.31844	217.31844	0	0	1	-360	360;
	4970	3326	0.00019	0.00124	0.13922	602.279203	602.279203	602.279203	0	0	1	-360	360;
	4970	2657	0.000331	0.002539	0.22112	541.255	541.255	541.255	0	0	1	-360	360;
	2861	2647	0.0016	0.02113	0.39058	1980.27264	1980.27264	1980.27264	0	0	1	-360	360;
	7048	4831	0.000909	0.004711	0.08412	469.242805	469.242805	469.242805	0	0	1	-360	360;
	7048	4831	0.00088	0.004969	0.08209	433.9769	433.9769	433.9769	0	0	1	-360	360;
	2	7870	0.001841	0.0088	0.01539	325.213843	325.213843	325.213843	0	0	1	-360	360;
	2	7423	0.01013	0.04238	0.07841	498.251088	498.251088	498.251088	0	0	1	-360	360;
	6588	2667	0.00059	0.00802	0.26604	825.882736	825.882736	825.882736	0	0	1	-360	360;
	6588	2667	0.00054	0.0079	0.24432	784.505636	784.505636	784.505636	0	0	1	-360	360;
	6588	3569	0.00171	0.02128	0.77791	1111.91101	1111.91101	1111.91101	0	0	1	-360	360;
	6588	3569	0.00174	0.02084	0.79287	975.494133	975.494133	975.494133	0	0	1	-360	360;
	6588	5968	0.00086	0.01433	0.53318	1265.56626	1265.56626	1265.56626	0	0	1	-360	360;
	6588	5968	0.00087	0.01404	0.54364	845.928903	845.928903	845.928903	0	0	1	-360	360;
	6588	3485	0.00126	0.01447	0.5739	1225.45549	1225.45549	1225.45549	0	0	1	-360	360;
	6588	5126	0.00122	0.01509	0.55265	1173.8674	1173.8674	1173.8674	0	0	1	-360	360;
	6588	5126	0.00154	0.019	0.37428	1384.1807	1384.1807	1384.1807	0	0	1	-360	360;
	4474	4464	0.00474	0.03564	0.05389	584.214329	584.214329	584.214329	0	0	1	-360	360;
	4474	5404	0.006	0.047849	0.07097	544.097261	544.097261	544.097261	0	0	1	-360	360;
	4474	4758	0.008039	0.03375	0.09851	398.687369	398.687369	398.687369	0	0	1	-360	360;
	6217	6626	2.1e-05	0.000211	0	614.640122	614.640122	614.640122	0	0	1	-360	360;
	7534	3347	0.00045	0.002341	0.0364	298.342114	298.342114	298.342114	0	0	1	-360	360;
	2509	7987	0.000669	0.004281	0.00705	537.199091	537.199091	537.199091	0	0	1	-360	360;
	8387	7987	0.00068	0.00419	0.00719	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2804	6463	0.000729	0.00431	0.00836	619.478329	619.478329	619.478329	0	0	1	-360	360;
	2804	6463	0.000591	0.004651	0.00697	569.23635	569.23635	569.23635	0	0	1	-360	360;
	2804	8054	0.00099	0.005081	0.00637	451.722072	451.722072	451.722072	0	0	1	-360	360;
	1427	6878	0.00212	0.013031	0.02462	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1427	3220	0.002159	0.01275	0.02509	246.765313	246.765313	246.765313	0	0	1	-360	360;
	7874	3363	0.00025	0.00095	0.20339	278.753035	278.753035	278.753035	0	0	1	-360	360;
	7874	3363	0.000271	0.00089	0.1299	288.927864	288.927864	288.927864	0	0	1	-360	360;
	7874	2537	0.001961	0.012409	0.02493	392.856143	392.856143	392.856143	0	0	1	-360	360;
	7874	7548	0.00044	0.00207	0.12342	572.216551	572.216551	572.216551	0	0	1	-360	360;
	7874	7716	0.000469	0.00143	0.19912	640.334403	640.334403	640.334403	0	0	1	-360	360;
	7874	5958	0.001271	0.01406	0.03741	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7874	5958	0.00124	0.013021	0.03746	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7874	8150	0.000409	0.00239	0.22086	364.053628	364.053628	364.053628	0	0	1	-360	360;
	7874	195	0.00094	0.005	0.02509	457.229405	457.229405	457.229405	0	0	1	-360	360;
	7874	2690	0.000219	0.0007	0.08927	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7874	6144	6e-05	0.00025	0.04364	276.368417	276.368417	276.368417	0	0	1	-360	360;
	7874	4797	5e-05	0.000229	0.04008	253.611279	253.611279	253.611279	0	0	1	-360	360;
	3430	2493	0.00019	0.000909	0.00107	357.480495	357.480495	357.480495	0	0	1	-360	360;
	2071	5034	0.001711	0.00782	0.01092	343.350355	343.350355	343.350355	0	0	1	-360	360;
	2071	6183	0.002651	0.01176	0.01876	437.907062	437.907062	437.907062	0	0	1	-360	360;
	7045	400	0.001409	0.010599	0.0158	229.347083	229.347083	229.347083	0	0	1	-360	360;
	7045	4847	0.001419	0.010219	0.01613	371.786299	371.786299	371.786299	0	0	1	-360	360;
	759	3569	0.00013	0.00212	0.05846	871.539976	871.539976	871.539976	0	0	1	-360	360;
	759	734	0.00026	0.00313	0.12476	1412.07434	1412.07434	1412.07434	0	0	1	-360	360;
	4102	2079	0.0047	0.0318	0.04833	479.263308	479.263308	479.263308	0	0	1	-360	360;
	4102	6170	0.00136	0.00768	0.01491	445.006045	445.006045	445.006045	0	0	1	-360	360;
	4102	3200	0.005669	0.027789	0.04732	414.465555	414.465555	414.465555	0	0	1	-360	360;
	6594	5690	0.004409	0.033091	0.05018	386.592424	386.592424	386.592424	0	0	1	-360	360;
	6594	1306	0.01056	0.044469	0.07696	252.950155	252.950155	252.950155	0	0	1	-360	360;
	6594	6483	0.01963	0.07736	0.14113	441.642125	441.642125	441.642125	0	0	1	-360	360;
	6594	6741	0.01157	0.04694	0.08944	467.568848	467.568848	467.568848	0	0	1	-360	360;
	5540	787	0.000479	0.00255	0.00356	459.609451	459.609451	459.609451	0	0	1	-360	360;
	6877	787	0.000469	0.00236	0.00348	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1081	1403	0.01237	0.067039	0.08816	389.154231	389.154231	389.154231	0	0	1	-360	360;
	1081	7720	0.0132	0.056229	0.09362	392.730777	392.730777	392.730777	0	0	1	-360	360;
	3619	3719	0.00026	0.001711	0.00302	346.036004	346.036004	346.036004	0	0	1	-360	360;
	3619	3719	0.000271	0.001669	0.00307	351.000339	351.000339	351.000339	0	0	1	-360	360;
	6210	4852	0.000211	0.000401	0.02316	295.64732	295.64732	295.64732	0	0	1	-360	360;
	7520	5276	0.00105	0.005159	0.00759	297.145995	297.145995	297.145995	0	0	1	-360	360;
	6454	6175	0.00212	0.00931	0.01208	428.374686	428.374686	428.374686	0	0	1	-360	360;
	561	6160	0.000151	0.00068	0.05924	377.209801	377.209801	377.209801	0	0	1	-360	360;
	4919	4465	0.00019	0.00082	0.07926	284.535106	284.535106	284.535106	0	0	1	-360	360;
	4656	6183	0.003219	0.01405	0.02045	402.455584	402.455584	402.455584	0	0	1	-360	360;
	6069	8309	0.007271	0.033159	0.04367	386.889263	386.889263	386.889263	0	0	1	-360	360;
	6069	7290	0.006659	0.03132	0.04937	415.800378	415.800378	415.800378	0	0	1	-360	360;
	3130	2579	0.00057	0.00668	0.14718	1581.85827	1581.85827	1581.85827	0	0	1	-360	360;
	3130	6400	0.00057	0.00758	0.14879	1516.30361	1516.30361	1516.30361	0	0	1	-360	360;
	3130	6400	0.00039	0.00607	0.904	952.215647	952.215647	952.215647	0	0	1	-360	360;
	3130	1496	0.00155	0.0179	0.48657	2132.72671	2132.72671	2132.72671	0	0	1	-360	360;
	3130	4180	0.00036	0.00485	0.10754	1652.5928	1652.5928	1652.5928	0	0	1	-360	360;
	5549	1543	0.000581	0.004229	0.00651	445.6565	445.6565	445.6565	0	0	1	-360	360;
	5549	7424	0.00056	0.00431	0.00637	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5549	5296	0.003099	0.020531	0.03456	582.983915	582.983915	582.983915	0	0	1	-360	360;
	5549	2904	0.00118	0.016919	0.02406	619.811749	619.811749	619.811749	0	0	1	-360	360;
	5549	4814	0.00188	0.01407	0.02137	278.789235	278.789235	278.789235	0	0	1	-360	360;
	5549	4449	0.00545	0.04082	0.0826	470.621449	470.621449	470.621449	0	0	1	-360	360;
	6585	400	0.00214	0.01639	0.02509	571.156848	571.156848	571.156848	0	0	1	-360	360;
	6585	4318	0.002479	0.01549	0.02818	279.059401	279.059401	279.059401	0	0	1	-360	360;
	7242	8183	0.000159	0.001229	0.10737	577.843915	577.843915	577.843915	0	0	1	-360	360;
	7242	6054	0.00026	0.00162	0.16913	556.787408	556.787408	556.787408	0	0	1	-360	360;
	7242	6268	0.000159	0.00113	0.09293	477.695283	477.695283	477.695283	0	0	1	-360	360;
	7242	8016	0.000461	0.003219	0.28994	581.150296	581.150296	581.150296	0	0	1	-360	360;
	7242	3814	0.000159	0.001229	0.10129	306.802975	306.802975	306.802975	0	0	1	-360	360;
	1462	454	0.000229	0.00139	0.00313	490.444493	490.444493	490.444493	0	0	1	-360	360;
	2573	454	0.0002	0.00156	0.00182	448.297946	448.297946	448.297946	0	0	1	-360	360;
	1232	464	0.00145	0.00824	0.01704	384.616292	384.616292	384.616292	0	0	1	-360	360;
	1232	454	0.00069	0.005	0.00759	608.363828	608.363828	608.363828	0	0	1	-360	360;
	806	3362	0.00036	0.00455	0.12014	1915.50187	1915.50187	1915.50187	0	0	1	-360	360;
	806	3362	0.00036	0.00453	0.12014	1890.26596	1890.26596	1890.26596	0	0	1	-360	360;
	7535	6841	0.00066	0.00997	0.29862	1462.66134	1462.66134	1462.66134	0	0	1	-360	360;
	7535	6841	0.00064	0.01016	0.29227	1094.77399	1094.77399	1094.77399	0	0	1	-360	360;
	5720	4390	9.9e-05	0.00044	0.04269	634.586246	634.586246	634.586246	0	0	1	-360	360;
	4390	4024	6e-05	0.000469	0.00093	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4390	1463	3.9e-05	0.000341	0.00093	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4390	6303	0.002531	0.01943	0.06916	547.744683	547.744683	547.744683	0	0	1	-360	360;
	4390	6463	0.00945	0.06418	0.11059	527.278805	527.278805	527.278805	0	0	1	-360	360;
	8282	196	0.000591	0.002831	0.00379	467.647725	467.647725	467.647725	0	0	1	-360	360;
	6360	196	0.00087	0.00387	0.00593	215.929128	215.929128	215.929128	0	0	1	-360	360;
	653	4860	0.0015	0.01785	0.25472	1295.82737	1295.82737	1295.82737	0	0	1	-360	360;
	653	4860	0.00147	0.0182	0.24952	1271.23713	1271.23713	1271.23713	0	0	1	-360	360;
	1561	578	0.00151	0.0113	0.01672	488.702327	488.702327	488.702327	0	0	1	-360	360;
	1561	578	0.001479	0.011521	0.01637	506.530949	506.530949	506.530949	0	0	1	-360	360;
	1561	8283	0.00174	0.010341	0.02091	532.042326	532.042326	532.042326	0	0	1	-360	360;
	1561	8283	0.001659	0.01206	0.01949	558.33219	558.33219	558.33219	0	0	1	-360	360;
	7440	1824	0.00069	0.000479	0.05645	230.537487	230.537487	230.537487	0	0	1	-360	360;
	7440	7235	0.00044	0.00139	0.27226	357.169938	357.169938	357.169938	0	0	1	-360	360;
	943	7414	0.002039	0.01601	0.02457	314.629004	314.629004	314.629004	0	0	1	-360	360;
	943	6769	0.01349	0.05662	0.09953	463.637543	463.637543	463.637543	0	0	1	-360	360;
	943	277	0.00082	0.00438	0.00637	400.643305	400.643305	400.643305	0	0	1	-360	360;
	943	277	0.0008	0.004461	0.00624	384.346889	384.346889	384.346889	0	0	1	-360	360;
	943	4644	0.001091	0.00888	0.01336	554.856241	554.856241	554.856241	0	0	1	-360	360;
	943	4644	0.00111	0.007039	0.01308	601.67257	601.67257	601.67257	0	0	1	-360	360;
	2287	7673	0.000849	0.00537	0.01283	545.608129	545.608129	545.608129	0	0	1	-360	360;
	7863	8074	0.00061	0.00275	0.25927	474.049385	474.049385	474.049385	0	0	1	-360	360;
	5631	6976	0.00141	0.0188	0.36135	1445.26632	1445.26632	1445.26632	0	0	1	-360	360;
	5631	6006	0.00121	0.01427	0.30849	1673.46762	1673.46762	1673.46762	0	0	1	-360	360;
	2946	7393	0.001711	0.007159	0.01232	456.156746	456.156746	456.156746	0	0	1	-360	360;
	2946	7979	0.00545	0.02381	0.03439	403.801076	403.801076	403.801076	0	0	1	-360	360;
	2946	3772	0.000771	0.004331	0.01597	332.685495	332.685495	332.685495	0	0	1	-360	360;
	2946	2533	0.00414	0.03282	0.04631	554.014499	554.014499	554.014499	0	0	1	-360	360;
	2946	5742	0.00264	0.011521	0.01913	450.519094	450.519094	450.519094	0	0	1	-360	360;
	2946	5793	0.00261	0.011	0.02509	483.29483	483.29483	483.29483	0	0	1	-360	360;
	2946	6824	0.00239	0.00919	0.01704	421.977217	421.977217	421.977217	0	0	1	-360	360;
	2946	2618	0.000539	0.00226	0.00732	379.86649	379.86649	379.86649	0	0	1	-360	360;
	2946	6249	0.00325	0.017979	0.03513	584.910509	584.910509	584.910509	0	0	1	-360	360;
	1134	4131	0.00076	0.0071	0.18714	1328.51914	1328.51914	1328.51914	0	0	1	-360	360;
	743	6824	0.000651	0.00343	0.00436	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2094	3964	0.001219	0.00812	0.0158	453.481004	453.481004	453.481004	0	0	1	-360	360;
	2094	4381	0.00243	0.01439	0.04077	600.419292	600.419292	600.419292	0	0	1	-360	360;
	8325	5668	0.0006	0.00839	0.63239	1277.47601	1277.47601	1277.47601	0	0	1	-360	360;
	8325	5668	0.00059	0.00856	0.61949	1217.21707	1217.21707	1217.21707	0	0	1	-360	360;
	8325	5135	0.00087	0.01328	0.40029	1778.51172	1778.51172	1778.51172	0	0	1	-360	360;
	8325	5135	0.00073	0.01595	0.33479	1366.67445	1366.67445	1366.67445	0	0	1	-360	360;
	2653	3315	0.00163	0.002039	0.17114	333.829792	333.829792	333.829792	0	0	1	-360	360;
	2653	3315	0.001659	0.002	0.17443	333.026155	333.026155	333.026155	0	0	1	-360	360;
	50	5231	0.00339	0.019669	0.03869	577.18965	577.18965	577.18965	0	0	1	-360	360;
	50	2144	0.00239	0.017281	0.02751	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5905	8203	0.00032	0.00112	0.1323	502.03988	502.03988	502.03988	0	0	1	-360	360;
	5905	5369	0.0018	0.01401	0.02092	507.550261	507.550261	507.550261	0	0	1	-360	360;
	5905	5369	0.00212	0.012469	0.02509	548.880596	548.880596	548.880596	0	0	1	-360	360;
	1160	2581	0.00063	0.00411	0.00692	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5828	3740	0.009729	0.03505	0.05018	312.522934	312.522934	312.522934	0	0	1	-360	360;
	5828	3237	0.01007	0.03056	0.04174	307.175262	307.175262	307.175262	0	0	1	-360	360;
	1607	2550	0.00042	0.00623	0.262	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	1607	317	0.00036	0.00722	0.22052	1864.82996	1864.82996	1864.82996	0	0	1	-360	360;
	1607	5968	0.00053	0.0078	0.3275	959.934774	959.934774	959.934774	0	0	1	-360	360;
	1607	1760	0.00109	0.0173	0.4942	832.524423	832.524423	832.524423	0	0	1	-360	360;
	1607	5297	0.00071	0.01144	0.2579	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	1607	3485	0.00157	0.01884	0.71634	1260.33373	1260.33373	1260.33373	0	0	1	-360	360;
	1607	3254	0.00032	0.00384	0.14694	857.689909	857.689909	857.689909	0	0	1	-360	360;
	7285	1642	0.00041	0.00649	0.24929	1565.71774	1565.71774	1565.71774	0	0	1	-360	360;
	7285	7081	0.0004	0.00658	0.24449	2196.77019	2196.77019	2196.77019	0	0	1	-360	360;
	7285	5968	0.00054	0.00763	0.33356	974.889924	974.889924	974.889924	0	0	1	-360	360;
	7285	1760	0.00107	0.01602	0.48345	929.910608	929.910608	929.910608	0	0	1	-360	360;
	7285	5297	0.0007	0.01165	0.25241	2195.72171	2195.72171	2195.72171	0	0	1	-360	360;
	7285	5126	0.0016	0.01844	0.72985	1350.24892	1350.24892	1350.24892	0	0	1	-360	360;
	7285	3254	0.00032	0.00376	0.14971	956.064021	956.064021	956.064021	0	0	1	-360	360;
	4197	151	0.00212	0.01207	0.02395	497.713806	497.713806	497.713806	0	0	1	-360	360;
	2856	2127	5e-05	0.00037	0.0006	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5690	7973	0.00355	0.024651	0.0389	433.725787	433.725787	433.725787	0	0	1	-360	360;
	5674	8290	0.00675	0.04145	0.07349	536.098997	536.098997	536.098997	0	0	1	-360	360;
	5023	6471	0.00195	0.014341	0.02547	456.865501	456.865501	456.865501	0	0	1	-360	360;
	7578	3966	0.00018	0.00043	0.0273	385.686284	385.686284	385.686284	0	0	1	-360	360;
	7578	2657	0.00037	0.000969	0.22368	458.323022	458.323022	458.323022	0	0	1	-360	360;
	552	1221	0.00257	0.017169	0.02962	492.175989	492.175989	492.175989	0	0	1	-360	360;
	552	4920	0.00076	0.00399	0.00546	510.629917	510.629917	510.629917	0	0	1	-360	360;
	552	4920	0.00075	0.00406	0.00535	496.374411	496.374411	496.374411	0	0	1	-360	360;
	552	1897	0.003901	0.023961	0.04412	586.979617	586.979617	586.979617	0	0	1	-360	360;
	552	701	0.00249	0.019211	0.02875	458.231569	458.231569	458.231569	0	0	1	-360	360;
	3323	2657	0.00025	0.001159	0.10555	490.205574	490.205574	490.205574	0	0	1	-360	360;
	1233	6601	0.00195	0.02924	0.88633	1991.76182	1991.76182	1991.76182	0	0	1	-360	360;
	1233	2000	0.00152	0.01527	0.38572	1513.14304	1513.14304	1513.14304	0	0	1	-360	360;
	1233	6933	0.00149	0.01558	0.37844	1632.93891	1632.93891	1632.93891	0	0	1	-360	360;
	1233	986	0.00253	0.0276	0.64512	1386.80684	1386.80684	1386.80684	0	0	1	-360	360;
	1233	986	0.00263	0.02648	0.67042	1582.07416	1582.07416	1582.07416	0	0	1	-360	360;
	1233	2143	0.00077	0.0092	0.34898	1426.81756	1426.81756	1426.81756	0	0	1	-360	360;
	1233	3823	0.00112	0.01705	0.50766	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5311	7748	0.00226	0.01101	0.01547	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5311	5434	0.004531	0.029039	0.05939	514.207606	514.207606	514.207606	0	0	1	-360	360;
	5311	7261	0.002081	0.030409	0.30294	267.782953	267.782953	267.782953	0	0	1	-360	360;
	5311	3576	0.0053	0.03257	0.05951	450.112893	450.112893	450.112893	0	0	1	-360	360;
	5311	3018	0.00513	0.041901	0.05878	557.505309	557.505309	557.505309	0	0	1	-360	360;
	5311	4319	0.001461	0.01056	0.01707	471.822522	471.822522	471.822522	0	0	1	-360	360;
	7491	4381	0.0005	0.002591	0.01692	371.718091	371.718091	371.718091	0	0	1	-360	360;
	5580	5727	0.00055	0.0033	0.00616	463.255348	463.255348	463.255348	0	0	1	-360	360;
	3648	5925	0.002289	0.01763	0	494.234428	494.234428	494.234428	0	0	1	-360	360;
	3648	1983	0.0075	0.043531	0	469.966803	469.966803	469.966803	0	0	1	-360	360;
	5195	1897	0.00138	0.019961	0.07928	270.022772	270.022772	270.022772	0	0	1	-360	360;
	5195	715	0.00126	0.018729	0.07188	605.776491	605.776491	605.776491	0	0	1	-360	360;
	2114	8338	0.00038	0.001331	0.161	512.857542	512.857542	512.857542	0	0	1	-360	360;
	2114	3090	0.00019	0.001461	0.12342	215.929128	215.929128	215.929128	0	0	1	-360	360;
	186	4318	0.002151	0.013401	0.0182	481.394527	481.394527	481.394527	0	0	1	-360	360;
	186	7066	0.000789	0.005669	0.00854	581.834283	581.834283	581.834283	0	0	1	-360	360;
	2916	6109	3e-05	0.00042	0.00797	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2916	603	0.00079	0.00899	0.35869	1386.46919	1386.46919	1386.46919	0	0	1	-360	360;
	2916	603	0.0008	0.00879	0.36533	1579.50923	1579.50923	1579.50923	0	0	1	-360	360;
	2916	4039	0.00334	0.03393	0.96124	1384.39658	1384.39658	1384.39658	0	0	1	-360	360;
	2916	1498	0.00063	0.00789	0.28834	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2916	3164	0.0006	0.00842	0.27454	1326.31029	1326.31029	1326.31029	0	0	1	-360	360;
	4256	7246	0.00011	0.00125	0.00546	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4256	5911	0.00012	0.001091	0.00604	367.910247	367.910247	367.910247	0	0	1	-360	360;
	4256	6982	9.9e-05	0.00125	0.00372	363.490434	363.490434	363.490434	0	0	1	-360	360;
	4256	374	0.00294	0.02443	0.03473	637.16863	637.16863	637.16863	0	0	1	-360	360;
	4256	2578	0.00144	0.011151	0.01637	507.419942	507.419942	507.419942	0	0	1	-360	360;
	4256	2321	0.001831	0.010849	0.02013	609.471163	609.471163	609.471163	0	0	1	-360	360;
	4256	2321	0.001599	0.01055	0.01777	576.22521	576.22521	576.22521	0	0	1	-360	360;
	4256	2723	0.00144	0.019591	0.04186	395.572657	395.572657	395.572657	0	0	1	-360	360;
	5165	910	0.0133	0.09551	0.14988	614.413016	614.413016	614.413016	0	0	1	-360	360;
	5165	4031	0.009591	0.073961	0.10803	531.827413	531.827413	531.827413	0	0	1	-360	360;
	4860	2467	0.0007	0.00671	0.0537	1578.30345	1578.30345	1578.30345	0	0	1	-360	360;
	4860	2467	0.00072	0.00658	0.05476	1608.90154	1608.90154	1608.90154	0	0	1	-360	360;
	2145	4490	0.001901	0.00964	0.26659	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2145	4711	0.001711	0.01189	0.23511	613.612808	613.612808	613.612808	0	0	1	-360	360;
	1867	4319	0.00025	0.00187	0.0027	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3824	5012	0.00043	0.002211	0.01278	225.206962	225.206962	225.206962	0	0	1	-360	360;
	5173	7864	0.000401	0.00243	0.01185	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7082	4648	0.00336	0.017	0.02562	460.561316	460.561316	460.561316	0	0	1	-360	360;
	6548	4381	0.005521	0.031031	0.10009	262.687918	262.687918	262.687918	0	0	1	-360	360;
	4009	3474	9.9e-05	0.00026	0.02045	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6220	2478	0.00045	0.00718	0.20724	1116.51234	1116.51234	1116.51234	0	0	1	-360	360;
	6220	4140	0.00044	0.00665	0.20273	949.136684	949.136684	949.136684	0	0	1	-360	360;
	6220	3951	0.00103	0.01516	0.47132	1314.72501	1314.72501	1314.72501	0	0	1	-360	360;
	6220	6814	0.00101	0.01546	0.4615	983.090838	983.090838	983.090838	0	0	1	-360	360;
	4563	21	0.00045	0.00608	0.1116	1516.1937	1516.1937	1516.1937	0	0	1	-360	360;
	4563	8361	0.0004	0.00501	0.10045	1373.0397	1373.0397	1373.0397	0	0	1	-360	360;
	5547	8174	0.001729	0.01337	0.02702	598.324654	598.324654	598.324654	0	0	1	-360	360;
	5547	5654	0.00264	0.01489	0.04259	520.051407	520.051407	520.051407	0	0	1	-360	360;
	5547	798	0.002341	0.018961	0.02582	524.119891	524.119891	524.119891	0	0	1	-360	360;
	5547	5209	0.001841	0.01363	0.06215	479.954154	479.954154	479.954154	0	0	1	-360	360;
	8041	6985	0.001219	0.00949	0.28237	506.666223	506.666223	506.666223	0	0	1	-360	360;
	8041	8174	0.002271	0.01151	0.02881	501.392474	501.392474	501.392474	0	0	1	-360	360;
	8041	5437	0.010781	0.05775	0.0906	273.24837	273.24837	273.24837	0	0	1	-360	360;
	8041	5852	0.00251	0.013919	0.0187	408.129055	408.129055	408.129055	0	0	1	-360	360;
	8041	7010	0.00101	0.007331	0.02277	522.067549	522.067549	522.067549	0	0	1	-360	360;
	2256	3740	0.00061	0.003909	0.01185	635.597556	635.597556	635.597556	0	0	1	-360	360;
	3630	5786	0.004289	0.019331	0.04391	370.029653	370.029653	370.029653	0	0	1	-360	360;
	3630	3740	0.0008	0.004401	0.0098	638.189847	638.189847	638.189847	0	0	1	-360	360;
	3630	4393	0.008099	0.03519	0.06156	617.404648	617.404648	617.404648	0	0	1	-360	360;
	3630	7128	0.00063	0.004591	0.00854	520.749874	520.749874	520.749874	0	0	1	-360	360;
	2542	7128	0.000651	0.00456	0.00789	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2256	8184	0.00114	0.00581	0.01044	317.036485	317.036485	317.036485	0	0	1	-360	360;
	2256	2033	0.001211	0.00538	0.00906	215.929128	215.929128	215.929128	0	0	1	-360	360;
	595	3453	0.00418	0.030159	0.04743	582.277446	582.277446	582.277446	0	0	1	-360	360;
	595	7702	0.00324	0.02413	0.03531	621.966212	621.966212	621.966212	0	0	1	-360	360;
	595	7702	0.003169	0.024591	0.03457	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3182	8174	0.000219	0.00088	0.09293	508.315031	508.315031	508.315031	0	0	1	-360	360;
	3182	8174	0.000289	0.00095	0.12545	526.658835	526.658835	526.658835	0	0	1	-360	360;
	5727	37	0.0035	0.022211	0.04027	557.795289	557.795289	557.795289	0	0	1	-360	360;
	5727	6878	0.00313	0.02256	0.03439	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5391	7991	0.000831	0.00162	0.19258	362.505036	362.505036	362.505036	0	0	1	-360	360;
	2896	7720	0.002099	0.01068	0.0158	400.868506	400.868506	400.868506	0	0	1	-360	360;
	914	7720	0.00224	0.00976	0.01611	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7973	1316	0.00169	0.01231	0.03531	637.28104	637.28104	637.28104	0	0	1	-360	360;
	6018	8220	0.00011	0.00076	0.00616	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8265	8220	0.00011	0.00068	0.00627	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6312	5786	0.005479	0.021909	0.04153	358.18201	358.18201	358.18201	0	0	1	-360	360;
	6312	8000	0.003331	0.017039	0.02416	447.086585	447.086585	447.086585	0	0	1	-360	360;
	4972	3964	0.000479	0.001271	0.08906	326.475885	326.475885	326.475885	0	0	1	-360	360;
	6774	4765	0.00109	0.01629	0.693	852.85624	852.85624	852.85624	0	0	1	-360	360;
	6774	7433	4e-05	0.00063	0.02356	1825.93617	1825.93617	1825.93617	0	0	1	-360	360;
	495	5706	0.0053	0.024531	0.04347	587.628928	587.628928	587.628928	0	0	1	-360	360;
	495	2082	0.004659	0.021919	0.03899	436.834022	436.834022	436.834022	0	0	1	-360	360;
	495	3325	0.001659	0.01043	0.02052	579.903116	579.903116	579.903116	0	0	1	-360	360;
	4550	3773	0.006961	0.053539	0.08688	513.689377	513.689377	513.689377	0	0	1	-360	360;
	1727	4936	0.00036	0.00288	0.00465	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1727	4936	0.00038	0.003	0.00455	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4981	4522	0.00048	0.00643	0.16791	2139.01364	2139.01364	2139.01364	0	0	1	-360	360;
	4981	4522	0.00051	0.0066	0.18021	2128.0306	2128.0306	2128.0306	0	0	1	-360	360;
	4981	4522	0.00046	0.00729	0.16288	1880.02206	1880.02206	1880.02206	0	0	1	-360	360;
	4981	4955	0.0006	0.00868	0.19685	1842.04972	1842.04972	1842.04972	0	0	1	-360	360;
	4981	7056	0.00068	0.01076	0.24161	1957.4687	1957.4687	1957.4687	0	0	1	-360	360;
	4981	6257	0.00066	0.01096	0.23647	1936.00152	1936.00152	1936.00152	0	0	1	-360	360;
	850	4151	0.00101	0.005099	0.00651	412.391875	412.391875	412.391875	0	0	1	-360	360;
	509	7681	0.003901	0.02437	0.04617	244.344495	244.344495	244.344495	0	0	1	-360	360;
	509	2051	0.00276	0.016461	0.03136	567.99298	567.99298	567.99298	0	0	1	-360	360;
	509	7172	0.00588	0.047081	0.07007	302.721917	302.721917	302.721917	0	0	1	-360	360;
	3943	1698	3.9e-05	0.000219	0	309.975226	309.975226	309.975226	0	0	1	-360	360;
	2607	3341	0.00131	0.00538	0.00923	372.542686	372.542686	372.542686	0	0	1	-360	360;
	2585	986	0.00679	0.04679	1.05162	1020.25296	1020.25296	1020.25296	0	0	1	-360	360;
	2585	7493	0.00138	0.01466	0.33876	1089.13932	1089.13932	1089.13932	0	0	1	-360	360;
	8268	8327	0.003531	0.01724	0.02562	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8268	1294	0.01324	0.06936	0.0967	493.246362	493.246362	493.246362	0	0	1	-360	360;
	8268	7720	0.00363	0.01743	0.02562	446.307335	446.307335	446.307335	0	0	1	-360	360;
	8268	7720	0.00257	0.01662	0.02863	533.502514	533.502514	533.502514	0	0	1	-360	360;
	8268	4595	0.01399	0.056289	0.1035	436.503269	436.503269	436.503269	0	0	1	-360	360;
	8268	958	0.00343	0.01751	0.02509	459.191437	459.191437	459.191437	0	0	1	-360	360;
	8268	1271	0.00301	0.015409	0.02178	325.259569	325.259569	325.259569	0	0	1	-360	360;
	8268	1444	0.00857	0.033091	0.06282	483.864882	483.864882	483.864882	0	0	1	-360	360;
	5597	6250	0.003341	0.01689	0.02366	353.368191	353.368191	353.368191	0	0	1	-360	360;
	6944	2745	3.9e-05	0.000219	0	380.255162	380.255162	380.255162	0	0	1	-360	360;
	6656	6438	0.00087	0.00714	0.01336	466.796076	466.796076	466.796076	0	0	1	-360	360;
	4273	7493	0.00011	0.001	0.06775	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1524	4873	0.000219	0.00101	0.0019	345.435467	345.435467	345.435467	0	0	1	-360	360;
	3532	4873	0.000219	0.001031	0.00186	345.389741	345.389741	345.389741	0	0	1	-360	360;
	6028	4873	0.000271	0.00101	0.00206	332.11087	332.11087	332.11087	0	0	1	-360	360;
	3284	4873	0.000271	0.000979	0.0021	326.313176	326.313176	326.313176	0	0	1	-360	360;
	2520	4873	0.000289	0.00107	0.00206	342.432784	342.432784	342.432784	0	0	1	-360	360;
	1687	4873	0.000289	0.00105	0.0021	347.149435	347.149435	347.149435	0	0	1	-360	360;
	3964	5493	0.001591	0.012289	0.0182	381.91845	381.91845	381.91845	0	0	1	-360	360;
	1973	4762	0.000271	0.001901	0.00284	507.504154	507.504154	507.504154	0	0	1	-360	360;
	1973	2615	0.00186	0.010711	0.0213	515.012387	515.012387	515.012387	0	0	1	-360	360;
	951	3193	1e-05	0.0002	0.00566	2097.1791	2097.1791	2097.1791	0	0	1	-360	360;
	3689	798	0.000169	0.000271	0.01777	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4113	798	0.000169	0.00031	0.0187	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1554	5276	0.00526	0.03787	0.05976	553.117885	553.117885	553.117885	0	0	1	-360	360;
	7350	1352	0.000229	0.000599	0.06709	371.510418	371.510418	371.510418	0	0	1	-360	360;
	5141	5012	0.00089	0.004091	0.01086	446.509673	446.509673	446.509673	0	0	1	-360	360;
	4078	7864	0.00093	0.004	0.01007	415.857154	415.857154	415.857154	0	0	1	-360	360;
	7979	6824	0.00289	0.013099	0.02074	405.591255	405.591255	405.591255	0	0	1	-360	360;
	2123	5832	0.002599	0.016479	0.02818	543.649907	543.649907	543.649907	0	0	1	-360	360;
	410	7053	0.00369	0.02712	0.04364	557.24124	557.24124	557.24124	0	0	1	-360	360;
	410	3511	0.0062	0.04864	0.07007	647.787002	647.787002	647.787002	0	0	1	-360	360;
	410	300	0.008021	0.059031	0.09486	647.787002	647.787002	647.787002	0	0	1	-360	360;
	410	6295	0.004979	0.0295	0.05645	573.836019	573.836019	573.836019	0	0	1	-360	360;
	7456	3325	0.001159	0.005281	0.00888	303.235574	303.235574	303.235574	0	0	1	-360	360;
	4915	7032	0.00361	0.01824	0.02602	371.390006	371.390006	371.390006	0	0	1	-360	360;
	4915	1773	0.001521	0.01075	0.02372	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4915	3856	0.00868	0.041711	0.06166	377.427762	377.427762	377.427762	0	0	1	-360	360;
	2348	6194	0.01118	0.05986	0.08642	370.230848	370.230848	370.230848	0	0	1	-360	360;
	2348	6868	0.003229	0.02543	0.03746	601.102898	601.102898	601.102898	0	0	1	-360	360;
	2348	6868	0.00218	0.01869	0.06377	336.0296	336.0296	336.0296	0	0	1	-360	360;
	2348	6868	0.001901	0.021909	0.05343	589.012525	589.012525	589.012525	0	0	1	-360	360;
	2348	7251	0.00974	0.043969	0.07405	408.998995	408.998995	408.998995	0	0	1	-360	360;
	2348	6551	0.003461	0.01814	0.02526	464.665238	464.665238	464.665238	0	0	1	-360	360;
	2348	1418	0.00487	0.027669	0.07401	435.802897	435.802897	435.802897	0	0	1	-360	360;
	8183	1052	0.000651	0.003961	0.15054	624.16221	624.16221	624.16221	0	0	1	-360	360;
	7880	4993	0.00169	0.00887	0	409.462353	409.462353	409.462353	0	0	1	-360	360;
	2806	8054	0.001711	0.01336	0.02912	520.834086	520.834086	520.834086	0	0	1	-360	360;
	1429	890	0.00031	0.000539	0.0331	414.850798	414.850798	414.850798	0	0	1	-360	360;
	2915	3025	0.0027	0.01157	0.01913	270.193102	270.193102	270.193102	0	0	1	-360	360;
	3236	578	0.000909	0.006901	0.0546	521.660586	521.660586	521.660586	0	0	1	-360	360;
	3236	4490	0.001729	0.01214	0.0645	528.719941	528.719941	528.719941	0	0	1	-360	360;
	374	3173	0.00614	0.03106	0.05643	453.491674	453.491674	453.491674	0	0	1	-360	360;
	374	7202	0.00418	0.02388	0.04792	515.124797	515.124797	515.124797	0	0	1	-360	360;
	2960	1147	0.00238	0.017771	0.02788	217.997473	217.997473	217.997473	0	0	1	-360	360;
	2960	1649	8.1e-05	0.00063	0.00095	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6683	330	0.001599	0.01618	0.0362	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3872	1944	0.00532	0.026961	0.03985	433.594706	433.594706	433.594706	0	0	1	-360	360;
	3872	1944	0.00361	0.02814	0.04531	558.62941	558.62941	558.62941	0	0	1	-360	360;
	3872	5055	0.002469	0.01025	0.023	538.21726	538.21726	538.21726	0	0	1	-360	360;
	5786	1653	0.005229	0.02675	0.0381	403.025637	403.025637	403.025637	0	0	1	-360	360;
	5786	1653	0.00512	0.02726	0.03731	421.280275	421.280275	421.280275	0	0	1	-360	360;
	5786	7128	0.00355	0.01694	0.03078	403.389922	403.389922	403.389922	0	0	1	-360	360;
	8335	6607	0.00139	0.007659	0.29899	474.404525	474.404525	474.404525	0	0	1	-360	360;
	8335	3643	0.000169	0.000669	0.07109	516.39827	516.39827	516.39827	0	0	1	-360	360;
	7849	2093	0.00111	0.00712	0.01208	490.060774	490.060774	490.060774	0	0	1	-360	360;
	821	5873	0.00031	0.001901	0.0041	536.780697	536.780697	536.780697	0	0	1	-360	360;
	6375	5873	0.000281	0.002091	0.00379	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3830	5944	0.000281	0.001651	0.00836	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6642	5944	0.000271	0.00169	0.00819	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5312	3081	0.001521	0.00843	0.01069	384.486735	384.486735	384.486735	0	0	1	-360	360;
	5312	3423	0.000841	0.0062	0.00871	537.806106	537.806106	537.806106	0	0	1	-360	360;
	3372	7953	0.00019	0.00126	0.00198	517.991445	517.991445	517.991445	0	0	1	-360	360;
	813	6002	0.000219	0.00143	0.00302	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2647	7352	0.00121	0.01796	0.33479	1899.1889	1899.1889	1899.1889	0	0	1	-360	360;
	2647	7352	0.0014	0.01512	0.39301	1985.79674	1985.79674	1985.79674	0	0	1	-360	360;
	5925	1983	0.008409	0.063651	0.10454	365.876196	365.876196	365.876196	0	0	1	-360	360;
	5925	4831	0.00918	0.07344	0.10919	624.859153	624.859153	624.859153	0	0	1	-360	360;
	5925	551	0.004219	0.0345	0.04809	518.067274	518.067274	518.067274	0	0	1	-360	360;
	5925	551	0.0042	0.031651	0.0474	469.815906	469.815906	469.815906	0	0	1	-360	360;
	4090	7951	0.00121	0.01381	0.56889	1572.44894	1572.44894	1572.44894	0	0	1	-360	360;
	4090	6400	0.00107	0.01326	0.54567	1600.15368	1600.15368	1600.15368	0	0	1	-360	360;
	6194	5276	0.003401	0.02607	0.03833	317.336753	317.336753	317.336753	0	0	1	-360	360;
	3337	16	0.000289	0.00138	0.12648	421.657896	421.657896	421.657896	0	0	1	-360	360;
	4669	6987	0.001031	0.006039	0.0115	363.614276	363.614276	363.614276	0	0	1	-360	360;
	1215	6987	0.00105	0.005909	0.01171	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7656	6918	0.001091	0.007901	0.01208	424.945606	424.945606	424.945606	0	0	1	-360	360;
	7656	1029	0.00332	0.024021	0.03795	565.090513	565.090513	565.090513	0	0	1	-360	360;
	7656	2792	0.00344	0.021419	0.03825	414.303228	414.303228	414.303228	0	0	1	-360	360;
	7656	278	0.00095	0.010419	0.02788	519.894795	519.894795	519.894795	0	0	1	-360	360;
	7656	278	0.00093	0.01062	0.02639	476.77352	476.77352	476.77352	0	0	1	-360	360;
	7656	6793	0.00474	0.025409	0.03549	365.289758	365.289758	365.289758	0	0	1	-360	360;
	4258	1905	0.000271	0.000841	0.20316	393.073342	393.073342	393.073342	0	0	1	-360	360;
	6787	5675	0.000271	0.00119	0.12102	392.802415	392.802415	392.802415	0	0	1	-360	360;
	6787	1424	0.000289	0.00076	0.05343	358.512382	358.512382	358.512382	0	0	1	-360	360;
	7054	1498	0.00031	0.00454	0.19026	1952.56197	1952.56197	1952.56197	0	0	1	-360	360;
	5418	3164	0.00032	0.0044	0.19061	2042.14807	2042.14807	2042.14807	0	0	1	-360	360;
	2524	1498	0.00028	0.00494	0.16981	1982.57034	1982.57034	1982.57034	0	0	1	-360	360;
	882	3164	0.00028	0.00508	0.17265	2227.30708	2227.30708	2227.30708	0	0	1	-360	360;
	8086	835	0.000531	0.00093	0.0969	306.723335	306.723335	306.723335	0	0	1	-360	360;
	8086	4471	0.0003	0.001031	0.12778	501.41267	501.41267	501.41267	0	0	1	-360	360;
	2408	8259	0.00139	0.02163	0.64598	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2408	6497	0.00164	0.02321	0.75001	1890.20212	1890.20212	1890.20212	0	0	1	-360	360;
	3199	8259	0.00136	0.02205	0.63253	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3199	6497	0.0016	0.02367	0.73193	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1623	8290	0.002091	0.01174	0.01603	386.824484	386.824484	386.824484	0	0	1	-360	360;
	1623	8290	0.001719	0.00961	0.01679	567.360435	567.360435	567.360435	0	0	1	-360	360;
	5119	934	0.012341	0.076781	0.10241	437.741305	437.741305	437.741305	0	0	1	-360	360;
	5119	8290	0.00176	0.009419	0.01712	607.733189	607.733189	607.733189	0	0	1	-360	360;
	1000	2618	0.00031	0.00182	0.00313	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6688	2478	0.00386	0.02348	0.34569	1284.41322	1284.41322	1284.41322	0	0	1	-360	360;
	6688	4353	0.00017	0.00104	0.01531	1164.70555	1164.70555	1164.70555	0	0	1	-360	360;
	3222	4822	0.00061	0.00808	0.1541	1635.04048	1635.04048	1635.04048	0	0	1	-360	360;
	3222	3485	0.00059	0.0067	0.14151	2032.34778	2032.34778	2032.34778	0	0	1	-360	360;
	2155	935	0.00059	0.00748	0.15075	1495.70853	1495.70853	1495.70853	0	0	1	-360	360;
	2155	5126	0.00061	0.0063	0.15018	2205.9643	2205.9643	2205.9643	0	0	1	-360	360;
	330	1066	0.001031	0.008789	0.0345	647.787002	647.787002	647.787002	0	0	1	-360	360;
	330	8340	0.00051	0.006169	0.01423	647.787002	647.787002	647.787002	0	0	1	-360	360;
	330	7911	0.0005	0.00619	0.01208	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1676	3285	0.00049	0.00376	0.0061	586.801285	586.801285	586.801285	0	0	1	-360	360;
	1676	3285	0.000539	0.003229	0.00627	536.059367	536.059367	536.059367	0	0	1	-360	360;
	1676	3610	0.001031	0.00882	0.03869	372.316342	372.316342	372.316342	0	0	1	-360	360;
	1676	1681	0.000711	0.00463	0.00789	459.750439	459.750439	459.750439	0	0	1	-360	360;
	1676	2032	0.000711	0.004539	0.00805	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3624	2237	0.00206	0.015341	0.02323	515.145754	515.145754	515.145754	0	0	1	-360	360;
	3624	2237	0.002021	0.01564	0.02275	525.266474	525.266474	525.266474	0	0	1	-360	360;
	5546	8034	0.00094	0.005781	0.0115	611.138262	611.138262	611.138262	0	0	1	-360	360;
	5546	910	0.002219	0.016969	0.02602	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6879	1653	0.00238	0.01187	0.02013	396.173194	396.173194	396.173194	0	0	1	-360	360;
	6879	2618	0.00707	0.03095	0.05336	384.746612	384.746612	384.746612	0	0	1	-360	360;
	21	6006	0.00256	0.03401	0.6483	1658.80207	1658.80207	1658.80207	0	0	1	-360	360;
	21	2248	0.00203	0.02416	0.53208	1161.19284	1161.19284	1161.19284	0	0	1	-360	360;
	21	4184	0.001	0.01459	0.45469	2047.42141	2047.42141	2047.42141	0	0	1	-360	360;
	21	3853	0.00103	0.00987	0.262	1625.62457	1625.62457	1625.62457	0	0	1	-360	360;
	21	4676	0.00038	0.00483	0.09617	1476.72729	1476.72729	1476.72729	0	0	1	-360	360;
	21	4676	0.00041	0.00398	0.10605	1335.01997	1335.01997	1335.01997	0	0	1	-360	360;
	6976	2248	0.00167	0.01721	0.48923	1402.08515	1402.08515	1402.08515	0	0	1	-360	360;
	6976	1941	0.00098	0.01487	0.44521	1003.40094	1003.40094	1003.40094	0	0	1	-360	360;
	6976	8361	0.00042	0.00342	0.06654	1354.02029	1354.02029	1354.02029	0	0	1	-360	360;
	6976	8361	0.00019	0.00248	0.08543	1612.09503	1612.09503	1612.09503	0	0	1	-360	360;
	6976	3853	0.00105	0.00966	0.26685	1910.21537	1910.21537	1910.21537	0	0	1	-360	360;
	6976	3193	0.00043	0.00473	0.109	1408.25624	1408.25624	1408.25624	0	0	1	-360	360;
	6976	3193	0.00041	0.00397	0.10605	1330.17446	1330.17446	1330.17446	0	0	1	-360	360;
	1717	3558	0.00051	0.00514	0.12858	1243.09338	1243.09338	1243.09338	0	0	1	-360	360;
	101	3558	0.00053	0.00503	0.13411	1417.78997	1417.78997	1417.78997	0	0	1	-360	360;
	7032	3421	0.00455	0.022771	0.03136	409.944002	409.944002	409.944002	0	0	1	-360	360;
	7032	3421	0.00557	0.02055	0.0394	393.584332	393.584332	393.584332	0	0	1	-360	360;
	7032	3421	0.00339	0.02026	0.05227	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4506	8151	0.00237	0.01068	0.01986	452.121033	452.121033	452.121033	0	0	1	-360	360;
	4506	7423	0.0037	0.019479	0.02701	391.486645	391.486645	391.486645	0	0	1	-360	360;
	5554	7157	0.01407	0.056849	0.07303	372.569741	372.569741	372.569741	0	0	1	-360	360;
	5554	4134	0.001669	0.015409	0.0453	618.888843	618.888843	618.888843	0	0	1	-360	360;
	5554	4134	0.001711	0.015091	0.04617	269.92065	269.92065	269.92065	0	0	1	-360	360;
	7174	3362	0.00172	0.01619	0.44793	1681.09789	1681.09789	1681.09789	0	0	1	-360	360;
	7174	6841	0.00029	0.00447	0.13019	1007.19863	1007.19863	1007.19863	0	0	1	-360	360;
	7174	6841	0.00033	0.00387	0.15	1146.92878	1146.92878	1146.92878	0	0	1	-360	360;
	7174	3770	0.00079	0.01154	0.36042	1008.68085	1008.68085	1008.68085	0	0	1	-360	360;
	7174	3770	0.00077	0.01177	0.35291	1009.14618	1009.14618	1009.14618	0	0	1	-360	360;
	7174	2748	0.00127	0.01498	0.32548	1548.04365	1548.04365	1548.04365	0	0	1	-360	360;
	1617	144	0.00214	0.012599	0.023	505.428568	505.428568	505.428568	0	0	1	-360	360;
	3866	2201	0.000159	0.00125	0.00178	462.409034	462.409034	462.409034	0	0	1	-360	360;
	3363	4811	0.000211	0.000669	0.09264	464.493765	464.493765	464.493765	0	0	1	-360	360;
	680	7548	0.000341	0.001169	0.14397	526.245394	526.245394	526.245394	0	0	1	-360	360;
	680	786	0.00024	0.001469	0.15805	590.382023	590.382023	590.382023	0	0	1	-360	360;
	680	2690	0.00018	0.000669	0.07063	542.037679	542.037679	542.037679	0	0	1	-360	360;
	3180	7245	0.00033	0.00683	0.16895	952.97716	952.97716	952.97716	0	0	1	-360	360;
	3180	2000	0.00121	0.01211	0.30613	1422.26888	1422.26888	1422.26888	0	0	1	-360	360;
	3180	6933	0.00123	0.01185	0.3119	1476.85761	1476.85761	1476.85761	0	0	1	-360	360;
	3180	3853	0.00058	0.0097	0.3	1204.93083	1204.93083	1204.93083	0	0	1	-360	360;
	3180	4676	0.00103	0.01292	0.25992	1266.52062	1266.52062	1266.52062	0	0	1	-360	360;
	3180	3193	0.00116	0.01266	0.29458	947.275353	947.275353	947.275353	0	0	1	-360	360;
	6607	3643	0.0013	0.006841	0.23104	474.73642	474.73642	474.73642	0	0	1	-360	360;
	2744	958	6e-05	0.000349	0.00091	456.894461	456.894461	456.894461	0	0	1	-360	360;
	2450	3121	0.000169	0.0007	0.12981	407.520517	407.520517	407.520517	0	0	1	-360	360;
	8284	1844	0.000169	0.000719	0.13243	298.105482	298.105482	298.105482	0	0	1	-360	360;
	5722	6061	0.000271	0.001031	0.00095	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4395	3347	0.008271	0.036841	0.04987	335.54757	335.54757	335.54757	0	0	1	-360	360;
	4913	2493	0.002969	0.02299	0.06006	436.880129	436.880129	436.880129	0	0	1	-360	360;
	2803	4205	0.00124	0.01273	0.30035	1220.38291	1220.38291	1220.38291	0	0	1	-360	360;
	2803	3951	0.00106	0.00989	0.26367	1507.9013	1507.9013	1507.9013	0	0	1	-360	360;
	2803	4805	7e-05	0.00029	0.06127	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2803	4922	9e-05	0.00079	0.01837	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2803	5874	0.00196	0.02049	0.48657	1638.5387	1638.5387	1638.5387	0	0	1	-360	360;
	2803	7675	0.00158	0.01519	0.393	1308.13861	1308.13861	1308.13861	0	0	1	-360	360;
	5648	2478	0.00172	0.01811	0.44152	1622.95828	1622.95828	1622.95828	0	0	1	-360	360;
	5648	6814	0.00096	0.01121	0.23491	1765.56599	1765.56599	1765.56599	0	0	1	-360	360;
	5648	5874	0.00147	0.01707	0.53053	1979.45057	1979.45057	1979.45057	0	0	1	-360	360;
	1980	4205	0.00129	0.0129	0.31837	1316.70745	1316.70745	1316.70745	0	0	1	-360	360;
	1980	4140	0.00179	0.01737	0.4585	1672.00449	1672.00449	1672.00449	0	0	1	-360	360;
	1980	5874	0.00131	0.01935	0.47265	1797.87799	1797.87799	1797.87799	0	0	1	-360	360;
	1900	8377	0.003901	0.01614	0.02771	435.064039	435.064039	435.064039	0	0	1	-360	360;
	1900	8377	0.002591	0.014771	0.04495	555.123739	555.123739	555.123739	0	0	1	-360	360;
	1900	8377	0.00224	0.0173	0.03912	493.213592	493.213592	493.213592	0	0	1	-360	360;
	1900	3693	0.00012	0.000849	0.0543	225.829218	225.829218	225.829218	0	0	1	-360	360;
	1900	3162	0.00394	0.03039	0.04531	505.185839	505.185839	505.185839	0	0	1	-360	360;
	1900	3277	0.00189	0.01	0.02672	380.840075	380.840075	380.840075	0	0	1	-360	360;
	1900	2530	0.00032	0.00257	0.00356	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1900	2530	0.00038	0.00224	0.00418	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1900	7674	0.00751	0.03955	0.05488	513.304515	513.304515	513.304515	0	0	1	-360	360;
	1900	1034	0.00086	0.006531	0.01425	252.079072	252.079072	252.079072	0	0	1	-360	360;
	1900	17	0.00087	0.006591	0.01692	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8034	910	0.003031	0.02276	0.03624	647.787002	647.787002	647.787002	0	0	1	-360	360;
	8034	4504	0.006331	0.03911	0.07286	267.036855	267.036855	267.036855	0	0	1	-360	360;
	8034	4017	0.00605	0.02986	0.0554	399.046701	399.046701	399.046701	0	0	1	-360	360;
	8034	6074	0.003919	0.01612	0.02767	443.995879	443.995879	443.995879	0	0	1	-360	360;
	8034	1054	0.002039	0.0108	0.01547	455.825993	455.825993	455.825993	0	0	1	-360	360;
	8034	6498	0.006289	0.026271	0.04617	519.987772	519.987772	519.987772	0	0	1	-360	360;
	2667	2579	0.00042	0.00547	0.19147	885.670428	885.670428	885.670428	0	0	1	-360	360;
	2667	2579	0.00043	0.00536	0.19523	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2667	6823	0.00147	0.01373	0.37804	1791.12375	1791.12375	1791.12375	0	0	1	-360	360;
	2667	6823	0.00131	0.01555	0.3368	1577.74268	1577.74268	1577.74268	0	0	1	-360	360;
	2667	1496	0.00221	0.02217	0.56327	1812.35925	1812.35925	1812.35925	0	0	1	-360	360;
	2667	734	0.00224	0.0217	0.57078	1832.20467	1832.20467	1832.20467	0	0	1	-360	360;
	8338	6160	0.00095	0.003289	0.30482	403.975217	403.975217	403.975217	0	0	1	-360	360;
	8338	3090	0.000229	0.00082	0.09827	283.554661	283.554661	283.554661	0	0	1	-360	360;
	8338	6433	0.00068	0.00306	0.40342	311.307762	311.307762	311.307762	0	0	1	-360	360;
	8338	3382	0.00049	0.00256	0.01274	437.917731	437.917731	437.917731	0	0	1	-360	360;
	8338	3366	0.000531	0.0022	0.00851	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8338	6808	0.00136	0.00956	0.02771	522.183389	522.183389	522.183389	0	0	1	-360	360;
	8338	6808	0.0012	0.010729	0.02457	479.382577	479.382577	479.382577	0	0	1	-360	360;
	8338	6808	0.00105	0.00861	0.03764	311.548967	311.548967	311.548967	0	0	1	-360	360;
	8338	2184	0.000591	0.006969	0.02405	275.48133	275.48133	275.48133	0	0	1	-360	360;
	2537	7716	0.00025	0.0008	0.19912	302.121761	302.121761	302.121761	0	0	1	-360	360;
	2537	195	0.00113	0.00632	0.01917	400.690174	400.690174	400.690174	0	0	1	-360	360;
	409	1941	0.0003	0.00275	0.07942	975.0288	975.0288	975.0288	0	0	1	-360	360;
	7882	6303	0.00836	0.051659	0.10261	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7882	6303	0.01338	0.073469	0.12588	605.624832	605.624832	605.624832	0	0	1	-360	360;
	7882	1077	0.0027	0.02089	0.03049	229.070059	229.070059	229.070059	0	0	1	-360	360;
	7882	4017	0.003419	0.02636	0.03833	617.015976	617.015976	617.015976	0	0	1	-360	360;
	7882	7931	0.009531	0.033711	0.05135	459.208966	459.208966	459.208966	0	0	1	-360	360;
	2220	7951	0.0004	0.00707	0.3152	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	2380	7951	0.00035	0.00701	0.20632	2138.8333	2138.8333	2138.8333	0	0	1	-360	360;
	4505	6154	0.00194	0.01751	0.0607	551.93053	551.93053	551.93053	0	0	1	-360	360;
	4505	6154	0.00156	0.01936	0.0543	519.458873	519.458873	519.458873	0	0	1	-360	360;
	4205	875	0.00078	0.00941	0.35817	1359.60428	1359.60428	1359.60428	0	0	1	-360	360;
	4205	875	0.00074	0.00969	0.34171	1582.9061	1582.9061	1582.9061	0	0	1	-360	360;
	4205	7566	0.00042	0.00532	0.09503	1694.37403	1694.37403	1694.37403	0	0	1	-360	360;
	4205	7566	0.0004	0.00473	0.09906	1860.19375	1860.19375	1860.19375	0	0	1	-360	360;
	4205	7566	0.0006	0.00499	0.10813	1572.8116	1572.8116	1572.8116	0	0	1	-360	360;
	1467	5941	0.00086	0.00638	0.02875	610.541155	610.541155	610.541155	0	0	1	-360	360;
	1467	5941	0.000969	0.00625	0.03258	616.423442	616.423442	616.423442	0	0	1	-360	360;
	1467	8171	0.002781	0.01725	0.03181	466.837229	466.837229	466.837229	0	0	1	-360	360;
	3978	4381	0.005031	0.028081	0.15972	564.321933	564.321933	564.321933	0	0	1	-360	360;
	4973	871	1e-05	0.00021	0.00555	2257.80447	2257.80447	2257.80447	0	0	1	-360	360;
	7180	871	1e-05	0.00025	0.00953	2266.56418	2266.56418	2266.56418	0	0	1	-360	360;
	457	871	2e-05	0.00031	0.00797	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2708	871	3e-05	0.0004	0.00831	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4143	5791	0.000841	0.00474	0.00958	533.507849	533.507849	533.507849	0	0	1	-360	360;
	4143	5791	0.00082	0.00487	0.00941	537.122881	537.122881	537.122881	0	0	1	-360	360;
	6209	1453	0.00186	0.011979	0.02074	237.236366	237.236366	237.236366	0	0	1	-360	360;
	3339	4847	0.00056	0.003021	0.00455	394.227165	394.227165	394.227165	0	0	1	-360	360;
	5643	4847	0.00055	0.003081	0.00445	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8169	5582	0.00114	0.01063	0.17472	2006.22728	2006.22728	2006.22728	0	0	1	-360	360;
	1693	5382	0.003719	0.02295	0.0431	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1693	5169	0.006669	0.04288	0.07651	527.460947	527.460947	527.460947	0	0	1	-360	360;
	2660	5263	0.00243	0.0198	0.02761	260.820767	260.820767	260.820767	0	0	1	-360	360;
	2660	6463	0.00845	0.04339	0.06356	537.431533	537.431533	537.431533	0	0	1	-360	360;
	2355	3341	0.00219	0.015169	0.04925	387.959636	387.959636	387.959636	0	0	1	-360	360;
	2355	1944	0.00357	0.0302	0.11918	420.597812	420.597812	420.597812	0	0	1	-360	360;
	2316	6868	0.005849	0.032419	0.04367	434.355284	434.355284	434.355284	0	0	1	-360	360;
	8015	1029	0.00406	0.01688	0.03032	218.208957	218.208957	218.208957	0	0	1	-360	360;
	6687	3018	0.00113	0.0065	0.01259	228.513724	228.513724	228.513724	0	0	1	-360	360;
	1150	4031	0.0068	0.04613	0.07998	532.165024	532.165024	532.165024	0	0	1	-360	360;
	1150	1795	0.00337	0.02011	0.03869	431.695928	431.695928	431.695928	0	0	1	-360	360;
	7312	6495	0.00011	0.0005	0.04966	246.960411	246.960411	246.960411	0	0	1	-360	360;
	6280	4626	0.003711	0.018521	0.02962	447.562518	447.562518	447.562518	0	0	1	-360	360;
	6280	1178	0.000271	0.00145	0.00851	458.417522	458.417522	458.417522	0	0	1	-360	360;
	7584	4282	0.00326	0.01824	0.02457	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8136	6471	0.00118	0.00712	0.01134	427.11493	427.11493	427.11493	0	0	1	-360	360;
	8136	6250	0.00482	0.02232	0.03949	389.630545	389.630545	389.630545	0	0	1	-360	360;
	3813	6365	0.00013	0.000469	0.0575	387.018058	387.018058	387.018058	0	0	1	-360	360;
	3813	5675	0.000229	0.00075	0.09304	437.840378	437.840378	437.840378	0	0	1	-360	360;
	7164	3081	0.004591	0.035021	0.05924	534.846862	534.846862	534.846862	0	0	1	-360	360;
	7164	1316	0.005469	0.03924	0.06166	625.749288	625.749288	625.749288	0	0	1	-360	360;
	7164	1316	0.00536	0.0403	0.05914	331.373155	331.373155	331.373155	0	0	1	-360	360;
	7164	102	0.00424	0.033031	0.0472	344.061397	344.061397	344.061397	0	0	1	-360	360;
	5076	3920	2e-05	0.00024	0.00601	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5076	1612	1e-05	0.00021	0.00566	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5076	3362	0.00286	0.03653	0.72754	1647.75979	1647.75979	1647.75979	0	0	1	-360	360;
	5076	3454	0.00117	0.01183	0.3	1745.74953	1745.74953	1745.74953	0	0	1	-360	360;
	5076	3287	0.00103	0.01336	0.26604	1504.5663	1504.5663	1504.5663	0	0	1	-360	360;
	5076	1368	0.00156	0.01969	0.39888	1566.26403	1566.26403	1566.26403	0	0	1	-360	360;
	5076	5943	0.00449	0.05772	1.14558	1777.87197	1777.87197	1777.87197	0	0	1	-360	360;
	5076	3283	0.00259	0.03469	0.66158	1615.59523	1615.59523	1615.59523	0	0	1	-360	360;
	910	3764	0.000219	0.00118	0.00175	215.929128	215.929128	215.929128	0	0	1	-360	360;
	910	1808	6e-05	0.00095	0.00284	215.929128	215.929128	215.929128	0	0	1	-360	360;
	910	467	6e-05	0.000651	0.00101	647.787002	647.787002	647.787002	0	0	1	-360	360;
	910	3257	8.1e-05	0.000841	0.00202	647.787002	647.787002	647.787002	0	0	1	-360	360;
	910	3466	0.00274	0.020831	0.06273	528.696696	528.696696	528.696696	0	0	1	-360	360;
	910	5094	0.004781	0.027729	0.03822	564.360419	564.360419	564.360419	0	0	1	-360	360;
	910	5094	0.002849	0.02093	0.03415	577.417138	577.417138	577.417138	0	0	1	-360	360;
	910	6074	0.00513	0.03014	0.04617	529.463371	529.463371	529.463371	0	0	1	-360	360;
	910	6498	0.00113	0.00651	0.01278	363.113575	363.113575	363.113575	0	0	1	-360	360;
	910	4899	0.00125	0.00707	0.06377	586.228946	586.228946	586.228946	0	0	1	-360	360;
	6742	4847	0.003349	0.01463	0.02668	356.974459	356.974459	356.974459	0	0	1	-360	360;
	6742	1391	0.00245	0.0122	0.02114	498.752933	498.752933	498.752933	0	0	1	-360	360;
	1184	6303	0.00587	0.036521	0.06544	404.413044	404.413044	404.413044	0	0	1	-360	360;
	1184	6303	0.00569	0.03774	0.06418	374.869003	374.869003	374.869003	0	0	1	-360	360;
	1184	6303	0.006031	0.03545	0.06795	638.255388	638.255388	638.255388	0	0	1	-360	360;
	1184	7154	0.000531	0.00301	0.00639	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1184	5832	0.00076	0.006099	0.00958	520.134095	520.134095	520.134095	0	0	1	-360	360;
	705	6537	0.008841	0.028919	0.03903	307.92136	307.92136	307.92136	0	0	1	-360	360;
	705	1747	0.00169	0.007091	0.01232	324.126323	324.126323	324.126323	0	0	1	-360	360;
	705	6187	0.001031	0.00361	0.00593	321.822107	321.822107	321.822107	0	0	1	-360	360;
	705	2793	0.00813	0.030961	0.04182	359.110632	359.110632	359.110632	0	0	1	-360	360;
	4811	5688	0.00013	0.00068	0.06097	548.819628	548.819628	548.819628	0	0	1	-360	360;
	3326	3509	0.000211	0.000729	0.67953	638.278251	638.278251	638.278251	0	0	1	-360	360;
	3326	3509	0.000169	0.000669	0.53409	603.583541	603.583541	603.583541	0	0	1	-360	360;
	3326	2657	0.000669	0.00251	0.48	500.108332	500.108332	500.108332	0	0	1	-360	360;
	3326	2657	0.000711	0.00238	0.50939	481.263065	481.263065	481.263065	0	0	1	-360	360;
	3326	1066	0.000229	0.00238	0.01132	593.590474	593.590474	593.590474	0	0	1	-360	360;
	3326	8340	0.0007	0.0065	0.02873	557.808625	557.808625	557.808625	0	0	1	-360	360;
	3326	7911	0.00061	0.006031	0.02656	526.635972	526.635972	526.635972	0	0	1	-360	360;
	5872	4870	0.000281	0.0018	0.18523	632.602494	632.602494	632.602494	0	0	1	-360	360;
	5872	3610	0.000211	0.002409	0.01336	275.072081	275.072081	275.072081	0	0	1	-360	360;
	1090	1223	0.00056	0.00773	0.26327	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1090	24	0.0023	0.01581	0.35639	1219.11789	1219.11789	1219.11789	0	0	1	-360	360;
	6776	3347	0.005039	0.032151	0.05529	363.985039	363.985039	363.985039	0	0	1	-360	360;
	6776	85	0.00364	0.01713	0.03513	460.479009	460.479009	460.479009	0	0	1	-360	360;
	6776	1709	0.002169	0.012409	0.0236	344.619637	344.619637	344.619637	0	0	1	-360	360;
	312	2897	0.00113	0.004719	0.00616	344.538092	344.538092	344.538092	0	0	1	-360	360;
	312	5958	0.00113	0.004979	0.00805	494.122018	494.122018	494.122018	0	0	1	-360	360;
	2897	4299	0.00264	0.01174	0.02215	231.266819	231.266819	231.266819	0	0	1	-360	360;
	2406	7157	0.00494	0.03181	0.05638	484.388447	484.388447	484.388447	0	0	1	-360	360;
	2406	1944	0.003711	0.023961	0.04229	626.278568	626.278568	626.278568	0	0	1	-360	360;
	3773	8055	0.001409	0.01101	0.01655	462.44409	462.44409	462.44409	0	0	1	-360	360;
	3773	1481	0.001169	0.00593	0.01022	382.408863	382.408863	382.408863	0	0	1	-360	360;
	3773	1481	0.00118	0.00587	0.00819	324.072214	324.072214	324.072214	0	0	1	-360	360;
	3773	3446	0.00264	0.012469	0.02555	538.831896	538.831896	538.831896	0	0	1	-360	360;
	3773	3446	0.002349	0.01414	0.02277	513.432167	513.432167	513.432167	0	0	1	-360	360;
	3110	3732	0.00014	0.0007	0.00099	286.096273	286.096273	286.096273	0	0	1	-360	360;
	6006	4025	3e-05	0.00054	0.01386	1889.22406	1889.22406	1889.22406	0	0	1	-360	360;
	6006	6492	3e-05	0.00055	0.01357	1830.69152	1830.69152	1830.69152	0	0	1	-360	360;
	6006	3329	4e-05	0.00058	0.01328	1778.91716	1778.91716	1778.91716	0	0	1	-360	360;
	6006	4977	4e-05	0.00054	0.013	1879.59292	1879.59292	1879.59292	0	0	1	-360	360;
	6006	3853	0.00171	0.02197	0.43707	1392.21246	1392.21246	1392.21246	0	0	1	-360	360;
	6006	3853	0.00164	0.02064	0.41847	1658.37096	1658.37096	1658.37096	0	0	1	-360	360;
	6006	6279	0.00136	0.01756	0.3502	1718.20604	1718.20604	1718.20604	0	0	1	-360	360;
	6006	389	0.00097	0.00942	0.25264	1505.89517	1505.89517	1505.89517	0	0	1	-360	360;
	6006	389	0.00086	0.01109	0.21989	1497.82655	1497.82655	1497.82655	0	0	1	-360	360;
	6006	389	0.00085	0.0113	0.21521	1452.03767	1452.03767	1452.03767	0	0	1	-360	360;
	3045	6249	0.000979	0.0058	0.01045	415.127822	415.127822	415.127822	0	0	1	-360	360;
	2248	1872	0.00052	0.00593	0.18425	1925.83989	1925.83989	1925.83989	0	0	1	-360	360;
	2248	1872	0.00046	0.00672	0.16415	1933.09105	1933.09105	1933.09105	0	0	1	-360	360;
	4626	7894	0.001151	0.00636	0.01278	574.538677	574.538677	574.538677	0	0	1	-360	360;
	4626	7894	0.001031	0.007211	0.01143	591.22186	591.22186	591.22186	0	0	1	-360	360;
	4626	4282	0.002341	0.02286	0.06464	534.216604	534.216604	534.216604	0	0	1	-360	360;
	4626	1178	0.002281	0.02518	0.01143	405.589349	405.589349	405.589349	0	0	1	-360	360;
	4626	1747	0.005479	0.02114	0.04046	386.46401	386.46401	386.46401	0	0	1	-360	360;
	4626	2713	0.011159	0.08856	0.13269	485.3647	485.3647	485.3647	0	0	1	-360	360;
	4626	5308	0.000669	0.00249	0.01673	409.147224	409.147224	409.147224	0	0	1	-360	360;
	4626	7096	0.00193	0.01643	0.03825	582.544181	582.544181	582.544181	0	0	1	-360	360;
	4626	2793	0.00024	0.0017	0.01301	451.808952	451.808952	451.808952	0	0	1	-360	360;
	4116	175	0.010159	0.065849	0.12243	340.040926	340.040926	340.040926	0	0	1	-360	360;
	8233	6679	0.00294	0.02131	0.00308	518.803465	518.803465	518.803465	0	0	1	-360	360;
	8233	8306	0.018841	0.07469	0.10019	241.746107	241.746107	241.746107	0	0	1	-360	360;
	8233	5768	0.015599	0.052169	0.08256	411.186229	411.186229	411.186229	0	0	1	-360	360;
	7157	2845	0.005831	0.03969	0.06912	507.534257	507.534257	507.534257	0	0	1	-360	360;
	7157	7765	0.00087	0.00661	0.01046	466.691668	466.691668	466.691668	0	0	1	-360	360;
	7157	5055	0.00637	0.024659	0.04685	394.024446	394.024446	394.024446	0	0	1	-360	360;
	5818	2791	0.0027	0.020419	0.03159	557.430242	557.430242	557.430242	0	0	1	-360	360;
	5818	3528	0.000409	0.00314	0.00436	454.361995	454.361995	454.361995	0	0	1	-360	360;
	5818	7987	0.001669	0.01301	0.01911	495.212586	495.212586	495.212586	0	0	1	-360	360;
	5818	7987	0.00164	0.01325	0.0187	514.358884	514.358884	514.358884	0	0	1	-360	360;
	7103	4140	0.00089	0.00556	0.14036	1297.62617	1297.62617	1297.62617	0	0	1	-360	360;
	7103	3755	0.00186	0.01809	0.45613	1635.83819	1635.83819	1635.83819	0	0	1	-360	360;
	7103	3038	0.00381	0.02853	0.59152	1022.12087	1022.12087	1022.12087	0	0	1	-360	360;
	5015	2158	0.012419	0.072151	0.1028	418.417437	418.417437	418.417437	0	0	1	-360	360;
	5015	1275	0.0042	0.02556	0.03659	417.244943	417.244943	417.244943	0	0	1	-360	360;
	5015	2581	0.010771	0.0493	0.06783	486.41983	486.41983	486.41983	0	0	1	-360	360;
	5015	2206	0.010159	0.04995	0.07399	416.372336	416.372336	416.372336	0	0	1	-360	360;
	5015	8315	0.013169	0.05113	0.09304	500.51377	500.51377	500.51377	0	0	1	-360	360;
	7716	941	0.000469	0.001159	0.08625	410.57769	410.57769	410.57769	0	0	1	-360	360;
	7716	5605	0.000229	0.00105	0.18562	472.456591	472.456591	472.456591	0	0	1	-360	360;
	3090	890	0.000271	0.001271	0.12465	537.160986	537.160986	537.160986	0	0	1	-360	360;
	1916	4184	0.00099	0.01194	0.35557	1658.71585	1658.71585	1658.71585	0	0	1	-360	360;
	1916	4131	0.00077	0.01184	0.28834	2054.01373	2054.01373	2054.01373	0	0	1	-360	360;
	1916	4676	0.00034	0.00426	0.12245	2224.49336	2224.49336	2224.49336	0	0	1	-360	360;
	3570	1941	0.00101	0.01168	0.36215	1765.08223	1765.08223	1765.08223	0	0	1	-360	360;
	3570	4131	0.00086	0.01046	0.3119	1826.82603	1826.82603	1826.82603	0	0	1	-360	360;
	3570	3193	0.00035	0.00417	0.12476	2277.38794	2277.38794	2277.38794	0	0	1	-360	360;
	6054	1352	0.000159	0.00069	0.06858	542.66527	542.66527	542.66527	0	0	1	-360	360;
	4283	5296	0.000781	0.00488	0.0082	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4283	520	0.000781	0.00511	0.00888	545.698819	545.698819	545.698819	0	0	1	-360	360;
	4283	7445	0.00064	0.003539	0.01491	469.574701	469.574701	469.574701	0	0	1	-360	360;
	4283	7445	0.00057	0.003781	0.00593	455.148484	455.148484	455.148484	0	0	1	-360	360;
	4283	2747	0.000531	0.00111	0.12193	352.546263	352.546263	352.546263	0	0	1	-360	360;
	4283	7228	0.000521	0.00113	0.11933	279.864562	279.864562	279.864562	0	0	1	-360	360;
	4283	5440	0.00176	0.00843	0.0151	454.605868	454.605868	454.605868	0	0	1	-360	360;
	4283	7987	0.002461	0.01975	0.02761	577.443049	577.443049	577.443049	0	0	1	-360	360;
	4109	6197	0.000591	0.00105	0.06273	349.130521	349.130521	349.130521	0	0	1	-360	360;
	1866	1294	0.00012	0.00064	0.00093	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6090	1653	0.000581	0.004419	0.0061	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5781	1653	0.000651	0.004331	0.00692	215.929128	215.929128	215.929128	0	0	1	-360	360;
	464	2597	0.001289	0.008289	0.0151	507.579221	507.579221	507.579221	0	0	1	-360	360;
	4522	8369	0.00233	0.03502	1.06047	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4522	8369	0.00238	0.03781	1.08404	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4522	1891	0.00052	0.00605	0.10447	1395.81994	1395.81994	1395.81994	0	0	1	-360	360;
	4522	4955	0.00023	0.00275	0.06481	1525.77679	1525.77679	1525.77679	0	0	1	-360	360;
	4522	7352	0.00134	0.01343	0.3398	1721.55815	1721.55815	1721.55815	0	0	1	-360	360;
	4522	7352	0.00116	0.01552	0.29758	1483.36832	1483.36832	1483.36832	0	0	1	-360	360;
	7218	641	0.004169	0.027919	0.04838	527.405314	527.405314	527.405314	0	0	1	-360	360;
	7218	4381	0.00063	0.008	0.01742	235.022459	235.022459	235.022459	0	0	1	-360	360;
	7218	1707	0.004841	0.028581	0.05233	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7218	4979	0.007419	0.05787	0.20563	602.967763	602.967763	602.967763	0	0	1	-360	360;
	4633	2237	0.002719	0.020331	0.02974	510.876457	510.876457	510.876457	0	0	1	-360	360;
	2115	6510	0.000789	0.005	0.00805	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8331	175	0.00238	0.013219	0.02271	462.355305	462.355305	462.355305	0	0	1	-360	360;
	8331	2852	0.003771	0.02701	0.04182	508.356185	508.356185	508.356185	0	0	1	-360	360;
	266	4134	0.001271	0.00939	0.01394	352.713164	352.713164	352.713164	0	0	1	-360	360;
	8174	6111	0.000841	0.00131	0.19486	304.552486	304.552486	304.552486	0	0	1	-360	360;
	7649	3421	0.005711	0.043039	0.07248	599.481144	599.481144	599.481144	0	0	1	-360	360;
	7649	3421	0.00562	0.041901	0.0697	514.029274	514.029274	514.029274	0	0	1	-360	360;
	4525	2669	0.00356	0.019781	0.0394	551.192053	551.192053	551.192053	0	0	1	-360	360;
	4525	144	0.004841	0.03149	0.0543	541.485536	541.485536	541.485536	0	0	1	-360	360;
	167	7712	0.00057	0.00107	0.06055	330.262771	330.262771	330.262771	0	0	1	-360	360;
	6268	4471	0.000539	0.00213	0.22396	521.885788	521.885788	521.885788	0	0	1	-360	360;
	400	4318	0.0007	0.00425	0.0082	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1327	5892	0.0012	0.009531	0.02761	280.25552	280.25552	280.25552	0	0	1	-360	360;
	1327	2493	0.00206	0.015901	0.04003	307.333779	307.333779	307.333779	0	0	1	-360	360;
	3276	4299	0.00063	0.00312	0.00465	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1136	5874	3e-05	0.00041	0.01132	1496.55429	1496.55429	1496.55429	0	0	1	-360	360;
	997	5874	2e-05	0.00043	0.01063	2122.41238	2122.41238	2122.41238	0	0	1	-360	360;
	5181	5874	3e-05	0.00039	0.0104	1284.40203	1284.40203	1284.40203	0	0	1	-360	360;
	3481	5874	3e-05	0.00038	0.01178	2081.13862	2081.13862	2081.13862	0	0	1	-360	360;
	1288	3183	0.0008	0.004151	0.0061	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4139	3183	0.000901	0.00381	0.00604	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2794	578	0.000591	0.004341	0.04704	478.85787	478.85787	478.85787	0	0	1	-360	360;
	2794	578	0.000591	0.00405	0.0575	513.876092	513.876092	513.876092	0	0	1	-360	360;
	4555	4131	0.00218	0.0275	0.43052	1169.57476	1169.57476	1169.57476	0	0	1	-360	360;
	4555	7429	0.00184	0.01677	0.39127	1108.78663	1108.78663	1108.78663	0	0	1	-360	360;
	686	875	0.00058	0.00807	0.26039	1091.57458	1091.57458	1091.57458	0	0	1	-360	360;
	686	875	0.00056	0.00823	0.25784	952.801426	952.801426	952.801426	0	0	1	-360	360;
	686	3558	0.00123	0.01146	0.31768	1434.22207	1434.22207	1434.22207	0	0	1	-360	360;
	686	3558	0.0011	0.01298	0.28302	1500.80415	1500.80415	1500.80415	0	0	1	-360	360;
	686	7566	0.00156	0.01585	0.40103	1783.56851	1783.56851	1783.56851	0	0	1	-360	360;
	686	2648	0.00226	0.02781	0.57667	1487.68335	1487.68335	1487.68335	0	0	1	-360	360;
	3421	168	0.00614	0.02831	0.06602	418.694842	418.694842	418.694842	0	0	1	-360	360;
	3421	2043	0.00513	0.03401	0.05522	239.49257	239.49257	239.49257	0	0	1	-360	360;
	3555	3744	0.013219	0.0597	0.08728	236.263162	236.263162	236.263162	0	0	1	-360	360;
	5214	6338	0.012969	0.06087	0.0855	240.085867	240.085867	240.085867	0	0	1	-360	360;
	7823	8107	0.00437	0.032831	0.05018	555.161463	555.161463	555.161463	0	0	1	-360	360;
	7823	2237	0.005651	0.033229	0.07114	596.715094	596.715094	596.715094	0	0	1	-360	360;
	1969	4979	0.008969	0.03625	0.06691	453.103383	453.103383	453.103383	0	0	1	-360	360;
	1969	7050	0.012409	0.05645	0.09084	421.345053	421.345053	421.345053	0	0	1	-360	360;
	4551	5791	0.00038	0.002271	0.00426	409.452827	409.452827	409.452827	0	0	1	-360	360;
	5873	3719	0.00138	0.01089	0.01547	574.794363	574.794363	574.794363	0	0	1	-360	360;
	5873	3719	0.001539	0.012521	0.01692	606.634237	606.634237	606.634237	0	0	1	-360	360;
	5873	45	0.00205	0.014229	0.10055	550.448241	550.448241	550.448241	0	0	1	-360	360;
	5873	6581	0.00026	0.00164	0.00307	511.214068	511.214068	511.214068	0	0	1	-360	360;
	5873	2245	0.000289	0.00169	0.00418	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4847	6874	0.003349	0.051211	0.02405	463.160848	463.160848	463.160848	0	0	1	-360	360;
	7224	2080	0.00038	0.001479	0.16032	379.885923	379.885923	379.885923	0	0	1	-360	360;
	1451	6438	0.00082	0.00625	0.01022	558.063549	558.063549	558.063549	0	0	1	-360	360;
	2900	1897	0.00275	0.015581	0.03088	572.708488	572.708488	572.708488	0	0	1	-360	360;
	5483	5375	0.000909	0.00739	0.01069	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3362	7692	0.00082	0.01038	0.20794	1565.17409	1565.17409	1565.17409	0	0	1	-360	360;
	3362	6830	0.00093	0.01017	0.23566	1410.85013	1410.85013	1410.85013	0	0	1	-360	360;
	3362	1368	0.00137	0.01431	0.34902	1878.21733	1878.21733	1878.21733	0	0	1	-360	360;
	3362	1416	0.00215	0.02021	0.55912	1685.37211	1685.37211	1685.37211	0	0	1	-360	360;
	6303	3453	0.01189	0.058841	0.08822	469.297677	469.297677	469.297677	0	0	1	-360	360;
	6112	2286	0.00104	0.01241	0.47144	1270.53617	1270.53617	1270.53617	0	0	1	-360	360;
	6112	2286	0.00092	0.014	0.41807	1274.77484	1274.77484	1274.77484	0	0	1	-360	360;
	6112	2149	0.00055	0.00845	0.25004	1171.18203	1171.18203	1171.18203	0	0	1	-360	360;
	6112	2149	0.00054	0.00861	0.24472	1003.75438	1003.75438	1003.75438	0	0	1	-360	360;
	6112	3454	0.00105	0.01057	0.26634	1831.44447	1831.44447	1831.44447	0	0	1	-360	360;
	6112	3287	0.00089	0.01122	0.22613	1630.41019	1630.41019	1630.41019	0	0	1	-360	360;
	6112	6261	0.00223	0.02584	0.56605	1794.35278	1794.35278	1794.35278	0	0	1	-360	360;
	7414	4644	0.00101	0.007841	0.01092	564.002993	564.002993	564.002993	0	0	1	-360	360;
	7414	4644	0.001081	0.008099	0.01208	578.823598	578.823598	578.823598	0	0	1	-360	360;
	5662	6498	0.011081	0.068479	0.12621	351.984213	351.984213	351.984213	0	0	1	-360	360;
	8369	7910	2e-05	0.00033	0.01132	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	8369	8053	1e-05	0.00029	0.00901	1798.15113	1798.15113	1798.15113	0	0	1	-360	360;
	8369	7866	0.00071	0.00892	0.32438	1601.74714	1601.74714	1601.74714	0	0	1	-360	360;
	8369	6497	0.00136	0.01526	0.62583	1467.10603	1467.10603	1467.10603	0	0	1	-360	360;
	5382	7050	0.005401	0.036021	0.07207	598.54795	598.54795	598.54795	0	0	1	-360	360;
	2853	3116	0.00437	0.033719	0.05924	470.166092	470.166092	470.166092	0	0	1	-360	360;
	2853	958	0.005151	0.034919	0.04987	522.354481	522.354481	522.354481	0	0	1	-360	360;
	258	725	0.00038	0.00139	0.00213	358.391588	358.391588	358.391588	0	0	1	-360	360;
	258	983	0.00037	0.0015	0.00206	215.929128	215.929128	215.929128	0	0	1	-360	360;
	258	4318	0.00069	0.004461	0.00805	547.833468	547.833468	547.833468	0	0	1	-360	360;
	3081	4741	0.00688	0.040591	0.07696	395.436241	395.436241	395.436241	0	0	1	-360	360;
	3081	102	0.001341	0.008349	0.01642	484.090465	484.090465	484.090465	0	0	1	-360	360;
	3081	102	0.001229	0.009669	0.01425	643.406819	643.406819	643.406819	0	0	1	-360	360;
	4430	5222	0.00137	0.006039	0.00987	436.028479	436.028479	436.028479	0	0	1	-360	360;
	6839	5944	0.001919	0.01261	0.02271	487.725312	487.725312	487.725312	0	0	1	-360	360;
	6839	787	0.00163	0.0095	0.01881	536.910636	536.910636	536.910636	0	0	1	-360	360;
	8074	8055	0.00488	0.0497	0.0558	583.316572	583.316572	583.316572	0	0	1	-360	360;
	8074	4101	0.001289	0.0087	0.08728	564.532273	564.532273	564.532273	0	0	1	-360	360;
	8074	3122	0.00211	0.01524	0.02467	576.331142	576.331142	576.331142	0	0	1	-360	360;
	8074	3122	0.00206	0.015539	0.02416	605.628643	605.628643	605.628643	0	0	1	-360	360;
	8074	3122	0.00276	0.01631	0.02582	516.448568	516.448568	516.448568	0	0	1	-360	360;
	8074	5944	0.00305	0.019521	0.03322	618.800058	618.800058	618.800058	0	0	1	-360	360;
	8074	7953	0.000409	0.0023	0.00426	509.545445	509.545445	509.545445	0	0	1	-360	360;
	8074	6002	0.00036	0.00268	0.00372	516.486674	516.486674	516.486674	0	0	1	-360	360;
	7400	6108	0.00032	0.00049	0.04838	215.929128	215.929128	215.929128	0	0	1	-360	360;
	984	6108	0.00013	0.000349	0.09477	288.756391	288.756391	288.756391	0	0	1	-360	360;
	7245	3853	0.0002	0.00327	0.10102	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	729	1951	0.000919	0.00561	0.01949	543.29591	543.29591	543.29591	0	0	1	-360	360;
	729	1951	0.00094	0.00549	0.023	539.357365	539.357365	539.357365	0	0	1	-360	360;
	729	6820	0.00175	0.01976	0.05187	541.369696	541.369696	541.369696	0	0	1	-360	360;
	729	6117	0.001729	0.019331	0.05096	484.208209	484.208209	484.208209	0	0	1	-360	360;
	2079	6170	0.00406	0.02581	0.04631	574.671283	574.671283	574.671283	0	0	1	-360	360;
	2079	698	0.00526	0.025961	0.0389	524.353475	524.353475	524.353475	0	0	1	-360	360;
	2945	986	0.00179	0.02098	0.43586	1360.13873	1360.13873	1360.13873	0	0	1	-360	360;
	2945	5430	0.00162	0.02148	0.39854	1411.55569	1411.55569	1411.55569	0	0	1	-360	360;
	2945	7493	0.00492	0.04991	1.20141	1060.16693	1060.16693	1060.16693	0	0	1	-360	360;
	2945	871	0.00109	0.01361	0.49859	1095.99031	1095.99031	1095.99031	0	0	1	-360	360;
	2945	871	0.00097	0.0153	0.44106	911.008357	911.008357	911.008357	0	0	1	-360	360;
	2945	3558	0.00322	0.03022	0.78911	1502.6227	1502.6227	1502.6227	0	0	1	-360	360;
	1403	6674	0.032831	0.074151	0.14161	257.629463	257.629463	257.629463	0	0	1	-360	360;
	1403	6215	0.01411	0.058711	0.11286	518.404123	518.404123	518.404123	0	0	1	-360	360;
	1403	7720	0.02725	0.102979	0.19486	438.393664	438.393664	438.393664	0	0	1	-360	360;
	1403	2880	0.012229	0.04049	0.08781	403.924918	403.924918	403.924918	0	0	1	-360	360;
	1403	2426	0.01994	0.066719	0.10772	357.425624	357.425624	357.425624	0	0	1	-360	360;
	1403	3453	0.018591	0.07462	0.13591	441.049972	441.049972	441.049972	0	0	1	-360	360;
	1403	4120	0.008979	0.03251	0.0607	356.029071	356.029071	356.029071	0	0	1	-360	360;
	1403	3856	0.016531	0.086031	0.11738	384.357178	384.357178	384.357178	0	0	1	-360	360;
	7825	37	0.003591	0.02137	0.04002	607.969059	607.969059	607.969059	0	0	1	-360	360;
	7825	6012	0.00124	0.008	0.01283	411.000276	411.000276	411.000276	0	0	1	-360	360;
	7825	6012	0.001271	0.007831	0.01308	455.553923	455.553923	455.553923	0	0	1	-360	360;
	7825	1409	0.000469	0.00294	0.00906	509.439132	509.439132	509.439132	0	0	1	-360	360;
	7825	3839	0.000841	0.003289	0.3457	500.49205	500.49205	500.49205	0	0	1	-360	360;
	6496	1754	0.00219	0.016539	0.02547	402.576378	402.576378	402.576378	0	0	1	-360	360;
	6496	5010	0.002901	0.020901	0.03252	403.514145	403.514145	403.514145	0	0	1	-360	360;
	6938	204	0.0002	0.001211	0.00202	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4427	1271	9.1e-05	0.000479	0.00091	335.513276	335.513276	335.513276	0	0	1	-360	360;
	6367	6458	0.005461	0.029469	0.04274	432.413447	432.413447	432.413447	0	0	1	-360	360;
	6367	3018	0.001979	0.011289	0.02236	595.247666	595.247666	595.247666	0	0	1	-360	360;
	1133	144	0.0023	0.01289	0.02345	252.431925	252.431925	252.431925	0	0	1	-360	360;
	8214	2227	0.000159	0.00037	0.02912	342.807738	342.807738	342.807738	0	0	1	-360	360;
	8055	3446	0.002789	0.01562	0.44832	543.451379	543.451379	543.451379	0	0	1	-360	360;
	5344	5062	0.002531	0.012331	0.01975	385.621887	385.621887	385.621887	0	0	1	-360	360;
	5344	6952	0.00357	0.023289	0.037	416.91762	416.91762	416.91762	0	0	1	-360	360;
	5344	4449	0.00076	0.006039	0.0091	527.690721	527.690721	527.690721	0	0	1	-360	360;
	3641	1391	0.000469	0.00075	0.04743	322.237072	322.237072	322.237072	0	0	1	-360	360;
	164	1391	0.00044	0.00069	0.04838	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2701	1904	0.00561	0.035521	0.06342	604.27553	604.27553	604.27553	0	0	1	-360	360;
	7870	844	0.004539	0.03468	0.04898	373.025859	373.025859	373.025859	0	0	1	-360	360;
	7870	844	0.00175	0.0067	0.0362	424.735266	424.735266	424.735266	0	0	1	-360	360;
	7870	3313	0.007461	0.031419	0.05235	401.11276	401.11276	401.11276	0	0	1	-360	360;
	7894	6099	0.00224	0.012909	0.02509	620.534984	620.534984	620.534984	0	0	1	-360	360;
	6391	16	0.0003	0.000961	0.24277	430.731868	430.731868	430.731868	0	0	1	-360	360;
	6391	16	0.00032	0.001409	0.14589	219.786509	219.786509	219.786509	0	0	1	-360	360;
	6391	1415	0.002031	0.0157	0.04269	568.47882	568.47882	568.47882	0	0	1	-360	360;
	6391	3735	0.00038	0.00306	0.02547	520.966692	520.966692	520.966692	0	0	1	-360	360;
	6391	7133	0.00118	0.0083	0.0351	552.860676	552.860676	552.860676	0	0	1	-360	360;
	6391	4952	0.001229	0.00787	0.02416	579.77051	579.77051	579.77051	0	0	1	-360	360;
	5062	5143	0.00301	0.01538	0.02137	525.66734	525.66734	525.66734	0	0	1	-360	360;
	6115	4952	0.00063	0.00114	0.14515	395.387847	395.387847	395.387847	0	0	1	-360	360;
	4790	7133	0.00064	0.00111	0.14799	266.970933	266.970933	266.970933	0	0	1	-360	360;
	7053	3511	0.004919	0.03837	0.05642	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5706	3122	0.001831	0.008349	0.02919	438.960287	438.960287	438.960287	0	0	1	-360	360;
	3116	362	0.00536	0.03188	0.06168	538.304521	538.304521	538.304521	0	0	1	-360	360;
	2696	8151	0.00038	0.001461	0.0021	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2696	8151	0.000289	0.001521	0.00182	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8151	2250	0.00305	0.016841	0.02366	487.972614	487.972614	487.972614	0	0	1	-360	360;
	8151	8264	0.001781	0.012729	0.02941	459.923437	459.923437	459.923437	0	0	1	-360	360;
	1878	6821	0.00041	0.00461	0.09908	985.302979	985.302979	985.302979	0	0	1	-360	360;
	1878	6821	0.0004	0.0047	0.09706	1044.36996	1044.36996	1044.36996	0	0	1	-360	360;
	3719	4852	0.00213	0.01401	0.02369	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3719	2108	0.001271	0.00887	0.01301	441.903145	441.903145	441.903145	0	0	1	-360	360;
	3719	2608	0.00462	0.02743	0.05227	544.428775	544.428775	544.428775	0	0	1	-360	360;
	2354	6728	0.001539	0.00776	0.01436	328.636826	328.636826	328.636826	0	0	1	-360	360;
	6514	8108	0.00025	0.00478	0.15203	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	4884	8108	0.00025	0.00487	0.14879	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	7367	8108	0.00025	0.00435	0.16202	843.64897	843.64897	843.64897	0	0	1	-360	360;
	5748	8108	0.00026	0.00426	0.16519	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1123	8108	0.00025	0.00479	0.152	2267.92595	2267.92595	2267.92595	0	0	1	-360	360;
	3757	8324	0.000169	0.00076	0.07144	449.317258	449.317258	449.317258	0	0	1	-360	360;
	3757	7235	0.00039	0.00126	0.16212	501.885173	501.885173	501.885173	0	0	1	-360	360;
	2208	4381	0.011229	0.074659	0.12737	629.055669	629.055669	629.055669	0	0	1	-360	360;
	2208	1707	0.014521	0.058081	0.07643	408.713969	408.713969	408.713969	0	0	1	-360	360;
	2208	7992	0.00064	0.003919	0.01013	533.867561	533.867561	533.867561	0	0	1	-360	360;
	2208	4678	0.014789	0.07768	0.1089	443.074497	443.074497	443.074497	0	0	1	-360	360;
	2208	6295	0.00938	0.05355	0.10755	580.516608	580.516608	580.516608	0	0	1	-360	360;
	2208	1591	0.00124	0.004289	0.00673	423.443883	423.443883	423.443883	0	0	1	-360	360;
	4741	7824	0.00968	0.04388	0.08518	248.199209	248.199209	248.199209	0	0	1	-360	360;
	4741	154	0.008531	0.067151	0.10106	465.063817	465.063817	465.063817	0	0	1	-360	360;
	4741	6741	0.00731	0.04106	0.05699	438.740421	438.740421	438.740421	0	0	1	-360	360;
	6657	5387	0.004349	0.02555	0.04344	536.298286	536.298286	536.298286	0	0	1	-360	360;
	6657	6797	0.00599	0.03462	0.06122	488.955726	488.955726	488.955726	0	0	1	-360	360;
	7991	4381	0.002961	0.021591	0.04182	564.023951	564.023951	564.023951	0	0	1	-360	360;
	7991	4381	0.002901	0.02206	0.04186	579.598656	579.598656	579.598656	0	0	1	-360	360;
	7991	5493	0.002979	0.016969	0.03301	490.70818	490.70818	490.70818	0	0	1	-360	360;
	5231	3817	0.00312	0.020021	0.03524	490.765719	490.765719	490.765719	0	0	1	-360	360;
	7712	1803	0.000219	0.000909	0.09676	342.38782	342.38782	342.38782	0	0	1	-360	360;
	1268	1803	0.000219	0.00093	0.09386	462.401031	462.401031	462.401031	0	0	1	-360	360;
	2627	1940	0.000151	0.000289	0.0236	375.192135	375.192135	375.192135	0	0	1	-360	360;
	4496	7483	0.00563	0.043599	0.06447	364.539468	364.539468	364.539468	0	0	1	-360	360;
	4496	4319	0.000969	0.00612	0.01007	523.535358	523.535358	523.535358	0	0	1	-360	360;
	3167	3269	0.005021	0.020719	0.046	441.499993	441.499993	441.499993	0	0	1	-360	360;
	3167	4252	0.00988	0.041039	0.12342	348.969336	348.969336	348.969336	0	0	1	-360	360;
	4230	3857	0.003159	0.02011	0.03624	480.934218	480.934218	480.934218	0	0	1	-360	360;
	4230	3857	0.003219	0.01969	0.03694	479.265976	479.265976	479.265976	0	0	1	-360	360;
	6467	4936	0.00049	0.00238	0.00379	307.409227	307.409227	307.409227	0	0	1	-360	360;
	6467	4936	0.000479	0.002419	0.00372	305.538266	305.538266	305.538266	0	0	1	-360	360;
	7479	5193	0.00264	0.019729	0.03903	544.554903	544.554903	544.554903	0	0	1	-360	360;
	7479	5193	0.002581	0.02011	0.03822	505.24414	505.24414	505.24414	0	0	1	-360	360;
	6197	6099	0.00037	0.002831	0.00455	570.722069	570.722069	570.722069	0	0	1	-360	360;
	6197	5161	0.00213	0.012789	0.01992	520.785312	520.785312	520.785312	0	0	1	-360	360;
	2158	730	0.018669	0.092281	0.16007	442.954085	442.954085	442.954085	0	0	1	-360	360;
	820	8251	3.9e-05	0.000401	0.00095	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3695	8251	6e-05	0.00044	0.00089	399.734117	399.734117	399.734117	0	0	1	-360	360;
	6203	1951	6e-05	0.0005	0.00093	520.109327	520.109327	520.109327	0	0	1	-360	360;
	5143	2393	0.001271	0.007031	0.0098	489.004882	489.004882	489.004882	0	0	1	-360	360;
	5143	6874	0.00055	0.003159	0.00532	616.945101	616.945101	616.945101	0	0	1	-360	360;
	5143	6874	0.00055	0.00338	0.00616	609.551565	609.551565	609.551565	0	0	1	-360	360;
	5143	6874	0.00105	0.004771	0.00941	495.662608	495.662608	495.662608	0	0	1	-360	360;
	5143	1484	0.003289	0.015039	0.02091	395.285725	395.285725	395.285725	0	0	1	-360	360;
	658	623	0.00036	0.004349	0.01069	617.130673	617.130673	617.130673	0	0	1	-360	360;
	3205	623	0.00038	0.003669	0.0151	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5791	2201	0.00118	0.00701	0.01359	300.927928	300.927928	300.927928	0	0	1	-360	360;
	5791	2201	0.001229	0.00686	0.01384	538.729774	538.729774	538.729774	0	0	1	-360	360;
	5791	5358	0.00301	0.0217	0.03159	285.137548	285.137548	285.137548	0	0	1	-360	360;
	5791	5358	0.003409	0.01924	0.03833	314.544029	314.544029	314.544029	0	0	1	-360	360;
	5791	5358	0.003409	0.01938	0.03764	573.819634	573.819634	573.819634	0	0	1	-360	360;
	7119	8365	0.002719	0.02131	0.03094	269.041946	269.041946	269.041946	0	0	1	-360	360;
	2709	1316	0.00295	0.01956	0.03357	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4648	5161	0.001091	0.00831	0.03284	489.757458	489.757458	489.757458	0	0	1	-360	360;
	2697	5358	0.00111	0.008211	0.03252	556.130476	556.130476	556.130476	0	0	1	-360	360;
	4936	1963	0.002469	0.015271	0.02873	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4936	7354	0.00212	0.01601	0.02416	242.804667	242.804667	242.804667	0	0	1	-360	360;
	4936	2294	0.00037	0.0028	0.00364	547.55149	547.55149	547.55149	0	0	1	-360	360;
	4936	6438	0.00162	0.01238	0.0183	234.559482	234.559482	234.559482	0	0	1	-360	360;
	4936	6438	0.001909	0.01162	0.02155	318.573264	318.573264	318.573264	0	0	1	-360	360;
	4936	6438	0.001909	0.010729	0.02236	561.117293	561.117293	561.117293	0	0	1	-360	360;
	4936	3401	0.00089	0.00526	0.00743	419.94088	419.94088	419.94088	0	0	1	-360	360;
	4936	5019	0.00087	0.00536	0.00728	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4936	5222	0.001521	0.01176	0.01742	548.813912	548.813912	548.813912	0	0	1	-360	360;
	8191	5951	0.00014	0.000591	0.09775	491.532013	491.532013	491.532013	0	0	1	-360	360;
	1221	1004	0.003531	0.028781	0.04097	266.599789	266.599789	266.599789	0	0	1	-360	360;
	2533	7517	0.004599	0.03657	0.05077	554.684768	554.684768	554.684768	0	0	1	-360	360;
	2533	7582	0.00307	0.019159	0.03322	603.388062	603.388062	603.388062	0	0	1	-360	360;
	2533	1003	0.000581	0.004461	0.00637	518.963125	518.963125	518.963125	0	0	1	-360	360;
	2533	6249	0.00164	0.01161	0.01707	609.086682	609.086682	609.086682	0	0	1	-360	360;
	4493	6242	0.00057	0.003281	0.00746	551.546049	551.546049	551.546049	0	0	1	-360	360;
	7020	6330	0.020419	0.063669	0.10465	322.593354	322.593354	322.593354	0	0	1	-360	360;
	5953	2840	0.01424	0.04751	0.08106	224.301584	224.301584	224.301584	0	0	1	-360	360;
	7897	255	0.00157	0.008849	0.03194	566.540794	566.540794	566.540794	0	0	1	-360	360;
	7897	2273	0.01218	0.07524	0.13852	339.407238	339.407238	339.407238	0	0	1	-360	360;
	1481	695	0.00037	0.00151	0.15705	364.999016	364.999016	364.999016	0	0	1	-360	360;
	7623	5655	0.000219	0.000531	0.03604	328.605199	328.605199	328.605199	0	0	1	-360	360;
	3600	1271	0.002849	0.0137	0.01992	341.119301	341.119301	341.119301	0	0	1	-360	360;
	875	6601	0.00116	0.01692	0.524	1293.38816	1293.38816	1293.38816	0	0	1	-360	360;
	875	6601	0.00133	0.01462	0.60359	1428.15761	1428.15761	1428.15761	0	0	1	-360	360;
	875	5874	0.00063	0.00821	0.28868	1019.4842	1019.4842	1019.4842	0	0	1	-360	360;
	875	5874	0.00064	0.00804	0.29434	1284.08216	1284.08216	1284.08216	0	0	1	-360	360;
	5135	4921	0.00126	0.01272	0.32143	1399.52944	1399.52944	1399.52944	0	0	1	-360	360;
	5135	7493	0.00337	0.0494	0.82527	1768.20726	1768.20726	1768.20726	0	0	1	-360	360;
	5135	6644	0.0014	0.0214	0.51846	2057.1414	2057.1414	2057.1414	0	0	1	-360	360;
	5135	6644	0.00137	0.02182	0.50766	2166.35375	2166.35375	2166.35375	0	0	1	-360	360;
	5135	6497	0.00212	0.02063	0.53959	1247.65259	1247.65259	1247.65259	0	0	1	-360	360;
	5038	369	0.00186	0.01045	0.08518	613.25462	613.25462	613.25462	0	0	1	-360	360;
	5038	6876	0.00426	0.021771	0.03066	454.707227	454.707227	454.707227	0	0	1	-360	360;
	5038	6876	0.002659	0.020539	0.03049	497.562148	497.562148	497.562148	0	0	1	-360	360;
	5038	3817	0.00606	0.049919	0.07035	462.840002	462.840002	462.840002	0	0	1	-360	360;
	6601	3823	0.00109	0.01192	0.48923	918.297693	918.297693	918.297693	0	0	1	-360	360;
	1391	3567	0.013021	0.049219	0.09157	453.632282	453.632282	453.632282	0	0	1	-360	360;
	1391	4318	0.00236	0.01239	0.02114	446.600744	446.600744	446.600744	0	0	1	-360	360;
	1391	4318	0.0027	0.015169	0.02091	431.264197	431.264197	431.264197	0	0	1	-360	360;
	4131	3038	0.00285	0.03283	1.0134	1517.03222	1517.03222	1517.03222	0	0	1	-360	360;
	4131	3038	0.00529	0.03969	0.82926	1039.84103	1039.84103	1039.84103	0	0	1	-360	360;
	1223	3038	0.00254	0.03721	0.90285	1573.65868	1573.65868	1573.65868	0	0	1	-360	360;
	3347	2465	0.000159	0.00069	0.00465	339.032664	339.032664	339.032664	0	0	1	-360	360;
	3347	2465	0.000229	0.000979	0.01781	314.355028	314.355028	314.355028	0	0	1	-360	360;
	3347	2465	0.00024	0.00082	0.01642	295.264745	295.264745	295.264745	0	0	1	-360	360;
	3347	720	0.000169	0.00075	0.00789	349.338193	349.338193	349.338193	0	0	1	-360	360;
	3347	4442	0.00368	0.01756	0.02439	338.946166	338.946166	338.946166	0	0	1	-360	360;
	3347	1709	0.002599	0.020659	0.03562	370.929696	370.929696	370.929696	0	0	1	-360	360;
	6767	5052	0.00018	0.00064	0.00616	336.905256	336.905256	336.905256	0	0	1	-360	360;
	6767	6380	0.000159	0.000771	0.00651	381.663908	381.663908	381.663908	0	0	1	-360	360;
	6767	5387	0.00357	0.01981	0.04617	485.920653	485.920653	485.920653	0	0	1	-360	360;
	6767	6797	0.006021	0.03205	0.05839	534.361022	534.361022	534.361022	0	0	1	-360	360;
	4413	2615	0.001781	0.01164	0.02114	536.38288	536.38288	536.38288	0	0	1	-360	360;
	4413	6393	0.002781	0.018831	0.03654	517.809302	517.809302	517.809302	0	0	1	-360	360;
	2273	5575	0.000349	0.00281	0.00436	266.219119	266.219119	266.219119	0	0	1	-360	360;
	79	1701	0.00313	0.02444	0.0364	620.69655	620.69655	620.69655	0	0	1	-360	360;
	79	1701	0.00368	0.02074	0.04259	604.304109	604.304109	604.304109	0	0	1	-360	360;
	1951	2978	9.1e-05	0.001539	0.00267	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1951	3224	0.00251	0.01111	0.01777	331.855946	331.855946	331.855946	0	0	1	-360	360;
	1951	6207	0.003081	0.017961	0.03113	436.032671	436.032671	436.032671	0	0	1	-360	360;
	1951	6117	0.005341	0.030419	0.05551	439.689238	439.689238	439.689238	0	0	1	-360	360;
	1951	6928	9.9e-05	0.00126	0.0019	518.000209	518.000209	518.000209	0	0	1	-360	360;
	8251	948	0.002169	0.017901	0.07124	431.883024	431.883024	431.883024	0	0	1	-360	360;
	2227	5605	0.00011	0.00038	0.02582	358.43922	358.43922	358.43922	0	0	1	-360	360;
	1294	1329	0.00212	0.00494	0.00836	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1294	7050	0.018901	0.077349	0.13545	483.138218	483.138218	483.138218	0	0	1	-360	360;
	7152	4873	0.00957	0.0395	0.0622	462.209744	462.209744	462.209744	0	0	1	-360	360;
	7152	6330	0.02324	0.063469	0.09766	407.673699	407.673699	407.673699	0	0	1	-360	360;
	6284	3021	0.000531	0.004331	0.01069	611.731559	611.731559	611.731559	0	0	1	-360	360;
	1789	557	0.002581	0.006229	0.01007	328.587289	328.587289	328.587289	0	0	1	-360	360;
	5857	1147	0.00031	0.002211	0.00348	429.794863	429.794863	429.794863	0	0	1	-360	360;
	728	6330	0.01087	0.06114	0.08104	448.355866	448.355866	448.355866	0	0	1	-360	360;
	728	5369	0.011581	0.061099	0.09202	430.563443	430.563443	430.563443	0	0	1	-360	360;
	6918	2792	0.002099	0.016271	0.02366	411.256723	411.256723	411.256723	0	0	1	-360	360;
	452	1411	0.00093	0.005461	0.02048	230.666663	230.666663	230.666663	0	0	1	-360	360;
	452	4762	0.009591	0.037039	0.07028	429.603576	429.603576	429.603576	0	0	1	-360	360;
	452	701	0.002591	0.017211	0.02962	558.3722	558.3722	558.3722	0	0	1	-360	360;
	452	8125	0.00164	0.013651	0.03295	505.145066	505.145066	505.145066	0	0	1	-360	360;
	2496	1897	0.003219	0.014719	0.02771	469.450097	469.450097	469.450097	0	0	1	-360	360;
	2496	3140	0.00401	0.022229	0.02938	473.398931	473.398931	473.398931	0	0	1	-360	360;
	2496	6393	0.00126	0.010211	0.01729	602.962428	602.962428	602.962428	0	0	1	-360	360;
	3748	7144	0.009169	0.03993	0.06745	349.170531	349.170531	349.170531	0	0	1	-360	360;
	3748	3519	0.001349	0.006081	0.00987	298.792898	298.792898	298.792898	0	0	1	-360	360;
	1603	7692	0.00156	0.01931	0.40201	1598.23509	1598.23509	1598.23509	0	0	1	-360	360;
	1603	6830	0.00152	0.01968	0.39363	1508.87475	1508.87475	1508.87475	0	0	1	-360	360;
	1603	7056	0.00094	0.01433	0.33547	1687.66915	1687.66915	1687.66915	0	0	1	-360	360;
	1603	6257	0.00092	0.01461	0.32848	1451.11754	1451.11754	1451.11754	0	0	1	-360	360;
	1603	871	0.00137	0.02012	0.62381	1134.94663	1134.94663	1134.94663	0	0	1	-360	360;
	1603	871	0.00134	0.02051	0.61081	953.339817	953.339817	953.339817	0	0	1	-360	360;
	3511	4087	0.00314	0.042151	0.0351	582.150937	582.150937	582.150937	0	0	1	-360	360;
	3511	5454	0.000651	0.00494	0.02695	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3511	6295	0.003961	0.02426	0.04515	591.024475	591.024475	591.024475	0	0	1	-360	360;
	776	1496	0.00013	0.0015	0.04679	1993.32367	1993.32367	1993.32367	0	0	1	-360	360;
	2754	734	0.00013	0.00147	0.04765	1889.27869	1889.27869	1889.27869	0	0	1	-360	360;
	4383	1496	0.00011	0.00172	0.04072	1879.09929	1879.09929	1879.09929	0	0	1	-360	360;
	1858	734	0.00012	0.00166	0.04245	1992.37853	1992.37853	1992.37853	0	0	1	-360	360;
	3523	1496	0.00012	0.00168	0.04245	1911.24937	1911.24937	1911.24937	0	0	1	-360	360;
	3405	734	0.00012	0.00172	0.04159	1924.76838	1924.76838	1924.76838	0	0	1	-360	360;
	672	623	0.00093	0.005669	0.05669	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2639	6438	0.00105	0.00506	0.0575	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5163	5654	0.00113	0.006539	0.02875	347.635276	347.635276	347.635276	0	0	1	-360	360;
	4713	4949	0.00037	0.002919	0.00364	215.929128	215.929128	215.929128	0	0	1	-360	360;
	641	5169	0.005169	0.03738	0.05882	477.636601	477.636601	477.636601	0	0	1	-360	360;
	3471	865	0.00125	0.006669	0.00801	465.157175	465.157175	465.157175	0	0	1	-360	360;
	2103	2213	0.001281	0.00661	0.00819	215.929128	215.929128	215.929128	0	0	1	-360	360;
	175	2713	0.00226	0.0138	0.02517	460.281625	460.281625	460.281625	0	0	1	-360	360;
	2280	3193	0.00213	0.02677	0.52763	1641.5841	1641.5841	1641.5841	0	0	1	-360	360;
	2280	3038	0.00154	0.01919	0.37948	1375.76588	1375.76588	1375.76588	0	0	1	-360	360;
	4134	545	0.00687	0.04413	0.07852	353.562908	353.562908	353.562908	0	0	1	-360	360;
	4134	2947	0.0045	0.04725	0.09322	475.193301	475.193301	475.193301	0	0	1	-360	360;
	4134	6653	0.00138	0.00837	0.01539	339.727702	339.727702	339.727702	0	0	1	-360	360;
	2791	1402	0.000419	0.001901	0.00296	323.609999	323.609999	323.609999	0	0	1	-360	360;
	2791	1402	0.00043	0.00186	0.00302	353.76677	353.76677	353.76677	0	0	1	-360	360;
	5550	4300	0.003331	0.015099	0.02369	603.009297	603.009297	603.009297	0	0	1	-360	360;
	3037	6001	0.002099	0.01631	0.02457	535.447399	535.447399	535.447399	0	0	1	-360	360;
	3466	5094	0.001331	0.01082	0.01514	593.612575	593.612575	593.612575	0	0	1	-360	360;
	3466	6068	0.00199	0.00818	0.01333	395.060524	395.060524	395.060524	0	0	1	-360	360;
	3466	6068	0.00169	0.008831	0.01183	380.35614	380.35614	380.35614	0	0	1	-360	360;
	1208	4899	0.001771	0.014281	0.0196	257.908393	257.908393	257.908393	0	0	1	-360	360;
	545	558	0.003099	0.01844	0.03554	501.494215	501.494215	501.494215	0	0	1	-360	360;
	3177	1415	0.003409	0.01819	0.03764	446.494431	446.494431	446.494431	0	0	1	-360	360;
	4480	2032	0.00057	0.00369	0.00593	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7309	5358	0.001081	0.00468	0.07341	352.494059	352.494059	352.494059	0	0	1	-360	360;
	4863	5358	0.00105	0.004771	0.07188	358.066552	358.066552	358.066552	0	0	1	-360	360;
	3341	558	0.001979	0.01549	0.02275	605.871372	605.871372	605.871372	0	0	1	-360	360;
	3341	6653	0.001469	0.015151	0.04429	348.592095	348.592095	348.592095	0	0	1	-360	360;
	5296	4624	0.000711	0.003969	0.00692	421.504714	421.504714	421.504714	0	0	1	-360	360;
	5296	6873	0.00051	0.00274	0.01001	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5296	2904	0.000521	0.00306	0.00523	495.490373	495.490373	495.490373	0	0	1	-360	360;
	5296	3169	0.00188	0.011841	0.02656	365.418172	365.418172	365.418172	0	0	1	-360	360;
	5296	1438	0.002469	0.010539	0.02769	215.929128	215.929128	215.929128	0	0	1	-360	360;
	520	4814	0.00045	0.00351	0.00455	523.680539	523.680539	523.680539	0	0	1	-360	360;
	3580	8324	0.00037	0.000659	0.04003	283.755094	283.755094	283.755094	0	0	1	-360	360;
	1905	3475	0.000401	0.0007	0.04458	339.768093	339.768093	339.768093	0	0	1	-360	360;
	1905	5655	0.00014	0.000531	0.1096	363.368879	363.368879	363.368879	0	0	1	-360	360;
	1905	1497	0.000289	0.0013	0.12557	512.03752	512.03752	512.03752	0	0	1	-360	360;
	1905	5675	7e-05	0.000289	0.03049	241.734294	241.734294	241.734294	0	0	1	-360	360;
	1905	6061	0.00051	0.002151	0.16622	396.553864	396.553864	396.553864	0	0	1	-360	360;
	1905	7386	0.00037	0.00157	0.257	534.993186	534.993186	534.993186	0	0	1	-360	360;
	4610	405	0.00022	0.00277	0.05526	1476.1889	1476.1889	1476.1889	0	0	1	-360	360;
	4610	405	0.00025	0.00271	0.06263	1563.93276	1563.93276	1563.93276	0	0	1	-360	360;
	6934	2762	0.000271	0.002081	0.00851	586.035372	586.035372	586.035372	0	0	1	-360	360;
	6934	2762	0.00025	0.00231	0.00789	536.218266	536.218266	536.218266	0	0	1	-360	360;
	2904	2126	0.00095	0.00331	0.07953	364.054009	364.054009	364.054009	0	0	1	-360	360;
	2904	7901	0.00093	0.00336	0.078	218.280213	218.280213	218.280213	0	0	1	-360	360;
	7349	7911	0.000341	0.002211	0.00395	605.829076	605.829076	605.829076	0	0	1	-360	360;
	2396	8340	0.00032	0.00232	0.00379	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3758	7726	3.9e-05	0.000271	0	517.282308	517.282308	517.282308	0	0	1	-360	360;
	548	6823	0.00075	0.0101	0.19131	1622.90431	1622.90431	1622.90431	0	0	1	-360	360;
	548	3644	0.00118	0.01455	0.29942	1557.52999	1557.52999	1557.52999	0	0	1	-360	360;
	1963	8193	0.0017	0.0098	0.02023	515.883088	515.883088	515.883088	0	0	1	-360	360;
	1963	8193	0.001521	0.01111	0.01802	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1963	1865	0.000169	0.001401	0.00178	218.591913	218.591913	218.591913	0	0	1	-360	360;
	1963	623	0.002591	0.01725	0.02962	618.957432	618.957432	618.957432	0	0	1	-360	360;
	1963	3200	0.000169	0.001159	0.00198	445.855789	445.855789	445.855789	0	0	1	-360	360;
	1963	4300	0.00307	0.01339	0.02215	550.866635	550.866635	550.866635	0	0	1	-360	360;
	2393	4449	0.002591	0.013521	0.02462	525.987041	525.987041	525.987041	0	0	1	-360	360;
	8234	8220	0.000581	0.005031	0.03056	564.422531	564.422531	564.422531	0	0	1	-360	360;
	6319	5061	0.001419	0.00595	0.00854	398.682416	398.682416	398.682416	0	0	1	-360	360;
	603	1360	0.00205	0.02664	0.49385	1480.24329	1480.24329	1480.24329	0	0	1	-360	360;
	603	2478	0.00295	0.03721	0.75117	1538.17031	1538.17031	1538.17031	0	0	1	-360	360;
	603	4140	0.00335	0.03648	0.85132	1502.22911	1502.22911	1502.22911	0	0	1	-360	360;
	603	1465	0.00022	0.00251	0.05377	1270.42625	1270.42625	1270.42625	0	0	1	-360	360;
	603	1908	0.00059	0.00651	0.14434	1660.62983	1660.62983	1660.62983	0	0	1	-360	360;
	3243	1983	0.008539	0.041979	0.06261	396.943298	396.943298	396.943298	0	0	1	-360	360;
	3243	2572	0.001349	0.0052	0.00888	408.121053	408.121053	408.121053	0	0	1	-360	360;
	3286	7233	0.00245	0.013909	0.02875	521.510071	521.510071	521.510071	0	0	1	-360	360;
	3286	2250	0.001469	0.008419	0.01811	563.429511	563.429511	563.429511	0	0	1	-360	360;
	4921	6497	0.00084	0.00836	0.20816	1380.21517	1380.21517	1380.21517	0	0	1	-360	360;
	6458	1896	0.002591	0.017789	0.02846	525.183024	525.183024	525.183024	0	0	1	-360	360;
	6458	7144	0.009651	0.05968	0.1067	434.687941	434.687941	434.687941	0	0	1	-360	360;
	6728	6820	0.00143	0.00799	0.01069	381.993898	381.993898	381.993898	0	0	1	-360	360;
	6728	6117	0.001969	0.014909	0.02184	572.3865	572.3865	572.3865	0	0	1	-360	360;
	934	1701	0.01039	0.06406	0.09099	498.481624	498.481624	498.481624	0	0	1	-360	360;
	4334	3021	0.000719	0.005331	0.00854	452.599252	452.599252	452.599252	0	0	1	-360	360;
	5478	6279	0.00052	0.00565	0.13551	1444.27115	1444.27115	1444.27115	0	0	1	-360	360;
	7864	6412	0.002409	0.019479	0.03117	502.097038	502.097038	502.097038	0	0	1	-360	360;
	7864	6412	0.00256	0.019151	0.03249	421.832037	421.832037	421.832037	0	0	1	-360	360;
	7864	2597	0.00656	0.02288	0.03949	392.086039	392.086039	392.086039	0	0	1	-360	360;
	3674	5852	0.0027	0.014901	0.0196	418.322174	418.322174	418.322174	0	0	1	-360	360;
	3674	5209	0.003591	0.01362	0.01742	419.048458	419.048458	419.048458	0	0	1	-360	360;
	6643	2250	0.002211	0.01188	0.01568	362.22001	362.22001	362.22001	0	0	1	-360	360;
	6643	7423	0.00068	0.00357	0.00445	408.790179	408.790179	408.790179	0	0	1	-360	360;
	6777	1444	0.000211	0.000969	0.0019	405.95211	405.95211	405.95211	0	0	1	-360	360;
	8122	3080	9.9e-05	0.000539	0.00093	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1653	5793	0.00645	0.0263	0.04809	423.549053	423.549053	423.549053	0	0	1	-360	360;
	4206	5473	0.00063	0.001901	0.00178	369.861991	369.861991	369.861991	0	0	1	-360	360;
	1571	3237	0.000281	0.000669	0.00101	290.966107	290.966107	290.966107	0	0	1	-360	360;
	6979	3519	0.000219	0.00064	0.00087	277.002867	277.002867	277.002867	0	0	1	-360	360;
	8332	6099	0.00014	0.00105	0.0019	238.797152	238.797152	238.797152	0	0	1	-360	360;
	7588	7936	0.003219	0.020729	0.03725	546.289067	546.289067	546.289067	0	0	1	-360	360;
	7588	7681	0.00051	0.003151	0.00593	573.621868	573.621868	573.621868	0	0	1	-360	360;
	7588	7681	0.000409	0.00331	0.00558	521.764994	521.764994	521.764994	0	0	1	-360	360;
	5108	6021	0.002289	0.017479	0.02547	593.966572	593.966572	593.966572	0	0	1	-360	360;
	5108	1080	0.002831	0.01895	0.03357	552.422848	552.422848	552.422848	0	0	1	-360	360;
	5108	6595	0.006289	0.036521	0.07454	364.589767	364.589767	364.589767	0	0	1	-360	360;
	5108	8166	0.005651	0.032169	0.06495	456.164367	456.164367	456.164367	0	0	1	-360	360;
	5108	6587	0.00414	0.02262	0.03185	489.37755	489.37755	489.37755	0	0	1	-360	360;
	3173	6108	0.01095	0.05814	0.08102	471.398412	471.398412	471.398412	0	0	1	-360	360;
	3173	924	0.009	0.0368	0.06691	510.461492	510.461492	510.461492	0	0	1	-360	360;
	3173	7533	0.000229	0.00181	0.00364	562.286739	562.286739	562.286739	0	0	1	-360	360;
	1809	5958	0.00119	0.00726	0.01394	616.286263	616.286263	616.286263	0	0	1	-360	360;
	8290	679	0.011771	0.051669	0.08658	638.394472	638.394472	638.394472	0	0	1	-360	360;
	6823	1987	0.00209	0.02279	0.53024	1719.74092	1719.74092	1719.74092	0	0	1	-360	360;
	6868	3567	0.002031	0.01443	0.02751	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6868	6252	0.00145	0.008479	0.01811	559.915838	559.915838	559.915838	0	0	1	-360	360;
	2286	5624	0.00124	0.01433	0.80471	1536.49327	1536.49327	1536.49327	0	0	1	-360	360;
	2286	5624	0.00127	0.01402	0.81961	1882.40137	1882.40137	1882.40137	0	0	1	-360	360;
	1349	3319	0.013039	0.052841	0.09722	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2149	5624	0.00235	0.02739	0.47138	980.011217	980.011217	980.011217	0	0	1	-360	360;
	2149	5624	0.0024	0.02681	0.48028	1234.29484	1234.29484	1234.29484	0	0	1	-360	360;
	2149	3770	0.00135	0.02056	0.5311	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2149	1194	0.00087	0.01124	0.39362	974.360748	974.360748	974.360748	0	0	1	-360	360;
	7539	154	0.0012	0.00818	0.01433	647.787002	647.787002	647.787002	0	0	1	-360	360;
	154	5832	0.006961	0.05561	0.0828	497.197482	497.197482	497.197482	0	0	1	-360	360;
	4852	7162	0.00105	0.00763	0.01044	492.936949	492.936949	492.936949	0	0	1	-360	360;
	4852	506	0.000219	0.00182	0.00649	273.047175	273.047175	273.047175	0	0	1	-360	360;
	4852	8024	3.9e-05	0.0002	0.03439	316.639049	316.639049	316.639049	0	0	1	-360	360;
	4852	8024	3.9e-05	0.000169	0.03456	309.608654	309.608654	309.608654	0	0	1	-360	360;
	4852	3860	0.0035	0.022651	0.04027	539.832155	539.832155	539.832155	0	0	1	-360	360;
	3520	1701	0.004281	0.02319	0.03185	474.715081	474.715081	474.715081	0	0	1	-360	360;
	3520	807	0.003729	0.014021	0.02555	433.553933	433.553933	433.553933	0	0	1	-360	360;
	6386	1983	0.00212	0.01582	0.02323	590.356874	590.356874	590.356874	0	0	1	-360	360;
	7866	6497	0.00056	0.00784	0.25472	786.395927	786.395927	786.395927	0	0	1	-360	360;
	1701	7050	0.004401	0.01813	0.0345	352.596181	352.596181	352.596181	0	0	1	-360	360;
	1701	7050	0.005021	0.02144	0.03624	356.565591	356.565591	356.565591	0	0	1	-360	360;
	4446	2657	0.00018	0.000219	0.02462	387.420067	387.420067	387.420067	0	0	1	-360	360;
	3786	2657	0.000151	0.000211	0.02404	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4036	2657	0.000151	0.000211	0.01597	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1891	5430	0.00282	0.02591	0.66429	1483.15112	1483.15112	1483.15112	0	0	1	-360	360;
	3135	5941	0.00131	0.008539	0.03949	548.193942	548.193942	548.193942	0	0	1	-360	360;
	3605	6808	0.002159	0.015969	0.02353	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4804	1681	0.00019	0.000281	0.01975	280.605325	280.605325	280.605325	0	0	1	-360	360;
	5059	2032	0.0002	0.000281	0.02013	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1896	7261	0.00139	0.00774	0.01597	590.386977	590.386977	590.386977	0	0	1	-360	360;
	578	5762	0.01101	0.06739	0.02052	647.787002	647.787002	647.787002	0	0	1	-360	360;
	578	2492	0.00063	0.00463	0.02178	438.563994	438.563994	438.563994	0	0	1	-360	360;
	578	8283	0.002341	0.009599	0.01778	491.232126	491.232126	491.232126	0	0	1	-360	360;
	369	3817	0.00375	0.021659	0.04366	493.700575	493.700575	493.700575	0	0	1	-360	360;
	1360	5247	0.00056	0.00749	0.14347	1295.94189	1295.94189	1295.94189	0	0	1	-360	360;
	527	6001	0.00118	0.00926	0.01336	549.824841	549.824841	549.824841	0	0	1	-360	360;
	527	5473	0.002039	0.013081	0.02074	518.538253	518.538253	518.538253	0	0	1	-360	360;
	1595	6808	0.00155	0.010961	0.03717	265.812918	265.812918	265.812918	0	0	1	-360	360;
	2978	6928	3.9e-05	0.00036	0	434.100361	434.100361	434.100361	0	0	1	-360	360;
	1008	176	0.00024	0.000409	0.12981	394.351388	394.351388	394.351388	0	0	1	-360	360;
	7726	6470	0.000341	0.001659	0.00319	566.951948	566.951948	566.951948	0	0	1	-360	360;
	7726	4861	0.005461	0.024919	0.04048	461.446117	461.446117	461.446117	0	0	1	-360	360;
	7951	6400	0.00158	0.02413	0.72211	974.381151	974.381151	974.381151	0	0	1	-360	360;
	7951	8006	0.00155	0.02458	0.70675	872.353486	872.353486	872.353486	0	0	1	-360	360;
	7951	6251	0.00021	0.00256	0.05545	1418.63836	1418.63836	1418.63836	0	0	1	-360	360;
	7951	6251	0.00024	0.00243	0.06735	1535.21903	1535.21903	1535.21903	0	0	1	-360	360;
	7951	8068	0.00144	0.02266	0.65096	1530.33534	1530.33534	1530.33534	0	0	1	-360	360;
	7951	8068	0.0014	0.02098	0.6368	1577.09503	1577.09503	1577.09503	0	0	1	-360	360;
	7951	7693	0.00086	0.01074	0.22093	2104.80806	2104.80806	2104.80806	0	0	1	-360	360;
	7936	7681	0.003341	0.02611	0.03822	554.079277	554.079277	554.079277	0	0	1	-360	360;
	196	6154	0.00387	0.02063	0.0313	516.222224	516.222224	516.222224	0	0	1	-360	360;
	196	6154	0.00405	0.018271	0.03181	515.374766	515.374766	515.374766	0	0	1	-360	360;
	196	1073	0.00101	0.00462	0.00789	417.640474	417.640474	417.640474	0	0	1	-360	360;
	196	2333	0.00088	0.00389	0.00604	426.348636	426.348636	426.348636	0	0	1	-360	360;
	196	6595	0.005081	0.039	0.0604	516.994615	516.994615	516.994615	0	0	1	-360	360;
	7233	7423	0.00257	0.0208	0.0285	566.961856	566.961856	566.961856	0	0	1	-360	360;
	7233	7423	0.00251	0.01926	0.02701	529.555205	529.555205	529.555205	0	0	1	-360	360;
	7233	947	0.002531	0.01314	0.0182	378.841462	378.841462	378.841462	0	0	1	-360	360;
	6575	7112	0.00132	0.00869	0	270.245687	270.245687	270.245687	0	0	1	-360	360;
	4002	8091	0.00082	0.004021	0.00705	434.432256	434.432256	434.432256	0	0	1	-360	360;
	4002	4415	0.00143	0.007479	0.01022	442.307441	442.307441	442.307441	0	0	1	-360	360;
	2531	1077	0.000659	0.005771	0.0205	363.512916	363.512916	363.512916	0	0	1	-360	360;
	2531	1077	0.000591	0.006539	0.01827	298.275049	298.275049	298.275049	0	0	1	-360	360;
	1188	5707	0.00589	0.025599	0.04328	477.416353	477.416353	477.416353	0	0	1	-360	360;
	1188	6330	0.00614	0.02343	0.04366	456.83235	456.83235	456.83235	0	0	1	-360	360;
	2478	1646	0.00155	0.01994	0.37729	1496.86692	1496.86692	1496.86692	0	0	1	-360	360;
	2478	3038	0.00205	0.03218	0.73298	1598.86431	1598.86431	1598.86431	0	0	1	-360	360;
	4140	3038	0.00233	0.03154	0.83071	1589.44774	1589.44774	1589.44774	0	0	1	-360	360;
	4296	5209	0.002341	0.01982	0.08305	259.809838	259.809838	259.809838	0	0	1	-360	360;
	988	7533	8.1e-05	0.000521	0.00101	548.968619	548.968619	548.968619	0	0	1	-360	360;
	1272	5055	9.9e-05	0.00069	0.00202	595.847821	595.847821	595.847821	0	0	1	-360	360;
	3898	4167	0.00276	0.016419	0.11399	566.049619	566.049619	566.049619	0	0	1	-360	360;
	6394	4465	0.00036	0.001229	0.15202	463.966771	463.966771	463.966771	0	0	1	-360	360;
	6394	3404	0.000229	0.000711	0.0969	520.140573	520.140573	520.140573	0	0	1	-360	360;
	5951	786	0.00013	0.00099	0.08102	514.944178	514.944178	514.944178	0	0	1	-360	360;
	8171	5892	0.001159	0.00689	0.01642	437.868957	437.868957	437.868957	0	0	1	-360	360;
	6250	5441	0.008531	0.03045	0.04181	383.38283	383.38283	383.38283	0	0	1	-360	360;
	4903	7261	0.000781	0.004531	0.00836	528.552659	528.552659	528.552659	0	0	1	-360	360;
	4903	7261	0.00075	0.00474	0.00805	542.558195	542.558195	542.558195	0	0	1	-360	360;
	5335	3673	0.002159	0.008349	0.01597	342.991024	342.991024	342.991024	0	0	1	-360	360;
	7429	7554	0.00162	0.01651	0.39946	1556.78427	1556.78427	1556.78427	0	0	1	-360	360;
	844	5768	0.01236	0.044211	0.07249	402.038714	402.038714	402.038714	0	0	1	-360	360;
	844	5768	0.00825	0.03989	0.07773	431.034423	431.034423	431.034423	0	0	1	-360	360;
	844	3437	0.02187	0.066961	0.12564	343.021127	343.021127	343.021127	0	0	1	-360	360;
	844	3623	0.01212	0.06661	0.09409	352.017364	352.017364	352.017364	0	0	1	-360	360;
	3569	1496	0.00036	0.00534	0.16115	905.377633	905.377633	905.377633	0	0	1	-360	360;
	6170	1050	0.000229	0.00063	0.04453	272.782726	272.782726	272.782726	0	0	1	-360	360;
	6170	6895	0.000219	0.000581	0.04269	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3667	3638	0.000281	0.00063	0.05332	363.773555	363.773555	363.773555	0	0	1	-360	360;
	1453	1287	0.005021	0.035659	0.05597	424.62438	424.62438	424.62438	0	0	1	-360	360;
	2833	3325	0.00174	0.007901	0.01283	374.491763	374.491763	374.491763	0	0	1	-360	360;
	4762	3140	0.003919	0.020581	0.02875	498.957557	498.957557	498.957557	0	0	1	-360	360;
	4762	8264	0.003341	0.016349	0.02467	509.523344	509.523344	509.523344	0	0	1	-360	360;
	4762	1344	0.003271	0.016669	0.02416	467.086056	467.086056	467.086056	0	0	1	-360	360;
	2535	7720	0.000151	0.000659	0.00103	437.813323	437.813323	437.813323	0	0	1	-360	360;
	5809	7144	0.00736	0.035659	0.04732	346.277971	346.277971	346.277971	0	0	1	-360	360;
	5809	3546	0.00944	0.042169	0.06071	350.2832	350.2832	350.2832	0	0	1	-360	360;
	4039	4305	0.00068	0.00652	0.31132	887.849001	887.849001	887.849001	0	0	1	-360	360;
	4039	3164	0.00188	0.02281	0.832	1387.04049	1387.04049	1387.04049	0	0	1	-360	360;
	1897	6396	0.00168	0.00956	0.01916	540.883094	540.883094	540.883094	0	0	1	-360	360;
	1897	6396	0.001669	0.00981	0.02509	524.000241	524.000241	524.000241	0	0	1	-360	360;
	1897	701	0.002229	0.01599	0.02562	599.663668	599.663668	599.663668	0	0	1	-360	360;
	1897	701	0.00218	0.01631	0.02695	614.721286	614.721286	614.721286	0	0	1	-360	360;
	986	5874	0.00238	0.02124	0.39058	1080.25127	1080.25127	1080.25127	0	0	1	-360	360;
	986	5874	0.00194	0.02023	0.48056	1257.92348	1257.92348	1257.92348	0	0	1	-360	360;
	6674	2880	0.013831	0.05955	0.07917	370.667914	370.667914	370.667914	0	0	1	-360	360;
	6674	383	0.00876	0.04875	0.06501	437.639945	437.639945	437.639945	0	0	1	-360	360;
	2594	1698	0.0002	0.00099	0.00186	404.126113	404.126113	404.126113	0	0	1	-360	360;
	1241	2206	0.00105	0.00826	0.01183	477.848846	477.848846	477.848846	0	0	1	-360	360;
	1983	679	0.0172	0.06637	0.12564	486.474702	486.474702	486.474702	0	0	1	-360	360;
	3269	2581	0.00305	0.015781	0.02048	477.069597	477.069597	477.069597	0	0	1	-360	360;
	5943	3283	0.00208	0.02105	0.53266	1485.15001	1485.15001	1485.15001	0	0	1	-360	360;
	7681	1795	0.00614	0.043729	0.06831	482.722491	482.722491	482.722491	0	0	1	-360	360;
	7681	5375	0.000229	0.001849	0.00267	556.787408	556.787408	556.787408	0	0	1	-360	360;
	5206	5493	0.00386	0.025789	0.04344	455.592409	455.592409	455.592409	0	0	1	-360	360;
	5206	2127	0.002031	0.012	0.04412	592.643181	592.643181	592.643181	0	0	1	-360	360;
	2382	6808	0.000591	0.004	0.01278	513.172671	513.172671	513.172671	0	0	1	-360	360;
	1035	2184	9.1e-05	0.00051	0.00091	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6874	5276	0.00732	0.039969	0.0546	434.795398	434.795398	434.795398	0	0	1	-360	360;
	6874	2937	0.004581	0.02031	0.03423	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6952	5276	0.00132	0.00537	0.00958	458.280725	458.280725	458.280725	0	0	1	-360	360;
	1077	1054	0.007521	0.0387	0.06443	491.79913	491.79913	491.79913	0	0	1	-360	360;
	6215	4595	0.005159	0.031281	0.04553	535.914187	535.914187	535.914187	0	0	1	-360	360;
	7531	7990	9.9e-05	0.001159	0.00284	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6300	7990	9.9e-05	0.00113	0.00302	283.453683	283.453683	283.453683	0	0	1	-360	360;
	4101	3968	0.00019	0.000909	0.08276	383.390451	383.390451	383.390451	0	0	1	-360	360;
	6480	7066	0.00107	0.00413	0.00746	269.057569	269.057569	269.057569	0	0	1	-360	360;
	6251	1902	9e-05	0.00058	0.01132	1255.79888	1255.79888	1255.79888	0	0	1	-360	360;
	6251	3666	9e-05	0.00059	0.01109	1177.67102	1177.67102	1177.67102	0	0	1	-360	360;
	6251	8232	8e-05	0.00053	0.01178	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6251	598	8e-05	0.00052	0.01201	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5997	8045	0.000271	0.00151	0.07926	497.671128	497.671128	497.671128	0	0	1	-360	360;
	8108	5624	0.00053	0.00585	0.34309	1910.45956	1910.45956	1910.45956	0	0	1	-360	360;
	8108	5624	0.00047	0.00663	0.30566	1450.87664	1450.87664	1450.87664	0	0	1	-360	360;
	8108	5624	0.00046	0.00676	0.29942	1282.81845	1282.81845	1282.81845	0	0	1	-360	360;
	8108	5624	0.00045	0.00689	0.29319	1277.63727	1277.63727	1277.63727	0	0	1	-360	360;
	8108	7699	0.00028	0.00338	0.06931	1598.811	1598.811	1598.811	0	0	1	-360	360;
	8108	7699	0.0003	0.00319	0.07364	1355.15236	1355.15236	1355.15236	0	0	1	-360	360;
	1497	4107	0.00025	0.00101	0.10191	491.671478	491.671478	491.671478	0	0	1	-360	360;
	176	1803	0.000281	0.00119	0.11988	254.631734	254.631734	254.631734	0	0	1	-360	360;
	2631	5061	0.00018	0.00063	0.00105	333.399204	333.399204	333.399204	0	0	1	-360	360;
	4504	4017	0.00382	0.02868	0.80483	325.516779	325.516779	325.516779	0	0	1	-360	360;
	4504	300	0.00361	0.0235	0.04229	573.380282	573.380282	573.380282	0	0	1	-360	360;
	4504	6267	0.00538	0.03569	0.06023	280.664769	280.664769	280.664769	0	0	1	-360	360;
	3069	7010	6e-05	0.000409	0.00093	523.830673	523.830673	523.830673	0	0	1	-360	360;
	935	2949	9e-05	0.00102	0.02356	767.876077	767.876077	767.876077	0	0	1	-360	360;
	935	2228	0.0014	0.01649	0.35378	1722.59873	1722.59873	1722.59873	0	0	1	-360	360;
	935	2648	0.00034	0.00427	0.08872	1149.59967	1149.59967	1149.59967	0	0	1	-360	360;
	4822	7762	0.0001	0.00109	0.02218	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4822	5829	0.00137	0.01682	0.34656	1808.57538	1808.57538	1808.57538	0	0	1	-360	360;
	4822	7566	0.00131	0.01262	0.33062	1518.96858	1518.96858	1518.96858	0	0	1	-360	360;
	6483	5091	0.00262	0.019669	0.02875	538.882575	538.882575	538.882575	0	0	1	-360	360;
	6483	2197	0.00082	0.00526	0.00906	479.019435	479.019435	479.019435	0	0	1	-360	360;
	6483	2197	0.000841	0.005151	0.00923	471.132057	471.132057	471.132057	0	0	1	-360	360;
	6483	2197	0.000669	0.005081	0.00784	480.371024	480.371024	480.371024	0	0	1	-360	360;
	6483	6337	0.001919	0.012669	0.02173	483.122213	483.122213	483.122213	0	0	1	-360	360;
	37	1409	0.002659	0.02	0.03224	303.44439	303.44439	303.44439	0	0	1	-360	360;
	37	3220	0.003031	0.02343	0.03295	577.874018	577.874018	577.874018	0	0	1	-360	360;
	5654	4282	0.01176	0.059159	0.0827	408.323392	408.323392	408.323392	0	0	1	-360	360;
	5654	7582	0.002469	0.017531	0.02751	512.560703	512.560703	512.560703	0	0	1	-360	360;
	5654	2080	0.000669	0.005729	0.04838	268.815602	268.815602	268.815602	0	0	1	-360	360;
	7517	7679	0.000781	0.00876	0.02265	386.075718	386.075718	386.075718	0	0	1	-360	360;
	7517	4949	0.00062	0.00482	0.02184	554.258372	554.258372	554.258372	0	0	1	-360	360;
	7517	6928	0.001349	0.012599	0.03524	543.338969	543.338969	543.338969	0	0	1	-360	360;
	6207	7679	0.000979	0.0082	0.02823	232.278128	232.278128	232.278128	0	0	1	-360	360;
	6207	798	0.008091	0.028831	0.04443	360.986547	360.986547	360.986547	0	0	1	-360	360;
	6207	2080	0.000711	0.005969	0.01611	530.452961	530.452961	530.452961	0	0	1	-360	360;
	3617	7010	0.00145	0.00857	0.01673	235.331492	235.331492	235.331492	0	0	1	-360	360;
	994	1029	0.00869	0.03768	0.06567	403.390684	403.390684	403.390684	0	0	1	-360	360;
	994	2792	0.001719	0.00837	0.01913	461.329135	461.329135	461.329135	0	0	1	-360	360;
	7720	4873	0.01582	0.067349	0.11276	442.79671	442.79671	442.79671	0	0	1	-360	360;
	7720	4873	0.01457	0.06469	0.10367	413.739272	413.739272	413.739272	0	0	1	-360	360;
	7720	4120	0.021711	0.090539	0.14218	386.852682	386.852682	386.852682	0	0	1	-360	360;
	7720	3056	0.00019	0.000729	0.00107	408.507058	408.507058	408.507058	0	0	1	-360	360;
	846	4888	0.001271	0.00638	0.18923	411.992533	411.992533	411.992533	0	0	1	-360	360;
	832	5958	3.9e-05	0.0002	0.02215	402.111495	402.111495	402.111495	0	0	1	-360	360;
	6044	4031	0.000281	0.00236	0.21362	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3951	1987	0.00019	0.00329	0.12078	1766.17086	1766.17086	1766.17086	0	0	1	-360	360;
	6814	3644	0.00035	0.00324	0.08578	2009.07588	2009.07588	2009.07588	0	0	1	-360	360;
	5958	7745	0.00049	0.00349	0.21656	426.31358	426.31358	426.31358	0	0	1	-360	360;
	5958	7745	0.000469	0.00331	0.20996	435.139487	435.139487	435.139487	0	0	1	-360	360;
	5958	1515	0.00313	0.01324	0.07773	309.786605	309.786605	309.786605	0	0	1	-360	360;
	5958	8274	0.001031	0.006719	0.01481	486.696474	486.696474	486.696474	0	0	1	-360	360;
	5958	1049	0.00038	0.00401	0.01044	235.003406	235.003406	235.003406	0	0	1	-360	360;
	454	3266	0.00037	0.002849	0.00436	500.823946	500.823946	500.823946	0	0	1	-360	360;
	7679	948	0.00032	0.00256	0.00465	478.152544	478.152544	478.152544	0	0	1	-360	360;
	7679	129	0.00037	0.00118	0.15288	395.767755	395.767755	395.767755	0	0	1	-360	360;
	7679	3885	0.00068	0.001151	0.15801	352.787469	352.787469	352.787469	0	0	1	-360	360;
	4490	4711	0.00168	0.009841	0.0213	504.366579	504.366579	504.366579	0	0	1	-360	360;
	4831	8035	0.000599	0.002401	0.25184	402.198375	402.198375	402.198375	0	0	1	-360	360;
	8190	3	0.002169	0.00911	0.01539	416.96487	416.96487	416.96487	0	0	1	-360	360;
	8190	4393	0.002849	0.01282	0.023	458.201085	458.201085	458.201085	0	0	1	-360	360;
	5675	4888	0.001229	0.01207	0.14032	558.268935	558.268935	558.268935	0	0	1	-360	360;
	5675	6808	0.001289	0.01089	0.15	623.369624	623.369624	623.369624	0	0	1	-360	360;
	8068	1498	0.00093	0.01036	0.42887	877.90457	877.90457	877.90457	0	0	1	-360	360;
	8068	3164	0.00083	0.01174	0.38208	843.22971	843.22971	843.22971	0	0	1	-360	360;
	6021	7112	0.00118	0.009469	0.01613	307.332636	307.332636	307.332636	0	0	1	-360	360;
	6021	7112	0.001151	0.009651	0.01578	297.698519	297.698519	297.698519	0	0	1	-360	360;
	4955	7352	0.00101	0.01327	0.24444	1605.52245	1605.52245	1605.52245	0	0	1	-360	360;
	5707	3817	0.00364	0.02486	0.04515	585.492756	585.492756	585.492756	0	0	1	-360	360;
	7617	4318	0.00056	0.00425	0.00651	551.668367	551.668367	551.668367	0	0	1	-360	360;
	7617	4167	0.00369	0.02681	0.04553	534.696728	534.696728	534.696728	0	0	1	-360	360;
	304	7096	0.000531	0.00276	0.00364	403.617409	403.617409	403.617409	0	0	1	-360	360;
	2238	7096	0.000521	0.00281	0.00356	395.824913	395.824913	395.824913	0	0	1	-360	360;
	4779	7096	0.000419	0.00232	0.00364	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1992	7096	0.000409	0.00237	0.00356	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7666	7493	0.00014	0.00193	0.02599	1539.96648	1539.96648	1539.96648	0	0	1	-360	360;
	2426	5434	0.003211	0.01481	0.02091	420.029284	420.029284	420.029284	0	0	1	-360	360;
	1080	5184	0.00381	0.01175	0.02091	378.080503	378.080503	378.080503	0	0	1	-360	360;
	1080	5915	0.003781	0.02676	0.04364	428.42346	428.42346	428.42346	0	0	1	-360	360;
	2144	7833	0.000919	0.005461	0.01045	557.507976	557.507976	557.507976	0	0	1	-360	360;
	2144	8010	0.000271	0.001211	0.11488	384.666591	384.666591	384.666591	0	0	1	-360	360;
	2144	3857	0.002229	0.017401	0.04964	575.610574	575.610574	575.610574	0	0	1	-360	360;
	2144	3857	0.00243	0.017659	0.05406	619.558731	619.558731	619.558731	0	0	1	-360	360;
	2144	3857	0.002531	0.01705	0.02818	593.40414	593.40414	593.40414	0	0	1	-360	360;
	8000	4638	0.012531	0.05451	0.09161	348.590571	348.590571	348.590571	0	0	1	-360	360;
	5247	4305	0.00087	0.01212	0.21989	1355.84279	1355.84279	1355.84279	0	0	1	-360	360;
	5247	818	0.0006	0.00802	0.1541	1740.99418	1740.99418	1740.99418	0	0	1	-360	360;
	5247	7554	0.00072	0.00667	0.18108	1870.4857	1870.4857	1870.4857	0	0	1	-360	360;
	5247	1908	0.00234	0.02626	0.57442	1436.32561	1436.32561	1436.32561	0	0	1	-360	360;
	6143	6337	0.00038	0.002969	0.00445	528.229909	528.229909	528.229909	0	0	1	-360	360;
	7156	3563	0.00024	0.00045	0.04833	347.961455	347.961455	347.961455	0	0	1	-360	360;
	6105	3563	0.000229	0.000409	0.04367	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3917	3563	0.000281	0.000401	0.04542	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4870	5091	0.001211	0.005271	0.51029	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4870	315	0.000281	0.00107	0.11244	443.519184	443.519184	443.519184	0	0	1	-360	360;
	4870	6134	0.00025	0.001091	0.10919	283.210572	283.210572	283.210572	0	0	1	-360	360;
	3509	1803	0.00019	0.00081	0.54263	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3509	1803	0.00045	0.00136	0.18846	609.337414	609.337414	609.337414	0	0	1	-360	360;
	3162	499	0.003461	0.023289	0.03949	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4954	7453	0.0013	0.01609	0.3327	1849.18899	1849.18899	1849.18899	0	0	1	-360	360;
	4954	7453	0.00128	0.0164	0.32305	1719.66127	1719.66127	1719.66127	0	0	1	-360	360;
	4954	2467	0.00089	0.01082	0.07679	1595.64713	1595.64713	1595.64713	0	0	1	-360	360;
	4954	2467	0.00087	0.01102	0.07519	1664.86587	1664.86587	1664.86587	0	0	1	-360	360;
	3446	7778	0.003349	0.02514	0.04275	538.659279	538.659279	538.659279	0	0	1	-360	360;
	443	2405	0.0003	0.00061	0.05537	330.986007	330.986007	330.986007	0	0	1	-360	360;
	132	6412	0.000289	0.001211	0.12193	518.223505	518.223505	518.223505	0	0	1	-360	360;
	6578	7931	0.01301	0.044091	0.07285	451.41304	451.41304	451.41304	0	0	1	-360	360;
	6578	1591	0.02162	0.065969	0.11626	403.948543	403.948543	403.948543	0	0	1	-360	360;
	8107	285	0.006461	0.05275	0.07391	530.840871	530.840871	530.840871	0	0	1	-360	360;
	6305	7235	0.0002	0.000271	0.02052	293.923445	293.923445	293.923445	0	0	1	-360	360;
	3019	7235	0.00018	0.00026	0.01812	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5105	1591	0.00032	0.00188	0.01052	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5091	6337	0.00105	0.006591	0.01208	480.451045	480.451045	480.451045	0	0	1	-360	360;
	16	5592	0.00013	0.00055	0.09918	305.308873	305.308873	305.308873	0	0	1	-360	360;
	3413	4873	0.006151	0.04013	0.07249	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3413	6330	0.007159	0.040591	0.08093	549.135901	549.135901	549.135901	0	0	1	-360	360;
	2082	2852	0.00874	0.048461	0.06824	416.264879	416.264879	416.264879	0	0	1	-360	360;
	2082	4309	0.004159	0.01606	0.03088	355.628967	355.628967	355.628967	0	0	1	-360	360;
	2082	3325	0.002919	0.01912	0.03423	538.050741	538.050741	538.050741	0	0	1	-360	360;
	285	3916	0.00094	0.00394	0.00719	436.962055	436.962055	436.962055	0	0	1	-360	360;
	285	5253	0.000901	0.00395	0.00705	215.929128	215.929128	215.929128	0	0	1	-360	360;
	285	2237	0.003159	0.017789	0.0362	588.924883	588.924883	588.924883	0	0	1	-360	360;
	285	2237	0.002419	0.018331	0.02701	308.880847	308.880847	308.880847	0	0	1	-360	360;
	285	4778	0.00375	0.018281	0.0607	419.420745	419.420745	419.420745	0	0	1	-360	360;
	2911	2405	0.00025	0.001031	0.10861	423.288033	423.288033	423.288033	0	0	1	-360	360;
	4252	4415	0.00264	0.01238	0.13103	429.510218	429.510218	429.510218	0	0	1	-360	360;
	4252	4415	0.001289	0.00795	0.35592	472.714944	472.714944	472.714944	0	0	1	-360	360;
	4252	1049	0.00087	0.003081	0.35051	480.110766	480.110766	480.110766	0	0	1	-360	360;
	7769	4743	0.000401	0.00224	0.00267	435.786893	435.786893	435.786893	0	0	1	-360	360;
	49	200	0.0003	0.00385	0.04615	2012.21737	2012.21737	2012.21737	0	0	1	-360	360;
	6175	770	0.00126	0.008219	0.0122	542.336423	542.336423	542.336423	0	0	1	-360	360;
	134	4638	0.01905	0.055729	0.09514	323.754036	323.754036	323.754036	0	0	1	-360	360;
	134	3237	0.00201	0.00444	0.00732	309.078231	309.078231	309.078231	0	0	1	-360	360;
	6537	2663	0.00101	0.00324	0.00513	247.014902	247.014902	247.014902	0	0	1	-360	360;
	1332	4353	0.00105	0.00677	0.12858	1234.39028	1234.39028	1234.39028	0	0	1	-360	360;
	1332	3485	0.00039	0.00502	0.17675	1159.20975	1159.20975	1159.20975	0	0	1	-360	360;
	1332	5126	0.0004	0.00492	0.18021	980.27054	980.27054	980.27054	0	0	1	-360	360;
	1332	3644	0.00051	0.00617	0.12199	1360.25457	1360.25457	1360.25457	0	0	1	-360	360;
	6837	3166	0.0003	0.00314	0.07364	1607.76026	1607.76026	1607.76026	0	0	1	-360	360;
	6837	3485	0.00051	0.00668	0.15203	1647.37476	1647.37476	1647.37476	0	0	1	-360	360;
	6837	5126	0.0005	0.00678	0.15145	1795.22618	1795.22618	1795.22618	0	0	1	-360	360;
	1147	2200	0.002909	0.017	0.04617	312.695169	312.695169	312.695169	0	0	1	-360	360;
	1147	1052	0.00245	0.011281	0.02052	476.917557	476.917557	476.917557	0	0	1	-360	360;
	1147	4471	0.0027	0.01532	0.02974	469.898213	469.898213	469.898213	0	0	1	-360	360;
	1147	4471	0.00105	0.011151	0.04174	323.558938	323.558938	323.558938	0	0	1	-360	360;
	1147	3489	0.00036	0.002669	0.00379	344.477505	344.477505	344.477505	0	0	1	-360	360;
	1147	501	0.0018	0.011169	0.01637	474.073772	474.073772	474.073772	0	0	1	-360	360;
	1147	2745	0.000711	0.007771	0.02091	269.049948	269.049948	269.049948	0	0	1	-360	360;
	1147	7190	0.00036	0.00257	0.00379	283.414053	283.414053	283.414053	0	0	1	-360	360;
	1415	3735	0.00201	0.01195	0.02258	617.614227	617.614227	617.614227	0	0	1	-360	360;
	1415	7133	0.001091	0.00731	0.01185	571.792823	571.792823	571.792823	0	0	1	-360	360;
	1415	4952	0.00105	0.007591	0.01233	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1415	1681	0.00299	0.02256	0.03276	487.232231	487.232231	487.232231	0	0	1	-360	360;
	2792	2617	0.00036	0.00288	0.00445	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4436	4100	0.00589	0.045599	0.06709	561.660291	561.660291	561.660291	0	0	1	-360	360;
	4436	3256	0.00157	0.0127	0.01781	485.475967	485.475967	485.475967	0	0	1	-360	360;
	4436	3256	0.00168	0.011969	0.01897	250.692046	250.692046	250.692046	0	0	1	-360	360;
	4436	7172	0.00289	0.021031	0.0332	443.547762	443.547762	443.547762	0	0	1	-360	360;
	5779	1005	0.00395	0.025211	0.04328	575.397567	575.397567	575.397567	0	0	1	-360	360;
	5779	8365	0.002771	0.021659	0.06187	494.959949	494.959949	494.959949	0	0	1	-360	360;
	5779	5130	0.002979	0.016901	0.03301	566.903555	566.903555	566.903555	0	0	1	-360	360;
	7684	5358	0.007419	0.034021	0.06023	220.611866	220.611866	220.611866	0	0	1	-360	360;
	4765	6279	0.00056	0.00813	0.7392	920.340681	920.340681	920.340681	0	0	1	-360	360;
	3	1657	0.00361	0.015211	0.02668	439.459083	439.459083	439.459083	0	0	1	-360	360;
	8125	6393	0.000409	0.003331	0.00535	645.433249	645.433249	645.433249	0	0	1	-360	360;
	6211	7974	0.004091	0.019599	0.03488	481.922664	481.922664	481.922664	0	0	1	-360	360;
	6211	6108	0.00294	0.012841	0.0302	507.799469	507.799469	507.799469	0	0	1	-360	360;
	4031	5094	0.020771	0.080479	0.15227	571.341658	571.341658	571.341658	0	0	1	-360	360;
	4031	1795	0.00662	0.0513	0.0758	566.839538	566.839538	566.839538	0	0	1	-360	360;
	4031	1616	0.00607	0.04581	0.06876	548.965952	548.965952	548.965952	0	0	1	-360	360;
	2762	1005	0.000669	0.0042	0.00705	491.441323	491.441323	491.441323	0	0	1	-360	360;
	2762	7691	0.001271	0.007419	0.12617	485.072814	485.072814	485.072814	0	0	1	-360	360;
	2762	5130	0.001599	0.012271	0.01742	588.349496	588.349496	588.349496	0	0	1	-360	360;
	807	4235	0.009599	0.05024	0.06824	430.740632	430.740632	430.740632	0	0	1	-360	360;
	807	471	0.00149	0.009219	0.02771	450.548054	450.548054	450.548054	0	0	1	-360	360;
	807	6183	0.01488	0.06724	0.10762	401.72549	401.72549	401.72549	0	0	1	-360	360;
	807	5369	0.003599	0.024031	0.05233	430.77607	430.77607	430.77607	0	0	1	-360	360;
	807	5369	0.00519	0.024	0.04495	447.119355	447.119355	447.119355	0	0	1	-360	360;
	7833	1255	0.003461	0.013349	0.02173	414.40916	414.40916	414.40916	0	0	1	-360	360;
	3122	3860	0.00105	0.01236	0.02416	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3122	3860	0.00199	0.012331	0.02258	503.398709	503.398709	503.398709	0	0	1	-360	360;
	3122	3325	0.00711	0.044711	0.06055	444.324345	444.324345	444.324345	0	0	1	-360	360;
	5628	6463	0.006729	0.02543	0.04685	336.250229	336.250229	336.250229	0	0	1	-360	360;
	8227	6463	0.005771	0.02999	0.04003	242.309301	242.309301	242.309301	0	0	1	-360	360;
	5441	85	0.0048	0.021771	0.03456	359.174649	359.174649	359.174649	0	0	1	-360	360;
	4593	278	0.00286	0.02212	0.04095	549.653368	549.653368	549.653368	0	0	1	-360	360;
	651	7609	0.00351	0.02175	0.02462	522.491278	522.491278	522.491278	0	0	1	-360	360;
	651	6330	0.008401	0.0448	0.08155	580.324558	580.324558	580.324558	0	0	1	-360	360;
	6531	7990	0.0018	0.01187	0.01975	272.935527	272.935527	272.935527	0	0	1	-360	360;
	6531	7990	0.001831	0.01163	0.02013	267.590522	267.590522	267.590522	0	0	1	-360	360;
	7791	2723	0.00075	0.00538	0.00854	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5968	746	2e-05	0.00033	0.00831	2061.61768	2061.61768	2061.61768	0	0	1	-360	360;
	5968	1532	1e-05	0.00034	0.00814	1835.36459	1835.36459	1835.36459	0	0	1	-360	360;
	5624	864	0.00023	0.00258	0.33685	854.127185	854.127185	854.127185	0	0	1	-360	360;
	5624	1661	0.00023	0.00252	0.34309	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2937	1370	0.002031	0.008531	0.03689	233.326781	233.326781	233.326781	0	0	1	-360	360;
	2937	6551	0.005349	0.020789	0.03833	426.951459	426.951459	426.951459	0	0	1	-360	360;
	2657	7136	0.0002	0.00082	0.08642	512.896028	512.896028	512.896028	0	0	1	-360	360;
	3567	6252	0.002091	0.011219	0.02048	353.414298	353.414298	353.414298	0	0	1	-360	360;
	5493	151	0.00262	0.016979	0.0302	547.989318	547.989318	547.989318	0	0	1	-360	360;
	5493	151	0.002669	0.01663	0.02975	604.444717	604.444717	604.444717	0	0	1	-360	360;
	8010	2769	0.00025	0.00118	0.10419	520.156196	520.156196	520.156196	0	0	1	-360	360;
	1416	1086	0.00056	0.00715	0.14388	1668.81956	1668.81956	1668.81956	0	0	1	-360	360;
	3563	3404	0.0007	0.00455	0.16386	442.938462	442.938462	442.938462	0	0	1	-360	360;
	3563	2443	0.000331	0.001711	0.0061	391.613154	391.613154	391.613154	0	0	1	-360	360;
	3563	2443	0.00011	0.000781	0.00637	557.364701	557.364701	557.364701	0	0	1	-360	360;
	3563	8058	0.00111	0.00664	0.01365	441.296131	441.296131	441.296131	0	0	1	-360	360;
	6821	7453	0.00066	0.00838	0.05678	1566.77149	1566.77149	1566.77149	0	0	1	-360	360;
	6821	7453	0.00064	0.00854	0.05557	1560.76231	1560.76231	1560.76231	0	0	1	-360	360;
	2197	3285	0.003091	0.017091	0.03301	572.922258	572.922258	572.922258	0	0	1	-360	360;
	2197	1287	0.00643	0.03687	0.07109	635.399409	635.399409	635.399409	0	0	1	-360	360;
	8060	3285	0.00275	0.01937	0.02941	225.056066	225.056066	225.056066	0	0	1	-360	360;
	7609	5369	0.00357	0.021289	0.05645	523.489632	523.489632	523.489632	0	0	1	-360	360;
	1751	6330	0.000289	0.002789	0.00356	333.769966	333.769966	333.769966	0	0	1	-360	360;
	7144	6020	0.002531	0.01345	0.0182	383.060079	383.060079	383.060079	0	0	1	-360	360;
	4151	6987	0.00162	0.012159	0.01859	530.693405	530.693405	530.693405	0	0	1	-360	360;
	4151	6987	0.00262	0.01168	0.02013	437.524867	437.524867	437.524867	0	0	1	-360	360;
	8016	1052	0.00099	0.006081	0.01129	261.344712	261.344712	261.344712	0	0	1	-360	360;
	8016	1120	0.000909	0.005979	0.00987	291.366972	291.366972	291.366972	0	0	1	-360	360;
	3080	368	0.00011	0.0008	0.00089	262.890256	262.890256	262.890256	0	0	1	-360	360;
	3080	6555	0.00011	0.00068	0.00103	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3080	5440	0.00207	0.008341	0.01568	505.112677	505.112677	505.112677	0	0	1	-360	360;
	1747	2308	0.00019	0.001031	0.00089	341.744987	341.744987	341.744987	0	0	1	-360	360;
	3623	6667	0.014039	0.0667	0.09478	332.583754	332.583754	332.583754	0	0	1	-360	360;
	2845	6797	0.004961	0.03043	0.0543	457.185584	457.185584	457.185584	0	0	1	-360	360;
	2438	5222	0.00056	0.003531	0.00906	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1093	3857	0.001461	0.00889	0.02155	525.122817	525.122817	525.122817	0	0	1	-360	360;
	6920	6419	6e-05	0.000419	0.00089	514.119584	514.119584	514.119584	0	0	1	-360	360;
	4387	3699	0.00139	0.005169	0	253.936315	253.936315	253.936315	0	0	1	-360	360;
	3770	1194	0.00069	0.00825	0.31259	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6463	4155	0.004969	0.030781	0.05643	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6463	8054	0.00012	0.000969	0.00186	609.674644	609.674644	609.674644	0	0	1	-360	360;
	6061	4888	0.001091	0.010159	0.02962	583.218642	583.218642	583.218642	0	0	1	-360	360;
	7386	4888	0.000969	0.00887	0.02764	558.36115	558.36115	558.36115	0	0	1	-360	360;
	3968	6002	0.00099	0.006789	0.20656	525.394507	525.394507	525.394507	0	0	1	-360	360;
	7425	7953	0.001031	0.00586	0.22751	387.443692	387.443692	387.443692	0	0	1	-360	360;
	6769	4962	0.00401	0.03345	0.0472	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3091	4861	9.1e-05	0.000349	0.00087	372.741976	372.741976	372.741976	0	0	1	-360	360;
	3423	204	0.00125	0.00745	0.01463	608.173303	608.173303	608.173303	0	0	1	-360	360;
	45	3673	0.003711	0.024219	0.04229	520.227834	520.227834	520.227834	0	0	1	-360	360;
	7453	6279	0.00169	0.02156	0.43164	1709.90442	1709.90442	1709.90442	0	0	1	-360	360;
	7453	6279	0.00171	0.02114	0.43805	1801.06226	1801.06226	1801.06226	0	0	1	-360	360;
	5944	3860	0.00088	0.00562	0.01007	543.182357	543.182357	543.182357	0	0	1	-360	360;
	5944	3860	0.0008	0.00618	0.00958	579.429088	579.429088	579.429088	0	0	1	-360	360;
	2669	6510	0.00161	0.01014	0.01812	247.961814	247.961814	247.961814	0	0	1	-360	360;
	3579	2321	0.00032	0.002031	0.00402	546.704794	546.704794	546.704794	0	0	1	-360	360;
	3579	5915	0.00011	0.000659	0.00105	509.395311	509.395311	509.395311	0	0	1	-360	360;
	1760	3485	0.00042	0.00505	0.1898	1115.74161	1115.74161	1115.74161	0	0	1	-360	360;
	1760	5126	0.00043	0.00494	0.19339	1282.65523	1282.65523	1282.65523	0	0	1	-360	360;
	2549	4471	0.00038	0.000711	0.04097	230.929969	230.929969	230.929969	0	0	1	-360	360;
	2549	4471	0.00037	0.000659	0.04008	254.570384	254.570384	254.570384	0	0	1	-360	360;
	3755	1498	0.00248	0.03096	0.61625	1863.23519	1863.23519	1863.23519	0	0	1	-360	360;
	144	3179	0.00375	0.019729	0.02788	511.803936	511.803936	511.803936	0	0	1	-360	360;
	144	2206	0.00518	0.020021	0.03833	450.908909	450.908909	450.908909	0	0	1	-360	360;
	7208	8264	0.000771	0.003711	0.00569	499.351945	499.351945	499.351945	0	0	1	-360	360;
	7208	1344	0.00076	0.003789	0.00558	466.410071	466.410071	466.410071	0	0	1	-360	360;
	2098	3220	0.000781	0.00574	0.00871	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3214	6878	0.00089	0.00562	0.00987	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1843	7745	0.000331	0.001401	0.14375	343.170118	343.170118	343.170118	0	0	1	-360	360;
	4706	7745	0.000401	0.00137	0.16389	285.980052	285.980052	285.980052	0	0	1	-360	360;
	7210	4743	0.00051	0.003349	0.00455	373.358898	373.358898	373.358898	0	0	1	-360	360;
	1879	8283	0.000461	0.00168	0.19747	253.946223	253.946223	253.946223	0	0	1	-360	360;
	1879	8283	0.000469	0.00164	0.20134	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4235	6183	0.006081	0.03114	0.04269	469.423043	469.423043	469.423043	0	0	1	-360	360;
	6820	704	0.005831	0.02218	0.03797	512.94366	512.94366	512.94366	0	0	1	-360	360;
	8163	3121	0.00013	0.000461	0.04458	340.200205	340.200205	340.200205	0	0	1	-360	360;
	7071	1844	0.00013	0.000461	0.04186	215.929128	215.929128	215.929128	0	0	1	-360	360;
	787	3673	0.00262	0.020289	0.02962	528.662021	528.662021	528.662021	0	0	1	-360	360;
	787	2608	0.00164	0.01194	0.01897	565.655993	565.655993	565.655993	0	0	1	-360	360;
	6667	8315	0.006479	0.02845	0.04553	531.941728	531.941728	531.941728	0	0	1	-360	360;
	7235	1836	0.00294	0.014419	0.18523	505.390082	505.390082	505.390082	0	0	1	-360	360;
	7235	1836	0.00238	0.01463	0.1682	513.540386	513.540386	513.540386	0	0	1	-360	360;
	7235	1599	0.002219	0.00993	0.19393	485.859685	485.859685	485.859685	0	0	1	-360	360;
	1275	2093	0.008651	0.04886	0.06882	404.672921	404.672921	404.672921	0	0	1	-360	360;
	4100	7172	0.00464	0.030979	0.05233	239.809605	239.809605	239.809605	0	0	1	-360	360;
	5297	405	0.00092	0.00889	0.23317	1563.4332	1563.4332	1563.4332	0	0	1	-360	360;
	5297	405	0.00093	0.00869	0.23749	1460.37944	1460.37944	1460.37944	0	0	1	-360	360;
	5297	1661	0.0022	0.03485	0.78184	1710.60407	1710.60407	1710.60407	0	0	1	-360	360;
	5297	2228	0.00074	0.00941	0.18974	1860.99607	1860.99607	1860.99607	0	0	1	-360	360;
	5297	5829	0.00087	0.00884	0.22347	1683.00398	1683.00398	1683.00398	0	0	1	-360	360;
	5297	1580	0.00139	0.01662	0.4928	2148.20709	2148.20709	2148.20709	0	0	1	-360	360;
	8365	538	0.002781	0.022979	0.03384	412.593832	412.593832	412.593832	0	0	1	-360	360;
	5762	4898	0.000521	0.00407	0.01742	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5620	6987	0.008479	0.032841	0.06282	222.503023	222.503023	222.503023	0	0	1	-360	360;
	2267	6793	0.005669	0.02901	0.04182	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3673	2603	0.001031	0.004099	0.00732	302.678477	302.678477	302.678477	0	0	1	-360	360;
	3673	7297	0.00301	0.018719	0.03322	489.924739	489.924739	489.924739	0	0	1	-360	360;
	1657	2597	0.00257	0.01713	0.02962	439.123377	439.123377	439.123377	0	0	1	-360	360;
	4873	4918	0.010461	0.03836	0.06141	350.440193	350.440193	350.440193	0	0	1	-360	360;
	4873	3732	0.001719	0.011531	0.01975	520.129904	520.129904	520.129904	0	0	1	-360	360;
	4873	1420	9.1e-05	0.00056	0.00089	409.033671	409.033671	409.033671	0	0	1	-360	360;
	3546	3519	0.00476	0.025289	0.03457	347.351393	347.351393	347.351393	0	0	1	-360	360;
	113	1418	0.000219	0.00164	0.0104	432.49423	432.49423	432.49423	0	0	1	-360	360;
	178	7251	0.002159	0.01163	0.01394	250.326236	250.326236	250.326236	0	0	1	-360	360;
	178	1418	0.00295	0.01539	0.01473	406.094623	406.094623	406.094623	0	0	1	-360	360;
	2713	2852	0.007789	0.033039	0.06273	417.564264	417.564264	417.564264	0	0	1	-360	360;
	6633	2793	0.00026	0.0013	0.00175	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5884	8166	0.00162	0.013229	0.01781	456.8796	456.8796	456.8796	0	0	1	-360	360;
	7691	538	0.007919	0.04194	0.05924	442.949131	442.949131	442.949131	0	0	1	-360	360;
	618	7683	0.001229	0.00401	0.00746	402.473875	402.473875	402.473875	0	0	1	-360	360;
	4678	4979	0.008021	0.042831	0.05823	442.663343	442.663343	442.663343	0	0	1	-360	360;
	6020	704	0.007711	0.03006	0.0513	450.90929	450.90929	450.90929	0	0	1	-360	360;
	7974	715	0.00261	0.015469	0.03032	432.991502	432.991502	432.991502	0	0	1	-360	360;
	5743	1420	0.00139	0.00536	0.00906	357.510598	357.510598	357.510598	0	0	1	-360	360;
	5743	1420	0.001419	0.00525	0.00923	364.323793	364.323793	364.323793	0	0	1	-360	360;
	1723	6987	0.006289	0.032271	0.05537	288.265597	288.265597	288.265597	0	0	1	-360	360;
	1723	1944	0.011031	0.05625	0.07992	374.236839	374.236839	374.236839	0	0	1	-360	360;
	151	2127	0.001841	0.01324	0.02087	574.699481	574.699481	574.699481	0	0	1	-360	360;
	6012	7	0.000599	0.00186	0.24881	566.621577	566.621577	566.621577	0	0	1	-360	360;
	4080	3860	0.00036	0.001591	0.15318	551.484319	551.484319	551.484319	0	0	1	-360	360;
	2769	3857	0.003849	0.027651	0.02708	566.138023	566.138023	566.138023	0	0	1	-360	360;
	770	4056	0.00186	0.00874	0.01811	594.792691	594.792691	594.792691	0	0	1	-360	360;
	3166	1987	0.00025	0.00285	0.05944	1581.3212	1581.3212	1581.3212	0	0	1	-360	360;
	1052	7190	0.00207	0.008841	0.01568	433.366456	433.366456	433.366456	0	0	1	-360	360;
	1052	1120	6e-05	0.00039	0.00101	557.764804	557.764804	557.764804	0	0	1	-360	360;
	8091	4415	0.002159	0.008909	0.01568	467.381751	467.381751	467.381751	0	0	1	-360	360;
	3681	1120	0.000219	0.00039	0.02353	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4471	2405	0.00011	0.000521	0.09083	411.803532	411.803532	411.803532	0	0	1	-360	360;
	4471	2405	0.00011	0.00049	0.0846	410.585311	410.585311	410.585311	0	0	1	-360	360;
	4471	501	0.000771	0.003789	0.00705	453.649048	453.649048	453.649048	0	0	1	-360	360;
	4471	2745	0.000229	0.002849	0.00637	274.540515	274.540515	274.540515	0	0	1	-360	360;
	7251	4778	0.00445	0.027599	0.04879	377.153024	377.153024	377.153024	0	0	1	-360	360;
	1559	3038	0.00218	0.02	0.5337	1513.3734	1513.3734	1513.3734	0	0	1	-360	360;
	1559	7675	0.00074	0.00865	0.18397	1227.31551	1227.31551	1227.31551	0	0	1	-360	360;
	6987	1121	0.002531	0.019419	0.02875	404.155454	404.155454	404.155454	0	0	1	-360	360;
	6987	7290	0.003	0.01631	0.02467	237.990086	237.990086	237.990086	0	0	1	-360	360;
	7871	3025	0.008461	0.025081	0.04578	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7423	3313	0.00382	0.013669	0.02555	505.518115	505.518115	505.518115	0	0	1	-360	360;
	2852	7683	0.002581	0.01814	0.02881	525.828143	525.828143	525.828143	0	0	1	-360	360;
	6117	5537	0.00201	0.012219	0.02258	542.453787	542.453787	542.453787	0	0	1	-360	360;
	7745	3926	0.00087	0.001599	0.21745	378.524808	378.524808	378.524808	0	0	1	-360	360;
	698	4758	0.00455	0.02394	0.08364	424.425852	424.425852	424.425852	0	0	1	-360	360;
	1795	5375	0.005651	0.041771	0.0664	463.267161	463.267161	463.267161	0	0	1	-360	360;
	8186	3643	0.00069	0.001271	0.16193	359.753465	359.753465	359.753465	0	0	1	-360	360;
	1754	7618	0.00068	0.00269	0.28754	404.417236	404.417236	404.417236	0	0	1	-360	360;
	1754	1845	0.00036	0.00261	0.01518	477.861802	477.861802	477.861802	0	0	1	-360	360;
	1754	7519	0.001669	0.012031	0.01859	419.733207	419.733207	419.733207	0	0	1	-360	360;
	4309	3325	0.009669	0.03513	0.05839	373.363089	373.363089	373.363089	0	0	1	-360	360;
	4309	7683	0.001099	0.00862	0.01307	312.429957	312.429957	312.429957	0	0	1	-360	360;
	7254	1498	0.0005	0.00711	0.19757	778.447753	778.447753	778.447753	0	0	1	-360	360;
	6108	924	0.003919	0.017781	0.02867	513.375391	513.375391	513.375391	0	0	1	-360	360;
	6108	8166	0.00161	0.019781	0.08342	259.388015	259.388015	259.388015	0	0	1	-360	360;
	6108	715	0.0007	0.01256	0.03036	622.37508	622.37508	622.37508	0	0	1	-360	360;
	4552	4898	0.000409	0.0018	0.14985	331.596451	331.596451	331.596451	0	0	1	-360	360;
	1558	3558	2e-05	0.00043	0.01109	2049.22021	2049.22021	2049.22021	0	0	1	-360	360;
	8302	3558	3e-05	0.0004	0.01224	1834.60439	1834.60439	1834.60439	0	0	1	-360	360;
	255	8309	0.00968	0.056659	0.07481	476.271675	476.271675	476.271675	0	0	1	-360	360;
	7297	3325	0.003969	0.02925	0.04743	367.464417	367.464417	367.464417	0	0	1	-360	360;
	5184	8166	0.00219	0.01468	0.02567	526.450781	526.450781	526.450781	0	0	1	-360	360;
	5184	7673	0.000401	0.003159	0.00455	494.560608	494.560608	494.560608	0	0	1	-360	360;
	5184	6242	0.000531	0.003	0.00426	493.552727	493.552727	493.552727	0	0	1	-360	360;
	4743	6587	0.00136	0.00576	0.01065	482.379164	482.379164	482.379164	0	0	1	-360	360;
	7990	8166	0.000849	0.006289	0.00929	466.170771	466.170771	466.170771	0	0	1	-360	360;
	7990	8166	0.000841	0.006409	0.0091	470.065876	470.065876	470.065876	0	0	1	-360	360;
	7990	8166	0.00045	0.00506	0.01742	246.977559	246.977559	246.977559	0	0	1	-360	360;
	197	1904	0.003669	0.02988	0.04453	508.503651	508.503651	508.503651	0	0	1	-360	360;
	471	5369	0.002	0.01568	0.02366	522.926057	522.926057	522.926057	0	0	1	-360	360;
	2405	1940	0.00031	0.0008	0.19236	419.490858	419.490858	419.490858	0	0	1	-360	360;
	1105	4758	0.00043	0.00038	0.09851	288.866134	288.866134	288.866134	0	0	1	-360	360;
	8155	5440	0.00043	0.002229	0.00348	376.199253	376.199253	376.199253	0	0	1	-360	360;
	4740	5440	0.00049	0.00218	0.00395	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3325	3699	0.00282	0.01844	0.02013	647.787002	647.787002	647.787002	0	0	1	-360	360;
	798	3025	0.00462	0.014849	0.02614	230.435746	230.435746	230.435746	0	0	1	-360	360;
	4676	3038	0.00263	0.04076	1.13395	1577.95725	1577.95725	1577.95725	0	0	1	-360	360;
	3179	2321	0.01062	0.056719	0.07734	466.875334	466.875334	466.875334	0	0	1	-360	360;
	1734	7433	0.0001	0.00105	0.02356	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8283	8045	0.00239	0.017341	0.02802	481.484456	481.484456	481.484456	0	0	1	-360	360;
	3183	2273	0.001341	0.010581	0.01547	539.026232	539.026232	539.026232	0	0	1	-360	360;
	3183	5575	0.00101	0.007599	0.01116	577.685017	577.685017	577.685017	0	0	1	-360	360;
	362	1444	0.00011	0.000289	0	513.765206	513.765206	513.765206	0	0	1	-360	360;
	6644	6279	0.00146	0.01769	0.54958	2069.51714	2069.51714	2069.51714	0	0	1	-360	360;
	6644	6279	0.0013	0.02004	0.48963	1765.76608	1765.76608	1765.76608	0	0	1	-360	360;
	6183	5061	0.007909	0.028349	0.04925	425.601014	425.601014	425.601014	0	0	1	-360	360;
	6703	7765	0.00012	0.00049	0.05628	255.689532	255.689532	255.689532	0	0	1	-360	360;
	6703	7765	0.00012	0.000479	0.05739	274.018474	274.018474	274.018474	0	0	1	-360	360;
	6279	389	0.00076	0.01018	0.19395	1643.72318	1643.72318	1643.72318	0	0	1	-360	360;
	6279	389	0.0009	0.00865	0.22769	1805.08505	1805.08505	1805.08505	0	0	1	-360	360;
	6279	389	0.00096	0.01294	0.24709	1661.78296	1661.78296	1661.78296	0	0	1	-360	360;
	6279	389	0.0004	0.00756	0.22925	1723.41882	1723.41882	1723.41882	0	0	1	-360	360;
	6279	389	0.00039	0.00771	0.22457	1547.16498	1547.16498	1547.16498	0	0	1	-360	360;
	6279	7433	0.00163	0.02291	0.74435	1668.65567	1668.65567	1668.65567	0	0	1	-360	360;
	2679	2840	0.00636	0.03855	0.08294	502.920871	502.920871	502.920871	0	0	1	-360	360;
	2679	2840	0.00639	0.03863	0.09182	529.349056	529.349056	529.349056	0	0	1	-360	360;
	864	1580	0.00113	0.01362	0.40548	812.853419	812.853419	812.853419	0	0	1	-360	360;
	1616	4899	0.00861	0.05318	0.09851	647.787002	647.787002	647.787002	0	0	1	-360	360;
	4247	4449	0.0012	0.004669	0.00851	446.233411	446.233411	446.233411	0	0	1	-360	360;
	4247	6419	0.00018	0.001091	0.00206	522.229496	522.229496	522.229496	0	0	1	-360	360;
	1121	7290	0.001401	0.00612	0.01045	460.583798	460.583798	460.583798	0	0	1	-360	360;
	102	204	0.00068	0.00387	0.00639	591.707319	591.707319	591.707319	0	0	1	-360	360;
	5193	7778	0.0043	0.033591	0.04822	530.276154	530.276154	530.276154	0	0	1	-360	360;
	3256	4962	0.003039	0.01911	0.03488	608.944169	608.944169	608.944169	0	0	1	-360	360;
	2952	110	6e-05	0.00111	0.0041	955.335694	955.335694	955.335694	0	0	1	-360	360;
	8027	4349	5e-05	0.001091	0.00348	956.475418	956.475418	956.475418	0	0	1	-360	360;
	2174	3002	3.9e-05	0.00089	0.00261	956.550866	956.550866	956.550866	0	0	1	-360	360;
	1733	4076	3.9e-05	0.000901	0.00273	956.641937	956.641937	956.641937	0	0	1	-360	360;
	614	4056	0.000169	0.00111	0.00178	610.364728	610.364728	610.364728	0	0	1	-360	360;
	911	4415	0.000331	0.0027	0.01232	271.829335	271.829335	271.829335	0	0	1	-360	360;
	5831	1836	0.00026	0.00369	0.00651	364.071537	364.071537	364.071537	0	0	1	-360	360;
	4449	6419	0.000341	0.00274	0.00445	266.040787	266.040787	266.040787	0	0	1	-360	360;
	4853	6793	0.00038	0.00506	0.00854	529.962548	529.962548	529.962548	0	0	1	-360	360;
	2321	5915	0.00012	0.000831	0.0014	494.08658	494.08658	494.08658	0	0	1	-360	360;
	1014	538	0.003031	0.0123	0.02196	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4318	7066	0.001521	0.00774	0.01069	510.246961	510.246961	510.246961	0	0	1	-360	360;
	3558	2748	0.00059	0.00552	0.15248	1743.31032	1743.31032	1743.31032	0	0	1	-360	360;
	3558	1086	0.00053	0.00625	0.13585	1664.78558	1664.78558	1664.78558	0	0	1	-360	360;
	5440	7987	0.00143	0.01076	0.01672	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1646	7693	0.00056	0.00593	0.13551	2151.29592	2151.29592	2151.29592	0	0	1	-360	360;
	2237	6510	0.00205	0.01461	0.02277	520.681666	520.681666	520.681666	0	0	1	-360	360;
	1944	2093	0.01068	0.054521	0.07806	252.776014	252.776014	252.776014	0	0	1	-360	360;
	8166	7673	0.00194	0.011039	0.02196	540.488325	540.488325	540.488325	0	0	1	-360	360;
	8166	6242	0.00131	0.010099	0.01568	542.276598	542.276598	542.276598	0	0	1	-360	360;
	8166	2723	0.001289	0.014229	0.02761	423.655367	423.655367	423.655367	0	0	1	-360	360;
	730	2929	0.00455	0.027031	0.06482	501.735039	501.735039	501.735039	0	0	1	-360	360;
	730	3021	0.00239	0.01414	0.02614	615.617518	615.617518	615.617518	0	0	1	-360	360;
	730	7533	0.004039	0.029669	0.0446	592.903058	592.903058	592.903058	0	0	1	-360	360;
	730	1904	0.001099	0.007771	0.01233	640.002126	640.002126	640.002126	0	0	1	-360	360;
	4888	1845	0.00082	0.00762	0.06097	516.058372	516.058372	516.058372	0	0	1	-360	360;
	4888	3744	0.000719	0.0033	0.00474	286.100464	286.100464	286.100464	0	0	1	-360	360;
	4888	5010	0.000599	0.00361	0.00705	401.781124	401.781124	401.781124	0	0	1	-360	360;
	4888	7519	0.000521	0.004039	0.00624	490.626635	490.626635	490.626635	0	0	1	-360	360;
	4888	3121	0.000901	0.0088	0.02818	552.409511	552.409511	552.409511	0	0	1	-360	360;
	6808	6338	0.000781	0.00306	0.00513	261.407967	261.407967	261.407967	0	0	1	-360	360;
	6808	2184	0.000419	0.00399	0.00888	275.677571	275.677571	275.677571	0	0	1	-360	360;
	6808	1844	0.000841	0.009841	0.02547	384.409763	384.409763	384.409763	0	0	1	-360	360;
	5876	5575	0.004021	0.03064	0.0474	515.323705	515.323705	515.323705	0	0	1	-360	360;
	4415	2443	0.0023	0.009031	0.01045	440.854493	440.854493	440.854493	0	0	1	-360	360;
	4415	2443	0.002081	0.009469	0.02114	492.750234	492.750234	492.750234	0	0	1	-360	360;
	4415	4993	0.00164	0.008831	0.01157	374.944451	374.944451	374.944451	0	0	1	-360	360;
	4415	8058	0.00094	0.00488	0.00697	430.17782	430.17782	430.17782	0	0	1	-360	360;
	4415	8274	0.000901	0.00551	0.01007	564.77767	564.77767	564.77767	0	0	1	-360	360;
	4415	1049	0.00051	0.005961	0.01547	227.473073	227.473073	227.473073	0	0	1	-360	360;
	1836	1599	0.00113	0.004349	0.00851	542.544477	542.544477	542.544477	0	0	1	-360	360;
	1344	6001	0.002081	0.013031	0.02315	510.63449	510.63449	510.63449	0	0	1	-360	360;
	734	4180	0.00116	0.01438	0.34466	2143.71238	2143.71238	2143.71238	0	0	1	-360	360;
	7851	4300	0.00013	0.001031	0.00175	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3732	557	0.00036	0.002659	0.00372	519.373898	519.373898	519.373898	0	0	1	-360	360;
	3699	4898	0.000771	0.004539	0.0213	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2995	3715	0.011902	0.04088	0.01812	130.951615	130.951615	130.951615	0	0	1	-360	360;
	156	2995	0.002231	0.010138	0.00339	153.817283	153.817283	153.817283	0	0	1	-360	360;
	1040	146	0.00023	0.0025	0.05403	1465.14926	1465.14926	1465.14926	0	0	1	-360	360;
	1040	146	0.00023	0.00254	0.0529	1304.93065	1304.93065	1304.93065	0	0	1	-360	360;
	4534	1040	0.00179	0.01478	0.40555	1440.15227	1440.15227	1440.15227	0	0	1	-360	360;
	4534	1040	0.00183	0.01446	0.41306	1318.22192	1318.22192	1318.22192	0	0	1	-360	360;
	1040	1397	0.00031	0.00312	0.07001	1216.89324	1216.89324	1216.89324	0	0	1	-360	360;
	965	1040	0.00031	0.00289	0.07025	1423.14952	1423.14952	1423.14952	0	0	1	-360	360;
	826	1040	0.00021	0.00194	0.0473	1233.78146	1233.78146	1233.78146	0	0	1	-360	360;
	5965	6916	0.001489	0.006249	0.00226	145.836253	145.836253	145.836253	0	0	1	-360	360;
	7022	6916	0.010209	0.042502	0.01453	139.322616	139.322616	139.322616	0	0	1	-360	360;
	5965	2163	0.001462	0.006369	0.00222	105.302627	105.302627	105.302627	0	0	1	-360	360;
	2163	8020	0.0048	0.01772	0.00684	106.55386	106.55386	106.55386	0	0	1	-360	360;
	5965	5225	0.00328	0.01472	0.005	228.07212	228.07212	228.07212	0	0	1	-360	360;
	416	5965	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8020	5965	0.002031	0.007591	0.2538	145.824562	145.824562	145.824562	0	0	1	-360	360;
	3766	6264	0.00372	0.014431	0.00567	224.7232	224.7232	224.7232	0	0	1	-360	360;
	3503	3766	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3766	4156	0.00108	0.003698	0.00164	147.684525	147.684525	147.684525	0	0	1	-360	360;
	3766	8180	0.000431	0.001658	0.02671	127.290925	127.290925	127.290925	0	0	1	-360	360;
	6264	2953	0.000782	0.002662	0.08238	211.722167	211.722167	211.722167	0	0	1	-360	360;
	5225	568	0.001729	0.005498	0.18189	171.295062	171.295062	171.295062	0	0	1	-360	360;
	7222	1765	0.00221	0.01891	0.49862	1128.91836	1128.91836	1128.91836	0	0	1	-360	360;
	7406	1765	0.00335	0.03107	0.75602	1105.40359	1105.40359	1105.40359	0	0	1	-360	360;
	1765	392	0.00109	0.01324	0.44736	1807.50254	1807.50254	1807.50254	0	0	1	-360	360;
	3133	2606	0.027218	0.118858	0.04142	105.90668	105.90668	105.90668	0	0	1	-360	360;
	5766	3133	0.002782	0.010778	0.00423	124.09763	124.09763	124.09763	0	0	1	-360	360;
	3133	2334	0.03336	0.110049	0.05079	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3133	2334	0.02972	0.124729	0.04525	130.277154	130.277154	130.277154	0	0	1	-360	360;
	3133	3322	0.00244	0.009351	0.00347	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3133	3322	0.002151	0.009649	0.00328	138.266498	138.266498	138.266498	0	0	1	-360	360;
	3133	3322	0.00244	0.009458	0.00372	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	5271	0.017102	0.080911	0.02603	129.048004	129.048004	129.048004	0	0	1	-360	360;
	2223	5847	0.030871	0.129529	0.04698	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	1718	0.008778	0.041551	0.01337	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	1718	0.008591	0.038471	0.01308	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	1718	0.01032	0.038049	0.01468	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	631	0.004151	0.017409	0.00632	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5847	631	0.004062	0.017751	0.00619	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2569	5552	0.000338	0.00156	0.00047	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7752	5552	0.002689	0.010018	0.0041	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5552	3940	0.020471	0.070289	0.03116	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3639	2672	4e-05	0.000271	7e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7739	3639	0.023289	0.097751	0.03546	118.391475	118.391475	118.391475	0	0	1	-360	360;
	4879	2672	0.005969	0.02292	0.00849	104.944872	104.944872	104.944872	0	0	1	-360	360;
	138	2672	0.00428	0.019138	0.0065	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1766	7752	0.01588	0.054529	0.02417	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5739	3878	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3878	2819	0.00368	0.007311	0.00262	130.122828	130.122828	130.122828	0	0	1	-360	360;
	6421	3878	0.006369	0.0324	0.00907	103.923048	103.923048	103.923048	0	0	1	-360	360;
	923	3878	0.000791	0.003569	0.00119	111.03762	111.03762	111.03762	0	0	1	-360	360;
	3802	3878	0.000809	0.003431	0.00124	125.219479	125.219479	125.219479	0	0	1	-360	360;
	2819	5366	0.008871	0.004431	0.07735	252.567043	252.567043	252.567043	0	0	1	-360	360;
	8301	6417	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6076	3144	0	0.000271	0	155.063581	155.063581	155.063581	0	0	1	-360	360;
	7051	3426	0	0.000271	0	107.487349	107.487349	107.487349	0	0	1	-360	360;
	5991	138	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	567	4071	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	883	7173	0	0.000271	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7932	4072	0.00096	0.003391	0.09439	129.49877	129.49877	129.49877	0	0	1	-360	360;
	1777	4072	0	0.000271	0	106.472021	106.472021	106.472021	0	0	1	-360	360;
	4072	5808	0.004471	0.021831	0.00635	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4072	3163	0.00624	0.03	0.00887	136.333789	136.333789	136.333789	0	0	1	-360	360;
	3015	2210	0.0008	0.0038	0.00122	140.23532	140.23532	140.23532	0	0	1	-360	360;
	156	3015	0.010058	0.035969	0.01532	142.253766	142.253766	142.253766	0	0	1	-360	360;
	8088	2210	0.000782	0.00352	0.0012	118.142839	118.142839	118.142839	0	0	1	-360	360;
	5386	8088	0.008471	0.040089	0.01289	137.884841	137.884841	137.884841	0	0	1	-360	360;
	7038	5337	0.00038	0.00312	0.08549	1467.88729	1467.88729	1467.88729	0	0	1	-360	360;
	7038	5337	0.00038	0.00305	0.08708	1215.64072	1215.64072	1215.64072	0	0	1	-360	360;
	547	7038	0.00255	0.02375	0.57789	1360.47572	1360.47572	1360.47572	0	0	1	-360	360;
	547	7038	0.0026	0.02326	0.58922	1528.51481	1528.51481	1528.51481	0	0	1	-360	360;
	7038	5512	0.00268	0.03039	0.60533	1353.05079	1353.05079	1353.05079	0	0	1	-360	360;
	7038	5512	0.00262	0.02814	0.59217	1216.08894	1216.08894	1216.08894	0	0	1	-360	360;
	2471	7418	0.000258	0.000929	0.06197	112.059704	112.059704	112.059704	0	0	1	-360	360;
	2879	7418	0.003049	0.013911	0.00465	142.987722	142.987722	142.987722	0	0	1	-360	360;
	2471	5235	0.000258	0.000911	0.06312	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7139	5235	0.000871	0.003991	0.00133	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5025	3003	0.024218	0.1016	0.03686	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3003	6237	0	0.000271	0	139.750519	139.750519	139.750519	0	0	1	-360	360;
	1937	4118	0	0.000222	0	136.960186	136.960186	136.960186	0	0	1	-360	360;
	5025	1937	0.02372	0.103591	0.03611	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1937	6378	0	0.000271	0	135.708432	135.708432	135.708432	0	0	1	-360	360;
	6901	4118	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	146	2283	0.00038	0.00396	0.08553	906.352396	906.352396	906.352396	0	0	1	-360	360;
	146	2283	0.00038	0.00404	0.08375	930.120567	930.120567	930.120567	0	0	1	-360	360;
	146	2283	0.00042	0.0036	0.09515	1011.24972	1011.24972	1011.24972	0	0	1	-360	360;
	146	2283	0.00043	0.00352	0.09694	977.12905	977.12905	977.12905	0	0	1	-360	360;
	4534	146	0.00177	0.01845	0.39814	1396.94872	1396.94872	1396.94872	0	0	1	-360	360;
	7173	8084	0.007831	0.035089	0.01192	133.396924	133.396924	133.396924	0	0	1	-360	360;
	7173	6654	0.002089	0.007182	0.00319	270.014683	270.014683	270.014683	0	0	1	-360	360;
	4672	8084	0.008871	0.034391	0.01351	248.201495	248.201495	248.201495	0	0	1	-360	360;
	4672	4416	0.002129	0.007031	0.00325	167.103066	167.103066	167.103066	0	0	1	-360	360;
	6802	4672	0.007222	0.032831	0.01099	229.769703	229.769703	229.769703	0	0	1	-360	360;
	4672	6978	0.006049	0.021431	0.0086	105.780673	105.780673	105.780673	0	0	1	-360	360;
	2568	7932	0.0114	0.005622	0.41898	143.162053	143.162053	143.162053	0	0	1	-360	360;
	2591	2295	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2295	8081	0.00976	0.043991	0.01389	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3327	1254	0.017658	0.0684	0.02688	156.751551	156.751551	156.751551	0	0	1	-360	360;
	3327	1254	0.018	0.067	0.0274	159.080466	159.080466	159.080466	0	0	1	-360	360;
	1254	7035	0.003378	0.009631	0.00241	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1254	4787	0.016409	0.07168	0.02498	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1254	1247	0.0182	0.04612	0.013	210.810502	210.810502	210.810502	0	0	1	-360	360;
	1254	2827	0.004818	0.017231	0.00734	144.936539	144.936539	144.936539	0	0	1	-360	360;
	6784	4987	0.000209	0.000911	0.00032	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6784	2424	0.000249	0.000831	0.00039	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7834	4529	0.00136	0.006458	0.00208	115.181552	115.181552	115.181552	0	0	1	-360	360;
	4529	4347	0.002782	0.013129	0.00422	112.48293	112.48293	112.48293	0	0	1	-360	360;
	5682	1662	0.008049	0.036031	0.01224	133.674658	133.674658	133.674658	0	0	1	-360	360;
	1662	2415	0.022311	0.0864	0.03396	126.189341	126.189341	126.189341	0	0	1	-360	360;
	607	1384	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5682	1384	0.00912	0.035311	0.01387	173.035513	173.035513	173.035513	0	0	1	-360	360;
	1384	1100	0.001471	0.005489	0.00225	136.189856	136.189856	136.189856	0	0	1	-360	360;
	6552	7837	0.00058	0.00639	0.13256	1083.94958	1083.94958	1083.94958	0	0	1	-360	360;
	6552	7837	0.00057	0.00651	0.12974	997.006065	997.006065	997.006065	0	0	1	-360	360;
	4666	3637	0.002111	0.007542	0.00321	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1095	3637	0.014049	0.037031	0.01003	126.033976	126.033976	126.033976	0	0	1	-360	360;
	235	2471	0	0.000222	0	147.589955	147.589955	147.589955	0	0	1	-360	360;
	235	6150	0.002071	0.005609	0.25839	164.723748	164.723748	164.723748	0	0	1	-360	360;
	1796	235	0.00368	0.016711	0.00559	189.962759	189.962759	189.962759	0	0	1	-360	360;
	1796	235	0.0036	0.017018	0.00547	150.870026	150.870026	150.870026	0	0	1	-360	360;
	235	7353	0.00012	0.000431	0.02737	258.105362	258.105362	258.105362	0	0	1	-360	360;
	2471	6696	0.00012	0.000431	0.02789	124.744811	124.744811	124.744811	0	0	1	-360	360;
	6417	5366	0.01328	0.031071	0.00949	249.134984	249.134984	249.134984	0	0	1	-360	360;
	7646	6417	0.031911	0.071711	0.0228	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5639	3144	0.000382	0.001289	0.04004	205.183329	205.183329	205.183329	0	0	1	-360	360;
	1796	5639	0.002582	0.00988	0.00366	193.979125	193.979125	193.979125	0	0	1	-360	360;
	2353	3426	0.00036	0.001218	0.03799	208.528352	208.528352	208.528352	0	0	1	-360	360;
	4176	2353	0.002458	0.00908	0.0035	195.74088	195.74088	195.74088	0	0	1	-360	360;
	6285	5538	0.000911	0.003271	0.00139	117.091917	117.091917	117.091917	0	0	1	-360	360;
	156	6285	0.01436	0.05344	0.02186	110.327826	110.327826	110.327826	0	0	1	-360	360;
	141	6017	0.000462	0.001631	0.0007	127.251694	127.251694	127.251694	0	0	1	-360	360;
	141	5538	0.000929	0.0032	0.00141	110.290673	110.290673	110.290673	0	0	1	-360	360;
	8243	6017	0	0.000222	0	148.535135	148.535135	148.535135	0	0	1	-360	360;
	5162	6017	0.013391	0.060849	0.02037	103.923048	103.923048	103.923048	0	0	1	-360	360;
	138	4560	0.010631	0.048342	0.01619	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5271	3008	0.005498	0.021302	0.00837	132.775464	132.775464	132.775464	0	0	1	-360	360;
	3008	1356	0.001449	0.006329	0.00221	117.441359	117.441359	117.441359	0	0	1	-360	360;
	760	8090	0.007951	0.029329	0.01131	103.923048	103.923048	103.923048	0	0	1	-360	360;
	760	1718	0.02072	0.086938	0.03154	118.26365	118.26365	118.26365	0	0	1	-360	360;
	211	760	0.007182	0.02672	0.01093	125.222337	125.222337	125.222337	0	0	1	-360	360;
	6569	2151	0.00424	0.019129	0.00604	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2606	6569	0.010542	0.047938	0.01606	145.965378	145.965378	145.965378	0	0	1	-360	360;
	6569	3322	0.016751	0.06488	0.0255	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4071	3258	0.016671	0.059071	0.02372	148.346774	148.346774	148.346774	0	0	1	-360	360;
	4071	5647	0.00536	0.024151	0.00762	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5114	2606	0	0.000222	0	124.581132	124.581132	124.581132	0	0	1	-360	360;
	2606	4818	0.007089	0.017938	0.00506	270.094704	270.094704	270.094704	0	0	1	-360	360;
	5087	2606	0.020849	0.052831	0.0149	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5224	4818	0.010609	0.005209	0.09247	255.170055	255.170055	255.170055	0	0	1	-360	360;
	2225	4675	0.020898	0.098071	0.02974	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5906	4675	0.002258	0.008409	0.00344	127.162061	127.162061	127.162061	0	0	1	-360	360;
	379	5732	0.00056	0.00236	0.00085	111.008522	111.008522	111.008522	0	0	1	-360	360;
	5732	923	0.014258	0.064311	0.0203	111.03762	111.03762	111.03762	0	0	1	-360	360;
	7938	1814	0	0.000222	0	122.874456	122.874456	122.874456	0	0	1	-360	360;
	6026	1814	0.040249	0.143902	0.06127	133.017865	133.017865	133.017865	0	0	1	-360	360;
	7925	1814	0.021778	0.071862	0.03316	164.73362	164.73362	164.73362	0	0	1	-360	360;
	1814	3486	0.022742	0.09932	0.03462	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4785	1814	0.021511	0.07092	0.03272	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4908	3875	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3875	8044	0.010818	0.0454	0.01646	118.102569	118.102569	118.102569	0	0	1	-360	360;
	2084	3875	0.013671	0.059698	0.0208	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5450	7080	0.00011	0.00059	0.92171	1006.14489	1006.14489	1006.14489	0	0	1	-360	360;
	1634	7080	0.00093	0.00993	0.20878	1444.78519	1444.78519	1444.78519	0	0	1	-360	360;
	1881	5450	0.00037	0.00305	0.08352	1491.50078	1491.50078	1491.50078	0	0	1	-360	360;
	2548	915	0.0001	0.00066	0.82116	1163.86373	1163.86373	1163.86373	0	0	1	-360	360;
	1634	2548	0.00105	0.00973	0.23662	1411.43261	1411.43261	1411.43261	0	0	1	-360	360;
	915	1504	0.00037	0.00298	0.08506	1618.60837	1618.60837	1618.60837	0	0	1	-360	360;
	1881	4534	0.00105	0.01196	0.23808	1487.03241	1487.03241	1487.03241	0	0	1	-360	360;
	1504	146	0.00073	0.00626	0.16502	1543.96821	1543.96821	1543.96821	0	0	1	-360	360;
	3531	6983	0.00112	0.002329	0.0008	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3531	8307	0.0068	0.032178	0.01035	225.030812	225.030812	225.030812	0	0	1	-360	360;
	5048	3531	0.013871	0.06308	0.02112	171.328317	171.328317	171.328317	0	0	1	-360	360;
	5927	3835	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3541	3835	0.00636	0.028689	0.00906	130.203369	130.203369	130.203369	0	0	1	-360	360;
	4744	3835	0.005231	0.020942	0.00745	109.818863	109.818863	109.818863	0	0	1	-360	360;
	4057	3835	0.019129	0.063111	0.02912	210.887665	210.887665	210.887665	0	0	1	-360	360;
	3835	2820	0.006649	0.0298	0.01013	182.355332	182.355332	182.355332	0	0	1	-360	360;
	6983	5176	0.0006	0.001609	0.00042	268.363866	268.363866	268.363866	0	0	1	-360	360;
	98	379	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7472	379	0.014849	0.066942	0.02113	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7800	5518	0.00392	0.009182	0.0001	244.150575	244.150575	244.150575	0	0	1	-360	360;
	5518	3368	0	0.00028	0	111.302624	111.302624	111.302624	0	0	1	-360	360;
	5518	273	0	0.00028	0	145.505778	145.505778	145.505778	0	0	1	-360	360;
	5518	5383	0	0.00028	0	136.560601	136.560601	136.560601	0	0	1	-360	360;
	2300	5586	0.00384	0.00936	0.0001	256.929213	256.929213	256.929213	0	0	1	-360	360;
	5586	6423	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5586	5904	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5586	1440	0	0.00028	0	110.400572	110.400572	110.400572	0	0	1	-360	360;
	2300	7800	0	0.000222	0	145.39484	145.39484	145.39484	0	0	1	-360	360;
	3368	6698	0.02008	0.09128	0.03056	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3704	6423	0.023991	0.092942	0.03653	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5210	273	0.01384	0.06552	0.02108	103.923048	103.923048	103.923048	0	0	1	-360	360;
	329	5904	0.00476	0.022538	0.00725	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5383	2288	0.00496	0.016382	0.00756	146.44758	146.44758	146.44758	0	0	1	-360	360;
	1440	7017	0.00244	0.008049	0.00372	112.170641	112.170641	112.170641	0	0	1	-360	360;
	8133	508	0.000671	0.002502	0.00103	111.160509	111.160509	111.160509	0	0	1	-360	360;
	508	4575	0.0066	0.029569	0.01005	110.946688	110.946688	110.946688	0	0	1	-360	360;
	8133	6680	0.000689	0.002449	0.00105	166.986672	166.986672	166.986672	0	0	1	-360	360;
	84	6680	0.007271	0.024969	0.01107	149.701411	149.701411	149.701411	0	0	1	-360	360;
	789	8133	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3541	2931	0.007649	0.031849	0.01089	164.852872	164.852872	164.852872	0	0	1	-360	360;
	6055	5377	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6055	6432	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	252	6055	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2031	5377	0.019471	0.074818	0.02772	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5611	6432	0.009231	0.041622	0.01314	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2486	7723	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2022	5766	0.000258	0.001089	0.00037	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2022	3278	0.019871	0.042889	0.01419	127.848472	127.848472	127.848472	0	0	1	-360	360;
	3278	7723	0.00076	0.001978	0.0006	263.609646	263.609646	263.609646	0	0	1	-360	360;
	5496	7723	0.000751	0.001831	0.00059	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7175	5496	0.039151	0.088	0.02798	226.763989	226.763989	226.763989	0	0	1	-360	360;
	357	5032	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5032	2590	0.032351	0.141289	0.04924	123.451488	123.451488	123.451488	0	0	1	-360	360;
	872	5032	0.022569	0.074449	0.03435	109.708185	109.708185	109.708185	0	0	1	-360	360;
	5032	4867	5.8e-05	0.000222	0.00011	122.217922	122.217922	122.217922	0	0	1	-360	360;
	6063	8287	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1380	8287	0.009542	0.031471	0.01452	125.944602	125.944602	125.944602	0	0	1	-360	360;
	6349	8287	0.021671	0.09856	0.03299	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2802	2121	0.005609	0.013978	0.00439	275.136271	275.136271	275.136271	0	0	1	-360	360;
	7599	2121	0.003751	0.01016	0.00268	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7327	3327	0.000529	0.00208	0.00082	123.286251	123.286251	123.286251	0	0	1	-360	360;
	7327	3709	0.008862	0.029231	0.01349	151.204138	151.204138	151.204138	0	0	1	-360	360;
	863	5556	0.000538	0.00204	0.00083	108.187271	108.187271	108.187271	0	0	1	-360	360;
	863	7450	0.009569	0.026289	0.00684	260.074703	260.074703	260.074703	0	0	1	-360	360;
	3327	3106	0.00028	0.001209	0.00042	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1095	3106	0.005489	0.014249	0.0043	125.633353	125.633353	125.633353	0	0	1	-360	360;
	5556	2802	0.000271	0.001231	0.00041	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3327	5236	0.000338	0.001311	0.00052	131.317424	131.317424	131.317424	0	0	1	-360	360;
	5236	5228	0.005791	0.02716	0.00824	146.481875	146.481875	146.481875	0	0	1	-360	360;
	3958	6552	0.00288	0.03132	0.64931	1268.18055	1268.18055	1268.18055	0	0	1	-360	360;
	3958	6552	0.00282	0.03191	0.6355	1245.53062	1245.53062	1245.53062	0	0	1	-360	360;
	6552	5337	0.00111	0.01111	0.2496	1341.82028	1341.82028	1341.82028	0	0	1	-360	360;
	6552	6114	0.00108	0.01133	0.2445	1322.54813	1322.54813	1322.54813	0	0	1	-360	360;
	5556	5771	0.008751	0.033911	0.01332	114.799115	114.799115	114.799115	0	0	1	-360	360;
	5556	267	0.002418	0.00936	0.00368	190.560057	190.560057	190.560057	0	0	1	-360	360;
	6222	4314	0	0.000222	0	129.276115	129.276115	129.276115	0	0	1	-360	360;
	6222	247	0.00944	0.044271	0.01342	140.564497	140.564497	140.564497	0	0	1	-360	360;
	4314	7278	0.006542	0.0272	0.0093	133.156862	133.156862	133.156862	0	0	1	-360	360;
	3258	2207	0.001591	0.005658	0.00228	136.694662	136.694662	136.694662	0	0	1	-360	360;
	2334	1011	0.000711	0.001822	0.00055	299.990767	299.990767	299.990767	0	0	1	-360	360;
	1011	4175	0.010369	0.021511	0.00741	302.506484	302.506484	302.506484	0	0	1	-360	360;
	2334	7175	0.001978	0.0068	0.00301	119.012415	119.012415	119.012415	0	0	1	-360	360;
	485	5456	0.009062	0.023902	0.00647	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5456	606	0.0092	0.025249	0.00657	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7652	819	0.0022	0.005942	0.00157	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4487	7652	0.016169	0.03352	0.01155	116.328343	116.328343	116.328343	0	0	1	-360	360;
	4655	606	0.006902	0.014898	0.00493	132.661668	132.661668	132.661668	0	0	1	-360	360;
	4301	4655	0.018609	0.041849	0.0133	114.748712	114.748712	114.748712	0	0	1	-360	360;
	3234	4470	0.00156	0.00696	0.00237	139.12828	139.12828	139.12828	0	0	1	-360	360;
	4821	3234	0.01592	0.07532	0.02422	126.479806	126.479806	126.479806	0	0	1	-360	360;
	2240	4470	0.00176	0.006822	0.00268	116.242087	116.242087	116.242087	0	0	1	-360	360;
	2240	7789	0.015542	0.069582	0.02364	106.931101	106.931101	106.931101	0	0	1	-360	360;
	6639	1486	0.00176	0.004631	0.00126	110.215849	110.215849	110.215849	0	0	1	-360	360;
	1997	1486	0.02272	0.061458	0.01624	273.862174	273.862174	273.862174	0	0	1	-360	360;
	5682	2553	0.00348	0.015809	0.00529	106.588934	106.588934	106.588934	0	0	1	-360	360;
	7940	2553	0.013449	0.056431	0.02047	106.530478	106.530478	106.530478	0	0	1	-360	360;
	4448	1161	0.002911	0.01272	0.00444	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5682	1161	0.0034	0.016102	0.00518	114.140503	114.140503	114.140503	0	0	1	-360	360;
	5970	4448	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4448	8145	0.030151	0.079538	0.02154	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4821	4448	0.017769	0.063511	0.02704	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4448	3726	0.031249	0.084538	0.02233	106.936297	106.936297	106.936297	0	0	1	-360	360;
	7872	5682	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5682	4121	0.013698	0.047049	0.02085	129.409397	129.409397	129.409397	0	0	1	-360	360;
	5682	4121	0.01396	0.046031	0.02124	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7397	5682	0.007938	0.037569	0.01209	144.953167	144.953167	144.953167	0	0	1	-360	360;
	7940	5682	0.016751	0.070271	0.02549	116.743515	116.743515	116.743515	0	0	1	-360	360;
	6639	5594	0.00172	0.00472	0.00123	121.65284	121.65284	121.65284	0	0	1	-360	360;
	5594	3094	0.007271	0.015689	0.00519	117.280278	117.280278	117.280278	0	0	1	-360	360;
	473	2344	0.001471	0.004	0.18414	111.603741	111.603741	111.603741	0	0	1	-360	360;
	473	2675	0.001991	0.007729	0.2484	113.244686	113.244686	113.244686	0	0	1	-360	360;
	7560	473	0.003151	0.047689	0.03121	133.01085	133.01085	133.01085	0	0	1	-360	360;
	473	322	0.000809	0.002769	0.10055	110.41694	110.41694	110.41694	0	0	1	-360	360;
	7600	473	0.002009	0.020209	0.01103	145.093203	145.093203	145.093203	0	0	1	-360	360;
	3828	5787	0	0.00028	0	140.888477	140.888477	140.888477	0	0	1	-360	360;
	5787	7927	0.001	0.002831	0.12539	191.006146	191.006146	191.006146	0	0	1	-360	360;
	5787	7315	0.001502	0.005169	0.18744	246.562629	246.562629	246.562629	0	0	1	-360	360;
	5400	5787	0.001969	0.0206	0.0108	311.769145	311.769145	311.769145	0	0	1	-360	360;
	6929	4744	0.000151	0.000649	0.00022	107.989038	107.989038	107.989038	0	0	1	-360	360;
	6929	3263	0.000142	0.000662	0.00022	132.305732	132.305732	132.305732	0	0	1	-360	360;
	2159	3263	0.004409	0.01628	0.00627	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6067	6089	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8293	6843	0	0.000222	0	134.244157	134.244157	134.244157	0	0	1	-360	360;
	6089	8014	0.00028	0.000791	0.03499	135.840675	135.840675	135.840675	0	0	1	-360	360;
	5120	6843	0.000622	0.001671	0.07722	137.570733	137.570733	137.570733	0	0	1	-360	360;
	3975	660	0.000471	0.0016	0.00071	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3929	660	0.013889	0.065751	0.02115	106.24365	106.24365	106.24365	0	0	1	-360	360;
	3975	2924	0.000471	0.00156	0.00073	127.359515	127.359515	127.359515	0	0	1	-360	360;
	4121	2924	0.018262	0.082978	0.02778	123.720389	123.720389	123.720389	0	0	1	-360	360;
	1100	2415	0.02168	0.077489	0.03299	126.614906	126.614906	126.614906	0	0	1	-360	360;
	7225	5025	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5025	7215	0.000538	0.001902	0.00076	142.405234	142.405234	142.405234	0	0	1	-360	360;
	2879	5025	0.004578	0.017031	0.00697	110.573604	110.573604	110.573604	0	0	1	-360	360;
	2879	5025	0.004662	0.016671	0.0071	111.756768	111.756768	111.756768	0	0	1	-360	360;
	3983	4121	0	0.000222	0	147.670495	147.670495	147.670495	0	0	1	-360	360;
	4121	2885	0.0242	0.093769	0.03684	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8044	1870	0.002289	0.008129	0.00327	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4392	907	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	907	3333	0.00668	0.02588	0.01017	116.10205	116.10205	116.10205	0	0	1	-360	360;
	1061	3333	0.00308	0.013778	0.00468	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2031	907	0.032169	0.124618	0.04897	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1214	7560	0.006778	0.02236	0.01032	184.731013	184.731013	184.731013	0	0	1	-360	360;
	7370	1214	0.00308	0.011929	0.00469	149.979665	149.979665	149.979665	0	0	1	-360	360;
	5314	6440	0.016911	0.070391	0.02406	115.438761	115.438761	115.438761	0	0	1	-360	360;
	2334	5314	0.03196	0.149969	0.04548	124.072169	124.072169	124.072169	0	0	1	-360	360;
	5264	7222	0.00016	0.00049	0.69896	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2780	5264	7e-05	0.00049	0.01418	1113.95071	1113.95071	1113.95071	0	0	1	-360	360;
	7369	7794	0.0022	0.00852	0.00335	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5803	7794	0.000431	0.000951	0.0003	133.746625	133.746625	133.746625	0	0	1	-360	360;
	7369	5292	0.002249	0.008351	0.00341	112.031125	112.031125	112.031125	0	0	1	-360	360;
	7016	5292	0.019431	0.053351	0.01388	262.167714	262.167714	262.167714	0	0	1	-360	360;
	4526	7369	0	0.000222	0	118.470976	118.470976	118.470976	0	0	1	-360	360;
	5265	4788	0.001929	0.008111	0.00294	131.096847	131.096847	131.096847	0	0	1	-360	360;
	7600	5265	0.00652	0.021511	0.00992	110.327566	110.327566	110.327566	0	0	1	-360	360;
	5827	4788	0.001889	0.008271	0.00288	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5827	7560	0.012289	0.051551	0.0187	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7583	4788	0.001431	0.005071	0.00203	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1062	4788	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7940	7739	0.007271	0.0344	0.01106	112.040218	112.040218	112.040218	0	0	1	-360	360;
	6561	7739	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1813	8044	0.003289	0.011329	0.00502	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8044	1247	0.012129	0.032831	0.00867	258.504686	258.504686	258.504686	0	0	1	-360	360;
	8044	156	0.006031	0.025289	0.00917	143.390164	143.390164	143.390164	0	0	1	-360	360;
	8044	156	0.005911	0.025778	0.00899	131.942001	131.942001	131.942001	0	0	1	-360	360;
	421	8044	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1988	7472	0.000529	0.001969	0.0008	139.865874	139.865874	139.865874	0	0	1	-360	360;
	7645	7472	0.000538	0.001929	0.00082	114.608936	114.608936	114.608936	0	0	1	-360	360;
	1568	6519	0.001151	0.004222	0.00163	117.57412	117.57412	117.57412	0	0	1	-360	360;
	4386	1568	0.003818	0.017222	0.00543	110.335101	110.335101	110.335101	0	0	1	-360	360;
	2288	7090	0.00416	0.019502	0.00591	110.222864	110.222864	110.222864	0	0	1	-360	360;
	4866	2412	0.001991	0.00372	0.00156	203.674366	203.674366	203.674366	0	0	1	-360	360;
	4866	2894	0.009889	0.026751	0.00707	205.001203	205.001203	205.001203	0	0	1	-360	360;
	3484	2412	0.002018	0.00364	0.00158	265.037289	265.037289	265.037289	0	0	1	-360	360;
	6553	3484	0.01644	0.036942	0.01174	112.097376	112.097376	112.097376	0	0	1	-360	360;
	8308	2412	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2334	5315	0.019071	0.07388	0.02903	108.361082	108.361082	108.361082	0	0	1	-360	360;
	468	5315	0.02804	0.1256	0.04268	131.083077	131.083077	131.083077	0	0	1	-360	360;
	1433	5315	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5315	5224	0.001911	0.000751	0.01661	265.201747	265.201747	265.201747	0	0	1	-360	360;
	6361	3871	0.008098	0.028689	0.01152	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6765	6361	0.017609	0.068191	0.0268	149.645033	149.645033	149.645033	0	0	1	-360	360;
	6361	4176	0.015018	0.058191	0.02287	122.216623	122.216623	122.216623	0	0	1	-360	360;
	6075	6359	0.007031	0.014578	0.00502	239.644211	239.644211	239.644211	0	0	1	-360	360;
	6075	3750	0.003791	0.012529	0.00577	129.266502	129.266502	129.266502	0	0	1	-360	360;
	6075	4190	0.010249	0.035209	0.01561	165.006678	165.006678	165.006678	0	0	1	-360	360;
	7022	848	0.002489	0.005818	0.00177	225.690464	225.690464	225.690464	0	0	1	-360	360;
	3085	7022	0.002071	0.009751	0.00296	118.116339	118.116339	118.116339	0	0	1	-360	360;
	3085	2963	0.001871	0.006902	0.00266	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3085	4905	0.003658	0.014062	0.00522	110.433308	110.433308	110.433308	0	0	1	-360	360;
	7017	7256	0.002289	0.008178	0.00348	110.391219	110.391219	110.391219	0	0	1	-360	360;
	1104	7256	0.002329	0.008	0.00354	110.128294	110.128294	110.128294	0	0	1	-360	360;
	1104	2288	0.002071	0.009249	0.00314	109.94461	109.94461	109.94461	0	0	1	-360	360;
	1380	6501	0.005951	0.021249	0.00905	110.089063	110.089063	110.089063	0	0	1	-360	360;
	329	6501	0.006058	0.0208	0.00922	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4302	993	0.000942	0.003129	0.00144	137.685828	137.685828	137.685828	0	0	1	-360	360;
	4302	3750	0.030849	0.10176	0.04696	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2044	993	0.00084	0.003551	0.00128	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2288	2044	0.002378	0.007831	0.00361	107.677269	107.677269	107.677269	0	0	1	-360	360;
	3492	1249	0.007231	0.01624	0.00517	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1383	3492	0.001311	0.005111	0.00201	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3492	5316	0.001338	0.005009	0.00205	127.529948	127.529948	127.529948	0	0	1	-360	360;
	3298	3492	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1249	515	0.000338	0.00116	0.00052	103.923048	103.923048	103.923048	0	0	1	-360	360;
	515	1173	0.025778	0.069751	0.01842	109.629203	109.629203	109.629203	0	0	1	-360	360;
	2016	515	0.004009	0.01376	0.0061	105.236116	105.236116	105.236116	0	0	1	-360	360;
	4472	515	0.00408	0.013462	0.00621	133.57853	133.57853	133.57853	0	0	1	-360	360;
	3399	515	0.000329	0.001262	0.0005	141.095543	141.095543	141.095543	0	0	1	-360	360;
	4356	515	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7845	1333	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2437	1183	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8229	1333	0.000711	0.002018	0.0621	114.217666	114.217666	114.217666	0	0	1	-360	360;
	1183	4609	0.00068	0.001502	0.05898	137.620876	137.620876	137.620876	0	0	1	-360	360;
	3427	847	0.00012	0.00056	0.00018	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2654	847	0.02516	0.086418	0.03831	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3427	5637	0.00012	0.00052	0.00018	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5637	6349	0.032591	0.116511	0.04961	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7412	3427	0	0.00028	0	122.668428	122.668428	122.668428	0	0	1	-360	360;
	2083	1988	0.01752	0.062098	0.02494	123.540862	123.540862	123.540862	0	0	1	-360	360;
	7645	2445	0.00076	0.003551	0.00114	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8386	2207	4e-05	0.00028	0.00918	111.127774	111.127774	111.127774	0	0	1	-360	360;
	8386	5116	0.000858	0.004458	0.0641	111.164666	111.164666	111.164666	0	0	1	-360	360;
	5116	5796	8e-05	0.000338	0.01726	116.14336	116.14336	116.14336	0	0	1	-360	360;
	5201	6158	4e-05	0.00028	0.00918	146.396658	146.396658	146.396658	0	0	1	-360	360;
	6158	2521	0.000982	0.004369	0.07264	121.920702	121.920702	121.920702	0	0	1	-360	360;
	2517	5201	0.01704	0.083271	0.02426	121.527613	121.527613	121.527613	0	0	1	-360	360;
	2521	2765	7.1e-05	0.000351	0.01689	130.006694	130.006694	130.006694	0	0	1	-360	360;
	4444	5501	0.000271	0.00068	0.0002	247.729944	247.729944	247.729944	0	0	1	-360	360;
	8252	4444	0.012809	0.05376	0.0195	129.826128	129.826128	129.826128	0	0	1	-360	360;
	4444	6689	0.00364	0.015262	0.00554	140.746622	140.746622	140.746622	0	0	1	-360	360;
	1059	4515	0.00012	0.000431	0.00018	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7606	4515	0.003982	0.015271	0.00566	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1059	3753	0.00012	0.000418	0.00018	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7933	3753	0.000431	0.00172	0.00061	103.923048	103.923048	103.923048	0	0	1	-360	360;
	572	2996	0.00024	0.00032	0.01062	133.493832	133.493832	133.493832	0	0	1	-360	360;
	3146	572	0.002418	0.011462	0.00369	111.88901	111.88901	111.88901	0	0	1	-360	360;
	6271	4057	0.00028	0.000311	0.01204	172.86352	172.86352	172.86352	0	0	1	-360	360;
	6271	7606	0.003658	0.01664	0.00557	209.212165	209.212165	209.212165	0	0	1	-360	360;
	7606	2996	0.003582	0.016951	0.00546	111.739621	111.739621	111.739621	0	0	1	-360	360;
	1716	4569	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4569	2132	0.006231	0.029249	0.00887	132.292742	132.292742	132.292742	0	0	1	-360	360;
	4569	7933	0.00144	0.005529	0.00205	130.333013	130.333013	130.333013	0	0	1	-360	360;
	4569	5397	0.011111	0.023978	0.00794	125.851851	125.851851	125.851851	0	0	1	-360	360;
	7924	1059	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2180	5075	0.001182	0.002409	0.00093	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5715	2773	0.006098	0.029822	0.00869	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2773	7482	0.00096	0.003378	0.00135	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3821	6026	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6026	1456	0.01044	0.03444	0.0159	111.602702	111.602702	111.602702	0	0	1	-360	360;
	6026	1456	0.009298	0.03904	0.01416	140.562678	140.562678	140.562678	0	0	1	-360	360;
	6026	4785	0.017311	0.078738	0.02636	103.923048	103.923048	103.923048	0	0	1	-360	360;
	814	8145	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8145	485	0.012449	0.029129	0.00889	105.96072	105.96072	105.96072	0	0	1	-360	360;
	657	1094	0.00015	0.00045	0.68377	811.107927	811.107927	811.107927	0	0	1	-360	360;
	6911	1094	0.00107	0.01215	0.24188	1342.21256	1342.21256	1342.21256	0	0	1	-360	360;
	2780	657	6e-05	0.00056	0.01263	1315.86103	1315.86103	1315.86103	0	0	1	-360	360;
	3791	7406	0.00035	0.00284	0.07794	1368.208	1368.208	1368.208	0	0	1	-360	360;
	6847	7406	0.00163	0.01512	0.36808	1466.30766	1466.30766	1466.30766	0	0	1	-360	360;
	4534	7406	0.00115	0.00913	0.26067	1422.46107	1422.46107	1422.46107	0	0	1	-360	360;
	2600	7406	0.00133	0.01513	0.30125	898.140951	898.140951	898.140951	0	0	1	-360	360;
	2600	7406	0.00131	0.01401	0.29471	1005.47222	1005.47222	1005.47222	0	0	1	-360	360;
	6911	7406	0.00013	0.00129	0.03155	1272.17372	1272.17372	1272.17372	0	0	1	-360	360;
	6461	7406	0.00077	0.00664	0.17486	1389.89962	1389.89962	1389.89962	0	0	1	-360	360;
	2288	3750	0.028391	0.12908	0.04321	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7812	2288	0.01216	0.05532	0.01852	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7916	5577	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2179	5577	0.023609	0.053071	0.01687	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4517	1610	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1610	6216	0.020471	0.051871	0.01462	262.889199	262.889199	262.889199	0	0	1	-360	360;
	2031	1610	0.01744	0.049862	0.01247	122.545539	122.545539	122.545539	0	0	1	-360	360;
	7789	2415	0.009111	0.039809	0.01387	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4821	6405	0.033022	0.13856	0.05026	120.26157	120.26157	120.26157	0	0	1	-360	360;
	6405	2415	0.009529	0.045089	0.0145	139.845869	139.845869	139.845869	0	0	1	-360	360;
	5352	8229	0	0.000222	0	149.728951	149.728951	149.728951	0	0	1	-360	360;
	8229	781	0.001889	0.003462	0.07129	266.308787	266.308787	266.308787	0	0	1	-360	360;
	8229	777	0.006569	0.0142	0.0047	287.803971	287.803971	287.803971	0	0	1	-360	360;
	6390	8229	0.003982	0.018102	0.00605	127.529689	127.529689	127.529689	0	0	1	-360	360;
	8229	573	0.00232	0.007969	0.00353	118.205972	118.205972	118.205972	0	0	1	-360	360;
	1303	2060	0.002502	0.011698	0.00355	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5041	2060	0.009351	0.039209	0.01422	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2060	7902	0.008622	0.0296	0.01312	108.407068	108.407068	108.407068	0	0	1	-360	360;
	1868	2339	0.008951	0.037289	0.01275	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2339	4119	0.005369	0.02144	0.00763	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6103	7448	0	0.000222	0	135.014486	135.014486	135.014486	0	0	1	-360	360;
	5041	7448	0.00608	0.025551	0.00927	117.945905	117.945905	117.945905	0	0	1	-360	360;
	5041	7448	0.00596	0.026049	0.00908	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7448	1868	0.001369	0.004898	0.00209	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7448	2180	0.00144	0.006031	0.00219	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4385	2989	0.0012	0.00328	0.00086	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7606	4385	0.00132	0.006271	0.00201	118.643748	118.643748	118.643748	0	0	1	-360	360;
	4385	4531	0.009089	0.040711	0.01383	115.774953	115.774953	115.774953	0	0	1	-360	360;
	5741	6590	5.8e-05	0.00028	4e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2879	6590	0.000302	0.00132	0.00045	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4059	2796	0.00368	0.016502	0.00561	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1796	4059	0.000329	0.00124	0.00051	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4601	1769	0.0012	0.002689	0.00085	204.078627	204.078627	204.078627	0	0	1	-360	360;
	2879	1769	0.000338	0.001289	0.00051	115.504493	115.504493	115.504493	0	0	1	-360	360;
	372	5741	0.000942	0.002151	0.00073	253.519757	253.519757	253.519757	0	0	1	-360	360;
	4812	7296	0.00316	0.010862	0.00481	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4812	6275	0.003222	0.010618	0.0049	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3304	7296	0.010169	0.037858	0.01549	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1668	6275	0.00552	0.024129	0.00841	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1957	3858	0.000449	0.001889	0.00069	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4166	5469	0.00044	0.001929	0.00067	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3480	4514	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4514	6739	0.000418	0.00192	0.00061	132.163617	132.163617	132.163617	0	0	1	-360	360;
	4514	6519	0.000409	0.00196	0.0006	110.746896	110.746896	110.746896	0	0	1	-360	360;
	7606	6739	0.00608	0.022431	0.00865	132.581648	132.581648	132.581648	0	0	1	-360	360;
	2857	3629	0	0.000222	0	151.721415	151.721415	151.721415	0	0	1	-360	360;
	3915	3629	0.00016	0.000751	0.00024	120.49176	120.49176	120.49176	0	0	1	-360	360;
	3915	7790	0.046622	0.173511	0.07097	121.3564	121.3564	121.3564	0	0	1	-360	360;
	1376	3629	0.000151	0.000698	0.00024	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2654	1376	0.01464	0.052329	0.02228	128.098667	128.098667	128.098667	0	0	1	-360	360;
	5484	5822	0.006129	0.023751	0.00933	159.316372	159.316372	159.316372	0	0	1	-360	360;
	4494	5822	0.001049	0.003311	0.10976	132.378998	132.378998	132.378998	0	0	1	-360	360;
	7743	4871	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4871	2796	0.003271	0.012089	0.00467	134.771047	134.771047	134.771047	0	0	1	-360	360;
	4871	1720	0.002951	0.013311	0.0042	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4871	4682	0.017489	0.082769	0.02662	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4871	2177	0.006231	0.029498	0.00949	135.662446	135.662446	135.662446	0	0	1	-360	360;
	2527	4772	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5782	7854	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6802	4772	0.00124	0.004458	0.15552	188.595391	188.595391	188.595391	0	0	1	-360	360;
	5324	7854	0.001169	0.001969	0.05146	220.838037	220.838037	220.838037	0	0	1	-360	360;
	4201	8014	0.000791	0.002929	0.09963	132.433817	132.433817	132.433817	0	0	1	-360	360;
	977	8014	0.000911	0.003138	0.11378	253.769952	253.769952	253.769952	0	0	1	-360	360;
	1796	1720	0.004671	0.01672	0.00711	130.881467	130.881467	130.881467	0	0	1	-360	360;
	1334	2344	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2344	7583	0.00064	0.00188	0.08014	103.923048	103.923048	103.923048	0	0	1	-360	360;
	547	6947	0.00014	0.00061	0.28157	888.824423	888.824423	888.824423	0	0	1	-360	360;
	4348	5354	0.000698	0.0024	0.00106	139.812613	139.812613	139.812613	0	0	1	-360	360;
	4348	2239	0.000711	0.002351	0.00108	145.096841	145.096841	145.096841	0	0	1	-360	360;
	7606	5354	0.001858	0.006391	0.00283	176.264142	176.264142	176.264142	0	0	1	-360	360;
	4468	2239	0.014129	0.06328	0.0215	178.335848	178.335848	178.335848	0	0	1	-360	360;
	4176	4468	0.0002	0.00092	0.00031	173.062273	173.062273	173.062273	0	0	1	-360	360;
	8266	638	4e-05	0.0004	0.00835	1385.12585	1385.12585	1385.12585	0	0	1	-360	360;
	8266	3791	4e-05	0.00039	0.00947	1251.2061	1251.2061	1251.2061	0	0	1	-360	360;
	4534	638	0.00066	0.00705	0.14833	1448.78495	1448.78495	1448.78495	0	0	1	-360	360;
	4233	262	0.003622	0.016209	0.00551	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8207	4233	5.8e-05	0.00028	4e-05	256.658233	256.658233	256.658233	0	0	1	-360	360;
	2343	4233	3.1e-05	0.00028	5e-05	105.671554	105.671554	105.671554	0	0	1	-360	360;
	7260	298	0.00084	0.002031	0.00066	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6479	7485	0.004511	0.01972	0.00687	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6479	7788	0.00064	0.0028	0.00097	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7707	7260	0.004422	0.020102	0.00673	147.378991	147.378991	147.378991	0	0	1	-360	360;
	7260	4344	0	0.00028	0	143.168029	143.168029	143.168029	0	0	1	-360	360;
	6234	670	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	670	2931	0.000822	0.003622	0.00126	175.409635	175.409635	175.409635	0	0	1	-360	360;
	670	2132	0.000809	0.003689	0.00123	166.348584	166.348584	166.348584	0	0	1	-360	360;
	3207	6081	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2373	6081	0.001471	0.00528	0.00225	167.814419	167.814419	167.814419	0	0	1	-360	360;
	2373	6081	0.001502	0.005169	0.00229	139.314822	139.314822	139.314822	0	0	1	-360	360;
	1972	6081	0.000769	0.002858	0.0972	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6081	2084	0.003698	0.013209	0.00563	138.244934	138.244934	138.244934	0	0	1	-360	360;
	6081	1957	0.00116	0.005271	0.00177	128.14855	128.14855	128.14855	0	0	1	-360	360;
	6081	4597	0.003378	0.006178	0.12747	263.431937	263.431937	263.431937	0	0	1	-360	360;
	6081	691	0.000849	0.004009	0.00129	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6081	149	0.000431	0.00112	0.03729	150.942772	150.942772	150.942772	0	0	1	-360	360;
	2461	8307	0.006778	0.002631	0.24938	216.664747	216.664747	216.664747	0	0	1	-360	360;
	1394	2820	0.003569	0.00468	0.16087	278.14952	278.14952	278.14952	0	0	1	-360	360;
	1394	2820	0.003178	0.005298	0.14333	267.960644	267.960644	267.960644	0	0	1	-360	360;
	5293	3880	0.00388	0.0174	0.00591	129.41901	129.41901	129.41901	0	0	1	-360	360;
	674	3880	0.003138	0.011689	0.00478	114.714678	114.714678	114.714678	0	0	1	-360	360;
	5293	3068	0.0044	0.017049	0.0067	143.613079	143.613079	143.613079	0	0	1	-360	360;
	3068	6969	0.000791	0.002631	0.00121	123.275079	123.275079	123.275079	0	0	1	-360	360;
	4886	2675	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	417	6216	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2083	6216	0.017698	0.0584	0.02695	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2083	6216	0.015769	0.066191	0.02401	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3958	7564	0.00225	0.02425	0.51035	832.427671	832.427671	832.427671	0	0	1	-360	360;
	3958	7564	0.00161	0.02118	0.66012	1156.95088	1156.95088	1156.95088	0	0	1	-360	360;
	1031	7564	0.00312	0.03268	0.70522	1258.80741	1258.80741	1258.80741	0	0	1	-360	360;
	1031	7564	0.00305	0.0333	0.69052	1387.72302	1387.72302	1387.72302	0	0	1	-360	360;
	5653	5041	0.000182	0.000831	0.00027	105.853159	105.853159	105.853159	0	0	1	-360	360;
	8008	5653	0.013649	0.054609	0.01944	105.853159	105.853159	105.853159	0	0	1	-360	360;
	8154	7097	0.007209	0.03252	0.01026	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8005	8154	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7097	5041	0.000169	0.000769	0.00026	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2484	5041	0.011649	0.026778	0.00911	118.998386	118.998386	118.998386	0	0	1	-360	360;
	7485	2484	0.004049	0.01112	0.0029	221.09005	221.09005	221.09005	0	0	1	-360	360;
	7485	5041	0.007898	0.037338	0.01202	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7485	5041	0.00772	0.034578	0.01175	107.643494	107.643494	107.643494	0	0	1	-360	360;
	5041	4271	0.042582	0.193609	0.06482	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5041	7594	0.027591	0.130569	0.042	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5041	5036	0.018431	0.065898	0.02806	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5041	26	0.02836	0.1134	0.04036	134.235843	134.235843	134.235843	0	0	1	-360	360;
	5041	872	0.046142	0.15844	0.07022	132.421087	132.421087	132.421087	0	0	1	-360	360;
	6835	5041	0.011409	0.027311	0.00893	281.707325	281.707325	281.707325	0	0	1	-360	360;
	5041	7902	0.019271	0.063569	0.02934	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5041	7902	0.02212	0.082329	0.03368	111.511769	111.511769	111.511769	0	0	1	-360	360;
	2693	4358	0.018249	0.04624	0.01303	103.923048	103.923048	103.923048	0	0	1	-360	360;
	605	4358	0.008929	0.01872	0.29574	251.062497	251.062497	251.062497	0	0	1	-360	360;
	605	4358	0.008742	0.017329	0.28931	253.469614	253.469614	253.469614	0	0	1	-360	360;
	1437	5394	0.001569	0.00412	0.00111	150.196085	150.196085	150.196085	0	0	1	-360	360;
	5394	7098	0.037409	0.106911	0.02673	167.568121	167.568121	167.568121	0	0	1	-360	360;
	1437	7599	0.01836	0.049671	0.01312	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3704	3275	0.007791	0.037502	0.01109	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3704	1356	0.014911	0.051218	0.0227	122.244162	122.244162	122.244162	0	0	1	-360	360;
	2223	4454	0.007071	0.0294	0.01006	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2223	6698	0.010169	0.039378	0.01547	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7485	7707	0	0.00028	0	112.528916	112.528916	112.528916	0	0	1	-360	360;
	7485	5890	0.004449	0.00924	0.00319	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3360	4843	0.002191	0.011142	0.00312	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4843	7325	0.00244	0.011951	0.00348	132.609187	132.609187	132.609187	0	0	1	-360	360;
	5611	6515	0.004662	0.020871	0.00709	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3360	6515	0.00528	0.020449	0.00804	119.839643	119.839643	119.839643	0	0	1	-360	360;
	5477	6515	0.0036	0.0144	0.00513	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2582	3517	0.00212	0.008871	0.00321	112.204157	112.204157	112.204157	0	0	1	-360	360;
	5576	2582	0.001991	0.008271	0.00283	111.964094	111.964094	111.964094	0	0	1	-360	360;
	345	3517	0.002071	0.009049	0.00315	125.994225	125.994225	125.994225	0	0	1	-360	360;
	345	3584	0.002031	0.008098	0.00288	125.79859	125.79859	125.79859	0	0	1	-360	360;
	2036	2444	0.002631	0.00868	0.004	132.477205	132.477205	132.477205	0	0	1	-360	360;
	2658	2444	0.00048	0.001769	0.00073	153.111906	153.111906	153.111906	0	0	1	-360	360;
	2036	5330	0.002338	0.009831	0.00357	250.554573	250.554573	250.554573	0	0	1	-360	360;
	5262	5330	0.006911	0.025742	0.01052	198.936774	198.936774	198.936774	0	0	1	-360	360;
	7524	2036	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2179	3105	0.005649	0.020209	0.00861	129.266502	129.266502	129.266502	0	0	1	-360	360;
	4175	3105	0.037822	0.075338	0.02702	171.309611	171.309611	171.309611	0	0	1	-360	360;
	8323	2179	0	0.000222	0	127.575415	127.575415	127.575415	0	0	1	-360	360;
	2179	6440	0.005609	0.020871	0.00854	117.689215	117.689215	117.689215	0	0	1	-360	360;
	2179	6906	0.042569	0.11516	0.03041	250.494817	250.494817	250.494817	0	0	1	-360	360;
	1401	6440	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7669	5210	0.00396	0.010458	0.00283	120.780406	120.780406	120.780406	0	0	1	-360	360;
	6349	5210	0.019671	0.088151	0.02995	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1106	7786	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1647	4948	0.00912	0.03364	0.01297	104.09686	104.09686	104.09686	0	0	1	-360	360;
	4948	6636	0.010942	0.051369	0.01558	104.519307	104.519307	104.519307	0	0	1	-360	360;
	2699	7786	0.000182	0.000822	0.00028	132.319762	132.319762	132.319762	0	0	1	-360	360;
	2699	7790	0.023209	0.118049	0.03304	132.742209	132.742209	132.742209	0	0	1	-360	360;
	1647	7786	0.0002	0.000809	0.00032	106.582959	106.582959	106.582959	0	0	1	-360	360;
	1943	5658	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1397	6847	0.00076	0.00811	0.17059	1473.54565	1473.54565	1473.54565	0	0	1	-360	360;
	7387	2568	5.8e-05	0.000231	0.0064	116.140242	116.140242	116.140242	0	0	1	-360	360;
	6906	7387	0.013471	0.053849	0.01917	116.406285	116.406285	116.406285	0	0	1	-360	360;
	6624	6906	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1492	1705	0.007489	0.03376	0.01065	126.377442	126.377442	126.377442	0	0	1	-360	360;
	9	1492	0.004271	0.016431	0.00608	134.938363	134.938363	134.938363	0	0	1	-360	360;
	3647	4650	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7925	4650	0.013289	0.04384	0.02023	161.569163	161.569163	161.569163	0	0	1	-360	360;
	7925	4650	0.01184	0.049689	0.01802	130.067749	130.067749	130.067749	0	0	1	-360	360;
	7925	4650	0.011338	0.0508	0.01726	163.775969	163.775969	163.775969	0	0	1	-360	360;
	3486	4650	0.008502	0.035671	0.01294	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4650	3265	0.02216	0.088622	0.03154	156.819361	156.819361	156.819361	0	0	1	-360	360;
	4650	4900	0.019911	0.101262	0.02834	153.32287	153.32287	153.32287	0	0	1	-360	360;
	4650	399	0.014551	0.068258	0.0207	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4650	4027	0.014129	0.050529	0.02151	125.373285	125.373285	125.373285	0	0	1	-360	360;
	4650	6694	0.018022	0.08072	0.02742	128.178948	128.178948	128.178948	0	0	1	-360	360;
	468	6906	0.005382	0.02004	0.0082	162.311434	162.311434	162.311434	0	0	1	-360	360;
	8245	584	0.007769	0.034791	0.01183	103.923048	103.923048	103.923048	0	0	1	-360	360;
	584	6349	0.068351	0.30616	0.10403	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1632	4905	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1270	8252	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	218	2083	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	11	392	0.00163	0.01819	0.66584	1597.08591	1597.08591	1597.08591	0	0	1	-360	360;
	5337	11	0.00241	0.02622	0.54384	1269.18756	1269.18756	1269.18756	0	0	1	-360	360;
	6114	11	0.00236	0.02672	0.53228	1354.90686	1354.90686	1354.90686	0	0	1	-360	360;
	11	2600	0.00188	0.02029	0.42683	1028.65857	1028.65857	1028.65857	0	0	1	-360	360;
	11	2600	0.00213	0.01988	0.48375	1044.98799	1044.98799	1044.98799	0	0	1	-360	360;
	1348	2654	0	0.000222	0	123.994746	123.994746	123.994746	0	0	1	-360	360;
	916	2654	0.009822	0.046449	0.01494	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6707	2654	0.00344	0.01628	0.00523	147.719599	147.719599	147.719599	0	0	1	-360	360;
	2654	1579	0.00648	0.025102	0.00986	123.977079	123.977079	123.977079	0	0	1	-360	360;
	3521	5758	0	0.00028	0	124.506567	124.506567	124.506567	0	0	1	-360	360;
	5758	163	0.02368	0.099338	0.03603	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2031	5758	0.017169	0.061382	0.02614	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1397	8026	0.00209	0.01796	0.47366	1236.43261	1236.43261	1236.43261	0	0	1	-360	360;
	1397	8026	0.00213	0.01758	0.4826	1219.59309	1219.59309	1219.59309	0	0	1	-360	360;
	4909	8026	0.0013	0.01389	0.29229	1185.7528	1185.7528	1185.7528	0	0	1	-360	360;
	4058	8026	0.00147	0.01361	0.33127	1295.40679	1295.40679	1295.40679	0	0	1	-360	360;
	547	3958	0.00052	0.00446	0.11749	1260.08231	1260.08231	1260.08231	0	0	1	-360	360;
	547	3958	0.00053	0.00436	0.11971	1192.67487	1192.67487	1192.67487	0	0	1	-360	360;
	2590	7016	0.014889	0.05544	0.02267	155.852357	155.852357	155.852357	0	0	1	-360	360;
	2590	7016	0.015178	0.054289	0.02311	123.481886	123.481886	123.481886	0	0	1	-360	360;
	2019	2590	0.003111	0.013569	0.00473	126.767933	126.767933	126.767933	0	0	1	-360	360;
	5736	2590	0.000791	0.00376	0.00121	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2590	6781	0.002751	0.009809	0.00418	244.691754	244.691754	244.691754	0	0	1	-360	360;
	2590	8240	0.007129	0.024471	0.01084	142.147245	142.147245	142.147245	0	0	1	-360	360;
	2590	7996	0.017951	0.0668	0.02732	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2590	4698	0.006178	0.02812	0.00942	149.714142	149.714142	149.714142	0	0	1	-360	360;
	2590	1539	0.006049	0.028649	0.00922	161.800132	161.800132	161.800132	0	0	1	-360	360;
	5512	4534	0.00396	0.03136	0.89602	1264.76723	1264.76723	1264.76723	0	0	1	-360	360;
	5512	4534	0.00353	0.03554	0.79827	1130.74152	1130.74152	1130.74152	0	0	1	-360	360;
	5512	4534	0.00387	0.0331	0.87257	1310.75619	1310.75619	1310.75619	0	0	1	-360	360;
	5512	4534	0.00394	0.03239	0.88903	1270.72375	1270.72375	1270.72375	0	0	1	-360	360;
	1802	1144	0.00225	0.01786	0.51036	1104.14581	1104.14581	1104.14581	0	0	1	-360	360;
	1144	2780	0.00124	0.01101	0.27878	1182.12953	1182.12953	1182.12953	0	0	1	-360	360;
	1802	4058	0.00201	0.02024	0.45468	1203.78494	1203.78494	1203.78494	0	0	1	-360	360;
	1802	5679	0.00197	0.02063	0.4454	1108.35487	1108.35487	1108.35487	0	0	1	-360	360;
	5679	2780	0.00126	0.01078	0.28414	1188.16569	1188.16569	1188.16569	0	0	1	-360	360;
	1802	4909	0.00193	0.02103	0.43612	1260.83132	1260.83132	1260.83132	0	0	1	-360	360;
	4475	2373	0.001209	0.004551	0.15228	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2373	1201	0.000449	0.002018	0.00068	142.419783	142.419783	142.419783	0	0	1	-360	360;
	962	2373	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5262	3441	0.007049	0.0252	0.01073	120.043332	120.043332	120.043332	0	0	1	-360	360;
	6942	5262	0.001191	0.003222	0.1485	136.473306	136.473306	136.473306	0	0	1	-360	360;
	5262	777	0.000511	0.001978	0.00078	154.819361	154.819361	154.819361	0	0	1	-360	360;
	5262	6390	0.001009	0.004782	0.00154	168.520576	168.520576	168.520576	0	0	1	-360	360;
	5262	7756	0.000991	0.004422	0.0015	118.231693	118.231693	118.231693	0	0	1	-360	360;
	3076	2338	0.000929	0.003098	0.00143	135.597754	135.597754	135.597754	0	0	1	-360	360;
	3436	3076	0.005858	0.0246	0.00893	152.936795	152.936795	152.936795	0	0	1	-360	360;
	1394	8084	0	0.00028	0	195.222824	195.222824	195.222824	0	0	1	-360	360;
	2461	8084	0	0.00028	0	209.206969	209.206969	209.206969	0	0	1	-360	360;
	4688	5170	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5170	1785	0.002031	0.009222	0.00308	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5414	5170	0.01036	0.045231	0.01576	120.196618	120.196618	120.196618	0	0	1	-360	360;
	5170	2285	0.026858	0.062822	0.01919	257.680057	257.680057	257.680057	0	0	1	-360	360;
	735	1785	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1785	2415	0.021569	0.098089	0.03284	114.176356	114.176356	114.176356	0	0	1	-360	360;
	7665	8084	0	0.000222	0	136.33301	136.33301	136.33301	0	0	1	-360	360;
	3795	4800	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4800	8073	0.000418	0.001769	0.00065	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4800	6636	0.000409	0.001809	0.00063	104.519307	104.519307	104.519307	0	0	1	-360	360;
	2736	4201	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5243	6150	0	0.00028	0	142.675174	142.675174	142.675174	0	0	1	-360	360;
	7572	4190	0.000569	0.00168	0.07155	132.951874	132.951874	132.951874	0	0	1	-360	360;
	7572	7919	0.000582	0.001649	0.0729	147.44706	147.44706	147.44706	0	0	1	-360	360;
	7919	4176	0.001858	0.006911	0.00283	149.08177	149.08177	149.08177	0	0	1	-360	360;
	8372	7572	0	0.000222	0	111.872122	111.872122	111.872122	0	0	1	-360	360;
	4584	247	0.0154	0.006022	0.37744	297.277336	297.277336	297.277336	0	0	1	-360	360;
	3155	5373	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5373	8257	8e-05	0.0004	0.00013	121.836264	121.836264	121.836264	0	0	1	-360	360;
	5373	921	9.8e-05	0.000329	0.00014	145.496165	145.496165	145.496165	0	0	1	-360	360;
	2334	8257	0.023129	0.076289	0.0352	194.581359	194.581359	194.581359	0	0	1	-360	360;
	468	921	0.011258	0.047249	0.01714	152.993693	152.993693	152.993693	0	0	1	-360	360;
	2334	5715	0.018822	0.072271	0.02678	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2334	7042	0.002009	0.006649	0.00307	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2334	4062	0.000942	0.00424	0.00134	132.586324	132.586324	132.586324	0	0	1	-360	360;
	2334	7482	0.017711	0.060809	0.02696	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2334	7482	0.01804	0.059489	0.02746	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5337	5512	0.0028	0.03055	0.63332	1255.50072	1255.50072	1255.50072	0	0	1	-360	360;
	5337	5512	0.00274	0.03113	0.61985	1405.45635	1405.45635	1405.45635	0	0	1	-360	360;
	1286	4717	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7790	262	0.0002	0.000942	0.0003	142.401077	142.401077	142.401077	0	0	1	-360	360;
	3562	5771	0.004351	0.014929	0.00662	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3562	7790	0.004831	0.017289	0.00736	135.264421	135.264421	135.264421	0	0	1	-360	360;
	4362	3845	0.014289	0.067609	0.02174	136.21168	136.21168	136.21168	0	0	1	-360	360;
	4362	3845	0.013978	0.0626	0.02127	171.034734	171.034734	171.034734	0	0	1	-360	360;
	3304	3709	0.004431	0.014609	0.00674	145.779615	145.779615	145.779615	0	0	1	-360	360;
	3304	3715	0.00664	0.025711	0.0101	131.61906	131.61906	131.61906	0	0	1	-360	360;
	3304	6446	0.001289	0.004422	0.00197	172.150608	172.150608	172.150608	0	0	1	-360	360;
	3304	1668	0.01448	0.065822	0.02204	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3304	2997	0.004631	0.017938	0.00705	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3304	2997	0.00492	0.01692	0.0075	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6349	4717	0.027511	0.120182	0.04188	147.797021	147.797021	147.797021	0	0	1	-360	360;
	5794	4717	0.004	0.014898	0.00609	115.942269	115.942269	115.942269	0	0	1	-360	360;
	6969	796	0.000351	0.001591	0.0005	123.096071	123.096071	123.096071	0	0	1	-360	360;
	6349	796	0.018929	0.084791	0.02881	123.096071	123.096071	123.096071	0	0	1	-360	360;
	1718	674	0.013502	0.063889	0.02055	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1705	4877	0.01632	0.057871	0.02323	147.719339	147.719339	147.719339	0	0	1	-360	360;
	7094	1967	0	0.000222	0	140.931605	140.931605	140.931605	0	0	1	-360	360;
	2388	1967	0.01176	0.04704	0.01674	126.372246	126.372246	126.372246	0	0	1	-360	360;
	2836	7098	0	0.000222	0	145.186474	145.186474	145.186474	0	0	1	-360	360;
	8207	7098	0.02892	0.057622	0.02067	259.813597	259.813597	259.813597	0	0	1	-360	360;
	2495	4246	0.001711	0.0078	0.00261	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2517	4246	0.01224	0.042022	0.01862	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5877	1997	0.007769	0.061249	0.45691	260.958049	260.958049	260.958049	0	0	1	-360	360;
	8134	5162	0.001391	0.0048	0.00213	110.657002	110.657002	110.657002	0	0	1	-360	360;
	4141	8134	0.004071	0.01988	0.00579	110.657002	110.657002	110.657002	0	0	1	-360	360;
	7109	5162	0.001418	0.004702	0.00217	136.648157	136.648157	136.648157	0	0	1	-360	360;
	7109	2807	0.012711	0.048809	0.01808	104.032168	104.032168	104.032168	0	0	1	-360	360;
	4865	1855	0	0.000222	0	106.583219	106.583219	106.583219	0	0	1	-360	360;
	7382	5162	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1855	5162	0.00448	0.016698	0.00683	112.794699	112.794699	112.794699	0	0	1	-360	360;
	6297	4141	0.014689	0.052049	0.0209	110.503716	110.503716	110.503716	0	0	1	-360	360;
	1855	156	0.025631	0.084538	0.03902	111.34809	111.34809	111.34809	0	0	1	-360	360;
	8005	2381	0.005431	0.026551	0.00774	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2381	4027	0.013151	0.058898	0.02002	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4606	7485	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3300	163	0.00036	0.001329	0.00052	121.803529	121.803529	121.803529	0	0	1	-360	360;
	5384	7621	0	0.00028	0	234.814128	234.814128	234.814128	0	0	1	-360	360;
	7695	5384	0	0.00028	0	234.814128	234.814128	234.814128	0	0	1	-360	360;
	7621	5248	0.00612	0.014671	0.38151	234.814128	234.814128	234.814128	0	0	1	-360	360;
	6557	7695	0.006942	0.014378	0.43238	288.451411	288.451411	288.451411	0	0	1	-360	360;
	5625	2974	4.9e-05	0.00028	7e-05	108.246507	108.246507	108.246507	0	0	1	-360	360;
	3739	5625	0.002071	0.008289	0.00295	108.246507	108.246507	108.246507	0	0	1	-360	360;
	4767	2974	4.9e-05	0.00028	7e-05	153.907437	153.907437	153.907437	0	0	1	-360	360;
	4767	6290	0.000729	0.003049	0.00104	134.178945	134.178945	134.178945	0	0	1	-360	360;
	2246	2974	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3094	8336	0.000409	0.00184	0.00062	112.219225	112.219225	112.219225	0	0	1	-360	360;
	2693	8336	0.0004	0.00188	0.00061	120.980198	120.980198	120.980198	0	0	1	-360	360;
	404	6247	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6247	6186	0.003982	0.020249	0.00567	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6247	7788	0.006498	0.031791	0.00926	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8019	5808	0.000129	0.000609	0.00019	123.799371	123.799371	123.799371	0	0	1	-360	360;
	8019	556	0.00036	0.000729	0.00025	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8019	7875	0.000129	0.000569	0.00019	126.333015	126.333015	126.333015	0	0	1	-360	360;
	982	8019	0	0.00028	0	152.298448	152.298448	152.298448	0	0	1	-360	360;
	7875	486	0.006862	0.030942	0.00977	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3232	4012	0.0016	0.00716	0.00243	138.212718	138.212718	138.212718	0	0	1	-360	360;
	3232	7338	0.001849	0.006542	0.00263	103.923048	103.923048	103.923048	0	0	1	-360	360;
	697	4012	0.001809	0.007018	0.00275	103.923048	103.923048	103.923048	0	0	1	-360	360;
	697	2689	0.004302	0.01588	0.00612	133.261304	133.261304	133.261304	0	0	1	-360	360;
	2958	4012	0	0.000222	0	121.347306	121.347306	121.347306	0	0	1	-360	360;
	7016	6218	0.008831	0.036751	0.01257	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7016	56	0.011031	0.045942	0.01571	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5950	7016	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	26	7016	0.043418	0.124111	0.03102	221.186699	221.186699	221.186699	0	0	1	-360	360;
	8236	8073	0.011698	0.048702	0.01666	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7925	3486	0.005058	0.02572	0.00719	245.897261	245.897261	245.897261	0	0	1	-360	360;
	7329	3486	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1245	5462	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7940	7397	0.008969	0.032062	0.01364	135.57723	135.57723	135.57723	0	0	1	-360	360;
	2138	6370	0.000489	0.002369	0.03105	110.000988	110.000988	110.000988	0	0	1	-360	360;
	2879	6370	0.0028	0.0096	0.00426	108.656484	108.656484	108.656484	0	0	1	-360	360;
	2138	4279	0.00048	0.002191	0.03038	120.645046	120.645046	120.645046	0	0	1	-360	360;
	2879	4279	0.002849	0.0094	0.00433	130.559565	130.559565	130.559565	0	0	1	-360	360;
	4295	2138	0	0.000222	0	124.719349	124.719349	124.719349	0	0	1	-360	360;
	3789	7940	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7404	7940	0.005138	0.023391	0.00783	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3032	7940	0.005031	0.023831	0.00766	119.060999	119.060999	119.060999	0	0	1	-360	360;
	5467	7940	0.00492	0.022062	0.0075	110.549182	110.549182	110.549182	0	0	1	-360	360;
	7940	4487	0.012449	0.03368	0.00889	141.554624	141.554624	141.554624	0	0	1	-360	360;
	3906	4821	0	0.000222	0	112.797557	112.797557	112.797557	0	0	1	-360	360;
	1250	4821	0.005582	0.021622	0.0085	128.576453	128.576453	128.576453	0	0	1	-360	360;
	4821	1766	0.011769	0.052729	0.01792	108.126216	108.126216	108.126216	0	0	1	-360	360;
	1639	6908	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7812	6908	0.005489	0.019618	0.00835	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6765	6908	0.005591	0.0192	0.00851	112.016056	112.016056	112.016056	0	0	1	-360	360;
	6378	1386	0	0.00028	0	118.716235	118.716235	118.716235	0	0	1	-360	360;
	6618	7404	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3032	7129	0	0.000222	0	118.724548	118.724548	118.724548	0	0	1	-360	360;
	1386	7129	0	0.00028	0	118.855232	118.855232	118.855232	0	0	1	-360	360;
	3250	967	0	0.000222	0	110.57932	110.57932	110.57932	0	0	1	-360	360;
	8299	3250	0	0.00028	0	110.580359	110.580359	110.580359	0	0	1	-360	360;
	967	6505	0	0.000222	0	110.372513	110.372513	110.372513	0	0	1	-360	360;
	6505	5467	0	0.000222	0	147.997333	147.997333	147.997333	0	0	1	-360	360;
	344	6237	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4682	447	0	0.000222	0	123.009815	123.009815	123.009815	0	0	1	-360	360;
	344	8299	0	0.00028	0	110.580359	110.580359	110.580359	0	0	1	-360	360;
	3642	1118	0	0.000222	0	128.62218	128.62218	128.62218	0	0	1	-360	360;
	289	3642	0.012969	0.046369	0.01975	134.301055	134.301055	134.301055	0	0	1	-360	360;
	1118	3353	0	0.000222	0	139.793388	139.793388	139.793388	0	0	1	-360	360;
	3353	1250	0	0.000222	0	128.420309	128.420309	128.420309	0	0	1	-360	360;
	447	6618	0	0.000222	0	134.316123	134.316123	134.316123	0	0	1	-360	360;
	7278	6773	0.003151	0.00128	0.07715	162.096053	162.096053	162.096053	0	0	1	-360	360;
	1397	282	0.00154	0.01649	0.3469	1421.64163	1421.64163	1421.64163	0	0	1	-360	360;
	2879	5313	0.001298	0.00568	0.00198	153.207255	153.207255	153.207255	0	0	1	-360	360;
	2879	1674	0.0002	0.000871	0.0003	148.804036	148.804036	148.804036	0	0	1	-360	360;
	7956	7188	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8245	7188	4e-05	0.00028	7e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2730	7188	4e-05	0.00028	7e-05	126.070349	126.070349	126.070349	0	0	1	-360	360;
	2730	7790	0.010071	0.039	0.01533	126.185444	126.185444	126.185444	0	0	1	-360	360;
	3940	4987	0.000218	0.000849	0.00034	131.164397	131.164397	131.164397	0	0	1	-360	360;
	1683	4987	0.000231	0.00084	0.00034	134.623736	134.623736	134.623736	0	0	1	-360	360;
	7813	4987	0.000231	0.000822	0.00035	128.246498	128.246498	128.246498	0	0	1	-360	360;
	90	4987	0.00024	0.0008	0.00036	110.255599	110.255599	110.255599	0	0	1	-360	360;
	6401	4987	0.00024	0.000782	0.00036	118.197918	118.197918	118.197918	0	0	1	-360	360;
	4987	3469	0.000209	0.000889	0.00032	140.242075	140.242075	140.242075	0	0	1	-360	360;
	8	4987	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2912	4987	0.000209	0.000911	0.00032	146.39432	146.39432	146.39432	0	0	1	-360	360;
	4347	1683	0.00784	0.025862	0.01193	133.505783	133.505783	133.505783	0	0	1	-360	360;
	7813	5321	0.001991	0.009391	0.00302	107.006445	107.006445	107.006445	0	0	1	-360	360;
	90	4560	0.003591	0.01184	0.00547	110.185451	110.185451	110.185451	0	0	1	-360	360;
	6997	485	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1397	1826	0.00013	0.00077	1.1635	954.363286	954.363286	954.363286	0	0	1	-360	360;
	965	1826	0.00013	0.00076	1.18504	1007.75282	1007.75282	1007.75282	0	0	1	-360	360;
	826	1826	0.0002	0.00127	1.74286	991.816321	991.816321	991.816321	0	0	1	-360	360;
	3718	4416	0.002449	0.006871	0.25648	160.788961	160.788961	160.788961	0	0	1	-360	360;
	2301	6654	0.00132	0.004929	0.16497	263.264362	263.264362	263.264362	0	0	1	-360	360;
	5374	2172	0	0.00028	0	129.114255	129.114255	129.114255	0	0	1	-360	360;
	2172	2285	0.036111	0.084471	0.0258	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2172	2499	0.043129	0.093129	0.03081	211.828428	211.828428	211.828428	0	0	1	-360	360;
	5462	7660	0.001071	0.004129	0.00153	165.402105	165.402105	165.402105	0	0	1	-360	360;
	1061	7660	0.01016	0.036298	0.01545	145.94797	145.94797	145.94797	0	0	1	-360	360;
	4386	5462	0.010951	0.055689	0.01558	133.312487	133.312487	133.312487	0	0	1	-360	360;
	3553	6429	0.005831	0.022591	0.00887	103.923048	103.923048	103.923048	0	0	1	-360	360;
	84	6429	0.016542	0.075222	0.02518	144.923289	144.923289	144.923289	0	0	1	-360	360;
	6429	4832	0.015938	0.00716	0.57456	127.225194	127.225194	127.225194	0	0	1	-360	360;
	4929	2368	0	0.000222	0	190.753354	190.753354	190.753354	0	0	1	-360	360;
	2368	492	0.009591	0.02332	0.5977	191.380009	191.380009	191.380009	0	0	1	-360	360;
	4929	5248	0	0.000222	0	190.753354	190.753354	190.753354	0	0	1	-360	360;
	2415	3345	0.03828	0.10096	0.02735	126.142576	126.142576	126.142576	0	0	1	-360	360;
	84	3345	0.03036	0.086778	0.02169	271.433752	271.433752	271.433752	0	0	1	-360	360;
	3013	1173	0	0.00028	0	151.013959	151.013959	151.013959	0	0	1	-360	360;
	7134	2019	0.002182	0.010329	0.00332	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7134	4271	0.002129	0.009569	0.00325	124.765855	124.765855	124.765855	0	0	1	-360	360;
	5136	2783	0.0018	0.00668	0.00274	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2350	2783	0.004569	0.01636	0.00696	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5136	2885	0.001831	0.006542	0.00279	132.532024	132.532024	132.532024	0	0	1	-360	360;
	2350	3929	0.006338	0.026609	0.00965	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6306	6349	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6349	5517	0.016231	0.070902	0.02471	121.61205	121.61205	121.61205	0	0	1	-360	360;
	6349	7528	0.010138	0.046102	0.01543	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5355	2031	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2031	8260	0.006431	0.028991	0.00915	135.175307	135.175307	135.175307	0	0	1	-360	360;
	2495	5647	0.01184	0.053022	0.01802	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2495	2517	0.012529	0.044422	0.01783	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6883	5112	0.00436	0.010689	0.00341	242.084844	242.084844	242.084844	0	0	1	-360	360;
	4575	6883	0.019529	0.055831	0.01396	242.767099	242.767099	242.767099	0	0	1	-360	360;
	2894	5112	0.004942	0.01048	0.00387	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7306	5112	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1157	2198	0.006431	0.01764	0.00459	230.379472	230.379472	230.379472	0	0	1	-360	360;
	4203	1157	4e-05	0.00028	7e-05	118.215585	118.215585	118.215585	0	0	1	-360	360;
	8081	2198	0.003431	0.013218	0.0049	127.741692	127.741692	127.741692	0	0	1	-360	360;
	4999	2198	0	0.00028	0	122.946422	122.946422	122.946422	0	0	1	-360	360;
	2198	5398	0.00708	0.014702	0.00507	265.50806	265.50806	265.50806	0	0	1	-360	360;
	3299	2198	0.065422	0.130311	0.04674	131.82249	131.82249	131.82249	0	0	1	-360	360;
	8081	3240	0.003049	0.013831	0.00463	131.231168	131.231168	131.231168	0	0	1	-360	360;
	8252	8081	0.011209	0.051009	0.01708	125.614906	125.614906	125.614906	0	0	1	-360	360;
	737	8081	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7258	8081	0.00504	0.024622	0.00717	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2083	3240	0.009049	0.033671	0.01377	144.230642	144.230642	144.230642	0	0	1	-360	360;
	5138	3240	0	0.00028	0	145.528121	145.528121	145.528121	0	0	1	-360	360;
	4203	6025	4e-05	0.00028	7e-05	144.768184	144.768184	144.768184	0	0	1	-360	360;
	4364	4203	0	0.00028	0	127.31197	127.31197	127.31197	0	0	1	-360	360;
	8252	5398	0.007631	0.015569	0.00598	253.829188	253.829188	253.829188	0	0	1	-360	360;
	8252	2225	0.000822	0.00316	0.00124	134.696742	134.696742	134.696742	0	0	1	-360	360;
	8252	6756	0.017751	0.060929	0.02701	128.441353	128.441353	128.441353	0	0	1	-360	360;
	2083	6025	0.007418	0.016022	0.0053	265.277091	265.277091	265.277091	0	0	1	-360	360;
	2083	8210	0.007542	0.035689	0.01148	131.800406	131.800406	131.800406	0	0	1	-360	360;
	8109	423	0	0.000222	0	108.830814	108.830814	108.830814	0	0	1	-360	360;
	423	1179	0.000338	0.001218	0.00052	126.692069	126.692069	126.692069	0	0	1	-360	360;
	7761	158	0	0.000222	0	117.221821	117.221821	117.221821	0	0	1	-360	360;
	158	7075	0.000351	0.001191	0.00053	183.81623	183.81623	183.81623	0	0	1	-360	360;
	6298	1061	0.004662	0.016009	0.0071	145.763507	145.763507	145.763507	0	0	1	-360	360;
	6298	7606	0.017489	0.084191	0.02489	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1061	25	0.003809	0.013111	0.00581	138.165433	138.165433	138.165433	0	0	1	-360	360;
	1061	5530	0.003489	0.013502	0.0053	134.966162	134.966162	134.966162	0	0	1	-360	360;
	3300	1061	0.004738	0.015658	0.00723	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2898	1061	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1061	2329	0.010351	0.035529	0.01575	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1061	5775	0.00388	0.012822	0.00592	126.834963	126.834963	126.834963	0	0	1	-360	360;
	1179	25	0.003729	0.013218	0.00531	150.547345	150.547345	150.547345	0	0	1	-360	360;
	5530	6435	0.002418	0.008298	0.00368	131.386792	131.386792	131.386792	0	0	1	-360	360;
	7075	6435	0.001538	0.007822	0.00219	127.458761	127.458761	127.458761	0	0	1	-360	360;
	1173	5724	0.005369	0.010702	0.00384	266.310346	266.310346	266.310346	0	0	1	-360	360;
	2283	6733	0.00088	0.0075	0.19784	1492.80464	1492.80464	1492.80464	0	0	1	-360	360;
	2283	6733	0.0009	0.00734	0.20157	1436.03272	1436.03272	1436.03272	0	0	1	-360	360;
	1910	6557	0	0.00028	0	288.451411	288.451411	288.451411	0	0	1	-360	360;
	1430	7560	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1910	7560	0	0.00028	0	288.451411	288.451411	288.451411	0	0	1	-360	360;
	5400	7560	0.018831	0.072951	0.02866	118.742215	118.742215	118.742215	0	0	1	-360	360;
	5663	7560	0.012351	0.042409	0.01879	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4575	3553	0.001938	0.008698	0.00295	112.425253	112.425253	112.425253	0	0	1	-360	360;
	1874	4575	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5839	7850	0.017871	0.047151	0.01277	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7850	605	0.00076	0.002098	0.00055	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4877	4584	0	0.000222	0	180.462893	180.462893	180.462893	0	0	1	-360	360;
	7325	2329	0.005249	0.024631	0.00747	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7340	7325	0	0.00028	0	116.163365	116.163365	116.163365	0	0	1	-360	360;
	1577	3636	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7711	3636	0.00016	0.000502	0.00023	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3455	7711	0.008658	0.037809	0.01318	121.681159	121.681159	121.681159	0	0	1	-360	360;
	5517	3636	0.000142	0.00056	0.00021	121.076327	121.076327	121.076327	0	0	1	-360	360;
	6136	5839	4e-05	0.00028	7e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6136	5087	4e-05	0.00028	7e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6781	2870	0.0002	0.00092	0.00031	127.253513	127.253513	127.253513	0	0	1	-360	360;
	2870	1193	0.002542	0.010649	0.00386	178.481081	178.481081	178.481081	0	0	1	-360	360;
	5736	26	0.054	0.121351	0.03857	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5978	6781	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3143	3904	0.001969	0.004271	0.00141	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3143	3299	0.009111	0.025009	0.0065	288.424911	288.424911	288.424911	0	0	1	-360	360;
	1585	3143	0.027502	0.074369	0.01964	129.670503	129.670503	129.670503	0	0	1	-360	360;
	7142	6145	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6145	3265	0.0002	0.000942	0.0003	166.139699	166.139699	166.139699	0	0	1	-360	360;
	6145	6186	0.0002	0.000871	0.0003	108.33718	108.33718	108.33718	0	0	1	-360	360;
	1813	8330	0.00052	0.00136	0.00036	133.171671	133.171671	133.171671	0	0	1	-360	360;
	2888	1813	0.020471	0.07928	0.03116	134.318202	134.318202	134.318202	0	0	1	-360	360;
	4404	765	3.1e-05	0.00028	5e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5650	765	0.015609	0.076289	0.02222	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4404	7258	3.1e-05	0.00028	5e-05	132.20077	132.20077	132.20077	0	0	1	-360	360;
	2156	4404	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7646	6608	0.008809	0.041342	0.01254	132.074763	132.074763	132.074763	0	0	1	-360	360;
	7885	6608	0.011698	0.017782	0.41382	110.235075	110.235075	110.235075	0	0	1	-360	360;
	5397	7646	0.029769	0.069649	0.02127	268.430376	268.430376	268.430376	0	0	1	-360	360;
	5287	5180	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5400	5180	0.006738	0.024089	0.01025	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5287	7370	0	0.00028	0	154.442121	154.442121	154.442121	0	0	1	-360	360;
	1252	5287	0	0.000222	0	127.1366	127.1366	127.1366	0	0	1	-360	360;
	7803	3369	0	0.000222	0	114.651025	114.651025	114.651025	0	0	1	-360	360;
	5591	2301	0	0.000222	0	109.248325	109.248325	109.248325	0	0	1	-360	360;
	3369	2301	0	0.000222	0	178.57643	178.57643	178.57643	0	0	1	-360	360;
	800	3369	0.001062	0.003649	0.1323	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2415	1628	0.00204	0.009129	0.0031	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2415	1628	0.00232	0.008951	0.00352	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2415	5427	0.00152	0.006911	0.00231	108.027489	108.027489	108.027489	0	0	1	-360	360;
	2415	5427	0.001489	0.00704	0.00226	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2415	5427	0.001462	0.00652	0.00221	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2415	7935	0.001391	0.003609	0.00109	269.768645	269.768645	269.768645	0	0	1	-360	360;
	2415	5415	0.001658	0.003311	0.00119	129.171673	129.171673	129.171673	0	0	1	-360	360;
	2415	7389	0.000991	0.004329	0.00151	113.547622	113.547622	113.547622	0	0	1	-360	360;
	2415	2439	0.010369	0.038569	0.01578	120.722729	120.722729	120.722729	0	0	1	-360	360;
	1213	1628	0	0.000222	0	161.471476	161.471476	161.471476	0	0	1	-360	360;
	1213	5474	0.026809	0.126889	0.04081	142.14023	142.14023	142.14023	0	0	1	-360	360;
	7829	1628	0	0.000222	0	168.693608	168.693608	168.693608	0	0	1	-360	360;
	7829	7668	0.026218	0.117489	0.03992	163.148534	163.148534	163.148534	0	0	1	-360	360;
	5427	1628	0.00064	0.002662	0.00097	130.938105	130.938105	130.938105	0	0	1	-360	360;
	5427	1628	0.000622	0.002711	0.00095	132.30781	132.30781	132.30781	0	0	1	-360	360;
	5427	75	0.000609	0.002271	0.00093	138.773903	138.773903	138.773903	0	0	1	-360	360;
	4217	7935	0.002098	0.005702	0.0015	269.768645	269.768645	269.768645	0	0	1	-360	360;
	1985	5415	0.002582	0.006031	0.00184	133.471489	133.471489	133.471489	0	0	1	-360	360;
	3726	1985	0.019458	0.055618	0.0139	129.126986	129.126986	129.126986	0	0	1	-360	360;
	1322	7389	0.00092	0.003289	0.0014	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1322	75	0.000942	0.003222	0.00143	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7467	1322	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4217	5558	0.03868	0.1062	0.02764	264.496369	264.496369	264.496369	0	0	1	-360	360;
	5279	8321	0	0.000222	0	127.13634	127.13634	127.13634	0	0	1	-360	360;
	2177	8321	0.000689	0.00312	0.00106	103.923048	103.923048	103.923048	0	0	1	-360	360;
	289	8321	0.0008	0.002991	0.00123	127.690509	127.690509	127.690509	0	0	1	-360	360;
	906	7612	0.000511	0.001191	0.00037	113.412782	113.412782	113.412782	0	0	1	-360	360;
	8023	606	0	0.00028	0	149.142565	149.142565	149.142565	0	0	1	-360	360;
	2285	2519	0.01216	0.048609	0.0173	108.02619	108.02619	108.02619	0	0	1	-360	360;
	3512	2439	0.000769	0.00348	0.00118	143.345477	143.345477	143.345477	0	0	1	-360	360;
	3512	5414	0.000871	0.003409	0.00134	120.196618	120.196618	120.196618	0	0	1	-360	360;
	5715	7482	0.007071	0.03316	0.01005	103.923048	103.923048	103.923048	0	0	1	-360	360;
	486	5113	0.009609	0.038418	0.01368	121.668169	121.668169	121.668169	0	0	1	-360	360;
	687	486	0	0.00028	0	144.419003	144.419003	144.419003	0	0	1	-360	360;
	7867	5113	0.00724	0.014951	0.25617	121.668169	121.668169	121.668169	0	0	1	-360	360;
	2950	3929	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7247	3929	0.016071	0.067418	0.02446	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7247	3929	0.015738	0.068751	0.02396	112.292751	112.292751	112.292751	0	0	1	-360	360;
	3929	3855	0.017431	0.037622	0.01244	223.982229	223.982229	223.982229	0	0	1	-360	360;
	6138	5815	0.001031	0.003378	0.1074	154.905358	154.905358	154.905358	0	0	1	-360	360;
	5815	7266	0.000169	0.00068	0.00028	180.218674	180.218674	180.218674	0	0	1	-360	360;
	1539	6138	0.006249	0.023262	0.00952	154.432768	154.432768	154.432768	0	0	1	-360	360;
	8240	5484	8.9e-05	0.000471	0.00013	134.187519	134.187519	134.187519	0	0	1	-360	360;
	2719	2388	0	0.000222	0	117.47981	117.47981	117.47981	0	0	1	-360	360;
	3868	2388	0.01564	0.065089	0.02226	114.830552	114.830552	114.830552	0	0	1	-360	360;
	2388	1551	0.018938	0.085391	0.02696	135.731036	135.731036	135.731036	0	0	1	-360	360;
	2388	1342	0.016511	0.044662	0.0118	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7804	468	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	468	7867	0.002689	0.012938	0.00382	152.506034	152.506034	152.506034	0	0	1	-360	360;
	4615	2604	0	0.000222	0	121.16726	121.16726	121.16726	0	0	1	-360	360;
	744	4615	0.007649	0.019631	0.45863	128.375103	128.375103	128.375103	0	0	1	-360	360;
	5677	327	0.008849	0.02096	0.53038	108.171423	108.171423	108.171423	0	0	1	-360	360;
	2604	5677	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	327	4398	0	0.00028	0	141.008508	141.008508	141.008508	0	0	1	-360	360;
	9	2159	0.00232	0.00928	0.00331	106.685843	106.685843	106.685843	0	0	1	-360	360;
	2445	1383	0	0.000222	0	130.409656	130.409656	130.409656	0	0	1	-360	360;
	6756	5316	0	0.000222	0	127.529948	127.529948	127.529948	0	0	1	-360	360;
	3082	2016	0	0.000222	0	108.992934	108.992934	108.992934	0	0	1	-360	360;
	5333	3082	0.0088	0.034102	0.0134	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6689	4472	0	0.000222	0	112.208573	112.208573	112.208573	0	0	1	-360	360;
	6366	2953	0	0.00028	0	134.717786	134.717786	134.717786	0	0	1	-360	360;
	7395	3718	0	0.00028	0	137.527605	137.527605	137.527605	0	0	1	-360	360;
	4916	1466	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7118	6978	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6989	4787	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6172	8020	0	0.00028	0	129.923556	129.923556	129.923556	0	0	1	-360	360;
	8020	5695	0.002542	0.00392	0.11078	284.003505	284.003505	284.003505	0	0	1	-360	360;
	6802	4156	0.00296	0.013462	0.00451	128.044627	128.044627	128.044627	0	0	1	-360	360;
	7204	6802	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6802	92	0.002489	0.003991	0.10852	268.990262	268.990262	268.990262	0	0	1	-360	360;
	696	4609	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7237	4609	0.000289	0.000742	0.03021	127.090873	127.090873	127.090873	0	0	1	-360	360;
	6846	3578	0	0.00028	0	120.61361	120.61361	120.61361	0	0	1	-360	360;
	2338	3578	0.000302	0.000729	0.03077	136.212459	136.212459	136.212459	0	0	1	-360	360;
	3578	6326	0.000662	0.002111	0.08262	129.438755	129.438755	129.438755	0	0	1	-360	360;
	4011	1097	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5924	1466	0.000822	0.002871	0.07999	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1466	7877	0.006391	0.028809	0.00909	106.503198	106.503198	106.503198	0	0	1	-360	360;
	2558	6978	0.000831	0.002809	0.08156	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6181	5924	0.014791	0.005969	0.54372	114.125954	114.125954	114.125954	0	0	1	-360	360;
	3933	2558	0.01508	0.005849	0.55418	106.359265	106.359265	106.359265	0	0	1	-360	360;
	3884	4787	0.011302	0.050578	0.01719	123.883289	123.883289	123.883289	0	0	1	-360	360;
	906	4787	0.01704	0.033938	0.01218	259.046645	259.046645	259.046645	0	0	1	-360	360;
	7653	4900	0.001262	0.00532	0.00193	146.333525	146.333525	146.333525	0	0	1	-360	360;
	7653	2689	0.00124	0.005431	0.00189	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5723	322	0	0.00028	0	107.024112	107.024112	107.024112	0	0	1	-360	360;
	322	2404	0.001538	0.004609	0.18216	136.534361	136.534361	136.534361	0	0	1	-360	360;
	1181	7927	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2449	5457	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7522	5457	0.00244	0.004071	0.10626	299.492196	299.492196	299.492196	0	0	1	-360	360;
	2013	4668	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5056	4668	0.002649	0.003631	0.11572	274.251106	274.251106	274.251106	0	0	1	-360	360;
	4524	1356	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2289	1247	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7507	5313	0.000849	0.001769	0.00062	107.980984	107.980984	107.980984	0	0	1	-360	360;
	5313	7139	0.000951	0.003991	0.00144	149.026691	149.026691	149.026691	0	0	1	-360	360;
	4263	4398	0	0.000222	0	116.552817	116.552817	116.552817	0	0	1	-360	360;
	4263	7178	0.0034	0.00872	0.20383	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6221	7247	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1097	4301	0.001978	0.007671	0.00302	126.877571	126.877571	126.877571	0	0	1	-360	360;
	1097	819	0.002018	0.00752	0.00308	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6332	3868	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3868	872	0.018351	0.070502	0.02612	123.216103	123.216103	123.216103	0	0	1	-360	360;
	7606	4905	0.0028	0.014209	0.00397	160.337675	160.337675	160.337675	0	0	1	-360	360;
	842	8180	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6690	1456	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6290	1456	0.02624	0.096858	0.03736	126.419011	126.419011	126.419011	0	0	1	-360	360;
	406	5386	5.8e-05	0.00024	0.00011	118.302621	118.302621	118.302621	0	0	1	-360	360;
	406	6446	5.8e-05	0.000231	0.00011	171.76947	171.76947	171.76947	0	0	1	-360	360;
	5920	3715	0.0012	0.005809	0.00172	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5920	6842	8.9e-05	0.00028	0.00013	118.517482	118.517482	118.517482	0	0	1	-360	360;
	5920	5789	8e-05	0.00032	0.00012	121.12673	121.12673	121.12673	0	0	1	-360	360;
	6842	2997	0.006889	0.032631	0.01049	116.546062	116.546062	116.546062	0	0	1	-360	360;
	5789	2997	0.006738	0.030209	0.01026	121.269104	121.269104	121.269104	0	0	1	-360	360;
	4770	2658	0.001218	0.004031	0.00186	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2658	906	0.005751	0.012929	0.00411	291.679261	291.679261	291.679261	0	0	1	-360	360;
	142	3441	0.000462	0.001791	0.0007	131.458759	131.458759	131.458759	0	0	1	-360	360;
	142	5658	0.00352	0.007302	0.00252	103.923048	103.923048	103.923048	0	0	1	-360	360;
	64	4475	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4079	1972	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1204	5120	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5120	899	0.001289	0.005022	0.16146	239.000408	239.000408	239.000408	0	0	1	-360	360;
	3245	7315	0	0.00028	0	149.490967	149.490967	149.490967	0	0	1	-360	360;
	7600	7877	0.00036	0.0016	0.00054	117.419535	117.419535	117.419535	0	0	1	-360	360;
	7600	4621	0.000409	0.001569	0.00061	139.373798	139.373798	139.373798	0	0	1	-360	360;
	7463	7600	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1111	7600	0.001658	0.004511	0.2079	105.829777	105.829777	105.829777	0	0	1	-360	360;
	5400	2359	0.00176	0.007911	0.0025	205.481848	205.481848	205.481848	0	0	1	-360	360;
	899	5400	0.00172	0.008058	0.00245	186.92275	186.92275	186.92275	0	0	1	-360	360;
	5819	5400	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1317	7877	0.0022	0.00852	0.00335	110.010601	110.010601	110.010601	0	0	1	-360	360;
	6852	4621	0.000111	0.000489	0.00014	143.333007	143.333007	143.333007	0	0	1	-360	360;
	6852	7890	0.00396	0.013058	0.00603	129.342886	129.342886	129.342886	0	0	1	-360	360;
	2359	7980	0.001138	0.002529	0.0992	223.367264	223.367264	223.367264	0	0	1	-360	360;
	7834	1513	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8160	7834	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7834	2424	0.001329	0.005978	0.00203	120.366013	120.366013	120.366013	0	0	1	-360	360;
	4887	8008	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4975	5926	0	0.000222	0	117.966949	117.966949	117.966949	0	0	1	-360	360;
	5926	7178	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4975	2994	0.010191	0.028329	0.61098	145.970314	145.970314	145.970314	0	0	1	-360	360;
	1675	6401	0.002391	0.00788	0.00363	107.798339	107.798339	107.798339	0	0	1	-360	360;
	961	1675	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5321	1675	0.002249	0.008351	0.00341	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5548	2765	0.000302	0.000969	0.00045	105.92019	105.92019	105.92019	0	0	1	-360	360;
	2219	5796	0.000271	0.001102	0.0004	108.412524	108.412524	108.412524	0	0	1	-360	360;
	3469	5548	0.000391	0.001738	0.00059	105.92019	105.92019	105.92019	0	0	1	-360	360;
	2219	2912	0.002289	0.008178	0.00348	108.381607	108.381607	108.381607	0	0	1	-360	360;
	6369	5321	0	0.000222	0	125.783522	125.783522	125.783522	0	0	1	-360	360;
	5321	4347	0.00112	0.00504	0.00171	126.686353	126.686353	126.686353	0	0	1	-360	360;
	5321	4347	0.001271	0.004942	0.00194	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3293	4347	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4347	6553	0.000418	0.00192	0.00064	112.184931	112.184931	112.184931	0	0	1	-360	360;
	4347	4075	0.000409	0.001951	0.00063	121.388616	121.388616	121.388616	0	0	1	-360	360;
	4347	84	0.03228	0.106458	0.04913	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4347	84	0.02876	0.120649	0.04377	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4075	84	0.05888	0.117289	0.04206	241.664735	241.664735	241.664735	0	0	1	-360	360;
	6290	7313	0.00324	0.01584	0.00462	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1166	7313	0	0.000222	0	111.705066	111.705066	111.705066	0	0	1	-360	360;
	3399	6995	0.00176	0.004449	0.00125	278.107691	278.107691	278.107691	0	0	1	-360	360;
	7462	916	0.009929	0.046969	0.01511	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7462	1966	0.009711	0.043489	0.01478	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5794	1966	0.005031	0.02252	0.00765	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1575	2404	0	0.000222	0	131.346003	131.346003	131.346003	0	0	1	-360	360;
	7741	5830	0	0.000222	0	123.770012	123.770012	123.770012	0	0	1	-360	360;
	6696	5830	0.000462	0.00108	0.05406	172.491216	172.491216	172.491216	0	0	1	-360	360;
	7353	2404	0.000471	0.001049	0.0551	282.530136	282.530136	282.530136	0	0	1	-360	360;
	1111	1317	0.00104	0.003729	0.1296	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7594	4750	0.002329	0.008	0.00354	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7996	4750	0.002378	0.007831	0.00361	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1834	4750	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	542	5794	0.006711	0.026018	0.01022	117.438501	117.438501	117.438501	0	0	1	-360	360;
	5794	1579	0.009298	0.04232	0.01417	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7902	5036	0.004751	0.01632	0.00724	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2516	1730	0	0.000222	0	142.965379	142.965379	142.965379	0	0	1	-360	360;
	1730	2807	0.000338	0.001151	0.00051	150.17504	150.17504	150.17504	0	0	1	-360	360;
	1730	5650	0.000338	0.00112	0.00052	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7879	6942	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3845	424	0.00484	0.023329	0.00689	240.237612	240.237612	240.237612	0	0	1	-360	360;
	4683	6181	0.004222	0.017751	0.00644	116.295087	116.295087	116.295087	0	0	1	-360	360;
	1047	3933	0.004138	0.018102	0.00631	118.648425	118.648425	118.648425	0	0	1	-360	360;
	4069	3011	0	0.000222	0	118.242086	118.242086	118.242086	0	0	1	-360	360;
	4069	1114	0.000622	0.002418	0.00095	118.104647	118.104647	118.104647	0	0	1	-360	360;
	3011	573	0.000631	0.002369	0.00097	130.665307	130.665307	130.665307	0	0	1	-360	360;
	7042	8037	0.00012	0.00032	9e-05	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7301	4785	0	0.00028	0	111.056586	111.056586	111.056586	0	0	1	-360	360;
	1002	5397	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2053	92	0	0.00028	0	114.719614	114.719614	114.719614	0	0	1	-360	360;
	7547	5695	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5348	568	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3096	6738	0	0.00028	0	143.524485	143.524485	143.524485	0	0	1	-360	360;
	2065	7790	0.001409	0.00464	0.00215	150.93186	150.93186	150.93186	0	0	1	-360	360;
	1585	2065	0.003098	0.014911	0.00441	155.047213	155.047213	155.047213	0	0	1	-360	360;
	7790	2343	0.004662	0.015369	0.0071	110.735464	110.735464	110.735464	0	0	1	-360	360;
	106	7450	9.8e-05	0.000369	0.00014	107.943312	107.943312	107.943312	0	0	1	-360	360;
	7790	106	0.001031	0.003658	0.00156	107.883556	107.883556	107.883556	0	0	1	-360	360;
	2997	5228	0.001049	0.003591	0.00159	182.667621	182.667621	182.667621	0	0	1	-360	360;
	5576	3357	0.003209	0.012338	0.00457	114.421355	114.421355	114.421355	0	0	1	-360	360;
	5775	5576	0.003311	0.012689	0.00471	105.856537	105.856537	105.856537	0	0	1	-360	360;
	7244	2285	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	84	335	0.003871	0.010458	0.00276	238.123817	238.123817	238.123817	0	0	1	-360	360;
	7668	84	0.000218	0.000849	0.00034	168.240504	168.240504	168.240504	0	0	1	-360	360;
	5474	84	0.000231	0.00084	0.00034	160.35898	160.35898	160.35898	0	0	1	-360	360;
	8314	84	0	0.00028	0	132.607109	132.607109	132.607109	0	0	1	-360	360;
	335	2574	0.000231	0.000822	0.00035	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2499	2574	0.00024	0.0008	0.00036	114.101791	114.101791	114.101791	0	0	1	-360	360;
	2140	2574	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5128	4386	0	0.00028	0	139.83132	139.83132	139.83132	0	0	1	-360	360;
	2306	5324	0.000151	0.00036	0.01768	184.20828	184.20828	184.20828	0	0	1	-360	360;
	2306	7522	8e-05	0.00028	0.0099	153.982001	153.982001	153.982001	0	0	1	-360	360;
	977	5056	8e-05	0.00028	0.0099	118.055284	118.055284	118.055284	0	0	1	-360	360;
	1263	2306	0	0.000222	0	115.943568	115.943568	115.943568	0	0	1	-360	360;
	2306	977	0	0.00028	0	121.036576	121.036576	121.036576	0	0	1	-360	360;
	2029	977	0	0.000222	0	143.86925	143.86925	143.86925	0	0	1	-360	360;
	1741	7893	0.001711	0.003058	0.06059	107.020475	107.020475	107.020475	0	0	1	-360	360;
	7139	1741	0.003369	0.008889	0.00241	260.303594	260.303594	260.303594	0	0	1	-360	360;
	1598	7893	0.002009	0.005311	0.00144	265.58886	265.58886	265.58886	0	0	1	-360	360;
	2682	1598	0	0.000222	0	134.623736	134.623736	134.623736	0	0	1	-360	360;
	7338	774	0.002551	0.011578	0.00388	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3739	774	0.002489	0.0118	0.0038	140.798843	140.798843	140.798843	0	0	1	-360	360;
	6573	7606	0	0.00028	0	130.745327	130.745327	130.745327	0	0	1	-360	360;
	7606	3584	0.010009	0.04892	0.01425	130.126726	130.126726	130.126726	0	0	1	-360	360;
	4494	3348	0.000182	0.000671	0.00028	171.817275	171.817275	171.817275	0	0	1	-360	360;
	7004	6715	0.006169	0.02036	0.00939	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6715	1342	0.007369	0.0322	0.01122	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5939	7004	0.005498	0.023071	0.00837	146.804296	146.804296	146.804296	0	0	1	-360	360;
	7266	5939	0.004809	0.021671	0.00684	173.493814	173.493814	173.493814	0	0	1	-360	360;
	3851	7004	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7004	872	0.016169	0.06784	0.02461	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5532	3299	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8029	26	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4997	26	0.01124	0.039822	0.01599	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6297	156	0.001169	0.00532	0.00178	110.697792	110.697792	110.697792	0	0	1	-360	360;
	424	156	0.001138	0.005418	0.00174	213.79725	213.79725	213.79725	0	0	1	-360	360;
	7760	3163	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6024	4997	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2084	7330	0	0.000222	0	151.847942	151.847942	151.847942	0	0	1	-360	360;
	4366	5403	0.00212	0.008871	0.00321	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1957	4366	0.002289	0.008902	0.0035	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5403	6649	0.00024	0.000782	0.00036	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2888	6649	0.000209	0.000889	0.00032	115.066197	115.066197	115.066197	0	0	1	-360	360;
	5140	7330	0	0.000222	0	114.410443	114.410443	114.410443	0	0	1	-360	360;
	89	7330	0	0.000222	0	108.664018	108.664018	108.664018	0	0	1	-360	360;
	1887	6649	0	0.000222	0	104.38031	104.38031	104.38031	0	0	1	-360	360;
	1594	4135	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4135	4897	0.001209	0.005529	0.00185	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4135	3855	0.001191	0.00564	0.00181	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2587	4897	0.014982	0.0428	0.0107	198.291412	198.291412	198.291412	0	0	1	-360	360;
	5724	7104	0.012849	0.022929	0.45441	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7104	1620	0.000431	0.000929	0.00031	254.876992	254.876992	254.876992	0	0	1	-360	360;
	7885	1620	0.000849	0.003831	0.00121	131.990585	131.990585	131.990585	0	0	1	-360	360;
	6865	1620	0	0.000222	0	119.682199	119.682199	119.682199	0	0	1	-360	360;
	3595	2827	0.000231	0.000769	0.00036	138.853144	138.853144	138.853144	0	0	1	-360	360;
	3595	3884	0.0002	0.000871	0.00032	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6709	3595	0	0.00028	0	128.372245	128.372245	128.372245	0	0	1	-360	360;
	3707	542	0.004049	0.01844	0.00618	127.430442	127.430442	127.430442	0	0	1	-360	360;
	3707	3455	0.003969	0.018791	0.00604	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4486	3707	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5558	352	0.002271	0.004258	0.00177	266.373479	266.373479	266.373479	0	0	1	-360	360;
	2587	3334	0.012449	0.0248	0.00889	130.02644	130.02644	130.02644	0	0	1	-360	360;
	3334	352	0.002311	0.004169	0.0018	248.704223	248.704223	248.704223	0	0	1	-360	360;
	3660	2587	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6521	352	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4321	3268	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7890	3268	0	0.000222	0	135.926671	135.926671	135.926671	0	0	1	-360	360;
	5663	3268	0	0.000222	0	125.616205	125.616205	125.616205	0	0	1	-360	360;
	2125	7528	0.002071	0.009049	0.00315	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2125	211	0.002031	0.009222	0.00308	125.279235	125.279235	125.279235	0	0	1	-360	360;
	8294	2125	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8219	872	0	0.000222	0	123.298202	123.298202	123.298202	0	0	1	-360	360;
	1342	872	0.015511	0.06772	0.0236	110.960457	110.960457	110.960457	0	0	1	-360	360;
	7336	8210	4e-05	0.00028	7e-05	132.30833	132.30833	132.30833	0	0	1	-360	360;
	7336	5333	4e-05	0.00028	7e-05	108.700651	108.700651	108.700651	0	0	1	-360	360;
	23	7336	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6773	1530	0.0052	0.013898	0.31199	132.331453	132.331453	132.331453	0	0	1	-360	360;
	8260	3802	0.009071	0.032151	0.01291	128.613346	128.613346	128.613346	0	0	1	-360	360;
	6003	8260	0	0.00028	0	124.865102	124.865102	124.865102	0	0	1	-360	360;
	492	7331	0	0.000222	0	155.264672	155.264672	155.264672	0	0	1	-360	360;
	1530	6308	0	0.000222	0	125.760399	125.760399	125.760399	0	0	1	-360	360;
	7331	744	0	0.000222	0	137.600611	137.600611	137.600611	0	0	1	-360	360;
	7331	2994	0	0.000222	0	114.38784	114.38784	114.38784	0	0	1	-360	360;
	7331	6308	0	0.000222	0	111.687399	111.687399	111.687399	0	0	1	-360	360;
	4867	6835	0.042889	0.088942	0.03064	129.301836	129.301836	129.301836	0	0	1	-360	360;
	691	5469	0.000609	0.001618	0.00044	245.958316	245.958316	245.958316	0	0	1	-360	360;
	399	8008	0.011289	0.057409	0.01607	106.854198	106.854198	106.854198	0	0	1	-360	360;
	3672	781	0	0.00028	0	109.960198	109.960198	109.960198	0	0	1	-360	360;
	4698	3436	8.9e-05	0.000449	0.00013	144.024355	144.024355	144.024355	0	0	1	-360	360;
	7460	4344	0.000169	0.000662	0.00024	150.079951	150.079951	150.079951	0	0	1	-360	360;
	2102	4597	0	0.00028	0	113.084904	113.084904	113.084904	0	0	1	-360	360;
	1193	5260	9.8e-05	0.000511	0.00015	161.273762	161.273762	161.273762	0	0	1	-360	360;
	7828	5260	0.006338	0.022662	0.00965	181.344421	181.344421	181.344421	0	0	1	-360	360;
	824	6390	0.000809	0.001729	0.00057	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7756	1114	0.00204	0.007898	0.0031	118.104647	118.104647	118.104647	0	0	1	-360	360;
	4560	8159	0.00188	0.00516	0.00134	241.218646	241.218646	241.218646	0	0	1	-360	360;
	7828	7237	0.000831	0.003511	0.00127	132.707135	132.707135	132.707135	0	0	1	-360	360;
	1201	149	0.00064	0.002871	0.0009	130.098666	130.098666	130.098666	0	0	1	-360	360;
	6326	3348	0.000182	0.00064	0.00028	132.378998	132.378998	132.378998	0	0	1	-360	360;
	6326	1551	0.000191	0.000631	0.00029	157.491483	157.491483	157.491483	0	0	1	-360	360;
	3211	6326	0	0.00028	0	114.410963	114.410963	114.410963	0	0	1	-360	360;
	1927	282	0.00043	0.00337	0.09639	1183.90924	1183.90924	1183.90924	0	0	1	-360	360;
	1927	6461	0.00038	0.00383	0.08588	1495.38207	1495.38207	1495.38207	0	0	1	-360	360;
	4531	4889	0.000431	0.00112	0.00031	274.544169	274.544169	274.544169	0	0	1	-360	360;
	1674	4531	0.006458	0.021302	0.00983	136.612823	136.612823	136.612823	0	0	1	-360	360;
	7179	2517	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6707	8236	0.000249	0.001138	0.00039	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7750	6707	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6707	1579	0.002449	0.011169	0.00374	108.108289	108.108289	108.108289	0	0	1	-360	360;
	3882	1579	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4776	3726	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7933	7945	0.000271	0.001089	0.00038	123.714154	123.714154	123.714154	0	0	1	-360	360;
	6078	3584	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	280	1342	0	0.000222	0	135.089051	135.089051	135.089051	0	0	1	-360	360;
	1260	800	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3046	1111	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6235	4027	0	0.000222	0	142.751557	142.751557	142.751557	0	0	1	-360	360;
	710	1766	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5397	3146	0.006729	0.031862	0.01025	121.473573	121.473573	121.473573	0	0	1	-360	360;
	5397	5048	0.006591	0.029498	0.01003	186.958344	186.958344	186.958344	0	0	1	-360	360;
	6738	7980	0.000258	0.000711	0.03267	191.148781	191.148781	191.148781	0	0	1	-360	360;
	4832	5877	0.020311	0.045649	0.01451	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7460	6694	0.004102	0.014089	0.00624	116.487085	116.487085	116.487085	0	0	1	-360	360;
	6694	1574	0	0.000222	0	125.219219	125.219219	125.219219	0	0	1	-360	360;
	5431	2704	0.038777	0.100653	0.00807	79.267981	79.267981	79.267981	0	0	1	-360	360;
	2976	5431	0.026711	0.08357	0.00954	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7061	2561	0.062719	0.144537	0.01303	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2673	7061	0.045033	0.112364	0.0094	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4086	6946	0.011659	0.070599	0.09889	392.012877	392.012877	392.012877	0	0	1	-360	360;
	4086	747	0.00968	0.04149	0.07777	375.523268	375.523268	375.523268	0	0	1	-360	360;
	4086	747	0.00862	0.047021	0.06929	387.835032	387.835032	387.835032	0	0	1	-360	360;
	1220	2561	0.077471	0.185529	0.01613	70.283175	70.283175	70.283175	0	0	1	-360	360;
	3390	2561	0.045157	0.126248	0.01486	60.925701	60.925701	60.925701	0	0	1	-360	360;
	4690	2561	0.01119	0.043099	0.00368	65.276353	65.276353	65.276353	0	0	1	-360	360;
	4690	2561	0.01095	0.043909	0.00361	64.872058	64.872058	64.872058	0	0	1	-360	360;
	2544	1220	0.03681	0.090496	0.00765	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2883	5652	0.056174	0.129223	0.01168	77.293755	77.293755	77.293755	0	0	1	-360	360;
	3491	2883	0.029256	0.116322	0.01003	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4998	7768	0.061347	0.097917	0.16601	80.330732	80.330732	80.330732	0	0	1	-360	360;
	7768	8082	0.088719	0.24762	0.05075	66.947262	66.947262	66.947262	0	0	1	-360	360;
	7332	244	0	0.000413	0	69.441433	69.441433	69.441433	0	0	1	-360	360;
	7332	5151	0.004612	0.01738	0.0016	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7332	5220	0.008364	0.02695	0.00296	78.656965	78.656965	78.656965	0	0	1	-360	360;
	7332	5220	0.014992	0.043636	0.00494	61.787067	61.787067	61.787067	0	0	1	-360	360;
	7543	244	0.003033	0.012851	0.00101	106.186007	106.186007	106.186007	0	0	1	-360	360;
	7543	244	0.002967	0.011901	0.00099	87.367223	87.367223	87.367223	0	0	1	-360	360;
	244	1392	0.015256	0.042686	0.00503	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1764	244	0.004397	0.025653	0.00627	126.258069	126.258069	126.258069	0	0	1	-360	360;
	3063	3592	0.0027	0.012961	0.02101	386.54022	386.54022	386.54022	0	0	1	-360	360;
	747	3592	0.038169	0.16812	0.30228	363.751073	363.751073	363.751073	0	0	1	-360	360;
	3592	4992	0.015479	0.073409	0.21497	517.58753	517.58753	517.58753	0	0	1	-360	360;
	5983	3148	0.027868	0.087843	0.00957	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4577	5983	0.013934	0.040479	0.00475	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1629	3148	0.019231	0.076471	0.00653	60.653631	60.653631	60.653631	0	0	1	-360	360;
	2087	5251	0.013421	0.037256	0.00462	71.546931	71.546931	71.546931	0	0	1	-360	360;
	5033	2087	0.022438	0.046851	0.00459	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7506	445	0.014562	0.045314	0.02465	100.683438	100.683438	100.683438	0	0	1	-360	360;
	445	2544	0.000826	0.002207	0.02371	61.45098	61.45098	61.45098	0	0	1	-360	360;
	1740	5194	0.040628	0.113562	0.01388	102.55116	102.55116	102.55116	0	0	1	-360	360;
	4995	1740	0.008058	0.024438	0.00276	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1740	5325	0.022471	0.041083	0.14109	66.861145	66.861145	66.861145	0	0	1	-360	360;
	3990	7779	0.047322	0.096669	0.00988	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7779	2976	0.01914	0.07614	0.00653	57.157677	57.157677	57.157677	0	0	1	-360	360;
	668	5499	0.007802	0.01924	0.0017	63.639929	63.639929	63.639929	0	0	1	-360	360;
	668	5499	0.005397	0.015554	0.00173	76.619103	76.619103	76.619103	0	0	1	-360	360;
	5499	6650	0.024719	0.040702	0.07219	68.373156	68.373156	68.373156	0	0	1	-360	360;
	5499	6650	0.02419	0.041471	0.07066	65.77896	65.77896	65.77896	0	0	1	-360	360;
	3867	2420	0.039603	0.109529	0.01353	92.919901	92.919901	92.919901	0	0	1	-360	360;
	3142	3867	0.013826	0.039826	0.00475	76.348367	76.348367	76.348367	0	0	1	-360	360;
	6219	3867	0.015174	0.04981	0.00519	77.193157	77.193157	77.193157	0	0	1	-360	360;
	7307	4177	0.003669	0.009008	0.00075	69.015037	69.015037	69.015037	0	0	1	-360	360;
	4177	1829	0.03276	0.106603	0.00954	112.964527	112.964527	112.964527	0	0	1	-360	360;
	4177	1829	0.03019	0.107612	0.01029	113.711959	113.711959	113.711959	0	0	1	-360	360;
	3344	4177	0.019033	0.051653	0.00562	79.409351	79.409351	79.409351	0	0	1	-360	360;
	166	4177	0.037603	0.124017	0.011	80.738838	80.738838	80.738838	0	0	1	-360	360;
	4177	8140	0.034868	0.097471	0.01188	58.685311	58.685311	58.685311	0	0	1	-360	360;
	4177	5837	0.043099	0.130603	0.01419	74.815017	74.815017	74.815017	0	0	1	-360	360;
	3990	7307	0.06105	0.110339	0.01265	80.307869	80.307869	80.307869	0	0	1	-360	360;
	2552	5692	0.041446	0.108843	0.01039	59.70729	59.70729	59.70729	0	0	1	-360	360;
	5692	213	0.063471	0.123207	0.01485	66.430748	66.430748	66.430748	0	0	1	-360	360;
	2812	4995	0.02343	0.073603	0.00774	74.149892	74.149892	74.149892	0	0	1	-360	360;
	2812	1726	0.022537	0.091587	0.00764	73.190595	73.190595	73.190595	0	0	1	-360	360;
	8378	4195	0.036744	0.106934	0.01255	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8378	748	0.006207	0.023959	0.00207	73.936884	73.936884	73.936884	0	0	1	-360	360;
	3142	3102	0.003157	0.011008	0.00117	78.454436	78.454436	78.454436	0	0	1	-360	360;
	3142	3102	0.003223	0.010793	0.0012	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6219	3102	0.005	0.035653	0.00715	87.52898	87.52898	87.52898	0	0	1	-360	360;
	6219	3102	0.004901	0.036347	0.00701	86.042499	86.042499	86.042499	0	0	1	-360	360;
	7308	1434	0.007917	0.029603	0.0027	65.15556	65.15556	65.15556	0	0	1	-360	360;
	656	1434	0.004876	0.01476	0.0017	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4963	4195	0.007182	0.023579	0.00245	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4963	4169	0.007322	0.023099	0.0025	80.588704	80.588704	80.588704	0	0	1	-360	360;
	5325	7747	0.010942	0.040562	0.00374	77.31738	77.31738	77.31738	0	0	1	-360	360;
	7747	823	0.030603	0.103702	0.13068	62.797996	62.797996	62.797996	0	0	1	-360	360;
	5251	794	0.01176	0.041421	0.00402	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5251	1292	0.011868	0.03581	0.00413	57.157677	57.157677	57.157677	0	0	1	-360	360;
	708	5251	0.01757	0.064579	0.00605	57.287806	57.287806	57.287806	0	0	1	-360	360;
	6451	6647	0.017099	0.064281	0.00585	67.474828	67.474828	67.474828	0	0	1	-360	360;
	721	3491	0.015207	0.057397	0.00522	87.073624	87.073624	87.073624	0	0	1	-360	360;
	1292	721	0.009198	0.036512	0.00313	79.11918	79.11918	79.11918	0	0	1	-360	360;
	721	6219	0.024281	0.239446	0.03888	71.315824	71.315824	71.315824	0	0	1	-360	360;
	8304	1195	0.052471	0.157347	0.01802	57.157677	57.157677	57.157677	0	0	1	-360	360;
	370	1195	0.009314	0.035909	0.00307	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7308	1195	0.039314	0.122777	0.01352	65.290071	65.290071	65.290071	0	0	1	-360	360;
	7543	408	0.001223	0.003711	0.03788	104.54844	104.54844	104.54844	0	0	1	-360	360;
	747	6691	0.00518	0.023091	0.04158	404.578039	404.578039	404.578039	0	0	1	-360	360;
	747	6691	0.005281	0.022591	0.04235	383.755498	383.755498	383.755498	0	0	1	-360	360;
	2390	7299	0.018198	0.037149	0.00374	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7183	5371	0.012893	0.040033	0.00447	82.619326	82.619326	82.619326	0	0	1	-360	360;
	3493	5371	0.007132	0.020331	0.00248	73.635092	73.635092	73.635092	0	0	1	-360	360;
	2919	1301	0.00479	0.05497	1.09267	1213.90576	1213.90576	1213.90576	0	0	1	-360	360;
	3990	1829	0.047322	0.139868	0.01851	87.529361	87.529361	87.529361	0	0	1	-360	360;
	3153	1829	0.005471	0.017471	0.00163	103.648778	103.648778	103.648778	0	0	1	-360	360;
	3153	1829	0.013826	0.040207	0.00454	70.506852	70.506852	70.506852	0	0	1	-360	360;
	2704	1829	0.010678	0.03814	0.00363	72.958345	72.958345	72.958345	0	0	1	-360	360;
	2704	1829	0.010603	0.034793	0.00357	74.931618	74.931618	74.931618	0	0	1	-360	360;
	4223	1829	0.014579	0.043331	0.00518	71.034608	71.034608	71.034608	0	0	1	-360	360;
	1829	8140	0.021413	0.079248	0.0073	137.635114	137.635114	137.635114	0	0	1	-360	360;
	1829	6415	0.031198	0.098298	0.0103	99.798828	99.798828	99.798828	0	0	1	-360	360;
	4755	3063	0.009651	0.045031	0.07653	385.332669	385.332669	385.332669	0	0	1	-360	360;
	5194	5009	0.050207	0.124066	0.01123	72.625497	72.625497	72.625497	0	0	1	-360	360;
	8351	4129	0.034314	0.070521	0.0057	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6022	5033	0.007463	0.022612	0.00255	64.1553	64.1553	64.1553	0	0	1	-360	360;
	6022	5033	0.007603	0.022132	0.0026	73.775128	73.775128	73.775128	0	0	1	-360	360;
	4995	3383	0.01081	0.033603	0.00374	65.332177	65.332177	65.332177	0	0	1	-360	360;
	8304	4995	0.011223	0.036446	0.00388	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4995	7183	0.014182	0.045669	0.0049	70.206584	70.206584	70.206584	0	0	1	-360	360;
	7183	4995	0.014463	0.044744	0.00499	71.359264	71.359264	71.359264	0	0	1	-360	360;
	166	3344	0.02124	0.069099	0.00621	71.273336	71.273336	71.273336	0	0	1	-360	360;
	8030	1483	0.008041	0.030504	0.00265	70.960493	70.960493	70.960493	0	0	1	-360	360;
	8030	2113	0.009107	0.029893	0.003	73.682533	73.682533	73.682533	0	0	1	-360	360;
	7846	6954	0.024802	0.074537	0.00859	57.157677	57.157677	57.157677	0	0	1	-360	360;
	116	7846	0.038711	0.136876	0.01323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	656	1726	0.069413	0.194223	0.02376	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8304	3383	0.01276	0.035281	0.0044	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3827	6799	0.008744	0.025802	0.00286	98.627286	98.627286	98.627286	0	0	1	-360	360;
	6799	2113	0.008909	0.025248	0.00292	89.589323	89.589323	89.589323	0	0	1	-360	360;
	4690	1483	0.009289	0.029281	0.00306	69.456294	69.456294	69.456294	0	0	1	-360	360;
	1523	3390	0.021421	0.081298	0.00729	65.889464	65.889464	65.889464	0	0	1	-360	360;
	1523	6451	0.044479	0.139058	0.0152	81.273072	81.273072	81.273072	0	0	1	-360	360;
	8213	6946	0.008479	0.049461	0.06704	404.373796	404.373796	404.373796	0	0	1	-360	360;
	747	6946	0.00394	0.02236	0.0311	456.983627	456.983627	456.983627	0	0	1	-360	360;
	668	5131	0.016174	0.044463	0.00539	87.938419	87.938419	87.938419	0	0	1	-360	360;
	668	5131	0.014413	0.05038	0.0048	73.740071	73.740071	73.740071	0	0	1	-360	360;
	700	668	0.020198	0.064678	0.00663	61.730481	61.730481	61.730481	0	0	1	-360	360;
	700	668	0.020587	0.063364	0.00676	62.254808	62.254808	62.254808	0	0	1	-360	360;
	700	668	0.031711	0.059223	0.00662	69.794667	69.794667	69.794667	0	0	1	-360	360;
	8351	3273	0.056678	0.178372	0.01955	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3273	2673	0.019496	0.078083	0.00662	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5131	6509	0.045198	0.111479	0.00984	65.413722	65.413722	65.413722	0	0	1	-360	360;
	4883	5033	0.015041	0.057983	0.00517	66.11276	66.11276	66.11276	0	0	1	-360	360;
	949	4883	0.03519	0.105653	0.01208	60.703358	60.703358	60.703358	0	0	1	-360	360;
	116	5652	0.044	0.099339	0.00902	81.351378	81.351378	81.351378	0	0	1	-360	360;
	1392	7543	0.013636	0.039512	0.00451	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1392	7543	0.012149	0.044777	0.00402	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5033	8345	0.035281	0.122702	0.01225	69.256623	69.256623	69.256623	0	0	1	-360	360;
	3577	8345	0.031967	0.090248	0.01123	59.503618	59.503618	59.503618	0	0	1	-360	360;
	6219	8345	0.023983	0.066421	0.00824	136.705539	136.705539	136.705539	0	0	1	-360	360;
	8351	6693	0.032901	0.126777	0.01128	60.697642	60.697642	60.697642	0	0	1	-360	360;
	6693	7967	0.015372	0.059083	0.01932	59.601929	59.601929	59.601929	0	0	1	-360	360;
	3827	6451	0.012636	0.024496	0.00439	170.595279	170.595279	170.595279	0	0	1	-360	360;
	3827	6451	0.012893	0.024	0.00447	171.47303	171.47303	171.47303	0	0	1	-360	360;
	3827	6755	0.001917	0.006091	0.05198	134.486488	134.486488	134.486488	0	0	1	-360	360;
	3827	2113	0.018149	0.049413	0.00594	71.710783	71.710783	71.710783	0	0	1	-360	360;
	1984	778	0.01276	0.037083	0.00418	113.802649	113.802649	113.802649	0	0	1	-360	360;
	8082	1984	0.069636	0.146636	0.01749	74.299645	74.299645	74.299645	0	0	1	-360	360;
	2401	8149	0.00219	0.02213	0.52937	1171.58352	1171.58352	1171.58352	0	0	1	-360	360;
	2249	747	0.00194	0.0087	0.01538	449.544746	449.544746	449.544746	0	0	1	-360	360;
	2249	747	0.001979	0.00851	0.01566	425.789253	425.789253	425.789253	0	0	1	-360	360;
	3491	2420	0.028893	0.108496	0.0099	58.1848	58.1848	58.1848	0	0	1	-360	360;
	794	3491	0.021008	0.05814	0.00715	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5131	3548	0.018661	0.056488	0.00615	83.877557	83.877557	83.877557	0	0	1	-360	360;
	7983	949	0.01381	0.03776	0.00704	57.649614	57.649614	57.649614	0	0	1	-360	360;
	7983	3577	0.012298	0.042793	0.00627	63.483317	63.483317	63.483317	0	0	1	-360	360;
	7658	5325	0.009281	0.03038	0.00316	67.378041	67.378041	67.378041	0	0	1	-360	360;
	6954	2552	0.011438	0.035711	0.00395	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7994	6954	0.05095	0.124661	0.01034	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5009	8351	0.058207	0.110223	0.01264	80.361979	80.361979	80.361979	0	0	1	-360	360;
	370	656	0.061876	0.195174	0.02038	61.190341	61.190341	61.190341	0	0	1	-360	360;
	7376	3007	0.013438	0.032033	0.00278	74.02586	74.02586	74.02586	0	0	1	-360	360;
	4169	7376	0.03619	0.085066	0.00739	149.118472	149.118472	149.118472	0	0	1	-360	360;
	3548	2423	0.05324	0.154826	0.0175	60.675351	60.675351	60.675351	0	0	1	-360	360;
	3493	3007	0.004579	0.026017	0.00226	78.630482	78.630482	78.630482	0	0	1	-360	360;
	2910	8238	0.00795	0.03657	0.06296	378.417733	378.417733	378.417733	0	0	1	-360	360;
	2552	1151	0.055967	0.167983	0.01919	60.786236	60.786236	60.786236	0	0	1	-360	360;
	7994	2552	0.070942	0.168066	0.0144	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2266	2401	0.00389	0.04149	0.88556	1263.5904	1263.5904	1263.5904	0	0	1	-360	360;
	6424	2266	0.003	0.02776	0.67411	1126.79047	1126.79047	1126.79047	0	0	1	-360	360;
	4693	4755	0.002159	0.01261	0.01683	382.678266	382.678266	382.678266	0	0	1	-360	360;
	4693	4755	0.002021	0.016849	0.02199	414.680468	414.680468	414.680468	0	0	1	-360	360;
	6356	4755	0.007151	0.046719	0.8722	615.561885	615.561885	615.561885	0	0	1	-360	360;
	4755	6282	0.000539	0.00368	0.01177	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4755	6282	0.000539	0.003591	0.01134	310.824208	310.824208	310.824208	0	0	1	-360	360;
	7183	7658	0.007446	0.010579	0.09655	83.480883	83.480883	83.480883	0	0	1	-360	360;
	7183	2302	0.011992	0.03276	0.01903	85.893508	85.893508	85.893508	0	0	1	-360	360;
	700	2390	0.034207	0.085124	0.00714	72.432113	72.432113	72.432113	0	0	1	-360	360;
	161	700	0.001248	0.004471	0.00048	109.117625	109.117625	109.117625	0	0	1	-360	360;
	161	700	0.001223	0.004562	0.00047	105.978716	105.978716	105.978716	0	0	1	-360	360;
	4992	2910	0.006961	0.04039	0.05461	396.711619	396.711619	396.711619	0	0	1	-360	360;
	4992	2910	0.004521	0.02305	0.03572	391.740806	391.740806	391.740806	0	0	1	-360	360;
	4992	4023	0.00395	0.022789	0.0313	457.657706	457.657706	457.657706	0	0	1	-360	360;
	116	3982	0.031934	0.123661	0.01141	57.157677	57.157677	57.157677	0	0	1	-360	360;
	116	3142	0.019802	0.070901	0.00675	57.157677	57.157677	57.157677	0	0	1	-360	360;
	116	3142	0.022438	0.069479	0.00765	57.157677	57.157677	57.157677	0	0	1	-360	360;
	116	1243	0.014727	0.04305	0.0053	61.77354	61.77354	61.77354	0	0	1	-360	360;
	5220	116	0.01343	0.06824	0.01274	78.81815	78.81815	78.81815	0	0	1	-360	360;
	1764	116	0.022	0.060388	0.00814	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1764	116	0.019603	0.068438	0.00725	66.525248	66.525248	66.525248	0	0	1	-360	360;
	1151	1107	0.048769	0.114711	0.00998	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3153	3990	0.046901	0.148033	0.0155	71.215607	71.215607	71.215607	0	0	1	-360	360;
	6270	3990	0.055744	0.139066	0.01166	81.118556	81.118556	81.118556	0	0	1	-360	360;
	708	5033	0.010917	0.032901	0.00371	58.046669	58.046669	58.046669	0	0	1	-360	360;
	6650	4998	0.026322	0.044388	0.04865	65.811349	65.811349	65.811349	0	0	1	-360	360;
	2673	7506	0.047438	0.143727	0.01562	86.180821	86.180821	86.180821	0	0	1	-360	360;
	6899	7506	0.026876	0.09776	0.02275	75.169775	75.169775	75.169775	0	0	1	-360	360;
	7967	6899	0.01705	0.060802	0.00588	61.15738	61.15738	61.15738	0	0	1	-360	360;
	823	7967	0.028727	0.075289	0.07038	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2673	5072	0.002388	0.015942	0.00343	171.47303	171.47303	171.47303	0	0	1	-360	360;
	2673	2544	0.088471	0.217198	0.01836	75.133385	75.133385	75.133385	0	0	1	-360	360;
	2423	6038	0.01714	0.056248	0.00571	57.724871	57.724871	57.724871	0	0	1	-360	360;
	5849	3153	0.004479	0.014008	0.00143	94.919467	94.919467	94.919467	0	0	1	-360	360;
	2704	5849	0.009851	0.032314	0.00324	58.847829	58.847829	58.847829	0	0	1	-360	360;
	4223	5849	0.001636	0.005587	0.05447	79.452981	79.452981	79.452981	0	0	1	-360	360;
	4223	2976	0.021603	0.084	0.00765	64.073946	64.073946	64.073946	0	0	1	-360	360;
	1629	4270	0.021727	0.065893	0.00742	59.575827	59.575827	59.575827	0	0	1	-360	360;
	1629	4169	0.025686	0.080256	0.00884	66.926495	66.926495	66.926495	0	0	1	-360	360;
	4169	1629	0.026182	0.078587	0.00901	67.81644	67.81644	67.81644	0	0	1	-360	360;
	6356	4693	0.00536	0.040919	0.05922	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2302	3493	0.007388	0.026628	0.01555	81.145229	81.145229	81.145229	0	0	1	-360	360;
	1392	4015	0.005826	0.017488	0.00201	60.398517	60.398517	60.398517	0	0	1	-360	360;
	1392	4015	0.005826	0.016653	0.00205	70.298798	70.298798	70.298798	0	0	1	-360	360;
	3577	4015	0.026826	0.082273	0.00936	74.035958	74.035958	74.035958	0	0	1	-360	360;
	8098	1392	0.003256	0.010587	0.00163	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8098	876	0.003331	0.010372	0.00166	94.78629	94.78629	94.78629	0	0	1	-360	360;
	7468	7308	0.025174	0.077661	0.00874	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2447	4270	0.030248	0.085413	0.01045	80.740744	80.740744	80.740744	0	0	1	-360	360;
	4577	4270	0.029471	0.109198	0.01008	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1107	166	0.058942	0.115529	0.0123	69.580898	69.580898	69.580898	0	0	1	-360	360;
	1243	3982	0.028413	0.083	0.01018	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7994	3982	0.005826	0.016	0.00205	62.624427	62.624427	62.624427	0	0	1	-360	360;
	7757	3982	0.005512	0.016612	0.00198	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5174	1392	0.014256	0.033909	0.00267	75.107283	75.107283	75.107283	0	0	1	-360	360;
	656	1392	0.052826	0.200702	0.01809	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1764	1392	0.01976	0.061058	0.00697	88.672514	88.672514	88.672514	0	0	1	-360	360;
	1764	1392	0.014008	0.053207	0.00526	67.843304	67.843304	67.843304	0	0	1	-360	360;
	1764	1392	0.013711	0.054207	0.00515	77.613838	77.613838	77.613838	0	0	1	-360	360;
	5220	3142	0.009132	0.024843	0.0022	62.934984	62.934984	62.934984	0	0	1	-360	360;
	6219	3142	0.010711	0.040256	0.00369	87.989861	87.989861	87.989861	0	0	1	-360	360;
	5923	3142	0.004562	0.012421	0.0011	74.876175	74.876175	74.876175	0	0	1	-360	360;
	7259	2447	0.025273	0.079777	0.00863	72.714663	72.714663	72.714663	0	0	1	-360	360;
	876	5174	0.009149	0.028992	0.00312	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4690	2113	0.027149	0.096802	0.00897	69.235856	69.235856	69.235856	0	0	1	-360	360;
	748	4169	0.063033	0.157099	0.0136	57.157677	57.157677	57.157677	0	0	1	-360	360;
	748	7259	0.009702	0.035397	0.00333	77.08913	77.08913	77.08913	0	0	1	-360	360;
	748	7259	0.009504	0.036091	0.00326	67.9902	67.9902	67.9902	0	0	1	-360	360;
	213	6270	0.029893	0.105364	0.01029	64.441279	64.441279	64.441279	0	0	1	-360	360;
	5220	5151	0.005397	0.020298	0.00189	98.404371	98.404371	98.404371	0	0	1	-360	360;
	6755	2605	0.001124	0.002942	0.03069	112.453918	112.453918	112.453918	0	0	1	-360	360;
	161	2605	0.001207	0.003174	0.03302	84.791889	84.791889	84.791889	0	0	1	-360	360;
	161	2605	0.00124	0.003099	0.03366	80.407705	80.407705	80.407705	0	0	1	-360	360;
	5923	5220	0.004066	0.014083	0.00098	61.14652	61.14652	61.14652	0	0	1	-360	360;
	2170	6356	0.005099	0.025479	0.03978	393.786289	393.786289	393.786289	0	0	1	-360	360;
	2170	6356	0.0052	0.024961	0.04056	379.871443	379.871443	379.871443	0	0	1	-360	360;
	7757	7994	0.003868	0.013901	0.00135	58.812201	58.812201	58.812201	0	0	1	-360	360;
	6424	1301	0.00126	0.01433	0.28656	1265.17727	1265.17727	1265.17727	0	0	1	-360	360;
	7468	1764	0.011752	0.044314	0.00414	68.895577	68.895577	68.895577	0	0	1	-360	360;
	8140	6415	0.021719	0.046248	0.00449	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8140	5837	0.05324	0.100281	0.01112	58.18823	58.18823	58.18823	0	0	1	-360	360;
	7628	6116	0.00109	0.01155	0.41153	1111.94129	1111.94129	1111.94129	0	0	1	-360	360;
	7628	6116	0.0011	0.01158	0.41253	1045.98053	1045.98053	1045.98053	0	0	1	-360	360;
	7628	7904	0.00096	0.01293	0.39742	1237.78188	1237.78188	1237.78188	0	0	1	-360	360;
	7628	7904	0.00098	0.01264	0.40478	1208.96942	1208.96942	1208.96942	0	0	1	-360	360;
	7628	514	0.00101	0.01059	0.20637	1406.77402	1406.77402	1406.77402	0	0	1	-360	360;
	7305	5629	0.00112	0.010969	0.02902	535.344896	535.344896	535.344896	0	0	1	-360	360;
	264	5629	0.00126	0.01075	0.03289	552.226607	552.226607	552.226607	0	0	1	-360	360;
	3351	7780	0.00197	0.0207	0.4876	986.330397	986.330397	986.330397	0	0	1	-360	360;
	3351	514	0.00303	0.0273	0.51885	1325.18019	1325.18019	1325.18019	0	0	1	-360	360;
	306	7316	0.00194	0.02339	0.48042	913.854324	913.854324	913.854324	0	0	1	-360	360;
	7904	306	0.00057	0.00492	0.12431	999.411052	999.411052	999.411052	0	0	1	-360	360;
	6062	2589	0.01105	0.056599	0.08685	351.772729	351.772729	351.772729	0	0	1	-360	360;
	4729	2589	0.00482	0.046521	0.11561	388.703447	388.703447	388.703447	0	0	1	-360	360;
	6116	7729	0.00194	0.01645	0.54766	985.064718	985.064718	985.064718	0	0	1	-360	360;
	7729	5306	0.00134	0.01459	0.38171	1046.11217	1046.11217	1046.11217	0	0	1	-360	360;
	4802	7734	0.000401	0.00318	0.00459	615.119104	615.119104	615.119104	0	0	1	-360	360;
	4802	6408	0.000349	0.00293	0.00412	571.646499	571.646499	571.646499	0	0	1	-360	360;
	4802	1276	0.000401	0.00324	0.00449	628.535915	628.535915	628.535915	0	0	1	-360	360;
	4802	2589	0.01214	0.058401	0.09615	341.784616	341.784616	341.784616	0	0	1	-360	360;
	4802	2589	0.012979	0.053281	0.12809	332.620716	332.620716	332.620716	0	0	1	-360	360;
	4802	2951	0.007919	0.053729	0.08698	549.153429	549.153429	549.153429	0	0	1	-360	360;
	4802	6167	0.014771	0.06813	0.11701	440.256242	440.256242	440.256242	0	0	1	-360	360;
	4802	6167	0.013789	0.067711	0.10063	449.015847	449.015847	449.015847	0	0	1	-360	360;
	1502	5629	0.005349	0.044469	0.06071	534.264997	534.264997	534.264997	0	0	1	-360	360;
	8176	5629	0.005229	0.04118	0.05939	534.356069	534.356069	534.356069	0	0	1	-360	360;
	5629	4469	0.0008	0.00676	0.00895	556.213926	556.213926	556.213926	0	0	1	-360	360;
	5629	1953	0.000781	0.00626	0.00876	258.673162	258.673162	258.673162	0	0	1	-360	360;
	5629	7138	0.00363	0.017409	0.05251	268.298515	268.298515	268.298515	0	0	1	-360	360;
	5629	7138	0.0037	0.01705	0.05352	271.007027	271.007027	271.007027	0	0	1	-360	360;
	5629	4544	0.00449	0.046469	0.09119	508.857648	508.857648	508.857648	0	0	1	-360	360;
	4466	3279	0.00052	0.00727	0.12854	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4466	5306	0.00047	0.0062	0.11428	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4466	5306	0.00046	0.00631	0.11184	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4547	6167	0.003531	0.013669	0.03479	497.933292	497.933292	497.933292	0	0	1	-360	360;
	6761	7316	0.00032	0.00322	0.07795	1131.80448	1131.80448	1131.80448	0	0	1	-360	360;
	6761	4382	0.00156	0.01603	0.38661	1009.23372	1009.23372	1009.23372	0	0	1	-360	360;
	6761	2152	0.0022	0.02389	0.63855	1444.98199	1444.98199	1444.98199	0	0	1	-360	360;
	6966	4544	0.00337	0.032669	0.08214	405.483798	405.483798	405.483798	0	0	1	-360	360;
	8333	4382	0.00199	0.0195	0.46272	1157.96052	1157.96052	1157.96052	0	0	1	-360	360;
	8333	2152	0.00144	0.01354	0.34334	1043.65321	1043.65321	1043.65321	0	0	1	-360	360;
	8333	2152	0.00147	0.01325	0.34982	1136.27878	1136.27878	1136.27878	0	0	1	-360	360;
	8333	4605	0.00211	0.02433	0.49124	1156.42302	1156.42302	1156.42302	0	0	1	-360	360;
	3779	2589	0.013901	0.088849	0.11777	326.736905	326.736905	326.736905	0	0	1	-360	360;
	4382	7904	0.00149	0.01564	0.3616	1073.55298	1073.55298	1073.55298	0	0	1	-360	360;
	4382	7239	0.00294	0.02278	0.414	1356.00404	1356.00404	1356.00404	0	0	1	-360	360;
	7904	7239	0.00133	0.01873	0.53568	1160.59456	1160.59456	1160.59456	0	0	1	-360	360;
	7138	6167	0.00581	0.034229	0.08461	585.003105	585.003105	585.003105	0	0	1	-360	360;
	7239	6159	0.00145	0.02024	0.36059	1625.62391	1625.62391	1625.62391	0	0	1	-360	360;
	7239	6159	0.00142	0.01874	0.35276	1772.13396	1772.13396	1772.13396	0	0	1	-360	360;
	7239	7780	0.00245	0.02367	0.59282	1178.74122	1178.74122	1178.74122	0	0	1	-360	360;
	7239	4605	0.00178	0.02175	0.43656	1133.70135	1133.70135	1133.70135	0	0	1	-360	360;
	5102	6487	0.000701	0.009299	0.0019	83.894791	83.894791	83.894791	0	0	1	-360	360;
	5102	6487	0.001042	0.007833	0.0024	69.281963	69.281963	69.281963	0	0	1	-360	360;
	5102	6487	0.001403	0.00534	0.0027	74.434468	74.434468	74.434468	0	0	1	-360	360;
	5306	2849	0.00022	0.00258	0.05115	830.598591	830.598591	830.598591	0	0	1	-360	360;
	2589	2951	0.02024	0.099349	0.14928	362.314891	362.314891	362.314891	0	0	1	-360	360;
	2589	5539	0.00044	0.003081	0.01938	559.573654	559.573654	559.573654	0	0	1	-360	360;
	2589	2100	0.00043	0.00314	0.01899	559.573654	559.573654	559.573654	0	0	1	-360	360;
	2589	7915	0.00461	0.027531	0.0363	354.394361	354.394361	354.394361	0	0	1	-360	360;
	8119	2951	0.01212	0.09901	0.13301	398.151992	398.151992	398.151992	0	0	1	-360	360;
	328	5921	0.000289	0.00261	0.06705	382.74495	382.74495	382.74495	0	0	1	-360	360;
	4160	328	0.00043	0.002831	0.04948	441.38987	441.38987	441.38987	0	0	1	-360	360;
	3187	328	0.000289	0.000599	0.04587	425.688275	425.688275	425.688275	0	0	1	-360	360;
	6791	1038	0.000289	0.00224	0.00387	533.728096	533.728096	533.728096	0	0	1	-360	360;
	3763	6791	0.000219	0.001591	0.00286	417.648476	417.648476	417.648476	0	0	1	-360	360;
	2442	6101	0.00045	0.00786	0.20766	1841.01243	1841.01243	1841.01243	0	0	1	-360	360;
	2442	6101	0.00044	0.00728	0.20315	1650.45175	1650.45175	1650.45175	0	0	1	-360	360;
	6101	7481	0.00049	0.00466	0.14017	1527.50385	1527.50385	1527.50385	0	0	1	-360	360;
	6101	7358	0.00125	0.01973	0.57488	2025.04265	2025.04265	2025.04265	0	0	1	-360	360;
	6178	6177	0.001461	0.00714	0.01432	422.673017	422.673017	422.673017	0	0	1	-360	360;
	8269	6178	0.00037	0.00094	0.1353	427.135888	427.135888	427.135888	0	0	1	-360	360;
	6178	1650	0.001039	0.00737	0.01181	339.214426	339.214426	339.214426	0	0	1	-360	360;
	4817	5738	8.1e-05	0.00061	0.00092	386.063525	386.063525	386.063525	0	0	1	-360	360;
	6628	7320	0.00053	0.00852	0.23849	1629.47755	1629.47755	1629.47755	0	0	1	-360	360;
	4969	7320	0.00053	0.00704	0.24029	1750.19487	1750.19487	1750.19487	0	0	1	-360	360;
	4647	3052	0.0001	0.00128	0.12197	833.732841	833.732841	833.732841	0	0	1	-360	360;
	3959	8239	0.0015	0.00705	0.01181	426.65805	426.65805	426.65805	0	0	1	-360	360;
	8239	6409	0.009771	0.05363	0.07652	412.437601	412.437601	412.437601	0	0	1	-360	360;
	3959	8239	0.00088	0.007771	0.01012	457.238931	457.238931	457.238931	0	0	1	-360	360;
	3777	7651	0.00539	0.032979	0.04202	401.522771	401.522771	401.522771	0	0	1	-360	360;
	1337	3777	0.00482	0.03086	0.05736	314.252144	314.252144	314.252144	0	0	1	-360	360;
	1949	3777	0.010539	0.065479	0.14325	473.383689	473.383689	473.383689	0	0	1	-360	360;
	3777	7163	0.01074	0.06405	0.1459	481.522942	481.522942	481.522942	0	0	1	-360	360;
	4052	3777	0.0063	0.048909	0.07462	535.182187	535.182187	535.182187	0	0	1	-360	360;
	1663	812	0.0002	0.00022	0.00277	1426.45819	1426.45819	1426.45819	0	0	1	-360	360;
	4530	812	0.0002	0.00022	0.00272	1466.19708	1466.19708	1466.19708	0	0	1	-360	360;
	6188	812	0.00019	0.00023	0.00266	1495.41103	1495.41103	1495.41103	0	0	1	-360	360;
	836	1164	0.000289	0.0017	0.00219	215.929128	215.929128	215.929128	0	0	1	-360	360;
	812	1085	0.00145	0.01937	0.68469	1999.46909	1999.46909	1999.46909	0	0	1	-360	360;
	2842	812	0.00048	0.00814	0.22346	1788.07901	1788.07901	1788.07901	0	0	1	-360	360;
	2842	812	0.00057	0.00758	0.26458	1636.48518	1636.48518	1636.48518	0	0	1	-360	360;
	7598	3464	0.0002	0.00187	0.00659	313.789929	313.789929	313.789929	0	0	1	-360	360;
	7598	3464	0.000169	0.00193	0.00564	308.383956	308.383956	308.383956	0	0	1	-360	360;
	1934	1806	0.00216	0.03591	1.05032	999.743433	999.743433	999.743433	0	0	1	-360	360;
	1934	3043	0.00092	0.01287	0.43829	1816.57686	1816.57686	1816.57686	0	0	1	-360	360;
	1146	3280	0.02412	0.10443	0.18558	476.4401	476.4401	476.4401	0	0	1	-360	360;
	1164	7651	0.00063	0.00401	0.01805	240.712315	240.712315	240.712315	0	0	1	-360	360;
	1164	3248	0.00038	0.002409	0.0109	276.250672	276.250672	276.250672	0	0	1	-360	360;
	4435	3727	0.010091	0.05524	0.09845	517.379476	517.379476	517.379476	0	0	1	-360	360;
	4984	4435	0.00668	0.04457	0.06427	450.043923	450.043923	450.043923	0	0	1	-360	360;
	6362	1722	0.00057	0.004961	0.05619	425.129654	425.129654	425.129654	0	0	1	-360	360;
	1722	4974	0.00025	0.002159	0.05179	361.400749	361.400749	361.400749	0	0	1	-360	360;
	476	4111	0.00443	0.01807	0.03672	441.326615	441.326615	441.326615	0	0	1	-360	360;
	476	5298	0.00037	0.001659	0.00331	495.916007	495.916007	495.916007	0	0	1	-360	360;
	6362	755	0.00037	0.003599	0.51014	494.850969	494.850969	494.850969	0	0	1	-360	360;
	6362	6796	0.000659	0.00836	0.01984	443.383148	443.383148	443.383148	0	0	1	-360	360;
	6362	6796	0.00151	0.012289	0.02276	383.067319	383.067319	383.067319	0	0	1	-360	360;
	6362	2169	0.000539	0.003711	0.0065	393.078296	393.078296	393.078296	0	0	1	-360	360;
	7611	1210	0.00036	0.00606	0.16471	1099.36282	1099.36282	1099.36282	0	0	1	-360	360;
	7611	1210	0.00035	0.00618	0.1612	949.731678	949.731678	949.731678	0	0	1	-360	360;
	1210	597	0.00101	0.01675	0.48261	2008.22222	2008.22222	2008.22222	0	0	1	-360	360;
	1210	4864	4e-05	0.00071	0.02437	1629.89286	1629.89286	1629.89286	0	0	1	-360	360;
	1210	4864	4e-05	0.00081	0.02107	1805.65569	1805.65569	1805.65569	0	0	1	-360	360;
	1210	893	0.00013	0.00152	0.038	2147.154	2147.154	2147.154	0	0	1	-360	360;
	1210	893	0.00014	0.00149	0.03875	989.791762	989.791762	989.791762	0	0	1	-360	360;
	1210	3806	0.00013	0.00164	0.0366	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1210	3806	0.00013	0.00168	0.03583	902.635658	902.635658	902.635658	0	0	1	-360	360;
	5460	4052	0.000669	0.00511	0.00806	476.804004	476.804004	476.804004	0	0	1	-360	360;
	8137	3147	3.9e-05	0.000341	0.00051	298.744504	298.744504	298.744504	0	0	1	-360	360;
	5339	5153	0.004219	0.02145	0.04075	325.757603	325.757603	325.757603	0	0	1	-360	360;
	5339	3677	0.000229	0.002849	0.00675	633.628664	633.628664	633.628664	0	0	1	-360	360;
	1146	46	0.0137	0.10436	0.17076	458.358078	458.358078	458.358078	0	0	1	-360	360;
	2425	1146	0.001521	0.01449	0.04952	528.627726	528.627726	528.627726	0	0	1	-360	360;
	694	393	8.1e-05	0.00057	0.00098	540.212063	540.212063	540.212063	0	0	1	-360	360;
	6177	393	0.001461	0.00813	0.01346	488.30832	488.30832	488.30832	0	0	1	-360	360;
	1295	393	0.00049	0.00381	0.00551	460.350595	460.350595	460.350595	0	0	1	-360	360;
	6840	2336	0.000401	0.00287	0.00499	539.15236	539.15236	539.15236	0	0	1	-360	360;
	5178	4384	8.3e-05	0.000347	0.00118	77.349301	77.349301	77.349301	0	0	1	-360	360;
	5948	6388	0.00023	0.00285	0.1104	1326.04109	1326.04109	1326.04109	0	0	1	-360	360;
	6388	2908	0.00016	0.00218	0.07368	2143.63208	2143.63208	2143.63208	0	0	1	-360	360;
	2714	6388	0.00057	0.00736	0.27483	1630.35292	1630.35292	1630.35292	0	0	1	-360	360;
	363	7126	0.000591	0.005789	0.00881	510.777384	510.777384	510.777384	0	0	1	-360	360;
	363	3030	0.002729	0.011461	0.04693	513.788831	513.788831	513.788831	0	0	1	-360	360;
	363	5543	0.00051	0.00205	0.27176	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7499	6122	0.00031	0.00374	0.1517	1876.49356	1876.49356	1876.49356	0	0	1	-360	360;
	2458	7499	0.00032	0.00518	0.15553	2224.46243	2224.46243	2224.46243	0	0	1	-360	360;
	1817	6623	0.00136	0.01038	0.0167	487.959277	487.959277	487.959277	0	0	1	-360	360;
	7735	3679	0.00151	0.02083	0.71503	1838.46461	1838.46461	1838.46461	0	0	1	-360	360;
	6953	3679	0.00015	0.00232	0.07104	1698.10919	1698.10919	1698.10919	0	0	1	-360	360;
	2007	7361	1e-05	0.00025	0.00413	1008.90266	1008.90266	1008.90266	0	0	1	-360	360;
	8178	7361	1e-05	0.00023	0.00469	1008.1214	1008.1214	1008.1214	0	0	1	-360	360;
	2199	7361	0.00047	0.00683	0.20319	1206.51968	1206.51968	1206.51968	0	0	1	-360	360;
	797	7361	0.00067	0.00858	0.31718	1310.09406	1310.09406	1310.09406	0	0	1	-360	360;
	2299	942	0.003669	0.025331	0.09971	518.573691	518.573691	518.573691	0	0	1	-360	360;
	331	2299	0.00525	0.02994	0.14222	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4245	6226	0.00595	0.03625	0.08592	506.116747	506.116747	506.116747	0	0	1	-360	360;
	4245	7113	0.004969	0.03411	0.07996	484.939447	484.939447	484.939447	0	0	1	-360	360;
	510	1700	0.005669	0.04293	0.0694	571.45102	571.45102	571.45102	0	0	1	-360	360;
	510	1700	0.00555	0.043771	0.06799	527.170968	527.170968	527.170968	0	0	1	-360	360;
	1700	3487	0.002831	0.022401	0.03479	360.361242	360.361242	360.361242	0	0	1	-360	360;
	1700	3487	0.00325	0.01938	0.03984	413.949612	413.949612	413.949612	0	0	1	-360	360;
	7302	143	0.00068	0.00082	1.15469	1038.6076	1038.6076	1038.6076	0	0	1	-360	360;
	7302	1862	0.00054	0.00868	0.17111	1362.94849	1362.94849	1362.94849	0	0	1	-360	360;
	1608	6162	0.00061	0.002909	0.00474	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1608	1336	0.002271	0.01014	0.01813	257.572306	257.572306	257.572306	0	0	1	-360	360;
	3547	1149	0.00083	0.00986	0.4002	2082.86568	2082.86568	2082.86568	0	0	1	-360	360;
	1149	7375	0.00031	0.00511	0.15071	1640.04198	1640.04198	1640.04198	0	0	1	-360	360;
	1149	7375	0.0003	0.00521	0.1475	1656.01994	1656.01994	1656.01994	0	0	1	-360	360;
	7805	1149	0.00081	0.01288	0.38613	2147.06515	2147.06515	2147.06515	0	0	1	-360	360;
	1149	8181	0.00039	0.00645	0.18093	1999.08932	1999.08932	1999.08932	0	0	1	-360	360;
	3431	2109	0.00243	0.012211	0.01893	328.037814	328.037814	328.037814	0	0	1	-360	360;
	3242	4816	0.00018	0.00226	0.08496	2142.90611	2142.90611	2142.90611	0	0	1	-360	360;
	2105	4816	0.0002	0.00347	0.097	1759.77006	1759.77006	1759.77006	0	0	1	-360	360;
	2908	4816	0.00018	0.0029	0.09086	1992.64509	1992.64509	1992.64509	0	0	1	-360	360;
	4989	6122	0.00045	0.00494	0.1307	1412.30141	1412.30141	1412.30141	0	0	1	-360	360;
	6122	2433	0.00042	0.00497	0.14413	1614.23938	1614.23938	1614.23938	0	0	1	-360	360;
	2936	7735	0.00019	0.00234	0.05305	1269.08818	1269.08818	1269.08818	0	0	1	-360	360;
	1806	7735	8e-05	0.00105	0.04323	1678.42766	1678.42766	1678.42766	0	0	1	-360	360;
	1806	7735	8e-05	0.00107	0.04235	1617.63822	1617.63822	1617.63822	0	0	1	-360	360;
	7735	7189	8e-05	0.001	0.40209	2100.72274	2100.72274	2100.72274	0	0	1	-360	360;
	7735	3620	0.00063	0.01015	0.29913	1867.9517	1867.9517	1867.9517	0	0	1	-360	360;
	795	6051	0.00064	0.005651	0.02099	215.929128	215.929128	215.929128	0	0	1	-360	360;
	795	6051	0.000581	0.006531	0.01907	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4651	7912	0.007099	0.04432	0.06825	400.579669	400.579669	400.579669	0	0	1	-360	360;
	5921	1650	0.00144	0.01157	0.01638	438.946188	438.946188	438.946188	0	0	1	-360	360;
	5921	7077	0.00114	0.00695	0.01238	435.365831	435.365831	435.365831	0	0	1	-360	360;
	3242	1536	0.00044	0.0053	0.21006	1642.4779	1642.4779	1642.4779	0	0	1	-360	360;
	6414	6005	0.00162	0.009039	0.03158	404.154692	404.154692	404.154692	0	0	1	-360	360;
	6414	6005	0.00187	0.0072	0.06935	403.31295	403.31295	403.31295	0	0	1	-360	360;
	1209	6005	0.00101	0.00564	0.00972	257.289947	257.289947	257.289947	0	0	1	-360	360;
	3370	6005	0.0007	0.00445	0.0112	258.04519	258.04519	258.04519	0	0	1	-360	360;
	2638	6005	0.001081	0.00619	0.03556	556.806842	556.806842	556.806842	0	0	1	-360	360;
	1806	7529	0.00022	0.00228	0.06341	1450.21451	1450.21451	1450.21451	0	0	1	-360	360;
	1806	5632	0.00023	0.00223	0.06643	1438.16457	1438.16457	1438.16457	0	0	1	-360	360;
	1806	7205	0.0002	0.00262	0.05709	1450.2336	1450.2336	1450.2336	0	0	1	-360	360;
	1806	1756	0.00021	0.00261	0.05868	1308.20311	1308.20311	1308.20311	0	0	1	-360	360;
	42	6340	0.00663	0.04181	0.08125	461.79859	461.79859	461.79859	0	0	1	-360	360;
	1806	2226	0.00059	0.00921	0.27853	1705.04772	1705.04772	1705.04772	0	0	1	-360	360;
	1806	3788	0.0014	0.02227	0.67283	1778.16354	1778.16354	1778.16354	0	0	1	-360	360;
	3524	2554	0.000401	0.002469	0.00474	442.871015	442.871015	442.871015	0	0	1	-360	360;
	5123	2554	0.00212	0.01218	0.01702	384.804151	384.804151	384.804151	0	0	1	-360	360;
	2554	1849	0.004169	0.02439	0.01892	448.956403	448.956403	448.956403	0	0	1	-360	360;
	7651	8235	0.02336	0.122419	0.22706	336.17821	336.17821	336.17821	0	0	1	-360	360;
	7651	620	0.002729	0.0137	0.08253	353.368572	353.368572	353.368572	0	0	1	-360	360;
	2610	7651	0.000919	0.003831	0.00829	376.875238	376.875238	376.875238	0	0	1	-360	360;
	2610	7651	0.00082	0.004349	0.00739	359.109108	359.109108	359.109108	0	0	1	-360	360;
	4111	7651	0.01325	0.055159	0.10796	454.823448	454.823448	454.823448	0	0	1	-360	360;
	4111	7651	0.0108	0.06074	0.08728	442.82186	442.82186	442.82186	0	0	1	-360	360;
	7651	4378	0.001099	0.00631	0.00879	422.346075	422.346075	422.346075	0	0	1	-360	360;
	7064	5738	0.00495	0.029151	0.03895	428.408599	428.408599	428.408599	0	0	1	-360	360;
	5738	8110	0.003909	0.0227	0.03087	401.96822	401.96822	401.96822	0	0	1	-360	360;
	1949	3235	0.00061	0.00424	0.00726	312.454345	312.454345	312.454345	0	0	1	-360	360;
	2602	1038	0.00068	0.003919	0.00535	532.10558	532.10558	532.10558	0	0	1	-360	360;
	4161	1038	0.00368	0.019901	0.02641	241.022491	241.022491	241.022491	0	0	1	-360	360;
	6726	2105	0.00025	0.00388	0.12151	1477.94163	1477.94163	1477.94163	0	0	1	-360	360;
	6726	337	0.00038	0.00623	0.18345	1224.21614	1224.21614	1224.21614	0	0	1	-360	360;
	6177	4337	0.000909	0.006031	0.00929	446.598458	446.598458	446.598458	0	0	1	-360	360;
	4337	1650	0.001211	0.0062	0.01223	517.71061	517.71061	517.71061	0	0	1	-360	360;
	2199	366	0.00022	0.00322	0.10273	1140.78468	1140.78468	1140.78468	0	0	1	-360	360;
	2199	4108	0.00018	0.00307	0.08602	1779.65629	1779.65629	1779.65629	0	0	1	-360	360;
	2199	4482	0.00068	0.00873	0.32682	1125.18319	1125.18319	1125.18319	0	0	1	-360	360;
	2199	7396	0.00124	0.02016	0.59343	1334.08207	1334.08207	1334.08207	0	0	1	-360	360;
	8291	6628	0.00025	0.00332	0.13434	1564.45799	1564.45799	1564.45799	0	0	1	-360	360;
	8291	6628	0.00025	0.00335	0.1165	1753.70955	1753.70955	1753.70955	0	0	1	-360	360;
	8291	6628	0.00025	0.00325	0.13692	1711.40178	1711.40178	1711.40178	0	0	1	-360	360;
	2790	8291	0.00076	0.01075	0.36484	1779.00206	1779.00206	1779.00206	0	0	1	-360	360;
	4428	8291	3e-05	0.00044	0.01014	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3147	6838	0.001651	0.010849	0.01913	549.79169	549.79169	549.79169	0	0	1	-360	360;
	7542	3147	0.00288	0.018469	0.03437	481.001283	481.001283	481.001283	0	0	1	-360	360;
	6255	1378	0.00072	0.00923	0.3409	1381.08857	1381.08857	1381.08857	0	0	1	-360	360;
	6255	5615	0.00069	0.00855	0.33299	1769.4453	1769.4453	1769.4453	0	0	1	-360	360;
	2893	4163	0.00124	0.01667	0.58961	2147.10595	2147.10595	2147.10595	0	0	1	-360	360;
	2893	751	0.00034	0.00571	0.1561	1829.53839	1829.53839	1829.53839	0	0	1	-360	360;
	6294	2893	0.0004	0.00638	0.19536	1930.75912	1930.75912	1930.75912	0	0	1	-360	360;
	6294	2893	0.00039	0.0065	0.19129	1897.63559	1897.63559	1897.63559	0	0	1	-360	360;
	7161	2893	0.00027	0.00451	0.12295	1684.5869	1684.5869	1684.5869	0	0	1	-360	360;
	7446	3544	0.012409	0.074961	0.11518	394.975931	394.975931	394.975931	0	0	1	-360	360;
	7446	7797	0.002419	0.01761	0.02279	442.990666	442.990666	442.990666	0	0	1	-360	360;
	7446	1295	0.006901	0.053409	0.06908	457.935112	457.935112	457.935112	0	0	1	-360	360;
	2602	10	0.000599	0.00356	0.0048	502.624413	502.624413	502.624413	0	0	1	-360	360;
	2602	4111	0.011841	0.049341	0.08708	435.254946	435.254946	435.254946	0	0	1	-360	360;
	4160	817	0.000521	0.004831	0.01	541.421519	541.421519	541.421519	0	0	1	-360	360;
	77	817	0.00055	0.00143	0.13971	395.15693	395.15693	395.15693	0	0	1	-360	360;
	7002	817	0.0005	0.00364	0.06374	346.468878	346.468878	346.468878	0	0	1	-360	360;
	7002	817	0.00056	0.003469	0.07625	350.424189	350.424189	350.424189	0	0	1	-360	360;
	817	4712	0.00082	0.00257	0.24	459.323281	459.323281	459.323281	0	0	1	-360	360;
	4117	4945	0.00249	0.01536	0.01915	475.615124	475.615124	475.615124	0	0	1	-360	360;
	4117	2252	0.00113	0.00651	0.00884	451.763988	451.763988	451.763988	0	0	1	-360	360;
	5853	4620	0.00643	0.045031	0.07877	520.635178	520.635178	520.635178	0	0	1	-360	360;
	7714	5853	0.003581	0.02706	0.05928	531.014631	531.014631	531.014631	0	0	1	-360	360;
	1658	3043	0.00046	0.00774	0.23407	1823.86949	1823.86949	1823.86949	0	0	1	-360	360;
	1658	5865	0.00031	0.00483	0.14581	1107.41038	1107.41038	1107.41038	0	0	1	-360	360;
	5392	3429	0.001099	0.00345	0.53179	467.185891	467.185891	467.185891	0	0	1	-360	360;
	1142	894	0.00913	0.053211	0.07112	368.983287	368.983287	368.983287	0	0	1	-360	360;
	894	7727	0.021151	0.095401	0.16513	369.54991	369.54991	369.54991	0	0	1	-360	360;
	5590	4163	0.00034	0.0055	0.16133	1816.26159	1816.26159	1816.26159	0	0	1	-360	360;
	2498	3225	0.00064	0.01945	0.55897	1821.05511	1821.05511	1821.05511	0	0	1	-360	360;
	2498	7857	0.0004	0.01026	0.34782	1474.75539	1474.75539	1474.75539	0	0	1	-360	360;
	6589	2498	0.00071	0.01087	0.34033	1268.70643	1268.70643	1268.70643	0	0	1	-360	360;
	7972	1335	0.00968	0.088169	0.13353	535.226008	535.226008	535.226008	0	0	1	-360	360;
	7972	1335	0.012349	0.086409	0.15134	475.021828	475.021828	475.021828	0	0	1	-360	360;
	1335	3723	0.00288	0.017211	0.03535	548.482779	548.482779	548.482779	0	0	1	-360	360;
	3723	1335	0.00257	0.0195	0.0315	501.717511	501.717511	501.717511	0	0	1	-360	360;
	2799	942	0.00051	0.00407	0.02631	568.535597	568.535597	568.535597	0	0	1	-360	360;
	525	5922	0.00019	0.00309	0.09344	1529.78971	1529.78971	1529.78971	0	0	1	-360	360;
	525	6448	0.00023	0.00357	0.11217	1616.91948	1616.91948	1616.91948	0	0	1	-360	360;
	2986	3536	0.00088	0.0092	0.19005	1190.0494	1190.0494	1190.0494	0	0	1	-360	360;
	2986	4253	0.00023	0.00179	0.05039	1588.66845	1588.66845	1588.66845	0	0	1	-360	360;
	5510	5192	0.00145	0.007919	0.01161	497.283218	497.283218	497.283218	0	0	1	-360	360;
	425	5192	0.00143	0.00812	0.01675	499.145797	499.145797	499.145797	0	0	1	-360	360;
	5192	5153	0.00524	0.02181	0.04255	525.24323	525.24323	525.24323	0	0	1	-360	360;
	1863	3431	0.00261	0.01913	0.14499	451.716356	451.716356	451.716356	0	0	1	-360	360;
	4111	3431	0.00164	0.016591	0.09122	572.382309	572.382309	572.382309	0	0	1	-360	360;
	5022	4775	0.00087	0.012719	0.57325	443.648741	443.648741	443.648741	0	0	1	-360	360;
	1	6932	0.00775	0.03501	0.01357	405.595827	405.595827	405.595827	0	0	1	-360	360;
	1	7107	0.01944	0.093479	0.15775	458.426287	458.426287	458.426287	0	0	1	-360	360;
	1	7107	0.01981	0.091539	0.16078	479.127654	479.127654	479.127654	0	0	1	-360	360;
	1388	5111	0.003771	0.020599	0.02939	428.908157	428.908157	428.908157	0	0	1	-360	360;
	7873	1388	0.001271	0.00811	0.01217	508.806206	508.806206	508.806206	0	0	1	-360	360;
	7913	978	0.00048	0.01459	0.41923	1160.86507	1160.86507	1160.86507	0	0	1	-360	360;
	3225	7913	0.00047	0.01351	0.41011	1740.23728	1740.23728	1740.23728	0	0	1	-360	360;
	2150	1337	0.021729	0.088281	0.08487	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1337	2610	0.00038	0.001771	0.00301	504.745725	504.745725	504.745725	0	0	1	-360	360;
	1337	1142	0.022901	0.101401	0.13759	487.958515	487.958515	487.958515	0	0	1	-360	360;
	4	4953	0.002909	0.01432	0.09089	495.301371	495.301371	495.301371	0	0	1	-360	360;
	4	2425	0.015729	0.08824	0.17591	422.498114	422.498114	422.498114	0	0	1	-360	360;
	2634	6309	0.00038	0.002169	0.00301	320.026213	320.026213	320.026213	0	0	1	-360	360;
	1650	2634	0.000479	0.00212	0.00394	314.030753	314.030753	314.030753	0	0	1	-360	360;
	2634	360	0.00019	0.00086	0.0016	263.996829	263.996829	263.996829	0	0	1	-360	360;
	692	126	0.00089	0.004901	0.0083	366.341459	366.341459	366.341459	0	0	1	-360	360;
	3526	978	0.00109	0.03151	0.95693	1823.39626	1823.39626	1823.39626	0	0	1	-360	360;
	510	5635	0.003409	0.020271	0.04168	540.98293	540.98293	540.98293	0	0	1	-360	360;
	510	5635	0.002909	0.02388	0.03564	480.818378	480.818378	480.818378	0	0	1	-360	360;
	5635	3487	0.00545	0.042969	0.06675	496.063093	496.063093	496.063093	0	0	1	-360	360;
	5635	3487	0.005341	0.0438	0.06536	488.474839	488.474839	488.474839	0	0	1	-360	360;
	4974	6796	0.001521	0.00863	0.05172	339.479256	339.479256	339.479256	0	0	1	-360	360;
	2907	4974	0.00057	0.005979	0.00852	395.763183	395.763183	395.763183	0	0	1	-360	360;
	8276	5217	0.00085	0.01449	0.40056	1752.92895	1752.92895	1752.92895	0	0	1	-360	360;
	797	5217	0.0006	0.00777	0.28066	1547.73958	1547.73958	1547.73958	0	0	1	-360	360;
	4482	5217	0.0003	0.00492	0.14491	1131.76433	1131.76433	1131.76433	0	0	1	-360	360;
	2434	2389	0.003349	0.02686	0.04071	647.787002	647.787002	647.787002	0	0	1	-360	360;
	2026	2389	0.00213	0.016849	0.02652	609.034097	609.034097	609.034097	0	0	1	-360	360;
	3468	8382	0.00077	0.0106	0.20935	1434.03844	1434.03844	1434.03844	0	0	1	-360	360;
	7466	8382	0.00155	0.0148	0.41791	1322.39873	1322.39873	1322.39873	0	0	1	-360	360;
	1470	510	0.016651	0.10632	0.21934	508.04639	508.04639	508.04639	0	0	1	-360	360;
	1470	510	0.014841	0.12049	0.19541	459.328235	459.328235	459.328235	0	0	1	-360	360;
	4240	510	9.9e-05	0.001229	0.00352	541.186792	541.186792	541.186792	0	0	1	-360	360;
	7025	510	9.9e-05	0.00114	0.00345	535.104834	535.104834	535.104834	0	0	1	-360	360;
	7492	2002	0.00048	0.00781	0.22253	1824.77251	1824.77251	1824.77251	0	0	1	-360	360;
	7492	3036	0.00076	0.01058	0.35945	1979.68357	1979.68357	1979.68357	0	0	1	-360	360;
	7787	7492	0.00073	0.01181	0.34268	979.775589	979.775589	979.775589	0	0	1	-360	360;
	6558	751	0.00059	0.00929	0.27667	1736.95165	1736.95165	1736.95165	0	0	1	-360	360;
	133	751	8e-05	0.00092	0.02228	1142.336	1142.336	1142.336	0	0	1	-360	360;
	268	942	0.00019	0.000849	0.00154	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2165	3597	0.003479	0.01538	0.02885	281.334276	281.334276	281.334276	0	0	1	-360	360;
	2165	3597	0.003539	0.01505	0.02939	289.180882	289.180882	289.180882	0	0	1	-360	360;
	2165	1336	0.01387	0.084099	0.10971	393.940996	393.940996	393.940996	0	0	1	-360	360;
	996	2165	0.00375	0.021341	0.02891	394.898577	394.898577	394.898577	0	0	1	-360	360;
	7590	4881	0.00023	0.00345	0.11002	1805.24828	1805.24828	1805.24828	0	0	1	-360	360;
	4881	1536	0.00054	0.00662	0.25038	1644.85591	1644.85591	1644.85591	0	0	1	-360	360;
	366	7774	7e-05	0.00099	0.02898	1231.41399	1231.41399	1231.41399	0	0	1	-360	360;
	5001	366	0.00012	0.002	0.05674	1769.70857	1769.70857	1769.70857	0	0	1	-360	360;
	208	3536	0.00075	0.00888	0.35551	1201.97824	1201.97824	1201.97824	0	0	1	-360	360;
	2818	3536	0.00054	0.00705	0.1472	1381.53021	1381.53021	1381.53021	0	0	1	-360	360;
	6226	3767	0.00574	0.0287	0.04472	503.921892	503.921892	503.921892	0	0	1	-360	360;
	6226	5979	0.001331	0.009479	0.04018	274.702461	274.702461	274.702461	0	0	1	-360	360;
	5979	6226	0.00539	0.024159	0.04882	416.512562	416.512562	416.512562	0	0	1	-360	360;
	940	6226	0.000539	0.0028	0.00795	451.052184	451.052184	451.052184	0	0	1	-360	360;
	7002	7269	0.000969	0.00724	0.09464	393.675022	393.675022	393.675022	0	0	1	-360	360;
	6029	7269	0.000349	0.00339	0.00514	468.072216	468.072216	468.072216	0	0	1	-360	360;
	2433	7590	0.00021	0.00349	0.10119	1962.95396	1962.95396	1962.95396	0	0	1	-360	360;
	1209	42	0.00318	0.020669	0.06141	403.173866	403.173866	403.173866	0	0	1	-360	360;
	1209	42	0.003599	0.02026	0.06959	425.485174	425.485174	425.485174	0	0	1	-360	360;
	42	1122	0.0053	0.035521	0.10017	382.366566	382.366566	382.366566	0	0	1	-360	360;
	42	2539	0.00486	0.024331	0.04752	426.884013	426.884013	426.884013	0	0	1	-360	360;
	6177	4160	0.000659	0.00645	0.00977	445.273924	445.273924	445.273924	0	0	1	-360	360;
	4953	3248	0.00239	0.01531	0.06831	258.454439	258.454439	258.454439	0	0	1	-360	360;
	5322	1432	0.00119	0.01215	0.40139	1708.67626	1708.67626	1708.67626	0	0	1	-360	360;
	1432	5626	0.00012	0.0015	0.06087	1926.23414	1926.23414	1926.23414	0	0	1	-360	360;
	1432	5862	0.00047	0.00738	0.22496	1434.10821	1434.10821	1434.10821	0	0	1	-360	360;
	7126	5406	0.00068	0.00189	0.25283	530.246432	530.246432	530.246432	0	0	1	-360	360;
	2176	8126	0.003669	0.022099	0.02912	491.890582	491.890582	491.890582	0	0	1	-360	360;
	7873	8126	0.004771	0.0218	0.03728	458.497543	458.497543	458.497543	0	0	1	-360	360;
	2638	6932	0.001469	0.006591	0.04685	549.721195	549.721195	549.721195	0	0	1	-360	360;
	3039	461	8.1e-05	0.00045	0.00069	269.765943	269.765943	269.765943	0	0	1	-360	360;
	461	425	0.001021	0.00782	0.01187	487.827814	487.827814	487.827814	0	0	1	-360	360;
	3677	461	0.0015	0.009591	0.01754	528.12893	528.12893	528.12893	0	0	1	-360	360;
	5585	755	0.00019	0.001581	0.2574	528.278683	528.278683	528.278683	0	0	1	-360	360;
	5585	7002	3.9e-05	0.000479	0.06686	612.120612	612.120612	612.120612	0	0	1	-360	360;
	6628	4964	0.00027	0.00281	0.07485	1291.06807	1291.06807	1291.06807	0	0	1	-360	360;
	4969	6628	0.00096	0.01609	0.43218	1802.84198	1802.84198	1802.84198	0	0	1	-360	360;
	6628	7787	0.00027	0.00339	0.12538	973.727579	973.727579	973.727579	0	0	1	-360	360;
	6628	7787	0.00023	0.00398	0.10693	1578.45417	1578.45417	1578.45417	0	0	1	-360	360;
	1202	6838	0.000901	0.00937	0.02245	588.378075	588.378075	588.378075	0	0	1	-360	360;
	6840	6838	0.001901	0.01549	0.02295	449.769185	449.769185	449.769185	0	0	1	-360	360;
	7542	6838	0.003591	0.02205	0.03251	489.677818	489.677818	489.677818	0	0	1	-360	360;
	6838	5177	0.0013	0.009901	0.01574	523.09372	523.09372	523.09372	0	0	1	-360	360;
	2419	2790	0.00022	0.00299	0.1103	2101.26574	2101.26574	2101.26574	0	0	1	-360	360;
	2790	7832	0.00047	0.00564	0.22482	1858.56804	1858.56804	1858.56804	0	0	1	-360	360;
	5	2790	0.00046	0.0075	0.22009	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	5717	642	0.00214	0.020979	0.03146	443.912047	443.912047	443.912047	0	0	1	-360	360;
	2610	642	0.004651	0.03332	0.06904	466.599072	466.599072	466.599072	0	0	1	-360	360;
	642	4052	0.00468	0.021281	0.03738	500.408219	500.408219	500.408219	0	0	1	-360	360;
	642	4378	0.00581	0.028331	0.04634	268.228021	268.228021	268.228021	0	0	1	-360	360;
	1609	6526	7e-05	0.00086	0.08756	1615.08514	1615.08514	1615.08514	0	0	1	-360	360;
	2387	4049	0.0002	0.00279	0.09258	1375.54736	1375.54736	1375.54736	0	0	1	-360	360;
	2387	4049	0.00017	0.00333	0.07644	1350.84326	1350.84326	1350.84326	0	0	1	-360	360;
	5166	5322	1e-05	0.00036	0.17353	1453.79896	1453.79896	1453.79896	0	0	1	-360	360;
	4712	6013	0.000401	0.004081	0.00588	462.014646	462.014646	462.014646	0	0	1	-360	360;
	6013	6500	0.001581	0.015719	0.0234	479.019816	479.019816	479.019816	0	0	1	-360	360;
	7011	7113	0.00012	0.000229	0.02988	285.161554	285.161554	285.161554	0	0	1	-360	360;
	7011	7113	0.00012	0.00024	0.02927	264.761217	264.761217	264.761217	0	0	1	-360	360;
	7113	7725	0.00132	0.005729	0.01972	456.818251	456.818251	456.818251	0	0	1	-360	360;
	7960	7113	0.00438	0.024901	0.06154	426.128008	426.128008	426.128008	0	0	1	-360	360;
	5111	93	0.005521	0.023711	0.04303	436.605391	436.605391	436.605391	0	0	1	-360	360;
	208	6672	0.0008	0.01014	0.28011	1310.49818	1310.49818	1310.49818	0	0	1	-360	360;
	208	6672	0.00068	0.01229	0.2374	1366.62114	1366.62114	1366.62114	0	0	1	-360	360;
	2434	1109	0.001901	0.01163	0.02246	505.100483	505.100483	505.100483	0	0	1	-360	360;
	1650	493	0.00101	0.00824	0.01219	226.64924	226.64924	226.64924	0	0	1	-360	360;
	7972	5232	0.00787	0.068219	0.10596	470.691943	470.691943	470.691943	0	0	1	-360	360;
	7972	5232	0.007711	0.069539	0.10375	487.643005	487.643005	487.643005	0	0	1	-360	360;
	1470	5444	0.00032	0.002091	0.00392	290.000523	290.000523	290.000523	0	0	1	-360	360;
	2541	4194	0.00012	0.00182	0.05714	1984.8924	1984.8924	1984.8924	0	0	1	-360	360;
	4194	5	0.00058	0.00736	0.28135	1415.73974	1415.73974	1415.73974	0	0	1	-360	360;
	2541	1609	0.00074	0.01143	0.35868	1484.26015	1484.26015	1484.26015	0	0	1	-360	360;
	6558	5063	0.00084	0.01378	0.39904	1852.0106	1852.0106	1852.0106	0	0	1	-360	360;
	6558	6513	0.00037	0.00356	0.10346	1602.75876	1602.75876	1602.75876	0	0	1	-360	360;
	6448	6558	0.00012	0.00206	0.05975	1487.34899	1487.34899	1487.34899	0	0	1	-360	360;
	6789	5716	0.00036	0.00151	0.00296	365.184969	365.184969	365.184969	0	0	1	-360	360;
	8276	1378	0.0012	0.01835	0.574	987.093885	987.093885	987.093885	0	0	1	-360	360;
	1378	5615	0.00088	0.01503	0.42549	1651.957	1651.957	1651.957	0	0	1	-360	360;
	1968	1378	0.00057	0.00939	0.27394	1774.0282	1774.0282	1774.0282	0	0	1	-360	360;
	3043	3361	0.00157	0.02006	0.12828	1757.73431	1757.73431	1757.73431	0	0	1	-360	360;
	3043	5065	0.00027	0.00349	0.13073	1816.62096	1816.62096	1816.62096	0	0	1	-360	360;
	6953	3043	0.00136	0.02325	0.65809	2020.26229	2020.26229	2020.26229	0	0	1	-360	360;
	3043	3547	0.00038	0.00478	0.18647	1756.20273	1756.20273	1756.20273	0	0	1	-360	360;
	3043	5846	0.00077	0.01171	0.36606	1620.71586	1620.71586	1620.71586	0	0	1	-360	360;
	8276	4969	0.00129	0.02076	0.59396	1215.66574	1215.66574	1215.66574	0	0	1	-360	360;
	8276	1609	0.00037	0.0046	0.17928	1302.49341	1302.49341	1302.49341	0	0	1	-360	360;
	1609	8276	0.00021	0.00375	0.10385	1499.0738	1499.0738	1499.0738	0	0	1	-360	360;
	8276	1968	0.00055	0.00885	0.26285	1385.92817	1385.92817	1385.92817	0	0	1	-360	360;
	6177	1650	0.001659	0.01138	0.01701	464.918256	464.918256	464.918256	0	0	1	-360	360;
	6177	4461	0.001169	0.010021	0.25763	467.643534	467.643534	467.643534	0	0	1	-360	360;
	6177	1295	0.002229	0.00975	0.02234	535.867317	535.867317	535.867317	0	0	1	-360	360;
	6177	6462	0.00049	0.004521	0.00728	446.617892	446.617892	446.617892	0	0	1	-360	360;
	6177	6029	0.0002	0.001969	0.00303	596.400345	596.400345	596.400345	0	0	1	-360	360;
	2714	4399	0.0001	0.00158	0.04849	1691.69853	1691.69853	1691.69853	0	0	1	-360	360;
	2714	4399	0.00011	0.00151	0.05064	1552.15267	1552.15267	1552.15267	0	0	1	-360	360;
	1085	4497	3e-05	0.00043	0.01664	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1085	621	0.00037	0.00589	0.1739	1848.08851	1848.08851	1848.08851	0	0	1	-360	360;
	1862	1085	0.00087	0.01218	0.41371	1462.29079	1462.29079	1462.29079	0	0	1	-360	360;
	6147	5232	0.00362	0.024409	0.04396	474.807296	474.807296	474.807296	0	0	1	-360	360;
	6147	3487	0.00488	0.03119	0.06434	450.004675	450.004675	450.004675	0	0	1	-360	360;
	1209	2104	0.00055	0.00264	0.29916	542.107412	542.107412	542.107412	0	0	1	-360	360;
	5543	2104	0.00031	0.001341	0.17141	542.903809	542.903809	542.903809	0	0	1	-360	360;
	5406	2104	0.00012	0.000669	0.06739	513.999553	513.999553	513.999553	0	0	1	-360	360;
	757	1540	3.1e-05	0.000469	0.0416	271.351497	271.351497	271.351497	0	0	1	-360	360;
	1818	1540	3.1e-05	0.000461	0.04257	277.353434	277.353434	277.353434	0	0	1	-360	360;
	1863	4098	0.00607	0.03457	0.04895	388.45805	388.45805	388.45805	0	0	1	-360	360;
	4098	4111	0.0087	0.047719	0.07619	462.434945	462.434945	462.434945	0	0	1	-360	360;
	3468	7466	0.00129	0.02159	0.59185	1399.39583	1399.39583	1399.39583	0	0	1	-360	360;
	3468	1112	0.00091	0.00903	0.24255	1673.12339	1673.12339	1673.12339	0	0	1	-360	360;
	7805	3468	0.0004	0.00518	0.19271	1758.42343	1758.42343	1758.42343	0	0	1	-360	360;
	3468	6110	0.00032	0.00557	0.15545	1985.01943	1985.01943	1985.01943	0	0	1	-360	360;
	6796	4291	3.1e-05	0.0002	0.0003	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6796	5617	3.1e-05	0.000211	0.00029	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4945	2328	0.00199	0.011591	0.01528	471.913974	471.913974	471.913974	0	0	1	-360	360;
	3364	4945	0.000211	0.000919	0.00194	353.292361	353.292361	353.292361	0	0	1	-360	360;
	3620	2226	0.00109	0.01525	0.5183	1578.29555	1578.29555	1578.29555	0	0	1	-360	360;
	678	2328	0.005789	0.02826	0.04643	449.369462	449.369462	449.369462	0	0	1	-360	360;
	5922	2613	0.00041	0.0064	0.18864	1848.05034	1848.05034	1848.05034	0	0	1	-360	360;
	5922	2433	0.00051	0.00603	0.17487	1679.29185	1679.29185	1679.29185	0	0	1	-360	360;
	3239	3900	0.00245	0.02726	0.08107	526.953006	526.953006	526.953006	0	0	1	-360	360;
	3239	6500	0.000271	0.00337	0.07711	438.970195	438.970195	438.970195	0	0	1	-360	360;
	3306	6253	0.000151	0.00064	0.00118	311.236124	311.236124	311.236124	0	0	1	-360	360;
	2351	3306	0.003341	0.022091	0.0325	304.802075	304.802075	304.802075	0	0	1	-360	360;
	3306	1122	0.022	0.125771	0.21415	461.42897	461.42897	461.42897	0	0	1	-360	360;
	3306	2106	0.0153	0.08863	0.14731	471.675055	471.675055	471.675055	0	0	1	-360	360;
	3306	2106	0.015599	0.08682	0.1502	492.257534	492.257534	492.257534	0	0	1	-360	360;
	639	1857	0.001169	0.00844	0.0315	220.406098	220.406098	220.406098	0	0	1	-360	360;
	5716	639	0.00081	0.004271	0.00851	375.271012	375.271012	375.271012	0	0	1	-360	360;
	639	5364	0.000349	0.0015	0.19046	353.49127	353.49127	353.49127	0	0	1	-360	360;
	3361	5846	0.00058	0.00717	0.27971	1943.82003	1943.82003	1943.82003	0	0	1	-360	360;
	3361	5101	0.0007	0.01203	0.34039	1813.8441	1813.8441	1813.8441	0	0	1	-360	360;
	5101	3361	0.00061	0.00957	0.3038	1918.8474	1918.8474	1918.8474	0	0	1	-360	360;
	7857	2346	2e-05	0.00075	0.02192	1522.90252	1522.90252	1522.90252	0	0	1	-360	360;
	6589	7857	0.00041	0.01221	0.36552	1974.9993	1974.9993	1974.9993	0	0	1	-360	360;
	942	1918	0.00157	0.007969	0.04235	267.079533	267.079533	267.079533	0	0	1	-360	360;
	3864	1918	0.00151	0.00625	0.01108	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4037	4040	2e-05	0.00023	0.22918	1222.54437	1222.54437	1222.54437	0	0	1	-360	360;
	4037	4040	2e-05	0.00022	0.23342	1186.40769	1186.40769	1186.40769	0	0	1	-360	360;
	3526	4037	0.00026	0.00395	0.11957	1637.19141	1637.19141	1637.19141	0	0	1	-360	360;
	3526	4037	0.00025	0.00403	0.11708	1727.97737	1727.97737	1727.97737	0	0	1	-360	360;
	5392	216	0.00111	0.005099	0.03261	459.778637	459.778637	459.778637	0	0	1	-360	360;
	216	4287	0.000729	0.0023	0.35453	479.571579	479.571579	479.571579	0	0	1	-360	360;
	4287	7291	0.003831	0.016271	0.11288	441.822744	441.822744	441.822744	0	0	1	-360	360;
	6782	8288	0.001849	0.014719	0.02237	518.119478	518.119478	518.119478	0	0	1	-360	360;
	6710	6782	0.00157	0.010169	0.01889	465.878123	465.878123	465.878123	0	0	1	-360	360;
	478	5717	0.00149	0.01513	0.02181	259.533195	259.533195	259.533195	0	0	1	-360	360;
	3781	478	0.003229	0.02918	0.04742	281.021814	281.021814	281.021814	0	0	1	-360	360;
	7466	1915	1e-05	0.00021	0.00579	1731.05831	1731.05831	1731.05831	0	0	1	-360	360;
	2092	2002	0.00054	0.00765	0.2522	1558.70418	1558.70418	1558.70418	0	0	1	-360	360;
	3678	5065	1e-05	0.0002	0.00591	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5065	5846	0.00053	0.00746	0.25789	1657.43042	1657.43042	1657.43042	0	0	1	-360	360;
	2944	3544	0.010469	0.057979	0.10448	324.763441	324.763441	324.763441	0	0	1	-360	360;
	2944	2402	0.00014	0.000781	0.00146	373.343274	373.343274	373.343274	0	0	1	-360	360;
	2944	808	0.00181	0.013039	0.01802	360.199295	360.199295	360.199295	0	0	1	-360	360;
	3767	7960	0.01107	0.04905	0.08621	422.292728	422.292728	422.292728	0	0	1	-360	360;
	2371	1468	0.00038	0.00468	0.18993	1619.73057	1619.73057	1619.73057	0	0	1	-360	360;
	2371	1468	0.00034	0.0053	0.16921	1440.17991	1440.17991	1440.17991	0	0	1	-360	360;
	3036	2371	0.00012	0.00185	0.0555	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3036	2371	0.00013	0.00181	0.0629	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4808	7033	0.000211	0.0018	0.00255	564.274683	564.274683	564.274683	0	0	1	-360	360;
	1346	3075	0.00037	0.00225	0.00456	572.895204	572.895204	572.895204	0	0	1	-360	360;
	2506	7161	0.00025	0.00317	0.12312	1920.88841	1920.88841	1920.88841	0	0	1	-360	360;
	1536	2506	0.00029	0.00484	0.13605	1887.04483	1887.04483	1887.04483	0	0	1	-360	360;
	88	6570	0.00057	0.00858	0.26443	1992.07379	1992.07379	1992.07379	0	0	1	-360	360;
	88	8181	0.0005	0.00665	0.23101	1918.80198	1918.80198	1918.80198	0	0	1	-360	360;
	7077	2750	0.00239	0.012289	0.02363	465.290924	465.290924	465.290924	0	0	1	-360	360;
	5954	2750	0.0008	0.005211	0.00788	389.685035	389.685035	389.685035	0	0	1	-360	360;
	5615	3661	0.00067	0.01081	0.31881	1605.32105	1605.32105	1605.32105	0	0	1	-360	360;
	2430	5615	0.00105	0.01455	0.50319	1748.11108	1748.11108	1748.11108	0	0	1	-360	360;
	439	6570	0.0002	0.00261	0.09805	819.0199	819.0199	819.0199	0	0	1	-360	360;
	5429	7898	1e-05	0.0002	0.00579	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1112	7898	0.00036	0.00601	0.1702	2142.25056	2142.25056	2142.25056	0	0	1	-360	360;
	8181	7898	0.0001	0.00151	0.04834	2084.53153	2084.53153	2084.53153	0	0	1	-360	360;
	4083	727	2e-05	0.00021	0.00486	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7466	727	0.00046	0.00568	0.12832	1253.35177	1253.35177	1253.35177	0	0	1	-360	360;
	597	727	0.00041	0.00531	0.11517	1289.04746	1289.04746	1289.04746	0	0	1	-360	360;
	2419	2092	0.0003	0.00482	0.1443	1880.2873	1880.2873	1880.2873	0	0	1	-360	360;
	1862	199	1e-05	0.0002	0.00567	1756.19549	1756.19549	1756.19549	0	0	1	-360	360;
	6953	1030	1e-05	0.00022	0.00563	1870.01839	1870.01839	1870.01839	0	0	1	-360	360;
	2457	7774	2e-05	0.00025	0.20796	791.282908	791.282908	791.282908	0	0	1	-360	360;
	2457	7774	2e-05	0.00026	0.20371	767.876077	767.876077	767.876077	0	0	1	-360	360;
	797	2887	0.00032	0.00528	0.15421	1722.22093	1722.22093	1722.22093	0	0	1	-360	360;
	2652	2887	0.0006	0.0084	0.28595	1887.96365	1887.96365	1887.96365	0	0	1	-360	360;
	5865	7375	0.00105	0.01735	0.50942	1544.95087	1544.95087	1544.95087	0	0	1	-360	360;
	3788	5865	0.00135	0.01758	0.65653	1483.5875	1483.5875	1483.5875	0	0	1	-360	360;
	6570	7375	0.0004	0.0068	0.19289	1823.63452	1823.63452	1823.63452	0	0	1	-360	360;
	6570	7375	0.00039	0.0063	0.18869	1714.24643	1714.24643	1714.24643	0	0	1	-360	360;
	7375	8181	0.00067	0.00834	0.32443	1427.77323	1427.77323	1427.77323	0	0	1	-360	360;
	6960	621	0.00084	0.01114	0.40082	1893.69178	1893.69178	1893.69178	0	0	1	-360	360;
	1862	621	0.00081	0.011	0.38053	1628.589	1628.589	1628.589	0	0	1	-360	360;
	2842	2874	0.00059	0.00726	0.27888	1454.73489	1454.73489	1454.73489	0	0	1	-360	360;
	3788	2874	0.001	0.01178	0.47118	1754.48488	1754.48488	1754.48488	0	0	1	-360	360;
	4040	2130	0.00031	0.00197	3.72164	1218.73878	1218.73878	1218.73878	0	0	1	-360	360;
	4040	2130	0.00035	0.00193	4.21786	1254.24755	1254.24755	1254.24755	0	0	1	-360	360;
	5812	2130	2e-05	0.0004	0.29919	965.606963	965.606963	965.606963	0	0	1	-360	360;
	5812	2130	2e-05	0.00041	0.29283	1021.07305	1021.07305	1021.07305	0	0	1	-360	360;
	4432	4153	3.1e-05	0.000271	0.03207	469.163928	469.163928	469.163928	0	0	1	-360	360;
	4153	4432	3.1e-05	0.000281	0.03091	470.743004	470.743004	470.743004	0	0	1	-360	360;
	4153	2935	0.000229	0.000271	0.02886	525.800707	525.800707	525.800707	0	0	1	-360	360;
	2034	5309	0.000289	0.001229	0.00234	216.486225	216.486225	216.486225	0	0	1	-360	360;
	2034	7960	0.002479	0.01106	0.01984	221.937924	221.937924	221.937924	0	0	1	-360	360;
	7727	678	0.00738	0.04069	0.05763	436.740283	436.740283	436.740283	0	0	1	-360	360;
	6294	5948	0.00058	0.00872	0.27698	1775.29256	1775.29256	1775.29256	0	0	1	-360	360;
	5948	5322	0.00116	0.01483	0.55559	1930.2701	1930.2701	1930.2701	0	0	1	-360	360;
	5948	5322	0.00118	0.01452	0.56607	1969.15533	1969.15533	1969.15533	0	0	1	-360	360;
	6570	2842	0.00139	0.02265	0.66615	1965.24904	1965.24904	1965.24904	0	0	1	-360	360;
	6570	2842	0.00149	0.02349	0.71021	1956.48735	1956.48735	1956.48735	0	0	1	-360	360;
	3682	2842	0.00086	0.01488	0.41499	1882.41322	1882.41322	1882.41322	0	0	1	-360	360;
	3682	2842	0.00095	0.01554	0.4511	1955.02422	1955.02422	1955.02422	0	0	1	-360	360;
	308	688	0.019651	0.09944	0.19161	465.171274	465.171274	465.171274	0	0	1	-360	360;
	6414	1154	0.0002	0.001831	0.28314	489.93236	489.93236	489.93236	0	0	1	-360	360;
	8038	2555	7e-05	0.00062	0.00093	461.105076	461.105076	461.105076	0	0	1	-360	360;
	2555	10	0.004039	0.023081	0.03444	474.606863	474.606863	474.606863	0	0	1	-360	360;
	4111	2555	0.0132	0.064331	0.10877	331.639128	331.639128	331.639128	0	0	1	-360	360;
	2555	8185	0.001401	0.008729	0.01075	315.65708	315.65708	315.65708	0	0	1	-360	360;
	5846	8285	0.00021	0.00247	0.09805	1139.82834	1139.82834	1139.82834	0	0	1	-360	360;
	5846	5101	0.00138	0.01791	0.65297	1811.36342	1811.36342	1811.36342	0	0	1	-360	360;
	7466	597	0.00056	0.00795	0.26896	1760.42166	1760.42166	1760.42166	0	0	1	-360	360;
	7466	4841	0.00081	0.00788	0.21886	1163.53135	1163.53135	1163.53135	0	0	1	-360	360;
	6914	6294	3e-05	0.00042	0.00901	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1597	6294	3e-05	0.00042	0.00883	859.617058	859.617058	859.617058	0	0	1	-360	360;
	7770	6294	3e-05	0.00035	0.00749	846.033554	846.033554	846.033554	0	0	1	-360	360;
	6149	6294	3e-05	0.00036	0.00733	779.859548	779.859548	779.859548	0	0	1	-360	360;
	7161	6294	0.00035	0.00423	0.16401	977.892538	977.892538	977.892538	0	0	1	-360	360;
	2818	4647	0.00076	0.00904	0.35756	1335.0305	1335.0305	1335.0305	0	0	1	-360	360;
	4647	4253	0.00029	0.00361	0.1378	1397.58452	1397.58452	1397.58452	0	0	1	-360	360;
	6443	5510	0.010349	0.056729	0.08363	478.505397	478.505397	478.505397	0	0	1	-360	360;
	6443	1857	0.001581	0.012771	0.04044	549.094747	549.094747	549.094747	0	0	1	-360	360;
	6443	942	0.001081	0.01136	0.02755	325.966419	325.966419	325.966419	0	0	1	-360	360;
	6443	3864	0.001669	0.00914	0.03423	456.348796	456.348796	456.348796	0	0	1	-360	360;
	6443	5153	0.01757	0.07575	0.14199	524.256688	524.256688	524.256688	0	0	1	-360	360;
	741	6513	0.00037	0.00558	0.17602	2297.7994	2297.7994	2297.7994	0	0	1	-360	360;
	741	7832	0.00041	0.0072	0.20119	1426.54507	1426.54507	1426.54507	0	0	1	-360	360;
	7064	4161	0.002479	0.017961	0.02974	431.212755	431.212755	431.212755	0	0	1	-360	360;
	1099	7064	0.00155	0.0127	0.01877	453.350304	453.350304	453.350304	0	0	1	-360	360;
	126	7107	0.001961	0.010419	0.0157	472.921092	472.921092	472.921092	0	0	1	-360	360;
	1947	1857	0.00138	0.007521	0.01219	307.314726	307.314726	307.314726	0	0	1	-360	360;
	8087	1947	0.00281	0.02543	0.08305	327.707442	327.707442	327.707442	0	0	1	-360	360;
	1947	7831	0.000211	0.000901	0.11428	500.756881	500.756881	500.756881	0	0	1	-360	360;
	6840	3473	0.000789	0.00519	0.00984	336.265852	336.265852	336.265852	0	0	1	-360	360;
	6840	91	0.00056	0.004281	0.00692	317.794396	317.794396	317.794396	0	0	1	-360	360;
	6840	7542	0.00163	0.012031	0.02414	602.762376	602.762376	602.762376	0	0	1	-360	360;
	3781	6840	0.000211	0.00161	0.00317	513.413115	513.413115	513.413115	0	0	1	-360	360;
	6840	5177	0.00018	0.00144	0.00212	479.810879	479.810879	479.810879	0	0	1	-360	360;
	6637	1650	9.1e-05	0.00081	0.003	268.28213	268.28213	268.28213	0	0	1	-360	360;
	7690	1650	9.1e-05	0.00099	0.00259	375.445153	375.445153	375.445153	0	0	1	-360	360;
	3030	4630	0.001219	0.01461	0.03999	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3030	2003	0.00769	0.035169	0.23743	647.787002	647.787002	647.787002	0	0	1	-360	360;
	6913	7476	5e-05	0.000219	0.02513	453.579697	453.579697	453.579697	0	0	1	-360	360;
	7476	5364	0.000281	0.00112	0.15427	444.791894	444.791894	444.791894	0	0	1	-360	360;
	5216	3900	0.003591	0.01825	0.03656	448.06017	448.06017	448.06017	0	0	1	-360	360;
	5216	3900	0.00125	0.01386	0.0417	472.581195	472.581195	472.581195	0	0	1	-360	360;
	308	3900	0.000961	0.00513	0.00989	530.872499	530.872499	530.872499	0	0	1	-360	360;
	1468	3989	0.00022	0.00295	0.06316	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5144	7637	0.00326	0.02562	0.03836	421.708576	421.708576	421.708576	0	0	1	-360	360;
	5144	6889	0.00164	0.013219	0.01969	585.57773	585.57773	585.57773	0	0	1	-360	360;
	3797	753	8.1e-05	0.00019	0.02282	346.293976	346.293976	346.293976	0	0	1	-360	360;
	3797	753	3.9e-05	0.00032	0.02349	396.518045	396.518045	396.518045	0	0	1	-360	360;
	3797	753	0.00011	0.0002	0.02538	327.707823	327.707823	327.707823	0	0	1	-360	360;
	3797	753	9.1e-05	0.000151	0.02223	350.43524	350.43524	350.43524	0	0	1	-360	360;
	3797	753	0.00011	0.00019	0.02586	341.962567	341.962567	341.962567	0	0	1	-360	360;
	3797	3473	8.1e-05	0.001	0.0514	453.897493	453.897493	453.897493	0	0	1	-360	360;
	3797	91	0.000281	0.00213	0.06809	428.661617	428.661617	428.661617	0	0	1	-360	360;
	3781	3797	0.00025	0.000831	0.15806	462.311103	462.311103	462.311103	0	0	1	-360	360;
	7501	3797	0.00013	0.000469	0.08448	432.021726	432.021726	432.021726	0	0	1	-360	360;
	2818	208	0.00088	0.01144	0.41827	1285.86977	1285.86977	1285.86977	0	0	1	-360	360;
	3942	208	0.00076	0.01445	0.32748	1025.27552	1025.27552	1025.27552	0	0	1	-360	360;
	1990	6960	0.00048	0.0076	0.23009	1808.4161	1808.4161	1808.4161	0	0	1	-360	360;
	6960	2146	0.00017	0.00273	0.07982	1879.37375	1879.37375	1879.37375	0	0	1	-360	360;
	8110	6863	0.00014	0.000849	0.00114	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2176	8110	0.0022	0.01289	0.01745	422.760659	422.760659	422.760659	0	0	1	-360	360;
	3233	3464	0.00582	0.03749	0.07206	487.167072	487.167072	487.167072	0	0	1	-360	360;
	3233	3464	0.00593	0.03669	0.07342	490.907089	490.907089	490.907089	0	0	1	-360	360;
	3464	5272	0.000281	0.003419	0.00945	645.076966	645.076966	645.076966	0	0	1	-360	360;
	3464	3723	0.00645	0.03995	0.0789	481.605249	481.605249	481.605249	0	0	1	-360	360;
	3464	3723	0.00657	0.039081	0.08037	459.212395	459.212395	459.212395	0	0	1	-360	360;
	7611	8050	8e-05	0.00125	0.03535	1749.53472	1749.53472	1749.53472	0	0	1	-360	360;
	7611	7929	8e-05	0.00127	0.03449	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7611	4192	7e-05	0.0013	0.03387	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7611	1354	7e-05	0.0012	0.03314	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8057	5063	5e-05	0.0006	0.62678	1521.50718	1521.50718	1521.50718	0	0	1	-360	360;
	5063	5	0.00034	0.00466	0.16464	2128.67956	2128.67956	2128.67956	0	0	1	-360	360;
	5979	5716	0.000591	0.00588	0.0155	510.378804	510.378804	510.378804	0	0	1	-360	360;
	5716	5979	0.00139	0.00818	0.01071	505.532214	505.532214	505.532214	0	0	1	-360	360;
	3864	5979	0.00169	0.013969	0.05121	574.066174	574.066174	574.066174	0	0	1	-360	360;
	3607	305	0.000229	0.00163	0.00283	450.055735	450.055735	450.055735	0	0	1	-360	360;
	3963	7725	0.00275	0.02836	0.08626	611.099776	611.099776	611.099776	0	0	1	-360	360;
	46	7725	0.004909	0.02468	0.08814	440.698262	440.698262	440.698262	0	0	1	-360	360;
	7725	7960	0.001419	0.02212	0.06483	420.155793	420.155793	420.155793	0	0	1	-360	360;
	4150	7611	0.0016	0.02118	0.74277	1924.34385	1924.34385	1924.34385	0	0	1	-360	360;
	7611	5730	0.00116	0.01869	0.52636	1576.787	1576.787	1576.787	0	0	1	-360	360;
	7396	7611	0.00169	0.02149	0.77557	1422.1458	1422.1458	1422.1458	0	0	1	-360	360;
	7611	4841	0.00171	0.02913	0.79419	862.275445	862.275445	862.275445	0	0	1	-360	360;
	7611	939	0.00138	0.0212	0.67168	1657.64565	1657.64565	1657.64565	0	0	1	-360	360;
	2252	1949	0.00511	0.02775	0.03978	496.254761	496.254761	496.254761	0	0	1	-360	360;
	3620	3788	0.0008	0.01262	0.38137	2014.60261	2014.60261	2014.60261	0	0	1	-360	360;
	4663	337	0.0005	0.00851	0.23944	1746.69731	1746.69731	1746.69731	0	0	1	-360	360;
	4663	2458	0.00032	0.00558	0.15828	1847.89764	1847.89764	1847.89764	0	0	1	-360	360;
	2442	5001	0.00039	0.00601	0.18896	1901.39182	1901.39182	1901.39182	0	0	1	-360	360;
	2442	4108	0.00036	0.00575	0.17204	1528.91828	1528.91828	1528.91828	0	0	1	-360	360;
	2442	7358	0.00117	0.01616	0.55541	1816.83092	1816.83092	1816.83092	0	0	1	-360	360;
	797	5590	0.0003	0.00471	0.1467	1453.87136	1453.87136	1453.87136	0	0	1	-360	360;
	3233	675	0.00011	0.00081	0.00141	481.117122	481.117122	481.117122	0	0	1	-360	360;
	3233	1470	0.01087	0.06457	0.13248	467.209897	467.209897	467.209897	0	0	1	-360	360;
	3233	1470	0.00968	0.073169	0.11803	465.22081	465.22081	465.22081	0	0	1	-360	360;
	1857	6913	0.00105	0.005781	0.00853	324.313419	324.313419	324.313419	0	0	1	-360	360;
	1712	2610	0.002961	0.029021	0.04381	448.665661	448.665661	448.665661	0	0	1	-360	360;
	3776	337	4e-05	0.00033	0.02775	1354.35728	1354.35728	1354.35728	0	0	1	-360	360;
	2119	337	4e-05	0.00035	0.02842	1423.53916	1423.53916	1423.53916	0	0	1	-360	360;
	4659	337	4e-05	0.0003	0.03169	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3040	337	4e-05	0.00028	0.03202	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4461	77	0.000211	0.00045	0.08014	347.595646	347.595646	347.595646	0	0	1	-360	360;
	8346	3385	0.00024	0.001099	0.00194	396.650651	396.650651	396.650651	0	0	1	-360	360;
	3763	8346	0.00544	0.029831	0.04403	394.631461	394.631461	394.631461	0	0	1	-360	360;
	2387	6513	0.00076	0.0101	0.33997	1945.503	1945.503	1945.503	0	0	1	-360	360;
	1863	5298	0.01106	0.06313	0.09	395.58104	395.58104	395.58104	0	0	1	-360	360;
	5168	5216	0.001909	0.011401	0.01939	478.056519	478.056519	478.056519	0	0	1	-360	360;
	5954	5168	0.00113	0.005479	0.01098	419.321672	419.321672	419.321672	0	0	1	-360	360;
	1209	3370	0.000151	0.000961	0.00189	497.846412	497.846412	497.846412	0	0	1	-360	360;
	1209	2106	0.00156	0.01125	0.05791	404.232045	404.232045	404.232045	0	0	1	-360	360;
	1209	2106	0.00344	0.021599	0.03313	372.8742	372.8742	372.8742	0	0	1	-360	360;
	1191	4630	0.00106	0.01276	0.03487	294.756423	294.756423	294.756423	0	0	1	-360	360;
	3959	5320	0.001349	0.008831	0.01483	316.916073	316.916073	316.916073	0	0	1	-360	360;
	3385	8038	0.001901	0.010031	0.01539	396.538241	396.538241	396.538241	0	0	1	-360	360;
	2026	6710	0.001719	0.012159	0.02086	491.312527	491.312527	491.312527	0	0	1	-360	360;
	2908	94	4e-05	0.00061	0.01938	2019.34611	2019.34611	2019.34611	0	0	1	-360	360;
	5507	171	0.00126	0.00595	0.20657	346.438775	346.438775	346.438775	0	0	1	-360	360;
	2652	5	0.00018	0.00224	0.08234	1955.22628	1955.22628	1955.22628	0	0	1	-360	360;
	5543	1122	0.000521	0.002039	0.28085	539.497592	539.497592	539.497592	0	0	1	-360	360;
	7163	8038	0.002169	0.016479	0.02942	437.484857	437.484857	437.484857	0	0	1	-360	360;
	4150	5001	0.00073	0.01168	0.3472	1878.01987	1878.01987	1878.01987	0	0	1	-360	360;
	7396	5001	0.00131	0.0162	0.62165	1484.81763	1484.81763	1484.81763	0	0	1	-360	360;
	6796	2169	0.00112	0.007591	0.01356	441.605926	441.605926	441.605926	0	0	1	-360	360;
	6462	6796	0.00162	0.01332	0.02451	467.182081	467.182081	467.182081	0	0	1	-360	360;
	5641	7396	0.00027	0.00344	0.07754	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7264	7396	0.0003	0.00313	0.08579	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1666	1609	0.00014	0.00243	0.06637	1760.44865	1760.44865	1760.44865	0	0	1	-360	360;
	1670	4396	0.00019	0.001781	0.34769	427.639257	427.639257	427.639257	0	0	1	-360	360;
	2901	4396	0.000151	0.001969	0.288	463.643258	463.643258	463.643258	0	0	1	-360	360;
	2351	1122	0.025461	0.14561	0.24792	443.990544	443.990544	443.990544	0	0	1	-360	360;
	6889	7637	0.004599	0.028581	0.0626	489.087951	489.087951	489.087951	0	0	1	-360	360;
	3727	1540	0.00987	0.05706	0.09142	467.498734	467.498734	467.498734	0	0	1	-360	360;
	94	2335	9e-05	0.00103	0.80805	1236.23581	1236.23581	1236.23581	0	0	1	-360	360;
	94	2335	0.0001	0.00101	0.82359	1114.31336	1114.31336	1114.31336	0	0	1	-360	360;
	2458	2335	0.00023	0.00381	0.1096	2103.41075	2103.41075	2103.41075	0	0	1	-360	360;
	112	1857	0.002219	0.01374	0.06016	565.846138	565.846138	565.846138	0	0	1	-360	360;
	7820	2818	0.00092	0.00895	0.25474	1983.25221	1983.25221	1983.25221	0	0	1	-360	360;
	112	5716	0.001151	0.009479	0.02574	561.421753	561.421753	561.421753	0	0	1	-360	360;
	901	2003	0.00188	0.012151	0.05653	616.561001	616.561001	616.561001	0	0	1	-360	360;
	3479	1468	0.00036	0.00471	0.16969	1654.03685	1654.03685	1654.03685	0	0	1	-360	360;
	3479	857	0.00085	0.01337	0.40092	1992.46738	1992.46738	1992.46738	0	0	1	-360	360;
	2760	3479	9e-05	0.00105	0.04004	1493.87023	1493.87023	1493.87023	0	0	1	-360	360;
	3036	3479	0.00037	0.00634	0.17607	1542.87826	1542.87826	1542.87826	0	0	1	-360	360;
	3723	2711	0.00056	0.004979	0.00728	388.777752	388.777752	388.777752	0	0	1	-360	360;
	3487	5232	0.00488	0.029039	0.06032	449.906744	449.906744	449.906744	0	0	1	-360	360;
	3487	5232	0.004349	0.032919	0.05327	449.35041	449.35041	449.35041	0	0	1	-360	360;
	2387	2613	0.00046	0.00571	0.21214	1948.52141	1948.52141	1948.52141	0	0	1	-360	360;
	8288	5123	0.010841	0.04989	0.08723	438.216475	438.216475	438.216475	0	0	1	-360	360;
	2940	901	0.00282	0.01436	0.02737	406.612091	406.612091	406.612091	0	0	1	-360	360;
	2940	901	0.00282	0.014789	0.02703	479.732001	479.732001	479.732001	0	0	1	-360	360;
	233	2402	0.00012	0.000841	0.00116	432.717907	432.717907	432.717907	0	0	1	-360	360;
	233	808	0.00175	0.01189	0.01617	386.521167	386.521167	386.521167	0	0	1	-360	360;
	233	1540	0.00851	0.04389	0.12431	495.991836	495.991836	495.991836	0	0	1	-360	360;
	6057	6127	0.00063	0.00374	0.00487	274.109165	274.109165	274.109165	0	0	1	-360	360;
	3526	143	0.00037	0.00484	0.17841	994.112709	994.112709	994.112709	0	0	1	-360	360;
	7805	4841	0.00161	0.02178	0.76796	1800.48833	1800.48833	1800.48833	0	0	1	-360	360;
	7797	1295	0.00745	0.03588	0.07352	431.586185	431.586185	431.586185	0	0	1	-360	360;
	1923	4540	0.00051	0.00829	0.24332	1692.3442	1692.3442	1692.3442	0	0	1	-360	360;
	1923	7358	0.00124	0.01751	0.53074	1403.63714	1403.63714	1403.63714	0	0	1	-360	360;
	1923	7161	0.0008	0.01242	0.3789	1018.22577	1018.22577	1018.22577	0	0	1	-360	360;
	1609	4969	0.00118	0.01502	0.54028	1855.05863	1855.05863	1855.05863	0	0	1	-360	360;
	3526	6589	0.00131	0.01723	0.6325	1173.76209	1173.76209	1173.76209	0	0	1	-360	360;
	1154	5406	0.000271	0.001081	0.14365	466.864284	466.864284	466.864284	0	0	1	-360	360;
	5626	2714	0.00014	0.00178	0.06384	1219.04615	1219.04615	1219.04615	0	0	1	-360	360;
	1154	1122	0.00051	0.00201	0.27689	532.129205	532.129205	532.129205	0	0	1	-360	360;
	7161	2649	9e-05	0.00091	0.02515	1755.43924	1755.43924	1755.43924	0	0	1	-360	360;
	1772	425	9.1e-05	0.000591	0.00085	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2935	4561	2.1e-05	0.00025	0.029	425.901663	425.901663	425.901663	0	0	1	-360	360;
	1862	5812	0.00035	0.00556	0.16404	1739.34479	1739.34479	1739.34479	0	0	1	-360	360;
	1862	5812	0.0004	0.00545	0.18038	1787.77099	1787.77099	1787.77099	0	0	1	-360	360;
	5730	4114	0.0004	0.00517	0.10593	1027.2369	1027.2369	1027.2369	0	0	1	-360	360;
	4114	4841	0.00037	0.00456	0.09974	1241.48282	1241.48282	1241.48282	0	0	1	-360	360;
	4864	5730	0.00133	0.01413	0.37462	1418.10524	1418.10524	1418.10524	0	0	1	-360	360;
	6409	3959	0.01187	0.05524	0.09269	368.591185	368.591185	368.591185	0	0	1	-360	360;
	3959	2901	0.00063	0.0065	0.40514	450.05726	450.05726	450.05726	0	0	1	-360	360;
	1540	3959	0.011841	0.052289	0.09246	402.781383	402.781383	402.781383	0	0	1	-360	360;
	597	1234	0.00055	0.00517	0.15398	928.207898	928.207898	928.207898	0	0	1	-360	360;
	1234	4864	0.00097	0.01163	0.27287	1389.3349	1389.3349	1389.3349	0	0	1	-360	360;
	1670	5022	0.000969	0.01039	0.49162	503.399471	503.399471	503.399471	0	0	1	-360	360;
	8087	6050	0.00326	0.015581	0.02676	463.025574	463.025574	463.025574	0	0	1	-360	360;
	2458	4253	0.00099	0.01348	0.47807	2230.05234	2230.05234	2230.05234	0	0	1	-360	360;
	5322	4253	0.00045	0.00368	0.14519	1988.96192	1988.96192	1988.96192	0	0	1	-360	360;
	942	3864	0.002021	0.011331	0.0661	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8270	942	0.00076	0.0073	0.02056	562.52642	562.52642	562.52642	0	0	1	-360	360;
	1990	3682	0.00101	0.01505	0.4751	2209.17226	2209.17226	2209.17226	0	0	1	-360	360;
	5773	627	0	0.000347	1e-05	302.47988	302.47988	302.47988	0	0	1	-360	360;
	2430	3661	0.00065	0.00768	0.28563	2132.70631	2132.70631	2132.70631	0	0	1	-360	360;
	4540	5590	0.00025	0.0035	0.11792	1423.63921	1423.63921	1423.63921	0	0	1	-360	360;
	3473	2332	0.00013	0.000331	0.06511	383.34663	383.34663	383.34663	0	0	1	-360	360;
	91	2332	3.9e-05	0.00036	0.00049	513.730911	513.730911	513.730911	0	0	1	-360	360;
	4111	2610	0.011021	0.061901	0.08894	462.30272	462.30272	462.30272	0	0	1	-360	360;
	2610	5298	0.007901	0.04495	0.06366	424.864442	424.864442	424.864442	0	0	1	-360	360;
	940	5716	0.00369	0.02186	0.03835	456.006231	456.006231	456.006231	0	0	1	-360	360;
	5716	5955	0.000281	0.001091	0.15031	466.111708	466.111708	466.111708	0	0	1	-360	360;
	5716	5364	0.000461	0.001841	0.2514	482.154725	482.154725	482.154725	0	0	1	-360	360;
	1142	3280	0.006271	0.02999	0.04894	405.312706	405.312706	405.312706	0	0	1	-360	360;
	4111	10	0.011219	0.05245	0.09075	397.361692	397.361692	397.361692	0	0	1	-360	360;
	7714	3963	0.000281	0.004091	0.01608	584.825154	584.825154	584.825154	0	0	1	-360	360;
	2457	3611	0.00041	0.00534	0.1924	1704.29015	1704.29015	1704.29015	0	0	1	-360	360;
	3271	7542	0.002659	0.01557	0.02216	226.858819	226.858819	226.858819	0	0	1	-360	360;
	7960	46	0.00795	0.0483	0.0683	319.866171	319.866171	319.866171	0	0	1	-360	360;
	6029	2907	0.00037	0.00361	0.00554	472.816684	472.816684	472.816684	0	0	1	-360	360;
	1849	4808	0.001211	0.0082	0.0146	602.763138	602.763138	602.763138	0	0	1	-360	360;
	1470	3504	9.9e-05	0.00101	0.00339	402.96505	402.96505	402.96505	0	0	1	-360	360;
	1470	3161	9.9e-05	0.00099	0.00343	343.952797	343.952797	343.952797	0	0	1	-360	360;
	1470	4722	0.000169	0.001539	0.00227	509.261943	509.261943	509.261943	0	0	1	-360	360;
	1470	3496	0.000169	0.00157	0.00222	486.790212	486.790212	486.790212	0	0	1	-360	360;
	808	1708	0.00011	0.000349	0.00059	399.803468	399.803468	399.803468	0	0	1	-360	360;
	808	1708	0.00012	0.000341	0.0006	416.429874	416.429874	416.429874	0	0	1	-360	360;
	2430	790	0.00017	0.00314	0.08502	801.880911	801.880911	801.880911	0	0	1	-360	360;
	360	3219	0.00012	0.00056	0.001	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1540	1423	0.000219	0.001469	0.00214	481.447112	481.447112	481.447112	0	0	1	-360	360;
	1540	1423	0.0002	0.001781	0.00197	445.55514	445.55514	445.55514	0	0	1	-360	360;
	7787	857	0.00038	0.00623	0.17728	1650.58733	1650.58733	1650.58733	0	0	1	-360	360;
	3271	6710	0.003669	0.01631	0.03237	439.333336	439.333336	439.333336	0	0	1	-360	360;
	632	4864	0.00013	0.00117	0.0362	1751.68499	1751.68499	1751.68499	0	0	1	-360	360;
	7625	4864	0.00012	0.0013	0.03461	1422.62627	1422.62627	1422.62627	0	0	1	-360	360;
	6409	2746	0.00082	0.004969	0.00645	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6409	2746	0.0008	0.004599	0.00631	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1540	6409	0.00914	0.060289	0.14843	508.320366	508.320366	508.320366	0	0	1	-360	360;
	4111	7291	0.005219	0.02681	0.03834	295.981502	295.981502	295.981502	0	0	1	-360	360;
	4111	7291	0.005919	0.026271	0.04345	303.921847	303.921847	303.921847	0	0	1	-360	360;
	4111	5298	0.00381	0.02086	0.03074	421.830132	421.830132	421.830132	0	0	1	-360	360;
	1122	6871	0.004539	0.01957	0.03635	419.817419	419.817419	419.817419	0	0	1	-360	360;
	2539	1122	0.00112	0.00632	0.01093	402.106923	402.106923	402.106923	0	0	1	-360	360;
	6110	4020	0.00027	0.00426	0.12823	1795.81789	1795.81789	1795.81789	0	0	1	-360	360;
	8181	4020	0.0002	0.00238	0.0943	987.831045	987.831045	987.831045	0	0	1	-360	360;
	987	4020	8e-05	0.00077	0.07688	1657.21454	1657.21454	1657.21454	0	0	1	-360	360;
	4111	1099	0.00149	0.012091	0.01773	593.245623	593.245623	593.245623	0	0	1	-360	360;
	8337	3864	0.00064	0.00356	0.00563	308.847314	308.847314	308.847314	0	0	1	-360	360;
	2176	1109	0.005729	0.033	0.04566	580.775723	580.775723	580.775723	0	0	1	-360	360;
	8185	6057	0.0017	0.007599	0.01317	320.741065	320.741065	320.741065	0	0	1	-360	360;
	7358	5862	0.00221	0.02589	0.56042	1568.40838	1568.40838	1568.40838	0	0	1	-360	360;
	3370	171	0.000331	0.000831	0.07725	377.56113	377.56113	377.56113	0	0	1	-360	360;
	3781	7501	9.9e-05	0.00045	0.06456	370.274288	370.274288	370.274288	0	0	1	-360	360;
	7831	3864	0.00039	0.00163	0.19294	485.314782	485.314782	485.314782	0	0	1	-360	360;
	7396	4841	0.00383	0.04628	1.7907	1461.96367	1461.96367	1461.96367	0	0	1	-360	360;
	5954	6500	0.00106	0.01101	0.01571	447.971004	447.971004	447.971004	0	0	1	-360	360;
	8270	6623	0.00131	0.012659	0.03469	513.017965	513.017965	513.017965	0	0	1	-360	360;
	2430	5101	0.00052	0.00735	0.24795	1439.37035	1439.37035	1439.37035	0	0	1	-360	360;
	5101	2430	0.00058	0.0074	0.27818	1104.30641	1104.30641	1104.30641	0	0	1	-360	360;
	939	2430	0.0013	0.01837	0.61988	1261.64285	1261.64285	1261.64285	0	0	1	-360	360;
	4240	7025	3.9e-05	0.000599	0.00178	246.662429	246.662429	246.662429	0	0	1	-360	360;
	1142	6871	0.00031	0.001461	0.00251	507.896637	507.896637	507.896637	0	0	1	-360	360;
	2935	6314	0.001151	0.008771	0.01387	489.786418	489.786418	489.786418	0	0	1	-360	360;
	4432	3756	0.00095	0.00457	0.02953	587.022295	587.022295	587.022295	0	0	1	-360	360;
	4432	1186	0.001419	0.00982	0.06946	450.821267	450.821267	450.821267	0	0	1	-360	360;
	7831	5955	0.00037	0.00144	0.19778	486.944157	486.944157	486.944157	0	0	1	-360	360;
	6623	1191	0.00143	0.01225	0.05644	647.787002	647.787002	647.787002	0	0	1	-360	360;
	996	6871	0.00044	0.00207	0.00346	392.576071	392.576071	392.576071	0	0	1	-360	360;
	3942	5862	0.00076	0.01509	0.31584	1640.32237	1640.32237	1640.32237	0	0	1	-360	360;
	2106	2638	0.0007	0.004039	0.02175	307.394747	307.394747	307.394747	0	0	1	-360	360;
	4052	7873	0.004419	0.01918	0.03587	405.440739	405.440739	405.440739	0	0	1	-360	360;
	6050	3677	0.006099	0.029539	0.0505	517.30098	517.30098	517.30098	0	0	1	-360	360;
	6917	6675	0.00106	0.00513	0.01653	282.58984	282.58984	282.58984	0	0	1	-360	360;
	7971	8379	0.001021	0.004711	0.01581	573.728562	573.728562	573.728562	0	0	1	-360	360;
	1505	4451	0.000211	0.00106	0.0034	533.632071	533.632071	533.632071	0	0	1	-360	360;
	7364	3625	7e-05	0.00087	0.08388	240.77938	240.77938	240.77938	0	0	1	-360	360;
	7364	4451	3.9e-05	0.000651	0.06049	237.717634	237.717634	237.717634	0	0	1	-360	360;
	8379	7371	0.002039	0.008271	0.15139	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8379	7371	0.002	0.00843	0.1483	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3702	6675	0.001219	0.00725	0.01913	609.403717	609.403717	609.403717	0	0	1	-360	360;
	6675	8379	0.001151	0.006789	0.01791	529.526245	529.526245	529.526245	0	0	1	-360	360;
	3702	6675	0.0012	0.006711	0.01871	591.093827	591.093827	591.093827	0	0	1	-360	360;
	6675	8379	0.00112	0.006289	0.01752	558.939204	558.939204	558.939204	0	0	1	-360	360;
	6956	6489	0.002021	0.01101	0.00315	293.388068	293.388068	293.388068	0	0	1	-360	360;
	6489	3702	0.0015	0.008521	0.02341	241.831463	241.831463	241.831463	0	0	1	-360	360;
	6489	3702	0.001469	0.00868	0.02291	240.468442	240.468442	240.468442	0	0	1	-360	360;
	8379	3625	0.000719	0.00443	0.09403	307.362739	307.362739	307.362739	0	0	1	-360	360;
	8379	4451	0.0005	0.00268	0.00764	602.736464	602.736464	602.736464	0	0	1	-360	360;
	5638	1767	0.029149	0.051802	0.00605	66.399882	66.399882	66.399882	0	0	1	-360	360;
	5172	1767	0.035777	0.063587	0.00743	63.377575	63.377575	63.377575	0	0	1	-360	360;
	5638	2810	0.025967	0.058702	0.00539	69.41495	69.41495	69.41495	0	0	1	-360	360;
	2810	72	0.01943	0.043942	0.00403	59.794741	59.794741	59.794741	0	0	1	-360	360;
	2810	1715	0.05776	0.115802	0.01198	85.304975	85.304975	85.304975	0	0	1	-360	360;
	192	2810	0.027388	0.054917	0.00568	57.157677	57.157677	57.157677	0	0	1	-360	360;
	72	1715	0.028438	0.104719	0.00981	72.038868	72.038868	72.038868	0	0	1	-360	360;
	4730	4567	0.008091	0.03755	0.06156	425.193289	425.193289	425.193289	0	0	1	-360	360;
	7701	4567	0.00556	0.02913	0.04236	447.559088	447.559088	447.559088	0	0	1	-360	360;
	7005	192	0.033512	0.069909	0.00695	60.853873	60.853873	60.853873	0	0	1	-360	360;
	7701	4376	0.00587	0.02949	0.04463	411.347414	411.347414	411.347414	0	0	1	-360	360;
	2475	5221	0.013314	0.040719	0.00459	75.109379	75.109379	75.109379	0	0	1	-360	360;
	2475	5221	0.01357	0.039876	0.00467	75.934545	75.934545	75.934545	0	0	1	-360	360;
	2475	5221	0.014579	0.041157	0.00503	59.8197	59.8197	59.8197	0	0	1	-360	360;
	2475	3888	0.031298	0.112364	0.01079	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2475	3888	0.030653	0.114521	0.01057	57.157677	57.157677	57.157677	0	0	1	-360	360;
	224	3540	0.00537	0.04394	0.05792	449.747084	449.747084	449.747084	0	0	1	-360	360;
	224	3540	0.005461	0.04726	0.05897	439.296374	439.296374	439.296374	0	0	1	-360	360;
	4730	224	0.01312	0.074419	0.09989	429.742278	429.742278	429.742278	0	0	1	-360	360;
	5221	6225	0	0.000413	0	63.643358	63.643358	63.643358	0	0	1	-360	360;
	6225	1823	0.052702	0.121281	0.01394	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8277	2966	0.00103	0.01397	0.24541	1165.36965	1165.36965	1165.36965	0	0	1	-360	360;
	3888	1715	0.002959	0.014851	0.00261	68.545582	68.545582	68.545582	0	0	1	-360	360;
	3888	1715	0.002769	0.016339	0.00245	88.317375	88.317375	88.317375	0	0	1	-360	360;
	8277	6962	0.00167	0.02105	0.40047	1229.74287	1229.74287	1229.74287	0	0	1	-360	360;
	5172	1715	0.067826	0.163702	0.01407	74.732329	74.732329	74.732329	0	0	1	-360	360;
	3292	2193	0.009463	0.038388	0.00495	77.058074	77.058074	77.058074	0	0	1	-360	360;
	4773	2193	0.021289	0.124008	0.01113	61.851465	61.851465	61.851465	0	0	1	-360	360;
	5353	5030	0.035587	0.104182	0.0121	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1076	5030	0.040678	0.119107	0.01382	59.339576	59.339576	59.339576	0	0	1	-360	360;
	6646	7563	0.00156	0.01963	0.35333	1310.43368	1310.43368	1310.43368	0	0	1	-360	360;
	6646	2964	0.00255	0.02314	0.6124	1435.99455	1435.99455	1435.99455	0	0	1	-360	360;
	6377	7676	0.01305	0.035686	0.00541	74.648497	74.648497	74.648497	0	0	1	-360	360;
	6377	7676	0.013314	0.034959	0.00551	75.451563	75.451563	75.451563	0	0	1	-360	360;
	1871	7676	0.005612	0.022636	0.00193	57.157677	57.157677	57.157677	0	0	1	-360	360;
	55	7676	0.029826	0.113909	0.01017	102.566021	102.566021	102.566021	0	0	1	-360	360;
	6123	7676	0.04824	0.109727	0.0097	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6377	6833	0.010802	0.04124	0.00369	72.659029	72.659029	72.659029	0	0	1	-360	360;
	1871	6377	0.021851	0.050934	0.00913	70.303561	70.303561	70.303561	0	0	1	-360	360;
	3196	6377	0.070223	0.192198	0.02927	59.801219	59.801219	59.801219	0	0	1	-360	360;
	3196	6377	0.071603	0.188281	0.02985	60.284202	60.284202	60.284202	0	0	1	-360	360;
	1871	6833	0.008339	0.021066	0.0035	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1970	1871	0.024744	0.092099	0.00845	59.618886	59.618886	59.618886	0	0	1	-360	360;
	2680	3527	0.029579	0.114661	0.01006	84.473902	84.473902	84.473902	0	0	1	-360	360;
	1076	3527	0.009298	0.035488	0.00315	58.160032	58.160032	58.160032	0	0	1	-360	360;
	4285	3527	0.020149	0.078107	0.00686	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2778	5353	0.017826	0.063777	0.00608	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2964	7563	0.00158	0.02095	0.35686	1324.03496	1324.03496	1324.03496	0	0	1	-360	360;
	2862	7563	0.00076	0.00957	0.18383	1272.65024	1272.65024	1272.65024	0	0	1	-360	360;
	4329	2680	0.005678	0.044198	0.00864	57.95636	57.95636	57.95636	0	0	1	-360	360;
	4329	2680	0.006438	0.043314	0.00979	58.280063	58.280063	58.280063	0	0	1	-360	360;
	5896	2680	0.013612	0.043157	0.00468	82.151204	82.151204	82.151204	0	0	1	-360	360;
	822	2680	0.056248	0.147934	0.0234	68.778404	68.778404	68.778404	0	0	1	-360	360;
	822	2680	0.057339	0.144851	0.02385	64.941409	64.941409	64.941409	0	0	1	-360	360;
	7671	2680	0.000463	0.003603	0.00086	242.998241	242.998241	242.998241	0	0	1	-360	360;
	2680	6857	0.037107	0.138157	0.01267	77.790264	77.790264	77.790264	0	0	1	-360	360;
	4329	7360	0.004058	0.015107	0.00134	67.85969	67.85969	67.85969	0	0	1	-360	360;
	5961	1168	0.011793	0.034521	0.004	73.13782	73.13782	73.13782	0	0	1	-360	360;
	1892	1168	0.019636	0.073132	0.00672	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3922	1586	0	0.000413	0	254.536471	254.536471	254.536471	0	0	1	-360	360;
	3292	829	0.014339	0.041983	0.00486	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4727	829	0.029314	0.118281	0.01003	87.839536	87.839536	87.839536	0	0	1	-360	360;
	7700	2473	0.016636	0.064488	0.00564	65.526132	65.526132	65.526132	0	0	1	-360	360;
	5799	2473	0.034273	0.127612	0.01171	85.059959	85.059959	85.059959	0	0	1	-360	360;
	2473	4773	0.00886	0.02495	0.00297	74.253919	74.253919	74.253919	0	0	1	-360	360;
	7700	5881	0.026587	0.084364	0.00905	74.576098	74.576098	74.576098	0	0	1	-360	360;
	5799	4148	0.039298	0.115058	0.01339	59.682331	59.682331	59.682331	0	0	1	-360	360;
	5799	5961	0.036066	0.134289	0.01229	63.693085	63.693085	63.693085	0	0	1	-360	360;
	5799	1546	0.000529	0.001554	0.00022	57.157677	57.157677	57.157677	0	0	1	-360	360;
	55	5896	0.03981	0.160628	0.01352	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7360	5896	0.008438	0.025744	0.00286	72.604729	72.604729	72.604729	0	0	1	-360	360;
	5896	3456	0.024223	0.093876	0.00827	74.669836	74.669836	74.669836	0	0	1	-360	360;
	5881	4277	0.019868	0.054017	0.0047	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6123	4277	0.051248	0.121347	0.01037	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5079	5049	0.042793	0.101314	0.00864	61.955683	61.955683	61.955683	0	0	1	-360	360;
	5079	5049	0.030091	0.099372	0.0103	62.69435	62.69435	62.69435	0	0	1	-360	360;
	2074	5049	0.005488	0.020959	0.00189	58.011993	58.011993	58.011993	0	0	1	-360	360;
	1884	4580	0.009198	0.025851	0.00308	98.108485	98.108485	98.108485	0	0	1	-360	360;
	4580	5557	0.019752	0.057843	0.0067	71.184552	71.184552	71.184552	0	0	1	-360	360;
	1892	4580	0.032934	0.096413	0.01123	98.492584	98.492584	98.492584	0	0	1	-360	360;
	6097	7718	0.05586	0.112686	0.01123	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7954	7718	0.029744	0.057636	0.00604	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3252	6097	0.004868	0.013686	0.00165	57.157677	57.157677	57.157677	0	0	1	-360	360;
	376	6097	0.005413	0.015207	0.00187	73.747121	73.747121	73.747121	0	0	1	-360	360;
	5961	6097	0.043669	0.084653	0.0088	99.710233	99.710233	99.710233	0	0	1	-360	360;
	376	3252	0.004818	0.01724	0.00167	57.157677	57.157677	57.157677	0	0	1	-360	360;
	41	3252	0.017719	0.054017	0.00604	83.145748	83.145748	83.145748	0	0	1	-360	360;
	5961	3549	0.004364	0.017719	0.00231	60.053856	60.053856	60.053856	0	0	1	-360	360;
	3549	1560	0.007463	0.041727	0.00395	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1884	4148	0.016289	0.065711	0.00552	80.806665	80.806665	80.806665	0	0	1	-360	360;
	2074	2778	0.02014	0.072066	0.00686	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2740	7661	0.003091	0.011826	0.00108	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3196	7661	0.00386	0.013793	0.00127	82.499866	82.499866	82.499866	0	0	1	-360	360;
	1970	2740	0.036	0.114223	0.01227	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4727	3196	0.027942	0.081793	0.0095	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5079	4266	0.027008	0.061041	0.00592	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6593	4266	0.025504	0.072198	0.00621	66.954502	66.954502	66.954502	0	0	1	-360	360;
	5557	5881	0.01686	0.060331	0.00578	67.165414	67.165414	67.165414	0	0	1	-360	360;
	6939	1560	0.010091	0.038529	0.00342	72.913381	72.913381	72.913381	0	0	1	-360	360;
	6593	6939	0.027099	0.082603	0.00922	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4275	5014	0.014463	0.051719	0.0049	90.213295	90.213295	90.213295	0	0	1	-360	360;
	5014	4773	0.014612	0.041066	0.00495	111.82023	111.82023	111.82023	0	0	1	-360	360;
	1673	55	0.001661	0.006446	0.00056	74.349563	74.349563	74.349563	0	0	1	-360	360;
	6605	55	0.005198	0.019331	0.00173	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6153	1673	0.003512	0.011587	0.00122	61.530048	61.530048	61.530048	0	0	1	-360	360;
	6153	6605	0.004339	0.015512	0.00147	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5079	7954	0.043628	0.081231	0.00875	70.717954	70.717954	70.717954	0	0	1	-360	360;
	2464	1076	0.031661	0.092669	0.0108	73.093618	73.093618	73.093618	0	0	1	-360	360;
	1560	41	0.017562	0.070818	0.00598	79.825078	79.825078	79.825078	0	0	1	-360	360;
	1892	4148	0.008843	0.033802	0.00306	64.479003	64.479003	64.479003	0	0	1	-360	360;
	6223	8040	0.01164	0.05205	0.09202	411.696457	411.696457	411.696457	0	0	1	-360	360;
	1560	5961	0.007934	0.042653	0.00413	70.334807	70.334807	70.334807	0	0	1	-360	360;
	8142	5961	0.00343	0.014488	0.00184	91.397983	91.397983	91.397983	0	0	1	-360	360;
	107	5961	0.004339	0.022421	0.00225	69.96957	69.96957	69.96957	0	0	1	-360	360;
	5961	3922	0.00143	0.007149	0.00125	71.749269	71.749269	71.749269	0	0	1	-360	360;
	5961	3922	0.001446	0.007	0.00127	72.782871	72.782871	72.782871	0	0	1	-360	360;
	5961	3922	0.001504	0.007256	0.00127	79.007913	79.007913	79.007913	0	0	1	-360	360;
	5961	3922	0.001529	0.007099	0.0013	71.867586	71.867586	71.867586	0	0	1	-360	360;
	1560	1892	0.015653	0.051669	0.0053	61.804596	61.804596	61.804596	0	0	1	-360	360;
	3833	2964	0.00046	0.00424	0.10354	1255.97198	1255.97198	1255.97198	0	0	1	-360	360;
	3292	4773	0.049339	0.138727	0.01683	64.906924	64.906924	64.906924	0	0	1	-360	360;
	4285	5079	0.005421	0.017876	0.00092	57.157677	57.157677	57.157677	0	0	1	-360	360;
	822	2464	0.001678	0.006421	0.00054	77.916964	77.916964	77.916964	0	0	1	-360	360;
	822	6857	0.024512	0.055752	0.0049	57.462899	57.462899	57.462899	0	0	1	-360	360;
	3534	1546	0.019347	0.072058	0.00662	60.255813	60.255813	60.255813	0	0	1	-360	360;
	4275	3534	0.002364	0.008777	0.00077	61.632551	61.632551	61.632551	0	0	1	-360	360;
	3534	4773	0.032331	0.102537	0.01102	84.312908	84.312908	84.312908	0	0	1	-360	360;
	107	8142	0.00305	0.017711	0.00156	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4990	6593	0.005289	0.009463	0.0011	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4990	7636	0.00181	0.007298	0.00064	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7636	6593	0.000893	0.00338	0.00027	76.177084	76.177084	76.177084	0	0	1	-360	360;
	3456	7636	0.012033	0.046579	0.00414	73.603083	73.603083	73.603083	0	0	1	-360	360;
	5446	1541	0.00017	0.01734	0	1203.26695	1203.26695	1203.26695	0	0	1	-360	360;
	3820	7459	0.00017	0.01767	0	1089.73958	1089.73958	1089.73958	0	0	1	-360	360;
	7997	1853	0.0006	0.040218	0	648.427861	648.427861	648.427861	0	0	1	-360	360;
	8013	1986	0.00076	0.002409	0.04476	104.057109	104.057109	104.057109	0	0	1	-360	360;
	8013	3433	0	0.000538	0	104.057109	104.057109	104.057109	0	0	1	-360	360;
	5802	1426	0.00528	0.027569	0.04458	150.278704	150.278704	150.278704	0	0	1	-360	360;
	5802	1426	0.005382	0.026978	0.04542	135.677256	135.677256	135.677256	0	0	1	-360	360;
	4777	5802	0.00104	0.009649	0.00974	241.083286	241.083286	241.083286	0	0	1	-360	360;
	4777	5802	0.001062	0.009449	0.00993	279.017537	279.017537	279.017537	0	0	1	-360	360;
	3087	4769	0.000289	0.000889	0.00033	222.321798	222.321798	222.321798	0	0	1	-360	360;
	1098	4769	0.003502	0.01524	0.00403	212.739833	212.739833	212.739833	0	0	1	-360	360;
	3087	7221	0.000289	0.00088	0.00033	122.53151	122.53151	122.53151	0	0	1	-360	360;
	1098	7221	0.003431	0.014111	0.00394	217.001198	217.001198	217.001198	0	0	1	-360	360;
	7603	5016	0.001502	0.1272	0	227.270614	227.270614	227.270614	0	0	1	-360	360;
	80	655	8e-05	0.000729	0.00047	315.658206	315.658206	315.658206	0	0	1	-360	360;
	1098	655	0.003818	0.012911	0.00432	182.65489	182.65489	182.65489	0	0	1	-360	360;
	8246	655	0.001618	0.013711	0.01077	242.555356	242.555356	242.555356	0	0	1	-360	360;
	3055	1711	0.001062	0.007791	0.00315	192.125657	192.125657	192.125657	0	0	1	-360	360;
	5930	1711	0.001809	0.008951	0.00459	135.105679	135.105679	135.105679	0	0	1	-360	360;
	5256	1711	0.007142	0.020311	0.00702	157.541626	157.541626	157.541626	0	0	1	-360	360;
	3055	7574	0.000942	0.008831	0.0028	215.314007	215.314007	215.314007	0	0	1	-360	360;
	5930	7574	0.001769	0.008289	0.00449	136.159978	136.159978	136.159978	0	0	1	-360	360;
	5256	7574	0.00728	0.019898	0.00716	160.031882	160.031882	160.031882	0	0	1	-360	360;
	8009	5211	0.003569	0.017951	0.01071	184.0459	184.0459	184.0459	0	0	1	-360	360;
	3128	8009	0.001991	0.007271	0.00598	169.833904	169.833904	169.833904	0	0	1	-360	360;
	6986	8009	0.000391	0.001462	0.00421	198.201518	198.201518	198.201518	0	0	1	-360	360;
	2374	7402	0.003502	0.018289	0.01048	182.312724	182.312724	182.312724	0	0	1	-360	360;
	222	2374	0.001769	0.00824	0.00533	176.097346	176.097346	176.097346	0	0	1	-360	360;
	6986	2374	0.0004	0.001431	0.0043	147.723236	147.723236	147.723236	0	0	1	-360	360;
	1655	4311	0	0.000413	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4798	6826	0.001471	0.1296	0	187.74582	187.74582	187.74582	0	0	1	-360	360;
	1986	6684	0.00056	0.015018	0	105.874204	105.874204	105.874204	0	0	1	-360	360;
	145	7696	0.000631	0.01472	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	145	1986	0	0.001169	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	145	6892	0.000942	0.003342	0.06995	112.281319	112.281319	112.281319	0	0	1	-360	360;
	145	2456	0.001582	0.002831	0.03332	228.462351	228.462351	228.462351	0	0	1	-360	360;
	2838	6403	0.0008	0.003422	0.0042	180.916777	180.916777	180.916777	0	0	1	-360	360;
	2838	6403	0.000822	0.003351	0.00428	175.967182	175.967182	175.967182	0	0	1	-360	360;
	6892	2456	0.002631	0.004889	0.05504	234.79724	234.79724	234.79724	0	0	1	-360	360;
	6892	2095	0.005418	0.022489	0.15481	134.606588	134.606588	134.606588	0	0	1	-360	360;
	6892	2095	0.00552	0.022018	0.15779	136.088791	136.088791	136.088791	0	0	1	-360	360;
	6701	6872	0.004091	0.020893	0.00189	139.485689	139.485689	139.485689	0	0	1	-360	360;
	6701	6872	0.004	0.019339	0.00185	131.823321	131.823321	131.823321	0	0	1	-360	360;
	6701	4021	0.016479	0.037058	0.03335	103.164843	103.164843	103.164843	0	0	1	-360	360;
	6701	4259	0.016347	0.05124	0.00521	69.598617	69.598617	69.598617	0	0	1	-360	360;
	5249	6872	0.007058	0.030587	0.00259	157.446346	157.446346	157.446346	0	0	1	-360	360;
	5249	6872	0.007149	0.029934	0.00264	146.6163	146.6163	146.6163	0	0	1	-360	360;
	6189	5627	0.002089	0.003449	0.05042	243.249562	243.249562	243.249562	0	0	1	-360	360;
	6189	5627	0.002129	0.003378	0.05139	248.50495	248.50495	248.50495	0	0	1	-360	360;
	2117	646	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	646	3890	0.007489	0.0516	0.01757	122.367571	122.367571	122.367571	0	0	1	-360	360;
	646	3890	0.008489	0.050569	0.01991	123.296383	123.296383	123.296383	0	0	1	-360	360;
	793	2917	0	0.047951	0.50558	118.686876	118.686876	118.686876	0	0	1	-360	360;
	6431	2917	0.001649	0.016591	0.37763	129.665827	129.665827	129.665827	0	0	1	-360	360;
	7116	2917	0.001618	0.016898	0.36958	132.87601	132.87601	132.87601	0	0	1	-360	360;
	3178	5627	0.006871	0.039911	0.01519	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3178	5627	0.007009	0.039089	0.01549	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3178	2895	0.001658	0.011871	0.00368	110.478514	110.478514	110.478514	0	0	1	-360	360;
	3178	2895	0.001631	0.010991	0.0036	112.490984	112.490984	112.490984	0	0	1	-360	360;
	4926	8275	0.00092	0.004689	0.00279	116.429148	116.429148	116.429148	0	0	1	-360	360;
	118	8275	0.002049	0.0038	0.06219	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5737	8275	0.005649	0.026	0.0176	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4926	2771	0.000902	0.004782	0.00274	167.836503	167.836503	167.836503	0	0	1	-360	360;
	250	2771	0.003471	0.014471	0.01055	143.937059	143.937059	143.937059	0	0	1	-360	360;
	118	2771	0.002089	0.00372	0.06341	189.279205	189.279205	189.279205	0	0	1	-360	360;
	3816	1721	0.003569	0.010671	0.00362	187.603966	187.603966	187.603966	0	0	1	-360	360;
	3550	3816	0.019782	0.061658	0.01962	116.406285	116.406285	116.406285	0	0	1	-360	360;
	440	5211	0.000409	0.0014	0.00438	174.887422	174.887422	174.887422	0	0	1	-360	360;
	5211	487	0.003302	0.017311	0.00992	200.598763	200.598763	200.598763	0	0	1	-360	360;
	440	7402	0.000409	0.001369	0.00446	110.855235	110.855235	110.855235	0	0	1	-360	360;
	7402	487	0.003231	0.016031	0.0097	189.504198	189.504198	189.504198	0	0	1	-360	360;
	8254	7670	0.000751	0.005711	0.12939	171.908468	171.908468	171.908468	0	0	1	-360	360;
	8254	8168	0.0018	0.00716	0.08656	311.769145	311.769145	311.769145	0	0	1	-360	360;
	5765	8254	0.001978	0.013462	0.21156	146.687903	146.687903	146.687903	0	0	1	-360	360;
	8254	1770	0.001502	0.0056	0.27143	142.66634	142.66634	142.66634	0	0	1	-360	360;
	2982	5825	0.005991	0.0284	0.03521	164.867421	164.867421	164.867421	0	0	1	-360	360;
	1877	5825	0.000302	0.000858	0.00034	119.890305	119.890305	119.890305	0	0	1	-360	360;
	5694	5825	0.002622	0.013231	0.00952	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6475	5096	0.004271	0.019502	0.01354	132.086974	132.086974	132.086974	0	0	1	-360	360;
	1877	5096	0.000271	0.000969	0.0003	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5694	5096	0.00356	0.014378	0.02094	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5705	3141	0.0022	0.009569	0.03479	141.468367	141.468367	141.468367	0	0	1	-360	360;
	2489	3141	0.000311	0.001169	0.0048	111.245726	111.245726	111.245726	0	0	1	-360	360;
	3000	3141	0.001889	0.006311	0.00214	259.272158	259.272158	259.272158	0	0	1	-360	360;
	2489	2982	0.000271	0.00132	0.00427	108.100495	108.100495	108.100495	0	0	1	-360	360;
	3000	2982	0.001929	0.006178	0.00218	272.441546	272.441546	272.441546	0	0	1	-360	360;
	716	5985	0.000339	0.001149	0.00841	57.157677	57.157677	57.157677	0	0	1	-360	360;
	426	8070	0.002591	0.014418	0	192.853638	192.853638	192.853638	0	0	1	-360	360;
	426	703	0	0.00028	0	247.750989	247.750989	247.750989	0	0	1	-360	360;
	5767	3128	0.0014	0.00652	0.00421	237.308541	237.308541	237.308541	0	0	1	-360	360;
	7399	3128	0.000418	0.001338	0.00454	108.435907	108.435907	108.435907	0	0	1	-360	360;
	5767	222	0.001369	0.006649	0.00413	210.186704	210.186704	210.186704	0	0	1	-360	360;
	7399	222	0.000369	0.00152	0.00405	125.21896	125.21896	125.21896	0	0	1	-360	360;
	7997	4834	0.004689	0.018151	0.01534	148.335083	148.335083	148.335083	0	0	1	-360	360;
	7997	4834	0.004591	0.018498	0.01502	148.66322	148.66322	148.66322	0	0	1	-360	360;
	1140	127	9.8e-05	0.00028	0	155.465503	155.465503	155.465503	0	0	1	-360	360;
	1140	3622	0.003698	0.0198	0	115.771575	115.771575	115.771575	0	0	1	-360	360;
	7271	1140	0.005471	0.0206	0.03082	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7271	1140	0.0062	0.020191	0.03493	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1140	3972	0.006098	0.030551	0.06659	129.394588	129.394588	129.394588	0	0	1	-360	360;
	861	5256	0.004809	0.019751	0.00546	103.923048	103.923048	103.923048	0	0	1	-360	360;
	861	5256	0.004711	0.018289	0.00534	270.260201	270.260201	270.260201	0	0	1	-360	360;
	3087	5486	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5705	6475	0.007929	0.044022	0.01896	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5497	1230	0.009782	0.033938	0.01107	257.757479	257.757479	257.757479	0	0	1	-360	360;
	5497	5947	0.003609	0.016889	0.00487	103.923048	103.923048	103.923048	0	0	1	-360	360;
	593	1266	0.00044	0.027591	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6201	3217	0.000419	0.027979	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	316	3217	0.000409	0.02851	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1266	3217	0	0.000461	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1266	2557	0.000531	0.005581	0.02446	315.63193	315.63193	315.63193	0	0	1	-360	360;
	959	1266	0.003841	0.022461	0.05954	481.782819	481.782819	481.782819	0	0	1	-360	360;
	3217	124	0.000539	0.005461	0.02494	407.752577	407.752577	407.752577	0	0	1	-360	360;
	3530	3217	0.00376	0.022901	0.05833	469.010745	469.010745	469.010745	0	0	1	-360	360;
	4549	8043	0.028843	0.060992	0.00616	72.650836	72.650836	72.650836	0	0	1	-360	360;
	4549	8043	0.029397	0.059744	0.00628	73.963558	73.963558	73.963558	0	0	1	-360	360;
	1655	5753	0.001727	0.00924	0.00142	142.688424	142.688424	142.688424	0	0	1	-360	360;
	1625	1655	0.004851	0.02724	0.00403	111.41041	111.41041	111.41041	0	0	1	-360	360;
	1230	5947	0.0138	0.040249	0.01501	261.833081	261.833081	261.833081	0	0	1	-360	360;
	1230	582	0.007098	0.022329	0.00775	128.886664	128.886664	128.886664	0	0	1	-360	360;
	1230	1679	0.001902	0.008391	0.00607	202.887668	202.887668	202.887668	0	0	1	-360	360;
	1230	1679	0.001929	0.008209	0.00619	191.922748	191.922748	191.922748	0	0	1	-360	360;
	1522	1230	0.001649	0.01228	0.00982	130.585546	130.585546	130.585546	0	0	1	-360	360;
	1522	1230	0.001471	0.01392	0.00875	159.769736	159.769736	159.769736	0	0	1	-360	360;
	1230	6922	0.000582	0.040991	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	1230	283	0.01116	0.033671	0.01217	290.989992	290.989992	290.989992	0	0	1	-360	360;
	6346	2251	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4556	7478	0.00031	0.00312	0.10854	1277.06728	1277.06728	1277.06728	0	0	1	-360	360;
	5508	7478	0.00069	0.00769	0.24736	1836.16494	1836.16494	1836.16494	0	0	1	-360	360;
	5570	4944	0.010471	0.029289	0.01153	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5570	4944	0.010671	0.028671	0.01175	127.607891	127.607891	127.607891	0	0	1	-360	360;
	310	5570	0.00624	0.021871	0.00687	311.769145	311.769145	311.769145	0	0	1	-360	360;
	310	5570	0.007071	0.021431	0.00778	311.769145	311.769145	311.769145	0	0	1	-360	360;
	8062	7031	9.8e-05	0.00028	0	144.333786	144.333786	144.333786	0	0	1	-360	360;
	8062	5947	0.001062	0.003351	0.0047	170.28415	170.28415	170.28415	0	0	1	-360	360;
	8062	4854	0.002111	0.02008	0.01412	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8062	5806	0.001009	0.010929	0.00675	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1522	8062	0.004711	0.050609	0.03014	133.144391	133.144391	133.144391	0	0	1	-360	360;
	8062	8246	0.011462	0.02396	0.02125	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4798	2135	0	0.0444	0.1406	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4798	1024	0.001698	0.1128	0	206.105126	206.105126	206.105126	0	0	1	-360	360;
	4798	1488	8.9e-05	0.000742	0	132.176088	132.176088	132.176088	0	0	1	-360	360;
	4798	1488	8.9e-05	0.000729	0	136.136076	136.136076	136.136076	0	0	1	-360	360;
	4798	2546	0.000302	0.000809	0.02492	137.198169	137.198169	137.198169	0	0	1	-360	360;
	4798	8200	0.001502	0.002182	0.06789	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4798	5578	0.001311	0.001871	0.06139	113.282878	113.282878	113.282878	0	0	1	-360	360;
	4798	5578	0.001369	0.00204	0.0638	200.169042	200.169042	200.169042	0	0	1	-360	360;
	4798	4198	0.001049	0.00308	0.07664	274.619254	274.619254	274.619254	0	0	1	-360	360;
	1625	5346	0.014967	0.09105	0.01258	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4810	5346	0.028959	0.12905	0.01817	85.674975	85.674975	85.674975	0	0	1	-360	360;
	2247	5346	0.006438	0.041256	0.00593	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2247	5346	0.006554	0.040364	0.00604	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2135	1145	0	0.012742	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5765	7670	0.001018	0.00868	0.17458	168.331956	168.331956	168.331956	0	0	1	-360	360;
	5765	8168	0.001871	0.00548	0.30536	218.160459	218.160459	218.160459	0	0	1	-360	360;
	8168	6779	0.00228	0.004338	0.13292	234.751514	234.751514	234.751514	0	0	1	-360	360;
	8168	6779	0.00224	0.004431	0.1302	244.563669	244.563669	244.563669	0	0	1	-360	360;
	2749	731	0.00076	0.00747	0.27098	1711.25369	1711.25369	1711.25369	0	0	1	-360	360;
	5150	731	0.00103	0.01061	0.3694	1489.86455	1489.86455	1489.86455	0	0	1	-360	360;
	2042	8165	0.002689	0.013818	0	227.846867	227.846867	227.846867	0	0	1	-360	360;
	487	2042	0.002111	0.003769	0.12183	116.955778	116.955778	116.955778	0	0	1	-360	360;
	487	2042	0.002151	0.003689	0.12422	277.833853	277.833853	277.833853	0	0	1	-360	360;
	7997	4926	0.00176	0.012702	0.01192	193.494324	193.494324	193.494324	0	0	1	-360	360;
	7997	4926	0.001729	0.012951	0.01167	250.289829	250.289829	250.289829	0	0	1	-360	360;
	5218	953	0.000151	0.000511	0.00034	255.059117	255.059117	255.059117	0	0	1	-360	360;
	5218	953	0.000151	0.000502	0.00034	263.546513	263.546513	263.546513	0	0	1	-360	360;
	5218	1679	0.01076	0.05508	0.03402	153.451474	153.451474	153.451474	0	0	1	-360	360;
	5218	1679	0.010542	0.05616	0.03332	150.815986	150.815986	150.815986	0	0	1	-360	360;
	5218	6190	0.003089	0.009	0.00268	224.382592	224.382592	224.382592	0	0	1	-360	360;
	5218	6190	0.003031	0.009169	0.00262	231.574847	231.574847	231.574847	0	0	1	-360	360;
	5218	5694	0.001849	0.00872	0.00591	137.274033	137.274033	137.274033	0	0	1	-360	360;
	5218	5694	0.00188	0.008538	0.00602	109.650767	109.650767	109.650767	0	0	1	-360	360;
	2703	3000	0.004151	0.01556	0.00468	257.106141	257.106141	257.106141	0	0	1	-360	360;
	1522	2703	0.006591	0.026778	0.00744	311.269535	311.269535	311.269535	0	0	1	-360	360;
	6346	8056	0.00056	0.038671	0	506.894542	506.894542	506.894542	0	0	1	-360	360;
	1596	537	0	0.00045	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	537	7459	0.00162	0.01968	0.5839	1616.22906	1616.22906	1616.22906	0	0	1	-360	360;
	6346	7958	9.8e-05	0.00028	0	123.145955	123.145955	123.145955	0	0	1	-360	360;
	6346	5767	0.002649	0.015111	0.01134	217.975996	217.975996	217.975996	0	0	1	-360	360;
	6346	5767	0.002702	0.014809	0.01156	246.781906	246.781906	246.781906	0	0	1	-360	360;
	6346	4576	0.0008	0.005711	0.02164	137.188816	137.188816	137.188816	0	0	1	-360	360;
	6346	4576	0.000782	0.005809	0.02118	142.379773	142.379773	142.379773	0	0	1	-360	360;
	6346	2998	0.001049	0.007738	0.0052	289.566246	289.566246	289.566246	0	0	1	-360	360;
	6346	2998	0.001071	0.007569	0.0053	311.769145	311.769145	311.769145	0	0	1	-360	360;
	6346	2838	0.005138	0.025338	0.16348	125.818596	125.818596	125.818596	0	0	1	-360	360;
	6346	2838	0.005031	0.025831	0.16015	139.961223	139.961223	139.961223	0	0	1	-360	360;
	6346	1331	0.003658	0.029071	0.01363	110.210653	110.210653	110.210653	0	0	1	-360	360;
	6346	1331	0.004151	0.028489	0.01545	126.925636	126.925636	126.925636	0	0	1	-360	360;
	1110	1228	0.001471	0.011471	0.00955	187.307265	187.307265	187.307265	0	0	1	-360	360;
	1110	1228	0.001502	0.011218	0.00973	206.377404	206.377404	206.377404	0	0	1	-360	360;
	2095	1228	0.001871	0.014511	0.01207	185.364424	185.364424	185.364424	0	0	1	-360	360;
	2095	1228	0.001902	0.0142	0.0123	232.805555	232.805555	232.805555	0	0	1	-360	360;
	179	6524	0.000351	0.004662	0.00321	169.898596	169.898596	169.898596	0	0	1	-360	360;
	44	179	0.00772	0.033818	0.02355	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4516	7643	0	0.00051	0	1919.93471	1919.93471	1919.93471	0	0	1	-360	360;
	7643	2903	0.00076	0.0119	0.47516	822.094256	822.094256	822.094256	0	0	1	-360	360;
	7643	3044	0.00025	0.00316	0.10448	1583.67287	1583.67287	1583.67287	0	0	1	-360	360;
	4516	3165	0.00077	0.01164	0.48396	841.706683	841.706683	841.706683	0	0	1	-360	360;
	4516	6137	0.00025	0.00322	0.10235	1499.7096	1499.7096	1499.7096	0	0	1	-360	360;
	758	5710	0.001369	0.021182	0.93656	504.076148	504.076148	504.076148	0	0	1	-360	360;
	2903	3165	0	0.00052	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2903	5685	0.00015	0.00328	0.1124	798.555789	798.555789	798.555789	0	0	1	-360	360;
	356	2903	0.00016	0.00314	0.11526	1365.64045	1365.64045	1365.64045	0	0	1	-360	360;
	3165	5508	0.00017	0.00321	0.12739	997.878153	997.878153	997.878153	0	0	1	-360	360;
	5301	3165	0.00018	0.00308	0.13063	979.61236	979.61236	979.61236	0	0	1	-360	360;
	2921	3663	0.00024	0.01948	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1522	6731	0.000631	0.037902	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2162	5238	0.00244	0.01532	0	235.644213	235.644213	235.644213	0	0	1	-360	360;
	256	7333	0	0.00053	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	256	5508	0.00013	0.00282	0.0914	988.91046	988.91046	988.91046	0	0	1	-360	360;
	3713	256	0.00038	0.00708	0.27003	1248.46479	1248.46479	1248.46479	0	0	1	-360	360;
	7333	5685	0.00013	0.00261	0.08942	1093.32073	1093.32073	1093.32073	0	0	1	-360	360;
	5342	7333	0.00038	0.00722	0.2644	1474.51384	1474.51384	1474.51384	0	0	1	-360	360;
	1343	4777	0.00016	0.002769	0.11027	254.46052	254.46052	254.46052	0	0	1	-360	360;
	3608	4777	0.00016	0.00272	0.11239	311.769145	311.769145	311.769145	0	0	1	-360	360;
	7292	4777	0.00016	0.002662	0.11451	311.769145	311.769145	311.769145	0	0	1	-360	360;
	1816	1455	0.024479	0.081727	0.00705	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6281	1455	0.024248	0.064893	0.00592	71.417564	71.417564	71.417564	0	0	1	-360	360;
	612	7896	0.0084	0.04056	0.02022	118.878874	118.878874	118.878874	0	0	1	-360	360;
	7896	184	0.003809	0.0248	0.01222	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7896	3972	0.002542	0.014569	0.04355	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7896	3972	0.002591	0.01428	0.04441	105.095041	105.095041	105.095041	0	0	1	-360	360;
	6880	7018	9.8e-05	0.00028	1.1	159.733623	159.733623	159.733623	0	0	1	-360	360;
	6880	5077	0.003071	0.01636	0.00764	128.917841	128.917841	128.917841	0	0	1	-360	360;
	6880	4872	0.003889	0.027022	0.00967	131.59204	131.59204	131.59204	0	0	1	-360	360;
	4341	6880	0.00624	0.036009	0.01551	210.170596	210.170596	210.170596	0	0	1	-360	360;
	4341	6880	0.00636	0.035258	0.0158	211.111879	211.111879	211.111879	0	0	1	-360	360;
	4777	2532	0.000982	0.001529	0.06484	114.269887	114.269887	114.269887	0	0	1	-360	360;
	4777	2532	0.000991	0.001502	0.06606	151.118662	151.118662	151.118662	0	0	1	-360	360;
	7839	4777	0.00052	0.004991	0.09797	262.289563	262.289563	262.289563	0	0	1	-360	360;
	4777	2451	0	0.00028	0	127.24364	127.24364	127.24364	0	0	1	-360	360;
	588	3736	0.012182	0.034826	0.01493	88.865707	88.865707	88.865707	0	0	1	-360	360;
	1816	588	0.025744	0.059	0.0192	125.447954	125.447954	125.447954	0	0	1	-360	360;
	3760	1490	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5596	3760	0	0.000222	0	118.01995	118.01995	118.01995	0	0	1	-360	360;
	5596	5326	0.003129	0.015031	0.04704	132.608408	132.608408	132.608408	0	0	1	-360	360;
	3760	449	0.00228	0.019871	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3760	5326	0.002791	0.017031	0.0419	125.410698	125.410698	125.410698	0	0	1	-360	360;
	5969	3145	0.001431	0.01252	0.1777	210.634352	210.634352	210.634352	0	0	1	-360	360;
	5969	3717	0.001649	0.007631	0.01346	140.970576	140.970576	140.970576	0	0	1	-360	360;
	5969	4768	0.001471	0.008649	0.012	115.539567	115.539567	115.539567	0	0	1	-360	360;
	449	5326	0.00656	0.009071	0.20079	266.038587	266.038587	266.038587	0	0	1	-360	360;
	5488	3331	0	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1895	8230	0	0.00028	0	241.371153	241.371153	241.371153	0	0	1	-360	360;
	2932	7298	0.000521	0.001529	0.01597	392.885294	392.885294	392.885294	0	0	1	-360	360;
	66	5213	0.000219	0.00095	0.0598	456.346509	456.346509	456.346509	0	0	1	-360	360;
	33	5213	0.000219	0.00088	0.0585	540.911292	540.911292	540.911292	0	0	1	-360	360;
	7997	8273	0.002391	0.015618	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4556	831	0	0.00054	0	1456.45735	1456.45735	1456.45735	0	0	1	-360	360;
	1636	4556	0.00079	0.01037	0.28387	1686.62396	1686.62396	1686.62396	0	0	1	-360	360;
	599	4556	0.00022	0.00296	0.08288	1654.48902	1654.48902	1654.48902	0	0	1	-360	360;
	7801	831	0.00077	0.01056	0.27783	1820.92743	1820.92743	1820.92743	0	0	1	-360	360;
	5765	3911	0.000649	0.03712	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5765	4732	0.003302	0.019138	0.02159	163.562667	163.562667	163.562667	0	0	1	-360	360;
	5765	4732	0.003369	0.018742	0.02201	150.921468	150.921468	150.921468	0	0	1	-360	360;
	5765	426	0.001769	0.018222	0.08625	230.358427	230.358427	230.358427	0	0	1	-360	360;
	5765	6779	0.000858	0.006809	0.01934	156.575921	156.575921	156.575921	0	0	1	-360	360;
	5765	6779	0.000969	0.006671	0.02192	168.948739	168.948739	168.948739	0	0	1	-360	360;
	5765	7977	0.005551	0.03908	0.0267	225.5881	225.5881	225.5881	0	0	1	-360	360;
	5765	7977	0.005431	0.039822	0.02613	229.617716	229.617716	229.617716	0	0	1	-360	360;
	5765	7977	0.005311	0.036871	0.02556	242.058084	242.058084	242.058084	0	0	1	-360	360;
	929	998	0.002169	0.016271	0.01291	129.886663	129.886663	129.886663	0	0	1	-360	360;
	2895	998	0.007418	0.07832	0.05206	143.978888	143.978888	143.978888	0	0	1	-360	360;
	4389	2749	0	0.0005	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4389	5756	0	0.00049	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4389	7356	0.00079	0.00848	0.27731	1589.51224	1589.51224	1589.51224	0	0	1	-360	360;
	4389	2921	0.00183	0.0174	0.6567	1447.83585	1447.83585	1447.83585	0	0	1	-360	360;
	2749	5756	0	0.00048	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2749	7959	0.0008	0.0083	0.28275	1691.52872	1691.52872	1691.52872	0	0	1	-360	360;
	7155	61	0.004449	0.02404	0.04483	144.243113	144.243113	144.243113	0	0	1	-360	360;
	7155	61	0.004529	0.023551	0.04571	129.99916	129.99916	129.99916	0	0	1	-360	360;
	7155	5627	0.006791	0.013271	0.54829	272.264877	272.264877	272.264877	0	0	1	-360	360;
	7155	1009	0.00424	0.030378	0.00933	110.414602	110.414602	110.414602	0	0	1	-360	360;
	7155	1009	0.004151	0.028129	0.00913	112.32107	112.32107	112.32107	0	0	1	-360	360;
	7155	6622	0.002311	0.003622	0.16224	123.69285	123.69285	123.69285	0	0	1	-360	360;
	7155	6622	0.002062	0.004102	0.14454	119.60088	119.60088	119.60088	0	0	1	-360	360;
	7155	6335	0.00452	0.029609	0.01032	129.868217	129.868217	129.868217	0	0	1	-360	360;
	7155	6335	0.004422	0.030169	0.0101	129.063333	129.063333	129.063333	0	0	1	-360	360;
	7155	4341	0.00416	0.023058	0.01033	257.29762	257.29762	257.29762	0	0	1	-360	360;
	7155	3294	0.002831	0.020742	0.00702	283.517405	283.517405	283.517405	0	0	1	-360	360;
	2838	1572	0.002338	0.01592	0	114.354844	114.354844	114.354844	0	0	1	-360	360;
	529	2864	9.8e-05	0.00028	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1009	529	0.004969	0.03636	0.01338	113.77963	113.77963	113.77963	0	0	1	-360	360;
	6335	529	0.003431	0.015698	0.05911	136.760653	136.760653	136.760653	0	0	1	-360	360;
	6596	529	0.00456	0.027582	0.01254	110.41694	110.41694	110.41694	0	0	1	-360	360;
	7356	7959	0	0.00047	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7605	7356	0.00093	0.01185	0.33039	1827.66389	1827.66389	1827.66389	0	0	1	-360	360;
	123	7959	0.00091	0.01207	0.32336	1817.23043	1817.23043	1817.23043	0	0	1	-360	360;
	612	5841	9.8e-05	0.00028	0	135.865876	135.865876	135.865876	0	0	1	-360	360;
	7926	612	0.000529	0.003151	0.00225	191.700092	191.700092	191.700092	0	0	1	-360	360;
	7926	612	0.00052	0.003209	0.0022	215.685792	215.685792	215.685792	0	0	1	-360	360;
	612	184	0.00352	0.020578	0.01363	135.665304	135.665304	135.665304	0	0	1	-360	360;
	8043	6274	0.039182	0.089413	0.01109	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8043	8354	0.033686	0.067124	0.00704	60.51912	60.51912	60.51912	0	0	1	-360	360;
	5800	3696	0.044992	0.093174	0.01229	70.244498	70.244498	70.244498	0	0	1	-360	360;
	5471	5800	0.013678	0.052479	0.01461	71.76756	71.76756	71.76756	0	0	1	-360	360;
	4434	5800	0.005686	0.022893	0.00602	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5800	6473	0.034281	0.094653	0.00913	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2508	5553	0	0.00045	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	66	2508	0.00032	0.001469	0.09518	449.496352	449.496352	449.496352	0	0	1	-360	360;
	33	5553	0.000289	0.001039	0.0853	470.258688	470.258688	470.258688	0	0	1	-360	360;
	2932	6456	0.00324	0.017628	0.00278	171.47303	171.47303	171.47303	0	0	1	-360	360;
	2932	7728	0.002678	0.019298	0.00229	171.47303	171.47303	171.47303	0	0	1	-360	360;
	2932	4857	0.000529	0.001504	0.01627	171.47303	171.47303	171.47303	0	0	1	-360	360;
	4542	3029	0.00042	0.0261	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6198	4363	0.00043	0.02555	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5886	4363	0.00044	0.02499	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7510	4363	0.00039	0.02833	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	522	1239	0.00019	0.01594	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	243	963	0.000419	0.026099	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2474	5426	0.00043	0.02555	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7856	7523	0.00038	0.02888	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6246	3698	0.00038	0.02944	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3387	3029	0.00037	0.02999	0	1891.53164	1891.53164	1891.53164	0	0	1	-360	360;
	1729	3029	0.00036	0.02777	0	2303.62757	2303.62757	2303.62757	0	0	1	-360	360;
	61	7377	9.8e-05	0.00028	1	154.204916	154.204916	154.204916	0	0	1	-360	360;
	6613	61	0.003951	0.02048	0.00715	103.923048	103.923048	103.923048	0	0	1	-360	360;
	61	6615	0.003858	0.020871	0.007	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4363	7523	0	0.00046	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4363	3698	0	0.00045	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4363	6695	0	0.00051	0	834.614143	834.614143	834.614143	0	0	1	-360	360;
	4363	1827	0	0.00052	0	841.047187	841.047187	841.047187	0	0	1	-360	360;
	2221	5426	0	0.00051	0	283.117596	283.117596	283.117596	0	0	1	-360	360;
	2221	963	0	0.000521	0	339.615292	339.615292	339.615292	0	0	1	-360	360;
	2221	5699	0	0.000531	0	388.587608	388.587608	388.587608	0	0	1	-360	360;
	2221	3952	0	0.000539	0	231.336551	231.336551	231.336551	0	0	1	-360	360;
	3029	7523	0	0.00053	0	1426.76161	1426.76161	1426.76161	0	0	1	-360	360;
	3029	3698	0	0.00054	0	974.284399	974.284399	974.284399	0	0	1	-360	360;
	3029	6695	0	0.0005	0	1133.87314	1133.87314	1133.87314	0	0	1	-360	360;
	3029	1827	0	0.00049	0	934.651474	934.651474	934.651474	0	0	1	-360	360;
	293	5426	0	0.0005	0	288.149757	288.149757	288.149757	0	0	1	-360	360;
	293	963	0	0.00049	0	379.415325	379.415325	379.415325	0	0	1	-360	360;
	293	5699	0	0.000479	0	374.199496	374.199496	374.199496	0	0	1	-360	360;
	293	3952	0	0.000469	0	253.335017	253.335017	253.335017	0	0	1	-360	360;
	4421	7523	0.00043	0.00544	0.28708	1608.43226	1608.43226	1608.43226	0	0	1	-360	360;
	2284	5426	0	0.000461	0	480.50744	480.50744	480.50744	0	0	1	-360	360;
	963	1239	0	0.00045	0	475.544249	475.544249	475.544249	0	0	1	-360	360;
	1787	6695	0.00038	0.00616	0.25576	1475.43397	1475.43397	1475.43397	0	0	1	-360	360;
	4813	1827	0.00041	0.02721	0	1098.28012	1098.28012	1098.28012	0	0	1	-360	360;
	3952	959	0.002081	0.023	0.08065	550.63648	550.63648	550.63648	0	0	1	-360	360;
	6042	758	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4697	2468	0	0.00048	0	1522.11271	1522.11271	1522.11271	0	0	1	-360	360;
	4697	1541	0.0011	0.01189	0.39796	1839.29721	1839.29721	1839.29721	0	0	1	-360	360;
	4697	2921	0.00084	0.00906	0.30326	1051.50594	1051.50594	1051.50594	0	0	1	-360	360;
	2468	1541	0.00112	0.01165	0.40561	1881.02644	1881.02644	1881.02644	0	0	1	-360	360;
	2468	7459	0.00115	0.0114	0.41326	1760.78629	1760.78629	1760.78629	0	0	1	-360	360;
	2468	5150	0.00086	0.00887	0.3091	1294.59328	1294.59328	1294.59328	0	0	1	-360	360;
	3055	6524	0.00116	0.00912	0.00927	261.909725	261.909725	261.909725	0	0	1	-360	360;
	3055	6524	0.001129	0.009298	0.00908	155.178416	155.178416	155.178416	0	0	1	-360	360;
	6524	8152	0.000609	0.005058	0.07718	152.475377	152.475377	152.475377	0	0	1	-360	360;
	44	6524	0.008009	0.04	0.02523	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6524	3522	0.002289	0.016218	0	193.847922	193.847922	193.847922	0	0	1	-360	360;
	3055	353	0.003098	0.032729	0.00789	218.544975	218.544975	218.544975	0	0	1	-360	360;
	3055	3209	0.002982	0.013058	0.01268	234.702151	234.702151	234.702151	0	0	1	-360	360;
	3055	3209	0.00304	0.0128	0.01293	233.04224	233.04224	233.04224	0	0	1	-360	360;
	3055	5289	0.008791	0.059622	0.01784	143.966937	143.966937	143.966937	0	0	1	-360	360;
	8152	5930	0.000511	0.004409	0.06356	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6963	5930	0.009391	0.030809	0.00759	247.943506	247.943506	247.943506	0	0	1	-360	360;
	6963	5930	0.009191	0.031391	0.00742	266.943757	266.943757	266.943757	0	0	1	-360	360;
	1552	2629	9.8e-05	0.000222	0	141.630487	141.630487	141.630487	0	0	1	-360	360;
	5627	1552	0.00276	0.014782	0.0063	106.409927	106.409927	106.409927	0	0	1	-360	360;
	5627	1552	0.002809	0.014458	0.00642	106.882777	106.882777	106.882777	0	0	1	-360	360;
	1009	1552	0.003138	0.017502	0.00719	160.330661	160.330661	160.330661	0	0	1	-360	360;
	1009	1552	0.0032	0.017138	0.00732	155.181014	155.181014	155.181014	0	0	1	-360	360;
	5367	5753	0.002033	0.005273	0.04891	101.825067	101.825067	101.825067	0	0	1	-360	360;
	3918	5753	0.036066	0.160107	0.01461	100.061943	100.061943	100.061943	0	0	1	-360	360;
	664	8095	0.027686	0.04614	0.04049	69.413235	69.413235	69.413235	0	0	1	-360	360;
	4923	8095	0.025107	0.051372	0.00485	75.269039	75.269039	75.269039	0	0	1	-360	360;
	1625	494	0.002479	0.01086	0.00305	77.956784	77.956784	77.956784	0	0	1	-360	360;
	494	6281	0.013603	0.051876	0.0061	62.286244	62.286244	62.286244	0	0	1	-360	360;
	8200	7713	0.00176	0.108	0	213.296341	213.296341	213.296341	0	0	1	-360	360;
	1450	5046	0	0.00047	0	1007.22759	1007.22759	1007.22759	0	0	1	-360	360;
	1450	7801	0.00038	0.00366	0.13832	1775.52227	1775.52227	1775.52227	0	0	1	-360	360;
	1450	7490	0.00052	0.00633	0.18396	1919.02445	1919.02445	1919.02445	0	0	1	-360	360;
	1025	7715	0	0.00051	0	254.903042	254.903042	254.903042	0	0	1	-360	360;
	1025	2962	0.001599	0.015599	0.07413	333.603066	333.603066	333.603066	0	0	1	-360	360;
	4825	1025	0.00212	0.022349	0.09805	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5046	1636	0.00034	0.00415	0.12323	1655.51446	1655.51446	1655.51446	0	0	1	-360	360;
	7442	5046	0.00051	0.00646	0.18012	1769.36566	1769.36566	1769.36566	0	0	1	-360	360;
	7715	332	0.00163	0.015271	0.07552	345.792131	345.792131	345.792131	0	0	1	-360	360;
	3543	7715	0.002159	0.021901	0.09998	269.837962	269.837962	269.837962	0	0	1	-360	360;
	5399	711	0.00224	0.015018	0	240.856474	240.856474	240.856474	0	0	1	-360	360;
	5399	2895	0.004111	0.00652	0.01292	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5399	2895	0.004191	0.006391	0.01318	175.765831	175.765831	175.765831	0	0	1	-360	360;
	6281	4646	0.00238	0.006744	0.05757	100.396125	100.396125	100.396125	0	0	1	-360	360;
	5289	353	0.003329	0.033129	0.00729	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1488	7603	0.005809	0.021182	0.08336	117.174536	117.174536	117.174536	0	0	1	-360	360;
	1488	7603	0.005689	0.021591	0.08166	138.31846	138.31846	138.31846	0	0	1	-360	360;
	7603	2466	0.00132	0.01052	0.31568	139.044102	139.044102	139.044102	0	0	1	-360	360;
	7603	221	0.001129	0.012329	0.26976	126.158684	126.158684	126.158684	0	0	1	-360	360;
	7603	4016	0.001569	0.1224	0	212.493795	212.493795	212.493795	0	0	1	-360	360;
	2838	2330	0.000662	0.036351	0	469.587206	469.587206	469.587206	0	0	1	-360	360;
	3931	6711	0.001777	0.009107	0.00205	75.198926	75.198926	75.198926	0	0	1	-360	360;
	3931	1218	0.002967	0.002901	0.03715	72.285027	72.285027	72.285027	0	0	1	-360	360;
	3931	7514	0.017893	0.057479	0.00824	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6711	8063	0.000554	0.001091	0.00842	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6711	6864	0.006339	0.022603	0.00218	143.126252	143.126252	143.126252	0	0	1	-360	360;
	1218	324	0.004174	0.006702	0.06492	60.32002	60.32002	60.32002	0	0	1	-360	360;
	324	6864	0.009818	0.040099	0.00793	87.207754	87.207754	87.207754	0	0	1	-360	360;
	250	5737	0.003729	0.01876	0.0117	110.15973	110.15973	110.15973	0	0	1	-360	360;
	5326	6554	0.000782	0.059938	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7428	5326	0.000769	0.061111	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	929	5909	0.000382	0.019378	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	929	348	0	0.000222	0	118.349126	118.349126	118.349126	0	0	1	-360	360;
	1009	6204	1.8e-05	0.00188	0	646.323939	646.323939	646.323939	0	0	1	-360	360;
	7686	5468	0.000372	0.001223	0.00896	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8200	5295	0.000809	0.001751	0.06632	167.703481	167.703481	167.703481	0	0	1	-360	360;
	3066	793	0	0.043511	0	144.540073	144.540073	144.540073	0	0	1	-360	360;
	6753	3066	0.003529	0.022178	0.02205	132.642962	132.642962	132.642962	0	0	1	-360	360;
	6753	3066	0.0036	0.021711	0.02246	173.189319	173.189319	173.189319	0	0	1	-360	360;
	5792	210	0	0.000413	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2030	5792	0.047479	0.102603	0.01748	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2030	5792	0.053727	0.091281	0.01976	57.157677	57.157677	57.157677	0	0	1	-360	360;
	919	5792	0.021678	0.045669	0.00626	71.558744	71.558744	71.558744	0	0	1	-360	360;
	2030	8059	0.004397	0.030678	0.00375	67.918181	67.918181	67.918181	0	0	1	-360	360;
	2030	8059	0.004314	0.031256	0.00367	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2030	6456	0.008347	0.055512	0.00715	171.47303	171.47303	171.47303	0	0	1	-360	360;
	2030	2446	0.006223	0.039744	0.00533	171.47303	171.47303	171.47303	0	0	1	-360	360;
	7503	2030	0.006124	0.041686	0.00522	70.813408	70.813408	70.813408	0	0	1	-360	360;
	7503	2030	0.006942	0.040851	0.00592	74.32689	74.32689	74.32689	0	0	1	-360	360;
	3696	5249	0.019248	0.039041	0.0181	194.614268	194.614268	194.614268	0	0	1	-360	360;
	6473	3696	0.015719	0.04943	0.00449	72.954153	72.954153	72.954153	0	0	1	-360	360;
	5037	1098	0.00296	0.011871	0.00357	126.914724	126.914724	126.914724	0	0	1	-360	360;
	5037	8246	0.01328	0.04144	0.01504	135.902769	135.902769	135.902769	0	0	1	-360	360;
	2489	6475	0.00736	0.033222	0.0115	153.783508	153.783508	153.783508	0	0	1	-360	360;
	2489	6475	0.007502	0.032529	0.01172	137.150625	137.150625	137.150625	0	0	1	-360	360;
	1522	2489	0.00192	0.02576	0.01769	311.769145	311.769145	311.769145	0	0	1	-360	360;
	1522	2489	0.002169	0.02524	0.02005	311.769145	311.769145	311.769145	0	0	1	-360	360;
	2489	7697	0.00656	0.017071	0.55531	141.600869	141.600869	141.600869	0	0	1	-360	360;
	2489	2491	0.002662	0.007471	0.2318	170.461339	170.461339	170.461339	0	0	1	-360	360;
	2489	2925	0.003191	0.009062	0.26848	187.703732	187.703732	187.703732	0	0	1	-360	360;
	6274	7862	0.00981	0.057298	0.00507	65.353897	65.353897	65.353897	0	0	1	-360	360;
	2890	5767	0.008342	0.037031	0.01238	186.713085	186.713085	186.713085	0	0	1	-360	360;
	1331	5767	0.0086	0.039791	0.00835	228.77464	228.77464	228.77464	0	0	1	-360	360;
	1541	7459	0	0.00046	0	1861.13034	1861.13034	1861.13034	0	0	1	-360	360;
	5508	1541	0.00055	0.00685	0.19546	1649.11696	1649.11696	1649.11696	0	0	1	-360	360;
	5685	7459	0.00054	0.00699	0.19138	1670.4446	1670.4446	1670.4446	0	0	1	-360	360;
	7997	60	0.0056	0.052889	0.02198	242.180973	242.180973	242.180973	0	0	1	-360	360;
	7997	5396	0.006058	0.023591	0.01851	186.931843	186.931843	186.931843	0	0	1	-360	360;
	7997	5396	0.0054	0.026729	0.01649	127.299239	127.299239	127.299239	0	0	1	-360	360;
	7997	7810	0.00168	0.01188	0.00681	311.769145	311.769145	311.769145	0	0	1	-360	360;
	7997	8179	0.003178	0.028658	0.00591	223.781657	223.781657	223.781657	0	0	1	-360	360;
	7997	4718	0.003178	0.028542	0.0116	311.769145	311.769145	311.769145	0	0	1	-360	360;
	1895	5844	0.00068	0.004449	0.08531	233.413245	233.413245	233.413245	0	0	1	-360	360;
	1895	3145	0.00076	0.004818	0.04133	166.055521	166.055521	166.055521	0	0	1	-360	360;
	2546	1626	0	0.000222	0	108.55334	108.55334	108.55334	0	0	1	-360	360;
	2546	1488	9e-06	0.000222	0	199.629681	199.629681	199.629681	0	0	1	-360	360;
	3628	1488	0	0.000222	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	6746	1488	1.8e-05	0.004329	0.73535	137.76429	137.76429	137.76429	0	0	1	-360	360;
	5488	4791	0.001831	0.001698	0.059	198.526278	198.526278	198.526278	0	0	1	-360	360;
	5488	4791	0.002209	0.001471	0.08081	196.002766	196.002766	196.002766	0	0	1	-360	360;
	190	2564	0.017653	0.070289	0.01037	84.685576	84.685576	84.685576	0	0	1	-360	360;
	3762	2564	0.013231	0.042207	0.00518	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3762	2564	0.013479	0.041331	0.00528	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2540	2564	0.013934	0.055702	0.02244	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3840	5471	0.000537	0.001471	0.0166	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5471	4434	0.009967	0.048959	0.00979	66.715202	66.715202	66.715202	0	0	1	-360	360;
	5471	8298	0.030612	0.055058	0.00594	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5471	8298	0.031207	0.053934	0.00605	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5498	5471	0.01314	0.040579	0.00386	71.278481	71.278481	71.278481	0	0	1	-360	360;
	5498	5471	0.012851	0.03757	0.00378	97.481655	97.481655	97.481655	0	0	1	-360	360;
	5471	6358	0.006901	0.021174	0.00202	58.986151	58.986151	58.986151	0	0	1	-360	360;
	5471	6358	0.007653	0.01924	0.00224	58.209187	58.209187	58.209187	0	0	1	-360	360;
	6613	5081	0.008982	0.038462	0.01685	108.275346	108.275346	108.275346	0	0	1	-360	360;
	5081	6615	0.009151	0.037649	0.01717	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5081	929	0.004258	0.03952	0.0266	251.414016	251.414016	251.414016	0	0	1	-360	360;
	2372	5081	0.003489	0.0264	0.02172	179.344941	179.344941	179.344941	0	0	1	-360	360;
	5081	2095	0.003849	0.05232	0.02447	180.254008	180.254008	180.254008	0	0	1	-360	360;
	1816	3736	0.032669	0.080273	0.00701	132.894075	132.894075	132.894075	0	0	1	-360	360;
	1728	3736	0.004901	0.014967	0.11843	140.194063	140.194063	140.194063	0	0	1	-360	360;
	284	5249	0.000562	0.001058	0.00858	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7605	123	0	0.00045	0	1548.45567	1548.45567	1548.45567	0	0	1	-360	360;
	7605	7490	0.001	0.01169	0.35326	1489.06421	1489.06421	1489.06421	0	0	1	-360	360;
	123	7442	0.00098	0.01192	0.34605	1462.15257	1462.15257	1462.15257	0	0	1	-360	360;
	5985	6864	0.002339	0.006223	0.00074	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5985	6864	0.002388	0.006091	0.00076	69.458961	69.458961	69.458961	0	0	1	-360	360;
	2313	5985	0.000504	0.001207	0.00764	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5249	6864	0.00105	0.007959	0.00218	247.971531	247.971531	247.971531	0	0	1	-360	360;
	5249	6864	0.001033	0.007372	0.00213	267.488781	267.488781	267.488781	0	0	1	-360	360;
	5249	2400	0.018752	0.042339	0.01616	75.353061	75.353061	75.353061	0	0	1	-360	360;
	5249	2400	0.018339	0.039231	0.01581	81.494082	81.494082	81.494082	0	0	1	-360	360;
	1816	6281	0.036793	0.081719	0.00902	59.071316	59.071316	59.071316	0	0	1	-360	360;
	1816	7686	0.003702	0.015132	0.00313	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2247	1816	0.00124	0.007314	0.1597	226.353164	226.353164	226.353164	0	0	1	-360	360;
	474	5227	0.002329	0.00356	0.13896	311.769145	311.769145	311.769145	0	0	1	-360	360;
	474	5227	0.002378	0.00348	0.14154	311.769145	311.769145	311.769145	0	0	1	-360	360;
	474	5227	0.00212	0.003951	0.1261	311.769145	311.769145	311.769145	0	0	1	-360	360;
	474	6779	0.001951	0.003791	0.11738	375.690654	375.690654	375.690654	0	0	1	-360	360;
	60	7123	0.002542	0.01472	0	149.022534	149.022534	149.022534	0	0	1	-360	360;
	1522	4854	0.003262	0.024151	0.02014	129.335871	129.335871	129.335871	0	0	1	-360	360;
	8355	4326	0.0036	0.01548	0.00517	265.621336	265.621336	265.621336	0	0	1	-360	360;
	8355	4326	0.00352	0.014329	0.00506	263.178625	263.178625	263.178625	0	0	1	-360	360;
	1721	3988	0.002591	0.014418	0	169.449648	169.449648	169.449648	0	0	1	-360	360;
	3209	1721	0.001071	0.003058	0.00107	163.590727	163.590727	163.590727	0	0	1	-360	360;
	3209	4260	9.8e-05	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3209	1211	0.017458	0.052129	0.01724	170.560326	170.560326	170.560326	0	0	1	-360	360;
	6613	7650	0.00136	0.00216	0.02863	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6615	7650	0.001391	0.00212	0.02919	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5806	8246	0.005858	0.016591	0.01086	115.50787	115.50787	115.50787	0	0	1	-360	360;
	811	7862	0.004562	0.011653	0.0085	94.437247	94.437247	94.437247	0	0	1	-360	360;
	811	5919	0.030868	0.064174	0.0067	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2962	332	0	0.000521	0	305.213229	305.213229	305.213229	0	0	1	-360	360;
	2162	929	0.005929	0.053271	0.04301	118.389397	118.389397	118.389397	0	0	1	-360	360;
	2162	270	0.00664	0.040849	0.0142	122.309374	122.309374	122.309374	0	0	1	-360	360;
	2162	2895	0.00968	0.049382	0.02356	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5289	4264	9.8e-05	0.000222	1	161.053185	161.053185	161.053185	0	0	1	-360	360;
	5289	60	0.00044	0.004489	0.00109	133.1592	133.1592	133.1592	0	0	1	-360	360;
	5289	3550	0.01064	0.043751	0.01106	110.785087	110.785087	110.785087	0	0	1	-360	360;
	5289	3550	0.0104	0.040511	0.01082	124.006697	124.006697	124.006697	0	0	1	-360	360;
	4718	5289	0.0024	0.025498	0.00897	130.386793	130.386793	130.386793	0	0	1	-360	360;
	3890	1096	0.00264	0.01412	0	106.924866	106.924866	106.924866	0	0	1	-360	360;
	426	5563	0.000609	0.03944	0	672.053467	672.053467	672.053467	0	0	1	-360	360;
	7766	426	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	426	5227	0.001182	0.003018	0.0975	311.769145	311.769145	311.769145	0	0	1	-360	360;
	426	5227	0.0012	0.002951	0.09931	303.24148	303.24148	303.24148	0	0	1	-360	360;
	426	5227	0.001071	0.003351	0.08847	292.202254	292.202254	292.202254	0	0	1	-360	360;
	426	6416	0.001031	0.01544	0.30937	172.067989	172.067989	172.067989	0	0	1	-360	360;
	426	314	0.001009	0.015738	0.30292	208.748149	208.748149	208.748149	0	0	1	-360	360;
	426	6491	0.000991	0.01604	0.29648	239.094718	239.094718	239.094718	0	0	1	-360	360;
	5627	6622	0.003898	0.008862	0.30829	256.984811	256.984811	256.984811	0	0	1	-360	360;
	5627	6622	0.003818	0.009031	0.30187	259.416611	259.416611	259.416611	0	0	1	-360	360;
	582	283	0.003831	0.012502	0.00417	112.187529	112.187529	112.187529	0	0	1	-360	360;
	5508	5685	0	0.00051	0	870.344065	870.344065	870.344065	0	0	1	-360	360;
	5461	5685	0.00127	0.0132	0.46307	1505.97876	1505.97876	1505.97876	0	0	1	-360	360;
	6753	2066	0	0.000222	0	309.919835	309.919835	309.919835	0	0	1	-360	360;
	6753	6955	0.001338	0.011209	0.29817	153.691796	153.691796	153.691796	0	0	1	-360	360;
	6753	6364	0.004222	0.030062	0.11194	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7970	6753	0.001182	0.0104	0.16346	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2066	3500	9e-06	0.000222	0	149.598527	149.598527	149.598527	0	0	1	-360	360;
	2066	6364	0.004138	0.03064	0.1096	119.619846	119.619846	119.619846	0	0	1	-360	360;
	3500	6364	0.004049	0.031209	0.10727	120.796254	120.796254	120.796254	0	0	1	-360	360;
	664	4923	0.022603	0.051959	0.03419	66.547349	66.547349	66.547349	0	0	1	-360	360;
	4196	664	0.013636	0.032041	0.09351	141.959283	141.959283	141.959283	0	0	1	-360	360;
	7697	2926	0.00228	0.009222	0.08473	136.488635	136.488635	136.488635	0	0	1	-360	360;
	2491	2926	0.002471	0.010302	0.0312	278.20304	278.20304	278.20304	0	0	1	-360	360;
	2925	2926	0.001831	0.007378	0.00439	273.315279	273.315279	273.315279	0	0	1	-360	360;
	8059	3918	0.006843	0.049248	0.00586	70.366244	70.366244	70.366244	0	0	1	-360	360;
	8059	3918	0.006686	0.045603	0.00573	63.192384	63.192384	63.192384	0	0	1	-360	360;
	2502	7271	0.004102	0.011529	0.1196	131.716488	131.716488	131.716488	0	0	1	-360	360;
	2502	7271	0.004182	0.01128	0.12181	110.269109	110.269109	110.269109	0	0	1	-360	360;
	1750	1933	0	0.000222	0	130.275335	130.275335	130.275335	0	0	1	-360	360;
	7390	1750	0.003649	0.0122	0.04313	130.123868	130.123868	130.123868	0	0	1	-360	360;
	7390	1750	0.00372	0.011938	0.04395	151.865609	151.865609	151.865609	0	0	1	-360	360;
	1750	6431	0.005031	0.028102	0.17151	144.201803	144.201803	144.201803	0	0	1	-360	360;
	1750	6431	0.005129	0.02752	0.17488	154.865347	154.865347	154.865347	0	0	1	-360	360;
	1750	7116	0.005218	0.026951	0.17824	134.189077	134.189077	134.189077	0	0	1	-360	360;
	4019	3145	0.00288	0.01992	0.12609	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4019	3145	0.003262	0.01952	0.1429	105.864591	105.864591	105.864591	0	0	1	-360	360;
	7926	1009	0.00956	0.061022	0.0234	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7926	2502	0.005382	0.033911	0.01853	137.307548	137.307548	137.307548	0	0	1	-360	360;
	7926	2502	0.005271	0.034569	0.01814	136.442649	136.442649	136.442649	0	0	1	-360	360;
	7926	5642	0.004102	0.023018	0.01105	129.118672	129.118672	129.118672	0	0	1	-360	360;
	1636	7801	0	0.00052	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	929	2122	0	0.000222	0	489.728792	489.728792	489.728792	0	0	1	-360	360;
	929	2895	0.011631	0.094591	0.06991	110.642193	110.642193	110.642193	0	0	1	-360	360;
	2372	929	0.0012	0.008769	0.0075	311.769145	311.769145	311.769145	0	0	1	-360	360;
	929	8237	0.001218	0.008582	0.00763	311.769145	311.769145	311.769145	0	0	1	-360	360;
	6475	6882	0.002689	0.013818	0	233.079392	233.079392	233.079392	0	0	1	-360	360;
	2400	7514	0.005	0.008934	0.16375	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2557	124	0	0.000531	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2557	4825	0.001281	0.0158	0.05906	251.837104	251.837104	251.837104	0	0	1	-360	360;
	124	3543	0.00125	0.016099	0.0578	302.29552	302.29552	302.29552	0	0	1	-360	360;
	190	5919	0.010058	0.048	0.00633	75.423174	75.423174	75.423174	0	0	1	-360	360;
	190	5919	0.010248	0.047	0.00646	75.252463	75.252463	75.252463	0	0	1	-360	360;
	190	2540	0.004314	0.024537	0.01472	77.119995	77.119995	77.119995	0	0	1	-360	360;
	1522	3000	0.010791	0.042151	0.0122	302.138856	302.138856	302.138856	0	0	1	-360	360;
	310	920	0.0006	0.040218	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	221	7818	0	0.000222	0	253.399986	253.399986	253.399986	0	0	1	-360	360;
	2921	5150	0	0.00053	0	1713.29142	1713.29142	1713.29142	0	0	1	-360	360;
	2921	3949	0	0.00054	0	1280.88933	1280.88933	1280.88933	0	0	1	-360	360;
	1522	7435	0.00196	0.010249	0.00356	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1522	1381	0.002182	0.010498	0.00348	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5396	5878	0.000582	0.040991	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5396	5132	0.000569	0.04176	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1331	2890	0.003818	0.013209	0.00468	258.743449	258.743449	258.743449	0	0	1	-360	360;
	5396	5737	0.00428	0.023511	0.01299	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5396	5737	0.004182	0.021769	0.01271	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1787	4421	0	0.0005	0	1114.05404	1114.05404	1114.05404	0	0	1	-360	360;
	1787	7490	0.00114	0.01661	0.77742	1510.86969	1510.86969	1510.86969	0	0	1	-360	360;
	3932	3869	0	0.000539	0	325.943556	325.943556	325.943556	0	0	1	-360	360;
	3932	66	0.00119	0.014081	0.07275	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4421	7442	0.00117	0.01626	0.79237	1716.57902	1716.57902	1716.57902	0	0	1	-360	360;
	3869	33	0.00106	0.01595	0.06482	243.303463	243.303463	243.303463	0	0	1	-360	360;
	3918	908	0.018058	0.065719	0.00603	116.884781	116.884781	116.884781	0	0	1	-360	360;
	3918	919	0.012826	0.025058	0.00285	78.962949	78.962949	78.962949	0	0	1	-360	360;
	3918	919	0.013331	0.024	0.00296	81.512944	81.512944	81.512944	0	0	1	-360	360;
	7435	1381	0.001089	0.003151	0.00139	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5256	2204	9.8e-05	0.000222	1	142.701155	142.701155	142.701155	0	0	1	-360	360;
	310	1162	0.000631	0.054471	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	310	1931	0.000622	0.055489	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	310	8076	0.0006	0.051382	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3762	5919	0.017702	0.064331	0.00483	78.984669	78.984669	78.984669	0	0	1	-360	360;
	3762	1635	0.012314	0.04924	0.00533	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2118	7595	0.001223	0.001248	0.00012	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4810	7595	0.001471	0.00481	0.00054	72.03601	72.03601	72.03601	0	0	1	-360	360;
	908	7595	0.011521	0.03138	0.00354	107.059186	107.059186	107.059186	0	0	1	-360	360;
	8200	7182	0	0.000222	0	113.042816	113.042816	113.042816	0	0	1	-360	360;
	8200	5578	0.000289	0.00056	0.02429	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3616	3062	0.00034	0.01805	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6783	2466	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1959	6064	0.00035	0.01766	0	1055.26481	1055.26481	1055.26481	0	0	1	-360	360;
	8131	3834	0.00035	0.01728	0	1079.40419	1079.40419	1079.40419	0	0	1	-360	360;
	4608	7540	0.00031	0.01958	0	1882.03543	1882.03543	1882.03543	0	0	1	-360	360;
	6591	3062	0.00031	0.01997	0	870.210454	870.210454	870.210454	0	0	1	-360	360;
	5826	3062	0.0003	0.02035	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5342	7454	0.00016	0.0025	1.39792	1037.56505	1037.56505	1037.56505	0	0	1	-360	360;
	4343	3062	0.00029	0.02074	0	1130.57763	1130.57763	1130.57763	0	0	1	-360	360;
	1212	7540	0.00029	0.0192	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7689	7540	0.00033	0.01882	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6064	3834	0	0.00049	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4304	6064	0	0.00048	0	1310.01508	1310.01508	1310.01508	0	0	1	-360	360;
	7127	3834	0	0.00047	0	1392.40597	1392.40597	1392.40597	0	0	1	-360	360;
	3062	5350	0	0.00046	0	1358.29582	1358.29582	1358.29582	0	0	1	-360	360;
	3062	378	0	0.00045	0	1117.24292	1117.24292	1117.24292	0	0	1	-360	360;
	3062	8078	0	0.00051	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7540	7127	0	0.00052	0	1005.39653	1005.39653	1005.39653	0	0	1	-360	360;
	7540	4304	0	0.00053	0	936.681957	936.681957	936.681957	0	0	1	-360	360;
	7540	8078	0	0.00054	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	5350	3721	0	0.0005	0	1461.33116	1461.33116	1461.33116	0	0	1	-360	360;
	3721	7127	0	0.00049	0	1869.72681	1869.72681	1869.72681	0	0	1	-360	360;
	3721	3713	0.00028	0.00403	0.19536	1402.9138	1402.9138	1402.9138	0	0	1	-360	360;
	5700	3721	0.00021	0.00396	0.33171	1798.37755	1798.37755	1798.37755	0	0	1	-360	360;
	378	2656	0	0.00048	0	1134.82552	1134.82552	1134.82552	0	0	1	-360	360;
	2656	4304	0	0.00047	0	1258.25059	1258.25059	1258.25059	0	0	1	-360	360;
	2656	5342	0.00029	0.00394	0.19898	977.352831	977.352831	977.352831	0	0	1	-360	360;
	4081	2656	0.0002	0.00404	0.32465	1771.32111	1771.32111	1771.32111	0	0	1	-360	360;
	2466	221	0	0.000222	0	115.435124	115.435124	115.435124	0	0	1	-360	360;
	221	503	0	0.000222	0	120.065935	120.065935	120.065935	0	0	1	-360	360;
	7697	2491	0.004471	0.020791	0.10896	121.848735	121.848735	121.848735	0	0	1	-360	360;
	7697	2925	0.002791	0.01132	0.09985	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5227	7604	0.000822	0.001649	0.04877	221.546532	221.546532	221.546532	0	0	1	-360	360;
	5227	7604	0.0008	0.00168	0.04773	216.064591	216.064591	216.064591	0	0	1	-360	360;
	5227	7604	0.000782	0.00156	0.04669	227.85648	227.85648	227.85648	0	0	1	-360	360;
	333	283	0.00432	0.012662	0.00518	103.923048	103.923048	103.923048	0	0	1	-360	360;
	333	283	0.003849	0.014351	0.00462	145.358467	145.358467	145.358467	0	0	1	-360	360;
	1009	2310	0.001418	0.004618	0.11025	148.924586	148.924586	148.924586	0	0	1	-360	360;
	1009	6596	0.001689	0.011631	0.00607	113.947985	113.947985	113.947985	0	0	1	-360	360;
	1009	5642	0.005991	0.03832	0.01384	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8355	1770	0.001169	0.007098	0.1704	407.092042	407.092042	407.092042	0	0	1	-360	360;
	6779	1770	0.0012	0.00696	0.17374	311.769145	311.769145	311.769145	0	0	1	-360	360;
	4021	8298	0.030066	0.046289	0.00582	74.512462	74.512462	74.512462	0	0	1	-360	360;
	4021	8298	0.026793	0.052463	0.00518	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7012	4021	0.023413	0.057397	0.03766	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5077	4872	0.001631	0.00832	0.00405	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5488	4035	0.012569	0.04732	0.01476	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5488	4035	0.012311	0.048231	0.01445	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7272	5913	0	0.000222	0	113.078669	113.078669	113.078669	0	0	1	-360	360;
	6364	7272	0.000982	0.004751	0.17983	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2694	1145	0.00144	0.12	0	104.359265	104.359265	104.359265	0	0	1	-360	360;
	310	4872	0.001249	0.0064	0.00311	148.979925	148.979925	148.979925	0	0	1	-360	360;
	310	4872	0.00112	0.007249	0.00277	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4825	3543	0	0.0005	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8354	7862	0.022843	0.098893	0.01078	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7862	5919	0.034909	0.072893	0.00815	58.24653	58.24653	58.24653	0	0	1	-360	360;
	6955	338	0.00188	0.092222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6955	1115	0.00184	0.093991	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6955	3649	0.0018	0.09576	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5326	6955	0.00132	0.01088	0.29214	135.075021	135.075021	135.075021	0	0	1	-360	360;
	5498	1728	0.01924	0.056603	0.00692	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5498	1728	0.017124	0.064149	0.00616	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3890	6119	9.8e-05	0.000222	1.5	247.83023	247.83023	247.83023	0	0	1	-360	360;
	3890	758	0.006889	0.03404	0.15314	174.407817	174.407817	174.407817	0	0	1	-360	360;
	3890	758	0.007018	0.033329	0.15608	129.478246	129.478246	129.478246	0	0	1	-360	360;
	3890	7214	0.006169	0.035689	0.01503	175.265442	175.265442	175.265442	0	0	1	-360	360;
	3890	7214	0.00628	0.034929	0.01531	153.896265	153.896265	153.896265	0	0	1	-360	360;
	3890	7977	0.004271	0.026418	0.01015	160.753108	160.753108	160.753108	0	0	1	-360	360;
	3890	7977	0.00436	0.02588	0.01035	151.461608	151.461608	151.461608	0	0	1	-360	360;
	3890	7977	0.00444	0.025338	0.01055	181.126442	181.126442	181.126442	0	0	1	-360	360;
	3550	1211	0.00852	0.02444	0.00838	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7810	44	0.002671	0.018929	0.01139	218.588103	218.588103	218.588103	0	0	1	-360	360;
	599	3044	0	0.00046	0	1615.18123	1615.18123	1615.18123	0	0	1	-360	360;
	599	6137	0	0.00045	0	1513.23781	1513.23781	1513.23781	0	0	1	-360	360;
	5461	3044	0	0.00051	0	1195.49846	1195.49846	1195.49846	0	0	1	-360	360;
	5461	6137	0	0.00052	0	1394.41802	1394.41802	1394.41802	0	0	1	-360	360;
	2310	6596	0.000329	0.000902	0.02604	133.37562	133.37562	133.37562	0	0	1	-360	360;
	7977	3376	0.00068	0.050351	0	120.947462	120.947462	120.947462	0	0	1	-360	360;
	3078	758	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1239	2284	0	0.00049	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1239	959	0.00212	0.022531	0.08223	558.121849	558.121849	558.121849	0	0	1	-360	360;
	1239	66	0.000919	0.012789	0.0563	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2284	3530	0.002159	0.02206	0.08381	438.808248	438.808248	438.808248	0	0	1	-360	360;
	2284	33	0.00094	0.012531	0.05741	215.929128	215.929128	215.929128	0	0	1	-360	360;
	174	6364	0.00032	0.014102	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	4409	6364	0.000329	0.0138	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	3358	6364	0.000329	0.013502	0	311.769145	311.769145	311.769145	0	0	1	-360	360;
	5737	487	0.005818	0.02396	0.01614	155.963814	155.963814	155.963814	0	0	1	-360	360;
	5737	487	0.005942	0.023462	0.01645	127.728441	127.728441	127.728441	0	0	1	-360	360;
	5326	6083	0.000671	0.001871	0.07215	267.429338	267.429338	267.429338	0	0	1	-360	360;
	4426	6083	0	0.001791	0	155.166205	155.166205	155.166205	0	0	1	-360	360;
	5326	3189	0.002249	0.003471	0.09006	224.441049	224.441049	224.441049	0	0	1	-360	360;
	5326	3189	0.001631	0.0042	0.0777	195.697493	195.697493	195.697493	0	0	1	-360	360;
	5326	3438	0.00092	0.002582	0.10491	184.633065	184.633065	184.633065	0	0	1	-360	360;
	6364	3189	0	0.000222	0	587.159508	587.159508	587.159508	0	0	1	-360	360;
	6364	5421	9.8e-05	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7970	6364	0.000822	0.00572	0.10682	158.482909	158.482909	158.482909	0	0	1	-360	360;
	6364	4154	0.001338	0.013622	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7310	6364	0.00136	0.004649	0.21168	251.24696	251.24696	251.24696	0	0	1	-360	360;
	3189	4426	0	0.001751	0	217.825308	217.825308	217.825308	0	0	1	-360	360;
	3438	3189	0	0.001711	0	184.633585	184.633585	184.633585	0	0	1	-360	360;
	3717	5844	0.000662	0.004529	0.00945	283.044555	283.044555	283.044555	0	0	1	-360	360;
	4768	5844	0.000649	0.004618	0.00925	235.994174	235.994174	235.994174	0	0	1	-360	360;
	5919	1635	0.005669	0.028992	0.00286	74.817684	74.817684	74.817684	0	0	1	-360	360;
	4198	6582	0.001631	0.1176	0	127.505526	127.505526	127.505526	0	0	1	-360	360;
	1110	7252	9.8e-05	0.000222	1.1	143.42316	143.42316	143.42316	0	0	1	-360	360;
	310	1110	0.006942	0.04516	0.01725	103.923048	103.923048	103.923048	0	0	1	-360	360;
	310	1110	0.0068	0.04604	0.0169	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7728	2446	0.003289	0.017909	0.00282	171.47303	171.47303	171.47303	0	0	1	-360	360;
	2925	6714	0.00056	0.038671	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4259	7012	0.02162	0.062612	0.00741	57.157677	57.157677	57.157677	0	0	1	-360	360;
	44	2259	9.8e-05	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	44	8179	0.000822	0.00656	0.00312	147.473041	147.473041	147.473041	0	0	1	-360	360;
	44	839	0.003009	0.004649	0.17188	281.091841	281.091841	281.091841	0	0	1	-360	360;
	44	839	0.003071	0.004551	0.17506	286.565468	286.565468	286.565468	0	0	1	-360	360;
	6335	4105	0.00464	0.029431	0.02262	108.443701	108.443701	108.443701	0	0	1	-360	360;
	6335	4105	0.005249	0.02884	0.02563	147.102815	147.102815	147.102815	0	0	1	-360	360;
	2838	5488	0.001902	0.012702	0.00812	129.95837	129.95837	129.95837	0	0	1	-360	360;
	2838	5488	0.001858	0.012938	0.00795	129.216879	129.216879	129.216879	0	0	1	-360	360;
	2838	6951	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5578	4937	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5694	3476	9.8e-05	0.000222	0.7	132.642962	132.642962	132.642962	0	0	1	-360	360;
	310	5459	9.8e-05	0.000222	0.5	120.10023	120.10023	120.10023	0	0	1	-360	360;
	8246	903	9.8e-05	0.000222	0.5	122.853411	122.853411	122.853411	0	0	1	-360	360;
	8355	2882	9.8e-05	0.000222	1.5	264.071844	264.071844	264.071844	0	0	1	-360	360;
	8355	758	0.004738	0.012529	0.18966	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8355	758	0.004831	0.012262	0.19323	103.923048	103.923048	103.923048	0	0	1	-360	360;
	8355	758	0.002711	0.018431	0.3923	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4923	8370	0.01205	0.03281	0.04818	105.364842	105.364842	105.364842	0	0	1	-360	360;
	3010	758	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	758	7057	0.000649	0.03712	0	393.895374	393.895374	393.895374	0	0	1	-360	360;
	758	4073	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5734	758	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4341	3294	0.00076	0.005329	0.00191	225.127461	225.127461	225.127461	0	0	1	-360	360;
	2895	270	0.005769	0.03528	0.01485	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6779	1330	0.002258	0.004849	0.13383	254.168237	254.168237	254.168237	0	0	1	-360	360;
	6779	1330	0.002209	0.004489	0.13092	261.243058	261.243058	261.243058	0	0	1	-360	360;
	959	3530	0	0.000479	0	250.392158	250.392158	250.392158	0	0	1	-360	360;
	959	5875	0.00256	0.01561	0.03976	548.981194	548.981194	548.981194	0	0	1	-360	360;
	3530	4809	0.00269	0.01451	0.04169	599.588601	599.588601	599.588601	0	0	1	-360	360;
	7503	5341	0.003372	0.028231	0.00994	74.803395	74.803395	74.803395	0	0	1	-360	360;
	7503	8281	0.007033	0.047083	0.01412	58.867453	58.867453	58.867453	0	0	1	-360	360;
	66	33	0	0.000469	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2895	1811	9.8e-05	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2502	2895	0.00392	0.025191	0.00916	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2502	2895	0.00444	0.024689	0.01038	138.02072	138.02072	138.02072	0	0	1	-360	360;
	8281	5341	0.003529	0.021223	0.00341	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3717	4768	0	0.000222	0	131.836779	131.836779	131.836779	0	0	1	-360	360;
	4081	356	0.00069	0.00134	0.04183	1499.75501	1499.75501	1499.75501	0	0	1	-360	360;
	5301	5700	0.00067	0.00136	0.04097	1459.88712	1459.88712	1459.88712	0	0	1	-360	360;
	2247	7686	0.012992	0.070207	0.01089	57.157677	57.157677	57.157677	0	0	1	-360	360;
	356	5301	0	0.00053	0	849.593646	849.593646	849.593646	0	0	1	-360	360;
	5875	4809	0	0.000461	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7490	7442	0	0.00054	0	810.24637	810.24637	810.24637	0	0	1	-360	360;
	6358	4196	0.002314	0.005661	0.09084	152.355121	152.355121	152.355121	0	0	1	-360	360;
	4196	6358	0.002512	0.005909	0.04977	125.833387	125.833387	125.833387	0	0	1	-360	360;
	5498	6358	0.006364	0.018488	0.00183	171.47303	171.47303	171.47303	0	0	1	-360	360;
	5498	6358	0.005959	0.017752	0.00172	171.47303	171.47303	171.47303	0	0	1	-360	360;
	5498	8370	0.018397	0.04176	0.00514	133.952064	133.952064	133.952064	0	0	1	-360	360;
	5498	8370	0.018058	0.042653	0.00505	149.306711	149.306711	149.306711	0	0	1	-360	360;
	6431	7116	0	0.000222	0	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2372	504	0	0.000529	0	123.323143	123.323143	123.323143	0	0	1	-360	360;
	8237	504	0	0.00052	0	119.572301	119.572301	119.572301	0	0	1	-360	360;
	2095	8237	0.003369	0.031329	0.02069	341.876432	341.876432	341.876432	0	0	1	-360	360;
	2272	5689	0.07449	0.527603	0	318.572502	318.572502	318.572502	0	0	1	-360	360;
	2764	5689	0.01093	0.074531	0.09945	440.586995	440.586995	440.586995	0	0	1	-360	360;
	4680	5689	0.00887	0.04576	0.07789	557.315164	557.315164	557.315164	0	0	1	-360	360;
	3560	5689	0.01099	0.06689	0.09638	305.6884	305.6884	305.6884	0	0	1	-360	360;
	7807	957	0.04732	0.42186	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7807	957	0.004469	0.03124	0.05173	241.251122	241.251122	241.251122	0	0	1	-360	360;
	957	388	0.500289	1.586157	0	217.044083	217.044083	217.044083	0	0	1	-360	360;
	957	3602	0.01089	0.060719	0.09686	473.369971	473.369971	473.369971	0	0	1	-360	360;
	957	4371	0.133469	0.882645	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7807	957	0.005531	0.032521	0.04721	495.166098	495.166098	495.166098	0	0	1	-360	360;
	957	4371	0.005219	0.025659	0.03021	506.640311	506.640311	506.640311	0	0	1	-360	360;
	7848	3502	0.00186	0.012789	0.0185	411.430483	411.430483	411.430483	0	0	1	-360	360;
	3635	3502	0.001	0.234938	0	242.206798	242.206798	242.206798	0	0	1	-360	360;
	4617	6077	0.00081	0.004781	0.00738	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2593	7410	0.00082	0.00468	0.00755	231.335027	231.335027	231.335027	0	0	1	-360	360;
	4168	4181	0.03374	0.614711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7740	4181	0.04612	0.366777	0	356.132336	356.132336	356.132336	0	0	1	-360	360;
	3027	4181	0.001461	0.01239	0.01711	233.779851	233.779851	233.779851	0	0	1	-360	360;
	7740	4168	0.078599	0.394649	0	389.862224	389.862224	389.862224	0	0	1	-360	360;
	7946	4168	0.00143	0.01263	0.01674	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6611	887	0.007841	0.05613	0.08965	553.892181	553.892181	553.892181	0	0	1	-360	360;
	4824	2224	0	0.000103	0	286.564584	286.564584	286.564584	0	0	1	-360	360;
	4824	2224	0.001	0.328946	0	577.670156	577.670156	577.670156	0	0	1	-360	360;
	3290	2224	0.224855	1.532376	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	665	2224	0.136781	1.457355	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	665	2224	0.01055	0.074979	0.10877	359.098819	359.098819	359.098819	0	0	1	-360	360;
	1347	2224	0.014909	0.523802	0	243.655555	243.655555	243.655555	0	0	1	-360	360;
	1680	2224	0.001	0.71781	0	229.254487	229.254487	229.254487	0	0	1	-360	360;
	6851	2224	0.134271	1.687004	0	241.277795	241.277795	241.277795	0	0	1	-360	360;
	3792	4824	0.002341	0.01263	0.02486	467.609239	467.609239	467.609239	0	0	1	-360	360;
	3290	4824	0.235227	1.541777	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	665	4824	0.16344	1.266136	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1347	4824	0.015479	0.562376	0	351.239639	351.239639	351.239639	0	0	1	-360	360;
	1680	4824	0.001	0.767789	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6851	4824	0.158281	1.596364	0	260.768182	260.768182	260.768182	0	0	1	-360	360;
	5318	3124	0.006169	0.039289	0.07504	561.653432	561.653432	561.653432	0	0	1	-360	360;
	5536	3124	0.08256	0.638182	0	290.353376	290.353376	290.353376	0	0	1	-360	360;
	2510	3124	0.221653	1.097397	0	269.75756	269.75756	269.75756	0	0	1	-360	360;
	3124	6084	0.0118	0.05445	0.09114	388.974375	388.974375	388.974375	0	0	1	-360	360;
	2729	3124	0.08986	0.756921	0	262.867774	262.867774	262.867774	0	0	1	-360	360;
	3645	577	0.03887	0.335083	0	225.618497	225.618497	225.618497	0	0	1	-360	360;
	577	3645	0.00455	0.036539	0.05724	465.617866	465.617866	465.617866	0	0	1	-360	360;
	153	577	0.002169	0.01688	0.02428	298.062423	298.062423	298.062423	0	0	1	-360	360;
	153	577	0.01045	0.244731	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5986	577	0.001	0.259525	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3282	577	0.036651	0.378657	0	354.140581	354.140581	354.140581	0	0	1	-360	360;
	5976	5704	0.004669	0.03839	0.05287	347.601743	347.601743	347.601743	0	0	1	-360	360;
	531	5704	0.000781	0.00674	0.00878	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3290	5704	0.001461	0.01201	0.0176	339.173653	339.173653	339.173653	0	0	1	-360	360;
	8380	5704	0.00087	0.00614	0.00967	605.821074	605.821074	605.821074	0	0	1	-360	360;
	8146	3792	0.00464	0.02919	0.05062	432.876805	432.876805	432.876805	0	0	1	-360	360;
	1680	3792	0.004719	0.02719	0.04704	484.823988	484.823988	484.823988	0	0	1	-360	360;
	5133	3792	0.00169	0.01361	0.0206	492.999441	492.999441	492.999441	0	0	1	-360	360;
	7243	3792	0.00751	0.04581	0.0563	564.403097	564.403097	564.403097	0	0	1	-360	360;
	1222	5976	0.00368	0.022469	0.04317	321.379325	321.379325	321.379325	0	0	1	-360	360;
	5133	5976	0.001659	0.012599	0.02016	512.365605	512.365605	512.365605	0	0	1	-360	360;
	6851	5976	0.007341	0.042419	0.05508	541.196699	541.196699	541.196699	0	0	1	-360	360;
	5976	6812	0.00186	0.01456	0.04413	527.331009	527.331009	527.331009	0	0	1	-360	360;
	5660	3060	0.00312	0.03151	0.803	1152.81883	1152.81883	1152.81883	0	0	1	-360	360;
	8004	5660	0.00073	0.00808	0.18882	1285.63546	1285.63546	1285.63546	0	0	1	-360	360;
	6506	2728	0.013479	0.27064	0	271.849531	271.849531	271.849531	0	0	1	-360	360;
	6506	2728	0.00187	0.020289	0.05356	219.981988	219.981988	219.981988	0	0	1	-360	360;
	8209	2728	0.012169	0.298636	0	284.578165	284.578165	284.578165	0	0	1	-360	360;
	6611	6292	0.00337	0.027831	0.03842	596.022343	596.022343	596.022343	0	0	1	-360	360;
	3560	6611	0.00461	0.03301	0.05272	537.723799	537.723799	537.723799	0	0	1	-360	360;
	6292	6611	0.005781	0.039789	0.06614	544.080876	544.080876	544.080876	0	0	1	-360	360;
	6292	6611	0.035151	0.987397	0	281.112504	281.112504	281.112504	0	0	1	-360	360;
	2077	6611	0.013099	0.631736	0	254.686986	254.686986	254.686986	0	0	1	-360	360;
	2764	2272	0.118409	1.072417	0	230.656756	230.656756	230.656756	0	0	1	-360	360;
	2764	2272	0.006281	0.051961	0.0717	475.209305	475.209305	475.209305	0	0	1	-360	360;
	1885	2272	0.284979	1.491054	0	228.636041	228.636041	228.636041	0	0	1	-360	360;
	3645	2957	0.006409	0.038081	0.07773	647.787002	647.787002	647.787002	0	0	1	-360	360;
	153	3645	0.225806	1.262624	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5986	3645	0.282252	1.121529	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3282	3645	0.561921	1.55469	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4971	3457	0.00322	0.02808	0.67532	1553.68491	1553.68491	1553.68491	0	0	1	-360	360;
	193	3457	0.00196	0.01843	0.4283	1284.45864	1284.45864	1284.45864	0	0	1	-360	360;
	6371	7657	0.552066	1.574607	0	252.431925	252.431925	252.431925	0	0	1	-360	360;
	3020	7657	0.010151	0.087021	0.1142	527.680814	527.680814	527.680814	0	0	1	-360	360;
	6616	7657	0.16999	0.991054	0	264.377499	264.377499	264.377499	0	0	1	-360	360;
	238	7657	0.003831	0.02386	0.04454	561.837479	561.837479	561.837479	0	0	1	-360	360;
	238	7657	0.00863	0.535785	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5598	7657	0.04357	0.617335	0	253.67339	253.67339	253.67339	0	0	1	-360	360;
	5204	2559	0.008081	0.03963	0.06647	441.086934	441.086934	441.086934	0	0	1	-360	360;
	2559	1680	0.00213	0.01488	0.02523	439.756303	439.756303	439.756303	0	0	1	-360	360;
	193	4123	0.00219	0.02232	0.47486	1249.01371	1249.01371	1249.01371	0	0	1	-360	360;
	4123	1197	0.00352	0.03947	0.74414	1011.82366	1011.82366	1011.82366	0	0	1	-360	360;
	4123	7613	0.00222	0.02857	0.57471	1489.5381	1489.5381	1489.5381	0	0	1	-360	360;
	325	5283	0.00104	0.00901	0.22878	1289.35549	1289.35549	1289.35549	0	0	1	-360	360;
	2500	325	0.00106	0.00881	0.23301	1474.34139	1474.34139	1474.34139	0	0	1	-360	360;
	6016	3690	0.00301	0.031271	0.08362	371.577102	371.577102	371.577102	0	0	1	-360	360;
	5283	2863	0.00133	0.01852	0.34169	1371.46402	1371.46402	1371.46402	0	0	1	-360	360;
	2863	4660	0.00253	0.0287	0.64674	965.460189	965.460189	965.460189	0	0	1	-360	360;
	5283	2863	0.06395	0.413802	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1197	8384	0.00075	0.00981	0.18449	1922.55953	1922.55953	1922.55953	0	0	1	-360	360;
	7160	2069	0.234814	1.148326	0	396.020773	396.020773	396.020773	0	0	1	-360	360;
	1373	2069	0.222417	1.185971	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7952	2069	0.08124	0.548574	0	426.18326	426.18326	426.18326	0	0	1	-360	360;
	7952	2069	0.01132	0.074211	0.12996	442.802426	442.802426	442.802426	0	0	1	-360	360;
	693	2069	0.115849	0.953099	0	357.041143	357.041143	357.041143	0	0	1	-360	360;
	4956	3877	0	3.5e-05	0	1283.8643	1283.8643	1283.8643	0	0	1	-360	360;
	4956	3877	0.001	0.517348	0	987.270277	987.270277	987.270277	0	0	1	-360	360;
	3877	6853	0.00051	0.00504	0.12794	1403.90963	1403.90963	1403.90963	0	0	1	-360	360;
	3877	4737	0.00258	0.02748	0.6791	1490.41216	1490.41216	1490.41216	0	0	1	-360	360;
	902	4956	0.00224	0.02852	0.57002	1489.29194	1489.29194	1489.29194	0	0	1	-360	360;
	4956	2500	0.00144	0.01774	0.36988	1433.99105	1433.99105	1433.99105	0	0	1	-360	360;
	5118	7601	0	0.000103	0	277.679233	277.679233	277.679233	0	0	1	-360	360;
	5118	7601	0.06794	0.65593	0	310.86498	310.86498	310.86498	0	0	1	-360	360;
	7601	6506	0.028479	0.136841	0.20849	479.877944	479.877944	479.877944	0	0	1	-360	360;
	5118	3690	0.02019	0.10464	0.17792	442.933508	442.933508	442.933508	0	0	1	-360	360;
	6853	2500	0.00128	0.01333	0.32562	1643.51256	1643.51256	1643.51256	0	0	1	-360	360;
	295	6195	0.667066	1.759628	0	265.442918	265.442918	265.442918	0	0	1	-360	360;
	4407	6195	0.0232	0.385289	0	307.122296	307.122296	307.122296	0	0	1	-360	360;
	4407	6195	0.00686	0.03882	0.07336	561.918643	561.918643	561.918643	0	0	1	-360	360;
	7125	6195	0.04143	0.328037	0	305.211324	305.211324	305.211324	0	0	1	-360	360;
	8363	6195	0.00107	0.00707	0.01049	469.100292	469.100292	469.100292	0	0	1	-360	360;
	4407	295	0.173021	1.092025	0	379.037703	379.037703	379.037703	0	0	1	-360	360;
	4407	295	0.010849	0.074771	0.12694	645.916803	645.916803	645.916803	0	0	1	-360	360;
	523	295	0.18632	0.723574	0	360.078883	360.078883	360.078883	0	0	1	-360	360;
	6992	295	0.077651	0.853368	0	321.281776	321.281776	321.281776	0	0	1	-360	360;
	6992	295	0.011169	0.07225	0.1297	488.506467	488.506467	488.506467	0	0	1	-360	360;
	7125	295	0.14268	0.506488	0	224.216229	224.216229	224.216229	0	0	1	-360	360;
	6616	295	0.352417	1.255207	0	451.830291	451.830291	451.830291	0	0	1	-360	360;
	238	295	0.391322	1.63	0	372.163921	372.163921	372.163921	0	0	1	-360	360;
	2928	2510	0.006271	0.037341	0.07279	559.869731	559.869731	559.869731	0	0	1	-360	360;
	3282	2928	0.005961	0.054719	0.18443	575.91351	575.91351	575.91351	0	0	1	-360	360;
	3195	4267	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3195	4267	0.001	0.333822	0	221.484854	221.484854	221.484854	0	0	1	-360	360;
	4267	3384	0.00587	0.040281	0.06694	428.632276	428.632276	428.632276	0	0	1	-360	360;
	3195	8209	0.003599	0.046841	0.09995	225.336138	225.336138	225.336138	0	0	1	-360	360;
	2678	4737	0.00224	0.02796	0.58727	1416.86786	1416.86786	1416.86786	0	0	1	-360	360;
	2510	5536	0.08693	0.670269	0	347.173442	347.173442	347.173442	0	0	1	-360	360;
	3290	5536	0.301839	1.820372	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1347	5536	0.00136	0.01051	0.01654	503.597236	503.597236	503.597236	0	0	1	-360	360;
	1347	5536	0.00445	0.344835	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7550	5536	0.197159	1.014752	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5204	5536	0.097341	0.78186	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2729	5536	0.013039	0.297211	0	447.344175	447.344175	447.344175	0	0	1	-360	360;
	5318	5536	0.035521	0.685351	0	299.113362	299.113362	299.113362	0	0	1	-360	360;
	3635	5536	0.367335	1.746467	0	250.782736	250.782736	250.782736	0	0	1	-360	360;
	2805	5536	0.04186	0.505517	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1347	5536	0.001021	0.284401	0	264.220124	264.220124	264.220124	0	0	1	-360	360;
	6084	5536	0.02575	0.548884	0	234.766393	234.766393	234.766393	0	0	1	-360	360;
	6812	5536	0.002211	0.016599	0.02619	495.947634	495.947634	495.947634	0	0	1	-360	360;
	4671	7848	0.0015	0.00876	0.01786	303.028282	303.028282	303.028282	0	0	1	-360	360;
	3635	4671	0.00062	0.00386	0.00743	531.831223	531.831223	531.831223	0	0	1	-360	360;
	6084	8146	0.00188	0.008651	0.01566	466.555632	466.555632	466.555632	0	0	1	-360	360;
	2805	8146	0.00651	0.278678	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5133	8146	0.080781	0.659587	0	259.504616	259.504616	259.504616	0	0	1	-360	360;
	4875	6224	0	0.000103	0	284.639133	284.639133	284.639133	0	0	1	-360	360;
	4875	6224	0.001	0.232851	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6224	2741	0.01614	0.08307	0.13914	450.481751	450.481751	450.481751	0	0	1	-360	360;
	388	6224	0.190479	0.901095	0	238.547945	238.547945	238.547945	0	0	1	-360	360;
	388	6224	0.00281	0.01663	0.02379	524.30089	524.30089	524.30089	0	0	1	-360	360;
	3602	6224	0.032841	0.593161	0	254.642403	254.642403	254.642403	0	0	1	-360	360;
	7740	6224	0.004599	0.03094	0.05184	485.52779	485.52779	485.52779	0	0	1	-360	360;
	4371	6224	0.100021	0.542231	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	388	4875	0.1945	0.883884	0	238.547945	238.547945	238.547945	0	0	1	-360	360;
	3602	4875	0.028409	0.617727	0	245.08221	245.08221	245.08221	0	0	1	-360	360;
	4875	3027	0.003281	0.04162	0.09109	349.383538	349.383538	349.383538	0	0	1	-360	360;
	4371	4875	0.09636	0.566219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4371	4875	0.007521	0.044719	0.04764	262.314869	262.314869	262.314869	0	0	1	-360	360;
	8209	6506	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8209	6506	0.001	0.19056	0	595.95985	595.95985	595.95985	0	0	1	-360	360;
	6506	3020	0.01356	0.08876	0.13214	552.562694	552.562694	552.562694	0	0	1	-360	360;
	8209	3020	0.005271	0.068591	0.14636	274.949383	274.949383	274.949383	0	0	1	-360	360;
	3092	7848	0.003591	0.03075	0.04138	647.787002	647.787002	647.787002	0	0	1	-360	360;
	3027	7848	0.01	0.056159	0.12256	410.839472	410.839472	410.839472	0	0	1	-360	360;
	1837	7848	0.011831	0.067969	0.10989	604.640196	604.640196	604.640196	0	0	1	-360	360;
	7946	7848	0.00937	0.05893	0.10475	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7848	6795	0.000909	0.004719	0.00955	500.196736	500.196736	500.196736	0	0	1	-360	360;
	2018	1197	0.00176	0.01759	0.1874	1450.10197	1450.10197	1450.10197	0	0	1	-360	360;
	2725	2018	0.00137	0.01801	0.33452	1544.30585	1544.30585	1544.30585	0	0	1	-360	360;
	2018	6919	0.00184	0.01756	0.45307	1295.29688	1295.29688	1295.29688	0	0	1	-360	360;
	6084	2805	0.00213	0.012539	0.02512	474.337079	474.337079	474.337079	0	0	1	-360	360;
	5133	2805	0.07664	0.597665	0	277.409068	277.409068	277.409068	0	0	1	-360	360;
	1347	2805	0.07132	0.544959	0	263.442018	263.442018	263.442018	0	0	1	-360	360;
	1347	2805	0.00405	0.02507	0.04892	508.119933	508.119933	508.119933	0	0	1	-360	360;
	6084	2805	0.001	0.348058	0	223.504044	223.504044	223.504044	0	0	1	-360	360;
	2729	2510	0.06064	0.52688	0	251.374889	251.374889	251.374889	0	0	1	-360	360;
	2510	2729	0.005271	0.03093	0.06438	524.328707	524.328707	524.328707	0	0	1	-360	360;
	3092	1222	0.002841	0.023409	0.03254	409.570191	409.570191	409.570191	0	0	1	-360	360;
	6322	1837	0.00226	0.015841	0.02118	468.890714	468.890714	468.890714	0	0	1	-360	360;
	5595	1837	0.01237	0.060909	0.09558	427.853789	427.853789	427.853789	0	0	1	-360	360;
	1837	3092	0.01637	0.07345	0.12628	495.440455	495.440455	495.440455	0	0	1	-360	360;
	3640	1837	0.00243	0.018341	0.02704	568.059664	568.059664	568.059664	0	0	1	-360	360;
	3139	717	0.0026	0.02813	0.66438	1194.14656	1194.14656	1194.14656	0	0	1	-360	360;
	4617	2593	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4617	2593	0.02631	0.515517	0	577.670156	577.670156	577.670156	0	0	1	-360	360;
	2593	6292	0.00394	0.041169	0.10702	216.570056	216.570056	216.570056	0	0	1	-360	360;
	2593	6292	0.343719	1.744855	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2593	7243	0.017159	0.080591	0.12617	471.929597	471.929597	471.929597	0	0	1	-360	360;
	4617	5004	0.005	0.066591	0.13873	439.265128	439.265128	439.265128	0	0	1	-360	360;
	4617	6292	0.065229	1.288988	0	264.64614	264.64614	264.64614	0	0	1	-360	360;
	665	3290	0	0.000103	0	265.193329	265.193329	265.193329	0	0	1	-360	360;
	665	3290	0.001	0.56	0	614.249926	614.249926	614.249926	0	0	1	-360	360;
	1347	3290	0.063599	0.651508	0	301.247248	301.247248	301.247248	0	0	1	-360	360;
	7550	3290	0.075159	1.271095	0	257.548299	257.548299	257.548299	0	0	1	-360	360;
	5204	3290	0.360475	1.617479	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1680	3290	0.001	0.740702	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2729	3290	0.363884	2.00314	0	266.469089	266.469089	266.469089	0	0	1	-360	360;
	5318	3290	0.3925	1.858533	0	263.333799	263.333799	263.333799	0	0	1	-360	360;
	1347	665	0.04769	0.602128	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7550	665	0.050341	1.259855	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1680	665	0.001	0.829566	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3020	7807	0.00049	0.00324	0.00582	540.646842	540.646842	540.646842	0	0	1	-360	360;
	7807	3444	0.018659	0.111289	0.21669	530.368368	530.368368	530.368368	0	0	1	-360	360;
	7807	4371	0.509835	1.831384	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3020	7807	0.0005	0.00318	0.00594	587.63312	587.63312	587.63312	0	0	1	-360	360;
	6084	1347	0.05936	0.582149	0	259.236356	259.236356	259.236356	0	0	1	-360	360;
	2729	1347	0.003331	0.0203	0.04069	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7550	1347	0.12286	0.888244	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5204	1347	0.033599	0.599959	0	255.796226	255.796226	255.796226	0	0	1	-360	360;
	1680	1347	0.100539	0.767066	0	307.166879	307.166879	307.166879	0	0	1	-360	360;
	2729	1347	0.0822	0.730331	0	270.740672	270.740672	270.740672	0	0	1	-360	360;
	5318	1347	0.15512	0.82874	0	268.337001	268.337001	268.337001	0	0	1	-360	360;
	4819	1358	0	0.000103	0	233.624382	233.624382	233.624382	0	0	1	-360	360;
	4819	1358	0.001	0.194781	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1358	2298	0.0132	0.071979	0.13541	486.451458	486.451458	486.451458	0	0	1	-360	360;
	1358	339	0.0148	0.073031	0.15524	511.920918	511.920918	511.920918	0	0	1	-360	360;
	4680	1358	0.005781	0.0378	0.07101	543.276858	543.276858	543.276858	0	0	1	-360	360;
	5595	1358	0.153961	1.590021	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5595	1358	0.00631	0.03855	0.0723	510.037382	510.037382	510.037382	0	0	1	-360	360;
	8102	1358	0.14124	1.705702	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4680	4819	0.00589	0.03699	0.07235	545.979654	545.979654	545.979654	0	0	1	-360	360;
	5595	4819	0.18032	1.449298	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8102	4819	0.150651	1.729298	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8102	4819	0.00636	0.045409	0.06238	455.830947	455.830947	455.830947	0	0	1	-360	360;
	2566	3425	0.00173	0.01602	0.37844	1318.88536	1318.88536	1318.88536	0	0	1	-360	360;
	4214	2566	0.00207	0.02233	0.54858	1230.7229	1230.7229	1230.7229	0	0	1	-360	360;
	1312	2566	0.00178	0.02178	0.46938	1307.29812	1307.29812	1307.29812	0	0	1	-360	360;
	3640	6322	0.00589	0.214174	0	257.809319	257.809319	257.809319	0	0	1	-360	360;
	5595	6322	0.105789	1.66719	0	223.90491	223.90491	223.90491	0	0	1	-360	360;
	8102	6322	0.108969	1.629731	0	222.696216	222.696216	222.696216	0	0	1	-360	360;
	7125	4407	0.024581	0.231446	0	216.906143	216.906143	216.906143	0	0	1	-360	360;
	7125	4407	0.004599	0.035341	0.05274	582.38414	582.38414	582.38414	0	0	1	-360	360;
	4660	1006	0.00106	0.01023	0.2705	771.521074	771.521074	771.521074	0	0	1	-360	360;
	1013	1006	0.00403	0.04889	1.04974	1816.48077	1816.48077	1816.48077	0	0	1	-360	360;
	5616	4996	0.00123	0.01589	0.33156	1424.65807	1424.65807	1424.65807	0	0	1	-360	360;
	5616	3060	0.00158	0.01592	0.38988	1929.01363	1929.01363	1929.01363	0	0	1	-360	360;
	226	6430	0.00219	0.02123	0.54184	1610.15735	1610.15735	1610.15735	0	0	1	-360	360;
	226	3865	0.00171	0.01708	0.42294	1568.15037	1568.15037	1568.15037	0	0	1	-360	360;
	226	3865	0.00174	0.0167	0.43077	1595.61949	1595.61949	1595.61949	0	0	1	-360	360;
	1013	226	0.00515	0.285942	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1013	226	0.00058	0.00775	0.14508	1678.43621	1678.43621	1678.43621	0	0	1	-360	360;
	1013	226	0.00057	0.0079	0.14199	1441.75888	1441.75888	1441.75888	0	0	1	-360	360;
	3092	6084	0.009341	0.6725	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3635	6084	0.171229	1.013678	0	265.199045	265.199045	265.199045	0	0	1	-360	360;
	6812	6084	0.00251	0.02049	0.02983	413.065192	413.065192	413.065192	0	0	1	-360	360;
	6084	6795	0.005909	0.036349	0.06605	593.802339	593.802339	593.802339	0	0	1	-360	360;
	8198	6919	0.00091	0.00886	0.23265	1282.47028	1282.47028	1282.47028	0	0	1	-360	360;
	8004	8198	0.00054	0.00532	0.13846	1577.81442	1577.81442	1577.81442	0	0	1	-360	360;
	238	6371	0.009531	0.051211	0.08945	487.799617	487.799617	487.799617	0	0	1	-360	360;
	238	6371	0.785124	1.943678	0	227.673887	227.673887	227.673887	0	0	1	-360	360;
	6371	3447	0.010789	0.05194	0.09025	529.003824	529.003824	529.003824	0	0	1	-360	360;
	5598	6371	0.173281	0.62781	0	263.204623	263.204623	263.204623	0	0	1	-360	360;
	5133	3092	0.00613	0.033659	0.05444	402.093205	402.093205	402.093205	0	0	1	-360	360;
	3635	3092	0.11576	0.840434	0	284.871955	284.871955	284.871955	0	0	1	-360	360;
	6992	523	0	0.000103	0	266.567781	266.567781	266.567781	0	0	1	-360	360;
	6992	523	0.05964	0.314938	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	523	1152	0.002169	0.017341	0.02439	454.597484	454.597484	454.597484	0	0	1	-360	360;
	2360	523	0.454215	1.265475	0	272.22258	272.22258	272.22258	0	0	1	-360	360;
	523	2360	0.007091	0.061521	0.08131	541.029418	541.029418	541.029418	0	0	1	-360	360;
	6992	1569	0.013021	0.087711	0.15121	594.917675	594.917675	594.917675	0	0	1	-360	360;
	877	6992	0.35781	1.954277	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2360	6992	0.0485	0.815537	0	331.698572	331.698572	331.698572	0	0	1	-360	360;
	1680	5004	0.415537	1.939483	0	253.033224	253.033224	253.033224	0	0	1	-360	360;
	5133	5004	0.097031	0.677913	0	280.847293	280.847293	280.847293	0	0	1	-360	360;
	5004	5133	0.003031	0.033021	0.08406	465.302355	465.302355	465.302355	0	0	1	-360	360;
	5004	5133	0.00637	0.038979	0.07806	581.49972	581.49972	581.49972	0	0	1	-360	360;
	6851	5004	0.024271	0.588967	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7243	5004	0.00439	0.372438	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	191	5004	0.001	0.198331	0	249.564896	249.564896	249.564896	0	0	1	-360	360;
	191	5004	0.00232	0.01226	0.02388	389.552048	389.552048	389.552048	0	0	1	-360	360;
	3447	1569	0.014211	0.079219	0.11648	647.787002	647.787002	647.787002	0	0	1	-360	360;
	153	1569	0.015151	0.07637	0.14567	539.436624	539.436624	539.436624	0	0	1	-360	360;
	2957	1569	0	0.000103	0	276.425956	276.425956	276.425956	0	0	1	-360	360;
	2957	1569	0.001	0.258698	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5986	1569	0.01544	0.074781	0.14847	560.79721	560.79721	560.79721	0	0	1	-360	360;
	877	2957	0.013909	0.073349	0.11395	644.851003	644.851003	644.851003	0	0	1	-360	360;
	4891	3581	0.006919	0.04764	0.08176	448.286134	448.286134	448.286134	0	0	1	-360	360;
	2049	3581	0.224318	1.85469	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7952	3581	0.015169	0.1077	0.17215	485.935514	485.935514	485.935514	0	0	1	-360	360;
	339	3581	0.130771	1.06564	0	265.831209	265.831209	265.831209	0	0	1	-360	360;
	4891	3581	0.249669	1.169917	0	226.300579	226.300579	226.300579	0	0	1	-360	360;
	3425	451	0.00127	0.01444	0.31414	1404.01296	1404.01296	1404.01296	0	0	1	-360	360;
	3425	5545	0.00318	0.04235	0.79571	1287.07227	1287.07227	1287.07227	0	0	1	-360	360;
	3425	7876	0.00131	0.01758	0.33385	1314.72962	1314.72962	1314.72962	0	0	1	-360	360;
	2077	2764	0.469607	1.302025	0	263.049154	263.049154	263.049154	0	0	1	-360	360;
	1885	2764	0.001	0.205711	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1885	2764	0.002969	0.02201	0.02957	519.457349	519.457349	519.457349	0	0	1	-360	360;
	8363	7125	0.00549	0.044831	0.06385	495.576109	495.576109	495.576109	0	0	1	-360	360;
	3099	2049	0	0.000103	0	374.681907	374.681907	374.681907	0	0	1	-360	360;
	3099	2049	0.08343	1.583099	0	605.110032	605.110032	605.110032	0	0	1	-360	360;
	2298	2049	0.094651	1.363719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2049	2741	0.00112	0.00895	0.01192	560.387961	560.387961	560.387961	0	0	1	-360	360;
	2741	2049	0.00151	0.00944	0.01193	561.801661	561.801661	561.801661	0	0	1	-360	360;
	339	2049	0.018669	0.563368	0	252.92653	252.92653	252.92653	0	0	1	-360	360;
	4891	2049	0.00538	0.652851	0	240.120162	240.120162	240.120162	0	0	1	-360	360;
	2298	3099	0	0.000103	0	252.133943	252.133943	252.133943	0	0	1	-360	360;
	2298	3099	0.001	0.495455	0	595.95985	595.95985	595.95985	0	0	1	-360	360;
	339	3099	0.036849	0.734215	0	231.537365	231.537365	231.537365	0	0	1	-360	360;
	4891	3099	0.001	0.459566	0	237.233699	237.233699	237.233699	0	0	1	-360	360;
	4891	3099	0.00095	0.00638	0.01135	547.459656	547.459656	547.459656	0	0	1	-360	360;
	2741	2298	0.00105	0.00818	0.01211	529.820035	529.820035	529.820035	0	0	1	-360	360;
	339	2298	0.040151	0.657893	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	339	2298	0.00193	0.01406	0.02108	521.559227	521.559227	521.559227	0	0	1	-360	360;
	4891	2298	0.001	0.49438	0	254.534947	254.534947	254.534947	0	0	1	-360	360;
	5204	7550	0	0.000103	0	323.310493	323.310493	323.310493	0	0	1	-360	360;
	5204	7550	0.08057	0.906178	0	577.670156	577.670156	577.670156	0	0	1	-360	360;
	2729	7550	0.0347	0.658202	0	303.977861	303.977861	303.977861	0	0	1	-360	360;
	5318	7550	0.032719	0.534318	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5318	7550	0.004281	0.033461	0.05199	567.927058	567.927058	567.927058	0	0	1	-360	360;
	2729	5204	0.001	0.395785	0	384.309927	384.309927	384.309927	0	0	1	-360	360;
	5318	5204	0.14007	0.838595	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7876	451	0.04194	0.384211	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	451	7876	0.00034	0.00341	0.08387	1440.45569	1440.45569	1440.45569	0	0	1	-360	360;
	699	7160	0.001969	0.014021	0.02462	377.121016	377.121016	377.121016	0	0	1	-360	360;
	1373	7160	0	0.000103	0	273.981513	273.981513	273.981513	0	0	1	-360	360;
	1373	7160	0.001	0.26219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7160	3589	0.03032	0.14413	0.26716	405.113035	405.113035	405.113035	0	0	1	-360	360;
	1373	693	0.01325	0.11038	0.15014	477.926962	477.926962	477.926962	0	0	1	-360	360;
	2331	2741	0.011901	0.07319	0.13775	491.149057	491.149057	491.149057	0	0	1	-360	360;
	2741	3444	0.00401	0.03175	0.04227	498.094476	498.094476	498.094476	0	0	1	-360	360;
	7876	2725	0.00368	0.03712	0.91191	1610.87345	1610.87345	1610.87345	0	0	1	-360	360;
	7876	2678	0.00169	0.02158	0.4329	1292.82081	1292.82081	1292.82081	0	0	1	-360	360;
	717	4214	0.00116	0.01472	0.2961	1286.22914	1286.22914	1286.22914	0	0	1	-360	360;
	193	6430	0.05575	0.602438	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6430	193	0.00163	0.01604	0.42269	1413.57762	1413.57762	1413.57762	0	0	1	-360	360;
	6430	2725	0.00269	0.02759	0.67961	1251.49307	1251.49307	1251.49307	0	0	1	-360	360;
	6430	1197	0.00256	0.02877	0.76243	1336.10663	1336.10663	1336.10663	0	0	1	-360	360;
	693	7952	0	0.000103	0	306.959587	306.959587	306.959587	0	0	1	-360	360;
	693	7952	0.09549	0.768967	0	614.249926	614.249926	614.249926	0	0	1	-360	360;
	3602	388	0	0.000103	0	267.716269	267.716269	267.716269	0	0	1	-360	360;
	3602	388	0.05899	0.428306	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3027	388	0.005591	0.04031	0.07402	612.017728	612.017728	612.017728	0	0	1	-360	360;
	4371	388	0.00714	0.15039	0	343.067615	343.067615	343.067615	0	0	1	-360	360;
	7946	3602	0.005469	0.041099	0.07251	606.732929	606.732929	606.732929	0	0	1	-360	360;
	4371	3602	0.036711	0.380269	0	281.411629	281.411629	281.411629	0	0	1	-360	360;
	4410	193	0.00056	0.124578	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4410	193	0.0007	0.00889	0.17364	1662.16668	1662.16668	1662.16668	0	0	1	-360	360;
	3020	2331	0.013479	0.11436	0.15718	542.738051	542.738051	542.738051	0	0	1	-360	360;
	3589	3020	0.012	0.066919	0.10486	555.1748	555.1748	555.1748	0	0	1	-360	360;
	3444	2331	0.3925	1.654463	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3640	6292	0.301178	1.948719	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3640	6292	0.009169	0.06301	0.10469	525.960368	525.960368	525.960368	0	0	1	-360	360;
	3060	717	0.00176	0.02342	0.42947	882.762592	882.762592	882.762592	0	0	1	-360	360;
	3343	717	0.00108	0.01167	0.27572	1497.68701	1497.68701	1497.68701	0	0	1	-360	360;
	3444	339	0.191961	1.234649	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4891	339	0.03699	0.374917	0	249.644917	249.644917	249.644917	0	0	1	-360	360;
	7946	7740	0.00361	0.03024	0.04313	625.025291	625.025291	625.025291	0	0	1	-360	360;
	238	6616	0.350806	1.554649	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6616	238	0.004831	0.06375	0.13403	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6616	2311	0.007099	0.07076	0.19694	256.472973	256.472973	256.472973	0	0	1	-360	360;
	5598	238	0.0022	0.01318	0.01917	391.693937	391.693937	391.693937	0	0	1	-360	360;
	238	877	0.01775	0.099419	0.15277	454.023621	454.023621	454.023621	0	0	1	-360	360;
	5598	238	0.186581	1.067335	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2725	2678	0.00373	0.05005	0.92844	1344.763	1344.763	1344.763	0	0	1	-360	360;
	3447	877	0	0.000103	0	235.279669	235.279669	235.279669	0	0	1	-360	360;
	3447	877	0.001	0.16038	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2360	877	0.010531	0.17819	0	273.79213	273.79213	273.79213	0	0	1	-360	360;
	2360	877	0.0025	0.017081	0.02869	505.007126	505.007126	505.007126	0	0	1	-360	360;
	2360	3447	0.0777	0.498492	0	236.015097	236.015097	236.015097	0	0	1	-360	360;
	6851	1680	0.01907	1.485992	0	257.996797	257.996797	257.996797	0	0	1	-360	360;
	7243	1680	0.120669	1.890888	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8102	4680	0.010271	0.06219	0.09021	416.671842	416.671842	416.671842	0	0	1	-360	360;
	5595	3640	0.298079	1.623781	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8102	3640	0.273223	1.612107	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2725	8004	0.00277	0.02823	0.68369	1480.20116	1480.20116	1480.20116	0	0	1	-360	360;
	1197	2725	0.00077	0.00747	0.38941	1251.62997	1251.62997	1251.62997	0	0	1	-360	360;
	6919	7613	0.00102	0.01145	0.25151	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6919	7613	0.00112	0.01049	0.29081	767.876077	767.876077	767.876077	0	0	1	-360	360;
	8102	5595	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8102	5595	0.001	0.22343	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	4996	3343	0	3.5e-05	0	1815.10254	1815.10254	1815.10254	0	0	1	-360	360;
	4996	3343	0.001	0.26178	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	3060	3343	0.00153	0.0214	0.37463	840.606207	840.606207	840.606207	0	0	1	-360	360;
	8004	3343	0.00468	0.06143	1.14105	1011.67688	1011.67688	1011.67688	0	0	1	-360	360;
	191	5133	0.271756	1.50345	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7243	6851	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7243	6851	0.015271	0.796736	0	577.670156	577.670156	577.670156	0	0	1	-360	360;
	191	6851	0.02776	0.935537	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	191	6851	0.004219	0.026599	0.03327	417.085282	417.085282	417.085282	0	0	1	-360	360;
	191	7243	0.001	0.620888	0	228.285474	228.285474	228.285474	0	0	1	-360	360;
	4891	3444	0.361364	1.973368	0	239.774167	239.774167	239.774167	0	0	1	-360	360;
	3560	2077	0.001599	0.00825	0.01395	478.218847	478.218847	478.218847	0	0	1	-360	360;
	1885	2077	0.659194	1.990062	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	5986	153	0	0.000103	0	262.053849	262.053849	262.053849	0	0	1	-360	360;
	5986	153	0.11788	0.822417	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3282	153	0.086341	0.745145	0	284.018782	284.018782	284.018782	0	0	1	-360	360;
	3282	5986	0.00569	0.033281	0.06485	540.443742	540.443742	540.443742	0	0	1	-360	360;
	3282	5986	0.03618	0.609897	0	287.947038	287.947038	287.947038	0	0	1	-360	360;
	22	4410	0.00087	0.00898	0.29637	1726.946	1726.946	1726.946	0	0	1	-360	360;
	125	3384	0.001	0.211322	0	641.690183	641.690183	641.690183	0	0	1	-360	360;
	125	3384	0	0.000103	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3589	3384	0.13407	1.029876	0	256.926424	256.926424	256.926424	0	0	1	-360	360;
	1678	3384	0.141419	0.956033	0	253.307962	253.307962	253.307962	0	0	1	-360	360;
	3589	125	0.12913	1.151054	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1678	125	0.14487	1.008244	0	258.958189	258.958189	258.958189	0	0	1	-360	360;
	125	1678	0.007599	0.05356	0.08699	378.584633	378.584633	378.584633	0	0	1	-360	360;
	5318	2729	0	0.000103	0	223.433931	223.433931	223.433931	0	0	1	-360	360;
	5318	2729	0.024219	0.488657	0	586.819957	586.819957	586.819957	0	0	1	-360	360;
	1678	3589	0.001	0.226901	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1678	3589	0	0.000103	0	265.734422	265.734422	265.734422	0	0	1	-360	360;
	6795	3635	0.00063	0.005159	0.00728	500.526345	500.526345	500.526345	0	0	1	-360	360;
	3690	2311	0	0.000103	0	249.578233	249.578233	249.578233	0	0	1	-360	360;
	3690	2311	0.001	0.214277	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3034	3879	0.019671	0.07004	0.02161	249.039894	249.039894	249.039894	0	0	1	-360	360;
	3034	3879	0.01812	0.06872	0.02006	247.96481	247.96481	247.96481	0	0	1	-360	360;
	5268	4293	0.001582	0.00724	0.00286	134.460057	134.460057	134.460057	0	0	1	-360	360;
	5268	4293	0.001551	0.007231	0.00276	121.158946	121.158946	121.158946	0	0	1	-360	360;
	5268	4293	0.001262	0.00796	0.0023	116.027745	116.027745	116.027745	0	0	1	-360	360;
	4946	2480	0.0264	0.075129	0.02967	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4946	4794	0.00824	0.044751	0.01738	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7150	4293	0.018271	0.070191	0.01993	143.597231	143.597231	143.597231	0	0	1	-360	360;
	7150	4293	0.021711	0.063129	0.0238	111.208314	111.208314	111.208314	0	0	1	-360	360;
	7150	6867	0.013191	0.0428	0.01355	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7344	2480	0.027631	0.102502	0.04257	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7344	6671	0.005369	0.052622	0.03551	115.511508	115.511508	115.511508	0	0	1	-360	360;
	7344	6671	0.013938	0.0534	0.01605	113.675187	113.675187	113.675187	0	0	1	-360	360;
	7344	6671	0.01364	0.049449	0.01571	117.107246	117.107246	117.107246	0	0	1	-360	360;
	7344	1280	0.007991	0.037738	0.06092	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2035	4293	0.007671	0.04096	0.01464	128.6757	128.6757	128.6757	0	0	1	-360	360;
	2035	4293	0.007511	0.041769	0.01434	121.205451	121.205451	121.205451	0	0	1	-360	360;
	2035	3115	0.004089	0.010978	0.01881	143.5315	143.5315	143.5315	0	0	1	-360	360;
	2035	3115	0.003	0.014769	0.02184	154.995251	154.995251	154.995251	0	0	1	-360	360;
	2035	553	0.001351	0.00784	0.00259	114.602441	114.602441	114.602441	0	0	1	-360	360;
	2035	859	0.003671	0.022089	0.00706	112.212211	112.212211	112.212211	0	0	1	-360	360;
	2035	2683	0.002378	0.00772	0.00257	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4293	553	0.006911	0.03268	0.01324	145.811831	145.811831	145.811831	0	0	1	-360	360;
	4293	1950	0.00944	0.03284	0.01053	156.430428	156.430428	156.430428	0	0	1	-360	360;
	4293	5907	0.009831	0.02848	0.01072	103.923048	103.923048	103.923048	0	0	1	-360	360;
	4293	859	0.003671	0.02092	0.00701	157.699329	157.699329	157.699329	0	0	1	-360	360;
	7975	2480	0.019378	0.10932	0.03713	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7975	2480	0.021969	0.107129	0.04209	127.20415	127.20415	127.20415	0	0	1	-360	360;
	7975	4794	0.009151	0.063249	0.01932	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7975	3746	0.029169	0.088102	0.0328	153.837029	153.837029	153.837029	0	0	1	-360	360;
	7975	4431	0.031351	0.115698	0.03597	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7975	6671	0.04936	0.182018	0.0567	118.57438	118.57438	118.57438	0	0	1	-360	360;
	7975	5740	0.001609	0.016889	0.01089	311.769145	311.769145	311.769145	0	0	1	-360	360;
	7975	2551	0	0.000222	0	135.397183	135.397183	135.397183	0	0	1	-360	360;
	637	8105	0.008698	0.033591	0.00959	201.960415	201.960415	201.960415	0	0	1	-360	360;
	5008	1703	0.01792	0.070169	0.01991	118.27612	118.27612	118.27612	0	0	1	-360	360;
	5008	7826	0.00308	0.008422	0.00345	311.769145	311.769145	311.769145	0	0	1	-360	360;
	5008	498	0.002031	0.012102	0.01557	155.210632	155.210632	155.210632	0	0	1	-360	360;
	5008	498	0.0068	0.03452	0.01063	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5008	3451	0.01176	0.055871	0.02218	184.905604	184.905604	184.905604	0	0	1	-360	360;
	3956	5907	0.04856	0.135111	0.05298	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2480	5614	0.039289	0.146129	0.04205	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2480	5614	0.038471	0.148938	0.04117	121.068273	121.068273	121.068273	0	0	1	-360	360;
	2480	3101	0.010911	0.04116	0.01192	250.80087	250.80087	250.80087	0	0	1	-360	360;
	5614	1280	0.002898	0.017409	0.02202	208.66527	208.66527	208.66527	0	0	1	-360	360;
	4554	1135	0.00544	0.051431	0.03503	126.147512	126.147512	126.147512	0	0	1	-360	360;
	4554	4844	0.003791	0.014342	0.0042	189.350133	189.350133	189.350133	0	0	1	-360	360;
	4554	5376	0.005969	0.026302	0.01126	108.136608	108.136608	108.136608	0	0	1	-360	360;
	4554	4338	0.00608	0.02572	0.01147	122.69259	122.69259	122.69259	0	0	1	-360	360;
	4554	2411	0.010818	0.066769	0.02286	260.439993	260.439993	260.439993	0	0	1	-360	360;
	1703	2185	0.006649	0.024138	0.0076	136.837556	136.837556	136.837556	0	0	1	-360	360;
	1703	2185	0.007542	0.023662	0.00861	147.089825	147.089825	147.089825	0	0	1	-360	360;
	1703	2185	0.007991	0.036622	0.01082	144.150361	144.150361	144.150361	0	0	1	-360	360;
	1703	2185	0.0116	0.046018	0.01267	106.441624	106.441624	106.441624	0	0	1	-360	360;
	3746	1950	0.011631	0.047049	0.01341	159.687377	159.687377	159.687377	0	0	1	-360	360;
	4431	6671	0.018822	0.064009	0.02166	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1135	5376	0.008071	0.049138	0.01524	108.365499	108.365499	108.365499	0	0	1	-360	360;
	1135	4338	0.007889	0.045502	0.01491	110.382645	110.382645	110.382645	0	0	1	-360	360;
	1135	7504	0.009822	0.057222	0.01832	183.818309	183.818309	183.818309	0	0	1	-360	360;
	2185	3375	0.00656	0.023831	0.00751	118.238448	118.238448	118.238448	0	0	1	-360	360;
	2185	3375	0.007431	0.023351	0.00851	118.866403	118.866403	118.866403	0	0	1	-360	360;
	2185	6717	0.007391	0.026978	0.00947	115.758585	115.758585	115.758585	0	0	1	-360	360;
	2185	3895	0.001871	0.012142	0.01489	134.153224	134.153224	134.153224	0	0	1	-360	360;
	2185	896	0.003969	0.026169	0.00779	119.319768	119.319768	119.319768	0	0	1	-360	360;
	2185	2009	0.008151	0.04664	0.06245	299.425166	299.425166	299.425166	0	0	1	-360	360;
	7430	34	0.00076	0.005622	0.00181	103.923048	103.923048	103.923048	0	0	1	-360	360;
	7430	3375	0.0094	0.035791	0.01011	146.850542	146.850542	146.850542	0	0	1	-360	360;
	7430	7067	0.00844	0.031031	0.00921	227.193711	227.193711	227.193711	0	0	1	-360	360;
	7430	1537	0.00028	0.002062	0.00069	141.617757	141.617757	141.617757	0	0	1	-360	360;
	7430	2623	0.004849	0.030031	0.01219	113.232735	113.232735	113.232735	0	0	1	-360	360;
	7430	3895	0.00268	0.022391	0.01774	152.859113	152.859113	152.859113	0	0	1	-360	360;
	7430	896	0.002391	0.025382	0.01581	139.891075	139.891075	139.891075	0	0	1	-360	360;
	498	3375	0.017551	0.078191	0.0223	163.445754	163.445754	163.445754	0	0	1	-360	360;
	5900	1407	0.003849	0.02475	0.12668	602.904127	602.904127	602.904127	0	0	1	-360	360;
	5900	6385	0.0042	0.02005	0.13817	327.881964	327.881964	327.881964	0	0	1	-360	360;
	5900	3262	0.00011	0.000581	0.00098	302.33172	302.33172	302.33172	0	0	1	-360	360;
	5900	707	0.000169	0.000651	0.00086	304.632126	304.632126	304.632126	0	0	1	-360	360;
	5900	3606	0.000169	0.000659	0.00085	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2441	5676	0.00219	0.01125	0.01789	513.709954	513.709954	513.709954	0	0	1	-360	360;
	2441	8132	0.00018	0.000599	0.05223	500.565593	500.565593	500.565593	0	0	1	-360	360;
	2441	8042	0.00018	0.000591	0.05321	401.476664	401.476664	401.476664	0	0	1	-360	360;
	2441	1584	0.00019	0.000581	0.0542	442.430139	442.430139	442.430139	0	0	1	-360	360;
	2441	2970	0.000169	0.000651	0.04828	501.006469	501.006469	501.006469	0	0	1	-360	360;
	291	1804	0.00231	0.0138	0.01967	463.354803	463.354803	463.354803	0	0	1	-360	360;
	291	1804	0.00082	0.003841	0.02703	266.607029	266.607029	266.607029	0	0	1	-360	360;
	291	157	0.00305	0.01762	0.02598	513.482085	513.482085	513.482085	0	0	1	-360	360;
	5139	4846	0.01101	0.055469	0.09147	490.293216	490.293216	490.293216	0	0	1	-360	360;
	5139	7170	0.009169	0.0462	0.07603	547.455465	547.455465	547.455465	0	0	1	-360	360;
	5139	6385	0.00201	0.00905	0.06553	244.61085	244.61085	244.61085	0	0	1	-360	360;
	5139	1831	0.005969	0.038841	0.04834	563.18983	563.18983	563.18983	0	0	1	-360	360;
	5139	7040	0.00474	0.02687	0.03965	438.613531	438.613531	438.613531	0	0	1	-360	360;
	5139	5047	0.005409	0.03438	0.04521	478.014222	478.014222	478.014222	0	0	1	-360	360;
	5139	2922	0.005401	0.03264	0.04378	577.213657	577.213657	577.213657	0	0	1	-360	360;
	5139	8138	0.00311	0.015919	0.02508	619.294281	619.294281	619.294281	0	0	1	-360	360;
	5139	1619	0.00462	0.02613	0.03855	422.724078	422.724078	422.724078	0	0	1	-360	360;
	2899	1012	0.01044	0.05187	0.08461	482.839855	482.839855	482.839855	0	0	1	-360	360;
	2899	1012	0.01001	0.05113	0.08454	479.889375	479.889375	479.889375	0	0	1	-360	360;
	2899	7698	0.008331	0.051729	0.06919	543.580555	543.580555	543.580555	0	0	1	-360	360;
	2899	5698	0.000169	0.00089	0.00138	452.987162	452.987162	452.987162	0	0	1	-360	360;
	2899	3632	0.00214	0.01099	0.01749	511.973884	511.973884	511.973884	0	0	1	-360	360;
	1804	5286	0.004669	0.042349	0.14328	479.903855	479.903855	479.903855	0	0	1	-360	360;
	1804	6574	0.00139	0.008789	0.01406	431.020324	431.020324	431.020324	0	0	1	-360	360;
	1804	6574	0.000219	0.001669	0.0088	514.058615	514.058615	514.058615	0	0	1	-360	360;
	1804	157	0.00119	0.01063	0.03653	288.121941	288.121941	288.121941	0	0	1	-360	360;
	2187	1831	0.004169	0.02724	0.03364	527.666715	527.666715	527.666715	0	0	1	-360	360;
	2187	1831	0.004081	0.025219	0.0329	498.422561	498.422561	498.422561	0	0	1	-360	360;
	2187	286	0.00168	0.01039	0.01372	547.939019	547.939019	547.939019	0	0	1	-360	360;
	2187	1355	0.0017	0.01055	0.01372	497.428399	497.428399	497.428399	0	0	1	-360	360;
	8104	5261	0.00139	0.008771	0.01153	499.966581	499.966581	499.966581	0	0	1	-360	360;
	8104	5790	0.000651	0.00261	0.29733	493.939113	493.939113	493.939113	0	0	1	-360	360;
	8104	5790	0.000729	0.00256	0.33698	493.012016	493.012016	493.012016	0	0	1	-360	360;
	8104	1125	0	0.008169	0	404.240047	404.240047	404.240047	0	0	1	-360	360;
	8104	5121	5e-05	0.00025	0.00043	450.019155	450.019155	450.019155	0	0	1	-360	360;
	5148	7170	0.003719	0.022669	0.03112	519.4181	519.4181	519.4181	0	0	1	-360	360;
	5148	5817	0.00732	0.042711	0.06424	482.82385	482.82385	482.82385	0	0	1	-360	360;
	5148	5817	0.003599	0.025331	0.1154	369.041207	369.041207	369.041207	0	0	1	-360	360;
	5148	8170	0.00494	0.027401	0.04149	535.492744	535.492744	535.492744	0	0	1	-360	360;
	4846	6164	0.0062	0.03519	0.05126	539.983432	539.983432	539.983432	0	0	1	-360	360;
	4846	6164	0.00739	0.04574	0.0607	455.013592	455.013592	455.013592	0	0	1	-360	360;
	4846	2922	0.006781	0.034289	0.05587	584.529077	584.529077	584.529077	0	0	1	-360	360;
	6164	2779	0.003669	0.03205	0.10948	258.910557	258.910557	258.910557	0	0	1	-360	360;
	6164	2779	0.002789	0.026289	0.08549	283.904466	283.904466	283.904466	0	0	1	-360	360;
	6164	2320	0.005979	0.031979	0.0487	535.395576	535.395576	535.395576	0	0	1	-360	360;
	2779	5676	0.00425	0.02132	0.03461	570.209936	570.209936	570.209936	0	0	1	-360	360;
	2779	1407	0.002469	0.015099	0.08158	350.8704	350.8704	350.8704	0	0	1	-360	360;
	2779	8132	0.004341	0.02662	0.03566	516.018743	516.018743	516.018743	0	0	1	-360	360;
	2779	2970	0.005219	0.02287	0.04417	612.858327	612.858327	612.858327	0	0	1	-360	360;
	2779	6910	0.0095	0.04693	0.07832	501.863072	501.863072	501.863072	0	0	1	-360	360;
	5286	157	0.01099	0.052909	0.09588	472.44173	472.44173	472.44173	0	0	1	-360	360;
	5286	5817	0.004901	0.02949	0.04115	551.501085	551.501085	551.501085	0	0	1	-360	360;
	5286	4315	0.00832	0.038531	0.07258	476.250718	476.250718	476.250718	0	0	1	-360	360;
	5286	5145	0.003909	0.018091	0.12873	394.368154	394.368154	394.368154	0	0	1	-360	360;
	5286	3104	0.000331	0.00189	0.00274	490.106501	490.106501	490.106501	0	0	1	-360	360;
	7170	5817	0.01101	0.05514	0.09527	507.503011	507.503011	507.503011	0	0	1	-360	360;
	7170	6804	0.00345	0.01925	0.02912	474.958192	474.958192	474.958192	0	0	1	-360	360;
	7667	157	0.00162	0.00889	0.05738	309.690962	309.690962	309.690962	0	0	1	-360	360;
	7667	8170	0.00039	0.00464	0.02069	309.170827	309.170827	309.170827	0	0	1	-360	360;
	7667	4869	0.00064	0.010099	0.03398	218.940575	218.940575	218.940575	0	0	1	-360	360;
	5676	5261	0.003151	0.01613	0.02679	571.658693	571.658693	571.658693	0	0	1	-360	360;
	5676	5261	0.002091	0.012409	0.01773	636.331842	636.331842	636.331842	0	0	1	-360	360;
	5676	8289	0	0.00799	0	341.562844	341.562844	341.562844	0	0	1	-360	360;
	5676	4225	0	0.00906	0	307.358928	307.358928	307.358928	0	0	1	-360	360;
	5676	6381	0.000909	0.010401	0.03008	267.854972	267.854972	267.854972	0	0	1	-360	360;
	5676	2262	0.003211	0.016081	0.02635	593.164078	593.164078	593.164078	0	0	1	-360	360;
	5676	8042	0.00314	0.017711	0.02587	459.552674	459.552674	459.552674	0	0	1	-360	360;
	5676	1584	0.00214	0.010591	0.01886	492.373374	492.373374	492.373374	0	0	1	-360	360;
	3159	2922	0.00056	0.003169	0.0046	215.929128	215.929128	215.929128	0	0	1	-360	360;
	6574	6560	0.001719	0.0108	0.01668	522.053831	522.053831	522.053831	0	0	1	-360	360;
	5261	3603	0.0005	0.00174	0.22866	529.35782	529.35782	529.35782	0	0	1	-360	360;
	5261	5121	0.00126	0.00587	0.01081	430.454082	430.454082	430.454082	0	0	1	-360	360;
	6311	2320	0.007909	0.03607	0.06443	546.744042	546.744042	546.744042	0	0	1	-360	360;
	6311	6381	0.00057	0.00457	0.02538	267.860687	267.860687	267.860687	0	0	1	-360	360;
	6311	5121	0.00132	0.008409	0.01076	402.230002	402.230002	402.230002	0	0	1	-360	360;
	157	2230	0.003581	0.017289	0.02899	620.550226	620.550226	620.550226	0	0	1	-360	360;
	157	4315	0.002711	0.01526	0.02361	494.053048	494.053048	494.053048	0	0	1	-360	360;
	157	5725	0.00039	0.006081	0.0198	215.929128	215.929128	215.929128	0	0	1	-360	360;
	157	4869	0.00037	0.005409	0.01987	261.895712	261.895712	261.895712	0	0	1	-360	360;
	157	6560	0.00244	0.013789	0.0212	490.970725	490.970725	490.970725	0	0	1	-360	360;
	3603	2262	0.000349	0.00149	0.16079	550.921506	550.921506	550.921506	0	0	1	-360	360;
	2230	5817	0.00212	0.010531	0.01776	597.959988	597.959988	597.959988	0	0	1	-360	360;
	341	4851	0.003039	0.031151	0.09065	215.929128	215.929128	215.929128	0	0	1	-360	360;
	341	5145	0.001831	0.0187	0.05439	279.968208	279.968208	279.968208	0	0	1	-360	360;
	1689	3583	0.006599	0.043169	0.06859	443.944818	443.944818	443.944818	0	0	1	-360	360;
	1689	4634	0.006729	0.046599	0.07008	455.289854	455.289854	455.289854	0	0	1	-360	360;
	1689	3297	0.002159	0.010789	0.01893	486.167193	486.167193	486.167193	0	0	1	-360	360;
	1689	712	0.0022	0.01057	0.0193	495.111989	495.111989	495.111989	0	0	1	-360	360;
	1689	944	0.003461	0.02049	0.02882	356.186064	356.186064	356.186064	0	0	1	-360	360;
	7276	633	0.00311	0.019469	0.0253	512.820199	512.820199	512.820199	0	0	1	-360	360;
	7276	569	0.01064	0.065969	0.0863	497.949677	497.949677	497.949677	0	0	1	-360	360;
	1012	286	0.003539	0.020479	0.02889	498.225939	498.225939	498.225939	0	0	1	-360	360;
	1012	1355	0.00357	0.020719	0.02875	513.647461	513.647461	513.647461	0	0	1	-360	360;
	1012	3583	0.00275	0.013831	0.02866	465.97491	465.97491	465.97491	0	0	1	-360	360;
	1012	4634	0.0028	0.013531	0.02919	464.268563	464.268563	464.268563	0	0	1	-360	360;
	5817	5145	0.005151	0.04851	0.15767	277.590067	277.590067	277.590067	0	0	1	-360	360;
	5817	7698	0.005401	0.025281	0.04781	545.909159	545.909159	545.909159	0	0	1	-360	360;
	5817	7698	0.0055	0.02475	0.04873	527.681957	527.681957	527.681957	0	0	1	-360	360;
	6804	5047	0.00093	0.00538	0.00796	452.917049	452.917049	452.917049	0	0	1	-360	360;
	6385	1831	0.010331	0.061849	0.08789	563.379594	563.379594	563.379594	0	0	1	-360	360;
	4284	3297	0.00686	0.032771	0.06007	479.713711	479.713711	479.713711	0	0	1	-360	360;
	4284	712	0.007	0.032081	0.06123	479.183669	479.183669	479.183669	0	0	1	-360	360;
	4851	3799	0.007729	0.038099	0.06307	376.577256	376.577256	376.577256	0	0	1	-360	360;
	3799	633	0.006581	0.035281	0.05409	576.896622	576.896622	576.896622	0	0	1	-360	360;
	3799	633	0.003711	0.02651	0.06934	545.241176	545.241176	545.241176	0	0	1	-360	360;
	3799	3945	0.00482	0.021979	0.03906	580.827927	580.827927	580.827927	0	0	1	-360	360;
	3799	3945	0.003979	0.026099	0.03226	546.596575	546.596575	546.596575	0	0	1	-360	360;
	1831	8138	0.0038	0.017271	0.0306	566.906984	566.906984	566.906984	0	0	1	-360	360;
	1831	6950	0.00626	0.03626	0.05054	647.787002	647.787002	647.787002	0	0	1	-360	360;
	7698	569	0.00538	0.02588	0.04673	576.167671	576.167671	576.167671	0	0	1	-360	360;
	7698	569	0.005479	0.025341	0.04763	555.734945	555.734945	555.734945	0	0	1	-360	360;
	115	2801	0.00134	0.01235	0.32442	1227.1615	1227.1615	1227.1615	0	0	1	-360	360;
	115	484	0.00127	0.01399	0.31268	1359.47857	1359.47857	1359.47857	0	0	1	-360	360;
	115	4000	0.00142	0.01374	0.33478	1363.18215	1363.18215	1363.18215	0	0	1	-360	360;
	7099	7230	0.00039	0.00365	0.10686	1227.41095	1227.41095	1227.41095	0	0	1	-360	360;
	7099	6131	0.00058	0.00507	0.1589	1300.89996	1300.89996	1300.89996	0	0	1	-360	360;
	933	5713	0.00129	0.01577	0.31825	1351.4659	1351.4659	1351.4659	0	0	1	-360	360;
	933	2733	0.00033	0.00373	0.08976	1342.58443	1342.58443	1342.58443	0	0	1	-360	360;
	933	2969	0.00086	0.00799	0.22698	1426.53256	1426.53256	1426.53256	0	0	1	-360	360;
	1219	2969	0.00189	0.02135	0.4886	1452.71296	1452.71296	1452.71296	0	0	1	-360	360;
	1219	1622	0.00106	0.01219	0.27997	1228.95833	1228.95833	1228.95833	0	0	1	-360	360;
	1219	2628	0.00147	0.01622	0.30007	1333.13626	1333.13626	1333.13626	0	0	1	-360	360;
	1219	1553	0.00215	0.02108	0.53373	1352.34193	1352.34193	1352.34193	0	0	1	-360	360;
	397	7335	0.00063	0.0052	0.16628	1412.02761	1412.02761	1412.02761	0	0	1	-360	360;
	397	2969	0.00159	0.01792	0.39884	1122.51625	1122.51625	1122.51625	0	0	1	-360	360;
	5713	1622	0.00157	0.01981	0.38575	1271.23318	1271.23318	1271.23318	0	0	1	-360	360;
	5713	2628	0.00158	0.02014	0.37007	1485.35931	1485.35931	1485.35931	0	0	1	-360	360;
	7335	2801	0.0018	0.02388	0.43689	1321.45095	1321.45095	1321.45095	0	0	1	-360	360;
	7335	490	0.00034	0.00318	0.0802	1376.97824	1376.97824	1376.97824	0	0	1	-360	360;
	7335	6422	0.00015	0.00132	0.04405	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7335	7798	0.00014	0.0015	0.03925	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7335	1051	0.00013	0.00153	0.0382	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7335	2961	0.00013	0.00156	0.0374	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7335	5277	0.0004	0.00449	0.1079	1293.30193	1293.30193	1293.30193	0	0	1	-360	360;
	7335	5277	0.00045	0.0044	0.12229	1176.52381	1176.52381	1176.52381	0	0	1	-360	360;
	2801	1737	0.0023	0.01984	0.55722	1211.83382	1211.83382	1211.83382	0	0	1	-360	360;
	2801	8202	0.0002	0.00235	0.04846	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2801	5449	0.00019	0.00239	0.04747	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2801	7083	0.00169	0.01602	0.43288	1343.87051	1343.87051	1343.87051	0	0	1	-360	360;
	2801	7827	0.00024	0.00253	0.06629	797.871283	797.871283	797.871283	0	0	1	-360	360;
	2801	7417	0.00022	0.00244	0.05933	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2801	5797	0.00025	0.00239	0.06724	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7455	7230	0.00066	0.0061	0.16047	1340.01292	1340.01292	1340.01292	0	0	1	-360	360;
	7455	7230	0.00062	0.00844	0.15693	1344.20223	1344.20223	1344.20223	0	0	1	-360	360;
	7455	7131	0.00124	0.01353	0.30148	1608.97789	1608.97789	1608.97789	0	0	1	-360	360;
	7455	5613	0.00046	0.00452	0.11391	1298.38505	1298.38505	1298.38505	0	0	1	-360	360;
	2733	2969	0.00045	0.00531	0.11598	1378.15178	1378.15178	1378.15178	0	0	1	-360	360;
	2733	490	0.00123	0.01433	0.2963	1158.83327	1158.83327	1158.83327	0	0	1	-360	360;
	2733	7083	0.00174	0.01696	0.42719	1453.28755	1453.28755	1453.28755	0	0	1	-360	360;
	4373	5613	0.00086	0.00877	0.22716	1406.84247	1406.84247	1406.84247	0	0	1	-360	360;
	4373	6131	0.00096	0.00999	0.25055	1245.01724	1245.01724	1245.01724	0	0	1	-360	360;
	4373	4481	0.00082	0.00761	0.21375	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7230	1553	0.00185	0.01803	0.45656	1301.67134	1301.67134	1301.67134	0	0	1	-360	360;
	7230	1854	0.00138	0.01246	0.37547	1493.9038	1493.9038	1493.9038	0	0	1	-360	360;
	7230	7240	0.00062	0.00605	0.15171	1516.85714	1516.85714	1516.85714	0	0	1	-360	360;
	7230	6131	0.00073	0.00817	0.20141	767.876077	767.876077	767.876077	0	0	1	-360	360;
	484	6810	0.00094	0.00999	0.25635	1341.33981	1341.33981	1341.33981	0	0	1	-360	360;
	484	3634	0.00058	0.00671	0.1588	1226.34206	1226.34206	1226.34206	0	0	1	-360	360;
	1622	1553	0.00108	0.01279	0.28634	1367.69265	1367.69265	1367.69265	0	0	1	-360	360;
	1622	7240	0.00225	0.02472	0.55439	1399.2326	1399.2326	1399.2326	0	0	1	-360	360;
	1622	4133	0.00019	0.00244	0.04648	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1622	5751	0.00018	0.00248	0.04549	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1622	237	0.0001	0.0011	0.02571	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1622	1880	0.00011	0.00108	0.0262	767.876077	767.876077	767.876077	0	0	1	-360	360;
	7131	5613	0.00089	0.00815	0.21353	1308.4598	1308.4598	1308.4598	0	0	1	-360	360;
	5519	2628	0.0009	0.00859	0.21851	1443.27928	1443.27928	1443.27928	0	0	1	-360	360;
	5519	6810	0.00167	0.02239	0.4094	1566.23968	1566.23968	1566.23968	0	0	1	-360	360;
	2628	4784	2e-05	0.00024	0.00465	767.876077	767.876077	767.876077	0	0	1	-360	360;
	1553	122	0.00157	0.01326	0.32339	1364.02659	1364.02659	1364.02659	0	0	1	-360	360;
	1737	5050	0.00155	0.0139	0.4163	1375.65794	1375.65794	1375.65794	0	0	1	-360	360;
	122	1854	0.00142	0.01472	0.27463	1370.53269	1370.53269	1370.53269	0	0	1	-360	360;
	4276	1854	0.00187	0.02232	0.46033	1539.64924	1539.64924	1539.64924	0	0	1	-360	360;
	4276	1431	0.00049	0.00572	0.13733	1355.40313	1355.40313	1355.40313	0	0	1	-360	360;
	1854	1825	0	0.00021	0	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6810	3634	0.00104	0.00939	0.28304	1374.41003	1374.41003	1374.41003	0	0	1	-360	360;
	6810	7200	0.00043	0.0049	0.11771	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6810	7200	0.00049	0.0048	0.13341	767.876077	767.876077	767.876077	0	0	1	-360	360;
	34	1068	0.002009	0.01464	0.00501	103.923048	103.923048	103.923048	0	0	1	-360	360;
	1068	3375	0.003302	0.026	0.00828	112.938893	112.938893	112.938893	0	0	1	-360	360;
	350	3879	0.006222	0.018089	0.00683	253.457923	253.457923	253.457923	0	0	1	-360	360;
	3375	6717	0	0.000222	0	128.072167	128.072167	128.072167	0	0	1	-360	360;
	3375	2623	8.9e-05	0.000911	0.00057	113.159469	113.159469	113.159469	0	0	1	-360	360;
	7957	7067	1.8e-05	0.000222	2e-05	227.157857	227.157857	227.157857	0	0	1	-360	360;
	5861	6717	0.00388	0.014289	0.00423	290.456607	290.456607	290.456607	0	0	1	-360	360;
	3655	553	0.001329	0.007898	0.0024	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3655	4164	0.00048	0.002951	0.00086	103.923048	103.923048	103.923048	0	0	1	-360	360;
	3655	7263	8e-05	0.00032	0.00013	103.923048	103.923048	103.923048	0	0	1	-360	360;
	2584	4844	0.000302	0.001102	0.00033	222.634087	222.634087	222.634087	0	0	1	-360	360;
	5907	276	0.003742	0.013098	0.00419	103.923048	103.923048	103.923048	0	0	1	-360	360;
	6867	5567	0.008911	0.032298	0.00997	103.923048	103.923048	103.923048	0	0	1	-360	360;
	5376	5931	0.00124	0.005409	0.00234	103.923048	103.923048	103.923048	0	0	1	-360	360;
	633	6444	0.0025	0.0123	0.02024	403.312569	403.312569	403.312569	0	0	1	-360	360;
	633	6444	0.001781	0.013729	0.01974	401.755593	401.755593	401.755593	0	0	1	-360	360;
	633	3945	0.00324	0.016039	0.0262	561.348972	561.348972	561.348972	0	0	1	-360	360;
	8170	5725	0.00037	0.00474	0.01951	215.929128	215.929128	215.929128	0	0	1	-360	360;
	7040	1619	9.1e-05	0.000521	0.00072	436.347038	436.347038	436.347038	0	0	1	-360	360;
	5725	4869	3.1e-05	0.00043	0.00154	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1619	6994	0.001281	0.005909	0.01151	217.253281	217.253281	217.253281	0	0	1	-360	360;
	1355	5254	0.000909	0.00476	0.00825	243.140754	243.140754	243.140754	0	0	1	-360	360;
	4634	4786	0.00011	0.000599	0.00118	236.54895	236.54895	236.54895	0	0	1	-360	360;
	3297	7193	0.000289	0.001719	0.0026	218.698608	218.698608	218.698608	0	0	1	-360	360;
	6560	3730	0.00026	0.001591	0.0021	528.884936	528.884936	528.884936	0	0	1	-360	360;
	5277	3634	0.00108	0.01382	0.26699	1483.31698	1483.31698	1483.31698	0	0	1	-360	360;
	6645	6950	0	0.0002	0	575.947042	575.947042	575.947042	0	0	1	-360	360;
	2107	4858	0.014281	0.033446	0.00353	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5772	4858	0.038273	0.11138	0.00969	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2534	194	0.039504	0.11795	0.01	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2644	2534	0.039893	0.095083	0.0101	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4193	4827	0.025107	0.052256	0.00463	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1611	4193	0.010124	0.017818	0.00172	74.058058	74.058058	74.058058	0	0	1	-360	360;
	463	6200	0.035198	0.083893	0.0089	68.904722	68.904722	68.904722	0	0	1	-360	360;
	6200	5912	0.026149	0.064851	0.00661	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6530	6725	0	0.000793	0	65.920711	65.920711	65.920711	0	0	1	-360	360;
	3416	6530	0.02086	0.051744	0.00528	57.157677	57.157677	57.157677	0	0	1	-360	360;
	189	6725	0.041992	0.117446	0.01062	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5097	4594	0.014031	0.079591	0.1339	465.081345	465.081345	465.081345	0	0	1	-360	360;
	4594	3054	0.0075	0.054099	0.07157	418.588529	418.588529	418.588529	0	0	1	-360	360;
	3126	4594	0.013841	0.0733	0.1173	444.287002	444.287002	444.287002	0	0	1	-360	360;
	4594	1903	0.006469	0.041401	0.06178	438.215713	438.215713	438.215713	0	0	1	-360	360;
	1822	2349	0.005628	0.015826	0.00148	60.587328	60.587328	60.587328	0	0	1	-360	360;
	7842	1822	0.026364	0.056017	0.00691	73.95346	73.95346	73.95346	0	0	1	-360	360;
	1822	3416	0.026099	0.062182	0.0066	78.776234	78.776234	78.776234	0	0	1	-360	360;
	1822	3312	0.04481	0.131231	0.01175	57.157677	57.157677	57.157677	0	0	1	-360	360;
	189	2349	0.005512	0.01614	0.00145	60.681257	60.681257	60.681257	0	0	1	-360	360;
	7842	189	0.023488	0.063479	0.00615	74.435109	74.435109	74.435109	0	0	1	-360	360;
	189	5672	0.044198	0.129446	0.01159	57.157677	57.157677	57.157677	0	0	1	-360	360;
	815	1296	0.006612	0.01657	0.00134	57.946834	57.946834	57.946834	0	0	1	-360	360;
	2488	815	0.003967	0.007521	0.00081	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5273	5749	0.027496	0.067438	0.00692	60.886453	60.886453	60.886453	0	0	1	-360	360;
	677	5273	0.044587	0.102702	0.01169	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5273	4082	0.051438	0.143868	0.01301	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5273	8189	0.029256	0.072579	0.0074	65.149272	65.149272	65.149272	0	0	1	-360	360;
	845	5273	0.049149	0.122653	0.01289	60.444815	60.444815	60.444815	0	0	1	-360	360;
	845	5273	0.043372	0.125157	0.01138	64.753741	64.753741	64.753741	0	0	1	-360	360;
	6339	5273	0.016471	0.042529	0.00417	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6788	53	0	0.00076	0	68.203017	68.203017	68.203017	0	0	1	-360	360;
	53	4054	0.038017	0.110653	0.00962	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4667	53	0.025174	0.07938	0.00637	60.621813	60.621813	60.621813	0	0	1	-360	360;
	4967	53	0.043017	0.128446	0.01089	57.157677	57.157677	57.157677	0	0	1	-360	360;
	611	6788	0.051636	0.189413	0.0171	58.680357	58.680357	58.680357	0	0	1	-360	360;
	6788	3206	0.04714	0.113008	0.01237	59.467037	59.467037	59.467037	0	0	1	-360	360;
	2562	6788	0.008149	0.025711	0.00206	89.141588	89.141588	89.141588	0	0	1	-360	360;
	4122	591	0.039686	0.102843	0.00902	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6449	591	0.006372	0.01857	0.00161	57.157677	57.157677	57.157677	0	0	1	-360	360;
	591	4228	0.045711	0.108901	0.01156	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1084	1339	0.04119	0.106331	0.01042	62.746364	62.746364	62.746364	0	0	1	-360	360;
	3061	1084	0.013752	0.03024	0.00348	60.766231	60.766231	60.766231	0	0	1	-360	360;
	5305	4268	0.00595	0.019008	0.03594	64.491388	64.491388	64.491388	0	0	1	-360	360;
	4268	7195	0.00576	0.020331	0.00191	72.233395	72.233395	72.233395	0	0	1	-360	360;
	2588	7273	0.03586	0.082107	0.00907	57.157677	57.157677	57.157677	0	0	1	-360	360;
	385	2588	0.022223	0.055099	0.00563	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2847	6023	0.008802	0.02862	0.00292	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6023	6849	0.032719	0.124942	0.01084	57.157677	57.157677	57.157677	0	0	1	-360	360;
	689	7464	0.019793	0.055331	0.00501	68.189299	68.189299	68.189299	0	0	1	-360	360;
	7464	7470	0.01395	0.034612	0.00354	60.571324	60.571324	60.571324	0	0	1	-360	360;
	3349	2012	0.004727	0.010107	0.00096	60.340788	60.340788	60.340788	0	0	1	-360	360;
	7909	3349	0.007289	0.014364	0.00147	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4748	3685	0.004818	0.009901	0.00098	57.157677	57.157677	57.157677	0	0	1	-360	360;
	422	4748	0.012223	0.031554	0.00309	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1237	2859	0.00108	0.01189	0.26132	1356.757	1356.757	1356.757	0	0	1	-360	360;
	881	1237	0.0032	0.03265	0.68253	1330.71416	1330.71416	1330.71416	0	0	1	-360	360;
	6086	1237	0.0007	0.00731	0.17234	1421.01702	1421.01702	1421.01702	0	0	1	-360	360;
	2296	5212	0	0.000744	0	124.957922	124.957922	124.957922	0	0	1	-360	360;
	2296	3837	0.009248	0.022041	0.00234	72.545476	72.545476	72.545476	0	0	1	-360	360;
	5212	7785	0.008124	0.02286	0.00213	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7682	6819	0.007702	0.019893	0.00195	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6819	6625	0.015579	0.04914	0.00394	68.778213	68.778213	68.778213	0	0	1	-360	360;
	8002	2839	0.00981	0.068021	0.09367	391.775101	391.775101	391.775101	0	0	1	-360	360;
	8002	1545	0.006341	0.042211	0.06052	402.213617	402.213617	402.213617	0	0	1	-360	360;
	900	5630	0.060281	0.144512	0.01581	59.368155	59.368155	59.368155	0	0	1	-360	360;
	900	5630	0.061438	0.141504	0.01611	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5630	2488	0.016579	0.039917	0.00336	65.171945	65.171945	65.171945	0	0	1	-360	360;
	1443	3397	0	0.000843	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2041	1443	0.010256	0.022711	0.00269	58.03238	58.03238	58.03238	0	0	1	-360	360;
	1443	4425	0.024983	0.073149	0.00655	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3397	3883	0.008388	0.02557	0.0022	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2323	2599	0.063512	0.200339	0.01607	72.928813	72.928813	72.928813	0	0	1	-360	360;
	4122	2323	0.037562	0.085967	0.0095	84.731111	84.731111	84.731111	0	0	1	-360	360;
	2756	676	0.002579	0.007752	0.00086	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1251	676	0.001777	0.00495	0.00059	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5999	6182	0.030248	0.084612	0.00765	59.990221	59.990221	59.990221	0	0	1	-360	360;
	6182	7860	0.013281	0.037124	0.00336	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4965	7536	0.009488	0.023521	0.0024	57.157677	57.157677	57.157677	0	0	1	-360	360;
	111	7536	0.022579	0.058314	0.00571	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4149	4397	0.02743	0.062793	0.00694	59.955545	59.955545	59.955545	0	0	1	-360	360;
	6547	4149	0.001719	0.003777	0.00044	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2513	1815	0.035992	0.085876	0.00756	64.022314	64.022314	64.022314	0	0	1	-360	360;
	7452	2513	0.008777	0.01962	0.00184	62.73417	62.73417	62.73417	0	0	1	-360	360;
	2755	7314	0.001793	0.001521	0.05227	64.767269	64.767269	64.767269	0	0	1	-360	360;
	7314	2070	0.004917	0.007702	0.05306	62.1866	62.1866	62.1866	0	0	1	-360	360;
	6465	3849	0.001521	0.004554	0.00039	57.157677	57.157677	57.157677	0	0	1	-360	360;
	459	3849	0.011198	0.041099	0.00372	57.865098	57.865098	57.865098	0	0	1	-360	360;
	1582	58	0.00144	0.01155	0.30699	1223.00707	1223.00707	1223.00707	0	0	1	-360	360;
	1582	5851	0.00123	0.01345	0.26344	1084.25958	1084.25958	1084.25958	0	0	1	-360	360;
	411	5669	0.007669	0.02419	0.00194	57.157677	57.157677	57.157677	0	0	1	-360	360;
	411	169	0.003661	0.008769	0.00096	80.225943	80.225943	80.225943	0	0	1	-360	360;
	7840	411	0.00943	0.036008	0.00395	71.538929	71.538929	71.538929	0	0	1	-360	360;
	5686	411	0.012661	0.037802	0.0032	79.671514	79.671514	79.671514	0	0	1	-360	360;
	1832	724	0.006083	0.024099	0.00201	60.668873	60.668873	60.668873	0	0	1	-360	360;
	1832	1053	0.002248	0.004248	0.00045	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1605	4317	0.019651	0.141771	0.18755	384.928373	384.928373	384.928373	0	0	1	-360	360;
	4317	3310	0.004841	0.02718	0.04644	410.355537	410.355537	410.355537	0	0	1	-360	360;
	4317	3310	0.00493	0.02661	0.04733	420.565423	420.565423	420.565423	0	0	1	-360	360;
	3939	4317	0.008729	0.04569	0.08329	403.289705	403.289705	403.289705	0	0	1	-360	360;
	8025	6410	0.002529	0.005793	0.00064	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6410	5542	0.003579	0.007876	0.0009	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7439	769	0.014182	0.03119	0.00359	59.346435	59.346435	59.346435	0	0	1	-360	360;
	7439	3462	0.061124	0.192777	0.01547	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4616	4255	0.114603	0.284256	0.029	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4255	6909	0.00786	0.019488	0.00199	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8025	7793	0.002669	0.006124	0.00068	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3302	4583	0.00895	0.022182	0.00227	59.347387	59.347387	59.347387	0	0	1	-360	360;
	4583	212	0.002281	0.008033	0.00067	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3302	769	0.005314	0.013719	0.00135	74.904564	74.904564	74.904564	0	0	1	-360	360;
	952	4942	0.031521	0.069298	0.00798	62.993475	62.993475	62.993475	0	0	1	-360	360;
	4942	7793	0.003719	0.013223	0.00109	65.413341	65.413341	65.413341	0	0	1	-360	360;
	4942	212	0.009719	0.027818	0.00249	72.679796	72.679796	72.679796	0	0	1	-360	360;
	4942	3601	0.042669	0.127397	0.0108	66.477617	66.477617	66.477617	0	0	1	-360	360;
	4942	6088	0.031157	0.077281	0.00788	70.565724	70.565724	70.565724	0	0	1	-360	360;
	5542	769	0.00319	0.008934	0.0008	73.116481	73.116481	73.116481	0	0	1	-360	360;
	769	5977	0.01462	0.046124	0.0037	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7480	4026	0.007587	0.018628	0.00172	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6704	7480	0.024529	0.052826	0.00551	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4642	4614	0.041223	0.115314	0.01044	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4642	2469	0.022512	0.051521	0.00569	58.044573	58.044573	58.044573	0	0	1	-360	360;
	6449	4642	0.063537	0.189702	0.01607	61.891475	61.891475	61.891475	0	0	1	-360	360;
	1671	5069	0.01524	0.045496	0.00385	57.157677	57.157677	57.157677	0	0	1	-360	360;
	433	5069	0.035876	0.092612	0.00908	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5669	4820	0.015471	0.046198	0.00392	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7635	5124	0.043364	0.152818	0.01437	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1159	5124	0.03095	0.100686	0.01025	71.008315	71.008315	71.008315	0	0	1	-360	360;
	1928	5124	0.035628	0.107959	0.00901	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1087	5999	0.010901	0.034397	0.00276	72.671794	72.671794	72.671794	0	0	1	-360	360;
	3129	2529	0.026099	0.060099	0.00685	64.832809	64.832809	64.832809	0	0	1	-360	360;
	2494	3129	0.025942	0.057008	0.00657	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8065	1476	0.001901	0.004636	0.05444	76.443629	76.443629	76.443629	0	0	1	-360	360;
	1476	7943	0.00176	0.003124	0.04966	100.00993	100.00993	100.00993	0	0	1	-360	360;
	3811	2599	0.003876	0.010909	0.00102	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2599	647	0.003802	0.011124	0.001	123.885454	123.885454	123.885454	0	0	1	-360	360;
	4173	2599	0.013339	0.029298	0.00337	83.537278	83.537278	83.537278	0	0	1	-360	360;
	1690	2599	0.031719	0.11181	0.01051	57.739351	57.739351	57.739351	0	0	1	-360	360;
	3065	3371	0.021521	0.05557	0.00545	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3846	3065	0.037008	0.107752	0.00937	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5908	1791	0.05743	0.142463	0.01454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5908	1412	0.036091	0.074041	0.00732	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7159	3815	0.001211	0.00545	0.01025	530.342456	530.342456	530.342456	0	0	1	-360	360;
	1922	3815	0.013711	0.078669	0.11622	256.521366	256.521366	256.521366	0	0	1	-360	360;
	1130	4136	0.050521	0.125331	0.01278	65.475262	65.475262	65.475262	0	0	1	-360	360;
	4136	4756	0.085818	0.221579	0.02172	58.640537	58.640537	58.640537	0	0	1	-360	360;
	4136	5093	0.014909	0.034132	0.00377	57.971602	57.971602	57.971602	0	0	1	-360	360;
	4136	4333	0.035826	0.113033	0.00907	58.796578	58.796578	58.796578	0	0	1	-360	360;
	5687	5752	0.008289	0.026017	0.00285	73.555452	73.555452	73.555452	0	0	1	-360	360;
	3631	5752	0.023339	0.089587	0.00801	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3395	2164	0.00661	0.039521	0.05609	422.507259	422.507259	422.507259	0	0	1	-360	360;
	4707	3395	0.00349	0.01574	0.02958	427.573716	427.573716	427.573716	0	0	1	-360	360;
	689	6168	0.03714	0.088496	0.0094	66.236411	66.236411	66.236411	0	0	1	-360	360;
	689	3545	0.037256	0.087669	0.0077	59.784453	59.784453	59.784453	0	0	1	-360	360;
	689	4170	0.03081	0.07957	0.0078	73.890396	73.890396	73.890396	0	0	1	-360	360;
	4574	4849	0.033248	0.096818	0.00842	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8253	4849	0.011479	0.034298	0.00291	62.691683	62.691683	62.691683	0	0	1	-360	360;
	4400	995	0	0.000893	0	73.883347	73.883347	73.883347	0	0	1	-360	360;
	4400	3083	0.019273	0.042628	0.00505	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4400	3396	0.019628	0.041702	0.00515	71.16931	71.16931	71.16931	0	0	1	-360	360;
	5402	4400	0.029512	0.062719	0.00774	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6171	995	0.034901	0.105744	0.00883	82.235607	82.235607	82.235607	0	0	1	-360	360;
	2045	995	0.026289	0.071083	0.0069	57.157677	57.157677	57.157677	0	0	1	-360	360;
	505	995	0.041893	0.121942	0.0106	79.374866	79.374866	79.374866	0	0	1	-360	360;
	1964	4614	0.024851	0.071711	0.00652	62.67263	62.67263	62.67263	0	0	1	-360	360;
	6677	5843	0.006512	0.01681	0.00165	82.122054	82.122054	82.122054	0	0	1	-360	360;
	3533	6677	0.00305	0.009099	0.00077	97.44355	97.44355	97.44355	0	0	1	-360	360;
	1314	6543	0.014678	0.04105	0.00371	65.903182	65.903182	65.903182	0	0	1	-360	360;
	5719	1314	0.054149	0.119066	0.01371	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7527	677	0.08557	0.197909	0.01736	60.643152	60.643152	60.643152	0	0	1	-360	360;
	8021	7527	0.038314	0.075496	0.00777	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5910	7497	0.036554	0.115289	0.00925	66.079799	66.079799	66.079799	0	0	1	-360	360;
	2234	5910	0.047636	0.11019	0.00966	62.689016	62.689016	62.689016	0	0	1	-360	360;
	2040	2512	0	0.000826	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4312	2512	0.013397	0.037488	0.00339	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2040	837	0.03095	0.079917	0.00783	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5057	2040	0.014058	0.032198	0.00356	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4325	8348	0.03086	0.093488	0.00781	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5719	4325	0.023669	0.071066	0.00784	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5719	4325	0.028083	0.078537	0.00711	57.157677	57.157677	57.157677	0	0	1	-360	360;
	714	2137	0.016661	0.052554	0.00421	61.464127	61.464127	61.464127	0	0	1	-360	360;
	2137	7366	0.022223	0.050868	0.00563	70.244308	70.244308	70.244308	0	0	1	-360	360;
	302	976	0.017	0.043909	0.0043	57.157677	57.157677	57.157677	0	0	1	-360	360;
	976	7384	0.017727	0.05595	0.00449	57.157677	57.157677	57.157677	0	0	1	-360	360;
	562	1400	0.022223	0.052967	0.00563	59.312521	59.312521	59.312521	0	0	1	-360	360;
	4753	1400	0.027628	0.071347	0.007	71.483296	71.483296	71.483296	0	0	1	-360	360;
	5503	4171	0	0.00081	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5503	8124	0.002983	0.010149	0.00103	66.738446	66.738446	66.738446	0	0	1	-360	360;
	2325	5503	0.008678	0.023463	0.00227	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4171	194	0.002917	0.010347	0.00101	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2325	4171	0.008512	0.023917	0.00223	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5197	3613	0.003603	0.011339	0.00091	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5197	2107	0.006347	0.015132	0.00161	61.53081	61.53081	61.53081	0	0	1	-360	360;
	5200	5197	0.027777	0.08295	0.00703	72.925384	72.925384	72.925384	0	0	1	-360	360;
	433	2798	0.045438	0.105124	0.00921	79.866041	79.866041	79.866041	0	0	1	-360	360;
	2798	7028	0.07486	0.187603	0.01518	61.222921	61.222921	61.222921	0	0	1	-360	360;
	1631	4503	0.01919	0.053198	0.00636	75.148436	75.148436	75.148436	0	0	1	-360	360;
	7515	4503	0.008488	0.032421	0.00281	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2203	6724	0.00365	0.03434	0.77731	1056.42386	1056.42386	1056.42386	0	0	1	-360	360;
	4460	6724	0.0049	0.04432	1.04447	1149.69379	1149.69379	1149.69379	0	0	1	-360	360;
	2839	3057	0.000901	0.005539	0.00767	379.8703	379.8703	379.8703	0	0	1	-360	360;
	1960	2839	0.00894	0.06106	0.08534	439.024304	439.024304	439.024304	0	0	1	-360	360;
	2839	1613	0.006919	0.049419	0.06639	393.182704	393.182704	393.182704	0	0	1	-360	360;
	2839	2050	0.006771	0.04576	0.06494	488.8014	488.8014	488.8014	0	0	1	-360	360;
	104	3022	0	0.000793	0	83.085351	83.085351	83.085351	0	0	1	-360	360;
	104	7567	0.019628	0.061364	0.0065	73.65624	73.65624	73.65624	0	0	1	-360	360;
	104	5764	0.011727	0.033826	0.00389	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3022	5411	0.006438	0.01943	0.00222	68.756303	68.756303	68.756303	0	0	1	-360	360;
	3022	5411	0.006562	0.019033	0.00226	72.275882	72.275882	72.275882	0	0	1	-360	360;
	4048	3022	0.006421	0.020893	0.00212	79.051734	79.051734	79.051734	0	0	1	-360	360;
	5226	4307	0.00031	0.00355	0.06244	1279.2011	1279.2011	1279.2011	0	0	1	-360	360;
	3899	1925	0.016752	0.046397	0.00554	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2956	3899	0.009347	0.035702	0.00309	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1925	7569	0.011298	0.040033	0.00388	71.567889	71.567889	71.567889	0	0	1	-360	360;
	7942	1180	0.001298	0.004322	0.03231	80.957371	80.957371	80.957371	0	0	1	-360	360;
	6988	1180	0.000347	0.000876	0.00862	71.515113	71.515113	71.515113	0	0	1	-360	360;
	2203	761	0.00028	0.00301	0.05904	767.876077	767.876077	767.876077	0	0	1	-360	360;
	58	2203	0.00297	0.02796	0.63316	1107.32613	1107.32613	1107.32613	0	0	1	-360	360;
	2203	2905	0.00274	0.03157	0.58502	1041.63325	1041.63325	1041.63325	0	0	1	-360	360;
	5687	752	0.00286	0.010554	0.00099	87.597378	87.597378	87.597378	0	0	1	-360	360;
	5687	8357	0.025818	0.091446	0.00886	60.415474	60.415474	60.415474	0	0	1	-360	360;
	1815	5687	0.027066	0.065554	0.00569	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5687	2576	0.086198	0.151504	0.01811	60.095962	60.095962	60.095962	0	0	1	-360	360;
	683	5687	0.010893	0.02719	0.00286	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5428	5687	0.03162	0.072818	0.00829	61.947871	61.947871	61.947871	0	0	1	-360	360;
	6141	752	0.002802	0.010752	0.00097	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6141	7895	0.054802	0.152636	0.01881	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6141	6325	0.014223	0.048397	0.00488	58.741897	58.741897	58.741897	0	0	1	-360	360;
	6528	6141	0.01814	0.043521	0.00476	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6141	3612	0.038248	0.07586	0.00804	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7474	6213	0.017331	0.061314	0.00752	81.248875	81.248875	81.248875	0	0	1	-360	360;
	3031	6213	0.009182	0.032339	0.00384	57.157677	57.157677	57.157677	0	0	1	-360	360;
	761	4003	0.00281	0.02539	0.59834	1059.34881	1059.34881	1059.34881	0	0	1	-360	360;
	6104	4003	0.00158	0.01426	0.33592	1083.2914	1083.2914	1083.2914	0	0	1	-360	360;
	5731	3247	0	0.000211	0	481.730234	481.730234	481.730234	0	0	1	-360	360;
	8366	5731	0.00888	0.061021	0.08533	430.049024	430.049024	430.049024	0	0	1	-360	360;
	5731	1613	0.006961	0.04406	0.06676	260.422568	260.422568	260.422568	0	0	1	-360	360;
	3247	8177	0.00869	0.06218	0.08352	372.454282	372.454282	372.454282	0	0	1	-360	360;
	3247	6592	0.002401	0.027581	0.05743	453.170067	453.170067	453.170067	0	0	1	-360	360;
	3247	2050	0.00682	0.04493	0.0654	446.563782	446.563782	446.563782	0	0	1	-360	360;
	3120	7280	0.00238	0.009446	0.00082	65.112882	65.112882	65.112882	0	0	1	-360	360;
	6704	3120	0.017653	0.060008	0.00766	101.485932	101.485932	101.485932	0	0	1	-360	360;
	6659	2757	0.01838	0.051397	0.00465	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2757	613	0.02995	0.071388	0.00758	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3137	3551	0.023223	0.042223	0.00471	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3551	574	0.018636	0.052124	0.00471	62.921076	62.921076	62.921076	0	0	1	-360	360;
	6104	2455	0.00096	0.01102	0.20429	1222.83397	1222.83397	1222.83397	0	0	1	-360	360;
	2455	4557	0.00364	0.02923	0.77721	1184.80042	1184.80042	1184.80042	0	0	1	-360	360;
	2455	2028	0.00248	0.02649	0.61253	1032.09821	1032.09821	1032.09821	0	0	1	-360	360;
	6363	2455	0.00384	0.04229	0.81596	1086.80015	1086.80015	1086.80015	0	0	1	-360	360;
	5164	2293	0	0.000219	0	215.929128	215.929128	215.929128	0	0	1	-360	360;
	8366	2293	0.00344	0.02181	0.03305	515.431162	515.431162	515.431162	0	0	1	-360	360;
	2293	8177	0.00337	0.02224	0.03237	479.544143	479.544143	479.544143	0	0	1	-360	360;
	2293	183	0.010521	0.05913	0.101	417.870248	417.870248	417.870248	0	0	1	-360	360;
	2293	491	0.00081	0.005341	0.00777	445.911423	445.911423	445.911423	0	0	1	-360	360;
	2755	2070	0.002579	0.00476	0.08746	98.981473	98.981473	98.981473	0	0	1	-360	360;
	8096	2755	0.004388	0.01595	0.00141	96.767184	96.767184	96.767184	0	0	1	-360	360;
	2755	6392	0.004752	0.017868	0.00158	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1070	4242	0.00338	0.012959	0.00123	70.15438	70.15438	70.15438	0	0	1	-360	360;
	7176	1070	0.007008	0.02338	0.00385	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5425	4047	0.025347	0.057942	0.0064	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3501	5425	0.06314	0.162959	0.01592	60.459295	60.459295	60.459295	0	0	1	-360	360;
	3488	2768	0.007967	0.023802	0.00202	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5959	3488	0.006636	0.016463	0.00168	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4538	6884	0.012091	0.022893	0.00245	61.636552	61.636552	61.636552	0	0	1	-360	360;
	4538	3365	0.012322	0.022388	0.0025	58.161747	58.161747	58.161747	0	0	1	-360	360;
	5573	6173	0.016529	0.049347	0.00419	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7592	7942	0.00119	0.007017	0.00066	115.882045	115.882045	115.882045	0	0	1	-360	360;
	7592	4199	0.002331	0.009636	0.0008	123.945279	123.945279	123.945279	0	0	1	-360	360;
	2758	1645	0.008091	0.02681	0.07328	124.341	124.341	124.341	0	0	1	-360	360;
	1285	1645	0.001347	0.006876	0.00074	84.456183	84.456183	84.456183	0	0	1	-360	360;
	6988	1645	0.002281	0.009818	0.00078	58.258343	58.258343	58.258343	0	0	1	-360	360;
	361	7279	0.001149	0.00243	0.0003	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3238	361	0.002909	0.006909	0.00073	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	2010	0.000413	0.000959	0.00011	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	2010	0.000421	0.000942	0.00011	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6543	6894	0.024628	0.073504	0.00623	69.471345	69.471345	69.471345	0	0	1	-360	360;
	6894	5750	0.041496	0.094992	0.0105	70.87571	70.87571	70.87571	0	0	1	-360	360;
	3006	403	0.01	0.021273	0.00519	61.70838	61.70838	61.70838	0	0	1	-360	360;
	403	1690	0.030479	0.08795	0.00799	75.318385	75.318385	75.318385	0	0	1	-360	360;
	403	1690	0.031157	0.094992	0.00817	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7773	5604	0.00962	0.023298	0.00202	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1719	7773	0.018248	0.040397	0.00105	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6161	5099	0.005868	0.01781	0.00149	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6161	8192	0.01	0.02814	0.00262	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6161	32	0.016562	0.052033	0.00569	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6161	32	0.016893	0.050967	0.0058	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5099	3309	0.013281	0.033149	0.00349	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4618	1189	0.005463	0.020008	0.0018	61.681897	61.681897	61.681897	0	0	1	-360	360;
	4618	4771	0.02695	0.066868	0.00682	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3309	4618	0.002752	0.008397	0.00073	66.387689	66.387689	66.387689	0	0	1	-360	360;
	5275	4618	0.002686	0.007777	0.00071	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4618	32	0.019099	0.057876	0.00483	70.213824	70.213824	70.213824	0	0	1	-360	360;
	4618	3559	0.017339	0.043008	0.00439	71.455098	71.455098	71.455098	0	0	1	-360	360;
	1189	32	0.017843	0.068107	0.00591	57.157677	57.157677	57.157677	0	0	1	-360	360;
	532	7923	0.022289	0.070314	0.00564	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7923	2101	0.035008	0.101917	0.00886	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1278	900	0.045231	0.147174	0.01498	78.665158	78.665158	78.665158	0	0	1	-360	360;
	4801	900	0.011413	0.03424	0.00377	64.061562	64.061562	64.061562	0	0	1	-360	360;
	6478	2057	0.001124	0.002579	0.00028	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6478	5541	0.046909	0.135298	0.01554	60.989337	60.989337	60.989337	0	0	1	-360	360;
	6478	4533	0.035446	0.075124	0.00838	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7580	6478	0.011107	0.025793	0.00221	63.352045	63.352045	63.352045	0	0	1	-360	360;
	6478	8242	0.005983	0.023752	0.00198	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4910	3138	0.016132	0.04695	0.00408	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4565	3138	0.008008	0.019083	0.00202	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3924	418	0.034281	0.088488	0.00867	57.157677	57.157677	57.157677	0	0	1	-360	360;
	418	4588	0.009149	0.036372	0.00303	62.690159	62.690159	62.690159	0	0	1	-360	360;
	418	4328	0.008711	0.024124	0.00288	75.969602	75.969602	75.969602	0	0	1	-360	360;
	418	1396	0.03624	0.109818	0.00917	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5541	2057	0.036777	0.116041	0.00931	60.759372	60.759372	60.759372	0	0	1	-360	360;
	7273	7736	0.003851	0.00995	0.00098	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7273	7688	0.025793	0.056702	0.00652	67.765189	67.765189	67.765189	0	0	1	-360	360;
	2759	7273	0.027942	0.061421	0.00707	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7736	8100	0.021488	0.047248	0.00543	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6807	2859	0.00015	0.00136	0.03769	1250.61177	1250.61177	1250.61177	0	0	1	-360	360;
	1087	8204	0.016909	0.037174	0.00428	91.630424	91.630424	91.630424	0	0	1	-360	360;
	3998	1087	0.040496	0.088512	0.00783	64.588365	64.588365	64.588365	0	0	1	-360	360;
	4512	511	0.035231	0.111132	0.00891	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2001	4512	0.031017	0.086752	0.00785	71.246472	71.246472	71.246472	0	0	1	-360	360;
	3613	2107	0.003983	0.014017	0.00131	58.836779	58.836779	58.836779	0	0	1	-360	360;
	6082	905	0.003521	0.010496	0.00089	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4312	905	0.002686	0.00805	0.00068	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6243	952	0.05195	0.145298	0.01315	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2800	6772	0.023207	0.053132	0.00588	58.835445	58.835445	58.835445	0	0	1	-360	360;
	1929	2800	0.049587	0.150256	0.01255	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2318	1339	0.036719	0.111298	0.0093	57.157677	57.157677	57.157677	0	0	1	-360	360;
	459	2318	0.008364	0.023207	0.00155	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1285	5479	0.005256	0.024653	0.00228	86.869952	86.869952	86.869952	0	0	1	-360	360;
	3554	5479	0.001347	0.004397	0.00058	75.992655	75.992655	75.992655	0	0	1	-360	360;
	7942	6278	0.005149	0.022826	0.00223	105.030089	105.030089	105.030089	0	0	1	-360	360;
	5196	6278	0.001207	0.004992	0.00052	90.801828	90.801828	90.801828	0	0	1	-360	360;
	3462	4616	0.003934	0.009744	0.001	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5066	6543	0.01438	0.04186	0.00364	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1296	1677	0.005744	0.014967	0.00116	75.897011	75.897011	75.897011	0	0	1	-360	360;
	5667	257	0.00176	0.01659	0.37579	1099.76431	1099.76431	1099.76431	0	0	1	-360	360;
	4557	257	0.00151	0.01378	0.37391	1199.45346	1199.45346	1199.45346	0	0	1	-360	360;
	4557	257	0.00128	0.01392	0.27253	1157.29708	1157.29708	1157.29708	0	0	1	-360	360;
	257	3393	0.00052	0.00397	0.12639	1193.91686	1193.91686	1193.91686	0	0	1	-360	360;
	2028	257	0.00221	0.02248	0.46986	1174.87705	1174.87705	1174.87705	0	0	1	-360	360;
	6449	6169	0.007182	0.017107	0.00181	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6169	3734	0.005149	0.011339	0.00131	70.686137	70.686137	70.686137	0	0	1	-360	360;
	7785	3837	0.00943	0.02157	0.00239	60.571324	60.571324	60.571324	0	0	1	-360	360;
	6499	4026	0.040636	0.11257	0.01346	64.009358	64.009358	64.009358	0	0	1	-360	360;
	6881	6499	0.014496	0.042231	0.00367	61.590445	61.590445	61.590445	0	0	1	-360	360;
	1815	752	0.002744	0.00995	0.00095	65.892894	65.892894	65.892894	0	0	1	-360	360;
	752	5967	0.003099	0.009752	0.00107	58.241196	58.241196	58.241196	0	0	1	-360	360;
	540	7998	0.007322	0.016744	0.00185	67.776049	67.776049	67.776049	0	0	1	-360	360;
	5357	7998	0.03024	0.069231	0.00766	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6107	1238	0.008438	0.018669	0.00221	75.165012	75.165012	75.165012	0	0	1	-360	360;
	1238	459	0.006727	0.027942	0.00292	79.609403	79.609403	79.609403	0	0	1	-360	360;
	1238	459	0.006587	0.028488	0.00286	69.294728	69.294728	69.294728	0	0	1	-360	360;
	611	2277	0.014579	0.040463	0.00356	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7168	2277	0.012826	0.036992	0.00424	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2277	4484	0.006909	0.019917	0.00229	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5664	7176	0.003421	0.011041	0.00282	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5664	8096	0.005364	0.016289	0.00213	107.708116	107.708116	107.708116	0	0	1	-360	360;
	4047	3807	0.012463	0.030909	0.00315	62.666915	62.666915	62.666915	0	0	1	-360	360;
	4047	3807	0.012702	0.030273	0.00321	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6384	3807	0.008331	0.02438	0.00218	63.783204	63.783204	63.783204	0	0	1	-360	360;
	6836	3807	0.008149	0.024843	0.00213	77.311473	77.311473	77.311473	0	0	1	-360	360;
	1521	1665	0.010789	0.071169	0.10359	403.743537	403.743537	403.743537	0	0	1	-360	360;
	3507	1665	0.011901	0.08499	0.11417	379.984615	379.984615	379.984615	0	0	1	-360	360;
	1964	3969	0.035942	0.113397	0.0091	77.378729	77.378729	77.378729	0	0	1	-360	360;
	2469	1964	0.035818	0.078752	0.00906	65.780293	65.780293	65.780293	0	0	1	-360	360;
	1964	4045	0.052587	0.147083	0.01331	67.02614	67.02614	67.02614	0	0	1	-360	360;
	5959	1512	0.007273	0.016661	0.00186	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2876	1512	0.023488	0.072488	0.0064	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8100	2851	0.004207	0.011777	0.00107	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7688	2851	0.005339	0.020388	0.00177	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7250	1185	0.007446	0.01638	0.00188	59.851327	59.851327	59.851327	0	0	1	-360	360;
	2782	1185	0.023033	0.057124	0.00582	57.157677	57.157677	57.157677	0	0	1	-360	360;
	169	2440	0.030198	0.095248	0.00765	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5702	4661	0	0.000744	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2758	5702	0.003628	0.012	0.00897	65.388954	65.388954	65.388954	0	0	1	-360	360;
	5702	3554	0.00681	0.028273	0.00295	59.498665	59.498665	59.498665	0	0	1	-360	360;
	4661	5196	0.006669	0.028826	0.00289	60.381179	60.381179	60.381179	0	0	1	-360	360;
	2712	1399	0.04543	0.100521	0.01191	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2712	1399	0.046273	0.098331	0.01213	68.304186	68.304186	68.304186	0	0	1	-360	360;
	2120	1399	0.005041	0.014554	0.00167	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4401	4346	0.007818	0.021653	0.00258	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4346	4848	0.041992	0.154033	0.01391	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3891	203	0.003711	0.009612	0.00094	60.694403	60.694403	60.694403	0	0	1	-360	360;
	3998	3891	0.009959	0.025719	0.00252	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6849	203	0.006744	0.014314	0.00177	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1033	6087	0.023909	0.052901	0.00627	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6087	6052	0.024347	0.051752	0.00639	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3137	860	0.04324	0.088702	0.00877	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3137	482	0.045223	0.125488	0.01057	63.044727	63.044727	63.044727	0	0	1	-360	360;
	4771	1198	0.00305	0.007893	0.00078	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3974	7614	0.01905	0.060107	0.00482	74.541231	74.541231	74.541231	0	0	1	-360	360;
	7614	7995	0.023471	0.055934	0.00594	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7171	2056	0.004893	0.014339	0.00129	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4965	6155	0.022934	0.050397	0.0058	57.157677	57.157677	57.157677	0	0	1	-360	360;
	51	6586	0.037248	0.081901	0.00943	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1203	6586	0.013347	0.039083	0.0035	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4537	6586	0.013058	0.039818	0.00342	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6539	7966	0.04038	0.11757	0.01022	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7966	6849	0.037397	0.111653	0.00946	57.157677	57.157677	57.157677	0	0	1	-360	360;
	40	2023	0.029273	0.069752	0.00741	57.391833	57.391833	57.391833	0	0	1	-360	360;
	40	2365	0.023678	0.070702	0.00599	72.881563	72.881563	72.881563	0	0	1	-360	360;
	1170	6168	0.001868	0.005463	0.00047	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3190	3947	0.00048	0.00423	0.10377	1166.34178	1166.34178	1166.34178	0	0	1	-360	360;
	5305	7963	0.003992	0.010289	0.00101	63.655742	63.655742	63.655742	0	0	1	-360	360;
	5305	3042	0.003777	0.009413	0.00096	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5750	5305	0.015091	0.047174	0.005	116.445048	116.445048	116.445048	0	0	1	-360	360;
	7195	3461	0.013826	0.039917	0.00363	66.956789	66.956789	66.956789	0	0	1	-360	360;
	7195	559	0.015678	0.039124	0.00411	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7195	5541	0.066851	0.192818	0.02215	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7195	8242	0.039653	0.128983	0.01313	59.370441	59.370441	59.370441	0	0	1	-360	360;
	7963	4527	0.005091	0.009702	0.00107	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4527	955	0.004066	0.010083	0.00103	57.157677	57.157677	57.157677	0	0	1	-360	360;
	955	3461	0.007661	0.024058	0.00263	58.90613	58.90613	58.90613	0	0	1	-360	360;
	7110	3461	0.045636	0.13286	0.01155	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3042	955	0.018421	0.042868	0.00387	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1706	5097	0.00805	0.04211	0.07677	340.24174	340.24174	340.24174	0	0	1	-360	360;
	5097	3984	0.003211	0.02194	0.03065	399.074517	399.074517	399.074517	0	0	1	-360	360;
	5097	3939	0.00974	0.05531	0.09304	452.473505	452.473505	452.473505	0	0	1	-360	360;
	8357	1644	0.019719	0.05495	0.00677	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5066	2183	0.03495	0.086686	0.00884	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5719	2183	0.041149	0.122851	0.01041	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3556	540	0.01124	0.027893	0.00285	70.457887	70.457887	70.457887	0	0	1	-360	360;
	4297	1323	0.02219	0.057281	0.00561	62.827528	62.827528	62.827528	0	0	1	-360	360;
	1323	775	0.011843	0.030529	0.00299	64.128627	64.128627	64.128627	0	0	1	-360	360;
	860	574	0.014298	0.039983	0.00362	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8116	4047	0.008083	0.028636	0.00277	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6384	4047	0.012942	0.02962	0.00327	58.407715	58.407715	58.407715	0	0	1	-360	360;
	2403	5278	0.01657	0.050017	0.00569	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1644	2403	0.006331	0.019107	0.00217	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7895	3631	0.003983	0.013537	0.00136	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1414	5667	0.00136	0.01515	0.32663	977.08561	977.08561	977.08561	0	0	1	-360	360;
	5275	8192	0.007876	0.01743	0.00206	68.791169	68.791169	68.791169	0	0	1	-360	360;
	6549	194	0.018636	0.055653	0.00472	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2440	6549	0.010554	0.031504	0.00267	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5185	1589	0.013488	0.039686	0.00383	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1589	4611	0.021934	0.048198	0.00554	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1699	7322	0	0.00086	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1699	7704	0.026463	0.075669	0.0078	58.151268	58.151268	58.151268	0	0	1	-360	360;
	1699	878	0.029446	0.081364	0.00876	70.906575	70.906575	70.906575	0	0	1	-360	360;
	4941	1699	0.055851	0.127826	0.01413	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2759	7322	0.047818	0.124744	0.0097	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4443	3618	0.02562	0.063256	0.00519	66.478951	66.478951	66.478951	0	0	1	-360	360;
	3618	2721	0.006372	0.012298	0.00144	76.822204	76.822204	76.822204	0	0	1	-360	360;
	1604	2490	0.009603	0.021099	0.00243	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8100	6324	0.014917	0.052579	0.00494	57.157677	57.157677	57.157677	0	0	1	-360	360;
	671	6324	0.03014	0.083512	0.00999	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6612	5521	0.03319	0.092818	0.0084	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5521	6157	0.024248	0.067826	0.00613	61.760013	61.760013	61.760013	0	0	1	-360	360;
	7137	7537	0.048529	0.153091	0.01228	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3136	7137	0.055397	0.137421	0.01402	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8373	6156	0	0.000219	0	418.489456	418.489456	418.489456	0	0	1	-360	360;
	8373	4978	0.0013	0.008539	0.01243	418.945193	418.945193	418.945193	0	0	1	-360	360;
	4288	8373	0.007169	0.04463	0.06183	410.887104	410.887104	410.887104	0	0	1	-360	360;
	8373	6550	0.01787	0.097419	0.1706	471.202933	471.202933	471.202933	0	0	1	-360	360;
	6156	4978	0.00164	0.007409	0.01415	393.902129	393.902129	393.902129	0	0	1	-360	360;
	6156	7830	0.0097	0.06274	0.08217	417.919784	417.919784	417.919784	0	0	1	-360	360;
	4288	6156	0.007021	0.045469	0.06052	430.680045	430.680045	430.680045	0	0	1	-360	360;
	2565	7357	0.007496	0.022397	0.0019	67.086918	67.086918	67.086918	0	0	1	-360	360;
	3734	7357	0.007959	0.024124	0.00201	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6940	2724	0.00405	0.011686	0.00134	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6940	1633	0.005471	0.01581	0.00181	71.405942	71.405942	71.405942	0	0	1	-360	360;
	3525	6449	0.012727	0.031579	0.00322	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4280	3525	0.009529	0.027496	0.00315	94.896033	94.896033	94.896033	0	0	1	-360	360;
	834	1883	0	0.000876	0	72.57253	72.57253	72.57253	0	0	1	-360	360;
	7079	1883	0.01243	0.040413	0.00411	61.274744	61.274744	61.274744	0	0	1	-360	360;
	4280	1883	0.002636	0.007967	0.0009	69.679971	69.679971	69.679971	0	0	1	-360	360;
	4280	1883	0.002686	0.007802	0.00092	74.785676	74.785676	74.785676	0	0	1	-360	360;
	834	4122	0.049421	0.127587	0.01251	67.451012	67.451012	67.451012	0	0	1	-360	360;
	4980	834	0.006587	0.018248	0.00218	64.696393	64.696393	64.696393	0	0	1	-360	360;
	834	1791	0.064438	0.166364	0.01631	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6682	1671	0.085322	0.155132	0.0173	57.157677	57.157677	57.157677	0	0	1	-360	360;
	59	4443	0.050488	0.130339	0.01277	70.636981	70.636981	70.636981	0	0	1	-360	360;
	8012	59	0.035471	0.111893	0.00898	66.772169	66.772169	66.772169	0	0	1	-360	360;
	6539	5028	0.013876	0.045132	0.00459	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5028	3998	0.042744	0.16319	0.01416	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4559	4234	0.027512	0.080083	0.00696	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3671	4559	0.013132	0.038223	0.00332	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7860	6817	0.017347	0.051802	0.00439	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3136	6817	0.036446	0.086851	0.00922	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3529	76	0	0.000893	0	102.53287	102.53287	102.53287	0	0	1	-360	360;
	5659	76	0.013223	0.056876	0.00553	57.157677	57.157677	57.157677	0	0	1	-360	360;
	76	5402	0.01662	0.035322	0.00436	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6171	3529	0.023752	0.069157	0.00601	81.376718	81.376718	81.376718	0	0	1	-360	360;
	3529	4723	0.029851	0.100967	0.0125	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3529	3778	0.014322	0.031504	0.00363	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3529	7072	0.01481	0.040017	0.00388	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3969	4045	0.038066	0.094413	0.00963	60.291442	60.291442	60.291442	0	0	1	-360	360;
	5117	5207	0.018702	0.058967	0.00473	66.096185	66.096185	66.096185	0	0	1	-360	360;
	7072	5117	0.008554	0.023917	0.00217	66.219836	66.219836	66.219836	0	0	1	-360	360;
	1605	1706	0.009849	0.05651	0.08348	411.435055	411.435055	411.435055	0	0	1	-360	360;
	7682	671	0.013554	0.032066	0.00359	73.749979	73.749979	73.749979	0	0	1	-360	360;
	7682	2270	0.013719	0.042876	0.00454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6732	1141	0.032512	0.094636	0.00823	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1141	3846	0.040008	0.116479	0.01013	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7039	385	0.02319	0.055273	0.00587	57.157677	57.157677	57.157677	0	0	1	-360	360;
	878	7039	0.013248	0.046413	0.00392	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4941	7039	0.073289	0.23843	0.02428	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5864	274	0.008083	0.029661	0.00268	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6516	274	0.014661	0.058273	0.00486	63.092358	63.092358	63.092358	0	0	1	-360	360;
	4397	3241	0.017273	0.044587	0.00437	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3241	1555	0.016413	0.049	0.00415	75.052602	75.052602	75.052602	0	0	1	-360	360;
	1563	6981	0.054587	0.120008	0.01382	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5772	1563	0.049521	0.144149	0.01253	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7842	2349	0.02757	0.063124	0.00698	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6669	4297	0.037686	0.089818	0.00954	72.55062	72.55062	72.55062	0	0	1	-360	360;
	5066	559	0.106033	0.321289	0.02683	66.577452	66.577452	66.577452	0	0	1	-360	360;
	8348	5066	0.066959	0.153273	0.01695	76.383233	76.383233	76.383233	0	0	1	-360	360;
	5066	6088	0.074421	0.208132	0.01884	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5066	8362	0.054157	0.123967	0.01371	76.995391	76.995391	76.995391	0	0	1	-360	360;
	613	3843	0.086802	0.24276	0.02196	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3843	5359	0.007628	0.018826	0.00155	61.077741	61.077741	61.077741	0	0	1	-360	360;
	2507	532	0.021256	0.050669	0.00538	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5936	97	0.015893	0.046223	0.00402	59.092273	59.092273	59.092273	0	0	1	-360	360;
	97	3270	0.01562	0.034331	0.00395	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7085	5455	0.011174	0.023008	0.00227	74.269732	74.269732	74.269732	0	0	1	-360	360;
	433	7085	0.014438	0.036174	0.00292	78.725745	78.725745	78.725745	0	0	1	-360	360;
	5455	3270	0.012636	0.029752	0.0025	57.16968	57.16968	57.16968	0	0	1	-360	360;
	2724	7079	0	0.000826	0	67.968289	67.968289	67.968289	0	0	1	-360	360;
	2724	2565	0.00086	0.003157	0.00029	60.699357	60.699357	60.699357	0	0	1	-360	360;
	2724	1633	0.01005	0.027843	0.00333	89.241233	89.241233	89.241233	0	0	1	-360	360;
	7079	931	0.000843	0.003223	0.00029	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5604	1815	0.022983	0.064256	0.00581	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1815	5502	0.008603	0.018273	0.00226	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1815	5502	0.007661	0.020702	0.00201	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1815	5967	0.002678	0.007174	0.00092	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1815	5967	0.002388	0.008132	0.00082	63.232013	63.232013	63.232013	0	0	1	-360	360;
	1613	652	0.00237	0.011781	0.0227	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2050	3215	0.001901	0.012831	0.01822	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2012	3685	0	0.00081	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2012	5106	0.017273	0.025289	0.00279	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3685	1007	0.030909	0.042397	0.00506	73.728068	73.728068	73.728068	0	0	1	-360	360;
	3685	960	0.020967	0.066149	0.00531	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4368	5182	0.049719	0.109298	0.01258	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5182	2107	0.065909	0.144901	0.01668	80.292246	80.292246	80.292246	0	0	1	-360	360;
	1101	6104	0.00614	0.05775	1.30752	1180.04639	1180.04639	1180.04639	0	0	1	-360	360;
	5185	7942	0.020678	0.079256	0.00897	60.659727	60.659727	60.659727	0	0	1	-360	360;
	2023	7942	0.020529	0.051702	0.00562	69.620527	69.620527	69.620527	0	0	1	-360	360;
	1285	3184	0.015066	0.050711	0.0043	64.759267	64.759267	64.759267	0	0	1	-360	360;
	225	5885	0.032364	0.074058	0.00819	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2291	225	0.024893	0.069612	0.0063	60.249907	60.249907	60.249907	0	0	1	-360	360;
	3501	6981	0.051289	0.112769	0.01298	68.747348	68.747348	68.747348	0	0	1	-360	360;
	7108	8100	0.024124	0.069562	0.00799	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8100	1159	0.026769	0.10219	0.01121	88.596114	88.596114	88.596114	0	0	1	-360	360;
	7198	8100	0.025909	0.077347	0.00655	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8100	702	0.01324	0.046628	0.00554	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7635	7086	0.01676	0.03686	0.00425	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2270	7635	0.017612	0.043678	0.00445	77.229547	77.229547	77.229547	0	0	1	-360	360;
	7250	4152	0.02043	0.057124	0.00516	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5733	6363	0.00084	0.00669	0.20687	1011.74138	1011.74138	1011.74138	0	0	1	-360	360;
	1694	3924	0.019893	0.062752	0.00503	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5810	68	0.058719	0.164223	0.01486	77.188584	77.188584	77.188584	0	0	1	-360	360;
	5810	2291	0.021661	0.049587	0.00549	67.210569	67.210569	67.210569	0	0	1	-360	360;
	3346	1842	0.008	0.030686	0.00347	73.318438	73.318438	73.318438	0	0	1	-360	360;
	3371	4124	0.024017	0.057248	0.00608	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4242	650	0.005636	0.020727	0.00187	83.714467	83.714467	83.714467	0	0	1	-360	360;
	4689	6392	0.004769	0.016826	0.00158	63.925146	63.925146	63.925146	0	0	1	-360	360;
	7388	4689	0.00895	0.031554	0.00297	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7211	3460	0.010919	0.07275	0.10428	369.323185	369.323185	369.323185	0	0	1	-360	360;
	3460	7401	0.004841	0.02861	0.04624	427.869412	427.869412	427.869412	0	0	1	-360	360;
	4827	2688	0.044909	0.092107	0.00911	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2876	2768	0.011901	0.039537	0.00337	62.938223	62.938223	62.938223	0	0	1	-360	360;
	365	8224	0.008421	0.020066	0.00213	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8224	1371	0.07076	0.184612	0.01435	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8224	4536	0.00624	0.018917	0.00158	61.838509	61.838509	61.838509	0	0	1	-360	360;
	8224	1835	0.055157	0.127587	0.01119	59.699669	59.699669	59.699669	0	0	1	-360	360;
	5207	685	0.03562	0.084876	0.00901	64.780987	64.780987	64.780987	0	0	1	-360	360;
	931	7663	0.024248	0.073463	0.00614	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1677	2041	0.006372	0.012686	0.00126	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6214	415	0.003397	0.009802	0.00112	70.815122	70.815122	70.815122	0	0	1	-360	360;
	6214	4484	0.002793	0.008397	0.00092	57.157677	57.157677	57.157677	0	0	1	-360	360;
	111	6745	0.035636	0.100248	0.00935	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2941	6745	0.034901	0.102182	0.00916	61.856609	61.856609	61.856609	0	0	1	-360	360;
	1089	5125	0.032388	0.094281	0.0082	61.285604	61.285604	61.285604	0	0	1	-360	360;
	4401	5125	0.031421	0.093802	0.00795	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7211	4528	0.02455	0.175521	0.23579	433.269288	433.269288	433.269288	0	0	1	-360	360;
	7100	4528	0.011281	0.058469	0.10824	330.686119	330.686119	330.686119	0	0	1	-360	360;
	4528	364	0.015979	0.102281	0.1526	336.733783	336.733783	336.733783	0	0	1	-360	360;
	6659	5866	0	0.000777	0	58.182133	58.182133	58.182133	0	0	1	-360	360;
	6659	7615	0.010868	0.037488	0.0029	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6659	7135	0.011967	0.024562	0.00243	61.868231	61.868231	61.868231	0	0	1	-360	360;
	5866	7615	0.010636	0.034711	0.00284	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5866	1593	0.055091	0.164496	0.01394	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5866	6902	0.073091	0.221488	0.0185	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7497	7509	0.01938	0.056421	0.00491	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2850	7509	0.018983	0.057496	0.0048	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4368	6384	0.031413	0.07795	0.00796	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4368	6621	0.046231	0.119372	0.0117	73.576791	73.576791	73.576791	0	0	1	-360	360;
	4820	2644	0.012066	0.029901	0.00305	57.157677	57.157677	57.157677	0	0	1	-360	360;
	946	302	0.022529	0.071066	0.0057	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4124	946	0.027471	0.065471	0.00695	57.157677	57.157677	57.157677	0	0	1	-360	360;
	650	3925	0	0.00076	0	69.352076	69.352076	69.352076	0	0	1	-360	360;
	7388	3925	0.005868	0.020678	0.00194	61.415924	61.415924	61.415924	0	0	1	-360	360;
	1396	562	0.022636	0.049769	0.00573	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6066	1007	0.033909	0.047901	0.0051	67.666306	67.666306	67.666306	0	0	1	-360	360;
	6066	1007	0.034537	0.04686	0.00519	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1007	5106	0.012959	0.028661	0.00249	70.513711	70.513711	70.513711	0	0	1	-360	360;
	2774	7159	0.003919	0.02719	0.03744	607.752622	607.752622	607.752622	0	0	1	-360	360;
	4774	2869	0	0.000744	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3572	4774	0.013909	0.038909	0.00352	80.001885	80.001885	80.001885	0	0	1	-360	360;
	1083	2869	0.007579	0.023	0.00192	59.437315	59.437315	59.437315	0	0	1	-360	360;
	1489	4122	0.032587	0.077653	0.00825	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4122	1412	0.035744	0.085851	0.00694	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1088	1977	0.011969	0.0687	0.10148	404.064764	404.064764	404.064764	0	0	1	-360	360;
	1977	2923	0.00431	0.024719	0.03652	447.593764	447.593764	447.593764	0	0	1	-360	360;
	301	4443	0.107463	0.250107	0.02257	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4443	780	0.043099	0.099678	0.00874	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4443	1555	0.033058	0.075669	0.00837	78.863685	78.863685	78.863685	0	0	1	-360	360;
	7706	4443	0.045273	0.099537	0.01146	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4443	2236	0.071182	0.141157	0.01496	69.037519	69.037519	69.037519	0	0	1	-360	360;
	4443	2236	0.072554	0.138223	0.01524	76.793434	76.793434	76.793434	0	0	1	-360	360;
	955	559	0.00781	0.02357	0.00268	60.600093	60.600093	60.600093	0	0	1	-360	360;
	559	5912	0.014298	0.042702	0.00362	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8017	7303	0.016793	0.041661	0.00425	65.203572	65.203572	65.203572	0	0	1	-360	360;
	7303	6884	0.011017	0.021711	0.00224	78.050141	78.050141	78.050141	0	0	1	-360	360;
	1922	7211	0.01255	0.08125	0.10642	499.847312	499.847312	499.847312	0	0	1	-360	360;
	8187	7124	0.008041	0.022488	0.00204	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8187	7124	0.009149	0.017438	0.00192	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8187	6027	0.032876	0.06481	0.00667	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7326	8187	0.022512	0.056769	0.00473	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7124	1562	0.01276	0.035702	0.00323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1615	5520	0.007793	0.02457	0.00197	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3472	5520	0.00762	0.022752	0.00193	74.719754	74.719754	74.719754	0	0	1	-360	360;
	2101	5966	0.010157	0.025198	0.00257	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2101	4591	0.018289	0.054603	0.00463	75.922351	75.922351	75.922351	0	0	1	-360	360;
	2101	5093	0.031876	0.082322	0.00807	57.947024	57.947024	57.947024	0	0	1	-360	360;
	8353	1615	0.011289	0.033917	0.00374	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4941	7108	0.055537	0.203711	0.01839	71.64448	71.64448	71.64448	0	0	1	-360	360;
	2848	7943	0.001488	0.003314	0.04138	121.457396	121.457396	121.457396	0	0	1	-360	360;
	7293	2848	0.009083	0.012124	0.0998	72.79354	72.79354	72.79354	0	0	1	-360	360;
	3352	2918	0	0.000876	0	95.559824	95.559824	95.559824	0	0	1	-360	360;
	5678	3190	0.0011	0.01267	0.23494	1079.22714	1079.22714	1079.22714	0	0	1	-360	360;
	3190	7373	0.00252	0.02751	0.5387	1151.60119	1151.60119	1151.60119	0	0	1	-360	360;
	4234	960	0.04314	0.128802	0.01092	77.189728	77.189728	77.189728	0	0	1	-360	360;
	3572	1083	0.011562	0.029843	0.00293	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7074	724	0.011256	0.041174	0.0033	61.902145	61.902145	61.902145	0	0	1	-360	360;
	7074	724	0.011017	0.04195	0.00323	61.61026	61.61026	61.61026	0	0	1	-360	360;
	6772	359	0.029496	0.076149	0.00747	60.32726	60.32726	60.32726	0	0	1	-360	360;
	7401	1088	0.009401	0.05757	0.07966	384.622008	384.622008	384.622008	0	0	1	-360	360;
	4401	1089	0.056934	0.185223	0.01886	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4848	1089	0.005479	0.016479	0.00181	65.303598	65.303598	65.303598	0	0	1	-360	360;
	7840	6135	0.011289	0.029157	0.00286	77.120948	77.120948	77.120948	0	0	1	-360	360;
	7840	5686	0.000504	0.001446	0.00012	67.676213	67.676213	67.676213	0	0	1	-360	360;
	4895	7840	0.014347	0.035802	0.00376	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4895	7840	0.014628	0.035066	0.00384	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7840	5573	0.019537	0.05462	0.00494	64.897969	64.897969	64.897969	0	0	1	-360	360;
	1633	4980	0.003281	0.009488	0.00109	83.183091	83.183091	83.183091	0	0	1	-360	360;
	5750	2124	0.033339	0.086091	0.00844	74.712704	74.712704	74.712704	0	0	1	-360	360;
	8362	2124	0.032628	0.095	0.00826	63.772535	63.772535	63.772535	0	0	1	-360	360;
	5278	6325	0.008281	0.029331	0.00284	58.306165	58.306165	58.306165	0	0	1	-360	360;
	183	491	0.008711	0.059841	0.08369	404.655393	404.655393	404.655393	0	0	1	-360	360;
	183	6550	0.0127	0.06389	0.12127	408.894968	408.894968	408.894968	0	0	1	-360	360;
	611	2255	0	0.000826	0	104.047548	104.047548	104.047548	0	0	1	-360	360;
	611	7168	0.016992	0.051281	0.00583	57.157677	57.157677	57.157677	0	0	1	-360	360;
	611	415	0.017322	0.050223	0.00595	57.157677	57.157677	57.157677	0	0	1	-360	360;
	611	3206	0.016901	0.06357	0.0056	59.447604	59.447604	59.447604	0	0	1	-360	360;
	611	3058	0.037231	0.112777	0.00942	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4611	611	0.039636	0.115388	0.01003	69.514023	69.514023	69.514023	0	0	1	-360	360;
	2365	611	0.067339	0.212438	0.01704	58.852021	58.852021	58.852021	0	0	1	-360	360;
	775	6135	0.061322	0.134818	0.01552	62.96623	62.96623	62.96623	0	0	1	-360	360;
	2157	415	0.04114	0.119752	0.01041	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1290	1509	0.03657	0.109198	0.00925	57.157677	57.157677	57.157677	0	0	1	-360	360;
	837	1290	0.042488	0.097248	0.01075	77.806078	77.806078	77.806078	0	0	1	-360	360;
	1290	4310	0.039099	0.089512	0.00989	61.646269	61.646269	61.646269	0	0	1	-360	360;
	1745	1567	0.013537	0.032471	0.00356	72.734477	72.734477	72.734477	0	0	1	-360	360;
	4557	4307	0.00158	0.01747	0.33681	1010.35657	1010.35657	1010.35657	0	0	1	-360	360;
	1851	4978	0.00894	0.04824	0.08579	426.965939	426.965939	426.965939	0	0	1	-360	360;
	4288	1851	0.003531	0.018341	0.03394	402.593906	402.593906	402.593906	0	0	1	-360	360;
	7569	6732	0.029818	0.068256	0.00755	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5697	5208	0	0.00081	0	67.934947	67.934947	67.934947	0	0	1	-360	360;
	5697	6988	0.003512	0.004661	0.0935	67.689169	67.689169	67.689169	0	0	1	-360	360;
	4045	3006	0.010959	0.027322	0.00568	100.320296	100.320296	100.320296	0	0	1	-360	360;
	4199	5208	0.003579	0.004562	0.09527	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4753	1694	0.045843	0.136851	0.0116	93.890058	93.890058	93.890058	0	0	1	-360	360;
	4753	3415	0.044678	0.135421	0.01131	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4753	1600	0.015793	0.04786	0.00399	78.324879	78.324879	78.324879	0	0	1	-360	360;
	6389	3846	0.036248	0.10138	0.00917	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1404	58	0.00336	0.02922	0.71685	1283.41016	1283.41016	1283.41016	0	0	1	-360	360;
	5026	7100	0.01663	0.097229	0.1596	429.390949	429.390949	429.390949	0	0	1	-360	360;
	8124	194	0	0.000777	0	82.491483	82.491483	82.491483	0	0	1	-360	360;
	194	6027	0.09357	0.184397	0.01897	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1848	194	0.090248	0.177719	0.01762	59.463227	59.463227	59.463227	0	0	1	-360	360;
	1848	194	0.091983	0.174017	0.01796	57.157677	57.157677	57.157677	0	0	1	-360	360;
	194	1139	0.018256	0.05314	0.00462	100.374596	100.374596	100.374596	0	0	1	-360	360;
	194	7326	0.112273	0.205446	0.02358	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2164	3126	0.003401	0.02362	0.03252	438.873789	438.873789	438.873789	0	0	1	-360	360;
	2164	6816	0.002719	0.012781	0.02344	405.740246	405.740246	405.740246	0	0	1	-360	360;
	2164	1989	0.00205	0.022719	0.08382	451.980044	451.980044	451.980044	0	0	1	-360	360;
	7265	7027	0.036579	0.090719	0.00926	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7265	6875	0.038802	0.092446	0.00982	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1758	4574	0.029397	0.067298	0.00744	70.830936	70.830936	70.830936	0	0	1	-360	360;
	1758	8253	0.054289	0.124289	0.01374	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4574	7171	0.007421	0.023438	0.00188	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4574	4781	0.006124	0.018554	0.00155	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2956	7569	0.010446	0.034017	0.00347	72.947294	72.947294	72.947294	0	0	1	-360	360;
	714	2562	0.045702	0.127802	0.01156	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4723	6387	0.057438	0.171496	0.01454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4473	6387	0.025182	0.057636	0.00637	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4216	2576	0.03224	0.075033	0.00677	57.157677	57.157677	57.157677	0	0	1	-360	360;
	301	780	0.075653	0.13295	0.01588	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	7074	0.005058	0.017017	0.0016	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	7074	0.005157	0.016669	0.00163	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2107	3705	0.046777	0.101124	0.01084	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5749	463	0.045364	0.135446	0.01148	62.218798	62.218798	62.218798	0	0	1	-360	360;
	6384	5966	0.037471	0.113529	0.00948	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4127	1955	0.058372	0.168446	0.01531	61.247499	61.247499	61.247499	0	0	1	-360	360;
	4127	1955	0.066157	0.165083	0.01735	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2850	1955	0.046628	0.120388	0.0118	80.106293	80.106293	80.106293	0	0	1	-360	360;
	4876	4173	0.029868	0.087488	0.00783	63.772154	63.772154	63.772154	0	0	1	-360	360;
	4333	2507	0.015562	0.047107	0.00394	57.545015	57.545015	57.545015	0	0	1	-360	360;
	7086	1928	0.045967	0.109562	0.01164	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	7279	0.00705	0.016876	0.00185	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	3238	0.004	0.009537	0.00102	57.157677	57.157677	57.157677	0	0	1	-360	360;
	613	2395	0.06243	0.161207	0.0158	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1053	7135	0.005942	0.014149	0.00151	78.588566	78.588566	78.588566	0	0	1	-360	360;
	7883	5879	0.004651	0.024159	0.04472	427.827496	427.827496	427.827496	0	0	1	-360	360;
	7883	1425	0.00474	0.02363	0.04555	462.497056	462.497056	462.497056	0	0	1	-360	360;
	6849	8204	0.036669	0.115669	0.00928	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7928	5255	0.043273	0.095132	0.01095	63.186287	63.186287	63.186287	0	0	1	-360	360;
	7928	5147	0.004521	0.011661	0.00114	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2905	5678	0.00207	0.01871	0.44089	1308.45651	1308.45651	1308.45651	0	0	1	-360	360;
	2923	1605	0.00419	0.02712	0.03551	387.146853	387.146853	387.146853	0	0	1	-360	360;
	2923	1605	0.002349	0.022039	0.05608	429.784575	429.784575	429.784575	0	0	1	-360	360;
	5541	4533	0.02	0.04314	0.0048	69.215279	69.215279	69.215279	0	0	1	-360	360;
	7213	3184	0.00676	0.016124	0.00172	58.179084	58.179084	58.179084	0	0	1	-360	360;
	6721	6855	0.007876	0.022529	0.00231	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6855	3472	0.025793	0.060298	0.00604	64.309055	64.309055	64.309055	0	0	1	-360	360;
	5940	6389	0.067587	0.204793	0.0171	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5502	6612	0.053008	0.116554	0.01342	73.292908	73.292908	73.292908	0	0	1	-360	360;
	6073	6027	0.01195	0.030868	0.00303	70.776255	70.776255	70.776255	0	0	1	-360	360;
	5147	6073	0.015463	0.04876	0.00391	57.157677	57.157677	57.157677	0	0	1	-360	360;
	459	2782	0.018702	0.071413	0.00619	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2834	4307	0.00142	0.01555	0.30433	767.876077	767.876077	767.876077	0	0	1	-360	360;
	2834	1404	0.00123	0.01312	0.30344	1109.64621	1109.64621	1109.64621	0	0	1	-360	360;
	2834	1404	0.00121	0.01337	0.29699	1216.92023	1216.92023	1216.92023	0	0	1	-360	360;
	1414	2834	0.00079	0.01226	0.35947	1096.49447	1096.49447	1096.49447	0	0	1	-360	360;
	2633	2834	0.00062	0.00673	0.13171	778.657713	778.657713	778.657713	0	0	1	-360	360;
	5551	4978	0	0.000211	0	325.825049	325.825049	325.825049	0	0	1	-360	360;
	4288	4978	0.00763	0.05451	0.07321	387.590397	387.590397	387.590397	0	0	1	-360	360;
	6449	7663	0	0.00076	0	72.497463	72.497463	72.497463	0	0	1	-360	360;
	8116	5300	0.00657	0.019603	0.00166	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6082	5774	0.011512	0.028562	0.00291	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7198	7704	0.014612	0.05362	0.00484	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8317	365	0.00976	0.02419	0.00246	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1396	4238	0.007876	0.022934	0.002	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4238	3415	0.01795	0.042769	0.00454	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5864	5340	0.023488	0.046364	0.007	71.582941	71.582941	71.582941	0	0	1	-360	360;
	5864	1567	0.013802	0.031802	0.00363	68.330097	68.330097	68.330097	0	0	1	-360	360;
	2756	6743	0.036678	0.080636	0.00928	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6743	3962	0.005917	0.017066	0.00197	61.493849	61.493849	61.493849	0	0	1	-360	360;
	7213	1688	0.003364	0.012901	0.00115	80.471912	80.471912	80.471912	0	0	1	-360	360;
	2157	7213	0.038107	0.120198	0.00964	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5185	1688	0.003289	0.011942	0.00113	63.852174	63.852174	63.852174	0	0	1	-360	360;
	837	3889	0.047818	0.109479	0.01211	71.027177	71.027177	71.027177	0	0	1	-360	360;
	3889	4093	0.023603	0.060744	0.00595	75.188447	75.188447	75.188447	0	0	1	-360	360;
	3889	4541	0.042851	0.125488	0.01124	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3889	4541	0.04376	0.123124	0.01148	57.641231	57.641231	57.641231	0	0	1	-360	360;
	4307	1404	0.00089	0.0071	0.18875	1185.35526	1185.35526	1185.35526	0	0	1	-360	360;
	7569	5940	0	0.000744	0	95.269844	95.269844	95.269844	0	0	1	-360	360;
	7569	7384	0.039554	0.086942	0.01001	57.157677	57.157677	57.157677	0	0	1	-360	360;
	459	3061	0.046364	0.078843	0.00892	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7343	3338	0.006248	0.007702	0.0675	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8065	7343	0.001273	0.002661	0.06447	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5885	3974	0.027711	0.061314	0.00727	62.407038	62.407038	62.407038	0	0	1	-360	360;
	5885	3974	0.028231	0.059967	0.0074	57.157677	57.157677	57.157677	0	0	1	-360	360;
	954	5255	0.029702	0.056231	0.00603	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2721	5824	0.006256	0.011397	0.00118	71.436998	71.436998	71.436998	0	0	1	-360	360;
	4170	4360	0.010298	0.032512	0.00261	59.626507	59.626507	59.626507	0	0	1	-360	360;
	4360	8300	0.01805	0.039686	0.00457	64.458427	64.458427	64.458427	0	0	1	-360	360;
	3191	6459	0.012849	0.07226	0.12346	592.330719	592.330719	592.330719	0	0	1	-360	360;
	3191	7783	0.013099	0.07075	0.12583	637.033738	637.033738	637.033738	0	0	1	-360	360;
	1631	7515	0.008314	0.033033	0.00275	93.660474	93.660474	93.660474	0	0	1	-360	360;
	6171	7978	0.031843	0.096463	0.00806	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7984	3545	0.002207	0.005273	0.00056	61.27322	61.27322	61.27322	0	0	1	-360	360;
	5899	7984	0.009124	0.027669	0.00212	86.819081	86.819081	86.819081	0	0	1	-360	360;
	1509	4876	0.010777	0.032653	0.00273	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1033	7236	0.00757	0.018901	0.00199	66.677097	66.677097	66.677097	0	0	1	-360	360;
	2695	6052	0.007719	0.018512	0.00203	67.191707	67.191707	67.191707	0	0	1	-360	360;
	3612	4216	0.016198	0.040818	0.0034	57.157677	57.157677	57.157677	0	0	1	-360	360;
	667	2611	0.054512	0.162752	0.0138	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2611	2494	0.030802	0.067719	0.0078	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4152	6465	0.012298	0.029281	0.00311	57.157677	57.157677	57.157677	0	0	1	-360	360;
	253	7516	0.074322	0.221901	0.01881	57.157677	57.157677	57.157677	0	0	1	-360	360;
	845	7516	0.009669	0.023033	0.00245	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4895	1848	0	0.000843	0	62.221275	62.221275	62.221275	0	0	1	-360	360;
	39	1171	0.073463	0.189678	0.01859	58.740944	58.740944	58.740944	0	0	1	-360	360;
	1171	960	0.058653	0.145488	0.01484	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1414	3393	0.00102	0.01773	0.46086	767.876077	767.876077	767.876077	0	0	1	-360	360;
	4723	3873	0.00838	0.02438	0.00212	60.19313	60.19313	60.19313	0	0	1	-360	360;
	4723	6542	0.01219	0.03024	0.00309	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4723	4473	0.048744	0.11614	0.01234	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6542	3083	0.012967	0.030934	0.00329	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3873	3396	0.006983	0.021149	0.00177	63.125319	63.125319	63.125319	0	0	1	-360	360;
	5659	3873	0.017347	0.058702	0.00726	59.428932	59.428932	59.428932	0	0	1	-360	360;
	2236	6547	0.02457	0.058562	0.00622	69.458009	69.458009	69.458009	0	0	1	-360	360;
	5916	4931	0.016083	0.044983	0.00407	73.934217	73.934217	73.934217	0	0	1	-360	360;
	4425	4931	0.004843	0.011512	0.00122	62.874778	62.874778	62.874778	0	0	1	-360	360;
	7110	7580	0.005058	0.011339	0.00099	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6351	2315	0.023727	0.074843	0.00601	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7978	2315	0.031157	0.098281	0.00788	57.157677	57.157677	57.157677	0	0	1	-360	360;
	702	671	0.014372	0.048661	0.00602	61.736197	61.736197	61.736197	0	0	1	-360	360;
	881	2905	0.00331	0.02754	0.70311	1082.05534	1082.05534	1082.05534	0	0	1	-360	360;
	6816	6598	0.002669	0.01732	0.02305	458.85154	458.85154	458.85154	0	0	1	-360	360;
	6598	7030	0.019719	0.103289	0.1883	442.757462	442.757462	442.757462	0	0	1	-360	360;
	6816	3126	0.00261	0.016039	0.02254	401.775789	401.775789	401.775789	0	0	1	-360	360;
	433	5936	0.008587	0.017446	0.00163	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8017	1278	0	0.000876	0	73.388171	73.388171	73.388171	0	0	1	-360	360;
	7567	8017	0.017413	0.05443	0.00577	61.23702	61.23702	61.23702	0	0	1	-360	360;
	8017	5764	0.024744	0.074289	0.00819	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1278	4801	0.036421	0.11381	0.01206	68.668471	68.668471	68.668471	0	0	1	-360	360;
	4588	4328	0.004198	0.012587	0.00139	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7366	6486	0.014686	0.044521	0.00372	57.733254	57.733254	57.733254	0	0	1	-360	360;
	2365	6486	0.015231	0.041256	0.00348	57.730397	57.730397	57.730397	0	0	1	-360	360;
	51	6155	0.01438	0.045364	0.00364	68.173485	68.173485	68.173485	0	0	1	-360	360;
	6660	917	0.025562	0.076298	0.00646	57.157677	57.157677	57.157677	0	0	1	-360	360;
	917	5772	0.028083	0.061744	0.00711	57.157677	57.157677	57.157677	0	0	1	-360	360;
	32	4665	0.05019	0.152107	0.0127	57.157677	57.157677	57.157677	0	0	1	-360	360;
	635	4665	0.027909	0.072033	0.00706	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1791	4228	0.020512	0.061248	0.00519	57.157677	57.157677	57.157677	0	0	1	-360	360;
	775	5489	0.025364	0.053893	0.00666	66.328816	66.328816	66.328816	0	0	1	-360	360;
	7537	6539	0.054479	0.165099	0.01379	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7537	6539	0.055636	0.161983	0.01408	59.309092	59.309092	59.309092	0	0	1	-360	360;
	7537	2847	0.051579	0.154	0.01305	57.157677	57.157677	57.157677	0	0	1	-360	360;
	960	482	0.035752	0.106752	0.00904	59.996698	59.996698	59.996698	0	0	1	-360	360;
	574	6066	0.021124	0.054537	0.00534	57.157677	57.157677	57.157677	0	0	1	-360	360;
	574	6066	0.021537	0.05343	0.00545	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6045	491	0.0082	0.04119	0.07819	391.023668	391.023668	391.023668	0	0	1	-360	360;
	3974	7995	0.040281	0.122058	0.01019	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4876	4093	0.031934	0.10086	0.0081	74.894085	74.894085	74.894085	0	0	1	-360	360;
	4876	1489	0.048347	0.145008	0.01238	57.944547	57.944547	57.944547	0	0	1	-360	360;
	4876	4310	0.056636	0.164901	0.01433	60.731937	60.731937	60.731937	0	0	1	-360	360;
	4127	6556	0.048091	0.119281	0.01217	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4127	4459	0.00305	0.008793	0.0008	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4127	4459	0.003463	0.00862	0.00091	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5359	4127	0.004223	0.009339	0.0011	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5359	4127	0.005281	0.009603	0.00107	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5749	4082	0.028033	0.066033	0.00705	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2397	8353	0.012983	0.051587	0.0043	75.226171	75.226171	75.226171	0	0	1	-360	360;
	6086	4001	0.00183	0.01416	0.45138	1014.67884	1014.67884	1014.67884	0	0	1	-360	360;
	3778	6404	0.012504	0.036397	0.00317	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7978	6404	0.024488	0.074198	0.00619	66.650614	66.650614	66.650614	0	0	1	-360	360;
	3533	7280	0.010579	0.045496	0.00443	65.112882	65.112882	65.112882	0	0	1	-360	360;
	6556	1593	0.025372	0.065512	0.00643	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4581	5977	0.015058	0.042124	0.00381	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3244	4581	0.038248	0.084099	0.00968	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7443	667	0.045719	0.113397	0.01158	58.887077	58.887077	58.887077	0	0	1	-360	360;
	7443	3559	0.046603	0.111041	0.0118	64.795466	64.795466	64.795466	0	0	1	-360	360;
	685	7512	0.013843	0.041934	0.0035	69.049903	69.049903	69.049903	0	0	1	-360	360;
	8012	7706	0.052083	0.134446	0.01318	57.157677	57.157677	57.157677	0	0	1	-360	360;
	677	253	0.013579	0.03981	0.00356	71.998096	71.998096	71.998096	0	0	1	-360	360;
	677	253	0.013868	0.039058	0.00364	72.752577	72.752577	72.752577	0	0	1	-360	360;
	677	6339	0.035099	0.077182	0.00889	69.549461	69.549461	69.549461	0	0	1	-360	360;
	8021	677	0.11376	0.27386	0.02307	57.157677	57.157677	57.157677	0	0	1	-360	360;
	253	7321	0.069628	0.153091	0.01762	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7373	4001	0.00296	0.03263	0.62953	992.511358	992.511358	992.511358	0	0	1	-360	360;
	3472	4910	0.048322	0.115149	0.01223	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1170	4628	0.009802	0.028678	0.00257	76.850211	76.850211	76.850211	0	0	1	-360	360;
	7293	3338	0.003132	0.004347	0.03379	66.00397	66.00397	66.00397	0	0	1	-360	360;
	3048	4269	0.00284	0.02937	0.7024	1431.58936	1431.58936	1431.58936	0	0	1	-360	360;
	4269	6363	0.00353	0.03885	0.74955	1045.54481	1045.54481	1045.54481	0	0	1	-360	360;
	1101	4269	0.00193	0.02044	0.41037	1214.46324	1214.46324	1214.46324	0	0	1	-360	360;
	3507	1521	0.00918	0.065591	0.08811	465.142314	465.142314	465.142314	0	0	1	-360	360;
	6959	3507	0	0.000211	0	430.082557	430.082557	430.082557	0	0	1	-360	360;
	3507	7883	0.01776	0.09207	0.17045	436.878224	436.878224	436.878224	0	0	1	-360	360;
	3507	7883	0.018081	0.09007	0.1736	428.356395	428.356395	428.356395	0	0	1	-360	360;
	3507	3566	0.0017	0.0117	0.01636	289.342829	289.342829	289.342829	0	0	1	-360	360;
	3507	1027	0.001669	0.011919	0.01601	286.093986	286.093986	286.093986	0	0	1	-360	360;
	3800	8317	0.009041	0.023331	0.00228	71.972565	71.972565	71.972565	0	0	1	-360	360;
	2232	3051	0.075851	0.15557	0.01538	66.307668	66.307668	66.307668	0	0	1	-360	360;
	1859	2232	0.012471	0.036587	0.0029	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2397	6721	0.003702	0.011537	0.00123	73.060276	73.060276	73.060276	0	0	1	-360	360;
	5402	2045	0.003661	0.009066	0.00093	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5402	7072	0	0.000893	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4716	4707	0.00537	0.032901	0.04552	377.858731	377.858731	377.858731	0	0	1	-360	360;
	8261	5899	0.016661	0.041331	0.00421	75.553875	75.553875	75.553875	0	0	1	-360	360;
	8261	6663	0.002628	0.007587	0.00087	102.923257	102.923257	102.923257	0	0	1	-360	360;
	8261	8300	0.009446	0.025388	0.00252	73.938027	73.938027	73.938027	0	0	1	-360	360;
	7166	7027	0.023248	0.060017	0.00589	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1198	667	0.053711	0.127983	0.01359	58.397617	58.397617	58.397617	0	0	1	-360	360;
	7512	505	0.009851	0.022537	0.00249	81.099503	81.099503	81.099503	0	0	1	-360	360;
	2515	7036	0.007256	0.021702	0.00184	78.329261	78.329261	78.329261	0	0	1	-360	360;
	2515	7470	0.001058	0.003149	0.00027	71.770227	71.770227	71.770227	0	0	1	-360	360;
	7036	6663	0.014744	0.048893	0.00379	100.255708	100.255708	100.255708	0	0	1	-360	360;
	3058	2518	0.040901	0.119017	0.01035	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6243	6909	0.026653	0.063504	0.00674	70.666131	70.666131	70.666131	0	0	1	-360	360;
	4128	1835	0.005198	0.011909	0.00132	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4128	888	0.006603	0.020793	0.00167	57.157677	57.157677	57.157677	0	0	1	-360	360;
	837	1229	0.023909	0.054744	0.00605	71.202461	71.202461	71.202461	0	0	1	-360	360;
	4541	1229	0.01286	0.038959	0.00325	71.14416	71.14416	71.14416	0	0	1	-360	360;
	5916	320	0	0.000826	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6884	320	0.005488	0.011256	0.00111	77.047024	77.047024	77.047024	0	0	1	-360	360;
	3365	320	0.018983	0.055579	0.00497	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5916	4048	0.013479	0.038876	0.00446	76.397141	76.397141	76.397141	0	0	1	-360	360;
	6704	4414	0.010041	0.02205	0.00254	76.375231	76.375231	76.375231	0	0	1	-360	360;
	4414	6881	0.01224	0.037099	0.0031	60.168553	60.168553	60.168553	0	0	1	-360	360;
	6704	7474	0	0.00081	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5104	6704	0.011	0.030959	0.00289	57.202069	57.202069	57.202069	0	0	1	-360	360;
	4635	7474	0.010777	0.031554	0.00283	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7474	5357	0.030521	0.069868	0.00772	64.01736	64.01736	64.01736	0	0	1	-360	360;
	7474	6277	0.016901	0.051198	0.00428	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5843	3556	0.008868	0.027967	0.00224	73.242799	73.242799	73.242799	0	0	1	-360	360;
	6881	3031	0.006322	0.02319	0.00265	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2529	5104	0.006488	0.019008	0.0017	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2975	4635	0.006347	0.019372	0.00167	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2494	2975	0.018339	0.038967	0.00481	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1130	4756	0.037793	0.093752	0.00957	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1169	683	0.018603	0.048017	0.0047	57.250082	57.250082	57.250082	0	0	1	-360	360;
	3819	5938	0.04086	0.128893	0.01034	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5938	1251	0.004587	0.010083	0.00117	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2518	4967	0.012529	0.028661	0.00316	57.157677	57.157677	57.157677	0	0	1	-360	360;
	39	3671	0.016339	0.035909	0.00414	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1169	6528	0.021702	0.058653	0.00569	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5672	7236	0.037339	0.109364	0.00979	66.487334	66.487334	66.487334	0	0	1	-360	360;
	3312	2695	0.036554	0.11143	0.00959	74.777293	74.777293	74.777293	0	0	1	-360	360;
	5489	5573	0.027149	0.062149	0.00687	65.802394	65.802394	65.802394	0	0	1	-360	360;
	4673	3501	0.034281	0.103876	0.00868	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4673	4591	0.023479	0.06843	0.00592	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6836	4673	0.053149	0.116868	0.01345	65.116121	65.116121	65.116121	0	0	1	-360	360;
	5340	6107	0.004802	0.013298	0.00159	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6107	459	0.012669	0.038231	0.00435	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4459	2234	0.040339	0.079496	0.00818	61.968257	61.968257	61.968257	0	0	1	-360	360;
	7706	1544	0.054521	0.171992	0.0138	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5110	1544	0.05514	0.122	0.01446	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2120	1544	0.056157	0.119339	0.01473	58.352844	58.352844	58.352844	0	0	1	-360	360;
	8189	4082	0.034248	0.075289	0.00867	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3048	802	0.0018	0.01865	0.44616	1452.26079	1452.26079	1452.26079	0	0	1	-360	360;
	6849	3998	0	0.000793	0	64.943124	64.943124	64.943124	0	0	1	-360	360;
	7111	6660	0.023231	0.064983	0.00588	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1545	1960	0.00163	0.01087	0.01558	400.356755	400.356755	400.356755	0	0	1	-360	360;
	5813	32	0.020727	0.053512	0.00524	57.157677	57.157677	57.157677	0	0	1	-360	360;
	459	6516	0.022364	0.064488	0.00741	57.157677	57.157677	57.157677	0	0	1	-360	360;
	837	3986	0.014579	0.04805	0.00416	61.694091	61.694091	61.694091	0	0	1	-360	360;
	837	3986	0.014281	0.048992	0.00407	66.69253	66.69253	66.69253	0	0	1	-360	360;
	5851	2743	0.00228	0.02412	0.48448	1102.47535	1102.47535	1102.47535	0	0	1	-360	360;
	6875	1371	0.004298	0.009132	0.00112	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3051	6875	0.003826	0.010347	0.001	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3051	1371	0	0.000777	0	64.243133	64.243133	64.243133	0	0	1	-360	360;
	7166	1371	0.042256	0.092917	0.01069	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3310	6357	0.005021	0.029341	0.04815	423.561247	423.561247	423.561247	0	0	1	-360	360;
	436	1192	0.066248	0.151661	0.01676	69.01275	69.01275	69.01275	0	0	1	-360	360;
	1192	4054	0.007587	0.016702	0.00193	70.255549	70.255549	70.255549	0	0	1	-360	360;
	2490	4536	0.008579	0.019636	0.00217	59.575827	59.575827	59.575827	0	0	1	-360	360;
	6836	5300	0.008942	0.024992	0.00226	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3591	3244	0.049124	0.154983	0.01243	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2001	3591	0.018289	0.04186	0.00462	57.157677	57.157677	57.157677	0	0	1	-360	360;
	802	6807	0.00064	0.00617	0.15715	1432.40484	1432.40484	1432.40484	0	0	1	-360	360;
	3472	1377	0	0.00076	0	65.390478	65.390478	65.390478	0	0	1	-360	360;
	3472	4565	0.024421	0.073331	0.00809	69.214898	69.214898	69.214898	0	0	1	-360	360;
	1377	5959	0	0.000744	0	63.084737	63.084737	63.084737	0	0	1	-360	360;
	6351	7978	0.025653	0.056388	0.00649	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3939	4850	0.003711	0.02356	0.03569	431.469964	431.469964	431.469964	0	0	1	-360	360;
	2756	3962	0.015777	0.089537	0.00941	69.524121	69.524121	69.524121	0	0	1	-360	360;
	1251	2756	0.001587	0.005612	0.00053	57.157677	57.157677	57.157677	0	0	1	-360	360;
	888	3800	0.008934	0.027091	0.00227	78.025944	78.025944	78.025944	0	0	1	-360	360;
	1859	1835	0.021298	0.052603	0.00432	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6949	7830	0.01036	0.05943	0.08779	408.612609	408.612609	408.612609	0	0	1	-360	360;
	7937	845	0.036438	0.114909	0.00922	67.990962	67.990962	67.990962	0	0	1	-360	360;
	511	7937	0.038876	0.113207	0.00984	70.485894	70.485894	70.485894	0	0	1	-360	360;
	3601	6102	0.045058	0.127901	0.01035	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6102	2001	0.014083	0.042661	0.00356	75.79489	75.79489	75.79489	0	0	1	-360	360;
	852	436	0.016537	0.052157	0.00419	61.656938	61.656938	61.656938	0	0	1	-360	360;
	6157	5428	0.045711	0.136504	0.01157	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4460	2905	0.00579	0.06394	1.23343	1088.19812	1088.19812	1088.19812	0	0	1	-360	360;
	1611	1203	0.012512	0.03381	0.00328	71.814048	71.814048	71.814048	0	0	1	-360	360;
	1203	111	0.053149	0.149512	0.01394	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1611	4537	0.012248	0.034471	0.00322	72.885564	72.885564	72.885564	0	0	1	-360	360;
	4537	2941	0.052041	0.152388	0.01365	60.708883	60.708883	60.708883	0	0	1	-360	360;
	845	3819	0.024471	0.056033	0.0062	66.76874	66.76874	66.76874	0	0	1	-360	360;
	1719	7452	0.008752	0.016322	0.00166	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4460	6363	0.00211	0.01832	0.44937	1345.96681	1345.96681	1345.96681	0	0	1	-360	360;
	6191	6625	0	0.00086	0	77.086844	77.086844	77.086844	0	0	1	-360	360;
	1694	68	0	0.000876	0	91.493055	91.493055	91.493055	0	0	1	-360	360;
	359	68	0.007248	0.021107	0.00183	71.679918	71.679918	71.679918	0	0	1	-360	360;
	1600	1694	0.036653	0.106661	0.00927	74.815207	74.815207	74.815207	0	0	1	-360	360;
	4312	3708	0.003107	0.007727	0.00079	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3708	5057	0.013612	0.029917	0.00344	58.379898	58.379898	58.379898	0	0	1	-360	360;
	6621	4756	0.064182	0.141124	0.01625	78.312685	78.312685	78.312685	0	0	1	-360	360;
	5633	4756	0.058331	0.150579	0.01476	69.944611	69.944611	69.944611	0	0	1	-360	360;
	6592	4539	0.00337	0.03668	0.08068	414.726194	414.726194	414.726194	0	0	1	-360	360;
	3705	7111	0.033959	0.080934	0.0086	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6384	6836	0	0.000826	0	65.546328	65.546328	65.546328	0	0	1	-360	360;
	5633	6836	0.07314	0.181438	0.01851	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6836	5200	0.01476	0.04295	0.00373	91.749693	91.749693	91.749693	0	0	1	-360	360;
	845	7321	0.018488	0.044091	0.00469	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6079	4460	0.00025	0.00227	0.04956	1306.10945	1306.10945	1306.10945	0	0	1	-360	360;
	6079	4460	0.00025	0.00222	0.05051	1416.97251	1416.97251	1416.97251	0	0	1	-360	360;
	4460	1101	0.00179	0.01494	0.38153	1033.68311	1033.68311	1033.68311	0	0	1	-360	360;
	1414	2633	0.00133	0.01403	0.31953	1240.36523	1240.36523	1240.36523	0	0	1	-360	360;
	954	1562	0.027438	0.076752	0.00695	67.84902	67.84902	67.84902	0	0	1	-360	360;
	6881	3533	0	0.00081	0	75.319338	75.319338	75.319338	0	0	1	-360	360;
	3533	6277	0.014347	0.03705	0.00363	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6949	4716	0.00286	0.015591	0.02729	421.498617	421.498617	421.498617	0	0	1	-360	360;
	5774	5057	0	0.000793	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5057	3986	0.008231	0.021273	0.00208	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1139	6669	0.013653	0.031231	0.00346	73.370071	73.370071	73.370071	0	0	1	-360	360;
	2876	2397	0	0.000777	0	65.288356	65.288356	65.288356	0	0	1	-360	360;
	7909	960	0.028364	0.063479	0.00642	61.152808	61.152808	61.152808	0	0	1	-360	360;
	3671	960	0.023636	0.051983	0.00598	71.228182	71.228182	71.228182	0	0	1	-360	360;
	422	960	0.022752	0.066248	0.00576	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5824	5110	0.004512	0.013099	0.00155	61.72648	61.72648	61.72648	0	0	1	-360	360;
	5824	2120	0.006331	0.015124	0.00133	72.877562	72.877562	72.877562	0	0	1	-360	360;
	4667	852	0.048058	0.143504	0.01216	62.156687	62.156687	62.156687	0	0	1	-360	360;
	6902	2395	0.02676	0.077901	0.00677	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2454	521	0	0.000744	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4249	1101	0.00031	0.00329	0.06108	1254.76224	1254.76224	1254.76224	0	0	1	-360	360;
	4249	1101	0.00035	0.00322	0.06923	1217.63567	1217.63567	1217.63567	0	0	1	-360	360;
	5357	635	0.01276	0.028033	0.00322	73.865247	73.865247	73.865247	0	0	1	-360	360;
	6682	7028	0.049983	0.088198	0.0954	64.238751	64.238751	64.238751	0	0	1	-360	360;
	4425	3365	0	0.000876	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3883	3365	0.020157	0.04643	0.00529	59.020064	59.020064	59.020064	0	0	1	-360	360;
	7268	7426	0.050554	0.09781	0.01052	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5572	7268	0.050157	0.093818	0.01032	94.134883	94.134883	94.134883	0	0	1	-360	360;
	6900	3565	0.03338	0.08238	0.00676	57.157677	57.157677	57.157677	0	0	1	-360	360;
	19	3565	0.041554	0.156107	0.01409	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3412	5960	0.04676	0.124372	0.00908	93.618749	93.618749	93.618749	0	0	1	-360	360;
	1958	5960	0.043868	0.116678	0.00852	83.014666	83.014666	83.014666	0	0	1	-360	360;
	6240	5199	0.015529	0.049017	0.00534	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6240	1299	0.019471	0.052802	0.00671	67.944473	67.944473	67.944473	0	0	1	-360	360;
	4332	5042	0.030339	0.062124	0.0062	64.076613	64.076613	64.076613	0	0	1	-360	360;
	1283	4332	0.043488	0.11219	0.00895	63.007955	63.007955	63.007955	0	0	1	-360	360;
	4332	5319	0.025934	0.052198	0.00517	75.178921	75.178921	75.178921	0	0	1	-360	360;
	7044	6834	0.024372	0.057347	0.00505	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7044	6053	0.023868	0.058446	0.00494	80.333209	80.333209	80.333209	0	0	1	-360	360;
	7006	1299	0.032636	0.103934	0.01125	68.210447	68.210447	68.210447	0	0	1	-360	360;
	8144	7006	0.02695	0.081719	0.00926	61.261407	61.261407	61.261407	0	0	1	-360	360;
	5205	4041	0.022521	0.13874	0.17178	457.786883	457.786883	457.786883	0	0	1	-360	360;
	5205	2822	0.0117	0.053091	0.11024	439.967024	439.967024	439.967024	0	0	1	-360	360;
	5205	4586	0.001401	0.006789	0.01068	477.31385	477.31385	477.31385	0	0	1	-360	360;
	5205	2181	0.00139	0.0113	0.01054	460.288484	460.288484	460.288484	0	0	1	-360	360;
	1794	5205	0.01818	0.095349	0.13858	467.182462	467.182462	467.182462	0	0	1	-360	360;
	7238	5205	0.005729	0.043599	0.06051	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5205	8319	0.00095	0.006409	0.01072	375.542321	375.542321	375.542321	0	0	1	-360	360;
	5205	1264	0.00093	0.006531	0.0105	647.787002	647.787002	647.787002	0	0	1	-360	360;
	5205	4457	0.025659	0.11894	0.19536	455.579834	455.579834	455.579834	0	0	1	-360	360;
	5205	629	0.01036	0.04869	0.07802	533.49756	533.49756	533.49756	0	0	1	-360	360;
	4685	2477	0.022231	0.051521	0.0044	57.157677	57.157677	57.157677	0	0	1	-360	360;
	992	2477	0.038099	0.092579	0.00752	85.467493	85.467493	85.467493	0	0	1	-360	360;
	3227	1174	0.018041	0.066793	0.0062	61.243498	61.243498	61.243498	0	0	1	-360	360;
	4350	1174	0.019942	0.075554	0.00688	71.076714	71.076714	71.076714	0	0	1	-360	360;
	165	4013	0.028347	0.064727	0.00569	99.975826	99.975826	99.975826	0	0	1	-360	360;
	165	4244	0.028124	0.066826	0.00565	68.025447	68.025447	68.025447	0	0	1	-360	360;
	5763	1182	0.017107	0.031438	0.00335	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6990	1182	0.028322	0.109331	0.00932	72.873942	72.873942	72.873942	0	0	1	-360	360;
	7641	6886	0	0.000413	0	60.670587	60.670587	60.670587	0	0	1	-360	360;
	7641	1393	0.002248	0.008413	0.00077	116.393416	116.393416	116.393416	0	0	1	-360	360;
	6886	1920	0.003421	0.010397	0.00119	106.897049	106.897049	106.897049	0	0	1	-360	360;
	2706	6886	0.039554	0.071554	0.00851	66.955646	66.955646	66.955646	0	0	1	-360	360;
	3089	6886	0.012612	0.036818	0.00439	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5380	137	0.004471	0.014041	0.00156	171.47303	171.47303	171.47303	0	0	1	-360	360;
	137	3575	0.013628	0.050934	0.0047	97.941013	97.941013	97.941013	0	0	1	-360	360;
	137	792	0.002438	0.007959	0.00083	57.157677	57.157677	57.157677	0	0	1	-360	360;
	137	3084	0.014579	0.030132	0.0031	57.157677	57.157677	57.157677	0	0	1	-360	360;
	137	7630	0.04186	0.08719	0.00828	57.157677	57.157677	57.157677	0	0	1	-360	360;
	137	7304	0.072653	0.128339	0.01538	81.413299	81.413299	81.413299	0	0	1	-360	360;
	7013	851	0.013612	0.049421	0.00467	80.929554	80.929554	80.929554	0	0	1	-360	360;
	7013	7231	0.037256	0.107727	0.00928	58.816773	58.816773	58.816773	0	0	1	-360	360;
	1794	2668	0.003349	0.0237	0.03732	269.926366	269.926366	269.926366	0	0	1	-360	360;
	1794	2668	0.003281	0.024159	0.03654	253.174213	253.174213	253.174213	0	0	1	-360	360;
	1794	2668	0.003599	0.023469	0.03965	257.275086	257.275086	257.275086	0	0	1	-360	360;
	1794	2668	0.003669	0.02299	0.04042	586.939988	586.939988	586.939988	0	0	1	-360	360;
	307	830	0.009124	0.02781	0.00303	94.539559	94.539559	94.539559	0	0	1	-360	360;
	307	830	0.008661	0.026479	0.00294	96.752133	96.752133	96.752133	0	0	1	-360	360;
	307	830	0.008488	0.027033	0.00289	122.427361	122.427361	122.427361	0	0	1	-360	360;
	307	1643	0.002248	0.006554	0.00077	72.805162	72.805162	72.805162	0	0	1	-360	360;
	307	4600	0.001612	0.004636	0.00055	72.786491	72.786491	72.786491	0	0	1	-360	360;
	307	1393	0.012223	0.027331	0.00324	139.158937	139.158937	139.158937	0	0	1	-360	360;
	307	4901	0.002099	0.01181	0.00146	91.714636	91.714636	91.714636	0	0	1	-360	360;
	307	4901	0.002058	0.012058	0.00144	97.017916	97.017916	97.017916	0	0	1	-360	360;
	830	1253	0.019008	0.055876	0.00634	73.785797	73.785797	73.785797	0	0	1	-360	360;
	830	2595	0.020727	0.05795	0.00708	71.217894	71.217894	71.217894	0	0	1	-360	360;
	2706	1311	0.004471	0.010471	0.00114	104.183965	104.183965	104.183965	0	0	1	-360	360;
	7814	1311	0.010479	0.025967	0.00213	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1643	4600	0.001636	0.004537	0.00056	66.024547	66.024547	66.024547	0	0	1	-360	360;
	1920	5380	0.001603	0.005256	0.00055	85.635917	85.635917	85.635917	0	0	1	-360	360;
	7677	275	0.001039	0.00812	0.01067	601.598265	601.598265	601.598265	0	0	1	-360	360;
	7677	275	0.001169	0.007961	0.0121	639.832178	639.832178	639.832178	0	0	1	-360	360;
	851	3575	0.001802	0.006628	0.00062	80.825718	80.825718	80.825718	0	0	1	-360	360;
	4840	8188	0.00022	0.00217	0.05257	1305.47496	1305.47496	1305.47496	0	0	1	-360	360;
	1538	8188	0.00067	0.0072	0.16078	1154.38398	1154.38398	1154.38398	0	0	1	-360	360;
	2326	2706	0.019826	0.06962	0.0067	74.486932	74.486932	74.486932	0	0	1	-360	360;
	1794	4041	0.005919	0.02874	0.04424	551.146708	551.146708	551.146708	0	0	1	-360	360;
	2131	4041	0.001271	0.009651	0.01349	528.38995	528.38995	528.38995	0	0	1	-360	360;
	7677	4041	0.00126	0.00955	0.0134	572.874246	572.874246	572.874246	0	0	1	-360	360;
	6676	4840	0.0009	0.01149	0.21324	1269.08159	1269.08159	1269.08159	0	0	1	-360	360;
	4840	7576	0.00115	0.01172	0.2706	1211.06177	1211.06177	1211.06177	0	0	1	-360	360;
	2131	1602	0.00431	0.033729	0.04495	465.937186	465.937186	465.937186	0	0	1	-360	360;
	7677	5761	0.001841	0.0115	0.01949	368.87926	368.87926	368.87926	0	0	1	-360	360;
	2822	7238	0.008401	0.053719	0.06392	583.388972	583.388972	583.388972	0	0	1	-360	360;
	3552	2822	0.0013	0.006021	0.01199	432.715621	432.715621	432.715621	0	0	1	-360	360;
	8205	7678	0.016083	0.036347	0.00333	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7678	4006	0.01576	0.037058	0.00326	64.012025	64.012025	64.012025	0	0	1	-360	360;
	5684	7678	0.039661	0.101099	0.00797	70.586301	70.586301	70.586301	0	0	1	-360	360;
	3409	340	0.015777	0.058099	0.00549	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4369	340	0.044901	0.08681	0.00894	64.209981	64.209981	64.209981	0	0	1	-360	360;
	4388	2586	0.00173	0.01919	0.41161	1280.57406	1280.57406	1280.57406	0	0	1	-360	360;
	2586	7764	0.00213	0.02194	0.5097	1503.10844	1503.10844	1503.10844	0	0	1	-360	360;
	202	1847	0.00395	0.008331	0.00078	109.364355	109.364355	109.364355	0	0	1	-360	360;
	7532	1847	0.157273	0.234413	0.02702	69.634626	69.634626	69.634626	0	0	1	-360	360;
	202	3518	0.003421	0.009099	0.00068	61.606259	61.606259	61.606259	0	0	1	-360	360;
	3518	5579	0.06795	0.151058	0.01395	69.017514	69.017514	69.017514	0	0	1	-360	360;
	202	2877	0.040099	0.154554	0.01253	64.459951	64.459951	64.459951	0	0	1	-360	360;
	202	4251	0.070769	0.175471	0.01365	78.969046	78.969046	78.969046	0	0	1	-360	360;
	202	2914	0.008504	0.028322	0.00301	69.368271	69.368271	69.368271	0	0	1	-360	360;
	202	8278	0.02738	0.069876	0.00626	63.693466	63.693466	63.693466	0	0	1	-360	360;
	898	7969	0.010347	0.023041	0.00189	68.913296	68.913296	68.913296	0	0	1	-360	360;
	8144	7969	0.044281	0.083967	0.00876	72.971301	72.971301	72.971301	0	0	1	-360	360;
	992	898	0.055744	0.101992	0.01097	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1549	4013	0.015446	0.026554	0.00312	133.616739	133.616739	133.616739	0	0	1	-360	360;
	4013	1361	0.035388	0.067603	0.00635	70.958779	70.958779	70.958779	0	0	1	-360	360;
	2752	6619	0.007273	0.013752	0.00147	92.451018	92.451018	92.451018	0	0	1	-360	360;
	6619	6336	0.00362	0.007033	0.00073	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3798	6336	0.001182	0.003727	0.00041	76.432579	76.432579	76.432579	0	0	1	-360	360;
	3798	4063	0.013314	0.052752	0.0046	62.718547	62.718547	62.718547	0	0	1	-360	360;
	2752	4137	0.028372	0.075868	0.00971	58.370181	58.370181	58.370181	0	0	1	-360	360;
	4856	2752	0.033917	0.080413	0.0069	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7922	3971	0.004669	0.03695	0.04845	215.929128	215.929128	215.929128	0	0	1	-360	360;
	3971	7238	0.004229	0.025271	0.04386	298.33678	298.33678	298.33678	0	0	1	-360	360;
	6900	2391	0.041909	0.072157	0.00848	69.176412	69.176412	69.176412	0	0	1	-360	360;
	2841	2391	0.031174	0.078554	0.0069	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2312	3227	0.030777	0.114017	0.01058	58.771047	58.771047	58.771047	0	0	1	-360	360;
	1517	6126	0.021479	0.05195	0.00423	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6126	1299	0.04881	0.09786	0.01012	73.720828	73.720828	73.720828	0	0	1	-360	360;
	6566	1072	0.043992	0.107769	0.00913	62.246044	62.246044	62.246044	0	0	1	-360	360;
	6566	1299	0.017347	0.059826	0.00598	62.191553	62.191553	62.191553	0	0	1	-360	360;
	2705	6336	0.006281	0.014802	0.00127	61.962732	61.962732	61.962732	0	0	1	-360	360;
	4137	2705	0.045818	0.086281	0.00966	62.5023	62.5023	62.5023	0	0	1	-360	360;
	6053	2705	0.022521	0.077653	0.0078	75.105187	75.105187	75.105187	0	0	1	-360	360;
	1935	5035	0.014132	0.053744	0.00486	83.361614	83.361614	83.361614	0	0	1	-360	360;
	726	1935	0.025901	0.082504	0.00896	69.004748	69.004748	69.004748	0	0	1	-360	360;
	2243	1935	0.05905	0.233421	0.0198	61.666274	61.666274	61.666274	0	0	1	-360	360;
	6	7619	0.035174	0.062992	0.00712	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6	6536	0.015901	0.059529	0.00584	60.652297	60.652297	60.652297	0	0	1	-360	360;
	7619	885	0.005174	0.015322	0.00191	90.444783	90.444783	90.444783	0	0	1	-360	360;
	7619	2133	0.021471	0.063207	0.00788	57.157677	57.157677	57.157677	0	0	1	-360	360;
	885	1860	0.030033	0.114521	0.01029	83.039244	83.039244	83.039244	0	0	1	-360	360;
	885	5107	0.050537	0.090521	0.01023	62.559077	62.559077	62.559077	0	0	1	-360	360;
	885	5107	0.026446	0.070959	0.00873	78.381275	78.381275	78.381275	0	0	1	-360	360;
	3765	3494	0.028157	0.075636	0.00961	59.785025	59.785025	59.785025	0	0	1	-360	360;
	3765	7575	0.021413	0.212132	0.01103	58.255676	58.255676	58.255676	0	0	1	-360	360;
	3765	7575	0.046446	0.096702	0.00969	61.165573	61.165573	61.165573	0	0	1	-360	360;
	7169	4215	0.042562	0.102702	0.00883	71.750031	71.750031	71.750031	0	0	1	-360	360;
	6828	4215	0.034314	0.071603	0.00712	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4298	7814	0.034033	0.077537	0.00689	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1283	4298	0.01919	0.035669	0.00383	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6900	7426	0.046826	0.094587	0.00948	118.139011	118.139011	118.139011	0	0	1	-360	360;
	8164	2677	0.059909	0.111868	0.01185	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6780	8164	0.020909	0.037934	0.0042	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5329	6208	0.009727	0.025347	0.002	90.438305	90.438305	90.438305	0	0	1	-360	360;
	4379	5329	0.01981	0.040331	0.00399	111.410791	111.410791	111.410791	0	0	1	-360	360;
	2436	5045	1.7e-05	0.000413	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	535	2436	0.017091	0.04114	0.00347	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2326	792	0.03014	0.111744	0.00973	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3308	1549	0.012628	0.026339	0.00262	134.181647	134.181647	134.181647	0	0	1	-360	360;
	1131	779	0.032934	0.078298	0.00705	61.830507	61.830507	61.830507	0	0	1	-360	360;
	1790	1131	0.025198	0.050942	0.00511	77.935445	77.935445	77.935445	0	0	1	-360	360;
	2538	6640	0.025529	0.049421	0.00482	58.041144	58.041144	58.041144	0	0	1	-360	360;
	5684	6640	0.026124	0.06962	0.00489	79.937488	79.937488	79.937488	0	0	1	-360	360;
	354	5863	0.00036	0.00188	0.00296	225.306797	225.306797	225.306797	0	0	1	-360	360;
	354	5863	0.00043	0.00174	0.00348	231.842968	231.842968	231.842968	0	0	1	-360	360;
	7597	4789	0.026091	0.052298	0.0052	75.201784	75.201784	75.201784	0	0	1	-360	360;
	7597	4789	0.015413	0.048636	0.00529	92.031099	92.031099	92.031099	0	0	1	-360	360;
	4789	6010	0.024463	0.08705	0.00841	73.741214	73.741214	73.741214	0	0	1	-360	360;
	5863	3355	0.00638	0.04836	0.06504	239.351963	239.351963	239.351963	0	0	1	-360	360;
	5863	4548	0.002081	0.01214	0.0213	537.828588	537.828588	537.828588	0	0	1	-360	360;
	5863	4548	0.002081	0.01257	0.02118	521.324499	521.324499	521.324499	0	0	1	-360	360;
	8118	2592	0.005314	0.015017	0.00194	57.157677	57.157677	57.157677	0	0	1	-360	360;
	8118	4244	0.004331	0.011752	0.00158	58.023234	58.023234	58.023234	0	0	1	-360	360;
	2592	4244	0.00386	0.013322	0.00141	71.323635	71.323635	71.323635	0	0	1	-360	360;
	4244	3494	0.02181	0.080802	0.00748	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4244	1564	0.036992	0.107727	0.01268	81.002335	81.002335	81.002335	0	0	1	-360	360;
	4244	4720	0.029273	0.105248	0.01011	72.86613	72.86613	72.86613	0	0	1	-360	360;
	6053	4063	0.016909	0.05057	0.00605	80.965564	80.965564	80.965564	0	0	1	-360	360;
	7680	4388	0.00302	0.03553	0.70958	1253.3544	1253.3544	1253.3544	0	0	1	-360	360;
	4251	6903	0.005983	0.023132	0.00207	74.634779	74.634779	74.634779	0	0	1	-360	360;
	6484	6903	0.003777	0.011421	0.00128	64.303339	64.303339	64.303339	0	0	1	-360	360;
	2914	6903	0.050041	0.180397	0.01652	91.021885	91.021885	91.021885	0	0	1	-360	360;
	6903	8278	0.056562	0.179397	0.01958	94.948046	94.948046	94.948046	0	0	1	-360	360;
	5850	1998	0.012818	0.046769	0.00438	60.135782	60.135782	60.135782	0	0	1	-360	360;
	7815	5761	0.00286	0.017581	0.03032	516.972514	516.972514	516.972514	0	0	1	-360	360;
	530	8205	0.012554	0.029083	0.00272	79.639125	79.639125	79.639125	0	0	1	-360	360;
	1839	2710	0.000909	0.002876	0.00032	140.731155	140.731155	140.731155	0	0	1	-360	360;
	6008	530	0.018719	0.033769	0.00401	62.755318	62.755318	62.755318	0	0	1	-360	360;
	2855	6008	0.032149	0.059041	0.00681	104.60636	104.60636	104.60636	0	0	1	-360	360;
	7342	4535	0.005198	0.013488	0.00103	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2677	7342	0.018636	0.044149	0.00378	69.637293	69.637293	69.637293	0	0	1	-360	360;
	2677	4535	0.014554	0.028438	0.00287	78.370796	78.370796	78.370796	0	0	1	-360	360;
	6780	2677	0.032934	0.155826	0.01172	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2677	5995	0.015231	0.035843	0.00317	65.397527	65.397527	65.397527	0	0	1	-360	360;
	2677	8342	0.037248	0.089901	0.00773	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2677	387	0.04643	0.09338	0.00924	58.830491	58.830491	58.830491	0	0	1	-360	360;
	223	7687	0.075612	0.155504	0.0155	59.267176	59.267176	59.267176	0	0	1	-360	360;
	7687	3955	0.04057	0.08876	0.00821	70.882568	70.882568	70.882568	0	0	1	-360	360;
	7687	296	0.114628	0.20105	0.02273	59.873809	59.873809	59.873809	0	0	1	-360	360;
	3887	8367	0.018521	0.090322	0.00906	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2595	8367	0.013702	0.052983	0.00445	72.732953	72.732953	72.732953	0	0	1	-360	360;
	1581	7120	0.006017	0.019198	0.0022	171.47303	171.47303	171.47303	0	0	1	-360	360;
	1581	6193	0.052686	0.090653	0.01071	99.210104	99.210104	99.210104	0	0	1	-360	360;
	1581	1998	0.065826	0.122818	0.01331	61.904241	61.904241	61.904241	0	0	1	-360	360;
	6053	6834	0.04443	0.105719	0.00916	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7021	6668	0.00838	0.018942	0.00173	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7530	6668	0.058017	0.142983	0.01253	61.32409	61.32409	61.32409	0	0	1	-360	360;
	1398	7102	0.00085	0.01148	0.20153	870.184127	870.184127	870.184127	0	0	1	-360	360;
	7764	7102	0.00172	0.0234	0.4111	1302.37033	1302.37033	1302.37033	0	0	1	-360	360;
	6900	7021	0.04805	0.152256	0.01649	59.51143	59.51143	59.51143	0	0	1	-360	360;
	7021	5290	0.028372	0.069471	0.00588	105.408282	105.408282	105.408282	0	0	1	-360	360;
	6208	7021	0.038364	0.12214	0.01323	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3412	5035	0.013463	0.05405	0.00463	73.970988	73.970988	73.970988	0	0	1	-360	360;
	3084	7630	0.025083	0.060273	0.00528	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3202	1248	0.012438	0.047124	0.00428	86.961976	86.961976	86.961976	0	0	1	-360	360;
	8144	3202	0.022388	0.084818	0.00772	93.658188	93.658188	93.658188	0	0	1	-360	360;
	2312	2737	0.005471	0.020281	0.00188	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4350	2737	0.107174	0.201248	0.02141	67.375373	67.375373	67.375373	0	0	1	-360	360;
	4685	2312	0.035058	0.124769	0.01205	65.50727	65.50727	65.50727	0	0	1	-360	360;
	1860	7120	0.015934	0.05876	0.00584	130.50698	130.50698	130.50698	0	0	1	-360	360;
	535	5045	0.020099	0.03862	0.00408	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2020	3676	0.04176	0.103083	0.00845	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3377	3676	0.066719	0.165537	0.0132	80.242329	80.242329	80.242329	0	0	1	-360	360;
	4548	4546	0.00013	0.00095	0.00127	215.929128	215.929128	215.929128	0	0	1	-360	360;
	2538	2999	0.033777	0.06776	0.00705	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7488	2999	0.057388	0.126289	0.01177	62.739314	62.739314	62.739314	0	0	1	-360	360;
	1398	8222	0.0019	0.0204	0.45561	1229.9219	1229.9219	1229.9219	0	0	1	-360	360;
	7530	7634	0.007529	0.025372	0.00256	61.620358	61.620358	61.620358	0	0	1	-360	360;
	7530	1514	0.023917	0.064636	0.00824	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7530	5319	0.039612	0.099818	0.00788	69.536315	69.536315	69.536315	0	0	1	-360	360;
	7530	6720	0.012388	0.038802	0.00424	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7634	6720	0.004281	0.015678	0.00146	61.059641	61.059641	61.059641	0	0	1	-360	360;
	6891	4856	0.013463	0.033289	0.00273	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7771	4856	0.015893	0.039901	0.00327	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7771	6891	0.005934	0.021686	0.00216	64.021933	64.021933	64.021933	0	0	1	-360	360;
	3432	6891	0.009107	0.02843	0.00311	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7771	3432	0.005802	0.020083	0.00211	82.090427	82.090427	82.090427	0	0	1	-360	360;
	7771	1651	0.10495	0.203661	0.02123	60.655345	60.655345	60.655345	0	0	1	-360	360;
	7771	927	0.026372	0.04995	0.00542	66.192591	66.192591	66.192591	0	0	1	-360	360;
	600	7922	0.00738	0.044	0.05611	262.673819	262.673819	262.673819	0	0	1	-360	360;
	7922	7238	0.010591	0.05825	0.08047	528.163225	528.163225	528.163225	0	0	1	-360	360;
	4424	600	0.01318	0.05764	0.09902	527.850001	527.850001	527.850001	0	0	1	-360	360;
	600	3999	0.0143	0.06869	0.11738	481.884178	481.884178	481.884178	0	0	1	-360	360;
	6900	2004	0.022099	0.080636	0.00757	58.728751	58.728751	58.728751	0	0	1	-360	360;
	992	1517	0.039843	0.08557	0.00784	58.680167	58.680167	58.680167	0	0	1	-360	360;
	1517	7647	0.007909	0.022124	0.00272	61.377056	61.377056	61.377056	0	0	1	-360	360;
	7562	19	0.051017	0.111273	0.01032	63.280788	63.280788	63.280788	0	0	1	-360	360;
	6208	1514	0.010331	0.038289	0.00355	57.157677	57.157677	57.157677	0	0	1	-360	360;
	779	2326	0.02343	0.058603	0.00485	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7764	1807	0.0009	0.00851	0.21417	1271.87293	1271.87293	1271.87293	0	0	1	-360	360;
	1807	5534	0.00157	0.01488	0.37481	1112.20851	1112.20851	1112.20851	0	0	1	-360	360;
	4424	2823	0.00682	0.03224	0.05114	304.838275	304.838275	304.838275	0	0	1	-360	360;
	6780	7169	0.01719	0.033322	0.00346	92.839499	92.839499	92.839499	0	0	1	-360	360;
	7169	6828	0.080446	0.14995	0.01599	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4369	223	0.055678	0.123917	0.01141	66.910491	66.910491	66.910491	0	0	1	-360	360;
	1116	1564	0.014463	0.040066	0.00478	136.900447	136.900447	136.900447	0	0	1	-360	360;
	1116	7203	0.013107	0.048314	0.00481	60.539887	60.539887	60.539887	0	0	1	-360	360;
	3955	1790	0.123174	0.23414	0.02442	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2877	5579	0.00557	0.011256	0.00114	68.815747	68.815747	68.815747	0	0	1	-360	360;
	2877	2386	0.007967	0.018719	0.00213	67.231146	67.231146	67.231146	0	0	1	-360	360;
	7532	2386	0.076711	0.159554	0.01609	83.594817	83.594817	83.594817	0	0	1	-360	360;
	497	2133	0.012669	0.049364	0.00465	57.157677	57.157677	57.157677	0	0	1	-360	360;
	497	4725	0.015719	0.058826	0.00576	69.406757	69.406757	69.406757	0	0	1	-360	360;
	6085	1281	0.007397	0.027686	0.00272	68.21521	68.21521	68.21521	0	0	1	-360	360;
	6085	1998	0.006967	0.02505	0.00255	72.946913	72.946913	72.946913	0	0	1	-360	360;
	5042	7859	0.037653	0.076876	0.00769	77.052359	77.052359	77.052359	0	0	1	-360	360;
	3409	6990	0.011	0.029826	0.00383	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4350	4685	0.042446	0.107058	0.00835	85.928946	85.928946	85.928946	0	0	1	-360	360;
	4251	6484	0.011008	0.025711	0.00217	59.853804	59.853804	59.853804	0	0	1	-360	360;
	5572	6828	0.014818	0.04719	0.00511	89.611615	89.611615	89.611615	0	0	1	-360	360;
	5572	6828	0.011802	0.040711	0.00407	71.222466	71.222466	71.222466	0	0	1	-360	360;
	6828	5572	0.010802	0.039769	0.00371	87.830963	87.830963	87.830963	0	0	1	-360	360;
	5572	7270	0.003331	0.011496	0.00117	72.418967	72.418967	72.418967	0	0	1	-360	360;
	5572	7562	0.041256	0.083702	0.00796	101.621967	101.621967	101.621967	0	0	1	-360	360;
	7270	6828	0.011223	0.030413	0.00393	80.823241	80.823241	80.823241	0	0	1	-360	360;
	6828	1839	0.105562	0.208901	0.02161	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6828	2020	0.021612	0.065496	0.0074	73.84848	73.84848	73.84848	0	0	1	-360	360;
	7764	2178	0.0056	0.03178	0.51847	1065.33758	1065.33758	1065.33758	0	0	1	-360	360;
	2595	1253	0.005364	0.01443	0.00182	72.606444	72.606444	72.606444	0	0	1	-360	360;
	5485	7581	0.007132	0.016347	0.00142	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7581	1998	0.035281	0.111719	0.01302	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1651	5485	0.027669	0.067686	0.00571	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5485	2855	0.086157	0.181149	0.01716	87.389515	87.389515	87.389515	0	0	1	-360	360;
	4144	7597	0.019983	0.054711	0.00647	125.159689	125.159689	125.159689	0	0	1	-360	360;
	4144	3887	0.007041	0.032174	0.00307	82.688296	82.688296	82.688296	0	0	1	-360	360;
	3402	5763	0.050587	0.128397	0.0103	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7597	5871	0.014983	0.031364	0.00303	62.308345	62.308345	62.308345	0	0	1	-360	360;
	5871	5696	0.010653	0.019	0.00227	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5159	4367	0.00189	0.02064	0.4428	1354.85091	1354.85091	1354.85091	0	0	1	-360	360;
	5159	4367	0.00227	0.01922	0.51608	1264.23345	1264.23345	1264.23345	0	0	1	-360	360;
	4367	810	0.00167	0.02044	0.38247	1304.47914	1304.47914	1304.47914	0	0	1	-360	360;
	1794	3064	0.016341	0.07818	0.12503	508.429727	508.429727	508.429727	0	0	1	-360	360;
	6822	3064	0.005409	0.024729	0.04162	545.693484	545.693484	545.693484	0	0	1	-360	360;
	3064	1340	0.00525	0.04099	0.05557	560.339948	560.339948	560.339948	0	0	1	-360	360;
	7597	3098	0.00714	0.025347	0.00246	61.319708	61.319708	61.319708	0	0	1	-360	360;
	7597	3419	0.00819	0.025802	0.00282	77.652514	77.652514	77.652514	0	0	1	-360	360;
	5696	7597	0.006322	0.014314	0.00134	64.86196	64.86196	64.86196	0	0	1	-360	360;
	4746	992	0.037678	0.087512	0.00743	93.141101	93.141101	93.141101	0	0	1	-360	360;
	4746	6529	0.010942	0.034529	0.00376	105.92232	105.92232	105.92232	0	0	1	-360	360;
	3098	6529	0.007041	0.02714	0.00243	81.035106	81.035106	81.035106	0	0	1	-360	360;
	6529	3419	0.002826	0.010496	0.00098	89.651816	89.651816	89.651816	0	0	1	-360	360;
	6676	5159	0.0001	0.00124	0.02313	1351.31188	1351.31188	1351.31188	0	0	1	-360	360;
	6676	4051	3e-05	0.00027	0.0031	1391.42528	1391.42528	1391.42528	0	0	1	-360	360;
	6676	5670	2e-05	0.00021	0.00447	767.876077	767.876077	767.876077	0	0	1	-360	360;
	6676	5159	0.00011	0.00133	0.02492	1261.80345	1261.80345	1261.80345	0	0	1	-360	360;
	6676	8222	0.00062	0.00694	0.14895	1261.87256	1261.87256	1261.87256	0	0	1	-360	360;
	1794	420	0.005281	0.04189	0.05579	609.779814	609.779814	609.779814	0	0	1	-360	360;
	1794	3272	0.00014	0.000651	0.00111	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1794	1037	0.000229	0.00107	0.00176	363.298765	363.298765	363.298765	0	0	1	-360	360;
	1794	2988	0.000229	0.001039	0.00179	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1794	809	0.00013	0.00086	0.00106	647.787002	647.787002	647.787002	0	0	1	-360	360;
	1794	6921	0.00013	0.00087	0.00104	403.82775	403.82775	403.82775	0	0	1	-360	360;
	1794	5203	0.001539	0.012651	0.01593	215.929128	215.929128	215.929128	0	0	1	-360	360;
	1794	629	0.00806	0.047091	0.0607	327.380119	327.380119	327.380119	0	0	1	-360	360;
	7647	1072	0.04943	0.099331	0.00959	57.157677	57.157677	57.157677	0	0	1	-360	360;
	160	296	0.050488	0.104149	0.01003	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6010	6771	0.003702	0.008934	0.00073	57.157677	57.157677	57.157677	0	0	1	-360	360;
	992	6771	0.074446	0.138033	0.01444	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3412	7630	0.008479	0.021653	0.00165	57.830423	57.830423	57.830423	0	0	1	-360	360;
	3412	7630	0.005314	0.019479	0.00177	65.49984	65.49984	65.49984	0	0	1	-360	360;
	7680	1538	0.00104	0.01121	0.2445	1463.24383	1463.24383	1463.24383	0	0	1	-360	360;
	1340	7815	0.01118	0.07193	0.08404	446.799272	446.799272	446.799272	0	0	1	-360	360;
	5290	2841	0.002777	0.008174	0.00078	66.752545	66.752545	66.752545	0	0	1	-360	360;
	6536	6193	0.006058	0.02362	0.00223	97.912243	97.912243	97.912243	0	0	1	-360	360;
	2178	6631	0.00192	0.01477	0.1866	1116.76245	1116.76245	1116.76245	0	0	1	-360	360;
	7120	6193	0.025331	0.098711	0.00929	145.315962	145.315962	145.315962	0	0	1	-360	360;
	4369	3568	0.05886	0.113157	0.01016	57.157677	57.157677	57.157677	0	0	1	-360	360;
	456	3568	0.046851	0.110231	0.00972	57.157677	57.157677	57.157677	0	0	1	-360	360;
	726	4350	0.02895	0.10819	0.00997	57.157677	57.157677	57.157677	0	0	1	-360	360;
	969	7532	0.079438	0.11481	0.01301	57.157677	57.157677	57.157677	0	0	1	-360	360;
	969	7532	0.077967	0.117314	0.01277	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7532	6208	0.032669	0.091388	0.0112	67.748613	67.748613	67.748613	0	0	1	-360	360;
	791	1839	0.008421	0.02657	0.00289	71.456813	71.456813	71.456813	0	0	1	-360	360;
	2020	791	0.082727	0.14876	0.01614	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3999	4548	0.00637	0.029669	0.02107	524.577533	524.577533	524.577533	0	0	1	-360	360;
	3761	969	0.020223	0.033198	0.0034	65.979011	65.979011	65.979011	0	0	1	-360	360;
	3761	969	0.017843	0.033876	0.003	68.196158	68.196158	68.196158	0	0	1	-360	360;
	6053	2538	0.022058	0.05405	0.0046	60.293537	60.293537	60.293537	0	0	1	-360	360;
	6053	8046	0.019826	0.073909	0.0066	62.78504	62.78504	62.78504	0	0	1	-360	360;
	8342	1361	0.063107	0.131711	0.0131	71.243043	71.243043	71.243043	0	0	1	-360	360;
	4662	1564	0.012099	0.029041	0.00248	61.15014	61.15014	61.15014	0	0	1	-360	360;
	4662	1564	0.006273	0.019033	0.00201	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4662	6973	0.007603	0.021273	0.0026	91.512108	91.512108	91.512108	0	0	1	-360	360;
	4662	6973	0.011041	0.027339	0.00227	72.397056	72.397056	72.397056	0	0	1	-360	360;
	4826	3389	0.011636	0.032562	0.00399	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4826	1564	0.007091	0.019058	0.00243	81.083118	81.083118	81.083118	0	0	1	-360	360;
	3389	6973	0.006653	0.020967	0.00227	81.064065	81.064065	81.064065	0	0	1	-360	360;
	2134	4548	0.00212	0.01695	0.02157	516.980516	516.980516	516.980516	0	0	1	-360	360;
	2134	4548	0.002349	0.0155	0.02411	233.279912	233.279912	233.279912	0	0	1	-360	360;
	1564	7575	0.053901	0.092744	0.01091	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1235	5107	0.003273	0.010033	0.0012	171.47303	171.47303	171.47303	0	0	1	-360	360;
	1235	2269	0.005066	0.013769	0.00186	57.948548	57.948548	57.948548	0	0	1	-360	360;
	5107	2269	0.007826	0.028587	0.00697	89.727836	89.727836	89.727836	0	0	1	-360	360;
	2269	4725	0.032769	0.058711	0.00661	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7304	7231	0.053702	0.134017	0.01304	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5159	2342	0.00147	0.01305	0.33639	1249.52709	1249.52709	1249.52709	0	0	1	-360	360;
	4188	5159	0.00019	0.00195	0.04578	1446.12985	1446.12985	1446.12985	0	0	1	-360	360;
	5432	5159	0.00019	0.00191	0.04666	1470.59701	1470.59701	1470.59701	0	0	1	-360	360;
	4380	2742	0.006413	0.019917	0.00217	66.759785	66.759785	66.759785	0	0	1	-360	360;
	3761	2742	0.011562	0.044182	0.00396	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5476	4380	0.002512	0.008421	0.00087	74.748142	74.748142	74.748142	0	0	1	-360	360;
	446	3402	0.041702	0.098711	0.00841	57.157677	57.157677	57.157677	0	0	1	-360	360;
	456	446	0.034347	0.061636	0.0069	85.137312	85.137312	85.137312	0	0	1	-360	360;
	4369	4763	0.004289	0.01243	0.00148	73.378644	73.378644	73.378644	0	0	1	-360	360;
	5442	4763	0.000686	0.001471	0.00016	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4369	5241	0.006612	0.01886	0.00245	74.371854	74.371854	74.371854	0	0	1	-360	360;
	7060	5241	0.008388	0.023339	0.00286	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4369	5442	0.004372	0.012174	0.00151	61.22254	61.22254	61.22254	0	0	1	-360	360;
	4369	7060	0.013413	0.047793	0.0046	71.658008	71.658008	71.658008	0	0	1	-360	360;
	992	1248	0.08438	0.183769	0.01638	57.157677	57.157677	57.157677	0	0	1	-360	360;
	2538	3657	0.006372	0.01681	0.00217	61.632551	61.632551	61.632551	0	0	1	-360	360;
	7203	4725	0.028579	0.084149	0.01048	103.447774	103.447774	103.447774	0	0	1	-360	360;
	7203	3861	0.007008	0.022388	0.00258	95.491235	95.491235	95.491235	0	0	1	-360	360;
	7304	7767	0.003	0.005058	0.00063	100.378788	100.378788	100.378788	0	0	1	-360	360;
	249	7767	0.002669	0.005727	0.00056	58.778097	58.778097	58.778097	0	0	1	-360	360;
	7767	2086	0.003661	0.008818	0.0011	81.776631	81.776631	81.776631	0	0	1	-360	360;
	7767	3761	0.029843	0.069744	0.0073	73.16316	73.16316	73.16316	0	0	1	-360	360;
	249	3152	0.006438	0.01824	0.00218	57.990083	57.990083	57.990083	0	0	1	-360	360;
	249	3152	0.006322	0.018636	0.00214	57.20588	57.20588	57.20588	0	0	1	-360	360;
	7576	7680	0.00068	0.00797	0.15917	1247.39459	1247.39459	1247.39459	0	0	1	-360	360;
	3152	7859	0.032364	0.088289	0.01098	117.505704	117.505704	117.505704	0	0	1	-360	360;
	3152	3761	0.027066	0.084496	0.00915	80.201556	80.201556	80.201556	0	0	1	-360	360;
	3152	3761	0.026554	0.086256	0.00898	65.376188	65.376188	65.376188	0	0	1	-360	360;
	3761	2086	0.029901	0.071322	0.00679	74.569429	74.569429	74.569429	0	0	1	-360	360;
	535	456	0.011901	0.023818	0.00242	76.754948	76.754948	76.754948	0	0	1	-360	360;
	6990	456	0.012132	0.023322	0.00247	62.553552	62.553552	62.553552	0	0	1	-360	360;
	4457	6822	0.0098	0.062349	0.07457	410.221407	410.221407	410.221407	0	0	1	-360	360;
	535	6258	0.004182	0.012678	0.00144	64.686676	64.686676	64.686676	0	0	1	-360	360;
	6990	6258	0.003636	0.014636	0.00125	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7859	4379	0.01919	0.045041	0.00385	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5199	4736	0.009793	0.037174	0.00337	81.246398	81.246398	81.246398	0	0	1	-360	360;
	4736	1299	0.004281	0.010339	0.00089	102.487334	102.487334	102.487334	0	0	1	-360	360;
	4736	1299	0.005521	0.021471	0.0019	62.74122	62.74122	62.74122	0	0	1	-360	360;
	810	231	0.0011	0.01325	0.26238	1375.13732	1375.13732	1375.13732	0	0	1	-360	360;
	992	7647	0.018421	0.051372	0.00634	69.199465	69.199465	69.199465	0	0	1	-360	360;
	992	912	0.011521	0.042529	0.00397	73.093999	73.093999	73.093999	0	0	1	-360	360;
	2326	3089	0.025182	0.071554	0.00867	69.405805	69.405805	69.405805	0	0	1	-360	360;
	8144	2938	0	0.000413	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1281	3844	0.014603	0.03386	0.003	94.753329	94.753329	94.753329	0	0	1	-360	360;
	3991	2326	0.022066	0.054521	0.00448	67.296877	67.296877	67.296877	0	0	1	-360	360;
	1790	3991	0.041471	0.090793	0.0084	61.636552	61.636552	61.636552	0	0	1	-360	360;
	3761	5476	0.008843	0.031901	0.00302	57.157677	57.157677	57.157677	0	0	1	-360	360;
	3414	4548	0.00038	0.003091	0.00406	536.215979	536.215979	536.215979	0	0	1	-360	360;
	3414	4548	0.0005	0.002901	0.00509	531.306135	531.306135	531.306135	0	0	1	-360	360;
	6973	7575	0.032132	0.318198	0.01653	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6973	1998	0.058554	0.146124	0.01127	74.532467	74.532467	74.532467	0	0	1	-360	360;
	5154	6973	0.023917	0.073289	0.00826	80.321587	80.321587	80.321587	0	0	1	-360	360;
	6973	3844	0.040818	0.089273	0.00807	82.866819	82.866819	82.866819	0	0	1	-360	360;
	2178	2052	2e-05	0.00025	0.00456	1213.87088	1213.87088	1213.87088	0	0	1	-360	360;
	2178	2052	2e-05	0.00026	0.00446	1101.24982	1101.24982	1101.24982	0	0	1	-360	360;
	2052	2482	0.00082	0.00822	0.19419	1110.17012	1110.17012	1110.17012	0	0	1	-360	360;
	2004	927	0.014339	0.031777	0.00294	57.157677	57.157677	57.157677	0	0	1	-360	360;
	4350	5769	0.011364	0.034364	0.00392	57.157677	57.157677	57.157677	0	0	1	-360	360;
	5769	2716	0.000347	0.001066	0.00012	79.85118	79.85118	79.85118	0	0	1	-360	360;
	3861	7575	0.012752	0.036041	0.00468	171.47303	171.47303	171.47303	0	0	1	-360	360;
	4720	7575	0.000669	0.001281	0.00014	171.47303	171.47303	171.47303	0	0	1	-360	360;
	4720	7575	0.000661	0.001314	0.00014	160.016155	160.016155	160.016155	0	0	1	-360	360;
	8046	3657	0.005289	0.014942	0.00183	61.598638	61.598638	61.598638	0	0	1	-360	360;
	1790	7488	0.004256	0.007364	0.00087	59.340909	59.340909	59.340909	0	0	1	-360	360;
	160	7488	0.003992	0.009107	0.00081	117.314036	117.314036	117.314036	0	0	1	-360	360;
	5995	3308	0.041488	0.08662	0.00826	68.869856	68.869856	68.869856	0	0	1	-360	360;
	3308	8342	0.00486	0.010132	0.00101	91.970512	91.970512	91.970512	0	0	1	-360	360;
	3308	387	0.00495	0.009934	0.00103	68.607121	68.607121	68.607121	0	0	1	-360	360;
	387	8342	0.00505	0.009719	0.00105	72.531186	72.531186	72.531186	0	0	1	-360	360;
	912	7647	0.00776	0.027521	0.00268	57.157677	57.157677	57.157677	0	0	1	-360	360;
	1958	2243	0.00262	0.007628	0.0009	78.157217	78.157217	78.157217	0	0	1	-360	360;
	5154	3844	0.001793	0.005471	0.00061	76.785813	76.785813	76.785813	0	0	1	-360	360;
	8278	3377	0.000537	0.015661	0.00184	90.598728	90.598728	90.598728	0	0	1	-360	360;
	4350	2716	0.01124	0.036826	0.00388	57.157677	57.157677	57.157677	0	0	1	-360	360;
	7758	1757	0.00156	0.01791	0.37415	1166.70509	1166.70509	1166.70509	0	0	1	-360	360;
	1757	6266	0.00155	0.01686	0.40732	1299.10049	1299.10049	1299.10049	0	0	1	-360	360;
	1757	6266	0.00152	0.01719	0.399	1246.28819	1246.28819	1246.28819	0	0	1	-360	360;
	7489	1757	0.00068	0.00644	0.23404	1435.51276	1435.51276	1435.51276	0	0	1	-360	360;
	7489	1757	0.00069	0.0063	0.23845	1449.74984	1449.74984	1449.74984	0	0	1	-360	360;
	1757	5285	0.00112	0.01064	0.27169	1725.27028	1725.27028	1725.27028	0	0	1	-360	360;
	3132	4518	0.00151	0.011539	0.01729	481.34461	481.34461	481.34461	0	0	1	-360	360;
	7381	3132	0.00788	0.05236	0.08792	479.500703	479.500703	479.500703	0	0	1	-360	360;
	1406	6266	0.00128	0.0126	0.30891	1282.33338	1282.33338	1282.33338	0	0	1	-360	360;
	740	1406	0.00037	0.00409	0.09538	1072.4367	1072.4367	1072.4367	0	0	1	-360	360;
	740	1406	0.00036	0.00417	0.09344	1152.64309	1152.64309	1152.64309	0	0	1	-360	360;
	1406	5285	0.00122	0.0112	0.29771	1041.28179	1041.28179	1041.28179	0	0	1	-360	360;
	7381	7546	0.006591	0.04075	0.073	436.572621	436.572621	436.572621	0	0	1	-360	360;
	2616	4518	0.007961	0.05756	0.08822	423.668322	423.668322	423.668322	0	0	1	-360	360;
	4632	2398	0.00059	0.0085	0.22559	1083.56257	1083.56257	1083.56257	0	0	1	-360	360;
	2398	4994	0.00084	0.0118	0.32178	970.642693	970.642693	970.642693	0	0	1	-360	360;
	2398	5848	0.00178	0.01804	0.63425	809.359803	809.359803	809.359803	0	0	1	-360	360;
	4513	2257	0.00937	0.073969	0.0931	426.43323	426.43323	426.43323	0	0	1	-360	360;
	4513	2909	0.011081	0.06593	0.07966	410.138338	410.138338	410.138338	0	0	1	-360	360;
	4513	2577	0.009669	0.04251	0.09891	385.804791	385.804791	385.804791	0	0	1	-360	360;
	6286	8201	0.00073	0.00713	0.20473	797.794934	797.794934	797.794934	0	0	1	-360	360;
	4783	4994	0.00034	0.00345	0.13025	1081.33397	1081.33397	1081.33397	0	0	1	-360	360;
	6751	6406	0.00012	0.00128	0.04699	1013.60271	1013.60271	1013.60271	0	0	1	-360	360;
	6751	232	0.00012	0.00123	0.04941	1033.59952	1033.59952	1033.59952	0	0	1	-360	360;
	1573	1873	9e-05	0.00112	0.02573	938.008188	938.008188	938.008188	0	0	1	-360	360;
	1217	766	0.00062	0.00866	0.23108	1456.0032	1456.0032	1456.0032	0	0	1	-360	360;
	6751	1217	0.00157	0.01614	0.59842	1471.35063	1471.35063	1471.35063	0	0	1	-360	360;
	6751	1217	0.0014	0.01829	0.53314	1434.35568	1434.35568	1434.35568	0	0	1	-360	360;
	261	2543	0.00078	0.00805	0.29614	1361.85855	1361.85855	1361.85855	0	0	1	-360	360;
	2543	3741	0.00094	0.01427	0.3911	1484.72088	1484.72088	1484.72088	0	0	1	-360	360;
	7037	4994	0.00024	0.00283	0.1308	1354.94767	1354.94767	1354.94767	0	0	1	-360	360;
	4994	766	0.00099	0.00865	0.27607	1267.80012	1267.80012	1267.80012	0	0	1	-360	360;
	6751	4994	0.0011	0.01426	0.32867	1321.73265	1321.73265	1321.73265	0	0	1	-360	360;
	2257	625	0.006919	0.050521	0.06011	420.229717	420.229717	420.229717	0	0	1	-360	360;
	3694	4645	0.0006	0.00646	0.22339	1370.4162	1370.4162	1370.4162	0	0	1	-360	360;
	3694	5280	0.00127	0.0128	0.32653	824.764489	824.764489	824.764489	0	0	1	-360	360;
	3694	8263	0.00052	0.00496	0.19362	1360.33157	1360.33157	1360.33157	0	0	1	-360	360;
	3694	1019	0.00064	0.00886	0.24217	1454.68026	1454.68026	1454.68026	0	0	1	-360	360;
	3694	1019	0.00072	0.00868	0.27446	1555.4383	1555.4383	1555.4383	0	0	1	-360	360;
	261	7475	0.00232	0.02522	0.6334	833.099014	833.099014	833.099014	0	0	1	-360	360;
	6751	261	0.0003	0.00304	0.08304	1468.67052	1468.67052	1468.67052	0	0	1	-360	360;
	6751	261	0.00032	0.00306	0.10801	1335.89074	1335.89074	1335.89074	0	0	1	-360	360;
	6751	261	0.00028	0.00307	0.10697	1460.45447	1460.45447	1460.45447	0	0	1	-360	360;
	8201	6861	0.00092	0.00969	0.33431	951.156636	951.156636	951.156636	0	0	1	-360	360;
	8201	5280	0.00181	0.01696	0.50006	1044.67602	1044.67602	1044.67602	0	0	1	-360	360;
	8201	5848	0.00132	0.01148	0.34945	1291.52485	1291.52485	1291.52485	0	0	1	-360	360;
	6861	3741	0.00108	0.01087	0.39137	1365.3956	1365.3956	1365.3956	0	0	1	-360	360;
	4645	7475	0.00209	0.01924	0.53053	849.76543	849.76543	849.76543	0	0	1	-360	360;
	8263	4645	0.00036	0.00457	0.13461	1061.05744	1061.05744	1061.05744	0	0	1	-360	360;
	4645	1573	0.0017	0.01986	0.48822	767.876077	767.876077	767.876077	0	0	1	-360	360;
	766	2085	0.00059	0.00982	0.23966	966.362553	966.362553	966.362553	0	0	1	-360	360;
	7948	2577	0.00494	0.03182	0.04985	276.410714	276.410714	276.410714	0	0	1	-360	360;
	7948	2577	0.004841	0.03245	0.04884	279.901524	279.901524	279.901524	0	0	1	-360	360;
	3741	5848	0.00036	0.00391	0.09835	1623.83037	1623.83037	1623.83037	0	0	1	-360	360;
	1019	1573	0.00071	0.00983	0.2685	944.591956	944.591956	944.591956	0	0	1	-360	360;
	1019	1573	0.00081	0.00963	0.3043	1038.41147	1038.41147	1038.41147	0	0	1	-360	360;
	4894	719	0.006271	0.03645	0.05313	359.078624	359.078624	359.078624	0	0	1	-360	360;
	7008	4479	0.186281	0.374281	0	73.493912	73.493912	73.493912	0	0	1	-360	360;
	7295	4479	0.089612	0.230339	0	67.299163	67.299163	67.299163	0	0	1	-360	360;
	7141	3818	0.00393	0.04552	0.95789	1344.74918	1344.74918	1344.74918	0	0	1	-360	360;
	3590	5527	0.004169	0.02651	0.15434	455.923543	455.923543	455.923543	0	0	1	-360	360;
	3590	5527	0.003831	0.02438	0.14132	469.075143	469.075143	469.075143	0	0	1	-360	360;
	3590	4894	0.00789	0.04725	0.08625	601.261415	601.261415	601.261415	0	0	1	-360	360;
	3590	4894	0.008039	0.046219	0.08785	395.129113	395.129113	395.129113	0	0	1	-360	360;
	867	7318	0.040174	0.084397	0	64.977037	64.977037	64.977037	0	0	1	-360	360;
	3198	550	0.00166	0.01276	0.14114	1428.94192	1428.94192	1428.94192	0	0	1	-360	360;
	6965	5527	0.00094	0.00599	0.03533	399.647618	399.647618	399.647618	0	0	1	-360	360;
	6965	5527	0.00107	0.00587	0.04005	411.050575	411.050575	411.050575	0	0	1	-360	360;
	981	3701	0.011901	0.028364	0.00244	70.917625	70.917625	70.917625	0	0	1	-360	360;
	3701	7318	0.059091	0.102636	0	68.529387	68.529387	68.529387	0	0	1	-360	360;
	7295	3701	0.090471	0.19919	0	74.524465	74.524465	74.524465	0	0	1	-360	360;
	543	1535	0.00755	0.050919	0.35194	456.220762	456.220762	456.220762	0	0	1	-360	360;
	543	1535	0.0077	0.04988	0.35884	519.602529	519.602529	519.602529	0	0	1	-360	360;
	3863	7008	0.238893	0.453901	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	402	3863	0.062719	0.104479	0	57.841092	57.841092	57.841092	0	0	1	-360	360;
	4894	1535	0.000979	0.074909	0.45818	491.050364	491.050364	491.050364	0	0	1	-360	360;
	4894	1535	0.000961	0.076349	0.44863	505.951371	505.951371	505.951371	0	0	1	-360	360;
	4596	1535	0.002159	0.079031	0.05909	430.889242	430.889242	430.889242	0	0	1	-360	360;
	4596	1535	0.001961	0.09125	0.05467	472.693605	472.693605	472.693605	0	0	1	-360	360;
	402	7008	0.066636	0.183934	0	57.157677	57.157677	57.157677	0	0	1	-360	360;
	6317	4211	0.00431	0.09324	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	6317	4211	0.0044	0.091339	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	1216	6469	0.006356	0.142878	0	119.999944	119.999944	119.999944	1.026667	0	1	-360	360;
	2244	6477	0.00523	0.10214	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	2244	6477	0.00532	0.099917	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	2244	6477	0.00442	0.105388	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	7486	5134	0.001002	0.059033	0	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	7486	5134	0.000812	0.056028	0	145.199621	145.199621	145.199621	1.052632	0	1	-360	360;
	6630	7196	0.000921	0.058347	0	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	7196	3934	0.0016	0.07243	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	7196	3934	0.00182	0.07098	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	7196	3934	0.00185	0.06953	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	248	7559	0.00349	0.07969	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	248	7559	0.00355	0.07799	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	1408	248	0.001101	0.048618	0	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	1408	248	0.000984	0.055107	0	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	5901	8007	0.00175	0.07752	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5901	8007	0.00171	0.07901	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5423	3941	0.00302	0.091446	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5423	3941	0.00295	0.084678	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5436	4207	0.0018	0.07815	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5436	4207	0.00185	0.07724	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	5436	4207	0.00193	0.07007	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	95	2305	0.00205	0.094653	0	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	4709	466	0.000454	0.028943	-0.00528	710.594074	710.594074	710.594074	1.013237	0.61829	1	-360	360;
	4709	466	0.0004	0.032807	-0.00528	710.594074	710.594074	710.594074	1.013237	0.61829	1	-360	360;
	2006	6931	0.000342	0.033341	-0.006005	733.486867	733.486867	733.486867	0.981613	0	1	-360	360;
	1240	6931	0.000342	0.03398	-0.006005	733.486867	733.486867	733.486867	0.981613	0	1	-360	360;
	4061	6931	0.000121	0.025184	-0.001154	720.000011	720.000011	720.000011	0.966268	14.9238	1	-360	360;
	4061	3482	0.000121	0.015705	-0.001154	720.000011	720.000011	720.000011	0.992741	6.90758	1	-360	360;
	6313	1672	0.000347	0.031821	-0.006411	706.071724	706.071724	706.071724	1.019726	1.22888	1	-360	360;
	6313	1672	0.000356	0.031175	-0.006411	706.071724	706.071724	706.071724	1.019726	1.22888	1	-360	360;
	3923	604	0	0.026157	0	360.000005	360.000005	360.000005	1	-1.357277	1	-360	360;
	3923	1672	0	0.025592	0	360.000005	360.000005	360.000005	1	1.35728	1	-360	360;
	7457	5080	0.000431	0.03159	-0.005565	737.077893	737.077893	737.077893	0.976831	-3.20845	1	-360	360;
	7457	5080	0.00038	0.034638	-0.005565	732.760237	732.760237	732.760237	0.982587	-2.55126	1	-360	360;
	3188	5303	0.0004	0.03334	-0.00514	715.090755	715.090755	715.090755	1.006865	0	1	-360	360;
	3188	5303	0.0004	0.033979	-0.00514	715.090755	715.090755	715.090755	1.006865	0	1	-360	360;
	3745	5303	0.000121	0.021666	-0.001154	842.44471	842.44471	842.44471	0.977039	12.3017	1	-360	360;
	3745	3367	0.000121	0.020058	-0.001154	842.44471	842.44471	842.44471	0.977039	12.3017	1	-360	360;
	311	271	0.000131	0.022842	-0.001154	810.322096	810.322096	810.322096	0.966268	14.9238	1	-360	360;
	311	6133	0.000141	0.019261	-0.001154	810.322096	810.322096	810.322096	0.977039	-12.30169	1	-360	360;
	2751	3539	0.000194	0.013059	-0.009884	1200.00041	1200.00041	1200.00041	1.093651	0	1	-360	360;
	7249	240	0.000463	0.031001	-0.006163	762.926569	762.926569	762.926569	0.943734	0	1	-360	360;
	7249	240	0.000472	0.030331	-0.006163	762.926569	762.926569	762.926569	0.943734	0	1	-360	360;
	413	6700	0.000163	0.015501	-0.012037	1439.99957	1439.99957	1439.99957	0.954768	0	1	-360	360;
	5190	3652	0.001987	0.071722	0	199.999664	199.999664	199.999664	0.9	0	1	-360	360;
	2275	5757	0.000756	0.051399	0	310.591524	310.591524	310.591524	1.159079	0	1	-360	360;
	2275	5757	0.000746	0.05345	0	310.591524	310.591524	310.591524	1.159079	0	1	-360	360;
	6584	8328	0.00069	0.04333	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	2483	6697	0.0022	0.065399	0	199.999664	199.999664	199.999664	0.9	0	1	-360	360;
	2483	6697	0.002235	0.064019	0	199.999664	199.999664	199.999664	0.9	0	1	-360	360;
	7809	4613	0.00168	0.07424	0	180.000193	180.000193	180.000193	1	0	1	-360	360;
	874	3149	0.002036	0.070048	0	163.63595	163.63595	163.63595	1.1	0	1	-360	360;
	874	3149	0.001997	0.071368	0	163.63595	163.63595	163.63595	1.1	0	1	-360	360;
	3296	874	0.00035	0.032859	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	3296	874	0.0004	0.0322	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2369	3459	0.00205	0.06915	0	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2369	3459	0.00082	0.07342	0	180.000193	180.000193	180.000193	1	0	1	-360	360;
	318	7777	0.002658	0.07732	0	195.513549	195.513549	195.513549	0.920652	0	1	-360	360;
	318	7777	0.002712	0.075649	0	195.513549	195.513549	195.513549	0.920652	0	1	-360	360;
	3664	3510	0.00065	0.04604	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	6445	2378	0.00062	0.03629	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	6445	7407	0.00019	0.03263	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2432	884	0.001797	0.07284	0	173.096689	173.096689	173.096689	1.039882	0	1	-360	360;
	2942	7610	0.002712	0.072578	0	190.958463	190.958463	190.958463	0.942614	0	1	-360	360;
	6604	318	0.00047	0.0307	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	6980	3587	0.002601	0.071481	0	191.999876	191.999876	191.999876	0.9375	0	1	-360	360;
	6980	3587	0.002653	0.06993	0	191.999876	191.999876	191.999876	0.9375	0	1	-360	360;
	5103	6980	0.00037	0.034169	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	5103	6980	0.00037	0.034831	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	6065	3450	0.0004	0.03524	0	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	6065	4726	0.00069	0.04525	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	8075	7047	0.00069	0.04231	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	2817	6831	0.00119	0.073146	0	167.44189	167.44189	167.44189	1.075	0	1	-360	360;
	2817	6831	0.002764	0.071607	0	175.609722	175.609722	175.609722	1.025	0	1	-360	360;
	8195	5933	0.000742	0.067219	-0.009597	353.959617	353.959617	353.959617	1.017065	0	1	-360	360;
	8195	5933	0.000793	0.06388	-0.009676	360.57037	360.57037	360.57037	0.998417	0	1	-360	360;
	5509	8341	0.000609	0.046809	-0.018347	707.746132	707.746132	707.746132	0.941014	0	1	-360	360;
	5509	1798	0.00059	0.04521	-0.01719	692.523103	692.523103	692.523103	0.901053	0	1	-360	360;
	7101	1798	0.519194	5.108058	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	7775	2089	0.000559	0.039982	-0.016839	609.157454	609.157454	609.157454	0.931778	0	1	-360	360;
	7775	5074	0.000537	0.045312	-0.016839	610.095359	610.095359	610.095359	0.930347	0	1	-360	360;
	4483	1659	0.000889	0.061549	-0.009676	358.807766	358.807766	358.807766	1.003323	0	1	-360	360;
	1659	718	1.467822	7.764888	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	6080	4172	0.000586	0.049031	-0.01719	667.676834	667.676834	667.676834	0.934584	0	1	-360	360;
	2216	3186	0.000417	0.047801	-0.018156	671.216523	671.216523	671.216523	0.911778	0	1	-360	360;
	1244	6192	0.000755	0.05143	-0.016859	618.265997	618.265997	618.265997	0.989866	0	1	-360	360;
	1528	207	0.000733	0.052399	-0.01686	662.331102	662.331102	662.331102	0.924009	0	1	-360	360;
	7101	6192	0.00078	0.064311	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	596	2981	0.00089	0.048632	-0.016223	568.416813	568.416813	568.416813	0.950007	0	1	-360	360;
	4738	718	0.00855	0.08972	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	1313	718	0.00223	0.04836	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	6511	1975	0.000581	0.043111	-0.01686	678.413055	678.413055	678.413055	0.902105	0	1	-360	360;
	6199	8139	0.000579	0.043272	-0.01719	692.559961	692.559961	692.559961	0.901005	0	1	-360	360;
	3328	2948	0.000586	0.049031	-0.018347	721.242757	721.242757	721.242757	0.923407	0	1	-360	360;
	7415	4032	0.000559	0.049192	-0.01686	654.555372	654.555372	654.555372	0.934985	0	1	-360	360;
	8196	6862	0.000573	0.051129	-0.01686	662.331102	662.331102	662.331102	0.924009	0	1	-360	360;
	2062	4738	0.000417	0.039229	-0.019195	739.615174	739.615174	739.615174	0.973479	0	1	-360	360;
	4738	6599	0.230542	1.197733	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	4738	7555	0.02284	0.2644	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	4738	2222	0.08537	0.736798	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	4738	4239	1.402768	3.640289	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	4738	4091	1.050227	4.450207	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	7413	6599	0.000716	0.04371	-0.01686	654.555372	654.555372	654.555372	0.934985	0	1	-360	360;
	1313	6599	0.518622	2.317689	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	5000	7555	0.00059	0.045662	-0.016839	586.169225	586.169225	586.169225	0.968322	0	1	-360	360;
	5000	1395	0.000603	0.050959	-0.01719	652.079959	652.079959	652.079959	0.956938	0	1	-360	360;
	1313	7555	0.047769	0.554178	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	7041	4687	0.000591	0.045153	-0.017622	646.447261	646.447261	646.447261	0.918868	0	1	-360	360;
	7041	246	0.0006	0.047778	-0.018347	719.929689	719.929689	719.929689	0.925091	0	1	-360	360;
	7101	5242	0.02436	0.270151	0	11687.1399	11687.1399	11687.1399	1	0	1	-360	360;
	7549	1313	0.000802	0.05727	-0.010077	390.891374	390.891374	390.891374	0.967021	0	1	-360	360;
	7549	1313	0.000823	0.056059	-0.010077	390.891374	390.891374	390.891374	0.967021	0	1	-360	360;
	1313	2222	0.188099	1.291901	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	1313	4239	2.466323	7.26343	0	11998.8001	11998.8001	11998.8001	1	0	1	-360	360;
	2063	2637	0.000523	0.050352	-0.018347	730.017604	730.017604	730.017604	0.912306	0	1	-360	360;
	8028	5242	0.000682	0.049638	-0.017851	654.634353	654.634353	654.634353	0.989866	0	1	-360	360;
	8028	7348	0.00099	0.04736	-0.006657	213.408058	213.408058	213.408058	1.012146	0	1	-360	360;
	288	2222	0.000616	0.048759	-0.018347	662.879365	662.879365	662.879365	1.004708	0	1	-360	360;
	6930	7781	9e-05	0.008834	-0.056	1680.00005	1680.00005	1680.00005	1	-4.394626	1	-360	360;
	5304	6827	9e-05	0.00866	-0.056	1680.00005	1680.00005	1680.00005	1	-4.394626	1	-360	360;
	78	4091	0.000664	0.045609	-0.01686	654.555372	654.555372	654.555372	0.934985	0	1	-360	360;
	2282	4239	0.000649	0.044862	-0.018347	713.786902	713.786902	713.786902	0.933053	0	1	-360	360;
	5443	7632	0.00067	0.04326	-0.018347	730.080131	730.080131	730.080131	0.91223	0	1	-360	360;
	1962	5443	9e-05	0.009385	-0.056	1680.00005	1680.00005	1680.00005	1	5.8573	1	-360	360;
	6347	3650	0.000246	0.018417	-0.015389	1394.52136	1394.52136	1394.52136	0.987971	3.94703	1	-360	360;
	3650	5011	0.00131	0.07578	-0.006612	229.565425	229.565425	229.565425	1.045455	0	1	-360	360;
	3650	5011	0.00128	0.07077	-0.006612	229.565425	229.565425	229.565425	1.045455	0	1	-360	360;
	2886	7955	0.00131	0.062651	-0.01093	240.000131	240.000131	240.000131	1	0	1	-360	360;
	2886	7955	0.00133	0.0577	-0.01091	240.000131	240.000131	240.000131	1	0	1	-360	360;
	4599	1318	0.001495	0.059475	-0.00682	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	4599	8383	0.001453	0.061392	-0.00327	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	5956	4924	0.001475	0.074554	0	329.999915	329.999915	329.999915	0.981267	0	1	-360	360;
	5956	4924	0.001309	0.084508	0	329.999915	329.999915	329.999915	0.981267	0	1	-360	360;
	803	4439	0.00119	0.084099	-0.00722	240.000131	240.000131	240.000131	1	0	1	-360	360;
	3995	979	0.00117	0.064859	-0.0109	240.000131	240.000131	240.000131	1	0	1	-360	360;
	3995	979	0.00114	0.066091	-0.0109	240.000131	240.000131	240.000131	1	0	1	-360	360;
	6869	5401	0.001225	0.083401	0	329.999915	329.999915	329.999915	0.973021	0	1	-360	360;
	6869	5401	0.001394	0.082574	0	329.999915	329.999915	329.999915	0.973021	0	1	-360	360;
	3125	4185	0.001415	0.080475	0	329.999915	329.999915	329.999915	0.973021	0	1	-360	360;
	3125	4185	0.001447	0.078793	0	329.999915	329.999915	329.999915	0.973021	0	1	-360	360;
	8182	8011	0.001444	0.074683	-0.00459	329.999915	329.999915	329.999915	0.981267	0	1	-360	360;
	8182	8011	0.001475	0.073051	-0.00459	329.999915	329.999915	329.999915	0.981267	0	1	-360	360;
	833	4158	0.001316	0.070415	-0.01036	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	7065	3123	0.00128	0.065919	-0.00353	240.000131	240.000131	240.000131	1	0	1	-360	360;
	8066	434	0.001056	0.075919	-0.01919	329.999915	329.999915	329.999915	1.052632	0	1	-360	360;
	1148	434	0.001451	0.078431	-0.00591	240.000131	240.000131	240.000131	0.905512	0	1	-360	360;
	1148	434	0.001415	0.075297	-0.00589	240.000131	240.000131	240.000131	0.905512	0	1	-360	360;
	2427	1534	0.001411	0.061844	-0.00683	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	2427	1534	0.001432	0.058938	-0.01037	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	2167	1888	0.00229	0.085231	-0.005908	220.557375	220.557375	220.557375	0.979339	0	1	-360	360;
	4692	517	0.001474	0.06058	-0.00514	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	1309	1206	0.00125	0.064651	-0.0035	240.000131	240.000131	240.000131	1	0	1	-360	360;
	1309	1206	0.00123	0.065919	-0.00352	240.000131	240.000131	240.000131	1	0	1	-360	360;
	1309	1206	0.0012	0.06718	-0.00352	240.000131	240.000131	240.000131	1	0	1	-360	360;
	2195	6652	0.001221	0.069631	-0.01052	240.000131	240.000131	240.000131	0.982906	0	1	-360	360;
	2195	6652	0.00119	0.065604	-0.01054	240.000131	240.000131	240.000131	0.982906	0	1	-360	360;
	1520	8297	0.001039	0.045925	-0.01168	329.999915	329.999915	329.999915	0.981267	0	1	-360	360;
	3537	8296	0.000406	0.016466	-0.00659	1322.86932	1322.86932	1322.86932	1.054276	2.34996	1	-360	360;
	8296	1714	0.001563	0.060356	-0.00697	240.000131	240.000131	240.000131	0.982906	0	1	-360	360;
	8296	1714	0.001459	0.05904	-0.01054	240.000131	240.000131	240.000131	0.982906	0	1	-360	360;
	5487	1001	0.001531	0.068989	-0.00644	240.000131	240.000131	240.000131	0.942623	0	1	-360	360;
	7544	5836	0.000318	0.014328	-0.00671	959.999933	959.999933	959.999933	1.002506	0	1	-360	360;
	1864	5451	0.000246	0.018417	-0.015389	959.999933	959.999933	959.999933	0.987971	3.94703	1	-360	360;
	5451	3336	0.00121	0.066091	-0.0109	240.000131	240.000131	240.000131	1	0	1	-360	360;
	5451	3336	0.00118	0.062289	-0.01091	240.000131	240.000131	240.000131	1	0	1	-360	360;
	7753	5680	0.001646	0.071174	-0.00895	240.000131	240.000131	240.000131	0.905512	0	1	-360	360;
	7753	5680	0.00229	0.087099	-0.004416	206.608616	206.608616	206.608616	1.045455	0	1	-360	360;
	1759	989	0.001421	0.061107	-0.01036	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	8343	2858	0.001573	0.080941	-0.00441	329.999915	329.999915	329.999915	0.956938	0	1	-360	360;
	8343	2858	0.001594	0.078385	-0.00436	329.999915	329.999915	329.999915	0.956938	0	1	-360	360;
	4988	7620	0.001369	0.06454	-0.00684	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	4988	7620	0.001348	0.065203	-0.00685	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	1041	4623	0.001295	0.071353	-0.00511	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	1041	4623	0.001295	0.070826	-0.00512	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	1372	3919	0.001441	0.084934	0	329.999915	329.999915	329.999915	0.956938	0	1	-360	360;
	1372	3919	0.001463	0.083626	0	329.999915	329.999915	329.999915	0.956938	0	1	-360	360;
	3627	31	0.000448	0.016447	-0.00528	755.99989	755.99989	755.99989	1.002506	0	1	-360	360;
	3627	1495	0.001134	0.053193	-0.00475	329.999915	329.999915	329.999915	0.998185	0	1	-360	360;
	31	1495	0.00136	0.057539	-0.00722	240.000131	240.000131	240.000131	1	0	1	-360	360;
	1924	3812	0.001432	0.06454	-0.00506	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	602	3670	0.00195	0.114628	0	220.557375	220.557375	220.557375	0.979339	0	1	-360	360;
	1664	6759	0.001295	0.067067	0	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	4354	8153	0.001242	0.074563	-0.00861	240.000131	240.000131	240.000131	0.974576	0	1	-360	360;
	5587	4112	0.00126	0.060099	-0.00535	240.000131	240.000131	240.000131	1	0	1	-360	360;
	5587	4112	0.00135	0.06051	-0.00721	240.000131	240.000131	240.000131	1	0	1	-360	360;
	6758	3185	0.00036	0.042469	-0.023005	718.511971	718.511971	718.511971	1.052174	0	1	-360	360;
	1091	5409	0.001278	0.05335	-0.013671	479.999638	479.999638	479.999638	0.973021	0	1	-360	360;
	6146	4933	0.000503	0.028042	-0.00085	479.999638	479.999638	479.999638	0.977126	-1.493575	1	-360	360;
	6146	4933	0.000503	0.02743	-0.00085	479.999638	479.999638	479.999638	0.977126	-1.493575	1	-360	360;
	3119	4130	0.000333	0.021103	-0.00228	720.000115	720.000115	720.000115	1.011235	0	1	-360	360;
	8085	4208	0.00044	0.030928	-0.00085	479.999638	479.999638	479.999638	0.977126	0	1	-360	360;
	3175	5270	0.000164	0.013858	-0.003881	1200.00041	1200.00041	1200.00041	1.030625	0	1	-360	360;
	4492	2715	0.000353	0.026265	-0.00236	943.750145	943.750145	943.750145	0.937567	-1.77827	1	-360	360;
	1421	3246	0.000159	0.016777	-0.00329	1200.00041	1200.00041	1200.00041	0.937567	0.86289	1	-360	360;
	2979	2254	0.000273	0.016238	-0.0112	990.270258	990.270258	990.270258	0.956938	0	1	-360	360;
	1246	5914	0.000289	0.015803	-0.011017	720.000115	720.000115	720.000115	0.972469	1.65217	1	-360	360;
	5244	5835	0.00142	0.02765	0	480.000261	480.000261	480.000261	1	0	1	-360	360;
	5244	5835	0.00145	0.02707	0	480.000261	480.000261	480.000261	1	0	1	-360	360;
	1531	342	0.000469	0.030842	0	479.999638	479.999638	479.999638	1.024958	0	1	-360	360;
	1472	342	0.00435	0.044132	0	480.000036	480.000036	480.000036	1	-5.914401	1	-360	360;
	3826	6545	0.000352	0.020852	-0.002327	720.000115	720.000115	720.000115	1.02738	0	1	-360	360;
	3714	135	0.0003	0.01099	-0.00127	214.038594	214.038594	214.038594	1	0	1	-360	360;
	3714	135	0.00029	0.0112	-0.00127	210.036206	210.036206	210.036206	1	0	1	-360	360;
	4394	3804	0.000375	0.018768	-0.00203	720.000115	720.000115	720.000115	1.061268	0	1	-360	360;
	6046	1994	0.000537	0.028553	-0.002479	720.000115	720.000115	720.000115	0.998185	0	1	-360	360;
	6046	1994	0.000537	0.028002	-0.001951	720.000115	720.000115	720.000115	0.998185	0	1	-360	360;
	1507	3621	0.000622	0.033296	-0.00278	360.000386	360.000386	360.000386	0.956938	0	1	-360	360;
	1507	3621	0.000622	0.029468	-0.00117	479.999638	479.999638	479.999638	0.956938	0.55119	1	-360	360;
	2463	343	0.00072	0.026242	-0.00704	479.999638	479.999638	479.999638	0.957131	0	1	-360	360;
	2463	343	0.000644	0.029735	-0.00704	479.999638	479.999638	479.999638	0.957131	0	1	-360	360;
	5785	4331	0.00314	0.168899	0	169.092984	169.092984	169.092984	1	1.43228	1	-360	360;
	5785	4331	0.00307	0.172155	0	165.899775	165.899775	165.899775	1	1.43228	1	-360	360;
	554	5356	0.000188	0.016506	-0.001192	959.999933	959.999933	959.999933	1.016322	0	1	-360	360;
	1961	5458	0.000259	0.027074	0	414.000049	414.000049	414.000049	1.002506	0	1	-360	360;
	2990	265	0.000437	0.02611	0	461.99975	461.99975	461.99975	1.047869	0	1	-360	360;
	4855	3033	0.000446	0.025573	0	405.600365	405.600365	405.600365	1.047869	0	1	-360	360;
	1724	2345	0.00032	0.02243	0	531.442274	531.442274	531.442274	1.034211	0	1	-360	360;
	950	5657	6.091942	8.676029	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	5657	6.444008	8.676035	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	5657	23.9876	40.776864	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	5657	35.24587	57.202482	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	5657	45.97521	69.597103	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	5657	0.081469	0.775971	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	5657	32.03719	50.568181	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	5657	0.628884	2.485744	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	6716	12.31426	19.110329	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	6716	13.03616	19.128307	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	6716	0.197581	1.393719	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	6716	1.415165	4.81467	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4904	2487	0.13108	0.821607	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5701	2487	0.129079	0.839681	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5691	2384	0.069591	0.599051	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5691	1042	0.04983	0.471267	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5691	3026	0.00502	0.310928	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	7063	5.68843	10.297518	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	7063	48.32851	56.749998	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	7063	23.27066	28.824377	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	7063	50.96074	55.43596	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	5636	5.967355	9.180988	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	5636	44.27686	58.522723	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	5636	20.49132	30.954553	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	5636	43.12397	61.993787	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	3654	-0.04093	0.365413	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	937	0.140533	1.138574	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	3117	0.151053	1.206163	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	3675	3.496607	8.23892	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	1124	3.21482	7.345568	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	394	0.136738	1.1482	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	1744	0.138539	1.094044	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	3024	0.077251	0.917036	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	1447	0.954778	3.73151	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	8114	0.745845	4.144945	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8032	1071	0.723892	4.195776	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	3654	0.47093	3.07252	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	3654	0.45124	2.967768	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	3654	0.460083	2.907231	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	3654	2.286777	7.419422	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	3654	3.438636	9.067975	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	3654	0.885579	3.747314	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	3654	0.788967	4.246901	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	3654	1.804731	7.772521	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6379	5599	0.000716	0.032505	-0.00219	396.000162	396.000162	396.000162	1.089681	0	1	-360	360;
	6379	5599	0.00056	0.04105	-0.005327	431.044919	431.044919	431.044919	1.113574	0	1	-360	360;
	6379	5599	0.00057	0.040211	-0.005327	431.044919	431.044919	431.044919	1.113574	0	1	-360	360;
	6803	5599	10.24442	15.278311	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3733	5599	10.4376	14.953303	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2274	5599	31.55372	31.927684	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1387	5599	0.02494	0.501983	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4429	5599	2.458264	5.693595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6100	5599	21.7686	29.266527	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7640	5599	0.260764	1.554236	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7088	5599	0.41095	2.453926	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4927	5599	1.358636	3.804546	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7950	7703	0.00445	0.11669	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2096	162	0.02812	0.471488	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1519	162	0.109781	0.670062	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3194	162	0.11731	0.676136	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6442	6996	0.000443	0.034282	-0.002535	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	4220	6996	0.000407	0.0336	-0.006729	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	81	6996	0.05005	0.961922	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	6996	20.2438	55.378103	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	6996	20.42293	51.698341	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	6996	7.97376	20.509093	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	6996	1.402252	5.36219	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	6996	4.842562	10.371072	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	6996	0.593884	2.602686	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	6996	9.390701	21.464872	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	6996	1.334504	5.801447	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	937	1.754986	5.357133	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	3117	1.766898	5.923268	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	3675	1.832964	7.959834	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	1124	1.611565	6.707825	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	394	-0.019578	0.602618	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	1744	-0.018109	0.575138	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	3024	2.305263	5.580056	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	1447	0.324709	1.981648	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	8114	11.49377	18.997923	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	1071	10.16551	21.393357	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7755	3514	0.00083	0.06663	-0.001184	433.38343	433.38343	433.38343	0.969119	0	1	-360	360;
	5971	4836	0.220599	1.426839	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8206	4836	0.215413	1.319979	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5894	4836	0.07932	0.491198	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4335	4836	0.05112	0.388202	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3912	4836	7.957025	11.110533	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5092	4836	8.10723	10.874171	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4222	4836	0.19505	0.651797	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6785	4836	0.273698	1.122666	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5971	6091	-0.016694	0.279194	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5971	6091	0.00066	0.04694	-0.001327	390.13381	390.13381	390.13381	1.076552	0	1	-360	360;
	5971	4815	0.262431	2.115858	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5971	2039	0.251911	2.002563	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5971	3961	0.086669	0.634127	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8206	6091	-0.018471	0.253657	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8206	6091	0.000299	0.012803	-0.00998	1576.19464	1576.19464	1576.19464	1.110134	0	1	-360	360;
	8206	4815	0.307389	1.800831	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8206	2039	0.307181	1.800693	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8206	3961	0.083082	0.659398	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5894	6091	2.852479	6.564463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4335	6091	0.610723	2.391322	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3912	6091	0.796095	2.828719	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5092	6091	0.778781	2.619215	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4222	6091	5.253512	8.690912	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6785	6091	0.14057	0.924359	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	381	4934	0.00012	0.01369	-0.01718	2712.78606	2712.78606	2712.78606	1	-23.00651	1	-360	360;
	5894	2024	0.00012	0.013772	-0.01718	2113.68756	2113.68756	2113.68756	1	-23.7279	1	-360	360;
	5894	4815	34.83864	37.361505	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5894	2039	31.4626	43.434901	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5894	3961	9.837258	14.382266	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1840	8106	0.12436	1.222541	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7722	8106	0.121719	1.24562	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1840	6760	0.11651	1.13624	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7722	6760	0.13205	1.113513	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3946	742	0.000447	0.03387	-0.007813	980.876065	980.876065	980.876065	0.905313	0	1	-360	360;
	3946	2487	0.188878	0.872161	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3946	6617	0.000399	0.010683	-0.016427	791.999665	791.999665	791.999665	1.202334	0	1	-360	360;
	7633	2364	0.01902	0.177292	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7633	8241	0.079771	0.509993	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7633	6638	0.00919	0.137159	0	148.80052	148.80052	148.80052	1	0	1	-360	360;
	81	4590	10.8518	37.942518	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	5156	20.41343	54.805408	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	2624	20.30263	51.706374	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	7277	0.787812	4.341828	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	236	0.590422	4.326316	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	220	0.712119	2.758379	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	973	0.891274	4.745083	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	3212	27.32618	53.325484	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	4355	0.072382	1.414889	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	6888	0.070083	1.331163	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	5823	0.02163	0.743629	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	1441	3.623961	20.259003	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	990	0.851039	4.324031	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	81	3564	1.040443	4.789612	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4335	4815	8.633058	14.86384	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3912	4815	0.05238	0.529835	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5092	4815	0.05337	0.518554	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4222	4815	56.80165	51.109504	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6785	4815	2.452686	6.967768	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4335	2039	8.052273	16.496276	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3912	2039	0.044141	0.60938	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5092	2039	0.0432	0.620868	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4222	2039	47.16736	58.301649	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6785	2039	2.566529	6.851652	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	4590	21.60319	36.731304	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	5156	38.25416	49.701504	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	2624	39.65235	49.572723	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	7277	13.37881	27.533241	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	236	11.70014	34.83726	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	220	0.03578	2.06572	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	6094	44.92382	57.412053	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	4274	5.792867	13.966068	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	6801	5.68705	12.979224	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	973	23.49307	56.524238	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	3212	0.817382	3.615513	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	4355	1.090651	5.780887	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	6888	1.04626	5.209279	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	4935	0.644758	1.803532	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	2361	0.580852	2.069945	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	5823	15.32341	41.617723	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	1801	37.75623	69.158594	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	950	1441	4.031925	12.920362	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	4590	19.41759	39.081025	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	5156	38.13227	52.87396	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	2624	39.55609	52.783939	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	7277	13.35596	29.188359	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	236	13.38227	31.89335	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	220	0.06002	1.808379	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	6094	48.46122	56.359411	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	4274	6.294598	13.758307	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	6801	6.185319	14.074795	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	973	21.89474	63.064402	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	3212	0.750554	3.860318	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	4355	1.114889	6.088781	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	6888	1.069114	5.491413	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	4935	0.653109	1.929778	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	2361	0.672888	1.912604	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	5823	18.13504	36.319253	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	1801	40.86842	68.019379	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5013	1441	4.398892	12.729917	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5388	2487	0.0384	0.48786	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	937	0.17575	1.589628	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	3117	0.16178	1.59079	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	3675	22.34695	35.318561	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	1124	20.61288	31.520779	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	394	1.071953	3.96018	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	1744	1.073546	3.773961	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	3024	1.477701	3.981233	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	1447	5.635665	13.348342	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	8114	13.41828	27.828945	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	172	1071	13.04432	28.183522	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	937	0.17575	1.589628	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	3117	0.161789	1.590785	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	3675	22.34695	35.318561	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	1124	20.61288	31.520779	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	394	1.071953	3.96018	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	1744	1.073546	3.773961	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	3024	1.477701	3.981233	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	1447	5.635665	13.348342	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	8114	13.41828	27.828945	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3041	1071	13.04432	28.183522	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	937	7.514876	17.076863	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	937	10.18967	19.679343	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	937	1.403533	4.891322	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	937	1.431054	4.791528	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	937	7.830992	14.143185	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	3117	9.209711	15.822724	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	3117	13.04525	19.283055	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	3117	1.367066	5.515496	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	3117	1.339153	5.62376	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	3117	7.23719	17.339673	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5942	1128	0.00696	0.20656	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1840	1128	0.085711	0.903781	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7722	1128	0.09713	0.885702	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4335	3961	2.771122	4.936011	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1840	6527	0.296032	1.447438	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1840	87	0.281413	1.320152	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7722	6527	0.307202	1.385873	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7722	87	0.25536	1.463643	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	985	6527	0.00046	0.03669	-0.031571	742.180099	742.180099	742.180099	0.970115	0	1	-360	360;
	6212	87	0.000808	0.04562	-0.013474	396.000162	396.000162	396.000162	0.975485	0	1	-360	360;
	6212	87	0.000438	0.03862	-0.007885	749.334508	749.334508	749.334508	0.960852	0	1	-360	360;
	2096	2641	0.037081	0.480186	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1519	2641	0.074531	0.583946	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3194	2641	0.08	0.590186	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4641	2526	0.105789	0.434903	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4641	8241	0.09937	0.415727	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1656	2526	0.095997	0.482181	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1656	8241	0.086717	0.480388	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	3675	0.873905	3.992149	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	3675	0.120211	1.173533	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	3675	7.082025	15.591937	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	3675	7.220868	15.273755	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	3675	0.872831	3.28905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	1124	0.921508	3.094834	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	1124	0.129919	0.960764	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	1124	6.142769	14.777069	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	1124	6.017355	15.066946	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	1124	0.694463	3.375206	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	394	7.254848	17.556091	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	1744	6.944598	15.819948	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	3024	0.962188	3.649169	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	1447	2.525	6.52029	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	8114	0.00154	0.39572	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4986	1071	0.0014	0.384758	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3912	3961	0.590782	1.406718	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5092	3961	0.526337	1.594252	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	4590	0.836859	5.618595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	4590	1.237665	7.884711	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	4590	3.449174	20.571071	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	4590	5.302273	16.162393	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	4590	0.789525	4.620455	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	4590	2.372521	10.507644	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1359	2624	0.000495	0.033881	-0.003463	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	2620	5156	2.23905	7.164463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	5156	3.327273	9.634503	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	5156	8.134091	27.90289	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	5156	10.55806	23.307847	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	5156	1.718843	7.201859	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	5156	4.571488	16.921905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	2624	1.894649	7.933677	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	2624	3.133678	10.68802	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	2624	8.769008	26.754135	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	2624	11.85517	21.464873	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	2624	2.005517	6.368182	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	2624	5.553718	14.366735	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	394	4.540289	12.081814	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	394	0.14887	1.547789	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	394	0.145641	1.433141	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	394	7.255165	15.148962	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	1744	4.918802	10.649998	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	1744	0.170281	1.312108	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	1744	0.17349	1.284174	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	1744	7.658058	13.540087	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6434	3026	1.803947	6.514197	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6434	5562	1.414404	5.42687	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6434	1143	0.097091	1.338712	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6434	3720	0.069813	0.61338	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6535	3026	1.656925	6.386704	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6535	5562	1.503047	5.113989	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6535	1143	0.107382	1.212396	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6535	3720	0.080478	0.53394	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	3024	1.566157	4.554339	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	3024	3.21405	6.727273	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	3024	2.86343	7.62438	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	3024	0.784091	3.66529	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	7277	-0.013061	0.68473	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	236	0.02627	0.92126	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	220	5.225831	17.344184	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	570	37.01108	55.722988	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	3111	38.53255	55.73961	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	973	0.00421	0.609432	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	3212	3.373338	12.718144	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	4355	5.841413	14.22646	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	6888	4.851247	14.847647	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	4935	15.98753	46.096945	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	2361	15.83449	47.580335	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	5823	5.167521	17.612876	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	1441	0.170402	3.131509	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	6035	13.13227	24.978533	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2620	5721	11.55748	21.738223	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	7277	0.071828	1.118006	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	236	0.01624	0.986219	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	220	1.180471	4.696329	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	973	0.01296	0.834834	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	3212	4.962535	20.238919	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	4355	0.872438	4.821676	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	6888	0.801593	4.526524	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	4935	20.20014	50.806099	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	2361	23.16066	50.423823	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	5823	0.96662	4.164474	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	1441	0.448781	4.337397	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	990	11.58657	23.771465	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	3564	13.00554	26.333102	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	6035	17.94044	35.608035	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4308	5721	15.18213	32.262468	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	1447	0.430249	2.466066	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	8114	0.0683	0.916828	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6254	1071	0.06612	0.843352	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	1447	1.444298	4.757438	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	1447	1.472603	4.660331	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	1447	2.284917	6.055578	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	444	8111	0.000504	0.031781	-0.002425	638.598472	638.598472	638.598472	1.127469	0	1	-360	360;
	444	3026	0.086683	1.502285	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	444	5562	0.05572	1.389266	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	444	1143	0.40606	1.910249	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	444	3720	0.0265	0.438137	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6845	5601	0.000669	0.032368	-0.007657	396.000162	396.000162	396.000162	1.225528	0	1	-360	360;
	8258	8114	16.48554	31.878094	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	8114	18.68347	31.239671	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	8114	-0.006983	0.380578	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8258	1071	19.27748	29.774788	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5189	1071	19.64132	29.142565	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	1071	-0.0075	0.354442	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	292	3026	0.979504	4.164463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5317	3026	0.241095	2.041488	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3691	3026	0.239525	2.084091	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	292	5562	0.733492	3.60186	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5317	5562	0.176539	1.602479	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3691	5562	0.203141	1.572934	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	7277	3.658471	11.068597	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	7277	0.10706	1.950682	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	7277	10.14339	24.185954	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	7277	0.07038	1.132707	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	7277	3.802686	12.010955	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	7277	3.583058	17.321278	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	236	2.938636	13.075416	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	236	0.00024	2.372521	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	236	9.23781	30.750005	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	236	0.041469	1.048347	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	236	3.576446	11.947316	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	236	4.138017	18.151449	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	292	2487	0.107147	0.58482	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	292	1143	7.159972	13.527699	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	292	3720	1.77874	5.115028	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	5360	7.195455	16.568181	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	5360	40.97313	63.811985	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	2560	7.423554	18.762807	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	2560	42.39256	65.737609	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	220	-0.016674	0.278698	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	973	5.349861	16.166202	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	3212	9.134349	25.279081	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	4355	0.197001	1.168768	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	6888	0.18955	1.05277	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	4935	6.030263	16.006928	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	2361	5.975346	16.528394	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	5823	3.107895	9.123269	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	1441	11.81025	37.443907	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	990	27.31856	49.591409	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	54	3564	34.37812	55.136426	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	220	0.231281	1.502872	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	220	7.048967	19.24607	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	220	0.229277	1.235537	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	220	2.544835	7.532231	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	5082	35.47521	51.481404	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	3731	34.64256	52.336781	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4222	3961	15.75277	20.093496	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5929	1776	0.000283	0.012832	-0.019399	791.999665	791.999665	791.999665	1.182689	0	1	-360	360;
	7553	1792	0.000292	0.01188	-0.01885	791.999665	791.999665	791.999665	1.182689	0	1	-360	360;
	6472	570	0.000398	0.034511	-0.005899	697.908984	697.908984	697.908984	1.031654	0	1	-360	360;
	7556	570	33.09298	47.6281	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	3111	34.44215	47.623971	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	973	0.161039	2.815236	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	3212	9.489612	27.308865	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	4355	0.246808	1.61939	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	6888	0.228082	1.521121	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	4935	18.28047	42.772855	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	2361	18.10111	44.135045	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	5823	0.331281	2.194459	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	1441	0.752078	9.020083	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	990	4.985803	12.659968	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8359	3564	5.581648	14.030467	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5883	4029	0.000476	0.03161	-0.010507	752.125188	752.125188	752.125188	0.957287	0	1	-360	360;
	5883	6518	0.000457	0.030901	-0.007587	752.125188	752.125188	752.125188	0.957287	0	1	-360	360;
	6425	7416	1.305041	5.872521	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	7416	12.81343	30.764465	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	6094	0.657355	3.229132	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	6094	6.467975	16.903509	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	4274	-0.017839	0.427271	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	6801	-0.020298	0.420457	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	3212	0.05596	1.32043	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	4935	0.430997	2.378602	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	2361	0.443573	2.357341	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	1801	0.073677	2.231025	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	1441	0.927909	4.907895	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	6035	42.82895	57.943892	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6425	5721	36.44598	52.531845	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4412	6801	0.00037	0.03817	-0.010751	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	4412	4274	0.000309	0.035711	-0.00968	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	1153	4274	0.529731	3.305372	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	6801	0.54188	3.249793	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	973	-0.008016	0.233843	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	3212	5.680748	16.603181	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	4355	6.552285	16.597649	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	6888	5.444182	17.322713	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	4935	24.32756	59.210516	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	2361	24.09903	61.117023	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	5823	5.625068	20.064407	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	1441	0.456198	4.138435	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	6035	11.52632	21.784622	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	5721	10.14543	18.959839	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7556	973	0.000476	0.031631	-0.003303	902.151896	902.151896	902.151896	0.969681	0	1	-360	360;
	2470	973	5.521694	14.730582	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	973	19.34525	41.508264	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	3212	32.73967	68.105374	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	3212	-0.083244	0.946591	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4125	4355	0.000447	0.035271	-0.002408	752.125188	752.125188	752.125188	0.957287	0	1	-360	360;
	2470	4355	-0.015103	0.337521	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	6888	-0.012686	0.286488	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	4935	27.97992	45.335871	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	2361	28.86773	44.978518	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	5823	0.994044	4.152008	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	1441	19.05609	41.630194	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	990	11.95776	23.687676	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2470	3564	11.73892	30.398894	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	4355	10.39773	25.376037	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	6888	9.457438	23.791321	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	4935	0.000409	0.036341	-0.003303	752.125188	752.125188	752.125188	0.957287	0	1	-360	360;
	1153	4935	-0.006942	0.494855	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	2361	-0.008161	0.491033	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	1801	4.701801	16.907201	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	1441	0.18313	2.973269	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	6035	30.6392	44.945983	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1153	5721	27.01524	39.081706	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5646	4674	0.000757	0.064121	-0.001346	417.676639	417.676639	417.676639	1.005563	0	1	-360	360;
	5317	1143	2.158264	7.671693	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3691	1143	2.128512	7.829339	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5317	3720	0.536593	2.637604	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3691	3720	0.529488	2.445638	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6785	6328	0.00039	0.017471	-0.005272	887.333647	887.333647	887.333647	1.049501	-5.042515	1	-360	360;
	6785	3961	0.856787	2.082133	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7427	1999	0.000515	0.03175	-0.009652	718.199994	718.199994	718.199994	1.002506	0	1	-360	360;
	7427	1999	0.000408	0.036729	-0.007329	718.199994	718.199994	718.199994	1.002506	0	1	-360	360;
	6941	3281	0.00648	0.148407	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2292	3281	0.00055	0.038349	-0.001095	412.748192	412.748192	412.748192	1.162937	0	1	-360	360;
	1516	2612	0.114331	0.894649	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7227	2612	3.931818	5.960743	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4639	2612	1.022872	2.477479	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1448	2612	1.145744	2.668595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2619	2612	0.115521	0.652645	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1516	5623	0.342335	1.036053	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	7087	0.08144	0.487603	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8036	48	0.492025	1.533306	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2196	4629	0.00079	0.058401	0	183.599776	183.599776	183.599776	1	0	1	-360	360;
	2196	4629	0.00077	0.05951	0	179.999535	179.999535	179.999535	1	0	1	-360	360;
	1516	3593	0.000209	0.016646	-0.00512	1200.00041	1200.00041	1200.00041	1.002506	0	1	-360	360;
	1516	358	-0.020062	0.263393	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1516	4523	0.02611	0.389661	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1516	1021	0.02953	0.274197	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1516	358	0.000249	0.015313	-0.00531	1200.00041	1200.00041	1200.00041	1.002506	0	1	-360	360;
	4213	1527	0.01135	0.389938	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4213	3218	0.03467	0.4241	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8036	1480	0.000159	0.018964	-0.00178	1200.00041	1200.00041	1200.00041	0.988772	0	1	-360	360;
	8036	1480	-0.01126	0.214855	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8036	6466	0.20036	1.026385	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8036	8316	0.117071	0.511302	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3490	375	0.093303	0.945014	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3976	375	0.01297	0.327971	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2166	375	0.05929	0.601787	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	4838	0.05013	0.919979	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	3796	0.0477	0.865888	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7227	3216	0.749308	2.45464	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7227	5323	2.902839	5.007549	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2635	1527	0.03282	0.444566	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7187	375	0.186479	0.852913	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1898	375	0.185979	0.818905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6730	375	1.207107	3.026033	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5619	375	0.414669	1.672355	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4639	3216	0.129612	0.934834	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4639	5323	0.796053	2.21392	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1448	3216	0.136461	1.06849	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1448	5323	0.838366	2.528809	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6912	1226	0.000298	0.01635	-0.022906	791.999665	791.999665	791.999665	1.009423	0.69199	1	-360	360;
	4210	5100	0.01604	0.374051	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1948	5100	0.01635	0.36626	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2635	5100	0.024141	0.327376	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5693	5100	0.02194	0.458409	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1821	8094	0.00046	0.03512	-0.002666	724.004478	724.004478	724.004478	0.99447	0	1	-360	360;
	1821	5973	0.000328	0.017373	-0.00429	791.999665	791.999665	791.999665	1.002506	0	1	-360	360;
	2635	3218	0.00056	0.05575	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	2414	0.09537	0.774091	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	4370	0.325393	1.201922	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5990	6898	0.367459	1.167149	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5693	2630	0.0008	0.04856	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7181	1482	0.00033	0.038031	-0.014431	725.453788	725.453788	725.453788	0.992481	0	1	-360	360;
	7181	1482	0.00034	0.037211	-0.014431	725.453788	725.453788	725.453788	0.992481	0	1	-360	360;
	2619	3216	-0.016405	0.249504	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2619	5323	0.087209	0.621302	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7914	4402	0.15538	0.761901	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6316	4402	0.15969	0.749463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6265	4402	1.428926	3.608884	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5223	4402	0.861798	2.342768	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5651	2632	0.119958	0.657978	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3445	5820	0.172789	1.348017	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2186	5820	0.00088	0.381756	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4165	5820	0.0061	0.418244	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5918	6437	0.05995	0.76688	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5416	6437	0.0653	0.750992	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2205	6437	-0.003347	0.292562	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2835	7029	1.83593	3.666115	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8311	7029	0.052409	0.497211	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1046	7355	0.436797	1.36533	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6455	7355	0.094151	0.896736	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1274	6905	0.00721	0.237881	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1274	4359	0.140762	0.856856	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7421	6905	0.05419	0.633158	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7421	4359	0.0291	0.36903	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2865	2632	0.143641	0.91531	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1103	2632	0.025969	0.445888	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1138	2632	0.270372	0.785723	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5018	2632	0.05175	0.49312	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4352	2632	0.08601	0.600496	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5470	2632	0.531591	1.712665	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3910	2632	0.43593	1.756797	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4885	3780	0.01926	0.358289	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6706	3557	0.01795	0.317397	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3445	140	0.00761	0.279277	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2186	140	0.260062	1.120476	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4165	140	0.158651	0.983822	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3445	4179	0.00395	0.366116	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	4179	0.083409	0.606219	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2186	4179	0.195669	1.120144	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4165	4179	0.104969	1.02188	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2875	4179	0.042831	0.74374	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2835	7447	0.000242	0.018374	-0.003495	1200.00041	1200.00041	1200.00041	1.009443	0.69399	1	-360	360;
	2835	5245	-0.003203	0.14994	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2835	5245	0.000451	0.034969	-0.007773	724.004478	724.004478	724.004478	0.99447	0	1	-360	360;
	2835	6345	0.113213	0.498338	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5870	4064	0.00055	0.0343	-0.007793	737.108169	737.108169	737.108169	0.97679	0	1	-360	360;
	6316	8130	0.000264	0.016372	-0.003379	1200.00041	1200.00041	1200.00041	0.988494	-1.359277	1	-360	360;
	7914	3585	0.01862	0.220201	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6316	3585	0.0168	0.250561	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1046	4359	0.202902	0.952701	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1046	7981	0.949862	3.139543	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1046	1511	0.00261	0.215422	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	3783	-0.02155	0.439897	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	3783	0.07045	1.003492	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	7835	0.00807	0.465263	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	5747	0.194723	1.065305	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	3160	0.7241	1.639128	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	4757	0.514127	1.973476	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	7889	0.06643	0.866482	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	3439	0.55937	2.088574	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	4782	0.522202	2.029502	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6748	6915	0.30876	1.923338	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2933	6754	0.06357	0.598203	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3751	6754	0.000445	0.035961	-0.008241	728.252367	728.252367	728.252367	0.988668	0	1	-360	360;
	5974	6754	0.01179	0.332211	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	723	6754	0.02131	0.372955	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4096	6754	0.00274	0.381756	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6265	3585	0.29314	1.107872	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5223	3585	0.144281	0.82905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2253	2217	0.00458	0.288905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7638	281	0.365888	1.746611	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	281	1.44438	5.252272	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	281	0.059581	0.615888	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	281	3.709298	8.934297	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	281	0.140919	0.944236	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5816	148	0.0005	0.0343	-0.003612	755.481903	755.481903	755.481903	0.953034	0	1	-360	360;
	8311	148	0.01962	0.216178	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	7835	0.398678	1.516364	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	801	0.09755	0.839897	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2875	801	0.499525	2.142562	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1205	3780	0.01537	0.348347	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3324	3780	0.01946	0.372727	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5974	5735	0.01679	0.350455	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	723	5735	0.0005	0.324174	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4096	5735	0.049469	0.557004	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1103	5349	0.0005	0.031271	0	334.799359	334.799359	334.799359	1	0	1	-360	360;
	7638	5349	0.422748	2.328099	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	5349	1.022934	4.279959	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	5349	0.256612	1.738141	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	5349	2.300207	9.353099	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	5349	0.0157	0.997789	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	7253	0.022021	0.625496	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2875	7253	0.223988	1.463946	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7638	5555	-0.037934	0.385868	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	5555	2.334711	7.437397	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	5555	0.137849	1.103905	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	5555	5.726446	13.392563	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	5555	0.183531	1.402418	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7878	7638	3.1e-05	0.002308	-0.00826	2687.34611	2687.34611	2687.34611	0.997785	-3.813636	1	-360	360;
	7638	4684	2.374861	8.881579	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7638	7372	0.158643	1.604501	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7638	7362	0.204162	1.92777	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7638	7659	0.169439	2.010873	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	4684	-0.011343	0.25657	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	4684	0.000423	0.033381	-0.024024	761.990638	761.990638	761.990638	0.944892	0	1	-360	360;
	1667	4684	6.533884	12.310123	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	4684	0.04751	0.576054	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	4684	0.21343	1.602582	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	7372	7.076177	15.615651	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	7362	6.591413	15.880881	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1982	7659	0.187507	2.209903	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	750	3775	0.0004	0.149419	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5223	3775	0.027469	0.320806	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6706	3775	0.00659	0.53155	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4097	5584	0.45281	1.969855	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8249	5584	0.561302	2.344629	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7602	5584	0.03438	0.839277	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7014	375	-0.00126	0.238202	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7014	375	0.000426	0.017203	-0.004064	938.505772	938.505772	938.505772	1.002506	0	1	-360	360;
	7014	114	0.01628	0.291863	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6327	7708	-0.010145	0.227045	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4097	3291	0.00079	0.680124	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8249	3291	-6.2e-05	0.76438	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7602	3291	0.288988	1.328967	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4097	269	0.289174	0.947893	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8249	269	0.357831	1.127893	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7602	269	0.00434	0.222459	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8244	7311	0.162281	0.824401	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1614	7311	0.510455	1.73717	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	565	7311	0.218822	1.471838	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1351	7311	0.029581	0.641756	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4290	7311	0.269194	1.398492	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	723	6248	0.01824	0.407147	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	723	6494	0.01992	0.388733	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6327	1780	0.504669	1.26314	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	528	5020	0.180042	0.913089	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	528	2025	0.28705	1.359279	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8305	4115	0.02939	0.690041	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7207	4115	0.106211	0.857975	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8305	1557	0.03057	0.640372	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7207	1557	0.11774	0.83438	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	7362	0.000416	0.031501	-0.041226	742.51248	742.51248	742.51248	0.969681	0	1	-360	360;
	1667	7372	-0.02392	0.308892	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	7362	-0.021129	0.40847	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1667	7659	0.801523	3.248477	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	7372	15.18223	29.541321	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	7372	0.810744	3.364049	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	7362	13.91921	30.6467	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	7362	0.716322	3.154132	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6706	488	0.00489	0.406095	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8244	5405	0.447909	1.098823	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8244	5988	0.01056	0.353927	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8244	1542	0.7509	1.841551	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1614	5405	0.37579	0.752493	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1614	5988	0.190388	1.261149	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1614	1542	0.05324	0.584501	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4653	5020	0.01773	0.615173	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4653	6329	0.01987	0.600173	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2058	5020	0.466918	1.296815	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2058	2025	0.815443	1.930956	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4097	3886	0.02061	0.308595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8249	3886	0.02549	0.367231	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7602	3886	0.00824	0.278492	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3260	7659	0.716621	3.572645	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2933	2965	0.045599	0.50595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8305	619	0.00887	0.706717	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8305	6673	0.04089	0.90831	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8305	4146	0.04322	0.911288	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	297	5020	0.032849	0.6625	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7058	5020	0.03726	0.637562	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	788	5020	0.074341	0.632975	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5754	5020	0.07781	0.62157	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	565	5405	0.534411	1.302908	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	565	5988	0.02355	0.61356	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	565	1542	0.420249	1.970499	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1351	5405	0.229855	0.963099	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4290	5405	0.074669	0.41343	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6327	745	0.06182	0.403213	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6227	7659	-0.029256	0.40562	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	5747	0.0033	0.455227	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2933	5747	0.00124	0.284731	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	69	6248	0.04584	0.400152	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	69	6494	0.03961	0.441877	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	2366	0.282149	1.82343	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2875	2366	-0.001467	0.288285	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7947	3160	0.000203	0.018291	-0.003547	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	5522	3160	3.246901	6.102273	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4982	5149	0.000187	0.016035	-0.003253	1200.00041	1200.00041	1200.00041	1.009443	-0.693988	1	-360	360;
	4982	5149	0.000198	0.015708	-0.003253	1200.00041	1200.00041	1200.00041	1.009443	-0.693988	1	-360	360;
	5365	5149	0.046659	0.29938	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5918	7721	0.02508	0.489778	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5918	6945	0.0359	0.39268	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5416	7721	0.01931	0.550478	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5416	6945	0.0365	0.48437	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1351	5988	-0.015393	0.242665	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1351	5988	0.00024	0.016571	-0.003032	1200.00041	1200.00041	1200.00041	1.031152	0	1	-360	360;
	4290	5988	0.778884	2.360744	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1351	1542	0.834349	2.163366	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2205	7721	0.02768	0.486653	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2205	7721	0.0005	0.0343	-0.003612	728.252367	728.252367	728.252367	0.988668	0	1	-360	360;
	2205	6945	0.170748	0.871399	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	4757	3.440496	6.966735	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	7889	-0.025	0.350744	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	3439	3.294252	7.269391	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	4782	3.075277	7.063712	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5522	6915	2.188573	6.820568	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5365	2025	0.03254	0.295443	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2875	7754	0.000216	0.016549	-0.003511	1200.00041	1200.00041	1200.00041	0.988772	0	1	-360	360;
	6455	4359	0.00914	0.394959	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7207	619	0.11943	0.839463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6706	619	0.02832	0.560682	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6706	3508	0.025289	0.560351	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6455	7981	0.203901	1.623181	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5593	7288	0.000184	0.017384	-0.004008	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	4096	6248	0.0086	0.46804	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4096	6494	0.00772	0.464148	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8311	7836	0.08261	0.429463	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8311	5068	0.09332	0.419277	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	297	6329	0.06593	0.705471	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7058	6329	0.070298	0.687417	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	8079	1374	0.000348	0.015847	-0.004506	791.999665	791.999665	791.999665	1.023085	-2.104465	1	-360	360;
	8311	4837	0.000239	0.01504	-0.003577	1200.00041	1200.00041	1200.00041	0.995521	0.68439	1	-360	360;
	7207	6673	0.0275	0.64187	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7207	4146	0.02423	0.621711	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6455	1511	0.021711	0.411839	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6455	1511	0.000216	0.018022	-0.003539	1200.00041	1200.00041	1200.00041	0.988772	0	1	-360	360;
	6455	7964	0.000173	0.017112	-0.003742	1200.00041	1200.00041	1200.00041	0.988772	0	1	-360	360;
	7076	6915	0.000179	0.016813	-0.00407	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	7076	6915	0.000209	0.016263	-0.00399	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	4290	1542	-0.024793	0.19232	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4097	1363	0.000274	0.015682	-0.036281	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	7602	1621	0.000476	0.030991	-0.015068	733.141323	733.141323	733.141323	0.982076	0	1	-360	360;
	788	2025	0.11386	1.025723	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5754	2025	0.111979	1.046363	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5754	2871	0.00022	0.018311	-0.003316	1200.00041	1200.00041	1200.00041	0.995521	-0.684388	1	-360	360;
	7965	7626	0.383282	2.185803	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7965	6163	0.497389	1.835595	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4243	6354	0.000615	0.029917	-0.000168	479.999638	479.999638	479.999638	1.193503	0	1	-360	360;
	4243	6354	0.000634	0.029307	-0.000168	479.999638	479.999638	479.999638	1.193503	0	1	-360	360;
	1738	8120	-0.02062	0.218202	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1738	2583	0.20759	1.109002	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1738	395	0.275429	1.015166	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1738	438	1.042036	2.256095	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	462	8120	0.10976	0.74593	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3086	8120	0.287831	1.163678	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7073	7431	0.07	0.46124	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6307	7431	0.068479	0.427066	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6967	6004	0.103857	0.477763	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6967	3752	1.274515	2.49626	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5867	6004	0.107943	0.458262	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5867	3752	1.323546	2.392243	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7484	7626	0.209488	1.002909	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7484	6163	0.126607	0.574751	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1995	3842	0.00044	0.03688	-0.024975	760.322154	760.322154	760.322154	0.946967	0	1	-360	360;
	1995	2194	0.01559	0.151898	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6260	2481	0.20449	0.679959	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5860	2583	0.105249	0.487299	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	462	2583	0.000246	0.018653	-0.015389	959.999933	959.999933	959.999933	0.9278	-2.469858	1	-360	360;
	462	2583	-0.01905	0.247025	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	462	395	1.418837	2.394528	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	462	438	0.205332	0.854986	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7627	2583	1.852727	2.747521	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3086	2583	1.359979	2.972315	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3086	395	0.02269	0.199771	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6260	3422	0.00688	0.14088	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	7199	6004	0.00021	0.037302	-0.015744	959.999933	959.999933	959.999933	0.884187	-6.484691	1	-360	360;
	7199	6004	0.00021	0.034539	-0.015744	959.999933	959.999933	959.999933	0.884187	-6.484691	1	-360	360;
	1459	173	0.021781	0.193349	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	3086	438	0.037729	0.222665	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5433	2194	0.05599	0.311752	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4303	5071	0.000687	0.037873	-0.00016	479.999638	479.999638	479.999638	1.027811	0	1	-360	360;
	2661	5071	0.000697	0.03705	-0.00016	479.999638	479.999638	479.999638	1.027811	0	1	-360	360;
	1459	5334	0.000258	0.015179	-0.015882	959.999933	959.999933	959.999933	1.031474	0	1	-360	360;
	1459	5334	-0.008698	0.24938	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	1459	5334	0.000218	0.015068	-0.015881	959.999933	959.999933	959.999933	1.031407	0	1	-360	360;
	7073	5334	0.055651	0.515413	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6307	5334	0.05444	0.477231	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	6260	2902	0.153248	0.626697	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2781	7023	-0.003823	0.150277	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2781	7023	0.00069	0.03981	-0.000161	417.175106	417.175106	417.175106	1.150597	0	1	-360	360;
	177	7023	0.056219	0.23343	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2099	7626	0.000243	0.023724	-0.0143	959.999933	959.999933	959.999933	1.004859	-6.027998	1	-360	360;
	2099	7626	-0.011591	0.366818	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	2099	6163	0.145519	0.713643	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	455	7626	0.000231	0.022812	-0.0143	959.999933	959.999933	959.999933	1.004859	-6.027998	1	-360	360;
	455	7626	-0.010888	0.388409	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	455	6163	0.157909	0.659799	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	5859	6815	0.00297	0.141711	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4094	2902	0.000245	0.011573	-0.015462	959.999933	959.999933	959.999933	1.108025	0	1	-360	360;
	4094	2902	0.000255	0.011329	-0.015462	959.999933	959.999933	959.999933	1.108025	0	1	-360	360;
	177	782	0.000697	0.027473	-0.00016	479.999638	479.999638	479.999638	1.193503	0	1	-360	360;
	177	782	0.000627	0.031139	-0.00016	479.999638	479.999638	479.999638	1.193503	0	1	-360	360;
	6977	3992	0.00203	0.101959	-0.00371	218.490935	218.490935	218.490935	0.939167	0	1	-360	360;
	6977	3992	0.00195	0.103917	-0.003729	214.366818	214.366818	214.366818	0.939167	0	1	-360	360;
	7115	4433	0.00341	0.25873	-0.00261	214.45827	214.45827	214.45827	0.923255	-9.72034	1	-360	360;
	5588	2290	0.001812	0.100668	-0.003932	182.124172	182.124172	182.124172	0.988338	2.88795	1	-360	360;
	5588	2290	0.001982	0.098657	-0.003952	182.124172	182.124172	182.124172	0.988338	2.88795	1	-360	360;
	4157	5175	0.00153	0.094124	-0.004491	220.971959	220.971959	220.971959	0.9775	0	1	-360	360;
	4157	5175	0.00156	0.092157	-0.004491	220.971959	220.971959	220.971959	0.9775	0	1	-360	360;
	1981	4110	0.00512	0.156934	-0.00202	109.477909	109.477909	109.477909	1.03	0	1	-360	360;
	1981	3050	0.00513	0.153636	-0.00203	111.839664	111.839664	111.839664	1.03	0	1	-360	360;
	7147	5258	0.001473	0.055275	-0.006378	240.000477	240.000477	240.000477	1.189316	0	1	-360	360;
	5084	6993	0.000628	0.026981	-0.012755	479.999638	479.999638	479.999638	1.189316	0	1	-360	360;
	6033	7949	0.000768	0.027532	-0.012755	479.999638	479.999638	479.999638	1.189316	0	1	-360	360;
	1491	3850	0.000778	0.05467	-0.012755	479.999638	479.999638	479.999638	0.850261	0	1	-360	360;
	1491	3850	0.000767	0.055473	-0.012755	479.999638	479.999638	479.999638	0.809534	0	1	-360	360;
	1784	7808	0.000905	0.055047	-0.012755	479.999638	479.999638	479.999638	0.809534	0	1	-360	360;
	6722	1889	0.000902	0.025704	-0.012755	479.999638	479.999638	479.999638	1.189316	0	1	-360	360;
	6722	1889	0.000902	0.024552	-0.012755	479.999638	479.999638	479.999638	1.189316	0	1	-360	360;
	4589	1889	0.000437	0.013336	-0.010039	360.000005	360.000005	360.000005	1.104097	0	1	-360	360;
	260	7287	0.001665	0.048767	-0.006378	295.200001	295.200001	295.200001	1.189316	0	1	-360	360;
	3841	2078	0.00083	0.05142	-0.012755	479.999638	479.999638	479.999638	0.850261	0	1	-360	360;
	3841	2078	0.000754	0.045701	-0.012755	479.999638	479.999638	479.999638	0.911396	0	1	-360	360;
	7437	6858	0.000798	0.052223	-0.012755	479.999638	479.999638	479.999638	0.864763	0	1	-360	360;
	5889	1525	0.000748	0.052717	-0.012755	533.228573	533.228573	533.228573	0.864763	0	1	-360	360;
	3204	201	0.000435	0.02158	-0.009566	360.000005	360.000005	360.000005	1.129615	0	1	-360	360;
	4640	7434	0.000871	0.0262	-0.012755	479.999638	479.999638	479.999638	1.189316	0	1	-360	360;
	5523	4187	0.000881	0.047183	-0.012755	479.999638	479.999638	479.999638	0.864763	0	1	-360	360;
	3862	2047	0.000214	0.018189	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	4289	5252	0.000544	0.028951	0	479.999638	479.999638	479.999638	1.027754	0	1	-360	360;
	4925	6736	0.000433	0.016641	0	720.000115	720.000115	720.000115	1.046093	0	1	-360	360;
	4925	6736	0.000271	0.019483	0	720.000115	720.000115	720.000115	1.033258	0	1	-360	360;
	6014	3320	0.000292	0.041885	0	450.000483	450.000483	450.000483	1.026094	0	1	-360	360;
	6014	3320	0.000283	0.025435	0	599.999547	599.999547	599.999547	1.026094	0	1	-360	360;
	6276	4511	0.00019	0.020772	0	720.000115	720.000115	720.000115	1.046072	0	1	-360	360;
	2691	8069	0.000567	0.0487	0	389.999541	389.999541	389.999541	0.988334	0	1	-360	360;
	2691	8069	0.000292	0.024683	0	599.999547	599.999547	599.999547	0.988343	0	1	-360	360;
	4132	6811	0.000238	0.020505	0	720.000115	720.000115	720.000115	0.988343	0	1	-360	360;
	4132	6811	0.000302	0.024253	0	599.999547	599.999547	599.999547	0.988324	0	1	-360	360;
	1413	2984	0.000394	0.025662	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	1413	2984	0.000403	0.026263	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	1413	2984	0.000358	0.02845	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	1092	3228	0.000184	0.020004	0	720.000115	720.000115	720.000115	1.046072	0	1	-360	360;
	6142	4372	0.000287	0.030676	0	599.999547	599.999547	599.999547	0.982307	0	1	-360	360;
	991	5845	0.000183	0.021447	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	991	5845	0.000183	0.019858	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	8310	5895	0.000196	0.019264	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	8310	5895	0.000205	0.018521	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	5160	3070	0.000205	0.018131	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	628	6342	0.000224	0.01875	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	628	6342	0.000685	0.016451	0	720.000115	720.000115	720.000115	1.043383	0	1	-360	360;
	4845	3305	0.00021	0.024641	0	720.000115	720.000115	720.000115	0.988343	0	1	-360	360;
	4845	3305	0.00021	0.025112	0	720.000115	720.000115	720.000115	0.988343	0	1	-360	360;
	7289	3662	0	0.00554	0	1523.9997	1523.9997	1523.9997	1	0	1	-360	360;
	1812	932	0	0.00471	0	900.000204	900.000204	900.000204	1	0	1	-360	360;
	431	6140	0.000205	0.016639	0	720.000115	720.000115	720.000115	1.077541	0	1	-360	360;
	431	6140	0.000214	0.017035	0	720.000115	720.000115	720.000115	1.077541	0	1	-360	360;
	431	6140	0.000214	0.016674	0	720.000115	720.000115	720.000115	1.077541	0	1	-360	360;
	1991	4702	0.000224	0.018397	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	2784	634	0.000522	0.027299	0	450.000483	450.000483	450.000483	1.026052	0	1	-360	360;
	2784	634	0.000646	0.037462	0	450.000483	450.000483	450.000483	1.026052	0	1	-360	360;
	5284	6450	0.000244	0.033945	0	381.599857	381.599857	381.599857	1.033258	0	1	-360	360;
	5284	6450	0.000196	0.023262	0	720.000115	720.000115	720.000115	0.979336	0	1	-360	360;
	4799	6450	0.00091	0.002213	0	462.000166	462.000166	462.000166	0.999807	1.12568	1	-360	360;
	1779	3930	0.000179	0.016955	0	720.000115	720.000115	720.000115	1.099594	0	1	-360	360;
	6620	386	0.000206	0.01883	0	720.000115	720.000115	720.000115	1.035987	0	1	-360	360;
	5481	7234	0.000233	0.018474	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	5481	7234	0.000421	0.031827	0	539.999921	539.999921	539.999921	1.033268	0	1	-360	360;
	966	8156	0.000219	0.016973	0	720.000115	720.000115	720.000115	1.056355	0	1	-360	360;
	7289	1601	0.000213	0.020004	0	720.000115	720.000115	720.000115	1.035987	0	1	-360	360;
	669	1267	0.000556	0.033053	0	479.999638	479.999638	479.999638	1.020472	0	1	-360	360;
	2037	7717	0.000508	0.028375	0	479.999638	479.999638	479.999638	1.037269	0	1	-360	360;
	3114	1310	0.000179	0.022003	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	2476	5559	0.000209	0.02019	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	2476	5559	0.000214	0.020311	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	6702	4930	0.000616	0.030505	0	479.999638	479.999638	479.999638	1.023738	0	1	-360	360;
	6702	4930	0.000626	0.029369	0	479.999638	479.999638	479.999638	1.023738	0	1	-360	360;
	2212	3452	0.000214	0.019409	0	720.000115	720.000115	720.000115	0.970491	0	1	-360	360;
	2212	3452	0.000192	0.022	0	720.000115	720.000115	720.000115	0.970491	0	1	-360	360;
	4764	7648	0.000531	0.03466	0	561.956125	561.956125	561.956125	1.001309	0	1	-360	360;
	5327	7383	0.000531	0.035228	0	479.999638	479.999638	479.999638	1.001309	0	1	-360	360;
	5327	7383	0.00049	0.031535	0	479.999638	479.999638	479.999638	0.998185	0	1	-360	360;
	5327	7383	0.000209	0.020964	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	6809	1126	0.000393	0.02663	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	6809	1126	0.000402	0.026088	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	2972	5804	0.000201	0.017787	0	720.000115	720.000115	720.000115	1.035987	0	1	-360	360;
	3907	722	0.000289	0.018121	0	720.000115	720.000115	720.000115	1.020734	0	1	-360	360;
	7194	4178	0.000643	0.024464	0	539.999921	539.999921	539.999921	1.056355	0	1	-360	360;
	7194	4178	0.000568	0.027771	0	539.999921	539.999921	539.999921	1.056355	0	1	-360	360;
	768	3495	0.000506	0.031833	0	479.999638	479.999638	479.999638	1.007616	0	1	-360	360;
	768	3495	0.000196	0.024118	0	720.000115	720.000115	720.000115	0.961805	0	1	-360	360;
	5370	8052	0.000205	0.022593	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	5370	8052	0.000186	0.01968	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	5370	8052	0.000205	0.01997	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	1063	2435	0.000215	0.018787	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	1063	2435	0.000215	0.018388	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	2977	5665	0.000224	0.018834	0	720.000115	720.000115	720.000115	0.997518	0	1	-360	360;
	3458	1022	0.000433	0.043546	0	360.000386	360.000386	360.000386	1.008509	0	1	-360	360;
	5338	7438	0.000234	0.018757	0	720.000115	720.000115	720.000115	1.054309	0	1	-360	360;
	5338	7438	0.000197	0.019207	0	720.000115	720.000115	720.000115	1.054309	0	1	-360	360;
	5338	7438	0.000197	0.019495	0	720.000115	720.000115	720.000115	1.054309	0	1	-360	360;
	6058	7091	0.00018	0.021226	0	720.000115	720.000115	720.000115	1.014585	0	1	-360	360;
	6058	7091	0.000812	0.03849	0	381.599857	381.599857	381.599857	1.046093	0	1	-360	360;
	2059	3829	0.000205	0.02035	0	720.000115	720.000115	720.000115	1.006865	0	1	-360	360;
	662	3392	0.000593	0.031521	0	479.999638	479.999638	479.999638	1.014004	0	1	-360	360;
	662	3392	0.000612	0.031123	0	479.999638	479.999638	479.999638	1.014004	0	1	-360	360;
	662	3392	0.000427	0.024994	0	539.999921	539.999921	539.999921	1.016389	0	1	-360	360;
	2505	373	0.000667	0.027207	0	539.999921	539.999921	539.999921	1.006865	0	1	-360	360;
	2505	373	0.000621	0.030836	0	539.999921	539.999921	539.999921	1.006865	0	1	-360	360;
	272	858	0.000192	0.020791	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	272	858	0.000192	0.02119	0	720.000115	720.000115	720.000115	1.026094	0	1	-360	360;
	7989	27	0.000686	0.030987	0	381.599857	381.599857	381.599857	1.072746	0	1	-360	360;
	2734	5146	0.000194	0.020764	0	720.000115	720.000115	720.000115	1.006865	0	1	-360	360;
	159	1691	0.000717	0.038087	0	360.000386	360.000386	360.000386	1.006865	0	1	-360	360;
	7185	6040	0.000206	0.01883	0	720.000115	720.000115	720.000115	1.035987	0	1	-360	360;
	7185	67	0.000206	0.018439	0	720.000115	720.000115	720.000115	1.035987	0	1	-360	360;
	6374	28	0.000219	0.017019	0	720.000115	720.000115	720.000115	1.066843	0	1	-360	360;
	6374	28	0.000224	0.016641	0	720.000115	720.000115	720.000115	1.066843	0	1	-360	360;
	5992	8208	0.000676	0.047085	0	389.999541	389.999541	389.999541	1.016389	0	1	-360	360;
	5992	8208	0.000667	0.048014	0	389.999541	389.999541	389.999541	1.016389	0	1	-360	360;
	6658	7558	0.000196	0.019215	0	720.000115	720.000115	720.000115	1.077541	0	1	-360	360;
	3614	5343	0.000253	0.025528	0	599.999547	599.999547	599.999547	0.997508	0	1	-360	360;
	7441	6291	0.00021	0.017884	0	720.000115	720.000115	720.000115	1.046072	0	1	-360	360;
	7441	6291	0.00021	0.017518	0	720.000115	720.000115	720.000115	1.046072	0	1	-360	360;
	1566	5935	0.000219	0.017358	0	720.000115	720.000115	720.000115	1.056355	0	1	-360	360;
	1566	5935	0.000219	0.016973	0	720.000115	720.000115	720.000115	1.056355	0	1	-360	360;
	3223	785	0.000295	0.027995	0	599.999547	599.999547	599.999547	0.979536	0	1	-360	360;
	3223	785	0.000277	0.027486	0	599.999547	599.999547	599.999547	0.979307	0	1	-360	360;
	3330	7471	0.000489	0.032074	0	479.999638	479.999638	479.999638	1.020618	0	1	-360	360;
	5528	2828	0.00018	0.02129	0	720.000115	720.000115	720.000115	1.033258	0	1	-360	360;
	980	1618	0.000366	0.029147	0	599.999547	599.999547	599.999547	0.988334	0	1	-360	360;
	980	1618	0.000366	0.028067	0	599.999547	599.999547	599.999547	0.993101	0	1	-360	360;
	980	1618	0.00042	0.026331	0	599.999547	599.999547	599.999547	1.013197	0	1	-360	360;
	4510	1379	0.000441	0.01916	0	720.000115	720.000115	720.000115	1.022456	0	1	-360	360;
	6734	5566	0.000391	0.024766	0	539.999921	539.999921	539.999921	1.035987	0	1	-360	360;
	6734	5566	0.000401	0.024225	0	539.999921	539.999921	539.999921	1.035987	0	1	-360	360;
	6734	4226	0.000408	0.028237	0	539.999921	539.999921	539.999921	1.026094	0	1	-360	360;
	6320	4209	0.000199	0.018609	0	720.000115	720.000115	720.000115	1.066843	0	1	-360	360;
	2448	8128	0.000196	0.021597	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	2448	8128	0.000186	0.022003	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	2580	2991	0.000186	0.019718	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	4050	1325	0.000205	0.022703	0	720.000115	720.000115	720.000115	0.953272	0	1	-360	360;
	738	7092	0.000228	0.021908	0	720.000115	720.000115	720.000115	1.006865	0	1	-360	360;
	6740	6923	0.000322	0.028734	0	599.999547	599.999547	599.999547	1.04286	0	1	-360	360;
	6740	8262	0.000319	0.026574	0	599.999547	599.999547	599.999547	1.072735	0	1	-360	360;
	3264	1127	0.000289	0.018298	0	720.000115	720.000115	720.000115	1.008509	0	1	-360	360;
	3264	1127	0.000199	0.02111	0	720.000115	720.000115	720.000115	1.008509	0	1	-360	360;
	6974	7526	0.000196	0.02119	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	6974	7526	0.000196	0.021597	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	4103	1369	0.001097	0.082091	0	240.000477	240.000477	240.000477	1.052931	0	1	-360	360;
	4103	1369	0.000406	0.048834	0	360.000386	360.000386	360.000386	1.052632	0	1	-360	360;
	4103	1369	7.2e-05	0.017798	0	720.000115	720.000115	720.000115	1.056355	0	1	-360	360;
	4103	1369	0.000424	0.04684	0	360.000386	360.000386	360.000386	1.052632	0	1	-360	360;
	7487	3716	0.000224	0.01875	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	7487	3716	0.000224	0.018334	0	720.000115	720.000115	720.000115	1.016389	0	1	-360	360;
	1165	3896	0.000379	0.018492	0	720.000115	720.000115	720.000115	1.052632	0	1	-360	360;
	1165	3896	0.000253	0.019711	0	720.000115	720.000115	720.000115	1.052632	0	1	-360	360;
	346	8199	0.000196	0.016483	0	1080.0005	1080.0005	1080.0005	1.006865	0	1	-360	360;
	346	8199	0.000186	0.022421	0	720.000115	720.000115	720.000115	1.006865	0	1	-360	360;
	6139	3729	0.00046	0.030794	0	479.999638	479.999638	479.999638	1.008509	0	1	-360	360;
	8352	2202	0.000695	0.049098	-0.0077	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	8352	2202	0.000355	0.024893	-0.01105	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	105	2202	0.000365	0.024372	-0.011029	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2861	7048	0.000411	0.023652	-0.01075	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6588	4474	0.000386	0.025673	-0.001838	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6588	4474	0.000344	0.029336	-0.002013	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	759	4102	0.000653	0.052104	-0.0077	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	759	4102	0.000767	0.043531	-0.00182	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3130	5549	0.000323	0.028314	-0.002516	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3130	5549	0.000323	0.025999	-0.002352	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3130	6585	0.000365	0.025478	-0.002932	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	806	5720	0.00086	0.039291	-0.00222	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	653	1561	0.000355	0.026586	-0.012724	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	653	1561	0.000375	0.026313	-0.001904	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5631	2946	0.000344	0.023329	-0.001411	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5631	2946	0.000302	0.026375	-0.001379	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	8325	2653	0.000788	0.04323	-0.00201	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	8325	2653	0.000767	0.04407	-0.00201	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1233	5311	0.000767	0.043562	-0.00157	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1233	5311	0.000788	0.041281	-0.00138	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1233	5311	0.000892	0.040524	-0.00123	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2916	4256	0.000975	0.043562	-0.00821	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2916	4256	0.000757	0.042649	-0.00905	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	4860	2145	0.000261	0.030762	-0.010678	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4860	2145	0.000261	0.030088	-0.010678	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4563	5547	0.000323	0.026447	-0.010864	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4563	8041	0.000323	0.026885	-0.010722	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6774	495	0.000323	0.02748	-0.013074	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6774	495	0.000292	0.028158	-0.010678	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	495	4550	0	0.008286	0	668.077215	668.077215	668.077215	1	0.69159	1	-360	360;
	4981	3474	0.000736	0.044526	-0.009	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2585	8268	0.00086	0.039488	-0.00194	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2647	5925	0.000365	0.024372	-0.009573	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2647	5925	0.000375	0.023986	-0.013775	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4090	6603	0.000386	0.025673	-0.001816	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	2408	1623	0.000809	0.042318	-0.00206	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3199	5119	0.000705	0.042038	-0.00174	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3222	330	0.000334	0.028022	-0.002615	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3222	1676	0.000353	0.022403	-0.00391	720.000115	720.000115	720.000115	0.982143	0	1	-360	360;
	2155	330	0.000313	0.026354	-0.002516	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2155	330	0.000323	0.025478	-0.00244	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1717	7032	0.000344	0.024622	-0.001543	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	101	7032	0.000355	0.024038	-0.001685	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3180	6607	0.000881	0.038244	-0.00201	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3180	6607	0.000757	0.040959	-0.00807	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2803	1900	0.000354	0.028765	-0.00951	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5648	1900	0.000354	0.027114	-0.01203	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1980	1900	0.000695	0.04748	-0.00661	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	4205	1467	0.000323	0.028314	-0.002309	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4205	1467	0.000323	0.026354	-0.001751	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4205	1467	0.000323	0.025696	-0.001707	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5076	910	0.000355	0.027305	-0.001783	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5076	910	0.000355	0.026666	-0.009179	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5076	910	0.000923	0.037456	-0.00171	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5076	910	0.000788	0.035217	-0.00325	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1090	6776	0.000323	0.026739	-0.001597	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1090	6776	0.000323	0.02749	-0.001433	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2248	4626	0.000313	0.027324	-0.013217	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2248	4626	0.000313	0.028074	-0.011357	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7103	5015	0.000601	0.024984	-0.00685	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7103	5015	0.000715	0.022797	-0.00574	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7103	5015	0.000344	0.024893	-0.009486	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4522	7218	0.000365	0.024163	-0.002298	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4522	7218	0.000344	0.02385	-0.00895	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4555	6679	0.000757	0.040182	-0.00771	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	686	3421	0.000323	0.028937	-0.010525	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	3362	6303	0.000354	0.026958	-0.00928	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3362	6303	0.000778	0.045656	-0.00178	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3362	6303	0.000695	0.048932	-0.00786	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3362	6303	0.000302	0.025926	-0.010438	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6112	7414	0.000355	0.027328	-0.002484	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6112	7414	0.000365	0.027156	-0.011182	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	8369	5382	0.000355	0.024299	-0.012538	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7245	729	0.000375	0.026313	-0.002538	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	7245	729	0.000375	0.025605	-0.012112	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	2945	1403	0.000334	0.026291	-0.001707	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2945	1403	0.000705	0.046735	-0.00703	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1878	3719	0.000313	0.026875	-0.011794	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1878	3719	0.000302	0.027376	-0.011794	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4184	4648	0.000664	0.044868	-0.00859	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1941	4648	0.000809	0.044526	-0.00642	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5135	5038	0.00086	0.034253	-0.00325	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5135	5038	0.000355	0.024372	-0.009311	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2579	1391	0.000365	0.023788	-0.009683	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4131	6767	0.000375	0.023204	-0.010799	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4131	6767	0.000302	0.026593	-0.010678	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1603	3511	0.000323	0.029256	-0.001958	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1603	3511	0.000334	0.029986	-0.003195	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1603	3511	0.000355	0.030979	-0.002341	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	2280	4134	0.000302	0.025707	-0.01267	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2280	4134	0.000344	0.025405	-0.011247	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2791	5550	0.00012	0.010515	0	603.600307	603.600307	603.600307	1	6.90708	1	-360	360;
	4610	6934	0.000365	0.023715	-0.010131	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5478	7864	0.000365	0.023517	-0.001411	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5478	5012	0.000334	0.026812	-0.001882	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7692	7976	0.000695	0.047253	-0.00681	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6830	7976	0.000778	0.043759	-0.00215	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6823	6868	0.000313	0.030636	-0.002626	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6823	6868	0.000313	0.028286	-0.003096	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6823	6868	0.000344	0.028115	-0.001521	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	2286	1349	0.000902	0.039425	-0.00145	599.999547	599.999547	599.999547	0.982143	0	1	-360	360;
	2149	7539	0.000355	0.024237	-0.010164	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7866	1701	0.000375	0.023986	-0.001915	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7866	1701	0.000375	0.023861	-0.001969	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1360	527	0.000819	0.04153	-0.00145	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1360	527	0.000809	0.042701	-0.00271	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1368	2531	0.000334	0.028022	-0.002407	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7429	844	0.000632	0.048237	-0.00767	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7429	844	0.000632	0.045159	-0.00813	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7429	844	0.000715	0.044039	-0.00692	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3569	6170	0.00074	0.044479	-0.00682	360.000386	360.000386	360.000386	0.986547	0	1	-360	360;
	3569	3667	0.00074	0.043811	-0.00745	360.000386	360.000386	360.000386	0.986547	0	1	-360	360;
	4039	1897	0.000375	0.02385	-0.014836	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4039	1897	0.000365	0.023267	-0.010569	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4039	1897	0.000323	0.026739	-0.000952	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	986	6674	0.000684	0.047025	-0.008	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	8259	1983	0.000313	0.02748	-0.0093	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5943	7681	0.000323	0.028554	-0.002407	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	5943	7681	0.000302	0.026437	-0.002418	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6400	6874	0.000365	0.02555	-0.001794	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6400	6874	0.000344	0.024893	-0.009694	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6400	8220	0.000375	0.024508	-0.001794	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6251	6154	0.000778	0.041157	-0.00807	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6251	6154	0.000923	0.03702	-0.00252	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6251	6154	0.000302	0.026593	-0.010678	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	935	6483	0.000344	0.029667	-0.001926	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	935	37	0.000334	0.02748	-0.002134	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4822	6483	0.000323	0.030636	-0.002134	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4822	37	0.000323	0.028526	-0.001926	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	8361	5654	0.000344	0.025332	-0.009595	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	8361	7517	0.000334	0.024674	-0.012145	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	8361	6207	0.000344	0.02457	-0.00139	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3283	6044	0.000365	0.023788	-0.013709	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3951	5958	0.000365	0.023601	-0.001313	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3951	5958	0.000344	0.02652	-0.002615	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6814	5958	0.000344	0.027272	-0.002385	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6814	5958	0.000334	0.027563	-0.002637	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	8068	6021	0.000313	0.030636	-0.010678	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	8068	6021	0.000313	0.028286	-0.010646	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4955	4381	0.000881	0.036803	-0.00289	431.999937	431.999937	431.999937	0.982143	0	1	-360	360;
	4955	4381	0.000902	0.036668	-0.00288	433.200456	433.200456	433.200456	0.982143	0	1	-360	360;
	4955	4381	0.000881	0.034014	-0.00297	466.799852	466.799852	466.799852	0.982143	0	1	-360	360;
	4954	3446	0.000365	0.026027	-0.013534	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4954	3446	0.000386	0.025673	-0.001619	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6578	4087	0.00012	0.017443	0	699.564236	699.564236	699.564236	1	-11.03971	1	-360	360;
	5582	49	0.00017	0.00629	0	1341.59979	1341.59979	1341.59979	1	-3.738837	1	-360	360;
	5582	6175	0.000664	0.047107	-0.00615	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	1332	1147	0.000313	0.028314	-0.001718	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1332	1147	0.000334	0.025707	-0.00233	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1332	1415	0.000344	0.025478	-0.001871	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6837	1415	0.000344	0.027236	-0.01198	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6837	1415	0.000375	0.026518	-0.002593	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6261	4436	0.000386	0.02646	-0.001422	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4765	3	0.000788	0.040317	-0.00684	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6821	506	0.000323	0.025863	-0.011794	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	679	3936	1e-05	0.041006	0	1053.59965	1053.59965	1053.59965	1	-42.80168	1	-360	360;
	3770	6463	0.000334	0.030306	-0.012166	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	3770	6463	0.000302	0.030636	-0.012615	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	6769	3319	0.00011	0.015095	0	590.399933	590.399933	590.399933	1	8.42776	1	-360	360;
	7453	5944	0.000344	0.024841	-0.011783	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7453	5944	0.000365	0.025029	-0.014201	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7453	5944	0.000355	0.024101	-0.012112	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3755	144	0.000344	0.02385	-0.012177	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3755	144	0.000411	0.02313	-0.01075	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3755	144	0.000829	0.042764	-0.00201	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5297	8365	0.000705	0.04748	-0.00871	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5297	1316	0.000323	0.029986	-0.001991	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5297	1316	0.000302	0.029952	-0.002319	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	7493	4873	0.000746	0.041064	-0.00233	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7493	4873	0.000881	0.042214	-0.00159	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2143	7261	0.00086	0.041209	-0.00184	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	2143	7261	0.000912	0.038669	-0.0021	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7352	151	0.000375	0.026244	-0.001772	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	7352	151	0.000375	0.025457	-0.010011	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	3166	1052	0.000302	0.029096	-0.001346	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	4305	3140	0.000302	0.029667	-0.010678	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1559	6987	0.000334	0.027563	-0.011915	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1559	6987	0.000323	0.028074	-0.013392	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	818	7423	0.000736	0.041064	-0.00163	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7554	7423	0.001234	0.040109	-0.00191	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7554	7423	0.000757	0.043209	-0.00937	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	7254	6108	0.000323	0.026586	-0.010394	957.600212	957.600212	957.600212	0.93617	0	1	-360	360;
	3823	255	0.000788	0.040296	-0.00613	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3823	255	0.000912	0.040514	-0.00706	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5874	3183	0.000798	0.042183	-0.00162	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	5874	3183	0.000778	0.043147	-0.00187	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	6644	3857	0.000323	0.027793	-0.002155	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6644	3857	0.000302	0.027918	-0.012626	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6644	3857	0.000302	0.025926	-0.013162	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6279	2679	0.000344	0.024956	-0.012068	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6279	2679	0.000355	0.024435	-0.011291	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3485	278	0.000386	0.026392	-0.001718	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	5126	278	0.000365	0.025525	-0.009256	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1580	102	0.000344	0.029986	-0.002451	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1580	102	0.000323	0.029838	-0.001827	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	2467	5193	0.000323	0.02748	-0.013074	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2467	5193	0.000313	0.028001	-0.013074	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	389	2952	0.000179	0.012988	-0.025464	1282.79937	1282.79937	1282.79937	1.03004	0	1	-360	360;
	389	8027	0.000198	0.012724	-0.025464	1282.79937	1282.79937	1282.79937	1.03004	0	1	-360	360;
	389	2174	0.000207	0.01247	-0.025464	1282.79937	1282.79937	1282.79937	1.03004	0	1	-360	360;
	389	1733	0.000207	0.012206	-0.025464	1282.79937	1282.79937	1282.79937	1.03004	0	1	-360	360;
	389	3100	0.000217	0.012912	-0.025464	1282.79937	1282.79937	1282.79937	0.990991	0	1	-360	360;
	389	5073	0.000217	0.012627	-0.025464	1282.79937	1282.79937	1282.79937	0.990991	0	1	-360	360;
	389	2831	0.000198	0.014317	-0.025464	1282.79937	1282.79937	1282.79937	0.990991	0	1	-360	360;
	389	2831	0.000189	0.014592	-0.025464	1282.79937	1282.79937	1282.79937	0.990991	0	1	-360	360;
	389	7128	0.000354	0.02725	-0.01155	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	871	7702	0.000628	0.042428	-0.00827	360.000386	360.000386	360.000386	1.009174	0	1	-360	360;
	871	7702	0.000726	0.044039	-0.00722	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3254	5358	0.000365	0.024893	-0.001586	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3254	5358	0.000438	0.02457	-0.001958	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3254	5358	0.000365	0.023986	-0.009486	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6497	5369	0.000375	0.023267	-0.014431	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	6497	5369	0.000323	0.026375	-0.011674	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1646	2237	0.000334	0.029735	-0.012713	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1646	2237	0.000313	0.030237	-0.012079	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	3038	1944	0.000767	0.044049	-0.00236	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3038	1944	0.000302	0.026072	-0.002188	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3038	1944	0.00085	0.039892	-0.00134	360.000386	360.000386	360.000386	0.982143	0	1	-360	360;
	3164	8166	0.000344	0.024956	-0.013447	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	8067	8166	0.000334	0.026666	-0.01221	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1908	730	0.000365	0.023788	-0.010153	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1908	730	0.000365	0.023465	-0.001663	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	200	1405	0.000334	0.026447	-0.011029	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	4888	0.000323	0.027042	-0.011773	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	4888	0.000313	0.02748	-0.008797	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	4888	0.000302	0.028314	-0.001893	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	6808	0.000323	0.025853	-0.002254	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	6808	0.000365	0.026197	-0.002155	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	7566	6808	0.000365	0.025165	-0.002221	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3644	4415	0.000355	0.024237	-0.010274	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	3644	1836	0.000375	0.024247	-0.001608	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1987	4415	0.000365	0.023267	-0.009486	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1987	1836	0.000342	0.023139	-0.01616	720.000115	720.000115	720.000115	0.982143	0	1	-360	360;
	1496	6438	0.000344	0.026666	-0.002287	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	1496	623	0.000334	0.027021	-0.002254	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	734	6438	0.000302	0.027615	-0.002418	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	734	623	0.000313	0.026135	-0.002221	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4180	7987	0.000365	0.025905	-0.002626	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	4180	7987	0.000355	0.025238	-0.002232	720.000115	720.000115	720.000115	0.979227	0	1	-360	360;
	2648	3839	0.000334	0.026814	-0.010678	720.000115	720.000115	720.000115	0.93617	0	1	-360	360;
	1040	5965	0.001242	0.083148	-0.00264	409.199948	409.199948	409.199948	0.945477	0	1	-360	360;
	1040	3766	0.0017	0.090508	-0.00122	374.400033	374.400033	374.400033	0.945477	0	1	-360	360;
	1765	3133	0.001648	0.071935	-0.00389	211.199867	211.199867	211.199867	1.018676	0	1	-360	360;
	1765	5847	0.001742	0.073131	-0.00388	119.999909	119.999909	119.999909	0.999334	0	1	-360	360;
	146	7173	0.000914	0.088676	-0.0012	202.800183	202.800183	202.800183	0.986842	0	1	-360	360;
	146	4672	0.000955	0.085404	-0.00187	265.200188	265.200188	265.200188	0.986842	0	1	-360	360;
	1634	3531	0.001479	0.088626	-0.00123	235.199717	235.199717	235.199717	0.986842	0	1	-360	360;
	1634	3835	0.00153	0.084928	-0.00123	274.800391	274.800391	274.800391	0.986842	0	1	-360	360;
	6552	5556	0.00151	0.102577	-0.00122	178.799674	178.799674	178.799674	0.945477	0	1	-360	360;
	7406	2288	0.000884	0.108054	-0.00173	163.200166	163.200166	163.200166	0.945477	0	1	-360	360;
	7564	5041	0.001099	0.069265	-0.00245	406.799568	406.799568	406.799568	1.018676	0	1	-360	360;
	11	2654	0.000904	0.088121	-0.00177	249.600022	249.600022	249.600022	0.986842	0	1	-360	360;
	3958	2590	0.001386	0.094158	-0.00123	391.20006	391.20006	391.20006	0.986842	0	1	-360	360;
	3958	2590	0.001355	0.096007	-0.00123	384.000237	384.000237	384.000237	0.986842	0	1	-360	360;
	4534	1061	0.001456	0.098925	-0.00123	266.400049	266.400049	266.400049	0.962773	0	1	-360	360;
	4534	1061	0.001424	0.100867	-0.00123	261.599947	261.599947	261.599947	0.962773	0	1	-360	360;
	1802	1628	0.001442	0.086424	-0.00122	119.999909	119.999909	119.999909	0.999334	0	1	-360	360;
	1031	7653	0.001317	0.070127	-0.00122	119.999909	119.999909	119.999909	1.074114	0	1	-360	360;
	547	2373	0.000991	0.083602	-0.00112	280.799696	280.799696	280.799696	0.999334	0	1	-360	360;
	547	2373	0.001272	0.097228	-0.00122	242.399541	242.399541	242.399541	0.999334	0	1	-360	360;
	547	5262	0.000968	0.088836	-0.0012	227.999894	227.999894	227.999894	0.974659	0	1	-360	360;
	547	5262	0.001537	0.08896	-0.00123	227.999894	227.999894	227.999894	0.974659	0	1	-360	360;
	547	5262	0.0018	0.078579	-0.0039	258.000365	258.000365	258.000365	0.974659	0	1	-360	360;
	392	2334	0.001071	0.079233	-0.00234	292.800279	292.800279	292.800279	0.999334	0	1	-360	360;
	5337	3562	0.001301	0.088365	-0.00123	119.999909	119.999909	119.999909	1.018676	0	1	-360	360;
	5337	4362	0.001436	0.082886	-0.00389	212.400386	212.400386	212.400386	1.018676	0	1	-360	360;
	5337	3304	0.001701	0.102782	-0.0039	370.799792	370.799792	370.799792	0.923361	0	1	-360	360;
	8026	7940	0.001352	0.091819	-0.00122	188.399878	188.399878	188.399878	0.999334	0	1	-360	360;
	1397	1796	0.001737	0.085791	-0.00391	266.400049	266.400049	266.400049	0.962773	0	1	-360	360;
	1397	2879	0.001331	0.087657	-0.00128	304.800204	304.800204	304.800204	0.945477	0	1	-360	360;
	1397	1796	0.001877	0.07879	-0.00391	290.399899	290.399899	290.399899	0.962773	0	1	-360	360;
	1397	4176	0.001672	0.089295	-0.00391	119.999909	119.999909	119.999909	0.962773	0	1	-360	360;
	1397	2879	0.001421	0.10862	-0.00122	245.999781	245.999781	245.999781	0.945477	0	1	-360	360;
	1397	4176	0.001338	0.09699	-0.00123	119.999909	119.999909	119.999909	0.962773	0	1	-360	360;
	2780	4987	0.001224	0.093571	-0.00123	260.399428	260.399428	260.399428	1.018676	0	1	-360	360;
	1826	2306	0.001667	0.092522	-0.00122	411.600328	411.600328	411.600328	0.945477	0	1	-360	360;
	1826	977	0.0017	0.090508	-0.00122	421.199873	421.199873	421.199873	0.945477	0	1	-360	360;
	2600	2031	0.001177	0.089984	-0.00123	218.39969	218.39969	218.39969	1.038781	0	1	-360	360;
	5512	8252	0.000975	0.083762	-0.00187	388.79968	388.79968	388.79968	0.986842	0	1	-360	360;
	5512	2083	0.001221	0.074424	-0.00112	469.200232	469.200232	469.200232	0.974659	0	1	-360	360;
	4752	7560	0.001205	0.088315	-0.00123	344.399562	344.399562	344.399562	1.038781	0	1	-360	360;
	2283	7600	0.000951	0.081618	-0.00112	289.200038	289.200038	289.200038	0.999334	0	1	-360	360;
	2283	5400	0.0011	0.085679	-0.00113	378.000274	378.000274	378.000274	0.962773	0	1	-360	360;
	2283	5400	0.000895	0.099866	-0.00129	324.000611	324.000611	324.000611	0.962773	0	1	-360	360;
	4086	2561	0	0.064	0	112.941283	112.941283	112.941283	1.0625	0	1	-360	360;
	4086	2561	0	0.06267	0	112.941283	112.941283	112.941283	1.0625	0	1	-360	360;
	4086	2561	0	0.06134	0	112.941283	112.941283	112.941283	1.0625	0	1	-360	360;
	1435	4177	0	0.06295	0	111.62818	111.62818	111.62818	1.075	0	1	-360	360;
	1435	4177	0	0.07135	0	111.62818	111.62818	111.62818	1.075	0	1	-360	360;
	2919	1829	0	0.04729	0	286.323795	286.323795	286.323795	0.972326	0	1	-360	360;
	2919	1829	0	0.0482	0	281.38745	281.38745	281.38745	0.972326	0	1	-360	360;
	2919	1829	0	0.04911	0	276.451105	276.451105	276.451105	0.972326	0	1	-360	360;
	3063	5194	0	0.06744	0	125.454622	125.454622	125.454622	0.956522	0	1	-360	360;
	6946	668	0	0.145661	0	114.049761	114.049761	114.049761	1.052174	0	1	-360	360;
	8149	747	0	0.03024	0	149.339568	149.339568	149.339568	1.052632	0	1	-360	360;
	8149	747	0	0.029729	0	151.62016	151.62016	151.62016	1.052632	0	1	-360	360;
	747	6451	0	0.07191	0	225.802926	225.802926	225.802926	0.8875	0	1	-360	360;
	747	6451	0	0.0815	0	198.760486	198.760486	198.760486	0.8875	0	1	-360	360;
	8238	2552	0	0.07843	0	133.818315	133.818315	133.818315	0.896739	0	1	-360	360;
	2266	4755	0	0.03389	0	369.093792	369.093792	369.093792	1.052632	0	1	-360	360;
	2266	4755	0	0.035419	0	353.161903	353.161903	353.161903	1.052632	0	1	-360	360;
	4755	7183	0	0.07323	0	118.074804	118.074804	118.074804	1.016304	0	1	-360	360;
	4755	7183	0	0.07177	0	118.074804	118.074804	118.074804	1.016304	0	1	-360	360;
	4992	116	0	0.06789	0	127.042844	127.042844	127.042844	0.944565	0	1	-360	360;
	4992	116	0	0.06524	0	127.042844	127.042844	127.042844	0.944565	0	1	-360	360;
	4992	116	0	0.06385	0	127.042844	127.042844	127.042844	0.944565	0	1	-360	360;
	2401	5072	0	0.04033	0	167.653408	167.653408	167.653408	1.052174	0	1	-360	360;
	4693	3493	0	0.07024	0	114.049761	114.049761	114.049761	1.052173	0	1	-360	360;
	4693	3493	0	0.07161	0	114.049761	114.049761	114.049761	1.052173	0	1	-360	360;
	5608	778	0	0.06784	0	118.588842	118.588842	118.588842	1.0119	0	1	-360	360;
	2910	7994	0	0.071916	0	119.999875	119.999875	119.999875	0.928269	0	1	-360	360;
	6356	4169	0	0.07323	0	119.480502	119.480502	119.480502	1.004348	0	1	-360	360;
	6356	4169	0	0.07177	0	119.480502	119.480502	119.480502	1.004348	0	1	-360	360;
	6424	1764	0	0.0447	0	114.049968	114.049968	114.049968	1.052174	0	1	-360	360;
	6424	1764	0	0.04377	0	114.049968	114.049968	114.049968	1.052174	0	1	-360	360;
	6424	1764	0	0.04123	0	114.049968	114.049968	114.049968	1.052174	0	1	-360	360;
	2170	7259	0	0.06295	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	2170	7259	0	0.07135	0	119.999875	119.999875	119.999875	1	0	1	-360	360;
	2249	161	0	0.07014	0	127.042844	127.042844	127.042844	0.944565	0	1	-360	360;
	2249	161	0	0.07149	0	127.042844	127.042844	127.042844	0.944565	0	1	-360	360;
	1301	4023	0	0.034531	0	113.683363	113.683363	113.683363	1.055562	0	1	-360	360;
	1301	6219	0	0.04656	0	239.317287	239.317287	239.317287	1.007866	0	1	-360	360;
	1301	6219	0	0.04563	0	244.079872	244.079872	244.079872	1.007866	0	1	-360	360;
	1654	7628	0.000229	0.01234	-0.016755	1492.67403	1492.67403	1492.67403	0.884318	0	1	-360	360;
	1654	7628	0.000229	0.0122	-0.016755	1518.19363	1518.19363	1518.19363	0.869455	0	1	-360	360;
	7729	2571	0.001485	0.065805	-0.00413	300.000103	300.000103	300.000103	0.956938	0	1	-360	360;
	7729	2571	0.001583	0.06586	-0.00413	300.000103	300.000103	300.000103	0.956938	0	1	-360	360;
	4802	2571	0.00175	0.070701	-0.002893	201.599698	201.599698	201.599698	0.952381	0	1	-360	360;
	4802	2571	0.00169	0.0715	-0.002893	201.599698	201.599698	201.599698	0.952381	0	1	-360	360;
	6116	5994	0.001228	0.060916	-0.004443	300.000103	300.000103	300.000103	1.072151	0	1	-360	360;
	6116	5994	0.00116	0.063798	-0.00194	300.000103	300.000103	300.000103	1.062377	0	1	-360	360;
	6116	5994	0.00113	0.059295	-0.00194	300.000103	300.000103	300.000103	1.062171	0	1	-360	360;
	4547	5994	0.00177	0.06853	-0.002893	326.417584	326.417584	326.417584	0.838191	0	1	-360	360;
	6761	1979	0.001363	0.063835	-0.004335	300.000103	300.000103	300.000103	0.992489	0	1	-360	360;
	6761	1979	0.001457	0.067155	-0.004335	300.000103	300.000103	300.000103	0.968502	0	1	-360	360;
	6966	1979	0.00193	0.06377	-0.002893	195.216329	195.216329	195.216329	0.983524	0	1	-360	360;
	6966	1979	0.00194	0.06188	-0.002893	195.216329	195.216329	195.216329	0.983524	0	1	-360	360;
	7904	4469	0.00045	0.02807	-0.009065	598.499556	598.499556	598.499556	1.002506	0	1	-360	360;
	7904	1953	0.00042	0.028021	-0.000884	598.499556	598.499556	598.499556	1.002506	0	1	-360	360;
	7780	5102	0.001208	0.06631	-0.004443	300.000103	300.000103	300.000103	1.030783	0	1	-360	360;
	7780	5102	0.001228	0.066264	-0.004443	300.000103	300.000103	300.000103	1.030783	0	1	-360	360;
	5306	2589	0.00047	0.026469	-0.009065	598.499556	598.499556	598.499556	1.002506	0	1	-360	360;
	8119	6487	0.00186	0.06826	-0.002893	229.064724	229.064724	229.064724	0.838191	0	1	-360	360;
	514	130	0.001431	0.068667	-0.00413	300.000103	300.000103	300.000103	0.956938	0	1	-360	360;
	514	130	0.001529	0.068787	-0.00413	300.000103	300.000103	300.000103	0.956938	0	1	-360	360;
	2951	130	0.00193	0.06377	-0.002893	201.599698	201.599698	201.599698	0.952381	0	1	-360	360;
	2951	130	0.00194	0.06188	-0.002893	201.599698	201.599698	201.599698	0.952381	0	1	-360	360;
	3052	4670	0	0.00898	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	2336	5745	0	0.00433	-0.00549	180.000193	180.000193	180.000193	1	0	1	-360	360;
	5745	4384	0.007955	0.151636	-0.004741	180.000193	180.000193	180.000193	0.929293	0	1	-360	360;
	6388	363	0.000271	0.030694	-0.016	479.999638	479.999638	479.999638	1.052632	0	1	-360	360;
	6388	363	0.000271	0.02842	-0.016	479.999638	479.999638	479.999638	1.052632	0	1	-360	360;
	2007	608	0.00064	0.0518	-0.006678	221.999931	221.999931	221.999931	1	0	1	-360	360;
	8178	6540	0.00072	0.05075	-0.006678	221.999931	221.999931	221.999931	1	0	1	-360	360;
	1608	7405	0	0.23657	0	72.000001	72.000001	72.000001	1	0	1	-360	360;
	7735	6051	0	0.020113	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	7735	7912	0	0.019669	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3242	6005	0.000326	0.028508	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3242	6005	0.000316	0.031545	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	8291	3147	0.000355	0.023171	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2498	1335	0.000296	0.032897	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2498	1335	0.000296	0.030461	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2105	1	5.9e-05	0.022767	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	978	5635	0.000335	0.029237	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	5217	8235	0.000385	0.027846	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3536	6226	0.000444	0.020064	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	7590	42	0.00033	0.02558	-0.016	479.999638	479.999638	479.999638	1	0	1	-360	360;
	6932	5186	0	0.026909	0	126.000288	126.000288	126.000288	1	0	1	-360	360;
	6628	6838	0.000385	0.022737	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2790	642	0.000375	0.022441	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	7189	5561	-0.002742	0.00908	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	7189	2685	-0.00268	0.00837	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	7189	6236	-0.00304	0.00824	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	6526	4902	-0.003927	0.00795	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	6526	5671	-0.00401	0.00794	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	4194	8172	0	0.00765	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	4194	1410	0	0.00748	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	6672	8318	-0.000111	0.00565	-0.012274	408.000087	408.000087	408.000087	1	0	1	-360	360;
	6672	3669	-0.000339	0.00645	-0.012274	408.000087	408.000087	408.000087	1	0	1	-360	360;
	6672	5299	-0.000332	0.00657	-0.012274	408.000087	408.000087	408.000087	1	0	1	-360	360;
	6672	4005	-0.000118	0.00525	-0.012274	408.000087	408.000087	408.000087	1	0	1	-360	360;
	4399	313	9.8e-05	0.060488	0	455.999787	455.999787	455.999787	1	0	1	-360	360;
	4399	2934	8e-05	0.00811	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	4399	4543	-0.003927	0.00835	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	2371	7033	0	0.028951	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2371	3075	0	0.02833	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	88	2750	0.000424	0.020212	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	4153	1592	0.00128	0.05539	-0.00429	159.599856	159.599856	159.599856	1	0	1	-360	360;
	4153	5060	0.00126	0.056469	-0.00429	159.599856	159.599856	159.599856	1	0	1	-360	360;
	4153	3711	0.00123	0.05756	-0.00429	159.599856	159.599856	159.599856	1	0	1	-360	360;
	1336	263	0	0.046831	0	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2842	688	0	0.03197	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	4647	6443	0	0.027151	0	459.496036	459.496036	459.496036	1.044622	0	1	-360	360;
	3364	838	0	0.0197	0	69.600141	69.600141	69.600141	1	0	1	-360	360;
	6570	3900	0.000335	0.028458	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	6570	3900	0.000375	0.027846	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	6863	1064	0.00073	0.01489	-0.002306	75.599792	75.599792	75.599792	1	0	1	-360	360;
	8057	7392	0	0.00864	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	8057	1242	0	0.00881	0	527.999996	527.999996	527.999996	1	0	1	-360	360;
	6513	1863	0.000326	0.028961	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2908	5507	0.000296	0.025084	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1112	6796	0.000326	0.025765	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1112	6796	0.000375	0.020301	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1112	6796	0.000345	0.021109	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2818	112	0.000385	0.02101	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1468	6889	0.000424	0.019472	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1468	6889	0.000434	0.019048	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	5730	3959	0.000326	0.025637	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	5730	3959	0.000355	0.026564	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	597	5022	0.000355	0.021553	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	597	5022	0.000128	0.022007	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	756	4253	6e-05	0.008289	0	1956.0003	1956.0003	1956.0003	0.993813	6.37679	1	-360	360;
	3134	4253	7e-05	0.007859	0	1956.0003	1956.0003	1956.0003	0.995845	5.22451	1	-360	360;
	4253	942	0.000395	0.021454	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3776	2985	-0.000388	0.0074	-0.009168	335.999878	335.999878	335.999878	1	0	1	-360	360;
	2119	4602	-0.000402	0.00724	-0.009168	335.999878	335.999878	335.999878	1	0	1	-360	360;
	4659	1477	-0.000589	0.01411	-0.009168	335.999878	335.999878	335.999878	1	0	1	-360	360;
	3040	7212	-0.00036	0.00803	-0.009168	335.999878	335.999878	335.999878	1	0	1	-360	360;
	7481	5773	0	0.040449	0	552.000505	552.000505	552.000505	0.956938	0	1	-360	360;
	3611	4984	0.000345	0.026071	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2387	7419	0	0.00046	0	371.999653	371.999653	371.999653	1	0	1	-360	360;
	2387	6269	0	0.00043	0	371.999653	371.999653	371.999653	1	0	1	-360	360;
	790	1708	0.000375	0.022007	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	857	6710	0.000385	0.020196	-0.014264	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	4864	6409	0.000375	0.026624	-0.014264	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	4020	1295	0.000425	0.021208	0	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	7358	2425	0.000466	0.020185	-0.014264	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	3526	1470	0.000324	0.022139	0	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	3526	1470	0.000324	0.022706	0	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	3526	1470	0.000314	0.023142	0	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	8181	6637	0.000344	0.024832	-0.014264	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	8181	7690	0.000334	0.023071	-0.014264	479.999638	479.999638	479.999638	0.993901	0	1	-360	360;
	2458	331	0.000335	0.022767	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	2430	305	0.00071	0.030845	-0.009149	300.000103	300.000103	300.000103	1.006865	0	1	-360	360;
	3620	6314	0.000385	0.019472	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3620	3756	0.000395	0.019057	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3620	1186	0.000404	0.018643	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3620	4651	0	0.027866	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	1536	6871	0.000326	0.032245	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3942	7714	0.000365	0.022599	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	3942	7714	0.000355	0.023023	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	7727	3659	0	0.032004	0	300.000068	300.000068	300.000068	1.045455	0	1	-360	360;
	5621	3659	0.00141	0.06647	0	386.864979	386.864979	386.864979	0.997125	4.34603	1	-360	360;
	2433	2106	0.000434	0.019018	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	5862	7960	0.000404	0.020497	-0.014639	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	5101	6927	0.00074	0.029553	-0.009149	300.000103	300.000103	300.000103	1.006865	0	1	-360	360;
	939	1540	0.000454	0.019669	0	479.999638	479.999638	479.999638	1.006865	0	1	-360	360;
	4376	2475	0.00206	0.09005	-0.00452	125	125	125	1	0	1	-360	360;
	2966	224	0.00039	0.03411	-0.007169	400	400	400	1.052632	0	1	-360	360;
	2966	224	0.0004	0.034469	-0.007158	400	400	400	1.052632	0	1	-360	360;
	7701	3888	0.00183	0.090471	-0.005433	150	150	150	1.026087	0	1	-360	360;
	7701	3888	0.00187	0.07429	-0.005464	150	150	150	1.026087	0	1	-360	360;
	6962	1715	0.00077	0.04355	-0.005363	300	300	300	0.999313	0	1	-360	360;
	6962	1715	0.00077	0.04266	-0.005363	300	300	300	0.999313	0	1	-360	360;
	6646	6377	0.00116	0.04106	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	6646	6377	0.00118	0.04119	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	7563	2680	0.0012	0.03931	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	7563	2680	0.00107	0.04567	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	6223	5961	0.002282	0.077224	0	185.400832	185.400832	185.400832	0.97087	0	1	-360	360;
	3833	5961	0.00102	0.04862	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	3833	5961	0.001	0.04954	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	2964	1560	0.00098	0.04477	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	2964	1560	0.00111	0.04387	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	2862	5079	0.00113	0.04193	0	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	8040	4773	0.002522	0.074933	0	170.300154	170.300154	170.300154	1.056957	0	1	-360	360;
	1266	4549	0.001252	0.073601	0	127.659766	127.659766	127.659766	0.94	0	1	-360	360;
	3217	4549	0.001271	0.072001	0	127.659766	127.659766	127.659766	0.94	0	1	-360	360;
	7478	5570	0.000442	0.040802	0	942.570029	942.570029	942.570029	0.895	0	1	-360	360;
	731	80	0.00048	0.041601	0	456.648752	456.648752	456.648752	0.925	0	1	-360	360;
	1596	6346	0.000543	0.04261	0	261.599947	261.599947	261.599947	1	0	1	-360	360;
	537	6346	0.000534	0.04342	0	253.199605	253.199605	253.199605	1	0	1	-360	360;
	7643	5710	0.003895	0.040564	0	222.408002	222.408002	222.408002	0.955	0	1	-360	360;
	7643	5710	0.004413	0.039762	0	227.434518	227.434518	227.434518	0.955	0	1	-360	360;
	7643	5710	0.004498	0.038954	0	231.203911	231.203911	231.203911	0.955	0	1	-360	360;
	2903	6431	0.000621	0.0376	0	216.494919	216.494919	216.494919	0.97	0	1	-360	360;
	3165	6431	0.00063	0.0368	0	218.969015	218.969015	218.969015	0.97	0	1	-360	360;
	3165	7116	0.00064	0.036	0	223.917207	223.917207	223.917207	0.97	0	1	-360	360;
	256	3608	0.000555	0.040801	0	310.515175	310.515175	310.515175	0.97	0	1	-360	360;
	7333	1343	0.000546	0.0416	0	304.330264	304.330264	304.330264	0.97	0	1	-360	360;
	7333	7292	0.000527	0.0424	0	299.380756	299.380756	299.380756	0.97	0	1	-360	360;
	4556	5765	0.000456	0.043201	0	354.345968	354.345968	354.345968	0.955	0	1	-360	360;
	4556	5765	0.000447	0.040001	0	381.989499	381.989499	381.989499	0.955	0	1	-360	360;
	831	5765	0.000505	0.039202	0	389.528943	389.528943	389.528943	0.955	0	1	-360	360;
	831	5765	0.000513	0.038402	0	397.068387	397.068387	397.068387	0.955	0	1	-360	360;
	4389	7155	0.000513	0.037602	0	382.978742	382.978742	382.978742	0.94	0	1	-360	360;
	2749	7155	0.00052	0.036802	0	389.362423	389.362423	389.362423	0.94	0	1	-360	360;
	5756	7155	0.000529	0.036002	0	395.744788	395.744788	395.744788	0.94	0	1	-360	360;
	7356	7926	0.000473	0.040802	0	380.733035	380.733035	380.733035	0.955	0	1	-360	360;
	7959	7926	0.000465	0.041601	0	374.450713	374.450713	374.450713	0.955	0	1	-360	360;
	3698	5699	7e-05	0.0212	0	808.086226	808.086226	808.086226	0.94	0	1	-360	360;
	1827	3952	7e-05	0.021599	0	805.53249	805.53249	805.53249	0.94	0	1	-360	360;
	4697	3055	0.00054	0.04	0	420.642395	420.642395	420.642395	0.984211	0	1	-360	360;
	4697	3055	0.000472	0.039202	0	513.19162	513.19162	513.19162	0.94	0	1	-360	360;
	2468	3055	0.00048	0.038402	0	528.511402	528.511402	528.511402	0.94	0	1	-360	360;
	2468	3055	0.00064	0.0376	0	451.123337	451.123337	451.123337	0.984211	0	1	-360	360;
	1450	1025	0.000333	0.029441	0	278.967325	278.967325	278.967325	1.015	0	1	-360	360;
	5046	7715	0.000353	0.028801	0	278.099186	278.099186	278.099186	1.015	0	1	-360	360;
	5046	7715	0.000312	0.032641	0	245.385042	245.385042	245.385042	1.015	0	1	-360	360;
	1541	7997	0.000529	0.041601	0	596.170918	596.170918	596.170918	0.94	0	1	-360	360;
	7459	7997	0.000511	0.042401	0	592.340973	592.340973	592.340973	0.94	0	1	-360	360;
	7605	6864	0.000633	0.051521	0	302.827641	302.827641	302.827641	0.955	0	1	-360	360;
	7605	6864	0.000616	0.047701	0	327.958243	327.958243	327.958243	0.955	0	1	-360	360;
	123	6864	0.000701	0.046752	0	335.497687	335.497687	335.497687	0.955	0	1	-360	360;
	2962	5498	0.000811	0.04579	0	241.297229	241.297229	241.297229	0.95484	0	1	-360	360;
	332	5498	0.000831	0.04484	0	245.067349	245.067349	245.067349	0.95484	0	1	-360	360;
	332	5498	0.000841	0.043881	0	250.094176	250.094176	250.094176	0.95484	0	1	-360	360;
	5508	6753	0.000621	0.036001	0	146.600884	146.600884	146.600884	1.015	0	1	-360	360;
	5508	2066	0.000555	0.040801	0	118.226774	118.226774	118.226774	1.015	0	1	-360	360;
	5685	3500	0.000563	0.0416	0	118.226774	118.226774	118.226774	1.015	0	1	-360	360;
	1636	929	0.000543	0.0424	0	559.162154	559.162154	559.162154	0.955	0	1	-360	360;
	2557	190	0.001026	0.086398	0	123.711313	123.711313	123.711313	0.97	0	1	-360	360;
	124	190	0.001007	0.080001	0	150.927894	150.927894	150.927894	0.97	0	1	-360	360;
	2921	1522	0.000522	0.039201	0	591.56827	591.56827	591.56827	0.925	0	1	-360	360;
	5150	1522	0.00053	0.038401	0	547.66047	547.66047	547.66047	0.94	0	1	-360	360;
	3949	1522	0.000548	0.037601	0	639.568629	639.568629	639.568629	0.925	0	1	-360	360;
	4345	1787	3e-05	0.00658	0	3026.31898	3026.31898	3026.31898	0.99548	5.44941	1	-360	360;
	3932	3918	0.000907	0.042929	0	180.886518	180.886518	180.886518	1.015	0	1	-360	360;
	4421	3869	7.7e-05	0.0204	0	729.011248	729.011248	729.011248	0.97	0	1	-360	360;
	6559	4421	3e-05	0.007638	0	3212.35205	3212.35205	3212.35205	0.989916	8.14356	1	-360	360;
	3869	3918	0.000773	0.05056	0	157.24115	157.24115	157.24115	1.015	0	1	-360	360;
	5350	2466	0.000728	0.043197	0	114.832544	114.832544	114.832544	1.045	0	1	-360	360;
	378	221	0.000447	0.040001	0	114.832544	114.832544	114.832544	1.045	0	1	-360	360;
	8078	5492	0.00061	0.0392	0	119.999909	119.999909	119.999909	1	0	1	-360	360;
	4825	7862	0.001203	0.076801	0	123.711313	123.711313	123.711313	0.97	0	1	-360	360;
	3543	7862	0.001223	0.075201	0	123.711313	123.711313	123.711313	0.97	0	1	-360	360;
	599	6416	0.004533	0.037356	0	212.783446	212.783446	212.783446	0.97	0	1	-360	360;
	5461	314	0.004615	0.036549	0	216.494919	216.494919	216.494919	0.97	0	1	-360	360;
	6137	6491	0.004113	0.041392	0	196.700835	196.700835	196.700835	0.97	0	1	-360	360;
	3713	6746	0.000627	0.041599	0	112.149805	112.149805	112.149805	1.07	0	1	-360	360;
	959	7503	0.001048	0.084802	0	114.83244	114.83244	114.83244	1.045	0	1	-360	360;
	3530	7503	0.001029	0.086397	0	114.83244	114.83244	114.83244	1.045	0	1	-360	360;
	66	4857	0.000808	0.047709	0	148.134026	148.134026	148.134026	1.045	0	1	-360	360;
	4081	2727	0.00061	0.0392	0	118.226774	118.226774	118.226774	1.015	0	1	-360	360;
	4081	2727	0.00062	0.0384	0	118.226774	118.226774	118.226774	1.015	0	1	-360	360;
	5700	2727	0.00064	0.0376	0	118.226774	118.226774	118.226774	1.015	0	1	-360	360;
	356	3717	0.00063	0.0368	0	217.731638	217.731638	217.731638	0.97	0	1	-360	360;
	5301	4768	0.00064	0.036	0	220.205734	220.205734	220.205734	0.97	0	1	-360	360;
	5301	4768	0.000573	0.0408	0	194.226739	194.226739	194.226739	0.97	0	1	-360	360;
	5875	2247	0.001067	0.083199	0	194.226739	194.226739	194.226739	0.97	0	1	-360	360;
	4809	2247	0.001048	0.084802	0	190.5153	190.5153	190.5153	0.97	0	1	-360	360;
	1312	5689	6e-05	0.11487	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4214	5689	0.20187	1.758347	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1197	3502	0.03911	0.342562	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	2725	4181	0.01564	0.58593	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	2725	4168	-0.000248	0.227727	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	5616	887	0.01414	0.513223	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3343	887	-0.007335	0.347624	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4996	887	-0.006901	0.404112	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	6919	3124	0.051021	0.452583	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3457	577	0.586797	1.843698	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	193	577	0.487562	1.572644	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4410	577	0.129359	0.588864	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	5660	3792	0.000246	0.025121	0	599.999547	599.999547	599.999547	0.996999	0	1	-360	360;
	4214	6611	0.10005	1.276798	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3343	6611	0.21936	1.401591	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4996	6611	0.22688	1.394834	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1312	2272	0.04986	0.379174	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3457	877	0.02303	0.62419	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	3457	2360	0.03435	0.639633	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	226	7657	0.01742	0.603616	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3865	7657	0.14245	1.97967	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	7657	-0.000145	0.32938	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4123	2928	-0.000298	0.102701	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	2863	6016	0.000708	0.03271	0	479.999638	479.999638	479.999638	1.002888	0	1	-360	360;
	2863	7125	0.06719	0.834339	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	5545	2069	0.050909	0.307686	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	902	2069	0.212521	0.943657	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3877	7601	-0.002652	0.490201	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	3877	5118	0.02628	0.438469	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	4956	7601	-0.008241	0.388947	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	6853	7601	0.000226	0.024837	0	396.000162	396.000162	396.000162	1.052632	0	1	-360	360;
	2500	7601	0.10611	0.699359	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	2500	5118	0.02855	0.418988	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	295	0.292479	1.229545	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	193	2928	0.757066	1.631798	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4410	2928	0.384752	1.021653	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4737	4267	-0.014297	0.212851	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4737	4267	0.000444	0.031171	0	599.999547	599.999547	599.999547	0.979739	0	1	-360	360;
	4737	3195	-0.014091	0.217459	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	6919	5536	0.00894	0.300393	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	8004	8146	0.049771	0.546674	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	2018	7848	0.000737	0.026925	-0.007642	396.000162	396.000162	396.000162	1.002888	0	1	-360	360;
	2018	7848	0.000845	0.026029	-0.007262	396.000162	396.000162	396.000162	1.014878	0	1	-360	360;
	8004	2805	0.05157	0.475537	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	6919	2510	0.054781	0.446446	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3139	1837	0.000452	0.023578	0	599.999547	599.999547	599.999547	0.952263	0	1	-360	360;
	3060	2593	-0.010579	0.27814	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3060	4617	-0.007004	0.324483	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3865	7807	-0.034029	0.423698	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	2566	1358	0.000384	0.0242	0	599.999547	599.999547	599.999547	1.008848	0	1	-360	360;
	1312	1358	0.299587	1.849794	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1312	4819	0.301095	1.740868	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1006	4407	0.000759	0.019191	0	943.199907	943.199907	943.199907	1.313163	0	1	-360	360;
	226	6371	0.170173	0.568199	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	6919	6084	0.000413	0.453868	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	8004	6084	0.057719	0.644008	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	6430	6371	0.07761	0.938223	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	6371	0.200409	0.618946	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	6919	3092	0.152771	0.863657	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	8004	3092	-0.001921	0.311818	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	523	0.305992	1.948119	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	8004	5004	0.17005	0.993099	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4971	1569	0.000257	0.029742	0	706.047372	706.047372	706.047372	0.916288	0	1	-360	360;
	4971	2957	0.000268	0.029134	0	707.242625	707.242625	707.242625	0.916288	0	1	-360	360;
	3425	3581	-0.014277	0.154039	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3425	3581	0.000738	0.031616	0	479.999638	479.999638	479.999638	1.002277	0	1	-360	360;
	451	3581	0.251178	0.958099	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3425	339	0.08599	0.668788	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	4660	7125	0.000241	0.02217	0	599.999547	599.999547	599.999547	1.093302	0	1	-360	360;
	451	2049	-0.003678	0.380909	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	7876	2049	-0.011302	0.627045	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	451	3099	0.00797	0.426157	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	7876	3099	0.058159	0.835806	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	451	2298	0.00907	0.406178	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	7876	2298	0.059789	0.790103	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	451	339	0.04738	0.355028	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	451	4891	0.00166	0.210118	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	902	699	0.000234	0.03334	0	396.000162	396.000162	396.000162	0.882932	0	1	-360	360;
	902	7160	-0.007978	0.33365	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	902	1373	-0.007673	0.339439	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	7876	2741	0.000213	0.026084	0	599.999547	599.999547	599.999547	1.027158	0	1	-360	360;
	7876	339	0.079162	0.676053	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	7876	3444	-0.012742	0.556163	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	7876	4891	0.05625	0.586697	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	1197	3602	0.11012	1.389876	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3865	3020	0.12839	0.634174	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	3865	3020	0.000265	0.029415	0	599.999547	599.999547	599.999547	0.882932	0	1	-360	360;
	2678	2331	0.01031	0.165651	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	717	6292	0.000906	0.032368	0	479.999638	479.999638	479.999638	1.03341	0	1	-360	360;
	3060	6292	0.258574	1.91281	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	2725	7740	0.114169	0.69781	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	6616	0.215475	0.921488	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	238	0.000388	0.026559	0	599.999547	599.999547	599.999547	0.985426	0	1	-360	360;
	1013	238	-0.012727	0.168081	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1013	5598	0.11171	0.548317	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	2678	3444	0.04893	0.453587	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	2725	3027	0.000427	0.025869	0	599.999547	599.999547	599.999547	0.985426	0	1	-360	360;
	2725	7946	0.000379	0.029319	0	599.999547	599.999547	599.999547	0.985426	0	1	-360	360;
	6919	2729	0.086233	0.611399	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	6919	3635	0.03295	0.399238	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	8004	5133	0.034401	0.560661	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	5283	2311	0.04848	0.4525	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	5283	3690	0.05627	0.448553	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	22	153	0.000262	0.020993	0	599.999547	599.999547	599.999547	1.079401	0	1	-360	360;
	4410	153	0.093521	0.726302	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	4410	5986	0.170831	0.824835	0	1439.99891	1439.99891	1439.99891	1	0	1	-360	360;
	1197	3635	0.05836	0.35759	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	7613	2729	0.000563	0.031553	-0.013861	479.999638	479.999638	479.999638	1.014878	0	1	-360	360;
	4410	3282	0.04447	0.619827	0	1439.99957	1439.99957	1439.99957	1	0	1	-360	360;
	5900	3034	0.001519	0.050701	-0.004109	119.999875	119.999875	119.999875	1.013514	0	1	-360	360;
	115	4946	0.000821	0.051524	-0.00383	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	115	4946	0.000803	0.047869	-0.00553	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	397	2035	0.000466	0.029318	-0.003795	360.000386	360.000386	360.000386	1.02529	0	1	-360	360;
	7335	4293	0.000463	0.028006	-0.003895	272.400012	272.400012	272.400012	1.038781	0	1	-360	360;
	7335	4293	0.000463	0.026569	-0.00973	286.800317	286.800317	286.800317	1.038781	0	1	-360	360;
	2801	7975	0.000556	0.030088	-0.00596	119.999909	119.999909	119.999909	1.081471	0	1	-360	360;
	2801	7975	0.000581	0.029019	-0.00285	119.999909	119.999909	119.999909	1.081471	0	1	-360	360;
	5286	637	0.0017	0.077979	-0.0006	119.999875	119.999875	119.999875	1	0	1	-360	360;
	7455	5008	0.000569	0.046371	-0.00583	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	7455	5008	0.00056	0.047255	-0.00547	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	484	3956	0.000799	0.050159	-0.00393	148.80052	148.80052	148.80052	1.066856	0	1	-360	360;
	5519	4554	0.000379	0.039836	-0.00999	340.79998	340.79998	340.79998	1.052632	0	1	-360	360;
	5519	4554	0.000433	0.039042	-0.00999	347.999803	347.999803	347.999803	1.052632	0	1	-360	360;
	1737	6671	0.000448	0.032183	-0.004109	119.999909	119.999909	119.999909	1.066856	0	1	-360	360;
	4284	1135	0.000498	0.034255	-0.00036	119.999875	119.999875	119.999875	1.041667	0	1	-360	360;
	4284	1135	0.000507	0.033509	-0.00036	119.999875	119.999875	119.999875	1.041667	0	1	-360	360;
	5613	2185	0.000415	0.040631	-0.00999	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	5613	2185	0.000406	0.041434	-0.00999	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	5050	1280	0.000397	0.037986	-0.004	251.999744	251.999744	251.999744	1.052632	0	1	-360	360;
	7099	1804	0.000322	0.035886	-0.00088	220.80007	220.80007	220.80007	1.043148	0	1	-360	360;
	933	8104	0.000408	0.039074	-0.0008	335.999878	335.999878	335.999878	0.989654	0	1	-360	360;
	5713	6164	0.000417	0.026115	-0.00463	119.999909	119.999909	119.999909	1.062289	0	1	-360	360;
	5713	6164	0.000443	0.02577	-0.004074	119.999909	119.999909	119.999909	1.062289	0	1	-360	360;
	2733	5676	0.000396	0.033918	-0.00086	199.199942	199.199942	199.199942	1.02924	0	1	-360	360;
	7230	157	0.000406	0.023438	-0.00405	219.600209	219.600209	219.600209	1.052632	0	1	-360	360;
	7230	157	0.000352	0.026759	-0.00333	192.000118	192.000118	192.000118	1.052632	0	1	-360	360;
	1553	6804	0.000334	0.036651	-0.0009	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	1553	6804	0.000334	0.037354	-0.0009	119.999909	119.999909	119.999909	1.052632	0	1	-360	360;
	4276	3799	0.000305	0.035675	-0.00096	241.200338	241.200338	241.200338	1.087225	0	1	-360	360;
	4276	3799	0.000296	0.033035	-0.00096	260.399428	260.399428	260.399428	1.087225	0	1	-360	360;
	4276	3799	0.000338	0.032376	-0.00096	265.200188	265.200188	265.200188	1.087225	0	1	-360	360;
	1854	7698	0.000285	0.02604	-0.00117	447.600103	447.600103	447.600103	1.199891	0	1	-360	360;
	933	1125	0.000419	0.02814	-0.00382	361.199589	361.199589	361.199589	0.989654	0	1	-360	360;
	2733	8289	0.000453	0.025129	-0.00361	204.000043	204.000043	204.000043	1.02924	0	1	-360	360;
	2733	4225	0.000453	0.02528	-0.00439	196.799562	196.799562	196.799562	1.02924	0	1	-360	360;
	4594	1822	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4594	189	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1237	5212	0.00172	0.099322	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	1237	5212	0.00168	0.101198	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	8002	5630	0.00363	0.146413	-0.007293	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	8002	5630	0.00182	0.07175	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1582	411	0.0019	0.08995	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	4317	4942	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4317	769	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3815	6800	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3395	689	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5026	5503	0.00167	0.0776	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6724	2839	0.000646	0.034336	-0.003628	199.199942	199.199942	199.199942	1.002441	-0.0065	1	-360	360;
	6724	2839	0.000636	0.034445	-0.003628	215.999969	215.999969	215.999969	1.002246	-0.0258	1	-360	360;
	2839	104	0.00182	0.07175	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2839	3022	0.00185	0.07028	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2203	5687	0.00194	0.088083	-0.003306	153.900093	153.900093	153.900093	1.052632	0	1	-360	360;
	2203	6141	0.00198	0.086207	-0.003306	152.759468	152.759468	152.759468	1.052632	0	1	-360	360;
	3247	6213	0.00196	0.06589	-0.003306	156.632992	156.632992	156.632992	1.049587	0	1	-360	360;
	4003	5731	0.000697	0.032338	-0.003628	278.400632	278.400632	278.400632	1.002506	0	1	-360	360;
	5731	3120	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2455	2293	0.000667	0.033601	-0.003628	291.59976	291.59976	291.59976	1.002506	0	1	-360	360;
	2455	2293	0.000647	0.034238	-0.003628	285.600456	285.600456	285.600456	1.002506	0	1	-360	360;
	2293	2755	0.0016	0.07553	-0.003306	112.996154	112.996154	112.996154	1.061983	0	1	-360	360;
	2293	1070	0.00182	0.07402	-0.003306	112.996154	112.996154	112.996154	1.061983	0	1	-360	360;
	3947	5305	0.0019	0.08995	-0.003306	178.980015	178.980015	178.980015	1.052632	0	1	-360	360;
	5097	3461	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5097	559	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5667	76	0.00201	0.084331	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	5667	76	0.00179	0.095579	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	8373	1883	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6156	834	0.00167	0.0776	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	761	1815	0.00168	0.101198	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	6104	7942	0.00165	0.093702	-0.003306	191.519647	191.519647	191.519647	1.052632	0	1	-360	360;
	6104	1285	0.00187	0.091826	-0.003306	196.080172	196.080172	196.080172	1.052632	0	1	-360	360;
	5733	8100	0.0019	0.08995	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	3460	8224	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3460	8224	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4528	6659	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4528	5866	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	7159	2869	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1977	4443	0.00167	0.0776	-0.006378	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1922	7124	0.00164	0.07907	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	8366	2848	0.00045	0.04299	-0.003199	300.586887	300.586887	300.586887	1.113821	0	1	-360	360;
	8177	2918	0.00051	0.04213	-0.003199	199.313772	199.313772	199.313772	1.113821	0	1	-360	360;
	1088	1089	0.00185	0.07028	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1088	1089	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	183	611	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	183	2255	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4557	4753	0.00179	0.095579	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	4557	4753	0.00176	0.097446	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	1851	1159	0.00167	0.0776	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1851	1159	0.00164	0.07907	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	58	5026	0.000637	0.031701	-0.003628	211.199867	211.199867	211.199867	1.002506	0	1	-360	360;
	58	7211	0.000716	0.031064	-0.003628	241.200338	241.200338	241.200338	1.002506	0	1	-360	360;
	7211	8124	0.00185	0.07028	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	7211	194	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2164	1758	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2164	1758	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5879	8204	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1425	4835	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5678	1605	0.00084	0.0322	-0.003628	264.48014	264.48014	264.48014	1.052632	0	1	-360	360;
	1605	2712	0.00238	0.123248	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4978	6449	0.00269	0.109562	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5551	7663	0.00275	0.107281	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4307	2774	0.000766	0.029164	-0.003628	172.799712	172.799712	172.799712	1.002506	0	1	-360	360;
	4307	2774	0.000776	0.028526	-0.003628	176.399952	176.399952	176.399952	1.002506	0	1	-360	360;
	2774	7569	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2774	5940	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3191	1631	0.00167	0.0776	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3393	3083	0.00168	0.101198	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	3393	3396	0.00165	0.093702	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	881	6598	0.000716	0.031064	-0.003628	250.800541	250.800541	250.800541	1.002506	0	1	-360	360;
	881	3126	0.000736	0.030437	-0.003628	151.199583	151.199583	151.199583	1.002506	0	1	-360	360;
	3126	433	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3126	433	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6045	3974	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	7100	4127	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1521	3467	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1521	3467	0.00167	0.0776	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4001	677	0.00168	0.101198	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	4269	3507	0.00062	0.03068	-0.003724	194.400498	194.400498	194.400498	1.015697	0	1	-360	360;
	4269	6959	0.000458	0.027443	-0.003628	207.599626	207.599626	207.599626	1.002506	0	1	-360	360;
	4269	6959	0.000468	0.026875	-0.003628	212.400386	212.400386	212.400386	1.002506	0	1	-360	360;
	4707	8261	0.001543	0.057853	-0.00572	194.400117	194.400117	194.400117	0.892857	0	1	-360	360;
	2028	4328	0.00198	0.086207	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	6816	3270	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3057	5411	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1545	6107	0.00171	0.07614	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3048	7883	0.000667	0.033601	-0.003628	135.865322	135.865322	135.865322	1.002506	0	1	-360	360;
	3048	6849	0.00168	0.101198	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	7883	6849	0.0016	0.07321	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	5851	6875	0.0019	0.08995	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	7373	3310	0.000746	0.029801	-0.003628	360.000386	360.000386	360.000386	1.002506	0	1	-360	360;
	6807	3472	0.00198	0.086207	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	4850	2756	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	7830	4541	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2905	6459	0.000677	0.032965	-0.003628	431.999937	431.999937	431.999937	1.002506	0	1	-360	360;
	2905	7783	0.000667	0.033601	-0.003628	435.59952	435.59952	435.59952	1.002506	0	1	-360	360;
	2905	1203	0.00168	0.101198	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	2905	4537	0.00165	0.093702	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	6363	4288	0.000716	0.031064	-0.003628	268.800429	268.800429	268.800429	1.002506	0	1	-360	360;
	6363	4288	0.000736	0.030437	-0.003628	274.800391	274.800391	274.800391	1.002506	0	1	-360	360;
	6363	4288	0.000746	0.029801	-0.003628	280.799696	280.799696	280.799696	1.002506	0	1	-360	360;
	4288	6191	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4288	6191	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1404	1907	0.00179	0.095578	-0.003306	177.840049	177.840049	177.840049	1.052632	0	1	-360	360;
	1404	3572	0.00176	0.097446	-0.003306	114.000605	114.000605	114.000605	1.052632	0	1	-360	360;
	7401	4756	0.00167	0.0776	-0.006378	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4539	32	0.00164	0.07907	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1960	459	0.0016	0.07321	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1960	459	0.00182	0.07175	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2743	6836	0.0019	0.08995	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	2743	6836	0.00194	0.088083	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	364	6836	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6592	6881	0.00196	0.06589	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4716	4312	0.00174	0.07467	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3939	3962	0.00387	0.152273	-0.007293	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	802	7030	0.000667	0.033601	-0.003628	119.999909	119.999909	119.999909	1.002506	0	1	-360	360;
	802	2876	0.00168	0.101198	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	802	2397	0.00165	0.093702	-0.009296	285.000196	285.000196	285.000196	1.052632	0	1	-360	360;
	2050	960	0.00182	0.07175	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2050	960	0.00185	0.07028	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6550	852	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6550	852	0.00192	0.06735	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	3566	3346	0.00196	0.07015	-0.003306	111.692578	111.692578	111.692578	1.07438	0	1	-360	360;
	1027	3346	0.00174	0.0795	-0.003306	111.692578	111.692578	111.692578	1.07438	0	1	-360	360;
	2633	2454	0.00136	0.05901	-0.003306	169.860283	169.860283	169.860283	1.052632	0	1	-360	360;
	2633	521	0.00133	0.06014	-0.003306	166.440383	166.440383	166.440383	1.052632	0	1	-360	360;
	2923	5110	0.00164	0.07907	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	2923	2120	0.0016	0.07321	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	1101	6949	0.000716	0.031064	-0.003628	301.199963	301.199963	301.199963	1.002506	0	1	-360	360;
	6949	3986	0.00185	0.07028	-0.006378	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	6357	2001	0.00189	0.06882	-0.003306	114.330976	114.330976	114.330976	1.049587	0	1	-360	360;
	4041	137	0.00158	0.05564	-0.00386	240.000131	240.000131	240.000131	1	0	1	-360	360;
	4041	7013	0.00179	0.0695	-0.00069	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2668	307	0.00084	0.04445	-0.00392	300.000068	300.000068	300.000068	1	0	1	-360	360;
	2668	307	0.0008	0.04546	-0.00517	300.000068	300.000068	300.000068	1	0	1	-360	360;
	2668	307	0.00219	0.08229	-0.00286	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2668	307	0.00069	0.04566	0	286.306994	286.306994	286.306994	1.047826	0	1	-360	360;
	8188	3575	0.00072	0.04835	-0.003815	360.247862	360.247862	360.247862	0.999313	0	1	-360	360;
	8188	3575	0.00082	0.04738	-0.003815	360.247862	360.247862	360.247862	0.999313	0	1	-360	360;
	4840	2131	0.00045	0.03151	-0.002925	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	4840	7677	0.00048	0.026419	-0.00297	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2822	7678	0.0022	0.07023	-0.00201	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2586	202	0.00241	0.0727	-0.00352	171.074624	171.074624	171.074624	1.052174	0	1	-360	360;
	2586	202	0.00081	0.04644	-0.007395	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	3971	2752	0.00158	0.08045	-0.00072	180.000193	180.000193	180.000193	1	0	1	-360	360;
	7102	7021	0.00069	0.04707	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	1398	7530	0.00075	0.04866	-0.002657	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	1398	7530	0.00076	0.04586	-0.002602	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	7922	7771	0.00189	0.07781	-0.00369	180.000193	180.000193	180.000193	1	0	1	-360	360;
	7922	7771	0.00205	0.07804	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	600	6900	0.00207	0.091083	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	600	6900	0.00211	0.06802	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	1807	6780	0.0008	0.03997	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	2823	6780	0.00159	0.086876	-0.0014	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2823	6780	0.0026	0.07998	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	4424	6828	0.00154	0.086488	-0.00288	180.000193	180.000193	180.000193	1	0	1	-360	360;
	7764	4424	0.00052	0.033159	-0.008266	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	7764	6828	0.00072	0.0451	-0.003598	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	7764	6828	0.00068	0.04531	-0.001738	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	4367	3064	0.00042	0.030281	-0.004199	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	4367	7597	0.00079	0.04086	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	3064	4746	0.00168	0.07343	-0.0032	180.000193	180.000193	180.000193	1	0	1	-360	360;
	6676	1794	0.00038	0.033591	-0.007479	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	6676	1794	0.00037	0.03405	-0.005861	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	1538	7815	0.00039	0.034849	-0.00585	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	1538	3412	0.00063	0.04764	-0.002015	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	1538	3412	0.00074	0.044626	-0.00328	360.000386	360.000386	360.000386	1.006865	0	1	-360	360;
	6631	7120	0.00069	0.04232	-0.002659	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	7238	6053	0.00228	0.07286	-0.00201	180.000193	180.000193	180.000193	1	0	1	-360	360;
	7238	6053	0.00189	0.07059	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2134	1564	0.00195	0.07374	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2134	1564	0.00199	0.07214	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	3355	5107	0.00177	0.07931	-0.0014	180.000193	180.000193	180.000193	1	0	1	-360	360;
	3355	5107	0.00174	0.08086	-0.0014	180.000193	180.000193	180.000193	1	0	1	-360	360;
	420	4369	0.0015	0.08239	-0.00083	180.000193	180.000193	180.000193	1	0	1	-360	360;
	420	4369	0.00149	0.091992	-0.0014	180.000193	180.000193	180.000193	1	0	1	-360	360;
	7576	3152	0.00066	0.04441	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	1602	3152	0.00171	0.07418	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	1602	3152	0.00175	0.07267	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	2342	4457	0.00053	0.03038	-0.005884	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	4457	535	0.00165	0.07003	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	4457	6990	0.00217	0.07323	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	231	1299	0.00072	0.0453	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	6822	992	0.0021	0.08257	-0.0037	180.000193	180.000193	180.000193	1	0	1	-360	360;
	6822	992	0.00186	0.087719	-0.00365	180.000193	180.000193	180.000193	1	0	1	-360	360;
	810	8144	0.00076	0.04796	-0.007395	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	810	2938	0.00075	0.04466	-0.007395	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	3414	6973	0.00401	0.10005	-0.00243	119.999875	119.999875	119.999875	1	0	1	-360	360;
	3414	6973	0.00272	0.07842	-0.00243	119.999875	119.999875	119.999875	1	0	1	-360	360;
	3414	6973	0.00278	0.07679	-0.00243	119.999875	119.999875	119.999875	1	0	1	-360	360;
	3414	6973	0.00283	0.07515	-0.00243	119.999875	119.999875	119.999875	1	0	1	-360	360;
	2178	4548	0.00045	0.029271	-0.005928	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2178	4548	0.0004	0.033169	-0.005928	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2178	4548	0.00039	0.03382	-0.005928	455.999787	455.999787	455.999787	1.052632	0	1	-360	360;
	2482	4720	0.00064	0.04577	-0.002659	342.000499	342.000499	342.000499	1.052632	0	1	-360	360;
	629	1790	0.00147	0.08395	-0.00083	180.000193	180.000193	180.000193	1	0	1	-360	360;
	629	160	0.00184	0.07633	-0.00201	180.000193	180.000193	180.000193	1	0	1	-360	360;
	5534	8342	0.00074	0.04352	-0.004229	342.149247	342.149247	342.149247	1.052174	0	1	-360	360;
	1340	4350	0.00176	0.08177	-0.00155	180.000193	180.000193	180.000193	1	0	1	-360	360;
	1340	4350	0.00242	0.06613	-0.00334	180.000193	180.000193	180.000193	1	0	1	-360	360;
	1757	3132	0.00042	0.030031	-0.006382	472.888668	472.888668	472.888668	1.015038	0	1	-360	360;
	1757	3132	0.00043	0.02938	-0.006382	472.888668	472.888668	472.888668	1.015038	0	1	-360	360;
	7758	8286	0.00014	0.01857	-0.01473	2302.02688	2302.02688	2302.02688	1	0	1	-360	360;
	5285	7381	0.00037	0.03395	-0.004935	472.888668	472.888668	472.888668	1.015038	0	1	-360	360;
	3741	3604	1.8e-05	0.052587	-0.00965	599.999547	599.999547	599.999547	1.052632	0	1	-360	360;
	2577	3604	5.5e-05	0.082396	-0.00321	600.000136	600.000136	600.000136	1.047619	0	1	-360	360;
	719	4479	0.001416	0.043162	-0.004272	150.000034	150.000034	150.000034	0.92437	0	1	-360	360;
	719	4479	0.001603	0.042296	-0.004272	150.000034	150.000034	150.000034	0.92437	0	1	-360	360;
	3818	3590	0.000806	0.025741	-0.0232	479.999638	479.999638	479.999638	1.002506	0	1	-360	360;
	3818	3590	0.000826	0.025204	-0.0232	479.999638	479.999638	479.999638	1.002506	0	1	-360	360;
	3818	3198	0.000107	0.01357	-0.0128	755.99989	755.99989	755.99989	1.05905	0	1	-360	360;
	3590	867	0.00228	0.07939	-0.00387	150.000034	150.000034	150.000034	1	0	1	-360	360;
	3198	3590	0.001087	0.065195	-0.0096	287.999814	287.999814	287.999814	0.995672	-2.273462	1	-360	360;
	3198	3590	0.001067	0.046326	-0.0096	287.999814	287.999814	287.999814	0.998557	-0.854786	1	-360	360;
	5527	3701	0.001301	0.061931	-0.00554	240.000131	240.000131	240.000131	0.948276	0	1	-360	360;
	5527	3701	0.001268	0.063098	-0.00554	240.000131	240.000131	240.000131	0.948276	0	1	-360	360;
	6965	981	0.002358	0.088053	-0.00675	150.000034	150.000034	150.000034	0.948276	0	1	-360	360;
	6965	981	0.00268	0.086295	-0.00675	150.000034	150.000034	150.000034	0.948276	0	1	-360	360;
	7141	543	0.00078	0.030576	-0.034641	479.999638	479.999638	479.999638	1.057818	-0.198597	1	-360	360;
	543	3863	0.001565	0.062128	-0.00582	240.000131	240.000131	240.000131	0.932203	0	1	-360	360;
	543	3863	0.001588	0.060806	-0.00582	240.000131	240.000131	240.000131	0.932203	0	1	-360	360;
	4894	7295	0.002713	0.092944	-0.004201	150.000034	150.000034	150.000034	0.916667	0	1	-360	360;
	4894	7295	0.002464	0.122138	-0.004201	150.000034	150.000034	150.000034	0.916667	0	1	-360	360;
	4596	402	0.002739	0.116189	-0.00218	150.000034	150.000034	150.000034	0.932203	0	1	-360	360;
	4596	402	0.002681	0.118434	-0.00218	150.000034	150.000034	150.000034	0.932203	0	1	-360	360;
	5805	550	0.000212	0.011512	-0.007263	1079.9995	1079.9995	1079.9995	0.970907	2.77215	1	-360	360;
];

%%-----  OPF Data  -----%%
%% generator cost data
%	1	startup	shutdown	n	x1	y1	...	xn	yn
%	2	startup	shutdown	n	c(n-1)	...	c0
mpc.gencost = [
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
	2	0	0	3	0	1	0;
];

if fixed
    [GEN_BUS, PG, QG, QMAX, QMIN, VG, MBASE, GEN_STATUS, PMAX, PMIN] = idx_gen;
    k = find(   isinf(mpc.gen(:, QMIN)) & ...
                isinf(mpc.gen(:, QMAX)) & ...
                isinf(mpc.gen(:, PMIN)) & ...
                isinf(mpc.gen(:, PMAX))  );
    mpc.gen(k, PMIN) = mpc.gen(k, PG);
    mpc.gen(k, PMAX) = mpc.gen(k, PG);
    mpc.gen(k, QMIN) = mpc.gen(k, QG);
    mpc.gen(k, QMAX) = mpc.gen(k, QG);
end
